// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See generate_spec_error_message.py for modifications
// Based on Vulkan specification version: 1.4.357

/***************************************************************************
 *
 * Copyright (c) 2016-2026 Google Inc.
 * Copyright (c) 2016-2026 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ****************************************************************************/

#include "vk_validation_error_messages.h"
#include <array>

// clang-format off

const vvl::unordered_map<std::string_view, vuid_info> &GetVuidMap() {
    static const std::array<std::pair<std::string_view, vuid_info>, 27617> vuid_array = {{
        { {"VUID-BaryCoordKHR-BaryCoordKHR-04154", 36}, { {"The BaryCoordKHR decoration must be used only within the Fragment Execution Model", 81}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordKHR-BaryCoordKHR-04155", 36}, { {"The variable decorated with BaryCoordKHR must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordKHR-BaryCoordKHR-04156", 36}, { {"The variable decorated with BaryCoordKHR must be declared as a three-component vector of 32-bit floating-point values", 117}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspAMD-BaryCoordNoPerspAMD-04157", 50}, { {"The BaryCoordNoPerspAMD decoration must be used only within the Fragment Execution Model", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspAMD-BaryCoordNoPerspAMD-04158", 50}, { {"The variable decorated with BaryCoordNoPerspAMD must be declared using the Input Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspAMD-BaryCoordNoPerspAMD-04159", 50}, { {"The variable decorated with BaryCoordNoPerspAMD must be declared as a two-component vector of 32-bit floating-point values", 122}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspCentroidAMD-BaryCoordNoPerspCentroidAMD-04163", 66}, { {"The BaryCoordNoPerspCentroidAMD decoration must be used only within the Fragment Execution Model", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspCentroidAMD-BaryCoordNoPerspCentroidAMD-04164", 66}, { {"The variable decorated with BaryCoordNoPerspCentroidAMD must be declared using the Input Storage Class", 102}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspCentroidAMD-BaryCoordNoPerspCentroidAMD-04165", 66}, { {"The variable decorated with BaryCoordNoPerspCentroidAMD must be declared as a three-component vector of 32-bit floating-point values", 132}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspKHR-BaryCoordNoPerspKHR-04160", 50}, { {"The BaryCoordNoPerspKHR decoration must be used only within the Fragment Execution Model", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspKHR-BaryCoordNoPerspKHR-04161", 50}, { {"The variable decorated with BaryCoordNoPerspKHR must be declared using the Input Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspKHR-BaryCoordNoPerspKHR-04162", 50}, { {"The variable decorated with BaryCoordNoPerspKHR must be declared as a three-component vector of 32-bit floating-point values", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspSampleAMD-BaryCoordNoPerspSampleAMD-04166", 62}, { {"The BaryCoordNoPerspSampleAMD decoration must be used only within the Fragment Execution Model", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspSampleAMD-BaryCoordNoPerspSampleAMD-04167", 62}, { {"The variable decorated with BaryCoordNoPerspSampleAMD must be declared using the Input Storage Class", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordNoPerspSampleAMD-BaryCoordNoPerspSampleAMD-04168", 62}, { {"The variable decorated with BaryCoordNoPerspSampleAMD must be declared as a two-component vector of 32-bit floating-point values", 128}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordPullModelAMD-BaryCoordPullModelAMD-04169", 54}, { {"The BaryCoordPullModelAMD decoration must be used only within the Fragment Execution Model", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordPullModelAMD-BaryCoordPullModelAMD-04170", 54}, { {"The variable decorated with BaryCoordPullModelAMD must be declared using the Input Storage Class", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordPullModelAMD-BaryCoordPullModelAMD-04171", 54}, { {"The variable decorated with BaryCoordPullModelAMD must be declared as a three-component vector of 32-bit floating-point values", 126}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothAMD-BaryCoordSmoothAMD-04172", 48}, { {"The BaryCoordSmoothAMD decoration must be used only within the Fragment Execution Model", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothAMD-BaryCoordSmoothAMD-04173", 48}, { {"The variable decorated with BaryCoordSmoothAMD must be declared using the Input Storage Class", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothAMD-BaryCoordSmoothAMD-04174", 48}, { {"The variable decorated with BaryCoordSmoothAMD must be declared as a two-component vector of 32-bit floating-point values", 121}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothCentroidAMD-BaryCoordSmoothCentroidAMD-04175", 64}, { {"The BaryCoordSmoothCentroidAMD decoration must be used only within the Fragment Execution Model", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothCentroidAMD-BaryCoordSmoothCentroidAMD-04176", 64}, { {"The variable decorated with BaryCoordSmoothCentroidAMD must be declared using the Input Storage Class", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothCentroidAMD-BaryCoordSmoothCentroidAMD-04177", 64}, { {"The variable decorated with BaryCoordSmoothCentroidAMD must be declared as a two-component vector of 32-bit floating-point values", 129}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothSampleAMD-BaryCoordSmoothSampleAMD-04178", 60}, { {"The BaryCoordSmoothSampleAMD decoration must be used only within the Fragment Execution Model", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothSampleAMD-BaryCoordSmoothSampleAMD-04179", 60}, { {"The variable decorated with BaryCoordSmoothSampleAMD must be declared using the Input Storage Class", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaryCoordSmoothSampleAMD-BaryCoordSmoothSampleAMD-04180", 60}, { {"The variable decorated with BaryCoordSmoothSampleAMD must be declared as a two-component vector of 32-bit floating-point values", 127}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseInstance-BaseInstance-04181", 36}, { {"The BaseInstance decoration must be used only within the Vertex Execution Model", 79}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseInstance-BaseInstance-04182", 36}, { {"The variable decorated with BaseInstance must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseInstance-BaseInstance-04183", 36}, { {"The variable decorated with BaseInstance must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseVertex-BaseVertex-04184", 32}, { {"The BaseVertex decoration must be used only within the Vertex Execution Model", 77}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseVertex-BaseVertex-04185", 32}, { {"The variable decorated with BaseVertex must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-BaseVertex-BaseVertex-04186", 32}, { {"The variable decorated with BaseVertex must be declared as a scalar 32-bit integer value", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistance-ClipDistance-04187", 36}, { {"The ClipDistance decoration must be used only within the MeshEXT, MeshNV, Vertex, Fragment, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 164}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistance-ClipDistance-04188", 36}, { {"The variable decorated with ClipDistance within the MeshEXT, MeshNV, or Vertex Execution Model must be declared using the Output Storage Class", 142}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistance-ClipDistance-04189", 36}, { {"The variable decorated with ClipDistance within the Fragment Execution Model must be declared using the Input Storage Class", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistance-ClipDistance-04190", 36}, { {"The variable decorated with ClipDistance within the TessellationControl, TessellationEvaluation, or Geometry Execution Model must not be declared in a Storage Class other than Input or Output", 191}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistance-ClipDistance-04191", 36}, { {"The variable decorated with ClipDistance must be declared as an array of 32-bit floating-point values", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistancePerViewNV-ClipDistancePerViewNV-04192", 54}, { {"The ClipDistancePerViewNV decoration must be used only within the MeshNV Execution Model", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistancePerViewNV-ClipDistancePerViewNV-04193", 54}, { {"The variable decorated with ClipDistancePerViewNV must be declared using the Output Storage Class", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistancePerViewNV-ClipDistancePerViewNV-04194", 54}, { {"The variable decorated with ClipDistancePerViewNV must also be decorated with the PerViewNV decoration", 102}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClipDistancePerViewNV-ClipDistancePerViewNV-04195", 54}, { {"The variable decorated with ClipDistancePerViewNV must be declared as a two-dimensional array of 32-bit floating-point values", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterIDHUAWEI-ClusterIDHUAWEI-07797", 42}, { {"The ClusterIDHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterIDHUAWEI-ClusterIDHUAWEI-07798", 42}, { {"The variable decorated with ClusterIDHUAWEI must be declared as a scalar 32-bit integer value", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterIDNV-ClusterIDNV-10531", 34}, { {"The ClusterIDNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterIDNV-ClusterIDNV-10532", 34}, { {"The variable decorated with ClusterIDNV must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterIDNV-ClusterIDNV-10533", 34}, { {"The variable decorated with ClusterIDNV must be declared as a scalar 32-bit integer value", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterShadingRateHUAWEI-ClusterShadingRateHUAWEI-09448", 60}, { {"The ClusterShadingRateHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 105}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ClusterShadingRateHUAWEI-ClusterShadingRateHUAWEI-09449", 60}, { {"The variable decorated with ClusterShadingRateHUAWEI must be declared as a scalar 32-bit integer value", 102}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoalescedInputCountAMDX-CoalescedInputCountAMDX-09172", 58}, { {"The variable decorated with CoalescedInputCountAMDX must be declared using the Input Storage Class", 98}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoalescedInputCountAMDX-CoalescedInputCountAMDX-09173", 58}, { {"If a variable is decorated with CoalescedInputCountAMDX, the CoalescingAMDX execution mode must be declared", 107}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoalescedInputCountAMDX-CoalescedInputCountAMDX-09174", 58}, { {"The variable decorated with CoalescedInputCountAMDX must be declared as a scalar 32-bit integer value", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreCountARM-CoreCountARM-07595", 36}, { {"The variable decorated with CoreCountARM must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreCountARM-CoreCountARM-07596", 36}, { {"The variable decorated with CoreCountARM must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreIDARM-CoreIDARM-07599", 30}, { {"The variable decorated with CoreIDARM must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreIDARM-CoreIDARM-07600", 30}, { {"The variable decorated with CoreIDARM must be declared as a scalar 32-bit integer value", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreMaxIDARM-CoreMaxIDARM-07597", 36}, { {"The variable decorated with CoreMaxIDARM must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CoreMaxIDARM-CoreMaxIDARM-07598", 36}, { {"The variable decorated with CoreMaxIDARM must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistance-CullDistance-04196", 36}, { {"The CullDistance decoration must be used only within the MeshEXT, MeshNV, Vertex, Fragment, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 164}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistance-CullDistance-04197", 36}, { {"The variable decorated with CullDistance within the MeshEXT, MeshNV or Vertex Execution Model must be declared using the Output Storage Class", 141}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistance-CullDistance-04198", 36}, { {"The variable decorated with CullDistance within the Fragment Execution Model must be declared using the Input Storage Class", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistance-CullDistance-04199", 36}, { {"The variable decorated with CullDistance within the TessellationControl, TessellationEvaluation, or Geometry Execution Model must not be declared using a Storage Class other than Input or Output", 194}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistance-CullDistance-04200", 36}, { {"The variable decorated with CullDistance must be declared as an array of 32-bit floating-point values", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistancePerViewNV-CullDistancePerViewNV-04201", 54}, { {"The CullDistancePerViewNV decoration must be used only within the MeshNV Execution Model", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistancePerViewNV-CullDistancePerViewNV-04202", 54}, { {"The variable decorated with CullDistancePerViewNV must be declared using the Output Storage Class", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistancePerViewNV-CullDistancePerViewNV-04203", 54}, { {"The variable decorated with CullDistancePerViewNV must also be decorated with the PerViewNV decoration", 102}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullDistancePerViewNV-CullDistancePerViewNV-04204", 54}, { {"The variable decorated with CullDistancePerViewNV must be declared as a two-dimensional array of 32-bit floating-point values", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullMaskKHR-CullMaskKHR-06735", 34}, { {"The CullMaskKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullMaskKHR-CullMaskKHR-06736", 34}, { {"The variable decorated with CullMaskKHR must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullMaskKHR-CullMaskKHR-06737", 34}, { {"The variable decorated with CullMaskKHR must be declared as a scalar 32-bit integer value", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-07034", 44}, { {"The CullPrimitiveEXT decoration must be used only within the MeshEXT Execution Model", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-07035", 44}, { {"The variable decorated with CullPrimitiveEXT must be declared using the Output Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-07036", 44}, { {"CullPrimitiveEXT must decorate a scalar boolean member of a structure decorated as Block, or decorate a variable of type OpTypeArray of boolean values", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-07038", 44}, { {"The variable decorated with CullPrimitiveEXT within the MeshEXT Execution Model must also be decorated with the PerPrimitiveEXT decoration", 138}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-10589", 44}, { {"If CullPrimitiveEXT is declared as an array of boolean values, the size of the array must match the value specified by OutputPrimitivesEXT", 138}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-10590", 44}, { {"If CullPrimitiveEXT decorates a member of a structure, the variable declaration of the containing Block type must have an array size that matches the value specified by OutputPrimitivesEXT", 188}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CullPrimitiveEXT-CullPrimitiveEXT-10591", 44}, { {"There must be only one declaration of the CullPrimitiveEXT associated with a entry point's interface", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CurrentRayTimeNV-CurrentRayTimeNV-04942", 44}, { {"The CurrentRayTimeNV decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 130}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CurrentRayTimeNV-CurrentRayTimeNV-04943", 44}, { {"The variable decorated with CurrentRayTimeNV must be declared using the Input Storage Class", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-CurrentRayTimeNV-CurrentRayTimeNV-04944", 44}, { {"The variable decorated with CurrentRayTimeNV must be declared as a scalar 32-bit floating-point value", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-DeviceIndex-DeviceIndex-04205", 34}, { {"The variable decorated with DeviceIndex must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-DeviceIndex-DeviceIndex-04206", 34}, { {"The variable decorated with DeviceIndex must be declared as a scalar 32-bit integer value", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-DrawIndex-DrawIndex-04207", 30}, { {"The DrawIndex decoration must be used only within the Vertex, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-DrawIndex-DrawIndex-04208", 30}, { {"The variable decorated with DrawIndex must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-DrawIndex-DrawIndex-04209", 30}, { {"The variable decorated with DrawIndex must be declared as a scalar 32-bit integer value", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstIndexHUAWEI-FirstIndexHUAWEI-07799", 44}, { {"The FirstIndexHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstIndexHUAWEI-FirstIndexHUAWEI-07800", 44}, { {"The variable decorated with FirstIndexHUAWEI must be declared as a scalar 32-bit integer value", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstInstanceHUAWEI-FirstInstanceHUAWEI-07801", 50}, { {"The FirstInstanceHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstInstanceHUAWEI-FirstInstanceHUAWEI-07802", 50}, { {"The variable decorated with FirstInstanceHUAWEI must be declared as a scalar 32-bit integer value", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstVertexHUAWEI-FirstVertexHUAWEI-07803", 46}, { {"The FirstVertexHUAWEI decoration must be used only within the FirstVertexHUAWEI Execution Model", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FirstVertexHUAWEI-FirstVertexHUAWEI-07804", 46}, { {"The variable decorated with FirstVertexHUAWEI must be declared as a scalar 32-bit integer value", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragCoord-FragCoord-04210", 30}, { {"The FragCoord decoration must be used only within the Fragment Execution Model", 78}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragCoord-FragCoord-04211", 30}, { {"The variable decorated with FragCoord must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragCoord-FragCoord-04212", 30}, { {"The variable decorated with FragCoord must be declared as a four-component vector of 32-bit floating-point values", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragDepth-FragDepth-04213", 30}, { {"The FragDepth decoration must be used only within the Fragment Execution Model", 78}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragDepth-FragDepth-04214", 30}, { {"The variable decorated with FragDepth must be declared using the Output Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragDepth-FragDepth-04215", 30}, { {"The variable decorated with FragDepth must be declared as a scalar 32-bit floating-point value", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragDepth-FragDepth-04216", 30}, { {"If the shader dynamically writes to the variable decorated with FragDepth, the DepthReplacing Execution Mode must be declared", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragInvocationCountEXT-FragInvocationCountEXT-04217", 56}, { {"The FragInvocationCountEXT decoration must be used only within the Fragment Execution Model", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragInvocationCountEXT-FragInvocationCountEXT-04218", 56}, { {"The variable decorated with FragInvocationCountEXT must be declared using the Input Storage Class", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragInvocationCountEXT-FragInvocationCountEXT-04219", 56}, { {"The variable decorated with FragInvocationCountEXT must be declared as a scalar 32-bit integer value", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragSizeEXT-FragSizeEXT-04220", 34}, { {"The FragSizeEXT decoration must be used only within the Fragment Execution Model", 80}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragSizeEXT-FragSizeEXT-04221", 34}, { {"The variable decorated with FragSizeEXT must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragSizeEXT-FragSizeEXT-04222", 34}, { {"The variable decorated with FragSizeEXT must be declared as a two-component vector of 32-bit integer values", 107}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragStencilRefEXT-FragStencilRefEXT-04223", 46}, { {"The FragStencilRefEXT decoration must be used only within the Fragment Execution Model", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragStencilRefEXT-FragStencilRefEXT-04224", 46}, { {"The variable decorated with FragStencilRefEXT must be declared using the Output Storage Class", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragStencilRefEXT-FragStencilRefEXT-04225", 46}, { {"The variable decorated with FragStencilRefEXT must be declared as a scalar integer value", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragmentSizeNV-FragmentSizeNV-04226", 40}, { {"The FragmentSizeNV decoration must be used only within the Fragment Execution Model", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragmentSizeNV-FragmentSizeNV-04227", 40}, { {"The variable decorated with FragmentSizeNV must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FragmentSizeNV-FragmentSizeNV-04228", 40}, { {"The variable decorated with FragmentSizeNV must be declared as a two-component vector of 32-bit integer values", 110}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FrontFacing-FrontFacing-04229", 34}, { {"The FrontFacing decoration must be used only within the Fragment Execution Model", 80}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FrontFacing-FrontFacing-04230", 34}, { {"The variable decorated with FrontFacing must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FrontFacing-FrontFacing-04231", 34}, { {"The variable decorated with FrontFacing must be declared as a boolean value", 75}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FullyCoveredEXT-FullyCoveredEXT-04232", 42}, { {"The FullyCoveredEXT decoration must be used only within the Fragment Execution Model", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FullyCoveredEXT-FullyCoveredEXT-04233", 42}, { {"The variable decorated with FullyCoveredEXT must be declared using the Input Storage Class", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FullyCoveredEXT-FullyCoveredEXT-04234", 42}, { {"The variable decorated with FullyCoveredEXT must be declared as a boolean value", 79}, { "chapters/interfaces.html", 24} } },
        { {"VUID-FullyCoveredEXT-conservativeRasterizationPostDepthCoverage-04235", 69}, { {"If VkPhysicalDeviceConservativeRasterizationPropertiesEXT::conservativeRasterizationPostDepthCoverage is not supported the PostDepthCoverage Execution Mode must not be declared, when a variable with the FullyCoveredEXT decoration is declared", 241}, { "chapters/interfaces.html", 24} } },
        { {"VUID-GlobalInvocationId-GlobalInvocationId-04236", 48}, { {"The GlobalInvocationId decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-GlobalInvocationId-GlobalInvocationId-04237", 48}, { {"The variable decorated with GlobalInvocationId must be declared using the Input Storage Class", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-GlobalInvocationId-GlobalInvocationId-04238", 48}, { {"The variable decorated with GlobalInvocationId must be declared as a three-component vector of 32-bit integer values", 116}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HelperInvocation-HelperInvocation-04239", 44}, { {"The HelperInvocation decoration must be used only within the Fragment Execution Model", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HelperInvocation-HelperInvocation-04240", 44}, { {"The variable decorated with HelperInvocation must be declared using the Input Storage Class", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HelperInvocation-HelperInvocation-04241", 44}, { {"The variable decorated with HelperInvocation must be declared as a boolean value", 80}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsLSSNV-HitIsLSSNV-10516", 32}, { {"The HitIsLSSNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 98}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsLSSNV-HitIsLSSNV-10517", 32}, { {"The variable decorated with HitIsLSSNV must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsLSSNV-HitIsLSSNV-10518", 32}, { {"The variable decorated with HitIsLSSNV must be declared as a boolean value", 74}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsSphereNV-HitIsSphereNV-10513", 38}, { {"The HitIsSphereNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsSphereNV-HitIsSphereNV-10514", 38}, { {"The variable decorated with HitIsSphereNV must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitIsSphereNV-HitIsSphereNV-10515", 38}, { {"The variable decorated with HitIsSphereNV must be declared as a boolean value", 77}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitKindKHR-HitKindKHR-04242", 32}, { {"The HitKindKHR decoration must be used only within the AnyHitKHR or ClosestHitKHR Execution Model", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitKindKHR-HitKindKHR-04243", 32}, { {"The variable decorated with HitKindKHR must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitKindKHR-HitKindKHR-04244", 32}, { {"The variable decorated with HitKindKHR must be declared as a scalar 32-bit integer value", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSPositionsNV-HitLSSPositionsNV-10525", 46}, { {"The HitLSSPositionsNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 105}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSPositionsNV-HitLSSPositionsNV-10526", 46}, { {"The variable decorated with HitLSSPositionsNV must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSPositionsNV-HitLSSPositionsNV-10527", 46}, { {"The variable decorated with HitLSSPositionsNV must be declared as an array of size two, containing three-component vector of 32-bit floating-point values", 153}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSRadiiNV-HitLSSRadiiNV-10528", 38}, { {"The HitLSSRadiiNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSRadiiNV-HitLSSRadiiNV-10529", 38}, { {"The variable decorated with HitLSSRadiiNV must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitLSSRadiiNV-HitLSSRadiiNV-10530", 38}, { {"The variable decorated with HitLSSRadiiNV must be declared as an array of size two, containing 32-bit floating-point values", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSpherePositionNV-HitSpherePositionNV-10519", 50}, { {"The HitSpherePositionNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 107}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSpherePositionNV-HitSpherePositionNV-10520", 50}, { {"The variable decorated with HitSpherePositionNV must be declared using the Input Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSpherePositionNV-HitSpherePositionNV-10521", 50}, { {"The variable decorated with HitSpherePositionNV must be declared as a three-component vector of 32-bit floating-point values", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSphereRadiusNV-HitSphereRadiusNV-10522", 46}, { {"The HitSphereRadiusNV decoration must be used only within the AnyHitKHR, or ClosestHitKHR Execution Model", 105}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSphereRadiusNV-HitSphereRadiusNV-10523", 46}, { {"The variable decorated with HitSphereRadiusNV must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitSphereRadiusNV-HitSphereRadiusNV-10524", 46}, { {"The variable decorated with HitSphereRadiusNV must be declared as a scalar 32-bit floating-point value", 102}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTNV-HitTNV-04245", 24}, { {"The HitTNV decoration must be used only within the AnyHitNV or ClosestHitNV Execution Model", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTNV-HitTNV-04246", 24}, { {"The variable decorated with HitTNV must be declared using the Input Storage Class", 81}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTNV-HitTNV-04247", 24}, { {"The variable decorated with HitTNV must be declared as a scalar 32-bit floating-point value", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTriangleVertexPositionsKHR-HitTriangleVertexPositionsKHR-08747", 70}, { {"The HitTriangleVertexPositionsKHR decoration must be used only within the AnyHitKHR or ClosestHitKHR Execution Model", 116}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTriangleVertexPositionsKHR-HitTriangleVertexPositionsKHR-08748", 70}, { {"The variable decorated with HitTriangleVertexPositionsKHR must be declared using the Input Storage Class", 104}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTriangleVertexPositionsKHR-HitTriangleVertexPositionsKHR-08749", 70}, { {"The variable decorated with HitTriangleVertexPositionsKHR must be declared as an array of three vectors of three 32-bit float values", 132}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTriangleVertexPositionsKHR-HitTriangleVertexPositionsKHR-08750", 70}, { {"The variable decorated with HitTriangleVertexPositionsKHR must be used only if the value of HitKindKHR is HitKindFrontFacingTriangleKHR or HitKindBackFacingTriangleKHR", 167}, { "chapters/interfaces.html", 24} } },
        { {"VUID-HitTriangleVertexPositionsKHR-None-08751", 45}, { {"The acceleration structure corresponding to the current intersection must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DATA_ACCESS_BIT_KHR", 152}, { "chapters/interfaces.html", 24} } },
        { {"VUID-IncomingRayFlagsKHR-IncomingRayFlagsKHR-04248", 50}, { {"The IncomingRayFlagsKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 133}, { "chapters/interfaces.html", 24} } },
        { {"VUID-IncomingRayFlagsKHR-IncomingRayFlagsKHR-04249", 50}, { {"The variable decorated with IncomingRayFlagsKHR must be declared using the Input Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-IncomingRayFlagsKHR-IncomingRayFlagsKHR-04250", 50}, { {"The variable decorated with IncomingRayFlagsKHR must be declared as a scalar 32-bit integer value", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-IndexCountHUAWEI-IndexCountHUAWEI-07805", 44}, { {"The IndexCountHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-IndexCountHUAWEI-IndexCountHUAWEI-07806", 44}, { {"The variable decorated with IndexCountHUAWEI must be declared as a scalar 32-bit integer value", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceCountHUAWEI-InstanceCountHUAWEI-07807", 50}, { {"The InstanceCountHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceCountHUAWEI-InstanceCountHUAWEI-07808", 50}, { {"The variable decorated with InstanceCountHUAWEI must be declared as a scalar 32-bit integer value", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceCustomIndexKHR-InstanceCustomIndexKHR-04251", 56}, { {"The InstanceCustomIndexKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 127}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceCustomIndexKHR-InstanceCustomIndexKHR-04252", 56}, { {"The variable decorated with InstanceCustomIndexKHR must be declared using the Input Storage Class", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceCustomIndexKHR-InstanceCustomIndexKHR-04253", 56}, { {"The variable decorated with InstanceCustomIndexKHR must be declared as a scalar 32-bit integer value", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceId-InstanceId-04254", 32}, { {"The InstanceId decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 115}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceId-InstanceId-04255", 32}, { {"The variable decorated with InstanceId must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceId-InstanceId-04256", 32}, { {"The variable decorated with InstanceId must be declared as a scalar 32-bit integer value", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceIndex-InstanceIndex-04263", 38}, { {"The InstanceIndex decoration must be used only within the Vertex Execution Model", 80}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceIndex-InstanceIndex-04264", 38}, { {"The variable decorated with InstanceIndex must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InstanceIndex-InstanceIndex-04265", 38}, { {"The variable decorated with InstanceIndex must be declared as a scalar 32-bit integer value", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationId-InvocationId-04257", 36}, { {"The InvocationId decoration must be used only within the TessellationControl or Geometry Execution Model", 104}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationId-InvocationId-04258", 36}, { {"The variable decorated with InvocationId must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationId-InvocationId-04259", 36}, { {"The variable decorated with InvocationId must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationsPerPixelNV-InvocationsPerPixelNV-04260", 54}, { {"The InvocationsPerPixelNV decoration must be used only within the Fragment Execution Model", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationsPerPixelNV-InvocationsPerPixelNV-04261", 54}, { {"The variable decorated with InvocationsPerPixelNV must be declared using the Input Storage Class", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-InvocationsPerPixelNV-InvocationsPerPixelNV-04262", 54}, { {"The variable decorated with InvocationsPerPixelNV must be declared as a scalar 32-bit integer value", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchIdKHR-LaunchIdKHR-04266", 34}, { {"The LaunchIdKHR decoration must be used only within the RayGenerationKHR, IntersectionKHR, AnyHitKHR, ClosestHitKHR, MissKHR, or CallableKHR Execution Model", 156}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchIdKHR-LaunchIdKHR-04267", 34}, { {"The variable decorated with LaunchIdKHR must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchIdKHR-LaunchIdKHR-04268", 34}, { {"The variable decorated with LaunchIdKHR must be declared as a three-component vector of 32-bit integer values", 109}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchSizeKHR-LaunchSizeKHR-04269", 38}, { {"The LaunchSizeKHR decoration must be used only within the RayGenerationKHR, IntersectionKHR, AnyHitKHR, ClosestHitKHR, MissKHR, or CallableKHR Execution Model", 158}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchSizeKHR-LaunchSizeKHR-04270", 38}, { {"The variable decorated with LaunchSizeKHR must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LaunchSizeKHR-LaunchSizeKHR-04271", 38}, { {"The variable decorated with LaunchSizeKHR must be declared as a three-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-04272", 22}, { {"The Layer decoration must be used only within the MeshEXT, MeshNV, Vertex, TessellationEvaluation, Geometry, or Fragment Execution Model", 136}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-04273", 22}, { {"If the shaderOutputLayer feature is not enabled then the Layer decoration must be used only within the Geometry or Fragment Execution Model", 139}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-04274", 22}, { {"The variable decorated with Layer within the MeshEXT, MeshNV, Vertex, TessellationEvaluation, or Geometry Execution Model must be declared using the Output Storage Class", 169}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-04275", 22}, { {"The variable decorated with Layer within the Fragment Execution Model must be declared using the Input Storage Class", 116}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-04276", 22}, { {"The variable decorated with Layer must be declared as a scalar 32-bit integer value for all supported execution models except MeshEXT", 133}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-07039", 22}, { {"The variable decorated with Layer within the MeshEXT Execution Model must also be decorated with the PerPrimitiveEXT decoration", 127}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-10592", 22}, { {"Layer within the MeshEXT Execution Model must decorate a scalar 32-bit integer member of a structure decorated as Block, or decorate a variable of type OpTypeArray of scalar 32-bit integer values", 195}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-10593", 22}, { {"If Layer is declared as an array of 32-bit integer values, within the MeshEXT Execution Model, size of the array must match the value specified by OutputPrimitivesEXT", 166}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Layer-Layer-10594", 22}, { {"If Layer decorates a member of a structure, the variable declaration of the containing Block type must have an array size that matches the value specified by OutputPrimitivesEXT", 177}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LayerPerViewNV-LayerPerViewNV-04277", 40}, { {"The LayerPerViewNV decoration must be used only within the MeshNV Execution Model", 81}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LayerPerViewNV-LayerPerViewNV-04278", 40}, { {"The variable decorated with LayerPerViewNV must be declared using the Output Storage Class", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LayerPerViewNV-LayerPerViewNV-04279", 40}, { {"The variable decorated with LayerPerViewNV must also be decorated with the PerViewNV decoration", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LayerPerViewNV-LayerPerViewNV-04280", 40}, { {"The variable decorated with LayerPerViewNV must be declared as an array of scalar 32-bit integer values", 103}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationId-LocalInvocationId-04281", 46}, { {"The LocalInvocationId decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationId-LocalInvocationId-04282", 46}, { {"The variable decorated with LocalInvocationId must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationId-LocalInvocationId-04283", 46}, { {"The variable decorated with LocalInvocationId must be declared as a three-component vector of 32-bit integer values", 115}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationIndex-LocalInvocationIndex-04284", 52}, { {"The LocalInvocationIndex decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 127}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationIndex-LocalInvocationIndex-04285", 52}, { {"The variable decorated with LocalInvocationIndex must be declared using the Input Storage Class", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-LocalInvocationIndex-LocalInvocationIndex-04286", 52}, { {"The variable decorated with LocalInvocationIndex must be declared as a scalar 32-bit integer value", 98}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewCountNV-MeshViewCountNV-04287", 42}, { {"The MeshViewCountNV decoration must be used only within the MeshNV or TaskNV Execution Model", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewCountNV-MeshViewCountNV-04288", 42}, { {"The variable decorated with MeshViewCountNV must be declared using the Input Storage Class", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewCountNV-MeshViewCountNV-04289", 42}, { {"The variable decorated with MeshViewCountNV must be declared as a scalar 32-bit integer value", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewIndicesNV-MeshViewIndicesNV-04290", 46}, { {"The MeshViewIndicesNV decoration must be used only within the MeshNV or TaskNV Execution Model", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewIndicesNV-MeshViewIndicesNV-04291", 46}, { {"The variable decorated with MeshViewIndicesNV must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-MeshViewIndicesNV-MeshViewIndicesNV-04292", 46}, { {"The variable decorated with MeshViewIndicesNV must be declared as an array of scalar 32-bit integer values", 106}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumSubgroups-NumSubgroups-04293", 36}, { {"The NumSubgroups decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 119}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumSubgroups-NumSubgroups-04294", 36}, { {"The variable decorated with NumSubgroups must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumSubgroups-NumSubgroups-04295", 36}, { {"The variable decorated with NumSubgroups must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumWorkgroups-NumWorkgroups-04296", 38}, { {"The NumWorkgroups decoration must be used only within the GLCompute, MeshEXT, or TaskEXT Execution Model", 104}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumWorkgroups-NumWorkgroups-04297", 38}, { {"The variable decorated with NumWorkgroups must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-NumWorkgroups-NumWorkgroups-04298", 38}, { {"The variable decorated with NumWorkgroups must be declared as a three-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayDirectionKHR-ObjectRayDirectionKHR-04299", 54}, { {"The ObjectRayDirectionKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 126}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayDirectionKHR-ObjectRayDirectionKHR-04300", 54}, { {"The variable decorated with ObjectRayDirectionKHR must be declared using the Input Storage Class", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayDirectionKHR-ObjectRayDirectionKHR-04301", 54}, { {"The variable decorated with ObjectRayDirectionKHR must be declared as a three-component vector of 32-bit floating-point values", 126}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayOriginKHR-ObjectRayOriginKHR-04302", 48}, { {"The ObjectRayOriginKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayOriginKHR-ObjectRayOriginKHR-04303", 48}, { {"The variable decorated with ObjectRayOriginKHR must be declared using the Input Storage Class", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectRayOriginKHR-ObjectRayOriginKHR-04304", 48}, { {"The variable decorated with ObjectRayOriginKHR must be declared as a three-component vector of 32-bit floating-point values", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectToWorldKHR-ObjectToWorldKHR-04305", 44}, { {"The ObjectToWorldKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 121}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectToWorldKHR-ObjectToWorldKHR-04306", 44}, { {"The variable decorated with ObjectToWorldKHR must be declared using the Input Storage Class", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ObjectToWorldKHR-ObjectToWorldKHR-04307", 44}, { {"The variable decorated with ObjectToWorldKHR must be declared as a matrix with four columns of three-component vectors of 32-bit floating-point values", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PFN_vkDebugUtilsMessengerCallbackEXT-None-04769", 52}, { {"The callback must not make calls to any Vulkan commands", 55}, { "chapters/debugging.html", 23} } },
        { {"VUID-PatchVertices-PatchVertices-04308", 38}, { {"The PatchVertices decoration must be used only within the TessellationControl or TessellationEvaluation Execution Model", 119}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PatchVertices-PatchVertices-04309", 38}, { {"The variable decorated with PatchVertices must be declared using the Input Storage Class", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PatchVertices-PatchVertices-04310", 38}, { {"The variable decorated with PatchVertices must be declared as a scalar 32-bit integer value", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointCoord-PointCoord-04311", 32}, { {"The PointCoord decoration must be used only within the Fragment Execution Model", 79}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointCoord-PointCoord-04312", 32}, { {"The variable decorated with PointCoord must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointCoord-PointCoord-04313", 32}, { {"The variable decorated with PointCoord must be declared as a two-component vector of 32-bit floating-point values", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointSize-PointSize-04314", 30}, { {"The PointSize decoration must be used only within the MeshEXT, MeshNV, Vertex, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 151}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointSize-PointSize-04315", 30}, { {"The variable decorated with PointSize within the MeshEXT, MeshNV, or Vertex Execution Model must be declared using the Output Storage Class", 139}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointSize-PointSize-04316", 30}, { {"The variable decorated with PointSize within the TessellationControl, TessellationEvaluation, or Geometry Execution Model must not be declared using a Storage Class other than Input or Output", 191}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PointSize-PointSize-04317", 30}, { {"The variable decorated with PointSize must be declared as a scalar 32-bit floating-point value", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Position-Position-04318", 28}, { {"The Position decoration must be used only within the MeshEXT, MeshNV, Vertex, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Position-Position-04319", 28}, { {"The variable decorated with Position within the MeshEXT, MeshNV, or Vertex Execution Model must be declared using the Output Storage Class", 138}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Position-Position-04320", 28}, { {"The variable decorated with Position within the TessellationControl, TessellationEvaluation, or Geometry Execution Model must not be declared using a Storage Class other than Input or Output", 190}, { "chapters/interfaces.html", 24} } },
        { {"VUID-Position-Position-04321", 28}, { {"The variable decorated with Position must be declared as a four-component vector of 32-bit floating-point values", 112}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PositionPerViewNV-PositionPerViewNV-04322", 46}, { {"The PositionPerViewNV decoration must be used only within the MeshNV, Vertex, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PositionPerViewNV-PositionPerViewNV-04323", 46}, { {"The variable decorated with PositionPerViewNV within the Vertex, or MeshNV Execution Model must be declared using the Output Storage Class", 138}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PositionPerViewNV-PositionPerViewNV-04324", 46}, { {"The variable decorated with PositionPerViewNV within the TessellationControl, TessellationEvaluation, or Geometry Execution Model must not be declared using a Storage Class other than Input or Output", 199}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PositionPerViewNV-PositionPerViewNV-04325", 46}, { {"The variable decorated with PositionPerViewNV must be declared as an array of four-component vector of 32-bit floating-point values with at least as many elements as the maximum view in the subpass's view mask plus one", 218}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PositionPerViewNV-PositionPerViewNV-04326", 46}, { {"The array variable decorated with PositionPerViewNV must only be indexed by a constant or specialization constant", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveCountNV-PrimitiveCountNV-04327", 44}, { {"The PrimitiveCountNV decoration must be used only within the MeshNV Execution Model", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveCountNV-PrimitiveCountNV-04328", 44}, { {"The variable decorated with PrimitiveCountNV must be declared using the Output Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveCountNV-PrimitiveCountNV-04329", 44}, { {"The variable decorated with PrimitiveCountNV must be declared as a scalar 32-bit integer value", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-Fragment-04331", 31}, { {"If pipeline contains both the Fragment and Geometry Execution Model and a variable decorated with PrimitiveId is read from Fragment shader, then the Geometry shader must write to the output variables decorated with PrimitiveId in all execution paths", 249}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-Fragment-04332", 31}, { {"If pipeline contains both the Fragment and MeshEXT or MeshNV Execution Model and a variable decorated with PrimitiveId is read from Fragment shader, then the MeshEXT or MeshNV shader must write to the output variables decorated with PrimitiveId in all execution paths", 267}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-Fragment-04333", 31}, { {"If Fragment Execution Model contains a variable decorated with PrimitiveId, then either the MeshShadingEXT, MeshShadingNV, Geometry or Tessellation capability must also be declared", 180}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-04330", 34}, { {"The PrimitiveId decoration must be used only within the MeshEXT, MeshNV, IntersectionKHR, AnyHitKHR, ClosestHitKHR, TessellationControl, TessellationEvaluation, Geometry, or Fragment Execution Model", 198}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-04334", 34}, { {"The variable decorated with PrimitiveId within the TessellationControl, TessellationEvaluation, Fragment, IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model must be declared using the Input Storage Class", 213}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-04335", 34}, { {"The variable decorated with PrimitiveId within the Geometry Execution Model must be declared using the Input or Output Storage Class", 132}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-04336", 34}, { {"The variable decorated with PrimitiveId within the MeshEXT or MeshNV Execution Model must be declared using the Output Storage Class", 132}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-04337", 34}, { {"The variable decorated with PrimitiveId must be declared as a scalar 32-bit integer value for all supported execution models except MeshEXT", 139}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-07040", 34}, { {"The variable decorated with PrimitiveId within the MeshEXT Execution Model must also be decorated with the PerPrimitiveEXT decoration", 133}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-10595", 34}, { {"PrimitiveId within the MeshEXT Execution Model must decorate a scalar 32-bit integer member of a structure decorated as Block, or decorate a variable of type OpTypeArray of 32-bit integer values", 194}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-10596", 34}, { {"If PrimitiveId is declared as an array of 32-bit integer values, within the MeshEXT Execution Model, size of the array must match the value specified by OutputPrimitivesEXT", 172}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveId-PrimitiveId-10597", 34}, { {"If PrimitiveId decorates a member of a structure, the variable declaration of the containing Block type must have an array size that matches the value specified by OutputPrimitivesEXT", 183}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-OutputLinesNV-04343", 43}, { {"If the Execution Mode is OutputLinesNV, then the array decorated with PrimitiveIndicesNV must be the size of two times the value specified by OutputPrimitivesNV", 160}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-OutputPoints-04342", 42}, { {"If the Execution Mode is OutputPoints, then the array decorated with PrimitiveIndicesNV must be the size of the value specified by OutputPrimitivesNV", 149}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-OutputTrianglesNV-04344", 47}, { {"If the Execution Mode is OutputTrianglesNV, then the array decorated with PrimitiveIndicesNV must be the size of three times the value specified by OutputPrimitivesNV", 166}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-PrimitiveIndicesNV-04338", 48}, { {"The PrimitiveIndicesNV decoration must be used only within the MeshNV Execution Model", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-PrimitiveIndicesNV-04339", 48}, { {"The variable decorated with PrimitiveIndicesNV must be declared using the Output Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-PrimitiveIndicesNV-04340", 48}, { {"The variable decorated with PrimitiveIndicesNV must be declared as an array of scalar 32-bit integer values", 107}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveIndicesNV-PrimitiveIndicesNV-04341", 48}, { {"All index values of the array decorated with PrimitiveIndicesNV must be in the range [0, N-1], where N is the value specified by the OutputVertices Execution Mode", 162}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07047", 58}, { {"The PrimitiveLineIndicesEXT decoration must be used only within the MeshEXT Execution Model", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07048", 58}, { {"The PrimitiveLineIndicesEXT decoration must be used with the OutputLinesEXT Execution Mode", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07049", 58}, { {"The variable decorated with PrimitiveLineIndicesEXT must be declared using the Output Storage Class", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07050", 58}, { {"The variable decorated with PrimitiveLineIndicesEXT must be declared as an array of two component vector 32-bit integer values", 126}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07051", 58}, { {"The index to access the array decorated with PrimitiveLineIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Primitive Count\" operand of OpSetMeshOutputsEXT", 188}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-07052", 58}, { {"The size of the array decorated with PrimitiveLineIndicesEXT must match the value specified by OutputPrimitivesEXT", 114}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveLineIndicesEXT-PrimitiveLineIndicesEXT-12336", 58}, { {"All index values written to the array decorated with PrimitiveLineIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Vertex Count\" operand of OpSetMeshOutputsEXT", 193}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07041", 60}, { {"The PrimitivePointIndicesEXT decoration must be used only within the MeshEXT Execution Model", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07042", 60}, { {"The PrimitivePointIndicesEXT decoration must be used with the OutputPoints Execution Mode", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07043", 60}, { {"The variable decorated with PrimitivePointIndicesEXT must be declared using the Output Storage Class", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07044", 60}, { {"The variable decorated with PrimitivePointIndicesEXT must be declared as an array of scalar 32-bit integer values", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07045", 60}, { {"The index to access the array decorated with PrimitivePointIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Primitive Count\" operand of OpSetMeshOutputsEXT", 189}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-07046", 60}, { {"The size of the array decorated with PrimitivePointIndicesEXT must match the value specified by OutputPrimitivesEXT", 115}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitivePointIndicesEXT-PrimitivePointIndicesEXT-12335", 60}, { {"All index values written to the array decorated with PrimitivePointIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Vertex Count\" operand of OpSetMeshOutputsEXT", 194}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04484", 58}, { {"The PrimitiveShadingRateKHR decoration must be used only within the MeshEXT, MeshNV, Vertex, or Geometry Execution Model", 120}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04485", 58}, { {"The variable decorated with PrimitiveShadingRateKHR must be declared using the Output Storage Class", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04486", 58}, { {"The variable decorated with PrimitiveShadingRateKHR must be declared as a scalar 32-bit integer value for all supported execution models except MeshEXT", 151}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04487", 58}, { {"The value written to PrimitiveShadingRateKHR must include no more than one of Vertical2Pixels and Vertical4Pixels", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04488", 58}, { {"The value written to PrimitiveShadingRateKHR must include no more than one of Horizontal2Pixels and Horizontal4Pixels", 117}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-04489", 58}, { {"The value written to PrimitiveShadingRateKHR must not have any bits set other than those defined by Fragment Shading Rate Flags enumerants in the SPIR-V specification", 166}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-07059", 58}, { {"The variable decorated with PrimitiveShadingRateKHR within the MeshEXT Execution Model must also be decorated with the PerPrimitiveEXT decoration", 145}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-10598", 58}, { {"PrimitiveShadingRateKHR within the MeshEXT Execution Model must decorate a scalar 32-bit integer member of a structure decorated as Block, or decorate a variable of type OpTypeArray of 32-bit integer values", 206}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-10599", 58}, { {"If PrimitiveShadingRateKHR is declared as an array of 32-bit integer values, within the MeshEXT Execution Model, size of the array must match the value specified by OutputPrimitivesEXT", 184}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-10600", 58}, { {"If PrimitiveShadingRateKHR decorates a member of a structure, the variable declaration of the containing Block type must have an array size that matches the value specified by OutputPrimitivesEXT", 195}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveShadingRateKHR-PrimitiveShadingRateKHR-12275", 58}, { {"If PrimitiveShadingRateKHR is declared in the MeshEXT Execution Model, the primitiveFragmentShadingRateMeshShader feature must be enabled", 137}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07053", 66}, { {"The PrimitiveTriangleIndicesEXT decoration must be used only within the MeshEXT Execution Model", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07054", 66}, { {"The PrimitiveTriangleIndicesEXT decoration must be used with the OutputTrianglesEXT Execution Mode", 98}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07055", 66}, { {"The variable decorated with PrimitiveTriangleIndicesEXT must be declared using the Output Storage Class", 103}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07056", 66}, { {"The variable decorated with PrimitiveTriangleIndicesEXT must be declared as an array of three component vector 32-bit integer values", 132}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07057", 66}, { {"The index to access the array decorated with PrimitiveTriangleIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Primitive Count\" operand of OpSetMeshOutputsEXT", 192}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-07058", 66}, { {"The size of the array decorated with PrimitiveTriangleIndicesEXT must match the value specified by OutputPrimitivesEXT", 118}, { "chapters/interfaces.html", 24} } },
        { {"VUID-PrimitiveTriangleIndicesEXT-PrimitiveTriangleIndicesEXT-12337", 66}, { {"All index values written to the array decorated with PrimitiveTriangleIndicesEXT must be in the range [0, N-1], where N is the value specified by the \"Vertex Count\" operand of OpSetMeshOutputsEXT", 197}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayGeometryIndexKHR-RayGeometryIndexKHR-04345", 50}, { {"The RayGeometryIndexKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayGeometryIndexKHR-RayGeometryIndexKHR-04346", 50}, { {"The variable decorated with RayGeometryIndexKHR must be declared using the Input Storage Class", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayGeometryIndexKHR-RayGeometryIndexKHR-04347", 50}, { {"The variable decorated with RayGeometryIndexKHR must be declared as a scalar 32-bit integer value", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTmaxKHR-RayTmaxKHR-04348", 32}, { {"The RayTmaxKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTmaxKHR-RayTmaxKHR-04349", 32}, { {"The variable decorated with RayTmaxKHR must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTmaxKHR-RayTmaxKHR-04350", 32}, { {"The variable decorated with RayTmaxKHR must be declared as a scalar 32-bit floating-point value", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTminKHR-RayTminKHR-04351", 32}, { {"The RayTminKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTminKHR-RayTminKHR-04352", 32}, { {"The variable decorated with RayTminKHR must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-RayTminKHR-RayTminKHR-04353", 32}, { {"The variable decorated with RayTminKHR must be declared as a scalar 32-bit floating-point value", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ResourceHeapEXT-ResourceHeapEXT-11241", 42}, { {"The variable decorated with ResourceHeapEXT must be declared using the UniformConstant Storage Class", 100}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-RuntimeSpirv-ConstOffset-10718", 35}, { {"If an OpTypeSampledImage variable backed by an image view and sampler that require sampler Y'CBCR conversion is statically used in a shader with a sampling instruction, it must not use the ConstOffset or Offset operands", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Coordinate-10713", 34}, { {"The Coordinate operand of any OpImageRead, OpImageSparseRead, OpImageWrite, OpUntypedImageTexelPointerEXT or OpImageTexelPointer instruction that consumes an OpTypeImage with an image Storage Class TileAttachmentQCOM must not result in any texels accessed outside the boundaries of the current tile, computed as described in Tile Attachments", 341}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Coordinate-10714", 34}, { {"The Coordinate operand(s) of any of the following instructions that consumes an OpTypeSampledImage with an image of Storage Class TileAttachmentQCOM must not result in any texels accessed outside boundaries of the current tile, computed as described in Tile Attachments: OpImageSample* OpImageSparseSample* OpImageFetch OpImageSparseFetch OpImage*Gather OpImageSparse*Gather OpImageSampleWeightedQCOM OpImageBoxFilterQCOM OpImageBlockMatch*QCOM OpImageGatherQCOM", 462}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-DescriptorSet-11385", 37}, { {"If an instruction accesses memory through any resource with a DescriptorSet and Binding that are mapped using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, the resource through which that memory is accessed must be determined by constant integral expressions", 519}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-DescriptorSet-11388", 37}, { {"All possible values of a variable pointer to a resource must not be resources with a DescriptorSet and Binding that are mapped using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT", 438}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-GraphARM-09922", 32}, { {"The GraphARM capability must not be declared in modules used to create a shader stage", 85}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Image-11379", 29}, { {"If the Image operand of OpImageTexelPointer or OpUntypedImageTexelPointerEXT is derived from a variable decorated with ResourceHeapEXT, it must correspond to accessing a descriptor matching one of the descriptor types listed in Heap Resource Type Correspondence for the OpTypeImage pointed to", 292}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Image-11380", 29}, { {"If the Image operand of OpImageTexelPointer or OpUntypedImageTexelPointerEXT is derived from a variable decorated with ResourceHeapEXT, the Image Format and Sampled Type of the OpTypeImage pointed to must correspond to a descriptor with a VkFormat that matches as described by Compatibility Between SPIR-V Image Formats and Vulkan Formats and Image Format and Type Matching", 373}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Image-11382", 29}, { {"If the Image operand of OpImageTexelPointer or OpUntypedImageTexelPointerEXT is derived from a variable decorated with ResourceHeapEXT, the operands of that OpTypeImage must correspond to a descriptor with operands that match as described by Heap Image Operand Correspondence and Heap Image View Type Dimensionality", 315}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-InputAttachmentArrayNonUniformIndexing-10138", 62}, { {"If the InputAttachmentArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through an input attachment, the input attachment through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 264}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-KSize-12353", 29}, { {"For OpExtractSubArrayQCOM, the length of the Result Type operand must match the VkCooperativeMatrixPropertiesKHR::KSize of one of the matrices in any of the supported VkCooperativeMatrixPropertiesKHR", 199}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-LocalSizeId-06434", 35}, { {"If Execution Mode LocalSizeId is used, maintenance4 must be enabled", 67}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Location-06272", 32}, { {"The sum of Location and the number of locations the variable it decorates consumes must be less than or equal to the value for the matching Execution Model defined in Shader Input and Output Locations", 200}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Location-06428", 32}, { {"The maximum number of storage buffers, storage images, and output Location decorated color attachments written to in the Fragment Execution Model must be less than or equal to maxFragmentCombinedOutputResources", 210}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07115", 31}, { {"For mesh shaders using the MeshEXT Execution Model the OutputVertices OpExecutionMode must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshOutputVertices", 178}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07116", 31}, { {"For mesh shaders using the MeshEXT Execution Model the OutputPrimitivesEXT OpExecutionMode must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshOutputPrimitives", 185}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07118", 31}, { {"In mesh shaders using the MeshEXT Execution Model OpSetMeshOutputsEXT must be called at most once under dynamically uniform conditions", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07295", 31}, { {"For mesh shaders using the MeshEXT Execution Model the x size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupSize[0]", 184}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07296", 31}, { {"For mesh shaders using the MeshEXT Execution Model the y size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupSize[1]", 184}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07297", 31}, { {"For mesh shaders using the MeshEXT Execution Model the z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupSize[2]", 184}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07298", 31}, { {"For mesh shaders using the MeshEXT Execution Model the product of x size, y size, and z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupInvocations", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07332", 31}, { {"For mesh shaders using the MeshEXT Execution Model the \"Vertex Count\" operand of OpSetMeshOutputsEXT must be less than or equal to OutputVertices OpExecutionMode", 163}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-07333", 31}, { {"For mesh shaders using the MeshEXT Execution Model the \"Primitive Count\" operand of OpSetMeshOutputsEXT must be less than or equal to OutputPrimitivesEXT OpExecutionMode", 171}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-09218", 31}, { {"In mesh shaders using the MeshEXT or MeshNV Execution Model and the OutputPoints Execution Mode, if the maintenance5 feature is not enabled, and if the number of output points is greater than 0, a PointSize decorated variable must be written to for each output point", 266}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-12333", 31}, { {"For mesh shaders using the MeshEXT Execution Model the index into the array of any variable decorated with the PerPrimitiveEXT decoration must be less than the \"Primitive Count\" operand of OpSetMeshOutputsEXT", 210}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshEXT-12380", 31}, { {"If the MeshEXT Execution Model dynamically accesses a variable with the TaskPayloadWorkgroupEXT Storage Class, there must be a matching TaskPayloadWorkgroupEXT Storage Class variable in the TaskEXT Execution Model passed as an argument to OpEmitMeshTasksEXT", 257}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshNV-07113", 30}, { {"For mesh shaders using the MeshNV Execution Model the OutputVertices OpExecutionMode must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesNV::maxMeshOutputVertices", 176}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-MeshNV-07114", 30}, { {"For mesh shaders using the MeshNV Execution Model the OutputPrimitivesNV OpExecutionMode must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesNV::maxMeshOutputPrimitives", 182}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-NSize-12352", 29}, { {"For OpExtractSubArrayQCOM, the length of the Source Array operand must match the VkCooperativeMatrixPropertiesKHR::NSize of one of the matrices in any of the supported VkCooperativeMatrixPropertiesKHR", 200}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-NodePayloadAMDX-09192", 39}, { {"Variables in the NodePayloadAMDX storage class must only be declared in the GLCompute or MeshEXT execution model", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-NonWritable-06340", 35}, { {"If the fragmentStoresAndAtomics feature is not enabled, then all storage image, storage texel buffer, storage tensor, and storage buffer variables in the fragment stage must be decorated with the NonWritable decoration", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-NonWritable-06341", 35}, { {"If the vertexPipelineStoresAndAtomics feature is not enabled, then all storage image, storage texel buffer, storage tensor, and storage buffer variables in the vertex, tessellation, and geometry stages must be decorated with the NonWritable decoration", 251}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-04745", 28}, { {"All block members in a variable with a Storage Class of PushConstant declared as an array must only be accessed by dynamically uniform indices", 142}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06275", 28}, { {"shaderSubgroupExtendedTypes must be enabled for group operations to use 8-bit integer, 16-bit integer, 64-bit integer, 16-bit floating-point, and vectors of these types", 168}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06278", 28}, { {"shaderBufferInt64Atomics must be enabled for 64-bit integer atomic operations to be supported on a Pointer with a Storage Class of StorageBuffer or Uniform", 155}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06279", 28}, { {"shaderSharedInt64Atomics must be enabled for 64-bit integer atomic operations to be supported on a Pointer with a Storage Class of Workgroup", 140}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06284", 28}, { {"shaderBufferFloat32Atomics, or shaderBufferFloat32AtomicAdd, or shaderBufferFloat64Atomics, or shaderBufferFloat64AtomicAdd, or shaderBufferFloat16Atomics, or shaderBufferFloat16AtomicAdd, or shaderBufferFloat16AtomicMinMax, or shaderBufferFloat32AtomicMinMax, or shaderBufferFloat64AtomicMinMax, or shaderFloat16VectorAtomics must be enabled for floating-point atomic operations to be supported on a Pointer with a Storage Class of StorageBuffer", 446}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06285", 28}, { {"shaderSharedFloat32Atomics, or shaderSharedFloat32AtomicAdd, or shaderSharedFloat64Atomics, or shaderSharedFloat64AtomicAdd, or shaderSharedFloat16Atomics, or shaderSharedFloat16AtomicAdd, or shaderSharedFloat16AtomicMinMax, or shaderSharedFloat32AtomicMinMax, or shaderSharedFloat64AtomicMinMax, or shaderFloat16VectorAtomics, must be enabled for floating-point atomic operations to be supported on a Pointer with a Storage Class of Workgroup", 443}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06286", 28}, { {"shaderImageFloat32Atomics, or shaderImageFloat32AtomicAdd, or shaderImageFloat32AtomicMinMax, must be enabled for 32-bit floating-point atomic operations to be supported on a Pointer with a Storage Class of Image", 212}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06287", 28}, { {"sparseImageFloat32Atomics, or sparseImageFloat32AtomicAdd, or sparseImageFloat32AtomicMinMax, must be enabled for 32-bit floating-point atomics to be supported on sparse images", 176}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06288", 28}, { {"shaderImageInt64Atomics must be enabled for 64-bit integer atomic operations to be supported on a Pointer with a Storage Class of Image", 135}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06337", 28}, { {"shaderBufferFloat16Atomics, or shaderBufferFloat16AtomicAdd, or shaderBufferFloat16AtomicMinMax, or shaderSharedFloat16Atomics, or shaderSharedFloat16AtomicAdd, or shaderSharedFloat16AtomicMinMax must be enabled for 16-bit floating-point atomic operations", 255}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06338", 28}, { {"shaderBufferFloat32Atomics, or shaderBufferFloat32AtomicAdd, or shaderSharedFloat32Atomics, or shaderSharedFloat32AtomicAdd, or shaderImageFloat32Atomics, or shaderImageFloat32AtomicAdd or shaderBufferFloat32AtomicMinMax, or shaderSharedFloat32AtomicMinMax, or shaderImageFloat32AtomicMinMax must be enabled for 32-bit floating-point atomic operations", 351}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06339", 28}, { {"shaderBufferFloat64Atomics, or shaderBufferFloat64AtomicAdd, or shaderSharedFloat64Atomics, or shaderSharedFloat64AtomicAdd, or shaderBufferFloat64AtomicMinMax, or shaderSharedFloat64AtomicMinMax, must be enabled for 64-bit floating-point atomic operations", 256}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06342", 28}, { {"If subgroupQuadOperationsInAllStages is VK_FALSE, then quad subgroup operations must not be used except for in fragment and compute stages", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-06343", 28}, { {"Group operations with subgroup scope must not be used if the shader stage is not in subgroupSupportedStages", 107}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-09558", 28}, { {"If the dynamicRenderingLocalRead feature is not enabled, any variable created with a \"Type\" of OpTypeImage that has a \"Dim\" operand of SubpassData must be decorated with InputAttachmentIndex", 194}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-10148", 28}, { {"If an instruction accesses memory through any resource, the effective subgroup size is 1, the DescriptorHeapEXT capability is not declared, and the resource through which that memory is accessed is not uniform within the invocation group, then the operand corresponding to that resource (e.g. the pointer or sampled image operand) must be decorated with NonUniform", 364}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-10824", 28}, { {"If the maintenance9 feature is not enabled, the Base operand of any OpBitCount, OpBitReverse, OpBitFieldInsert, OpBitFieldSExtract, or OpBitFieldUExtract instruction must be a 32-bit integer scalar or a vector of 32-bit integers", 228}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-10834", 28}, { {"Buffer indexing calculations must not wrap 32 bits if the pipeline or shader was not created with the VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT or VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT flags and the entry point does not set the Shader64BitIndexingEXT execution mode", 271}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-10980", 28}, { {"Objects containing 8-bit integer or 8-bit floating-point elements must not have storage class of Input or Output", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-None-12205", 28}, { {"If an image view or sampler that requires sampler Y'CBCR conversion is accessed in a shader, it must be determined by constant integral expressions", 147}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-06308", 30}, { {"The Offset plus size of the type of each variable, in the output interface of the entry point being compiled, decorated with XfbBuffer must not be greater than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBufferDataSize", 242}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-06344", 30}, { {"The first element of the Offset operand of InterpolateAtOffset must be greater than or equal to: fragwidth x minInterpolationOffset where fragwidth is the width of the current fragment in pixels", 194}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-06345", 30}, { {"The first element of the Offset operand of InterpolateAtOffset must be less than or equal to fragwidth x (maxInterpolationOffset + ULP ) - ULP where fragwidth is the width of the current fragment in pixels and ULP = 1 / 2^subPixelInterpolationOffsetBits^", 254}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-06346", 30}, { {"The second element of the Offset operand of InterpolateAtOffset must be greater than or equal to fragheight x minInterpolationOffset where fragheight is the height of the current fragment in pixels", 197}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-06347", 30}, { {"The second element of the Offset operand of InterpolateAtOffset must be less than or equal to fragheight x (maxInterpolationOffset + ULP ) - ULP where fragheight is the height of the current fragment in pixels and ULP = 1 / 2^subPixelInterpolationOffsetBits^", 258}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Offset-10213", 30}, { {"If the maintenance8 feature is not enabled, image operand Offset must only be used with OpImage*Gather or OpImageGatherQCOM instructions", 136}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpArrayLength-11807", 37}, { {"OpArrayLength and OpUntypedArrayLengthKHR result type must be a 32-bit integer type if the pipeline or shader was not compiled with the VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT or VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT flags and the entry point does not set the Shader64BitIndexingEXT execution mode", 305}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpConstantSizeOfEXT-11475", 43}, { {"The result type of OpConstantSizeOfEXT must not be a 64-bit integer type if the pipeline or shader was not compiled with the VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT or VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT flags and the entry point does not set the Shader64BitIndexingEXT execution mode", 294}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpControlBarrierArriveEXT-13558", 49}, { {"If a OpControlBarrierArriveEXT or OpControlBarrierWaitEXT instruction is executed with an execution scope equal to Workgroup, its execution must be inside uniform control flow within the invocation group", 203}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpControlBarrierArriveEXT-13561", 49}, { {"If a OpControlBarrierArriveEXT instruction is executed, another OpControlBarrierArriveEXT instruction with the same execution scope must not have been previously executed unless a OpControlBarrierWaitEXT instruction with the same execution scope has been executed after it", 272}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpControlBarrierArriveEXT-13562", 49}, { {"If the module uses OpControlBarrierArriveEXT or OpControlBarrierWaitEXT, the sum of size in bytes for variables and padding in the Workgroup Storage Class in the GLCompute Execution Model must be less than or equal to maxComputeSharedMemorySize minus splitBarrierReservedSharedMemory", 283}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpControlBarrierWaitEXT-13559", 47}, { {"If a OpControlBarrierWaitEXT instruction is executed, a OpControlBarrierArriveEXT instruction with the same execution scope must have been executed previously", 158}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpControlBarrierWaitEXT-13560", 47}, { {"If a OpControlBarrierWaitEXT instruction is executed, another OpControlBarrierWaitEXT instruction with the same execution scope must not have been previously executed prior to the last executed OpControlBarrierArriveEXT instruction with the same execution scope", 261}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeMatrix-12379", 43}, { {"If any OpCooperativeMatrix* instructions accesses a descriptor, it must not be a null descriptor", 96}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeMatrixLoadKHR-08986", 50}, { {"For OpCooperativeMatrixLoadKHR and OpCooperativeMatrixStoreKHR instructions, the Pointer and Stride operands must be aligned to at least the lesser of 16 bytes or the natural alignment of a row or column (depending on ColumnMajor) of the matrix (where the natural alignment is the number of columns/rows multiplied by the component size)", 337}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeMatrixLoadNV-06324", 49}, { {"For OpCooperativeMatrixLoadNV and OpCooperativeMatrixStoreNV instructions, the Pointer and Stride operands must be aligned to at least the lesser of 16 bytes or the natural alignment of a row or column (depending on ColumnMajor) of the matrix (where the natural alignment is the number of columns/rows multiplied by the component size)", 335}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeMatrixLoadTensorNV-12429", 55}, { {"For each OpCooperativeMatrixLoadTensorNV instruction that includes DecodeVectorFunc in Tensor Addressing Operands, the tensor block size in the innermost dimension must be a multiple of the number of components of the DecodeVectorFunc function's return type.", 258}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeMatrixMulAddKHR-10060", 52}, { {"For OpCooperativeMatrixMulAddKHR, the operands must match a supported VkCooperativeMatrixPropertiesKHR", 102}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorLoadNV-10099", 49}, { {"For OpCooperativeVectorLoadNV and OpCooperativeVectorStoreNV instructions, the Pointer and Offset must be aligned to 16 bytes", 125}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorLoadNV-11809", 49}, { {"OpCooperativeVectorLoadNV, OpCooperativeVectorStoreNV, OpCooperativeVectorOuterProductAccumulateNV, and OpCooperativeVectorReduceSumAccumulateNV Offset parameters must be 32-bit integer types if the pipeline or shader was not compiled with the VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT or VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT flags and the entry point does not set the Shader64BitIndexingEXT execution mode", 413}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulAddNV-10098", 57}, { {"For OpCooperativeVectorMatrixMulAddNV instructions, the Bias and BiasOffset must be aligned to 16 bytes", 103}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulAddNV-11808", 57}, { {"OpCooperativeVectorMatrixMulAddNV and OpCooperativeVectorMatrixMulNV MatrixOffset and BiasOffset parameters must be 32-bit integer types if the pipeline or shader was not compiled with the VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT or VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT flags and the entry point does not set the Shader64BitIndexingEXT execution mode", 358}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulNV-10089", 54}, { {"For OpCooperativeVectorMatrixMulNV and OpCooperativeVectorMatrixMulAddNV, the following must be satisfied by the same entry in the VkCooperativeVectorPropertiesNV array returned by vkGetPhysicalDeviceCooperativeVectorPropertiesNV: The component type of Input must match VkCooperativeVectorPropertiesNV::inputType The InputInterpretation must match VkCooperativeVectorPropertiesNV::inputInterpretation The MatrixInterpretation must match VkCooperativeVectorPropertiesNV::matrixInterpretation For OpCooperativeVectorMatrixMulAddNV, the BiasInterpretation must match VkCooperativeVectorPropertiesNV::biasInterpretation The Result Type must match VkCooperativeVectorPropertiesNV::resultType If Transpose is true, VkCooperativeVectorPropertiesNV::transpose must be VK_TRUE", 767}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulNV-10090", 54}, { {"For OpCooperativeVectorMatrixMulNV and OpCooperativeVectorMatrixMulAddNV, if MatrixInterpretation is either VK_COMPONENT_TYPE_FLOAT_E4M3_NV or VK_COMPONENT_TYPE_FLOAT_E5M2_NV then MemoryLayout must be either VK_COOPERATIVE_VECTOR_MATRIX_LAYOUT_INFERENCING_OPTIMAL_NV or VK_COOPERATIVE_VECTOR_MATRIX_LAYOUT_TRAINING_OPTIMAL_NV", 325}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulNV-10096", 54}, { {"For OpCooperativeVectorMatrixMulNV and OpCooperativeVectorMatrixMulAddNV instructions using non-optimal layouts, the Stride operand must be aligned to 16 bytes", 159}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorMatrixMulNV-10097", 54}, { {"For OpCooperativeVectorMatrixMulNV and OpCooperativeVectorMatrixMulAddNV instructions, the Matrix and MatrixOffset must be aligned to 64 bytes", 142}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorOuterProductAccumulateNV-10093", 67}, { {"For OpCooperativeVectorOuterProductAccumulateNV: MatrixInterpretation must be VK_COMPONENT_TYPE_FLOAT16_KHR or VK_COMPONENT_TYPE_FLOAT32_KHR If MatrixInterpretation is VK_COMPONENT_TYPE_FLOAT16_KHR, cooperativeVectorTrainingFloat16Accumulation must be supported If MatrixInterpretation is VK_COMPONENT_TYPE_FLOAT32_KHR, cooperativeVectorTrainingFloat32Accumulation must be supported The component types of A and B must be 16-bit floating-point The matrix layout must be training-optimal The pointer's storage class must be StorageBuffer or PhysicalStorageBuffer", 561}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorOuterProductAccumulateNV-10101", 67}, { {"For OpCooperativeVectorOuterProductAccumulateNV instructions, the Pointer and Offset must be aligned to 64 bytes", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorReduceSumAccumulateNV-10092", 64}, { {"For OpCooperativeVectorReduceSumAccumulateNV: The component type of V must be either 16- or 32-bit floating-point If the component type of V is 16-bit floating-point, cooperativeVectorTrainingFloat16Accumulation must be supported If the component type of V is 32-bit floating-point, cooperativeVectorTrainingFloat32Accumulation must be supported The pointer's storage class must be StorageBuffer or PhysicalStorageBuffer", 420}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpCooperativeVectorReduceSumAccumulateNV-10100", 64}, { {"For OpCooperativeVectorReduceSumAccumulateNV instructions, the Pointer and Offset must be aligned to 16 bytes", 109}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpEmitStreamVertex-06310", 42}, { {"The Stream value to OpEmitStreamVertex and OpEndStreamPrimitive must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 157}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpEntryPoint-07754", 36}, { {"Any user-defined variables between the OpEntryPoint of two shader stages must have the same type and width for each Component", 125}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpEntryPoint-08727", 36}, { {"Each OpEntryPoint must not have more than one variable decorated with InputAttachmentIndex per image aspect of the attachment image bound to it, either explicitly or implicitly as described by input attachment interface", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpEntryPoint-08743", 36}, { {"Any user-defined variables shared between the OpEntryPoint of two shader stages, and declared with Input as its Storage Class for the subsequent shader stage, must have all Location slots and Component words declared in the preceding shader stage's OpEntryPoint with Output as the Storage Class", 294}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpExtractSubArrayQCOM-12354", 45}, { {"For OpExtractSubArrayQCOM, the Start Index operand must be a multiple of the length of the Result Type operand", 110}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11878", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, all components of the RayOrigin and RayDirection operands must be finite floating-point values", 244}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11879", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, the RayTmin and RayTmax operands must be non-negative floating-point values", 225}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11880", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, the RayTmin operand must be less than or equal to the RayTmax operand", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11881", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, RayOrigin, RayDirection, RayTmin, and RayTmax operands must not contain NaNs", 226}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11882", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions the time operand must be between 0.0 and 1.0", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11883", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, the Rayflags operand must not contain both SkipTrianglesKHR and SkipAABBsKHR", 226}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11884", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, the Rayflags operand must not contain more than one of SkipTrianglesKHR, CullBackFacingTrianglesKHR, and CullFrontFacingTrianglesKHR", 282}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11885", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, the Rayflags operand must not contain more than one of OpaqueKHR, NoOpaqueKHR, CullOpaqueKHR, and CullNoOpaqueKHR", 263}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11886", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, if the Rayflags operand contains SkipTrianglesKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR set", 299}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayEXT-11887", 46}, { {"For OpHitObjectTraceRayEXT, OpHitObjectTraceRayMotionEXT, OpHitObjectTraceMotionReorderExecuteEXT and OpHitObjectTraceReorderExecuteEXT instructions, if the Rayflags operand contains SkipAABBsKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR set", 299}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayMotionEXT-11877", 52}, { {"For OpHitObjectTraceRayMotionEXT and OpHitObjectTraceMotionReorderExecuteEXT instructions, the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 186}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayMotionNV-07704", 51}, { {"For OpHitObjectTraceRayMotionNV instructions, if Acceleration Structure was built with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags, the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 238}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayMotionNV-07709", 51}, { {"For OpHitObjectTraceRayMotionNV instructions, Acceleration Structure must be an acceleration structure built as a top-level acceleration structure with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags", 206}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayMotionNV-07711", 51}, { {"For OpHitObjectTraceRayMotionNV instructions the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 140}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07705", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, all components of the RayOrigin and RayDirection operands must be finite floating-point values", 166}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07706", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, the RayTmin and RayTmax operands must be non-negative floating-point values", 147}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07707", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, the RayTmin operand must be less than or equal to the RayTmax operand", 141}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07708", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, RayOrigin, RayDirection, RayTmin, and RayTmax operands must not contain NaNs", 148}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07710", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions the time operand must be between 0.0 and 1.0", 115}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07712", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, the Rayflags operand must not contain both SkipTrianglesKHR and SkipAABBsKHR", 148}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07713", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, the Rayflags operand must not contain more than one of SkipTrianglesKHR, CullBackFacingTrianglesKHR, and CullFrontFacingTrianglesKHR", 204}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07714", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, the Rayflags operand must not contain more than one of OpaqueKHR, NoOpaqueKHR, CullOpaqueKHR, and CullNoOpaqueKHR", 185}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07715", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, if the Rayflags operand contains SkipTrianglesKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR set", 221}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpHitObjectTraceRayNV-07716", 45}, { {"For OpHitObjectTraceRayNV and OpHitObjectTraceRayMotionNV instructions, if the Rayflags operand contains SkipAABBsKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR set", 221}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImage-06376", 31}, { {"If an OpImage*Gather or OpImageGatherQCOM operation has an image operand of Offset, ConstOffset, or ConstOffsets the offset value must be greater than or equal to minTexelGatherOffset", 183}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImage-06377", 31}, { {"If an OpImage*Gather or OpImageGatherQCOM operation has an image operand of Offset, ConstOffset, or ConstOffsets the offset value must be less than or equal to maxTexelGatherOffset", 180}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImage-10706", 31}, { {"An OpImage with Storage Class TileAttachmentQCOM must not be consumed by OpImageTexelPointer or OpUntypedImageTexelPointerEXT unless the tileShadingAtomicOps feature is enabled", 176}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSADQCOM-06982", 48}, { {"If an OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM operation is used, then the target sampled image, reference sampled image, and Block Size parameters must both be dynamically uniform for the quad", 205}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSSDQCOM-06983", 48}, { {"If an OpImageBlockMatchSSDQCOM or OpImageBlockMatchSADQCOM operation is used, then target sampled image and reference sampled image parameters must be of storage class UniformConstant and type OpTypeImage with Depth=0, Dim=2D, Arrayed=0, MS=0, and Sampled=1", 257}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSSDQCOM-06984", 48}, { {"If an OpImageBlockMatchSSDQCOM or OpImageBlockMatchSADQCOM operation is used, then the target sampled image and reference sampled image parameters must be decorated with BlockMatchTextureQCOM", 191}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSSDQCOM-06985", 48}, { {"If an OpImageBlockMatchSSDQCOM or OpImageBlockMatchSADQCOM operation is used, then target sampled image and reference sampled image parameters must have been created using an identical sampler object", 199}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSSDQCOM-06986", 48}, { {"If an OpImageBlockMatchSSDQCOM or OpImageBlockMatchSADQCOM operation is used, then target sampled image and reference sampled image parameters must have been created with a sampler object with unnormalizedCoordinates equal to VK_TRUE", 233}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchSSDQCOM-06988", 48}, { {"If an OpImageBlockMatchSSDQCOM or OpImageBlockMatchSADQCOM operation is used, then Block Size less than or equal to maxBlockMatchRegion", 135}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchWindow-09219", 47}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then the target sampled image, reference sampled image, and Block Size parameters must both be dynamically uniform for the quad", 212}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchWindow-09220", 47}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then target sampled image and reference sampled image parameters must be of storage class UniformConstant and type OpTypeImage with Depth=0, Dim=2D, Arrayed=0, MS=0, and Sampled=1", 264}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchWindow-09221", 47}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then the target sampled image and reference sampled image parameters must be decorated with BlockMatchTextureQCOM", 198}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchWindow-09222", 47}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then target sampled image and reference sampled image parameters must have been created using an identical sampler object", 206}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBlockMatchWindow-09223", 47}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then target sampled image and reference sampled image parameters must have been created with a sampler object with unnormalizedCoordinates equal to VK_TRUE", 240}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBoxFilterQCOM-06989", 44}, { {"If an OpImageBoxFilterQCOM operation is used, then Box Size.y must be equal to or greater than 1.0 and less than or equal to maxBoxFilterBlockSize.height", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageBoxFilterQCOM-06990", 44}, { {"If an OpImageBoxFilterQCOM operation is used, then Sampled Texture Image and Box Size parameters must be dynamically uniform", 124}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageSample-06435", 37}, { {"If an OpImageSample* or OpImageFetch* operation has an image operand of Offset or ConstOffset then the offset value must be greater than or equal to minTexelOffset", 163}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageSample-06436", 37}, { {"If an OpImageSample* or OpImageFetch* operation has an image operand of Offset or ConstOffset then the offset value must be less than or equal to maxTexelOffset", 160}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageSampleWeightedQCOM-06979", 49}, { {"If an OpImageSampleWeightedQCOM operation is used, then the Texture Sampled Image and Weight Image parameters must both be dynamically uniform for the quad", 155}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageSampleWeightedQCOM-06980", 49}, { {"If an OpImageSampleWeightedQCOM operation is used, then the Weight Image parameter must be of Storage Class UniformConstant and type OpTypeImage with Depth=0, Dim=2D, Arrayed=1, MS=0, and Sampled=1", 197}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageSampleWeightedQCOM-06981", 49}, { {"If an OpImageSampleWeightedQCOM operation is used, then the Weight Image parameter must be decorated with WeightTextureQCOM", 123}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpImageWrite-07112", 36}, { {"OpImageWrite to any Image whose Image Format is not Unknown must have the Texel operand contain at least as many components as the corresponding VkFormat as given in the SPIR-V Image Format compatibility table", 209}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryGenerateIntersectionKHR-06353", 57}, { {"For OpRayQueryGenerateIntersectionKHR instructions, Hit T must satisfy the condition RayTmin <= Hit T <= RayTmax, where RayTmin is equal to the value returned by OpRayQueryGetRayTMinKHR with the same ray query object, and RayTmax is equal to the value of OpRayQueryGetIntersectionTKHR for the current committed intersection with the same ray query object", 354}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryGenerateIntersectionKHR-06354", 57}, { {"For OpRayQueryGenerateIntersectionKHR instructions, Acceleration Structure must not be built with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags", 152}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06348", 47}, { {"For OpRayQueryInitializeKHR instructions, all components of the RayOrigin and RayDirection operands must be finite floating-point values", 136}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06349", 47}, { {"For OpRayQueryInitializeKHR instructions, the RayTmin and RayTmax operands must be non-negative floating-point values", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06350", 47}, { {"For OpRayQueryInitializeKHR instructions, the RayTmin operand must be less than or equal to the RayTmax operand", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06351", 47}, { {"For OpRayQueryInitializeKHR instructions, RayOrigin, RayDirection, RayTmin, and RayTmax operands must not contain NaNs", 118}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06352", 47}, { {"For OpRayQueryInitializeKHR instructions, Acceleration Structure must be an acceleration structure built as a top-level acceleration structure", 142}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06889", 47}, { {"For OpRayQueryInitializeKHR instructions, the Rayflags operand must not contain both SkipTrianglesKHR and SkipAABBsKHR", 118}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06890", 47}, { {"For OpRayQueryInitializeKHR instructions, the Rayflags operand must not contain more than one of SkipTrianglesKHR, CullBackFacingTrianglesKHR, and CullFrontFacingTrianglesKHR", 174}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpRayQueryInitializeKHR-06891", 47}, { {"For OpRayQueryInitializeKHR instructions, the Rayflags operand must not contain more than one of OpaqueKHR, NoOpaqueKHR, CullOpaqueKHR, and CullNoOpaqueKHR", 155}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpReportIntersectionKHR-06998", 47}, { {"The value of the \"Hit Kind\" operand of OpReportIntersectionKHR must be in the range [0,127]", 93}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06355", 37}, { {"For OpTraceRayKHR instructions, all components of the RayOrigin and RayDirection operands must be finite floating-point values", 126}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06356", 37}, { {"For OpTraceRayKHR instructions, the RayTmin and RayTmax operands must be non-negative floating-point values", 107}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06357", 37}, { {"For OpTraceRayKHR instructions, the RayTmin operand must be less than or equal to the RayTmax operand", 101}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06358", 37}, { {"For OpTraceRayKHR instructions, RayOrigin, RayDirection, RayTmin, and RayTmax operands must not contain NaNs", 108}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06359", 37}, { {"For OpTraceRayKHR instructions, Acceleration Structure must be an acceleration structure built as a top-level acceleration structure", 132}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06360", 37}, { {"For OpTraceRayKHR instructions, if Acceleration Structure was built with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags, the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 224}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06552", 37}, { {"For OpTraceRayKHR instructions, the Rayflags operand must not contain both SkipTrianglesKHR and SkipAABBsKHR", 108}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06553", 37}, { {"For OpTraceRayKHR instructions, if the Rayflags operand contains SkipTrianglesKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR set", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06554", 37}, { {"For OpTraceRayKHR instructions, if the Rayflags operand contains SkipAABBsKHR, the pipeline must not have been created with VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR set", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06892", 37}, { {"For OpTraceRayKHR instructions, the Rayflags operand must not contain more than one of SkipTrianglesKHR, CullBackFacingTrianglesKHR, and CullFrontFacingTrianglesKHR", 164}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-06893", 37}, { {"For OpTraceRayKHR instructions, the Rayflags operand must not contain more than one of OpaqueKHR, NoOpaqueKHR, CullOpaqueKHR, and CullNoOpaqueKHR", 145}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayKHR-11855", 37}, { {"For modules which contain OpTraceRayKHR or OpTraceRayMotionNV instructions that declare a variable in the RayPayloadKHR Storage Class, all shaders which may be invoked as part of that shader call must declare an identical variable in the IncomingRayPayloadKHR Storage Class", 273}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06361", 42}, { {"For OpTraceRayMotionNV instructions, all components of the RayOrigin and RayDirection operands must be finite floating-point values", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06362", 42}, { {"For OpTraceRayMotionNV instructions, the RayTmin and RayTmax operands must be non-negative floating-point values", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06363", 42}, { {"For OpTraceRayMotionNV instructions, the RayTmin operand must be less than or equal to the RayTmax operand", 106}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06364", 42}, { {"For OpTraceRayMotionNV instructions, RayOrigin, RayDirection, RayTmin, and RayTmax operands must not contain NaNs", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06365", 42}, { {"For OpTraceRayMotionNV instructions, Acceleration Structure must be an acceleration structure built as a top-level acceleration structure with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags", 197}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06366", 42}, { {"For OpTraceRayMotionNV instructions the time operand must be between 0.0 and 1.0", 80}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTraceRayMotionNV-06367", 42}, { {"For OpTraceRayMotionNV instructions the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixKHR-10163", 50}, { {"For OpTypeCooperativeMatrixKHR, if the cooperativeMatrixFlexibleDimensions feature is not enabled, the component type, scope, number of rows, and number of columns must match one of the matrices in any of the supported VkCooperativeMatrixPropertiesKHR", 251}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixKHR-10770", 50}, { {"Any pipeline containing a shader with OpTypeCooperativeMatrixKHR or OpCooperativeMatrix*KHR instructions must be created with the VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT flag or the shader module must be version 1.6 or greater", 245}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixKHR-10771", 50}, { {"Any shader object containing OpTypeCooperativeMatrixKHR or OpCooperativeMatrix*KHR instructions must be created with the VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT flag or the shader module must be version 1.6 or greater", 225}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixMulAddNV-10059", 55}, { {"For OpTypeCooperativeMatrixMulAddNV, the operands must match a supported VkCooperativeMatrixPropertiesNV", 104}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixNV-06316", 49}, { {"For OpTypeCooperativeMatrixNV, the component type, scope, number of rows, and number of columns must match one of the matrices in any of the supported VkCooperativeMatrixPropertiesNV", 182}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeMatrixNV-06322", 49}, { {"OpTypeCooperativeMatrixNV and OpCooperativeMatrix* instructions must not be used in shader stages not included in VkPhysicalDeviceCooperativeMatrixPropertiesNV::cooperativeMatrixSupportedStages", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeCooperativeVector-10095", 47}, { {"OpTypeCooperativeVector instructions must have Component Type that is any supported type reported by vkGetPhysicalDeviceCooperativeVectorPropertiesNV", 149}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-09644", 35}, { {"Any variable declared as an OpTypeArray where the Element Type is an OpTypeImage with a \"Dim\" operand of SubpassData must be decorated with InputAttachmentIndex", 162}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-10707", 35}, { {"An OpTypeImage with Storage Class TileAttachmentQCOM must not be backed by a view equivalent to the color attachment of the current subpass instance unless the tileShadingColorAttachments feature is enabled", 206}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-10708", 35}, { {"An OpTypeImage with Storage Class TileAttachmentQCOM must not be backed by a view equivalent to the depth aspect of the depth/stencil attachment of the current subpass instance unless the tileShadingDepthAttachments", 215}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-10709", 35}, { {"An OpTypeImage with Storage Class TileAttachmentQCOM must not be backed by a view equivalent to the stencil aspect of the depth/stencil attachment of the current subpass instance unless the tileShadingStencilAttachments feature is enabled", 238}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-10710", 35}, { {"An OpTypeImage with Storage Class TileAttachmentQCOM must not be backed by a view equivalent to the input attachment of the current subpass instance unless the tileShadingInputAttachments feature is enabled", 206}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-12204", 35}, { {"If a variable with type OpTypeImage is declared in the Shader Resource Interface, it must not be backed by an image view that requires sampler Y'CBCR conversion", 160}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeImage-12207", 35}, { {"If a OpTypeImage variable backed by an image view that requires sampler Y'CBCR conversion is statically used in a shader, it must only be used with OpImageQueryLevels or OpImageQuerySizeLod instructions", 202}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeSampledImage-10711", 42}, { {"An OpTypeSampledImage with Storage Class TileAttachmentQCOM must not be backed by a view equivalent to an attachment of the current subpass instance unless the tileShadingSampledAttachments feature is enabled", 208}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeSampledImage-12206", 42}, { {"If an OpTypeSampledImage variable backed by an image view and sampler that require sampler Y'CBCR conversion is statically used in a shader, it must only be used with OpImageSample*, OpImageSparseSample*, or OpImage instructions", 228}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeSampledImage-12208", 42}, { {"If an OpTypeSampledImage variable backed by an image view and sampler that require sampler Y'CBCR conversion is statically used in a shader, the image view and sampler must have been created with an identically defined VkSamplerYcbcrConversion set via VkSamplerYcbcrConversionInfo", 280}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeSampler-12203", 37}, { {"If a variable with type OpTypeSampler is declared in the Shader Resource Interface, it must not be backed by a sampler that requires sampler Y'CBCR conversion", 158}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeTensorARM-09902", 39}, { {"OpTypeTensorARM with a Shape must not be used in shader stages", 62}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeTensorARM-09907", 39}, { {"OpTypeTensorARM without a Rank must not be used in shader stages", 64}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpTypeUntypedPointerKHR-11166", 47}, { {"Any memory access made using an OpTypeUntypedPointerKHR must have an alignment that satisfies Offset and Stride Assignment", 122}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-OpVariable-08746", 34}, { {"Any variable, Block-decorated OpTypeStruct, or Block-decorated OpTypeStruct members shared between the OpEntryPoint of two shader stages must have matching decorations as defined in interface matching", 200}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-PhysicalStorageBuffer64-06315", 47}, { {"If the PhysicalStorageBuffer64 addressing model is enabled the pointer value of a memory access instruction must be at least as aligned as specified by the Aligned memory access operand", 185}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-PhysicalStorageBuffer64-11819", 47}, { {"If the PhysicalStorageBuffer64 addressing model is enabled the pointer value of a memory access instruction in the PhysicalStorageBuffer Storage Class must reference a buffer created with the VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT usage flag set", 248}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-RayTracingOpacityMicromapKHR-11542", 52}, { {"If the RayTracingOpacityMicromapKHR capability is not enabled, any Rayflags operand provided to an instruction must not contain ForceOpacityMicromap2StateKHR", 157}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11340", 30}, { {"If the Result Type operand of OpLoad is OpTypeSampler, and Pointer is derived from a variable decorated with SamplerHeapEXT, it must correspond to loading a descriptor matching one of the descriptor types listed in Heap Resource Type Correspondence for OpTypeSampler", 266}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11341", 30}, { {"If the Result Type operand of OpLoad is OpTypeImage, and Pointer is derived from a variable decorated with ResourceHeapEXT, it must correspond to loading a descriptor matching one of the descriptor types listed in Heap Resource Type Correspondence for OpTypeImage", 263}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11342", 30}, { {"The Buffer operand of OpBufferPointerEXT must correspond to accessing a descriptor matching one of the descriptor types listed in Heap Resource Type Correspondence for OpTypePointer", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11343", 30}, { {"If the Result Type operand of OpLoad is OpTypeImage, and Pointer is derived from a variable decorated with ResourceHeapEXT, the Image Format and Sampled Type of that OpTypeImage must correspond to a descriptor with a VkFormat that matches as described by Compatibility Between SPIR-V Image Formats and Vulkan Formats and Image Format and Type Matching Between SPIR-V Image Formats and Vulkan Formats", 399}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11345", 30}, { {"If the Result Type operand of OpLoad is OpTypeImage, and Pointer is derived from a variable decorated with ResourceHeapEXT, the operands of that OpTypeImage must correspond to a descriptor with operands that match as described by Heap Image Operand Correspondence and Heap Image View Type Dimensionality", 303}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Result-11350", 30}, { {"If the Result Type operand of OpLoad is OpTypeAccelerationStructureKHR and Pointer is derived from a variable decorated with ResourceHeapEXT, it must be at an offset from the base that is a multiple of bufferDescriptorAlignment", 227}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-SampleRateShading-06378", 41}, { {"If the subpass description contains VK_SUBPASS_DESCRIPTION_FRAGMENT_REGION_BIT_EXT, then the SPIR-V fragment shader Capability SampleRateShading must not be enabled", 164}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-SampledImageArrayNonUniformIndexing-10135", 59}, { {"If the SampledImageArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a sampled image or sampler, the sampled image or sampler through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 276}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-ShaderEnqueueAMDX-09191", 41}, { {"The ShaderEnqueueAMDX capability must only be used in shaders with the GLCompute or MeshEXT execution model", 107}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Size-11165", 28}, { {"The Size operand of OpCopyMemorySized must be a multiple of 4 If 16-bit storage is enabled for the storage classes of both the Target and Source operands the Size operand may instead be a multiple of 2 If 8-bit storage is enabled for the storage classes of both the Target and Source operands the Size operand may instead be any value", 334}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-StorageBufferArrayNonUniformIndexing-10136", 60}, { {"If the StorageBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a storage buffer, the storage buffer through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 257}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-StorageImageArrayNonUniformIndexing-10137", 59}, { {"If the StorageImageArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a storage image, the storage image through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 254}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-StorageTexelBufferArrayNonUniformIndexing-10140", 65}, { {"If the StorageTexelBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a storage texel buffer, the storage texel buffer through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 274}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Stream-06312", 30}, { {"The stream number value to Stream must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 127}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-SubgroupUniformControlFlowKHR-06379", 53}, { {"The Execution Mode SubgroupUniformControlFlowKHR must not be applied to an entry point unless the shaderSubgroupUniformControlFlow feature is enabled, the corresponding shader stage bit is set in subgroupSupportedStages, and the entry point does not execute any invocation repack instructions", 292}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07117", 31}, { {"In task shaders using the TaskEXT Execution Model OpEmitMeshTasksEXT must be called exactly once under dynamically uniform conditions", 133}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07291", 31}, { {"In task shaders using the TaskEXT Execution Model the x size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupSize[0]", 183}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07292", 31}, { {"In task shaders using the TaskEXT Execution Model the y size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupSize[1]", 183}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07293", 31}, { {"In task shaders using the TaskEXT Execution Model the z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupSize[2]", 183}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07294", 31}, { {"In task shaders using the TaskEXT Execution Model the product of x size, y size, and z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupInvocations", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07299", 31}, { {"In task shaders using the TaskEXT Execution Model the value of the \"Group Count X\" operand of OpEmitMeshTasksEXT must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[0]", 210}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07300", 31}, { {"In task shaders using the TaskEXT Execution Model the value of the \"Group Count Y\" operand of OpEmitMeshTasksEXT must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[1]", 210}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07301", 31}, { {"In task shaders using the TaskEXT Execution Model the value of the \"Group Count Z\" operand of OpEmitMeshTasksEXT must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[2]", 210}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TaskEXT-07302", 31}, { {"In task shaders using the TaskEXT Execution Model the product of the \"Group Count\" operands of OpEmitMeshTasksEXT must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupTotalCount", 213}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TileShadingQCOM-10698", 39}, { {"TileShadingQCOM capability must not be declared in the compute stage unless the tileShading feature is enabled", 110}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TileShadingQCOM-10699", 39}, { {"The TileShadingQCOM capability must not be declared in the fragment stage unless the tileShadingFragmentStage feature is enabled", 128}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TileShadingQCOM-10700", 39}, { {"A shader that enables SPIR-V capability TileShadingQCOM must not be invoked outside a tile shading render pass", 110}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-TileShadingQCOM-10701", 39}, { {"A compute shader that enables SPIR-V capability TileShadingQCOM must only be invoked inside those portions of a command buffer where per-tile execution model is enabled", 168}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-UniformBufferArrayNonUniformIndexing-10134", 60}, { {"If the UniformBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a uniform buffer, the uniform buffer through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 257}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-UniformTexelBufferArrayNonUniformIndexing-10139", 65}, { {"If the UniformTexelBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a uniform texel buffer, the uniform texel buffer through which that memory is accessed must be dynamically uniform within the invocation group or subgroup", 274}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-Workgroup-06530", 33}, { {"The sum of size in bytes for variables and padding in the Workgroup Storage Class in the GLCompute Execution Model must be less than or equal to maxComputeSharedMemorySize", 171}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-XfbBuffer-06309", 33}, { {"For any given XfbBuffer value, define the buffer data size to be smallest number of bytes such that, for all outputs decorated with the same XfbBuffer value, the size of the output interface variable plus the Offset is less than or equal to the buffer data size. For a given Stream, the sum of all the buffer data sizes for all buffers writing to that stream the must not exceed VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreamDataSize", 461}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-XfbStride-06313", 33}, { {"The XFB Stride value to XfbStride must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBufferDataStride", 148}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-apiVersion-07954", 34}, { {"If VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.3, the VK_KHR_format_feature_flags2 extension is not supported, and the shaderStorageImageWriteWithoutFormat feature is not enabled, any variable created with a \"Type\" of OpTypeImage that has a \"Sampled\" operand of 2 and an \"Image Format\" operand of Unknown must be decorated with NonWritable", 362}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-apiVersion-07955", 34}, { {"If VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.3, the VK_KHR_format_feature_flags2 extension is not supported, and the shaderStorageImageReadWithoutFormat feature is not enabled, any variable created with a \"Type\" of OpTypeImage that has a \"Sampled\" operand of 2 and an \"Image Format\" operand of Unknown must be decorated with NonReadable", 361}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-bufferDescriptorAlignment-11384", 49}, { {"The Buffer operand of OpBufferPointerEXT must be at an offset from the base of the ResourceHeapEXT that is a multiple of bufferDescriptorAlignment", 146}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-bufferDescriptorAlignment-11478", 49}, { {"If a OpTypeBufferEXT member of a struct is decorated with Offset or OffsetIdEXT, the Byte Offset value must be a multiple of bufferDescriptorAlignment", 150}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-bufferDescriptorAlignment-11479", 49}, { {"If a OpTypeAccelerationStructureKHR member of a struct is decorated with Offset or OffsetIdEXT, the Byte Offset value must be a multiple of bufferDescriptorAlignment", 165}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeMatrixFlexibleDimensions-10165", 59}, { {"For OpTypeCooperativeMatrixKHR, if the cooperativeMatrixFlexibleDimensions feature is enabled, the component type, scope, number of rows, and number of columns must match either one of the matrices in one of the supported VkCooperativeMatrixPropertiesKHR or VkCooperativeMatrixFlexibleDimensionsPropertiesNV", 307}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeMatrixFlexibleDimensions-10166", 59}, { {"For OpCooperativeMatrixMulAddKHR, if the cooperativeMatrixFlexibleDimensions feature is enabled, the operands must match either one of the supported VkCooperativeMatrixPropertiesKHR or VkCooperativeMatrixFlexibleDimensionsPropertiesNV", 234}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeMatrixFlexibleDimensionsMaxDimension-10167", 71}, { {"All OpTypeCooperativeMatrixKHR dimensions must be less than or equal to cooperativeMatrixFlexibleDimensionsMaxDimension", 119}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeMatrixSupportedStages-08985", 56}, { {"OpTypeCooperativeMatrixKHR and OpCooperativeMatrix* instructions must not be used in shader stages not included in VkPhysicalDeviceCooperativeMatrixPropertiesKHR::cooperativeMatrixSupportedStages", 195}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeMatrixWorkgroupScope-10164", 55}, { {"If the cooperativeMatrixWorkgroupScope feature is not enabled, the scope of all OpTypeCooperativeMatrixKHR must not be VkScopeKHR::VK_SCOPE_WORKGROUP_KHR", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-cooperativeVectorSupportedStages-10091", 56}, { {"OpTypeCooperativeVectorNV and OpCooperativeVector* instructions must not be used in shader stages not included in VkPhysicalDeviceCooperativeVectorPropertiesNV::cooperativeVectorSupportedStages", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-denormBehaviorIndependence-06289", 50}, { {"If denormBehaviorIndependence is VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_32_BIT_ONLY, then the entry point must use the same denormals Execution Mode for both 16-bit and 64-bit floating-point types", 195}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-denormBehaviorIndependence-06290", 50}, { {"If denormBehaviorIndependence is VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_NONE, then the entry point must use the same denormals Execution Mode for all floating-point types", 169}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-flags-08761", 29}, { {"For OpRayQueryGetIntersectionTriangleVertexPositionsKHR instructions, Acceleration Structure must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DATA_ACCESS_BIT_KHR in flags", 185}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-flags-11875", 29}, { {"For OpHitObjectTraceRayMotionEXT and OpHitObjectTraceMotionReorderExecuteEXT instructions, if Acceleration Structure was built with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags, the pipeline must have been created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV set", 283}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-flags-11876", 29}, { {"For OpHitObjectTraceRayMotionEXT and OpHitObjectTraceMotionReorderExecuteEXT instructions, Acceleration Structure must be an acceleration structure built as a top-level acceleration structure with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV in flags", 251}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-imageDescriptorAlignment-11349", 48}, { {"If the Result Type operand of OpLoad is OpTypeImage, and Pointer is derived from a variable decorated with ResourceHeapEXT, it must be at an offset from the base that is a multiple of imageDescriptorAlignment", 208}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-imageDescriptorAlignment-11383", 48}, { {"If the Image operand of OpImageTexelPointer or OpUntypedImageTexelPointerEXT is derived from a variable decorated with ResourceHeapEXT, it must be at an offset from the base that is a multiple of imageDescriptorAlignment", 220}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-imageDescriptorAlignment-11477", 48}, { {"If a OpTypeImage member of a struct is decorated with Offset or OffsetIdEXT, the Byte Offset value must be a multiple of imageDescriptorAlignment", 145}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-longVector-12296", 34}, { {"If the longVector feature is enabled, the Component Count of any vector type must be less than or equal to maxVectorComponents", 126}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maintenance4-06817", 36}, { {"If the maintenance4 feature is not enabled, any vector type output interface variables must not have a higher Component Count than a matching vector type input interface variable", 178}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maintenance5-09190", 36}, { {"If the maintenance5 feature is enabled and a PointSize decorated variable is written to, all execution paths must write to a PointSize decorated variable", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maintenance5-10934", 36}, { {"If the maintenance5 feature is enabled and a PointSize decorated variable is written to for any output vertex in the Geometry Execution Model, all execution paths for all output vertices must write to a PointSize decorated variable", 231}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxBlockMatchRegion-09225", 43}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM operation is used, then Block Size less than or equal to maxBlockMatchRegion", 142}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxComputeSharedMemorySize-10168", 50}, { {"If the module uses OpTypeCooperativeMatrixKHR with Scope equal to Workgroup, the sum of size in bytes for variables and padding in the Workgroup Storage Class in the GLCompute Execution Model must be less than or equal to maxComputeSharedMemorySize minus cooperativeMatrixWorkgroupScopeReservedSharedMemory", 306}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxCooperativeVectorComponents-10094", 54}, { {"OpTypeCooperativeVector instructions must have Component Count less than or equal to maxCooperativeVectorComponents", 115}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxExecutionGraphShaderOutputNodes-09197", 58}, { {"Shaders must not include more than maxExecutionGraphShaderOutputNodes instances of OpInitializeNodePayloadsAMDX", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxExecutionGraphShaderPayloadCount-09196", 59}, { {"Shaders must not statically initialize more than maxExecutionGraphShaderPayloadCount variables in the NodeOutputPayloadAMDX storage class", 137}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxExecutionGraphShaderPayloadSize-09193", 58}, { {"Variables declared in the NodePayloadAMDX storage class must not be larger than the maxExecutionGraphShaderPayloadSize limit", 124}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxExecutionGraphShaderPayloadSize-09194", 58}, { {"Variables declared in the NodeOutputPayloadAMDX storage class must not be larger than the maxExecutionGraphShaderPayloadSize limit", 130}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxExecutionGraphShaderPayloadSize-09195", 58}, { {"For a given entry point, the sum of the size of any variable in the NodePayloadAMDX storage class, and the combined size of all statically initialized variables in the NodeOutputPayloadAMDX storage class must not be greater than maxExecutionGraphShaderPayloadSize", 263}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxMeshOutputMemorySize-08756", 47}, { {"The sum of size in bytes for variables in the Output Storage Class in the MeshEXT Execution Model must be less than or equal to maxMeshOutputMemorySize according to the formula in Mesh Shader Output", 198}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxMeshPayloadAndOutputMemorySize-08757", 57}, { {"The sum of size in bytes for variables and in the TaskPayloadWorkgroupEXT or Output Storage Class in the MeshEXT Execution Model must be less than or equal to maxMeshPayloadAndOutputMemorySize according to the formula in Mesh Shader Output", 239}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxMeshPayloadAndSharedMemorySize-08755", 57}, { {"The sum of size in bytes for variables and padding in the TaskPayloadWorkgroupEXT or Workgroup Storage Class in the MeshEXT Execution Model must be less than or equal to maxMeshPayloadAndSharedMemorySize", 203}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxMeshSharedMemorySize-08754", 47}, { {"The sum of size in bytes for variables and padding in the Workgroup Storage Class in the MeshEXT Execution Model must be less than or equal to maxMeshSharedMemorySize", 166}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxPushDataSize-12455", 39}, { {"The size of the PushConstant Storage Class block must be less than or equal to maxPushDataSize", 94}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxShaderBindingTableRecordIndex-11888", 56}, { {"The SBT Index passed in to OpHitObjectSetShaderBindingTableRecordIndexEXT must be less than or equal to VkPhysicalDeviceRayTracingInvocationReorderPropertiesEXT::maxShaderBindingTableRecordIndex", 194}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxShaderWaitQueues-12426", 43}, { {"The literal operand of the MultipleWaitQueuesQCOM Loop Control must be less than or equal to VkPhysicalDeviceShaderMultipleWaitQueuesPropertiesQCOM::maxShaderWaitQueues", 168}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxTaskPayloadAndSharedMemorySize-08760", 57}, { {"The sum of size in bytes for variables and padding in the TaskPayloadWorkgroupEXT or Workgroup Storage Class in the TaskEXT Execution Model must be less than or equal to maxTaskPayloadAndSharedMemorySize", 203}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxTaskPayloadSize-08758", 42}, { {"The sum of size in bytes for variables and in the TaskPayloadWorkgroupEXT Storage Class in the TaskEXT Execution Model must be less than or equal to maxTaskPayloadSize", 167}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxTaskSharedMemorySize-08759", 47}, { {"The sum of size in bytes for variables and padding in the Workgroup Storage Class in the TaskEXT Execution Model must be less than or equal to maxTaskSharedMemorySize", 166}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxTensorShaderAccessArrayLength-09903", 56}, { {"The length of an array returned by OpTensorReadARM or passed as the Object operand to OpTensorWriteARM must be less than or equal to maxTensorShaderAccessArrayLength", 165}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-maxTensorShaderAccessSize-09904", 49}, { {"The total size of the data (number of tensor elements x size of an element) read or written by one OpTensorReadARM, or OpTensorWriteARM instruction, respectively, must be less than or equal to maxTensorShaderAccessSize", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-meshAndTaskShaderDerivatives-10153", 52}, { {"If meshAndTaskShaderDerivatives is VK_FALSE, the DerivativeGroupLinearKHR and DerivativeGroupQuadsKHR execution modes must not be used in the MeshEXT, MeshNV, TaskEXT, or TaskNV Execution Model", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-micromap-11630", 32}, { {"If the OpacityMicromapIdKHR execution mode is specified with its specialization constant operand equal to VK_TRUE, the pipeline must not have been created with VK_PIPELINE_CREATE_2_DISALLOW_OPACITY_MICROMAP_BIT_ARM", 214}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-minSampleShading-08731", 40}, { {"If sample shading is enabled and minSampleShading is 1.0, the sample operand of any OpColorAttachmentReadEXT, OpDepthAttachmentReadEXT, or OpStencilAttachmentReadEXT operation must evaluate to the value of the coverage index for any given fragment invocation", 258}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-minSampleShading-08732", 40}, { {"If sample shading is enabled and any of the OpColorAttachmentReadEXT, OpDepthAttachmentReadEXT, or OpStencilAttachmentReadEXT operations are used, then minSampleShading must be 1.0", 180}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-pNext-09226", 29}, { {"If a OpImageBlockMatchWindow*QCOM operation is used, then target sampled image must have been created using a sampler object that included VkSamplerBlockMatchWindowCreateInfoQCOM in the pNext chain", 197}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-pNext-09919", 29}, { {"For each data graph pipeline created with a VkDataGraphPipelineShaderModuleCreateInfoARM structure included in the pNext chain of VkDataGraphPipelineCreateInfoARM, the OpGraph that is used by the OpGraphEntryPointARM the pipeline is being created for must have an OpTypeGraphARM that only uses OpTypeTensorARM with Shape present", 328}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-pNext-09920", 29}, { {"For each data graph pipeline created with a VkDataGraphPipelineShaderModuleCreateInfoARM structure included in the pNext chain of VkDataGraphPipelineCreateInfoARM, all the OpGraphConstantARM with OpTypeTensorARM type used by the OpGraph that is used by the OpGraphEntryPointARM the pipeline is being created for must have an OpTypeTensorARM with Shape present", 359}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-pNext-09921", 29}, { {"The pConstants in the data graph pipeline must satisfy all constraints", 70}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-pNext-09923", 29}, { {"The pResourceInfos in the data graph pipeline must satisfy all constraints", 74}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-protectedNoFault-09645", 40}, { {"If protectedNoFault is not supported, the Storage Class of the PhysicalStorageBuffer must not be used if the buffer being accessed is protected", 143}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-roundingModeIndependence-06291", 48}, { {"If roundingModeIndependence is VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_32_BIT_ONLY, then the entry point must use the same rounding Execution Mode for both 16-bit and 64-bit floating-point types", 192}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-roundingModeIndependence-06292", 48}, { {"If roundingModeIndependence is VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_NONE, then the entry point must use the same rounding Execution Mode for all floating-point types", 166}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-samplerDescriptorAlignment-11348", 50}, { {"If the Result Type operand of OpLoad is OpTypeSampler, and Pointer is derived from a variable decorated with SamplerHeapEXT, it must be at an offset from the base that is a multiple of samplerDescriptorAlignment", 211}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-samplerDescriptorAlignment-11476", 50}, { {"If a OpTypeSampler member of a struct is decorated with Offset or OffsetIdEXT, the Byte Offset value must be a multiple of samplerDescriptorAlignment", 149}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-samples-08725", 31}, { {"If an OpTypeImage has an MS operand 0, its bound image must have been created with VkImageCreateInfo::samples as VK_SAMPLE_COUNT_1_BIT", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-samples-08726", 31}, { {"If an OpTypeImage has an MS operand 1, its bound image must not have been created with VkImageCreateInfo::samples as VK_SAMPLE_COUNT_1_BIT", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormFlushToZeroFloat16-06299", 54}, { {"If shaderDenormFlushToZeroFloat16 is VK_FALSE, then DenormFlushToZero for 16-bit floating-point type must not be used", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormFlushToZeroFloat32-06300", 54}, { {"If shaderDenormFlushToZeroFloat32 is VK_FALSE, then DenormFlushToZero for 32-bit floating-point type must not be used", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormFlushToZeroFloat64-06301", 54}, { {"If shaderDenormFlushToZeroFloat64 is VK_FALSE, then DenormFlushToZero for 64-bit floating-point type must not be used", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormPreserveFloat16-06296", 51}, { {"If shaderDenormPreserveFloat16 is VK_FALSE, then DenormPreserve for 16-bit floating-point type must not be used", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormPreserveFloat32-06297", 51}, { {"If shaderDenormPreserveFloat32 is VK_FALSE, then DenormPreserve for 32-bit floating-point type must not be used", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDenormPreserveFloat64-06298", 51}, { {"If shaderDenormPreserveFloat64 is VK_FALSE, then DenormPreserve for 64-bit floating-point type must not be used", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderDeviceClock-06268", 41}, { {"If the shaderDeviceClock feature is not enabled, the Device scope must not be used for OpReadClockKHR", 101}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06767", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the EarlyAndLateFragmentTestsAMD Execution Mode must not be used", 127}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06768", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefUnchangedFrontAMD Execution Mode must not be used", 126}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06769", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefUnchangedBackAMD Execution Mode must not be used", 125}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06770", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefGreaterFrontAMD Execution Mode must not be used", 124}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06771", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefGreaterBackAMD Execution Mode must not be used", 123}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06772", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefLessFrontAMD Execution Mode must not be used", 121}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderEarlyAndLateFragmentTests-06773", 55}, { {"If the shaderEarlyAndLateFragmentTests feature is not enabled, the StencilRefLessBackAMD Execution Mode must not be used", 120}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderFloat16VectorAtomics-09581", 50}, { {"shaderFloat16VectorAtomics, must be enabled for 16-bit floating-point, 2- and 4-component vector atomic operations to be supported", 130}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderFmaFloat16-10977", 40}, { {"shaderFmaFloat16 must be enabled for OpFmaKHR to be supported with a result type that includes a 16-bit float", 109}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderFmaFloat32-10978", 40}, { {"shaderFmaFloat32 must be enabled for OpFmaKHR to be supported with a result type that includes a 32-bit float", 109}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderFmaFloat64-10979", 40}, { {"shaderFmaFloat64 must be enabled for OpFmaKHR to be supported with a result type that includes a 64-bit float", 109}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRelaxedExtendedInstruction-10773", 56}, { {"If the shaderRelaxedExtendedInstruction feature is not enabled, the OpExtInstWithForwardRefsKHR must not be used", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTEFloat16-06302", 52}, { {"If shaderRoundingModeRTEFloat16 is VK_FALSE, then RoundingModeRTE for 16-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTEFloat32-06303", 52}, { {"If shaderRoundingModeRTEFloat32 is VK_FALSE, then RoundingModeRTE for 32-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTEFloat64-06304", 52}, { {"If shaderRoundingModeRTEFloat64 is VK_FALSE, then RoundingModeRTE for 64-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTZFloat16-06305", 52}, { {"If shaderRoundingModeRTZFloat16 is VK_FALSE, then RoundingModeRTZ for 16-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTZFloat32-06306", 52}, { {"If shaderRoundingModeRTZFloat32 is VK_FALSE, then RoundingModeRTZ for 32-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderRoundingModeRTZFloat64-06307", 52}, { {"If shaderRoundingModeRTZFloat64 is VK_FALSE, then RoundingModeRTZ for 64-bit floating-point type must not be used", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSampleRateInterpolationFunctions-06325", 62}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::shaderSampleRateInterpolationFunctions is VK_FALSE, then GLSL.std.450 fragment interpolation functions are not supported by the implementation and OpCapability must not be InterpolationFunction", 298}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat16-06293", 61}, { {"If shaderSignedZeroInfNanPreserveFloat16 is VK_FALSE, then SignedZeroInfNanPreserve for 16-bit floating-point type must not be used", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat16-09559", 61}, { {"If shaderSignedZeroInfNanPreserveFloat16 is VK_FALSE then any FPFastMathDefault execution mode with a type of 16-bit float must include the NSZ, NotInf, and NotNaN flags", 169}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat16-09560", 61}, { {"If shaderSignedZeroInfNanPreserveFloat16 is VK_FALSE then any FPFastMathMode decoration on an instruction with result type or any operand type that includes a 16-bit float must include the NSZ, NotInf, and NotNaN flags", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat32-06294", 61}, { {"If shaderSignedZeroInfNanPreserveFloat32 is VK_FALSE, then SignedZeroInfNanPreserve for 32-bit floating-point type must not be used", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat32-09561", 61}, { {"If shaderSignedZeroInfNanPreserveFloat32 is VK_FALSE then any FPFastMathDefault execution mode with a type of 32-bit float must include the NSZ, NotInf, and NotNaN flags", 169}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat32-09562", 61}, { {"If shaderSignedZeroInfNanPreserveFloat32 is VK_FALSE then any FPFastMathMode decoration on an instruction with result type or any operand type that includes a 32-bit float must include the NSZ, NotInf, and NotNaN flags", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat64-06295", 61}, { {"If shaderSignedZeroInfNanPreserveFloat64 is VK_FALSE, then SignedZeroInfNanPreserve for 64-bit floating-point type must not be used", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat64-09563", 61}, { {"If shaderSignedZeroInfNanPreserveFloat64 is VK_FALSE then any FPFastMathDefault execution mode with a type of 64-bit float must include the NSZ, NotInf, and NotNaN flags", 169}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSignedZeroInfNanPreserveFloat64-09564", 61}, { {"If shaderSignedZeroInfNanPreserveFloat64 is VK_FALSE then any FPFastMathMode decoration on an instruction with result type or any operand type that includes a 64-bit float must include the NSZ, NotInf, and NotNaN flags", 218}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSubgroupClock-06267", 43}, { {"If the shaderSubgroupClock feature is not enabled, the Subgroup scope must not be used for OpReadClockKHR", 105}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderSubgroupRotateClustered-09566", 53}, { {"If shaderSubgroupRotateClustered is VK_FALSE, then the ClusterSize operand to OpGroupNonUniformRotateKHR must not be used", 121}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderTensorSupportedStages-09901", 51}, { {"OpTypeTensorARM, OpTensorReadARM, OpTensorWriteARM, or OpTensorQuerySizeARM must not be used in shader stages not in shaderTensorSupportedStages", 144}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderUniformBufferUnsizedArray-11806", 55}, { {"If the shaderUniformBufferUnsizedArray feature is not enabled, OpTypeRuntimeArray must not be used for the last member of a Block-decorated OpTypeStruct in the Uniform storage Storage Class", 189}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-shaderZeroInitializeWorkgroupMemory-06372", 59}, { {"If the shaderZeroInitializeWorkgroupMemory feature is not enabled, any variable with Workgroup as its Storage Class must not have an Initializer operand", 152}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-source-11387", 30}, { {"All possible values of a variable pointer to a resource that are not OpConstantNull must either all be pointers to resources with DescriptorSet and Binding decorations with the same mapping types as specified by VkDescriptorSetAndBindingMappingEXT::source, or all be pointers to resources without DescriptorSet and Binding decorations", 334}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-storageBuffer16BitAccess-11161", 48}, { {"If storageBuffer16BitAccess is VK_FALSE, then objects containing 16-bit integer or 16-bit floating-point elements must not have Storage Class of StorageBuffer, ShaderRecordBufferKHR, or PhysicalStorageBuffer unless: storageBuffer8BitAccess is VK_TRUE, or the elements are accessed in 32-bit multiples if shaderUntypedPointers is VK_TRUE.", 337}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-storageBuffer8BitAccess-06328", 47}, { {"If storageBuffer8BitAccess is VK_FALSE, then objects containing an 8-bit integer element must not have Storage Class of StorageBuffer, ShaderRecordBufferKHR, or PhysicalStorageBuffer unless shaderUntypedPointers is VK_TRUE and they are accessed in: 32-bit multiples, or 16-bit multiples if storageBuffer16BitAccess is VK_TRUE.", 326}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-storageInputOutput16-11162", 44}, { {"If storageInputOutput16 is VK_FALSE, then objects containing 16-bit integer or 16-bit floating-point elements must not have storage class of Input or Output", 156}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-storagePushConstant16-06333", 45}, { {"If storagePushConstant16 is VK_FALSE, then objects containing 16-bit integer or 16-bit floating-point elements must not have Storage Class of PushConstant unless: StoragePushConstant8 is VK_TRUE, or elements are accessed in 32-bit multiples if shaderUntypedPointers is VK_TRUE", 276}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-storagePushConstant8-06330", 44}, { {"If storagePushConstant8 is VK_FALSE, then objects containing an 8-bit integer element must not have Storage Class of PushConstant unless shaderUntypedPointers is VK_TRUE and they are accessed in: 32-bit multiples, or 16-bit multiples if storagePushConstant16 is VK_TRUE.", 270}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupBroadcastDynamicId-06276", 50}, { {"If subgroupBroadcastDynamicId is VK_TRUE, and the shader module version is 1.5 or higher, the \"Index\" for OpGroupNonUniformQuadBroadcast must be dynamically uniform within the derivative group. Otherwise, \"Index\" must be a constant", 235}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupBroadcastDynamicId-06277", 50}, { {"If subgroupBroadcastDynamicId is VK_TRUE, and the shader module version is 1.5 or higher, the \"Id\" for OpGroupNonUniformBroadcast must be dynamically uniform within the subgroup. Otherwise, \"Id\" must be a constant", 217}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10141", 36}, { {"If the effective subgroup size is 1, the UniformBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a uniform buffer, the uniform buffer through which that memory is accessed must be dynamically uniform within the invocation group", 279}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10142", 36}, { {"If the effective subgroup size is 1, the SampledImageArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a sampled image or sampler, the sampled image or sampler through which that memory is accessed must be dynamically uniform within the invocation group", 298}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10143", 36}, { {"If the effective subgroup size is 1, the StorageBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a storage buffer, the storage buffer through which that memory is accessed must be dynamically uniform within the invocation group", 279}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10144", 36}, { {"If the effective subgroup size is 1, the StorageImageArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a storage image, the storage image through which that memory is accessed must be dynamically uniform within the invocation group", 276}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10145", 36}, { {"If the effective subgroup size is 1, the InputAttachmentArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through an input attachment, the input attachment through which that memory is accessed must be dynamically uniform within the invocation group", 286}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10146", 36}, { {"If the effective subgroup size is 1, the UniformTexelBufferArrayNonUniformIndexing capability is not declared, and an instruction accesses memory through a uniform texel buffer, the uniform texel buffer through which that memory is accessed must be dynamically uniform within the invocation group", 296}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10147", 36}, { {"If the effective subgroup size is 1, the StorageTexelBufferArrayNonUniformIndexing capability is not is not declared, and an instruction accesses memory through a storage texel buffer, the storage texel buffer through which that memory is accessed must be dynamically uniform within the invocation group", 303}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-subgroupSize-10149", 36}, { {"If an instruction accesses memory through any resource, the effective subgroup size is greater than 1, the DescriptorHeapEXT capability is not declared, and the resource through which that memory is accessed is not uniform within the invocation group, and not uniform within the subgroup, then the operand corresponding to that resource (e.g. the pointer or sampled image operand) must be decorated with NonUniform", 414}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tensorDescriptorAlignment-11480", 49}, { {"If a OpTypeTensorARM member of a struct is decorated with Offset or OffsetIdEXT, the Byte Offset value must be a multiple of tensorDescriptorAlignment", 150}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tensorDescriptorAlignment-11481", 49}, { {"If the Result Type operand of OpLoad is OpTypeTensorARM and Pointer is derived from a variable decorated with ResourceHeapEXT, it must be at an offset from the base that is a multiple of tensorDescriptorAlignment", 212}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tessellationShader-06326", 42}, { {"If the tessellationShader feature is enabled, and the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::tessellationIsolines is VK_FALSE, then OpExecutionMode must not be IsoLines", 227}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tessellationShader-06327", 42}, { {"If the tessellationShader feature is enabled, and the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::tessellationPointMode is VK_FALSE, then OpExecutionMode must not be PointMode", 229}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tileShadingImageProcessing-10712", 50}, { {"If an OpTypeSampledImage with Storage Class TileAttachmentQCOM is consumed by any argument of the following operations, tileShadingImageProcessing must be enabled: OpImageSampleWeightedQCOM OpImageBoxFilterQCOM OpImageBlockMatch*QCOM", 233}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-tileSize-10705", 32}, { {"In compute shaders that enables SPIR-V capability TileShadingQCOM, VkTilePropertiesQCOM.tileSize.z % TileShadingRateQCOM::z must equal 0", 136}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-transformFeedbackStreamsLinesTriangles-06311", 62}, { {"If the geometry shader emits to more than one vertex stream and VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackStreamsLinesTriangles is VK_FALSE, then Execution Mode must be OutputPoints", 204}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-uniformAndStorageBuffer16BitAccess-06332", 58}, { {"If uniformAndStorageBuffer16BitAccess is VK_FALSE, then objects in the Uniform Storage Class with the Block decoration must not have 16-bit integer or 16-bit floating-point members unless: uniformAndStorageBuffer8BitAccess is VK_TRUE, or members are accessed in 32-bit multiples and shaderUntypedPointers is VK_TRUE", 315}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-uniformAndStorageBuffer8BitAccess-06329", 57}, { {"If uniformAndStorageBuffer8BitAccess is VK_FALSE, then objects in the Uniform Storage Class with the Block decoration must not have an 8-bit integer member unless shaderUntypedPointers is VK_TRUE and they are accessed in: 32-bit multiples, or 16-bit multiples if uniformAndStorageBuffer16BitAccess is VK_TRUE.", 309}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-vulkanMemoryModel-06265", 41}, { {"If the vulkanMemoryModel feature is enabled and the vulkanMemoryModelDeviceScope feature is not enabled, Device memory scope must not be used", 141}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-vulkanMemoryModel-06266", 41}, { {"If the vulkanMemoryModel feature is not enabled, QueueFamily memory scope must not be used", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-workgroupMemoryExplicitLayout16BitAccess-10757", 64}, { {"If workgroupMemoryExplicitLayout16BitAccess is VK_FALSE, then objects in the Workgroup Storage Class with the Block decoration must not have an 16-bit integer or 16-bit floating-point elements unless: workgroupMemoryExplicitLayout8BitAccess is VK_TRUE, or elements are accessed in 32-bit multiples if shaderUntypedPointers is VK_TRUE", 333}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-workgroupMemoryExplicitLayout8BitAccess-10756", 63}, { {"If workgroupMemoryExplicitLayout8BitAccess is VK_FALSE, then objects in the Workgroup Storage Class with the Block decoration must not have an 8-bit integer element unless shaderUntypedPointers is VK_TRUE and they are accessed in: 32-bit multiples, or 16-bit multiples if workgroupMemoryExplicitLayout16BitAccess is VK_TRUE", 323}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-x-06429", 25}, { {"In compute shaders using the GLCompute Execution Model the x size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupSize[0]", 174}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-x-06432", 25}, { {"In compute shaders using the GLCompute Execution Model the product of x size, y size, and z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupInvocations", 209}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-x-10702", 25}, { {"In compute shaders that enables SPIR-V capability TileShadingQCOM, the x size in TileShadingRateQCOM must be less than or equal to VkPhysicalDeviceTileShadingPropertiesQCOM.maxTileShadingRate.x", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-y-06430", 25}, { {"In compute shaders using the GLCompute Execution Model the y size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupSize[1]", 174}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-y-10703", 25}, { {"In compute shaders that enables SPIR-V capability TileShadingQCOM, the y size in TileShadingRateQCOM must be less than or equal to VkPhysicalDeviceTileShadingPropertiesQCOM.maxTileShadingRate.y", 193}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-z-06431", 25}, { {"In compute shaders using the GLCompute Execution Model the z size in LocalSize or LocalSizeId must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupSize[2]", 174}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-RuntimeSpirv-z-10704", 25}, { {"In compute shaders that enables SPIR-V capability TileShadingQCOM, the z size in TileShadingRateQCOM must be less than or equal to VkTilePropertiesQCOM.tileSize.z", 162}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-SMCountNV-SMCountNV-04363", 30}, { {"The variable decorated with SMCountNV must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SMCountNV-SMCountNV-04364", 30}, { {"The variable decorated with SMCountNV must be declared as a scalar 32-bit integer value", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SMIDNV-SMIDNV-04365", 24}, { {"The variable decorated with SMIDNV must be declared using the Input Storage Class", 81}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SMIDNV-SMIDNV-04366", 24}, { {"The variable decorated with SMIDNV must be declared as a scalar 32-bit integer value", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleId-SampleId-04354", 28}, { {"The SampleId decoration must be used only within the Fragment Execution Model", 77}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleId-SampleId-04355", 28}, { {"The variable decorated with SampleId must be declared using the Input Storage Class", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleId-SampleId-04356", 28}, { {"The variable decorated with SampleId must be declared as a scalar 32-bit integer value", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleMask-SampleMask-04357", 32}, { {"The SampleMask decoration must be used only within the Fragment Execution Model", 79}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleMask-SampleMask-04358", 32}, { {"The variable decorated with SampleMask must be declared using the Input or Output Storage Class", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SampleMask-SampleMask-04359", 32}, { {"The variable decorated with SampleMask must be declared as an array of 32-bit integer values", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SamplePosition-SamplePosition-04360", 40}, { {"The SamplePosition decoration must be used only within the Fragment Execution Model", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SamplePosition-SamplePosition-04361", 40}, { {"The variable decorated with SamplePosition must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SamplePosition-SamplePosition-04362", 40}, { {"The variable decorated with SamplePosition must be declared as a two-component vector of 32-bit floating-point values", 117}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SamplerHeapEXT-SamplerHeapEXT-11239", 40}, { {"The variable decorated with SamplerHeapEXT must be declared using the UniformConstant Storage Class", 99}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-ShaderIndexAMDX-ShaderIndexAMDX-09175", 42}, { {"The variable decorated with ShaderIndexAMDX must be declared using the Input Storage Class", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ShaderIndexAMDX-ShaderIndexAMDX-09176", 42}, { {"The variable decorated with ShaderIndexAMDX must be declared as a scalar 32-bit integer value", 93}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ShadingRateKHR-ShadingRateKHR-04490", 40}, { {"The ShadingRateKHR decoration must be used only within the Fragment Execution Model", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ShadingRateKHR-ShadingRateKHR-04491", 40}, { {"The variable decorated with ShadingRateKHR must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ShadingRateKHR-ShadingRateKHR-04492", 40}, { {"The variable decorated with ShadingRateKHR must be declared as a scalar 32-bit integer value", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-StandaloneSpirv-Base-07650", 31}, { {"The Base operand of OpPtrAccessChain or OpUntypedPtrAccessChainKHR must have a storage class of Workgroup, StorageBuffer, or PhysicalStorageBuffer", 146}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Base-07651", 31}, { {"If the Base operand of OpPtrAccessChain or OpUntypedPtrAccessChainKHR has a Workgroup Storage Class, then the VariablePointers capability must be declared", 154}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Base-07652", 31}, { {"If the Base operand of OpPtrAccessChain or OpUntypedPtrAccessChainKHR has a StorageBuffer Storage Class, then the VariablePointers or VariablePointersStorageBuffer capability must be declared", 191}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Base-12469", 31}, { {"The Base operand of OpBitcastExtractEXT must be a scalar or vector integer type", 79}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-BuiltIn-04668", 34}, { {"Any BuiltIn decoration not listed in Built-In Variables must not be used", 72}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-CallableDataKHR-04704", 42}, { {"CallableDataKHR Storage Class must only be used in ray generation, closest hit, miss, and callable shaders", 106}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-04920", 36}, { {"The Component decoration value must not be greater than 3", 57}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-04921", 36}, { {"If the Component decoration is used on a variable that has a vector type with a Component Type with a Width that is less than or equal to 32, the sum of its Component Count and the Component decoration value must be less than or equal to 4", 239}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-04922", 36}, { {"If the Component decoration is used on a variable that has a vector type with a Component Type with a Width that is equal to 64, the sum of two times its Component Count and the Component decoration value must be less than or equal to 4", 236}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-04923", 36}, { {"The Component decorations value must not be 1 or 3 for scalar or two-component 64-bit data types", 96}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-07703", 36}, { {"The Component decorations must not be used for a 64-bit vector type with more than two components", 97}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Component-10583", 36}, { {"The Component decorations must not be used with any type that is not a scalar, vector, array of scalars or vectors, or an array of arrays of scalars or vectors", 159}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DerivativeGroupLinearKHR-10152", 51}, { {"For compute, mesh, or task shaders using the DerivativeGroupLinearKHR execution mode, the product of the dimensions of the local workgroup size must be a multiple of four", 170}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DerivativeGroupLinearNV-04778", 50}, { {"For compute shaders using the DerivativeGroupLinearNV execution mode, the product of the dimensions of the local workgroup size must be a multiple of four", 154}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DerivativeGroupQuadsKHR-10151", 50}, { {"For compute, mesh, or task shaders using the DerivativeGroupQuadsKHR execution mode, the first two dimensions of the local workgroup size must be a multiple of two", 163}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DerivativeGroupQuadsNV-04684", 49}, { {"For compute shaders using the DerivativeGroupQuadsNV execution mode, the first two dimensions of the local workgroup size must be a multiple of two", 147}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DescriptorHeapEXT-11482", 44}, { {"If the DescriptorHeapEXT capability is declared, structure types must not contain opaque types other than descriptors", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-DescriptorSet-06491", 40}, { {"If a variable is decorated by DescriptorSet or Binding, the Storage Class must correspond to an entry in Shader Resource and Storage Class Correspondence", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Execution-10584", 36}, { {"If the Execution Model of an entry point is not TessellationControl or Geometry, its interface definition must not include any variables in the Input storage class decorated with Component that are arrays of arrays of scalars or vectors", 236}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Execution-10585", 36}, { {"If the Execution Model of an entry point is not MeshNV, its interface definition must not include any variables in the Output storage class decorated with Component that are arrays of arrays of scalars or vectors", 212}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Execution-10687", 36}, { {"Execution Mode NonCoherentTileAttachmentReadQCOM must not be used in any stage other than fragment", 98}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Execution-10688", 36}, { {"Execution Mode TileShadingRateQCOM must not be used in any stage other than compute", 83}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-ExecutionModel-07320", 41}, { {"If the Execution Model is TessellationControl, and the MemoryModel is GLSL450, the Scope for memory must not be Workgroup", 121}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-FPRoundingMode-04675", 41}, { {"Rounding modes other than round-to-nearest-even and round-towards-zero must not be used for the FPRoundingMode decoration", 121}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Flat-04670", 31}, { {"The Flat, NoPerspective, Sample, and Centroid decorations must only be used on variables with the Output or Input Storage Class", 127}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Flat-04744", 31}, { {"Any variable with integer or double-precision floating-point type and with Input Storage Class in a fragment shader, must be decorated Flat", 139}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Flat-06201", 31}, { {"The Flat, NoPerspective, Sample, and Centroid decorations must not be used on variables with the Output storage class in a fragment shader", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Flat-06202", 31}, { {"The Flat, NoPerspective, Sample, and Centroid decorations must not be used on variables with the Input storage class in a vertex shader", 135}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Function-04681", 35}, { {"A type T that is an array sized with a specialization constant must neither be, nor be contained in, the type T2 of a variable V, unless either: a) T is equal to T2, b) V is declared in the Function, or Private Storage Class, c) V is a non-Block variable in the Workgroup Storage Class, or d) V is an interface variable with an additional level of arrayness, as described in interface matching, and T is the member type of the array type T2", 440}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Function-12294", 35}, { {"A type T that is a vector sized with a specialization constant must neither be, nor be contained in, the type T2 of a variable V, unless either: a) T is equal to T2, b) V is declared in the Function, or Private Storage Class, c) V is a non-Block variable in the Workgroup Storage Class, or d) V is an interface variable with an additional level of arrayness, as described in interface matching, and T is the member type of the array type T2", 440}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-GLSLShared-04669", 37}, { {"The GLSLShared and GLSLPacked decorations must not be used", 58}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-HitAttributeKHR-04701", 42}, { {"HitAttributeKHR Storage Class must only be used in intersection, any-hit, or closest hit shaders", 96}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-HitAttributeKHR-04702", 42}, { {"There must be at most one variable with the HitAttributeKHR Storage Class in the input interface of an entry point", 114}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-HitAttributeKHR-04703", 42}, { {"A variable with HitAttributeKHR Storage Class must only be written to in an intersection shader", 95}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Image-04965", 32}, { {"The SPIR-V Type of the Image Format operand of an OpTypeImage must match the Sampled Type, as defined in Image Format and Type Matching", 135}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-IncomingCallableDataKHR-04705", 50}, { {"IncomingCallableDataKHR Storage Class must only be used in callable shaders", 75}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-IncomingCallableDataKHR-04706", 50}, { {"There must be at most one variable with the IncomingCallableDataKHR Storage Class in the input interface of an entry point", 122}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-IncomingRayPayloadKHR-04699", 48}, { {"IncomingRayPayloadKHR Storage Class must only be used in closest hit, any-hit, or miss shaders", 94}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-IncomingRayPayloadKHR-04700", 48}, { {"There must be at most one variable with the IncomingRayPayloadKHR Storage Class in the input interface of an entry point", 120}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Input-06778", 32}, { {"Variables with a Storage Class of Input in a fragment shader stage that are decorated with PerVertexKHR must be declared as arrays", 130}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Input-07290", 32}, { {"Variables with a Storage Class of Input or Output and a type of OpTypeBool must be decorated with the BuiltIn decoration", 120}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Input-09557", 32}, { {"The pointers of any Input or Output Interface user-defined variables must not contain any PhysicalStorageBuffer Storage Class pointers", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Input-10588", 32}, { {"Variables in the Input storage class in the TessellationControl, TessellationEvaluation, Geometry, or Fragment execution model must not have overlapping Component and Location decorations as defined by Location and Component Assignment", 235}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Input-10604", 32}, { {"Any variable in the Input storage class in the Fragment execution model must not have an assigned Location shared with another variable with different decorations for PerVertexKHR, Flat, NoPerspective, Sample, or Centroid", 221}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-InputAttachmentArrayDynamicIndexing-10131", 62}, { {"If the InputAttachmentArrayDynamicIndexing capability is not declared, and an instruction accesses memory through an input attachment, the input attachment through which that memory is accessed must be determined by constant integral expressions", 245}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-InputAttachmentIndex-06678", 47}, { {"Variables decorated with InputAttachmentIndex must be in the UniformConstant Storage Class", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Invariant-04677", 36}, { {"Variables decorated with Invariant and variables with structure types that have any members decorated with Invariant must be in the Output or Input Storage Class, Invariant used on an Input Storage Class variable or structure member has no effect", 246}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-04915", 35}, { {"The Location or Component decorations must not be used with BuiltIn", 67}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-04916", 35}, { {"The Location decorations must be used on user-defined variables", 63}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-04917", 35}, { {"If a user-defined variable is not a pointer to a Block decorated OpTypeStruct, then the variable must have a Location decoration", 128}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-04918", 35}, { {"If a user-defined variable has a Location decoration, and the variable is a pointer to a OpTypeStruct, then the members of that structure must not have Location decorations", 172}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-04919", 35}, { {"If a user-defined variable does not have a Location decoration, and the variable is a pointer to a Block decorated OpTypeStruct, then each member of the struct must have a Location decoration", 191}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Location-06672", 35}, { {"The Location or Component decorations must only be used with the Input, Output, RayPayloadKHR, IncomingRayPayloadKHR, HitAttributeKHR, HitObjectAttributeNV, CallableDataKHR, IncomingCallableDataKHR, or ShaderRecordBufferKHR storage classes", 239}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MaximallyReconvergesKHR-09565", 50}, { {"The MaximallyReconvergesKHR Execution Mode must not be applied to an entry point if an invocation repack instruction is statically used", 135}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10865", 42}, { {"MemorySemantics must have at most one non-relaxed memory order bit set (Acquire, Release, or AcquireRelease)", 108}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10866", 42}, { {"MemorySemantics with SequentiallyConsistent memory order must not be used in the Vulkan API", 91}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10867", 42}, { {"MemorySemantics must not use Acquire or AcquireRelease memory order with OpAtomicStore", 86}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10868", 42}, { {"MemorySemantics must not use Release or AcquireRelease memory order with OpAtomicLoad", 85}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10869", 42}, { {"MemorySemantics must not use Relaxed memory order with OpMemoryBarrier", 70}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10870", 42}, { {"MemorySemantics with a non-relaxed memory order (Acquire, Release, or AcquireRelease) must have at least one Vulkan-supported storage class semantics bit set (UniformMemory, WorkgroupMemory, ImageMemory, or OutputMemory)", 220}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10871", 42}, { {"MemorySemantics with at least one Vulkan-supported storage class semantics bit set (UniformMemory, WorkgroupMemory, ImageMemory, or OutputMemory) must use a non-relaxed memory order (Acquire, Release, or AcquireRelease)", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10872", 42}, { {"MemorySemantics with MakeAvailable bit set must use Release or AcquireRelease memory order", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-10873", 42}, { {"MemorySemantics with MakeVisible bit set must use Acquire or AcquireRelease memory order", 88}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-13551", 42}, { {"MemorySemantics with Volatile bit set must not be used with barrier instructions (OpControlBarrier, OpControlBarrierArriveEXT, OpControlBarrierWaitEXT, or OpMemoryBarrier)", 171}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-13556", 42}, { {"The MemorySemantics for OpControlBarrierArriveEXT must not have any non-relaxed memory order set other than Release", 115}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MemorySemantics-13557", 42}, { {"The MemorySemantics for OpControlBarrierWaitEXT must not have any non-relaxed memory order set other than Acquire", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07102", 34}, { {"The module must not contain both an entry point that uses the TaskEXT or MeshEXT Execution Model and an entry point that uses the TaskNV or MeshNV Execution Model", 162}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07106", 34}, { {"In mesh shaders using the MeshEXT Execution Model OpSetMeshOutputsEXT must be called before any outputs are written", 115}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07107", 34}, { {"In mesh shaders using the MeshEXT Execution Model all variables declared in the Output Storage Class must not be read", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07108", 34}, { {"In mesh shaders using the MeshEXT Execution Model for OpSetMeshOutputsEXT instructions, the \"Vertex Count\" and \"Primitive Count\" operands must not depend on ViewIndex", 170}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07109", 34}, { {"In mesh shaders using the MeshEXT Execution Model variables decorated with PrimitivePointIndicesEXT, PrimitiveLineIndicesEXT, or PrimitiveTriangleIndicesEXT declared as an array must not be accessed by indices that depend on ViewIndex", 234}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07110", 34}, { {"In mesh shaders using the MeshEXT Execution Model any values stored in variables decorated with PrimitivePointIndicesEXT, PrimitiveLineIndicesEXT, or PrimitiveTriangleIndicesEXT must not depend on ViewIndex", 206}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07111", 34}, { {"In mesh shaders using the MeshEXT Execution Model variables in workgroup or private Storage Class declared as or containing a composite type must not be accessed by indices that depend on ViewIndex", 197}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07330", 34}, { {"In mesh shaders using the MeshEXT Execution Model the OutputVertices Execution Mode must be greater than 0", 106}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-MeshEXT-07331", 34}, { {"In mesh shaders using the MeshEXT Execution Model the OutputPrimitivesEXT Execution Mode must be greater than 0", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-NonCoherentTileAttachmentReadQCOM-10690", 60}, { {"If the NonCoherentTileAttachmentReadQCOM Execution Mode is used, the TileShadingQCOM capability must be enabled", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04633", 31}, { {"Every entry point must have no return value and accept no arguments", 67}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04634", 31}, { {"The static function-call graph for an entry point must not contain cycles; that is, static recursion is not allowed", 115}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04635", 31}, { {"The Logical or PhysicalStorageBuffer64 addressing model must be selected", 72}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04636", 31}, { {"Scope for execution must be limited to Workgroup or Subgroup", 60}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04637", 31}, { {"If the Scope for execution is Workgroup, then it must only be used in the task, mesh, tessellation control, or compute Execution Model", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04638", 31}, { {"Scope for memory must be limited to Device, QueueFamily, Workgroup, ShaderCallKHR, Subgroup, or Invocation", 106}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04640", 31}, { {"If the Scope for memory is ShaderCallKHR, then it must only be used in ray generation, intersection, closest hit, any-hit, miss, and callable Execution Model", 157}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04641", 31}, { {"If the Scope for memory is Invocation, then MemorySemantics must use Relaxed memory order", 89}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04642", 31}, { {"Scope for group operations must be limited to Subgroup", 54}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04643", 31}, { {"Storage Class must be limited to UniformConstant, Input, Uniform, Output, Workgroup, Private, Function, PushConstant, Image, StorageBuffer, RayPayloadKHR, IncomingRayPayloadKHR, HitAttributeKHR, CallableDataKHR, IncomingCallableDataKHR, ShaderRecordBufferKHR, PhysicalStorageBuffer, or TileImageEXT", 298}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04644", 31}, { {"If the Storage Class is Output, then it must not be used in the GlCompute, RayGenerationKHR, IntersectionKHR, AnyHitKHR, ClosestHitKHR, MissKHR, or CallableKHR Execution Model", 175}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04645", 31}, { {"If the Storage Class is Workgroup, then it must only be used in the task, mesh, or compute Execution Model", 106}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04667", 31}, { {"If the DescriptorHeapEXT capability is not declared, structure types must not contain opaque types", 98}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-04686", 31}, { {"The Pointer operand of all atomic instructions must have a Storage Class limited to Uniform, Workgroup, Image, StorageBuffer, PhysicalStorageBuffer, or TaskPayloadWorkgroupEXT", 175}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-07321", 31}, { {"If the Scope for memory is Workgroup, then it must only be used in the task, mesh, tessellation control, or compute Execution Model", 131}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-08720", 31}, { {"If the Storage Class is TileImageEXT, then it must only be used in the fragment execution model", 95}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-08724", 31}, { {"The TileImageEXT Storage Class must only be used for declaring tile image variables", 83}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-10684", 31}, { {"All variables must have valid explicit layout decorations as described in Shader Interfaces", 91}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-10685", 31}, { {"Either a TileShadingRateQCOM, LocalSize, or LocalSizeId Execution Mode, or an object decorated with the WorkgroupSize decoration must be specified for each entry point with a task, mesh, or compute Execution Model", 213}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-None-12295", 31}, { {"If the LongVectorEXT capability is not declared, the Component Count of any vector type must be less than or equal to 4 and greater than 1", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04687", 33}, { {"Output variables or block members decorated with Offset that have a 64-bit type, or a composite type containing a 64-bit type, must specify an Offset value aligned to a 8 byte boundary", 184}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04689", 33}, { {"The size of any output block containing any member decorated with Offset that is a 64-bit type must be a multiple of 8", 118}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04690", 33}, { {"The first member of an output block specifying a Offset decoration must specify a Offset value that is aligned to an 8 byte boundary if that block contains any member decorated with Offset and is a 64-bit type", 209}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04691", 33}, { {"Output variables or block members decorated with Offset that have a 32-bit type, or a composite type contains a 32-bit type, must specify an Offset value aligned to a 4 byte boundary", 182}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04692", 33}, { {"Output variables, blocks, or block members decorated with Offset must only contain base types that have components that are either 32-bit or 64-bit in size", 155}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04716", 33}, { {"Only variables or block members in the output interface decorated with Offset can be captured for transform feedback, and those variables or block members must also be decorated with XfbBuffer and XfbStride, or inherit XfbBuffer and XfbStride decorations from a block containing them", 283}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Offset-04865", 33}, { {"Any image instruction which uses an Offset, ConstOffset, or ConstOffsets image operand, must only consume a \"Sampled Image\" operand whose type has its \"Sampled\" operand set to 1", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpArrayLength-11805", 40}, { {"OpArrayLength and OpUntypedArrayLengthKHR must not be used with an OpTypeRuntimeArray that is the last member of a Block-decorated OpTypeStruct in the Uniform storage Storage Class", 180}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpControlBarrier-04682", 43}, { {"If OpControlBarrier is used in ray generation, intersection, any-hit, closest hit, miss, fragment, vertex, tessellation evaluation, or geometry shaders, the execution Scope must be Subgroup", 189}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpControlBarrierArriveEXT-13553", 52}, { {"The execution Scope for OpControlBarrierArriveEXT and OpControlBarrierWaitEXT must be Workgroup or Subgroup", 107}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpConvertFToU-12465", 40}, { {"Floating-point objects consumed or produced by OpConvertFToU, OpConvertFToS, OpConvertSToF, or OpConvertUToF must not have a Float6E2M3EXT, Float6E3M2EXT, Float4E2M1EXT, Float8UnsignedE8M0EXT, or MXInt8EXT encoding", 214}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpEntryPoint-06674", 39}, { {"Each OpEntryPoint must statically use at most one variable in the PushConstant Storage Class unless the PushConstantBanksNV capability is declared", 146}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpEntryPoint-08721", 39}, { {"Each OpEntryPoint must not have more than one Input variable assigned the same Component word inside a Location slot, either explicitly or implicitly", 149}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpEntryPoint-08722", 39}, { {"Each OpEntryPoint must not have more than one Output variable assigned the same Component word inside a Location slot, either explicitly or implicitly", 150}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpEntryPoint-09658", 39}, { {"For a given OpEntryPoint, any BuiltIn decoration must not be used more than once by the Input interface", 103}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpEntryPoint-09659", 39}, { {"For a given OpEntryPoint, any BuiltIn decoration must not be used more than once by the Output interface", 104}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpFConvert-12466", 37}, { {"Floating-point objects produced by OpFConvert must not have a Float6E2M3EXT, Float6E3M2EXT, Float4E2M1EXT, Float8UnsignedE8M0EXT, or MXInt8EXT encoding", 151}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpFConvert-12467", 37}, { {"If the object consumed by OpFConvert has a Float6E2M3EXT, Float6E3M2EXT, Float4E2M1EXT, Float8UnsignedE8M0EXT, or MXInt8EXT encoding, then the Result Type must have an IEEE 754, Float8E4M3EXT, Float8E5M2EXT, or BFloat16KHR encoding", 231}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpGraphInputARM-09931", 42}, { {"The InputIndex and ElementIndex operands to OpGraphInputARM must be the of a constant instruction", 97}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpGraphSetOutputARM-09932", 46}, { {"The OutputIndex and ElementIndex operands to OpGraphSetOutputARM must be the of a constant instruction", 102}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpGroupNonUniformBallotBitCount-04685", 58}, { {"If OpGroupNonUniformBallotBitCount is used, the group operation must be limited to Reduce, InclusiveScan, or ExclusiveScan", 122}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpImage-04777", 34}, { {"OpImage*Dref* instructions must not consume an image whose Dim is 3D", 68}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpImageGather-04664", 40}, { {"The \"Component\" operand of OpImageGather, OpImageGatherQCOM, and OpImageSparseGather must be the of a constant instruction", 124}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpImageQuerySizeLod-04659", 46}, { {"OpImageQuerySizeLod, OpImageQueryLod, and OpImageQueryLevels must only consume an \"Image\" operand whose type has its \"Sampled\" operand set to 1", 147}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpImageTexelPointer-04658", 46}, { {"If an OpImageTexelPointer is used in an atomic operation, the image type of the image parameter to OpImageTexelPointer must have an image format of R64i, R64ui, R32f, R32i, or R32ui", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpReadClockKHR-04652", 41}, { {"Scope for OpReadClockKHR must be limited to Subgroup or Device", 62}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeFloat-10370", 38}, { {"Variables with a type of OpTypeFloat and an encoding of BFloat16KHR must not be declared with a Storage Class of Input or Output", 128}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeFloat-10823", 38}, { {"Variables with a type of OpTypeFloat and an encoding of Float8E4M3EXT or Float8E5M2EXT must not be declared with a Storage Class of Input or Output", 147}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeForwardPointer-04711", 47}, { {"OpTypeForwardPointer must have a Storage Class of PhysicalStorageBuffer", 71}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-04656", 38}, { {"OpTypeImage must declare a scalar 32-bit float, 64-bit integer, or 32-bit integer type for the \"Sampled Type\" (RelaxedPrecision can be applied to a sampling instruction and to the variable holding the result of a sampling instruction)", 236}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-04657", 38}, { {"OpTypeImage must have a \"Sampled\" operand of 1 (sampled image) or 2 (storage image)", 85}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-06214", 38}, { {"An OpTypeImage with a \"Dim\" operand of SubpassData must have an \"Arrayed\" operand of 0 (non-arrayed) and a \"Sampled\" operand of 2 (storage image)", 151}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-06924", 38}, { {"Objects of types OpTypeImage, OpTypeSampler, OpTypeSampledImage, OpTypeAccelerationStructureKHR, OpTypeTensorARM, and arrays of these types must not be stored to or modified", 173}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-09638", 38}, { {"An OpTypeImage must not have a \"Dim\" operand of Rect", 54}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-10693", 38}, { {"OpTypeImage variables in the TileAttachmentQCOM Storage Class must have Dim equal to 2D", 87}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeImage-10694", 38}, { {"OpTypeImage variables in the TileAttachmentQCOM Storage Class must Sampled equal to 1 or 2", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeRuntimeArray-04680", 45}, { {"OpTypeRuntimeArray must only be instantiated by a variable as: the last member of a Block-decorated OpTypeStruct in StorageBuffer or PhysicalStorageBuffer storage Storage Class BufferBlock-decorated OpTypeStruct in the Uniform storage Storage Class the last member of a Block-decorated OpTypeStruct in the Uniform storage Storage Class the outermost dimension of an arrayed variable in the StorageBuffer, Uniform, or UniformConstant storage Storage Class the NodePayloadAMDX storage Storage Class when the CoalescingAMDX Execution Mode is specified", 548}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeSampledImage-06671", 45}, { {"OpTypeSampledImage must have a OpTypeImage with a \"Sampled\" operand of 1 (sampled image)", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpTypeUntypedPointerKHR-11417", 50}, { {"The Storage Class of OpTypeUntypedPointerKHR must be Image, UniformConstant,Workgroup, StorageBuffer, Uniform, PushConstant, or PhysicalStorageBuffer", 149}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpUndef-12464", 34}, { {"Any OpUndef instruction must not have a Result Type of OpTypeImage, OpTypeSampler, OpTypeSampledImage, OpTypeAccelerationStructureKHR, or OpTypeTensorARM", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpUntypedImageTexelPointerEXT-11416", 56}, { {"If an OpUntypedImageTexelPointerEXT instruction is used in an atomic operation, the image type operand must have an image format of R64i, R64ui, R32f, R32i, or R32ui", 165}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpUntypedVariableKHR-11167", 47}, { {"Any OpUntypedVariableKHR with a Storage Class other than UniformConstant must have a Data Type operand specified", 112}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpUntypedVariableKHR-11347", 47}, { {"Any OpUntypedVariableKHR in the UniformConstant storage class without a Data Type must be decorated with SamplerHeapEXT or ResourceHeapEXT", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpVariable-04651", 37}, { {"Any variable with an Initializer operand must have Output, Private, Function, or Workgroup as its Storage Class operand", 119}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpVariable-04734", 37}, { {"Any variable with an Initializer operand and Workgroup as its Storage Class operand must use OpConstantNull as the initializer", 126}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OpVariable-06673", 37}, { {"There must be at most one variable in the PushConstant Storage Class listed in the Interface for each OpEntryPoint unless the PushConstantBanksNV capability is declared", 168}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-OriginLowerLeft-04653", 42}, { {"The OriginLowerLeft Execution Mode must not be used; fragment entry points must declare OriginUpperLeft", 103}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Output-10586", 33}, { {"Variables in the Output storage class in the Vertex, TessellationControl, TessellationEvaluation, or Geometry execution model must not have overlapping Component and Location decorations as defined by Location and Component Assignment", 234}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Output-10587", 33}, { {"Variables in the Output storage class in the Fragment execution model must not have both identical Index decorations and overlapping Component and Location decorations as defined by Location and Component Assignment", 215}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PerVertexKHR-06777", 39}, { {"The PerVertexKHR decoration must only be used on variables with the Input Storage Class in a fragment shader", 108}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PhysicalStorageBuffer64-04708", 50}, { {"If the PhysicalStorageBuffer64 addressing model is enabled, all instructions that support memory access operands and that use a physical pointer must include the Aligned operand", 177}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PhysicalStorageBuffer64-04709", 50}, { {"If the PhysicalStorageBuffer64 addressing model is enabled, any access chain instruction that accesses into a RowMajor matrix must only be used as the Pointer operand to OpLoad or OpStore", 187}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PhysicalStorageBuffer64-04710", 50}, { {"If the PhysicalStorageBuffer64 addressing model is enabled, OpConvertUToPtr and OpConvertPtrToU must use an integer type whose Width is 64", 138}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PhysicalStorageBuffer64-06314", 50}, { {"If the PhysicalStorageBuffer64 addressing model is enabled any load or store through a physical pointer type must have an Aligned operand which is a multiple of the size of the largest scalar type in the pointed-to type", 219}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PixelCenterInteger-04654", 45}, { {"The PixelCenterInteger Execution Mode must not be used (pixels are always centered at half-integer coordinates)", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Pointer-08973", 34}, { {"The Storage Class of the Pointer operand to OpCooperativeMatrixLoadKHR or OpCooperativeMatrixStoreKHR must be limited to Workgroup, StorageBuffer, or PhysicalStorageBuffer", 171}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PushConstant-06675", 39}, { {"Any variable in the PushConstant or StorageBuffer storage class must be decorated as Block", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-PushConstant-06808", 39}, { {"Any variable in the PushConstant Storage Class must be typed as OpTypeStruct", 76}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-RayPayloadKHR-04698", 40}, { {"RayPayloadKHR Storage Class must only be used in ray generation, closest hit or miss shaders", 92}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Result-04780", 33}, { {"The Result Type operand of any OpImageRead or OpImageSparseRead instruction must be a vector of four components", 111}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Result-11336", 33}, { {"If the Result Type operand of OpLoad is OpTypeSampler, Pointer must be derived from a variable decorated with Binding and DescriptorSet, or decorated with BuiltIn and SamplerHeapEXT", 181}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Result-11337", 33}, { {"If the Result Type operand of OpLoad is OpTypeImage, Pointer must be derived from a variable decorated with Binding and DescriptorSet, or decorated with BuiltIn and ResourceHeapEXT", 180}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Result-11339", 33}, { {"If the Result Type operand of OpLoad is OpTypeAccelerationStructureKHR, and Pointer is not in the Private or Function Storage Class, Pointer must be derived from a variable decorated with Binding and DescriptorSet, or decorated with BuiltIn and ResourceHeapEXT", 260}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Result-12468", 33}, { {"The Result Type operand of OpBitcastExtractEXT must be OpTypeFloat with a Float6E2M3EXT, Float6E3M2EXT, or Float4E2M1EXT encoding", 129}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-SampledImageArrayDynamicIndexing-10128", 59}, { {"If the SampledImageArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a sampled image or sampler, the sampled image or sampler through which that memory is accessed must be determined by constant integral expressions", 257}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Scope-12243", 32}, { {"The Scope operand of OpTypeCooperativeMatrixKHR must be limited to Workgroup or Subgroup", 88}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-ShaderRecordBufferKHR-07119", 48}, { {"ShaderRecordBufferKHR Storage Class must only be used in ray generation, intersection, any-hit, closest hit, callable, or miss shaders", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-SplitBarrierEXT-13552", 42}, { {"The SplitBarrierEXT capability must not be enabled in any stage other than compute", 82}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-StorageBufferArrayDynamicIndexing-10129", 60}, { {"If the StorageBufferArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a storage buffer, the storage buffer through which that memory is accessed must be determined by constant integral expressions", 238}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-StorageImageArrayDynamicIndexing-10130", 59}, { {"If the StorageImageArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a storage image, the storage image through which that memory is accessed must be determined by constant integral expressions", 235}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-StorageTexelBufferArrayDynamicIndexing-10133", 65}, { {"If the StorageTexelBufferArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a storage texel buffer, the storage texel buffer through which that memory is accessed must be determined by constant integral expressions", 255}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Stream-04694", 33}, { {"If any variables or block members in the output interface of the entry point being compiled are decorated with Stream, then all variables belonging to the same XfbBuffer must specify the same Stream value", 204}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-SubgroupVoteKHR-07951", 42}, { {"If none of the SubgroupVoteKHR, GroupNonUniform, or SubgroupBallotKHR capabilities are declared, Scope for memory must not be Subgroup", 134}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-SubpassData-04660", 38}, { {"The (u,v) coordinates used for a SubpassData must be the of a constant vector (0,0)", 83}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TessLevelInner-10880", 41}, { {"Any variable decorated with TessLevelInner or TessLevelOuter must also be decorated with Patch", 94}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileAttachmentQCOM-10689", 45}, { {"If the TileAttachmentQCOM Storage Class is used, the TileShadingQCOM capability must be enabled", 95}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileAttachmentQCOM-10695", 45}, { {"Any variable in the TileAttachmentQCOM Storage Class must be decorated with DescriptorSet and Binding", 101}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileAttachmentQCOM-10696", 45}, { {"Any variable in the TileAttachmentQCOM Storage Class must not be decorated with Component decoration", 100}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileAttachmentQCOM-10697", 45}, { {"An OpTypeImage variables in the TileAttachmentQCOM Storage Class must not be consumed by an OpImageQuery* instruction", 117}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileImageEXT-08723", 39}, { {"The tile image variable declarations must obey the constraints on the TileImageEXT Storage Class and the Location decoration described in Fragment Tile Image Interface", 167}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileShadingQCOM-10686", 42}, { {"TileShadingQCOM capability must not be enabled in any stage other than compute or fragment", 90}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileShadingRateQCOM-10691", 46}, { {"If the TileShadingRateQCOM Execution Mode is used, the TileShadingQCOM capability must be enabled", 97}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-TileShadingRateQCOM-10692", 46}, { {"If the TileShadingRateQCOM Execution Mode is used, LocalSize and LocalSizeId Execution Mode must not be specified", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Type-12297", 31}, { {"Any pointer type whose Type parameter is a vector type with more than four components (or an aggregate containing such a type) must have Storage Class of Function, Private, Uniform, Workgroup, StorageBuffer, PhysicalStorageBuffer, PushConstant, or ShaderRecordBufferKHR", 269}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UnequalMemorySemantics-10875", 49}, { {"UnequalMemorySemantics of OpAtomicCompareExchange must not use Release or AcquireRelease memory order", 101}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UnequalMemorySemantics-10876", 49}, { {"UnequalMemorySemantics of OpAtomicCompareExchange must not use a stronger memory order than the corresponding EqualMemorySemantics", 130}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UnequalMemorySemantics-10877", 49}, { {"UnequalMemorySemantics of OpAtomicCompareExchange must not have any Vulkan-supported storage class semantics bit set (UniformMemory, WorkgroupMemory, ImageMemory, or OutputMemory) unless this bit is also set in the corresponding EqualMemorySemantics", 249}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UnequalMemorySemantics-10878", 49}, { {"UnequalMemorySemantics of OpAtomicCompareExchange must not have MakeVisible bit set unless this bit is also set in the corresponding EqualMemorySemantics", 153}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UnequalMemorySemantics-10879", 49}, { {"UnequalMemorySemantics of OpAtomicCompareExchange must have Volatile bit set if and only if this bit is also set in the corresponding EqualMemorySemantics", 154}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Uniform-06676", 34}, { {"Any variable in the Uniform Storage Class must be decorated as Block or BufferBlock", 83}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Uniform-06807", 34}, { {"Any variable in the Uniform or StorageBuffer Storage Class must be typed as OpTypeStruct or an array of this type", 113}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-Uniform-06925", 34}, { {"Any variable in the Uniform Storage Class decorated as Block must not be stored to or modified", 94}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UniformBufferArrayDynamicIndexing-10127", 60}, { {"If the UniformBufferArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a uniform buffer, the uniform buffer through which that memory is accessed must be determined by constant integral expressions", 238}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UniformConstant-04655", 42}, { {"Any variable in the UniformConstant Storage Class must be typed as either OpTypeImage, OpTypeSampler, OpTypeSampledImage, OpTypeAccelerationStructureKHR, OpTypeTensorARM or an array of one of these types", 203}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UniformConstant-06677", 42}, { {"Any variable in the UniformConstant, StorageBuffer, or Uniform Storage Class must be decorated with either DescriptorSet and Binding, or BuiltIn with SamplerHeapEXT or ResourceHeapEXT", 183}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-UniformTexelBufferArrayDynamicIndexing-10132", 65}, { {"If the UniformTexelBufferArrayDynamicIndexing capability is not declared, and an instruction accesses memory through a uniform texel buffer, the uniform texel buffer through which that memory is accessed must be determined by constant integral expressions", 255}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-ViewportMaskNV-04674", 41}, { {"The ViewportMaskNV and ViewportIndex decorations must not both be statically used by one or more OpEntryPoint's that form the pre-rasterization shader stages of a graphics pipeline", 180}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-ViewportRelativeNV-04672", 45}, { {"The ViewportRelativeNV decoration must only be used on a variable decorated with Layer in the vertex, tessellation evaluation, or geometry shader stages", 152}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-ViewportRelativeNV-04673", 45}, { {"The ViewportRelativeNV decoration must not be used unless a variable decorated with one of ViewportIndex or ViewportMaskNV is also statically used by the same OpEntryPoint", 171}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-VulkanMemoryModel-04678", 44}, { {"If the VulkanMemoryModel capability is not declared, the Volatile decoration must be used on any variable declaration that includes one of the SMIDNV, WarpIDNV, SubgroupSize, SubgroupLocalInvocationId, SubgroupEqMask, SubgroupGeMask, SubgroupGtMask, SubgroupLeMask, or SubgroupLtMask BuiltIn decorations when used in the ray generation, closest hit, miss, intersection, or callable shaders, or with the RayTmaxKHR Builtin decoration when used in an intersection shader", 468}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-VulkanMemoryModel-04679", 44}, { {"If the VulkanMemoryModel capability is declared, the OpLoad instruction must use MemorySemantics with the Volatile flag when it accesses into any variable that includes one of the SMIDNV, WarpIDNV, SubgroupSize, SubgroupLocalInvocationId, SubgroupEqMask, SubgroupGeMask, SubgroupGtMask, SubgroupLeMask, or SubgroupLtMask BuiltIn decorations when used in the ray generation, closest hit, miss, intersection, or callable shaders, or with the RayTmaxKHR Builtin decoration when used in an intersection shader", 505}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-XfbBuffer-04693", 36}, { {"All variables or block members in the output interface of the entry point being compiled decorated with a specific XfbBuffer value must all be decorated with identical XfbStride values", 184}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-XfbBuffer-04696", 36}, { {"For any two variables or block members in the output interface of the entry point being compiled with the same XfbBuffer value, the ranges determined by the Offset decoration and the size of the type must not overlap", 216}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-StandaloneSpirv-XfbBuffer-04697", 36}, { {"All block members in the output interface of the entry point being compiled that are in the same block and have a declared or inherited XfbBuffer decoration must specify the same XfbBuffer value", 194}, { "appendices/spirvenv.html", 24} } },
        { {"VUID-SubgroupEqMask-SubgroupEqMask-04370", 40}, { {"The variable decorated with SubgroupEqMask must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupEqMask-SubgroupEqMask-04371", 40}, { {"The variable decorated with SubgroupEqMask must be declared as a four-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupGeMask-SubgroupGeMask-04372", 40}, { {"The variable decorated with SubgroupGeMask must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupGeMask-SubgroupGeMask-04373", 40}, { {"The variable decorated with SubgroupGeMask must be declared as a four-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupGtMask-SubgroupGtMask-04374", 40}, { {"The variable decorated with SubgroupGtMask must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupGtMask-SubgroupGtMask-04375", 40}, { {"The variable decorated with SubgroupGtMask must be declared as a four-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupId-SubgroupId-04367", 32}, { {"The SubgroupId decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 117}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupId-SubgroupId-04368", 32}, { {"The variable decorated with SubgroupId must be declared using the Input Storage Class", 85}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupId-SubgroupId-04369", 32}, { {"The variable decorated with SubgroupId must be declared as a scalar 32-bit integer value", 88}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLeMask-SubgroupLeMask-04376", 40}, { {"The variable decorated with SubgroupLeMask must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLeMask-SubgroupLeMask-04377", 40}, { {"The variable decorated with SubgroupLeMask must be declared as a four-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLocalInvocationId-SubgroupLocalInvocationId-04380", 62}, { {"The variable decorated with SubgroupLocalInvocationId must be declared using the Input Storage Class", 100}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLocalInvocationId-SubgroupLocalInvocationId-04381", 62}, { {"The variable decorated with SubgroupLocalInvocationId must be declared as a scalar 32-bit integer value", 103}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLtMask-SubgroupLtMask-04378", 40}, { {"The variable decorated with SubgroupLtMask must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupLtMask-SubgroupLtMask-04379", 40}, { {"The variable decorated with SubgroupLtMask must be declared as a four-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupSize-SubgroupSize-04382", 36}, { {"The variable decorated with SubgroupSize must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-SubgroupSize-SubgroupSize-04383", 36}, { {"The variable decorated with SubgroupSize must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TaskCountNV-TaskCountNV-04384", 34}, { {"The TaskCountNV decoration must be used only within the TaskNV Execution Model", 78}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TaskCountNV-TaskCountNV-04385", 34}, { {"The variable decorated with TaskCountNV must be declared using the Output Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TaskCountNV-TaskCountNV-04386", 34}, { {"The variable decorated with TaskCountNV must be declared as a scalar 32-bit integer value", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessCoord-TessCoord-04387", 30}, { {"The TessCoord decoration must be used only within the TessellationEvaluation Execution Model", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessCoord-TessCoord-04388", 30}, { {"The variable decorated with TessCoord must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessCoord-TessCoord-04389", 30}, { {"The variable decorated with TessCoord must be declared as a three-component vector of 32-bit floating-point values", 114}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelInner-TessLevelInner-04394", 40}, { {"The TessLevelInner decoration must be used only within the TessellationControl or TessellationEvaluation Execution Model", 120}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelInner-TessLevelInner-04395", 40}, { {"The variable decorated with TessLevelInner within the TessellationControl Execution Model must be declared using the Output Storage Class", 137}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelInner-TessLevelInner-04396", 40}, { {"The variable decorated with TessLevelInner within the TessellationEvaluation Execution Model must be declared using the Input Storage Class", 139}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelInner-TessLevelInner-04397", 40}, { {"The variable decorated with TessLevelInner must be declared as an array of size two, containing 32-bit floating-point values", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelOuter-TessLevelOuter-04390", 40}, { {"The TessLevelOuter decoration must be used only within the TessellationControl or TessellationEvaluation Execution Model", 120}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelOuter-TessLevelOuter-04391", 40}, { {"The variable decorated with TessLevelOuter within the TessellationControl Execution Model must be declared using the Output Storage Class", 137}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelOuter-TessLevelOuter-04392", 40}, { {"The variable decorated with TessLevelOuter within the TessellationEvaluation Execution Model must be declared using the Input Storage Class", 139}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TessLevelOuter-TessLevelOuter-04393", 40}, { {"The variable decorated with TessLevelOuter must be declared as an array of size four, containing 32-bit floating-point values", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileApronSizeQCOM-TileApronSizeQCOM-10632", 46}, { {"The TileApronSizeQCOM decoration must be used only within the Fragment Execution Model or GLCompute Execution Model", 115}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileApronSizeQCOM-TileApronSizeQCOM-10633", 46}, { {"The variable decorated with TileApronSizeQCOM must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileApronSizeQCOM-TileApronSizeQCOM-10634", 46}, { {"The variable decorated with TileApronSizeQCOM must be declared as a two-component vector of 32-bit integer values", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileDimensionQCOM-TileDimensionQCOM-10629", 46}, { {"The TileDimensionQCOM decoration must be used only within the Fragment Execution Model or GLCompute Execution Model", 115}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileDimensionQCOM-TileDimensionQCOM-10630", 46}, { {"The variable decorated with TileDimensionQCOM must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileDimensionQCOM-TileDimensionQCOM-10631", 46}, { {"The variable decorated with TileDimensionQCOM must be declared as a two-component vector of 32-bit integer values", 113}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileOffsetQCOM-TileOffsetQCOM-10626", 40}, { {"The TileOffsetQCOM decoration must be used only within the Fragment Execution Model or GLCompute Execution Model", 112}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileOffsetQCOM-TileOffsetQCOM-10627", 40}, { {"The variable decorated with TileOffsetQCOM must be declared using the Input Storage Class", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-TileOffsetQCOM-TileOffsetQCOM-10628", 40}, { {"The variable decorated with TileOffsetQCOM must be declared as a two-component vector of 32-bit integer values", 110}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexCountHUAWEI-VertexCountHUAWEI-07809", 46}, { {"The VertexCountHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 98}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexCountHUAWEI-VertexCountHUAWEI-07810", 46}, { {"The variable decorated with VertexCountHUAWEI must be declared as a scalar 32-bit integer value", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexIndex-VertexIndex-04398", 34}, { {"The VertexIndex decoration must be used only within the Vertex Execution Model", 78}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexIndex-VertexIndex-04399", 34}, { {"The variable decorated with VertexIndex must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexIndex-VertexIndex-04400", 34}, { {"The variable decorated with VertexIndex must be declared as a scalar 32-bit integer value", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexOffsetHUAWEI-VertexOffsetHUAWEI-07811", 48}, { {"The VertexOffsetHUAWEI decoration must be used only within the ClusterCullingHUAWEI Execution Model", 99}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VertexOffsetHUAWEI-VertexOffsetHUAWEI-07812", 48}, { {"The variable decorated with VertexOffsetHUAWEI must be declared as a scalar 32-bit integer value", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewIndex-ViewIndex-04401", 30}, { {"The ViewIndex decoration must be used only within the MeshEXT, Vertex, Geometry, TessellationControl, TessellationEvaluation or Fragment Execution Model", 152}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewIndex-ViewIndex-04402", 30}, { {"The variable decorated with ViewIndex must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewIndex-ViewIndex-04403", 30}, { {"The variable decorated with ViewIndex must be declared as a scalar 32-bit integer value", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-04404", 38}, { {"The ViewportIndex decoration must be used only within the MeshEXT, MeshNV, Vertex, TessellationEvaluation, Geometry, or Fragment Execution Model", 144}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-04405", 38}, { {"If the shaderOutputViewportIndex feature is not enabled then the ViewportIndex decoration must be used only within the Geometry or Fragment Execution Model", 155}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-04406", 38}, { {"The variable decorated with ViewportIndex within the MeshEXT, MeshNV, Vertex, TessellationEvaluation, or Geometry Execution Model must be declared using the Output Storage Class", 177}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-04407", 38}, { {"The variable decorated with ViewportIndex within the Fragment Execution Model must be declared using the Input Storage Class", 124}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-04408", 38}, { {"The variable decorated with ViewportIndex must be declared as a scalar 32-bit integer value for all supported execution models except MeshEXT", 141}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-07060", 38}, { {"The variable decorated with ViewportIndex within the MeshEXT Execution Model must also be decorated with the PerPrimitiveEXT decoration", 135}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-10601", 38}, { {"ViewportIndex within the MeshEXT Execution Model must decorate a scalar 32-bit integer member of a structure decorated as Block, or decorate a variable of type OpTypeArray of scalar 32-bit integer values", 203}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-10602", 38}, { {"If ViewportIndex is declared as an array of 32-bit integer values, within the MeshEXT Execution Model, size of the array must match the value specified by OutputPrimitivesEXT", 174}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportIndex-ViewportIndex-10603", 38}, { {"If ViewportIndex decorates a member of a structure, the variable declaration of the containing Block type must have an array size that matches the value specified by OutputPrimitivesEXT", 185}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskNV-ViewportMaskNV-04409", 40}, { {"The ViewportMaskNV decoration must be used only within the Vertex, MeshNV, TessellationEvaluation, or Geometry Execution Model", 126}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskNV-ViewportMaskNV-04410", 40}, { {"The variable decorated with ViewportMaskNV must be declared using the Output Storage Class", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskNV-ViewportMaskNV-04411", 40}, { {"The variable decorated with ViewportMaskNV must be declared as an array of 32-bit integer values", 96}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04412", 54}, { {"The ViewportMaskPerViewNV decoration must be used only within the Vertex, MeshNV, TessellationControl, TessellationEvaluation, or Geometry Execution Model", 154}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04413", 54}, { {"The variable decorated with ViewportMaskPerViewNV must be declared using the Output Storage Class", 97}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04414", 54}, { {"The variable decorated with ViewportMaskPerViewNV must be declared as an array of 32-bit integer values", 103}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04415", 54}, { {"The array decorated with ViewportMaskPerViewNV must be a size less than or equal to 32", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04416", 54}, { {"The array decorated with ViewportMaskPerViewNV must be a size greater than the maximum view in the subpass's view mask", 118}, { "chapters/interfaces.html", 24} } },
        { {"VUID-ViewportMaskPerViewNV-ViewportMaskPerViewNV-04417", 54}, { {"The array variable decorated with ViewportMaskPerViewNV must only be indexed by a constant or specialization constant", 117}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkAabbPositionsKHR-minX-03546", 34}, { {"minX must be less than or equal to maxX", 39}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAabbPositionsKHR-minY-03547", 34}, { {"minY must be less than or equal to maxY", 39}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAabbPositionsKHR-minZ-03548", 34}, { {"minZ must be less than or equal to maxZ", 39}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-commonparent", 61}, { {"Both of dstAccelerationStructure, and srcAccelerationStructure that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 182}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-dstAccelerationStructure-04927", 79}, { {"If dstAccelerationStructure was created with VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV set in VkAccelerationStructureCreateInfoKHR::createFlags, VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV must be set in flags", 216}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-03796", 60}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_TRACE_BIT_KHR bit set, then it must not have the VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_BUILD_BIT_KHR bit set", 175}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-04928", 60}, { {"If VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV is set in flags, dstAccelerationStructure must have been created with VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV set in VkAccelerationStructureCreateInfoKHR::createFlags", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-04929", 60}, { {"If VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV is set in flags, type must not be VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 125}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-11558", 60}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR bit set then it must not have the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_DATA_UPDATE_BIT_EXT bit set", 203}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-11562", 60}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, flags must only include the following bits: VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_TRACE_BIT_KHR VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_BUILD_BIT_KHR VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR", 335}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-11709", 60}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_DATA_UPDATE_BIT_EXT bit set, the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-11710", 60}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR bit set, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap or VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 232}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-11711", 60}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISABLE_OPACITY_MICROMAPS_BIT_KHR bit set, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap or VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 234}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-flags-parameter", 64}, { {"flags must be a valid combination of VkBuildAccelerationStructureFlagBitsKHR values", 83}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-format-11712", 61}, { {"If flags includes VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, each element of its pUsageCounts or ppUsageCounts members must have its format member equal to VK_OPACITY_MICROMAP_FORMAT_4_STATE_KHR", 350}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-format-11713", 61}, { {"If flags includes VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, each element of its triangleArray member must have its format member equal to VK_OPACITY_MICROMAP_FORMAT_4_STATE_KHR", 333}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-geometryCount-11560", 68}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, geometryCount must be 1", 87}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-geometryType-11561", 67}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, the geometryType member of elements of either pGeometries or ppGeometries must be VK_GEOMETRY_TYPE_MICROMAP_KHR", 175}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-geometryType-11707", 67}, { {"If type is not VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, the geometryType member of elements of either pGeometries or ppGeometries must not be VK_GEOMETRY_TYPE_MICROMAP_KHR", 183}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-micromap-11559", 63}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature must be enabled", 144}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-pGeometries-03788", 66}, { {"If geometryCount is not 0, exactly one of pGeometries or ppGeometries must be a valid pointer, the other must be NULL", 117}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-pGeometries-parameter", 70}, { {"If geometryCount is not 0, and pGeometries is not NULL, pGeometries must be a valid pointer to an array of geometryCount valid VkAccelerationStructureGeometryKHR structures", 172}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-ppGeometries-parameter", 71}, { {"If geometryCount is not 0, and ppGeometries is not NULL, ppGeometries must be a valid pointer to an array of geometryCount valid pointers to valid VkAccelerationStructureGeometryKHR structures", 192}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_GEOMETRY_INFO_KHR", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-subdivisionLevel-11564", 71}, { {"If flags does not include VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, for each element of its pUsageCounts or ppUsageCounts members, if its format member is VK_OPACITY_MICROMAP_FORMAT_4_STATE_KHR, its subdivisionLevel member must be less than or equal to maxOpacity4StateSubdivisionLevel", 442}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-subdivisionLevel-11565", 71}, { {"If flags includes VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, for each element of its pUsageCounts or ppUsageCounts members, its subdivisionLevel member must be less than or equal to maxOpacityLossy4StateSubdivisionLevel", 375}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-subdivisionLevel-11566", 71}, { {"If flags does not include VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, for each element of its triangleArray member, if its format member is VK_OPACITY_MICROMAP_FORMAT_4_STATE_KHR, its subdivisionLevel member must be less than or equal to maxOpacity4StateSubdivisionLevel", 425}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-subdivisionLevel-11567", 71}, { {"If flags includes VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, for each element of pGeometries or ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, for each element of its triangleArray member, its subdivisionLevel member must be less than or equal to maxOpacityLossy4StateSubdivisionLevel", 358}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03654", 59}, { {"type must not be VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 59}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03789", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, the geometryType member of elements of either pGeometries or ppGeometries must be VK_GEOMETRY_TYPE_INSTANCES_KHR", 169}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03790", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, geometryCount must be 1", 80}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03791", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR the geometryType member of elements of either pGeometries or ppGeometries must not be VK_GEOMETRY_TYPE_INSTANCES_KHR", 175}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03792", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR then the geometryType member of each geometry in either pGeometries or ppGeometries must be the same", 159}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03793", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR then geometryCount must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxGeometryCount", 176}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03794", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR and the geometryType member of either pGeometries or ppGeometries is VK_GEOMETRY_TYPE_AABBS_KHR, the total number of AABBs in all geometries must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxPrimitiveCount", 299}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-03795", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR and the geometryType member of either pGeometries or ppGeometries is VK_GEOMETRY_TYPE_TRIANGLES_KHR, the total number of triangles in all geometries must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxPrimitiveCount", 307}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-10884", 59}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR and the geometryType member of either pGeometries or ppGeometries is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, then geometryCount must be 1", 211}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-11563", 59}, { {"If flags includes VK_BUILD_ACCELERATION_STRUCTURE_MICROMAP_LOSSY_BIT_KHR, type must be VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 138}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildGeometryInfoKHR-type-parameter", 63}, { {"type must be a valid VkAccelerationStructureTypeKHR value", 57}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-None-10775", 56}, { {"For geometries of type VK_GEOMETRY_TYPE_TRIANGLES_KHR, if the geometry does not use indices, then VkAccelerationStructureGeometryTrianglesDataKHR::maxVertex must be greater than or equal to firstVertex + primitiveCount x 3 - 1", 226}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-maxVertex-10774", 61}, { {"For geometries of type VK_GEOMETRY_TYPE_TRIANGLES_KHR, if the geometry uses indices, then VkAccelerationStructureGeometryTrianglesDataKHR::maxVertex must be greater than or equal to firstVertex plus the maximum index value found in the VkAccelerationStructureGeometryTrianglesDataKHR::indexData in the range [primitiveOffset, primitiveOffset primitiveCount x 3]", 361}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-primitiveOffset-03656", 67}, { {"For geometries of type VK_GEOMETRY_TYPE_TRIANGLES_KHR, if the geometry uses indices, the offset primitiveOffset from VkAccelerationStructureGeometryTrianglesDataKHR::indexData must be a multiple of the element size of VkAccelerationStructureGeometryTrianglesDataKHR::indexType", 276}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-primitiveOffset-03657", 67}, { {"For geometries of type VK_GEOMETRY_TYPE_TRIANGLES_KHR, if the geometry does not use indices, the offset primitiveOffset from VkAccelerationStructureGeometryTrianglesDataKHR::vertexData must be a multiple of: the size of the format specified in VkAccelerationStructureGeometryTrianglesDataKHR::vertexFormat, if that format is a packed format the component size of the VkAccelerationStructureGeometryTrianglesDataKHR::vertexFormat, if that format is not a packed format", 467}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-primitiveOffset-03659", 67}, { {"For geometries of type VK_GEOMETRY_TYPE_AABBS_KHR, the offset primitiveOffset from VkAccelerationStructureGeometryAabbsDataKHR::data must be a multiple of 8", 156}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-primitiveOffset-03660", 67}, { {"For geometries of type VK_GEOMETRY_TYPE_INSTANCES_KHR, the offset primitiveOffset from VkAccelerationStructureGeometryInstancesDataKHR::data must be a multiple of 16", 165}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-transformOffset-03658", 67}, { {"For geometries of type VK_GEOMETRY_TYPE_TRIANGLES_KHR, the offset transformOffset from VkAccelerationStructureGeometryTrianglesDataKHR::transformData must be a multiple of 16", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildRangeInfoKHR-vertexData-10418", 62}, { {"The number of vertices consumed from VkAccelerationStructureGeometryTrianglesDataKHR::vertexData must be less than or equal to VkAccelerationStructureGeometryTrianglesDataKHR::maxVertex + 1", 189}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureBuildSizesInfoKHR-pNext-pNext", 57}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureBuildSizesInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_SIZES_INFO_KHR", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-None-12430", 67}, { {"If the accelerationStructure feature is not enabled, accelerationStructure must be VK_NULL_HANDLE", 97}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructure-08091", 84}, { {"If accelerationStructure is not VK_NULL_HANDLE then accelerationStructure must have been created with VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkAccelerationStructureCreateInfoKHR::createFlags", 232}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructure-08093", 84}, { {"If accelerationStructure is not VK_NULL_HANDLE then accelerationStructureNV must be VK_NULL_HANDLE", 98}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructure-parameter", 88}, { {"If accelerationStructure is not VK_NULL_HANDLE, accelerationStructure must be a valid VkAccelerationStructureKHR handle", 119}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructureNV-08092", 86}, { {"If accelerationStructureNV is not VK_NULL_HANDLE then accelerationStructureNV must have been created with VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkAccelerationStructureCreateInfoNV::info.flags", 234}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructureNV-08094", 86}, { {"If accelerationStructureNV is not VK_NULL_HANDLE then accelerationStructure must be VK_NULL_HANDLE", 98}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructureNV-12431", 86}, { {"If the VK_NV_ray_tracing extension is not enabled, then accelerationStructureNV must be VK_NULL_HANDLE", 102}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-accelerationStructureNV-parameter", 90}, { {"If accelerationStructureNV is not VK_NULL_HANDLE, accelerationStructureNV must be a valid VkAccelerationStructureNV handle", 122}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-commonparent", 69}, { {"Both of accelerationStructure, and accelerationStructureNV that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 178}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-pNext-pNext", 68}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCaptureDescriptorDataInfoEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT", 87}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressFlags-13100", 61}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressFlags-13101", 61}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressFlags-parameter", 65}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-11602", 61}, { {"addressRange.address must be a valid VkDeviceAddress", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-11603", 61}, { {"addressRange must specify a range within a valid address retrieved from buffer that was created with a usage value containing VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR", 180}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-11604", 61}, { {"addressRange must specify a range within a valid address retrieved from buffer that was not created with a flags value containing VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT", 167}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-11605", 61}, { {"addressRange.address must be a multiple of 256 bytes", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-11608", 61}, { {"addressRange.size must be greater than zero", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13097", 61}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13098", 61}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13099", 61}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13122", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13123", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13124", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-addressRange-13125", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-03613", 60}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR, VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureCaptureReplay must be VK_TRUE", 196}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-04954", 60}, { {"If VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV is set in createFlags and type is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, one member of the pNext chain must be a pointer to a valid instance of VkAccelerationStructureMotionInfoNV", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-04955", 60}, { {"If any geometry includes VkAccelerationStructureGeometryMotionTrianglesDataNV then createFlags must contain VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV", 154}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-08108", 60}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay feature must be enabled", 156}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-11606", 60}, { {"If the address member of addressRange was retrieved from a VkBuffer created with VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT, createFlags must include VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR", 228}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-11607", 60}, { {"If the address member of addressRange was retrieved from a VkBuffer not created with VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT, createFlags must not include VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-11610", 60}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, createFlags must not include VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 166}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-11611", 60}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, createFlags must not include VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV", 139}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-11673", 60}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR and device was created with multiple physical devices, the bufferDeviceAddressMultiDevice feature must be enabled", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-createFlags-parameter", 64}, { {"createFlags must be a valid combination of VkAccelerationStructureCreateFlagBitsKHR values", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-micromap-11609", 57}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature must be enabled", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-pNext-08109", 54}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, createFlags must contain VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-pNext-pNext", 54}, { {"pNext must be NULL or a pointer to a valid instance of VkOpaqueCaptureDescriptorDataCreateInfoEXT", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_2_KHR", 72}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfo2KHR-type-parameter", 57}, { {"type must be a valid VkAccelerationStructureTypeKHR value", 57}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-buffer-03614", 54}, { {"buffer must have been created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR usage flag set", 108}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-buffer-03615", 54}, { {"buffer must not have been created with VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-buffer-parameter", 58}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-03613", 59}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR, VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureCaptureReplay must be VK_TRUE", 196}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-04954", 59}, { {"If VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV is set in createFlags and type is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, one member of the pNext chain must be a pointer to a valid instance of VkAccelerationStructureMotionInfoNV", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-04955", 59}, { {"If any geometry includes VkAccelerationStructureGeometryMotionTrianglesDataNV then createFlags must contain VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV", 154}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-08108", 59}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay feature must be enabled", 156}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-11673", 59}, { {"If createFlags includes VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR and device was created with multiple physical devices, the bufferDeviceAddressMultiDevice feature must be enabled", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-createFlags-parameter", 63}, { {"createFlags must be a valid combination of VkAccelerationStructureCreateFlagBitsKHR values", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-deviceAddress-03612", 61}, { {"If deviceAddress is not zero, createFlags must include VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR", 125}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-deviceAddress-09488", 61}, { {"If deviceAddress is not zero, it must have been retrieved from an identically created acceleration structure, except for buffer and deviceAddress", 145}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-deviceAddress-09489", 61}, { {"If deviceAddress is not zero, buffer must have been created identically to the buffer used to create the acceleration structure from which deviceAddress was retrieved, except for VkBufferOpaqueCaptureAddressCreateInfo::opaqueCaptureAddress", 239}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-deviceAddress-09490", 61}, { {"If deviceAddress is not zero, buffer must have been created with a VkBufferOpaqueCaptureAddressCreateInfo::opaqueCaptureAddress that was retrieved from vkGetBufferOpaqueCaptureAddress for the buffer that was used to create the acceleration structure from which deviceAddress was retrieved", 288}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-deviceAddress-parameter", 65}, { {"If deviceAddress is not 0, deviceAddress must be a valid VkDeviceAddress value", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-offset-03616", 54}, { {"The sum of offset and size must be less than or equal to the size of buffer", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-offset-03734", 54}, { {"offset must be a multiple of 256 bytes", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-pNext-08109", 53}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, createFlags must contain VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-pNext-pNext", 53}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAccelerationStructureMotionInfoNV or VkOpaqueCaptureDescriptorDataCreateInfoEXT", 212}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_KHR", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-sType-unique", 54}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-type-11600", 52}, { {"type must not equal VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoKHR-type-parameter", 56}, { {"type must be a valid VkAccelerationStructureTypeKHR value", 57}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoNV-compactedSize-02421", 60}, { {"If compactedSize is not 0 then both info.geometryCount and info.instanceCount must be 0", 87}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoNV-info-parameter", 55}, { {"info must be a valid VkAccelerationStructureInfoNV structure", 60}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoNV-pNext-pNext", 52}, { {"pNext must be NULL or a pointer to a valid instance of VkOpaqueCaptureDescriptorDataCreateInfoEXT", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoNV-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_NV", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureCreateInfoNV-sType-unique", 53}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-compressedData-10885", 85}, { {"The buffer from which compressedData.deviceAddress is queried must have been created with the VK_BUFFER_USAGE_2_COMPRESSED_DATA_DGF1_BIT_AMDX usage flag set", 156}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-compressedData-parameter", 89}, { {"compressedData must be a valid VkDeviceOrHostAddressConstKHR union", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-denseGeometryFormat-10886", 90}, { {"The VkPhysicalDeviceDenseGeometryFormatFeaturesAMDX::denseGeometryFormat feature must be enabled", 96}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-format-10887", 77}, { {"If format is VK_COMPRESSED_TRIANGLE_FORMAT_DGF1_AMDX, then compressedData must be aligned to VK_COMPRESSED_TRIANGLE_FORMAT_DGF1_BYTE_ALIGNMENT_AMDX (128) bytes", 159}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-format-10888", 77}, { {"If format is VK_COMPRESSED_TRIANGLE_FORMAT_DGF1_AMDX, then dataSize must be a multiple of VK_COMPRESSED_TRIANGLE_FORMAT_DGF1_BYTE_STRIDE_AMDX (128) bytes", 153}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-format-parameter", 81}, { {"format must be a valid VkCompressedTriangleFormatAMDX value", 59}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX-sType-sType", 76}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DENSE_GEOMETRY_FORMAT_TRIANGLES_DATA_AMDX", 96}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureDeviceAddressInfoKHR-accelerationStructure-parameter", 80}, { {"accelerationStructure must be a valid VkAccelerationStructureKHR handle", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureDeviceAddressInfoKHR-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureDeviceAddressInfoKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DEVICE_ADDRESS_INFO_KHR", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureGeometryAabbsDataKHR-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryAabbsDataKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_AABBS_DATA_KHR", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryAabbsDataKHR-stride-03545", 61}, { {"stride must be a multiple of 8", 30}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryAabbsDataKHR-stride-03820", 61}, { {"stride must be less than or equal to 2^32-1", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryInstancesDataKHR-pNext-pNext", 64}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryInstancesDataKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_INSTANCES_DATA_KHR", 82}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-aabbs-parameter", 55}, { {"If geometryType is VK_GEOMETRY_TYPE_AABBS_KHR, the aabbs member of geometry must be a valid VkAccelerationStructureGeometryAabbsDataKHR structure", 145}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-flags-11569", 51}, { {"If geometryType is VK_GEOMETRY_TYPE_MICROMAP_KHR, flags must be 0", 65}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-flags-parameter", 55}, { {"flags must be a valid combination of VkGeometryFlagBitsKHR values", 65}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-geometryType-parameter", 62}, { {"geometryType must be a valid VkGeometryTypeKHR value", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-instances-parameter", 59}, { {"If geometryType is VK_GEOMETRY_TYPE_INSTANCES_KHR, the instances member of geometry must be a valid VkAccelerationStructureGeometryInstancesDataKHR structure", 157}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-micromap-11568", 54}, { {"If geometryType is VK_GEOMETRY_TYPE_MICROMAP_KHR, pNext chain must include a single VkAccelerationStructureGeometryMicromapDataKHR structure", 140}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-pNext-pNext", 51}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX, VkAccelerationStructureGeometryLinearSweptSpheresDataNV, VkAccelerationStructureGeometryMicromapDataKHR, or VkAccelerationStructureGeometrySpheresDataNV", 344}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_KHR", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryKHR-triangles-parameter", 59}, { {"If geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, the triangles member of geometry must be a valid VkAccelerationStructureGeometryTrianglesDataKHR structure", 157}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-None-10419", 71}, { {"The linearSweptSpheres feature must be enabled", 46}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-endCapsMode-parameter", 82}, { {"endCapsMode must be a valid VkRayTracingLssPrimitiveEndCapsModeNV value", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-indexData-10428", 76}, { {"indexType must be VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, or VK_INDEX_TYPE_NONE_KHR", 87}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-indexData-parameter", 80}, { {"indexData must be a valid VkDeviceOrHostAddressConstKHR union", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-indexType-parameter", 80}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-indexingMode-10427", 79}, { {"If indexingMode is VK_RAY_TRACING_LSS_INDEXING_MODE_SUCCESSIVE_NV, indexData must not be NULL", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-indexingMode-parameter", 83}, { {"indexingMode must be a valid VkRayTracingLssIndexingModeNV value", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-radiusData-10426", 77}, { {"All values referenced in radiusData must be greater than or equal to 0", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-radiusData-parameter", 81}, { {"radiusData must be a valid VkDeviceOrHostAddressConstKHR union", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-radiusFormat-10424", 79}, { {"The format features of radiusFormat must contain VK_FORMAT_FEATURE_2_ACCELERATION_STRUCTURE_RADIUS_BUFFER_BIT_NV", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-radiusFormat-parameter", 83}, { {"radiusFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_LINEAR_SWEPT_SPHERES_DATA_NV", 92}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-vertexData-parameter", 81}, { {"vertexData must be a valid VkDeviceOrHostAddressConstKHR union", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-vertexFormat-10423", 79}, { {"The format features of vertexFormat must contain VK_FORMAT_FEATURE_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-vertexFormat-parameter", 83}, { {"vertexFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-vertexStride-10421", 79}, { {"vertexStride must be a multiple of: the size of the format specified in vertexFormat if that format is a packed format the component size specified in vertexFormat if that format is not a packed format", 201}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryLinearSweptSpheresDataNV-vertexStride-10422", 79}, { {"vertexStride and radiusStride must be less than or equal to 2^32-1", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-count-11643", 63}, { {"The total sum of count from all elements of pUsageCounts and ppUsageCounts must be less than or equal to maxMicromapTriangles", 125}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-pUsageCounts-11642", 70}, { {"Only one of pUsageCounts or ppUsageCounts can be a valid pointer, the other must be NULL", 88}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-pUsageCounts-parameter", 74}, { {"If usageCountsCount is not 0, and pUsageCounts is not NULL, pUsageCounts must be a valid pointer to an array of usageCountsCount valid VkMicromapUsageKHR structures", 164}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-ppUsageCounts-parameter", 75}, { {"If usageCountsCount is not 0, and ppUsageCounts is not NULL, ppUsageCounts must be a valid pointer to an array of usageCountsCount valid pointers to valid VkMicromapUsageKHR structures", 184}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MICROMAP_DATA_KHR", 81}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-subdivisionLevel-11645", 74}, { {"For each member of pUsageCounts or ppUsageCounts, if format is VK_OPACITY_MICROMAP_FORMAT_2_STATE_KHR, subdivisionLevel must be less than or equal to maxOpacity2StateSubdivisionLevel", 182}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-triangleArrayStride-11646", 77}, { {"triangleArrayStride must be greater than or equal to 8 bytes", 60}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMicromapDataKHR-triangleArrayStride-11647", 77}, { {"triangleArrayStride must be a multiple of 4 bytes", 49}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkAccelerationStructureGeometryMotionTrianglesDataNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MOTION_TRIANGLES_DATA_NV", 88}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-None-10429", 60}, { {"The spheres feature must be enabled", 35}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-indexData-10437", 65}, { {"indexType must be VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, VK_INDEX_TYPE_NONE_KHR", 84}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-indexData-parameter", 69}, { {"indexData must be a valid VkDeviceOrHostAddressConstKHR union", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-indexType-parameter", 69}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-radiusData-10436", 66}, { {"All values referenced in radiusData must be greater than or equal to 0", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-radiusData-parameter", 70}, { {"radiusData must be a valid VkDeviceOrHostAddressConstKHR union", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-radiusFormat-10435", 68}, { {"The format features of radiusFormat must contain VK_FORMAT_FEATURE_2_ACCELERATION_STRUCTURE_RADIUS_BUFFER_BIT_NV", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-radiusFormat-parameter", 72}, { {"radiusFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_SPHERES_DATA_NV", 79}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-vertexData-parameter", 70}, { {"vertexData must be a valid VkDeviceOrHostAddressConstKHR union", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-vertexFormat-10434", 68}, { {"The format features of vertexFormat must contain VK_FORMAT_FEATURE_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-vertexFormat-parameter", 72}, { {"vertexFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-vertexStride-10431", 68}, { {"vertexStride must be a multiple of: the size of the format specified in vertexFormat if that format is a packed format the smallest component size specified in vertexFormat if that format is not a packed format", 210}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometrySpheresDataNV-vertexStride-10432", 68}, { {"vertexStride and radiusStride must be less than or equal to 2^32-1", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-indexType-03798", 68}, { {"indexType must be VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, or VK_INDEX_TYPE_NONE_KHR", 87}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-indexType-parameter", 72}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-pNext-pNext", 64}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAccelerationStructureGeometryMotionTrianglesDataNV, VkAccelerationStructureTrianglesDisplacementMicromapNV, VkAccelerationStructureTrianglesOpacityMicromapEXT, or VkAccelerationStructureTrianglesOpacityMicromapKHR", 346}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_TRIANGLES_DATA_KHR", 82}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-sType-unique", 65}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-vertexFormat-03797", 71}, { {"The format features of vertexFormat must contain VK_FORMAT_FEATURE_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-vertexFormat-parameter", 75}, { {"vertexFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-vertexStride-03735", 71}, { {"vertexStride must be a multiple of the size in bytes of the smallest component of vertexFormat", 94}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureGeometryTrianglesDataKHR-vertexStride-03819", 71}, { {"vertexStride must be less than or equal to 2^32-1", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureInfoNV-flags-02592", 46}, { {"If flags has the VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_TRACE_BIT_NV bit set, then it must not have the VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_BUILD_BIT_NV bit set", 173}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-flags-parameter", 50}, { {"flags must be a valid combination of VkBuildAccelerationStructureFlagBitsKHR values", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-geometryCount-02422", 54}, { {"geometryCount must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxGeometryCount", 100}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-instanceCount-02423", 54}, { {"instanceCount must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxInstanceCount", 100}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-instanceData-02782", 53}, { {"If instanceData is not VK_NULL_HANDLE, instanceData must have been created with the VK_BUFFER_USAGE_RAY_TRACING_BIT_NV usage flag set", 133}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-maxTriangleCount-02424", 57}, { {"The total number of triangles in all geometries must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxTriangleCount", 134}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-pGeometries-parameter", 56}, { {"If geometryCount is not 0, pGeometries must be a valid pointer to an array of geometryCount valid VkGeometryNV structures", 121}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_INFO_NV", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-scratch-02781", 48}, { {"scratch must have been created with the VK_BUFFER_USAGE_RAY_TRACING_BIT_NV usage flag set", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-type-02425", 45}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_NV then geometryCount must be 0", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-type-02426", 45}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_NV then instanceCount must be 0", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-type-02786", 45}, { {"If type is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_NV then the geometryType member of each geometry in pGeometries must be the same", 135}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-type-04623", 45}, { {"type must not be VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInfoNV-type-parameter", 49}, { {"type must be a valid VkAccelerationStructureTypeNV value", 56}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureInstanceKHR-flags-11581", 51}, { {"If flags includes VK_GEOMETRY_INSTANCE_DISABLE_OPACITY_MICROMAPS_BIT_KHR, the flags used to build the acceleration structure referenced by accelerationStructureReference must have included VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISABLE_OPACITY_MICROMAPS_BIT_KHR", 260}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureInstanceKHR-flags-parameter", 55}, { {"flags must be a valid combination of VkGeometryInstanceFlagBitsKHR values", 73}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureInstanceKHR-micromap-11580", 54}, { {"If flags includes VK_GEOMETRY_INSTANCE_FORCE_OPACITY_MICROMAP_2_STATE_BIT_KHR or VK_GEOMETRY_INSTANCE_DISABLE_OPACITY_MICROMAPS_BIT_KHR, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap or VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 273}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMatrixMotionInstanceNV-flags-parameter", 66}, { {"flags must be a valid combination of VkGeometryInstanceFlagBitsKHR values", 73}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMemoryRequirementsInfoNV-accelerationStructure-parameter", 84}, { {"accelerationStructure must be a valid VkAccelerationStructureNV handle", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMemoryRequirementsInfoNV-pNext-pNext", 64}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMemoryRequirementsInfoNV-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_INFO_NV", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMemoryRequirementsInfoNV-type-parameter", 67}, { {"type must be a valid VkAccelerationStructureMemoryRequirementsTypeNV value", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMotionInfoNV-flags-zerobitmask", 58}, { {"flags must be 0", 15}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMotionInfoNV-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MOTION_INFO_NV", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkAccelerationStructureMotionInstanceNV-flags-zerobitmask", 62}, { {"flags must be 0", 15}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMotionInstanceNV-matrixMotionInstance-parameter", 75}, { {"If type is VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_MATRIX_MOTION_NV, the matrixMotionInstance member of data must be a valid VkAccelerationStructureMatrixMotionInstanceNV structure", 187}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMotionInstanceNV-srtMotionInstance-parameter", 72}, { {"If type is VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_SRT_MOTION_NV, the srtMotionInstance member of data must be a valid VkAccelerationStructureSRTMotionInstanceNV structure", 178}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMotionInstanceNV-staticInstance-parameter", 69}, { {"If type is VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_STATIC_NV, the staticInstance member of data must be a valid VkAccelerationStructureInstanceKHR structure", 163}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureMotionInstanceNV-type-parameter", 59}, { {"type must be a valid VkAccelerationStructureMotionInstanceTypeNV value", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureSRTMotionInstanceNV-flags-parameter", 63}, { {"flags must be a valid combination of VkGeometryInstanceFlagBitsKHR values", 73}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-displacementBiasAndScaleFormat-09501", 96}, { {"displacementBiasAndScaleFormat must not be VK_FORMAT_UNDEFINED", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-displacementBiasAndScaleFormat-parameter", 100}, { {"displacementBiasAndScaleFormat must be a valid VkFormat value", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-displacementVectorFormat-09502", 90}, { {"displacementVectorFormat must not be VK_FORMAT_UNDEFINED", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-displacementVectorFormat-parameter", 94}, { {"displacementVectorFormat must be a valid VkFormat value", 55}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-indexType-parameter", 79}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-micromap-parameter", 78}, { {"If micromap is not VK_NULL_HANDLE, micromap must be a valid VkMicromapEXT handle", 80}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-pUsageCounts-07992", 78}, { {"Only one of pUsageCounts or ppUsageCounts can be a valid pointer, the other must be NULL", 88}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-pUsageCounts-parameter", 82}, { {"If usageCountsCount is not 0, and pUsageCounts is not NULL, pUsageCounts must be a valid pointer to an array of usageCountsCount VkMicromapUsageEXT structures", 158}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-ppUsageCounts-parameter", 83}, { {"If usageCountsCount is not 0, and ppUsageCounts is not NULL, ppUsageCounts must be a valid pointer to an array of usageCountsCount valid pointers to VkMicromapUsageEXT structures", 178}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesDisplacementMicromapNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_DISPLACEMENT_MICROMAP_NV", 89}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-format-11679", 68}, { {"If micromap is not VK_NULL_HANDLE, the type used to construct micromap is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, the format member of each element of pUsageCounts or ppUsageCounts must be a valid value from VkOpacityMicromapFormatEXT", 233}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-format-11682", 68}, { {"If micromap is not VK_NULL_HANDLE, the type used to construct micromap is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, the format member of each element of pUsageCounts or ppUsageCounts must be a valid value from VkDisplacementMicromapFormatNV", 241}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-indexType-10719", 71}, { {"indexType must be VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, or VK_INDEX_TYPE_NONE_KHR", 87}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-indexType-10722", 71}, { {"If indexType is not VK_INDEX_TYPE_NONE_KHR, then indexStride must be a multiple of the size in bytes of indexType", 113}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-indexType-10723", 71}, { {"If indexType is not VK_INDEX_TYPE_NONE_KHR, then indexStride must be less than or equal to 2^32-1", 97}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-indexType-parameter", 75}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-micromap-parameter", 74}, { {"If micromap is not VK_NULL_HANDLE, micromap must be a valid VkMicromapEXT handle", 80}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-pUsageCounts-07335", 74}, { {"Only one of pUsageCounts or ppUsageCounts can be a valid pointer, the other must be NULL", 88}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-pUsageCounts-parameter", 78}, { {"If usageCountsCount is not 0, and pUsageCounts is not NULL, pUsageCounts must be a valid pointer to an array of usageCountsCount VkMicromapUsageEXT structures", 158}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-ppUsageCounts-parameter", 79}, { {"If usageCountsCount is not 0, and ppUsageCounts is not NULL, ppUsageCounts must be a valid pointer to an array of usageCountsCount valid pointers to VkMicromapUsageEXT structures", 178}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_EXT", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-subdivisionLevel-11680", 78}, { {"If micromap is not VK_NULL_HANDLE, the type used to build micromap is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of pUsageCounts or ppUsageCounts with format equal to VK_OPACITY_MICROMAP_FORMAT_2_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity2StateSubdivisionLevel", 344}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-subdivisionLevel-11681", 78}, { {"If micromap is not VK_NULL_HANDLE, the type used to build micromap is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of pUsageCounts or ppUsageCounts with format equal to VK_OPACITY_MICROMAP_FORMAT_4_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity4StateSubdivisionLevel", 344}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapEXT-subdivisionLevel-11683", 78}, { {"If micromap is not VK_NULL_HANDLE, the type used to construct micromap is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, for each member of pUsageCounts or ppUsageCounts subdivisionLevel must be less than or equal to VkPhysicalDeviceDisplacementMicromapPropertiesNV::maxDisplacementMicromapSubdivisionLevel", 302}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-geometry-11576", 70}, { {"If indexType is VK_INDEX_TYPE_NONE_KHR, for each triangle index in the geometry, index plus baseTriangle must be less than or equal to numTriangles, where numTriangles is given by the sum of all count parameters in the pUsageCounts or ppUsageCounts provided to the micromap's build command", 289}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexBuffer-11571", 73}, { {"If indexType is VK_INDEX_TYPE_NONE_KHR, indexBuffer must be 0", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexBuffer-11572", 73}, { {"If indexType is not VK_INDEX_TYPE_NONE_KHR, indexBuffer must be a valid VkDeviceAddress", 87}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexStride-11573", 73}, { {"If indexType is not VK_INDEX_TYPE_NONE_KHR, indexStride must be a multiple of the size in bytes of indexType", 108}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexStride-11574", 73}, { {"If indexType is not VK_INDEX_TYPE_NONE_KHR, indexStride must be less than or equal to 2^32-1", 92}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexType-11570", 71}, { {"indexType must be VK_INDEX_TYPE_UINT8, VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, or VK_INDEX_TYPE_NONE_KHR", 108}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-indexType-parameter", 75}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-micromap-11579", 70}, { {"If indexType is VK_INDEX_TYPE_NONE_KHR, micromap must not be VK_NULL_HANDLE", 75}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-micromap-parameter", 74}, { {"If micromap is not VK_NULL_HANDLE, micromap must be a valid VkAccelerationStructureKHR handle", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureTrianglesOpacityMicromapKHR-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_KHR", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureVersionInfoKHR-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureVersionInfoKHR-pVersionData-parameter", 65}, { {"pVersionData must be a valid pointer to an array of \\(2 \\times \\mathtt{VK\\_UUID\\_SIZE}\\) uint8_t values", 109}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAccelerationStructureVersionInfoKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_VERSION_INFO_KHR", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkAcquireNextImageInfoKHR-commonparent", 43}, { {"Each of fence, semaphore, and swapchain that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 159}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-deviceMask-01290", 47}, { {"deviceMask must be a valid device mask", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-deviceMask-01291", 47}, { {"deviceMask must not be zero", 27}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-fence-01289", 42}, { {"If fence is not VK_NULL_HANDLE, fence must be unsignaled", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-fence-10067", 42}, { {"If fence is not VK_NULL_HANDLE, fence must not be associated with any other queue command that has not yet completed execution on that queue", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-fence-parameter", 46}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_ACQUIRE_NEXT_IMAGE_INFO_KHR", 59}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-semaphore-01288", 46}, { {"If semaphore is not VK_NULL_HANDLE, it must be unsignaled", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-semaphore-01781", 46}, { {"If semaphore is not VK_NULL_HANDLE, it must not have any uncompleted signal or wait operations pending", 102}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-semaphore-01782", 46}, { {"semaphore and fence must not both be equal to VK_NULL_HANDLE", 60}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-semaphore-03266", 46}, { {"semaphore must have a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-semaphore-parameter", 50}, { {"If semaphore is not VK_NULL_HANDLE, semaphore must be a valid VkSemaphore handle", 80}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-swapchain-01675", 46}, { {"swapchain must not be in the retired state", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireNextImageInfoKHR-swapchain-parameter", 50}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAcquireProfilingLockInfoKHR-flags-zerobitmask", 52}, { {"flags must be 0", 15}, { "chapters/queries.html", 21} } },
        { {"VUID-VkAcquireProfilingLockInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkAcquireProfilingLockInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_ACQUIRE_PROFILING_LOCK_INFO_KHR", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-VkAllocationCallbacks-pfnAllocation-00632", 46}, { {"pfnAllocation must be a valid pointer to a valid application-defined PFN_vkAllocationFunction", 93}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAllocationCallbacks-pfnFree-00634", 40}, { {"pfnFree must be a valid pointer to a valid application-defined PFN_vkFreeFunction", 81}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAllocationCallbacks-pfnInternalAllocation-00635", 54}, { {"If either of pfnInternalAllocation or pfnInternalFree is not NULL, both must be valid callbacks", 95}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAllocationCallbacks-pfnReallocation-00633", 48}, { {"pfnReallocation must be a valid pointer to a valid application-defined PFN_vkReallocationFunction", 97}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAmigoProfilingSubmitInfoSEC-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_AMIGO_PROFILING_SUBMIT_INFO_SEC", 63}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkAndroidHardwareBufferFormatProperties2ANDROID-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_2_ANDROID", 83}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferFormatPropertiesANDROID-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_ANDROID", 81}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferFormatResolvePropertiesANDROID-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_RESOLVE_PROPERTIES_ANDROID", 89}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferPropertiesANDROID-pNext-pNext", 57}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAndroidHardwareBufferFormatProperties2ANDROID, VkAndroidHardwareBufferFormatPropertiesANDROID, or VkAndroidHardwareBufferFormatResolvePropertiesANDROID", 284}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferPropertiesANDROID-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_PROPERTIES_ANDROID", 74}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferPropertiesANDROID-sType-unique", 58}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkAndroidHardwareBufferUsageANDROID-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_USAGE_ANDROID", 69}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkAndroidSurfaceCreateInfoKHR-flags-zerobitmask", 52}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAndroidSurfaceCreateInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAndroidSurfaceCreateInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_ANDROID_SURFACE_CREATE_INFO_KHR", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAndroidSurfaceCreateInfoKHR-window-01248", 47}, { {"window must point to a valid Android ANativeWindow", 50}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAntiLagDataAMD-mode-parameter", 36}, { {"mode must be a valid VkAntiLagModeAMD value", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAntiLagDataAMD-pPresentationInfo-parameter", 49}, { {"If pPresentationInfo is not NULL, pPresentationInfo must be a valid pointer to a valid VkAntiLagPresentationInfoAMD structure", 125}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAntiLagDataAMD-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_ANTI_LAG_DATA_AMD", 49}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAntiLagPresentationInfoAMD-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_ANTI_LAG_PRESENTATION_INFO_AMD", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkAntiLagPresentationInfoAMD-stage-parameter", 49}, { {"stage must be a valid VkAntiLagStageAMD value", 45}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkApplicationInfo-apiVersion-04010", 39}, { {"If apiVersion is not 0, then it must be greater than or equal to VK_API_VERSION_1_0", 83}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkApplicationInfo-pApplicationName-parameter", 49}, { {"If pApplicationName is not NULL, pApplicationName must be a null-terminated UTF-8 string", 88}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkApplicationInfo-pEngineName-parameter", 44}, { {"If pEngineName is not NULL, pEngineName must be a null-terminated UTF-8 string", 78}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkApplicationInfo-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkApplicationInfo-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_APPLICATION_INFO", 48}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkAttachmentDescription-attachmentFeedbackLoopLayout-07309", 63}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-attachmentFeedbackLoopLayout-07310", 63}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-dynamicRenderingLocalRead-09544", 60}, { {"If the dynamicRenderingLocalRead feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 119}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-dynamicRenderingLocalRead-09545", 60}, { {"If the dynamicRenderingLocalRead feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-finalLayout-00843", 46}, { {"finalLayout must not be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-finalLayout-parameter", 50}, { {"finalLayout must be a valid VkImageLayout value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-11773", 40}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR, flags must not include VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-11774", 40}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveSrgbFormatSupportsTransferFunctionControl must be VK_TRUE", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-11775", 40}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, maintenance10 must be enabled", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-11776", 40}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, format must use sRGB encoding", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-11777", 40}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, samples must be VK_SAMPLE_COUNT_1_BIT", 191}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-flags-parameter", 44}, { {"flags must be a valid combination of VkAttachmentDescriptionFlagBits values", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03280", 41}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 154}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03281", 41}, { {"If format is a depth/stencil format, initialLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 103}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03282", 41}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 152}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03283", 41}, { {"If format is a depth/stencil format, finalLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03286", 41}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 226}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03287", 41}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 224}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03290", 41}, { {"If format is a depth/stencil format which includes only the depth component, initialLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 190}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03291", 41}, { {"If format is a depth/stencil format which includes only the depth component, finalLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03292", 41}, { {"If format is a depth/stencil format which includes only the stencil component, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-03293", 41}, { {"If format is a depth/stencil format which includes only the stencil component, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06242", 41}, { {"If format is a depth/stencil format which includes both depth and stencil components, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06243", 41}, { {"If format is a depth/stencil format which includes both depth and stencil components, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 193}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06487", 41}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 175}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06488", 41}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06698", 41}, { {"format must not be VK_FORMAT_UNDEFINED", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06699", 41}, { {"If format includes a color or depth component and loadOp is VK_ATTACHMENT_LOAD_OP_LOAD, then initialLayout must not be VK_IMAGE_LAYOUT_UNDEFINED", 144}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06700", 41}, { {"If format includes a stencil component and stencilLoadOp is VK_ATTACHMENT_LOAD_OP_LOAD, then initialLayout must not be VK_IMAGE_LAYOUT_UNDEFINED", 144}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06906", 41}, { {"If format is a depth/stencil format which includes both depth and stencil components, initialLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 199}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-06907", 41}, { {"If format is a depth/stencil format which includes both depth and stencil components, finalLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 197}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-format-parameter", 45}, { {"format must be a valid VkFormat value", 37}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-initialLayout-parameter", 52}, { {"initialLayout must be a valid VkImageLayout value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-loadOp-parameter", 45}, { {"loadOp must be a valid VkAttachmentLoadOp value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-samples-08745", 42}, { {"samples must be a valid VkSampleCountFlagBits value that is set in imageCreateSampleCounts (as defined in Image Creation Limits) for the given format", 149}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-samples-parameter", 46}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-separateDepthStencilLayouts-03284", 62}, { {"If the separateDepthStencilLayouts feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 257}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-separateDepthStencilLayouts-03285", 62}, { {"If the separateDepthStencilLayouts feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 255}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-stencilLoadOp-parameter", 52}, { {"stencilLoadOp must be a valid VkAttachmentLoadOp value", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-stencilStoreOp-parameter", 53}, { {"stencilStoreOp must be a valid VkAttachmentStoreOp value", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-storeOp-parameter", 46}, { {"storeOp must be a valid VkAttachmentStoreOp value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-synchronization2-06908", 51}, { {"If the synchronization2 feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 153}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription-synchronization2-06909", 51}, { {"If the synchronization2 feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-attachmentFeedbackLoopLayout-07309", 64}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-attachmentFeedbackLoopLayout-07310", 64}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-dynamicRenderingLocalRead-09544", 61}, { {"If the dynamicRenderingLocalRead feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 119}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-dynamicRenderingLocalRead-09545", 61}, { {"If the dynamicRenderingLocalRead feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-finalLayout-00843", 47}, { {"finalLayout must not be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-finalLayout-parameter", 51}, { {"finalLayout must be a valid VkImageLayout value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-11773", 41}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR, flags must not include VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-11774", 41}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveSrgbFormatSupportsTransferFunctionControl must be VK_TRUE", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-11775", 41}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, maintenance10 must be enabled", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-11776", 41}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, format must use sRGB encoding", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-11777", 41}, { {"If flags includes VK_ATTACHMENT_DESCRIPTION_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_ATTACHMENT_DESCRIPTION_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, samples must be VK_SAMPLE_COUNT_1_BIT", 191}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-flags-parameter", 45}, { {"flags must be a valid combination of VkAttachmentDescriptionFlagBits values", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03280", 42}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 154}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03281", 42}, { {"If format is a depth/stencil format, initialLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 103}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03282", 42}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 152}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03283", 42}, { {"If format is a depth/stencil format, finalLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03286", 42}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 226}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03287", 42}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 224}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03290", 42}, { {"If format is a depth/stencil format which includes only the depth component, initialLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 190}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-03291", 42}, { {"If format is a depth/stencil format which includes only the depth component, finalLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06247", 42}, { {"If the pNext chain does not include a VkAttachmentDescriptionStencilLayout structure and format only includes a stencil component, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 240}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06248", 42}, { {"If the pNext chain does not include a VkAttachmentDescriptionStencilLayout structure and format only includes a stencil component, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 238}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06249", 42}, { {"If format is a depth/stencil format which includes both depth and stencil components, and initialLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, the pNext chain must include a VkAttachmentDescriptionStencilLayout structure", 269}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06250", 42}, { {"If format is a depth/stencil format which includes both depth and stencil components, and finalLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, the pNext chain must include a VkAttachmentDescriptionStencilLayout structure", 267}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06487", 42}, { {"If format is a color format, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 175}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06488", 42}, { {"If format is a color format, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06699", 42}, { {"If format includes a color or depth component and loadOp is VK_ATTACHMENT_LOAD_OP_LOAD, then initialLayout must not be VK_IMAGE_LAYOUT_UNDEFINED", 144}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06906", 42}, { {"If format is a depth/stencil format which includes both depth and stencil components, initialLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 199}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-06907", 42}, { {"If format is a depth/stencil format which includes both depth and stencil components, finalLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 197}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-09332", 42}, { {"If the externalFormatResolve feature is not enabled, format must not be VK_FORMAT_UNDEFINED", 91}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-09334", 42}, { {"If format is VK_FORMAT_UNDEFINED, there must be a VkExternalFormatANDROID structure in the pNext chain with a externalFormat that is not equal to 0", 147}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-format-parameter", 46}, { {"format must be a valid VkFormat value", 37}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-initialLayout-parameter", 53}, { {"initialLayout must be a valid VkImageLayout value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-loadOp-parameter", 46}, { {"loadOp must be a valid VkAttachmentLoadOp value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-pNext-06704", 41}, { {"If the pNext chain does not include a VkAttachmentDescriptionStencilLayout structure, format includes a stencil component, and stencilLoadOp is VK_ATTACHMENT_LOAD_OP_LOAD, then initialLayout must not be VK_IMAGE_LAYOUT_UNDEFINED", 228}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-pNext-06705", 41}, { {"If the pNext chain includes a VkAttachmentDescriptionStencilLayout structure, format includes a stencil component, and stencilLoadOp is VK_ATTACHMENT_LOAD_OP_LOAD, then VkAttachmentDescriptionStencilLayout::stencilInitialLayout must not be VK_IMAGE_LAYOUT_UNDEFINED", 265}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-pNext-pNext", 41}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAttachmentDescriptionStencilLayout, VkExternalFormatANDROID, or VkExternalFormatOHOS", 217}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_2", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-samples-08745", 43}, { {"samples must be a valid VkSampleCountFlagBits value that is set in imageCreateSampleCounts (as defined in Image Creation Limits) for the given format", 149}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-samples-parameter", 47}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-separateDepthStencilLayouts-03284", 63}, { {"If the separateDepthStencilLayouts feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 257}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-separateDepthStencilLayouts-03285", 63}, { {"If the separateDepthStencilLayouts feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 255}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-stencilLoadOp-parameter", 53}, { {"stencilLoadOp must be a valid VkAttachmentLoadOp value", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-stencilStoreOp-parameter", 54}, { {"stencilStoreOp must be a valid VkAttachmentStoreOp value", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-storeOp-parameter", 47}, { {"storeOp must be a valid VkAttachmentStoreOp value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-synchronization2-06908", 52}, { {"If the synchronization2 feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 153}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescription2-synchronization2-06909", 52}, { {"If the synchronization2 feature is not enabled, finalLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_STENCIL_LAYOUT", 69}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-stencilFinalLayout-03309", 66}, { {"stencilFinalLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 376}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-stencilFinalLayout-03310", 66}, { {"stencilFinalLayout must not be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-stencilFinalLayout-parameter", 70}, { {"stencilFinalLayout must be a valid VkImageLayout value", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-stencilInitialLayout-03308", 68}, { {"stencilInitialLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 378}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentDescriptionStencilLayout-stencilInitialLayout-parameter", 72}, { {"stencilInitialLayout must be a valid VkImageLayout value", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentFeedbackLoopInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_FEEDBACK_LOOP_INFO_EXT", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentFeedbackLoopInfoEXT-unifiedImageLayouts-10782", 62}, { {"If the unifiedImageLayouts feature is not enabled, feedbackLoopEnable must be VK_FALSE", 86}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-attachmentFeedbackLoopLayout-07311", 61}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, layout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 131}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-dynamicRenderingLocalRead-09546", 58}, { {"If the dynamicRenderingLocalRead feature is not enabled, layout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 112}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-layout-03077", 39}, { {"If attachment is not VK_ATTACHMENT_UNUSED, layout must not be VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, VK_IMAGE_LAYOUT_PREINITIALIZED, or VK_IMAGE_LAYOUT_PRESENT_SRC_KHR", 193}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-layout-parameter", 43}, { {"layout must be a valid VkImageLayout value", 42}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-separateDepthStencilLayouts-03313", 60}, { {"If the separateDepthStencilLayouts feature is not enabled, and attachment is not VK_ATTACHMENT_UNUSED, layout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 294}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference-synchronization2-06910", 49}, { {"If the synchronization2 feature is not enabled, layout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-attachmentFeedbackLoopLayout-07311", 62}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, layout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 131}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-dynamicRenderingLocalRead-09546", 59}, { {"If the dynamicRenderingLocalRead feature is not enabled, layout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 112}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-layout-03077", 40}, { {"If attachment is not VK_ATTACHMENT_UNUSED, layout must not be VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, VK_IMAGE_LAYOUT_PREINITIALIZED, or VK_IMAGE_LAYOUT_PRESENT_SRC_KHR", 193}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-layout-parameter", 44}, { {"layout must be a valid VkImageLayout value", 42}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-pNext-pNext", 39}, { {"pNext must be NULL or a pointer to a valid instance of VkAttachmentReferenceStencilLayout", 89}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_2", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-separateDepthStencilLayouts-03313", 61}, { {"If the separateDepthStencilLayouts feature is not enabled, and attachment is not VK_ATTACHMENT_UNUSED, layout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL,", 294}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReference2-synchronization2-06910", 50}, { {"If the synchronization2 feature is not enabled, layout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReferenceStencilLayout-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_STENCIL_LAYOUT", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReferenceStencilLayout-stencilLayout-03318", 59}, { {"stencilLayout must not be VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_PREINITIALIZED, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_PRESENT_SRC_KHR", 463}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentReferenceStencilLayout-stencilLayout-parameter", 63}, { {"stencilLayout must be a valid VkImageLayout value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentSampleCountInfoAMD-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_ATTACHMENT_SAMPLE_COUNT_INFO_AMD", 64}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkAttachmentSampleLocationsEXT-attachmentIndex-01531", 57}, { {"attachmentIndex must be less than the attachmentCount specified in VkRenderPassCreateInfo the render pass specified by VkRenderPassBeginInfo::renderPass was created with", 169}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkAttachmentSampleLocationsEXT-sampleLocationsInfo-parameter", 65}, { {"sampleLocationsInfo must be a valid VkSampleLocationsInfoEXT structure", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkBeginCustomResolveInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_BEGIN_CUSTOM_RESOLVE_INFO_EXT", 61}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-accelerationStructure-03620", 72}, { {"accelerationStructure must not already be backed by a memory object", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-accelerationStructure-parameter", 76}, { {"accelerationStructure must be a valid VkAccelerationStructureNV handle", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-commonparent", 57}, { {"Both of accelerationStructure, and memory must have been created, allocated, or retrieved from the same VkDevice", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-memory-03622", 57}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetAccelerationStructureMemoryRequirementsNV with accelerationStructure and type of VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_OBJECT_NV", 307}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-memory-parameter", 61}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-memoryOffset-03621", 63}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-memoryOffset-03623", 63}, { {"memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetAccelerationStructureMemoryRequirementsNV with accelerationStructure and type of VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_OBJECT_NV", 273}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-pDeviceIndices-parameter", 69}, { {"If deviceIndexCount is not 0, pDeviceIndices must be a valid pointer to an array of deviceIndexCount uint32_t values", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-pNext-pNext", 56}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_BIND_ACCELERATION_STRUCTURE_MEMORY_INFO_NV", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindAccelerationStructureMemoryInfoNV-size-03624", 55}, { {"The size member of the VkMemoryRequirements structure returned from a call to vkGetAccelerationStructureMemoryRequirementsNV with accelerationStructure and type of VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_OBJECT_NV must be less than or equal to the size of memory minus memoryOffset", 292}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryDeviceGroupInfo-deviceIndexCount-01606", 61}, { {"deviceIndexCount must either be zero or equal to the number of physical devices in the logical device", 101}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryDeviceGroupInfo-pDeviceIndices-01607", 59}, { {"All elements of pDeviceIndices must be valid device indices", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryDeviceGroupInfo-pDeviceIndices-parameter", 63}, { {"If deviceIndexCount is not 0, pDeviceIndices must be a valid pointer to an array of deviceIndexCount uint32_t values", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryDeviceGroupInfo-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_DEVICE_GROUP_INFO", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-None-01898", 38}, { {"If buffer was created with the VK_BUFFER_CREATE_PROTECTED_BIT bit set, the buffer must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-None-01899", 38}, { {"If buffer was created with the VK_BUFFER_CREATE_PROTECTED_BIT bit not set, the buffer must not be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 197}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-None-10739", 38}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer", 274}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-None-10741", 38}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer must be less than or equal to the size of memory minus memoryOffset", 289}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-apiVersion-07920", 44}, { {"If the VK_KHR_dedicated_allocation extension is not enabled, VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and buffer was not created with VkDedicatedAllocationBufferCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must not have been allocated dedicated for a specific buffer or image", 312}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-01030", 40}, { {"buffer must not have been created with any sparse memory binding flags", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-01038", 40}, { {"If buffer was created with VkDedicatedAllocationBufferCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must have been allocated with VkDedicatedAllocationMemoryAllocateInfoNV::buffer equal to a buffer handle created with identical creation parameters to buffer and memoryOffset must be zero", 300}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-01444", 40}, { {"If buffer requires a dedicated allocation (as reported by vkGetBufferMemoryRequirements2 in VkMemoryDedicatedRequirements::requiresDedicatedAllocation for buffer), memory must have been allocated with VkMemoryDedicatedAllocateInfo::buffer equal to buffer", 254}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-06408", 40}, { {"If buffer was created with VkBufferCollectionBufferCreateInfoFUCHSIA chained to VkBufferCreateInfo::pNext, memory must be allocated with a VkImportMemoryBufferCollectionFUCHSIA chained to VkMemoryAllocateInfo::pNext", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-07459", 40}, { {"buffer must not have been bound to a memory object", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-09201", 40}, { {"If the buffer was created with the VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 203}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-11408", 40}, { {"If the buffer was created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT or VK_BUFFER_USAGE_2_DESCRIPTOR_HEAP_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-buffer-parameter", 44}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-bufferDeviceAddress-03339", 53}, { {"If the VkPhysicalDeviceBufferDeviceAddressFeatures::bufferDeviceAddress feature is enabled and buffer was created with the VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT usage flag set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 267}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-bufferDeviceAddressCaptureReplay-09200", 66}, { {"If the VkPhysicalDeviceBufferDeviceAddressFeatures::bufferDeviceAddressCaptureReplay feature is enabled and buffer was created with the VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 297}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-commonparent", 40}, { {"Both of buffer, and memory must have been created, allocated, or retrieved from the same VkDevice", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-descriptorBufferCaptureReplay-08112", 63}, { {"If the buffer was created with the VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 188}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-01035", 40}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer", 202}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-01508", 40}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::buffer was not VK_NULL_HANDLE, then buffer must equal VkMemoryDedicatedAllocateInfo::buffer, and memoryOffset must be zero", 291}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-02726", 40}, { {"If the value of VkExportMemoryAllocateInfo::handleTypes used to allocate memory is not 0, it must include at least one of the handles set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 210}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-02985", 40}, { {"If memory was allocated by a memory import operation, that is not VkImportAndroidHardwareBufferInfoANDROID with a non-NULL buffer value, the external handle type of the imported memory must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 281}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-02986", 40}, { {"If memory was allocated with the VkImportAndroidHardwareBufferInfoANDROID memory import operation with a non-NULL buffer value, VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 291}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-10740", 40}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetBufferMemoryRequirements with buffer", 278}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-10742", 40}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetBufferMemoryRequirements with buffer must be less than or equal to the size of memory minus memoryOffset", 293}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-10925", 40}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, VkMemoryDedicatedAllocateInfo::image must have been VK_NULL_HANDLE", 200}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memory-parameter", 44}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-memoryOffset-01031", 46}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-pNext-01605", 39}, { {"If the pNext chain includes a VkBindBufferMemoryDeviceGroupInfo structure, all instances of memory specified by VkBindBufferMemoryDeviceGroupInfo::pDeviceIndices must have been allocated", 186}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-pNext-pNext", 39}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBindBufferMemoryDeviceGroupInfo or VkBindMemoryStatus", 186}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_INFO", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindBufferMemoryInfo-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-bindPoint-09786", 64}, { {"bindPoint must have been returned as part of a VkDataGraphPipelineSessionBindPointRequirementARM whose bindPointType member is VK_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_TYPE_MEMORY_ARM by a prior call to vkGetDataGraphPipelineSessionMemoryRequirementsARM for session", 266}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-bindPoint-parameter", 68}, { {"bindPoint must be a valid VkDataGraphPipelineSessionBindPointARM value", 70}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-commonparent", 61}, { {"Both of memory, and session must have been created, allocated, or retrieved from the same VkDevice", 98}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-memory-09788", 61}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetDataGraphPipelineSessionMemoryRequirementsARM with session", 224}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-memory-parameter", 65}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-memoryOffset-09787", 67}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-memoryOffset-09789", 67}, { {"memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetDataGraphPipelineSessionMemoryRequirementsARM with session", 190}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-objectIndex-09805", 66}, { {"objectIndex must be less than the value of numObjects returned by vkGetDataGraphPipelineSessionBindPointRequirementsARM for bindPoint", 133}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_BIND_DATA_GRAPH_PIPELINE_SESSION_MEMORY_INFO_ARM", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-session-09785", 62}, { {"session must not have been bound to a memory object for bindPoint", 65}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-session-09791", 62}, { {"If session was created with the VK_DATA_GRAPH_PIPELINE_SESSION_CREATE_PROTECTED_BIT_ARM bit set, the session must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 216}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-session-09792", 62}, { {"If session was created with the VK_DATA_GRAPH_PIPELINE_SESSION_CREATE_PROTECTED_BIT_ARM bit not set, the session must not be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 224}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-session-parameter", 66}, { {"session must be a valid VkDataGraphPipelineSessionARM handle", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDataGraphPipelineSessionMemoryInfoARM-size-09790", 59}, { {"The size member of the VkMemoryRequirements structure returned from a call to vkGetDataGraphPipelineSessionMemoryRequirementsARM with session must be less than or equal to the size of memory minus memoryOffset", 209}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-None-09495", 61}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-layout-09496", 63}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-layout-parameter", 67}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-pNext-pNext", 62}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_BUFFER_EMBEDDED_SAMPLERS_INFO_EXT", 81}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-sType-unique", 63}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-set-08070", 60}, { {"The VkDescriptorSetLayout at index set when layout was created must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT bit set", 170}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-set-08071", 60}, { {"set must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 109}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-stageFlags-parameter", 71}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorBufferEmbeddedSamplersInfoEXT-stageFlags-requiredbitmask", 77}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBindDescriptorSetsInfo-None-09495", 40}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-commonparent", 42}, { {"Both of layout, and the elements of pDescriptorSets that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 171}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-descriptorSetCount-arraylength", 60}, { {"descriptorSetCount must be greater than 0", 41}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-dynamicOffsetCount-00359", 54}, { {"dynamicOffsetCount must be equal to the total number of dynamic descriptors in pDescriptorSets", 94}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-firstSet-00360", 44}, { {"The sum of firstSet and descriptorSetCount must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 148}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-layout-09496", 42}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-layout-parameter", 46}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-00358", 51}, { {"Each element of pDescriptorSets that is not VK_NULL_HANDLE must have been allocated with a VkDescriptorSetLayout that matches (is the same as, or identically defined as) the VkDescriptorSetLayout at set n in layout, where n is the sum of firstSet and the index into pDescriptorSets", 281}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-01979", 51}, { {"For each dynamic uniform or storage buffer binding in pDescriptorSets, the sum of the effective offset and the range of the binding must be less than or equal to the size of the buffer", 184}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-04616", 51}, { {"Each element of pDescriptorSets must not have been allocated from a VkDescriptorPool with the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT flag set", 146}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-06563", 51}, { {"If the graphicsPipelineLibrary feature is not enabled, each element of pDescriptorSets must be a valid VkDescriptorSet", 118}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-06715", 51}, { {"For each dynamic uniform or storage buffer binding in pDescriptorSets, if the range was set with VK_WHOLE_SIZE then pDynamicOffsets which corresponds to the descriptor binding must be 0", 185}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-08010", 51}, { {"Each element of pDescriptorSets must have been allocated with a VkDescriptorSetLayout which was not created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 170}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-09914", 51}, { {"If any element of pDescriptorSets was allocated from a descriptor pool created with a VkDescriptorPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure specifying foreign data processing engines in its pNext chain, then the command pool from which commandBuffer was allocated must have been created with a VkCommandPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain specifying a superset of all the foreign data processing engines specified when creating the descriptor pools from which the elements of pDescriptorSets were allocated", 621}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-09915", 51}, { {"If none of the elements of pDescriptorSets were allocated from a descriptor pool created with a VkDescriptorPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure specifying foreign data processing engines in its pNext chain, then the command pool from which commandBuffer was allocated must not have been created with a VkCommandPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain", 467}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDescriptorSets-parameter", 55}, { {"pDescriptorSets must be a valid pointer to an array of descriptorSetCount valid VkDescriptorSet handles", 103}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDynamicOffsets-01971", 51}, { {"Each element of pDynamicOffsets which corresponds to a descriptor binding with type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC must be a multiple of VkPhysicalDeviceLimits::minUniformBufferOffsetAlignment", 203}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDynamicOffsets-01972", 51}, { {"Each element of pDynamicOffsets which corresponds to a descriptor binding with type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC must be a multiple of VkPhysicalDeviceLimits::minStorageBufferOffsetAlignment", 203}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pDynamicOffsets-parameter", 55}, { {"If dynamicOffsetCount is not 0, and pDynamicOffsets is not NULL, pDynamicOffsets must be a valid pointer to an array of dynamicOffsetCount or VK_NULL_HANDLE uint32_t values", 172}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-pNext-pNext", 41}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_SETS_INFO", 57}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-stageFlags-parameter", 50}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindDescriptorSetsInfo-stageFlags-requiredbitmask", 56}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkBindHeapInfoEXT-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkBindHeapInfoEXT-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_BIND_HEAP_INFO_EXT", 50}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-deviceIndexCount-01633", 60}, { {"At least one of deviceIndexCount and splitInstanceBindRegionCount must be zero", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-deviceIndexCount-01634", 60}, { {"deviceIndexCount must either be zero or equal to the number of physical devices in the logical device", 101}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-extent-01640", 50}, { {"The extent.width member of any element of pSplitInstanceBindRegions must either be a multiple of the sparse image block width of all non-metadata aspects of the image, or else extent.width + offset.x must equal the width of the image subresource", 245}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-extent-01641", 50}, { {"The extent.height member of any element of pSplitInstanceBindRegions must either be a multiple of the sparse image block height of all non-metadata aspects of the image, or else extent.height + offset.y must equal the height of the image subresource", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-offset-01638", 50}, { {"The offset.x member of any element of pSplitInstanceBindRegions must be a multiple of the sparse image block width (VkSparseImageFormatProperties::imageGranularity.width) of all non-metadata aspects of the image", 211}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-offset-01639", 50}, { {"The offset.y member of any element of pSplitInstanceBindRegions must be a multiple of the sparse image block height (VkSparseImageFormatProperties::imageGranularity.height) of all non-metadata aspects of the image", 213}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-pDeviceIndices-01635", 58}, { {"All elements of pDeviceIndices must be valid device indices", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-pDeviceIndices-parameter", 62}, { {"If deviceIndexCount is not 0, pDeviceIndices must be a valid pointer to an array of deviceIndexCount uint32_t values", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-pSplitInstanceBindRegions-01637", 69}, { {"Elements of pSplitInstanceBindRegions that correspond to the same instance of an image must not overlap", 103}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-pSplitInstanceBindRegions-parameter", 73}, { {"If splitInstanceBindRegionCount is not 0, pSplitInstanceBindRegions must be a valid pointer to an array of splitInstanceBindRegionCount VkRect2D structures", 155}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_DEVICE_GROUP_INFO", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryDeviceGroupInfo-splitInstanceBindRegionCount-01636", 72}, { {"splitInstanceBindRegionCount must either be zero or equal to the number of physical devices in the logical device squared", 121}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-None-01901", 37}, { {"If image was created with the VK_IMAGE_CREATE_PROTECTED_BIT bit set, the image must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 186}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-None-01902", 37}, { {"If image was created with the VK_IMAGE_CREATE_PROTECTED_BIT bit not set, the image must not be bound to a memory object created with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 192}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-apiVersion-07921", 43}, { {"If the VK_KHR_dedicated_allocation extension is not enabled, VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and image was not created with VkDedicatedAllocationImageCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must not have been allocated dedicated for a specific buffer or image", 310}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-commonparent", 39}, { {"Both of image, and memory that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 145}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-01045", 38}, { {"image must not have been created with any sparse memory binding flags", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-01050", 38}, { {"If image was created with VkDedicatedAllocationImageCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must have been created with VkDedicatedAllocationMemoryAllocateInfoNV::image equal to an image handle created with identical creation parameters to image and memoryOffset must be zero", 294}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-01445", 38}, { {"If image requires a dedicated allocation (as reported by vkGetImageMemoryRequirements2 in VkMemoryDedicatedRequirements::requiresDedicatedAllocation for image), memory must have been created with VkMemoryDedicatedAllocateInfo::image equal to image", 247}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-01630", 38}, { {"If image was created with a valid swapchain handle in VkImageSwapchainCreateInfoKHR::swapchain, then the pNext chain must include a VkBindImageMemorySwapchainInfoKHR structure containing the same swapchain handle", 212}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-07460", 38}, { {"image must not have been bound to a memory object", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-07736", 38}, { {"If image was created with the VK_IMAGE_CREATE_DISJOINT_BIT bit set, then the pNext chain must include a VkBindImagePlaneMemoryInfo structure", 140}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-08113", 38}, { {"If the image was created with the VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-09202", 38}, { {"If the image was created with the VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 199}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-image-parameter", 42}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-02628", 39}, { {"If the dedicatedAllocationImageAliasing feature is not enabled, and the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::image was not VK_NULL_HANDLE, then image must equal VkMemoryDedicatedAllocateInfo::image and memoryOffset must be zero", 352}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-02629", 39}, { {"If the dedicatedAllocationImageAliasing feature is enabled, and the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::image was not VK_NULL_HANDLE, then memoryOffset must be zero, and image must be either equal to VkMemoryDedicatedAllocateInfo::image or an image that was created using the same parameters in VkImageCreateInfo, with the exception that extent and arrayLayers may differ subject to the following restrictions: every dimension in the extent parameter of the image being bound must be equal to or smaller than the original image for which the allocation was created; and the arrayLayers parameter of the image being bound must be equal to or smaller than the original image for which the allocation was created", 835}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-02728", 39}, { {"If the value of VkExportMemoryAllocateInfo::handleTypes used to allocate memory is not 0, it must include at least one of the handles set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-02989", 39}, { {"If memory was created by a memory import operation, that is not VkImportAndroidHardwareBufferInfoANDROID with a non-NULL buffer value, the external handle type of the imported memory must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 277}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-02990", 39}, { {"If memory was created with the VkImportAndroidHardwareBufferInfoANDROID memory import operation with a non-NULL buffer value, VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 287}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memory-10926", 39}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, VkMemoryDedicatedAllocateInfo::buffer must have been VK_NULL_HANDLE", 201}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-memoryOffset-01046", 45}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01615", 38}, { {"If the pNext chain does not include a VkBindImagePlaneMemoryInfo structure, and memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, then memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with image", 389}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01616", 38}, { {"If the pNext chain does not include a VkBindImagePlaneMemoryInfo structure, and memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, then memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with image", 355}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01617", 38}, { {"If the pNext chain does not include a VkBindImagePlaneMemoryInfo structure, and memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, then the difference of the size of memory and memoryOffset must be greater than or equal to the size member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with the same image", 402}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01618", 38}, { {"If the pNext chain includes a VkBindImagePlaneMemoryInfo structure, image must have been created with the VK_IMAGE_CREATE_DISJOINT_BIT bit set", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01619", 38}, { {"If the pNext chain includes a VkBindImagePlaneMemoryInfo structure, memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with image and where VkBindImagePlaneMemoryInfo::planeAspect corresponds to the VkImagePlaneMemoryRequirementsInfo::planeAspect in the VkImageMemoryRequirementsInfo2 structure's pNext chain", 448}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01620", 38}, { {"If the pNext chain includes a VkBindImagePlaneMemoryInfo structure, memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with image and where VkBindImagePlaneMemoryInfo::planeAspect corresponds to the VkImagePlaneMemoryRequirementsInfo::planeAspect in the VkImageMemoryRequirementsInfo2 structure's pNext chain", 414}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01621", 38}, { {"If the pNext chain includes a VkBindImagePlaneMemoryInfo structure, the difference of the size of memory and memoryOffset must be greater than or equal to the size member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements2 with the same image and where VkBindImagePlaneMemoryInfo::planeAspect corresponds to the VkImagePlaneMemoryRequirementsInfo::planeAspect in the VkImageMemoryRequirementsInfo2 structure's pNext chain", 461}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01626", 38}, { {"If the pNext chain includes a VkBindImageMemoryDeviceGroupInfo structure, all instances of memory specified by VkBindImageMemoryDeviceGroupInfo::pDeviceIndices must have been allocated", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01627", 38}, { {"If the pNext chain includes a VkBindImageMemoryDeviceGroupInfo structure, and VkBindImageMemoryDeviceGroupInfo::splitInstanceBindRegionCount is not zero, then image must have been created with the VK_IMAGE_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT bit set", 252}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01628", 38}, { {"If the pNext chain includes a VkBindImageMemoryDeviceGroupInfo structure, all elements of VkBindImageMemoryDeviceGroupInfo::pSplitInstanceBindRegions must be valid rectangles contained within the dimensions of image", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01629", 38}, { {"If the pNext chain includes a VkBindImageMemoryDeviceGroupInfo structure, the union of the areas of all elements of VkBindImageMemoryDeviceGroupInfo::pSplitInstanceBindRegions that correspond to the same instance of image must cover the entire image", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01631", 38}, { {"If the pNext chain includes a VkBindImageMemorySwapchainInfoKHR structure, memory must be VK_NULL_HANDLE", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-01632", 38}, { {"If the pNext chain does not include a VkBindImageMemorySwapchainInfoKHR structure, memory must be a valid VkDeviceMemory handle", 127}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-12329", 38}, { {"If the pNext chain does not include a VkBindImagePlaneMemoryInfo structure, and memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, then memoryOffset must be an integer multiple of the alignment member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetImageMemoryRequirements2 with image", 359}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-12330", 38}, { {"If the pNext chain does not include a VkBindImagePlaneMemoryInfo structure, and memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, then size member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetImageMemoryRequirements2 with image must be less than or equal to the size of memory minus memoryOffset", 374}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBindImageMemoryDeviceGroupInfo, VkBindImageMemorySwapchainInfoKHR, VkBindImagePlaneMemoryInfo, or VkBindMemoryStatus", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO", 54}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemoryInfo-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemorySwapchainInfoKHR-imageIndex-01644", 55}, { {"imageIndex must be less than the number of images in swapchain", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemorySwapchainInfoKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_SWAPCHAIN_INFO_KHR", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemorySwapchainInfoKHR-swapchain-07756", 54}, { {"If the swapchain has been created with VK_SWAPCHAIN_CREATE_DEFERRED_MEMORY_ALLOCATION_BIT_KHR, imageIndex must be one that has previously been returned by vkAcquireNextImageKHR or vkAcquireNextImage2KHR", 202}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImageMemorySwapchainInfoKHR-swapchain-parameter", 58}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImagePlaneMemoryInfo-planeAspect-02283", 49}, { {"If the image's tiling is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, then planeAspect must be a single valid multi-planar aspect mask bit", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImagePlaneMemoryInfo-planeAspect-02284", 49}, { {"If the image's tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then planeAspect must be a single valid memory plane for the image (that is, aspectMask must specify a plane index that is less than the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with the image's format and VkImageDrmFormatModifierPropertiesEXT::drmFormatModifier)", 361}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImagePlaneMemoryInfo-planeAspect-parameter", 53}, { {"planeAspect must be a valid VkImageAspectFlagBits value", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindImagePlaneMemoryInfo-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_BIND_IMAGE_PLANE_MEMORY_INFO", 60}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-None-13055", 41}, { {"If the nullDescriptor feature is not enabled, addressRange.size must be greater than 0", 86}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressFlags-13100", 49}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressFlags-13101", 49}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressFlags-parameter", 53}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13051", 49}, { {"If addressRange.address is not 0, the buffer from which addressRange was queried must have been created with VK_BUFFER_USAGE_INDEX_BUFFER_BIT", 141}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13052", 49}, { {"addressRange.address must be a multiple of the size of the type indicated by indexType", 86}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13056", 49}, { {"If addressRange.size is 0, addressRange.address must be 0", 57}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13097", 49}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13098", 49}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13099", 49}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13122", 49}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13123", 49}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13124", 49}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-addressRange-13125", 49}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-indexType-13053", 46}, { {"indexType must not be VK_INDEX_TYPE_NONE_KHR", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-indexType-13054", 46}, { {"If indexType is VK_INDEX_TYPE_UINT8_KHR, the indexTypeUint8 feature must be enabled", 83}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-indexType-parameter", 50}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBuffer3InfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_BIND_INDEX_BUFFER_3_INFO_KHR", 60}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkBindIndexBufferIndirectCommandEXT-None-11117", 51}, { {"The buffer's usage flags from which the address was acquired must have the VK_BUFFER_USAGE_INDEX_BUFFER_BIT bit set", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandEXT-bufferAddress-11118", 60}, { {"The bufferAddress must be aligned to the VkIndexType of the indexType used", 74}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandEXT-bufferAddress-parameter", 64}, { {"bufferAddress must be a valid VkDeviceAddress value", 51}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandEXT-indexType-parameter", 60}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandNV-None-02946", 50}, { {"The buffer's usage flag from which the address was acquired must have the VK_BUFFER_USAGE_INDEX_BUFFER_BIT bit set", 114}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandNV-bufferAddress-02947", 59}, { {"The bufferAddress must be aligned to the indexType used", 55}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandNV-bufferAddress-parameter", 63}, { {"bufferAddress must be a valid VkDeviceAddress value", 51}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindIndexBufferIndirectCommandNV-indexType-parameter", 59}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindMemoryStatus-pResult-parameter", 41}, { {"pResult must be a valid pointer to a VkResult value", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindMemoryStatus-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_BIND_MEMORY_STATUS", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindPipelineIndirectCommandNV-None-09092", 47}, { {"The referenced pipeline must have been created with VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 95}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindPipelineIndirectCommandNV-None-09093", 47}, { {"The referenced pipeline must have been updated with vkCmdUpdatePipelineIndirectBufferNV", 87}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindPipelineIndirectCommandNV-None-09094", 47}, { {"The referenced pipeline's address must have been queried with vkGetPipelineIndirectDeviceAddressNV", 98}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindPipelineIndirectCommandNV-deviceGeneratedComputePipelines-09091", 74}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindPipelineIndirectCommandNV-pipelineAddress-parameter", 62}, { {"pipelineAddress must be a valid VkDeviceAddress value", 53}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindShaderGroupIndirectCommandNV-None-02944", 50}, { {"The current bound graphics pipeline, as well as the pipelines it may reference, must have been created with VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindShaderGroupIndirectCommandNV-index-02945", 51}, { {"The index must be within range of the accessible shader groups of the current bound graphics pipeline. See vkCmdBindPipelineShaderGroupNV for further details", 157}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindSparseInfo-commonparent", 34}, { {"Both of the elements of pSignalSemaphores, and the elements of pWaitSemaphores that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 198}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pBufferBinds-parameter", 44}, { {"If bufferBindCount is not 0, pBufferBinds must be a valid pointer to an array of bufferBindCount valid VkSparseBufferMemoryBindInfo structures", 142}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pImageBinds-parameter", 43}, { {"If imageBindCount is not 0, pImageBinds must be a valid pointer to an array of imageBindCount valid VkSparseImageMemoryBindInfo structures", 138}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pImageOpaqueBinds-parameter", 49}, { {"If imageOpaqueBindCount is not 0, pImageOpaqueBinds must be a valid pointer to an array of imageOpaqueBindCount valid VkSparseImageOpaqueMemoryBindInfo structures", 162}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pNext-03247", 33}, { {"If the pNext chain of this structure includes a VkTimelineSemaphoreSubmitInfo structure and any element of pWaitSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE then its waitSemaphoreValueCount member must equal waitSemaphoreCount", 257}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pNext-03248", 33}, { {"If the pNext chain of this structure includes a VkTimelineSemaphoreSubmitInfo structure and any element of pSignalSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE then its signalSemaphoreValueCount member must equal signalSemaphoreCount", 263}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pNext-09753", 33}, { {"If the pNext chain of this structure includes a VkFrameBoundaryTensorsARM structure then it must also include a VkFrameBoundaryEXT structure", 140}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pNext-pNext", 33}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceGroupBindSparseInfo, VkFrameBoundaryEXT, VkFrameBoundaryTensorsARM, or VkTimelineSemaphoreSubmitInfo", 239}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pSignalSemaphores-03249", 45}, { {"For each element of pSignalSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pSignalSemaphoreValues must have a value greater than the current value of the semaphore when the semaphore signal operation is executed", 295}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pSignalSemaphores-03251", 45}, { {"For each element of pSignalSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pSignalSemaphoreValues must have a value which does not differ from the current value of the semaphore or from the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 400}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pSignalSemaphores-parameter", 49}, { {"If signalSemaphoreCount is not 0, pSignalSemaphores must be a valid pointer to an array of signalSemaphoreCount valid VkSemaphore handles", 137}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pWaitSemaphores-03246", 43}, { {"If any element of pWaitSemaphores or pSignalSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE then the pNext chain must include a VkTimelineSemaphoreSubmitInfo structure", 195}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pWaitSemaphores-03250", 43}, { {"For each element of pWaitSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pWaitSemaphoreValues must have a value which does not differ from the current value of the semaphore or from the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 396}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-pWaitSemaphores-parameter", 47}, { {"If waitSemaphoreCount is not 0, pWaitSemaphores must be a valid pointer to an array of waitSemaphoreCount valid VkSemaphore handles", 131}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_BIND_SPARSE_INFO", 48}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindSparseInfo-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-commonparent", 43}, { {"Both of memory, and tensor must have been created, allocated, or retrieved from the same VkDevice", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-09714", 43}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetTensorMemoryRequirementsARM with tensor", 205}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-09806", 43}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfoTensorARM structure in its pNext chain, and VkMemoryDedicatedAllocateInfoTensorARM::tensor was not VK_NULL_HANDLE, then tensor must equal VkMemoryDedicatedAllocateInfoTensorARM::tensor, and memoryOffset must be zero", 318}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-09895", 43}, { {"If the value of VkExportMemoryAllocateInfo::handleTypes used to allocate memory is not 0, it must include at least one of the handles set in VkExternalMemoryTensorCreateInfoARM::handleTypes when tensor was created", 213}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-09896", 43}, { {"If memory was allocated by a memory import operation, that is not VkImportAndroidHardwareBufferInfoANDROID with a non-NULL buffer value, the external handle type of the imported memory must also have been set in VkExternalMemoryTensorCreateInfoARM::handleTypes when tensor was created", 284}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-09897", 43}, { {"If memory was allocated with the VkImportAndroidHardwareBufferInfoANDROID memory import operation with a non-NULL buffer value, VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must also have been set in VkExternalMemoryTensorCreateInfoARM::handleTypes when tensor was created", 294}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memory-parameter", 47}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memoryOffset-09713", 49}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-memoryOffset-09715", 49}, { {"memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetTensorMemoryRequirementsARM with tensor", 171}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_BIND_TENSOR_MEMORY_INFO_ARM", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-size-09716", 41}, { {"The size member of the VkMemoryRequirements structure returned from a call to vkGetTensorMemoryRequirementsARM with tensor must be less than or equal to the size of memory minus memoryOffset", 190}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09712", 43}, { {"tensor must not already be backed by a memory object", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09717", 43}, { {"If tensor requires a dedicated allocation (as reported by vkGetTensorMemoryRequirementsARM in VkMemoryDedicatedRequirements::requiresDedicatedAllocation for tensor), memory must have been created with VkMemoryDedicatedAllocateInfoTensorARM::tensor equal to tensor", 263}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09718", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_PROTECTED_BIT_ARM bit set, the tensor must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 193}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09719", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_PROTECTED_BIT_ARM bit not set, the tensor must not be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 201}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09943", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-09944", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 199}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-11406", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_ARM bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-11407", 43}, { {"If tensor was created with the VK_TENSOR_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_ARM bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 197}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTensorMemoryInfoARM-tensor-parameter", 47}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressFlags-13100", 61}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressFlags-13101", 61}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressFlags-parameter", 65}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13097", 61}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13098", 61}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13099", 61}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13122", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13123", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13124", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-addressRange-13125", 61}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindTransformFeedbackBuffer2InfoEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_BIND_TRANSFORM_FEEDBACK_BUFFER_2_INFO_EXT", 73}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressFlags-13100", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressFlags-13101", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressFlags-parameter", 54}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13074", 50}, { {"If addressRange.address is not 0, the buffer from which addressRange was queried must have been created with the VK_BUFFER_USAGE_VERTEX_BUFFER_BIT usage flag set", 161}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13075", 50}, { {"If addressRange.size is 0, addressRange.address must be 0", 57}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13097", 50}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13098", 50}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13099", 50}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13122", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13123", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13124", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-addressRange-13125", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_BIND_VERTEX_BUFFER_3_INFO_KHR", 61}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-setStride-13126", 47}, { {"If setStride is VK_TRUE, addressRange.stride must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindingStride", 126}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-setStride-13127", 47}, { {"If setStride is VK_FALSE, addressRange.stride must be 0", 55}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBuffer3InfoKHR-size-13072", 42}, { {"If the nullDescriptor feature is not enabled, addressRange.size must not be 0", 77}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkBindVertexBufferIndirectCommandEXT-None-11120", 52}, { {"The buffer's usage flag from which the address was acquired must have the VK_BUFFER_USAGE_VERTEX_BUFFER_BIT bit set", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindVertexBufferIndirectCommandEXT-bufferAddress-parameter", 65}, { {"bufferAddress must be a valid VkDeviceAddress value", 51}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindVertexBufferIndirectCommandNV-None-02949", 51}, { {"The buffer's usage flag from which the address was acquired must have the VK_BUFFER_USAGE_VERTEX_BUFFER_BIT bit set", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindVertexBufferIndirectCommandNV-bufferAddress-parameter", 64}, { {"bufferAddress must be a valid VkDeviceAddress value", 51}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkBindVideoSessionMemoryInfoKHR-memory-parameter", 53}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkBindVideoSessionMemoryInfoKHR-memoryOffset-07201", 55}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkBindVideoSessionMemoryInfoKHR-memorySize-07202", 53}, { {"memorySize must be less than or equal to the size of memory minus memoryOffset", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkBindVideoSessionMemoryInfoKHR-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkBindVideoSessionMemoryInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_BIND_VIDEO_SESSION_MEMORY_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkBlitImageCubicWeightsInfoQCOM-cubicWeights-parameter", 59}, { {"cubicWeights must be a valid VkCubicFilterWeightsQCOM value", 59}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageCubicWeightsInfoQCOM-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_BLIT_IMAGE_CUBIC_WEIGHTS_INFO_QCOM", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-aspectMask-00241", 38}, { {"For each element of pRegions, srcSubresource.aspectMask must specify aspects present in srcImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-aspectMask-00242", 38}, { {"For each element of pRegions, dstSubresource.aspectMask must specify aspects present in dstImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-commonparent", 34}, { {"Both of dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-00224", 36}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-00225", 36}, { {"If dstImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-00234", 36}, { {"dstImage must have been created with a samples value of VK_SAMPLE_COUNT_1_BIT", 77}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-00250", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffsets[0].y must be 0 and dstOffsets[1].y must be 1", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-00252", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffsets[0].z must be 0 and dstOffsets[1].z must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-02000", 36}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_BLIT_DST_BIT", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-02545", 36}, { {"dstImage and srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-06422", 36}, { {"dstImage must not use a format that requires a sampler Y'CBCR conversion", 72}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImage-parameter", 40}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImageLayout-00226", 42}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImageLayout-01399", 42}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstImageLayout-parameter", 46}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstOffset-00248", 37}, { {"For each element of pRegions, dstOffsets[0].x and dstOffsets[1].x must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstOffset-00249", 37}, { {"For each element of pRegions, dstOffsets[0].y and dstOffsets[1].y must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 186}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstOffset-00251", 37}, { {"For each element of pRegions, dstOffsets[0].z and dstOffsets[1].z must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstSubresource-01706", 42}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-dstSubresource-01708", 42}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-filter-00237", 34}, { {"If filter is VK_FILTER_CUBIC_EXT, srcImage must be of type VK_IMAGE_TYPE_2D", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-filter-02001", 34}, { {"If filter is VK_FILTER_LINEAR, then the format features of srcImage must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-filter-02002", 34}, { {"If filter is VK_FILTER_CUBIC_EXT, then the format features of srcImage must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-filter-09204", 34}, { {"If filter is VK_FILTER_CUBIC_EXT and if the selectableCubicWeights feature is not enabled then the cubic weights must be VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 161}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-filter-parameter", 38}, { {"filter must be a valid VkFilter value", 37}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-maintenance8-10207", 40}, { {"If the maintenance8 feature is enabled and srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer must be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 231}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-maintenance8-10208", 40}, { {"If the maintenance8 feature is enabled and dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 231}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-maintenance8-10579", 40}, { {"If the maintenance8 feature is enabled, dstImage is VK_IMAGE_TYPE_3D, and srcImage is not of type VK_IMAGE_TYPE_3D, then for each element of pRegions, the absolute difference of the z member of each member of dstOffsets must equal srcSubresource.layerCount", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-maintenance8-10580", 40}, { {"If the maintenance8 feature is enabled, srcImage is VK_IMAGE_TYPE_3D, and dstImage is not of type VK_IMAGE_TYPE_3D, then for each element of pRegions, the absolute difference of the z member of each member of srcOffsets must equal dstSubresource.layerCount", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-pNext-pNext", 33}, { {"pNext must be NULL or a pointer to a valid instance of VkBlitImageCubicWeightsInfoQCOM", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-pRegions-00217", 36}, { {"The union of all destination regions, specified by the elements of pRegions, must not overlap in memory with any texel that may be sampled during the blit operation", 164}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-pRegions-04561", 36}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage and dstImage must not be block-compressed images", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-pRegions-parameter", 40}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageBlit2 structures", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-regionCount-arraylength", 45}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_BLIT_IMAGE_INFO_2", 49}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00219", 36}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00220", 36}, { {"If srcImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00229", 36}, { {"If either of srcImage or dstImage was created with a signed integer VkFormat, the other must also have been created with a signed integer VkFormat", 146}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00230", 36}, { {"If either of srcImage or dstImage was created with an unsigned integer color VkFormat, the other must also have been created with an unsigned integer color VkFormat", 164}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00231", 36}, { {"If either of srcImage or dstImage was created with a depth/stencil format, the other must have exactly the same format", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00232", 36}, { {"If srcImage was created with a depth/stencil format, filter must be VK_FILTER_NEAREST", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00233", 36}, { {"srcImage must have been created with a samples value of VK_SAMPLE_COUNT_1_BIT", 77}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00240", 36}, { {"If the maintenance8 feature is not enabled and either srcImage or dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer and dstSubresource.baseArrayLayer must each be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 293}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00245", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffsets[0].y must be 0 and srcOffsets[1].y must be 1", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-00247", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffsets[0].z must be 0 and srcOffsets[1].z must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-01999", 36}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_BLIT_SRC_BIT", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-06421", 36}, { {"srcImage must not use a format that requires a sampler Y'CBCR conversion", 72}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-09459", 36}, { {"If srcImage and dstImage are the same, and an element of pRegions contains the srcSubresource and dstSubresource with matching mipLevel and overlapping array layers, then the srcImageLayout and dstImageLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR", 278}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImage-parameter", 40}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImageLayout-00221", 42}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImageLayout-01398", 42}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcImageLayout-parameter", 46}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcOffset-00243", 37}, { {"For each element of pRegions, srcOffsets[0].x and srcOffsets[1].x must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcOffset-00244", 37}, { {"For each element of pRegions, srcOffsets[0].y and srcOffsets[1].y must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 186}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcOffset-00246", 37}, { {"For each element of pRegions, srcOffsets[0].z and srcOffsets[1].z must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcSubresource-01705", 42}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2-srcSubresource-01707", 42}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2KHR-pRegions-06207", 39}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage must be of type VK_IMAGE_TYPE_2D", 133}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBlitImageInfo2KHR-pRegions-06208", 39}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage must not have a multi-planar format", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferCaptureDescriptorDataInfoEXT-buffer-08075", 54}, { {"buffer must have been created with VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkBufferCreateInfo::flags", 125}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBufferCaptureDescriptorDataInfoEXT-buffer-parameter", 58}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBufferCaptureDescriptorDataInfoEXT-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBufferCaptureDescriptorDataInfoEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT", 71}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkBufferCollectionBufferCreateInfoFUCHSIA-collection-parameter", 67}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionBufferCreateInfoFUCHSIA-index-06388", 58}, { {"index must be less than VkBufferCollectionPropertiesFUCHSIA::bufferCount", 72}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionBufferCreateInfoFUCHSIA-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COLLECTION_BUFFER_CREATE_INFO_FUCHSIA", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionConstraintsInfoFUCHSIA-pNext-pNext", 57}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionConstraintsInfoFUCHSIA-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CONSTRAINTS_INFO_FUCHSIA", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionCreateInfoFUCHSIA-collectionToken-06393", 62}, { {"collectionToken must be a valid zx_handle_t to a Zircon channel allocated from Sysmem (fuchsia.sysmem.Allocator/AllocateSharedCollection) with ZX_DEFAULT_CHANNEL_RIGHTS rights", 175}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionCreateInfoFUCHSIA-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionCreateInfoFUCHSIA-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CREATE_INFO_FUCHSIA", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionImageCreateInfoFUCHSIA-collection-parameter", 66}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionImageCreateInfoFUCHSIA-index-06391", 57}, { {"index must be less than VkBufferCollectionPropertiesFUCHSIA::bufferCount", 72}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionImageCreateInfoFUCHSIA-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COLLECTION_IMAGE_CREATE_INFO_FUCHSIA", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionPropertiesFUCHSIA-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCollectionPropertiesFUCHSIA-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COLLECTION_PROPERTIES_FUCHSIA", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-bufferCollectionConstraints-parameter", 73}, { {"bufferCollectionConstraints must be a valid VkBufferCollectionConstraintsInfoFUCHSIA structure", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-createInfo-parameter", 56}, { {"createInfo must be a valid VkBufferCreateInfo structure", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-requiredFormatFeatures-06404", 64}, { {"The requiredFormatFeatures bitmask of VkFormatFeatureFlagBits must be chosen from among the buffer compatible format features listed in buffer compatible format features", 169}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-requiredFormatFeatures-parameter", 68}, { {"requiredFormatFeatures must be a valid combination of VkFormatFeatureFlagBits values", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferConstraintsInfoFUCHSIA-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_CONSTRAINTS_INFO_FUCHSIA", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCopy-size-01988", 28}, { {"The size must be greater than 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferCopy2-pNext-pNext", 30}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferCopy2-sType-sType", 30}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_COPY_2", 45}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferCopy2-size-01988", 29}, { {"The size must be greater than 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferCreateInfo-None-01888", 34}, { {"If any of the bits VK_BUFFER_CREATE_SPARSE_BINDING_BIT, VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT, or VK_BUFFER_CREATE_SPARSE_ALIASED_BIT are set, VK_BUFFER_CREATE_PROTECTED_BIT must not also be set", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-None-09499", 34}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must be a valid combination of VkBufferUsageFlagBits values", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-None-09500", 34}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must not be 0", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-deviceAddress-02604", 43}, { {"If VkBufferDeviceAddressCreateInfoEXT::deviceAddress is not zero, flags must include VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT", 135}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-00915", 35}, { {"If the sparseBinding feature is not enabled, flags must not contain VK_BUFFER_CREATE_SPARSE_BINDING_BIT", 103}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-00916", 35}, { {"If the sparseResidencyBuffer feature is not enabled, flags must not contain VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT", 113}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-00917", 35}, { {"If the sparseResidencyAliased feature is not enabled, flags must not contain VK_BUFFER_CREATE_SPARSE_ALIASED_BIT", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-00918", 35}, { {"If flags contains VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT or VK_BUFFER_CREATE_SPARSE_ALIASED_BIT, it must also contain VK_BUFFER_CREATE_SPARSE_BINDING_BIT", 152}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-01887", 35}, { {"If the protectedMemory feature is not enabled, flags must not contain VK_BUFFER_CREATE_PROTECTED_BIT", 100}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-03338", 35}, { {"If flags includes VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT, the VkPhysicalDeviceBufferDeviceAddressFeaturesEXT::bufferDeviceAddressCaptureReplay feature or the bufferDeviceAddressCaptureReplay feature must be enabled", 226}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-08099", 35}, { {"If flags includes VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay feature must be enabled", 134}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-08325", 35}, { {"If flags includes VK_BUFFER_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then videoMaintenance1 must be enabled", 108}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-09641", 35}, { {"If flags includes VK_BUFFER_CREATE_PROTECTED_BIT, then the effective usage flags must not contain bits other than VK_BUFFER_USAGE_TRANSFER_SRC_BIT VK_BUFFER_USAGE_TRANSFER_DST_BIT VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT VK_BUFFER_USAGE_STORAGE_BUFFER_BIT VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT VK_BUFFER_USAGE_2_VIDEO_DECODE_SRC_BIT_KHR VK_BUFFER_USAGE_2_VIDEO_ENCODE_DST_BIT_KHR VK_BUFFER_USAGE_2_DESCRIPTOR_HEAP_BIT_EXT", 501}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-11277", 35}, { {"If the protectedDescriptorHeaps property is not supported and the effective usage flags include the VK_BUFFER_USAGE_2_DESCRIPTOR_HEAP_BIT_EXT flag, flags must not include the VK_BUFFER_CREATE_PROTECTED_BIT flag", 210}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-11279", 35}, { {"If the sparseDescriptorHeaps property is not supported and the effective usage flags include the VK_BUFFER_USAGE_2_DESCRIPTOR_HEAP_BIT_EXT flag, flags must not include any of the VK_BUFFER_CREATE_SPARSE_BINDING_BIT, VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT, or VK_BUFFER_CREATE_SPARSE_ALIASED_BIT flags", 299}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-flags-parameter", 39}, { {"flags must be a valid combination of VkBufferCreateFlagBits values", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-maintenance11-13353", 43}, { {"If the maintenance11 feature is enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 0", 136}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-opaqueCaptureAddress-03337", 50}, { {"If VkBufferOpaqueCaptureAddressCreateInfo::opaqueCaptureAddress is not zero, flags must include VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT", 146}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-00920", 35}, { {"If the pNext chain includes a VkExternalMemoryBufferCreateInfo structure, its handleTypes member must only contain bits that are also in VkExternalBufferProperties::externalMemoryProperties.compatibleHandleTypes, as returned by vkGetPhysicalDeviceExternalBufferProperties with pExternalBufferInfo->handleType equal to any one of the handle types specified in VkExternalMemoryBufferCreateInfo::handleTypes", 404}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-01571", 35}, { {"If the pNext chain includes a VkDedicatedAllocationBufferCreateInfoNV structure, and the dedicatedAllocation member of the chained structure is VK_TRUE, then flags must not include VK_BUFFER_CREATE_SPARSE_BINDING_BIT, VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT, or VK_BUFFER_CREATE_SPARSE_ALIASED_BIT", 295}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-08100", 35}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-10249", 35}, { {"If the pNext chain includes a VkVideoProfileListInfoKHR structure and for any element of its pProfiles member videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the videoEncodeAV1 feature must be enabled", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-10783", 35}, { {"If the pNext chain includes a VkVideoProfileListInfoKHR structure and for any element of its pProfiles member videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the videoDecodeVP9 feature must be enabled", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-10919", 35}, { {"If the pNext chain includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be enabled", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-pNext-pNext", 35}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBufferCollectionBufferCreateInfoFUCHSIA, VkBufferDeviceAddressCreateInfoEXT, VkBufferOpaqueCaptureAddressCreateInfo, VkBufferUsageFlags2CreateInfo, VkDedicatedAllocationBufferCreateInfoNV, VkExternalMemoryBufferCreateInfo, VkOpaqueCaptureDescriptorDataCreateInfoEXT, or VkVideoProfileListInfoKHR", 428}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sType-unique", 36}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sharingMode-00913", 41}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, pQueueFamilyIndices must be a valid pointer to an array of queueFamilyIndexCount uint32_t values", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sharingMode-00914", 41}, { {"If the maintenance11 feature is not enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 1", 140}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sharingMode-01419", 41}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, each element of pQueueFamilyIndices must be unique and must be less than pQueueFamilyPropertyCount returned by either vkGetPhysicalDeviceQueueFamilyProperties2 or vkGetPhysicalDeviceQueueFamilyProperties for the physicalDevice that was used to create device", 303}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-sharingMode-parameter", 45}, { {"sharingMode must be a valid VkSharingMode value", 47}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-size-00912", 34}, { {"size must be greater than 0", 27}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-size-06409", 34}, { {"size must be less than or equal to VkPhysicalDeviceMaintenance4Properties::maxBufferSize", 88}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-tileMemoryHeap-10762", 44}, { {"If the tileMemoryHeap feature is not enabled, usage must not include VK_BUFFER_USAGE_TILE_MEMORY_BIT_QCOM", 105}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-04813", 35}, { {"If usage includes VK_BUFFER_USAGE_VIDEO_DECODE_SRC_BIT_KHR or VK_BUFFER_USAGE_VIDEO_DECODE_DST_BIT_KHR, and flags does not include VK_BUFFER_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then the pNext chain must include a VkVideoProfileListInfoKHR structure with profileCount greater than 0 and pProfiles including at least one VkVideoProfileInfoKHR structure with a videoCodecOperation member specifying a decode operation", 420}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-04814", 35}, { {"If usage includes VK_BUFFER_USAGE_VIDEO_ENCODE_SRC_BIT_KHR or VK_BUFFER_USAGE_VIDEO_ENCODE_DST_BIT_KHR, and flags does not include VK_BUFFER_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then the pNext chain must include a VkVideoProfileListInfoKHR structure with profileCount greater than 0 and pProfiles including at least one VkVideoProfileInfoKHR structure with a videoCodecOperation member specifying an encode operation", 421}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-08097", 35}, { {"If usage includes VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT, creating this VkBuffer must not cause the total required space for all currently valid buffers using this flag on the device to exceed VkPhysicalDeviceDescriptorBufferPropertiesEXT::samplerDescriptorBufferAddressSpaceSize or VkPhysicalDeviceDescriptorBufferPropertiesEXT::descriptorBufferAddressSpaceSize", 373}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-08098", 35}, { {"If usage includes VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT, creating this VkBuffer must not cause the total required space for all currently valid buffers using this flag on the device to exceed VkPhysicalDeviceDescriptorBufferPropertiesEXT::resourceDescriptorBufferAddressSpaceSize or VkPhysicalDeviceDescriptorBufferPropertiesEXT::descriptorBufferAddressSpaceSize", 375}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-08101", 35}, { {"If usage includes VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT, the descriptorBufferPushDescriptors feature must be enabled", 137}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-08102", 35}, { {"If usage includes VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT VkPhysicalDeviceDescriptorBufferPropertiesEXT::bufferlessPushDescriptors must be VK_FALSE", 166}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-08103", 35}, { {"If usage includes VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT, usage must contain at least one of VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT or VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT", 216}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-10763", 35}, { {"If usage includes VK_BUFFER_USAGE_TILE_MEMORY_BIT_QCOM, then flags must not contain any of the VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT, VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, VK_BUFFER_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, VK_BUFFER_CREATE_SPARSE_BINDING_BIT, VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT, VK_BUFFER_CREATE_SPARSE_ALIASED_BIT, or VK_BUFFER_CREATE_PROTECTED_BIT flags", 410}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferCreateInfo-usage-10764", 35}, { {"If usage includes VK_BUFFER_USAGE_TILE_MEMORY_BIT_QCOM, then only the following usages may be set: VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT VK_BUFFER_USAGE_STORAGE_BUFFER_BIT VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT and if VkPhysicalDeviceTileMemoryHeapPropertiesQCOM::tileBufferTransfers is VK_TRUE then additionally VK_BUFFER_USAGE_TRANSFER_SRC_BIT or VK_BUFFER_USAGE_TRANSFER_DST_BIT", 463}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressCreateInfoEXT-deviceAddress-parameter", 63}, { {"If deviceAddress is not 0, deviceAddress must be a valid VkDeviceAddress value", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressCreateInfoEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_CREATE_INFO_EXT", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressInfo-buffer-02601", 43}, { {"buffer must have been created with the VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT usage flag set", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressInfo-buffer-parameter", 47}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressInfo-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferDeviceAddressInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_INFO", 58}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferImageCopy-aspectMask-09103", 39}, { {"The aspectMask member of imageSubresource must only have a single bit set", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-bufferImageHeight-09102", 46}, { {"bufferImageHeight must be 0, or greater than or equal to the height member of imageExtent", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-bufferRowLength-09101", 44}, { {"bufferRowLength must be 0, or greater than or equal to the width member of imageExtent", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-imageExtent-06659", 40}, { {"imageExtent.width must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-imageExtent-06660", 40}, { {"imageExtent.height must not be 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-imageExtent-06661", 40}, { {"imageExtent.depth must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy-imageSubresource-parameter", 49}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-aspectMask-09103", 40}, { {"The aspectMask member of imageSubresource must only have a single bit set", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-bufferImageHeight-09102", 47}, { {"bufferImageHeight must be 0, or greater than or equal to the height member of imageExtent", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-bufferRowLength-09101", 45}, { {"bufferRowLength must be 0, or greater than or equal to the width member of imageExtent", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-imageExtent-06659", 41}, { {"imageExtent.width must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-imageExtent-06660", 41}, { {"imageExtent.height must not be 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-imageExtent-06661", 41}, { {"imageExtent.depth must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-imageSubresource-parameter", 50}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-pNext-pNext", 35}, { {"pNext must be NULL or a pointer to a valid instance of VkCopyCommandTransformInfoQCOM", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_IMAGE_COPY_2", 51}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferImageCopy2-sType-unique", 36}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkBufferMemoryBarrier-None-09049", 37}, { {"If the synchronization2 feature is not enabled, and buffer was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, at least one of srcQueueFamilyIndex and dstQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED", 213}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-None-09050", 37}, { {"If the synchronization2 feature is not enabled, and buffer was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED or VK_QUEUE_FAMILY_EXTERNAL", 201}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-None-09051", 37}, { {"If the synchronization2 feature is not enabled, and buffer was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED or VK_QUEUE_FAMILY_EXTERNAL", 201}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-None-09097", 37}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-None-09098", 37}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-buffer-01931", 39}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-buffer-09095", 39}, { {"If buffer was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 241}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-buffer-09096", 39}, { {"If buffer was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 241}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-buffer-parameter", 43}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-dstQueueFamilyIndex-09100", 52}, { {"If the VK_EXT_queue_family_foreign extension is not enabled dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-offset-01187", 39}, { {"offset must be less than the size of buffer", 43}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-pNext-pNext", 38}, { {"pNext must be NULL or a pointer to a valid instance of VkExternalMemoryAcquireUnmodifiedEXT", 91}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-size-01188", 37}, { {"If size is not equal to VK_WHOLE_SIZE, size must be greater than 0", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-size-01189", 37}, { {"If size is not equal to VK_WHOLE_SIZE, size must be less than or equal to than the size of buffer minus offset", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier-srcQueueFamilyIndex-09099", 52}, { {"If the VK_EXT_queue_family_foreign extension is not enabled srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-None-09097", 38}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-None-09098", 38}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-buffer-01931", 40}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-buffer-09095", 40}, { {"If buffer was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 241}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-buffer-09096", 40}, { {"If buffer was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 241}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-buffer-parameter", 44}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03900", 47}, { {"If dstAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03901", 47}, { {"If dstAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03902", 47}, { {"If dstAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03903", 47}, { {"If dstAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03904", 47}, { {"If dstAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03905", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03906", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03907", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03909", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03910", 47}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03911", 47}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03912", 47}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03913", 47}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03914", 47}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03915", 47}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03916", 47}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03917", 47}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03918", 47}, { {"If dstAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03919", 47}, { {"If dstAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03920", 47}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03922", 47}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03923", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03924", 47}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03925", 47}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03926", 47}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03927", 47}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-03928", 47}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04747", 47}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04858", 47}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04859", 47}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04860", 47}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04861", 47}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-04994", 47}, { {"If dstAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, dstStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-06256", 47}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07272", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07454", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07455", 47}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07456", 47}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07457", 47}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-07458", 47}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-08118", 47}, { {"If dstAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-10670", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-10671", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-11294", 47}, { {"If dstAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-11771", 47}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-11772", 47}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstAccessMask-parameter", 51}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstQueueFamilyIndex-09100", 53}, { {"If the VK_EXT_queue_family_foreign extension is not enabled dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03929", 46}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03930", 46}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03931", 46}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03932", 46}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03933", 46}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03934", 46}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-03935", 46}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-04957", 46}, { {"If the subpassShading feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-04995", 46}, { {"If the invocationMask feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-07316", 46}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-07946", 46}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-10751", 46}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-10752", 46}, { {"If the rayTracingMaintenance1 feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-11541", 46}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-dstStageMask-parameter", 50}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-offset-01187", 40}, { {"offset must be less than the size of buffer", 43}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-pNext-pNext", 39}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExternalMemoryAcquireUnmodifiedEXT or VkMemoryBarrierAccessFlags3KHR", 201}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2", 55}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-size-01188", 38}, { {"If size is not equal to VK_WHOLE_SIZE, size must be greater than 0", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-size-01189", 38}, { {"If size is not equal to VK_WHOLE_SIZE, size must be less than or equal to than the size of buffer minus offset", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03900", 47}, { {"If srcAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03901", 47}, { {"If srcAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03902", 47}, { {"If srcAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03903", 47}, { {"If srcAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03904", 47}, { {"If srcAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03905", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03906", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03907", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03909", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03910", 47}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03911", 47}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03912", 47}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03913", 47}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03914", 47}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03915", 47}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03916", 47}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03917", 47}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03918", 47}, { {"If srcAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03919", 47}, { {"If srcAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03920", 47}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03922", 47}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03923", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03924", 47}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03925", 47}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03926", 47}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03927", 47}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-03928", 47}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04747", 47}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04858", 47}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04859", 47}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04860", 47}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04861", 47}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-04994", 47}, { {"If srcAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, srcStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-06256", 47}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07272", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07454", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07455", 47}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07456", 47}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07457", 47}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-07458", 47}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-08118", 47}, { {"If srcAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-10670", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-10671", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-11294", 47}, { {"If srcAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-11771", 47}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-11772", 47}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcAccessMask-parameter", 51}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcQueueFamilyIndex-09099", 53}, { {"If the VK_EXT_queue_family_foreign extension is not enabled srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03851", 46}, { {"If either srcStageMask or dstStageMask includes VK_PIPELINE_STAGE_2_HOST_BIT, srcQueueFamilyIndex and dstQueueFamilyIndex must be equal", 135}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03929", 46}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03930", 46}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03931", 46}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03932", 46}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03933", 46}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03934", 46}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-03935", 46}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-04957", 46}, { {"If the subpassShading feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-04995", 46}, { {"If the invocationMask feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-07316", 46}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-07946", 46}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-10751", 46}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-10752", 46}, { {"If the rayTracingMaintenance1 feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-11541", 46}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryBarrier2-srcStageMask-parameter", 50}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkBufferMemoryRequirementsInfo2-buffer-parameter", 53}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferMemoryRequirementsInfo2-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferMemoryRequirementsInfo2-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_MEMORY_REQUIREMENTS_INFO_2", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferOpaqueCaptureAddressCreateInfo-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_OPAQUE_CAPTURE_ADDRESS_CREATE_INFO", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferUsageFlags2CreateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_USAGE_FLAGS_2_CREATE_INFO", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferUsageFlags2CreateInfo-usage-parameter", 50}, { {"usage must be a valid combination of VkBufferUsageFlagBits2 values", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferUsageFlags2CreateInfo-usage-requiredbitmask", 56}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-None-12278", 38}, { {"If Vulkan 1.3 is not supported and the ycbcr2plane444Formats feature is not enabled, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 286}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-buffer-00932", 40}, { {"buffer must have been created with at least one of the VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT or VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT usage flags set", 155}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-buffer-00935", 40}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-buffer-02750", 40}, { {"If buffer was created with the VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT usage flag set, offset must be a multiple of the effective alignment requirement of format for VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER as defined by minTexelBufferOffsetAlignment", 250}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-buffer-02751", 40}, { {"If buffer was created with the VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT usage flag set, offset must be a multiple of the effective alignment requirement of format for VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER as defined by minTexelBufferOffsetAlignment", 250}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-buffer-parameter", 44}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-flags-zerobitmask", 45}, { {"flags must be 0", 15}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-format-08778", 40}, { {"If the buffer view usage contains VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT, then format features of format must contain VK_FORMAT_FEATURE_UNIFORM_TEXEL_BUFFER_BIT", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-format-08779", 40}, { {"If the buffer view usage contains VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT, then format features of format must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_BIT", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-format-parameter", 44}, { {"format must be a valid VkFormat value", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-offset-00925", 40}, { {"offset must be less than the size of buffer", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-offset-00931", 40}, { {"If range is not equal to VK_WHOLE_SIZE, the sum of offset and range must be less than or equal to the size of buffer", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-pNext-06782", 39}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-pNext-08780", 39}, { {"If the pNext chain includes a VkBufferUsageFlags2CreateInfo, its usage must not contain any other bit than VK_BUFFER_USAGE_2_UNIFORM_TEXEL_BUFFER_BIT or VK_BUFFER_USAGE_2_STORAGE_TEXEL_BUFFER_BIT", 195}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-pNext-08781", 39}, { {"If the pNext chain includes a VkBufferUsageFlags2CreateInfo, its usage must be a subset of the VkBufferCreateInfo::usage specified or VkBufferUsageFlags2CreateInfo::usage from VkBufferCreateInfo::pNext when creating buffer", 222}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-pNext-pNext", 39}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBufferUsageFlags2CreateInfo or VkExportMetalObjectCreateInfoEXT", 196}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-range-00928", 39}, { {"If range is not equal to VK_WHOLE_SIZE, range must be greater than 0", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-range-00929", 39}, { {"If range is not equal to VK_WHOLE_SIZE, range must be an integer multiple of the texel block size of format", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-range-00930", 39}, { {"If range is not equal to VK_WHOLE_SIZE, the number of texel buffer elements given by (floor(range / (texel block size)) x (texels per block)) where texel block size and texels per block are as defined in the Compatible Formats table for format, must be less than or equal to VkPhysicalDeviceLimits::maxTexelBufferElements", 321}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-range-04059", 39}, { {"If range is equal to VK_WHOLE_SIZE, the number of texel buffer elements given by (floor((size - offset) / (texel block size)) x (texels per block)) where size is the size of buffer, and texel block size and texels per block are as defined in the Compatible Formats table for format, must be less than or equal to VkPhysicalDeviceLimits::maxTexelBufferElements", 359}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_BUFFER_VIEW_CREATE_INFO", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBufferViewCreateInfo-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureIndirectCommandNV-argData-10565", 75}, { {"An instance index must not be referenced by more than one structure in argData", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureIndirectCommandNV-opType-parameter", 78}, { {"opType must be a valid VkPartitionedAccelerationStructureOpTypeNV value", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-dstAccelerationStructureData-10562", 85}, { {"Memory at dstAccelerationStructureData must be equal or larger than the VkAccelerationStructureBuildSizesInfoKHR::accelerationStructureSize value returned from vkGetPartitionedAccelerationStructuresBuildSizesNV with the same build parameters", 241}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-dstAccelerationStructureData-parameter", 89}, { {"dstAccelerationStructureData must be a valid VkDeviceAddress value", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-input-parameter", 66}, { {"input must be a valid VkPartitionedAccelerationStructureInstancesInputNV structure", 82}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-pNext-pNext", 62}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_BUILD_PARTITIONED_ACCELERATION_STRUCTURE_INFO_NV", 80}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-scratchData-10558", 68}, { {"If the buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetPartitionedAccelerationStructuresBuildSizesNV is not 0, then scratchData must be a valid VkDeviceAddress", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-scratchData-10559", 68}, { {"Memory at scratchData must be equal or larger than the VkAccelerationStructureBuildSizesInfoKHR::buildScratchSize value returned from vkGetPartitionedAccelerationStructuresBuildSizesNV with the same build parameters", 215}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcAccelerationStructureData-10560", 85}, { {"If srcAccelerationStructureData is not NULL, it must have previously been built as a PTLAS", 90}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcAccelerationStructureData-parameter", 89}, { {"If srcAccelerationStructureData is not 0, srcAccelerationStructureData must be a valid VkDeviceAddress value", 108}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcInfos-10564", 65}, { {"Each element of srcInfos array must have a unique VkBuildPartitionedAccelerationStructureIndirectCommandNV::opType", 114}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcInfos-parameter", 69}, { {"srcInfos must be a valid VkDeviceAddress value", 46}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcInfosCount-10563", 70}, { {"srcInfosCount must be 4-byte aligned", 36}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkBuildPartitionedAccelerationStructureInfoNV-srcInfosCount-parameter", 74}, { {"srcInfosCount must be a valid VkDeviceAddress value", 51}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-pNext-pNext", 45}, { {"pNext must be NULL or a pointer to a valid instance of VkSwapchainCalibratedTimestampInfoEXT", 92}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_CALIBRATED_TIMESTAMP_INFO_KHR", 61}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-sType-unique", 46}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-timeDomain-02354", 50}, { {"timeDomain must be one of the VkTimeDomainKHR values returned by vkGetPhysicalDeviceCalibrateableTimeDomainsKHR", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-timeDomain-12227", 50}, { {"If timeDomain is VK_TIME_DOMAIN_SWAPCHAIN_LOCAL_EXT or VK_TIME_DOMAIN_PRESENT_STAGE_LOCAL_EXT, the pNext chain must include a VkSwapchainCalibratedTimestampInfoEXT structure", 173}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCalibratedTimestampInfoKHR-timeDomain-parameter", 54}, { {"timeDomain must be a valid VkTimeDomainKHR value", 48}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkCheckpointData2NV-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkCheckpointData2NV-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_CHECKPOINT_DATA_2_NV", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkCheckpointDataNV-pNext-pNext", 35}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkCheckpointDataNV-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_CHECKPOINT_DATA_NV", 50}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkClearAttachment-aspectMask-00019", 39}, { {"If aspectMask includes VK_IMAGE_ASPECT_COLOR_BIT, it must not include VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 126}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClearAttachment-aspectMask-00020", 39}, { {"aspectMask must not include VK_IMAGE_ASPECT_METADATA_BIT", 56}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClearAttachment-aspectMask-02246", 39}, { {"aspectMask must not include VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT for any index i", 82}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClearAttachment-aspectMask-parameter", 43}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClearAttachment-aspectMask-requiredbitmask", 49}, { {"aspectMask must not be 0", 24}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClearDepthStencilValue-depth-00022", 41}, { {"Unless the VK_EXT_depth_range_unrestricted extension is enabled depth must be between 0.0 and 1.0, inclusive", 108}, { "chapters/clears.html", 20} } },
        { {"VUID-VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV-clusterReferences-10484", 89}, { {"All cluster references in clusterReferences must be unique", 58}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV-clusterReferences-10485", 89}, { {"clusterReferences must have at least clusterReferencesCount values", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV-clusterReferences-parameter", 93}, { {"clusterReferences must be a valid VkDeviceAddress value", 55}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV-clusterReferencesStride-10486", 95}, { {"clusterReferencesStride must be greater than or equal to 8", 58}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-baseGeometryIndex-10496", 85}, { {"The maximum geometry index after using the values in baseGeometryIndex and geometryIndexBuffer must be less than VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxClusterGeometryIndex", 194}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-clusterFlags-parameter", 84}, { {"clusterFlags must be a valid combination of VkClusterAccelerationStructureClusterFlagBitsNV values", 98}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-clusterID-10488", 77}, { {"clusterID must not be 0xFFFFFFFF", 32}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-geometryIndexAndFlagsBuffer-parameter", 99}, { {"If geometryIndexAndFlagsBuffer is not 0, geometryIndexAndFlagsBuffer must be a valid VkDeviceAddress value", 106}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-indexBuffer-parameter", 83}, { {"indexBuffer must be a valid VkDeviceAddress value", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-indexBufferStride-10494", 85}, { {"indexBufferStride must be 0 or a multiple of indexType", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-indexType-10491", 77}, { {"indexType must only have a single bit set", 41}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-opacityMicromapArray-10881", 88}, { {"If opacityMicromapArray is not 0, then the cluster acceleration structure must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_CLUSTER_OPACITY_MICROMAPS_BIT_NV flag set in VkClusterAccelerationStructureInputInfoNV::flags", 231}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-opacityMicromapArray-parameter", 92}, { {"If opacityMicromapArray is not 0, opacityMicromapArray must be a valid VkDeviceAddress value", 92}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-opacityMicromapIndexBuffer-parameter", 98}, { {"If opacityMicromapIndexBuffer is not 0, opacityMicromapIndexBuffer must be a valid VkDeviceAddress value", 104}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-opacityMicromapIndexType-10492", 92}, { {"opacityMicromapIndexType must only have a single bit set", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-positionTruncateBitCount-10493", 92}, { {"positionTruncateBitCount must be greater than or equal to VkClusterAccelerationStructureTriangleClusterInputNV::minPositionTruncateBitCount and less than or equal to 32", 168}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-triangleCount-10489", 81}, { {"triangleCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxTrianglesPerCluster", 124}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-vertexBuffer-parameter", 84}, { {"vertexBuffer must be a valid VkDeviceAddress value", 50}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-vertexBufferStride-10495", 86}, { {"vertexBufferStride must be 0 or a multiple of value specified in VkClusterAccelerationStructureTriangleClusterInputNV::vertexFormat", 131}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterInfoNV-vertexCount-10490", 79}, { {"vertexCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxVerticesPerCluster", 121}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-baseGeometryIndex-10506", 93}, { {"The maximum geometry index after using the values in baseGeometryIndex and geometryIndexBuffer must be less than VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxClusterGeometryIndex", 194}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-clusterFlags-parameter", 92}, { {"clusterFlags must be a valid combination of VkClusterAccelerationStructureClusterFlagBitsNV values", 98}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-clusterID-10497", 85}, { {"clusterID must not be 0xFFFFFFFF", 32}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-geometryIndexAndFlagsBuffer-parameter", 107}, { {"If geometryIndexAndFlagsBuffer is not 0, geometryIndexAndFlagsBuffer must be a valid VkDeviceAddress value", 106}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-indexBuffer-parameter", 91}, { {"indexBuffer must be a valid VkDeviceAddress value", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-indexBufferStride-10503", 93}, { {"indexBufferStride must be 0 or a multiple of indexType", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-indexType-10500", 85}, { {"indexType must only have a single bit set", 41}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-instantiationBoundingBoxLimit-10505", 105}, { {"instantiationBoundingBoxLimit must be aligned to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::clusterTemplateBoundsByteAlignment", 141}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-instantiationBoundingBoxLimit-parameter", 109}, { {"If instantiationBoundingBoxLimit is not 0, instantiationBoundingBoxLimit must be a valid VkDeviceAddress value", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-opacityMicromapArray-10882", 96}, { {"If opacityMicromapArray is not 0, then the template cluster acceleration structure must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_CLUSTER_OPACITY_MICROMAPS_BIT_NV flag set in VkClusterAccelerationStructureInputInfoNV::flags", 240}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-opacityMicromapArray-parameter", 100}, { {"If opacityMicromapArray is not 0, opacityMicromapArray must be a valid VkDeviceAddress value", 92}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-opacityMicromapIndexBuffer-parameter", 106}, { {"If opacityMicromapIndexBuffer is not 0, opacityMicromapIndexBuffer must be a valid VkDeviceAddress value", 104}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-opacityMicromapIndexType-10501", 100}, { {"opacityMicromapIndexType must only have a single bit set", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-positionTruncateBitCount-10502", 100}, { {"positionTruncateBitCount must be greater than or equal to VkClusterAccelerationStructureTriangleClusterInputNV::minPositionTruncateBitCount and less than or equal to 32", 168}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-triangleCount-10498", 89}, { {"triangleCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxTrianglesPerCluster", 124}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-vertexBuffer-parameter", 92}, { {"If vertexBuffer is not 0, vertexBuffer must be a valid VkDeviceAddress value", 76}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-vertexBufferStride-10504", 94}, { {"vertexBufferStride must be 0 or a multiple of value specified in VkClusterAccelerationStructureTriangleClusterInputNV::vertexFormat", 131}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV-vertexCount-10499", 87}, { {"vertexCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxVerticesPerCluster", 121}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureClustersBottomLevelInputNV-pNext-pNext", 73}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureClustersBottomLevelInputNV-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_CLUSTERS_BOTTOM_LEVEL_INPUT_NV", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-addressResolutionFlags-parameter", 82}, { {"addressResolutionFlags must be a valid combination of VkClusterAccelerationStructureAddressResolutionFlagBitsNV values", 118}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-buildScratchSize-12321", 72}, { {"If the buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetClusterAccelerationStructureBuildSizesNV is not 0, then VkClusterAccelerationStructureCommandsInfoNV::scratchData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV, must be a valid VkDeviceAddress", 530}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-dstAddressesArray-10474", 73}, { {"The stride in dstAddressesArray must be greater than or equal to 8", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-dstImplicitData-parameter", 75}, { {"If dstImplicitData is not 0, dstImplicitData must be a valid VkDeviceAddress value", 82}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-dstSizesArray-10475", 69}, { {"The stride in dstSizesArray must be greater than or equal to 4", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-10469", 61}, { {"If input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_MOVE_OBJECTS_NV, the total memory moved must not be larger than the size provided in VkClusterAccelerationStructureMoveObjectsInputNV::maxMovedBytes", 209}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-12317", 61}, { {"If input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_NV, then depending on the VkClusterAccelerationStructureInputInfoNV::opMode, VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, or addresses specified in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the addresses indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be aligned to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::clusterByteAlignment", 966}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-12318", 61}, { {"If input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_TEMPLATE_NV, then depending on the VkClusterAccelerationStructureInputInfoNV::opMode, VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, or addresses specified in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the addresses indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be aligned to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::clusterTemplateByteAlignment", 983}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-12319", 61}, { {"If input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_INSTANTIATE_TRIANGLE_CLUSTER_NV, then depending on the VkClusterAccelerationStructureInputInfoNV::opMode, VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, or addresses specified in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the addresses indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be aligned to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::clusterByteAlignment", 972}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-12323", 61}, { {"If input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_CLUSTERS_BOTTOM_LEVEL_NV, the total and per argument number of cluster acceleration structures referenced in VkClusterAccelerationStructureCommandsInfoNV::srcInfosArray, or the structures indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::srcInfosArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SRC_INFOS_ARRAY_BIT_NV, must be equal or less than the maximum values with which memory requirements were queried in vkGetClusterAccelerationStructureBuildSizesNV with VkClusterAccelerationStructureOpInputNV::pClustersBottomLevel", 716}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-input-parameter", 65}, { {"input must be a valid VkClusterAccelerationStructureInputInfoNV structure", 73}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12309", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_IMPLICIT_DESTINATIONS_NV, VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, must be a valid VkDeviceAddress", 493}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12310", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_IMPLICIT_DESTINATIONS_NV and input.opType is not VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_MOVE_OBJECTS_NV, the memory in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV must be equal to or larger than the VkAccelerationStructureBuildSizesInfoKHR::accelerationStructureSize value returned from vkGetClusterAccelerationStructureBuildSizesNV with same input parameters", 752}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12311", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_IMPLICIT_DESTINATIONS_NV and input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_MOVE_OBJECTS_NV, the memory in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, must be equal to or larger than the sum of all the built acceleration structures that are being moved", 654}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12312", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_COMPUTE_SIZES_NV, VkClusterAccelerationStructureCommandsInfoNV::dstSizesArray, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstSizesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_SIZES_ARRAY_BIT_NV, must be a valid VkDeviceAddress", 479}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12313", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_EXPLICIT_DESTINATIONS_NV, the address in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be a valid VkDeviceAddress", 512}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12314", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_EXPLICIT_DESTINATIONS_NV, the addresses in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be large enough to accommodate built or moved clusters", 541}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12315", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_EXPLICIT_DESTINATIONS_NV, the buffers in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the buffers indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must not overlap", 497}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-opMode-12316", 62}, { {"If VkClusterAccelerationStructureInputInfoNV::opMode is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_MODE_EXPLICIT_DESTINATIONS_NV, the addresses in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the addresses indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be aligned based on the cluster acceleration structure type and its alignment properties as described in VkPhysicalDeviceClusterAccelerationStructurePropertiesNV", 651}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-pNext-pNext", 61}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_COMMANDS_INFO_NV", 79}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-scratchData-12320", 67}, { {"VkClusterAccelerationStructureCommandsInfoNV::scratchData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV, must be aligned to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::clusterScratchByteAlignment", 429}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-scratchData-parameter", 71}, { {"scratchData must be a valid VkDeviceAddress value", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-srcInfosArray-10476", 69}, { {"The stride in srcInfosArray must be greater than the type of structure the address is describing", 96}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-srcInfosCount-12322", 69}, { {"VkClusterAccelerationStructureCommandsInfoNV::srcInfosCount, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::srcInfosCount when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SRC_INFOS_COUNT_BIT_NV, must be 4-byte aligned", 354}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureCommandsInfoNV-srcInfosCount-parameter", 73}, { {"If srcInfosCount is not 0, srcInfosCount must be a valid VkDeviceAddress value", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureGeometryIndexAndGeometryFlagsNV-reserved-10487", 81}, { {"reserved must be 0", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureGetTemplateIndicesInfoNV-clusterTemplateAddress-10833", 88}, { {"clusterTemplateAddress must be a template cluster acceleration structure", 72}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureGetTemplateIndicesInfoNV-clusterTemplateAddress-parameter", 92}, { {"clusterTemplateAddress must be a valid VkDeviceAddress value", 60}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-flags-parameter", 62}, { {"flags must be a valid combination of VkBuildAccelerationStructureFlagBitsKHR values", 83}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-opMode-parameter", 63}, { {"opMode must be a valid VkClusterAccelerationStructureOpModeNV value", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-opType-parameter", 63}, { {"opType must be a valid VkClusterAccelerationStructureOpTypeNV value", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-pClustersBottomLevel-parameter", 77}, { {"If opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_CLUSTERS_BOTTOM_LEVEL_NV, the pClustersBottomLevel member of opInput must be a valid pointer to a VkClusterAccelerationStructureClustersBottomLevelInputNV structure", 225}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-pMoveObjects-parameter", 69}, { {"If opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_MOVE_OBJECTS_NV, the pMoveObjects member of opInput must be a valid pointer to a VkClusterAccelerationStructureMoveObjectsInputNV structure", 194}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-pNext-pNext", 58}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-pTriangleClusters-parameter", 74}, { {"If opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_NV, VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_TEMPLATE_NV, VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_INSTANTIATE_TRIANGLE_CLUSTER_NV, or VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_GET_CLUSTER_TEMPLATE_INDICES_NV, the pTriangleClusters member of opInput must be a valid pointer to a VkClusterAccelerationStructureTriangleClusterInputNV structure", 444}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInputInfoNV-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_INPUT_INFO_NV", 76}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInstantiateClusterInfoNV-clusterTemplateAddress-parameter", 92}, { {"clusterTemplateAddress must be a valid VkDeviceAddress value", 60}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInstantiateClusterInfoNV-geometryIndexOffset-10510", 85}, { {"The maximum geometry index after using the value in geometryIndexOffset must be less than VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxClusterGeometryIndex", 171}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInstantiateClusterInfoNV-reserved-10509", 74}, { {"reserved must be 0", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInstantiateClusterInfoNV-vertexBuffer-10507", 78}, { {"vertexBuffer must not be 0 if the template was built without vertex data", 72}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureInstantiateClusterInfoNV-vertexBuffer-10508", 78}, { {"The format in vertexBuffer must match the original format specified in VkClusterAccelerationStructureTriangleClusterInputNV", 123}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureMoveObjectsInfoNV-srcAccelerationStructure-10483", 83}, { {"srcAccelerationStructure must be a type of cluster acceleration structure", 73}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureMoveObjectsInfoNV-srcAccelerationStructure-parameter", 87}, { {"srcAccelerationStructure must be a valid VkDeviceAddress value", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureMoveObjectsInputNV-pNext-pNext", 65}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureMoveObjectsInputNV-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_MOVE_OBJECTS_INPUT_NV", 84}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureMoveObjectsInputNV-type-parameter", 68}, { {"type must be a valid VkClusterAccelerationStructureTypeNV value", 63}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-maxClusterTriangleCount-10440", 87}, { {"maxClusterTriangleCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxTrianglesPerCluster", 134}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-maxClusterVertexCount-10441", 85}, { {"maxClusterVertexCount must be less than or equal to VkPhysicalDeviceClusterAccelerationStructurePropertiesNV::maxVerticesPerCluster", 131}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-minPositionTruncateBitCount-10442", 91}, { {"minPositionTruncateBitCount must be less than or equal to 32", 60}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-pNext-pNext", 69}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_TRIANGLE_CLUSTER_INPUT_NV", 88}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-vertexFormat-10439", 76}, { {"The format features of vertexFormat must contain VK_FORMAT_FEATURE_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkClusterAccelerationStructureTriangleClusterInputNV-vertexFormat-parameter", 80}, { {"vertexFormat must be a valid VkFormat value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCoarseSampleLocationNV-pixelX-02078", 42}, { {"pixelX must be less than the width (in pixels) of the fragment", 62}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleLocationNV-pixelY-02079", 42}, { {"pixelY must be less than the height (in pixels) of the fragment", 63}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleLocationNV-sample-02080", 42}, { {"sample must be less than the number of coverage samples in each pixel belonging to the fragment", 95}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-pSampleLocations-02077", 55}, { {"The array pSampleLocations must contain exactly one entry for every combination of valid values for pixelX, pixelY, and sample in the structure VkCoarseSampleOrderCustomNV", 171}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-pSampleLocations-parameter", 59}, { {"pSampleLocations must be a valid pointer to an array of sampleLocationCount VkCoarseSampleLocationNV structures", 111}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-sampleCount-02074", 50}, { {"sampleCount must correspond to a sample count enumerated in VkSampleCountFlags whose corresponding bit is set in VkPhysicalDeviceLimits::framebufferNoAttachmentsSampleCounts", 173}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-sampleLocationCount-02075", 58}, { {"sampleLocationCount must be equal to the product of sampleCount, the fragment width for shadingRate, and the fragment height for shadingRate", 140}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-sampleLocationCount-02076", 58}, { {"sampleLocationCount must be less than or equal to the value of VkPhysicalDeviceShadingRateImagePropertiesNV::shadingRateMaxCoarseSamples", 136}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-sampleLocationCount-arraylength", 64}, { {"sampleLocationCount must be greater than 0", 42}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-shadingRate-02073", 50}, { {"shadingRate must be a shading rate that generates fragments with more than one pixel", 84}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkCoarseSampleOrderCustomNV-shadingRate-parameter", 54}, { {"shadingRate must be a valid VkShadingRatePaletteEntryNV value", 61}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkColorBlendAdvancedEXT-advancedBlendOp-parameter", 54}, { {"advancedBlendOp must be a valid VkBlendOp value", 47}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendAdvancedEXT-blendOverlap-07507", 47}, { {"If the correlated overlap property is not supported, blendOverlap must be VK_BLEND_OVERLAP_UNCORRELATED_EXT", 107}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendAdvancedEXT-blendOverlap-parameter", 51}, { {"blendOverlap must be a valid VkBlendOverlapEXT value", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendAdvancedEXT-dstPremultiplied-07506", 51}, { {"If the non-premultiplied destination color property is not supported, dstPremultiplied must be VK_TRUE", 102}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendAdvancedEXT-srcPremultiplied-07505", 51}, { {"If the non-premultiplied source color property is not supported, srcPremultiplied must be VK_TRUE", 97}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-alphaBlendOp-parameter", 51}, { {"alphaBlendOp must be a valid VkBlendOp value", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-colorBlendOp-07361", 47}, { {"colorBlendOp and alphaBlendOp must not be a VkBlendOp from VK_EXT_blend_operation_advanced", 90}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-colorBlendOp-parameter", 51}, { {"colorBlendOp must be a valid VkBlendOp value", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-constantAlphaColorBlendFactors-07362", 65}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::constantAlphaColorBlendFactors is VK_FALSE, srcColorBlendFactor must not be VK_BLEND_FACTOR_CONSTANT_ALPHA or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA", 255}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-constantAlphaColorBlendFactors-07363", 65}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::constantAlphaColorBlendFactors is VK_FALSE, dstColorBlendFactor must not be VK_BLEND_FACTOR_CONSTANT_ALPHA or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA", 255}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dstAlphaBlendFactor-parameter", 58}, { {"dstAlphaBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dstColorBlendFactor-parameter", 58}, { {"dstColorBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dualSrcBlend-07357", 47}, { {"If the dualSrcBlend feature is not enabled, srcColorBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dualSrcBlend-07358", 47}, { {"If the dualSrcBlend feature is not enabled, dstColorBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dualSrcBlend-07359", 47}, { {"If the dualSrcBlend feature is not enabled, srcAlphaBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-dualSrcBlend-07360", 47}, { {"If the dualSrcBlend feature is not enabled, dstAlphaBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-srcAlphaBlendFactor-parameter", 58}, { {"srcAlphaBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkColorBlendEquationEXT-srcColorBlendFactor-parameter", 58}, { {"srcColorBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkCommandBufferAllocateInfo-commandPool-parameter", 54}, { {"commandPool must be a valid VkCommandPool handle", 48}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferAllocateInfo-level-parameter", 48}, { {"level must be a valid VkCommandBufferLevel value", 48}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferAllocateInfo-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferAllocateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO", 60}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-00055", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT, the framebuffer member of pInheritanceInfo must be either VK_NULL_HANDLE, or a valid VkFramebuffer that is compatible with the renderPass member of pInheritanceInfo", 232}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-06000", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT and the renderPass member of pInheritanceInfo is not VK_NULL_HANDLE, the renderPass member of pInheritanceInfo must be a valid VkRenderPass", 206}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-06001", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT and the renderPass member of pInheritanceInfo is not VK_NULL_HANDLE, the subpass member of pInheritanceInfo must be a valid subpass index within the renderPass member of pInheritanceInfo", 253}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-06002", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT and the renderPass member of pInheritanceInfo is VK_NULL_HANDLE, the pNext chain of pInheritanceInfo must include a VkCommandBufferInheritanceRenderingInfo structure", 232}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-06003", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT, the renderPass member of pInheritanceInfo is VK_NULL_HANDLE, and the pNext chain of pInheritanceInfo includes a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, the colorAttachmentCount member of that structure must be equal to the value of VkCommandBufferInheritanceRenderingInfo::colorAttachmentCount", 396}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-09123", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 156}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-09240", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT and the dynamicRendering feature is not enabled, the renderPass member of pInheritanceInfo must not be VK_NULL_HANDLE", 184}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-10617", 41}, { {"If flags contains VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT , the renderPass member of pInheritanceInfo is not VK_NULL_HANDLE, and renderPass was created with tile shading enabled, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 296}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-10618", 41}, { {"If flags does not contain VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT , the renderPass member of pInheritanceInfo is VK_NULL_HANDLE, or renderPass was not created with tile shading enabled, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 307}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-10619", 41}, { {"If VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM is included in VkRenderPassTileShadingCreateInfoQCOM::flags, VkRenderPassTileShadingCreateInfoQCOM::tileApronSize must be equal to the tileApronSize used to create renderPass", 221}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-flags-parameter", 45}, { {"flags must be a valid combination of VkCommandBufferUsageFlagBits values", 72}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-pNext-pNext", 41}, { {"pNext must be NULL or a pointer to a valid instance of VkDeviceGroupCommandBufferBeginInfo", 90}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO", 57}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferBeginInfo-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceConditionalRenderingInfoEXT-conditionalRenderingEnable-01977", 91}, { {"If the inheritedConditionalRendering feature is not enabled, conditionalRenderingEnable must be VK_FALSE", 104}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceConditionalRenderingInfoEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_CONDITIONAL_RENDERING_INFO_EXT", 89}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-descriptorHeap-11200", 73}, { {"If the descriptorHeap feature is not enabled, pSamplerHeapBindInfo must be NULL", 79}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-descriptorHeap-11201", 73}, { {"If the descriptorHeap feature is not enabled, pResourceHeapBindInfo must be NULL", 80}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-pResourceHeapBindInfo-11471", 80}, { {"If pResourceHeapBindInfo is not NULL, pResourceHeapBindInfo->heapRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT usage flag set", 216}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-pResourceHeapBindInfo-parameter", 84}, { {"If pResourceHeapBindInfo is not NULL, pResourceHeapBindInfo must be a valid pointer to a valid VkBindHeapInfoEXT structure", 122}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-pSamplerHeapBindInfo-11470", 79}, { {"If pSamplerHeapBindInfo is not NULL, pSamplerHeapBindInfo->heapRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT usage flag set", 214}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-pSamplerHeapBindInfo-parameter", 83}, { {"If pSamplerHeapBindInfo is not NULL, pSamplerHeapBindInfo must be a valid pointer to a valid VkBindHeapInfoEXT structure", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceDescriptorHeapInfoEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_DESCRIPTOR_HEAP_INFO_EXT", 83}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-commonparent", 48}, { {"Both of framebuffer, and renderPass that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 155}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-occlusionQueryEnable-00056", 62}, { {"If the inheritedQueries feature is not enabled, occlusionQueryEnable must be VK_FALSE", 85}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-pNext-pNext", 47}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAttachmentSampleCountInfoAMD, VkCommandBufferInheritanceConditionalRenderingInfoEXT, VkCommandBufferInheritanceDescriptorHeapInfoEXT, VkCommandBufferInheritanceRenderPassTransformInfoQCOM, VkCommandBufferInheritanceRenderingInfo, VkCommandBufferInheritanceViewportScissorInfoNV, VkCustomResolveCreateInfoEXT, VkExternalFormatANDROID, VkExternalFormatOHOS, VkMultiviewPerViewAttributesInfoNVX, VkRenderPassTileShadingCreateInfoQCOM, VkRenderingAttachmentLocationInfo, VkRenderingInputAttachmentIndexInfo, or VkTileMemoryBindInfoQCOM", 664}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-pipelineStatistics-00058", 60}, { {"If the pipelineStatisticsQuery feature is not enabled, pipelineStatistics must be 0", 83}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-pipelineStatistics-02789", 60}, { {"If the pipelineStatisticsQuery feature is enabled, pipelineStatistics must be a valid combination of VkQueryPipelineStatisticFlagBits values", 140}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-queryFlags-00057", 52}, { {"If the inheritedQueries feature is enabled, queryFlags must be a valid combination of VkQueryControlFlagBits values", 115}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-queryFlags-02788", 52}, { {"If the inheritedQueries feature is not enabled, queryFlags must be 0", 68}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_INFO", 63}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceInfo-sType-unique", 48}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderPassTransformInfoQCOM-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDER_PASS_TRANSFORM_INFO_QCOM", 90}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderPassTransformInfoQCOM-transform-02864", 74}, { {"transform must be VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR", 180}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-colorAttachmentCount-06004", 71}, { {"If colorAttachmentCount is not 0, rasterizationSamples must be a valid VkSampleCountFlagBits value", 98}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-depthAttachmentFormat-06007", 72}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 163}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-depthAttachmentFormat-06200", 72}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED and stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, depthAttachmentFormat must equal stencilAttachmentFormat", 164}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-depthAttachmentFormat-06540", 72}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a depth component", 104}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-depthAttachmentFormat-parameter", 76}, { {"depthAttachmentFormat must be a valid VkFormat value", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-flags-parameter", 60}, { {"flags must be a valid combination of VkRenderingFlagBits values", 63}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-multiview-06008", 60}, { {"If the multiview feature is not enabled, viewMask must be 0", 59}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-pColorAttachmentFormats-06492", 74}, { {"If any element of pColorAttachmentFormats is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT , or VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV if the linearColorAttachment feature is enabled", 276}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-pColorAttachmentFormats-parameter", 78}, { {"If colorAttachmentCount is not 0, pColorAttachmentFormats must be a valid pointer to an array of colorAttachmentCount valid VkFormat values", 139}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-rasterizationSamples-parameter", 75}, { {"If rasterizationSamples is not 0, rasterizationSamples must be a valid VkSampleCountFlagBits value", 98}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDERING_INFO", 73}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-stencilAttachmentFormat-06199", 74}, { {"If stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 165}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-stencilAttachmentFormat-06541", 74}, { {"If stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a stencil aspect", 105}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-stencilAttachmentFormat-parameter", 78}, { {"stencilAttachmentFormat must be a valid VkFormat value", 54}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-variableMultisampleRate-06005", 74}, { {"If the variableMultisampleRate feature is not enabled, rasterizationSamples must be a valid VkSampleCountFlagBits value", 119}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceRenderingInfo-viewMask-06009", 59}, { {"The index of the most significant bit in viewMask must be less than maxMultiviewViewCount", 89}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_VIEWPORT_SCISSOR_INFO_NV", 83}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-viewportScissor2D-04782", 76}, { {"If the inheritedViewportScissor2D feature is not enabled, viewportScissor2D must be VK_FALSE", 92}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-viewportScissor2D-04783", 76}, { {"If the multiViewport feature is not enabled and viewportScissor2D is VK_TRUE, then viewportDepthCount must be 1", 111}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-viewportScissor2D-04784", 76}, { {"If viewportScissor2D is VK_TRUE, then viewportDepthCount must be greater than 0", 79}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-viewportScissor2D-04785", 76}, { {"If viewportScissor2D is VK_TRUE, then pViewportDepths must be a valid pointer to an array of viewportDepthCount valid VkViewport structures, except any requirements on x, y, width, and height do not apply", 204}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferInheritanceViewportScissorInfoNV-viewportScissor2D-04786", 76}, { {"If viewportScissor2D is VK_TRUE, then the command buffer must be recorded with the VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT", 131}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-commandBuffer-03890", 50}, { {"commandBuffer must not have been allocated with VK_COMMAND_BUFFER_LEVEL_SECONDARY", 81}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-commandBuffer-09445", 50}, { {"If any render pass instance in commandBuffer was recorded with a VkRenderPassStripeBeginInfoARM structure in its pNext chain and did not specify the VK_RENDERING_RESUMING_BIT flag, a VkRenderPassStripeSubmitInfoARM must be included in the pNext chain", 250}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-deviceMask-03891", 47}, { {"If deviceMask is not 0, it must be a valid device mask", 54}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-pNext-09446", 42}, { {"If a VkRenderPassStripeSubmitInfoARM is included in the pNext chain, the value of VkRenderPassStripeSubmitInfoARM::stripeSemaphoreInfoCount must be equal to the sum of the VkRenderPassStripeBeginInfoARM::stripeInfoCount parameters provided to render pass instances recorded in commandBuffer that did not specify the VK_RENDERING_RESUMING_BIT flag", 346}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-pNext-pNext", 42}, { {"pNext must be NULL or a pointer to a valid instance of VkRenderPassStripeSubmitInfoARM", 86}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO", 58}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandBufferSubmitInfo-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-flags-02860", 40}, { {"If the protectedMemory feature is not enabled, the VK_COMMAND_POOL_CREATE_PROTECTED_BIT bit of flags must not be set", 116}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-flags-parameter", 44}, { {"flags must be a valid combination of VkCommandPoolCreateFlagBits values", 71}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-pNext-09908", 40}, { {"If the pNext chain includes a VkDataGraphProcessingEngineCreateInfoARM structure, then queueFamilyIndex must designate a queue family that supports VK_QUEUE_DATA_GRAPH_BIT_ARM", 175}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-pNext-09909", 40}, { {"If the pNext chain includes a VkDataGraphProcessingEngineCreateInfoARM structure, each member of pProcessingEngines must be identical to VkQueueFamilyDataGraphPropertiesARM::engine retrieved from vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with queueFamilyIndex and the physicalDevice that was used to create device", 324}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-pNext-pNext", 40}, { {"pNext must be NULL or a pointer to a valid instance of VkDataGraphProcessingEngineCreateInfoARM", 95}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-queueFamilyIndex-11830", 51}, { {"If queueFamilyIndex designates a queue family that supports VK_QUEUE_DATA_GRAPH_BIT_ARM and enumerates a foreign engine through vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with type VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_NEURAL_QCOM or VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_COMPUTE_QCOM, the pNext chain must include VkDataGraphProcessingEngineCreateInfoARM with VkPhysicalDeviceDataGraphProcessingEngineARM::isForeign set to VK_TRUE for all elements of pProcessingEngines", 510}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO", 56}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkCommandPoolCreateInfo-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkComponentMapping-a-parameter", 35}, { {"a must be a valid VkComponentSwizzle value", 42}, { "chapters/resources.html", 23} } },
        { {"VUID-VkComponentMapping-b-parameter", 35}, { {"b must be a valid VkComponentSwizzle value", 42}, { "chapters/resources.html", 23} } },
        { {"VUID-VkComponentMapping-g-parameter", 35}, { {"g must be a valid VkComponentSwizzle value", 42}, { "chapters/resources.html", 23} } },
        { {"VUID-VkComponentMapping-r-parameter", 35}, { {"r must be a valid VkComponentSwizzle value", 42}, { "chapters/resources.html", 23} } },
        { {"VUID-VkComputeOccupancyPriorityParametersNV-occupancyPriority-12298", 67}, { {"occupancyPriority must be between 0 and 1, inclusive", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkComputeOccupancyPriorityParametersNV-occupancyThrottling-12299", 69}, { {"occupancyThrottling must be between 0 and 1, inclusive", 54}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkComputeOccupancyPriorityParametersNV-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkComputeOccupancyPriorityParametersNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_COMPUTE_OCCUPANCY_PRIORITY_PARAMETERS_NV", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkComputePipelineCreateInfo-None-09497", 43}, { {"If the pNext chain does not include a VkPipelineCreateFlags2CreateInfo structure, flags must be a valid combination of VkPipelineCreateFlagBits values", 150}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-None-10391", 43}, { {"If a resource variables is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-None-11367", 43}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must not be VK_NULL_HANDLE", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-commonparent", 45}, { {"Both of basePipelineHandle, and layout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 158}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03365", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03366", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03367", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03368", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03369", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03370", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-03576", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 96}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-04945", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07367", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07369", 44}, { {"flags must not include both VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT and VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07984", 44}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineIndex is -1, basePipelineHandle must be a valid compute VkPipeline handle", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07985", 44}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineHandle is VK_NULL_HANDLE, basePipelineIndex must be a valid index into the calling command's pCreateInfos parameter", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07986", 44}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, basePipelineIndex must be -1 or basePipelineHandle must be VK_NULL_HANDLE", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-07996", 44}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-09007", 44}, { {"If the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature is not enabled, flags must not include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-09008", 44}, { {"If flags includes VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV, then the pNext chain must include a pointer to a valid instance of VkComputePipelineIndirectBufferInfoNV specifying the address where the pipeline's metadata will be saved", 234}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-09178", 44}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, the shader specified by stage must not declare the ShaderEnqueueAMDX capability", 141}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-11007", 44}, { {"If flags includes VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT, then the VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-11311", 44}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be VK_NULL_HANDLE", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-11312", 44}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 287}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-11593", 44}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature is not enabled, flags must not include VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 181}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-flags-11798", 44}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-01687", 45}, { {"If layout is not VK_NULL_HANDLE, the number of resources in layout accessible to the compute shader stage must be less than or equal to VkPhysicalDeviceLimits::maxPerStageResources", 180}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-07987", 45}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, a push constant range in layout must match the shader stage", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-07988", 45}, { {"If a resource variable is declared in a shader and layout is not VK_NULL_HANDLE, the corresponding descriptor set in layout must match the shader stage", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-07990", 45}, { {"If a resource variable is declared in a shader, layout is not VK_NULL_HANDLE, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in layout must match the descriptor type", 214}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-07991", 45}, { {"If a resource variable is declared in a shader as an array and layout is not VK_NULL_HANDLE, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 236}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-10069", 45}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, the block must be contained inside the push constant range in layout that matches the stage", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-12396", 45}, { {"If layout is not VK_NULL_HANDLE, it must not have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-layout-parameter", 49}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkComputePipelineIndirectBufferInfoNV, VkPipelineBinaryInfoKHR, VkPipelineCompilerControlCreateInfoAMD, VkPipelineCreateFlags2CreateInfo, VkPipelineCreationFeedbackCreateInfo, VkPipelineRobustnessCreateInfo, VkSubpassShadingPipelineCreateInfoHUAWEI, or VkValidationFeaturesEXT", 407}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-pipelineCreationCacheControl-02878", 67}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT nor VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-pipelineProtectedAccess-07368", 62}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-pipelineStageCreationFeedbackCount-06566", 73}, { {"If VkPipelineCreationFeedbackCreateInfo::pipelineStageCreationFeedbackCount is not 0, it must be 1", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO", 60}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-shaderEnqueue-09177", 52}, { {"If the shaderEnqueue feature is not enabled, flags must not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 102}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-stage-00701", 44}, { {"The stage member of stage must be VK_SHADER_STAGE_COMPUTE_BIT", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-stage-00702", 44}, { {"The shader code for the entry point identified by stage and the rest of the state identified by this structure must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 195}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineCreateInfo-stage-parameter", 48}, { {"stage must be a valid VkPipelineShaderStageCreateInfo structure", 63}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-deviceAddress-09011", 62}, { {"deviceAddress must be aligned to the VkMemoryRequirements2::memoryRequirements.alignment, as returned by vkGetPipelineIndirectMemoryRequirementsNV", 146}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-deviceAddress-09012", 62}, { {"deviceAddress must have been allocated from a buffer that was created with both the VK_BUFFER_USAGE_TRANSFER_DST_BIT and VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flags set", 172}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-deviceAddress-parameter", 66}, { {"deviceAddress must be a valid VkDeviceAddress value", 51}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-deviceGeneratedComputePipelines-09009", 80}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-flags-09010", 54}, { {"The pipeline creation flags in VkComputePipelineCreateInfo::flags must include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 122}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-pipelineDeviceAddressCaptureReplay-09014", 83}, { {"If pipelineDeviceAddressCaptureReplay is non-zero then the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputeCaptureReplay feature must be enabled", 176}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-pipelineDeviceAddressCaptureReplay-09015", 83}, { {"If pipelineDeviceAddressCaptureReplay is non-zero then that address must have been allocated with flag VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT set", 159}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-pipelineDeviceAddressCaptureReplay-09016", 83}, { {"If pipelineDeviceAddressCaptureReplay is non-zero, the pipeline must have been recreated for replay", 99}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-pipelineDeviceAddressCaptureReplay-09017", 83}, { {"pipelineDeviceAddressCaptureReplay must satisfy the alignment and size requirements similar to deviceAddress", 108}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-pipelineDeviceAddressCaptureReplay-parameter", 87}, { {"If pipelineDeviceAddressCaptureReplay is not 0, pipelineDeviceAddressCaptureReplay must be a valid VkDeviceAddress value", 120}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_INDIRECT_BUFFER_INFO_NV", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkComputePipelineIndirectBufferInfoNV-size-09013", 53}, { {"size must be greater than or equal to the VkMemoryRequirements2::memoryRequirements.size, as returned by vkGetPipelineIndirectMemoryRequirementsNV", 146}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressFlags-13100", 59}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressFlags-13101", 59}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressFlags-parameter", 63}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13064", 59}, { {"The VkBuffer that addressRange was queried from must have been created with VK_BUFFER_USAGE_CONDITIONAL_RENDERING_BIT_EXT", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13065", 59}, { {"addressRange.address must be a multiple of 4", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13066", 59}, { {"addressRange.size must be greater than or equal to 4", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13097", 59}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13098", 59}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13099", 59}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13122", 59}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13123", 59}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13124", 59}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-addressRange-13125", 59}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-flags-parameter", 56}, { {"flags must be a valid combination of VkConditionalRenderingFlagBitsEXT values", 77}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfo2EXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_2_EXT", 70}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-buffer-01981", 52}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-buffer-01982", 52}, { {"buffer must have been created with the VK_BUFFER_USAGE_CONDITIONAL_RENDERING_BIT_EXT usage flag set", 99}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-buffer-parameter", 56}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-flags-parameter", 55}, { {"flags must be a valid combination of VkConditionalRenderingFlagBitsEXT values", 77}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-offset-01983", 52}, { {"offset must be less than the size of buffer by at least 32 bits", 63}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-offset-01984", 52}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConditionalRenderingBeginInfoEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_EXT", 68}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstComponentType-10080", 66}, { {"If dstComponentType is not a supported VkCooperativeVectorPropertiesNV::matrixInterpretation value as reported by vkGetPhysicalDeviceCooperativeVectorPropertiesNV, then dstComponentType must be VK_COMPONENT_TYPE_FLOAT32_KHR", 223}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstComponentType-10082", 66}, { {"If dstComponentType is VK_COMPONENT_TYPE_FLOAT_E4M3_NV or VK_COMPONENT_TYPE_FLOAT_E5M2_NV, then dstLayout must be VK_COOPERATIVE_VECTOR_MATRIX_LAYOUT_INFERENCING_OPTIMAL_NV or VK_COOPERATIVE_VECTOR_MATRIX_LAYOUT_TRAINING_OPTIMAL_NV", 231}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstComponentType-parameter", 70}, { {"dstComponentType must be a valid VkComponentTypeKHR value", 57}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstData-parameter", 61}, { {"dstData must be a valid VkDeviceOrHostAddressKHR union", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstLayout-10078", 59}, { {"If dstLayout is row-major or column-major, then dstStride must be greater than the length of a row/column, and a multiple of the element size", 141}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-dstLayout-parameter", 63}, { {"dstLayout must be a valid VkCooperativeVectorMatrixLayoutNV value", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-pDstSize-parameter", 62}, { {"pDstSize must be a valid pointer to a size_t value", 50}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_CONVERT_COOPERATIVE_VECTOR_MATRIX_INFO_NV", 73}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcComponentType-10079", 66}, { {"If srcComponentType is not a supported VkCooperativeVectorPropertiesNV::matrixInterpretation value as reported by vkGetPhysicalDeviceCooperativeVectorPropertiesNV, then srcComponentType must be VK_COMPONENT_TYPE_FLOAT32_KHR", 223}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcComponentType-10081", 66}, { {"If srcComponentType and dstComponentType are not equal, then one must be VK_COMPONENT_TYPE_FLOAT32_KHR or VK_COMPONENT_TYPE_FLOAT16_KHR and the other must be a lower-precision floating-point type", 195}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcComponentType-parameter", 70}, { {"srcComponentType must be a valid VkComponentTypeKHR value", 57}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcData-parameter", 61}, { {"srcData must be a valid VkDeviceOrHostAddressConstKHR union", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcLayout-10077", 59}, { {"If srcLayout is row-major or column-major, then srcStride must be greater than the length of a row/column, and a multiple of the element size", 141}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkConvertCooperativeVectorMatrixInfoNV-srcLayout-parameter", 63}, { {"srcLayout must be a valid VkCooperativeVectorMatrixLayoutNV value", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixFlexibleDimensionsPropertiesNV-pNext-pNext", 66}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixFlexibleDimensionsPropertiesNV-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_FLEXIBLE_DIMENSIONS_PROPERTIES_NV", 84}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixPropertiesKHR-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixPropertiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_KHR", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixPropertiesNV-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeMatrixPropertiesNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_NV", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-biasInterpretation-parameter", 65}, { {"biasInterpretation must be a valid VkComponentTypeKHR value", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-inputInterpretation-parameter", 66}, { {"inputInterpretation must be a valid VkComponentTypeKHR value", 60}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-inputType-parameter", 56}, { {"inputType must be a valid VkComponentTypeKHR value", 50}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-matrixInterpretation-parameter", 67}, { {"matrixInterpretation must be a valid VkComponentTypeKHR value", 61}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-resultType-parameter", 57}, { {"resultType must be a valid VkComponentTypeKHR value", 51}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCooperativeVectorPropertiesNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_COOPERATIVE_VECTOR_PROPERTIES_NV", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-buffer-03718", 52}, { {"The range of src accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-buffer-03719", 52}, { {"The range of dst accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-commonparent", 52}, { {"Both of dst, and src must have been created, allocated, or retrieved from the same VkDevice", 91}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-dst-07791", 49}, { {"The range of memory backing dst that is accessed by this command must not overlap the memory backing src that is accessed by this command", 137}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-dst-parameter", 53}, { {"dst must be a valid VkAccelerationStructureKHR handle", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-mode-03410", 50}, { {"mode must be VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR or VK_COPY_ACCELERATION_STRUCTURE_MODE_CLONE_KHR", 109}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-mode-12418", 50}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_CLONE_KHR, the memory pointed to by dst must be at least as large as the size of src, as reported by vkWriteAccelerationStructuresPropertiesKHR or vkCmdWriteAccelerationStructuresPropertiesKHR with a query type of VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR", 306}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-mode-12419", 50}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR, the memory pointed to by dst must be at least as large as the compacted size of src, as reported by vkWriteAccelerationStructuresPropertiesKHR or vkCmdWriteAccelerationStructuresPropertiesKHR with a query type of VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR", 328}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-mode-parameter", 54}, { {"mode must be a valid VkCopyAccelerationStructureModeKHR value", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_INFO_KHR", 68}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-src-03411", 49}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR, src must have been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR in the build", 165}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureInfoKHR-src-parameter", 53}, { {"src must be a valid VkAccelerationStructureKHR handle", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-dst-03561", 57}, { {"The memory pointed to by dst must be at least as large as the serialization size of src, as reported by vkWriteAccelerationStructuresPropertiesKHR or vkCmdWriteAccelerationStructuresPropertiesKHR with a query type of VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR", 276}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-mode-03412", 58}, { {"mode must be VK_COPY_ACCELERATION_STRUCTURE_MODE_SERIALIZE_KHR", 62}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-mode-11714", 58}, { {"The range of src accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-mode-11715", 58}, { {"The range of dst accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-mode-parameter", 62}, { {"mode must be a valid VkCopyAccelerationStructureModeKHR value", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-pNext-pNext", 59}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_TO_MEMORY_INFO_KHR", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-src-04959", 57}, { {"The source acceleration structure src must have been constructed prior to the execution of this command", 103}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-src-11582", 57}, { {"If the source acceleration structure src was constructed through deserialization, all micromap arrays the acceleration structure references that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command", 382}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyAccelerationStructureToMemoryInfoKHR-src-parameter", 61}, { {"src must be a valid VkAccelerationStructureKHR handle", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyBufferInfo2-commonparent", 35}, { {"Both of dstBuffer, and srcBuffer must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-dstBuffer-00120", 38}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-dstBuffer-00121", 38}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-dstBuffer-parameter", 42}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-dstOffset-00114", 38}, { {"The dstOffset member of each element of pRegions must be less than the size of dstBuffer", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-pRegions-00117", 37}, { {"The union of the source regions, and the union of the destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-pRegions-parameter", 41}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferCopy2 structures", 90}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-regionCount-arraylength", 46}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_COPY_BUFFER_INFO_2", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-size-00115", 33}, { {"The size member of each element of pRegions must be less than or equal to the size of srcBuffer minus srcOffset", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-size-00116", 33}, { {"The size member of each element of pRegions must be less than or equal to the size of dstBuffer minus dstOffset", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-srcBuffer-00118", 38}, { {"srcBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-srcBuffer-00119", 38}, { {"If srcBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-srcBuffer-parameter", 42}, { {"srcBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferInfo2-srcOffset-00113", 38}, { {"The srcOffset member of each element of pRegions must be less than the size of srcBuffer", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-bufferImageHeight-09107", 53}, { {"For each element of pRegions, bufferImageHeight must be a multiple of the texel block extent height of the VkFormat of dstImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-bufferRowLength-09106", 51}, { {"For each element of pRegions, bufferRowLength must be a multiple of the texel block extent width of the VkFormat of dstImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-bufferRowLength-09108", 51}, { {"For each element of pRegions, bufferRowLength divided by the texel block extent width and then multiplied by the texel block size of dstImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-commonparent", 42}, { {"Both of dstImage, and srcBuffer must have been created, allocated, or retrieved from the same VkDevice", 102}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-00177", 44}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-00207", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-00208", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-00209", 44}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of dstImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-01997", 44}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07274", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07275", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07276", 44}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07966", 44}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07969", 44}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07973", 44}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07975", 44}, { {"If dstImage does not have either a depth/stencil format or a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the texel block size", 172}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07976", 44}, { {"If dstImage has a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 275}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07978", 44}, { {"If dstImage has a depth/stencil format, the bufferOffset member of any element of pRegions must be a multiple of 4", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07979", 44}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07980", 44}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07981", 44}, { {"If dstImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-07983", 44}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImage-parameter", 48}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImageLayout-00180", 50}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImageLayout-01396", 50}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-dstImageLayout-parameter", 54}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-09104", 47}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of dstImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10051", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10052", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10053", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10054", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10055", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10056", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10057", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageOffset-10058", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageSubresource-07967", 52}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageSubresource-07968", 52}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-imageSubresource-09105", 52}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in dstImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-00171", 44}, { {"srcBuffer must be large enough to contain all buffer locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-00173", 44}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-04565", 44}, { {"The image region specified by each element of pRegions that does not contain VkCopyCommandTransformInfoQCOM in its pNext chain must be contained within the specified imageSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-06223", 44}, { {"For each element of pRegions not containing VkCopyCommandTransformInfoQCOM in its pNext chain, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of dstImage", 270}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-06224", 44}, { {"For each element of pRegions not containing VkCopyCommandTransformInfoQCOM in its pNext chain, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of dstImage", 272}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-07931", 44}, { {"If VK_EXT_depth_range_unrestricted is not enabled, for each element of pRegions whose imageSubresource contains a depth aspect, the data in srcBuffer must be in the range [0,1]", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-pRegions-parameter", 48}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferImageCopy2 structures", 95}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-regionCount-arraylength", 53}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_COPY_BUFFER_TO_IMAGE_INFO_2", 59}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-srcBuffer-00174", 45}, { {"srcBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-srcBuffer-00176", 45}, { {"If srcBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2-srcBuffer-parameter", 49}, { {"srcBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2KHR-pRegions-04554", 47}, { {"If the image region specified by each element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, the rotated destination region must be contained within dstImage", 182}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2KHR-pRegions-04555", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then dstImage must have a 1x1x1 texel block extent", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2KHR-pRegions-06203", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then dstImage must be of type VK_IMAGE_TYPE_2D", 133}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyBufferToImageInfo2KHR-pRegions-06204", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then dstImage must not have a multi-planar format", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyCommandTransformInfoQCOM-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_COPY_COMMAND_TRANSFORM_INFO_QCOM", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyCommandTransformInfoQCOM-transform-04560", 51}, { {"transform must be VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR", 180}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDescriptorSet-commonparent", 37}, { {"Both of dstSet, and srcSet must have been created, allocated, or retrieved from the same VkDevice", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstArrayElement-00348", 46}, { {"The sum of dstArrayElement and descriptorCount must be less than or equal to the number of array elements in the descriptor set binding specified by dstBinding, and all applicable consecutive bindings", 200}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstBinding-00347", 41}, { {"dstBinding must be a valid binding within dstSet", 48}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstBinding-02224", 41}, { {"If the descriptor type of the descriptor set binding specified by dstBinding is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, dstArrayElement must be an integer multiple of 4", 169}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstBinding-02632", 41}, { {"The type of dstBinding within dstSet must be equal to the type of srcBinding within srcSet", 90}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstBinding-02753", 41}, { {"If the descriptor type of the descriptor set binding specified by dstBinding is VK_DESCRIPTOR_TYPE_SAMPLER, then dstSet must not have been allocated with a layout that included immutable samplers for dstBinding", 210}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstSet-04612", 37}, { {"If VkDescriptorSetLayoutBinding for dstSet at dstBinding is VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the new active descriptor type must exist in the corresponding pMutableDescriptorTypeLists list for dstBinding if the new active descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 274}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstSet-04614", 37}, { {"If VkDescriptorSetLayoutBinding for dstSet at dstBinding is VK_DESCRIPTOR_TYPE_MUTABLE_EXT, and the new active descriptor type is VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the pMutableDescriptorTypeLists for srcBinding and dstBinding must match exactly", 242}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-dstSet-parameter", 41}, { {"dstSet must be a valid VkDescriptorSet handle", 45}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_COPY_DESCRIPTOR_SET", 51}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcArrayElement-00346", 46}, { {"The sum of srcArrayElement and descriptorCount must be less than or equal to the number of array elements in the descriptor set binding specified by srcBinding, and all applicable consecutive bindings", 200}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcBinding-00345", 41}, { {"srcBinding must be a valid binding within srcSet", 48}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcBinding-02223", 41}, { {"If the descriptor type of the descriptor set binding specified by srcBinding is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, srcArrayElement must be an integer multiple of 4", 169}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcBinding-02225", 41}, { {"If the descriptor type of the descriptor set binding specified by either srcBinding or dstBinding is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, descriptorCount must be an integer multiple of 4", 190}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-00349", 37}, { {"If srcSet is equal to dstSet, then the source and destination ranges of descriptors must not overlap, where the ranges may include array elements from consecutive bindings", 171}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-01918", 37}, { {"If srcSet's layout was created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT flag set, then dstSet's layout must also have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT flag set", 234}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-01920", 37}, { {"If the descriptor pool from which srcSet was allocated was created with the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT flag set, then the descriptor pool from which dstSet was allocated must also have been created with the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT flag set", 284}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-04613", 37}, { {"If VkDescriptorSetLayoutBinding for srcSet at srcBinding is VK_DESCRIPTOR_TYPE_MUTABLE_EXT and the VkDescriptorSetLayoutBinding for dstSet at dstBinding is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the active descriptor type for the source descriptor must match the descriptor type of dstBinding", 289}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-04885", 37}, { {"If srcSet's layout was created without either the VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT flag or the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT flag set, then dstSet's layout must have been created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT flag set", 309}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-04887", 37}, { {"If the descriptor pool from which srcSet was allocated was created without either the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT flag or the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT flag set, then the descriptor pool from which dstSet was allocated must have been created without the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT flag set", 348}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDescriptorSet-srcSet-parameter", 41}, { {"srcSet must be a valid VkDescriptorSet handle", 45}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-addressImageHeight-09107", 60}, { {"For each element of pRegions, addressImageHeight must be a multiple of the texel block extent height of the VkFormat of image", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-addressRange-13026", 54}, { {"The range of memory backing the address range defined by the addressRange member of any element of pRegions must not overlap the memory backing the address range defined by the addressRange of any other element of pRegions", 222}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-addressRange-13027", 54}, { {"The range of memory backing the address range defined by the addressRange member of any element of pRegions must not overlap the memory backing any texels in image defined by the imageSubresource, imageOffset, and imageExtent members of any element of pRegions", 260}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-addressRowLength-09106", 58}, { {"For each element of pRegions, addressRowLength must be a multiple of the texel block extent width of the VkFormat of image", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-addressRowLength-09108", 58}, { {"For each element of pRegions, addressRowLength divided by the texel block extent width and then multiplied by the texel block size of image must be less than or equal to 2^31-1", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-00207", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of image", 331}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-00208", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of image", 339}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-00209", 47}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of image", 236}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07274", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of image", 258}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07275", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of image", 258}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07276", 47}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of image", 119}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07966", 47}, { {"If image is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 146}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07969", 47}, { {"image must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07973", 47}, { {"image must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 61}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07979", 47}, { {"If image is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07980", 47}, { {"If image is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07981", 47}, { {"If image has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 150}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-07983", 47}, { {"If image is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-13029", 47}, { {"If image does not have either a depth/stencil format or a multi-planar format, then for each element of pRegions, addressRange.address must be a multiple of the texel block size", 177}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-13030", 47}, { {"If image has a multi-planar format, then for each element of pRegions, addressRange.address must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 280}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-13031", 47}, { {"If image has a depth/stencil format, the addressRange.address member of any element of pRegions must be a multiple of 4", 119}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-image-parameter", 51}, { {"image must be a valid VkImage handle", 36}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageLayout-13028", 53}, { {"The imageLayout member of each element of pRegions must specify the layout of the image subresource specified for that element of pRegions at the time this command is executed on a VkDevice", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-09104", 53}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of image", 202}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10051", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of image", 352}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10052", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of image", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10053", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of image", 252}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10054", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of image", 252}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10055", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of image", 335}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10056", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of image", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10057", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of image", 254}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageOffset-10058", 53}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of image", 253}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageSubresource-07967", 58}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when image was created", 150}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageSubresource-07968", 58}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 254}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-imageSubresource-09105", 58}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in image", 95}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-13032", 50}, { {"The image region specified by each element of pRegions that does not contain VkCopyCommandTransformInfoQCOM in its pNext chain must be contained within the specified imageSubresource of image", 191}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-13033", 50}, { {"If the image region specified by each element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, the rotated image region must be contained within image", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-13034", 50}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then image must have a 1x1x1 texel block extent", 134}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-13035", 50}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then image must be of type VK_IMAGE_TYPE_2D", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-13036", 50}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then image must not have a multi-planar format", 133}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-pRegions-parameter", 54}, { {"pRegions must be a valid pointer to an array of regionCount valid VkDeviceMemoryImageCopyKHR structures", 103}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-regionCount-arraylength", 59}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryImageInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_COPY_DEVICE_MEMORY_IMAGE_INFO_KHR", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryInfoKHR-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryInfoKHR-pRegions-parameter", 49}, { {"pRegions must be a valid pointer to an array of regionCount valid VkDeviceMemoryCopyKHR structures", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryInfoKHR-regionCount-arraylength", 54}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_COPY_DEVICE_MEMORY_INFO_KHR", 59}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyDeviceMemoryInfoKHR-srcRange-13015", 45}, { {"The range of memory backing the address range defined by the srcRange member of any element of pRegions must not overlap the memory backing the address range defined by the srcRange or dstRange of any element of pRegions", 220}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-None-01549", 32}, { {"In a copy to or from a plane of a multi-planar image, the VkFormat of the image and plane must be compatible according to the description of compatible planes for the plane being copied", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-apiVersion-07932", 38}, { {"If the VK_KHR_maintenance1 extension is not enabled, or VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and either srcImage or dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer and dstSubresource.baseArrayLayer must both be 0, and srcSubresource.layerCount and dstSubresource.layerCount must both be 1", 370}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-apiVersion-07933", 38}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, srcImage and dstImage must have the same VkImageType", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-apiVersion-08969", 38}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, srcImage or dstImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, extent.depth must be 1", 228}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspect-06662", 34}, { {"If the aspect member of any element of pRegions includes any flag other than VK_IMAGE_ASPECT_STENCIL_BIT or srcImage was not created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 249}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspect-06663", 34}, { {"If the aspect member of any element of pRegions includes any flag other than VK_IMAGE_ASPECT_STENCIL_BIT or dstImage was not created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 249}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspect-06664", 34}, { {"If the aspect member of any element of pRegions includes VK_IMAGE_ASPECT_STENCIL_BIT, and srcImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 227}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspect-06665", 34}, { {"If the aspect member of any element of pRegions includes VK_IMAGE_ASPECT_STENCIL_BIT, and dstImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 227}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspectMask-00142", 38}, { {"For each element of pRegions, srcSubresource.aspectMask must specify aspects present in srcImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-aspectMask-00143", 38}, { {"For each element of pRegions, dstSubresource.aspectMask must specify aspects present in dstImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-commonparent", 34}, { {"Both of dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-00152", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.y must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-01557", 36}, { {"If dstImage has a multi-planar format and the srcImage does not have a multi-planar image format, then for each element of pRegions, srcSubresource.aspectMask must be VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-01786", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-01788", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-01792", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_2D, and srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, extent.depth must equal dstSubresource.layerCount", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-01996", 36}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-04444", 36}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0 and dstSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-07966", 36}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-07969", 36}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-08714", 36}, { {"If dstImage has a multi-planar format, then for each element of pRegions, dstSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImage-parameter", 40}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImageLayout-00133", 42}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImageLayout-01395", 42}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstImageLayout-parameter", 46}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstOffset-00150", 37}, { {"For each element of pRegions, dstOffset.x and (extent.width + dstOffset.x), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 244}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstOffset-00151", 37}, { {"For each element of pRegions, dstOffset.y and (extent.height + dstOffset.y), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 246}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstOffset-00153", 37}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstOffset.z and (extent.depth + dstOffset.z), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 290}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstOffset-01784", 37}, { {"The dstOffset and extent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstSubresource-07967", 42}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstSubresource-07968", 42}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-dstSubresource-10215", 42}, { {"If dstSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then srcSubresource.aspectMask must not contain both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 169}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pNext-pNext", 33}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-00124", 36}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07278", 36}, { {"For each element of pRegions, srcOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07279", 36}, { {"For each element of pRegions, srcOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 121}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07280", 36}, { {"For each element of pRegions, srcOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07281", 36}, { {"For each element of pRegions, dstOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07282", 36}, { {"For each element of pRegions, dstOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 121}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-07283", 36}, { {"For each element of pRegions, dstOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-12201", 36}, { {"For each element of pRegions where srcSubresource.aspectMask and dstSubresource.aspectMask each contain at least one of VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, srcSubresource.aspectMask and dstSubresource.aspectMask must match", 244}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-pRegions-parameter", 40}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageCopy2 structures", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-regionCount-arraylength", 45}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_COPY_IMAGE_INFO_2", 49}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-00136", 36}, { {"The sample count of srcImage and dstImage must match", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-00146", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01548", 36}, { {"If the VkFormat of each of srcImage and dstImage is not a multi-planar format, the VkFormat of each of srcImage and dstImage must be size-compatible", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01551", 36}, { {"If neither srcImage nor dstImage has a multi-planar format and the maintenance8 feature is not enabled then for each element of pRegions, srcSubresource.aspectMask and dstSubresource.aspectMask must match", 204}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01556", 36}, { {"If srcImage has a multi-planar format and the dstImage does not have a multi-planar image format, then for each element of pRegions, dstSubresource.aspectMask must be VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01728", 36}, { {"For each element of pRegions, if the sum of srcOffset.x and extent.width does not equal the width of the subresource specified by srcSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01729", 36}, { {"For each element of pRegions, if the sum of srcOffset.y and extent.height does not equal the height of the subresource specified by srcSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 241}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01730", 36}, { {"For each element of pRegions, if the sum of srcOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01785", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.z must be 0 and extent.depth must be 1", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01787", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01790", 36}, { {"If srcImage and dstImage are both of type VK_IMAGE_TYPE_2D, then for each element of pRegions, extent.depth must be 1", 117}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01791", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_2D, and dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, extent.depth must equal srcSubresource.layerCount", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-01995", 36}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-04443", 36}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer must be 0 and srcSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-07743", 36}, { {"If srcImage and dstImage have a different VkImageType, and the maintenance5 feature is not enabled, one must be VK_IMAGE_TYPE_3D and the other must be VK_IMAGE_TYPE_2D", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-07966", 36}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-07969", 36}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-08713", 36}, { {"If srcImage has a multi-planar format, then for each element of pRegions, srcSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-08793", 36}, { {"If srcImage and dstImage have the same VkImageType, for each element of pRegions, if neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount members of srcSubresource or dstSubresource must match", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-08794", 36}, { {"If srcImage and dstImage have the same VkImageType, and one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 315}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-09247", 36}, { {"If the VkFormat of each of srcImage and dstImage is a compressed image format, the formats must have the same texel block extent", 128}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-09460", 36}, { {"If srcImage and dstImage are the same, and any elements of pRegions contains the srcSubresource and dstSubresource with matching mipLevel and overlapping array layers, then the srcImageLayout and dstImageLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR", 280}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-10907", 36}, { {"If either the VkFormat of each of srcImage and dstImage is not a compressed image format, and dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, extent.depth must be 1", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-10908", 36}, { {"If either the VkFormat of each of srcImage and dstImage is not a compressed image format, and dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, extent.height must be 1, where extent is adjusted for size-compatibility", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImage-parameter", 40}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImageLayout-00128", 42}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImageLayout-01917", 42}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcImageLayout-parameter", 46}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcOffset-00144", 37}, { {"For each element of pRegions, srcOffset.x and (extent.width + srcOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcOffset-00145", 37}, { {"For each element of pRegions, srcOffset.y and (extent.height + srcOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcOffset-00147", 37}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcOffset.z and (extent.depth + srcOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 240}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcOffset-01783", 37}, { {"The srcOffset and extent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcSubresource-07967", 42}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcSubresource-07968", 42}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcSubresource-10211", 42}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT and dstSubresource.aspectMask is VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, then the VkFormat values of srcImage and dstImage must be compatible according to the list of compatible depth-stencil and color formats", 285}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcSubresource-10212", 42}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT and dstSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then the VkFormat values of srcImage and dstImage must be compatible according to the list of compatible depth-stencil and color formats", 285}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageInfo2-srcSubresource-10214", 42}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then dstSubresource.aspectMask must not contain both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 169}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-bufferImageHeight-09107", 53}, { {"For each element of pRegions, bufferImageHeight must be a multiple of the texel block extent height of the VkFormat of srcImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-bufferRowLength-09106", 51}, { {"For each element of pRegions, bufferRowLength must be a multiple of the texel block extent width of the VkFormat of srcImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-bufferRowLength-09108", 51}, { {"For each element of pRegions, bufferRowLength divided by the texel block extent width and then multiplied by the texel block size of srcImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-commonparent", 42}, { {"Both of dstBuffer, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 102}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-dstBuffer-00191", 45}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-dstBuffer-00192", 45}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-dstBuffer-parameter", 49}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-00197", 47}, { {"For each element of pRegions not containing VkCopyCommandTransformInfoQCOM in its pNext chain, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of srcImage", 270}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-00198", 47}, { {"For each element of pRegions not containing VkCopyCommandTransformInfoQCOM in its pNext chain, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of srcImage", 272}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-09104", 47}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of srcImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10051", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10052", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10053", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10054", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10055", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10056", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10057", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageOffset-10058", 47}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageSubresource-07967", 52}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageSubresource-07968", 52}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-imageSubresource-09105", 52}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in srcImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-pRegions-00183", 44}, { {"dstBuffer must be large enough to contain all buffer locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-pRegions-00184", 44}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-pRegions-04566", 44}, { {"The image region specified by each element of pRegions that does not contain VkCopyCommandTransformInfoQCOM in its pNext chain must be contained within the specified imageSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-pRegions-parameter", 48}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferImageCopy2 structures", 95}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-regionCount-arraylength", 53}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_COPY_IMAGE_TO_BUFFER_INFO_2", 59}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-00186", 44}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-00207", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-00208", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-00209", 44}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-01998", 44}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07274", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07275", 44}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07276", 44}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07966", 44}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07969", 44}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07973", 44}, { {"srcImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07975", 44}, { {"If srcImage does not have either a depth/stencil format or a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the texel block size", 172}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07976", 44}, { {"If srcImage has a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 275}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07978", 44}, { {"If srcImage has a depth/stencil format, the bufferOffset member of any element of pRegions must be a multiple of 4", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07979", 44}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07980", 44}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07981", 44}, { {"If srcImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-07983", 44}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImage-parameter", 48}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImageLayout-00189", 50}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImageLayout-01397", 50}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2-srcImageLayout-parameter", 54}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2KHR-pRegions-04557", 47}, { {"If the image region specified by each element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, the rotated source region must be contained within srcImage", 177}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2KHR-pRegions-04558", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage must have a 1x1x1 texel block extent", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2KHR-pRegions-06205", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage must be of type VK_IMAGE_TYPE_2D", 133}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToBufferInfo2KHR-pRegions-06206", 47}, { {"If any element of pRegions contains VkCopyCommandTransformInfoQCOM in its pNext chain, then srcImage must not have a multi-planar format", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-commonparent", 40}, { {"Both of dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-00207", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of dstOffset.x and extent.width does not equal the width of the subresource specified by dstSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 330}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-00208", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of dstOffset.y and extent.height does not equal the height of the subresource specified by dstSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-00209", 42}, { {"For each element of pRegions, if the sum of dstOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of dstImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07274", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, dstOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 259}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07275", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, dstOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 259}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07276", 42}, { {"For each element of pRegions, dstOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07966", 42}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07969", 42}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07979", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07980", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07981", 42}, { {"If dstImage has a multi-planar format, then for each element of pRegions, dstSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-07983", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, dstSubresource.baseArrayLayer must be 0 and dstSubresource.layerCount must be 1", 150}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-09109", 42}, { {"If dstImage is sparse then all memory ranges accessed by the copy command must be bound as described in Binding Resource Memory", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-09111", 42}, { {"If the stencil aspect of dstImage is accessed, and dstImage was not created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 193}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-09112", 42}, { {"If the stencil aspect of dstImage is accessed, and dstImage was created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-09113", 42}, { {"If non-stencil aspects of dstImage are accessed, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-09115", 42}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the extent member of each element of pRegions must equal the extents of dstImage identified by dstSubresource", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImage-parameter", 46}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImageLayout-09071", 48}, { {"dstImageLayout must specify the current layout of the image subresources of dstImage specified in pRegions", 106}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImageLayout-09073", 48}, { {"dstImageLayout must be one of the image layouts returned in VkPhysicalDeviceHostImageCopyProperties::pCopyDstLayouts", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstImageLayout-parameter", 52}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-09104", 43}, { {"For each element of pRegions, dstOffset.z and (extent.depth + dstOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-09114", 43}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the x, y, and z members of the dstOffset member of each element of pRegions must be 0", 134}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10051", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and dstOffset.x does not equal the width of the subresource specified by dstSubresource, dstOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 349}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10052", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and dstOffset.y does not equal the height of the subresource specified by dstSubresource, dstOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 352}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10053", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of dstOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 253}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10054", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of dstOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 253}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10055", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of dstOffset.x and extent.height does not equal the width of the subresource specified by dstSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10056", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of dstOffset.y and extent.width does not equal the height of the subresource specified by dstSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 333}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10057", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of dstOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstOffset-10058", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of dstOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 254}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstSubresource-07967", 48}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstSubresource-07968", 48}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstSubresource-07971", 48}, { {"For each element of pRegions, dstOffset.x and (extent.width + dstOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstSubresource-07972", 48}, { {"For each element of pRegions, dstOffset.y and (extent.height + dstOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-dstSubresource-09105", 48}, { {"For each element of pRegions, dstSubresource.aspectMask must specify aspects present in dstImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-flags-parameter", 43}, { {"flags must be a valid combination of VkHostImageCopyFlagBits values", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-pRegions-parameter", 46}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageCopy2 structures", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-regionCount-arraylength", 51}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_COPY_IMAGE_TO_IMAGE_INFO", 56}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-00207", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of srcOffset.x and extent.width does not equal the width of the subresource specified by srcSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 330}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-00208", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of srcOffset.y and extent.height does not equal the height of the subresource specified by srcSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-00209", 42}, { {"For each element of pRegions, if the sum of srcOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07274", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, srcOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 259}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07275", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, srcOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 259}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07276", 42}, { {"For each element of pRegions, srcOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07966", 42}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07969", 42}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07979", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07980", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07981", 42}, { {"If srcImage has a multi-planar format, then for each element of pRegions, srcSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-07983", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, srcSubresource.baseArrayLayer must be 0 and srcSubresource.layerCount must be 1", 150}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09069", 42}, { {"srcImage and dstImage must have been created with identical image creation parameters", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09109", 42}, { {"If srcImage is sparse then all memory ranges accessed by the copy command must be bound as described in Binding Resource Memory", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09111", 42}, { {"If the stencil aspect of srcImage is accessed, and srcImage was not created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 193}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09112", 42}, { {"If the stencil aspect of srcImage is accessed, and srcImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09113", 42}, { {"If non-stencil aspects of srcImage are accessed, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-09115", 42}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the extent member of each element of pRegions must equal the extents of srcImage identified by srcSubresource", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImage-parameter", 46}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImageLayout-09070", 48}, { {"srcImageLayout must specify the current layout of the image subresources of srcImage specified in pRegions", 106}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImageLayout-09072", 48}, { {"srcImageLayout must be one of the image layouts returned in VkPhysicalDeviceHostImageCopyProperties::pCopySrcLayouts", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcImageLayout-parameter", 52}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-09104", 43}, { {"For each element of pRegions, srcOffset.z and (extent.depth + srcOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-09114", 43}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the x, y, and z members of the srcOffset member of each element of pRegions must be 0", 134}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10051", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and srcOffset.x does not equal the width of the subresource specified by srcSubresource, srcOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 349}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10052", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and srcOffset.y does not equal the height of the subresource specified by srcSubresource, srcOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 352}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10053", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of srcOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 253}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10054", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of srcOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 253}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10055", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of srcOffset.x and extent.height does not equal the width of the subresource specified by srcSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10056", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of srcOffset.y and extent.width does not equal the height of the subresource specified by srcSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 333}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10057", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of srcOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcOffset-10058", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of srcOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 254}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcSubresource-07967", 48}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcSubresource-07968", 48}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcSubresource-07971", 48}, { {"For each element of pRegions, srcOffset.x and (extent.width + srcOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcSubresource-07972", 48}, { {"For each element of pRegions, srcOffset.y and (extent.height + srcOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToImageInfo-srcSubresource-09105", 48}, { {"For each element of pRegions, srcSubresource.aspectMask must specify aspects present in srcImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-flags-09394", 40}, { {"If flags includes VK_HOST_IMAGE_COPY_MEMCPY_BIT, for each region in pRegions, memoryRowLength and memoryImageHeight must both be 0", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-flags-parameter", 44}, { {"flags must be a valid combination of VkHostImageCopyFlagBits values", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-09104", 46}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of srcImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-09114", 46}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the x, y, and z members of the imageOffset member of each element of pRegions must be 0", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10051", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10052", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10053", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10054", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10055", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10056", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10057", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageOffset-10058", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageSubresource-07967", 51}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageSubresource-07968", 51}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageSubresource-07971", 51}, { {"For each element of pRegions, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of srcImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageSubresource-07972", 51}, { {"For each element of pRegions, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of srcImage", 207}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-imageSubresource-09105", 51}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in srcImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-memoryImageHeight-09107", 52}, { {"For each element of pRegions, memoryImageHeight must be a multiple of the texel block extent height of the VkFormat of srcImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-memoryRowLength-09106", 50}, { {"For each element of pRegions, memoryRowLength must be a multiple of the texel block extent width of the VkFormat of srcImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-memoryRowLength-09108", 50}, { {"For each element of pRegions, memoryRowLength divided by the texel block extent width and then multiplied by the texel block size of srcImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-pRegions-parameter", 47}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageToMemoryCopy structures", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-regionCount-arraylength", 52}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_COPY_IMAGE_TO_MEMORY_INFO", 57}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-00207", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-00208", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-00209", 43}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07274", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07275", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07276", 43}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07966", 43}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07969", 43}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07973", 43}, { {"srcImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07979", 43}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07980", 43}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07981", 43}, { {"If srcImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-07983", 43}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-09109", 43}, { {"If srcImage is sparse then all memory ranges accessed by the copy command must be bound as described in Binding Resource Memory", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-09111", 43}, { {"If the stencil aspect of srcImage is accessed, and srcImage was not created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 193}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-09112", 43}, { {"If the stencil aspect of srcImage is accessed, and srcImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-09113", 43}, { {"If non-stencil aspects of srcImage are accessed, srcImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-09115", 43}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the imageExtent member of each element of pRegions must equal the extents of srcImage identified by imageSubresource", 165}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImage-parameter", 47}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImageLayout-09064", 49}, { {"srcImageLayout must specify the current layout of the image subresources of srcImage specified in pRegions", 106}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImageLayout-09065", 49}, { {"srcImageLayout must be one of the image layouts returned in VkPhysicalDeviceHostImageCopyProperties::pCopySrcLayouts", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyImageToMemoryInfo-srcImageLayout-parameter", 53}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-dstAddress-10959", 52}, { {"The dstAddress must be 4 byte aligned", 37}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-dstAddress-10962", 52}, { {"The memory in range [dstAddress, dstAddress + size - 1] must be within the bounds of the memory allocation backing dstAddress", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-dstAddress-12212", 52}, { {"The range of memory defined by dstAddress and size must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-dstAddress-parameter", 56}, { {"dstAddress must be a valid VkDeviceAddress value", 48}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-size-10960", 46}, { {"The size must be 4 byte aligned", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-srcAddress-10958", 52}, { {"The srcAddress must be 4 byte aligned", 37}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-srcAddress-10961", 52}, { {"The memory in range [srcAddress, srcAddress + size - 1] must be within the bounds of the memory allocation backing srcAddress", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-srcAddress-12211", 52}, { {"The range of memory defined by srcAddress and size must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectCommandKHR-srcAddress-parameter", 56}, { {"srcAddress must be a valid VkDeviceAddress value", 48}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-copyAddressRange-10942", 55}, { {"copyAddressRange.address must be 4 byte aligned", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-copyAddressRange-10943", 55}, { {"copyAddressRange.stride must be a multiple of 4 and must be greater than or equal to sizeof(VkCopyMemoryIndirectCommandKHR)", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-copyAddressRange-10945", 55}, { {"Any of the source or destination memory regions specified in copyAddressRange must not overlap with any of the specified destination memory regions", 147}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-copyAddressRange-12210", 55}, { {"copyAddressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-copyCount-10944", 48}, { {"copyCount must be less than or equal to copyAddressRange.size / copyAddressRange.stride", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-dstCopyFlags-10939", 51}, { {"If dstCopyFlags contains VK_ADDRESS_COPY_SPARSE_BIT_KHR, the destination memory regions accessed must be bound to memory", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-dstCopyFlags-10941", 51}, { {"dstCopyFlags must not contain VK_ADDRESS_COPY_PROTECTED_BIT_KHR", 63}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-dstCopyFlags-parameter", 55}, { {"dstCopyFlags must be a valid combination of VkAddressCopyFlagBitsKHR values", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MEMORY_INDIRECT_INFO_KHR", 61}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-srcCopyFlags-10938", 51}, { {"If srcCopyFlags contains VK_ADDRESS_COPY_SPARSE_BIT_KHR, the source memory regions accessed must be bound to memory", 115}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-srcCopyFlags-10940", 51}, { {"srcCopyFlags must not contain VK_ADDRESS_COPY_PROTECTED_BIT_KHR", 63}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryIndirectInfoKHR-srcCopyFlags-parameter", 55}, { {"srcCopyFlags must be a valid combination of VkAddressCopyFlagBitsKHR values", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-03746", 57}, { {"dst must have been created with a size greater than or equal to that used to serialize the data in src", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-11584", 57}, { {"If the serialized acceleration structure in src is bottom-level, and any block in the header has a type of VK_ACCELERATION_STRUCTURE_SERIALIZED_BLOCK_TYPE_OPACITY_MICROMAP_KHR the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature must be enabled", 256}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-11585", 57}, { {"If the serialized acceleration structure in src is bottom-level, every device address in each block in the header with type VK_ACCELERATION_STRUCTURE_SERIALIZED_BLOCK_TYPE_OPACITY_MICROMAP_KHR must point to a valid VkDeviceAddress that was retrieved by vkGetBufferDeviceAddressKHR for the underlying buffer objects of compatible micromaps constructed on the VkDevice when the command is executed", 395}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-11716", 57}, { {"The range of src accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-11717", 57}, { {"The range of dst accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-dst-parameter", 61}, { {"dst must be a valid VkAccelerationStructureKHR handle", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-mode-03413", 58}, { {"mode must be VK_COPY_ACCELERATION_STRUCTURE_MODE_DESERIALIZE_KHR", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-mode-parameter", 62}, { {"mode must be a valid VkCopyAccelerationStructureModeKHR value", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-pInfo-03414", 59}, { {"The data in src must have a format compatible with the destination physical device as returned by vkGetDeviceAccelerationStructureCompatibilityKHR", 146}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-pNext-pNext", 59}, { {"pNext must be NULL", 18}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MEMORY_TO_ACCELERATION_STRUCTURE_INFO_KHR", 78}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-src-04960", 57}, { {"The source memory pointed to by src must contain data previously serialized using vkCmdCopyAccelerationStructureToMemoryKHR, potentially modified to relocate acceleration structure references as described in that command", 220}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToAccelerationStructureInfoKHR-src-11583", 57}, { {"If the serialized acceleration structure in src is bottom-level, each block in the header must have a type that corresponds to a valid value in VkAccelerationStructureSerializedBlockTypeKHR", 189}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-bufferImageHeight-10965", 66}, { {"bufferImageHeight must be 0, or greater than or equal to the height member of imageExtent", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-bufferRowLength-10964", 64}, { {"bufferRowLength must be 0, or greater than or equal to the width member of imageExtent", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-dstImage-12282", 57}, { {"If dstImage does not have either a depth/stencil format or a multi-planar format, srcAddress must be a multiple of the texel block size", 135}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-dstImage-12283", 57}, { {"If dstImage has a multi-planar format, srcAddress must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 238}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-dstImage-12285", 57}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, imageOffset.y must be 0 and imageExtent.height must be 1", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-dstImage-12286", 57}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, imageOffset.z must be 0 and imageExtent.depth must be 1", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageExtent-10967", 60}, { {"imageExtent must specify a valid region in the destination image and can be 0", 77}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageOffset-10966", 60}, { {"imageOffset must specify a valid offset in the destination image", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageOffset-10969", 60}, { {"The imageOffset and imageExtent members of each region must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 199}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageOffset-10970", 60}, { {"For each destination region, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified subresource", 187}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageOffset-10971", 60}, { {"For each destination region, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified subresource", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageSubresource-12284", 65}, { {"The members of imageSubresource must be identical to the members of the VkImageSubresourceLayers structure specified in the corresponding index of the pCopyMemoryToImageIndirectInfo->pImageSubresources array of vkCmdCopyMemoryToImageIndirectKHR during command recording", 269}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-imageSubresource-parameter", 69}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-srcAddress-10963", 59}, { {"The srcAddress must be 4 byte aligned", 37}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-srcAddress-10968", 59}, { {"The memory region starting at srcAddress and described by bufferRowLength and bufferImageHeight must not exceed the bounds of the memory allocation backing memory at srcAddress", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-srcAddress-12214", 59}, { {"srcAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 143}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectCommandKHR-srcAddress-parameter", 63}, { {"srcAddress must be a valid VkDeviceAddress value", 48}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-aspectMask-07662", 56}, { {"The aspectMask member for every subresource in pImageSubresources must only have a single bit set", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-aspectMask-12287", 56}, { {"The aspectMask member for every subresource in pImageSubresources must specify an aspect present in dstImage", 108}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-commandBuffer-07674", 59}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, for each region, the aspectMask member of pImageSubresources must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyAddressRange-10952", 62}, { {"copyAddressRange.address must be 4 byte aligned", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyAddressRange-10953", 62}, { {"copyAddressRange.stride must be a multiple of 4 and must be greater than or equal to sizeof(VkCopyMemoryToImageIndirectCommandKHR)", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyAddressRange-10975", 62}, { {"Any of the source or destination memory regions specified in copyAddressRange must not overlap with any of the specified destination memory regions at the time this command is executed on device", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyAddressRange-12213", 62}, { {"copyAddressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyCount-10951", 55}, { {"copyCount must be less than or equal to copyAddressRange.size / copyAddressRange.stride", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-copyCount-arraylength", 61}, { {"copyCount must be greater than 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-07661", 54}, { {"dstImage must not be a protected image", 38}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-07664", 54}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-07665", 54}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-07673", 54}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-07973", 54}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-10955", 54}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_2_COPY_IMAGE_INDIRECT_DST_BIT_KHR", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-10974", 54}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-12288", 54}, { {"If dstImage is not of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, the specified baseArrayLayer + layerCount of each region in pImageSubresources must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 326}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-12289", 54}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, for each destination region, (imageSubresource.baseArrayLayer + imageSubresource.layerCount) must be less than or equal to the depth of the specified subresource", 302}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-12290", 54}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, for each destination region, if (imageSubresource.baseArrayLayer + imageSubresource.layerCount) does not equal the depth of the specified subresource, imageSubresource.layerCount must be a multiple of the texel block extent depth of the VkFormat of dstImage", 398}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-12291", 54}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each destination region, imageSubresource.baseArrayLayer must be a multiple of the texel block extent depth of the VkFormat of dstImage", 180}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-12292", 54}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each destination region, imageSubresource.baseArrayLayer must be less than or equal to the depth of the specified subresource", 170}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImage-parameter", 58}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImageLayout-07667", 60}, { {"dstImageLayout must specify the layout of the image subresources of dstImage at the time this command is executed on a VkDevice", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImageLayout-07669", 60}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-dstImageLayout-parameter", 64}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-mipLevel-07670", 54}, { {"The specified mipLevel of each region in pImageSubresources must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-pImageSubresources-parameter", 68}, { {"pImageSubresources must be a valid pointer to an array of copyCount valid VkImageSubresourceLayers structures", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INDIRECT_INFO_KHR", 70}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-srcCopyFlags-10950", 58}, { {"If srcCopyFlags contains VK_ADDRESS_COPY_SPARSE_BIT_KHR, the source memory regions accessed must be bound to memory", 115}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageIndirectInfoKHR-srcCopyFlags-parameter", 62}, { {"srcCopyFlags must be a valid combination of VkAddressCopyFlagBitsKHR values", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-00207", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-00208", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-00209", 43}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of dstImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07274", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07275", 43}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07276", 43}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07966", 43}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07969", 43}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07973", 43}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07979", 43}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07980", 43}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07981", 43}, { {"If dstImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-07983", 43}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-09109", 43}, { {"If dstImage is sparse then all memory ranges accessed by the copy command must be bound as described in Binding Resource Memory", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-09111", 43}, { {"If the stencil aspect of dstImage is accessed, and dstImage was not created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 193}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-09112", 43}, { {"If the stencil aspect of dstImage is accessed, and dstImage was created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-09113", 43}, { {"If non-stencil aspects of dstImage are accessed, dstImage must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 137}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-09115", 43}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the imageExtent member of each element of pRegions must equal the extents of dstImage identified by imageSubresource", 165}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImage-parameter", 47}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImageLayout-09059", 49}, { {"dstImageLayout must specify the current layout of the image subresources of dstImage specified in pRegions", 106}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImageLayout-09060", 49}, { {"dstImageLayout must be one of the image layouts returned in VkPhysicalDeviceHostImageCopyProperties::pCopyDstLayouts", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-dstImageLayout-parameter", 53}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-flags-09393", 40}, { {"If flags includes VK_HOST_IMAGE_COPY_MEMCPY_BIT, for each region in pRegions, memoryRowLength and memoryImageHeight must both be 0", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-flags-parameter", 44}, { {"flags must be a valid combination of VkHostImageCopyFlagBits values", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-09104", 46}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of dstImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-09114", 46}, { {"If flags contains VK_HOST_IMAGE_COPY_MEMCPY_BIT, the x, y, and z members of the imageOffset member of each element of pRegions must be 0", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10051", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10052", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10053", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10054", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10055", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10056", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10057", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageOffset-10058", 46}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageSubresource-07967", 51}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageSubresource-07968", 51}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageSubresource-07971", 51}, { {"For each element of pRegions, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of dstImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageSubresource-07972", 51}, { {"For each element of pRegions, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of dstImage", 207}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-imageSubresource-09105", 51}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in dstImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-memoryImageHeight-09107", 52}, { {"For each element of pRegions, memoryImageHeight must be a multiple of the texel block extent height of the VkFormat of dstImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-memoryRowLength-09106", 50}, { {"For each element of pRegions, memoryRowLength must be a multiple of the texel block extent width of the VkFormat of dstImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-memoryRowLength-09108", 50}, { {"For each element of pRegions, memoryRowLength divided by the texel block extent width and then multiplied by the texel block size of dstImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-pRegions-parameter", 47}, { {"pRegions must be a valid pointer to an array of regionCount valid VkMemoryToImageCopy structures", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-regionCount-arraylength", 52}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToImageInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INFO", 57}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-dst-07550", 44}, { {"dst must have been created with a size greater than or equal to that used to serialize the data in src", 102}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-dst-parameter", 48}, { {"dst must be a valid VkMicromapEXT handle", 40}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-mode-07548", 45}, { {"mode must be VK_COPY_MICROMAP_MODE_DESERIALIZE_EXT", 50}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-mode-parameter", 49}, { {"mode must be a valid VkCopyMicromapModeEXT value", 48}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MEMORY_TO_MICROMAP_INFO_EXT", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-src-07547", 44}, { {"The source memory pointed to by src must contain data previously serialized using vkCmdCopyMicromapToMemoryEXT", 110}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMemoryToMicromapInfoEXT-src-07549", 44}, { {"The data in src must have a format compatible with the destination physical device as returned by vkGetDeviceMicromapCompatibilityEXT", 133}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-commonparent", 39}, { {"Both of dst, and src must have been created, allocated, or retrieved from the same VkDevice", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-dst-parameter", 40}, { {"dst must be a valid VkMicromapEXT handle", 40}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-mode-07531", 37}, { {"mode must be VK_COPY_MICROMAP_MODE_COMPACT_EXT or VK_COPY_MICROMAP_MODE_CLONE_EXT", 81}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-mode-07533", 37}, { {"If mode is VK_COPY_MICROMAP_MODE_COMPACT_EXT, src must have been constructed with VK_BUILD_MICROMAP_ALLOW_COMPACTION_BIT_EXT in the build", 137}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MICROMAP_INFO_EXT", 54}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapInfoEXT-src-parameter", 40}, { {"src must be a valid VkMicromapEXT handle", 40}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-dst-07541", 44}, { {"The memory pointed to by dst must be at least as large as the serialization size of src, as reported by vkWriteMicromapsPropertiesEXT or vkCmdWriteMicromapsPropertiesEXT with a query type of VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT", 236}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-mode-07542", 45}, { {"mode must be VK_COPY_MICROMAP_MODE_SERIALIZE_EXT", 48}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-mode-parameter", 49}, { {"mode must be a valid VkCopyMicromapModeEXT value", 48}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_COPY_MICROMAP_TO_MEMORY_INFO_EXT", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-src-07540", 44}, { {"The source micromap src must have been constructed prior to the execution of this command", 89}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyMicromapToMemoryInfoEXT-src-parameter", 48}, { {"src must be a valid VkMicromapEXT handle", 40}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkCopyTensorInfoARM-commonparent", 37}, { {"Both of dstTensor, and srcTensor must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-dimensionCount-09684", 45}, { {"srcTensor and dstTensor must have been created with equal values for VkTensorDescriptionARM::dimensionCount", 107}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-dstTensor-09692", 40}, { {"The format features of dstTensor must contain VK_FORMAT_FEATURE_2_TRANSFER_DST_BIT", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-dstTensor-09693", 40}, { {"dstTensor must have been created with the VK_TENSOR_USAGE_TRANSFER_DST_BIT_ARM usage flag set", 93}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-dstTensor-09695", 40}, { {"If dstTensor is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-dstTensor-parameter", 44}, { {"dstTensor must be a valid VkTensorARM handle", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pDimensions-09685", 42}, { {"For each of the elements of VkTensorDescriptionARM::pDimensions, srcTensor and dstTensor must be the same size", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pRegions-09687", 39}, { {"Each element of pRegions must be a VkTensorCopyARM structure whose pSrcOffset is NULL or has all its elements equal to 0", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pRegions-09688", 39}, { {"Each element of pRegions must be a VkTensorCopyARM structure whose pDstOffset is NULL or has all its elements equal to 0", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pRegions-09689", 39}, { {"Each element of pRegions must be a VkTensorCopyARM structure whose pExtent is NULL or equal to the VkTensorDescriptionARM::pDimensions array specified when srcTensor and dstTensor were created", 192}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pRegions-09954", 39}, { {"Each element of pRegions must be a VkTensorCopyARM structure whose dimensionCount, if it is not equal to 0, is equal to the largest of the VkTensorDescriptionARM::dimensionCount of srcTensor or dstTensor", 203}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-pRegions-parameter", 43}, { {"pRegions must be a valid pointer to an array of regionCount valid VkTensorCopyARM structures", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-regionCount-09686", 42}, { {"regionCount must be equal to 1", 30}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-regionCount-arraylength", 48}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_COPY_TENSOR_INFO_ARM", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-srcTensor-09690", 40}, { {"The format features of srcTensor must contain VK_FORMAT_FEATURE_2_TRANSFER_SRC_BIT", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-srcTensor-09691", 40}, { {"srcTensor must have been created with the VK_TENSOR_USAGE_TRANSFER_SRC_BIT_ARM usage flag set", 93}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-srcTensor-09694", 40}, { {"If srcTensor is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCopyTensorInfoARM-srcTensor-parameter", 44}, { {"srcTensor must be a valid VkTensorARM handle", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-VkCuFunctionCreateInfoNVX-module-parameter", 47}, { {"module must be a valid VkCuModuleNVX handle", 43}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuFunctionCreateInfoNVX-pName-parameter", 46}, { {"pName must be a null-terminated UTF-8 string", 44}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuFunctionCreateInfoNVX-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuFunctionCreateInfoNVX-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_CU_FUNCTION_CREATE_INFO_NVX", 59}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuLaunchInfoNVX-function-parameter", 41}, { {"function must be a valid VkCuFunctionNVX handle", 47}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuLaunchInfoNVX-pExtras-parameter", 40}, { {"If extraCount is not 0, pExtras must be a valid pointer to an array of extraCount bytes", 87}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuLaunchInfoNVX-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuLaunchInfoNVX-pParams-parameter", 40}, { {"If paramCount is not 0, pParams must be a valid pointer to an array of paramCount bytes", 87}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuLaunchInfoNVX-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_CU_LAUNCH_INFO_NVX", 50}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuModuleCreateInfoNVX-pData-parameter", 44}, { {"If dataSize is not 0, pData must be a valid pointer to an array of dataSize bytes", 81}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuModuleCreateInfoNVX-pNext-pNext", 40}, { {"pNext must be NULL or a pointer to a valid instance of VkCuModuleTexturingModeCreateInfoNVX", 91}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuModuleCreateInfoNVX-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_CU_MODULE_CREATE_INFO_NVX", 57}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuModuleCreateInfoNVX-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCuModuleTexturingModeCreateInfoNVX-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_CU_MODULE_TEXTURING_MODE_CREATE_INFO_NVX", 72}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkCudaFunctionCreateInfoNV-module-parameter", 48}, { {"module must be a valid VkCudaModuleNV handle", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaFunctionCreateInfoNV-pName-parameter", 47}, { {"pName must be a null-terminated UTF-8 string", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaFunctionCreateInfoNV-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaFunctionCreateInfoNV-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_CUDA_FUNCTION_CREATE_INFO_NV", 60}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaLaunchInfoNV-extraCount-09411", 40}, { {"extraCount must be 0", 20}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-function-parameter", 42}, { {"function must be a valid VkCudaFunctionNV handle", 48}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-gridDimX-09406", 38}, { {"gridDimX must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0]", 90}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-gridDimY-09407", 38}, { {"gridDimY must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1]", 90}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-gridDimZ-09408", 38}, { {"gridDimZ must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2]", 90}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-pExtras-09412", 37}, { {"pExtras must be NULL", 20}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-pNext-pNext", 35}, { {"pNext must be NULL", 18}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-pParams-09410", 37}, { {"pParams must be a pointer to a table of paramCount parameters, corresponding to the arguments of function", 105}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-paramCount-09409", 40}, { {"paramCount must be the total amount of parameters listed in the pParams table", 77}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaLaunchInfoNV-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_CUDA_LAUNCH_INFO_NV", 51}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkCudaModuleCreateInfoNV-dataSize-09413", 44}, { {"dataSize must be the total size in bytes of the PTX files or binary cache passed to pData", 89}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaModuleCreateInfoNV-dataSize-arraylength", 50}, { {"dataSize must be greater than 0", 31}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaModuleCreateInfoNV-pData-parameter", 45}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaModuleCreateInfoNV-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCudaModuleCreateInfoNV-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_CUDA_MODULE_CREATE_INFO_NV", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-colorAttachmentCount-11507", 60}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-depthAttachmentFormat-11508", 61}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a depth component", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-depthAttachmentFormat-11509", 61}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 163}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-depthAttachmentFormat-11513", 61}, { {"If depthAttachmentFormat is not VK_FORMAT_UNDEFINED and stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, depthAttachmentFormat must equal stencilAttachmentFormat", 164}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-pColorAttachmentFormats-11510", 63}, { {"If any element of pColorAttachmentFormats is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT , or VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV if the linearColorAttachment feature is enabled", 276}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_CUSTOM_RESOLVE_CREATE_INFO_EXT", 62}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-stencilAttachmentFormat-11511", 63}, { {"If stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a stencil aspect", 105}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkCustomResolveCreateInfoEXT-stencilAttachmentFormat-11512", 63}, { {"If stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 165}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkD3D12FenceSubmitInfoKHR-pSignalSemaphoreValues-parameter", 63}, { {"If signalSemaphoreValuesCount is not 0, and pSignalSemaphoreValues is not NULL, pSignalSemaphoreValues must be a valid pointer to an array of signalSemaphoreValuesCount uint64_t values", 184}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkD3D12FenceSubmitInfoKHR-pWaitSemaphoreValues-parameter", 61}, { {"If waitSemaphoreValuesCount is not 0, and pWaitSemaphoreValues is not NULL, pWaitSemaphoreValues must be a valid pointer to an array of waitSemaphoreValuesCount uint64_t values", 176}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkD3D12FenceSubmitInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_D3D12_FENCE_SUBMIT_INFO_KHR", 59}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkD3D12FenceSubmitInfoKHR-signalSemaphoreValuesCount-00080", 63}, { {"signalSemaphoreValuesCount must be the same value as VkSubmitInfo::signalSemaphoreCount, where this structure is in the pNext chain of a VkSubmitInfo structure", 159}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkD3D12FenceSubmitInfoKHR-waitSemaphoreValuesCount-00079", 61}, { {"waitSemaphoreValuesCount must be the same value as VkSubmitInfo::waitSemaphoreCount, where this structure is in the pNext chain of a VkSubmitInfo structure", 155}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDataGraphOpticalFlowImageFormatInfoARM-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_OPTICAL_FLOW_IMAGE_FORMAT_INFO_ARM", 77}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphOpticalFlowImageFormatInfoARM-usage-parameter", 61}, { {"usage must be a valid combination of VkDataGraphOpticalFlowImageUsageFlagBitsARM values", 87}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphOpticalFlowImageFormatInfoARM-usage-requiredbitmask", 67}, { {"usage must not be 0", 19}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphOpticalFlowImageFormatPropertiesARM-pNext-pNext", 63}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphOpticalFlowImageFormatPropertiesARM-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_ARM", 83}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineBuiltinModelCreateInfoQCOM-pOperation-11842", 67}, { {"All members of pOperation must be identical to a VkQueueFamilyDataGraphPropertiesARM::operation retrieved from vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with the physicalDevice that was used to create device and paired in the retrieved results with a VkQueueFamilyDataGraphPropertiesARM::engine identical to an element of VkDataGraphProcessingEngineCreateInfoARM::pProcessingEngines provided in the pNext chain", 421}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineBuiltinModelCreateInfoQCOM-pOperation-parameter", 71}, { {"pOperation must be a valid pointer to a valid VkPhysicalDeviceDataGraphOperationSupportARM structure", 100}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineBuiltinModelCreateInfoQCOM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_BUILTIN_MODEL_CREATE_INFO_QCOM", 82}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCompilerControlCreateInfoARM-pVendorOptions-parameter", 77}, { {"pVendorOptions must be a null-terminated UTF-8 string", 53}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCompilerControlCreateInfoARM-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_COMPILER_CONTROL_CREATE_INFO_ARM", 84}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-id-09850", 44}, { {"If the pNext chain of this structure includes a VkTensorDescriptionARM structure, then its usage member must contain VK_TENSOR_USAGE_DATA_GRAPH_BIT_ARM", 151}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pConstantData-parameter", 59}, { {"pConstantData must be a pointer value", 37}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-09775", 47}, { {"If the pNext chain of this structure includes one or more VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM structures then it must also include a VkTensorDescriptionARM structure", 192}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-09776", 47}, { {"If the pNext chain of this structure includes one or more VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM structures then, for each structure, VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM::dimension must be less than VkTensorDescriptionARM::dimensionCount", 288}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-09777", 47}, { {"If the pNext chain of this structure includes a VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM structure then, for each structure, VkTensorDescriptionARM::pDimensions[VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM::dimension] must be a multiple of VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM::groupSize", 353}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-09870", 47}, { {"If the pNext chain of this structure includes multiple VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM structures then no two structures may have their VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM::dimension member set to the same value", 269}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-09917", 47}, { {"If the pNext chain of this structure includes a VkTensorDescriptionARM structure, then its tiling member must be VK_TENSOR_TILING_LINEAR_ARM", 140}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-pNext-pNext", 47}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM or VkTensorDescriptionARM", 219}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CONSTANT_ARM", 64}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantARM-sType-unique", 48}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM", 172}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM-sType-sType", 79}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CONSTANT_TENSOR_SEMI_STRUCTURED_SPARSITY_INFO_ARM", 101}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-None-11840", 48}, { {"If a VkDataGraphPipelineIdentifierCreateInfoARM or a VkDataGraphPipelineBuiltinModelCreateInfoQCOM structure is included in the pNext chain, then flags must contain VK_PIPELINE_CREATE_2_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT", 223}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-None-12363", 48}, { {"If a VkDataGraphPipelineIdentifierCreateInfoARM or a VkDataGraphPipelineBuiltinModelCreateInfoQCOM structure is included in the pNext chain, then resourceInfoCount must be 0", 173}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-None-12365", 48}, { {"If neither a VkDataGraphPipelineIdentifierCreateInfoARM nor a VkDataGraphPipelineBuiltinModelCreateInfoQCOM structure are included in the pNext chain, then resourceInfoCount must be greater than 0", 196}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-dataGraphDescriptorBuffer-09885", 69}, { {"If the dataGraphDescriptorBuffer feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_DESCRIPTOR_BUFFER_BIT_EXT", 126}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-dataGraphShaderModule-09886", 65}, { {"If the dataGraphShaderModule feature is not enabled, a VkDataGraphPipelineShaderModuleCreateInfoARM structure must not be included in the pNext chain", 149}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-dataGraphUpdateAfterBind-09768", 68}, { {"If the dataGraphUpdateAfterBind feature is not enabled, layout must not use any VkDescriptorSetLayout object created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set", 192}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-flags-09764", 49}, { {"flags may only contain VK_PIPELINE_CREATE_2_NO_PROTECTED_ACCESS_BIT_EXT, VK_PIPELINE_CREATE_2_PROTECTED_ACCESS_ONLY_BIT_EXT, VK_PIPELINE_CREATE_2_DISABLE_OPTIMIZATION_BIT, VK_PIPELINE_CREATE_2_DESCRIPTOR_BUFFER_BIT_EXT, VK_PIPELINE_CREATE_2_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_KHR or VK_PIPELINE_CREATE_2_EARLY_RETURN_ON_FAILURE_BIT_KHR", 338}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-flags-09773", 49}, { {"flags must not include both VK_PIPELINE_CREATE_2_NO_PROTECTED_ACCESS_BIT_EXT and VK_PIPELINE_CREATE_2_PROTECTED_ACCESS_ONLY_BIT_EXT", 131}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-flags-09848", 49}, { {"If flags contains VK_PIPELINE_CREATE_2_PROTECTED_ACCESS_ONLY_BIT_EXT then a VkDataGraphPipelineNeuralStatisticsCreateInfoARM structure must not be included in the pNext chain.", 175}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-flags-parameter", 53}, { {"flags must be a valid combination of VkPipelineCreateFlagBits2 values", 69}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-layout-09767", 50}, { {"layout must have been created with pushConstantRangeCount equal to 0 and pPushConstantRanges equal to NULL", 106}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-layout-09769", 50}, { {"If a VkDataGraphPipelineShaderModuleCreateInfoARM structure is included in the pNext chain and a resource variable is declared in the shader module, the corresponding descriptor binding used to create layout must have a descriptorType that corresponds to the type of the resource variable", 288}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-layout-09934", 50}, { {"If a VkDataGraphPipelineShaderModuleCreateInfoARM structure is included in the pNext chain and an array resource variable is declared in the shader module, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 299}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-layout-parameter", 54}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pNext-09804", 49}, { {"If the pNext chain includes an VkPipelineCreationFeedbackCreateInfo structure, then its pipelineStageCreationFeedbackCount must be 0", 132}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pNext-09948", 49}, { {"If a VkDataGraphProcessingEngineCreateInfoARM structure is included in the pNext chain, each member of pProcessingEngines must be identical to an VkQueueFamilyDataGraphPropertiesARM::engine retrieved from vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with the physicalDevice that was used to create device", 312}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pNext-09949", 49}, { {"If a VkDataGraphProcessingEngineCreateInfoARM structure is not included in the pNext chain, VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_DEFAULT_ARM must be set in an VkQueueFamilyDataGraphPropertiesARM::engine retrieved from vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with the physicalDevice that was used to create device", 341}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pNext-09977", 49}, { {"One and only one of the following structures must be included in the pNext chain: VkDataGraphPipelineShaderModuleCreateInfoARM VkDataGraphPipelineIdentifierCreateInfoARM VkDataGraphPipelineBuiltinModelCreateInfoQCOM VkDataGraphPipelineSingleNodeCreateInfoARM", 258}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pNext-pNext", 49}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDataGraphPipelineBuiltinModelCreateInfoQCOM, VkDataGraphPipelineCompilerControlCreateInfoARM, VkDataGraphPipelineIdentifierCreateInfoARM, VkDataGraphPipelineNeuralStatisticsCreateInfoARM, VkDataGraphPipelineOpticalFlowCreateInfoARM, VkDataGraphPipelineShaderModuleCreateInfoARM, VkDataGraphPipelineSingleNodeCreateInfoARM, VkDataGraphProcessingEngineCreateInfoARM, VkPipelineCreationFeedbackCreateInfo, or VkShaderModuleCreateInfo", 563}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pResourceInfos-parameter", 62}, { {"If resourceInfoCount is not 0, pResourceInfos must be a valid pointer to an array of resourceInfoCount valid VkDataGraphPipelineResourceInfoARM structures", 154}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pSetLayouts-09770", 55}, { {"The descriptor set layouts in VkPipelineLayoutCreateInfo::pSetLayouts used to create layout must not include any VkDescriptorSetLayoutBinding whose descriptor type is VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 197}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pipelineCreationCacheControl-09871", 72}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_2_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_KHR or VK_PIPELINE_CREATE_2_EARLY_RETURN_ON_FAILURE_BIT_KHR", 201}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-pipelineProtectedAccess-09772", 67}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_2_NO_PROTECTED_ACCESS_BIT_EXT or VK_PIPELINE_CREATE_2_PROTECTED_ACCESS_ONLY_BIT_EXT", 180}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-resourceInfoCount-12364", 61}, { {"If resourceInfoCount is equal to 0, then pResourceInfos must equal NULL", 71}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CREATE_INFO_ARM", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineCreateInfoARM-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineDispatchInfoARM-flags-zerobitmask", 57}, { {"flags must be 0", 15}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineDispatchInfoARM-pNext-pNext", 51}, { {"pNext must be NULL or a pointer to a valid instance of VkDataGraphPipelineOpticalFlowDispatchInfoARM", 100}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineDispatchInfoARM-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_DISPATCH_INFO_ARM", 69}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineDispatchInfoARM-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineIdentifierCreateInfoARM-identifierSize-arraylength", 74}, { {"identifierSize must be greater than 0", 37}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineIdentifierCreateInfoARM-pIdentifier-parameter", 69}, { {"pIdentifier must be a valid pointer to an array of identifierSize uint8_t values", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineIdentifierCreateInfoARM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_IDENTIFIER_CREATE_INFO_ARM", 78}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineInfoARM-dataGraphPipeline-09803", 55}, { {"dataGraphPipeline must have been created with vkCreateDataGraphPipelinesARM", 75}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineInfoARM-dataGraphPipeline-parameter", 59}, { {"dataGraphPipeline must be a valid VkPipeline handle", 51}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineInfoARM-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineInfoARM-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_INFO_ARM", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineNeuralStatisticsCreateInfoARM-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_NEURAL_STATISTICS_CREATE_INFO_ARM", 85}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-costFormat-09970", 65}, { {"costFormat must be one of the formats returned by vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM for VK_DATA_GRAPH_OPTICAL_FLOW_IMAGE_USAGE_COST_BIT_ARM if VK_DATA_GRAPH_OPTICAL_FLOW_CREATE_ENABLE_COST_BIT_ARM is set in flags", 244}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-costFormat-parameter", 69}, { {"If costFormat is not 0, costFormat must be a valid VkFormat value", 65}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-flags-09974", 60}, { {"VK_DATA_GRAPH_OPTICAL_FLOW_CREATE_ENABLE_HINT_BIT_ARM must not be set in flags if VkQueueFamilyDataGraphOpticalFlowPropertiesARM::hintSupported is VK_FALSE", 155}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-flags-09975", 60}, { {"VK_DATA_GRAPH_OPTICAL_FLOW_CREATE_ENABLE_COST_BIT_ARM must not be set in flags if VkQueueFamilyDataGraphOpticalFlowPropertiesARM::costSupported is VK_FALSE", 155}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-flags-parameter", 64}, { {"flags must be a valid combination of VkDataGraphOpticalFlowCreateFlagBitsARM values", 83}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-flowVectorFormat-09969", 71}, { {"flowVectorFormat must be one of the formats returned by vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM for VK_DATA_GRAPH_OPTICAL_FLOW_IMAGE_USAGE_OUTPUT_BIT_ARM", 179}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-flowVectorFormat-parameter", 75}, { {"flowVectorFormat must be a valid VkFormat value", 47}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-height-09967", 61}, { {"height must be greater than or equal to VkQueueFamilyDataGraphOpticalFlowPropertiesARM::minHeight and less than or equal to VkQueueFamilyDataGraphOpticalFlowPropertiesARM::maxHeight", 181}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-hintGridSize-09972", 67}, { {"hintGridSize must be 0 or exactly one of the bits reported in VkQueueFamilyDataGraphOpticalFlowPropertiesARM::supportedHintGridSizes if VK_DATA_GRAPH_OPTICAL_FLOW_CREATE_ENABLE_HINT_BIT_ARM is set in flags", 205}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-hintGridSize-09973", 67}, { {"hintGridSize must be the same as outputGridSize if hintGridSize is not 0", 72}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-imageFormat-09968", 66}, { {"imageFormat must be one of the formats returned by vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM for VK_DATA_GRAPH_OPTICAL_FLOW_IMAGE_USAGE_INPUT_BIT_ARM", 173}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-imageFormat-parameter", 70}, { {"imageFormat must be a valid VkFormat value", 42}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-outputGridSize-09971", 69}, { {"outputGridSize must be exactly one of the bits reported in VkQueueFamilyDataGraphOpticalFlowPropertiesARM::supportedOutputGridSizes", 131}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-performanceLevel-parameter", 75}, { {"If performanceLevel is not 0, performanceLevel must be a valid VkDataGraphOpticalFlowPerformanceLevelARM value", 110}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_OPTICAL_FLOW_CREATE_INFO_ARM", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowCreateInfoARM-width-09966", 60}, { {"width must be greater than or equal to VkQueueFamilyDataGraphOpticalFlowPropertiesARM::minWidth and less than or equal to VkQueueFamilyDataGraphOpticalFlowPropertiesARM::maxWidth", 178}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowDispatchInfoARM-flags-parameter", 66}, { {"flags must be a valid combination of VkDataGraphOpticalFlowExecuteFlagBitsARM values", 84}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowDispatchInfoARM-meanFlowL1NormHint-09976", 75}, { {"meanFlowL1NormHint, when different from 0, must be less than or equal to the maximum of the width or height of the input image provided at pipeline creation time via VkDataGraphPipelineOpticalFlowCreateInfoARM::width or VkDataGraphPipelineOpticalFlowCreateInfoARM::height, respectively", 285}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineOpticalFlowDispatchInfoARM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_OPTICAL_FLOW_DISPATCH_INFO_ARM", 82}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelinePropertyQueryResultARM-pData-parameter", 62}, { {"If dataSize is not 0, and pData is not NULL, pData must be a valid pointer to an array of dataSize bytes", 104}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelinePropertyQueryResultARM-pNext-pNext", 58}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelinePropertyQueryResultARM-property-parameter", 65}, { {"property must be a valid VkDataGraphPipelinePropertyARM value", 61}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelinePropertyQueryResultARM-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_PROPERTY_QUERY_RESULT_ARM", 77}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoARM-descriptorSet-09851", 59}, { {"If the pNext chain of this structure includes a VkTensorDescriptionARM structure, then its usage must contain VK_TENSOR_USAGE_DATA_GRAPH_BIT_ARM", 144}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoARM-descriptorSet-09962", 59}, { {"If descriptorSet and binding identify an image resource or an array of image resources, then a VkDataGraphPipelineResourceInfoImageLayoutARM structure must be included in the pNext chain , unless the unifiedImageLayouts feature is enabled", 238}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoARM-pNext-pNext", 51}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDataGraphPipelineResourceInfoImageLayoutARM or VkTensorDescriptionARM", 202}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoARM-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_RESOURCE_INFO_ARM", 69}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoARM-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoImageLayoutARM-layout-parameter", 67}, { {"layout must be a valid VkImageLayout value", 42}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineResourceInfoImageLayoutARM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_RESOURCE_INFO_IMAGE_LAYOUT_ARM", 82}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionBindPointRequirementARM-pNext-pNext", 66}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionBindPointRequirementARM-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_REQUIREMENT_ARM", 86}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionBindPointRequirementsInfoARM-pNext-pNext", 71}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionBindPointRequirementsInfoARM-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_REQUIREMENTS_INFO_ARM", 92}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionBindPointRequirementsInfoARM-session-parameter", 77}, { {"session must be a valid VkDataGraphPipelineSessionARM handle", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-dataGraphPipeline-09781", 68}, { {"dataGraphPipeline must have been obtained via a call to vkCreateDataGraphPipelinesARM", 85}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-dataGraphPipeline-parameter", 72}, { {"dataGraphPipeline must be a valid VkPipeline handle", 51}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-flags-09853", 56}, { {"If flags contains VK_DATA_GRAPH_PIPELINE_SESSION_CREATE_PROTECTED_BIT_ARM then a VkDataGraphPipelineNeuralStatisticsCreateInfoARM structure must not be included in the pNext chain", 179}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-flags-parameter", 60}, { {"flags must be a valid combination of VkDataGraphPipelineSessionCreateFlagBitsARM values", 87}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-pNext-09852", 56}, { {"A VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM structure may only be included in the pNext chain if and only if a VkDataGraphPipelineNeuralStatisticsCreateInfoARM structure whose allowNeuralStatistics member was VK_TRUE was included in the pNext chain of the VkDataGraphPipelineCreateInfoARM structure used to create dataGraphPipeline", 345}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-pNext-pNext", 56}, { {"pNext must be NULL or a pointer to a valid instance of VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM", 110}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-protectedMemory-09782", 66}, { {"If the protectedMemory feature is not enabled, flags must not contain VK_DATA_GRAPH_PIPELINE_SESSION_CREATE_PROTECTED_BIT_ARM", 125}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_CREATE_INFO_ARM", 75}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionCreateInfoARM-sType-unique", 57}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionMemoryRequirementsInfoARM-bindPoint-parameter", 76}, { {"bindPoint must be a valid VkDataGraphPipelineSessionBindPointARM value", 70}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionMemoryRequirementsInfoARM-objectIndex-09855", 74}, { {"objectIndex must be less than the number of objects returned by vkGetDataGraphPipelineSessionBindPointRequirementsARM via VkDataGraphPipelineSessionBindPointRequirementARM::numObjects with VkDataGraphPipelineSessionMemoryRequirementsInfoARM::bindPoint equal to bindPoint", 270}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionMemoryRequirementsInfoARM-pNext-pNext", 68}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionMemoryRequirementsInfoARM-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_MEMORY_REQUIREMENTS_INFO_ARM", 88}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionMemoryRequirementsInfoARM-session-parameter", 74}, { {"session must be a valid VkDataGraphPipelineSessionARM handle", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM-mode-parameter", 75}, { {"mode must be a valid VkNeuralAcceleratorStatisticsModeARM value", 63}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_NEURAL_STATISTICS_CREATE_INFO_ARM", 93}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-dataGraphSpecializationConstants-09849", 88}, { {"If the dataGraphSpecializationConstants feature is not enabled then pSpecializationInfo must be NULL and module must not contain any OpSpec* instructions", 153}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-module-parameter", 66}, { {"If module is not VK_NULL_HANDLE, module must be a valid VkShaderModule handle", 77}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pConstants-parameter", 70}, { {"If constantCount is not 0, and pConstants is not NULL, pConstants must be a valid pointer to an array of constantCount valid VkDataGraphPipelineConstantARM structures", 166}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pName-09872", 61}, { {"pName must be the name of an OpGraphEntryPointARM in module", 59}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pName-parameter", 65}, { {"pName must be a null-terminated UTF-8 string", 44}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pNext-09873", 61}, { {"If the pNext chain includes a VkShaderModuleCreateInfo structure, then module must be VK_NULL_HANDLE", 100}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pNext-09874", 61}, { {"If the pNext chain does not include a VkShaderModuleCreateInfo structure, then module must be a valid VkShaderModule", 116}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-pSpecializationInfo-parameter", 79}, { {"If pSpecializationInfo is not NULL, pSpecializationInfo must be a valid pointer to a valid VkSpecializationInfo structure", 121}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineShaderModuleCreateInfoARM-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SHADER_MODULE_CREATE_INFO_ARM", 81}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeConnectionARM-connection-parameter", 68}, { {"connection must be a valid VkDataGraphPipelineNodeConnectionTypeARM value", 73}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeConnectionARM-pNext-pNext", 59}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeConnectionARM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SINGLE_NODE_CONNECTION_ARM", 78}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-connectionCount-arraylength", 75}, { {"connectionCount must be greater than 0", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-nodeType-09963", 62}, { {"If nodeType is VK_DATA_GRAPH_PIPELINE_NODE_TYPE_OPTICAL_FLOW_ARM, then a VkDataGraphPipelineOpticalFlowCreateInfoARM structure must be included in the pNext chain of this structure", 180}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-nodeType-09978", 62}, { {"If nodeType is VK_DATA_GRAPH_PIPELINE_NODE_TYPE_OPTICAL_FLOW_ARM, then one and only VkDataGraphPipelineSingleNodeConnectionARM structure must be present in the pConnections array for each of the following values of its connection member: VK_DATA_GRAPH_PIPELINE_NODE_CONNECTION_TYPE_OPTICAL_FLOW_INPUT_ARM VK_DATA_GRAPH_PIPELINE_NODE_CONNECTION_TYPE_OPTICAL_FLOW_REFERENCE_ARM VK_DATA_GRAPH_PIPELINE_NODE_CONNECTION_TYPE_OPTICAL_FLOW_FLOW_VECTOR_ARM", 448}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-nodeType-09979", 62}, { {"If nodeType is VK_DATA_GRAPH_PIPELINE_NODE_TYPE_OPTICAL_FLOW_ARM and VkDataGraphPipelineOpticalFlowCreateInfoARM::hintGridSize is not 0, then one and only VkDataGraphPipelineSingleNodeConnectionARM structure whose connection member is VK_DATA_GRAPH_PIPELINE_NODE_CONNECTION_TYPE_OPTICAL_FLOW_HINT_ARM must be present in the pConnections array", 342}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-nodeType-parameter", 66}, { {"nodeType must be a valid VkDataGraphPipelineNodeTypeARM value", 61}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-pConnections-parameter", 70}, { {"pConnections must be a valid pointer to an array of connectionCount valid VkDataGraphPipelineSingleNodeConnectionARM structures", 127}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphPipelineSingleNodeCreateInfoARM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SINGLE_NODE_CREATE_INFO_ARM", 79}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-dataGraph-09953", 61}, { {"The dataGraph feature must be enabled", 37}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-pProcessingEngines-09918", 70}, { {"pProcessingEngines must not contain identical VkPhysicalDeviceDataGraphProcessingEngineARM structures", 101}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-pProcessingEngines-09956", 70}, { {"For each element of pProcessingEngines, its type member must be a valid VkPhysicalDeviceDataGraphProcessingEngineTypeARM value", 126}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-pProcessingEngines-11843", 70}, { {"If any element of pProcessingEngines has a type of VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_NEURAL_QCOM or VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_COMPUTE_QCOM and isForeign set to VK_TRUE, processingEngineCount must equal 1", 249}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-pProcessingEngines-11844", 70}, { {"If any element of pProcessingEngines has a type of VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_NEURAL_QCOM or VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_COMPUTE_QCOM, the dataGraphModel feature must be enabled", 228}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-pProcessingEngines-parameter", 74}, { {"pProcessingEngines must be a valid pointer to an array of processingEngineCount VkPhysicalDeviceDataGraphProcessingEngineARM structures", 135}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-processingEngineCount-arraylength", 79}, { {"processingEngineCount must be greater than 0", 44}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDataGraphProcessingEngineCreateInfoARM-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_DATA_GRAPH_PROCESSING_ENGINE_CREATE_INFO_ARM", 76}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkDebugMarkerMarkerInfoEXT-pMarkerName-parameter", 53}, { {"pMarkerName must be a null-terminated UTF-8 string", 50}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerMarkerInfoEXT-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerMarkerInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_MARKER_MARKER_INFO_EXT", 60}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-object-01491", 48}, { {"object must not be VK_NULL_HANDLE", 33}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-object-01492", 48}, { {"object must be a Vulkan object of the type associated with objectType as defined in VkDebugReportObjectTypeEXT and Vulkan Handle Relationship", 141}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-objectType-01490", 52}, { {"objectType must not be VK_DEBUG_REPORT_OBJECT_TYPE_UNKNOWN_EXT", 62}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-objectType-parameter", 56}, { {"objectType must be a valid VkDebugReportObjectTypeEXT value", 59}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-pObjectName-parameter", 57}, { {"pObjectName must be a null-terminated UTF-8 string", 50}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectNameInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_NAME_INFO_EXT", 65}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-object-01494", 47}, { {"object must not be VK_NULL_HANDLE", 33}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-object-01495", 47}, { {"object must be a Vulkan object of the type associated with objectType as defined in VkDebugReportObjectTypeEXT and Vulkan Handle Relationship", 141}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-objectType-01493", 51}, { {"objectType must not be VK_DEBUG_REPORT_OBJECT_TYPE_UNKNOWN_EXT", 62}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-objectType-parameter", 55}, { {"objectType must be a valid VkDebugReportObjectTypeEXT value", 59}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-pTag-parameter", 49}, { {"pTag must be a valid pointer to an array of tagSize bytes", 57}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_TAG_INFO_EXT", 64}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugMarkerObjectTagInfoEXT-tagSize-arraylength", 54}, { {"tagSize must be greater than 0", 30}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugReportCallbackCreateInfoEXT-flags-parameter", 55}, { {"flags must be a valid combination of VkDebugReportFlagBitsEXT values", 68}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugReportCallbackCreateInfoEXT-pfnCallback-parameter", 61}, { {"pfnCallback must be a valid PFN_vkDebugReportCallbackEXT value", 62}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugReportCallbackCreateInfoEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_REPORT_CALLBACK_CREATE_INFO_EXT", 69}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsLabelEXT-pLabelName-parameter", 46}, { {"pLabelName must be a null-terminated UTF-8 string", 49}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsLabelEXT-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsLabelEXT-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT", 53}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-flags-zerobitmask", 59}, { {"flags must be 0", 15}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pCmdBufLabels-parameter", 65}, { {"If cmdBufLabelCount is not 0, pCmdBufLabels must be a valid pointer to an array of cmdBufLabelCount valid VkDebugUtilsLabelEXT structures", 137}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pMessage-parameter", 60}, { {"If pMessage is not NULL, pMessage must be a null-terminated UTF-8 string", 72}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pMessageIdName-parameter", 66}, { {"If pMessageIdName is not NULL, pMessageIdName must be a null-terminated UTF-8 string", 84}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pNext-pNext", 53}, { {"pNext must be NULL or a pointer to a valid instance of VkDeviceAddressBindingCallbackDataEXT", 92}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pObjects-parameter", 60}, { {"If objectCount is not 0, pObjects must be a valid pointer to an array of objectCount valid VkDebugUtilsObjectNameInfoEXT structures", 131}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-pQueueLabels-parameter", 64}, { {"If queueLabelCount is not 0, pQueueLabels must be a valid pointer to an array of queueLabelCount valid VkDebugUtilsLabelEXT structures", 134}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CALLBACK_DATA_EXT", 71}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCallbackDataEXT-sType-unique", 54}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-flags-zerobitmask", 57}, { {"flags must be 0", 15}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-messageSeverity-parameter", 65}, { {"messageSeverity must be a valid combination of VkDebugUtilsMessageSeverityFlagBitsEXT values", 92}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-messageSeverity-requiredbitmask", 71}, { {"messageSeverity must not be 0", 29}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-messageType-parameter", 61}, { {"messageType must be a valid combination of VkDebugUtilsMessageTypeFlagBitsEXT values", 84}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-messageType-requiredbitmask", 67}, { {"messageType must not be 0", 25}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-pfnUserCallback-parameter", 65}, { {"pfnUserCallback must be a valid PFN_vkDebugUtilsMessengerCallbackEXT value", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsMessengerCreateInfoEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT", 69}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectNameInfoEXT-objectType-02589", 51}, { {"If objectType is VK_OBJECT_TYPE_UNKNOWN, objectHandle must not be VK_NULL_HANDLE", 80}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectNameInfoEXT-objectType-02590", 51}, { {"If objectType is not VK_OBJECT_TYPE_UNKNOWN, objectHandle must be VK_NULL_HANDLE or a valid Vulkan handle of the type associated with objectType as defined in the VkObjectType and Vulkan Handle Relationship table", 212}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectNameInfoEXT-objectType-parameter", 55}, { {"objectType must be a valid VkObjectType value", 45}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectNameInfoEXT-pObjectName-parameter", 56}, { {"If pObjectName is not NULL, pObjectName must be a null-terminated UTF-8 string", 78}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectNameInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT", 64}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-objectHandle-01910", 52}, { {"objectHandle must be a valid Vulkan handle of the type associated with objectType as defined in the VkObjectType and Vulkan Handle Relationship table", 149}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-objectType-01908", 50}, { {"objectType must not be VK_OBJECT_TYPE_UNKNOWN", 45}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-objectType-parameter", 54}, { {"objectType must be a valid VkObjectType value", 45}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-pTag-parameter", 48}, { {"pTag must be a valid pointer to an array of tagSize bytes", 57}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_TAG_INFO_EXT", 63}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDebugUtilsObjectTagInfoEXT-tagSize-arraylength", 53}, { {"tagSize must be greater than 0", 30}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDecompressMemoryInfoEXT-decompressionMethod-07690", 56}, { {"The decompressionMethod must have a single bit set", 50}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-decompressionMethod-11762", 56}, { {"If decompressionMethod is VK_MEMORY_DECOMPRESSION_METHOD_GDEFLATE_1_0_BIT_EXT, then for each element of pRegions, decompressedSize must be less than or equal to 65536 bytes", 172}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-decompressionMethod-11763", 56}, { {"decompressionMethod must be a valid bit specified in VkPhysicalDeviceMemoryDecompressionPropertiesEXT::decompressionMethods", 123}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-decompressionMethod-parameter", 60}, { {"decompressionMethod must be a valid combination of VkMemoryDecompressionMethodFlagBitsEXT values", 96}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-decompressionMethod-requiredbitmask", 66}, { {"decompressionMethod must not be 0", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-pRegions-parameter", 49}, { {"pRegions must be a valid pointer to an array of regionCount VkDecompressMemoryRegionEXT structures", 98}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-regionCount-arraylength", 54}, { {"regionCount must be greater than 0", 34}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryInfoEXT-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_DECOMPRESS_MEMORY_INFO_EXT", 58}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-compressedSize-11795", 53}, { {"compressedSize must not be zero", 31}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-decompressedSize-07689", 55}, { {"decompressedSize must be large enough to hold the decompressed data based on the decompressionMethod", 100}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-decompressedSize-11796", 55}, { {"decompressedSize must not be zero", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-dstAddress-07687", 49}, { {"dstAddress must be 4 byte aligned", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-dstAddress-07688", 49}, { {"The memory range defined by dstAddress and decompressedSize must be contained within the size of the buffer bound to dstAddress, minus the offset of dstAddress from the base address of that buffer", 196}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-dstAddress-11765", 49}, { {"dstAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_2_MEMORY_DECOMPRESSION_BIT_EXT usage flag set", 157}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-dstAddress-parameter", 53}, { {"dstAddress must be a valid VkDeviceAddress value", 48}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-srcAddress-07685", 49}, { {"srcAddress must be 4 byte aligned", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-srcAddress-07686", 49}, { {"The memory range defined by srcAddress and compressedSize must be contained within the size of the buffer bound to srcAddress, minus the offset of srcAddress from the base address of that buffer", 194}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-srcAddress-07691", 49}, { {"The memory range defined by srcAddress and compressedSize must not overlap the memory range defined by dstAddress and decompressedSize", 134}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-srcAddress-11764", 49}, { {"srcAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_2_MEMORY_DECOMPRESSION_BIT_EXT usage flag set", 157}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionEXT-srcAddress-parameter", 53}, { {"srcAddress must be a valid VkDeviceAddress value", 48}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-compressedSize-11795", 52}, { {"compressedSize must not be zero", 31}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-decompressedSize-07689", 54}, { {"decompressedSize must be large enough to hold the decompressed data based on the decompressionMethod", 100}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-decompressedSize-11796", 54}, { {"decompressedSize must not be zero", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-decompressionMethod-07690", 57}, { {"The decompressionMethod must have a single bit set", 50}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-decompressionMethod-09395", 57}, { {"If decompressionMethod is VK_MEMORY_DECOMPRESSION_METHOD_GDEFLATE_1_0_BIT_EXT, then decompressedSize must be less than or equal to 65536 bytes", 142}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-dstAddress-07687", 48}, { {"dstAddress must be 4 byte aligned", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-dstAddress-07688", 48}, { {"The memory range defined by dstAddress and decompressedSize must be contained within the size of the buffer bound to dstAddress, minus the offset of dstAddress from the base address of that buffer", 196}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-srcAddress-07685", 48}, { {"srcAddress must be 4 byte aligned", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-srcAddress-07686", 48}, { {"The memory range defined by srcAddress and compressedSize must be contained within the size of the buffer bound to srcAddress, minus the offset of srcAddress from the base address of that buffer", 194}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDecompressMemoryRegionNV-srcAddress-07691", 48}, { {"The memory range defined by srcAddress and compressedSize must not overlap the memory range defined by dstAddress and decompressedSize", 134}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-VkDedicatedAllocationBufferCreateInfoNV-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_BUFFER_CREATE_INFO_NV", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDedicatedAllocationImageCreateInfoNV-dedicatedAllocation-00994", 69}, { {"If dedicatedAllocation is VK_TRUE, VkImageCreateInfo::flags must not include VK_IMAGE_CREATE_SPARSE_BINDING_BIT, VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT, or VK_IMAGE_CREATE_SPARSE_ALIASED_BIT", 188}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDedicatedAllocationImageCreateInfoNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_IMAGE_CREATE_INFO_NV", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-buffer-00651", 59}, { {"If buffer is not VK_NULL_HANDLE, the buffer must have been created with VkDedicatedAllocationBufferCreateInfoNV::dedicatedAllocation equal to VK_TRUE", 149}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-buffer-00653", 59}, { {"If buffer is not VK_NULL_HANDLE, VkMemoryAllocateInfo::allocationSize must equal the VkMemoryRequirements::size of the buffer", 125}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-buffer-00655", 59}, { {"If buffer is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation, the memory being imported must also be a dedicated buffer allocation and buffer must be identical to the buffer associated with the imported memory", 239}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-buffer-parameter", 63}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-commonparent", 59}, { {"Both of buffer, and image that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 145}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-image-00649", 58}, { {"At least one of image and buffer must be VK_NULL_HANDLE", 55}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-image-00650", 58}, { {"If image is not VK_NULL_HANDLE, the image must have been created with VkDedicatedAllocationImageCreateInfoNV::dedicatedAllocation equal to VK_TRUE", 146}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-image-00652", 58}, { {"If image is not VK_NULL_HANDLE, VkMemoryAllocateInfo::allocationSize must equal the VkMemoryRequirements::size of the image", 123}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-image-00654", 58}, { {"If image is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation, the memory being imported must also be a dedicated image allocation and image must be identical to the image associated with the imported memory", 235}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-image-parameter", 62}, { {"If image is not VK_NULL_HANDLE, image must be a valid VkImage handle", 68}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDedicatedAllocationMemoryAllocateInfoNV-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_MEMORY_ALLOCATE_INFO_NV", 76}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDependencyInfo-dependencyFlags-parameter", 47}, { {"dependencyFlags must be a valid combination of VkDependencyFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pBufferMemoryBarriers-parameter", 53}, { {"If bufferMemoryBarrierCount is not 0, pBufferMemoryBarriers must be a valid pointer to an array of bufferMemoryBarrierCount valid VkBufferMemoryBarrier2 structures", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pImageMemoryBarriers-parameter", 52}, { {"If imageMemoryBarrierCount is not 0, pImageMemoryBarriers must be a valid pointer to an array of imageMemoryBarrierCount valid VkImageMemoryBarrier2 structures", 159}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pMemoryBarriers-10605", 43}, { {"For each element of pMemoryBarriers, the sType value of each structure in the pNext chain must be unique", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pMemoryBarriers-10606", 43}, { {"For each element of pMemoryBarriers, pNext must be either NULL or a pointer to a valid instance of VkMemoryBarrierAccessFlags3KHR", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pMemoryBarriers-parameter", 47}, { {"If memoryBarrierCount is not 0, pMemoryBarriers must be a valid pointer to an array of memoryBarrierCount valid VkMemoryBarrier2 structures", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pNext-09754", 33}, { {"If a VkTensorDependencyInfoARM structure is included in the pNext chain, a VkTensorMemoryBarrierARM structure must not be included in the pNext chain", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-pNext-pNext", 33}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkMemoryRangeBarriersInfoKHR, VkTensorDependencyInfoARM, or VkTensorMemoryBarrierARM", 215}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_DEPENDENCY_INFO", 47}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDependencyInfo-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDepthBiasInfoEXT-depthBiasClamp-08950", 44}, { {"If the depthBiasClamp feature is not enabled, depthBiasClamp must be 0.0", 72}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasInfoEXT-pNext-pNext", 35}, { {"pNext must be NULL or a pointer to a valid instance of VkDepthBiasRepresentationInfoEXT", 87}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasInfoEXT-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_DEPTH_BIAS_INFO_EXT", 51}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasInfoEXT-sType-unique", 36}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasRepresentationInfoEXT-depthBiasExact-08949", 58}, { {"If the depthBiasExact feature is not enabled, depthBiasExact must be VK_FALSE", 77}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasRepresentationInfoEXT-depthBiasRepresentation-parameter", 71}, { {"depthBiasRepresentation must be a valid VkDepthBiasRepresentationEXT value", 74}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasRepresentationInfoEXT-floatRepresentation-08948", 63}, { {"If the floatRepresentation feature is not enabled, depthBiasRepresentation must not be VK_DEPTH_BIAS_REPRESENTATION_FLOAT_EXT", 125}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasRepresentationInfoEXT-leastRepresentableValueForceUnormRepresentation-08947", 91}, { {"If the leastRepresentableValueForceUnormRepresentation feature is not enabled, depthBiasRepresentation must not be VK_DEPTH_BIAS_REPRESENTATION_LEAST_REPRESENTABLE_VALUE_FORCE_UNORM_EXT", 185}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthBiasRepresentationInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_DEPTH_BIAS_REPRESENTATION_INFO_EXT", 66}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkDepthClampRangeEXT-pDepthClampRange-00999", 48}, { {"minDepthClamp must be less than or equal to maxDepthClamp", 57}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkDepthClampRangeEXT-pDepthClampRange-09648", 48}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled, minDepthClamp must be greater than or equal to 0.0", 115}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkDepthClampRangeEXT-pDepthClampRange-09649", 48}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled, maxDepthClamp must be less than or equal to 1.0", 112}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkDescriptorAddressInfoEXT-None-09508", 42}, { {"If address is not zero, and the descriptor is of type VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, then format must not be VK_FORMAT_UNDEFINED", 181}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-None-12271", 42}, { {"If Vulkan 1.3 is not supported and the ycbcr2plane444Formats feature is not enabled, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 286}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-address-08043", 45}, { {"If the nullDescriptor feature is not enabled, address must not be zero", 70}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-address-parameter", 49}, { {"If address is not 0, address must be a valid VkDeviceAddress value", 66}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-format-parameter", 48}, { {"format must be a valid VkFormat value", 37}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-nullDescriptor-08938", 52}, { {"If address is zero, range must be VK_WHOLE_SIZE", 47}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-nullDescriptor-08939", 52}, { {"If address is not zero, range must not be VK_WHOLE_SIZE", 55}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-range-08045", 43}, { {"If address is not zero, then range must be less than or equal to the size of the buffer containing address minus the offset of address from the base address of the buffer", 170}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-range-08940", 43}, { {"range must not be zero", 22}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorAddressInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_ADDRESS_INFO_EXT", 59}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-None-09499", 48}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must be a valid combination of VkBufferUsageFlagBits values", 144}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-None-09500", 48}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must not be 0", 98}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-address-08057", 51}, { {"address must be aligned to VkPhysicalDeviceDescriptorBufferPropertiesEXT::descriptorBufferOffsetAlignment", 105}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-address-parameter", 55}, { {"address must be a valid VkDeviceAddress value", 45}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-bufferlessPushDescriptors-08056", 69}, { {"If VkPhysicalDeviceDescriptorBufferPropertiesEXT::bufferlessPushDescriptors is VK_FALSE, and usage contains VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT, then the pNext chain must include a VkDescriptorBufferBindingPushDescriptorBufferHandleEXT structure", 268}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-pNext-pNext", 49}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBufferUsageFlags2CreateInfo or VkDescriptorBufferBindingPushDescriptorBufferHandleEXT", 218}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_INFO_EXT", 66}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-usage-08122", 49}, { {"If usage includes VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT, address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT usage flag set", 226}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-usage-08123", 49}, { {"If usage includes VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT, address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT usage flag set", 228}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-usage-08124", 49}, { {"If usage includes VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT, address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT usage flag set", 244}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingInfoEXT-usage-10998", 49}, { {"The usage must include at least one of VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT, VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT, or VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT", 203}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingPushDescriptorBufferHandleEXT-buffer-parameter", 76}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingPushDescriptorBufferHandleEXT-bufferlessPushDescriptors-08059", 91}, { {"VkPhysicalDeviceDescriptorBufferPropertiesEXT::bufferlessPushDescriptors must be VK_FALSE", 89}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferBindingPushDescriptorBufferHandleEXT-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_PUSH_DESCRIPTOR_BUFFER_HANDLE_EXT", 91}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorBufferInfo-buffer-02998", 40}, { {"If the nullDescriptor feature is not enabled, buffer must not be VK_NULL_HANDLE", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorBufferInfo-buffer-02999", 40}, { {"If buffer is VK_NULL_HANDLE, offset must be zero and range must be VK_WHOLE_SIZE", 80}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorBufferInfo-buffer-parameter", 44}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorBufferInfo-offset-00340", 40}, { {"offset must be less than the size of buffer", 43}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorBufferInfo-range-00341", 39}, { {"If range is not equal to VK_WHOLE_SIZE, range must be greater than 0", 68}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorBufferInfo-range-00342", 39}, { {"If range is not equal to VK_WHOLE_SIZE, range must be less than or equal to the size of buffer minus offset", 107}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorDataEXT-type-08034", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and the nullDescriptor feature is not enabled, pCombinedImageSampler->imageView must not be VK_NULL_HANDLE", 184}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08035", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and the nullDescriptor feature is not enabled, pSampledImage must not be NULL and pSampledImage->imageView must not be VK_NULL_HANDLE", 202}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08036", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and the nullDescriptor feature is not enabled, pStorageImage must not be NULL and pStorageImage->imageView must not be VK_NULL_HANDLE", 202}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08037", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, and the nullDescriptor feature is not enabled, pUniformTexelBuffer must not be NULL", 159}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08038", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, and the nullDescriptor feature is not enabled, pStorageTexelBuffer must not be NULL", 159}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08039", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, and the nullDescriptor feature is not enabled, pUniformBuffer must not be NULL", 148}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08040", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, and the nullDescriptor feature is not enabled, pStorageBuffer must not be NULL", 148}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08041", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, and the nullDescriptor feature is not enabled, accelerationStructure must not be 0", 164}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorDataEXT-type-08042", 35}, { {"If VkDescriptorGetInfoEXT::type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, and the nullDescriptor feature is not enabled, accelerationStructure must not be 0", 163}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-accelerationStructure-parameter", 59}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR or VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, the accelerationStructure member of data must be a valid VkDeviceAddress value", 184}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pCombinedImageSampler-parameter", 59}, { {"If type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, the pCombinedImageSampler member of data must be a valid pointer to a valid VkDescriptorImageInfo structure", 161}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pInputAttachmentImage-parameter", 59}, { {"If type is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the pInputAttachmentImage member of data must be a valid pointer to a valid VkDescriptorImageInfo structure", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pNext-pNext", 39}, { {"pNext must be NULL or a pointer to a valid instance of VkDescriptorGetTensorInfoARM", 83}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pSampledImage-parameter", 51}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and if pSampledImage is not NULL, the pSampledImage member of data must be a valid pointer to a valid VkDescriptorImageInfo structure", 178}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pSampler-parameter", 46}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLER, the pSampler member of data must be a valid pointer to a valid VkSampler handle", 118}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pStorageBuffer-parameter", 52}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, and if pStorageBuffer is not NULL, the pStorageBuffer member of data must be a valid pointer to a valid VkDescriptorAddressInfoEXT structure", 186}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pStorageImage-parameter", 51}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and if pStorageImage is not NULL, the pStorageImage member of data must be a valid pointer to a valid VkDescriptorImageInfo structure", 178}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pStorageTexelBuffer-parameter", 57}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, and if pStorageTexelBuffer is not NULL, the pStorageTexelBuffer member of data must be a valid pointer to a valid VkDescriptorAddressInfoEXT structure", 202}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pUniformBuffer-parameter", 52}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, and if pUniformBuffer is not NULL, the pUniformBuffer member of data must be a valid pointer to a valid VkDescriptorAddressInfoEXT structure", 186}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-pUniformTexelBuffer-parameter", 57}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, and if pUniformTexelBuffer is not NULL, the pUniformTexelBuffer member of data must be a valid pointer to a valid VkDescriptorAddressInfoEXT structure", 202}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_GET_INFO_EXT", 55}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08018", 38}, { {"type must not be VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 144}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08019", 38}, { {"If type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, the pCombinedImageSampler->sampler member of data must be a VkSampler created on device", 141}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08020", 38}, { {"If type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, the pCombinedImageSampler->imageView member of data must be a VkImageView created on device, or VK_NULL_HANDLE", 164}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08021", 38}, { {"If type is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the pInputAttachmentImage->imageView member of data must be a VkImageView created on device", 139}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08022", 38}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and if pSampledImage is not NULL, the pSampledImage->imageView member of data must be a VkImageView created on device, or VK_NULL_HANDLE", 181}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08023", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and if pStorageImage is not NULL, the pStorageImage->imageView member of data must be a VkImageView created on device, or VK_NULL_HANDLE", 181}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08028", 38}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR and accelerationStructure is not 0, accelerationStructure must contain the address of a VkAccelerationStructureKHR created on device", 189}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-08029", 38}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV and accelerationStructure is not 0, accelerationStructure must contain the handle of a VkAccelerationStructureNV created on device, returned by vkGetAccelerationStructureHandleNV", 234}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-09427", 38}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, pUniformBuffer is not NULL , the number of texel buffer elements given by (floor(pUniformBuffer->range / (texel block size)) x (texels per block)) where texel block size and texels per block are as defined in the Compatible Formats table for pUniformBuffer->format, must be less than or equal to VkPhysicalDeviceLimits::maxTexelBufferElements", 394}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-09428", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, pStorageBuffer is not NULL , the number of texel buffer elements given by (floor(pStorageBuffer->range / (texel block size)) x (texels per block)) where texel block size and texels per block are as defined in the Compatible Formats table for pStorageBuffer->format, must be less than or equal to VkPhysicalDeviceLimits::maxTexelBufferElements", 394}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-09701", 38}, { {"If type is VK_DESCRIPTOR_TYPE_TENSOR_ARM, a VkDescriptorGetTensorInfoARM structure must be included in the pNext chain and data is ignored", 138}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12216", 38}, { {"If type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER and pCombinedImageSampler->imageView is not VK_NULL_HANDLE, the pCombinedImageSampler->imageView member of data must have been created with VK_IMAGE_USAGE_SAMPLED_BIT set", 223}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12217", 38}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, pSampledImage is not NULL, and pSampledImage->imageView is not VK_NULL_HANDLE, the pSampledImage->imageView member of data must have been created with VK_IMAGE_USAGE_SAMPLED_BIT set", 226}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12218", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, pStorageImage is not NULL, and pStorageImage->imageView is not VK_NULL_HANDLE, the pStorageImage->imageView member of data must have been created with VK_IMAGE_USAGE_STORAGE_BIT set", 226}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12219", 38}, { {"If type is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the pInputAttachmentImage->imageView member of data must have been created with VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT set", 167}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12220", 38}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, pUniformBuffer is not NULL and pUniformBuffer->address is not zero, pUniformBuffer->address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 272}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12221", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, pStorageBuffer is not NULL and pStorageBuffer->address is not zero, pStorageBuffer->address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 272}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12222", 38}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, pUniformTexelBuffer is not NULL and pUniformTexelBuffer->address is not zero, pUniformTexelBuffer->address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT usage flag set", 299}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12223", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, pStorageTexelBuffer is not NULL and pStorageTexelBuffer->address is not zero, pStorageTexelBuffer->address must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT usage flag set", 299}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12265", 38}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER and pUniformBuffer is not NULL, pUniformBuffer->address must be aligned to VkPhysicalDeviceLimits::minUniformBufferOffsetAlignment", 175}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12266", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER and pStorageBuffer is not NULL, pStorageBuffer->address must be aligned to VkPhysicalDeviceLimits::minStorageBufferOffsetAlignment", 175}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12269", 38}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, and pUniformTexelBuffer is not NULL, pUniformTexelBuffer->address must be a multiple of the effective alignment requirement as determined by minTexelBufferOffsetAlignment", 222}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-12270", 38}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, and pStorageTexelBuffer is not NULL, pStorageTexelBuffer->address must be a multiple of the effective alignment requirement as determined by minTexelBufferOffsetAlignment", 222}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetInfoEXT-type-parameter", 42}, { {"type must be a valid VkDescriptorType value", 43}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetTensorInfoARM-nullDescriptor-09899", 54}, { {"If the nullDescriptor feature is not enabled, tensorView must not be VK_NULL_HANDLE", 83}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetTensorInfoARM-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_GET_TENSOR_INFO_ARM", 62}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorGetTensorInfoARM-tensorView-parameter", 54}, { {"If tensorView is not VK_NULL_HANDLE, tensorView must be a valid VkTensorViewARM handle", 86}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkDescriptorImageInfo-commonparent", 39}, { {"Both of imageView, and sampler that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 150}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-descriptorType-06713", 47}, { {"If the image2DViewOf3D feature is not enabled or descriptorType is not VK_DESCRIPTOR_TYPE_STORAGE_IMAGE then imageView must not be a 2D view created from a 3D image", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-descriptorType-06714", 47}, { {"If the sampler2DViewOf3D feature is not enabled or descriptorType is not VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER then imageView must not be a 2D view created from a 3D image", 211}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageLayout-09425", 44}, { {"If imageLayout is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, then the aspectMask used to create imageView must not include either VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 185}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageLayout-09426", 44}, { {"If imageLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, then the aspectMask used to create imageView must not include VK_IMAGE_ASPECT_COLOR_BIT", 496}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageView-01976", 42}, { {"If imageView is created from a depth/stencil image, the aspectMask used to create the imageView must include either VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT but not both", 185}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageView-06712", 42}, { {"imageView must not be a 2D array image view created from a 3D image", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageView-07795", 42}, { {"If imageView is a 2D view created from a 3D image, then descriptorType must be VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER", 191}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-imageView-07796", 42}, { {"If imageView is a 2D view created from a 3D image, then the image must have been created with VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT set", 140}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-mutableComparisonSamplers-04450", 58}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::mutableComparisonSamplers is VK_FALSE, then sampler must have been created with VkSamplerCreateInfo::compareEnable set to VK_FALSE", 235}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorImageInfo-sampler-01564", 40}, { {"If sampler is used and the VkFormat of the image is a multi-planar format, the image must have been created with VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT, and the aspectMask of the imageView must be a valid multi-planar aspect mask bit", 229}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorMappingSourceConstantOffsetEXT-pEmbeddedSampler-11415", 70}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, and there is a VkDebugUtilsObjectNameInfoEXT structure in its pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 185}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceConstantOffsetEXT-pEmbeddedSampler-11445", 70}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, its borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 159}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceConstantOffsetEXT-pEmbeddedSampler-12432", 70}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, the VkSamplerCreateInfo::pNext chain must not contain VkOpaqueCaptureDescriptorDataCreateInfoEXT", 161}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceConstantOffsetEXT-pEmbeddedSampler-parameter", 74}, { {"If pEmbeddedSampler is not NULL, pEmbeddedSampler must be a valid pointer to a valid VkSamplerCreateInfo structure", 114}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceHeapDataEXT-heapOffset-11263", 58}, { {"heapOffset must be a multiple of minUniformBufferOffsetAlignment", 64}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceHeapDataEXT-pushOffset-11264", 58}, { {"pushOffset must be a multiple of 4", 34}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceHeapDataEXT-pushOffset-11265", 58}, { {"pushOffset must be less than or equal to maxPushDataSize - 4", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectAddressEXT-addressOffset-11268", 68}, { {"addressOffset must be a multiple of 8", 37}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectAddressEXT-pushOffset-11266", 65}, { {"pushOffset must be a multiple of 8", 34}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectAddressEXT-pushOffset-11267", 65}, { {"pushOffset must be less than or equal to maxPushDataSize - 8", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-addressOffset-11361", 71}, { {"addressOffset must be a multiple of 4", 37}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pEmbeddedSampler-11415", 74}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, and there is a VkDebugUtilsObjectNameInfoEXT structure in its pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 185}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pEmbeddedSampler-11445", 74}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, its borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 159}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pEmbeddedSampler-12432", 74}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, the VkSamplerCreateInfo::pNext chain must not contain VkOpaqueCaptureDescriptorDataCreateInfoEXT", 161}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pEmbeddedSampler-parameter", 78}, { {"If pEmbeddedSampler is not NULL, pEmbeddedSampler must be a valid pointer to a valid VkSamplerCreateInfo structure", 114}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pushOffset-11359", 68}, { {"pushOffset must be a multiple of 8", 34}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexArrayEXT-pushOffset-11360", 68}, { {"pushOffset must be less than or equal to maxPushDataSize - 8", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-addressOffset-11262", 66}, { {"addressOffset must be a multiple of 4", 37}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pEmbeddedSampler-11415", 69}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, and there is a VkDebugUtilsObjectNameInfoEXT structure in its pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 185}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pEmbeddedSampler-11445", 69}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, its borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 159}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pEmbeddedSampler-12432", 69}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, the VkSamplerCreateInfo::pNext chain must not contain VkOpaqueCaptureDescriptorDataCreateInfoEXT", 161}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pEmbeddedSampler-parameter", 73}, { {"If pEmbeddedSampler is not NULL, pEmbeddedSampler must be a valid pointer to a valid VkSamplerCreateInfo structure", 114}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pushOffset-11260", 63}, { {"pushOffset must be a multiple of 8", 34}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceIndirectIndexEXT-pushOffset-11261", 63}, { {"pushOffset must be less than or equal to maxPushDataSize - 8", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pEmbeddedSampler-11415", 65}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, and there is a VkDebugUtilsObjectNameInfoEXT structure in its pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 185}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pEmbeddedSampler-11445", 65}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, its borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 159}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pEmbeddedSampler-12432", 65}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, the VkSamplerCreateInfo::pNext chain must not contain VkOpaqueCaptureDescriptorDataCreateInfoEXT", 161}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pEmbeddedSampler-parameter", 69}, { {"If pEmbeddedSampler is not NULL, pEmbeddedSampler must be a valid pointer to a valid VkSamplerCreateInfo structure", 114}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pushOffset-11258", 59}, { {"pushOffset must be a multiple of 4", 34}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourcePushIndexEXT-pushOffset-11259", 59}, { {"pushOffset must be less than or equal to maxPushDataSize - 4", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-pEmbeddedSampler-11415", 73}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, and there is a VkDebugUtilsObjectNameInfoEXT structure in its pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 185}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-pEmbeddedSampler-11445", 73}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, its borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 159}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-pEmbeddedSampler-12432", 73}, { {"If pEmbeddedSampler is a valid pointer to a VkSamplerCreateInfo, the VkSamplerCreateInfo::pNext chain must not contain VkOpaqueCaptureDescriptorDataCreateInfoEXT", 161}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-pEmbeddedSampler-parameter", 77}, { {"If pEmbeddedSampler is not NULL, pEmbeddedSampler must be a valid pointer to a valid VkSamplerCreateInfo structure", 114}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-shaderRecordOffset-11269", 75}, { {"shaderRecordOffset must be a multiple of 4", 42}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorMappingSourceShaderRecordIndexEXT-shaderRecordOffset-11270", 75}, { {"shaderRecordOffset must be less than or equal to maxShaderGroupStride - 4", 73}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorPoolCreateInfo-descriptorPoolOverallocation-09227", 66}, { {"If the descriptorPoolOverallocation feature is not enabled, or flags does not have VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_SETS_BIT_NV set, maxSets must be greater than 0", 177}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-flags-04607", 43}, { {"If flags has the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT bit set, then the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT bit must not be set", 146}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-flags-04609", 43}, { {"If flags has the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT bit set, VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT::mutableDescriptorType must be enabled", 157}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-flags-09228", 43}, { {"If flags has the VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_SETS_BIT_NV or VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_POOLS_BIT_NV bits set, then descriptorPoolOverallocation must be enabled", 198}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-flags-parameter", 47}, { {"flags must be a valid combination of VkDescriptorPoolCreateFlagBits values", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-mutableDescriptorType-04608", 59}, { {"If VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT::mutableDescriptorType is not enabled, pPoolSizes must not contain a descriptorType of VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 169}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-pNext-09946", 43}, { {"If a VkDataGraphProcessingEngineCreateInfoARM structure is included in the pNext chain, each member of pProcessingEngines must be identical to an VkQueueFamilyDataGraphPropertiesARM::engine retrieved from vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with the physicalDevice that was used to create device", 312}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-pNext-pNext", 43}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDataGraphProcessingEngineCreateInfoARM, VkDescriptorPoolInlineUniformBlockCreateInfo, or VkMutableDescriptorTypeCreateInfoEXT", 258}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-pPoolSizes-04787", 48}, { {"If pPoolSizes contains a descriptorType of VK_DESCRIPTOR_TYPE_MUTABLE_EXT, any other VK_DESCRIPTOR_TYPE_MUTABLE_EXT element in pPoolSizes must not have sets of supported descriptor types which partially overlap", 210}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-pPoolSizes-09424", 48}, { {"If pPoolSizes contains a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, the pNext chain must include a VkDescriptorPoolInlineUniformBlockCreateInfo structure whose maxInlineUniformBlockBindings member is not zero", 224}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-pPoolSizes-parameter", 52}, { {"If poolSizeCount is not 0, pPoolSizes must be a valid pointer to an array of poolSizeCount valid VkDescriptorPoolSize structures", 128}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO", 59}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolCreateInfo-sType-unique", 44}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolInlineUniformBlockCreateInfo-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_INLINE_UNIFORM_BLOCK_CREATE_INFO", 80}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolSize-descriptorCount-00302", 47}, { {"descriptorCount must be greater than 0", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolSize-type-02218", 36}, { {"If type is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK then descriptorCount must be a multiple of 4", 95}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorPoolSize-type-parameter", 40}, { {"type must be a valid VkDescriptorType value", 43}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-apiVersion-07895", 49}, { {"If the VK_KHR_maintenance1 extension is not enabled and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, descriptorSetCount must not be greater than the number of sets that are currently available for allocation in descriptorPool", 244}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-apiVersion-07896", 49}, { {"If the VK_KHR_maintenance1 extension is not enabled and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, descriptorPool must have enough free descriptor capacity remaining to allocate the descriptor sets of the specified layouts", 243}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-commonparent", 45}, { {"Both of descriptorPool, and the elements of pSetLayouts must have been created, allocated, or retrieved from the same VkDevice", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-descriptorPool-parameter", 57}, { {"descriptorPool must be a valid VkDescriptorPool handle", 54}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-descriptorSetCount-arraylength", 63}, { {"descriptorSetCount must be greater than 0", 41}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pNext-pNext", 44}, { {"pNext must be NULL or a pointer to a valid instance of VkDescriptorSetVariableDescriptorCountAllocateInfo", 105}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-00308", 50}, { {"Each element of pSetLayouts must not have been created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT set", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-03044", 50}, { {"If any element of pSetLayouts was created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set, descriptorPool must have been created with the VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT flag set", 222}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-04610", 50}, { {"If any element of pSetLayouts was created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT bit set, descriptorPool must have been created with the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT flag set", 214}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-08009", 50}, { {"Each element of pSetLayouts must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT bit set", 129}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-09380", 50}, { {"If pSetLayouts[i] was created with an element of pBindingFlags that includes VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT, and VkDescriptorSetVariableDescriptorCountAllocateInfo is included in the pNext chain, and VkDescriptorSetVariableDescriptorCountAllocateInfo::descriptorSetCount is not zero, then VkDescriptorSetVariableDescriptorCountAllocateInfo::pDescriptorCounts[i] must be less than or equal to VkDescriptorSetLayoutBinding::descriptorCount for the corresponding binding used to create pSetLayouts[i]", 518}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-pSetLayouts-parameter", 54}, { {"pSetLayouts must be a valid pointer to an array of descriptorSetCount valid VkDescriptorSetLayout handles", 105}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO", 60}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAllocateInfo-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-constantOffset-parameter", 65}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, the constantOffset member of sourceData must be a valid VkDescriptorMappingSourceConstantOffsetEXT structure", 181}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-indirectIndex-parameter", 64}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, the indirectIndex member of sourceData must be a valid VkDescriptorMappingSourceIndirectIndexEXT structure", 178}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-indirectIndexArray-parameter", 69}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the indirectIndexArray member of sourceData must be a valid VkDescriptorMappingSourceIndirectIndexArrayEXT structure", 194}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-pNext-pNext", 52}, { {"pNext must be NULL or a pointer to a valid instance of VkPushConstantBankInfoNV", 79}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-pushIndex-parameter", 60}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, the pushIndex member of sourceData must be a valid VkDescriptorMappingSourcePushIndexEXT structure", 166}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-resourceMask-parameter", 63}, { {"resourceMask must be a valid combination of VkSpirvResourceTypeFlagBitsEXT values", 81}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-resourceMask-requiredbitmask", 69}, { {"resourceMask must not be 0", 26}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_AND_BINDING_MAPPING_EXT", 70}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-sType-unique", 53}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-shaderRecordIndex-parameter", 68}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, the shaderRecordIndex member of sourceData must be a valid VkDescriptorMappingSourceShaderRecordIndexEXT structure", 191}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11245", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, bindingCount must be 1", 233}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11246", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, sourceData.pushDataOffset must be a multiple of 4", 106}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11247", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, sourceData.pushAddressOffset must be a multiple of 8", 112}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11248", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, bindingCount must be 1", 146}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11249", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, sourceData.shaderRecordDataOffset must be a multiple of 4", 123}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11250", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, sourceData.shaderRecordAddressOffset must be a multiple of 8", 129}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11251", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeImage variables, any heapOffset, and heapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of imageDescriptorAlignment", 542}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11252", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeStruct variables, any heapOffset, and heapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of bufferDescriptorAlignment", 544}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11253", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampler variables, any heapOffset and heapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of samplerDescriptorAlignment", 545}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11254", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, then any samplerHeapOffset and samplerHeapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of samplerDescriptorAlignment", 599}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11356", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, resourceMask must include VK_SPIRV_RESOURCE_TYPE_UNIFORM_BUFFER_BIT_EXT", 237}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11357", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, resourceMask must include at least one of VK_SPIRV_RESOURCE_TYPE_UNIFORM_BUFFER_BIT_EXT, VK_SPIRV_RESOURCE_TYPE_READ_ONLY_STORAGE_BUFFER_BIT_EXT, VK_SPIRV_RESOURCE_TYPE_READ_WRITE_STORAGE_BUFFER_BIT_EXT, or VK_SPIRV_RESOURCE_TYPE_ACCELERATION_STRUCTURE_BIT_EXT", 378}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11358", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and the mapping sets useCombinedImageSamplerIndex to VK_TRUE, resourceMask must include at least one of VK_SPIRV_RESOURCE_TYPE_COMBINED_SAMPLED_IMAGE_BIT_EXT, VK_SPIRV_RESOURCE_TYPE_SAMPLED_IMAGE_BIT_EXT, or VK_SPIRV_RESOURCE_TYPE_SAMPLER_BIT_EXT", 505}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11389", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and bindingCount is not 1, the pEmbeddedSampler member of the corresponding mapping structure must be NULL", 425}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-11390", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeTensorARM variables, the heapOffset, and heapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of tensorDescriptorAlignment", 547}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12406", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, any heapOffset and heapArrayStride members of the corresponding member of sourceData must be 0 or a multiple of imageDescriptorAlignment", 548}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12452", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, sourceData.pushDataOffset must be less than or equal to maxPushDataSize - 4", 132}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12453", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, sourceData.pushAddressOffset must be less than or equal to maxPushDataSize - 8", 138}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12456", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, samplerPushOffset must be a multiple of 4", 232}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12457", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, samplerPushOffset must be less than or equal to maxPushDataSize - 4", 258}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12458", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, samplerPushOffset must be a multiple of 8", 303}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12459", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, samplerPushOffset must be less than or equal to maxPushDataSize - 8", 329}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-12460", 53}, { {"If source is VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and descriptorSet, firstBinding, and bindingCount identify any OpTypeSampledImage variables, and pEmbeddedSampler is NULL, samplerAddressOffset must be a multiple of 4", 306}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetAndBindingMappingEXT-source-parameter", 57}, { {"source must be a valid VkDescriptorMappingSourceEXT value", 57}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkDescriptorSetBindingReferenceVALVE-descriptorSetLayout-parameter", 71}, { {"descriptorSetLayout must be a valid VkDescriptorSetLayout handle", 64}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkDescriptorSetBindingReferenceVALVE-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkDescriptorSetBindingReferenceVALVE-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_BINDING_REFERENCE_VALVE", 70}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorCount-09465", 55}, { {"If descriptorCount is not 0, stageFlags must be VK_SHADER_STAGE_ALL or a valid combination of other VkShaderStageFlagBits values", 128}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-00282", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and descriptorCount is not 0 and pImmutableSamplers is not NULL, pImmutableSamplers must be a valid pointer to an array of descriptorCount valid VkSampler handles", 256}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-01510", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT and descriptorCount is not 0, then stageFlags must be 0 or VK_SHADER_STAGE_FRAGMENT_BIT", 144}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-02209", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK then descriptorCount must be a multiple of 4", 105}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-04604", 54}, { {"If the inlineUniformBlock feature is not enabled, descriptorType must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 116}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-04605", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_MUTABLE_EXT, then pImmutableSamplers must be NULL", 89}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-08004", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK and VkDescriptorSetLayoutCreateInfo::flags does not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT then descriptorCount must be less than or equal to VkPhysicalDeviceInlineUniformBlockProperties::maxInlineUniformBlockSize", 301}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-12200", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and descriptorCount is not 0 and pImmutableSamplers is not NULL, either each element of pImmutableSamplers must be a VkSampler that enables sampler Y'CBCR conversion or none of them enable sampler Y'CBCR conversion", 278}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-12215", 54}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER, each element of pImmutableSamplers must not be a VkSampler object that enables sampler Y'CBCR conversion", 153}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-descriptorType-parameter", 58}, { {"descriptorType must be a valid VkDescriptorType value", 53}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-flags-08005", 45}, { {"If VkDescriptorSetLayoutCreateInfo::flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT, descriptorType must be VK_DESCRIPTOR_TYPE_SAMPLER", 169}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-flags-08006", 45}, { {"If VkDescriptorSetLayoutCreateInfo::flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT, descriptorCount must be less than or equal to 1", 167}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-flags-08007", 45}, { {"If VkDescriptorSetLayoutCreateInfo::flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT, and descriptorCount is equal to 1, pImmutableSamplers must not be NULL", 190}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-flags-09466", 45}, { {"If VkDescriptorSetLayoutCreateInfo::flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PER_STAGE_BIT_NV, and descriptorCount is not 0, then stageFlags must be a valid combination of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, VK_SHADER_STAGE_FRAGMENT_BIT and VK_SHADER_STAGE_COMPUTE_BIT values", 390}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBinding-pImmutableSamplers-04009", 58}, { {"The sampler objects indicated by pImmutableSamplers must not have a borderColor with one of the values VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 169}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-None-03011", 59}, { {"All bindings with descriptor type VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 215}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-bindingCount-03002", 67}, { {"If bindingCount is not zero, bindingCount must equal VkDescriptorSetLayoutCreateInfo::bindingCount", 98}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingAccelerationStructureUpdateAfterBind-03570", 108}, { {"If VkPhysicalDeviceAccelerationStructureFeaturesKHR::descriptorBindingAccelerationStructureUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR or VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 307}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingInlineUniformBlockUpdateAfterBind-02211", 105}, { {"If VkPhysicalDeviceInlineUniformBlockFeatures::descriptorBindingInlineUniformBlockUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 244}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingPartiallyBound-03013", 86}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingPartiallyBound is not enabled, all elements of pBindingFlags must not include VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT", 183}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingSampledImageUpdateAfterBind-03006", 99}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingSampledImageUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_SAMPLER, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, or VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 305}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingStorageBufferUpdateAfterBind-03008", 100}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingStorageBufferUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 233}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingStorageImageUpdateAfterBind-03007", 99}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingStorageImageUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_STORAGE_IMAGE must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 231}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingStorageTensorUpdateAfterBind-09697", 100}, { {"If VkPhysicalDeviceTensorFeaturesARM::descriptorBindingStorageTensorUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_TENSOR_ARM must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 220}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingStorageTexelBufferUpdateAfterBind-03010", 105}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingStorageTexelBufferUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 244}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingUniformBufferUpdateAfterBind-03005", 100}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingUniformBufferUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 233}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingUniformTexelBufferUpdateAfterBind-03009", 105}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingUniformTexelBufferUpdateAfterBind is not enabled, all bindings with descriptor type VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER must not use VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT", 244}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingUpdateUnusedWhilePending-03012", 96}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingUpdateUnusedWhilePending is not enabled, all elements of pBindingFlags must not include VK_DESCRIPTOR_BINDING_UPDATE_UNUSED_WHILE_PENDING_BIT", 205}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-descriptorBindingVariableDescriptorCount-03014", 95}, { {"If VkPhysicalDeviceDescriptorIndexingFeatures::descriptorBindingVariableDescriptorCount is not enabled, all elements of pBindingFlags must not include VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT", 202}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-flags-03003", 60}, { {"If VkDescriptorSetLayoutCreateInfo::flags includes VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, then all elements of pBindingFlags must not include VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT, VK_DESCRIPTOR_BINDING_UPDATE_UNUSED_WHILE_PENDING_BIT, or VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT", 310}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-pBindingFlags-03004", 68}, { {"If an element of pBindingFlags includes VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT, then it must be the element with the highest binding number", 152}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-pBindingFlags-03015", 68}, { {"If an element of pBindingFlags includes VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT, that element's descriptorType must not be VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 221}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-pBindingFlags-parameter", 72}, { {"If bindingCount is not 0, pBindingFlags must be a valid pointer to an array of bindingCount valid combinations of VkDescriptorBindingFlagBits values", 148}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutBindingFlagsCreateInfo-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_BINDING_FLAGS_CREATE_INFO", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-binding-00279", 50}, { {"If the perStageDescriptorSet feature is not enabled, or flags does not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_PER_STAGE_BIT_NV, then the VkDescriptorSetLayoutBinding::binding members of the elements of the pBindings array must each have different values", 254}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-descriptorType-03001", 57}, { {"If any binding has the VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT bit set, then all bindings must not have descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 212}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-descriptorType-04594", 57}, { {"If a binding has a descriptorType value of VK_DESCRIPTOR_TYPE_MUTABLE_EXT, then pImmutableSamplers must be NULL", 111}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-00280", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, then all elements of pBindings must not have a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 222}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-00281", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, then the total number of elements of all bindings must be less than or equal to VkPhysicalDevicePushDescriptorProperties::maxPushDescriptors", 211}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-02208", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, then all elements of pBindings must not have a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 175}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-03000", 48}, { {"If any binding has the VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT bit set, flags must include VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT", 153}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-04590", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, flags must not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT", 148}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-04591", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, pBindings must not have a descriptorType of VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 145}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-04592", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT, flags must not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-04596", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT, VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT::mutableDescriptorType must be enabled", 161}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-08000", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, then all elements of pBindings must not have a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 228}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-08001", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT, flags must also contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 168}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-08002", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, then flags must not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT", 163}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-08003", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, then flags must not contain VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT", 159}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-09463", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PER_STAGE_BIT_NV, then perStageDescriptorSet must be enabled", 110}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-09464", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PER_STAGE_BIT_NV, then there must not be any two elements of the pBindings array with the same VkDescriptorSetLayoutBinding::binding value and their VkDescriptorSetLayoutBinding::stageFlags containing the same bit", 263}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-10354", 48}, { {"If flags contains VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT, and the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 158}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-flags-parameter", 52}, { {"flags must be a valid combination of VkDescriptorSetLayoutCreateFlagBits values", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-mutableDescriptorType-04595", 64}, { {"If VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT::mutableDescriptorType is not enabled, pBindings must not contain a descriptorType of VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 168}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-pBindings-07303", 52}, { {"If any element pBindings[i] has a descriptorType of VK_DESCRIPTOR_TYPE_MUTABLE_EXT, then the pNext chain must include a VkMutableDescriptorTypeCreateInfoEXT structure, and mutableDescriptorTypeListCount must be greater than i", 225}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-pBindings-parameter", 56}, { {"If bindingCount is not 0, pBindings must be a valid pointer to an array of bindingCount valid VkDescriptorSetLayoutBinding structures", 133}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-pNext-pNext", 48}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDescriptorSetLayoutBindingFlagsCreateInfo or VkMutableDescriptorTypeCreateInfoEXT", 214}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO", 65}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutCreateInfo-sType-unique", 49}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutHostMappingInfoVALVE-pNext-pNext", 58}, { {"pNext must be NULL", 18}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkDescriptorSetLayoutHostMappingInfoVALVE-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_HOST_MAPPING_INFO_VALVE", 77}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkDescriptorSetLayoutSupport-pNext-pNext", 45}, { {"pNext must be NULL or a pointer to a valid instance of VkDescriptorSetVariableDescriptorCountLayoutSupport", 106}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutSupport-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_SUPPORT", 61}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetLayoutSupport-sType-unique", 46}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetVariableDescriptorCountAllocateInfo-descriptorSetCount-03045", 80}, { {"If descriptorSetCount is not zero, descriptorSetCount must equal VkDescriptorSetAllocateInfo::descriptorSetCount", 112}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetVariableDescriptorCountAllocateInfo-pDescriptorCounts-parameter", 83}, { {"If descriptorSetCount is not 0, pDescriptorCounts must be a valid pointer to an array of descriptorSetCount uint32_t values", 123}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetVariableDescriptorCountAllocateInfo-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_ALLOCATE_INFO", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorSetVariableDescriptorCountLayoutSupport-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_LAYOUT_SUPPORT", 87}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-commonparent", 54}, { {"Both of descriptorSetLayout, and pipelineLayout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 167}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-descriptorUpdateEntryCount-arraylength", 80}, { {"descriptorUpdateEntryCount must be greater than 0", 49}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-flags-zerobitmask", 59}, { {"flags must be 0", 15}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-pDescriptorUpdateEntries-parameter", 76}, { {"pDescriptorUpdateEntries must be a valid pointer to an array of descriptorUpdateEntryCount valid VkDescriptorUpdateTemplateEntry structures", 139}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_DESCRIPTOR_UPDATE_TEMPLATE_CREATE_INFO", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-00350", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_DESCRIPTOR_SET, descriptorSetLayout must be a valid VkDescriptorSetLayout handle", 134}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-00351", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS, pipelineBindPoint must be a valid VkPipelineBindPoint value", 131}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-00352", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS, pipelineLayout must be a valid VkPipelineLayout handle", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-00353", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS, set must be the unique set number in the pipeline layout that uses a descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT", 236}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-04615", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_DESCRIPTOR_SET, descriptorSetLayout must not contain a binding with type VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 157}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-10355", 60}, { {"If templateType is VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS, and the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 159}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateCreateInfo-templateType-parameter", 64}, { {"templateType must be a valid VkDescriptorUpdateTemplateType value", 65}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateEntry-descriptor-02226", 53}, { {"If descriptor type is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, dstArrayElement must be an integer multiple of 4", 111}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateEntry-descriptor-02227", 53}, { {"If descriptor type is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, descriptorCount must be an integer multiple of 4", 111}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateEntry-descriptorType-parameter", 61}, { {"descriptorType must be a valid VkDescriptorType value", 53}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateEntry-dstArrayElement-00355", 58}, { {"dstArrayElement and descriptorCount must be less than or equal to the number of array elements in the descriptor set binding implicitly specified when using a descriptor update template to update descriptors, and all applicable consecutive bindings", 248}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDescriptorUpdateTemplateEntry-dstBinding-00354", 53}, { {"dstBinding must be a valid binding in the descriptor set layout implicitly specified when using a descriptor update template to update descriptors", 146}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkDeviceAddress-None-10894", 31}, { {"If a VkDeviceAddress was retrieved from a non-sparse buffer, that buffer must be bound completely and contiguously to a single VkDeviceMemory object", 148}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkDeviceAddress-size-11364", 31}, { {"A valid VkDeviceAddress must be equal to the sum of an address retrieved from a VkBuffer via vkGetBufferDeviceAddress, and any offset in the range [0, size), where size is the value of VkBufferCreateInfo::size used to create that VkBuffer", 238}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkDeviceAddressBindingCallbackDataEXT-baseAddress-parameter", 64}, { {"baseAddress must be a valid VkDeviceAddress value", 49}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceAddressBindingCallbackDataEXT-bindingType-parameter", 64}, { {"bindingType must be a valid VkDeviceAddressBindingTypeEXT value", 63}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceAddressBindingCallbackDataEXT-flags-parameter", 58}, { {"flags must be a valid combination of VkDeviceAddressBindingFlagBitsEXT values", 77}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceAddressBindingCallbackDataEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_ADDRESS_BINDING_CALLBACK_DATA_EXT", 72}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceAddressRangeKHR-address-11365", 42}, { {"The sum of address and size must be less than or equal to the sum of an address retrieved from a VkBuffer and the value of VkBufferCreateInfo::size used to create that VkBuffer", 176}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkDeviceAddressRangeKHR-address-parameter", 46}, { {"If address is not 0, address must be a valid VkDeviceAddress value", 66}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkDeviceAddressRangeKHR-size-11411", 39}, { {"If size is not 0, address must not be 0", 39}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkDeviceBufferMemoryRequirements-pCreateInfo-parameter", 59}, { {"pCreateInfo must be a valid pointer to a valid VkBufferCreateInfo structure", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceBufferMemoryRequirements-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceBufferMemoryRequirements-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_BUFFER_MEMORY_REQUIREMENTS", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceCreateInfo-None-04896", 34}, { {"If the sparseImageInt64Atomics feature is enabled, shaderImageInt64Atomics must be enabled", 90}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-None-04897", 34}, { {"If the sparseImageFloat32Atomics feature is enabled, shaderImageFloat32Atomics must be enabled", 94}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-None-04898", 34}, { {"If the sparseImageFloat32AtomicAdd feature is enabled, shaderImageFloat32AtomicAdd must be enabled", 98}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-None-08095", 34}, { {"If the descriptorBuffer feature is enabled, ppEnabledExtensionNames must not contain VK_AMD_shader_fragment_mask", 112}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-None-10778", 34}, { {"If the maintenance9 feature is not supported, queueCreateInfoCount must be greater than 0", 89}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-enabledLayerCount-12384", 47}, { {"enabledLayerCount must be 0", 27}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-flags-zerobitmask", 41}, { {"flags must be 0", 15}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-fragmentDensityMap-04481", 48}, { {"If the fragmentDensityMap feature is enabled, the pipelineFragmentShadingRate feature must not be enabled", 105}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-fragmentDensityMap-04482", 48}, { {"If the fragmentDensityMap feature is enabled, the primitiveFragmentShadingRate feature must not be enabled", 106}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-fragmentDensityMap-04483", 48}, { {"If the fragmentDensityMap feature is enabled, the attachmentFragmentShadingRate feature must not be enabled", 107}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pEnabledFeatures-parameter", 50}, { {"If pEnabledFeatures is not NULL, pEnabledFeatures must be a valid pointer to a valid VkPhysicalDeviceFeatures structure", 119}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-00373", 35}, { {"If the pNext chain includes a VkPhysicalDeviceFeatures2 structure, then pEnabledFeatures must be NULL", 101}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-02829", 35}, { {"If the pNext chain includes a VkPhysicalDeviceVulkan11Features structure, then it must not include a VkPhysicalDevice16BitStorageFeatures, VkPhysicalDeviceMultiviewFeatures, VkPhysicalDeviceVariablePointersFeatures, VkPhysicalDeviceProtectedMemoryFeatures, VkPhysicalDeviceSamplerYcbcrConversionFeatures, or VkPhysicalDeviceShaderDrawParametersFeatures structure", 362}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-02830", 35}, { {"If the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then it must not include a VkPhysicalDevice8BitStorageFeatures, VkPhysicalDeviceShaderAtomicInt64Features, VkPhysicalDeviceShaderFloat16Int8Features, VkPhysicalDeviceDescriptorIndexingFeatures, VkPhysicalDeviceScalarBlockLayoutFeatures, VkPhysicalDeviceImagelessFramebufferFeatures, VkPhysicalDeviceUniformBufferStandardLayoutFeatures, VkPhysicalDeviceShaderSubgroupExtendedTypesFeatures, VkPhysicalDeviceSeparateDepthStencilLayoutsFeatures, VkPhysicalDeviceHostQueryResetFeatures, VkPhysicalDeviceTimelineSemaphoreFeatures, VkPhysicalDeviceBufferDeviceAddressFeatures, or VkPhysicalDeviceVulkanMemoryModelFeatures structure", 698}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-04748", 35}, { {"If the pNext chain includes a VkPhysicalDeviceVulkan12Features structure and VkPhysicalDeviceVulkan12Features::bufferDeviceAddress is VK_TRUE, ppEnabledExtensionNames must not contain VK_EXT_buffer_device_address", 212}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-06532", 35}, { {"If the pNext chain includes a VkPhysicalDeviceVulkan13Features structure, then it must not include a VkPhysicalDeviceDynamicRenderingFeatures, VkPhysicalDeviceImageRobustnessFeatures, VkPhysicalDeviceInlineUniformBlockFeatures, VkPhysicalDeviceMaintenance4Features, VkPhysicalDevicePipelineCreationCacheControlFeatures, VkPhysicalDevicePrivateDataFeatures, VkPhysicalDeviceShaderDemoteToHelperInvocationFeatures, VkPhysicalDeviceShaderIntegerDotProductFeatures, VkPhysicalDeviceShaderTerminateInvocationFeatures, VkPhysicalDeviceSubgroupSizeControlFeatures, VkPhysicalDeviceSynchronization2Features, VkPhysicalDeviceTextureCompressionASTCHDRFeatures, or VkPhysicalDeviceZeroInitializeWorkgroupMemoryFeatures structure", 717}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-09396", 35}, { {"If the pNext chain includes a VkDeviceQueueShaderCoreControlCreateInfoARM structure, then it must not be included in the pNext chain of any of the VkDeviceQueueCreateInfo structures in pQueueCreateInfos", 202}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-09397", 35}, { {"If the pNext chain includes a VkDeviceQueueShaderCoreControlCreateInfoARM structure then VkPhysicalDeviceSchedulingControlsPropertiesARM::schedulingControlsFlags must contain VK_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_SHADER_CORE_COUNT_ARM", 235}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-10360", 35}, { {"If the pNext chain includes a VkPhysicalDeviceVulkan14Features structure, then it must not include a VkPhysicalDeviceGlobalPriorityQueryFeatures, VkPhysicalDeviceShaderSubgroupRotateFeatures, VkPhysicalDeviceShaderFloatControls2Features, VkPhysicalDeviceShaderExpectAssumeFeatures, VkPhysicalDeviceLineRasterizationFeatures, VkPhysicalDeviceVertexAttributeDivisorFeatures, VkPhysicalDeviceIndexTypeUint8Features, VkPhysicalDeviceDynamicRenderingLocalReadFeatures, VkPhysicalDeviceMaintenance5Features, VkPhysicalDeviceMaintenance6Features, VkPhysicalDevicePipelineProtectedAccessFeatures, VkPhysicalDevicePipelineRobustnessFeatures, or VkPhysicalDeviceHostImageCopyFeatures structure", 683}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pNext-pNext", 35}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid struct for extending VkDeviceCreateInfo", 158}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pProperties-04451", 41}, { {"If the VK_KHR_portability_subset extension is included in pProperties of vkEnumerateDeviceExtensionProperties, ppEnabledExtensionNames must include \"VK_KHR_portability_subset\"", 177}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pQueueCreateInfos-06654", 47}, { {"If multiple elements of pQueueCreateInfos share the same queueFamilyIndex, then all of such elements must have the same global priority level, which can be specified explicitly by the including a VkDeviceQueueGlobalPriorityCreateInfo structure in the pNext chain, or by the implicit default value", 296}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pQueueCreateInfos-06755", 47}, { {"If multiple elements of pQueueCreateInfos share the same queueFamilyIndex, the sum of their queueCount members must be less than or equal to the queueCount member of the VkQueueFamilyProperties structure, as returned by vkGetPhysicalDeviceQueueFamilyProperties in the pQueueFamilyProperties[queueFamilyIndex]", 308}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-pQueueCreateInfos-parameter", 51}, { {"If queueCreateInfoCount is not 0, pQueueCreateInfos must be a valid pointer to an array of queueCreateInfoCount valid VkDeviceQueueCreateInfo structures", 152}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-00374", 53}, { {"ppEnabledExtensionNames must not contain both VK_KHR_maintenance1 and VK_AMD_negative_viewport_height", 101}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-01840", 53}, { {"If VkPhysicalDeviceProperties::apiVersion advertises Vulkan 1.1 or later, ppEnabledExtensionNames must not contain VK_AMD_negative_viewport_height", 146}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-02831", 53}, { {"If ppEnabledExtensionNames contains \"VK_KHR_draw_indirect_count\" and the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then VkPhysicalDeviceVulkan12Features::drawIndirectCount must be VK_TRUE", 214}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-02832", 53}, { {"If ppEnabledExtensionNames contains \"VK_KHR_sampler_mirror_clamp_to_edge\" and the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then VkPhysicalDeviceVulkan12Features::samplerMirrorClampToEdge must be VK_TRUE", 230}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-02833", 53}, { {"If ppEnabledExtensionNames contains \"VK_EXT_descriptor_indexing\" and the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then VkPhysicalDeviceVulkan12Features::descriptorIndexing must be VK_TRUE", 215}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-02834", 53}, { {"If ppEnabledExtensionNames contains \"VK_EXT_sampler_filter_minmax\" and the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then VkPhysicalDeviceVulkan12Features::samplerFilterMinmax must be VK_TRUE", 218}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-02835", 53}, { {"If ppEnabledExtensionNames contains \"VK_EXT_shader_viewport_index_layer\" and the pNext chain includes a VkPhysicalDeviceVulkan12Features structure, then VkPhysicalDeviceVulkan12Features::shaderOutputViewportIndex and VkPhysicalDeviceVulkan12Features::shaderOutputLayer must both be VK_TRUE", 291}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-03328", 53}, { {"ppEnabledExtensionNames must not contain both VK_KHR_buffer_device_address and VK_EXT_buffer_device_address", 107}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-04476", 53}, { {"If ppEnabledExtensionNames contains \"VK_KHR_shader_draw_parameters\" and the pNext chain includes a VkPhysicalDeviceVulkan11Features structure, then VkPhysicalDeviceVulkan11Features::shaderDrawParameters must be VK_TRUE", 220}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-10858", 53}, { {"If ppEnabledExtensionNames contains \"VK_KHR_push_descriptor\" and the pNext chain includes a VkPhysicalDeviceVulkan14Features structure, then VkPhysicalDeviceVulkan14Features::pushDescriptor must be VK_TRUE", 207}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledExtensionNames-parameter", 57}, { {"If enabledExtensionCount is not 0, ppEnabledExtensionNames must be a valid pointer to an array of enabledExtensionCount null-terminated UTF-8 strings", 149}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-ppEnabledLayerNames-12385", 49}, { {"ppEnabledLayerNames must be NULL", 32}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-queueFamilyIndex-02802", 46}, { {"The combination of the values in the queueFamilyIndex and flags members of each element of pQueueCreateInfos must be unique within pQueueCreateInfos", 148}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-queueFamilyIndex-11831", 46}, { {"If any element of pQueueCreateInfos specifies a queueFamilyIndex that supports VK_QUEUE_DATA_GRAPH_BIT_ARM and that queueFamilyIndex enumerates an engine through vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with type VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_NEURAL_QCOM or VK_PHYSICAL_DEVICE_DATA_GRAPH_PROCESSING_ENGINE_TYPE_COMPUTE_QCOM, a VkPhysicalDeviceDataGraphModelFeaturesQCOM structure must be included in pNext with dataGraphModel set to VK_TRUE", 475}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-robustBufferAccess-10247", 48}, { {"If the robustBufferAccess feature is enabled, and robustBufferAccessUpdateAfterBind is VK_FALSE, then descriptorBindingUniformBufferUpdateAfterBind, descriptorBindingStorageBufferUpdateAfterBind, descriptorBindingUniformTexelBufferUpdateAfterBind, and descriptorBindingStorageTexelBufferUpdateAfterBind must not be enabled", 322}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO", 50}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-sType-unique", 36}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkDeviceDeviceMemoryReportCreateInfoEXT or VkDevicePrivateDataCreateInfo", 182}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-shadingRateImage-04478", 46}, { {"If the shadingRateImage feature is enabled, the pipelineFragmentShadingRate feature must not be enabled", 103}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-shadingRateImage-04479", 46}, { {"If the shadingRateImage feature is enabled, the primitiveFragmentShadingRate feature must not be enabled", 104}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-shadingRateImage-04480", 46}, { {"If the shadingRateImage feature is enabled, the attachmentFragmentShadingRate feature must not be enabled", 105}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceCreateInfo-sparseImageFloat32AtomicMinMax-04975", 60}, { {"If the sparseImageFloat32AtomicMinMax feature is enabled, shaderImageFloat32AtomicMinMax must be enabled", 104}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceDeviceMemoryReportCreateInfoEXT-flags-zerobitmask", 62}, { {"flags must be 0", 15}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceDeviceMemoryReportCreateInfoEXT-pfnUserCallback-parameter", 70}, { {"pfnUserCallback must be a valid PFN_vkDeviceMemoryReportCallbackEXT value", 73}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceDeviceMemoryReportCreateInfoEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_DEVICE_MEMORY_REPORT_CREATE_INFO_EXT", 75}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceDiagnosticsConfigCreateInfoNV-flags-parameter", 58}, { {"flags must be a valid combination of VkDeviceDiagnosticsConfigFlagBitsNV values", 79}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceDiagnosticsConfigCreateInfoNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_DIAGNOSTICS_CONFIG_CREATE_INFO_NV", 72}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceEventInfoEXT-deviceEvent-parameter", 47}, { {"deviceEvent must be a valid VkDeviceEventTypeEXT value", 54}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDeviceEventInfoEXT-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDeviceEventInfoEXT-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_EVENT_INFO_EXT", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDeviceFaultCountsEXT-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultCountsEXT-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_FAULT_COUNTS_EXT", 55}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultDebugInfoKHR-pNext-pNext", 42}, { {"pNext must be NULL or a pointer to a valid instance of VkDeviceFaultShaderAbortMessageInfoKHR", 93}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultDebugInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_FAULT_DEBUG_INFO_KHR", 59}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultDebugInfoKHR-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultInfoEXT-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultInfoEXT-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_EXT", 53}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultInfoKHR-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultInfoKHR-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_KHR", 53}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultShaderAbortMessageInfoKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_FAULT_SHADER_ABORT_MESSAGE_INFO_KHR", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultVendorBinaryHeaderVersionOneEXT-headerSize-07340", 66}, { {"headerSize must be 56", 21}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultVendorBinaryHeaderVersionOneEXT-headerVersion-07341", 69}, { {"headerVersion must be VK_DEVICE_FAULT_VENDOR_BINARY_HEADER_VERSION_ONE_KHR", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceFaultVendorBinaryHeaderVersionOneKHR-headerVersion-parameter", 73}, { {"headerVersion must be a valid VkDeviceFaultVendorBinaryHeaderVersionKHR value", 77}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkDeviceGroupBindSparseInfo-memoryDeviceIndex-01119", 56}, { {"Each memory allocation bound in this batch must have allocated an instance for memoryDeviceIndex", 96}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkDeviceGroupBindSparseInfo-resourceDeviceIndex-01118", 58}, { {"resourceDeviceIndex and memoryDeviceIndex must both be valid device indices", 75}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkDeviceGroupBindSparseInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_BIND_SPARSE_INFO", 61}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkDeviceGroupCommandBufferBeginInfo-deviceMask-00106", 57}, { {"deviceMask must be a valid device mask value", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupCommandBufferBeginInfo-deviceMask-00107", 57}, { {"deviceMask must not be zero", 27}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupCommandBufferBeginInfo-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_COMMAND_BUFFER_BEGIN_INFO", 70}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupDeviceCreateInfo-pPhysicalDevices-00375", 57}, { {"Each element of pPhysicalDevices must be unique", 47}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceGroupDeviceCreateInfo-pPhysicalDevices-00376", 57}, { {"All elements of pPhysicalDevices must be in the same device group as enumerated by vkEnumeratePhysicalDeviceGroups", 114}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceGroupDeviceCreateInfo-pPhysicalDevices-parameter", 61}, { {"If physicalDeviceCount is not 0, pPhysicalDevices must be a valid pointer to an array of physicalDeviceCount valid VkPhysicalDevice handles", 139}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceGroupDeviceCreateInfo-physicalDeviceCount-00377", 60}, { {"If physicalDeviceCount is not 0, the physicalDevice parameter of vkCreateDevice must be an element of pPhysicalDevices", 118}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceGroupDeviceCreateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_DEVICE_CREATE_INFO", 63}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceGroupPresentCapabilitiesKHR-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentCapabilitiesKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_CAPABILITIES_KHR", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-01298", 43}, { {"If mode is VK_DEVICE_GROUP_PRESENT_MODE_LOCAL_BIT_KHR, then each element of pDeviceMasks must have exactly one bit set, and the corresponding element of VkDeviceGroupPresentCapabilitiesKHR::presentMask must be non-zero", 218}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-01299", 43}, { {"If mode is VK_DEVICE_GROUP_PRESENT_MODE_REMOTE_BIT_KHR, then each element of pDeviceMasks must have exactly one bit set, and some physical device in the logical device must include that bit in its VkDeviceGroupPresentCapabilitiesKHR::presentMask", 245}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-01300", 43}, { {"If mode is VK_DEVICE_GROUP_PRESENT_MODE_SUM_BIT_KHR, then each element of pDeviceMasks must have a value for which all set bits are set in one of the elements of VkDeviceGroupPresentCapabilitiesKHR::presentMask", 210}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-01301", 43}, { {"If mode is VK_DEVICE_GROUP_PRESENT_MODE_LOCAL_MULTI_DEVICE_BIT_KHR, then for each bit set in each element of pDeviceMasks, the corresponding element of VkDeviceGroupPresentCapabilitiesKHR::presentMask must be non-zero", 217}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-01303", 43}, { {"mode must have exactly one bit set, and that bit must have been included in VkDeviceGroupSwapchainCreateInfoKHR::modes", 118}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-mode-parameter", 47}, { {"mode must be a valid VkDeviceGroupPresentModeFlagBitsKHR value", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-pDeviceMasks-01302", 51}, { {"The value of each element of pDeviceMasks must be equal to the device mask passed in VkAcquireNextImageInfoKHR::deviceMask when the image index was last acquired", 161}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-pDeviceMasks-parameter", 55}, { {"If swapchainCount is not 0, pDeviceMasks must be a valid pointer to an array of swapchainCount uint32_t values", 110}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_INFO_KHR", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupPresentInfoKHR-swapchainCount-01297", 53}, { {"swapchainCount must equal 0 or VkPresentInfoKHR::swapchainCount", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-deviceMask-00905", 54}, { {"deviceMask must be a valid device mask value", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-deviceMask-00906", 54}, { {"deviceMask must not be zero", 27}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-deviceMask-00907", 54}, { {"deviceMask must be a subset of the command buffer's initial device mask", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-deviceRenderAreaCount-00908", 65}, { {"deviceRenderAreaCount must either be zero or equal to the number of physical devices in the logical device", 106}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-extent-08998", 50}, { {"The extent.width member of any element of pDeviceRenderAreas must be greater than 0", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-extent-08999", 50}, { {"The extent.height member of any element of pDeviceRenderAreas must be greater than 0", 84}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-offset-06166", 50}, { {"The offset.x member of any element of pDeviceRenderAreas must be greater than or equal to 0", 91}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-offset-06167", 50}, { {"The offset.y member of any element of pDeviceRenderAreas must be greater than or equal to 0", 91}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-offset-06168", 50}, { {"The sum of the offset.x and extent.width members of any element of pDeviceRenderAreas must be less than or equal to maxFramebufferWidth", 135}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-offset-06169", 50}, { {"The sum of the offset.y and extent.height members of any element of pDeviceRenderAreas must be less than or equal to maxFramebufferHeight", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-pDeviceRenderAreas-parameter", 66}, { {"If deviceRenderAreaCount is not 0, pDeviceRenderAreas must be a valid pointer to an array of deviceRenderAreaCount VkRect2D structures", 134}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupRenderPassBeginInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_RENDER_PASS_BEGIN_INFO", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-commandBufferCount-00083", 53}, { {"commandBufferCount must equal VkSubmitInfo::commandBufferCount", 62}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-pCommandBufferDeviceMasks-00086", 60}, { {"All elements of pCommandBufferDeviceMasks must be valid device masks", 68}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-pCommandBufferDeviceMasks-parameter", 64}, { {"If commandBufferCount is not 0, pCommandBufferDeviceMasks must be a valid pointer to an array of commandBufferCount uint32_t values", 131}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-pSignalSemaphoreDeviceIndices-parameter", 68}, { {"If signalSemaphoreCount is not 0, pSignalSemaphoreDeviceIndices must be a valid pointer to an array of signalSemaphoreCount uint32_t values", 139}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-pWaitSemaphoreDeviceIndices-00085", 62}, { {"All elements of pWaitSemaphoreDeviceIndices and pSignalSemaphoreDeviceIndices must be valid device indices", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-pWaitSemaphoreDeviceIndices-parameter", 66}, { {"If waitSemaphoreCount is not 0, pWaitSemaphoreDeviceIndices must be a valid pointer to an array of waitSemaphoreCount uint32_t values", 133}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_SUBMIT_INFO", 56}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-signalSemaphoreCount-00084", 55}, { {"signalSemaphoreCount must equal VkSubmitInfo::signalSemaphoreCount", 66}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSubmitInfo-waitSemaphoreCount-00082", 53}, { {"waitSemaphoreCount must equal VkSubmitInfo::waitSemaphoreCount", 62}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkDeviceGroupSwapchainCreateInfoKHR-modes-parameter", 56}, { {"modes must be a valid combination of VkDeviceGroupPresentModeFlagBitsKHR values", 79}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupSwapchainCreateInfoKHR-modes-requiredbitmask", 62}, { {"modes must not be 0", 19}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceGroupSwapchainCreateInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_GROUP_SWAPCHAIN_CREATE_INFO_KHR", 70}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-06416", 54}, { {"The pCreateInfo->pNext chain must not contain a VkImageSwapchainCreateInfoKHR structure", 87}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-06417", 54}, { {"If pCreateInfo->format specifies a multi-planar format and pCreateInfo->flags has VK_IMAGE_CREATE_DISJOINT_BIT set then planeAspect must not be VK_IMAGE_ASPECT_NONE_KHR", 168}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-06419", 54}, { {"If pCreateInfo->flags has VK_IMAGE_CREATE_DISJOINT_BIT set and if the pCreateInfo->tiling is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, then planeAspect must be a single valid multi-planar aspect mask bit", 212}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-06420", 54}, { {"If pCreateInfo->tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then planeAspect must be a single valid memory plane for the image (that is, aspectMask must specify a plane index that is less than the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with the image's format and VkImageDrmFormatModifierPropertiesEXT::drmFormatModifier)", 362}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-06776", 54}, { {"The pCreateInfo->pNext chain must not contain a VkImageDrmFormatModifierExplicitCreateInfoEXT structure", 103}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkImageCreateInfo structure", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pNext-06996", 48}, { {"Applications also must not call vkGetDeviceImageMemoryRequirements with a VkImageCreateInfo whose pNext chain includes a VkExternalFormatANDROID structure with non-zero externalFormat", 183}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pNext-08962", 48}, { {"Applications also must not call vkGetDeviceImageMemoryRequirements with a VkImageCreateInfo whose pNext chain includes a VkExternalFormatQNX structure with non-zero externalFormat", 179}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-planeAspect-12399", 54}, { {"If planeAspect is not 0, pCreateInfo->tiling is not VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, and pCreateInfo->flags does not have VK_IMAGE_CREATE_DISJOINT_BIT set, planeAspect must be a valid VkImageAspectFlagBits value", 219}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageMemoryRequirements-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_IMAGE_MEMORY_REQUIREMENTS", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-arrayLayer-01717", 50}, { {"The arrayLayer member of pSubresource must be less than the arrayLayers specified in pCreateInfo", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-aspectMask-00997", 50}, { {"The aspectMask member of pSubresource must only have a single bit set", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-format-04462", 46}, { {"If format of the pCreateInfo has a depth component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_DEPTH_BIT", 128}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-format-04463", 46}, { {"If format of the pCreateInfo has a stencil component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_STENCIL_BIT", 132}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-format-04464", 46}, { {"If format of the pCreateInfo does not contain a stencil or depth component, the aspectMask member of pSubresource must not contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 187}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-format-08886", 46}, { {"If format of the image is a color format that is not a multi-planar format, and tiling of the pCreateInfo is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, the aspectMask member of pSubresource must be VK_IMAGE_ASPECT_COLOR_BIT", 231}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-mipLevel-01716", 48}, { {"The mipLevel member of pSubresource must be less than the mipLevels specified in pCreateInfo", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-pCreateInfo-parameter", 55}, { {"pCreateInfo must be a valid pointer to a valid VkImageCreateInfo structure", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-pSubresource-parameter", 56}, { {"pSubresource must be a valid pointer to a valid VkImageSubresource2 structure", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_IMAGE_SUBRESOURCE_INFO", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceImageSubresourceInfo-tiling-08717", 46}, { {"If the tiling of the pCreateInfo is VK_IMAGE_TILING_LINEAR and has a multi-planar format, then the aspectMask member of pSubresource must be a single valid multi-planar aspect mask bit", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstFlags-13100", 41}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstFlags-13101", 41}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstFlags-parameter", 45}, { {"dstFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13018", 41}, { {"The buffer from which dstRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT", 103}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13097", 41}, { {"If the range specified by dstRange is not bound completely to memory when accessed, dstFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13098", 41}, { {"If the buffer from which the range specified by dstRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 203}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13099", 41}, { {"If the buffer from which the range specified by dstRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 210}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13122", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 282}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13123", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 230}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13124", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 319}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-dstRange-13125", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_MEMORY_COPY_KHR", 54}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-size-13016", 37}, { {"The size member of dstRange must be greater than or equal to the size member of srcRange", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcFlags-13100", 41}, { {"srcFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcFlags-13101", 41}, { {"srcFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcFlags-parameter", 45}, { {"srcFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13017", 41}, { {"The buffer from which srcRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_SRC_BIT", 103}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13097", 41}, { {"If the range specified by srcRange is not bound completely to memory when accessed, srcFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13098", 41}, { {"If the buffer from which the range specified by srcRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, srcFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 203}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13099", 41}, { {"If the buffer from which the range specified by srcRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, srcFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 210}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13122", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing srcRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, srcFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 282}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13123", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing srcRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, srcFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 230}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13124", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing srcRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, srcFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 319}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryCopyKHR-srcRange-13125", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing srcRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, srcFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressFlags-13100", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressFlags-13101", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressFlags-parameter", 54}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressImageHeight-09102", 56}, { {"addressImageHeight must be 0, or greater than or equal to the height member of imageExtent", 90}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13097", 50}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13098", 50}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13099", 50}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13122", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13123", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13124", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRange-13125", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-addressRowLength-09101", 54}, { {"addressRowLength must be 0, or greater than or equal to the width member of imageExtent", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-aspectMask-09103", 48}, { {"The aspectMask member of imageSubresource must only have a single bit set", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-imageExtent-06659", 49}, { {"imageExtent.width must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-imageExtent-06660", 49}, { {"imageExtent.height must not be 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-imageExtent-06661", 49}, { {"imageExtent.depth must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-imageLayout-parameter", 53}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-imageSubresource-parameter", 58}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-pNext-pNext", 43}, { {"pNext must be NULL or a pointer to a valid instance of VkCopyCommandTransformInfoQCOM", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_MEMORY_IMAGE_COPY_KHR", 60}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-sType-unique", 44}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryImageCopyKHR-size-13037", 42}, { {"The size member of addressRange must be large enough to contain all address locations that are accessed by this copy according to Buffer and Image Addressing", 157}, { "chapters/copies.html", 20} } },
        { {"VUID-VkDeviceMemoryOpaqueCaptureAddressInfo-memory-03336", 56}, { {"memory must have been allocated with VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT", 74}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDeviceMemoryOpaqueCaptureAddressInfo-memory-parameter", 60}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDeviceMemoryOpaqueCaptureAddressInfo-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDeviceMemoryOpaqueCaptureAddressInfo-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_MEMORY_OPAQUE_CAPTURE_ADDRESS_INFO", 73}, { "chapters/memory.html", 20} } },
        { {"VUID-VkDeviceMemoryOverallocationCreateInfoAMD-overallocationBehavior-parameter", 79}, { {"overallocationBehavior must be a valid VkMemoryOverallocationBehaviorAMD value", 78}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceMemoryOverallocationCreateInfoAMD-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_MEMORY_OVERALLOCATION_CREATE_INFO_AMD", 76}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceMemoryReportCallbackDataEXT-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceMemoryReportCallbackDataEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_MEMORY_REPORT_CALLBACK_DATA_EXT", 70}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDevicePipelineBinaryInternalCacheControlKHR-disableInternalCache-09602", 77}, { {"If VkPhysicalDevicePipelineBinaryPropertiesKHR::pipelineBinaryInternalCacheControl is VK_FALSE, disableInternalCache must be VK_FALSE", 133}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDevicePipelineBinaryInternalCacheControlKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_PIPELINE_BINARY_INTERNAL_CACHE_CONTROL_KHR", 81}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDevicePrivateDataCreateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_PRIVATE_DATA_CREATE_INFO", 63}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-flags-02861", 40}, { {"If the protectedMemory feature is not enabled, the VK_DEVICE_QUEUE_CREATE_PROTECTED_BIT bit of flags must not be set", 116}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-flags-06449", 40}, { {"If flags includes VK_DEVICE_QUEUE_CREATE_PROTECTED_BIT, queueFamilyIndex must be the index of a queue family that includes the VK_QUEUE_PROTECTED_BIT capability", 160}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-flags-parameter", 44}, { {"flags must be a valid combination of VkDeviceQueueCreateFlagBits values", 71}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-internallySynchronizedQueues-12348", 63}, { {"If the internallySynchronizedQueues feature is not enabled, flags must not include VK_DEVICE_QUEUE_CREATE_INTERNALLY_SYNCHRONIZED_BIT_KHR", 137}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-pNext-09398", 40}, { {"If the pNext chain includes a VkDeviceQueueShaderCoreControlCreateInfoARM structure then VkPhysicalDeviceSchedulingControlsPropertiesARM::schedulingControlsFlags must contain VK_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_SHADER_CORE_COUNT_ARM", 235}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-pNext-pNext", 40}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceQueueGlobalPriorityCreateInfo or VkDeviceQueueShaderCoreControlCreateInfoARM", 215}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-pQueuePriorities-00383", 51}, { {"Each element of pQueuePriorities must be between 0.0 and 1.0 inclusive", 70}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-pQueuePriorities-parameter", 55}, { {"pQueuePriorities must be a valid pointer to an array of queueCount float values", 79}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-queueCount-00382", 45}, { {"queueCount must be less than or equal to the queueCount member of the VkQueueFamilyProperties structure, as returned by vkGetPhysicalDeviceQueueFamilyProperties in the pQueueFamilyProperties[queueFamilyIndex]", 208}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-queueCount-arraylength", 51}, { {"queueCount must be greater than 0", 33}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-queueFamilyIndex-00381", 51}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties", 113}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO", 56}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueCreateInfo-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueGlobalPriorityCreateInfo-globalPriority-parameter", 67}, { {"globalPriority must be a valid VkQueueGlobalPriority value", 58}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueGlobalPriorityCreateInfo-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_QUEUE_GLOBAL_PRIORITY_CREATE_INFO", 72}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-flags-06225", 35}, { {"flags must be equal to VkDeviceQueueCreateInfo::flags for a VkDeviceQueueCreateInfo structure for the queue family indicated by queueFamilyIndex when device was created", 168}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-flags-parameter", 39}, { {"flags must be a valid combination of VkDeviceQueueCreateFlagBits values", 71}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-pNext-pNext", 35}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-queueFamilyIndex-01842", 46}, { {"queueFamilyIndex must be one of the queue family indices specified when device was created, via the VkDeviceQueueCreateInfo structure", 133}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-queueIndex-01843", 40}, { {"queueIndex must be less than VkDeviceQueueCreateInfo::queueCount for the corresponding queue family and flags indicated by queueFamilyIndex and flags when device was created", 173}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueInfo2-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_QUEUE_INFO_2", 51}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueShaderCoreControlCreateInfoARM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_QUEUE_SHADER_CORE_CONTROL_CREATE_INFO_ARM", 80}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceQueueShaderCoreControlCreateInfoARM-shaderCoreCount-09399", 70}, { {"shaderCoreCount must be greater than 0 and less than or equal to the total number of shader cores as reported via VkPhysicalDeviceShaderCoreBuiltinsPropertiesARM::shaderCoreCount", 178}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkDeviceTensorMemoryRequirementsARM-pCreateInfo-parameter", 62}, { {"pCreateInfo must be a valid pointer to a valid VkTensorCreateInfoARM structure", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceTensorMemoryRequirementsARM-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDeviceTensorMemoryRequirementsARM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_DEVICE_TENSOR_MEMORY_REQUIREMENTS_ARM", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkDirectDriverLoadingInfoLUNARG-flags-zerobitmask", 54}, { {"flags must be 0", 15}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectDriverLoadingInfoLUNARG-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_INFO_LUNARG", 65}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectDriverLoadingListLUNARG-driverCount-arraylength", 60}, { {"driverCount must be greater than 0", 34}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectDriverLoadingListLUNARG-mode-parameter", 51}, { {"mode must be a valid VkDirectDriverLoadingModeLUNARG value", 58}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectDriverLoadingListLUNARG-pDrivers-parameter", 55}, { {"pDrivers must be a valid pointer to an array of driverCount valid VkDirectDriverLoadingInfoLUNARG structures", 108}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectDriverLoadingListLUNARG-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_LIST_LUNARG", 65}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkDirectFBSurfaceCreateInfoEXT-dfb-04117", 45}, { {"dfb must point to a valid DirectFB IDirectFB", 44}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDirectFBSurfaceCreateInfoEXT-flags-zerobitmask", 53}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDirectFBSurfaceCreateInfoEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDirectFBSurfaceCreateInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_DIRECTFB_SURFACE_CREATE_INFO_EXT", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDirectFBSurfaceCreateInfoEXT-surface-04118", 49}, { {"surface must point to a valid DirectFB IDirectFBSurface", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDispatchGraphInfoAMDX-payloadCount-09171", 47}, { {"payloadCount must be no greater than maxExecutionGraphShaderPayloadCount", 72}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressFlags-13100", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressFlags-13101", 50}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressFlags-parameter", 54}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13097", 50}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13098", 50}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13099", 50}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13107", 50}, { {"The buffer from which addressRange was queried must have been created with VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag", 121}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13109", 50}, { {"addressRange.address must be a multiple of 4", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13122", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13123", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13124", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-addressRange-13125", 50}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-protectedNoFault-13108", 54}, { {"If protectedNoFault is not supported, the buffer from which addressRange was queried must not have been created with VK_BUFFER_CREATE_PROTECTED_BIT", 147}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirect2InfoKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DISPATCH_INDIRECT_2_INFO_KHR", 60}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirectCommand-x-00417", 38}, { {"x must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0]", 83}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirectCommand-y-00418", 38}, { {"y must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1]", 83}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchIndirectCommand-z-00419", 38}, { {"z must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2]", 83}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchParametersARM-maxQueuedWorkGroupBatches-12393", 60}, { {"maxQueuedWorkGroupBatches must be less than or equal to VkPhysicalDeviceSchedulingControlsDispatchParametersPropertiesARM::schedulingControlsMaxQueuedBatchesCount", 162}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchParametersARM-maxWarpsPerShaderCore-12392", 56}, { {"maxWarpsPerShaderCore must be less than or equal to VkPhysicalDeviceSchedulingControlsDispatchParametersPropertiesARM::schedulingControlsMaxWarpCount", 149}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchParametersARM-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchParametersARM-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_DISPATCH_PARAMETERS_ARM", 55}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchParametersARM-workGroupBatchSize-12394", 53}, { {"workGroupBatchSize must be less than or equal to VkPhysicalDeviceSchedulingControlsPropertiesARM::schedulingControlsMaxWorkGroupBatchSize", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchTileInfoQCOM-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDispatchTileInfoQCOM-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_DISPATCH_TILE_INFO_QCOM", 55}, { "chapters/dispatch.html", 22} } },
        { {"VUID-VkDisplayEventInfoEXT-displayEvent-parameter", 49}, { {"displayEvent must be a valid VkDisplayEventTypeEXT value", 56}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDisplayEventInfoEXT-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDisplayEventInfoEXT-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_EVENT_INFO_EXT", 54}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkDisplayModeCreateInfoKHR-flags-zerobitmask", 49}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeCreateInfoKHR-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeCreateInfoKHR-parameters-parameter", 52}, { {"parameters must be a valid VkDisplayModeParametersKHR structure", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeCreateInfoKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_MODE_CREATE_INFO_KHR", 60}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeParametersKHR-height-01991", 44}, { {"The height member of visibleRegion must be greater than 0", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeParametersKHR-refreshRate-01992", 49}, { {"refreshRate must be greater than 0", 34}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeParametersKHR-width-01990", 43}, { {"The width member of visibleRegion must be greater than 0", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeProperties2KHR-pNext-pNext", 44}, { {"pNext must be NULL or a pointer to a valid instance of VkDisplayModeStereoPropertiesNV", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeProperties2KHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_MODE_PROPERTIES_2_KHR", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeProperties2KHR-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayModeStereoPropertiesNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_MODE_STEREO_PROPERTIES_NV", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayNativeHdrSurfaceCapabilitiesAMD-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_NATIVE_HDR_SURFACE_CAPABILITIES_AMD", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneCapabilities2KHR-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneCapabilities2KHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_PLANE_CAPABILITIES_2_KHR", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneInfo2KHR-mode-parameter", 42}, { {"mode must be a valid VkDisplayModeKHR handle", 44}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneInfo2KHR-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneInfo2KHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_PLANE_INFO_2_KHR", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneProperties2KHR-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPlaneProperties2KHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_PLANE_PROPERTIES_2_KHR", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPowerInfoEXT-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPowerInfoEXT-powerState-parameter", 47}, { {"powerState must be a valid VkDisplayPowerStateEXT value", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPowerInfoEXT-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_POWER_INFO_EXT", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPresentInfoKHR-dstRect-01258", 42}, { {"dstRect must specify a rectangular region that is a subset of the visibleRegion parameter of the display mode the swapchain being presented uses", 144}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPresentInfoKHR-persistentContent-01259", 52}, { {"If the persistentContent member of the VkDisplayPropertiesKHR structure returned by vkGetPhysicalDeviceDisplayPropertiesKHR for the display the present operation targets is VK_FALSE, then persistent must be VK_FALSE", 215}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPresentInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_PRESENT_INFO_KHR", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayPresentInfoKHR-srcRect-01257", 42}, { {"srcRect must specify a rectangular region that is a subset of the image being presented", 87}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayProperties2KHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplayProperties2KHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_PROPERTIES_2_KHR", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-alphaMode-01254", 50}, { {"If alphaMode is VK_DISPLAY_PLANE_ALPHA_GLOBAL_BIT_KHR then globalAlpha must be between 0 and 1, inclusive", 105}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-alphaMode-01255", 50}, { {"alphaMode must be one of the bits present in the supportedAlpha member of VkDisplayPlaneCapabilitiesKHR for the display plane corresponding to displayMode", 154}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-alphaMode-parameter", 54}, { {"alphaMode must be a valid VkDisplayPlaneAlphaFlagBitsKHR value", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-displayMode-parameter", 56}, { {"displayMode must be a valid VkDisplayModeKHR handle", 51}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-flags-zerobitmask", 52}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-pNext-10284", 46}, { {"If the pNext chain includes a VkDisplaySurfaceStereoCreateInfoNV structure whose stereoType member is VK_DISPLAY_SURFACE_STEREO_TYPE_HDMI_3D_NV, then the hdmi3DSupported member of the VkDisplayModeStereoPropertiesNV structure in the pNext chain of the VkDisplayModeProperties2KHR structure returned by vkGetDisplayModeProperties2KHR for the display mode corresponding to displayMode must be VK_TRUE", 398}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL or a pointer to a valid instance of VkDisplaySurfaceStereoCreateInfoNV", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-planeIndex-01252", 51}, { {"planeIndex must be less than the number of display planes supported by the device as determined by calling vkGetPhysicalDeviceDisplayPlanePropertiesKHR", 151}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-planeReorderPossible-01253", 61}, { {"If the planeReorderPossible member of the VkDisplayPropertiesKHR structure returned by vkGetPhysicalDeviceDisplayPropertiesKHR for the display corresponding to displayMode is VK_TRUE then planeStackIndex must be less than the number of display planes supported by the device as determined by calling vkGetPhysicalDeviceDisplayPlanePropertiesKHR; otherwise planeStackIndex must equal the currentStackIndex member of VkDisplayPlanePropertiesKHR returned by vkGetPhysicalDeviceDisplayPlanePropertiesKHR for the display plane corresponding to displayMode", 550}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_SURFACE_CREATE_INFO_KHR", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-sType-unique", 47}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-transform-06740", 50}, { {"transform must be one of the bits present in the supportedTransforms member of VkDisplayPropertiesKHR for the display corresponding to displayMode", 146}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-transform-parameter", 54}, { {"transform must be a valid VkSurfaceTransformFlagBitsKHR value", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceCreateInfoKHR-width-01256", 46}, { {"The width and height members of imageExtent must be less than or equal to VkPhysicalDeviceLimits::maxImageDimension2D", 117}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceStereoCreateInfoNV-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_DISPLAY_SURFACE_STEREO_CREATE_INFO_NV", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDisplaySurfaceStereoCreateInfoNV-stereoType-parameter", 60}, { {"stereoType must be a valid VkDisplaySurfaceStereoTypeNV value", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkDrawIndexedIndirectCommand-None-09462", 44}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndexedIndirectCommand-firstInstance-00554", 53}, { {"If the drawIndirectFirstInstance feature is not enabled, firstInstance must be 0", 80}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndexedIndirectCommand-pNext-09461", 45}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndexedIndirectCommand-robustBufferAccess2-08798", 59}, { {"If the robustBufferAccess2 feature is not enabled, (indexSize x (firstIndex + indexCount)) must be less than or equal to the size of the bound index buffer range, with indexSize being based on the type specified by indexType, and the other parameters sourced from this command", 276}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressFlags-13100", 46}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressFlags-13101", 46}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressFlags-parameter", 50}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13097", 46}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13098", 46}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13099", 46}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13107", 46}, { {"The buffer from which addressRange was queried must have been created with VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13109", 46}, { {"addressRange.address must be a multiple of 4", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13122", 46}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13123", 46}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13124", 46}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-addressRange-13125", 46}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-drawCount-02718", 43}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-drawCount-02719", 43}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-protectedNoFault-13108", 50}, { {"If protectedNoFault is not supported, the buffer from which addressRange was queried must not have been created with VK_BUFFER_CREATE_PROTECTED_BIT", 147}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirect2InfoKHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_DRAW_INDIRECT_2_INFO_KHR", 56}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCommand-None-09462", 37}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCommand-firstInstance-00501", 46}, { {"If the drawIndirectFirstInstance feature is not enabled, firstInstance must be 0", 80}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCommand-pNext-09461", 38}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressFlags-13100", 51}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressFlags-13101", 51}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressFlags-parameter", 55}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13097", 51}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13098", 51}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13099", 51}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13107", 51}, { {"The buffer from which addressRange was queried must have been created with VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13109", 51}, { {"addressRange.address must be a multiple of 4", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13122", 51}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13123", 51}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13124", 51}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-addressRange-13125", 51}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressFlags-13100", 56}, { {"countAddressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 147}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressFlags-13101", 56}, { {"countAddressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 169}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressFlags-parameter", 60}, { {"countAddressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 83}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13097", 56}, { {"If the range specified by countAddressRange is not bound completely to memory when accessed, countAddressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 166}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13098", 56}, { {"If the buffer from which the range specified by countAddressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, countAddressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 221}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13099", 56}, { {"If the buffer from which the range specified by countAddressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, countAddressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 228}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13114", 56}, { {"The buffer from which countAddressRange was queried must have been created with VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13115", 56}, { {"countAddressRange.address must be a multiple of 4", 49}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13116", 56}, { {"The count stored in countAddressRange must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13117", 56}, { {"countAddressRange.size must be greater than or equal to 4", 57}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13122", 56}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing countAddressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, countAddressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 300}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13123", 56}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing countAddressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, countAddressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13124", 56}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing countAddressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, countAddressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-countAddressRange-13125", 56}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing countAddressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, countAddressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 274}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-protectedNoFault-13108", 55}, { {"If protectedNoFault is not supported, the buffer from which addressRange was queried must not have been created with VK_BUFFER_CREATE_PROTECTED_BIT", 147}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCount2InfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_DRAW_INDIRECT_COUNT_2_INFO_KHR", 62}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawIndirectCountIndirectCommandEXT-None-11122", 53}, { {"The buffer's usage flag from which the address was acquired must have the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkDrawIndirectCountIndirectCommandEXT-bufferAddress-parameter", 66}, { {"bufferAddress must be a valid VkDeviceAddress value", 51}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07322", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountX must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[0]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07323", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountY must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[1]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07324", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[2]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07325", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, The product of groupCountX, groupCountY and groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupTotalCount", 271}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07326", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountX must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[0]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07327", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountY must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[1]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07328", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[2]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandEXT-TaskEXT-07329", 52}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, The product of groupCountX, groupCountY and groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupTotalCount", 279}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrawMeshTasksIndirectCommandNV-taskCount-02175", 53}, { {"taskCount must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesNV::maxDrawMeshTasksCount", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkDrmFormatModifierPropertiesList2EXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_2_EXT", 73}, { "chapters/formats.html", 21} } },
        { {"VUID-VkDrmFormatModifierPropertiesListEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_EXT", 71}, { "chapters/formats.html", 21} } },
        { {"VUID-VkEventCreateInfo-flags-parameter", 38}, { {"flags must be a valid combination of VkEventCreateFlagBits values", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkEventCreateInfo-pNext-06790", 34}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkEventCreateInfo-pNext-pNext", 34}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExportMetalObjectCreateInfoEXT or VkImportMetalSharedEventInfoEXT", 198}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkEventCreateInfo-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_EVENT_CREATE_INFO", 49}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkEventCreateInfo-sType-unique", 35}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT", 142}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-CoalescedInputCountAMDX-09137", 73}, { {"There must be no two nodes in the pipeline that share the same name where one includes CoalescedInputCountAMDX and the other does not", 133}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-09134", 54}, { {"There must be no two nodes in the pipeline that share both the same shader name and index, as specified by VkPipelineShaderStageNodeCreateInfoAMDX", 146}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-09135", 54}, { {"There must be no two nodes in the pipeline that share the same shader name and have input payload declarations with different sizes", 131}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-09136", 54}, { {"There must be no two nodes in the pipeline that share the same name but have different execution models", 103}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-09497", 54}, { {"If the pNext chain does not include a VkPipelineCreateFlags2CreateInfo structure, flags must be a valid combination of VkPipelineCreateFlagBits values", 150}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-10391", 54}, { {"If a resource variables is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 169}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-None-11363", 54}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must not be VK_NULL_HANDLE", 142}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-PayloadNodeNameAMDX-09139", 69}, { {"If an output payload declared in any shader in the pipeline has a PayloadNodeNameAMDX decoration with a Node Name that matches the shader name of any other node in the graph, the size of the output payload must match the size of the input payload in the matching node", 267}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-StaticNumWorkgroupsAMDX-09138", 73}, { {"There must be no two nodes in the pipeline that share the same name where one includes StaticNumWorkgroupsAMDX and the other does not", 133}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-commonparent", 56}, { {"Both of basePipelineHandle, and layout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 158}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03365", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR", 85}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03366", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR", 89}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03367", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR", 82}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03368", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR", 90}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03369", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR", 76}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03370", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 72}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-03576", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 96}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-04945", 55}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV", 73}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-07369", 55}, { {"flags must not include both VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT and VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 119}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-07984", 55}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineIndex is -1, basePipelineHandle must be a valid execution graph VkPipeline handle", 159}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-07985", 55}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineHandle is VK_NULL_HANDLE, basePipelineIndex must be a valid index into the calling command's pCreateInfos parameter", 193}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-07986", 55}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, basePipelineIndex must be -1 or basePipelineHandle must be VK_NULL_HANDLE", 135}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-09007", 55}, { {"If the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature is not enabled, flags must not include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-09008", 55}, { {"If flags includes VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV, then the pNext chain must include a pointer to a valid instance of VkComputePipelineIndirectBufferInfoNV specifying the address where the pipeline's metadata will be saved", 234}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-10184", 55}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, and an output payload declared in any shader in the pipeline does not have a PayloadNodeSparseArrayAMDX decoration, there must be a node in the graph corresponding to every index from 0 to its PayloadNodeArraySizeAMDX decoration", 290}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11007", 55}, { {"If flags includes VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT, then the VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 174}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11271", 55}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also have that flag set", 167}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11272", 55}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also not have that flag set", 179}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11311", 55}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be VK_NULL_HANDLE", 130}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11312", 55}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 287}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-11798", 55}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT", 114}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-flags-12334", 55}, { {"flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 114}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-07987", 56}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, a push constant range in layout must match the shader stage", 142}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-07988", 56}, { {"If a resource variable is declared in a shader and layout is not VK_NULL_HANDLE, the corresponding descriptor set in layout must match the shader stage", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-07990", 56}, { {"If a resource variable is declared in a shader, layout is not VK_NULL_HANDLE, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in layout must match the descriptor type", 214}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-07991", 56}, { {"If a resource variable is declared in a shader as an array and layout is not VK_NULL_HANDLE, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 236}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-09130", 56}, { {"If layout is not VK_NULL_HANDLE, it must be consistent with the layout of the shaders specified in pStages", 106}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-09132", 56}, { {"If layout is not VK_NULL_HANDLE, the number of resources in layout accessible to each shader stage that is used by the pipeline must be less than or equal to VkPhysicalDeviceLimits::maxPerStageResources", 202}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-10069", 56}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, the block must be contained inside the push constant range in layout that matches the stage", 174}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-layout-parameter", 60}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pLibraryInfo-09131", 62}, { {"If pLibraryInfo is not NULL and layout is not VK_NULL_HANDLE, each element of pLibraryInfo->pLibraries must have been created with a layout that is compatible with the layout in this pipeline", 191}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pLibraryInfo-09133", 62}, { {"If pLibraryInfo is not NULL, each element of pLibraryInfo->pLibraries must be either a compute pipeline, an execution graph pipeline, or a graphics pipeline", 156}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pLibraryInfo-10181", 62}, { {"If pLibraryInfo is not NULL, each element of pLibraryInfo->pLibraries that is a compute pipeline or a graphics pipeline must have been created with VK_PIPELINE_CREATE_2_EXECUTION_GRAPH_BIT_AMDX set", 197}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pLibraryInfo-10183", 62}, { {"Any element of pLibraryInfo->pLibraries identifying a graphics pipeline must have been created with all possible state subsets", 126}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pLibraryInfo-parameter", 66}, { {"If pLibraryInfo is not NULL, pLibraryInfo must be a valid pointer to a valid VkPipelineLibraryCreateInfoKHR structure", 117}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pNext-pNext", 55}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineCompilerControlCreateInfoAMD or VkPipelineCreationFeedbackCreateInfo", 209}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pStages-09129", 57}, { {"The shader code for the entry point identified by each element of pStages and the rest of the state identified by this structure must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 213}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pStages-parameter", 61}, { {"If stageCount is not 0, and pStages is not NULL, pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 155}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pipelineCreationCacheControl-02878", 78}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT nor VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 190}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-pipelineProtectedAccess-07368", 73}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 169}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_CREATE_INFO_AMDX", 73}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-sType-unique", 56}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-shaderMeshEnqueue-10182", 67}, { {"If the shaderMeshEnqueue feature is not enabled, and pLibraryInfo->pLibraries is not NULL, pLibraryInfo->pLibraries must not contain any graphics pipelines", 155}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineCreateInfoAMDX-stage-09128", 55}, { {"The stage member of any element of pStages must be VK_SHADER_STAGE_COMPUTE_BIT", 78}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExecutionGraphPipelineScratchSizeAMDX-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_SCRATCH_SIZE_AMDX", 74}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkExportFenceCreateInfo-handleTypes-01446", 46}, { {"The bits in handleTypes must be supported and compatible, as reported by VkExternalFenceProperties", 98}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportFenceCreateInfo-handleTypes-parameter", 50}, { {"handleTypes must be a valid combination of VkExternalFenceHandleTypeFlagBits values", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportFenceCreateInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_FENCE_CREATE_INFO", 56}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportFenceWin32HandleInfoKHR-handleTypes-01447", 54}, { {"If VkExportFenceCreateInfo::handleTypes does not include VK_EXTERNAL_FENCE_HANDLE_TYPE_OPAQUE_WIN32_BIT, a VkExportFenceWin32HandleInfoKHR structure must not be included in the pNext chain of VkFenceCreateInfo", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportFenceWin32HandleInfoKHR-pAttributes-parameter", 58}, { {"If pAttributes is not NULL, pAttributes must be a valid pointer to a valid SECURITY_ATTRIBUTES value", 100}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportFenceWin32HandleInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_FENCE_WIN32_HANDLE_INFO_KHR", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportMemoryAllocateInfo-handleTypes-09860", 49}, { {"The bits in handleTypes must be supported and compatible, as reported by VkExternalTensorPropertiesARM, VkExternalImageFormatProperties, or VkExternalBufferProperties", 166}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryAllocateInfo-handleTypes-parameter", 53}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBits values", 84}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryAllocateInfo-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO", 59}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryAllocateInfoNV-handleTypes-parameter", 55}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBitsNV values", 86}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryAllocateInfoNV-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO_NV", 62}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryWin32HandleInfoKHR-handleTypes-00657", 55}, { {"If VkExportMemoryAllocateInfo::handleTypes does not include VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT, a VkExportMemoryWin32HandleInfoKHR structure must not be included in the pNext chain of VkMemoryAllocateInfo", 217}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryWin32HandleInfoKHR-pAttributes-parameter", 59}, { {"If pAttributes is not NULL, pAttributes must be a valid pointer to a valid SECURITY_ATTRIBUTES value", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryWin32HandleInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_KHR", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryWin32HandleInfoNV-pAttributes-parameter", 58}, { {"If pAttributes is not NULL, pAttributes must be a valid pointer to a valid SECURITY_ATTRIBUTES value", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMemoryWin32HandleInfoNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_NV", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalBufferInfoEXT-memory-parameter", 48}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalBufferInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_BUFFER_INFO_EXT", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalCommandQueueInfoEXT-queue-parameter", 53}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalCommandQueueInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_COMMAND_QUEUE_INFO_EXT", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalDeviceInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_DEVICE_INFO_EXT", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalIOSurfaceInfoEXT-image-parameter", 50}, { {"image must be a valid VkImage handle", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalIOSurfaceInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_IO_SURFACE_INFO_EXT", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectCreateInfoEXT-exportObjectType-parameter", 64}, { {"If exportObjectType is not 0, exportObjectType must be a valid VkExportMetalObjectTypeFlagBitsEXT value", 103}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectCreateInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECT_CREATE_INFO_EXT", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06791", 44}, { {"If the pNext chain includes a VkExportMetalDeviceInfoEXT structure, the VkInstance must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_DEVICE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkInstanceCreateInfo structure in the vkCreateInstance command", 327}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06792", 44}, { {"If the pNext chain includes a VkExportMetalCommandQueueInfoEXT structure, the VkInstance must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_COMMAND_QUEUE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkInstanceCreateInfo structure in the vkCreateInstance command", 340}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06793", 44}, { {"If the pNext chain includes a VkExportMetalBufferInfoEXT structure, the VkDeviceMemory in its memory member must have been allocated with VK_EXPORT_METAL_OBJECT_TYPE_METAL_BUFFER_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkMemoryAllocateInfo structure in the vkAllocateMemory command", 354}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06794", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, exactly one of its image, imageView, or bufferView members must not be VK_NULL_HANDLE", 154}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06795", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and its image member is not VK_NULL_HANDLE, the VkImage in its image member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkImageCreateInfo structure in the vkCreateImage command", 384}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06796", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and its imageView member is not VK_NULL_HANDLE, the VkImageView in its imageView member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkImageViewCreateInfo structure in the vkCreateImageView command", 404}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06797", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and its bufferView member is not VK_NULL_HANDLE, the VkBufferView in its bufferView member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkBufferViewCreateInfo structure in the vkCreateBufferView command", 409}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06798", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and if either its image or imageView member is not VK_NULL_HANDLE, then plane must be VK_IMAGE_ASPECT_PLANE_0_BIT, VK_IMAGE_ASPECT_PLANE_1_BIT, or VK_IMAGE_ASPECT_PLANE_2_BIT", 243}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06799", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and if the VkImage in its image member does not have a multi-planar format, then its plane member must be VK_IMAGE_ASPECT_PLANE_0_BIT", 202}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06800", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and if the VkImage in its image member has a multi-planar format with only two planes, then its plane member must not be VK_IMAGE_ASPECT_PLANE_2_BIT", 217}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06801", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and if the VkImageView in its imageView member does not have a multi-planar format, then its plane member must be VK_IMAGE_ASPECT_PLANE_0_BIT", 210}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06802", 44}, { {"If the pNext chain includes a VkExportMetalTextureInfoEXT structure, and if the VkImageView in its imageView member has a multi-planar format with only two planes, then its plane member must not be VK_IMAGE_ASPECT_PLANE_2_BIT", 225}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06803", 44}, { {"If the pNext chain includes a VkExportMetalIOSurfaceInfoEXT structure, the VkImage in its image member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_IOSURFACE_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkImageCreateInfo structure in the vkCreateImage command", 344}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06804", 44}, { {"If the pNext chain includes a VkExportMetalSharedEventInfoEXT structure, exactly one of its semaphore or event members must not be VK_NULL_HANDLE", 145}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06805", 44}, { {"If the pNext chain includes a VkExportMetalSharedEventInfoEXT structure, and its semaphore member is not VK_NULL_HANDLE, the VkSemaphore in its semaphore member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkSemaphoreCreateInfo structure in the vkCreateSemaphore command", 413}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-06806", 44}, { {"If the pNext chain includes a VkExportMetalSharedEventInfoEXT structure, and its event member is not VK_NULL_HANDLE, the VkEvent in its event member must have been created with VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT in the exportObjectType member of a VkExportMetalObjectCreateInfoEXT structure in the pNext chain of the VkEventCreateInfo structure in the vkCreateEvent command", 393}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExportMetalBufferInfoEXT, VkExportMetalCommandQueueInfoEXT, VkExportMetalDeviceInfoEXT, VkExportMetalIOSurfaceInfoEXT, VkExportMetalSharedEventInfoEXT, or VkExportMetalTextureInfoEXT", 315}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECTS_INFO_EXT", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalObjectsInfoEXT-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalBufferInfoEXT, VkExportMetalCommandQueueInfoEXT, VkExportMetalIOSurfaceInfoEXT, VkExportMetalSharedEventInfoEXT, or VkExportMetalTextureInfoEXT", 266}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalSharedEventInfoEXT-commonparent", 49}, { {"Both of event, and semaphore that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 148}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalSharedEventInfoEXT-event-parameter", 52}, { {"If event is not VK_NULL_HANDLE, event must be a valid VkEvent handle", 68}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalSharedEventInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_SHARED_EVENT_INFO_EXT", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalSharedEventInfoEXT-semaphore-parameter", 56}, { {"If semaphore is not VK_NULL_HANDLE, semaphore must be a valid VkSemaphore handle", 80}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-bufferView-parameter", 53}, { {"If bufferView is not VK_NULL_HANDLE, bufferView must be a valid VkBufferView handle", 83}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-commonparent", 45}, { {"Each of bufferView, image, and imageView that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 160}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-image-parameter", 48}, { {"If image is not VK_NULL_HANDLE, image must be a valid VkImage handle", 68}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-imageView-parameter", 52}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-plane-parameter", 48}, { {"plane must be a valid VkImageAspectFlagBits value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportMetalTextureInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_METAL_TEXTURE_INFO_EXT", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-VkExportSemaphoreCreateInfo-handleTypes-01124", 50}, { {"The bits in handleTypes must be supported and compatible, as reported by VkExternalSemaphoreProperties", 102}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportSemaphoreCreateInfo-handleTypes-parameter", 54}, { {"handleTypes must be a valid combination of VkExternalSemaphoreHandleTypeFlagBits values", 87}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportSemaphoreCreateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportSemaphoreWin32HandleInfoKHR-handleTypes-01125", 58}, { {"If VkExportSemaphoreCreateInfo::handleTypes does not include VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_BIT or VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_D3D12_FENCE_BIT, VkExportSemaphoreWin32HandleInfoKHR must not be included in the pNext chain of VkSemaphoreCreateInfo", 266}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportSemaphoreWin32HandleInfoKHR-pAttributes-parameter", 62}, { {"If pAttributes is not NULL, pAttributes must be a valid pointer to a valid SECURITY_ATTRIBUTES value", 100}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExportSemaphoreWin32HandleInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExternalBufferProperties-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalBufferProperties-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_BUFFER_PROPERTIES", 58}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalComputeQueueCreateInfoNV-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalComputeQueueCreateInfoNV-preferredQueue-parameter", 64}, { {"preferredQueue must be a valid VkQueue handle", 45}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalComputeQueueCreateInfoNV-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_CREATE_INFO_NV", 69}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalComputeQueueDataParamsNV-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalComputeQueueDataParamsNV-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_DATA_PARAMS_NV", 69}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalComputeQueueDeviceCreateInfoNV-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_DEVICE_CREATE_INFO_NV", 76}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkExternalFenceProperties-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalFenceProperties-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_FENCE_PROPERTIES", 57}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalFormatANDROID-externalFormat-01894", 49}, { {"externalFormat must be 0 or a value returned in the externalFormat member of VkAndroidHardwareBufferFormatPropertiesANDROID by an earlier call to vkGetAndroidHardwareBufferPropertiesANDROID", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalFormatANDROID-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_ANDROID", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalFormatOHOS-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_OHOS", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalFormatQNX-externalFormat-08956", 45}, { {"externalFormat must be 0 or a value returned in the externalFormat member of VkScreenBufferFormatPropertiesQNX by an earlier call to vkGetScreenBufferPropertiesQNX", 163}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalFormatQNX-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_QNX", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalImageFormatProperties-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_IMAGE_FORMAT_PROPERTIES", 64}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalMemoryAcquireUnmodifiedEXT-acquireUnmodifiedMemory-08922", 71}, { {"If acquireUnmodifiedMemory is VK_TRUE, and the memory barrier's srcQueueFamilyIndex is a special queue family reserved for external memory ownership transfers (as described in Queue Family Ownership Transfer), then each range of VkDeviceMemory bound to the resource must have remained unmodified during all time since the resource's most recent release of ownership to the queue family", 385}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExternalMemoryAcquireUnmodifiedEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_ACQUIRE_UNMODIFIED_EXT", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkExternalMemoryBufferCreateInfo-handleTypes-parameter", 59}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBits values", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryBufferCreateInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_BUFFER_CREATE_INFO", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryImageCreateInfo-handleTypes-parameter", 58}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBits values", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryImageCreateInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryImageCreateInfoNV-handleTypes-parameter", 60}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBitsNV values", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryImageCreateInfoNV-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO_NV", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryTensorCreateInfoARM-handleTypes-parameter", 62}, { {"handleTypes must be a valid combination of VkExternalMemoryHandleTypeFlagBits values", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalMemoryTensorCreateInfoARM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_TENSOR_CREATE_INFO_ARM", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkExternalSemaphoreProperties-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalSemaphoreProperties-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_SEMAPHORE_PROPERTIES", 61}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalTensorPropertiesARM-externalMemoryProperties-parameter", 69}, { {"externalMemoryProperties must be a valid VkExternalMemoryProperties structure", 77}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalTensorPropertiesARM-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkExternalTensorPropertiesARM-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_EXTERNAL_TENSOR_PROPERTIES_ARM", 62}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkFenceCreateInfo-flags-parameter", 38}, { {"flags must be a valid combination of VkFenceCreateFlagBits values", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceCreateInfo-pNext-pNext", 34}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExportFenceCreateInfo or VkExportFenceWin32HandleInfoKHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceCreateInfo-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_FENCE_CREATE_INFO", 49}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceCreateInfo-sType-unique", 35}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-fence-01455", 36}, { {"fence must not currently have its payload replaced by an imported payload as described below in Importing Fence Payloads unless that imported payload's handle type was included in VkExternalFenceProperties::exportFromImportedHandleTypes for handleType", 251}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-fence-parameter", 40}, { {"fence must be a valid VkFence handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-handleType-01453", 41}, { {"handleType must have been included in VkExportFenceCreateInfo::handleTypes when fence's current payload was created", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-handleType-01454", 41}, { {"If handleType refers to a handle type with copy payload transference semantics, fence must be signaled, or have an associated fence signal operation pending execution", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-handleType-01456", 41}, { {"handleType must be defined as a POSIX file descriptor handle", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-handleType-parameter", 45}, { {"handleType must be a valid VkExternalFenceHandleTypeFlagBits value", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetFdInfoKHR-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_FENCE_GET_FD_INFO_KHR", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-fence-01450", 45}, { {"fence must not currently have its payload replaced by an imported payload as described below in Importing Fence Payloads unless that imported payload's handle type was included in VkExternalFenceProperties::exportFromImportedHandleTypes for handleType", 251}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-fence-parameter", 49}, { {"fence must be a valid VkFence handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-handleType-01448", 50}, { {"handleType must have been included in VkExportFenceCreateInfo::handleTypes when the fence's current payload was created", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-handleType-01449", 50}, { {"If handleType is defined as an NT handle, vkGetFenceWin32HandleKHR must be called no more than once for each valid unique combination of fence and handleType", 157}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-handleType-01451", 50}, { {"If handleType refers to a handle type with copy payload transference semantics, fence must be signaled, or have an associated fence signal operation pending execution", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-handleType-01452", 50}, { {"handleType must be defined as an NT handle or a global share handle", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-handleType-parameter", 54}, { {"handleType must be a valid VkExternalFenceHandleTypeFlagBits value", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFenceGetWin32HandleInfoKHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_FENCE_GET_WIN32_HANDLE_INFO_KHR", 63}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkFilterCubicImageViewImageFormatPropertiesEXT-pNext-02627", 63}, { {"If the pNext chain of the VkImageFormatProperties2 structure includes a VkFilterCubicImageViewImageFormatPropertiesEXT structure, the pNext chain of the VkPhysicalDeviceImageFormatInfo2 structure must include a VkPhysicalDeviceImageViewImageFormatInfoEXT structure with an imageViewType that is compatible with imageType", 320}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkFilterCubicImageViewImageFormatPropertiesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_FILTER_CUBIC_IMAGE_VIEW_IMAGE_FORMAT_PROPERTIES_EXT", 83}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkFormatProperties2-pNext-pNext", 36}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDrmFormatModifierPropertiesList2EXT, VkDrmFormatModifierPropertiesListEXT, VkFormatProperties3, VkFormatProperties4KHR, VkSubpassResolvePerformanceQueryEXT, VkTensorExplicitTilingFormatPropertiesARM, or VkTensorFormatPropertiesARM", 363}, { "chapters/formats.html", 21} } },
        { {"VUID-VkFormatProperties2-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2", 51}, { "chapters/formats.html", 21} } },
        { {"VUID-VkFormatProperties2-sType-unique", 37}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/formats.html", 21} } },
        { {"VUID-VkFormatProperties3-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_3", 51}, { "chapters/formats.html", 21} } },
        { {"VUID-VkFormatProperties4KHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_4_KHR", 55}, { "chapters/formats.html", 21} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04524", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, its layout member must be equal to VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR", 223}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04525", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.width must be a power of two value", 166}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04526", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.width must be less than or equal to maxFragmentShadingRateAttachmentTexelSize.width", 215}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04527", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.width must be greater than or equal to minFragmentShadingRateAttachmentTexelSize.width", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04528", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.height must be a power of two value", 167}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04529", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.height must be less than or equal to maxFragmentShadingRateAttachmentTexelSize.height", 217}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04530", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, shadingRateAttachmentTexelSize.height must be greater than or equal to minFragmentShadingRateAttachmentTexelSize.height", 220}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04531", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, the quotient of shadingRateAttachmentTexelSize.width and shadingRateAttachmentTexelSize.height must be less than or equal to maxFragmentShadingRateAttachmentTexelSizeAspectRatio", 278}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-04532", 80}, { {"If pFragmentShadingRateAttachment is not NULL and its attachment member is not VK_ATTACHMENT_UNUSED, the quotient of shadingRateAttachmentTexelSize.height and shadingRateAttachmentTexelSize.width must be less than or equal to maxFragmentShadingRateAttachmentTexelSizeAspectRatio", 278}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-pFragmentShadingRateAttachment-parameter", 84}, { {"If pFragmentShadingRateAttachment is not NULL, pFragmentShadingRateAttachment must be a valid pointer to a valid VkAttachmentReference2 structure", 145}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFragmentShadingRateAttachmentInfoKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR", 73}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFrameBoundaryEXT-commonparent", 36}, { {"Both of the elements of pBuffers, and the elements of pImages that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 181}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryEXT-flags-parameter", 39}, { {"flags must be a valid combination of VkFrameBoundaryFlagBitsEXT values", 70}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryEXT-pBuffers-parameter", 42}, { {"If bufferCount is not 0, and pBuffers is not NULL, pBuffers must be a valid pointer to an array of bufferCount valid VkBuffer handles", 133}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryEXT-pImages-parameter", 41}, { {"If imageCount is not 0, and pImages is not NULL, pImages must be a valid pointer to an array of imageCount valid VkImage handles", 128}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryEXT-pTag-parameter", 38}, { {"If tagSize is not 0, and pTag is not NULL, pTag must be a valid pointer to an array of tagSize bytes", 100}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryEXT-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_FRAME_BOUNDARY_EXT", 50}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryTensorsARM-pTensors-parameter", 49}, { {"pTensors must be a valid pointer to an array of tensorCount valid VkTensorARM handles", 85}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryTensorsARM-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_FRAME_BOUNDARY_TENSORS_ARM", 58}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFrameBoundaryTensorsARM-tensorCount-arraylength", 54}, { {"tensorCount must be greater than 0", 34}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-flags-parameter", 53}, { {"flags must be a valid combination of VkImageCreateFlagBits values", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-pNext-pNext", 49}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkImageCreateFlags2CreateInfoKHR or VkImageUsageFlags2CreateInfoKHR", 198}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-pViewFormats-parameter", 60}, { {"If viewFormatCount is not 0, pViewFormats must be a valid pointer to an array of viewFormatCount valid VkFormat values", 118}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENT_IMAGE_INFO", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-usage-parameter", 53}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-usage-requiredbitmask", 59}, { {"usage must not be 0", 19}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentImageInfo-viewFormatCount-09536", 59}, { {"If viewFormatCount is not 0, and the render pass is not being used with an external format resolve attachment, each element of pViewFormats must not be VK_FORMAT_UNDEFINED", 171}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentsCreateInfo-pAttachmentImageInfos-parameter", 71}, { {"If attachmentImageInfoCount is not 0, pAttachmentImageInfos must be a valid pointer to an array of attachmentImageInfoCount valid VkFramebufferAttachmentImageInfo structures", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferAttachmentsCreateInfo-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENTS_CREATE_INFO", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-attachmentCount-00876", 50}, { {"attachmentCount must be equal to the attachment count specified in renderPass", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-commonparent", 41}, { {"Both of renderPass, and the elements of pAttachments that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 172}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-02778", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT and attachmentCount is not 0, pAttachments must be a valid pointer to an array of attachmentCount valid VkImageView handles", 185}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03189", 40}, { {"If the imagelessFramebuffer feature is not enabled, flags must not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT", 110}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03190", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the pNext chain must include a VkFramebufferAttachmentsCreateInfo structure", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03191", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the attachmentImageInfoCount member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain must be equal to either zero or attachmentCount", 207}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03196", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the width member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is referenced by VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment in renderPass must be greater than or equal to the ceiling of width/maxFragmentDensityTexelSize.width", 389}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03197", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the height member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that is referenced by VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment in renderPass must be greater than or equal to the ceiling of height/maxFragmentDensityTexelSize.height", 401}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03201", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the usage member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that refers to an attachment used as a color attachment or resolve attachment by renderPass must include VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT", 341}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03202", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the usage member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that refers to an attachment used as a depth/stencil attachment by renderPass must include VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT", 335}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03203", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the usage member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that refers to an attachment used as a depth/stencil resolve attachment by renderPass must include VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT", 343}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03204", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the usage member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that refers to an attachment used as an input attachment by renderPass must include VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT", 320}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-03205", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, at least one element of the pViewFormats member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain must be equal to the corresponding value of VkAttachmentDescription::format used to create renderPass", 333}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04113", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments must have been created with VkImageViewCreateInfo::viewType not equal to VK_IMAGE_VIEW_TYPE_3D", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04533", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as an input, color, resolve, or depth/stencil attachment by renderPass must have been created with a VkImageCreateInfo::extent.width greater than or equal to width", 268}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04534", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as an input, color, resolve, or depth/stencil attachment by renderPass must have been created with a VkImageCreateInfo::extent.height greater than or equal to height", 270}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04535", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as an input, color, resolve, or depth/stencil attachment by renderPass must have been created with a VkImageViewCreateInfo::subresourceRange.layerCount greater than or equal to layers", 288}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04537", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, and renderPass was specified with non-zero view masks, each element of pAttachments that is used as a fragment shading rate attachment by renderPass must have a layerCount that is either 1, or greater than the index of the most significant bit set in any of those view masks", 337}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04538", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, and renderPass was not specified with non-zero view masks, each element of pAttachments that is used as a fragment shading rate attachment by renderPass must have a layerCount that is either 1, or greater than layers", 279}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04539", 40}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, an element of pAttachments that is used as a fragment shading rate attachment must have a width at least as large as ceil(width / texelWidth), where texelWidth is the largest value of shadingRateAttachmentTexelSize.width in a VkFragmentShadingRateAttachmentInfoKHR which references that attachment", 625}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04540", 40}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, an element of pAttachments that is used as a fragment shading rate attachment must have a height at least as large as ceil(height / texelHeight), where texelHeight is the largest value of shadingRateAttachmentTexelSize.height in a VkFragmentShadingRateAttachmentInfoKHR which references that attachment", 630}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04541", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the width member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as an input, color, resolve, or depth/stencil attachment in renderPass must be greater than or equal to width", 314}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04542", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the height member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as an input, color, resolve, or depth/stencil attachment in renderPass must be greater than or equal to height", 316}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04543", 40}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, and flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the width member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as a fragment shading rate attachment must be greater than or equal to ceil(width / texelWidth), where texelWidth is the largest value of shadingRateAttachmentTexelSize.width in a VkFragmentShadingRateAttachmentInfoKHR which references that attachment", 725}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04544", 40}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, and flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the height member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as a fragment shading rate attachment must be greater than or equal to ceil(height / texelHeight), where texelHeight is the largest value of shadingRateAttachmentTexelSize.height in a VkFragmentShadingRateAttachmentInfoKHR which references that attachment", 730}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04545", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the layerCount member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as a fragment shading rate attachment must be either 1, or greater than or equal to layers", 300}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04548", 40}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as a fragment shading rate attachment by renderPass must have been created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set", 259}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04549", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the usage member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain that refers to an attachment used as a fragment shading rate attachment by renderPass must include VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 355}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-04587", 40}, { {"If flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT and renderPass was specified with non-zero view masks, the layerCount member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure in the pNext chain that is used as a fragment shading rate attachment must be either 1, or greater than the index of the most significant bit set in any of those view masks", 404}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-flags-parameter", 44}, { {"flags must be a valid combination of VkFramebufferCreateFlagBits values", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-height-00887", 41}, { {"height must be greater than 0", 29}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-height-00888", 41}, { {"height must be less than or equal to maxFramebufferHeight", 57}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-layers-00889", 41}, { {"layers must be greater than 0", 29}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-layers-00890", 41}, { {"layers must be less than or equal to maxFramebufferLayers", 57}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-nullColorAttachmentWithExternalFormatResolve-09349", 79}, { {"If the nullColorAttachmentWithExternalFormatResolve is VK_FALSE, and flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the format of the color attachment for each subpass in renderPass that includes an external format image as a resolve attachment must have a format equal to the value of VkAndroidHardwareBufferFormatResolvePropertiesANDROID::colorAttachmentFormat as returned by a call to vkGetAndroidHardwareBufferPropertiesANDROID for the Android hardware buffer that was used to create the image view use as its resolve attachment", 545}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00877", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as a color attachment or resolve attachment by renderPass must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 245}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00879", 47}, { {"If renderpass is not VK_NULL_HANDLE, flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as an input attachment by renderPass must have been created with the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 258}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00880", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments must have been created with a VkFormat value that matches the VkFormat specified by the corresponding VkAttachmentDescription in renderPass", 231}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00881", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments must have been created with a samples value that matches the samples value specified by the corresponding VkAttachmentDescription in renderPass", 235}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00883", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments must only specify a single mip level", 128}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00884", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments must have been created with the identity swizzle", 140}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-00891", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is a 2D or 2D array image view taken from a 3D image must not be a depth/stencil format", 184}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-02552", 47}, { {"Each element of pAttachments that is used as a fragment density map attachment by renderPass must not have been created with a flags value including VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-02555", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, an element of pAttachments that is referenced by fragmentDensityMapAttachment must have a width at least as large as the ceiling of width/maxFragmentDensityTexelSize.width", 234}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-02556", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, an element of pAttachments that is referenced by fragmentDensityMapAttachment must have a height at least as large as the ceiling of height/maxFragmentDensityTexelSize.height", 237}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-02633", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as a depth/stencil attachment by renderPass must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 239}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-02634", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of pAttachments that is used as a depth/stencil resolve attachment by renderPass must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 247}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-09350", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, then if an element of pAttachments has a format of VK_FORMAT_UNDEFINED, it must have been created with a VkExternalFormatANDROID::externalFormat value identical to that provided in the VkExternalFormatANDROID::externalFormat specified by the corresponding VkAttachmentDescription2 in renderPass", 357}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pAttachments-12327", 47}, { {"If flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, then each element of pAttachments that is used as a resolve attachment by renderPass must not be bound to a VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 276}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-pNext-pNext", 40}, { {"pNext must be NULL or a pointer to a valid instance of VkFramebufferAttachmentsCreateInfo", 89}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-02531", 45}, { {"If renderPass was specified with non-zero view masks, layers must be 1", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-02553", 45}, { {"If renderPass has a fragment density map attachment and the fragmentDensityMapNonSubsampledImages feature is not enabled, each element of pAttachments must have been created with a flags value including VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT unless that element is the fragment density map attachment", 296}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-02746", 45}, { {"Each element of pAttachments that is referenced by fragmentDensityMapAttachment must have a layerCount equal to 1 or if renderPass was specified with non-zero view masks, greater than the index of the most significant bit set in any of those view masks", 252}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-03198", 45}, { {"If multiview is enabled for renderPass and flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the layerCount member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain used as an input, color, resolve, or depth/stencil attachment in renderPass must be greater than the maximum bit index set in the view mask in the subpasses in which it is used in renderPass", 436}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-04536", 45}, { {"If renderPass was specified with non-zero view masks, each element of pAttachments that is used as an input, color, resolve, or depth/stencil attachment by renderPass must have a layerCount greater than the index of the most significant bit set in any of those view masks", 271}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-04546", 45}, { {"If multiview is not enabled for renderPass and flags includes VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the layerCount member of any element of the pAttachmentImageInfos member of a VkFramebufferAttachmentsCreateInfo structure included in the pNext chain used as an input, color, resolve, or depth/stencil attachment in renderPass must be greater than or equal to layers", 365}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-10830", 45}, { {"If renderPass was created with VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then layers must be less than or equal to maxFragmentDensityMapLayers", 160}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-renderPass-parameter", 49}, { {"renderPass must be a valid VkRenderPass handle", 46}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO", 55}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-samples-06881", 42}, { {"If multisampled-render-to-single-sampled is enabled for any subpass, all color, depth/stencil, and input attachments used in that subpass which have VkAttachmentDescription::samples or VkAttachmentDescription2::samples equal to VK_SAMPLE_COUNT_1_BIT must have been created with VK_IMAGE_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT in their VkImageCreateInfo::flags", 373}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-samples-07009", 42}, { {"If multisampled-render-to-single-sampled is enabled for any subpass, all color, depth/stencil, and input attachments used in that subpass which have VkAttachmentDescription::samples or VkAttachmentDescription2::samples equal to VK_SAMPLE_COUNT_1_BIT must have a format that supports the sample count specified in VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 377}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-width-00885", 40}, { {"width must be greater than 0", 28}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferCreateInfo-width-00886", 40}, { {"width must be less than or equal to maxFramebufferWidth", 55}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkFramebufferMixedSamplesCombinationNV-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkFramebufferMixedSamplesCombinationNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_FRAMEBUFFER_MIXED_SAMPLES_COMBINATION_NV", 72}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-commonparent", 44}, { {"Both of indirectCommandsLayout, and indirectExecutionSet that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 176}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectAddress-11074", 53}, { {"indirectAddress must be aligned to 4", 36}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectAddress-12407", 53}, { {"The underlying buffer for indirectAddress must have the VK_BUFFER_USAGE_2_INDIRECT_BUFFER_BIT_KHR bit set in its usage flag", 123}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectAddress-parameter", 57}, { {"indirectAddress must be a valid VkDeviceAddress value", 53}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectAddressSize-11077", 57}, { {"indirectAddressSize must be greater than zero", 45}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-10241", 60}, { {"If the token sequence of the passed indirectCommandsLayout does not contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the indirectExecutionSet must be VK_NULL_HANDLE", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11002", 60}, { {"If indirectCommandsLayout was created with a token sequence that contained the VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the shader stages used to create the initial shader state of indirectExecutionSet must equal the VkIndirectCommandsExecutionSetTokenEXT::shaderStages used to create indirectCommandsLayout", 323}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11065", 60}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT, then the indirectExecutionSet's push constant layout must contain the updateRange specified in VkIndirectCommandsPushConstantTokenEXT", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11066", 60}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT, then the indirectExecutionSet's push constant layout must contain the updateRange specified in VkIndirectCommandsPushConstantTokenEXT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11079", 60}, { {"If indirectCommandsLayout was created using a VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_EXT token and shader objects are not bound then the bound graphics pipeline must have been created with VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE in pDynamicStates", 258}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11083", 60}, { {"If the token sequence of the passed indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the indirectExecutionSet must not be VK_NULL_HANDLE", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11144", 60}, { {"If the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, and there is a descriptor and push constant layout info provided either by pipelineLayout or through a VkPipelineLayoutCreateInfo in pNext of the VkIndirectCommandsLayoutCreateInfoEXT used to create indirectCommandsLayout, the pipeline layout must be compatible with the descriptor and push constant layout info used by indirectExecutionSet", 438}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11328", 60}, { {"If the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, and there was no descriptor and push constant layout info provided either by pipelineLayout or through a VkPipelineLayoutCreateInfo in pNext of the VkIndirectCommandsLayoutCreateInfoEXT used to create indirectCommandsLayout, pipelines in indirectExecutionSet must have been created with VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT", 429}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11329", 60}, { {"If the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, and there was a descriptor and push constant layout info provided either by pipelineLayout or through a VkPipelineLayoutCreateInfo in pNext of the VkIndirectCommandsLayoutCreateInfoEXT used to create indirectCommandsLayout, pipelines in indirectExecutionSet must have been created without VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT", 431}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11330", 60}, { {"If the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, and there was no descriptor and push constant layout info provided either by pipelineLayout or through a VkPipelineLayoutCreateInfo in pNext of the VkIndirectCommandsLayoutCreateInfoEXT used to create indirectCommandsLayout, shaders in indirectExecutionSet must have been created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 423}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-11331", 60}, { {"If the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, and there was a descriptor and push constant layout info provided either by pipelineLayout or through a VkPipelineLayoutCreateInfo in pNext of the VkIndirectCommandsLayoutCreateInfoEXT used to create indirectCommandsLayout, shaders in indirectExecutionSet must have been created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 425}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectCommandsLayout-parameter", 64}, { {"indirectCommandsLayout must be a valid VkIndirectCommandsLayoutEXT handle", 73}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectExecutionSet-11080", 58}, { {"If indirectExecutionSet is VK_NULL_HANDLE, a VkGeneratedCommandsPipelineInfoEXT or VkGeneratedCommandsShaderInfoEXT must be included in the pNext chain", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-indirectExecutionSet-parameter", 62}, { {"If indirectExecutionSet is not VK_NULL_HANDLE, indirectExecutionSet must be a valid VkIndirectExecutionSetEXT handle", 116}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-maxDrawCount-11078", 50}, { {"When not ignored, maxDrawCount x maxSequenceCount must be less than 2^24", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-maxSequenceCount-10246", 54}, { {"maxSequenceCount must not be zero", 33}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-maxSequenceCount-11067", 54}, { {"maxSequenceCount must be less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectSequenceCount and VkGeneratedCommandsMemoryRequirementsInfoEXT::maxSequenceCount that was used to determine the preprocessSize", 233}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-preprocessAddress-11063", 55}, { {"If vkGetGeneratedCommandsMemoryRequirementsEXT returns a non-zero size, preprocessAddress must not be 0", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-preprocessAddress-11064", 55}, { {"VkDeviceMemory objects bound to the underlying buffer for preprocessAddress must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned by vkGetGeneratedCommandsMemoryRequirementsEXT", 261}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-preprocessAddress-11069", 55}, { {"preprocessAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_2_PREPROCESS_BUFFER_BIT_EXT usage flag set", 161}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-preprocessAddress-parameter", 59}, { {"If preprocessAddress is not 0, preprocessAddress must be a valid VkDeviceAddress value", 86}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-preprocessSize-11071", 52}, { {"preprocessSize must be greater than or equal to the memory requirement's size returned by vkGetGeneratedCommandsMemoryRequirementsEXT using the matching inputs (indirectCommandsLayout, ...) as within this structure", 214}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_EXT", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-sequenceCountAddress-11068", 58}, { {"If sequenceCountAddress is not NULL, the value contained in the address must be less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectSequenceCount and VkGeneratedCommandsMemoryRequirementsInfoEXT::maxSequenceCount that was used to determine the preprocessSize", 288}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-sequenceCountAddress-11072", 58}, { {"The underlying buffer for sequenceCountAddress must have the VK_BUFFER_USAGE_2_INDIRECT_BUFFER_BIT_KHR bit set in its usage flag", 128}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-sequenceCountAddress-11073", 58}, { {"If sequenceCountAddress is not NULL, sequenceCountAddress must be aligned to 4", 78}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-sequenceCountAddress-parameter", 62}, { {"If sequenceCountAddress is not 0, sequenceCountAddress must be a valid VkDeviceAddress value", 92}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-shaderStages-parameter", 54}, { {"shaderStages must be a valid combination of VkShaderStageFlagBits values", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoEXT-shaderStages-requiredbitmask", 60}, { {"shaderStages must not be 0", 26}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-commonparent", 43}, { {"Each of indirectCommandsLayout, pipeline, preprocessBuffer, sequencesCountBuffer, and sequencesIndexBuffer that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 226}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-02913", 59}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_SHADER_GROUP_NV, then the pipeline must have been created with multiple shader groups", 163}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-02914", 59}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_SHADER_GROUP_NV, then the pipeline must have been created with VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV set in VkGraphicsPipelineCreateInfo::flags", 227}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-02915", 59}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, then the pipeline's VkPipelineLayout must match the VkIndirectCommandsLayoutTokenNV::pushconstantPipelineLayout", 207}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-07078", 59}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_TASKS_NV, then the pipeline must contain a shader stage using the MeshNV Execution Model", 171}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-07079", 59}, { {"If the indirectCommandsLayout uses a token of VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV, then the pipeline must contain a shader stage using the MeshEXT Execution Model", 177}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-indirectCommandsLayout-parameter", 63}, { {"indirectCommandsLayout must be a valid VkIndirectCommandsLayoutNV handle", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pStreams-parameter", 49}, { {"pStreams must be a valid pointer to an array of streamCount valid VkIndirectCommandsStreamNV structures", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipeline-02912", 45}, { {"The provided pipeline must match the pipeline bound at execution time", 69}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipeline-parameter", 49}, { {"If pipeline is not VK_NULL_HANDLE, pipeline must be a valid VkPipeline handle", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipelineBindPoint-09084", 54}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, then the pipeline must have been created with the flag VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 162}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipelineBindPoint-09085", 54}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, then the pipeline must have been created with a VkComputePipelineIndirectBufferInfoNV structure specifying a valid address where its metadata will be saved", 219}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipelineBindPoint-09086", 54}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, then vkCmdUpdatePipelineIndirectBufferNV must have been called on that pipeline to save its metadata to a device address", 184}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipelineBindPoint-09087", 54}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, and if VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV is used, then pipeline must be VK_NULL_HANDLE", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-pipelineBindPoint-parameter", 58}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-preprocessBuffer-02918", 53}, { {"preprocessBuffer must have the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set in its usage flag", 92}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-preprocessBuffer-02971", 53}, { {"If preprocessBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-preprocessBuffer-parameter", 57}, { {"preprocessBuffer must be a valid VkBuffer handle", 48}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-preprocessOffset-02919", 53}, { {"preprocessOffset must be aligned to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::minIndirectCommandsBufferOffsetAlignment", 129}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-preprocessSize-02920", 51}, { {"preprocessSize must be at least equal to the memory requirement's size returned by vkGetGeneratedCommandsMemoryRequirementsNV using the matching inputs (indirectCommandsLayout, ...) as within this structure", 206}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_NV", 58}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCount-02917", 51}, { {"sequencesCount must be less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectSequenceCount and VkGeneratedCommandsMemoryRequirementsInfoNV::maxSequencesCount that was used to determine the preprocessSize", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCountBuffer-02921", 57}, { {"sequencesCountBuffer can be set if the actual used count of sequences is sourced from the provided buffer. In that case the sequencesCount serves as upper bound", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCountBuffer-02922", 57}, { {"If sequencesCountBuffer is not VK_NULL_HANDLE, its usage flag must have the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set", 119}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCountBuffer-02923", 57}, { {"If sequencesCountBuffer is not VK_NULL_HANDLE, sequencesCountOffset must be aligned to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::minSequencesCountBufferOffsetAlignment", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCountBuffer-02972", 57}, { {"If sequencesCountBuffer is not VK_NULL_HANDLE and is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesCountBuffer-parameter", 61}, { {"If sequencesCountBuffer is not VK_NULL_HANDLE, sequencesCountBuffer must be a valid VkBuffer handle", 99}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesIndexBuffer-02924", 57}, { {"If indirectCommandsLayout's VK_INDIRECT_COMMANDS_LAYOUT_USAGE_INDEXED_SEQUENCES_BIT_NV is set, sequencesIndexBuffer must be set otherwise it must be VK_NULL_HANDLE", 163}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesIndexBuffer-02925", 57}, { {"If sequencesIndexBuffer is not VK_NULL_HANDLE, its usage flag must have the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set", 119}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesIndexBuffer-02926", 57}, { {"If sequencesIndexBuffer is not VK_NULL_HANDLE, sequencesIndexOffset must be aligned to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::minSequencesIndexBufferOffsetAlignment", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesIndexBuffer-02973", 57}, { {"If sequencesIndexBuffer is not VK_NULL_HANDLE and is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-sequencesIndexBuffer-parameter", 61}, { {"If sequencesIndexBuffer is not VK_NULL_HANDLE, sequencesIndexBuffer must be a valid VkBuffer handle", 99}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-streamCount-02916", 48}, { {"streamCount must match the indirectCommandsLayout's streamCount", 63}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsInfoNV-streamCount-arraylength", 54}, { {"streamCount must be greater than 0", 34}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-commonparent", 62}, { {"Both of indirectCommandsLayout, and indirectExecutionSet that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 176}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectCommandsLayout-11010", 78}, { {"If indirectCommandsLayout was created with a token sequence that contained the VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, indirectExecutionSet must not be VK_NULL_HANDLE", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectCommandsLayout-11011", 78}, { {"If indirectCommandsLayout was not created with a token sequence that contained the VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, indirectExecutionSet must be VK_NULL_HANDLE", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectCommandsLayout-11151", 78}, { {"If indirectCommandsLayout was created with a token sequence that contained the VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the shader stages used to create the initial shader state of indirectExecutionSet must equal the VkIndirectCommandsExecutionSetTokenEXT::shaderStages used to create indirectCommandsLayout", 323}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectCommandsLayout-parameter", 82}, { {"indirectCommandsLayout must be a valid VkIndirectCommandsLayoutEXT handle", 73}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectExecutionSet-11012", 76}, { {"If indirectExecutionSet is VK_NULL_HANDLE, either a VkGeneratedCommandsPipelineInfoEXT or a VkGeneratedCommandsShaderInfoEXT must be included in the pNext chain", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-indirectExecutionSet-parameter", 80}, { {"If indirectExecutionSet is not VK_NULL_HANDLE, indirectExecutionSet must be a valid VkIndirectExecutionSetEXT handle", 116}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-maxDrawCount-11146", 68}, { {"When not ignored, maxDrawCount x maxSequenceCount must be less than 2^24", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-maxSequencesCount-11009", 73}, { {"maxSequencesCount must be less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectSequenceCount", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-pNext-pNext", 61}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkGeneratedCommandsPipelineInfoEXT or VkGeneratedCommandsShaderInfoEXT", 201}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_EXT", 79}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoEXT-sType-unique", 62}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-commonparent", 61}, { {"Both of indirectCommandsLayout, and pipeline that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 164}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-indirectCommandsLayout-parameter", 81}, { {"indirectCommandsLayout must be a valid VkIndirectCommandsLayoutNV handle", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-maxSequencesCount-02907", 72}, { {"maxSequencesCount must be less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectSequenceCount", 120}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pipeline-parameter", 67}, { {"If pipeline is not VK_NULL_HANDLE, pipeline must be a valid VkPipeline handle", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pipelineBindPoint-09075", 72}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_GRAPHICS, then pipeline must be a valid VkPipeline handle", 112}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pipelineBindPoint-09076", 72}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, and the indirectCommandsLayout was not created with a VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV token, then the pipeline must be a valid VkPipeline handle", 220}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pipelineBindPoint-09077", 72}, { {"If pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, and the indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV token, then the pipeline must be VK_NULL_HANDLE", 197}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-pipelineBindPoint-parameter", 76}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsMemoryRequirementsInfoNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_NV", 78}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsPipelineInfoEXT-pipeline-parameter", 58}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsPipelineInfoEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_PIPELINE_INFO_EXT", 68}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsShaderInfoEXT-pShaders-11127", 52}, { {"pShaders must not contain more than one shader object for a given VkShaderStageFlagBits stage", 93}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsShaderInfoEXT-pShaders-parameter", 56}, { {"pShaders must be a valid pointer to an array of shaderCount valid VkShaderEXT handles", 85}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsShaderInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_GENERATED_COMMANDS_SHADER_INFO_EXT", 66}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeneratedCommandsShaderInfoEXT-shaderCount-arraylength", 61}, { {"shaderCount must be greater than 0", 34}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkGeometryAABBNV-aabbData-parameter", 40}, { {"If aabbData is not VK_NULL_HANDLE, aabbData must be a valid VkBuffer handle", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryAABBNV-offset-02439", 34}, { {"offset must be less than the size of aabbData", 45}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryAABBNV-offset-02440", 34}, { {"offset must be a multiple of 8", 30}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryAABBNV-pNext-pNext", 33}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryAABBNV-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_GEOMETRY_AABB_NV", 48}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryAABBNV-stride-02441", 34}, { {"stride must be a multiple of 8", 30}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryDataNV-aabbs-parameter", 37}, { {"aabbs must be a valid VkGeometryAABBNV structure", 48}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryDataNV-triangles-parameter", 41}, { {"triangles must be a valid VkGeometryTrianglesNV structure", 57}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-flags-parameter", 33}, { {"flags must be a valid combination of VkGeometryFlagBitsKHR values", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-geometry-parameter", 36}, { {"geometry must be a valid VkGeometryDataNV structure", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-geometryType-03503", 36}, { {"geometryType must be VK_GEOMETRY_TYPE_TRIANGLES_NV or VK_GEOMETRY_TYPE_AABBS_NV", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-geometryType-parameter", 40}, { {"geometryType must be a valid VkGeometryTypeKHR value", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-pNext-pNext", 29}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryNV-sType-sType", 29}, { {"sType must be VK_STRUCTURE_TYPE_GEOMETRY_NV", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-commonparent", 39}, { {"Each of indexData, transformData, and vertexData that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 168}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexCount-02436", 43}, { {"indexCount must be 0 if indexType is VK_INDEX_TYPE_NONE_NV", 58}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexData-02434", 42}, { {"indexData must be VK_NULL_HANDLE if indexType is VK_INDEX_TYPE_NONE_NV", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexData-02435", 42}, { {"indexData must be a valid VkBuffer handle if indexType is not VK_INDEX_TYPE_NONE_NV", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexData-parameter", 46}, { {"If indexData is not VK_NULL_HANDLE, indexData must be a valid VkBuffer handle", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexOffset-02431", 44}, { {"indexOffset must be less than the size of indexData", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexOffset-02432", 44}, { {"indexOffset must be a multiple of the element size of indexType", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexType-02433", 42}, { {"indexType must be VK_INDEX_TYPE_UINT16, VK_INDEX_TYPE_UINT32, or VK_INDEX_TYPE_NONE_NV", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-indexType-parameter", 46}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_GEOMETRY_TRIANGLES_NV", 53}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-transformData-parameter", 50}, { {"If transformData is not VK_NULL_HANDLE, transformData must be a valid VkBuffer handle", 85}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-transformOffset-02437", 48}, { {"transformOffset must be less than the size of transformData", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-transformOffset-02438", 48}, { {"transformOffset must be a multiple of 16", 40}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexData-parameter", 47}, { {"If vertexData is not VK_NULL_HANDLE, vertexData must be a valid VkBuffer handle", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexFormat-02430", 45}, { {"vertexFormat must be one of VK_FORMAT_R32G32B32_SFLOAT, VK_FORMAT_R32G32_SFLOAT, VK_FORMAT_R16G16B16_SFLOAT, VK_FORMAT_R16G16_SFLOAT, VK_FORMAT_R16G16_SNORM, or VK_FORMAT_R16G16B16_SNORM", 186}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexFormat-parameter", 49}, { {"vertexFormat must be a valid VkFormat value", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexOffset-02428", 45}, { {"vertexOffset must be less than the size of vertexData", 53}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexOffset-02429", 45}, { {"vertexOffset must be a multiple of the component size of vertexFormat", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGeometryTrianglesNV-vertexStride-03818", 45}, { {"vertexStride must be less than or equal to 2^32-1", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkGetLatencyMarkerInfoNV-pTimings-parameter", 48}, { {"If timingCount is not 0, and pTimings is not NULL, pTimings must be a valid pointer to an array of timingCount VkLatencyTimingsFrameReportNV structures", 151}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkGetLatencyMarkerInfoNV-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_GET_LATENCY_MARKER_INFO_NV", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkGpaDeviceClockModeInfoAMD-clockMode-parameter", 52}, { {"clockMode must be a valid VkGpaDeviceClockModeAMD value", 55}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaDeviceClockModeInfoAMD-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaDeviceClockModeInfoAMD-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_GPA_DEVICE_CLOCK_MODE_INFO_AMD", 62}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaDeviceGetClockInfoAMD-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaDeviceGetClockInfoAMD-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_GPA_DEVICE_GET_CLOCK_INFO_AMD", 61}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaPerfBlockPropertiesAMD-blockType-parameter", 52}, { {"blockType must be a valid VkGpaPerfBlockAMD value", 49}, { "chapters/limits.html", 20} } },
        { {"VUID-VkGpaPerfBlockPropertiesAMD-flags-zerobitmask", 50}, { {"flags must be 0", 15}, { "chapters/limits.html", 20} } },
        { {"VUID-VkGpaPerfCounterAMD-blockType-parameter", 44}, { {"blockType must be a valid VkGpaPerfBlockAMD value", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-pPerfCounters-parameter", 52}, { {"pPerfCounters must be a valid pointer to an array of perfCounterCount valid VkGpaPerfCounterAMD structures", 106}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-perfCounterCount-arraylength", 57}, { {"perfCounterCount must be greater than 0", 39}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_GPA_SAMPLE_BEGIN_INFO_AMD", 57}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sampleType-12412", 45}, { {"If sampleType is VK_GPA_SAMPLE_TYPE_CUMULATIVE_AMD, the VkPhysicalDeviceGpaFeaturesAMD::perfCounters feature must be enabled", 124}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sampleType-12413", 45}, { {"If sampleType is VK_GPA_SAMPLE_TYPE_TRACE_AMD, at least one of the VkPhysicalDeviceGpaFeaturesAMD::streamingPerfCounters or VkPhysicalDeviceGpaFeaturesAMD::sqThreadTracing features must be enabled", 196}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sampleType-parameter", 49}, { {"sampleType must be a valid VkGpaSampleTypeAMD value", 51}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sqShaderMask-parameter", 51}, { {"sqShaderMask must be a valid combination of VkGpaSqShaderStageFlagBitsAMD values", 80}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-sqThreadTraceEnable-12414", 54}, { {"If sqThreadTraceEnable is VK_TRUE, the VkPhysicalDeviceGpaFeaturesAMD::sqThreadTracing feature must be enabled", 110}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-timingPostSample-parameter", 55}, { {"timingPostSample must be a valid combination of VkPipelineStageFlagBits values", 78}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSampleBeginInfoAMD-timingPreSample-parameter", 54}, { {"timingPreSample must be a valid combination of VkPipelineStageFlagBits values", 77}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSessionCreateInfoAMD-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSessionCreateInfoAMD-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_GPA_SESSION_CREATE_INFO_AMD", 59}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSessionCreateInfoAMD-secondaryCopySource-12461", 56}, { {"secondaryCopySource must have been used previously to begin and end a session, and it must not have been reset since then", 121}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGpaSessionCreateInfoAMD-secondaryCopySource-parameter", 60}, { {"If secondaryCopySource is not VK_NULL_HANDLE, secondaryCopySource must be a valid VkGpaSessionAMD handle", 104}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-Geometry-07726", 48}, { {"If the pipeline is being created with a Geometry Execution Model, uses the OutputPoints Execution Mode, and the shaderTessellationAndGeometryPointSize feature is not enabled, a PointSize decorated variable must not be written to", 228}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-Input-06632", 45}, { {"If the pipeline requires fragment shader state with a fragment shader that either enables sample shading or decorates any variable in the Input storage class with Sample, and the VK_EXT_extended_dynamic_state3 extension is not enabled or any of the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT, VK_DYNAMIC_STATE_SAMPLE_MASK_EXT, or VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic states is not set, or the alphaToOne feature is enabled and VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT is not set, then pMultisampleState must be a valid pointer to a valid VkPipelineMultisampleStateCreateInfo structure", 601}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-Input-07904", 45}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and the pipeline is being created with vertex input state and pVertexInputState is not dynamic, then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription::location", 383}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-Input-08733", 45}, { {"If the pipeline requires vertex input state and pVertexInputState is not dynamic, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription::format", 274}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-02322", 44}, { {"If the pipeline requires pre-rasterization shader state, and there are any mesh shader stages in the pipeline there must not be any shader stage in the pipeline with a Xfb execution mode", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-04574", 44}, { {"If the pipeline requires fragment shader state, and the supersampleFragmentShadingRates feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRate must not be equal to VK_FRAGMENT_SHADING_RATE_2_INVOCATIONS_PER_PIXEL_NV, VK_FRAGMENT_SHADING_RATE_4_INVOCATIONS_PER_PIXEL_NV, VK_FRAGMENT_SHADING_RATE_8_INVOCATIONS_PER_PIXEL_NV, or VK_FRAGMENT_SHADING_RATE_16_INVOCATIONS_PER_PIXEL_NV", 411}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-04575", 44}, { {"If the pipeline requires fragment shader state, and the noInvocationFragmentShadingRates feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRate must not be equal to VK_FRAGMENT_SHADING_RATE_NO_INVOCATIONS_NV", 240}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-04889", 44}, { {"If the pipeline requires pre-rasterization shader state and fragment shader state, the fragment shader and last pre-rasterization shader stage and any relevant state must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 250}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-06569", 44}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRateType must be a valid VkFragmentShadingRateTypeNV value", 261}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-07826", 44}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, the pipeline includes a complete set of state, and there are no libraries included in VkPipelineLibraryCreateInfoKHR::pLibraries, then VkPipelineLayout must be a valid pipeline layout", 292}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-09043", 44}, { {"If pDynamicState->pDynamicStates does not include VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the colorWriteMask member of the corresponding element of pColorBlendState->pAttachments must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 404}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-09497", 44}, { {"If the pNext chain does not include a VkPipelineCreateFlags2CreateInfo structure, flags must be a valid combination of VkPipelineCreateFlagBits values", 150}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-09526", 44}, { {"If the dynamicRenderingLocalRead feature is not enabled, the pipeline requires fragment output interface state, and the flags member of VkPipelineDepthStencilStateCreateInfo includes VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT or VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT, renderPass must not be VK_NULL_HANDLE", 408}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-None-10391", 44}, { {"If a resource variables is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-PrimitiveId-06264", 51}, { {"If the pipeline requires pre-rasterization shader state, it includes a mesh shader and the fragment shader code reads from an input variable that is decorated with PrimitiveId, then the mesh shader code must write to a matching output variable, decorated with PrimitiveId, in all execution paths", 295}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-TaskNV-07063", 46}, { {"The shader stages for VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT must use either the TaskNV and MeshNV Execution Model or the TaskEXT and MeshEXT Execution Model, but must not use both", 202}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-TessellationEvaluation-07723", 62}, { {"If the pipeline is being created with a TessellationEvaluation Execution Model, no Geometry Execution Model, uses the PointMode Execution Mode, and the shaderTessellationAndGeometryPointSize feature is enabled, a PointSize decorated variable must be written to if the maintenance5 feature is not enabled", 303}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-TessellationEvaluation-07724", 62}, { {"If the pipeline is being created with a TessellationEvaluation Execution Model, no Geometry Execution Model, uses the PointMode Execution Mode, and the shaderTessellationAndGeometryPointSize feature is not enabled, a PointSize decorated variable must not be written to", 268}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-alphaToCoverageEnable-08891", 61}, { {"If the pipeline is being created with fragment shader state, the VkPipelineMultisampleStateCreateInfo::alphaToCoverageEnable is not ignored and is VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 268}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-commonparent", 46}, { {"Each of basePipelineHandle, layout, and renderPass that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 170}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-conservativePointAndLineRasterization-08892", 77}, { {"If conservativePointAndLineRasterization is not supported and the effective rasterization input topology is in line or point topology class, then VkPipelineRasterizationConservativeStateCreateInfoEXT::conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 286}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-coverageReductionMode-02722", 61}, { {"If the pipeline requires fragment output interface state, the coverageReductionMode feature is enabled, and rasterizationSamples is not dynamic, the coverage reduction mode specified by VkPipelineCoverageReductionStateCreateInfoNV::coverageReductionMode, the rasterizationSamples member of pMultisampleState and the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 505}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-customResolve-11858", 53}, { {"If one element of VkPipelineLibraryCreateInfoKHR includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VkCustomResolveCreateInfoEXT is included and VkCustomResolveCreateInfoEXT::customResolve is VK_TRUE, and another element of VkPipelineLibraryCreateInfoKHR::pLibraries also includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, the other library must also include VkCustomResolveCreateInfoEXT and the VkCustomResolveCreateInfoEXT::customResolve specified by the library must be VK_TRUE", 650}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-customResolve-11859", 53}, { {"If one element of VkPipelineLibraryCreateInfoKHR includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VkCustomResolveCreateInfoEXT is not included or VkCustomResolveCreateInfoEXT::customResolve is VK_FALSE, and another element of VkPipelineLibraryCreateInfoKHR::pLibraries also includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, either the other library must not include VkCustomResolveCreateInfoEXT or the VkCustomResolveCreateInfoEXT::customResolve specified by the library must be VK_FALSE", 660}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-depthStencilAttachmentSamples-06593", 69}, { {"If the fragment output interface state and the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV is not 0, it must be a valid VkSampleCountFlagBits value", 207}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-dynamicPrimitiveTopologyUnrestricted-09031", 76}, { {"If the pipeline requires vertex input state, and related dynamic state is not set, pInputAssemblyState must be a valid pointer to a valid VkPipelineInputAssemblyStateCreateInfo structure", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-dynamicRendering-06576", 56}, { {"If the dynamicRendering feature is not enabled and the pipeline requires pre-rasterization shader state, fragment shader state, or fragment output interface state, renderPass must not be VK_NULL_HANDLE", 201}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3AlphaToCoverageEnable-07375", 82}, { {"If the extendedDynamicState3AlphaToCoverageEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT", 197}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3AlphaToOneEnable-07376", 77}, { {"If the extendedDynamicState3AlphaToOneEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ColorBlendAdvanced-07386", 79}, { {"If the extendedDynamicState3ColorBlendAdvanced feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ColorBlendEnable-07378", 77}, { {"If the extendedDynamicState3ColorBlendEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ColorBlendEquation-07379", 79}, { {"If the extendedDynamicState3ColorBlendEquation feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ColorWriteMask-07380", 75}, { {"If the extendedDynamicState3ColorWriteMask feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT", 182}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ConservativeRasterizationMode-07382", 90}, { {"If the extendedDynamicState3ConservativeRasterizationMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT", 212}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageModulationMode-07395", 83}, { {"If the extendedDynamicState3CoverageModulationMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV", 197}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageModulationTable-07397", 84}, { {"If the extendedDynamicState3CoverageModulationTable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV", 199}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageModulationTableEnable-07396", 90}, { {"If the extendedDynamicState3CoverageModulationTableEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV", 212}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageReductionMode-07398", 82}, { {"If the extendedDynamicState3CoverageReductionMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV", 195}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageToColorEnable-07393", 82}, { {"If the extendedDynamicState3CoverageToColorEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV", 196}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3CoverageToColorLocation-07394", 84}, { {"If the extendedDynamicState3CoverageToColorLocation feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV", 200}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3DepthClampEnable-07371", 77}, { {"If the extendedDynamicState3DepthClampEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3DepthClipEnable-07384", 76}, { {"If the extendedDynamicState3DepthClipEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT", 184}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3DepthClipNegativeOneToOne-07390", 86}, { {"If the extendedDynamicState3DepthClipNegativeOneToOne feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT", 207}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ExtraPrimitiveOverestimationSize-07383", 93}, { {"If the extendedDynamicState3ExtraPrimitiveOverestimationSize feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT", 219}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3LineRasterizationMode-07388", 82}, { {"If the extendedDynamicState3LineRasterizationMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT", 196}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3LineStippleEnable-07389", 78}, { {"If the extendedDynamicState3LineStippleEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3LogicOpEnable-07377", 74}, { {"If the extendedDynamicState3LogicOpEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT", 180}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3PolygonMode-07372", 72}, { {"If the extendedDynamicState3PolygonMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_POLYGON_MODE_EXT", 175}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ProvokingVertexMode-07387", 80}, { {"If the extendedDynamicState3ProvokingVertexMode feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT", 192}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3RasterizationSamples-07373", 81}, { {"If the extendedDynamicState3RasterizationSamples feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3RasterizationStream-07381", 80}, { {"If the extendedDynamicState3RasterizationStream feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3RepresentativeFragmentTestEnable-07399", 93}, { {"If the extendedDynamicState3RepresentativeFragmentTestEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV", 218}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3SampleLocationsEnable-07385", 82}, { {"If the extendedDynamicState3SampleLocationsEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT", 196}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3SampleMask-07374", 71}, { {"If the extendedDynamicState3SampleMask feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_SAMPLE_MASK_EXT", 173}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ShadingRateImageEnable-07400", 83}, { {"If the extendedDynamicState3ShadingRateImageEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV", 198}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3TessellationDomainOrigin-07370", 85}, { {"If the extendedDynamicState3TessellationDomainOrigin feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT", 202}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ViewportSwizzle-07392", 76}, { {"If the extendedDynamicState3ViewportSwizzle feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV", 182}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-extendedDynamicState3ViewportWScalingEnable-07391", 83}, { {"If the extendedDynamicState3ViewportWScalingEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV", 198}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09301", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, VkPipelineRenderingCreateInfo::viewMask must be 0", 237}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09304", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, VkExternalFormatANDROID::externalFormat is not 0, and rasterizationSamples is not dynamic, VkPipelineMultisampleStateCreateInfo::rasterizationSamples must be 1", 293}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09305", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, and blendEnable is not dynamic, the blendEnable member of each element of pColorBlendState->pAttachments must be VK_FALSE", 309}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09306", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, and pDynamicState->pDynamicStates does not include VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width must be 1", 360}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09307", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, and pDynamicState->pDynamicStates does not include VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must be 1", 361}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09308", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires pre-rasterization shader state and fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, the last pre-rasterization shader stage must not statically use a variable with the PrimitiveShadingRateKHR built-in", 339}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09309", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, VkPipelineRenderingCreateInfo::colorAttachmentCount must be 1", 249}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09310", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment shader state and fragment output interface state, renderPass is VK_NULL_HANDLE, and VkExternalFormatANDROID::externalFormat is not 0, the fragment shader must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 311}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09313", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is not VK_NULL_HANDLE, subpass includes an external format resolve attachment, and rasterizationSamples is not dynamic, VkPipelineMultisampleStateCreateInfo::rasterizationSamples must be VK_SAMPLE_COUNT_1_BIT", 323}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09314", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is not VK_NULL_HANDLE, subpass includes an external format resolve attachment, and blendEnable is not dynamic, the blendEnable member of each element of pColorBlendState->pAttachments must be VK_FALSE", 315}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09315", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is not VK_NULL_HANDLE, subpass includes an external format resolve attachment, and pDynamicState->pDynamicStates does not include VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width must be 1", 366}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09316", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires fragment output interface state, renderPass is not VK_NULL_HANDLE, subpass includes an external format resolve attachment, and pDynamicState->pDynamicStates does not include VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must be 1", 367}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-externalFormatResolve-09317", 61}, { {"If the externalFormatResolve feature is enabled, the pipeline requires pre-rasterization shader state and fragment output interface state, renderPass is not VK_NULL_HANDLE, and subpass includes an external format resolve attachment, the last pre-rasterization shader stage must not statically use a variable with the PrimitiveShadingRateKHR built-in", 349}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-00764", 45}, { {"flags must not contain the VK_PIPELINE_CREATE_DISPATCH_BASE_BIT flag", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-02877", 45}, { {"If flags includes VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV, then the VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 170}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-02966", 45}, { {"If the pipeline requires pre-rasterization shader state and flags includes VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV, then all stages must not specify Xfb execution mode", 171}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03372", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03373", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03374", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03375", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03376", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03377", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-03577", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 96}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-04947", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06482", 45}, { {"If the dynamicRenderingLocalRead feature is not enabled, the pipeline requires fragment output interface state, and the flags member of VkPipelineColorBlendStateCreateInfo includes VK_PIPELINE_COLOR_BLEND_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_BIT_EXT, renderPass must not be VK_NULL_HANDLE", 302}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06591", 45}, { {"If pStages includes a fragment shader stage, and the fragment shader declares the EarlyFragmentTests execution mode, the flags member of VkPipelineDepthStencilStateCreateInfo must not include VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT or VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT", 378}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06596", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, the value of VkMultiviewPerViewAttributesInfoNVX::perViewAttributes specified in both this pipeline and the library must be equal", 405}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06598", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, the value of VkMultiviewPerViewAttributesInfoNVX::perViewAttributesPositionXOnly specified in both this pipeline and the library must be equal", 418}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06608", 45}, { {"If the shaderMeshEnqueue feature is not enabled, and the pipeline is being created with all possible state subsets, flags must not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 173}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06609", 45}, { {"If flags includes VK_PIPELINE_CREATE_LINK_TIME_OPTIMIZATION_BIT_EXT, pipeline libraries included via VkPipelineLibraryCreateInfoKHR must have been created with VK_PIPELINE_CREATE_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT", 221}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06610", 45}, { {"If flags includes VK_PIPELINE_CREATE_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT, pipeline libraries included via VkPipelineLibraryCreateInfoKHR must have been created with VK_PIPELINE_CREATE_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT", 233}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06612", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, and layout was not created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then the layout used by this pipeline and the library must be identically defined", 547}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06614", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other subset, and layout was created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then the layout used by the library must also have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 579}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06616", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other subset, and layout was created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, elements of the pSetLayouts array which layout was created with that are not VK_NULL_HANDLE must be identically defined to the element at the same index of pSetLayouts used to create the library's layout", 663}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06618", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, any descriptor set layout N specified by layout in both this pipeline and the library which include bindings accessed by shader stages in each must be identically defined", 552}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06620", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, push constants specified in layout in both this pipeline and the library which are available to shader stages in each must be identically defined", 527}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06626", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes either VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and the other element is included in VkPipelineLibraryCreateInfoKHR::pLibraries, and renderPass is VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::viewMask used by this pipeline and that specified by the library must be identical", 426}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06633", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT with a pMultisampleState that was not NULL, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, pMultisampleState must be identically defined to that used to create the library", 377}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06637", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, pMultisampleState->sampleShadingEnable is VK_TRUE, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, pMultisampleState must be identically defined to that used to create the library", 385}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06638", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other flag, values specified in VkPipelineFragmentShadingRateStateCreateInfoKHR for both this pipeline and that library must be identical", 401}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06640", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 284}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06642", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, layout must be a valid VkPipelineLayout handle", 330}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06643", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and renderPass is not VK_NULL_HANDLE, renderPass must be a valid VkRenderPass handle", 327}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06644", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, stageCount must be greater than 0", 211}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06645", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags is non-zero, if flags includes VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR, any libraries must have also been created with VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR", 247}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06679", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, and VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other subset, VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT was not used, and any element of the pSetLayouts array when layout was created and the corresponding element of the pSetLayouts array used to create the library's layout must not both be VK_NULL_HANDLE", 636}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06682", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, and VkGraphicsPipelineLibraryCreateInfoEXT::flags includes both VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, layout must have been created with no elements of the pSetLayouts array set to VK_NULL_HANDLE", 387}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06683", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and pRasterizationState->rasterizerDiscardEnable is VK_TRUE, layout must have been created with no elements of the pSetLayouts array set to VK_NULL_HANDLE", 382}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06684", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes at least one of and no more than two of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes one of the other flags, the value of subpass must be equal to that used to create the library", 446}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06729", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, flags includes VK_PIPELINE_CREATE_LINK_TIME_OPTIMIZATION_BIT_EXT, the pipeline includes a complete set of state specified entirely by libraries, and each library was created with a VkPipelineLayout created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then layout must be compatible with the union of the libraries' pipeline layouts other than the inclusion/exclusion of VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 541}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06730", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, flags does not include VK_PIPELINE_CREATE_LINK_TIME_OPTIMIZATION_BIT_EXT, the pipeline includes a complete set of state specified entirely by libraries, and each library was created with a VkPipelineLayout created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then layout must be compatible with the union of the libraries' pipeline layouts", 460}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06756", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other subset, and any element of the pSetLayouts array which layout was created with was VK_NULL_HANDLE, then the corresponding element of the pSetLayouts array used to create the library's layout must not have shader bindings for shaders in the other subset", 624}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-06757", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes only one of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes the other subset, and any element of the pSetLayouts array used to create the library's layout was VK_NULL_HANDLE, then the corresponding element of the pSetLayouts array used to create this pipeline's layout must not have shader bindings for shaders in the other subset", 632}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07369", 45}, { {"flags must not include both VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT and VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07401", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07984", 45}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineIndex is -1, basePipelineHandle must be a valid graphics VkPipeline handle", 152}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07985", 45}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineHandle is VK_NULL_HANDLE, basePipelineIndex must be a valid index into the calling command's pCreateInfos parameter", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07986", 45}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, basePipelineIndex must be -1 or basePipelineHandle must be VK_NULL_HANDLE", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-07997", 45}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08897", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_VERTEX_INPUT_INTERFACE_BIT_EXT, pre-rasterization shader state is specified either in a library or by the inclusion of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and that state includes a vertex shader stage in pStages, the pipeline must define vertex input state", 371}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08898", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_VERTEX_INPUT_INTERFACE_BIT_EXT, and pre-rasterization shader state is not specified, the pipeline must define vertex input state", 215}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08899", 45}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, pre-rasterization shader state is specified either in a library or by the inclusion of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and that state includes a vertex shader stage in pStages, the pipeline must either define vertex input state or include that state in a linked pipeline library", 372}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08900", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT the pipeline must define pre-rasterization shader state", 176}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08901", 45}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, the pipeline must either define pre-rasterization shader state or include that state in a linked pipeline library", 175}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08903", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, pre-rasterization shader state is specified either in a library or by the inclusion of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and that state either includes VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE or has pRasterizationState->rasterizerDiscardEnable set to VK_FALSE, the pipeline must define fragment shader state", 452}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08904", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and pre-rasterization shader state is not specified, the pipeline must define fragment shader state", 211}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08906", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, pre-rasterization shader state is specified either in a library or by the inclusion of VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and that state either includes VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE or has pRasterizationState->rasterizerDiscardEnable set to VK_FALSE, the pipeline must define fragment output interface state", 462}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08907", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and pre-rasterization shader state is not specified, the pipeline must define fragment output interface state", 221}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-08909", 45}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, pre-rasterization shader state is specified either in a library or by the inclusion of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and that state either includes VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE or has pRasterizationState->rasterizerDiscardEnable set to VK_FALSE, the pipeline must define fragment output interface state and fragment shader state or include those states in linked pipeline libraries", 491}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-09245", 45}, { {"If flags includes VK_PIPELINE_CREATE_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT, flags must also include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 139}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-10188", 45}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, shaders specified by pStages must not declare the ShaderEnqueueAMDX capability", 140}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-10190", 45}, { {"If VK_PIPELINE_CREATE_2_EXECUTION_GRAPH_BIT_AMDX is included in flags, and the pipeline requires pre-rasterization shader state, there must not be a task or vertex shader specified in pStages", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-10191", 45}, { {"If VK_PIPELINE_CREATE_2_EXECUTION_GRAPH_BIT_AMDX is included in flags, all elements of VkPipelineLibraryCreateInfoKHR::pLibraries must have been created with VK_PIPELINE_CREATE_2_EXECUTION_GRAPH_BIT_AMDX", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11000", 45}, { {"If flags includes VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT, then the VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11001", 45}, { {"If the pipeline requires pre-rasterization shader state and flags includes VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT, then all stages must not specify Xfb execution mode", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11273", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also have that flag set", 167}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11274", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also not have that flag set", 179}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11311", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be VK_NULL_HANDLE", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11312", 45}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 287}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11595", 45}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature is not enabled, flags must not include VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 181}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11798", 45}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11856", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VkCustomResolveCreateInfoEXT is included and VkCustomResolveCreateInfoEXT::customResolve is VK_TRUE, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries also includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, the library must also include VkCustomResolveCreateInfoEXT and the VkCustomResolveCreateInfoEXT::customResolve specified by the library must be VK_TRUE", 639}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-11857", 45}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VkCustomResolveCreateInfoEXT is not included or VkCustomResolveCreateInfoEXT::customResolve is VK_FALSE, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries also includes either VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, either the library must not include VkCustomResolveCreateInfoEXT or the VkCustomResolveCreateInfoEXT::customResolve specified by the library must be VK_FALSE", 649}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12355", 45}, { {"If flags includes VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, all libraries linked to this pipeline must also have that flag set", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12356", 45}, { {"If flags does not include VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, all libraries linked to this pipeline must also not have that flag set", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12357", 45}, { {"If flags includes VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT, all libraries linked to this pipeline must also have that flag set", 128}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12358", 45}, { {"If flags does not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT, all libraries linked to this pipeline must also not have that flag set", 140}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12359", 45}, { {"If flags includes VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT, all libraries linked to this pipeline must also have that flag set", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-flags-12360", 45}, { {"If flags does not include VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT, all libraries linked to this pipeline must also not have that flag set", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-fragmentShadingRateNonTrivialCombinerOps-04506", 80}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state, the fragmentShadingRateNonTrivialCombinerOps limit is not supported, and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, elements of VkPipelineFragmentShadingRateStateCreateInfoKHR::combinerOps must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR or VK_FRAGMENT_SHADING_RATE_COMBINER_OP_REPLACE_KHR", 426}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-fragmentShadingRateNonTrivialCombinerOps-04573", 80}, { {"If the pipeline requires fragment shader state, and the fragmentShadingRateNonTrivialCombinerOps limit is not supported and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, elements of VkPipelineFragmentShadingRateEnumStateCreateInfoNV::combinerOps must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR or VK_FRAGMENT_SHADING_RATE_COMBINER_OP_REPLACE_KHR", 398}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-geometryStreams-02321", 55}, { {"If the pipeline requires pre-rasterization shader state, and the last pre-rasterization shader stage is a geometry shader, and that geometry shader uses the GeometryStreams capability, then VkPhysicalDeviceTransformFeedbackFeaturesEXT::geometryStreams feature must be enabled", 275}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-graphicsPipelineLibrary-06606", 63}, { {"If the graphicsPipelineLibrary feature is not enabled, and if the shaderMeshEnqueue feature is not enabled, flags must not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 165}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-01688", 46}, { {"The number of resources in layout accessible to each shader stage that is used by the pipeline must be less than or equal to VkPhysicalDeviceLimits::maxPerStageResources", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-06602", 46}, { {"If the pipeline requires fragment shader state or pre-rasterization shader state, and VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be a valid VkPipelineLayout handle", 238}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-07827", 46}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, the pipeline includes a complete set of state specified entirely by libraries, and each library was created with a VkPipelineLayout created without VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then layout must be compatible with the layouts in those libraries", 375}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-07987", 46}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, a push constant range in layout must match the shader stage", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-07988", 46}, { {"If a resource variable is declared in a shader and layout is not VK_NULL_HANDLE, the corresponding descriptor set in layout must match the shader stage", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-07990", 46}, { {"If a resource variable is declared in a shader, layout is not VK_NULL_HANDLE, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in layout must match the descriptor type", 214}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-07991", 46}, { {"If a resource variable is declared in a shader as an array and layout is not VK_NULL_HANDLE, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 236}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-layout-10069", 46}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, the block must be contained inside the push constant range in layout that matches the stage", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-libraryCount-06648", 52}, { {"If the pipeline is not created with a complete set of state, or VkPipelineLibraryCreateInfoKHR::libraryCount is not 0, VkGraphicsPipelineShaderGroupsCreateInfoNV::groupCount and VkGraphicsPipelineShaderGroupsCreateInfoNV::pipelineCount must be 0", 245}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-libraryCount-06649", 52}, { {"If the pipeline is created with a complete set of state, and VkPipelineLibraryCreateInfoKHR::libraryCount is 0, and the pNext chain includes an instance of VkGraphicsPipelineShaderGroupsCreateInfoNV, VkGraphicsPipelineShaderGroupsCreateInfoNV::groupCount must be greater than 0", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-lineRasterizationMode-02766", 61}, { {"If the pipeline requires pre-rasterization shader state and at least one of fragment output interface state or fragment shader state, and pMultisampleState is not NULL, the lineRasterizationMode member of a VkPipelineRasterizationLineStateCreateInfo structure included in the pNext chain of pRasterizationState is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the alphaToCoverageEnable, alphaToOneEnable, and sampleShadingEnable members of pMultisampleState must all be VK_FALSE", 524}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-multisampledRenderToSingleSampled-06853", 73}, { {"If the pipeline requires fragment output interface state, the multisampledRenderToSingleSampled feature is not enabled, the VK_AMD_mixed_attachment_samples extension is not enabled, the VK_NV_framebuffer_mixed_samples extension is not enabled, rasterizationSamples is not dynamic, and subpass uses color and/or depth/stencil attachments, then the rasterizationSamples member of pMultisampleState must be the same as the sample count for those subpass attachments", 462}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-multiview-06577", 49}, { {"If the multiview feature is not enabled, the pipeline requires pre-rasterization shader state or fragment shader state, and renderPass is VK_NULL_HANDLE, VkPipelineRenderingCreateInfo::viewMask must be 0", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-multiviewPerViewViewports-12249", 65}, { {"If the multiviewPerViewViewports feature is enabled, renderpass is VK_NULL_HANDLE, the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT, then the index of the most significant bit in VkPipelineRenderingCreateInfo::viewMask must be less than pViewportState->viewportCount", 372}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-multiviewPerViewViewports-12250", 65}, { {"If the multiviewPerViewViewports feature is enabled, renderpass is VK_NULL_HANDLE, the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT, then the index of the most significant bit in VkPipelineRenderingCreateInfo::viewMask must be less than pViewportState->scissorCount", 370}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pColorAttachmentSamples-06592", 63}, { {"If the fragment output interface state, elements of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV must be valid VkSampleCountFlagBits values", 196}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pColorBlendState-09038", 56}, { {"If pColorBlendState is not NULL it must be a valid pointer to a valid VkPipelineColorBlendStateCreateInfo structure", 115}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDepthStencilState-09029", 58}, { {"If pDepthStencilState is not NULL it must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDepthStencilState-09034", 58}, { {"If pDepthStencilState is not NULL it must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDepthStencilState-09036", 58}, { {"If pDepthStencilState is not NULL it must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04494", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width must be greater than or equal to 1", 280}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04495", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must be greater than or equal to 1", 281}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04496", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width must be a power-of-two value", 274}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04497", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must be a power-of-two value", 275}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04498", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width must be less than or equal to 4", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04499", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must be less than or equal to 4", 278}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04500", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the pipelineFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.width and VkPipelineFragmentShadingRateStateCreateInfoKHR::fragmentSize.height must both be equal to 1", 402}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04501", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the primitiveFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateStateCreateInfoKHR::combinerOps[0] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 356}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04502", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the attachmentFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateStateCreateInfoKHR::combinerOps[1] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 357}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04569", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the fragmentShadingRateEnums feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRateType must be equal to VK_FRAGMENT_SHADING_RATE_TYPE_FRAGMENT_SIZE_NV", 332}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04570", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the pipelineFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRate must be equal to VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_PIXEL_NV", 335}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04571", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the primitiveFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::combinerOps[0] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 325}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-04572", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, and the attachmentFragmentShadingRate feature is not enabled, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::combinerOps[1] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 326}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-06567", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::combinerOps[0] must be a valid VkFragmentShadingRateCombinerOpKHR value", 298}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-06568", 53}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateStateCreateInfoKHR::combinerOps[1] must be a valid VkFragmentShadingRateCombinerOpKHR value", 298}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-06570", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::shadingRate must be a valid VkFragmentShadingRateNV value", 253}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-06571", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::combinerOps[0] must be a valid VkFragmentShadingRateCombinerOpKHR value", 267}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-06572", 53}, { {"If the pipeline requires fragment shader state and VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR is not included in pDynamicState->pDynamicStates, VkPipelineFragmentShadingRateEnumStateCreateInfoNV::combinerOps[1] must be a valid VkFragmentShadingRateCombinerOpKHR value", 267}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-09639", 53}, { {"If the pipeline requires pre-rasterization shader state, pDynamicState includes VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT, and pDynamicState does not include VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT, pRasterizationState must include a VkPipelineRasterizationConservativeStateCreateInfoEXT in its pNext chain", 334}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicState-parameter", 57}, { {"If pDynamicState is not NULL, pDynamicState must be a valid pointer to a valid VkPipelineDynamicStateCreateInfo structure", 121}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-00749", 54}, { {"If the pipeline requires pre-rasterization shader state, and the wideLines feature is not enabled, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_LINE_WIDTH, the lineWidth member of pRasterizationState must be 1.0", 247}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-00754", 54}, { {"If the pipeline requires pre-rasterization shader state, the depthBiasClamp feature is not enabled, no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_DEPTH_BIAS, and the depthBiasEnable member of pRasterizationState is VK_TRUE, the depthBiasClamp member of pRasterizationState must be 0.0", 315}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-01715", 54}, { {"If the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV, and the viewportWScalingEnable member of a VkPipelineViewportWScalingStateCreateInfoNV structure, included in the pNext chain of pViewportState, is VK_TRUE, the pViewportWScalings member of the VkPipelineViewportWScalingStateCreateInfoNV must be a pointer to an array of VkPipelineViewportWScalingStateCreateInfoNV::viewportCount valid VkViewportWScalingNV structures", 528}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-02510", 54}, { {"If the pipeline requires fragment shader state, the VK_EXT_depth_range_unrestricted extension is not enabled and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_DEPTH_BOUNDS, and the depthBoundsTestEnable member of pDepthStencilState is VK_TRUE, the minDepthBounds and maxDepthBounds members of pDepthStencilState must be between 0.0 and 1.0, inclusive", 381}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-03378", 54}, { {"If the extendedDynamicState feature is not enabled, and the minimum value of VkApplicationInfo::apiVersion used to create the VkInstance and apiVersion supported by the physical device is less than Version 1.3 there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_CULL_MODE, VK_DYNAMIC_STATE_FRONT_FACE, VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY, VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT, VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT, VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE, VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE, VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE, VK_DYNAMIC_STATE_DEPTH_COMPARE_OP, VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE, VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE, or VK_DYNAMIC_STATE_STENCIL_OP", 722}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-03379", 54}, { {"If the pipeline requires pre-rasterization shader state, and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT is included in the pDynamicStates array then viewportCount must be zero", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-03380", 54}, { {"If the pipeline requires pre-rasterization shader state, and VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT is included in the pDynamicStates array then scissorCount must be zero", 167}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-03578", 54}, { {"All elements of the pDynamicStates member of pDynamicState must not be VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR", 123}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04056", 54}, { {"If the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV, and if pViewportState->pNext chain includes a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure, and if its exclusiveScissorCount member is not 0, then its pExclusiveScissors member must be a valid pointer to an array of exclusiveScissorCount VkRect2D structures", 434}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04057", 54}, { {"If the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV, and if pViewportState->pNext chain includes a VkPipelineViewportShadingRateImageStateCreateInfoNV structure, then its pShadingRatePalettes member must be a valid pointer to an array of viewportCount valid VkShadingRatePaletteNV structures", 410}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04058", 54}, { {"If the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT, and if pNext chain includes a VkPipelineDiscardRectangleStateCreateInfoEXT structure, and if its discardRectangleCount member is not 0, then its pDiscardRectangles member must be a valid pointer to an array of discardRectangleCount VkRect2D structures", 412}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04130", 54}, { {"If the pipeline requires pre-rasterization shader state, and pViewportState->pViewports is not dynamic, then pViewportState->pViewports must be a valid pointer to an array of pViewportState->viewportCount valid VkViewport structures", 232}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04131", 54}, { {"If the pipeline requires pre-rasterization shader state, and pViewportState->pScissors is not dynamic, then pViewportState->pScissors must be a valid pointer to an array of pViewportState->scissorCount VkRect2D structures", 221}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04132", 54}, { {"If the pipeline requires pre-rasterization shader state, and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT is included in the pDynamicStates array then VK_DYNAMIC_STATE_VIEWPORT must not be present", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04133", 54}, { {"If the pipeline requires pre-rasterization shader state, and VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT is included in the pDynamicStates array then VK_DYNAMIC_STATE_SCISSOR must not be present", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04800", 54}, { {"If the colorWriteEnable feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT", 165}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04807", 54}, { {"If the pipeline requires pre-rasterization shader state and the vertexInputDynamicState feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_VERTEX_INPUT_EXT", 223}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04868", 54}, { {"If the extendedDynamicState2 feature is not enabled, and the minimum value of VkApplicationInfo::apiVersion used to create the VkInstance and apiVersion supported by the physical device is less than Version 1.3 there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE, VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE, or VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE", 413}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04869", 54}, { {"If the extendedDynamicState2LogicOp feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_LOGIC_OP_EXT", 167}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-04870", 54}, { {"If the extendedDynamicState2PatchControlPoints feature is not enabled, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07065", 54}, { {"If the pipeline requires pre-rasterization shader state, and includes a mesh shader, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY, or VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE", 247}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07066", 54}, { {"If the pipeline requires pre-rasterization shader state, and includes a mesh shader, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE, or VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT", 250}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07067", 54}, { {"If the pipeline requires pre-rasterization shader state, and includes a mesh shader, there must be no element of the pDynamicStates member of pDynamicState set to VK_DYNAMIC_STATE_VERTEX_INPUT_EXT", 196}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07610", 54}, { {"If the pipeline requires fragment shader state or fragment output interface state, and rasterizationSamples and sampleLocationsInfo are not dynamic, and VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable included in the pNext chain of pMultisampleState is VK_TRUE, sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 522}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07611", 54}, { {"If the pipeline requires fragment shader state or fragment output interface state, and rasterizationSamples and sampleLocationsInfo are not dynamic, and VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable the included in the pNext chain of pMultisampleState is VK_TRUE or VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT is used, sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 584}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07612", 54}, { {"If the pipeline requires fragment shader state or fragment output interface state, and rasterizationSamples and sampleLocationsInfo are not dynamic, and VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable included in the pNext chain of pMultisampleState is VK_TRUE or VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT is used, sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples", 412}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07730", 54}, { {"If the multiviewPerViewViewports feature is enabled, renderpass is not VK_NULL_HANDLE, the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT, then the index of the most significant bit in each element of VkRenderPassMultiviewCreateInfo::pViewMasks must be less than pViewportState->viewportCount", 396}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07731", 54}, { {"If the multiviewPerViewViewports feature is enabled, renderpass is not VK_NULL_HANDLE, the pipeline requires pre-rasterization shader state, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT, then the index of the most significant bit in each element of VkRenderPassMultiviewCreateInfo::pViewMasks must be less than pViewportState->scissorCount", 394}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07854", 54}, { {"If VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV is included in the pDynamicStates array then the implementation must support at least specVersion 2 of the VK_NV_scissor_exclusive extension", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07855", 54}, { {"If VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT is included in the pDynamicStates array then the implementation must support at least specVersion 2 of the VK_EXT_discard_rectangles extension", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-07856", 54}, { {"If VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT is included in the pDynamicStates array then the implementation must support at least specVersion 2 of the VK_EXT_discard_rectangles extension", 189}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pDynamicStates-10913", 54}, { {"If the pipeline requires fragment shader state, no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_DEPTH_BOUNDS, and pDynamicStates includes VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE or the depthBoundsTestEnable member of pDepthStencilState is VK_TRUE, minDepthBounds must be less than or equal to maxDepthBounds", 338}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pInputAssemblyState-09032", 59}, { {"If pInputAssemblyState is not NULL it must be a valid pointer to a valid VkPipelineInputAssemblyStateCreateInfo structure", 121}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06597", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, the value of VkMultiviewPerViewAttributesInfoNVX::perViewAttributes specified in both libraries must be equal", 325}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06599", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, the value of VkMultiviewPerViewAttributesInfoNVX::perViewAttributesPositionXOnly specified in both libraries must be equal", 338}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06611", 50}, { {"Any pipeline libraries included via VkPipelineLibraryCreateInfoKHR::pLibraries must not include any state subset already defined by this structure or defined by any other pipeline library in VkPipelineLibraryCreateInfoKHR::pLibraries", 233}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06613", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and the layout specified by either library was not created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then the layout used by each library must be identically defined", 502}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06615", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and the layout specified by either library was created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, then the layout used by both libraries must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 551}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06617", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and the layout specified by either library was created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT, elements of the pSetLayouts array which either layout was created with that are not VK_NULL_HANDLE must be identically defined to the element at the same index of pSetLayouts used to create the other library's layout", 650}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06619", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, any descriptor set layout N specified by layout in both libraries which include bindings accessed by shader stages in each must be identically defined", 472}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06621", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, and one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, push constants specified in layout in both this pipeline and the library which are available to shader stages in each must be identically defined", 471}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06623", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes at least one of and no more than two of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and another element of VkPipelineLibraryCreateInfoKHR::pLibraries includes one of the other flags, the value of subpass used to create each library must be identical", 460}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06627", 50}, { {"If VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT are included in two elements of VkPipelineLibraryCreateInfoKHR::pLibraries, and renderPass was VK_NULL_HANDLE for both libraries, the value of VkPipelineRenderingCreateInfo::viewMask set by each library must be identical", 343}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06628", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes at least one of and no more than two of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and another element of VkPipelineLibraryCreateInfoKHR::pLibraries includes one of the other flags, the renderPass objects used to create each library must be compatible or all equal to VK_NULL_HANDLE", 494}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06634", 50}, { {"If an element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT with a pMultisampleState that was not NULL, and if VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, pMultisampleState must be identically defined to that used to create the library", 380}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06635", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT with a pMultisampleState that was not NULL, and if a different element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, the pMultisampleState used to create each library must be identically defined", 406}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06636", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT and a value of pMultisampleState->sampleShadingEnable equal VK_TRUE, and if a different element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, the pMultisampleState used to create each library must be identically defined", 431}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06639", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, values specified in VkPipelineFragmentShadingRateStateCreateInfoKHR for both this pipeline and that library must be identical", 341}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06646", 50}, { {"If VkPipelineLibraryCreateInfoKHR::pLibraries includes more than one library, and any library was created with VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR, all libraries must have also been created with VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR", 278}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06647", 50}, { {"If VkPipelineLibraryCreateInfoKHR::pLibraries includes at least one library, VkGraphicsPipelineLibraryCreateInfoEXT::flags is non-zero, and any library was created with VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR, flags must include VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR", 308}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06681", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT was not used, and any element of the pSetLayouts array used to create each library's layout was VK_NULL_HANDLE, then the corresponding element of the pSetLayouts array used to create the other library's layout must not be VK_NULL_HANDLE", 609}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-06758", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, one element of VkPipelineLibraryCreateInfoKHR::pLibraries includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT and another element includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, and any element of the pSetLayouts array used to create each library's layout was VK_NULL_HANDLE, then the corresponding element of the pSetLayouts array used to create the other library's layout must not have shader bindings for shaders in the other subset", 579}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-09567", 50}, { {"If one element of VkPipelineLibraryCreateInfoKHR::pLibraries was created with VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT and a value of pMultisampleState->sampleShadingEnable equal VK_TRUE, and if VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, pMultisampleState must be identically defined to that used to create the library", 406}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pLibraries-11594", 50}, { {"If flags includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR, each element of VkPipelineLibraryCreateInfoKHR::pLibraries must have been created with the VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR bit set", 267}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pMultisampleState-09026", 57}, { {"If the pipeline requires fragment output interface state, and the VK_EXT_extended_dynamic_state3 extension is not enabled or any of the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT, VK_DYNAMIC_STATE_SAMPLE_MASK_EXT, or VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic states is not set, or the alphaToOne feature is enabled and VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT is not set, pMultisampleState must be a valid pointer to a valid VkPipelineMultisampleStateCreateInfo structure", 483}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pMultisampleState-09027", 57}, { {"If pMultisampleState is not NULL it must be a valid pointer to a valid VkPipelineMultisampleStateCreateInfo structure", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pNext-12427", 45}, { {"If VkPipelineRenderingCreateInfo is included in the pNext chain, then renderPass must be VK_NULL_HANDLE", 103}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pNext-pNext", 45}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAttachmentSampleCountInfoAMD, VkCustomResolveCreateInfoEXT, VkExternalFormatANDROID, VkExternalFormatOHOS, VkGraphicsPipelineLibraryCreateInfoEXT, VkGraphicsPipelineShaderGroupsCreateInfoNV, VkMultiviewPerViewAttributesInfoNVX, VkPipelineBinaryInfoKHR, VkPipelineCompilerControlCreateInfoAMD, VkPipelineCreateFlags2CreateInfo, VkPipelineCreationFeedbackCreateInfo, VkPipelineDiscardRectangleStateCreateInfoEXT, VkPipelineFragmentDensityMapLayeredCreateInfoVALVE, VkPipelineFragmentShadingRateEnumStateCreateInfoNV, VkPipelineFragmentShadingRateStateCreateInfoKHR, VkPipelineLibraryCreateInfoKHR, VkPipelineRenderingCreateInfo, VkPipelineRepresentativeFragmentTestStateCreateInfoNV, VkPipelineRobustnessCreateInfo, VkRenderingAttachmentLocationInfo, VkRenderingInputAttachmentIndexInfo, or VkValidationFeaturesEXT", 945}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pRasterizationState-06601", 59}, { {"If the pipeline requires pre-rasterization shader state, and related dynamic state is not set, pRasterizationState must be a valid pointer to a valid VkPipelineRasterizationStateCreateInfo structure", 198}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pRasterizationState-09039", 59}, { {"If VkGraphicsPipelineLibraryCreateInfoEXT::flags includes VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, and related dynamic state is not set, then pMultisampleState must be a valid pointer to a valid VkPipelineMultisampleStateCreateInfo structure", 264}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pRasterizationState-09040", 59}, { {"If pRasterizationState is not NULL it must be a valid pointer to a valid VkPipelineRasterizationStateCreateInfo structure", 121}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00729", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes a tessellation control shader stage, it must include a tessellation evaluation shader stage", 168}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00730", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes a tessellation evaluation shader stage, it must include a tessellation control shader stage", 168}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00732", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes tessellation shader stages, the shader code of at least one stage must contain an OpExecutionMode instruction specifying the type of subdivision in the pipeline", 237}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00733", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes tessellation shader stages, and the shader code of both stages contain an OpExecutionMode instruction specifying the type of subdivision in the pipeline, they must both specify the same subdivision mode", 279}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00734", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes tessellation shader stages, the shader code of at least one stage must contain an OpExecutionMode instruction specifying the output patch size in the pipeline", 235}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00735", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes tessellation shader stages, and the shader code of both contain an OpExecutionMode instruction specifying the out patch size in the pipeline, they must both specify the same patch size", 261}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00738", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes a geometry shader stage, and does not include any tessellation shader stages, its shader code must contain an OpExecutionMode instruction specifying an input primitive type that is compatible with the primitive topology specified in pInputAssembly", 324}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00739", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes a geometry shader stage, and also includes tessellation shader stages, its shader code must contain an OpExecutionMode instruction specifying an input primitive type that is compatible with the primitive topology that is output by the tessellation stages", 331}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00740", 47}, { {"If the pipeline requires pre-rasterization shader state and fragment shader state, it includes both a fragment shader and a geometry shader, and the fragment shader code reads from an input variable that is decorated with PrimitiveId, then the geometry shader code must write to a matching output variable, decorated with PrimitiveId, in all execution paths", 357}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-00742", 47}, { {"If the pipeline requires pre-rasterization shader state and multiple pre-rasterization shader stages are included in pStages, the shader code for the entry points identified by those pStages and the rest of the state identified by this structure must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 330}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-01565", 47}, { {"If the pipeline requires fragment shader state and an input attachment was referenced by an aspectMask at renderPass creation time, the fragment shader must only read from the aspects that were specified for that input attachment", 229}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-02095", 47}, { {"If the pipeline requires pre-rasterization shader state the geometric shader stages provided in pStages must be either from the mesh shading pipeline (stage is VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT) or from the primitive shading pipeline (stage is VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, or VK_SHADER_STAGE_GEOMETRY_BIT)", 418}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-02097", 47}, { {"If the pipeline requires vertex input state, and pVertexInputState is not dynamic, then pVertexInputState must be a valid pointer to a valid VkPipelineVertexInputStateCreateInfo structure", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-02317", 47}, { {"If the pipeline requires pre-rasterization shader state, the Xfb execution mode can be specified by no more than one shader stage in pStages", 140}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-02318", 47}, { {"If the pipeline requires pre-rasterization shader state, and any shader stage in pStages specifies Xfb execution mode it must be the last pre-rasterization shader stage", 168}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-06600", 47}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state, pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-06894", 47}, { {"If the pipeline requires pre-rasterization shader state but not fragment shader state, elements of pStages must not have stage set to VK_SHADER_STAGE_FRAGMENT_BIT", 162}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-06895", 47}, { {"If the pipeline requires fragment shader state but not pre-rasterization shader state, elements of pStages must not have stage set to a shader stage which participates in pre-rasterization", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-06896", 47}, { {"If the pipeline requires pre-rasterization shader state, all elements of pStages must have a stage set to a shader stage which participates in fragment shader state or pre-rasterization shader state", 198}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-08711", 47}, { {"If pStages includes a fragment shader stage, VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE is not set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable member of VkPipelineDepthStencilStateCreateInfo must be VK_FALSE", 336}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-08712", 47}, { {"If pStages includes a fragment shader stage, VK_DYNAMIC_STATE_STENCIL_WRITE_MASK is not set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the value of VkStencilOpState::writeMask for both front and back in VkPipelineDepthStencilStateCreateInfo must be 0", 368}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-08888", 47}, { {"If the pipeline is being created with pre-rasterization shader state and vertex input state and pStages includes tessellation shader stages, and either VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state is not enabled or dynamicPrimitiveTopologyUnrestricted is VK_FALSE, the topology member of pInputAssembly must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST", 348}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-09022", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages includes a tessellation control shader stage, and the VK_EXT_extended_dynamic_state3 extension is not enabled or the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state is not set, pTessellationState must be a valid pointer to a valid VkPipelineTessellationStateCreateInfo structure", 354}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-09631", 47}, { {"If the pipeline requires pre-rasterization shader state and pStages contains both VK_SHADER_STAGE_TASK_BIT_EXT and VK_SHADER_STAGE_MESH_BIT_EXT, then the mesh shader's entry point must not declare a variable with a DrawIndex BuiltIn decoration", 243}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pStages-10189", 47}, { {"If any shader stages in pStages declare the ShaderEnqueueAMDX capability, VK_PIPELINE_CREATE_2_EXECUTION_GRAPH_BIT_AMDX and VK_PIPELINE_CREATE_2_LIBRARY_BIT_KHR must be included in flags", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pTessellationState-09023", 58}, { {"If pTessellationState is not NULL it must be a pointer to a valid VkPipelineTessellationStateCreateInfo structure", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pVertexInputState-08929", 57}, { {"If the pipeline is being created with vertex input state and pVertexInputState is not dynamic, and VkVertexInputAttributeDescription::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 310}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pVertexInputState-08930", 57}, { {"If the pipeline is being created with vertex input state and pVertexInputState is not dynamic, and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription::format must have a 64-bit component", 315}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pVertexInputState-09198", 57}, { {"If the pipeline is being created with vertex input state and pVertexInputState is not dynamic, and VkVertexInputAttributeDescription::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 321}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pViewportState-09025", 54}, { {"If pViewportState is not NULL it must be a valid pointer to a valid VkPipelineViewportStateCreateInfo structure", 111}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pipelineCreationCacheControl-02878", 68}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT nor VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pipelineProtectedAccess-07368", 63}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-pipelineStageCreationFeedbackCount-06594", 74}, { {"If VkPipelineCreationFeedbackCreateInfo::pipelineStageCreationFeedbackCount is not 0, it must be equal to stageCount", 116}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-primitiveFragmentShadingRateWithMultipleViewports-04503", 89}, { {"If the pipeline requires pre-rasterization shader state and the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT is not included in pDynamicState->pDynamicStates, and VkPipelineViewportStateCreateInfo::viewportCount is greater than 1, entry points specified in pStages must not write to the PrimitiveShadingRateKHR built-in", 385}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-primitiveFragmentShadingRateWithMultipleViewports-04504", 89}, { {"If the pipeline requires pre-rasterization shader state and the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and entry points specified in pStages write to the ViewportIndex built-in, they must not also write to the PrimitiveShadingRateKHR built-in", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-primitiveFragmentShadingRateWithMultipleViewports-04505", 89}, { {"If the pipeline requires pre-rasterization shader state and the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and entry points specified in pStages write to the ViewportMaskNV built-in, they must not also write to the PrimitiveShadingRateKHR built-in", 278}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-rasterizationSamples-04899", 60}, { {"If the pipeline requires fragment shader state, and the VK_QCOM_render_pass_shader_resolve extension or the customResolve feature is enabled, rasterizationSamples is not dynamic, and if subpass has any input attachments, and if the subpass description contains VK_SUBPASS_DESCRIPTION_FRAGMENT_REGION_BIT_EXT, then the sample count of the input attachments must equal rasterizationSamples", 387}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-rasterizationStream-02319", 59}, { {"If the pipeline requires pre-rasterization shader state, and a VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream value other than zero is specified, all variables in the output interface of the entry point being compiled decorated with Position, PointSize, ClipDistance, or CullDistance must be decorated with identical Stream values that match the rasterizationStream", 387}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-rasterizationStream-02320", 59}, { {"If the pipeline requires pre-rasterization shader state, and VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream is zero, or not specified, all variables in the output interface of the entry point being compiled decorated with Position, PointSize, ClipDistance, or CullDistance must be decorated with a Stream value of zero, or must not specify the Stream decoration", 383}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-rasterizerDiscardEnable-09024", 63}, { {"If the pipeline requires pre-rasterization shader state, and the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state is enabled or the rasterizerDiscardEnable member of pRasterizationState is VK_FALSE, and related dynamic state is not set, pViewportState must be a valid pointer to a valid VkPipelineViewportStateCreateInfo structure", 338}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06038", 50}, { {"If renderPass is not VK_NULL_HANDLE and the pipeline is being created with fragment shader state the fragment shader must not read from any input attachment that is defined as VK_ATTACHMENT_UNUSED in subpass", 207}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06041", 50}, { {"If renderPass is not VK_NULL_HANDLE, and the pipeline is being created with fragment output interface state, then for each color attachment in the subpass, if the potential format features of the format of the corresponding attachment description do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the blendEnable member of the corresponding element of the pAttachments member of pColorBlendState must be VK_FALSE", 428}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06046", 50}, { {"If renderPass is not VK_NULL_HANDLE, subpass must be a valid subpass within renderPass", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06047", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, subpass viewMask is not 0, and multiviewTessellationShader is not enabled, then pStages must not include tessellation shaders", 229}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06048", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, subpass viewMask is not 0, and multiviewGeometryShader is not enabled, then pStages must not include a geometry shader", 222}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06050", 50}, { {"If renderPass is not VK_NULL_HANDLE and the pipeline is being created with pre-rasterization shader state, and subpass viewMask is not 0, then all of the shaders in the pipeline must not include variables decorated with the Layer built-in decoration in their interfaces", 269}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06055", 50}, { {"If renderPass is VK_NULL_HANDLE, pColorBlendState is not dynamic, and the pipeline is being created with fragment output interface state, pColorBlendState->attachmentCount must be equal to VkPipelineRenderingCreateInfo::colorAttachmentCount", 240}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06057", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, VkPipelineRenderingCreateInfo::viewMask is not 0, and the multiviewTessellationShader feature is not enabled, then pStages must not include tessellation shaders", 260}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06058", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, VkPipelineRenderingCreateInfo::viewMask is not 0, and the multiviewGeometryShader feature is not enabled, then pStages must not include a geometry shader", 253}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06059", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, and VkPipelineRenderingCreateInfo::viewMask is not 0, all of the shaders in the pipeline must not include variables decorated with the Layer built-in decoration in their interfaces", 280}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06061", 50}, { {"If the dynamicRenderingLocalRead feature is not enabled, the pipeline requires fragment shader state, and renderPass is VK_NULL_HANDLE, fragment shaders in pStages must not include the InputAttachment capability", 211}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06062", 50}, { {"If the pipeline requires fragment output interface state and renderPass is VK_NULL_HANDLE, for each color attachment format defined by the pColorAttachmentFormats member of VkPipelineRenderingCreateInfo, if its potential format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the blendEnable member of the corresponding element of the pAttachments member of pColorBlendState must be VK_FALSE", 418}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06063", 50}, { {"If the pipeline requires fragment output interface state and renderPass is VK_NULL_HANDLE, if the pNext chain includes VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV, the colorAttachmentCount member of that structure must be equal to the value of VkPipelineRenderingCreateInfo::colorAttachmentCount", 315}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06578", 50}, { {"If the pipeline requires pre-rasterization shader state or fragment shader state, and renderPass is VK_NULL_HANDLE, the index of the most significant bit in VkPipelineRenderingCreateInfo::viewMask must be less than maxMultiviewViewCount", 236}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06579", 50}, { {"If the pipeline requires fragment output interface state, and renderPass is VK_NULL_HANDLE, and VkPipelineRenderingCreateInfo::colorAttachmentCount is not 0, VkPipelineRenderingCreateInfo::pColorAttachmentFormats must be a valid pointer to an array of colorAttachmentCount valid VkFormat values", 294}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06580", 50}, { {"If the pipeline requires fragment output interface state, and renderPass is VK_NULL_HANDLE, each element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats must be a valid VkFormat value", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06582", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and any element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats is not VK_FORMAT_UNDEFINED, that format must be a format with potential format features that include VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 355}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06583", 50}, { {"If the pipeline requires fragment output interface state, and renderPass is VK_NULL_HANDLE, VkPipelineRenderingCreateInfo::depthAttachmentFormat must be a valid VkFormat value", 175}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06584", 50}, { {"If the pipeline requires fragment output interface state, and renderPass is VK_NULL_HANDLE, VkPipelineRenderingCreateInfo::stencilAttachmentFormat must be a valid VkFormat value", 177}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06585", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkPipelineRenderingCreateInfo::depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 283}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06586", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkPipelineRenderingCreateInfo::stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format with potential format features that include VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 285}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06587", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkPipelineRenderingCreateInfo::depthAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a depth component", 224}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06588", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, and VkPipelineRenderingCreateInfo::stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, it must be a format that includes a stencil component", 228}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06589", 50}, { {"If the pipeline requires fragment output interface state, renderPass is VK_NULL_HANDLE, VkPipelineRenderingCreateInfo::depthAttachmentFormat is not VK_FORMAT_UNDEFINED, and VkPipelineRenderingCreateInfo::stencilAttachmentFormat is not VK_FORMAT_UNDEFINED, depthAttachmentFormat must equal stencilAttachmentFormat", 312}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06595", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state or fragment shader state, and VkMultiviewPerViewAttributesInfoNVX::perViewAttributesPositionXOnly is VK_TRUE then VkMultiviewPerViewAttributesInfoNVX::perViewAttributes must also be VK_TRUE", 288}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06603", 50}, { {"If the pipeline requires pre-rasterization shader state, fragment shader state, or fragment output state, and renderPass is not VK_NULL_HANDLE, renderPass must be a valid VkRenderPass handle", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-06854", 50}, { {"If renderPass is not VK_NULL_HANDLE, the multisampledRenderToSingleSampled feature is not enabled, rasterizationSamples is not dynamic, and subpass has a VkMultisampledRenderToSingleSampledInfoEXT structure included in the VkSubpassDescription2::pNext chain with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples member of pMultisampleState must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 460}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-07064", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, subpass viewMask is not 0, and multiviewMeshShader is not enabled, then pStages must not include a mesh shader", 214}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-07609", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with fragment output interface state, the pColorBlendState pointer is not NULL, the attachmentCount member of pColorBlendState is not ignored, and the subpass uses color attachments, the attachmentCount member of pColorBlendState must be equal to the colorAttachmentCount used to create subpass", 361}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-07720", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, and VkPipelineRenderingCreateInfo::viewMask is not 0, and multiviewMeshShader is not enabled, then pStages must not include a mesh shader", 237}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-08710", 50}, { {"If the pipeline requires fragment shader state and renderPass is not VK_NULL_HANDLE, fragment shaders in pStages must not include any of the TileImageColorReadAccessEXT, TileImageDepthReadAccessEXT, or TileImageStencilReadAccessEXT capabilities", 244}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-08744", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline requires fragment output state or fragment shader state, the pipeline enables sample shading, rasterizationSamples is not dynamic, and the pNext chain includes a VkPipelineRenderingCreateInfo structure, rasterizationSamples must be a valid VkSampleCountFlagBits value that is set in imageCreateSampleCounts (as defined in Image Creation Limits) for every element of depthAttachmentFormat, stencilAttachmentFormat and the pColorAttachmentFormats array which is not VK_FORMAT_UNDEFINED", 529}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09028", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with fragment shader state, and subpass uses a depth/stencil attachment, and related dynamic state is not set, pDepthStencilState must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 279}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09030", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with fragment output interface state, and subpass uses color attachments, and related dynamic state is not set, pColorBlendState must be a valid pointer to a valid VkPipelineColorBlendStateCreateInfo structure", 276}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09033", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with fragment shader state and fragment output interface state, and either of VkPipelineRenderingCreateInfo::depthAttachmentFormat or VkPipelineRenderingCreateInfo::stencilAttachmentFormat are not VK_FORMAT_UNDEFINED, and the VK_EXT_extended_dynamic_state3 extension is not enabled or any of the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE, VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE, VK_DYNAMIC_STATE_DEPTH_COMPARE_OP, VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE, VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE, VK_DYNAMIC_STATE_STENCIL_OP, or VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic states are not set, pDepthStencilState must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 739}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09035", 50}, { {"If renderPass is VK_NULL_HANDLE and the pipeline is being created with fragment shader state but not fragment output interface state, and the VK_EXT_extended_dynamic_state3 extension is not enabled, or any of the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE, VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE, VK_DYNAMIC_STATE_DEPTH_COMPARE_OP, VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE, VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE, VK_DYNAMIC_STATE_STENCIL_OP, or VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic states are not set, pDepthStencilState must be a valid pointer to a valid VkPipelineDepthStencilStateCreateInfo structure", 593}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09037", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with fragment output interface state, and any element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats is not VK_FORMAT_UNDEFINED, and the VK_EXT_extended_dynamic_state3 extension is not enabled, or any of the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT, VK_DYNAMIC_STATE_LOGIC_OP_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT, VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, or VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic states are not set, pColorBlendState must be a valid pointer to a valid VkPipelineColorBlendStateCreateInfo structure", 635}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09527", 50}, { {"If the pipeline requires fragment output interface state, renderPass is not VK_NULL_HANDLE, and the flags member of VkPipelineColorBlendStateCreateInfo includes VK_PIPELINE_COLOR_BLEND_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_BIT_EXT subpass must have been created with VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_COLOR_ACCESS_BIT_EXT", 354}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09528", 50}, { {"If the pipeline requires fragment shader state, renderPass is not VK_NULL_HANDLE, and the flags member of VkPipelineDepthStencilStateCreateInfo includes VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT, subpass must have been created with VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT", 355}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09529", 50}, { {"If the pipeline requires fragment shader state, renderPass is not VK_NULL_HANDLE, and the flags member of VkPipelineDepthStencilStateCreateInfo includes VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT, subpass must have been created with VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT", 359}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09531", 50}, { {"If the pipeline is being created with fragment shader state and fragment output state, the value of renderPass is VK_NULL_HANDLE, and VkRenderingInputAttachmentIndexInfo is included, VkRenderingInputAttachmentIndexInfo::colorAttachmentCount must be equal to VkPipelineRenderingCreateInfo::colorAttachmentCount", 309}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09532", 50}, { {"If the pipeline is being created with fragment output state, and the value of renderPass is VK_NULL_HANDLE, VkRenderingAttachmentLocationInfo::colorAttachmentCount must be equal to VkPipelineRenderingCreateInfo::colorAttachmentCount", 232}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-09652", 50}, { {"If the pipeline is being created with fragment shader state and fragment output state, the value of renderPass is VK_NULL_HANDLE, and VkRenderingInputAttachmentIndexInfo is not included, the fragment shader must not contain any input attachments with a InputAttachmentIndex greater than or equal to VkPipelineRenderingCreateInfo::colorAttachmentCount", 350}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-11504", 50}, { {"If renderPass is VK_NULL_HANDLE, a VkCustomResolveCreateInfoEXT is in the pNext chain, and the pipeline is being created with fragment output interface state, VkPipelineRenderingCreateInfo::colorAttachmentCount must be equal to VkCustomResolveCreateInfoEXT::colorAttachmentCount", 278}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-12325", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, pStages include a mesh shader, and subpass viewMask is not 0, then the index of the most significant bit in viewMask must be less than maxMeshMultiviewViewCount", 264}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderPass-12326", 50}, { {"If renderPass is VK_NULL_HANDLE, the pipeline is being created with pre-rasterization shader state, pStages include a mesh shader, and VkPipelineRenderingCreateInfo::viewMask is not 0, then the index of the most significant bit in viewMask must be less than maxMeshMultiviewViewCount", 283}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderpass-06624", 50}, { {"If renderpass is not VK_NULL_HANDLE, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes at least one of and no more than two of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes one of the other flags, renderPass must be compatible with that used to create the library", 477}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderpass-06625", 50}, { {"If renderpass is VK_NULL_HANDLE, VkGraphicsPipelineLibraryCreateInfoEXT::flags includes at least one of and no more than two of VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT, VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT, or VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT, and an element of VkPipelineLibraryCreateInfoKHR::pLibraries includes one of the other flags, the value of renderPass used to create that library must also be VK_NULL_HANDLE", 486}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-renderpass-06631", 50}, { {"If renderPass is not VK_NULL_HANDLE, the pipeline requires fragment shader state, and the VK_EXT_extended_dynamic_state3 extension is not enabled or any of the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT, VK_DYNAMIC_STATE_SAMPLE_MASK_EXT, or VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic states is not set, or the alphaToOne feature is enabled and VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT is not set, then pMultisampleState must be a valid pointer to a valid VkPipelineMultisampleStateCreateInfo structure", 512}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-sType-unique", 46}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-sampleLocationsEnable-01524", 61}, { {"If the pipeline requires fragment shader state, and the sampleLocationsEnable member of a VkPipelineSampleLocationsStateCreateInfoEXT structure included in the pNext chain of pMultisampleState is VK_TRUE, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 298}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-sampleShadingEnable-04900", 59}, { {"If the pipeline requires fragment shader state, and the VK_QCOM_render_pass_shader_resolve extension or the customResolve feature is enabled, and if the subpass description contains VK_SUBPASS_DESCRIPTION_FRAGMENT_REGION_BIT_EXT, then sampleShadingEnable must be false", 268}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-shaderMeshEnqueue-10187", 57}, { {"If the shaderMeshEnqueue feature is not enabled, shaders specified by pStages must not declare the ShaderEnqueueAMDX capability", 127}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-shaderTessellationAndGeometryPointSize-08776", 78}, { {"If the pipeline is being created with a Geometry Execution Model, uses the OutputPoints Execution Mode, and the shaderTessellationAndGeometryPointSize feature is enabled, a PointSize decorated variable must be written to for every vertex emitted if the maintenance5 feature is not enabled", 288}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-stage-02096", 45}, { {"If the pipeline requires pre-rasterization shader state the stage member of one element of pStages must be VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT", 165}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-stage-06897", 45}, { {"If the pipeline requires fragment shader state and/or pre-rasterization shader state, any value of stage must not be set in more than one element of pStages", 156}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-stageCount-09530", 50}, { {"If the pipeline requires pre-rasterization shader state, stageCount must be greater than 0", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-stageCount-09587", 50}, { {"If the pipeline does not require pre-rasterization shader state or fragment shader state, stageCount must be zero", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-stippledLineEnable-02767", 58}, { {"If the pipeline requires pre-rasterization shader state, the stippledLineEnable member of VkPipelineRasterizationLineStateCreateInfo is VK_TRUE, and no element of the pDynamicStates member of pDynamicState is VK_DYNAMIC_STATE_LINE_STIPPLE, then the lineStippleFactor member of VkPipelineRasterizationLineStateCreateInfo must be in the range [1,256]", 348}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-subpass-00758", 47}, { {"If the pipeline requires fragment output interface state, rasterizationSamples is not dynamic, and subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples member of pMultisampleState must follow the rules for a zero-attachment subpass", 273}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-subpass-01411", 47}, { {"If the pipeline requires fragment output interface state, the VK_NV_framebuffer_mixed_samples extension is enabled, rasterizationSamples is not dynamic, and if subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled, then the rasterizationSamples member of pMultisampleState must be the same as the sample count of the depth/stencil attachment", 389}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-subpass-01412", 47}, { {"If the pipeline requires fragment output interface state, the VK_NV_framebuffer_mixed_samples extension is enabled, rasterizationSamples is not dynamic, and if subpass has any color attachments, then the rasterizationSamples member of pMultisampleState must be greater than or equal to the sample count for those subpass attachments", 332}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-subpass-01505", 47}, { {"If the pipeline requires fragment output interface state, and the VK_AMD_mixed_attachment_samples extension is enabled, rasterizationSamples is not dynamic, and if subpass uses color and/or depth/stencil attachments, then the rasterizationSamples member of pMultisampleState must equal the maximum of the sample counts of those subpass attachments", 347}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-topology-08773", 48}, { {"If the pipeline is being created with a Vertex Execution Model and no TessellationEvaluation or Geometry Execution Model, and the topology member of pInputAssembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, and either VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state is not enabled or dynamicPrimitiveTopologyUnrestricted is VK_FALSE, a PointSize decorated variable must be written to if the maintenance5 feature is not enabled", 422}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineCreateInfo-topology-08889", 48}, { {"If the pipeline is being created with pre-rasterization shader state and vertex input state and the topology member of pInputAssembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, and either VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state is not enabled or dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then pStages must include tessellation shader stages", 352}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineLibraryCreateInfoEXT-flags-parameter", 59}, { {"flags must be a valid combination of VkGraphicsPipelineLibraryFlagBitsEXT values", 80}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineLibraryCreateInfoEXT-flags-requiredbitmask", 65}, { {"flags must not be 0", 19}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineLibraryCreateInfoEXT-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_LIBRARY_CREATE_INFO_EXT", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-deviceGeneratedCommands-02887", 77}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-groupCount-02879", 64}, { {"groupCount must be at least 1 and as maximum VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxGraphicsShaderGroupCount", 125}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-groupCount-02880", 64}, { {"The sum of groupCount including those groups added from referenced pPipelines must also be as maximum VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxGraphicsShaderGroupCount", 182}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pGroups-02881", 61}, { {"The state of the first element of pGroups must match its equivalent within the parent's VkGraphicsPipelineCreateInfo", 116}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pGroups-02882", 61}, { {"Each element of pGroups must in combination with the rest of the pipeline state yield a valid state configuration", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pGroups-02884", 61}, { {"All elements of pGroups must use the same shader stage combinations unless any mesh shader stage is used, then either combination of task and mesh or just mesh shader is valid", 175}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pGroups-02885", 61}, { {"Mesh and regular primitive shading stages cannot be mixed across pGroups", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pGroups-parameter", 65}, { {"If groupCount is not 0, pGroups must be a valid pointer to an array of groupCount valid VkGraphicsShaderGroupCreateInfoNV structures", 132}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pPipelines-02886", 64}, { {"Each element of pPipelines must have been created with identical state to the pipeline currently created except the state that can be overridden by VkGraphicsShaderGroupCreateInfoNV", 181}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-pPipelines-parameter", 68}, { {"If pipelineCount is not 0, pPipelines must be a valid pointer to an array of pipelineCount valid VkPipeline handles", 115}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsPipelineShaderGroupsCreateInfoNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_SHADER_GROUPS_CREATE_INFO_NV", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-pStages-02889", 52}, { {"For pStages, the same restrictions as in VkGraphicsPipelineCreateInfo::pStages apply", 84}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-pStages-parameter", 56}, { {"pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-pTessellationState-02891", 63}, { {"For pTessellationState, the same restrictions as in VkGraphicsPipelineCreateInfo::pTessellationState apply", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-pVertexInputState-02890", 62}, { {"For pVertexInputState, the same restrictions as in VkGraphicsPipelineCreateInfo::pVertexInputState apply", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_GRAPHICS_SHADER_GROUP_CREATE_INFO_NV", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-stageCount-02888", 55}, { {"For stageCount, the same restrictions as in VkGraphicsPipelineCreateInfo::stageCount apply", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkGraphicsShaderGroupCreateInfoNV-stageCount-arraylength", 61}, { {"stageCount must be greater than 0", 33}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkHdrMetadataEXT-pNext-pNext", 33}, { {"pNext must be NULL or a pointer to a valid instance of VkHdrVividDynamicMetadataHUAWEI", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHdrMetadataEXT-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_HDR_METADATA_EXT", 48}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHdrMetadataEXT-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHdrVividDynamicMetadataHUAWEI-dynamicMetadataSize-arraylength", 68}, { {"dynamicMetadataSize must be greater than 0", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHdrVividDynamicMetadataHUAWEI-pDynamicMetadata-parameter", 63}, { {"pDynamicMetadata must be a valid pointer to an array of dynamicMetadataSize bytes", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHdrVividDynamicMetadataHUAWEI-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_HDR_VIVID_DYNAMIC_METADATA_HUAWEI", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHeadlessSurfaceCreateInfoEXT-flags-zerobitmask", 53}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHeadlessSurfaceCreateInfoEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHeadlessSurfaceCreateInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkHostAddressRangeConstEXT-address-parameter", 49}, { {"address must be a valid pointer to an array of size bytes", 57}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkHostAddressRangeConstEXT-size-arraylength", 48}, { {"size must be greater than 0", 27}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkHostAddressRangeEXT-address-parameter", 44}, { {"address must be a valid pointer to an array of size bytes", 57}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkHostAddressRangeEXT-size-arraylength", 43}, { {"size must be greater than 0", 27}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkHostImageCopyDevicePerformanceQuery-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_HOST_IMAGE_COPY_DEVICE_PERFORMANCE_QUERY", 72}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-aspectMask-08702", 53}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_DEPTH_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-aspectMask-08703", 53}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_STENCIL_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-attachmentFeedbackLoopLayout-07313", 71}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 134}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-dynamicRenderingLocalRead-09551", 68}, { {"If the dynamicRenderingLocalRead feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-dynamicRenderingLocalRead-09552", 68}, { {"If the dynamicRenderingLocalRead feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-01672", 48}, { {"If image has a multi-planar format and the image is disjoint, then the aspectMask member of subresourceRange must include at least one multi-planar aspect mask bit or VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-01932", 48}, { {"If image is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 146}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-03319", 48}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is enabled, then the aspectMask member of subresourceRange must include either or both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 252}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-03320", 48}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is not enabled, then the aspectMask member of subresourceRange must include both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 246}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-09055", 48}, { {"image must have been created with the VK_IMAGE_USAGE_HOST_TRANSFER_BIT usage flag set", 85}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-09241", 48}, { {"If image has a color format that is single-plane, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-09242", 48}, { {"If image has a color format and is not disjoint, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-10749", 48}, { {"If image has a depth-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_DEPTH_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-10750", 48}, { {"If image has a stencil-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_STENCIL_BIT", 117}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-image-parameter", 52}, { {"image must be a valid VkImage handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-maintenance9-10798", 55}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer must be less than the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 338}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-maintenance9-10799", 55}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set and either subresourceRange.baseArrayLayer is not equal to 0 or subresourceRange.layerCount is not equal to VK_REMAINING_ARRAY_LAYERS, subresourceRange.levelCount must be 1", 292}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-maintenance9-10800", 55}, { {"If the maintenance9 feature is enabled, subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 443}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-newLayout-09057", 52}, { {"newLayout must be one of the layouts in VkPhysicalDeviceHostImageCopyProperties::pCopyDstLayouts", 96}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-newLayout-parameter", 56}, { {"newLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01208", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01209", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 179}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01210", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 178}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01211", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 193}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01212", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 155}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01213", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 155}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01658", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-01659", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-02088", 52}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR then image must have been created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set", 203}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-09229", 52}, { {"oldLayout must be either VK_IMAGE_LAYOUT_UNDEFINED or the current layout of the image subresources as specified in subresourceRange", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-09230", 52}, { {"If oldLayout is not VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, or VK_IMAGE_LAYOUT_PREINITIALIZED, it must be one of the layouts in VkPhysicalDeviceHostImageCopyProperties::pCopySrcLayouts", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-10767", 52}, { {"If the zeroInitializeDeviceMemory feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT", 116}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-10768", 52}, { {"If oldLayout is VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, then all subresources must be included in the barrier", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-oldLayout-parameter", 56}, { {"oldLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_HOST_IMAGE_LAYOUT_TRANSITION_INFO", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-03938", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL, image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 200}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-03939", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL, image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 245}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-04065", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-04066", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 171}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-04067", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 257}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-04068", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 173}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07006", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then image must have been created with either the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set, and the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT usage flags set, and the VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT usage flag set", 393}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07120", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07121", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07122", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07123", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07124", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-07125", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR usage flag set", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-09550", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ then image must have been created with either the VK_IMAGE_USAGE_STORAGE_BIT usage flag set, or with both the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag and either of the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set", 339}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-srcQueueFamilyIndex-10287", 62}, { {"If oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_QUANTIZATION_MAP_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR usage flags set", 248}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-01486", 59}, { {"subresourceRange.baseMipLevel must be less than the mipLevels specified in VkImageCreateInfo when image was created", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-01488", 59}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, subresourceRange.baseArrayLayer must be less than the arrayLayers specified in VkImageCreateInfo when image was created", 258}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-01724", 59}, { {"If subresourceRange.levelCount is not VK_REMAINING_MIP_LEVELS, subresourceRange.baseMipLevel + subresourceRange.levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-01725", 59}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 366}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-09601", 59}, { {"subresourceRange.aspectMask must be valid for the format the image was created with", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-subresourceRange-parameter", 63}, { {"subresourceRange must be a valid VkImageSubresourceRange structure", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-synchronization2-07793", 59}, { {"If the synchronization2 feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkHostImageLayoutTransitionInfo-synchronization2-07794", 59}, { {"If the synchronization2 feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkIOSSurfaceCreateInfoMVK-flags-zerobitmask", 48}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkIOSSurfaceCreateInfoMVK-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkIOSSurfaceCreateInfoMVK-pView-01316", 42}, { {"If pView is a UIView object, it must be a valid UIView, must be backed by a CALayer object of type CAMetalLayer, and vkCreateIOSSurfaceMVK must be called on the main thread", 172}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkIOSSurfaceCreateInfoMVK-pView-04143", 42}, { {"If pView is a CAMetalLayer object, it must be a valid CAMetalLayer", 66}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkIOSSurfaceCreateInfoMVK-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_IOS_SURFACE_CREATE_INFO_MVK", 59}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkImageAlignmentControlCreateInfoMESA-imageAlignmentControl-09657", 70}, { {"imageAlignmentControl must be enabled", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageAlignmentControlCreateInfoMESA-maximumRequestedAlignment-09655", 74}, { {"If maximumRequestedAlignment is not 0, maximumRequestedAlignment must be a power of two", 87}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageAlignmentControlCreateInfoMESA-maximumRequestedAlignment-09656", 74}, { {"If maximumRequestedAlignment is not 0, the bitwise-and of maximumRequestedAlignment and supportedImageAlignmentMask must be non-zero", 132}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageAlignmentControlCreateInfoMESA-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_ALIGNMENT_CONTROL_CREATE_INFO_MESA", 72}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageBlit-aspectMask-00238", 33}, { {"The aspectMask member of srcSubresource and dstSubresource must match", 69}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit-dstSubresource-parameter", 41}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit-layerCount-08800", 33}, { {"If neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount members of srcSubresource or dstSubresource must match", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit-layerCount-08801", 33}, { {"If one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 262}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit-srcSubresource-parameter", 41}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-aspectMask-00238", 34}, { {"The aspectMask member of srcSubresource and dstSubresource must match", 69}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-dstSubresource-parameter", 42}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-layerCount-08800", 34}, { {"If neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount members of srcSubresource or dstSubresource must match", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-layerCount-08801", 34}, { {"If one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 262}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-pNext-pNext", 29}, { {"pNext must be NULL or a pointer to a valid instance of VkCopyCommandTransformInfoQCOM", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-sType-sType", 29}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_BLIT_2", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-sType-unique", 30}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageBlit2-srcSubresource-parameter", 42}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCaptureDescriptorDataInfoEXT-image-08079", 52}, { {"image must have been created with VK_IMAGE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkImageCreateInfo::flags", 122}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageCaptureDescriptorDataInfoEXT-image-parameter", 56}, { {"image must be a valid VkImage handle", 36}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageCaptureDescriptorDataInfoEXT-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageCaptureDescriptorDataInfoEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT", 70}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageCompressionControlEXT-flags-06747", 45}, { {"flags must be one of VK_IMAGE_COMPRESSION_DEFAULT_EXT, VK_IMAGE_COMPRESSION_FIXED_RATE_DEFAULT_EXT, VK_IMAGE_COMPRESSION_FIXED_RATE_EXPLICIT_EXT, or VK_IMAGE_COMPRESSION_DISABLED_EXT", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCompressionControlEXT-flags-06748", 45}, { {"If flags includes VK_IMAGE_COMPRESSION_FIXED_RATE_EXPLICIT_EXT, pFixedRateFlags must not be NULL", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCompressionControlEXT-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_CONTROL_EXT", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCompressionPropertiesEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_PROPERTIES_EXT", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-attachmentFragmentShadingRate-06401", 70}, { {"If the attachmentFragmentShadingRate feature is enabled, and pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR, then pFormatConstraints->requiredFormatFeatures must contain VK_FORMAT_FEATURE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 289}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-bufferCollectionConstraints-parameter", 72}, { {"bufferCollectionConstraints must be a valid VkBufferCollectionConstraintsInfoFUCHSIA structure", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-flags-parameter", 50}, { {"flags must be a valid combination of VkImageConstraintsInfoFlagBitsFUCHSIA values", 81}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-formatConstraintsCount-arraylength", 69}, { {"formatConstraintsCount must be greater than 0", 45}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06395", 59}, { {"All elements of pFormatConstraints must have at least one bit set in its VkImageFormatConstraintsInfoFUCHSIA::requiredFormatFeatures", 132}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06396", 59}, { {"If pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_SAMPLED_BIT, then pFormatConstraints->requiredFormatFeatures must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT", 179}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06397", 59}, { {"If pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_STORAGE_BIT, then pFormatConstraints->requiredFormatFeatures must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT", 179}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06398", 59}, { {"If pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, then pFormatConstraints->requiredFormatFeatures must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT", 191}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06399", 59}, { {"If pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, then pFormatConstraints->requiredFormatFeatures must contain VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 207}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-06400", 59}, { {"If pFormatConstraints->imageCreateInfo->usage contains VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, then pFormatConstraints->requiredFormatFeatures must contain at least one of VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 257}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pFormatConstraints-parameter", 63}, { {"pFormatConstraints must be a valid pointer to an array of formatConstraintsCount valid VkImageFormatConstraintsInfoFUCHSIA structures", 133}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageConstraintsInfoFUCHSIA-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_CONSTRAINTS_INFO_FUCHSIA", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCopy-apiVersion-07940", 33}, { {"If the VK_KHR_sampler_ycbcr_conversion extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, the aspectMask member of srcSubresource and dstSubresource must match", 202}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-apiVersion-07941", 33}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, the layerCount member of srcSubresource and dstSubresource must match", 190}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-dstSubresource-parameter", 41}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-extent-06668", 29}, { {"extent.width must not be 0", 26}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-extent-06669", 29}, { {"extent.height must not be 0", 27}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-extent-06670", 29}, { {"extent.depth must not be 0", 26}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy-srcSubresource-parameter", 41}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-apiVersion-07940", 34}, { {"If the VK_KHR_sampler_ycbcr_conversion extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, the aspectMask member of srcSubresource and dstSubresource must match", 202}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-apiVersion-07941", 34}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, the layerCount member of srcSubresource and dstSubresource must match", 190}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-dstSubresource-parameter", 42}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-extent-06668", 30}, { {"extent.width must not be 0", 26}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-extent-06669", 30}, { {"extent.height must not be 0", 27}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-extent-06670", 30}, { {"extent.depth must not be 0", 26}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-pNext-pNext", 29}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-sType-sType", 29}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_COPY_2", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCopy2-srcSubresource-parameter", 42}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageCreateFlags2CreateInfoKHR-flags-parameter", 53}, { {"flags must be a valid combination of VkImageCreateFlagBits2KHR values", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateFlags2CreateInfoKHR-flags-requiredbitmask", 59}, { {"flags must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateFlags2CreateInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_CREATE_FLAGS_2_CREATE_INFO_KHR", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-Format-02536", 35}, { {"If Format is a depth-stencil format and the pNext chain includes a VkImageStencilUsageCreateInfo structure with its stencilUsage member including VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, extent.width must be less than or equal to VkPhysicalDeviceLimits::maxFramebufferWidth", 269}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-None-01891", 33}, { {"If any of the bits VK_IMAGE_CREATE_SPARSE_BINDING_BIT, VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT, or VK_IMAGE_CREATE_SPARSE_ALIASED_BIT are set, VK_IMAGE_CREATE_PROTECTED_BIT must not also be set", 190}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-None-01925", 33}, { {"If any of the bits VK_IMAGE_CREATE_SPARSE_BINDING_BIT, VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT, or VK_IMAGE_CREATE_SPARSE_ALIASED_BIT are set, VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT must not also be set", 200}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-None-12279", 33}, { {"If Vulkan 1.3 is not supported and the ycbcr2plane444Formats feature is not enabled, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 286}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-arrayLayers-00948", 40}, { {"arrayLayers must be greater than 0", 34}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-arrayLayers-02256", 40}, { {"arrayLayers must be less than or equal to imageCreateMaxArrayLayers (as defined in Image Creation Limits)", 105}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-00944", 35}, { {"extent.width must be greater than 0", 35}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-00945", 35}, { {"extent.height must be greater than 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-00946", 35}, { {"extent.depth must be greater than 0", 35}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-02252", 35}, { {"extent.width must be less than or equal to imageCreateMaxExtent.width (as defined in Image Creation Limits)", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-02253", 35}, { {"extent.height must be less than or equal to imageCreateMaxExtent.height (as defined in Image Creation Limits)", 109}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-extent-02254", 35}, { {"extent.depth must be less than or equal to imageCreateMaxExtent.depth (as defined in Image Creation Limits)", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-00949", 34}, { {"If flags contains VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT, imageType must be VK_IMAGE_TYPE_2D", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-00950", 34}, { {"If flags contains VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT, imageType must be VK_IMAGE_TYPE_3D", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-00969", 34}, { {"If the sparseBinding feature is not enabled, flags must not contain VK_IMAGE_CREATE_SPARSE_BINDING_BIT", 102}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-00987", 34}, { {"If flags contains VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT or VK_IMAGE_CREATE_SPARSE_ALIASED_BIT, it must also contain VK_IMAGE_CREATE_SPARSE_BINDING_BIT", 149}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-01533", 34}, { {"If flags contains VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT format must be a depth or depth/stencil format", 122}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-01572", 34}, { {"If flags contains VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT, then format must be a compressed image format", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-01573", 34}, { {"If flags contains VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT, then flags must also contain VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT", 130}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-01890", 34}, { {"If the protectedMemory feature is not enabled, flags must not contain VK_IMAGE_CREATE_PROTECTED_BIT", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-01924", 34}, { {"If the sparseResidencyAliased feature is not enabled, flags must not contain VK_IMAGE_CREATE_SPARSE_ALIASED_BIT", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02050", 34}, { {"If flags contains VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV, imageType must be VK_IMAGE_TYPE_2D or VK_IMAGE_TYPE_3D", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02051", 34}, { {"If flags contains VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV, it must not contain VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT and the format must not be a depth/stencil format", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02052", 34}, { {"If flags contains VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV and imageType is VK_IMAGE_TYPE_2D, extent.width and extent.height must be greater than 1", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02053", 34}, { {"If flags contains VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV and imageType is VK_IMAGE_TYPE_3D, extent.width, extent.height, and extent.depth must be greater than 1", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02259", 34}, { {"If flags contains VK_IMAGE_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT, then mipLevels must be one, arrayLayers must be one, imageType must be VK_IMAGE_TYPE_2D, and imageCreateMaybeLinear (as defined in Image Creation Limits) must be VK_FALSE", 237}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02557", 34}, { {"If flags contains VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT, imageType must be VK_IMAGE_TYPE_2D", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02565", 34}, { {"If flags contains VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT, tiling must be VK_IMAGE_TILING_OPTIMAL", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02566", 34}, { {"If flags contains VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT, imageType must be VK_IMAGE_TYPE_2D", 88}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02567", 34}, { {"If flags contains VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT, flags must not contain VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-02568", 34}, { {"If flags contains VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT, mipLevels must be 1", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-04738", 34}, { {"If flags does not contain VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT and the pNext chain includes a VkImageFormatListCreateInfo structure, then VkImageFormatListCreateInfo::viewFormatCount must be 0 or 1", 195}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-06883", 34}, { {"If flags contains VK_IMAGE_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT, samples must be VK_SAMPLE_COUNT_1_BIT", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-07755", 34}, { {"If flags contains VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT, imageType must be VK_IMAGE_TYPE_3D", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08104", 34}, { {"If flags includes VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay or descriptorHeapCaptureReplay feature must be enabled", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08328", 34}, { {"If flags includes VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then videoMaintenance1 must be enabled", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08329", 34}, { {"If flags includes VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR and usage does not include VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR, then usage must not include VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR", 203}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08331", 34}, { {"If flags includes VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then usage must not include VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR, VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, or VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08865", 34}, { {"If flags contains VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT, extent.width and extent.height must be equal", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-08866", 34}, { {"If flags contains VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT, arrayLayers must be greater than or equal to 6", 101}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-09403", 34}, { {"If flags contains VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT, flags must not include VK_IMAGE_CREATE_SPARSE_ALIASED_BIT, VK_IMAGE_CREATE_SPARSE_BINDING_BIT, or VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 193}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-11281", 34}, { {"If VkOpaqueCaptureDataCreateInfoEXT::pData is not NULL, flags must contain VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT", 129}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-13355", 34}, { {"If flags contains VK_IMAGE_CREATE_ALIAS_SINGLE_LAYER_DESCRIPTOR_BIT_KHR, the maintenance11 feature must be enabled", 114}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-13356", 34}, { {"If flags contains VK_IMAGE_CREATE_ALIAS_SINGLE_LAYER_DESCRIPTOR_BIT_KHR, imageType must be VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D", 127}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-flags-parameter", 38}, { {"flags must be a valid combination of VkImageCreateFlagBits values", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-01577", 35}, { {"If format is not a multi-planar format, and flags does not include VK_IMAGE_CREATE_ALIAS_BIT, flags must not contain VK_IMAGE_CREATE_DISJOINT_BIT", 145}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02537", 35}, { {"If format is a depth-stencil format and the pNext chain includes a VkImageStencilUsageCreateInfo structure with its stencilUsage member including VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, extent.height must be less than or equal to VkPhysicalDeviceLimits::maxFramebufferHeight", 271}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02538", 35}, { {"If the shaderStorageImageMultisample feature is not enabled, format is a depth-stencil format and the pNext chain includes a VkImageStencilUsageCreateInfo structure with its stencilUsage including VK_IMAGE_USAGE_STORAGE_BIT, samples must be VK_SAMPLE_COUNT_1_BIT", 262}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02795", 35}, { {"If format is a depth-stencil format, usage includes VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, and the pNext chain includes a VkImageStencilUsageCreateInfo structure, then its VkImageStencilUsageCreateInfo::stencilUsage member must also include VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT", 290}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02796", 35}, { {"If format is a depth-stencil format, usage does not include VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, and the pNext chain includes a VkImageStencilUsageCreateInfo structure, then its VkImageStencilUsageCreateInfo::stencilUsage member must also not include VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT", 302}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02797", 35}, { {"If format is a depth-stencil format, usage includes VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, and the pNext chain includes a VkImageStencilUsageCreateInfo structure, then its VkImageStencilUsageCreateInfo::stencilUsage member must also include VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT", 282}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-02798", 35}, { {"If format is a depth-stencil format, usage does not include VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, and the pNext chain includes a VkImageStencilUsageCreateInfo structure, then its VkImageStencilUsageCreateInfo::stencilUsage member must also not include VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT", 294}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-04712", 35}, { {"If format has a _422 or _420 suffix, extent.width must be a multiple of 2", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-04713", 35}, { {"If format has a _420 suffix, extent.height must be a multiple of 2", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-06410", 35}, { {"If the image format is one of the formats that require a sampler Y'CBCR conversion, mipLevels must be 1", 103}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-06411", 35}, { {"If the image format is one of the formats that require a sampler Y'CBCR conversion, samples must be VK_SAMPLE_COUNT_1_BIT", 121}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-06412", 35}, { {"If the image format is one of the formats that require a sampler Y'CBCR conversion, imageType must be VK_IMAGE_TYPE_2D", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-09583", 35}, { {"If format is one of the VK_FORMAT_PVRTC1_*_IMG formats, extent.width must be a power of 2", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-09584", 35}, { {"If format is one of the VK_FORMAT_PVRTC1_*_IMG formats, extent.height must be a power of 2", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-format-parameter", 39}, { {"format must be a valid VkFormat value", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-fragmentDensityMapOffset-06514", 53}, { {"If the fragmentDensityMapOffset feature is not enabled and usage includes VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT, extent.width must be less than or equal to the ceiling of maxFramebufferWidth/minFragmentDensityTexelSize.width", 230}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-fragmentDensityMapOffset-06515", 53}, { {"If the fragmentDensityMapOffset feature is not enabled and usage includes VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT, extent.height must be less than or equal to the ceiling of maxFramebufferHeight/minFragmentDensityTexelSize.height", 233}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageCreateFormatFeatures-02260", 54}, { {"If format is a multi-planar format, and if imageCreateFormatFeatures (as defined in Image Creation Limits) does not contain VK_FORMAT_FEATURE_DISJOINT_BIT, then flags must not contain VK_IMAGE_CREATE_DISJOINT_BIT", 212}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageCreateFormatFeatures-09048", 54}, { {"If imageCreateFormatFeatures (as defined in Image Creation Limits) does not contain VK_FORMAT_FEATURE_2_HOST_IMAGE_TRANSFER_BIT, then usage must not contain VK_IMAGE_USAGE_HOST_TRANSFER_BIT", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageCreateMaxMipLevels-02251", 52}, { {"Each of the following values (as described in Image Creation Limits) must not be undefined : imageCreateMaxMipLevels, imageCreateMaxArrayLayers, imageCreateMaxExtent, and imageCreateSampleCounts", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00956", 38}, { {"If imageType is VK_IMAGE_TYPE_1D, both extent.height and extent.depth must be 1", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00957", 38}, { {"If imageType is VK_IMAGE_TYPE_2D, extent.depth must be 1", 56}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00961", 38}, { {"If imageType is VK_IMAGE_TYPE_3D, arrayLayers must be 1", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00970", 38}, { {"If imageType is VK_IMAGE_TYPE_1D, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00971", 38}, { {"If the sparseResidencyImage2D feature is not enabled, and imageType is VK_IMAGE_TYPE_2D, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 148}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00972", 38}, { {"If the sparseResidencyImage3D feature is not enabled, and imageType is VK_IMAGE_TYPE_3D, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 148}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00973", 38}, { {"If the sparseResidency2Samples feature is not enabled, imageType is VK_IMAGE_TYPE_2D, and samples is VK_SAMPLE_COUNT_2_BIT, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 183}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00974", 38}, { {"If the sparseResidency4Samples feature is not enabled, imageType is VK_IMAGE_TYPE_2D, and samples is VK_SAMPLE_COUNT_4_BIT, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 183}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00975", 38}, { {"If the sparseResidency8Samples feature is not enabled, imageType is VK_IMAGE_TYPE_2D, and samples is VK_SAMPLE_COUNT_8_BIT, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 183}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-00976", 38}, { {"If the sparseResidency16Samples feature is not enabled, imageType is VK_IMAGE_TYPE_2D, and samples is VK_SAMPLE_COUNT_16_BIT, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 185}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-02082", 38}, { {"If usage includes VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR, imageType must be VK_IMAGE_TYPE_2D", 109}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-10197", 38}, { {"If flags contains VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT and either the maintenance9 feature is not enabled on the device or image2DViewOf3DSparse is VK_FALSE , flags must not include VK_IMAGE_CREATE_SPARSE_ALIASED_BIT, VK_IMAGE_CREATE_SPARSE_BINDING_BIT, or VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 299}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageType-parameter", 42}, { {"imageType must be a valid VkImageType value", 43}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-imageView2DOn3DImage-04459", 49}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::imageView2DOn3DImage is VK_FALSE, flags must not contain VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT", 201}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-initialLayout-00993", 42}, { {"initialLayout must be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 121}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-initialLayout-10765", 42}, { {"If the zeroInitializeDeviceMemory feature is not enabled, initialLayout must not be VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT", 120}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-initialLayout-parameter", 46}, { {"initialLayout must be a valid VkImageLayout value", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-maintenance11-13354", 42}, { {"If the maintenance11 feature is enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 0", 136}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-mipLevels-00947", 38}, { {"mipLevels must be greater than 0", 32}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-mipLevels-00958", 38}, { {"mipLevels must be less than or equal to the number of levels in the complete mipmap chain based on extent.width, extent.height, and extent.depth", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-mipLevels-02255", 38}, { {"mipLevels must be less than or equal to imageCreateMaxMipLevels (as defined in Image Creation Limits)", 101}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-multisampleArrayImage-04460", 50}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::multisampleArrayImage is VK_FALSE, and samples is not VK_SAMPLE_COUNT_1_BIT, then arrayLayers must be 1", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-multisampledRenderToSingleSampled-06882", 62}, { {"If the multisampledRenderToSingleSampled feature is not enabled, flags must not contain VK_IMAGE_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT", 149}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pData-11286", 34}, { {"If flags contains VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT, and VkOpaqueCaptureDataCreateInfoEXT::pData is not NULL, VkOpaqueCaptureDataCreateInfoEXT::pData->size must be equal to imageCaptureReplayOpaqueDataSize", 226}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-00988", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfoNV structure, it must not contain a VkExternalMemoryImageCreateInfo structure", 138}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-00990", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo structure, its handleTypes member must only contain bits that are also in VkExternalImageFormatProperties::externalMemoryProperties.compatibleHandleTypes, as returned by vkGetPhysicalDeviceImageFormatProperties2 with format, imageType, tiling, usage, and flags equal to those in this structure, and with a VkPhysicalDeviceExternalImageFormatInfo structure included in the pNext chain, with a handleType equal to any one of the handle types specified in VkExternalMemoryImageCreateInfo::handleTypes", 559}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-00991", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfoNV structure, its handleTypes member must only contain bits that are also in VkExternalImageFormatPropertiesNV::externalMemoryFeatures.compatibleHandleTypes, as returned by vkGetPhysicalDeviceExternalImageFormatPropertiesNV with format, imageType, tiling, usage, and flags equal to those in this structure, and with externalHandleType equal to any one of the handle types specified in VkExternalMemoryImageCreateInfoNV::handleTypes", 492}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-01443", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo or VkExternalMemoryImageCreateInfoNV structure whose handleTypes member is not 0, initialLayout must be VK_IMAGE_LAYOUT_UNDEFINED", 191}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-01974", 34}, { {"If the pNext chain includes a VkExternalFormatANDROID structure, and its externalFormat member is non-zero the format must be VK_FORMAT_UNDEFINED", 145}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-01975", 34}, { {"If the pNext chain does not include a VkExternalFormatANDROID structure, or does and its externalFormat member is 0, the format must not be VK_FORMAT_UNDEFINED", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02262", 34}, { {"If the pNext chain includes a VkImageDrmFormatModifierListCreateInfoEXT or VkImageDrmFormatModifierExplicitCreateInfoEXT structure, then tiling must be VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT", 191}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02393", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo structure whose handleTypes member includes VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, imageType must be VK_IMAGE_TYPE_2D", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02394", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo structure whose handleTypes member includes VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, mipLevels must either be 1 or equal to the number of levels in the complete mipmap chain based on extent.width, extent.height, and extent.depth", 317}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02396", 34}, { {"If the pNext chain includes a VkExternalFormatANDROID structure whose externalFormat member is not 0, flags must not include VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02397", 34}, { {"If the pNext chain includes a VkExternalFormatANDROID structure whose externalFormat member is not 0, usage must not include any usages except VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, or VK_IMAGE_USAGE_SAMPLED_BIT", 246}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-02398", 34}, { {"If the pNext chain includes a VkExternalFormatANDROID structure whose externalFormat member is not 0, tiling must be VK_IMAGE_TILING_OPTIMAL", 140}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06390", 34}, { {"If the VkImage is to be used to import memory from a VkBufferCollectionFUCHSIA, a VkBufferCollectionImageCreateInfoFUCHSIA structure must be chained to pNext", 157}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06722", 34}, { {"If a VkImageFormatListCreateInfo structure was included in the pNext chain and format is not a multi-planar format and VkImageFormatListCreateInfo::viewFormatCount is not zero, then each format in VkImageFormatListCreateInfo::pViewFormats must either be compatible with the format as described in the compatibility table or, if flags contains VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT, be an uncompressed format that is size-compatible with format", 453}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06743", 34}, { {"If the pNext chain includes a VkImageCompressionControlEXT structure, format is a multi-planar format, and VkImageCompressionControlEXT::flags includes VK_IMAGE_COMPRESSION_FIXED_RATE_EXPLICIT_EXT, then VkImageCompressionControlEXT::compressionControlPlaneCount must be equal to the number of planes in format", 309}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06744", 34}, { {"If the pNext chain includes a VkImageCompressionControlEXT structure, format is not a multi-planar format, and VkImageCompressionControlEXT::flags includes VK_IMAGE_COMPRESSION_FIXED_RATE_EXPLICIT_EXT, then VkImageCompressionControlEXT::compressionControlPlaneCount must be 1", 275}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06746", 34}, { {"If the pNext chain includes a VkImageCompressionControlEXT structure, it must not contain a VkImageDrmFormatModifierExplicitCreateInfoEXT structure", 147}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06783", 34}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be either VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT or VK_EXPORT_METAL_OBJECT_TYPE_METAL_IOSURFACE_BIT_EXT", 221}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06784", 34}, { {"If the pNext chain includes a VkImportMetalTextureInfoEXT structure its plane member must be VK_IMAGE_ASPECT_PLANE_0_BIT, VK_IMAGE_ASPECT_PLANE_1_BIT, or VK_IMAGE_ASPECT_PLANE_2_BIT", 181}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06785", 34}, { {"If the pNext chain includes a VkImportMetalTextureInfoEXT structure and the image does not have a multi-planar format, then VkImportMetalTextureInfoEXT::plane must be VK_IMAGE_ASPECT_PLANE_0_BIT", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06786", 34}, { {"If the pNext chain includes a VkImportMetalTextureInfoEXT structure and the image has a multi-planar format with only two planes, then VkImportMetalTextureInfoEXT::plane must not be VK_IMAGE_ASPECT_PLANE_2_BIT", 209}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-06811", 34}, { {"If the pNext chain includes a VkVideoProfileListInfoKHR structure with profileCount greater than 0, then supportedVideoFormat must be VK_TRUE", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08105", 34}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_IMAGE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08951", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo structure whose handleTypes member includes VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX, imageType must be VK_IMAGE_TYPE_2D", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08952", 34}, { {"If the pNext chain includes a VkExternalMemoryImageCreateInfo structure whose handleTypes member includes VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX, mipLevels must either be 1 or equal to the number of levels in the complete mipmap chain based on extent.width, extent.height, and extent.depth", 303}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08953", 34}, { {"If the pNext chain includes a VkExternalFormatQNX structure whose externalFormat member is not 0, flags must not include VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT", 155}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08954", 34}, { {"If the pNext chain includes a VkExternalFormatQNX structure whose externalFormat member is not 0, usage must not include any usages except VK_IMAGE_USAGE_SAMPLED_BIT", 165}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-08955", 34}, { {"If the pNext chain includes a VkExternalFormatQNX structure whose externalFormat member is not 0, tiling must be VK_IMAGE_TILING_OPTIMAL", 136}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-09457", 34}, { {"If the pNext chain includes a VkExternalFormatANDROID structure whose externalFormat member is not 0, and externalFormatResolve feature is not enabled, usage must not include VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-09653", 34}, { {"If the pNext chain contains a VkImageAlignmentControlCreateInfoMESA structure, tiling must be VK_IMAGE_TILING_OPTIMAL", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-09654", 34}, { {"If the pNext chain contains a VkImageAlignmentControlCreateInfoMESA structure, it must not contain a VkExternalMemoryImageCreateInfo structure", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-10062", 34}, { {"If a VkImageFormatListCreateInfo structure was included in the pNext chain and format is a multi-planar format and flags contains VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT and VkImageFormatListCreateInfo::viewFormatCount is not zero, then each format in VkImageFormatListCreateInfo::pViewFormats must be compatible with the VkFormat for the plane of the image format", 359}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-10250", 34}, { {"If the pNext chain includes a VkVideoProfileListInfoKHR structure and for any element of its pProfiles member videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the videoEncodeAV1 feature must be enabled", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-10784", 34}, { {"If the pNext chain includes a VkVideoProfileListInfoKHR structure and for any element of its pProfiles member videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the videoDecodeVP9 feature must be enabled", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-10920", 34}, { {"If the pNext chain includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be enabled", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-12442", 34}, { {"pNext must contain at most one of the structures VkImageStencilUsageCreateInfo and VkImageStencilUsage2CreateInfoKHR", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-pNext-pNext", 34}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkBufferCollectionImageCreateInfoFUCHSIA, VkDataGraphOpticalFlowImageFormatInfoARM, VkDedicatedAllocationImageCreateInfoNV, VkExportMetalObjectCreateInfoEXT, VkExternalFormatANDROID, VkExternalFormatOHOS, VkExternalFormatQNX, VkExternalMemoryImageCreateInfo, VkExternalMemoryImageCreateInfoNV, VkImageAlignmentControlCreateInfoMESA, VkImageCompressionControlEXT, VkImageCreateFlags2CreateInfoKHR, VkImageDrmFormatModifierExplicitCreateInfoEXT, VkImageDrmFormatModifierListCreateInfoEXT, VkImageFormatListCreateInfo, VkImageStencilUsage2CreateInfoKHR, VkImageStencilUsageCreateInfo, VkImageSwapchainCreateInfoKHR, VkImageUsageFlags2CreateInfoKHR, VkImportMetalIOSurfaceInfoEXT, VkImportMetalTextureInfoEXT, VkOpaqueCaptureDataCreateInfoEXT, VkOpaqueCaptureDescriptorDataCreateInfoEXT, VkOpticalFlowImageFormatInfoNV, or VkVideoProfileListInfoKHR", 975}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-physicalDeviceCount-01421", 48}, { {"If the logical device was created with VkDeviceGroupDeviceCreateInfo::physicalDeviceCount equal to 1, flags must not contain VK_IMAGE_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT", 172}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sType-unique", 35}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT or VkImportMetalTextureInfoEXT", 173}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-samples-02083", 36}, { {"If usage includes VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR, samples must be VK_SAMPLE_COUNT_1_BIT", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-samples-02257", 36}, { {"If samples is not VK_SAMPLE_COUNT_1_BIT, then imageType must be VK_IMAGE_TYPE_2D, flags must not contain VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT, mipLevels must be equal to 1, and imageCreateMaybeLinear (as defined in Image Creation Limits) must be VK_FALSE,", 254}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-samples-02258", 36}, { {"samples must be a valid VkSampleCountFlagBits value that is set in imageCreateSampleCounts (as defined in Image Creation Limits)", 128}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-samples-02558", 36}, { {"If samples is not VK_SAMPLE_COUNT_1_BIT, usage must not contain VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-samples-parameter", 40}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-shadingRateImage-07727", 45}, { {"If the shadingRateImage feature is enabled and usage includes VK_IMAGE_USAGE_SHADING_RATE_IMAGE_BIT_NV, tiling must be VK_IMAGE_TILING_OPTIMAL", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sharingMode-00941", 40}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, pQueueFamilyIndices must be a valid pointer to an array of queueFamilyIndexCount uint32_t values", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sharingMode-00942", 40}, { {"If the maintenance11 feature is not enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 1", 140}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sharingMode-01420", 40}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, each element of pQueueFamilyIndices must be unique and must be less than pQueueFamilyPropertyCount returned by either vkGetPhysicalDeviceQueueFamilyProperties or vkGetPhysicalDeviceQueueFamilyProperties2 for the physicalDevice that was used to create device", 303}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-sharingMode-parameter", 44}, { {"sharingMode must be a valid VkSharingMode value", 47}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tileMemoryHeap-10766", 43}, { {"If the tileMemoryHeap feature is not enabled, usage must not include VK_IMAGE_USAGE_TILE_MEMORY_BIT_QCOM", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tiling-02261", 35}, { {"If tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then the pNext chain must include exactly one of VkImageDrmFormatModifierListCreateInfoEXT or VkImageDrmFormatModifierExplicitCreateInfoEXT structures", 204}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tiling-02353", 35}, { {"If tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT and flags contains VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT, then the pNext chain must include a VkImageFormatListCreateInfo structure with non-zero viewFormatCount", 211}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tiling-04121", 35}, { {"If tiling is VK_IMAGE_TILING_LINEAR, flags must not contain VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tiling-09711", 35}, { {"If tiling is VK_IMAGE_TILING_LINEAR then VK_IMAGE_USAGE_TENSOR_ALIASING_BIT_ARM must not be set in usage", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-tiling-parameter", 39}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-00963", 34}, { {"If usage includes VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, then bits other than VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, and VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT must not be set", 217}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-00964", 34}, { {"If usage includes VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, extent.width must be less than or equal to VkPhysicalDeviceLimits::maxFramebufferWidth", 267}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-00965", 34}, { {"If usage includes VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, extent.height must be less than or equal to VkPhysicalDeviceLimits::maxFramebufferHeight", 269}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-00966", 34}, { {"If usage includes VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, usage must also contain at least one of VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT", 219}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-00968", 34}, { {"If the shaderStorageImageMultisample feature is not enabled, and usage contains VK_IMAGE_USAGE_STORAGE_BIT, samples must be VK_SAMPLE_COUNT_1_BIT", 145}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-04815", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR, VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR, or VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR, and flags does not include VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then the pNext chain must include a VkVideoProfileListInfoKHR structure with profileCount greater than 0 and pProfiles including at least one VkVideoProfileInfoKHR structure with a videoCodecOperation member specifying a decode operation", 459}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-04816", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR, VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR, or VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR, and flags does not include VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR, then the pNext chain must include a VkVideoProfileListInfoKHR structure with profileCount greater than 0 and pProfiles including at least one VkVideoProfileInfoKHR structure with a videoCodecOperation member specifying an encode operation", 460}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-04992", 34}, { {"If usage includes VK_IMAGE_USAGE_INVOCATION_MASK_BIT_HUAWEI, tiling must be VK_IMAGE_TILING_LINEAR", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10245", 34}, { {"If usage includes VK_IMAGE_USAGE_HOST_TRANSFER_BIT, then the hostImageCopy feature must be enabled", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10251", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, then the videoEncodeQuantizationMap feature must be enabled", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10252", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, imageType must be VK_IMAGE_TYPE_2D", 164}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10253", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, samples must be VK_SAMPLE_COUNT_1_BIT", 167}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10254", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, then the pNext chain must include a VkVideoProfileListInfoKHR structure with profileCount equal to 1 and pProfiles pointing to a VkVideoProfileInfoKHR structure with a videoCodecOperation member specifying an encode operation", 355}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10255", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR, then VkVideoEncodeCapabilitiesKHR::flags must include VK_VIDEO_ENCODE_CAPABILITY_QUANTIZATION_DELTA_MAP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in the pProfiles member of the VkVideoProfileListInfoKHR structure included in the pNext chain", 372}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10256", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, then VkVideoEncodeCapabilitiesKHR::flags must include VK_VIDEO_ENCODE_CAPABILITY_EMPHASIS_MAP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in the pProfiles member of the VkVideoProfileListInfoKHR structure included in the pNext chain", 352}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10257", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, extent.width must be less than or equal to VkVideoEncodeQuantizationMapCapabilitiesKHR::maxQuantizationMapExtent.width, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in the pProfiles member of the VkVideoProfileListInfoKHR structure included in the pNext chain", 431}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-10258", 34}, { {"If usage includes VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, extent.height must be less than or equal to VkVideoEncodeQuantizationMapCapabilitiesKHR::maxQuantizationMapExtent.height, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in the pProfiles member of the VkVideoProfileListInfoKHR structure included in the pNext chain", 433}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-parameter", 38}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageCreateInfo-usage-requiredbitmask", 44}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDescriptorInfoEXT-layout-11219", 42}, { {"layout must be VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ_KHR, VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT, or VK_IMAGE_LAYOUT_GENERAL", 503}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-layout-11221", 42}, { {"If layout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, then pView->aspectMask must not include VK_IMAGE_ASPECT_COLOR_BIT", 334}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-layout-parameter", 46}, { {"layout must be a valid VkImageLayout value", 42}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-pView-11426", 41}, { {"pView->viewType must not be VK_IMAGE_VIEW_TYPE_2D_ARRAY if pView->image was created with an imageType of VK_IMAGE_TYPE_3D", 121}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-pView-11427", 41}, { {"If pView->viewType is VK_IMAGE_VIEW_TYPE_2D and pView->image was created with an imageType of VK_IMAGE_TYPE_3D, pView->image must have been created with VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT set", 199}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-pView-11430", 41}, { {"If pView->image is a depth/stencil image, pView->subresourceRange.aspectMask must include either VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT but not both", 166}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-pView-parameter", 45}, { {"pView must be a valid pointer to a valid VkImageViewCreateInfo structure", 72}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDescriptorInfoEXT-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_DESCRIPTOR_INFO_EXT", 57}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-arrayPitch-02268", 67}, { {"For each element of pPlaneLayouts, arrayPitch must be 0 if VkImageCreateInfo::arrayLayers is 1", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-depthPitch-02269", 67}, { {"For each element of pPlaneLayouts, depthPitch must be 0 if VkImageCreateInfo::extent.depth is 1", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-drmFormatModifier-02264", 74}, { {"drmFormatModifier must be compatible with the parameters in VkImageCreateInfo and its pNext chain, as determined by querying VkPhysicalDeviceImageFormatInfo2 extended with VkPhysicalDeviceImageDrmFormatModifierInfoEXT", 217}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-drmFormatModifierPlaneCount-02265", 84}, { {"drmFormatModifierPlaneCount must be equal to the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with VkImageCreateInfo::format and drmFormatModifier, as found by querying VkDrmFormatModifierPropertiesListEXT", 233}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-drmFormatModifierPlaneCount-arraylength", 90}, { {"drmFormatModifierPlaneCount must be greater than 0", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-pPlaneLayouts-parameter", 74}, { {"pPlaneLayouts must be a valid pointer to an array of drmFormatModifierPlaneCount VkSubresourceLayout structures", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_EXPLICIT_CREATE_INFO_EXT", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierExplicitCreateInfoEXT-size-02267", 61}, { {"For each element of pPlaneLayouts, size must be 0", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierListCreateInfoEXT-drmFormatModifierCount-arraylength", 81}, { {"drmFormatModifierCount must be greater than 0", 45}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierListCreateInfoEXT-pDrmFormatModifiers-02263", 72}, { {"Each modifier in pDrmFormatModifiers must be compatible with the parameters in VkImageCreateInfo and its pNext chain, as determined by querying VkPhysicalDeviceImageFormatInfo2 extended with VkPhysicalDeviceImageDrmFormatModifierInfoEXT", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierListCreateInfoEXT-pDrmFormatModifiers-parameter", 76}, { {"pDrmFormatModifiers must be a valid pointer to an array of drmFormatModifierCount uint64_t values", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierListCreateInfoEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_LIST_CREATE_INFO_EXT", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierPropertiesEXT-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageDrmFormatModifierPropertiesEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_PROPERTIES_EXT", 72}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-colorSpaceCount-arraylength", 68}, { {"colorSpaceCount must be greater than 0", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-flags-zerobitmask", 58}, { {"flags must be 0", 15}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-imageCreateInfo-parameter", 66}, { {"imageCreateInfo must be a valid VkImageCreateInfo structure", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-pColorSpaces-parameter", 63}, { {"pColorSpaces must be a valid pointer to an array of colorSpaceCount valid VkSysmemColorSpaceFUCHSIA structures", 110}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-requiredFormatFeatures-parameter", 73}, { {"requiredFormatFeatures must be a valid combination of VkFormatFeatureFlagBits values", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-requiredFormatFeatures-requiredbitmask", 79}, { {"requiredFormatFeatures must not be 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatConstraintsInfoFUCHSIA-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_FORMAT_CONSTRAINTS_INFO_FUCHSIA", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatListCreateInfo-pViewFormats-parameter", 55}, { {"If viewFormatCount is not 0, pViewFormats must be a valid pointer to an array of viewFormatCount valid VkFormat values", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatListCreateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_FORMAT_LIST_CREATE_INFO", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatListCreateInfo-viewFormatCount-09540", 54}, { {"If viewFormatCount is not 0, each element of pViewFormats must not be VK_FORMAT_UNDEFINED", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageFormatProperties2-pNext-pNext", 41}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAndroidHardwareBufferUsageANDROID, VkExternalImageFormatProperties, VkFilterCubicImageViewImageFormatPropertiesEXT, VkHostImageCopyDevicePerformanceQuery, VkImageCompressionPropertiesEXT, VkNativeBufferUsageOHOS, VkSamplerYcbcrConversionImageFormatProperties, VkSubsampledImageFormatPropertiesEXT, or VkTextureLODGatherFormatPropertiesAMD", 471}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkImageFormatProperties2-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_FORMAT_PROPERTIES_2", 57}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkImageFormatProperties2-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkImageMemoryBarrier-None-09052", 36}, { {"If the synchronization2 feature is not enabled, and image was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, at least one of srcQueueFamilyIndex and dstQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED", 212}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-None-09053", 36}, { {"If the synchronization2 feature is not enabled, and image was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED or VK_QUEUE_FAMILY_EXTERNAL", 200}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-None-09054", 36}, { {"If the synchronization2 feature is not enabled, and image was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_IGNORED or VK_QUEUE_FAMILY_EXTERNAL", 200}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-None-09097", 36}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-None-09098", 36}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-aspectMask-08702", 42}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_DEPTH_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-aspectMask-08703", 42}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_STENCIL_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-attachmentFeedbackLoopLayout-07313", 60}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 134}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-dstQueueFamilyIndex-09100", 51}, { {"If the VK_EXT_queue_family_foreign extension is not enabled dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-dstQueueFamilyIndex-12331", 51}, { {"If dstQueueFamilyIndex is VK_QUEUE_FAMILY_EXTERNAL and image was created with VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT in VkExternalMemoryImageCreateInfo::handleTypes, newLayout must be VK_IMAGE_LAYOUT_GENERAL", 273}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-dynamicRenderingLocalRead-09551", 57}, { {"If the dynamicRenderingLocalRead feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-dynamicRenderingLocalRead-09552", 57}, { {"If the dynamicRenderingLocalRead feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-01672", 37}, { {"If image has a multi-planar format and the image is disjoint, then the aspectMask member of subresourceRange must include at least one multi-planar aspect mask bit or VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-01932", 37}, { {"If image is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 146}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-03319", 37}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is enabled, then the aspectMask member of subresourceRange must include either or both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 252}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-03320", 37}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is not enabled, then the aspectMask member of subresourceRange must include both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 246}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-09117", 37}, { {"If image was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 240}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-09118", 37}, { {"If image was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 240}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-09241", 37}, { {"If image has a color format that is single-plane, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-09242", 37}, { {"If image has a color format and is not disjoint, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-10749", 37}, { {"If image has a depth-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_DEPTH_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-10750", 37}, { {"If image has a stencil-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_STENCIL_BIT", 117}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-image-parameter", 41}, { {"image must be a valid VkImage handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-maintenance9-10798", 44}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer must be less than the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 338}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-maintenance9-10799", 44}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set and either subresourceRange.baseArrayLayer is not equal to 0 or subresourceRange.layerCount is not equal to VK_REMAINING_ARRAY_LAYERS, subresourceRange.levelCount must be 1", 292}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-maintenance9-10800", 44}, { {"If the maintenance9 feature is enabled, subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 443}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-newLayout-01198", 41}, { {"If layouts are not ignored, newLayout must not be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-newLayout-parameter", 45}, { {"newLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01197", 41}, { {"If layouts are not ignored, oldLayout must be VK_IMAGE_LAYOUT_UNDEFINED or the current layout of the image subresources affected by the barrier", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01208", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01209", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 204}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01210", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 203}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01211", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 218}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01212", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 180}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01213", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 180}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01658", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-01659", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-02088", 41}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR then image must have been created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-10767", 41}, { {"If the zeroInitializeDeviceMemory feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT", 116}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-10768", 41}, { {"If oldLayout is VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, then all subresources must be included in the barrier", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-oldLayout-parameter", 45}, { {"oldLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExternalMemoryAcquireUnmodifiedEXT or VkSampleLocationsInfoEXT", 195}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-03938", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL, image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 225}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-03939", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL, image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 270}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-04065", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 280}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-04066", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 196}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-04067", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 282}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-04068", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 198}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07006", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then image must have been created with either the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set, and the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT usage flags set, and the VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT usage flag set", 418}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07120", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07121", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07122", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07123", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07124", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-07125", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-09099", 51}, { {"If the VK_EXT_queue_family_foreign extension is not enabled srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-09550", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ then image must have been created with either the VK_IMAGE_USAGE_STORAGE_BIT usage flag set, or with both the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag and either of the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set", 364}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-10287", 51}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_QUANTIZATION_MAP_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR usage flags set", 273}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-srcQueueFamilyIndex-12332", 51}, { {"If srcQueueFamilyIndex is VK_QUEUE_FAMILY_EXTERNAL and image was created with VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT in VkExternalMemoryImageCreateInfo::handleTypes, oldLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_UNDEFINED", 302}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-01486", 48}, { {"subresourceRange.baseMipLevel must be less than the mipLevels specified in VkImageCreateInfo when image was created", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-01488", 48}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, subresourceRange.baseArrayLayer must be less than the arrayLayers specified in VkImageCreateInfo when image was created", 258}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-01724", 48}, { {"If subresourceRange.levelCount is not VK_REMAINING_MIP_LEVELS, subresourceRange.baseMipLevel + subresourceRange.levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-01725", 48}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 366}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-09601", 48}, { {"subresourceRange.aspectMask must be valid for the format the image was created with", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-subresourceRange-parameter", 52}, { {"subresourceRange must be a valid VkImageSubresourceRange structure", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-synchronization2-07793", 48}, { {"If the synchronization2 feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier-synchronization2-07794", 48}, { {"If the synchronization2 feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-None-09097", 37}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-None-09098", 37}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-aspectMask-08702", 43}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_DEPTH_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-aspectMask-08703", 43}, { {"If the aspectMask member of subresourceRange includes VK_IMAGE_ASPECT_STENCIL_BIT, oldLayout and newLayout must not be one of VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-attachmentFeedbackLoopLayout-07313", 61}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT", 134}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03900", 46}, { {"If dstAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03901", 46}, { {"If dstAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03902", 46}, { {"If dstAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03903", 46}, { {"If dstAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03904", 46}, { {"If dstAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03905", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03906", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03907", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03909", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03910", 46}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03911", 46}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03912", 46}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03913", 46}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03914", 46}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03915", 46}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03916", 46}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03917", 46}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03918", 46}, { {"If dstAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03919", 46}, { {"If dstAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03920", 46}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03922", 46}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03923", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03924", 46}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03925", 46}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03926", 46}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03927", 46}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-03928", 46}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04747", 46}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04858", 46}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04859", 46}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04860", 46}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04861", 46}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-04994", 46}, { {"If dstAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, dstStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-06256", 46}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07272", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07454", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07455", 46}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07456", 46}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07457", 46}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-07458", 46}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-08118", 46}, { {"If dstAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-10670", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-10671", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-11294", 46}, { {"If dstAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-11771", 46}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-11772", 46}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstAccessMask-parameter", 50}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstQueueFamilyIndex-09100", 52}, { {"If the VK_EXT_queue_family_foreign extension is not enabled dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03929", 45}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03930", 45}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03931", 45}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03932", 45}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03933", 45}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03934", 45}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-03935", 45}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-04957", 45}, { {"If the subpassShading feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-04995", 45}, { {"If the invocationMask feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-07316", 45}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-07946", 45}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-10751", 45}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-10752", 45}, { {"If the rayTracingMaintenance1 feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-11541", 45}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dstStageMask-parameter", 49}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dynamicRenderingLocalRead-09551", 58}, { {"If the dynamicRenderingLocalRead feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-dynamicRenderingLocalRead-09552", 58}, { {"If the dynamicRenderingLocalRead feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-01672", 38}, { {"If image has a multi-planar format and the image is disjoint, then the aspectMask member of subresourceRange must include at least one multi-planar aspect mask bit or VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-01932", 38}, { {"If image is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 146}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-03319", 38}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is enabled, then the aspectMask member of subresourceRange must include either or both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 252}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-03320", 38}, { {"If image has a depth/stencil format with both depth and stencil and the separateDepthStencilLayouts feature is not enabled, then the aspectMask member of subresourceRange must include both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 246}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-09117", 38}, { {"If image was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 240}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-09118", 38}, { {"If image was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 240}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-09241", 38}, { {"If image has a color format that is single-plane, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-09242", 38}, { {"If image has a color format and is not disjoint, then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_COLOR_BIT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-10749", 38}, { {"If image has a depth-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_DEPTH_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-10750", 38}, { {"If image has a stencil-only format then the aspectMask member of subresourceRange must be VK_IMAGE_ASPECT_STENCIL_BIT", 117}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-image-parameter", 42}, { {"image must be a valid VkImage handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-maintenance9-10798", 45}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer must be less than the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 338}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-maintenance9-10799", 45}, { {"If the maintenance9 feature is enabled and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set and either subresourceRange.baseArrayLayer is not equal to 0 or subresourceRange.layerCount is not equal to VK_REMAINING_ARRAY_LAYERS, subresourceRange.levelCount must be 1", 292}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-maintenance9-10800", 45}, { {"If the maintenance9 feature is enabled, subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, and image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 443}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-newLayout-01198", 42}, { {"If layouts are not ignored, newLayout must not be VK_IMAGE_LAYOUT_UNDEFINED or VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT or VK_IMAGE_LAYOUT_PREINITIALIZED", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-newLayout-parameter", 46}, { {"newLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01197", 42}, { {"If layouts are not ignored, oldLayout must be VK_IMAGE_LAYOUT_UNDEFINED or the current layout of the image subresources affected by the barrier", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01208", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01209", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 204}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01210", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 203}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01211", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 218}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01212", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 180}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01213", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 180}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01658", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-01659", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-02088", 42}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR then image must have been created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-10767", 42}, { {"If the zeroInitializeDeviceMemory feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT", 116}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-10768", 42}, { {"If oldLayout is VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, then all subresources must be included in the barrier", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-oldLayout-parameter", 46}, { {"oldLayout must be a valid VkImageLayout value", 45}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExternalMemoryAcquireUnmodifiedEXT, VkMemoryBarrierAccessFlags3KHR, or VkSampleLocationsInfoEXT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2", 54}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03900", 46}, { {"If srcAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03901", 46}, { {"If srcAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03902", 46}, { {"If srcAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03903", 46}, { {"If srcAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03904", 46}, { {"If srcAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03905", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03906", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03907", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03909", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03910", 46}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03911", 46}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03912", 46}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03913", 46}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03914", 46}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03915", 46}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03916", 46}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03917", 46}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03918", 46}, { {"If srcAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03919", 46}, { {"If srcAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03920", 46}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03922", 46}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03923", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03924", 46}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03925", 46}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03926", 46}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03927", 46}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-03928", 46}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04747", 46}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04858", 46}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04859", 46}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04860", 46}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04861", 46}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-04994", 46}, { {"If srcAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, srcStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-06256", 46}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07272", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07454", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07455", 46}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07456", 46}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07457", 46}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-07458", 46}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-08118", 46}, { {"If srcAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-10670", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-10671", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-11294", 46}, { {"If srcAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-11771", 46}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-11772", 46}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcAccessMask-parameter", 50}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-03938", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL, image must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 225}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-03939", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL, image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 270}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-04065", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 280}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-04066", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 196}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-04067", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then image must have been created with at least one of the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, VK_IMAGE_USAGE_SAMPLED_BIT, or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 282}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-04068", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL then image must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 198}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07006", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then image must have been created with either the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set, and the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT usage flags set, and the VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT usage flag set", 418}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07120", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07121", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07122", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07123", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07124", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DST_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-07125", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR usage flag set", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-09099", 52}, { {"If the VK_EXT_queue_family_foreign extension is not enabled srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-09550", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ then image must have been created with either the VK_IMAGE_USAGE_STORAGE_BIT usage flag set, or with both the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag and either of the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set", 364}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcQueueFamilyIndex-10287", 52}, { {"If layouts are not ignored, oldLayout or newLayout is VK_IMAGE_LAYOUT_VIDEO_ENCODE_QUANTIZATION_MAP_KHR then image must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR usage flags set", 273}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03854", 45}, { {"If either srcStageMask or dstStageMask includes VK_PIPELINE_STAGE_2_HOST_BIT, srcQueueFamilyIndex and dstQueueFamilyIndex must be equal", 135}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03855", 45}, { {"If srcStageMask includes VK_PIPELINE_STAGE_2_HOST_BIT, and srcQueueFamilyIndex and dstQueueFamilyIndex define a queue family ownership transfer or oldLayout and newLayout define an image layout transition, oldLayout must be one of VK_IMAGE_LAYOUT_PREINITIALIZED, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, VK_IMAGE_LAYOUT_UNDEFINED, or VK_IMAGE_LAYOUT_GENERAL", 354}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03929", 45}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03930", 45}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03931", 45}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03932", 45}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03933", 45}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03934", 45}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-03935", 45}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-04957", 45}, { {"If the subpassShading feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-04995", 45}, { {"If the invocationMask feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-07316", 45}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-07946", 45}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-10751", 45}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-10752", 45}, { {"If the rayTracingMaintenance1 feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-11541", 45}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-srcStageMask-parameter", 49}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-01486", 49}, { {"subresourceRange.baseMipLevel must be less than the mipLevels specified in VkImageCreateInfo when image was created", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-01488", 49}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, subresourceRange.baseArrayLayer must be less than the arrayLayers specified in VkImageCreateInfo when image was created", 258}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-01724", 49}, { {"If subresourceRange.levelCount is not VK_REMAINING_MIP_LEVELS, subresourceRange.baseMipLevel + subresourceRange.levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-01725", 49}, { {"If image is not a 3D image or was created without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, or the maintenance9 feature is not enabled, and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 366}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-09601", 49}, { {"subresourceRange.aspectMask must be valid for the format the image was created with", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-subresourceRange-parameter", 53}, { {"subresourceRange must be a valid VkImageSubresourceRange structure", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-synchronization2-07793", 49}, { {"If the synchronization2 feature is not enabled, oldLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryBarrier2-synchronization2-07794", 49}, { {"If the synchronization2 feature is not enabled, newLayout must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR or VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-01589", 47}, { {"If image was created with a multi-planar format and the VK_IMAGE_CREATE_DISJOINT_BIT flag, there must be a VkImagePlaneMemoryRequirementsInfo included in the pNext chain of the VkImageMemoryRequirementsInfo2 structure", 217}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-01590", 47}, { {"If image was not created with the VK_IMAGE_CREATE_DISJOINT_BIT flag, there must not be a VkImagePlaneMemoryRequirementsInfo included in the pNext chain of the VkImageMemoryRequirementsInfo2 structure", 199}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-01897", 47}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID external memory handle type, then image must be bound to memory", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-02279", 47}, { {"If image was created with VK_IMAGE_CREATE_DISJOINT_BIT and with VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then there must be a VkImagePlaneMemoryRequirementsInfo included in the pNext chain of the VkImageMemoryRequirementsInfo2 structure", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-02280", 47}, { {"If image was created with a single-plane format and with any tiling other than VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then there must not be a VkImagePlaneMemoryRequirementsInfo included in the pNext chain of the VkImageMemoryRequirementsInfo2 structure", 255}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-08961", 47}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX external memory handle type, then image must be bound to memory", 146}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-image-parameter", 51}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-pNext-pNext", 47}, { {"pNext must be NULL or a pointer to a valid instance of VkImagePlaneMemoryRequirementsInfo", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_MEMORY_REQUIREMENTS_INFO_2", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageMemoryRequirementsInfo2-sType-unique", 48}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImagePipeSurfaceCreateInfoFUCHSIA-flags-zerobitmask", 58}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkImagePipeSurfaceCreateInfoFUCHSIA-imagePipeHandle-04863", 62}, { {"imagePipeHandle must be a valid zx_handle_t", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkImagePipeSurfaceCreateInfoFUCHSIA-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkImagePipeSurfaceCreateInfoFUCHSIA-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_IMAGEPIPE_SURFACE_CREATE_INFO_FUCHSIA", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkImagePlaneMemoryRequirementsInfo-planeAspect-02281", 57}, { {"If the image's tiling is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, then planeAspect must be a single valid multi-planar aspect mask bit", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImagePlaneMemoryRequirementsInfo-planeAspect-02282", 57}, { {"If the image's tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then planeAspect must be a single valid memory plane for the image (that is, aspectMask must specify a plane index that is less than the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with the image's format and VkImageDrmFormatModifierPropertiesEXT::drmFormatModifier)", 361}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImagePlaneMemoryRequirementsInfo-planeAspect-parameter", 61}, { {"planeAspect must be a valid VkImageAspectFlagBits value", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImagePlaneMemoryRequirementsInfo-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_PLANE_MEMORY_REQUIREMENTS_INFO", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageResolve-aspectMask-10981", 36}, { {"The aspectMask member of srcSubresource and dstSubresource must only contain VK_IMAGE_ASPECT_COLOR_BIT", 102}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve-dstSubresource-parameter", 44}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve-layerCount-08803", 36}, { {"If neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount member of srcSubresource and dstSubresource must match", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve-layerCount-08804", 36}, { {"If one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 262}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve-srcSubresource-parameter", 44}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-aspectMask-10993", 37}, { {"The aspectMask member of srcSubresource and dstSubresource must only contain VK_IMAGE_ASPECT_DEPTH_BIT, VK_IMAGE_ASPECT_STENCIL_BIT, or VK_IMAGE_ASPECT_COLOR_BIT", 161}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-dstSubresource-parameter", 45}, { {"dstSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-layerCount-08803", 37}, { {"If neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount member of srcSubresource and dstSubresource must match", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-layerCount-08804", 37}, { {"If one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 262}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-maintenance10-10994", 40}, { {"If maintenance10 feature is not enabled, srcSubresource and dstSubresource must not contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-pNext-pNext", 32}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_RESOLVE_2", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageResolve2-srcSubresource-parameter", 45}, { {"srcSubresource must be a valid VkImageSubresourceLayers structure", 65}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSparseMemoryRequirementsInfo2-image-parameter", 57}, { {"image must be a valid VkImage handle", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkImageSparseMemoryRequirementsInfo2-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkImageSparseMemoryRequirementsInfo2-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_SPARSE_MEMORY_REQUIREMENTS_INFO_2", 71}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkImageStencilUsage2CreateInfoKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_2_CREATE_INFO_KHR", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsage2CreateInfoKHR-stencilUsage-12443", 57}, { {"If stencilUsage includes VK_IMAGE_USAGE_2_TRANSIENT_ATTACHMENT_BIT_KHR, it must not include bits other than VK_IMAGE_USAGE_2_DEPTH_STENCIL_ATTACHMENT_BIT_KHR or VK_IMAGE_USAGE_2_INPUT_ATTACHMENT_BIT_KHR", 202}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsage2CreateInfoKHR-stencilUsage-parameter", 61}, { {"stencilUsage must be a valid combination of VkImageUsageFlagBits2KHR values", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsage2CreateInfoKHR-stencilUsage-requiredbitmask", 67}, { {"stencilUsage must not be 0", 26}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsageCreateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_CREATE_INFO", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsageCreateInfo-stencilUsage-02539", 53}, { {"If stencilUsage includes VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT, it must not include bits other than VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsageCreateInfo-stencilUsage-parameter", 57}, { {"stencilUsage must be a valid combination of VkImageUsageFlagBits values", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageStencilUsageCreateInfo-stencilUsage-requiredbitmask", 63}, { {"stencilUsage must not be 0", 26}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresource-aspectMask-parameter", 44}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresource-aspectMask-requiredbitmask", 50}, { {"aspectMask must not be 0", 24}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresource2-imageSubresource-parameter", 51}, { {"imageSubresource must be a valid VkImageSubresource structure", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresource2-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresource2-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_SUBRESOURCE_2", 51}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceLayers-aspectMask-00167", 46}, { {"If aspectMask contains VK_IMAGE_ASPECT_COLOR_BIT, it must not contain either of VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-aspectMask-00168", 46}, { {"aspectMask must not contain VK_IMAGE_ASPECT_METADATA_BIT", 56}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-aspectMask-02247", 46}, { {"aspectMask must not include VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT for any index i", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-aspectMask-parameter", 50}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-aspectMask-requiredbitmask", 56}, { {"aspectMask must not be 0", 24}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-layerCount-01700", 46}, { {"If layerCount is not VK_REMAINING_ARRAY_LAYERS, it must be greater than 0", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceLayers-layerCount-09243", 46}, { {"If the maintenance5 feature is not enabled, layerCount must not be VK_REMAINING_ARRAY_LAYERS", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageSubresourceRange-aspectMask-01670", 45}, { {"If aspectMask includes VK_IMAGE_ASPECT_COLOR_BIT, then it must not include any of VK_IMAGE_ASPECT_PLANE_0_BIT, VK_IMAGE_ASPECT_PLANE_1_BIT, or VK_IMAGE_ASPECT_PLANE_2_BIT", 170}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceRange-aspectMask-02278", 45}, { {"aspectMask must not include VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT for any index i", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceRange-aspectMask-parameter", 49}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceRange-aspectMask-requiredbitmask", 55}, { {"aspectMask must not be 0", 24}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceRange-layerCount-01721", 45}, { {"If layerCount is not VK_REMAINING_ARRAY_LAYERS, it must be greater than 0", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSubresourceRange-levelCount-01720", 45}, { {"If levelCount is not VK_REMAINING_MIP_LEVELS, it must be greater than 0", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSwapchainCreateInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_SWAPCHAIN_CREATE_INFO_KHR", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSwapchainCreateInfoKHR-swapchain-00995", 50}, { {"If swapchain is not VK_NULL_HANDLE, the fields of VkImageCreateInfo must match the implied image creation parameters of the swapchain", 133}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageSwapchainCreateInfoKHR-swapchain-parameter", 54}, { {"If swapchain is not VK_NULL_HANDLE, swapchain must be a valid VkSwapchainKHR handle", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageToMemoryCopy-aspectMask-09103", 41}, { {"The aspectMask member of imageSubresource must only have a single bit set", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-imageExtent-06659", 42}, { {"imageExtent.width must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-imageExtent-06660", 42}, { {"imageExtent.height must not be 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-imageExtent-06661", 42}, { {"imageExtent.depth must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-imageSubresource-parameter", 51}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-memoryImageHeight-09102", 48}, { {"memoryImageHeight must be 0, or greater than or equal to the height member of imageExtent", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-memoryRowLength-09101", 46}, { {"memoryRowLength must be 0, or greater than or equal to the width member of imageExtent", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-pHostPointer-09066", 43}, { {"pHostPointer must point to memory that is large enough to contain all memory locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 175}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-pHostPointer-parameter", 47}, { {"pHostPointer must be a pointer value", 36}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-pRegions-09067", 39}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageToMemoryCopy-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_TO_MEMORY_COPY", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-VkImageUsageFlags2CreateInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_USAGE_FLAGS_2_CREATE_INFO_KHR", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageUsageFlags2CreateInfoKHR-usage-parameter", 52}, { {"usage must be a valid combination of VkImageUsageFlagBits2KHR values", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageUsageFlags2CreateInfoKHR-usage-requiredbitmask", 58}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-decodeMode-02230", 50}, { {"decodeMode must be one of VK_FORMAT_R16G16B16A16_SFLOAT, VK_FORMAT_R8G8B8A8_UNORM, or VK_FORMAT_E5B9G9R9_UFLOAT_PACK32", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-decodeMode-02231", 50}, { {"If the decodeModeSharedExponent feature is not enabled, decodeMode must not be VK_FORMAT_E5B9G9R9_UFLOAT_PACK32", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-decodeMode-02232", 50}, { {"If decodeMode is VK_FORMAT_R8G8B8A8_UNORM the image view must not include blocks using any of the ASTC HDR modes", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-decodeMode-parameter", 54}, { {"decodeMode must be a valid VkFormat value", 41}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-format-04084", 46}, { {"format of the image view must be one of the ASTC Compressed Image Formats", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewASTCDecodeModeEXT-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_ASTC_DECODE_MODE_EXT", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewAddressPropertiesNVX-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewAddressPropertiesNVX-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_ADDRESS_PROPERTIES_NVX", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCaptureDescriptorDataInfoEXT-imageView-08083", 60}, { {"imageView must have been created with VK_IMAGE_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkImageViewCreateInfo::flags", 135}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageViewCaptureDescriptorDataInfoEXT-imageView-parameter", 64}, { {"imageView must be a valid VkImageView handle", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageViewCaptureDescriptorDataInfoEXT-pNext-pNext", 56}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageViewCaptureDescriptorDataInfoEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_EXT", 75}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkImageViewCreateInfo-None-02273", 37}, { {"The format features of the resultant image view must contain at least one bit", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-None-12280", 37}, { {"If Vulkan 1.3 is not supported and the ycbcr2plane444Formats feature is not enabled, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 286}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-attachmentFragmentShadingRate-12386", 62}, { {"If the attachmentFragmentShadingRate feature is not enabled, the usage for the image view must not contain VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-components-parameter", 47}, { {"components must be a valid VkComponentMapping structure", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-02572", 38}, { {"If the fragmentDensityMapDynamic feature is not enabled, flags must not contain VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DYNAMIC_BIT_EXT", 137}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-03567", 38}, { {"If the fragmentDensityMapDeferred feature is not enabled, flags must not contain VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DEFERRED_BIT_EXT", 139}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-03568", 38}, { {"If flags contains VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DEFERRED_BIT_EXT, flags must not contain VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DYNAMIC_BIT_EXT", 158}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-04116", 38}, { {"If flags does not contain VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DYNAMIC_BIT_EXT, and image was created with the VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT usage flag set, its flags must not contain any of VK_IMAGE_CREATE_PROTECTED_BIT, VK_IMAGE_CREATE_SPARSE_BINDING_BIT, VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT, or VK_IMAGE_CREATE_SPARSE_ALIASED_BIT", 352}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-08106", 38}, { {"If flags includes VK_IMAGE_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay feature must be enabled", 138}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-flags-parameter", 42}, { {"flags must be a valid combination of VkImageViewCreateFlagBits values", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-format-04714", 39}, { {"If format has a _422 or _420 suffix then image must have been created with a width that is a multiple of 2", 106}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-format-04715", 39}, { {"If format has a _420 suffix then image must have been created with a height that is a multiple of 2", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-format-06415", 39}, { {"If the image view's format requires a sampler Y'CBCR conversion and usage contains VK_IMAGE_USAGE_SAMPLED_BIT, then the pNext chain must include a VkSamplerYcbcrConversionInfo structure with a conversion value other than VK_NULL_HANDLE", 235}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-format-12398", 39}, { {"If the format of the image is a multi-planar format and subresourceRange.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, format must be identical to the format used to create image", 173}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-format-parameter", 43}, { {"format must be a valid VkFormat value", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-01003", 38}, { {"If image was not created with VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT then viewType must not be VK_IMAGE_VIEW_TYPE_CUBE or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 148}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-01020", 38}, { {"If image is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 146}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-01583", 38}, { {"If image was created with the VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT flag, format must be compatible with, or must be an uncompressed format that is size-compatible with, the format used to create image", 211}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-01586", 38}, { {"If image was created with the VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT flag, if the format of the image is a multi-planar format, and if subresourceRange.aspectMask is one of the multi-planar aspect mask bits, then format must be compatible with the VkFormat for the plane of the image format indicated by subresourceRange.aspectMask, as defined in Compatible Formats of Planes of Multi-Planar Formats", 395}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-01761", 38}, { {"If image was created with the VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT flag, but without the VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT flag, and if the format of the image is not a multi-planar format, format must be compatible with the format used to create image, as defined in Format Compatibility Classes", 308}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02086", 38}, { {"If image was created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set, viewType must be VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 171}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02087", 38}, { {"If the shadingRateImage feature is enabled, and image was created with the VK_IMAGE_USAGE_SHADING_RATE_IMAGE_BIT_NV usage flag set, format must be VK_FORMAT_R8_UINT", 164}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02399", 38}, { {"If image has an Android external format, format must be VK_FORMAT_UNDEFINED", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02400", 38}, { {"If image has an Android external format, the pNext chain must include a VkSamplerYcbcrConversionInfo structure with a conversion object created with the same external format as image", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02401", 38}, { {"If image has an Android external format, all members of components must be the identity swizzle", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02571", 38}, { {"If image was created with the VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT usage flag set, subresourceRange.levelCount must be 1", 127}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-02724", 38}, { {"If image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, and viewType is VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY, subresourceRange.baseArrayLayer must be less than the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 368}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-03569", 38}, { {"If image was created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT and the VK_IMAGE_USAGE_SAMPLED_BIT usage flag set, subresourceRange.layerCount must be less than or equal to VkPhysicalDeviceFragmentDensityMap2PropertiesEXT::maxSubsampledArrayLayers", 261}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-04441", 38}, { {"image must have been created with a usage value containing at least one of the following: VK_IMAGE_USAGE_SAMPLED_BIT VK_IMAGE_USAGE_STORAGE_BIT VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR VK_IMAGE_USAGE_SAMPLE_WEIGHT_BIT_QCOM VK_IMAGE_USAGE_SAMPLE_BLOCK_MATCH_BIT_QCOM VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR", 748}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-04817", 38}, { {"If image was created with the VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR usage flag set, VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR, or VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR, then the viewType must be VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-04818", 38}, { {"If image was created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR usage flag set, VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR, or VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR, then the viewType must be VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 249}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-04970", 38}, { {"If image was created with VK_IMAGE_TYPE_3D and viewType is VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY then subresourceRange.levelCount must be 1", 154}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-04972", 38}, { {"If image was created with a samples value not equal to VK_SAMPLE_COUNT_1_BIT then viewType must be either VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 158}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-06723", 38}, { {"If image was created with VK_IMAGE_TYPE_3D but without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set then viewType must not be VK_IMAGE_VIEW_TYPE_2D_ARRAY", 152}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-06724", 38}, { {"If image is not a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT or VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT set, or viewType is not VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY, subresourceRange.baseArrayLayer must be less than the arrayLayers specified in VkImageCreateInfo when image was created", 323}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-06728", 38}, { {"If image was created with VK_IMAGE_TYPE_3D but without VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT or VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT set, then viewType must not be VK_IMAGE_VIEW_TYPE_2D", 193}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-07072", 38}, { {"If image was created with the VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT flag and format is a non-compressed format, the levelCount member of subresourceRange must be 1", 173}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08333", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_VIDEO_DECODE_OUTPUT_BIT_KHR", 155}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08334", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_VIDEO_DECODE_DPB_BIT_KHR", 152}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08335", 38}, { {"usage must not include VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08336", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_VIDEO_ENCODE_INPUT_BIT_KHR", 154}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08337", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_VIDEO_ENCODE_DPB_BIT_KHR", 152}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08338", 38}, { {"usage must not include VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08957", 38}, { {"If image has an QNX Screen external format, format must be VK_FORMAT_UNDEFINED", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08958", 38}, { {"If image has an QNX Screen external format, the pNext chain must include a VkSamplerYcbcrConversionInfo structure with a conversion object created with the same external format as image", 185}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-08959", 38}, { {"If image has an QNX Screen external format, all members of components must be the identity swizzle", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-09487", 38}, { {"If image was created with the VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT flag, the VkPhysicalDeviceMaintenance6Properties::blockTexelViewCompatibleMultipleLayers property is not VK_TRUE, and format is a non-compressed format, then the layerCount member of subresourceRange must be 1", 287}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-10261", 38}, { {"If image was created with the VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR usage flags set, then viewType must be VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 232}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-12397", 38}, { {"If image was not created with the VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT flag format must be identical to the format used to create image", 133}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-13357", 38}, { {"If image was created with VK_IMAGE_CREATE_ALIAS_SINGLE_LAYER_DESCRIPTOR_BIT_KHR, format must not be a multi-planar format", 121}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-image-parameter", 42}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-imageViewFormatReinterpretation-04466", 64}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::imageViewFormatReinterpretation is VK_FALSE, the VkFormat in format must not contain a different number of components, or a different number of bits in each component, than the format of the VkImage in image", 312}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-imageViewFormatSwizzle-04465", 55}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::imageViewFormatSwizzle is VK_FALSE, all elements of components must have the identity swizzle", 198}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-imageViewType-04973", 46}, { {"If viewType is VK_IMAGE_VIEW_TYPE_1D, VK_IMAGE_VIEW_TYPE_2D, or VK_IMAGE_VIEW_TYPE_3D; and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, then subresourceRange.layerCount must be 1", 195}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-imageViewType-04974", 46}, { {"If viewType is VK_IMAGE_VIEW_TYPE_1D, VK_IMAGE_VIEW_TYPE_2D, or VK_IMAGE_VIEW_TYPE_3D; and subresourceRange.layerCount is VK_REMAINING_ARRAY_LAYERS, then the remaining number of layers must be 1", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-invocationMask-04993", 47}, { {"If the invocationMask feature is enabled, and image was created with the VK_IMAGE_USAGE_INVOCATION_MASK_BIT_HUAWEI usage flag set, format must be VK_FORMAT_R8_UINT", 163}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-01585", 38}, { {"If a VkImageFormatListCreateInfo structure was included in the pNext chain of the VkImageCreateInfo structure used when creating image and VkImageFormatListCreateInfo::viewFormatCount is not zero then format must be one of the formats in VkImageFormatListCreateInfo::pViewFormats", 279}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-01970", 38}, { {"If the pNext chain includes a VkSamplerYcbcrConversionInfo structure with a conversion value other than VK_NULL_HANDLE, all members of components must have the identity swizzle", 176}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-02662", 38}, { {"If the pNext chain includes a VkImageViewUsageCreateInfo structure, and image was not created with a VkImageStencilUsageCreateInfo structure included in the pNext chain of VkImageCreateInfo, its usage member must not include any bits that were not set in the usage member of the VkImageCreateInfo structure used to create image", 327}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-02663", 38}, { {"If the pNext chain includes a VkImageViewUsageCreateInfo structure, image was created with a VkImageStencilUsageCreateInfo structure included in the pNext chain of VkImageCreateInfo, and subresourceRange.aspectMask includes VK_IMAGE_ASPECT_STENCIL_BIT, the usage member of the VkImageViewUsageCreateInfo structure must not include any bits that were not set in the usage member of the VkImageStencilUsageCreateInfo structure used to create image", 445}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-02664", 38}, { {"If the pNext chain includes a VkImageViewUsageCreateInfo structure, image was created with a VkImageStencilUsageCreateInfo structure included in the pNext chain of VkImageCreateInfo, and subresourceRange.aspectMask includes bits other than VK_IMAGE_ASPECT_STENCIL_BIT, the usage member of the VkImageViewUsageCreateInfo structure must not include any bits that were not set in the usage member of the VkImageCreateInfo structure used to create image", 449}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06658", 38}, { {"If the pNext chain includes a VkSamplerYcbcrConversionInfo structure with a conversion value other than VK_NULL_HANDLE, format must be the same used in VkSamplerYcbcrConversionCreateInfo::format", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06787", 38}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT", 159}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06944", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then textureSampleWeighted feature must be enabled", 127}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06945", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then image must have been created with the VK_IMAGE_USAGE_SAMPLE_WEIGHT_BIT_QCOM usage flag set", 172}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06946", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then components must be VK_COMPONENT_SWIZZLE_IDENTITY for all components", 149}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06947", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then subresourceRange.aspectMask must be VK_IMAGE_ASPECT_COLOR_BIT", 143}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06948", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then subresourceRange.levelCount must be 1", 119}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06949", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure, then viewType must be VK_IMAGE_VIEW_TYPE_1D_ARRAY or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 157}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06950", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and if viewType is VK_IMAGE_VIEW_TYPE_1D_ARRAY, then image must have been created with imageType VK_IMAGE_TYPE_1D", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06951", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and viewType is VK_IMAGE_VIEW_TYPE_1D_ARRAY, then subresourceRange.layerCount must be equal to 2", 172}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06952", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and viewType is VK_IMAGE_VIEW_TYPE_1D_ARRAY, then image must have been created with width equal to or greater than \\((numPhases \\times \\mathbin{max}\\left( \\mathbin{align}\\left(filterSize.width,4\\right), filterSize.height\\right))\\)", 315}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06953", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and if viewType is VK_IMAGE_VIEW_TYPE_2D_ARRAY, then image must have been created with imageType VK_IMAGE_TYPE_2D", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06954", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and viewType is VK_IMAGE_VIEW_TYPE_2D_ARRAY, then subresourceRange.layerCount must be equal or greater than numPhases", 193}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06955", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and viewType is VK_IMAGE_VIEW_TYPE_2D_ARRAY, then image must have been created with width equal to or greater than filterSize.width", 207}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06956", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure and viewType is VK_IMAGE_VIEW_TYPE_2D_ARRAY, then image must have been created with height equal to or greater than filterSize.height", 209}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-06957", 38}, { {"If the pNext chain includes VkImageViewSampleWeightCreateInfoQCOM structure then VkImageViewSampleWeightCreateInfoQCOM::filterSize.height must be less than or equal to VkPhysicalDeviceImageProcessingPropertiesQCOM::maxWeightFilterDimension.height", 246}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-08107", 38}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_IMAGE_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 164}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExportMetalObjectCreateInfoEXT, VkImageViewASTCDecodeModeEXT, VkImageViewMinLodCreateInfoEXT, VkImageViewSampleWeightCreateInfoQCOM, VkImageViewSlicedCreateInfoEXT, VkImageViewUsage2CreateInfoKHR, VkImageViewUsageCreateInfo, VkOpaqueCaptureDescriptorDataCreateInfoEXT, or VkSamplerYcbcrConversionInfo", 433}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO", 54}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subResourceRange-01021", 49}, { {"viewType must be compatible with the type of image as shown in the view type compatibility table", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-01478", 49}, { {"subresourceRange.baseMipLevel must be less than the mipLevels specified in VkImageCreateInfo when image was created", 115}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-01718", 49}, { {"If subresourceRange.levelCount is not VK_REMAINING_MIP_LEVELS, subresourceRange.baseMipLevel + subresourceRange.levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 220}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-02725", 49}, { {"If subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, image is a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT set, and viewType is VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY, subresourceRange.layerCount must be non-zero and subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the depth computed from baseMipLevel and extent.depth specified in VkImageCreateInfo when image was created, according to the formula defined in Image Mip Level Sizing", 521}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-06725", 49}, { {"If subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, image is not a 3D image created with VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT or VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT set, or viewType is not VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY, subresourceRange.layerCount must be non-zero and subresourceRange.baseArrayLayer + subresourceRange.layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 476}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-07818", 49}, { {"subresourceRange.aspectMask must only have at most 1 valid multi-planar aspect mask bit", 87}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-09594", 49}, { {"subresourceRange.aspectMask must be valid for the format the image was created with", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-subresourceRange-parameter", 53}, { {"subresourceRange must be a valid VkImageSubresourceRange structure", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-02274", 38}, { {"If usage contains VK_IMAGE_USAGE_SAMPLED_BIT, then the format features of the resultant image view must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT", 147}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-02275", 38}, { {"If usage contains VK_IMAGE_USAGE_STORAGE_BIT, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT", 132}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-02277", 38}, { {"If usage contains VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, then the image view's format features must contain VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-04550", 38}, { {"If the usage for the image view includes VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 207}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-04551", 38}, { {"If the usage for the image view includes VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR, and layeredShadingRateAttachments is VK_FALSE, subresourceRange.layerCount must be 1", 182}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-08931", 38}, { {"If usage contains VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, then the image view's format features must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 198}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-08932", 38}, { {"If usage contains VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT, and any of the following is true: the externalFormatResolve feature is not enabled the nullColorAttachmentWithExternalFormatResolve property is VK_FALSE image was created with an VkExternalFormatANDROID::externalFormat value of 0 then the image view's format features must contain at least one of VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT or VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 494}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-10259", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_2_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR", 192}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-usage-10260", 38}, { {"If usage contains VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR, then the image view's format features must contain VK_FORMAT_FEATURE_2_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR", 172}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-01004", 41}, { {"If the imageCubeArray feature is not enabled, viewType must not be VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-02960", 41}, { {"If viewType is VK_IMAGE_VIEW_TYPE_CUBE and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, subresourceRange.layerCount must be 6", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-02961", 41}, { {"If viewType is VK_IMAGE_VIEW_TYPE_CUBE_ARRAY and subresourceRange.layerCount is not VK_REMAINING_ARRAY_LAYERS, subresourceRange.layerCount must be a multiple of 6", 162}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-02962", 41}, { {"If viewType is VK_IMAGE_VIEW_TYPE_CUBE and subresourceRange.layerCount is VK_REMAINING_ARRAY_LAYERS, the remaining number of layers must be 6", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-02963", 41}, { {"If viewType is VK_IMAGE_VIEW_TYPE_CUBE_ARRAY and subresourceRange.layerCount is VK_REMAINING_ARRAY_LAYERS, the remaining number of layers must be a multiple of 6", 161}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewCreateInfo-viewType-parameter", 45}, { {"viewType must be a valid VkImageViewType value", 46}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-commonparent", 42}, { {"Both of imageView, and sampler that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 150}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-descriptorType-02654", 50}, { {"descriptorType must be VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER", 135}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-descriptorType-parameter", 54}, { {"descriptorType must be a valid VkDescriptorType value", 53}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-imageView-02656", 45}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, the image that imageView was created from must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT usage flag set", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-imageView-02657", 45}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, the image that imageView was created from must have been created with the VK_IMAGE_USAGE_STORAGE_BIT usage flag set", 170}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-imageView-parameter", 49}, { {"imageView must be a valid VkImageView handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_HANDLE_INFO_NVX", 58}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-sampler-02655", 43}, { {"sampler must be a valid VkSampler if descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewHandleInfoNVX-sampler-parameter", 47}, { {"If sampler is not VK_NULL_HANDLE, sampler must be a valid VkSampler handle", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewMinLodCreateInfoEXT-minLod-06455", 48}, { {"If the minLod feature is not enabled, minLod must be 0.0", 56}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewMinLodCreateInfoEXT-minLod-06456", 48}, { {"minLod must be less or equal to the index of the last mipmap level accessible to the view", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewMinLodCreateInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_MIN_LOD_CREATE_INFO_EXT", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-filterCenter-06960", 61}, { {"filterCenter.x must be less than or equal to (filterSize.width - 1)", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-filterCenter-06961", 61}, { {"filterCenter.y must be less than or equal to (filterSize.height - 1)", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-filterSize-06958", 59}, { {"filterSize.width must be less than or equal to VkPhysicalDeviceImageProcessingPropertiesQCOM::maxWeightFilterDimension.width", 124}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-filterSize-06959", 59}, { {"filterSize.height must be less than or equal to VkPhysicalDeviceImageProcessingPropertiesQCOM::maxWeightFilterDimension.height", 126}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-numPhases-06962", 58}, { {"numPhases must be a power of two squared value (i.e., 1, 4, 16, 64, 256, etc.)", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-numPhases-06963", 58}, { {"numPhases must be less than or equal to VkPhysicalDeviceImageProcessingPropertiesQCOM::maxWeightFilterPhases", 108}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSampleWeightCreateInfoQCOM-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_SAMPLE_WEIGHT_CREATE_INFO_QCOM", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-None-07870", 46}, { {"The image view must reference exactly 1 mip level", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-None-07871", 46}, { {"The imageSlicedViewOf3D feature must be enabled", 47}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-image-07869", 47}, { {"image must have been created with imageType equal to VK_IMAGE_TYPE_3D", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_SLICED_CREATE_INFO_EXT", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-sliceCount-07868", 52}, { {"If sliceCount is not VK_REMAINING_3D_SLICES_EXT, it must be non-zero and sliceOffset + sliceCount must be less than or equal to the effective view depth as specified in Image Mip Level Sizing", 191}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-sliceOffset-07867", 53}, { {"sliceOffset must be less than the effective view depth as specified in Image Mip Level Sizing", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewSlicedCreateInfoEXT-viewType-07909", 50}, { {"viewType must be VK_IMAGE_VIEW_TYPE_3D", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsage2CreateInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_2_CREATE_INFO_KHR", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsage2CreateInfoKHR-usage-parameter", 51}, { {"usage must be a valid combination of VkImageUsageFlagBits2KHR values", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsage2CreateInfoKHR-usage-requiredbitmask", 57}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsageCreateInfo-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_CREATE_INFO", 60}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsageCreateInfo-usage-parameter", 47}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImageViewUsageCreateInfo-usage-requiredbitmask", 53}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportAndroidHardwareBufferInfoANDROID-buffer-01881", 58}, { {"If buffer is not NULL, it must be a valid Android hardware buffer object with AHardwareBuffer_Desc::usage compatible with Vulkan as described in Android Hardware Buffers", 169}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportAndroidHardwareBufferInfoANDROID-buffer-09863", 58}, { {"If buffer is not NULL, Android hardware buffers must be supported for import, as reported by VkExternalTensorPropertiesARM, VkExternalImageFormatProperties, or VkExternalBufferProperties", 186}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportAndroidHardwareBufferInfoANDROID-buffer-parameter", 62}, { {"buffer must be a valid pointer to an AHardwareBuffer value", 58}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportAndroidHardwareBufferInfoANDROID-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_ANDROID_HARDWARE_BUFFER_INFO_ANDROID", 75}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportFenceFdInfoKHR-fd-01541", 36}, { {"fd must obey any requirements listed for handleType in external fence handle types compatibility", 96}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-fence-parameter", 43}, { {"fence must be a valid VkFence handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-flags-parameter", 43}, { {"flags must be a valid combination of VkFenceImportFlagBits values", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-handleType-01464", 44}, { {"handleType must be a value included in the Handle Types Supported by VkImportFenceFdInfoKHR table", 97}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-handleType-07306", 44}, { {"If handleType refers to a handle type with copy payload transference semantics, flags must contain VK_FENCE_IMPORT_TEMPORARY_BIT", 128}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-handleType-parameter", 48}, { {"handleType must be a valid VkExternalFenceHandleTypeFlagBits value", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceFdInfoKHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_FENCE_FD_INFO_KHR", 56}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-fence-parameter", 52}, { {"fence must be a valid VkFence handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-flags-parameter", 52}, { {"flags must be a valid combination of VkFenceImportFlagBits values", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handle-01462", 49}, { {"If handle is not NULL, name must be NULL", 40}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handle-01539", 49}, { {"If handle is not NULL, it must obey any requirements listed for handleType in external fence handle types compatibility", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handleType-01457", 53}, { {"handleType must be a value included in the Handle Types Supported by VkImportFenceWin32HandleInfoKHR table", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handleType-01459", 53}, { {"If handleType is not VK_EXTERNAL_FENCE_HANDLE_TYPE_OPAQUE_WIN32_BIT, name must be NULL", 86}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handleType-01460", 53}, { {"If handle is NULL, name must name a valid synchronization primitive of the type specified by handleType", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-handleType-01461", 53}, { {"If name is NULL, handle must be a valid handle of the type specified by handleType", 82}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-name-01540", 47}, { {"If name is not NULL, it must obey any requirements listed for handleType in external fence handle types compatibility", 117}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportFenceWin32HandleInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_FENCE_WIN32_HANDLE_INFO_KHR", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportMemoryBufferCollectionFUCHSIA-collection-parameter", 63}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportMemoryBufferCollectionFUCHSIA-index-06406", 54}, { {"index must be less than the value retrieved as VkBufferCollectionPropertiesFUCHSIA::bufferCount", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportMemoryBufferCollectionFUCHSIA-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_BUFFER_COLLECTION_FUCHSIA", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportMemoryFdInfoKHR-fd-00668", 37}, { {"The memory from which fd was exported must have been created on the same underlying physical device as device", 109}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-fd-01520", 37}, { {"fd must obey any requirements listed for handleType in external memory handle types compatibility", 97}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-fd-01746", 37}, { {"The memory represented by fd must have been created from a physical device and driver that is compatible with device and handleType, as described in external memory handle types compatibility", 191}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-handleType-00669", 45}, { {"If handleType is not 0, it must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT", 129}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-handleType-00670", 45}, { {"If handleType is not 0, fd must be a valid handle of the type specified by handleType", 85}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-handleType-09862", 45}, { {"If handleType is not 0, it must be supported for import, as reported by VkExternalTensorPropertiesARM, VkExternalImageFormatProperties or VkExternalBufferProperties", 164}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-handleType-parameter", 49}, { {"If handleType is not 0, handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryFdInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_FD_INFO_KHR", 57}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-handleType-01747", 54}, { {"If handleType is not 0, it must be supported for import, as reported in VkExternalMemoryProperties", 98}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-handleType-01748", 54}, { {"If handleType is not 0, it must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT", 158}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-handleType-01750", 54}, { {"If handleType is VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT, pHostPointer must be a pointer to allocationSize number of bytes of host memory, where allocationSize is the member of the VkMemoryAllocateInfo structure this structure is chained to", 255}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-handleType-01751", 54}, { {"If handleType is VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT, pHostPointer must be a pointer to allocationSize number of bytes of host mapped foreign memory, where allocationSize is the member of the VkMemoryAllocateInfo structure this structure is chained to", 281}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-handleType-parameter", 58}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-pHostPointer-01749", 56}, { {"pHostPointer must be a pointer aligned to an integer multiple of VkPhysicalDeviceExternalMemoryHostPropertiesEXT::minImportedHostPointerAlignment", 145}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-pHostPointer-parameter", 60}, { {"pHostPointer must be a pointer value", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryHostPointerInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_HOST_POINTER_INFO_EXT", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handle-10409", 50}, { {"The memory from which handle was exported must have been created on the same underlying physical device as device", 113}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handle-10412", 50}, { {"handle must obey any requirements listed for handleType in external memory handle types compatibility", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handleType-10408", 54}, { {"If handleType is not 0, it must be supported for import, as reported by VkExternalImageFormatProperties or VkExternalBufferProperties", 133}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handleType-10410", 54}, { {"If handleType is not 0, it must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLBUFFER_BIT_EXT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLTEXTURE_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLHEAP_BIT_EXT", 184}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handleType-10411", 54}, { {"If handleType is not 0 , handle must be a valid non-NULL handle of the type specified by handleType", 99}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-handleType-parameter", 58}, { {"If handleType is not 0, handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryMetalHandleInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_METAL_HANDLE_INFO_EXT", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handle-00659", 50}, { {"The memory from which handle was exported, or the memory named by name must have been created on the same underlying physical device as device", 142}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handle-01441", 50}, { {"If handle is not NULL, name must be NULL", 40}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handle-01518", 50}, { {"If handle is not NULL, it must obey any requirements listed for handleType in external memory handle types compatibility", 120}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-00660", 54}, { {"If handleType is not 0, it must be defined as an NT handle or a global share handle", 83}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-00661", 54}, { {"If handleType is not 0 and name is NULL, handle must be a valid handle of the type specified by handleType", 106}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-01439", 54}, { {"If handleType is not VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_HEAP_BIT, or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_RESOURCE_BIT, name must be NULL", 238}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-01440", 54}, { {"If handleType is not 0 and handle is NULL, name must name a valid memory resource of the type specified by handleType", 117}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-09861", 54}, { {"If handleType is not 0, it must be supported for import, as reported by VkExternalTensorPropertiesARM, VkExternalImageFormatProperties, or VkExternalBufferProperties", 165}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-handleType-parameter", 58}, { {"If handleType is not 0, handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-name-01519", 48}, { {"If name is not NULL, it must obey any requirements listed for handleType in external memory handle types compatibility", 118}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_KHR", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoNV-handle-01328", 49}, { {"handle must be a valid handle to memory, obtained as specified by handleType", 76}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoNV-handleType-01327", 53}, { {"handleType must not have more than one bit set", 46}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoNV-handleType-parameter", 57}, { {"handleType must be a valid combination of VkExternalMemoryHandleTypeFlagBitsNV values", 85}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryWin32HandleInfoNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_NV", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryZirconHandleInfoFUCHSIA-handle-04772", 55}, { {"handle must be a valid VMO handle", 33}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryZirconHandleInfoFUCHSIA-handleType-04771", 59}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA", 72}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryZirconHandleInfoFUCHSIA-handleType-parameter", 63}, { {"If handleType is not 0, handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMemoryZirconHandleInfoFUCHSIA-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_MEMORY_ZIRCON_HANDLE_INFO_FUCHSIA", 72}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMetalBufferInfoEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_METAL_BUFFER_INFO_EXT", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMetalIOSurfaceInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_METAL_IO_SURFACE_INFO_EXT", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMetalSharedEventInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_METAL_SHARED_EVENT_INFO_EXT", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMetalTextureInfoEXT-plane-parameter", 48}, { {"plane must be a valid VkImageAspectFlagBits value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportMetalTextureInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_METAL_TEXTURE_INFO_EXT", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportNativeBufferInfoOHOS-buffer-parameter", 50}, { {"buffer must be a valid pointer to an OH_NativeBuffer value", 58}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportNativeBufferInfoOHOS-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_NATIVE_BUFFER_INFO_OHOS", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkImportScreenBufferInfoQNX-buffer-08966", 45}, { {"If buffer is not NULL, QNX Screen Buffers must be supported for import, as reported by VkExternalImageFormatProperties or VkExternalBufferProperties", 148}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportScreenBufferInfoQNX-buffer-08967", 45}, { {"buffer is not NULL, it must be a pointer to valid QNX Screen buffer", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportScreenBufferInfoQNX-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_SCREEN_BUFFER_INFO_QNX", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-fd-01544", 40}, { {"fd must obey any requirements listed for handleType in external semaphore handle types compatibility", 100}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-flags-03323", 43}, { {"If flags contains VK_SEMAPHORE_IMPORT_TEMPORARY_BIT, the VkSemaphoreTypeCreateInfo::semaphoreType field of the semaphore from which fd was exported must not be VK_SEMAPHORE_TYPE_TIMELINE", 186}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-flags-parameter", 47}, { {"flags must be a valid combination of VkSemaphoreImportFlagBits values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-handleType-01143", 48}, { {"handleType must be a value included in the Handle Types Supported by VkImportSemaphoreFdInfoKHR table", 101}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-handleType-03263", 48}, { {"If handleType is VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT, the VkSemaphoreCreateInfo::flags field must match that of the semaphore from which fd was exported", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-handleType-03264", 48}, { {"If handleType is VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT, the VkSemaphoreTypeCreateInfo::semaphoreType field must match that of the semaphore from which fd was exported", 176}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-handleType-07307", 48}, { {"If handleType refers to a handle type with copy payload transference semantics, flags must contain VK_SEMAPHORE_IMPORT_TEMPORARY_BIT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-handleType-parameter", 52}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_FD_INFO_KHR", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreFdInfoKHR-semaphore-parameter", 51}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-flags-03322", 52}, { {"If flags contains VK_SEMAPHORE_IMPORT_TEMPORARY_BIT, the VkSemaphoreTypeCreateInfo::semaphoreType field of the semaphore from which handle or name was exported must not be VK_SEMAPHORE_TYPE_TIMELINE", 198}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-flags-parameter", 56}, { {"flags must be a valid combination of VkSemaphoreImportFlagBits values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handle-01469", 53}, { {"If handle is not NULL, name must be NULL", 40}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handle-01542", 53}, { {"If handle is not NULL, it must obey any requirements listed for handleType in external semaphore handle types compatibility", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-01140", 57}, { {"handleType must be a value included in the Handle Types Supported by VkImportSemaphoreWin32HandleInfoKHR table", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-01466", 57}, { {"If handleType is not VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_BIT or VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_D3D12_FENCE_BIT, name must be NULL", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-01467", 57}, { {"If handle is NULL, name must name a valid synchronization primitive of the type specified by handleType", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-01468", 57}, { {"If name is NULL, handle must be a valid handle of the type specified by handleType", 82}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-03261", 57}, { {"If handleType is VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_BIT or VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT, the VkSemaphoreCreateInfo::flags field must match that of the semaphore from which handle or name was exported", 237}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-handleType-03262", 57}, { {"If handleType is VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_BIT or VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT, the VkSemaphoreTypeCreateInfo::semaphoreType field must match that of the semaphore from which handle or name was exported", 249}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-name-01543", 51}, { {"If name is not NULL, it must obey any requirements listed for handleType in external semaphore handle types compatibility", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreWin32HandleInfoKHR-semaphore-parameter", 60}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-flags-parameter", 61}, { {"flags must be a valid combination of VkSemaphoreImportFlagBits values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-handleType-04765", 62}, { {"handleType must be a value included in the Handle Types Supported by VkImportSemaphoreZirconHandleInfoFUCHSIA table", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-handleType-parameter", 66}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-pNext-pNext", 57}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_ZIRCON_HANDLE_INFO_FUCHSIA", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-semaphore-parameter", 65}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-semaphoreType-04768", 65}, { {"The VkSemaphoreTypeCreateInfo::semaphoreType field must not be VK_SEMAPHORE_TYPE_TIMELINE", 89}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-zirconHandle-04766", 64}, { {"zirconHandle must obey any requirements listed for handleType in external semaphore handle types compatibility", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkImportSemaphoreZirconHandleInfoFUCHSIA-zirconHandle-04767", 64}, { {"zirconHandle must have ZX_RIGHTS_BASIC and ZX_RIGHTS_SIGNAL rights", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkIndirectCommandsExecutionSetTokenEXT-shaderStages-11137", 62}, { {"Each bit in shaderStages must be supported by VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesPipelineBinding or VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesShaderBinding", 260}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsExecutionSetTokenEXT-shaderStages-parameter", 66}, { {"shaderStages must be a valid combination of VkShaderStageFlagBits values", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsExecutionSetTokenEXT-shaderStages-requiredbitmask", 72}, { {"shaderStages must not be 0", 26}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsExecutionSetTokenEXT-type-parameter", 58}, { {"type must be a valid VkIndirectExecutionSetInfoTypeEXT value", 60}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsIndexBufferTokenEXT-mode-11135", 53}, { {"mode must be non-zero", 21}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsIndexBufferTokenEXT-mode-11136", 53}, { {"mode must be one of the bits supported in VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsInputModes", 131}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsIndexBufferTokenEXT-mode-parameter", 57}, { {"mode must be a valid VkIndirectCommandsInputModeFlagBitsEXT value", 65}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-flags-parameter", 58}, { {"flags must be a valid combination of VkIndirectCommandsLayoutUsageFlagBitsEXT values", 84}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-indirectStride-11090", 63}, { {"indirectStride must be less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectCommandsIndirectStride", 132}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pNext-pNext", 54}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11093", 56}, { {"The number of tokens in the pTokens array with type equal to VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT must be less than or equal to 1", 142}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11094", 56}, { {"The number of tokens in the pTokens array with type equal to VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT must be less than or equal to 1", 207}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11095", 56}, { {"If the action command token in the pTokens array is not an indexed draw token, then pTokens must not contain a member with type set to VK_INDIRECT_COMMANDS_TOKEN_TYPE_INDEX_BUFFER_EXT", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11096", 56}, { {"If the action command token in the pTokens array is not a non-mesh draw token, then pTokens must not contain a member with type set to VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_EXT", 184}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11097", 56}, { {"If the pTokens array contains multiple tokens with type equal to VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_EXT, then there must be no duplicate VkIndirectCommandsVertexBufferTokenEXT::vertexBindingUnit values", 212}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11099", 56}, { {"For all VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT, and VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT type tokens in pTokens, there must be no overlapping ranges between any specified push constant ranges", 325}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11100", 56}, { {"The action command token must be the last token in the pTokens array", 68}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11101", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT and the dynamicGeneratedPipelineLayout feature is not enabled, then the pipelineLayout must not be VK_NULL_HANDLE", 256}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11102", 56}, { {"For any element of pTokens, if type is either VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT and pipelineLayout is VK_NULL_HANDLE, then the pNext chain must include a VkPipelineLayoutCreateInfo struct", 257}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11103", 56}, { {"For any element of pTokens, the offset must be greater than or equal to the offset member of the previous tokens", 112}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11104", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_INDEX_BUFFER_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_NV_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_INDEXED_COUNT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_INDEXED_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_COUNT_EXT, or VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_EXT, then shaderStages must contain graphics stages", 615}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11105", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_EXT, then shaderStages must be VK_SHADER_STAGE_COMPUTE_BIT", 138}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11106", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_EXT, then shaderStages must contain VK_SHADER_STAGE_MESH_BIT_EXT", 212}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11107", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_NV_EXT, then the shaderStages must contain VK_SHADER_STAGE_MESH_BIT_NV", 221}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11108", 56}, { {"For any element of pTokens, if type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_TRACE_RAYS2_EXT, then shaderStages must contain ray tracing stages", 137}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11139", 56}, { {"If the pTokens array contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, then this token must be the first token in the array", 141}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-11145", 56}, { {"The number of tokens in the pTokens array with type equal to VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT must be less than or equal to 1", 143}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pTokens-parameter", 60}, { {"pTokens must be a valid pointer to an array of tokenCount valid VkIndirectCommandsLayoutTokenEXT structures", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-pipelineLayout-parameter", 67}, { {"If pipelineLayout is not VK_NULL_HANDLE, pipelineLayout must be a valid VkPipelineLayout handle", 95}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_EXT", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11091", 61}, { {"shaderStages must only contain stages supported by VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStages", 142}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11109", 61}, { {"If shaderStages contains graphics stages then the state tokens in pTokens must not include VK_INDIRECT_COMMANDS_TOKEN_TYPE_TRACE_RAYS2_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_EXT", 184}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11110", 61}, { {"If shaderStages is VK_SHADER_STAGE_COMPUTE_BIT then the state tokens in pTokens must only include VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT, or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT", 408}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11111", 61}, { {"If shaderStages contains ray tracing stages then the state tokens in pTokens must only include VK_INDIRECT_COMMANDS_TOKEN_TYPE_TRACE_RAYS2_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT, or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT", 408}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11112", 61}, { {"The shaderStages must only contain stages from one of the following: VK_SHADER_STAGE_ALL (if the descriptorHeap feature is enabled) graphics stages VK_SHADER_STAGE_COMPUTE_BIT mesh stages and VK_SHADER_STAGE_FRAGMENT_BIT ray tracing stages", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-11113", 61}, { {"If shaderStages contains VK_SHADER_STAGE_FRAGMENT_BIT, then shaderStages must also contain VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT", 149}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-parameter", 65}, { {"shaderStages must be a valid combination of VkShaderStageFlagBits values", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-shaderStages-requiredbitmask", 71}, { {"shaderStages must not be 0", 26}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-tokenCount-11092", 59}, { {"tokenCount must be less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectCommandsTokenCount", 124}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoEXT-tokenCount-arraylength", 65}, { {"tokenCount must be greater than 0", 33}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-flags-parameter", 57}, { {"flags must be a valid combination of VkIndirectCommandsLayoutUsageFlagBitsNV values", 83}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pStreamStrides-02937", 62}, { {"each element of pStreamStrides must be greater than 0 and less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectCommandsStreamStride. Furthermore the alignment of each token input must be ensured", 227}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pStreamStrides-parameter", 66}, { {"pStreamStrides must be a valid pointer to an array of streamCount uint32_t values", 81}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-02932", 55}, { {"If pTokens contains an entry of VK_INDIRECT_COMMANDS_TOKEN_TYPE_SHADER_GROUP_NV it must be the first element of the array and there must be only a single element of such token type", 180}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-02933", 55}, { {"If pTokens contains an entry of VK_INDIRECT_COMMANDS_TOKEN_TYPE_STATE_FLAGS_NV there must be only a single element of such token type", 133}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-02934", 55}, { {"All state tokens in pTokens must occur before any action command tokens (VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_NV, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_INDEXED_NV, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_TASKS_NV, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV , VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_NV )", 308}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-02935", 55}, { {"The content of pTokens must include one single action command token that is compatible with the pipelineBindPoint", 113}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-09585", 55}, { {"If pTokens contains an entry of VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV it must be the first element of the array and there must be only a single element of such token type", 176}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pTokens-parameter", 59}, { {"pTokens must be a valid pointer to an array of tokenCount valid VkIndirectCommandsLayoutTokenNV structures", 106}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pipelineBindPoint-02930", 65}, { {"The pipelineBindPoint must be VK_PIPELINE_BIND_POINT_GRAPHICS or VK_PIPELINE_BIND_POINT_COMPUTE", 95}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pipelineBindPoint-09088", 65}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_COMPUTE then the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedCompute feature must be enabled", 168}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pipelineBindPoint-09089", 65}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_COMPUTE then the state tokens in pTokens must only include VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_NV, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_NV, or VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV", 293}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pipelineBindPoint-09090", 65}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_COMPUTE and pTokens includes VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV, then the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 243}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-pipelineBindPoint-parameter", 69}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_NV", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-streamCount-02936", 59}, { {"streamCount must be greater than 0 and less or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectCommandsStreamCount", 139}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-streamCount-arraylength", 65}, { {"streamCount must be greater than 0", 34}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-tokenCount-02931", 58}, { {"tokenCount must be greater than 0 and less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectCommandsTokenCount", 142}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutCreateInfoNV-tokenCount-arraylength", 64}, { {"tokenCount must be greater than 0", 33}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutPushDataTokenNV-pushDataOffset-11335", 65}, { {"The sum of pushDataOffset and pushDataSize must be less than maxPushDataSize", 76}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutPushDataTokenNV-pushDataOffset-11420", 65}, { {"pushDataOffset must be a multiple of 4", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutPushDataTokenNV-pushDataSize-11421", 63}, { {"pushDataSize must be a multiple of 4", 36}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutPushDataTokenNV-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_PUSH_DATA_TOKEN_NV", 75}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-descriptorHeap-11332", 58}, { {"If the descriptorHeap feature is not enabled, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT", 172}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-deviceGeneratedCommandsMultiDrawIndirectCount-11129", 89}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::deviceGeneratedCommandsMultiDrawIndirectCount is not supported, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_INDEXED_COUNT_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_COUNT_EXT", 242}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-deviceGeneratedCommandsMultiDrawIndirectCount-11130", 89}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::deviceGeneratedCommandsMultiDrawIndirectCount is not supported, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_EXT", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-deviceGeneratedCommandsMultiDrawIndirectCount-11131", 89}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::deviceGeneratedCommandsMultiDrawIndirectCount is not supported, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_NV_EXT", 198}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-meshShader-11126", 54}, { {"If meshShader or taskShader are not enabled, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_EXT VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_COUNT_NV_EXT", 291}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-offset-11124", 50}, { {"offset must be less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectCommandsTokenOffset", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-offset-11125", 50}, { {"offset must be aligned to 4", 27}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-pExecutionSet-parameter", 61}, { {"If type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT, the pExecutionSet member of data must be a valid pointer to a valid VkIndirectCommandsExecutionSetTokenEXT structure", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-pIndexBuffer-parameter", 60}, { {"If type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_INDEX_BUFFER_EXT, the pIndexBuffer member of data must be a valid pointer to a valid VkIndirectCommandsIndexBufferTokenEXT structure", 175}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-pPushConstant-parameter", 61}, { {"If type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT, VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT, or VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT, the pPushConstant member of data must be a valid pointer to a valid VkIndirectCommandsPushConstantTokenEXT structure", 342}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-pVertexBuffer-parameter", 61}, { {"If type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_EXT, the pVertexBuffer member of data must be a valid pointer to a valid VkIndirectCommandsVertexBufferTokenEXT structure", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-rayTracingMaintenance1-11128", 66}, { {"If the rayTracingMaintenance1 feature is not enabled, type must not be VK_INDIRECT_COMMANDS_TOKEN_TYPE_TRACE_RAYS2_EXT", 118}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_EXT", 66}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-type-11333", 48}, { {"If type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT VkIndirectCommandsPushConstantTokenEXT::updateRange.shaderStages must be VK_SHADER_STAGE_ALL", 213}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenEXT-type-parameter", 52}, { {"type must be a valid VkIndirectCommandsTokenTypeEXT value", 57}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-indirectStateFlags-parameter", 65}, { {"indirectStateFlags must be a valid combination of VkIndirectStateFlagBitsNV values", 82}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-offset-02952", 49}, { {"offset must be less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::maxIndirectCommandsTokenOffset", 120}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-offset-06888", 49}, { {"offset must be aligned to the scalar alignment of tokenType or minIndirectCommandsBufferOffsetAlignment, whichever is lower", 123}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-pIndexTypeValues-parameter", 63}, { {"If indexTypeCount is not 0, pIndexTypeValues must be a valid pointer to an array of indexTypeCount uint32_t values", 114}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-pIndexTypes-parameter", 58}, { {"If indexTypeCount is not 0, pIndexTypes must be a valid pointer to an array of indexTypeCount valid VkIndexType values", 118}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-pNext-pNext", 48}, { {"pNext must be NULL or a pointer to a valid instance of VkIndirectCommandsLayoutPushDataTokenNV", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-pushconstantPipelineLayout-parameter", 73}, { {"If pushconstantPipelineLayout is not VK_NULL_HANDLE, pushconstantPipelineLayout must be a valid VkPipelineLayout handle", 119}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-pushconstantShaderStageFlags-parameter", 75}, { {"pushconstantShaderStageFlags must be a valid combination of VkShaderStageFlagBits values", 88}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_NV", 65}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-sType-unique", 49}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-stream-02951", 49}, { {"stream must be smaller than VkIndirectCommandsLayoutCreateInfoNV::streamCount", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02976", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_NV, vertexBindingUnit must stay within device supported limits for the appropriate commands", 153}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02977", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, pushconstantPipelineLayout must be valid", 106}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02978", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, pushconstantOffset must be a multiple of 4", 108}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02979", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, pushconstantSize must be a multiple of 4", 106}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02980", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, pushconstantOffset must be less than VkPhysicalDeviceLimits::maxPushConstantsSize", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02981", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, pushconstantSize must be less than or equal to VkPhysicalDeviceLimits::maxPushConstantsSize minus pushconstantOffset", 182}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02982", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, for each byte in the range specified by pushconstantOffset and pushconstantSize and for each shader stage in pushconstantShaderStageFlags, there must be a push constant range in pushconstantPipelineLayout that includes that byte and that stage", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02983", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV, for each byte in the range specified by pushconstantOffset and pushconstantSize and for each push constant range that overlaps that byte, pushconstantShaderStageFlags must include all stages in that push constant range's VkPushConstantRange::stageFlags", 318}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-02984", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_STATE_FLAGS_NV, indirectStateFlags must not be 0", 96}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-11334", 52}, { {"If tokenType is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_NV, VkIndirectCommandsLayoutPushDataTokenNV::pushDataSize must be greater than 0", 138}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsLayoutTokenNV-tokenType-parameter", 56}, { {"tokenType must be a valid VkIndirectCommandsTokenTypeNV value", 61}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsPushConstantTokenEXT-size-11133", 54}, { {"If the token type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_DATA_SEQUENCE_INDEX_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT, the size member of updateRange must be 4", 177}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsPushConstantTokenEXT-updateRange-11132", 61}, { {"If the token type is VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_EXT or VK_INDIRECT_COMMANDS_TOKEN_TYPE_SEQUENCE_INDEX_EXT, updateRange must be contained within the push constant info used by VkIndirectCommandsLayoutCreateInfoEXT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsPushConstantTokenEXT-updateRange-parameter", 65}, { {"updateRange must be a valid VkPushConstantRange structure", 57}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsStreamNV-buffer-02942", 44}, { {"The buffer's usage flag must have the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set", 81}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsStreamNV-buffer-02975", 44}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsStreamNV-buffer-parameter", 48}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsStreamNV-offset-02943", 44}, { {"The offset must be aligned to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV::minIndirectCommandsBufferOffsetAlignment", 123}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectCommandsVertexBufferTokenEXT-vertexBindingUnit-11134", 67}, { {"vertexBindingUnit must be less than the total number of vertex input bindings in use by the current graphics state", 114}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetCreateInfoEXT-maxIndirectShaderObjectCount-11014", 75}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectShaderObjectCount is zero or the shaderObject feature is not enabled type must not be VK_INDIRECT_EXECUTION_SET_INFO_TYPE_SHADER_OBJECTS_EXT", 208}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetCreateInfoEXT-pPipelineInfo-parameter", 64}, { {"If type is VK_INDIRECT_EXECUTION_SET_INFO_TYPE_PIPELINES_EXT, the pPipelineInfo member of info must be a valid pointer to a valid VkIndirectExecutionSetPipelineInfoEXT structure", 177}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetCreateInfoEXT-pShaderInfo-parameter", 62}, { {"If type is VK_INDIRECT_EXECUTION_SET_INFO_TYPE_SHADER_OBJECTS_EXT, the pShaderInfo member of info must be a valid pointer to a valid VkIndirectExecutionSetShaderInfoEXT structure", 178}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetCreateInfoEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_CREATE_INFO_EXT", 70}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetCreateInfoEXT-type-parameter", 55}, { {"type must be a valid VkIndirectExecutionSetInfoTypeEXT value", 60}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-initialPipeline-11019", 64}, { {"initialPipeline must not use descriptors of type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 135}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-initialPipeline-11153", 64}, { {"initialPipeline must have been created with VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT", 90}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-initialPipeline-parameter", 68}, { {"initialPipeline must be a valid VkPipeline handle", 49}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-maxPipelineCount-11018", 65}, { {"maxPipelineCount must be between 1 and VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectPipelineCount", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_PIPELINE_INFO_EXT", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-supportedIndirectCommandsShaderStagesPipelineBinding-11015", 101}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesPipelineBinding does not contain VK_SHADER_STAGE_COMPUTE_BIT, the VkPipelineBindPoint of initialPipeline must not be VK_PIPELINE_BIND_POINT_COMPUTE", 241}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-supportedIndirectCommandsShaderStagesPipelineBinding-11016", 101}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesPipelineBinding does not contain VK_SHADER_STAGE_FRAGMENT_BIT, the VkPipelineBindPoint of initialPipeline must not be VK_PIPELINE_BIND_POINT_GRAPHICS", 243}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetPipelineInfoEXT-supportedIndirectCommandsShaderStagesPipelineBinding-11017", 101}, { {"If VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesPipelineBinding does not contain ray tracing stages, the VkPipelineBindPoint of initialPipeline must not be VK_PIPELINE_BIND_POINT_RAY_TRACING_KHR", 240}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-maxShaderCount-11021", 61}, { {"maxShaderCount must not be zero", 31}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-maxShaderCount-11022", 61}, { {"maxShaderCount must be less than or equal to VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::maxIndirectShaderObjectCount", 127}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-maxShaderCount-11036", 61}, { {"maxShaderCount must be greater than or equal to shaderCount", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-11020", 62}, { {"All members of pInitialShaders must have a stage supported by VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesShaderBinding", 166}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-11154", 62}, { {"Each member of pInitialShaders must have been created with VK_SHADER_CREATE_INDIRECT_BINDABLE_BIT_EXT", 101}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-11321", 62}, { {"If any element of pInitialShaders was created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, all elements of pInitialShaders must have been created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 193}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-11322", 62}, { {"If any element of pInitialShaders was created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, all elements of pInitialShaders must have been created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 199}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-11323", 62}, { {"If elements of pInitialShaders were created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, pSetLayoutInfos must be NULL", 119}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pInitialShaders-parameter", 66}, { {"pInitialShaders must be a valid pointer to an array of shaderCount valid VkShaderEXT handles", 92}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pPushConstantRanges-parameter", 70}, { {"If pushConstantRangeCount is not 0, pPushConstantRanges must be a valid pointer to an array of pushConstantRangeCount valid VkPushConstantRange structures", 154}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pSetLayoutInfos-10929", 62}, { {"If pSetLayoutInfos is not NULL, the descriptor layout values specified must be compatible with the descriptor set layouts defined at the creation of the shader object", 166}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-pSetLayoutInfos-parameter", 66}, { {"If pSetLayoutInfos is not NULL, pSetLayoutInfos must be a valid pointer to an array of shaderCount valid VkIndirectExecutionSetShaderLayoutInfoEXT structures", 157}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_SHADER_INFO_EXT", 70}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-shaderCount-arraylength", 64}, { {"shaderCount must be greater than 0", 34}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderInfoEXT-stage-11023", 52}, { {"The stage of each element in the pInitialShaders array must be unique", 69}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderLayoutInfoEXT-pSetLayouts-11024", 64}, { {"All members of pSetLayouts must not contain descriptors of type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 150}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderLayoutInfoEXT-pSetLayouts-parameter", 68}, { {"If setLayoutCount is not 0, pSetLayouts must be a valid pointer to an array of setLayoutCount valid or VK_NULL_HANDLE VkDescriptorSetLayout handles", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkIndirectExecutionSetShaderLayoutInfoEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_SHADER_LAYOUT_INFO_EXT", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkInitializePerformanceApiInfoINTEL-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkInitializePerformanceApiInfoINTEL-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_INITIALIZE_PERFORMANCE_API_INFO_INTEL", 69}, { "chapters/queries.html", 21} } },
        { {"VUID-VkInputAttachmentAspectReference-aspectMask-01964", 54}, { {"aspectMask must not include VK_IMAGE_ASPECT_METADATA_BIT", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkInputAttachmentAspectReference-aspectMask-02250", 54}, { {"aspectMask must not include VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT for any index i", 82}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkInputAttachmentAspectReference-aspectMask-parameter", 58}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkInputAttachmentAspectReference-aspectMask-requiredbitmask", 64}, { {"aspectMask must not be 0", 24}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkInstanceCreateInfo-flags-06559", 37}, { {"If flags has the VK_INSTANCE_CREATE_ENUMERATE_PORTABILITY_BIT_KHR bit set, the list of enabled extensions in ppEnabledExtensionNames must contain VK_KHR_portability_enumeration", 176}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-flags-parameter", 41}, { {"flags must be a valid combination of VkInstanceCreateFlagBits values", 68}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pApplicationInfo-parameter", 52}, { {"If pApplicationInfo is not NULL, pApplicationInfo must be a valid pointer to a valid VkApplicationInfo structure", 112}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-04925", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkDebugReportCallbackCreateInfoEXT structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_EXT_debug_report", 190}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-04926", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkDebugUtilsMessengerCreateInfoEXT structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_EXT_debug_utils", 189}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-06779", 37}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be either VK_EXPORT_METAL_OBJECT_TYPE_METAL_DEVICE_BIT_EXT or VK_EXPORT_METAL_OBJECT_TYPE_METAL_COMMAND_QUEUE_BIT_EXT", 224}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-09400", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkDirectDriverLoadingListLUNARG structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_LUNARG_direct_driver_loading", 199}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-10242", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkLayerSettingsCreateInfoEXT structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_EXT_layer_settings", 186}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-10243", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkValidationFeaturesEXT structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_EXT_validation_features", 186}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-10244", 37}, { {"If the pNext chain of VkInstanceCreateInfo includes a VkValidationFlagsEXT structure, the list of enabled extensions in ppEnabledExtensionNames must contain VK_EXT_validation_flags", 180}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDebugReportCallbackCreateInfoEXT, VkDebugUtilsMessengerCreateInfoEXT, VkDirectDriverLoadingListLUNARG, VkExportMetalObjectCreateInfoEXT, VkLayerSettingsCreateInfoEXT, VkValidationFeaturesEXT, or VkValidationFlagsEXT", 348}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-ppEnabledExtensionNames-parameter", 59}, { {"If enabledExtensionCount is not 0, ppEnabledExtensionNames must be a valid pointer to an array of enabledExtensionCount null-terminated UTF-8 strings", 149}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-ppEnabledLayerNames-parameter", 55}, { {"If enabledLayerCount is not 0, ppEnabledLayerNames must be a valid pointer to an array of enabledLayerCount null-terminated UTF-8 strings", 137}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO", 52}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkInstanceCreateInfo-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkDebugUtilsMessengerCreateInfoEXT, VkExportMetalObjectCreateInfoEXT, or VkLayerSettingsCreateInfoEXT", 211}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLatencySleepInfoNV-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_LATENCY_SLEEP_INFO_NV", 53}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySleepInfoNV-signalSemaphore-09361", 47}, { {"signalSemaphore must be a timeline semaphore", 44}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySleepInfoNV-signalSemaphore-parameter", 51}, { {"signalSemaphore must be a valid VkSemaphore handle", 50}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySleepModeInfoNV-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_LATENCY_SLEEP_MODE_INFO_NV", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySubmissionPresentIdNV-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_LATENCY_SUBMISSION_PRESENT_ID_NV", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySurfaceCapabilitiesNV-pPresentModes-parameter", 59}, { {"If presentModeCount is not 0, and pPresentModes is not NULL, pPresentModes must be a valid pointer to an array of presentModeCount VkPresentModeKHR values", 154}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencySurfaceCapabilitiesNV-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_LATENCY_SURFACE_CAPABILITIES_NV", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLatencyTimingsFrameReportNV-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_LATENCY_TIMINGS_FRAME_REPORT_NV", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkLayerSettingEXT-pLayerName-parameter", 43}, { {"pLayerName must be a null-terminated UTF-8 string", 49}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLayerSettingEXT-pSettingName-parameter", 45}, { {"pSettingName must be a null-terminated UTF-8 string", 51}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLayerSettingEXT-type-parameter", 37}, { {"type must be a valid VkLayerSettingTypeEXT value", 48}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLayerSettingEXT-valueCount-10070", 39}, { {"If valueCount is not 0, pValues must be a valid pointer to an array of valueCount values of the type indicated by type", 118}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLayerSettingsCreateInfoEXT-pSettings-parameter", 53}, { {"If settingCount is not 0, pSettings must be a valid pointer to an array of settingCount valid VkLayerSettingEXT structures", 122}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkLayerSettingsCreateInfoEXT-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_LAYER_SETTINGS_CREATE_INFO_EXT", 62}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkMacOSSurfaceCreateInfoMVK-flags-zerobitmask", 50}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMacOSSurfaceCreateInfoMVK-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMacOSSurfaceCreateInfoMVK-pView-01317", 44}, { {"If pView is an NSView object, it must be a valid NSView, must be backed by a CALayer object of type CAMetalLayer, and vkCreateMacOSSurfaceMVK must be called on the main thread", 175}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMacOSSurfaceCreateInfoMVK-pView-04144", 44}, { {"If pView is a CAMetalLayer object, it must be a valid CAMetalLayer", 66}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMacOSSurfaceCreateInfoMVK-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMappedMemoryRange-memory-00684", 37}, { {"memory must be currently host mapped", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-memory-parameter", 41}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-offset-00687", 37}, { {"offset must be a multiple of VkPhysicalDeviceLimits::nonCoherentAtomSize", 72}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE", 51}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-size-00685", 35}, { {"If size is not equal to VK_WHOLE_SIZE, offset and size must specify a range contained within the currently mapped range of memory", 129}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-size-00686", 35}, { {"If size is equal to VK_WHOLE_SIZE, offset must be within the currently mapped range of memory", 93}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-size-01389", 35}, { {"If size is equal to VK_WHOLE_SIZE, the end of the current mapping of memory must either be a multiple of VkPhysicalDeviceLimits::nonCoherentAtomSize bytes from the beginning of the memory object, or be equal to the end of the memory object", 239}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMappedMemoryRange-size-01390", 35}, { {"If size is not equal to VK_WHOLE_SIZE, size must either be a multiple of VkPhysicalDeviceLimits::nonCoherentAtomSize, or offset plus size must equal the size of memory", 167}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-deviceMask-00675", 47}, { {"If VK_MEMORY_ALLOCATE_DEVICE_MASK_BIT is set, deviceMask must be a valid device mask", 84}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-deviceMask-00676", 47}, { {"If VK_MEMORY_ALLOCATE_DEVICE_MASK_BIT is set, deviceMask must not be zero", 73}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-flags-10760", 42}, { {"If the allocation is performing a memory import operation, then flags must not contain VK_MEMORY_ALLOCATE_ZERO_INITIALIZE_BIT_EXT", 129}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-flags-10761", 42}, { {"If the allocation uses protected memory, then flags must not contain VK_MEMORY_ALLOCATE_ZERO_INITIALIZE_BIT_EXT", 111}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-flags-parameter", 46}, { {"flags must be a valid combination of VkMemoryAllocateFlagBits values", 68}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateFlagsInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_FLAGS_INFO", 58}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-None-00643", 36}, { {"If the parameters define an import operation and the external handle specified was created by the Vulkan API, the device mask specified by VkMemoryAllocateFlagsInfo must match the mask specified when the payload being imported was allocated", 240}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-None-00644", 36}, { {"If the parameters define an import operation and the external handle specified was created by the Vulkan API, the list of physical devices that comprise the logical device passed to vkAllocateMemory must match the list of physical devices that comprise the logical device on which the payload was originally allocated", 317}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-None-04749", 36}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA, the value of memoryTypeIndex must be an index identifying a memory type from the memoryTypeBits field of the VkMemoryZirconHandlePropertiesFUCHSIA structure populated by a call to vkGetMemoryZirconHandlePropertiesFUCHSIA", 352}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-None-06657", 36}, { {"The parameters must not define more than one import operation", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-00647", 46}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_HEAP_BIT, allocationSize must match the size specified when creating the Direct3D 12 heap from which the payload was extracted", 240}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-01742", 46}, { {"If the parameters define an import operation, the external handle specified was created by the Vulkan API, and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT, then the values of allocationSize and memoryTypeIndex must match those specified when the payload being imported was created", 309}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-01743", 46}, { {"If the parameters define an import operation, the external handle was created by the Vulkan API, and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT, then the values of allocationSize and memoryTypeIndex must match those specified when the payload being imported was created", 357}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-01745", 46}, { {"If the parameters define an import operation and the external handle is a host pointer, allocationSize must be an integer multiple of VkPhysicalDeviceExternalMemoryHostPropertiesEXT::minImportedHostPointerAlignment", 214}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-02383", 46}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, allocationSize must be the size returned by vkGetAndroidHardwareBufferPropertiesANDROID for the Android hardware buffer", 264}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-06382", 46}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA, allocationSize must match VkMemoryRequirements::size value retrieved by vkGetImageMemoryRequirements or vkGetBufferMemoryRequirements for image-based or buffer-based collections respectively", 270}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-07897", 46}, { {"If the parameters do not define an import or export operation, allocationSize must be greater than 0", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-07899", 46}, { {"If the parameters define an export operation and the handle type is not VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID , allocationSize must be greater than 0", 178}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-07902", 46}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA, the value of allocationSize must be greater than 0", 182}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-07903", 46}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA, the value of allocationSize must be less than or equal to the size of the VMO as determined by zx_vmo_get_size(handle) where handle is the VMO handle to the imported external memory", 313}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-allocationSize-08942", 46}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX, allocationSize must be the size returned by vkGetScreenBufferPropertiesQNX for the QNX Screen buffer", 231}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-buffer-06380", 38}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA, and VkMemoryDedicatedAllocateInfo::buffer is present and non-NULL, VkImportMemoryBufferCollectionFUCHSIA::collection and VkImportMemoryBufferCollectionFUCHSIA::index must match VkBufferCollectionBufferCreateInfoFUCHSIA::collection and VkBufferCollectionBufferCreateInfoFUCHSIA::index, respectively, of the VkBufferCollectionBufferCreateInfoFUCHSIA structure used to create the VkMemoryDedicatedAllocateInfo::buffer", 494}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-buffer-06385", 38}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA and VkMemoryDedicatedAllocateInfo::buffer is not VK_NULL_HANDLE, the buffer must be created with a VkBufferCollectionBufferCreateInfoFUCHSIA structure chained to its VkBufferCreateInfo::pNext pointer", 278}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-flags-03330", 37}, { {"If VkMemoryAllocateFlagsInfo::flags includes VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT, the bufferDeviceAddressCaptureReplay feature must be enabled", 159}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-flags-03331", 37}, { {"If VkMemoryAllocateFlagsInfo::flags includes VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT, the bufferDeviceAddress feature must be enabled", 131}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-image-06381", 37}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA, and VkMemoryDedicatedAllocateInfo::image is present and non-NULL, VkImportMemoryBufferCollectionFUCHSIA::collection and VkImportMemoryBufferCollectionFUCHSIA::index must match VkBufferCollectionImageCreateInfoFUCHSIA::collection and VkBufferCollectionImageCreateInfoFUCHSIA::index, respectively, of the VkBufferCollectionImageCreateInfoFUCHSIA structure used to create the VkMemoryDedicatedAllocateInfo::image", 489}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-image-06384", 37}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA and VkMemoryDedicatedAllocateInfo::image is not VK_NULL_HANDLE, the image must be created with a VkBufferCollectionImageCreateInfoFUCHSIA structure chained to its VkImageCreateInfo::pNext pointer", 274}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-00645", 47}, { {"If the parameters define an import operation and the external handle is an NT handle or a global share handle created outside of the Vulkan API, the value of memoryTypeIndex must be one of those returned by vkGetMemoryWin32HandlePropertiesKHR", 242}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-00648", 47}, { {"If the parameters define an import operation and the external handle is a POSIX file descriptor created outside of the Vulkan API, the value of memoryTypeIndex must be one of those returned by vkGetMemoryFdPropertiesKHR", 219}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-01744", 47}, { {"If the parameters define an import operation and the external handle is a host pointer, the value of memoryTypeIndex must be one of those returned by vkGetMemoryHostPointerPropertiesEXT", 185}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-01872", 47}, { {"If the protectedMemory feature is not enabled, the VkMemoryAllocateInfo::memoryTypeIndex must not indicate a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 166}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-02385", 47}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, memoryTypeIndex must be one of those returned by vkGetAndroidHardwareBufferPropertiesANDROID for the Android hardware buffer", 269}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-06386", 47}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA, memoryTypeIndex must be from VkBufferCollectionPropertiesFUCHSIA as retrieved by vkGetBufferCollectionPropertiesFUCHSIA", 199}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-memoryTypeIndex-08943", 47}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX, memoryTypeIndex must be one of those returned by vkGetScreenBufferPropertiesQNX for the QNX Screen buffer", 236}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-opaqueCaptureAddress-03329", 52}, { {"If VkMemoryOpaqueCaptureAddressAllocateInfo::opaqueCaptureAddress is not zero, VkMemoryAllocateFlagsInfo::flags must include VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT", 177}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-opaqueCaptureAddress-03333", 52}, { {"If the parameters define an import operation, VkMemoryOpaqueCaptureAddressAllocateInfo::opaqueCaptureAddress must be zero", 121}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-00639", 37}, { {"If the pNext chain includes a VkExportMemoryAllocateInfo structure, and any of the handle types specified in VkExportMemoryAllocateInfo::handleTypes require a dedicated allocation, as reported by vkGetPhysicalDeviceImageFormatProperties2 in VkExternalImageFormatProperties::externalMemoryProperties.externalMemoryFeatures, or by vkGetPhysicalDeviceExternalBufferProperties in VkExternalBufferProperties::externalMemoryProperties.externalMemoryFeatures, the pNext chain must include a VkMemoryDedicatedAllocateInfo or VkDedicatedAllocationMemoryAllocateInfoNV structure with either its image or buffer member set to a value other than VK_NULL_HANDLE", 648}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-00640", 37}, { {"If the pNext chain includes a VkExportMemoryAllocateInfo structure, it must not include a VkExportMemoryAllocateInfoNV or VkExportMemoryWin32HandleInfoNV structure", 163}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-00641", 37}, { {"If the pNext chain includes a VkImportMemoryWin32HandleInfoKHR structure, it must not include a VkImportMemoryWin32HandleInfoNV structure", 137}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-01874", 37}, { {"If the parameters do not define an import operation, and the pNext chain includes a VkExportMemoryAllocateInfo structure with VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID included in its handleTypes member, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image not equal to VK_NULL_HANDLE, then allocationSize must be 0", 369}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02384", 37}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, and the pNext chain does not include a VkMemoryDedicatedAllocateInfo structure or VkMemoryDedicatedAllocateInfo::image is VK_NULL_HANDLE, the Android hardware buffer must have a AHardwareBuffer_Desc::format of AHARDWAREBUFFER_FORMAT_BLOB and a AHardwareBuffer_Desc::usage that includes AHARDWAREBUFFER_USAGE_GPU_DATA_BUFFER", 468}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02386", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo with image that is not VK_NULL_HANDLE, the Android hardware buffer's AHardwareBuffer::usage must include at least one of AHARDWAREBUFFER_USAGE_GPU_FRAMEBUFFER, AHARDWAREBUFFER_USAGE_GPU_SAMPLED_IMAGE or AHARDWAREBUFFER_USAGE_GPU_DATA_BUFFER", 398}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02387", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo with image that is not VK_NULL_HANDLE, the format of image must be VK_FORMAT_UNDEFINED or the format returned by vkGetAndroidHardwareBufferPropertiesANDROID in VkAndroidHardwareBufferFormatPropertiesANDROID::format for the Android hardware buffer", 404}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02388", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, the width, height, and array layer dimensions of image and the Android hardware buffer's AHardwareBuffer_Desc must be identical", 334}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02389", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, and the Android hardware buffer's AHardwareBuffer::usage includes AHARDWAREBUFFER_USAGE_GPU_MIPMAP_COMPLETE, the image must have a complete mipmap chain", 359}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02390", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, each bit set in the usage of image must be listed in AHardwareBuffer Usage Equivalence, and if there is a corresponding AHARDWAREBUFFER_USAGE bit listed that bit must be included in the Android hardware buffer's AHardwareBuffer_Desc::usage", 446}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02586", 37}, { {"If the parameters define an import operation, the external handle is an Android hardware buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, and the Android hardware buffer's AHardwareBuffer::usage does not include AHARDWAREBUFFER_USAGE_GPU_MIPMAP_COMPLETE, the image must have exactly one mipmap level", 368}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02805", 37}, { {"If the parameters define an import operation and the external handle is a host pointer, the pNext chain must not include a VkDedicatedAllocationMemoryAllocateInfoNV structure with either its image or buffer field set to a value other than VK_NULL_HANDLE", 253}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-02806", 37}, { {"If the parameters define an import operation and the external handle is a host pointer, the pNext chain must not include a VkMemoryDedicatedAllocateInfo structure with either its image or buffer field set to a value other than VK_NULL_HANDLE", 241}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-03332", 37}, { {"If the pNext chain includes a VkImportMemoryHostPointerInfoEXT structure, VkMemoryOpaqueCaptureAddressAllocateInfo::opaqueCaptureAddress must be zero", 149}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-06383", 37}, { {"If the parameters define an import operation from an VkBufferCollectionFUCHSIA, the pNext chain must include a VkMemoryDedicatedAllocateInfo structure with either its image or buffer field set to a value other than VK_NULL_HANDLE", 229}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-06780", 37}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be VK_EXPORT_METAL_OBJECT_TYPE_METAL_BUFFER_BIT_EXT", 158}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-07900", 37}, { {"If the parameters define an export operation, the handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, and the pNext does not include a VkMemoryDedicatedAllocateInfo structure, allocationSize must be greater than 0", 244}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-07901", 37}, { {"If the parameters define an export operation, the handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with buffer set to a valid VkBuffer object, allocationSize must be greater than 0", 285}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-08944", 37}, { {"If the parameters define an import operation, the external handle is a QNX Screen buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo with image that is not VK_NULL_HANDLE, the QNX Screen's buffer must be a valid QNX Screen buffer", 247}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-08945", 37}, { {"If the parameters define an import operation, the external handle is an QNX Screen buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo with image that is not VK_NULL_HANDLE, the format of image must be VK_FORMAT_UNDEFINED or the format returned by vkGetScreenBufferPropertiesQNX in VkScreenBufferFormatPropertiesQNX::format for the QNX Screen buffer", 366}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-08946", 37}, { {"If the parameters define an import operation, the external handle is a QNX Screen buffer, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, the width, height, and array layer dimensions of image and the QNX Screen buffer's _screen_buffer must be identical", 315}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-09858", 37}, { {"If the pNext chain includes a VkExportMemoryAllocateInfo structure, and any of the handle types specified in VkExportMemoryAllocateInfo::handleTypes require a dedicated allocation, as reported by vkGetPhysicalDeviceExternalTensorPropertiesARM in VkExternalTensorPropertiesARM::externalMemoryProperties.externalMemoryFeatures, the pNext chain must include a VkMemoryDedicatedAllocateInfoTensorARM structure with its tensor member set to a value other than VK_NULL_HANDLE", 469}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-10395", 37}, { {"If the parameters define an import operation and the external handle is a VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLTEXTURE_BIT_EXT, then pNext must include a VkMemoryDedicatedAllocateInfo with image that is not VK_NULL_HANDLE", 218}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-10396", 37}, { {"If the parameters define an import operation, the external handle is a Metal MTLTexture, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, the width, height, array layer dimensions, and mipmap levels of image and the Metal MTLTexture's must be identical", 313}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-10397", 37}, { {"If the parameters define an import operation, the external handle is a Metal MTLTexture, and the pNext chain includes a VkMemoryDedicatedAllocateInfo structure with image that is not VK_NULL_HANDLE, allocationSize must be 0", 223}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDedicatedAllocationMemoryAllocateInfoNV, VkExportMemoryAllocateInfo, VkExportMemoryAllocateInfoNV, VkExportMemoryWin32HandleInfoKHR, VkExportMemoryWin32HandleInfoNV, VkExportMetalObjectCreateInfoEXT, VkImportAndroidHardwareBufferInfoANDROID, VkImportMemoryBufferCollectionFUCHSIA, VkImportMemoryFdInfoKHR, VkImportMemoryHostPointerInfoEXT, VkImportMemoryMetalHandleInfoEXT, VkImportMemoryWin32HandleInfoKHR, VkImportMemoryWin32HandleInfoNV, VkImportMemoryZirconHandleInfoFUCHSIA, VkImportMetalBufferInfoEXT, VkImportNativeBufferInfoOHOS, VkImportScreenBufferInfoQNX, VkMemoryAllocateFlagsInfo, VkMemoryDedicatedAllocateInfo, VkMemoryDedicatedAllocateInfoTensorARM, VkMemoryOpaqueCaptureAddressAllocateInfo, or VkMemoryPriorityAllocateInfoEXT", 874}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO", 52}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT", 142}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryAllocateInfo-screenBufferImport-08941", 50}, { {"If the parameters define an import operation and the external handle type is VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX, VkPhysicalDeviceExternalMemoryScreenBufferFeaturesQNX::screenBufferImport must be enabled", 220}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryBarrier-dstAccessMask-parameter", 44}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier-pNext-pNext", 32}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_BARRIER", 46}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier-srcAccessMask-parameter", 44}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03900", 41}, { {"If dstAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03901", 41}, { {"If dstAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03902", 41}, { {"If dstAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03903", 41}, { {"If dstAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03904", 41}, { {"If dstAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03905", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03906", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03907", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03909", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03910", 41}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03911", 41}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03912", 41}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03913", 41}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03914", 41}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03915", 41}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03916", 41}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03917", 41}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03918", 41}, { {"If dstAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03919", 41}, { {"If dstAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03920", 41}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03922", 41}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03923", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03924", 41}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03925", 41}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03926", 41}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03927", 41}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-03928", 41}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04747", 41}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04858", 41}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04859", 41}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04860", 41}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04861", 41}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-04994", 41}, { {"If dstAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, dstStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-06256", 41}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07272", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07454", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07455", 41}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07456", 41}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07457", 41}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-07458", 41}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-08118", 41}, { {"If dstAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-10670", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-10671", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-11294", 41}, { {"If dstAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-11771", 41}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-11772", 41}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstAccessMask-parameter", 45}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03929", 40}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03930", 40}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03931", 40}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03932", 40}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03933", 40}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03934", 40}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-03935", 40}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-04957", 40}, { {"If the subpassShading feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-04995", 40}, { {"If the invocationMask feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-07316", 40}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-07946", 40}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-10751", 40}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-10752", 40}, { {"If the rayTracingMaintenance1 feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-11541", 40}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-dstStageMask-parameter", 44}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_BARRIER_2", 48}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03900", 41}, { {"If srcAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03901", 41}, { {"If srcAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03902", 41}, { {"If srcAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03903", 41}, { {"If srcAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03904", 41}, { {"If srcAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03905", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03906", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03907", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03909", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03910", 41}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03911", 41}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03912", 41}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03913", 41}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03914", 41}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03915", 41}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03916", 41}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03917", 41}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03918", 41}, { {"If srcAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03919", 41}, { {"If srcAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03920", 41}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03922", 41}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03923", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03924", 41}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03925", 41}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03926", 41}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03927", 41}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-03928", 41}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04747", 41}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04858", 41}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04859", 41}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04860", 41}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04861", 41}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-04994", 41}, { {"If srcAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, srcStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-06256", 41}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07272", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07454", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07455", 41}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07456", 41}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07457", 41}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-07458", 41}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-08118", 41}, { {"If srcAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-10670", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-10671", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-11294", 41}, { {"If srcAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-11771", 41}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-11772", 41}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcAccessMask-parameter", 45}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03929", 40}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03930", 40}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03931", 40}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03932", 40}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03933", 40}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03934", 40}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-03935", 40}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-04957", 40}, { {"If the subpassShading feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-04995", 40}, { {"If the invocationMask feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-07316", 40}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-07946", 40}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-10751", 40}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-10752", 40}, { {"If the rayTracingMaintenance1 feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-11541", 40}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrier2-srcStageMask-parameter", 44}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrierAccessFlags3KHR-dstAccessMask3-parameter", 60}, { {"dstAccessMask3 must be a valid combination of VkAccessFlagBits3KHR values", 73}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrierAccessFlags3KHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_BARRIER_ACCESS_FLAGS_3_KHR", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryBarrierAccessFlags3KHR-srcAccessMask3-parameter", 60}, { {"srcAccessMask3 must be a valid combination of VkAccessFlagBits3KHR values", 73}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-01436", 47}, { {"If buffer is not VK_NULL_HANDLE, buffer must have been created without VK_BUFFER_CREATE_SPARSE_BINDING_BIT set in VkBufferCreateInfo::flags", 139}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-01877", 47}, { {"If buffer is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_HEAP_BIT, or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_RESOURCE_BIT, and the external handle was created by the Vulkan API, then the memory being imported must also be a dedicated buffer allocation and buffer must be identical to the buffer associated with the imported memory", 622}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-01879", 47}, { {"If buffer is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT, the memory being imported must also be a dedicated buffer allocation and buffer must be identical to the buffer associated with the imported memory", 301}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-02965", 47}, { {"If buffer is not VK_NULL_HANDLE and the memory is not an imported Android Hardware Buffer or an imported QNX Screen buffer , VkMemoryAllocateInfo::allocationSize must be greater than or equal to the VkMemoryRequirements::size of the buffer", 239}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-04752", 47}, { {"If buffer is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA, the memory being imported must also be a dedicated buffer allocation and buffer must be identical to the buffer associated with the imported memory", 310}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-buffer-parameter", 51}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-commonparent", 47}, { {"Both of buffer, and image that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 145}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-01432", 46}, { {"At least one of image and buffer must be VK_NULL_HANDLE", 55}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-01434", 46}, { {"If image is not VK_NULL_HANDLE, image must have been created without VK_IMAGE_CREATE_SPARSE_BINDING_BIT set in VkImageCreateInfo::flags", 135}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-01797", 46}, { {"If image is not VK_NULL_HANDLE, image must not have been created with VK_IMAGE_CREATE_DISJOINT_BIT set in VkImageCreateInfo::flags", 130}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-01876", 46}, { {"If image is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_HEAP_BIT, or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_RESOURCE_BIT, and the external handle was created by the Vulkan API, then the memory being imported must also be a dedicated image allocation and image must be identical to the image associated with the imported memory", 618}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-01878", 46}, { {"If image is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT, the memory being imported must also be a dedicated image allocation and image must be identical to the image associated with the imported memory", 297}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-02964", 46}, { {"If image is not VK_NULL_HANDLE and the memory is not an imported Android Hardware Buffer or an imported QNX Screen buffer , VkMemoryAllocateInfo::allocationSize must be greater than or equal to the VkMemoryRequirements::size of the image", 237}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-04751", 46}, { {"If image is not VK_NULL_HANDLE and VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA, the memory being imported must also be a dedicated image allocation and image must be identical to the image associated with the imported memory", 306}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-image-parameter", 50}, { {"If image is not VK_NULL_HANDLE, image must be a valid VkImage handle", 68}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO", 62}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfoTensorARM-allocationSize-09710", 64}, { {"VkMemoryAllocateInfo::allocationSize must equal the VkMemoryRequirements::size of the tensor", 92}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfoTensorARM-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO_TENSOR_ARM", 73}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfoTensorARM-tensor-09859", 56}, { {"If VkMemoryAllocateInfo defines a memory import operation with handle type VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT, the memory being imported must also be a dedicated tensor allocation and tensor must be identical to the tensor associated with the imported memory", 268}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedAllocateInfoTensorARM-tensor-parameter", 60}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryDedicatedRequirements-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_DEDICATED_REQUIREMENTS", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryFdPropertiesKHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryFdPropertiesKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_FD_PROPERTIES_KHR", 56}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetAndroidHardwareBufferInfoANDROID-handleTypes-01882", 66}, { {"VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must have been included in VkExportMemoryAllocateInfo::handleTypes when memory was created", 157}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetAndroidHardwareBufferInfoANDROID-memory-parameter", 65}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetAndroidHardwareBufferInfoANDROID-pNext-01883", 60}, { {"If the pNext chain of the VkMemoryAllocateInfo used to allocate memory included a VkMemoryDedicatedAllocateInfo with non-NULL image member, then that image must already be bound to memory", 187}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetAndroidHardwareBufferInfoANDROID-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetAndroidHardwareBufferInfoANDROID-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_ANDROID_HARDWARE_BUFFER_INFO_ANDROID", 79}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-handleType-00671", 42}, { {"handleType must have been included in VkExportMemoryAllocateInfo::handleTypes when memory was created", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-handleType-00672", 42}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT", 113}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-handleType-parameter", 46}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-memory-parameter", 42}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetFdInfoKHR-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_FD_INFO_KHR", 54}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-handleType-10414", 51}, { {"handleType must have been included in VkExportMemoryAllocateInfo::handleTypes when memory was created", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-handleType-10415", 51}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLBUFFER_BIT_EXT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLTEXTURE_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLHEAP_BIT_EXT", 168}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-handleType-parameter", 55}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-memory-10413", 47}, { {"memory must have been created with a valid VkExportMemoryAllocateInfo", 69}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-memory-parameter", 51}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetMetalHandleInfoEXT-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_METAL_HANDLE_INFO_EXT", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetNativeBufferInfoOHOS-memory-parameter", 53}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryGetNativeBufferInfoOHOS-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryGetNativeBufferInfoOHOS-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_NATIVE_BUFFER_INFO_OHOS", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryGetRemoteAddressInfoNV-handleType-04966", 52}, { {"handleType must have been included in VkExportMemoryAllocateInfo::handleTypes when memory was created", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetRemoteAddressInfoNV-handleType-parameter", 56}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetRemoteAddressInfoNV-memory-parameter", 52}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetRemoteAddressInfoNV-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetRemoteAddressInfoNV-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_REMOTE_ADDRESS_INFO_NV", 65}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-handleType-00662", 51}, { {"handleType must have been included in VkExportMemoryAllocateInfo::handleTypes when memory was created", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-handleType-00663", 51}, { {"If handleType is defined as an NT handle, vkGetMemoryWin32HandleKHR must be called no more than once for each valid unique combination of memory and handleType", 159}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-handleType-00664", 51}, { {"handleType must be defined as an NT handle or a global share handle", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-handleType-parameter", 55}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-memory-parameter", 51}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetWin32HandleInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_WIN32_HANDLE_INFO_KHR", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-handleType-04775", 56}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA", 72}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-handleType-04776", 56}, { {"handleType must have been included in the handleTypes field of the VkExportMemoryAllocateInfo structure when the external memory was allocated", 142}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-handleType-parameter", 60}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-memory-parameter", 56}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryGetZirconHandleInfoFUCHSIA-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_GET_ZIRCON_HANDLE_INFO_FUCHSIA", 69}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryHostPointerPropertiesEXT-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryHostPointerPropertiesEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_HOST_POINTER_PROPERTIES_EXT", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09569", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags, the memoryMapPlaced feature must be enabled", 92}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09570", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags, the pNext chain must include a VkMemoryMapPlacedInfoEXT structure and VkMemoryMapPlacedInfoEXT::pPlacedAddress must not be NULL", 176}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09571", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags and the memoryMapRangePlaced feature is not enabled, offset must be zero", 120}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09572", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags and the memoryMapRangePlaced feature is not enabled, size must be VK_WHOLE_SIZE or VkMemoryAllocateInfo::allocationSize", 167}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09573", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags and the memoryMapRangePlaced feature is enabled, offset must be aligned to an integer multiple of VkPhysicalDeviceMapMemoryPlacedPropertiesEXT::minPlacedMemoryMapAlignment", 219}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09574", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags and size is not VK_WHOLE_SIZE, size must be aligned to an integer multiple of VkPhysicalDeviceMapMemoryPlacedPropertiesEXT::minPlacedMemoryMapAlignment", 199}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09575", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags, the memory object must not have been imported from a handle type of VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT", 240}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-09651", 32}, { {"If VK_MEMORY_MAP_PLACED_BIT_EXT is set in flags and size is VK_WHOLE_SIZE, VkMemoryAllocateInfo::allocationSize must be aligned to an integer multiple of VkPhysicalDeviceMapMemoryPlacedPropertiesEXT::minPlacedMemoryMapAlignment", 227}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-flags-parameter", 36}, { {"flags must be a valid combination of VkMemoryMapFlagBits values", 63}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-memory-07958", 33}, { {"memory must not be currently host mapped", 40}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-memory-07962", 33}, { {"memory must have been created with a memory type that reports VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT", 97}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-memory-07963", 33}, { {"memory must not have been allocated with multiple instances", 59}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-memory-parameter", 37}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-offset-07959", 33}, { {"offset must be less than the size of memory", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-pNext-pNext", 32}, { {"pNext must be NULL or a pointer to a valid instance of VkMemoryMapPlacedInfoEXT", 79}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_MAP_INFO", 47}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-sType-unique", 33}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-size-07960", 31}, { {"If size is not equal to VK_WHOLE_SIZE, size must be greater than 0", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapInfo-size-07961", 31}, { {"If size is not equal to VK_WHOLE_SIZE, size must be less than or equal to the size of the memory minus offset", 109}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapPlacedInfoEXT-pPlacedAddress-09577", 50}, { {"pPlacedAddress must be aligned to an integer multiple of VkPhysicalDeviceMapMemoryPlacedPropertiesEXT::minPlacedMemoryMapAlignment", 130}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapPlacedInfoEXT-pPlacedAddress-09578", 50}, { {"The address range specified by pPlacedAddress and VkMemoryMapInfo::size must not overlap any existing Vulkan memory object mapping", 130}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMapPlacedInfoEXT-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_MAP_PLACED_INFO_EXT", 58}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstFlags-13100", 41}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstFlags-13101", 41}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstFlags-parameter", 45}, { {"dstFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13039", 41}, { {"The buffer from which dstRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13040", 41}, { {"dstRange.address must be a multiple of 4", 40}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13041", 41}, { {"dstRange.size must be greater than or equal to 4", 48}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13097", 41}, { {"If the range specified by dstRange is not bound completely to memory when accessed, dstFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13098", 41}, { {"If the buffer from which the range specified by dstRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 203}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13099", 41}, { {"If the buffer from which the range specified by dstRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 210}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13122", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 282}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13123", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 230}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13124", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 319}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-dstRange-13125", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing dstRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_MARKER_INFO_AMD", 54}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03929", 38}, { {"If the geometryShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 108}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03930", 38}, { {"If the tessellationShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 182}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03931", 38}, { {"If the conditionalRendering feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03932", 38}, { {"If the fragmentDensityMap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03933", 38}, { {"If the transformFeedback feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03934", 38}, { {"If the meshShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-03935", 38}, { {"If the taskShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-04957", 38}, { {"If the subpassShading feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-04995", 38}, { {"If the invocationMask feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 115}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-07316", 38}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 177}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-07946", 38}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stage must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 159}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-10751", 38}, { {"If the accelerationStructure feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-10752", 38}, { {"If the rayTracingMaintenance1 feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-11541", 38}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-13038", 38}, { {"stage must include only a single pipeline stage", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-parameter", 42}, { {"stage must be a valid combination of VkPipelineStageFlagBits2KHR values", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMarkerInfoAMD-stage-requiredbitmask", 48}, { {"stage must not be 0", 19}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryMetalHandlePropertiesEXT-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryMetalHandlePropertiesEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_METAL_HANDLE_PROPERTIES_EXT", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryOpaqueCaptureAddressAllocateInfo-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_OPAQUE_CAPTURE_ADDRESS_ALLOCATE_INFO", 75}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryPriorityAllocateInfoEXT-priority-02602", 51}, { {"priority must be between 0 and 1, inclusive", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryPriorityAllocateInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_PRIORITY_ALLOCATE_INFO_EXT", 65}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryRangeBarrierKHR-None-09097", 39}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-None-09098", 39}, { {"If the VK_KHR_external_memory extension is not enabled, and the value of VkApplicationInfo::apiVersion used to create the VkInstance is not greater than or equal to Version 1.1, dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_EXTERNAL", 234}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-address-13087", 42}, { {"If the buffer from which address was queried was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, srcQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 276}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-address-13088", 42}, { {"If the buffer from which address was queried was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not equal, dstQueueFamilyIndex must be VK_QUEUE_FAMILY_EXTERNAL, VK_QUEUE_FAMILY_FOREIGN_EXT, or a valid queue family", 276}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressFlags-13100", 47}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 142}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressFlags-13101", 47}, { {"addressFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressFlags-parameter", 51}, { {"addressFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 78}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13097", 47}, { {"If the range specified by addressRange is not bound completely to memory when accessed, addressFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13098", 47}, { {"If the buffer from which the range specified by addressRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13099", 47}, { {"If the buffer from which the range specified by addressRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, addressFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 218}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13122", 47}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 290}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13123", 47}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, addressFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 238}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13124", 47}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 327}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-addressRange-13125", 47}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing addressRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, addressFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 264}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03900", 48}, { {"If dstAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03901", 48}, { {"If dstAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03902", 48}, { {"If dstAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03903", 48}, { {"If dstAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03904", 48}, { {"If dstAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03905", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03906", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03907", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03909", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03910", 48}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03911", 48}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03912", 48}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03913", 48}, { {"If dstAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03914", 48}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03915", 48}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03916", 48}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03917", 48}, { {"If dstAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03918", 48}, { {"If dstAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03919", 48}, { {"If dstAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03920", 48}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03922", 48}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03923", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03924", 48}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03925", 48}, { {"If dstAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03926", 48}, { {"If dstAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03927", 48}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-03928", 48}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04747", 48}, { {"If dstAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04858", 48}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04859", 48}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04860", 48}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04861", 48}, { {"If dstAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-04994", 48}, { {"If dstAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, dstStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-06256", 48}, { {"If dstAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, dstStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07272", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07454", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07455", 48}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07456", 48}, { {"If dstAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, dstStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07457", 48}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-07458", 48}, { {"If dstAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-08118", 48}, { {"If dstAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-10670", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-10671", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, dstStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-11294", 48}, { {"If dstAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-11771", 48}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-11772", 48}, { {"If dstAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, dstStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstAccessMask-parameter", 52}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstQueueFamilyIndex-09100", 54}, { {"If the VK_EXT_queue_family_foreign extension is not enabled dstQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03929", 47}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03930", 47}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03931", 47}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03932", 47}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03933", 47}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03934", 47}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-03935", 47}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-04957", 47}, { {"If the subpassShading feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-04995", 47}, { {"If the invocationMask feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-07316", 47}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-07946", 47}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-10751", 47}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-10752", 47}, { {"If the rayTracingMaintenance1 feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-11541", 47}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-dstStageMask-parameter", 51}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_RANGE_BARRIER_KHR", 56}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03900", 48}, { {"If srcAccessMask includes VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_COPY_INDIRECT_BIT_KHR, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03901", 48}, { {"If srcAccessMask includes VK_ACCESS_2_INDEX_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03902", 48}, { {"If srcAccessMask includes VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT, VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 254}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03903", 48}, { {"If srcAccessMask includes VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT, VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03904", 48}, { {"If srcAccessMask includes VK_ACCESS_2_UNIFORM_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03905", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_SAMPLED_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03906", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03907", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03909", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 209}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03910", 48}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03911", 48}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 217}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03912", 48}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 267}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03913", 48}, { {"If srcAccessMask includes VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 268}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03914", 48}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 430}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03915", 48}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFER_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_COPY_BIT, VK_PIPELINE_STAGE_2_BLIT_BIT, VK_PIPELINE_STAGE_2_RESOLVE_BIT, VK_PIPELINE_STAGE_2_CLEAR_BIT, VK_PIPELINE_STAGE_2_ALL_TRANSFER_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_CONVERT_COOPERATIVE_VECTOR_MATRIX_BIT_NV, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 462}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03916", 48}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03917", 48}, { {"If srcAccessMask includes VK_ACCESS_2_HOST_WRITE_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_HOST_BIT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03918", 48}, { {"If srcAccessMask includes VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03919", 48}, { {"If srcAccessMask includes VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03920", 48}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 223}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03922", 48}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03923", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 220}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03924", 48}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03925", 48}, { {"If srcAccessMask includes VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03926", 48}, { {"If srcAccessMask includes VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03927", 48}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 390}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-03928", 48}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04747", 48}, { {"If srcAccessMask includes VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT, VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT, VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 269}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04858", 48}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04859", 48}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04860", 48}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04861", 48}, { {"If srcAccessMask includes VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-04994", 48}, { {"If srcAccessMask includes VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI, srcStageMask must include VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 143}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-06256", 48}, { {"If srcAccessMask includes VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_KHR, and the rayQuery feature is not enabled, srcStageMask must include VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR, or VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT", 344}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07272", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT or VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07454", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_READ_BIT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT, or one of the VK_PIPELINE_STAGE_*_SHADER_BIT stages", 310}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07455", 48}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07456", 48}, { {"If srcAccessMask includes VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV, srcStageMask must include VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07457", 48}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-07458", 48}, { {"If srcAccessMask includes VK_ACCESS_2_MICROMAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT or VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-08118", 48}, { {"If srcAccessMask includes VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 219}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-10670", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_READ_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-10671", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SHADER_TILE_ATTACHMENT_WRITE_BIT_QCOM, srcStageMask must include VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT or VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-11294", 48}, { {"If srcAccessMask includes VK_ACCESS_2_SAMPLER_HEAP_READ_BIT_EXT or VK_ACCESS_2_RESOURCE_HEAP_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, or one of VK_PIPELINE_STAGE_*_SHADER_BIT stages", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-11771", 48}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_READ_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-11772", 48}, { {"If srcAccessMask includes VK_ACCESS_2_MEMORY_DECOMPRESSION_WRITE_BIT_EXT, srcStageMask must include VK_PIPELINE_STAGE_2_MEMORY_DECOMPRESSION_BIT_EXT", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcAccessMask-parameter", 52}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcQueueFamilyIndex-09099", 54}, { {"If the VK_EXT_queue_family_foreign extension is not enabled srcQueueFamilyIndex must not be VK_QUEUE_FAMILY_FOREIGN_EXT", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03929", 47}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03930", 47}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 189}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03931", 47}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03932", 47}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 132}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03933", 47}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 125}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03934", 47}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-03935", 47}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 111}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-04957", 47}, { {"If the subpassShading feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-04995", 47}, { {"If the invocationMask feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-07316", 47}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 184}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-07946", 47}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 166}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-10751", 47}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-10752", 47}, { {"If the rayTracingMaintenance1 feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-11541", 47}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-13089", 47}, { {"If either srcStageMask or dstStageMask includes VK_PIPELINE_STAGE_2_HOST_BIT, srcQueueFamilyIndex and dstQueueFamilyIndex must be equal", 135}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarrierKHR-srcStageMask-parameter", 51}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarriersInfoKHR-pMemoryRangeBarriers-parameter", 64}, { {"If memoryRangeBarrierCount is not 0, pMemoryRangeBarriers must be a valid pointer to an array of memoryRangeBarrierCount valid VkMemoryRangeBarrierKHR structures", 161}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRangeBarriersInfoKHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_RANGE_BARRIERS_INFO_KHR", 62}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkMemoryRequirements2-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkMemoryDedicatedRequirements or VkTileMemoryRequirementsQCOM", 192}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryRequirements2-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2", 53}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryRequirements2-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMemoryToImageCopy-aspectMask-09103", 41}, { {"The aspectMask member of imageSubresource must only have a single bit set", 73}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-imageExtent-06659", 42}, { {"imageExtent.width must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-imageExtent-06660", 42}, { {"imageExtent.height must not be 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-imageExtent-06661", 42}, { {"imageExtent.depth must not be 0", 31}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-imageSubresource-parameter", 51}, { {"imageSubresource must be a valid VkImageSubresourceLayers structure", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-memoryImageHeight-09102", 48}, { {"memoryImageHeight must be 0, or greater than or equal to the height member of imageExtent", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-memoryRowLength-09101", 46}, { {"memoryRowLength must be 0, or greater than or equal to the width member of imageExtent", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-pHostPointer-09061", 43}, { {"pHostPointer must point to memory that is large enough to contain all memory locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 175}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-pHostPointer-parameter", 47}, { {"pHostPointer must be a pointer value", 36}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-pRegions-09062", 39}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryToImageCopy-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_TO_IMAGE_COPY", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-flags-09579", 34}, { {"If VK_MEMORY_UNMAP_RESERVE_BIT_EXT is set in flags, the memoryUnmapReserve must be enabled", 90}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-flags-09580", 34}, { {"If VK_MEMORY_UNMAP_RESERVE_BIT_EXT is set in flags, the memory object must not have been imported from a handle type of VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT", 243}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-flags-parameter", 38}, { {"flags must be a valid combination of VkMemoryUnmapFlagBits values", 65}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-memory-07964", 35}, { {"memory must be currently host mapped", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-memory-parameter", 39}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryUnmapInfo-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_UNMAP_INFO", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryWin32HandlePropertiesKHR-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryWin32HandlePropertiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_WIN32_HANDLE_PROPERTIES_KHR", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryZirconHandlePropertiesFUCHSIA-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMemoryZirconHandlePropertiesFUCHSIA-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_MEMORY_ZIRCON_HANDLE_PROPERTIES_FUCHSIA", 71}, { "chapters/memory.html", 20} } },
        { {"VUID-VkMetalSurfaceCreateInfoEXT-flags-zerobitmask", 50}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMetalSurfaceCreateInfoEXT-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMetalSurfaceCreateInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_METAL_SURFACE_CREATE_INFO_EXT", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkMicromapBuildInfoEXT-flags-parameter", 43}, { {"flags must be a valid combination of VkBuildMicromapFlagBitsEXT values", 70}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-format-11652", 40}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of pUsageCounts or ppUsageCounts, format must be a valid value from VkOpacityMicromapFormatEXT", 160}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-format-11653", 40}, { {"If type is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, for each member of pUsageCounts or ppUsageCounts, format must be a valid value from VkDisplacementMicromapFormatNV", 168}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-format-11657", 40}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of triangleArray, format must be a valid value from VkOpacityMicromapFormatEXT", 144}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-format-11658", 40}, { {"If type is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, for each member of triangleArray, format must be a valid value from VkDisplacementMicromapFormatNV", 152}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-pUsageCounts-07516", 46}, { {"Only one of pUsageCounts or ppUsageCounts can be a valid pointer, the other must be NULL", 88}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-pUsageCounts-parameter", 50}, { {"If usageCountsCount is not 0, and pUsageCounts is not NULL, pUsageCounts must be a valid pointer to an array of usageCountsCount VkMicromapUsageEXT structures", 158}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-ppUsageCounts-parameter", 51}, { {"If usageCountsCount is not 0, and ppUsageCounts is not NULL, ppUsageCounts must be a valid pointer to an array of usageCountsCount valid pointers to VkMicromapUsageEXT structures", 178}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_MICROMAP_BUILD_INFO_EXT", 55}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11654", 50}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of pUsageCounts or ppUsageCounts, if format is VK_OPACITY_MICROMAP_FORMAT_2_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity2StateSubdivisionLevel", 278}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11655", 50}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of pUsageCounts or ppUsageCounts, if format is VK_OPACITY_MICROMAP_FORMAT_4_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity4StateSubdivisionLevel", 278}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11656", 50}, { {"If type is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, for each member of pUsageCounts or ppUsageCounts, subdivisionLevel must be less than or equal to VkPhysicalDeviceDisplacementMicromapPropertiesNV::maxDisplacementMicromapSubdivisionLevel", 240}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11659", 50}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of triangleArray, if format is VK_OPACITY_MICROMAP_FORMAT_2_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity2StateSubdivisionLevel", 262}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11660", 50}, { {"If type is VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT, for each member of triangleArray, if format is VK_OPACITY_MICROMAP_FORMAT_4_STATE_EXT, subdivisionLevel must be less than or equal to VkPhysicalDeviceOpacityMicromapPropertiesEXT::maxOpacity4StateSubdivisionLevel", 262}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-subdivisionLevel-11661", 50}, { {"If type is VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV, for each member of triangleArray, subdivisionLevel must be less than or equal to VkPhysicalDeviceDisplacementMicromapPropertiesNV::maxDisplacementMicromapSubdivisionLevel", 224}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildInfoEXT-type-parameter", 42}, { {"type must be a valid VkMicromapTypeEXT value", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapBuildSizesInfoEXT-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapBuildSizesInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_MICROMAP_BUILD_SIZES_INFO_EXT", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-buffer-07435", 41}, { {"buffer must have been created with the VK_BUFFER_USAGE_MICROMAP_STORAGE_BIT_EXT usage flag set", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-buffer-07436", 41}, { {"buffer must not have been created with VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-buffer-parameter", 45}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-bufferDeviceAddressMultiDevice-11617", 65}, { {"If createFlags includes VK_MICROMAP_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_EXT and device was created with multiple physical devices, the bufferDeviceAddressMultiDevice feature must be enabled", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-createFlags-parameter", 50}, { {"createFlags must be a valid combination of VkMicromapCreateFlagBitsEXT values", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-deviceAddress-07433", 48}, { {"If deviceAddress is not zero, createFlags must include VK_MICROMAP_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_EXT", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-deviceAddress-parameter", 52}, { {"If deviceAddress is not 0, deviceAddress must be a valid VkDeviceAddress value", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-micromapCaptureReplay-11616", 56}, { {"If createFlags includes VK_MICROMAP_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_EXT, VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapCaptureReplay must be VK_TRUE", 163}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-offset-07437", 41}, { {"The sum of offset and size must be less than or equal to the size of buffer", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-offset-07438", 41}, { {"offset must be a multiple of 256 bytes", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_MICROMAP_CREATE_INFO_EXT", 56}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapCreateInfoEXT-type-parameter", 43}, { {"type must be a valid VkMicromapTypeEXT value", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-VkMicromapUsageKHR-format-parameter", 40}, { {"format must be a valid VkOpacityMicromapFormatKHR value", 55}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapVersionInfoEXT-pNext-pNext", 41}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapVersionInfoEXT-pVersionData-parameter", 52}, { {"pVersionData must be a valid pointer to an array of \\(2 \\times \\mathtt{VK\\_UUID\\_SIZE}\\) uint8_t values", 109}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMicromapVersionInfoEXT-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_MICROMAP_VERSION_INFO_EXT", 57}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-VkMultisamplePropertiesEXT-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/limits.html", 20} } },
        { {"VUID-VkMultisamplePropertiesEXT-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_MULTISAMPLE_PROPERTIES_EXT", 58}, { "chapters/limits.html", 20} } },
        { {"VUID-VkMultisampledRenderToSingleSampledInfoEXT-pNext-06880", 59}, { {"If added to the pNext chain of VkRenderingInfo, each imageView member of any element of VkRenderingInfo::pColorAttachments, VkRenderingInfo::pDepthAttachment, or VkRenderingInfo::pStencilAttachment that is not VK_NULL_HANDLE must have a format that supports the sample count specified in rasterizationSamples", 308}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultisampledRenderToSingleSampledInfoEXT-rasterizationSamples-06878", 74}, { {"The value of rasterizationSamples must not be VK_SAMPLE_COUNT_1_BIT", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultisampledRenderToSingleSampledInfoEXT-rasterizationSamples-parameter", 78}, { {"rasterizationSamples must be a valid VkSampleCountFlagBits value", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultisampledRenderToSingleSampledInfoEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_INFO_EXT", 78}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewAttributesInfoNVX-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_ATTRIBUTES_INFO_NVX", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-offset-07861", 70}, { {"The offset.x member of any element of pPerViewRenderAreas must be greater than or equal to 0", 92}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-offset-07862", 70}, { {"The offset.y member of any element of pPerViewRenderAreas must be greater than or equal to 0", 92}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-offset-07863", 70}, { {"The sum of the offset.x and extent.width members of any element of pPerViewRenderAreas must be less than or equal to maxFramebufferWidth", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-offset-07864", 70}, { {"The sum of the offset.y and extent.height members of any element of pPerViewRenderAreas must be less than or equal to maxFramebufferHeight", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-pNext-07865", 69}, { {"If this structure is in the pNext chain of VkRenderPassBeginInfo and if the render pass object included an element in VkRenderPassMultiviewCreateInfo::pViewMasks that set bit n, then perViewRenderAreaCount must be at least equal to n+1", 235}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-pNext-07866", 69}, { {"If this structure is in the pNext chain of VkRenderingInfo and if VkRenderingInfo::viewMask set bit n, then perViewRenderAreaCount must be at least equal to n+1", 160}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-pPerViewRenderAreas-parameter", 87}, { {"If perViewRenderAreaCount is not 0, pPerViewRenderAreas must be a valid pointer to an array of perViewRenderAreaCount VkRect2D structures", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_RENDER_AREAS_RENDER_PASS_BEGIN_INFO_QCOM", 91}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkMutableDescriptorTypeCreateInfoEXT-pMutableDescriptorTypeLists-parameter", 79}, { {"If mutableDescriptorTypeListCount is not 0, pMutableDescriptorTypeLists must be a valid pointer to an array of mutableDescriptorTypeListCount valid VkMutableDescriptorTypeListEXT structures", 189}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeCreateInfoEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_MUTABLE_DESCRIPTOR_TYPE_CREATE_INFO_EXT", 71}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-descriptorTypeCount-04597", 61}, { {"descriptorTypeCount must not be 0 if the corresponding binding is of VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-descriptorTypeCount-04599", 61}, { {"descriptorTypeCount must be 0 if the corresponding binding is not of VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-04598", 58}, { {"pDescriptorTypes must be a valid pointer to an array of descriptorTypeCount valid, unique VkDescriptorType values if the given binding is of VK_DESCRIPTOR_TYPE_MUTABLE_EXT type", 176}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-04600", 58}, { {"pDescriptorTypes must not contain VK_DESCRIPTOR_TYPE_MUTABLE_EXT", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-04601", 58}, { {"pDescriptorTypes must not contain VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC", 75}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-04602", 58}, { {"pDescriptorTypes must not contain VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC", 75}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-04603", 58}, { {"pDescriptorTypes must not contain VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 73}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-09696", 58}, { {"pDescriptorTypes must not contain VK_DESCRIPTOR_TYPE_TENSOR_ARM", 63}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkMutableDescriptorTypeListEXT-pDescriptorTypes-parameter", 62}, { {"If descriptorTypeCount is not 0, pDescriptorTypes must be a valid pointer to an array of descriptorTypeCount valid VkDescriptorType values", 138}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkNativeBufferFormatPropertiesOHOS-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_NATIVE_BUFFER_FORMAT_PROPERTIES_OHOS", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkNativeBufferPropertiesOHOS-pNext-pNext", 45}, { {"pNext must be NULL or a pointer to a valid instance of VkNativeBufferFormatPropertiesOHOS", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-VkNativeBufferPropertiesOHOS-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_NATIVE_BUFFER_PROPERTIES_OHOS", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkNativeBufferPropertiesOHOS-sType-unique", 46}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkNativeBufferUsageOHOS-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_NATIVE_BUFFER_USAGE_OHOS", 56}, { "chapters/resources.html", 23} } },
        { {"VUID-VkOpaqueCaptureDataCreateInfoEXT-pData-parameter", 53}, { {"If pData is not NULL, pData must be a valid pointer to a valid VkHostAddressRangeConstEXT structure", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-VkOpaqueCaptureDataCreateInfoEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DATA_CREATE_INFO_EXT", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkOpaqueCaptureDescriptorDataCreateInfoEXT-opaqueCaptureDescriptorData-parameter", 85}, { {"opaqueCaptureDescriptorData must be a pointer value", 51}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkOpaqueCaptureDescriptorDataCreateInfoEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DESCRIPTOR_DATA_CREATE_INFO_EXT", 78}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkOpticalFlowExecuteInfoNV-flags-parameter", 47}, { {"flags must be a valid combination of VkOpticalFlowExecuteFlagBitsNV values", 74}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowExecuteInfoNV-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowExecuteInfoNV-pRegions-parameter", 50}, { {"If regionCount is not 0, pRegions must be a valid pointer to an array of regionCount VkRect2D structures", 104}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowExecuteInfoNV-regionCount-07593", 49}, { {"regionCount must be 0 if VK_OPTICAL_FLOW_SESSION_CREATE_ALLOW_REGIONS_BIT_NV was not set for VkOpticalFlowSessionNV on which this command is operating", 150}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowExecuteInfoNV-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_OPTICAL_FLOW_EXECUTE_INFO_NV", 60}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowImageFormatInfoNV-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_INFO_NV", 65}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowImageFormatInfoNV-usage-parameter", 51}, { {"usage must be a valid combination of VkOpticalFlowUsageFlagBitsNV values", 72}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowImageFormatInfoNV-usage-requiredbitmask", 57}, { {"usage must not be 0", 19}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowImageFormatPropertiesNV-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowImageFormatPropertiesNV-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_NV", 71}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-costFormat-07585", 54}, { {"costFormat must be one of the formats returned by vkGetPhysicalDeviceOpticalFlowImageFormatsNV for VK_OPTICAL_FLOW_USAGE_COST_BIT_NV if VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_COST_BIT_NV is set in flags", 201}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-costFormat-parameter", 58}, { {"If costFormat is not 0, costFormat must be a valid VkFormat value", 65}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-07588", 49}, { {"VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_HINT_BIT_NV must not be set in flags if VkPhysicalDeviceOpticalFlowPropertiesNV::hintSupported is VK_FALSE", 144}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-07589", 49}, { {"VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_COST_BIT_NV must not be set in flags if VkPhysicalDeviceOpticalFlowPropertiesNV::costSupported is VK_FALSE", 144}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-07590", 49}, { {"VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_GLOBAL_FLOW_BIT_NV must not be set in flags if VkPhysicalDeviceOpticalFlowPropertiesNV::globalFlowSupported is VK_FALSE", 157}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-07591", 49}, { {"VK_OPTICAL_FLOW_SESSION_CREATE_ALLOW_REGIONS_BIT_NV must not be set in flags if VkPhysicalDeviceOpticalFlowPropertiesNV::maxNumRegionsOfInterest is 0", 149}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-07592", 49}, { {"VK_OPTICAL_FLOW_SESSION_CREATE_BOTH_DIRECTIONS_BIT_NV must not be set in flags if VkPhysicalDeviceOpticalFlowPropertiesNV::bidirectionalFlowSupported is VK_FALSE", 161}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flags-parameter", 53}, { {"flags must be a valid combination of VkOpticalFlowSessionCreateFlagBitsNV values", 80}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flowVectorFormat-07584", 60}, { {"flowVectorFormat must be one of the formats returned by vkGetPhysicalDeviceOpticalFlowImageFormatsNV for VK_OPTICAL_FLOW_USAGE_OUTPUT_BIT_NV", 140}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-flowVectorFormat-parameter", 64}, { {"flowVectorFormat must be a valid VkFormat value", 47}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-height-07582", 50}, { {"height must be greater than or equal to VkPhysicalDeviceOpticalFlowPropertiesNV::minHeight and less than or equal to VkPhysicalDeviceOpticalFlowPropertiesNV::maxHeight", 167}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-hintGridSize-07587", 56}, { {"hintGridSize must be exactly one of the bits reported in VkPhysicalDeviceOpticalFlowPropertiesNV::supportedHintGridSizes if VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_HINT_BIT_NV is set in flags", 189}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-hintGridSize-parameter", 60}, { {"hintGridSize must be a valid combination of VkOpticalFlowGridSizeFlagBitsNV values", 82}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-imageFormat-07583", 55}, { {"imageFormat must be one of the formats returned by vkGetPhysicalDeviceOpticalFlowImageFormatsNV for VK_OPTICAL_FLOW_USAGE_INPUT_BIT_NV", 134}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-imageFormat-parameter", 59}, { {"imageFormat must be a valid VkFormat value", 42}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-outputGridSize-07586", 58}, { {"outputGridSize must be exactly one of the bits reported in VkPhysicalDeviceOpticalFlowPropertiesNV::supportedOutputGridSizes", 124}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-outputGridSize-parameter", 62}, { {"outputGridSize must be a valid combination of VkOpticalFlowGridSizeFlagBitsNV values", 84}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-outputGridSize-requiredbitmask", 68}, { {"outputGridSize must not be 0", 28}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-pNext-pNext", 49}, { {"pNext must be NULL or a pointer to a valid instance of VkOpticalFlowSessionCreatePrivateDataInfoNV", 98}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-performanceLevel-parameter", 64}, { {"If performanceLevel is not 0, performanceLevel must be a valid VkOpticalFlowPerformanceLevelNV value", 100}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_INFO_NV", 67}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreateInfoNV-width-07581", 49}, { {"width must be greater than or equal to VkPhysicalDeviceOpticalFlowPropertiesNV::minWidth and less than or equal to VkPhysicalDeviceOpticalFlowPropertiesNV::maxWidth", 164}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreatePrivateDataInfoNV-pPrivateData-parameter", 71}, { {"pPrivateData must be a pointer value", 36}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOpticalFlowSessionCreatePrivateDataInfoNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_PRIVATE_DATA_INFO_NV", 80}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-VkOutOfBandQueueTypeInfoNV-queueType-parameter", 51}, { {"queueType must be a valid VkOutOfBandQueueTypeNV value", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkOutOfBandQueueTypeInfoNV-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_OUT_OF_BAND_QUEUE_TYPE_INFO_NV", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPartitionedAccelerationStructureFlagsNV-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_FLAGS_NV", 75}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureInstancesInputNV-flags-parameter", 71}, { {"flags must be a valid combination of VkBuildAccelerationStructureFlagBitsKHR values", 83}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureInstancesInputNV-pNext-pNext", 67}, { {"pNext must be NULL or a pointer to a valid instance of VkPartitionedAccelerationStructureFlagsNV", 96}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureInstancesInputNV-partitionCount-10535", 76}, { {"The sum of partitionCount and maxInstanceInGlobalPartitionCount must be less than or equal to VkPhysicalDevicePartitionedAccelerationStructurePropertiesNV::maxPartitionCount", 173}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureInstancesInputNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_INSTANCES_INPUT_NV", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureInstancesInputNV-sType-unique", 68}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureUpdateInstanceDataNV-None-10572", 70}, { {"The instance must have either been created with flag VK_PARTITIONED_ACCELERATION_STRUCTURE_INSTANCE_FLAG_ENABLE_EXPLICIT_BOUNDING_BOX_NV or did not have an acceleration structure assigned with VkPartitionedAccelerationStructureWriteInstanceDataNV", 246}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureUpdateInstanceDataNV-accelerationStructure-parameter", 91}, { {"accelerationStructure must be a valid VkDeviceAddress value", 59}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureUpdateInstanceDataNV-instanceContributionToHitGroupIndex-10571", 101}, { {"The most significant 8 bits of instanceContributionToHitGroupIndex must be 0", 76}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureUpdateInstanceDataNV-instanceIndex-10573", 79}, { {"instanceIndex must be less than VkBuildPartitionedAccelerationStructureInfoNV::input.instanceCount", 98}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-accelerationStructure-parameter", 90}, { {"accelerationStructure must be a valid VkDeviceAddress value", 59}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-explicitAABB-10570", 77}, { {"explicitAABB must be a valid bounding box if instance was created with flag VK_PARTITIONED_ACCELERATION_STRUCTURE_INSTANCE_FLAG_ENABLE_EXPLICIT_BOUNDING_BOX_NV set", 163}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-instanceContributionToHitGroupIndex-10567", 100}, { {"The most significant 8 bits of instanceContributionToHitGroupIndex must be 0", 76}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-instanceFlags-parameter", 82}, { {"instanceFlags must be a valid combination of VkPartitionedAccelerationStructureInstanceFlagBitsNV values", 104}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-instanceIndex-10568", 78}, { {"instanceIndex must be less than VkBuildPartitionedAccelerationStructureInfoNV::input.instanceCount", 98}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-instanceMask-10566", 77}, { {"The most significant 24 bits of instanceMask must be 0", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWriteInstanceDataNV-partitionIndex-10569", 79}, { {"partitionIndex must be less than VkBuildPartitionedAccelerationStructureInfoNV::input.partitionCount", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWritePartitionTranslationDataNV-enablePartitionTranslation-10575", 103}, { {"The partitioned acceleration structure must have the VkPartitionedAccelerationStructureFlagsNV::enablePartitionTranslation flag set", 131}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPartitionedAccelerationStructureWritePartitionTranslationDataNV-partitionIndex-10574", 91}, { {"partitionIndex must be less than VkBuildPartitionedAccelerationStructureInfoNV::input.partitionCount", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkPastPresentationTimingEXT-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_EXT", 60}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingInfoEXT-flags-parameter", 52}, { {"flags must be a valid combination of VkPastPresentationTimingFlagBitsEXT values", 79}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingInfoEXT-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_INFO_EXT", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingInfoEXT-swapchain-parameter", 56}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingPropertiesEXT-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPastPresentationTimingPropertiesEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_PROPERTIES_EXT", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPerTileBeginInfoQCOM-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkPerTileBeginInfoQCOM-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_PER_TILE_BEGIN_INFO_QCOM", 56}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkPerTileEndInfoQCOM-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkPerTileEndInfoQCOM-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_PER_TILE_END_INFO_QCOM", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkPerfHintInfoQCOM-pNext-pNext", 35}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerfHintInfoQCOM-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_PERF_HINT_INFO_QCOM", 51}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerfHintInfoQCOM-scale-12390", 35}, { {"scale must be less than or equal to 100", 39}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerfHintInfoQCOM-type-12389", 34}, { {"If type is not VK_PERF_HINT_TYPE_FREQUENCY_SCALED_QCOM, scale must equal 0", 74}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerfHintInfoQCOM-type-parameter", 38}, { {"type must be a valid VkPerfHintTypeQCOM value", 45}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceConfigurationAcquireInfoINTEL-pNext-pNext", 59}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceConfigurationAcquireInfoINTEL-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_CONFIGURATION_ACQUIRE_INFO_INTEL", 76}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceConfigurationAcquireInfoINTEL-type-parameter", 62}, { {"type must be a valid VkPerformanceConfigurationTypeINTEL value", 62}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceCounterARM-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterARM-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_ARM", 55}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterDescriptionARM-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterDescriptionARM-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_ARM", 67}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterDescriptionKHR-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterDescriptionKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_KHR", 67}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterKHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceCounterKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_KHR", 55}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPerformanceMarkerInfoINTEL-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceMarkerInfoINTEL-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_MARKER_INFO_INTEL", 61}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceOverrideInfoINTEL-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceOverrideInfoINTEL-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_OVERRIDE_INFO_INTEL", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceOverrideInfoINTEL-type-parameter", 50}, { {"type must be a valid VkPerformanceOverrideTypeINTEL value", 57}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceQuerySubmitInfoKHR-counterPassIndex-03221", 59}, { {"counterPassIndex must be less than the number of counter passes required by any queries within the batch. The required number of counter passes for a performance query is obtained by calling vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR", 246}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkPerformanceQuerySubmitInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_QUERY_SUBMIT_INFO_KHR", 65}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkPerformanceStreamMarkerInfoINTEL-marker-02735", 52}, { {"The value written by the application into marker must only used the valid bits as reported by vkGetPerformanceParameterINTEL with the VK_PERFORMANCE_PARAMETER_TYPE_STREAM_MARKER_VALID_BITS_INTEL", 194}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceStreamMarkerInfoINTEL-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPerformanceStreamMarkerInfoINTEL-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PERFORMANCE_STREAM_MARKER_INFO_INTEL", 68}, { "chapters/queries.html", 21} } },
        { {"VUID-VkPhysicalDevice16BitStorageFeatures-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_16BIT_STORAGE_FEATURES", 70}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevice4444FormatsFeaturesEXT-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevice8BitStorageFeatures-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_8BIT_STORAGE_FEATURES", 69}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceASTCDecodeFeaturesEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ASTC_DECODE_FEATURES_EXT", 72}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceAccelerationStructureFeaturesKHR-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_FEATURES_KHR", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceAccelerationStructurePropertiesKHR-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_PROPERTIES_KHR", 85}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceAddressBindingReportFeaturesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ADDRESS_BINDING_REPORT_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceAmigoProfilingFeaturesSEC-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_AMIGO_PROFILING_FEATURES_SEC", 76}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkPhysicalDeviceAntiLagFeaturesAMD-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ANTI_LAG_FEATURES_AMD", 69}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceAttachmentFeedbackLoopDynamicStateFeaturesEXT-sType-sType", 78}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_DYNAMIC_STATE_FEATURES_EXT", 99}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceAttachmentFeedbackLoopLayoutFeaturesEXT-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_LAYOUT_FEATURES_EXT", 92}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceBlendOperationAdvancedFeaturesEXT-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_FEATURES_EXT", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_PROPERTIES_EXT", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceBorderColorSwizzleFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BORDER_COLOR_SWIZZLE_FEATURES_EXT", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceBufferDeviceAddressFeatures-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceBufferDeviceAddressFeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_EXT", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceClusterAccelerationStructureFeaturesNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_ACCELERATION_STRUCTURE_FEATURES_NV", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceClusterAccelerationStructurePropertiesNV-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_ACCELERATION_STRUCTURE_PROPERTIES_NV", 92}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceClusterCullingShaderFeaturesHUAWEI-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_FEATURES_HUAWEI", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_PROPERTIES_HUAWEI", 88}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceClusterCullingShaderVrsFeaturesHUAWEI-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_VRS_FEATURES_HUAWEI", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCoherentMemoryFeaturesAMD-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COHERENT_MEMORY_FEATURES_AMD", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceColorWriteEnableFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COLOR_WRITE_ENABLE_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCommandBufferInheritanceFeaturesNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMMAND_BUFFER_INHERITANCE_FEATURES_NV", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceComputeOccupancyPriorityFeaturesNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_OCCUPANCY_PRIORITY_FEATURES_NV", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceComputeShaderDerivativesFeaturesKHR-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_FEATURES_KHR", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceComputeShaderDerivativesPropertiesKHR-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_PROPERTIES_KHR", 89}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceConditionalRenderingFeaturesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceConservativeRasterizationPropertiesEXT-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONSERVATIVE_RASTERIZATION_PROPERTIES_EXT", 89}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrix2FeaturesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_2_FEATURES_NV", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrix2PropertiesNV-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_2_PROPERTIES_NV", 82}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixConversionFeaturesQCOM-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_CONVERSION_FEATURES_QCOM", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixDecodeVectorFeaturesNV-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_DECODE_VECTOR_FEATURES_NV", 92}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixFeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_KHR", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixFeaturesNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixPropertiesKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_KHR", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCooperativeMatrixPropertiesNV-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCooperativeVectorFeaturesNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_VECTOR_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCooperativeVectorPropertiesNV-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_VECTOR_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCopyMemoryIndirectFeaturesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_KHR", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCopyMemoryIndirectFeaturesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_NV", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_PROPERTIES_KHR", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCornerSampledImageFeaturesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CORNER_SAMPLED_IMAGE_FEATURES_NV", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCoverageReductionModeFeaturesNV-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COVERAGE_REDUCTION_MODE_FEATURES_NV", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCubicClampFeaturesQCOM-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_CLAMP_FEATURES_QCOM", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCubicWeightsFeaturesQCOM-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_WEIGHTS_FEATURES_QCOM", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCudaKernelLaunchFeaturesNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCudaKernelLaunchPropertiesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCustomBorderColorFeaturesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_FEATURES_EXT", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceCustomBorderColorPropertiesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_PROPERTIES_EXT", 82}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceCustomResolveFeaturesEXT-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_RESOLVE_FEATURES_EXT", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDataGraphFeaturesARM-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_FEATURES_ARM", 71}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDataGraphModelFeaturesQCOM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_MODEL_FEATURES_QCOM", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDataGraphNeuralAcceleratorStatisticsFeaturesARM-sType-sType", 80}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_NEURAL_ACCELERATOR_STATISTICS_FEATURES_ARM", 101}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDataGraphOpticalFlowFeaturesARM-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_OPTICAL_FLOW_FEATURES_ARM", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDedicatedAllocationImageAliasingFeaturesNV-sType-sType", 75}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEDICATED_ALLOCATION_IMAGE_ALIASING_FEATURES_NV", 95}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDenseGeometryFormatFeaturesAMDX-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DENSE_GEOMETRY_FORMAT_FEATURES_AMDX", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthBiasControlFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_BIAS_CONTROL_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthClampControlFeaturesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_CONTROL_FEATURES_EXT", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthClampZeroOneFeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_ZERO_ONE_FEATURES_KHR", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthClipControlFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_CONTROL_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthClipEnableFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_ENABLE_FEATURES_EXT", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDepthStencilResolveProperties-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_STENCIL_RESOLVE_PROPERTIES", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorBufferDensityMapPropertiesEXT-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_DENSITY_MAP_PROPERTIES_EXT", 92}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorBufferFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_FEATURES_EXT", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDescriptorBufferPropertiesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_PROPERTIES_EXT", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorBufferTensorFeaturesARM-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_TENSOR_FEATURES_ARM", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDescriptorBufferTensorPropertiesARM-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_TENSOR_PROPERTIES_ARM", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorHeapFeaturesEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_FEATURES_EXT", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDescriptorHeapPropertiesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_PROPERTIES_EXT", 78}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorHeapTensorPropertiesARM-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_TENSOR_PROPERTIES_ARM", 85}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorIndexingFeatures-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDescriptorIndexingProperties-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_PROPERTIES", 78}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDescriptorSetHostMappingFeaturesVALVE-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_SET_HOST_MAPPING_FEATURES_VALVE", 90}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkPhysicalDeviceDeviceAddressCommandsFeaturesKHR-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_ADDRESS_COMMANDS_FEATURES_KHR", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_COMPUTE_FEATURES_NV", 93}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_NV", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_EXT", 88}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_NV", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDeviceMemoryReportFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_MEMORY_REPORT_FEATURES_EXT", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDiagnosticsConfigFeaturesNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DIAGNOSTICS_CONFIG_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDiscardRectanglePropertiesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISCARD_RECTANGLE_PROPERTIES_EXT", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDisplacementMicromapFeaturesNV-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_FEATURES_NV", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDisplacementMicromapPropertiesNV-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_PROPERTIES_NV", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceDriverProperties-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRIVER_PROPERTIES", 65}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceDrmPropertiesEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRM_PROPERTIES_EXT", 66}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceDynamicRenderingFeatures-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_FEATURES", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDynamicRenderingLocalReadFeatures-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_LOCAL_READ_FEATURES", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceDynamicRenderingUnusedAttachmentsFeaturesEXT-sType-sType", 77}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_UNUSED_ATTACHMENTS_FEATURES_EXT", 97}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceElapsedTimerQueryFeaturesQCOM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ELAPSED_TIMER_QUERY_FEATURES_QCOM", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExclusiveScissorFeaturesNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXCLUSIVE_SCISSOR_FEATURES_NV", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedDynamicState2FeaturesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_2_FEATURES_EXT", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedDynamicState3FeaturesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_FEATURES_EXT", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedDynamicState3PropertiesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_PROPERTIES_EXT", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceExtendedDynamicStateFeaturesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedFlagsFeaturesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_FLAGS_FEATURES_KHR", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedSparseAddressSpaceFeaturesNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_FEATURES_NV", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_PROPERTIES_NV", 91}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-None-09499", 50}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must be a valid combination of VkBufferUsageFlagBits values", 144}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-None-09500", 50}, { {"If the pNext chain does not include a VkBufferUsageFlags2CreateInfo structure, usage must not be 0", 98}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-flags-parameter", 55}, { {"flags must be a valid combination of VkBufferCreateFlagBits values", 66}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-handleType-parameter", 60}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-pNext-pNext", 51}, { {"pNext must be NULL or a pointer to a valid instance of VkBufferUsageFlags2CreateInfo", 84}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_BUFFER_INFO", 68}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalBufferInfo-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalComputeQueuePropertiesNV-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_COMPUTE_QUEUE_PROPERTIES_NV", 84}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-VkPhysicalDeviceExternalFenceInfo-handleType-parameter", 59}, { {"handleType must be a valid VkExternalFenceHandleTypeFlagBits value", 66}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalFenceInfo-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalFenceInfo-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FENCE_INFO", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalFormatResolveFeaturesANDROID-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_FEATURES_ANDROID", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExternalFormatResolvePropertiesANDROID-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_PROPERTIES_ANDROID", 90}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceExternalImageFormatInfo-handleType-parameter", 65}, { {"If handleType is not 0, handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 91}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalImageFormatInfo-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_IMAGE_FORMAT_INFO", 74}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalMemoryHostPropertiesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_HOST_PROPERTIES_EXT", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceExternalMemoryRDMAFeaturesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_RDMA_FEATURES_NV", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExternalMemoryScreenBufferFeaturesQNX-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_SCREEN_BUFFER_FEATURES_QNX", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceExternalSemaphoreInfo-handleType-parameter", 63}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalSemaphoreInfo-pNext-pNext", 54}, { {"pNext must be NULL or a pointer to a valid instance of VkSemaphoreTypeCreateInfo", 80}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalSemaphoreInfo-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_SEMAPHORE_INFO", 71}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalSemaphoreInfo-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalTensorInfoARM-flags-parameter", 58}, { {"flags must be a valid combination of VkTensorCreateFlagBitsARM values", 69}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalTensorInfoARM-handleType-parameter", 63}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalTensorInfoARM-pDescription-parameter", 65}, { {"pDescription must be a valid pointer to a valid VkTensorDescriptionARM structure", 80}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalTensorInfoARM-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceExternalTensorInfoARM-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_TENSOR_INFO_ARM", 72}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceFaultFeaturesEXT-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_EXT", 66}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFaultFeaturesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_KHR", 66}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFaultPropertiesKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_PROPERTIES_KHR", 68}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFeatures2-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2", 58}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFloatControlsProperties-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT_CONTROLS_PROPERTIES", 73}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFormatPackFeaturesARM-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FORMAT_PACK_FEATURES_ARM", 72}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMap2FeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMap2PropertiesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_PROPERTIES_EXT", 85}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_FEATURES_EXT", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapLayeredFeaturesVALVE-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_LAYERED_FEATURES_VALVE", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapLayeredPropertiesVALVE-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_LAYERED_PROPERTIES_VALVE", 93}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapOffsetFeaturesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_FEATURES_EXT", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapOffsetPropertiesEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_PROPERTIES_EXT", 90}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentDensityMapPropertiesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_PROPERTIES_EXT", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentShaderBarycentricFeaturesKHR-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_FEATURES_KHR", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentShaderBarycentricPropertiesKHR-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_PROPERTIES_KHR", 90}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentShaderInterlockFeaturesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_INTERLOCK_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRateEnumsFeaturesNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_FEATURES_NV", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRateEnumsPropertiesNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_PROPERTIES_NV", 89}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRateFeaturesKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_FEATURES_KHR", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRateKHR-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRateKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_KHR", 73}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPhysicalDeviceFragmentShadingRatePropertiesKHR-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_PROPERTIES_KHR", 84}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceFrameBoundaryFeaturesEXT-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAME_BOUNDARY_FEATURES_EXT", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceGlobalPriorityQueryFeatures-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GLOBAL_PRIORITY_QUERY_FEATURES", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceGpaFeaturesAMD-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_FEATURES_AMD", 64}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceGpaProperties2AMD-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_PROPERTIES_2_AMD", 68}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceGpaPropertiesAMD-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_PROPERTIES_AMD", 66}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceGraphicsPipelineLibraryFeaturesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceGraphicsPipelineLibraryPropertiesEXT-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_PROPERTIES_EXT", 88}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceGroupProperties-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceGroupProperties-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GROUP_PROPERTIES", 64}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceHdrVividFeaturesHUAWEI-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HDR_VIVID_FEATURES_HUAWEI", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceHostImageCopyFeatures-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_FEATURES", 72}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceHostImageCopyProperties-pCopyDstLayouts-parameter", 70}, { {"If copyDstLayoutCount is not 0, and pCopyDstLayouts is not NULL, pCopyDstLayouts must be a valid pointer to an array of copyDstLayoutCount VkImageLayout values", 159}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceHostImageCopyProperties-pCopySrcLayouts-parameter", 70}, { {"If copySrcLayoutCount is not 0, and pCopySrcLayouts is not NULL, pCopySrcLayouts must be a valid pointer to an array of copySrcLayoutCount VkImageLayout values", 159}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceHostImageCopyProperties-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_PROPERTIES", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceHostQueryResetFeatures-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceIDProperties-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ID_PROPERTIES", 61}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceImage2DViewOf3DFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_2D_VIEW_OF_3D_FEATURES_EXT", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageAlignmentControlFeaturesMESA-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ALIGNMENT_CONTROL_FEATURES_MESA", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageAlignmentControlPropertiesMESA-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ALIGNMENT_CONTROL_PROPERTIES_MESA", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceImageCompressionControlFeaturesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageCompressionControlSwapchainFeaturesEXT-sType-sType", 76}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_SWAPCHAIN_FEATURES_EXT", 96}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-maintenance11-13351", 70}, { {"If the maintenance11 feature is supported and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 0", 138}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT", 82}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-sharingMode-02314", 68}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, then pQueueFamilyIndices must be a valid pointer to an array of queueFamilyIndexCount uint32_t values", 147}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-sharingMode-02315", 68}, { {"If the maintenance11 feature is not supported and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 1", 142}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-sharingMode-02316", 68}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, each element of pQueueFamilyIndices must be unique and must be less than the pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties2 for the physicalDevice that was used to create device", 256}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageDrmFormatModifierInfoEXT-sharingMode-parameter", 72}, { {"sharingMode must be a valid VkSharingMode value", 47}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-flags-parameter", 53}, { {"flags must be a valid combination of VkImageCreateFlagBits values", 65}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-format-parameter", 54}, { {"format must be a valid VkFormat value", 37}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-pNext-pNext", 49}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDataGraphOpticalFlowImageFormatInfoARM, VkImageCompressionControlEXT, VkImageCreateFlags2CreateInfoKHR, VkImageFormatListCreateInfo, VkImageStencilUsage2CreateInfoKHR, VkImageStencilUsageCreateInfo, VkImageUsageFlags2CreateInfoKHR, VkOpticalFlowImageFormatInfoNV, VkPhysicalDeviceExternalImageFormatInfo, VkPhysicalDeviceImageDrmFormatModifierInfoEXT, VkPhysicalDeviceImageViewImageFormatInfoEXT, or VkVideoProfileListInfoKHR", 558}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-tiling-02249", 50}, { {"tiling must be VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT if and only if the pNext chain includes VkPhysicalDeviceImageDrmFormatModifierInfoEXT", 140}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-tiling-02313", 50}, { {"If tiling is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT and flags contains VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT, then the pNext chain must include a VkImageFormatListCreateInfo structure with non-zero viewFormatCount", 211}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-tiling-parameter", 54}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-type-parameter", 52}, { {"type must be a valid VkImageType value", 38}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-usage-parameter", 53}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageFormatInfo2-usage-requiredbitmask", 59}, { {"usage must not be 0", 19}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageProcessing2FeaturesQCOM-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_FEATURES_QCOM", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageProcessing2PropertiesQCOM-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_PROPERTIES_QCOM", 82}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceImageProcessing3FeaturesQCOM-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_3_FEATURES_QCOM", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageProcessingFeaturesQCOM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_FEATURES_QCOM", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageProcessingPropertiesQCOM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_PROPERTIES_QCOM", 80}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceImageRobustnessFeatures-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ROBUSTNESS_FEATURES", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageSlicedViewOf3DFeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_SLICED_VIEW_OF_3D_FEATURES_EXT", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImageViewImageFormatInfoEXT-imageViewType-parameter", 72}, { {"imageViewType must be a valid VkImageViewType value", 51}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageViewImageFormatInfoEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_IMAGE_FORMAT_INFO_EXT", 80}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkPhysicalDeviceImageViewMinLodFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_MIN_LOD_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceImagelessFramebufferFeatures-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGELESS_FRAMEBUFFER_FEATURES", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceIndexTypeUint8Features-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INDEX_TYPE_UINT8_FEATURES", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceInheritedViewportScissorFeaturesNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INHERITED_VIEWPORT_SCISSOR_FEATURES_NV", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceInlineUniformBlockFeatures-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_FEATURES", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceInlineUniformBlockProperties-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_PROPERTIES", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceInternallySynchronizedQueuesFeaturesKHR-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INTERNALLY_SYNCHRONIZED_QUEUES_FEATURES_KHR", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceInvocationMaskFeaturesHUAWEI-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INVOCATION_MASK_FEATURES_HUAWEI", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceLayeredApiPropertiesKHR-pNext-pNext", 56}, { {"pNext must be NULL or a pointer to a valid instance of VkPhysicalDeviceLayeredApiVulkanPropertiesKHR", 100}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiPropertiesKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_KHR", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiPropertiesKHR-sType-unique", 57}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiPropertiesListKHR-pLayeredApis-parameter", 71}, { {"If layeredApiCount is not 0, and pLayeredApis is not NULL, pLayeredApis must be a valid pointer to an array of layeredApiCount VkPhysicalDeviceLayeredApiPropertiesKHR structures", 177}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiPropertiesListKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_LIST_KHR", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiVulkanPropertiesKHR-pNext-10011", 62}, { {"Only VkPhysicalDeviceDriverProperties and VkPhysicalDeviceIDProperties are allowed in the pNext chain of properties", 115}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredApiVulkanPropertiesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_VULKAN_PROPERTIES_KHR", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLayeredDriverPropertiesMSFT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_DRIVER_PROPERTIES_MSFT", 78}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceLegacyDitheringFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_DITHERING_FEATURES_EXT", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceLegacyVertexAttributesFeaturesEXT-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_VERTEX_ATTRIBUTES_FEATURES_EXT", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceLegacyVertexAttributesPropertiesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_VERTEX_ATTRIBUTES_PROPERTIES_EXT", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLineRasterizationFeatures-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_FEATURES", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceLineRasterizationProperties-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_PROPERTIES", 77}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceLinearColorAttachmentFeaturesNV-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINEAR_COLOR_ATTACHMENT_FEATURES_NV", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance10FeaturesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_10_FEATURES_KHR", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance10PropertiesKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_10_PROPERTIES_KHR", 77}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance11FeaturesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_11_FEATURES_KHR", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance3Properties-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_3_PROPERTIES", 72}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance4Features-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES", 70}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance4Properties-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_PROPERTIES", 72}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance5Features-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_FEATURES", 70}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance5Properties-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_PROPERTIES", 72}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance6Features-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_FEATURES", 70}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance6Properties-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_PROPERTIES", 72}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance7FeaturesKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_7_FEATURES_KHR", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance7PropertiesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_7_PROPERTIES_KHR", 76}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMaintenance8FeaturesKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_8_FEATURES_KHR", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance9FeaturesKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_9_FEATURES_KHR", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMaintenance9PropertiesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_9_PROPERTIES_KHR", 76}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMapMemoryPlacedFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAP_MEMORY_PLACED_FEATURES_EXT", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMapMemoryPlacedPropertiesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAP_MEMORY_PLACED_PROPERTIES_EXT", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMemoryBudgetPropertiesEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_BUDGET_PROPERTIES_EXT", 76}, { "chapters/memory.html", 20} } },
        { {"VUID-VkPhysicalDeviceMemoryDecompressionFeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_FEATURES_EXT", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMemoryDecompressionPropertiesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_PROPERTIES_EXT", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMemoryPriorityFeaturesEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PRIORITY_FEATURES_EXT", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMemoryProperties2-pNext-pNext", 50}, { {"pNext must be NULL or a pointer to a valid instance of VkPhysicalDeviceMemoryBudgetPropertiesEXT", 96}, { "chapters/memory.html", 20} } },
        { {"VUID-VkPhysicalDeviceMemoryProperties2-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PROPERTIES_2", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkPhysicalDeviceMemoryProperties2-sType-unique", 51}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkPhysicalDeviceMeshShaderFeaturesEXT-multiviewMeshShader-07032", 68}, { {"If multiviewMeshShader is enabled then VkPhysicalDeviceMultiviewFeaturesKHR::multiview must also be enabled", 107}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMeshShaderFeaturesEXT-primitiveFragmentShadingRateMeshShader-07033", 87}, { {"If primitiveFragmentShadingRateMeshShader is enabled then VkPhysicalDeviceFragmentShadingRateFeaturesKHR::primitiveFragmentShadingRate must also be enabled", 155}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMeshShaderFeaturesEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_EXT", 72}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMeshShaderFeaturesNV-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_NV", 71}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMeshShaderPropertiesEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_EXT", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMeshShaderPropertiesNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_NV", 73}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMultiDrawFeaturesEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_FEATURES_EXT", 71}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiDrawPropertiesEXT-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_PROPERTIES_EXT", 73}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMultisampledRenderToSingleSampledFeaturesEXT-sType-sType", 77}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_FEATURES_EXT", 98}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultisampledRenderToSwapchainFeaturesEXT-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SWAPCHAIN_FEATURES_EXT", 93}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewFeatures-multiviewGeometryShader-00580", 68}, { {"If multiviewGeometryShader is enabled then multiview must also be enabled", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewFeatures-multiviewTessellationShader-00581", 72}, { {"If multiviewTessellationShader is enabled then multiview must also be enabled", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewFeatures-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_FEATURES", 66}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewPerViewAttributesPropertiesNVX-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_ATTRIBUTES_PROPERTIES_NVX", 92}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMultiviewPerViewRenderAreasFeaturesQCOM-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_RENDER_AREAS_FEATURES_QCOM", 93}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewPerViewViewportsFeaturesQCOM-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_VIEWPORTS_FEATURES_QCOM", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceMultiviewProperties-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PROPERTIES", 68}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MUTABLE_DESCRIPTOR_TYPE_FEATURES_EXT", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceNestedCommandBufferFeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_FEATURES_EXT", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceNestedCommandBufferPropertiesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_PROPERTIES_EXT", 84}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceNonSeamlessCubeMapFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceOpacityMicromapFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_EXT", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceOpacityMicromapFeaturesKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_KHR", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceOpacityMicromapPropertiesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_EXT", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceOpacityMicromapPropertiesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_KHR", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceOpticalFlowFeaturesNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_FEATURES_NV", 72}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceOpticalFlowPropertiesNV-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_PROPERTIES_NV", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePCIBusInfoPropertiesEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PCI_BUS_INFO_PROPERTIES_EXT", 75}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDevicePageableDeviceLocalMemoryFeaturesEXT-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PAGEABLE_DEVICE_LOCAL_MEMORY_FEATURES_EXT", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePartitionedAccelerationStructureFeaturesNV-sType-sType", 75}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PARTITIONED_ACCELERATION_STRUCTURE_FEATURES_NV", 94}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePartitionedAccelerationStructurePropertiesNV-sType-sType", 77}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PARTITIONED_ACCELERATION_STRUCTURE_PROPERTIES_NV", 96}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePerStageDescriptorSetFeaturesNV-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PER_STAGE_DESCRIPTOR_SET_FEATURES_NV", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePerformanceCountersByRegionFeaturesARM-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_COUNTERS_BY_REGION_FEATURES_ARM", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePerformanceCountersByRegionPropertiesARM-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_COUNTERS_BY_REGION_PROPERTIES_ARM", 93}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePerformanceQueryFeaturesKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_FEATURES_KHR", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePerformanceQueryPropertiesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_PROPERTIES_KHR", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePipelineBinaryFeaturesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_BINARY_FEATURES_KHR", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineBinaryPropertiesKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_BINARY_PROPERTIES_KHR", 78}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePipelineCacheIncrementalModeFeaturesSEC-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CACHE_INCREMENTAL_MODE_FEATURES_SEC", 92}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkPhysicalDevicePipelineCreationCacheControlFeatures-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CREATION_CACHE_CONTROL_FEATURES", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineExecutablePropertiesFeaturesKHR-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_EXECUTABLE_PROPERTIES_FEATURES_KHR", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineLibraryGroupHandlesFeaturesEXT-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_LIBRARY_GROUP_HANDLES_FEATURES_EXT", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineOpacityMicromapFeaturesARM-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_OPACITY_MICROMAP_FEATURES_ARM", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelinePropertiesFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROPERTIES_FEATURES_EXT", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineProtectedAccessFeatures-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROTECTED_ACCESS_FEATURES", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineRobustnessFeatures-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePipelineRobustnessProperties-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_PROPERTIES", 78}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePointClippingProperties-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_POINT_CLIPPING_PROPERTIES", 73}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePortabilitySubsetFeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_FEATURES_KHR", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePortabilitySubsetPropertiesKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_PROPERTIES_KHR", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePresentBarrierFeaturesNV-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_BARRIER_FEATURES_NV", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentId2FeaturesKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_2_FEATURES_KHR", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentIdFeaturesKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_FEATURES_KHR", 71}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentMeteringFeaturesNV-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_METERING_FEATURES_NV", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentModeFifoLatestReadyFeaturesKHR-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_MODE_FIFO_LATEST_READY_FEATURES_KHR", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentTimingFeaturesEXT-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_TIMING_FEATURES_EXT", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentWait2FeaturesKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_2_FEATURES_KHR", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePresentWaitFeaturesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_FEATURES_KHR", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePrimitiveRestartIndexFeaturesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_RESTART_INDEX_FEATURES_EXT", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePrimitiveTopologyListRestartFeaturesEXT-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_TOPOLOGY_LIST_RESTART_FEATURES_EXT", 92}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePrimitivesGeneratedQueryFeaturesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVES_GENERATED_QUERY_FEATURES_EXT", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePrivateDataFeatures-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIVATE_DATA_FEATURES", 69}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceProperties2-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid struct for extending VkPhysicalDeviceProperties2", 167}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceProperties2-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2", 60}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceProperties2-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceProtectedMemoryFeatures-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_FEATURES", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceProtectedMemoryProperties-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_PROPERTIES", 75}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceProvokingVertexFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_FEATURES_EXT", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceProvokingVertexPropertiesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_PROPERTIES_EXT", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePushConstantBankFeaturesNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_CONSTANT_BANK_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDevicePushConstantBankPropertiesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_CONSTANT_BANK_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDevicePushDescriptorProperties-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_DESCRIPTOR_PROPERTIES", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM-engineType-parameter", 85}, { {"engineType must be a valid VkPhysicalDeviceDataGraphProcessingEngineTypeARM value", 81}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM-pNext-pNext", 76}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM-sType-sType", 76}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_FAMILY_DATA_GRAPH_PROCESSING_ENGINE_INFO_ARM", 98}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkPhysicalDeviceQueuePerfHintFeaturesQCOM-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_PERF_HINT_FEATURES_QCOM", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceQueuePerfHintPropertiesQCOM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_PERF_HINT_PROPERTIES_QCOM", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRGBA10X6FormatsFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RGBA10X6_FORMATS_FEATURES_EXT", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRasterizationOrderAttachmentAccessFeaturesEXT-sType-sType", 78}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_FEATURES_EXT", 98}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRawAccessChainsFeaturesNV-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAW_ACCESS_CHAINS_FEATURES_NV", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayQueryFeaturesKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_QUERY_FEATURES_KHR", 70}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingInvocationReorderFeaturesEXT-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_EXT", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingInvocationReorderFeaturesNV-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_NV", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingInvocationReorderPropertiesEXT-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_EXT", 93}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRayTracingInvocationReorderPropertiesNV-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_NV", 92}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRayTracingLinearSweptSpheresFeaturesNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_LINEAR_SWEPT_SPHERES_FEATURES_NV", 92}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingMaintenance1FeaturesKHR-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MAINTENANCE_1_FEATURES_KHR", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingMotionBlurFeaturesNV-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MOTION_BLUR_FEATURES_NV", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingPipelineFeaturesKHR-rayTracingPipelineShaderGroupHandleCaptureReplayMixed-03575", 110}, { {"If rayTracingPipelineShaderGroupHandleCaptureReplayMixed is VK_TRUE, rayTracingPipelineShaderGroupHandleCaptureReplay must also be VK_TRUE", 138}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingPipelineFeaturesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_FEATURES_KHR", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingPipelinePropertiesKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_PROPERTIES_KHR", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRayTracingPositionFetchFeaturesKHR-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_POSITION_FETCH_FEATURES_KHR", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRayTracingPropertiesNV-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PROPERTIES_NV", 73}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRayTracingValidationFeaturesNV-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_VALIDATION_FEATURES_NV", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRelaxedLineRasterizationFeaturesIMG-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RELAXED_LINE_RASTERIZATION_FEATURES_IMG", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRenderPassStripedFeaturesARM-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_FEATURES_ARM", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRenderPassStripedPropertiesARM-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_PROPERTIES_ARM", 82}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceRepresentativeFragmentTestFeaturesNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_REPRESENTATIVE_FRAGMENT_TEST_FEATURES_NV", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRobustness2FeaturesKHR-robustBufferAccess2-04000", 69}, { {"If robustBufferAccess2 is enabled then robustBufferAccess must also be enabled", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRobustness2FeaturesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_FEATURES_KHR", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceRobustness2PropertiesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_PROPERTIES_KHR", 75}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSampleLocationsPropertiesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLE_LOCATIONS_PROPERTIES_EXT", 79}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSamplerFilterMinmaxProperties-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_FILTER_MINMAX_PROPERTIES", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSamplerYcbcrConversionFeatures-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_YCBCR_CONVERSION_FEATURES", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceScalarBlockLayoutFeatures-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCALAR_BLOCK_LAYOUT_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSchedulingControlsDispatchParametersPropertiesARM-sType-sType", 82}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_DISPATCH_PARAMETERS_PROPERTIES_ARM", 102}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceSchedulingControlsFeaturesARM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_FEATURES_ARM", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSchedulingControlsPropertiesARM-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_PROPERTIES_ARM", 82}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceSeparateDepthStencilLayoutsFeatures-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SEPARATE_DEPTH_STENCIL_LAYOUTS_FEATURES", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShader64BitIndexingFeaturesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_64_BIT_INDEXING_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderAbortFeaturesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ABORT_FEATURES_KHR", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderAbortPropertiesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ABORT_PROPERTIES_KHR", 75}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderAtomicFloat16VectorFeaturesNV-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT16_VECTOR_FEATURES_NV", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderAtomicFloat2FeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_2_FEATURES_EXT", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderAtomicFloatFeaturesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_FEATURES_EXT", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderAtomicInt64Features-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_INT64_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderBfloat16FeaturesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_BFLOAT16_FEATURES_KHR", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderClockFeaturesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CLOCK_FEATURES_KHR", 73}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderConstantDataFeaturesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CONSTANT_DATA_FEATURES_KHR", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderCoreBuiltinsFeaturesARM-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_FEATURES_ARM", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderCoreBuiltinsPropertiesARM-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_PROPERTIES_ARM", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderCoreProperties2AMD-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_2_AMD", 76}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderCorePropertiesAMD-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_AMD", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderCorePropertiesARM-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_ARM", 74}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderDemoteToHelperInvocationFeatures-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DEMOTE_TO_HELPER_INVOCATION_FEATURES", 91}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderDrawParametersFeatures-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETERS_FEATURES", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderEarlyAndLateFragmentTestsFeaturesAMD-sType-sType", 75}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EARLY_AND_LATE_FRAGMENT_TESTS_FEATURES_AMD", 97}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderEnqueueFeaturesAMDX-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_FEATURES_AMDX", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderEnqueuePropertiesAMDX-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_PROPERTIES_AMDX", 78}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderExpectAssumeFeatures-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EXPECT_ASSUME_FEATURES", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderFloat16Int8Features-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT16_INT8_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderFloat8FeaturesEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT8_FEATURES_EXT", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderFloatControls2Features-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT_CONTROLS_2_FEATURES", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderFmaFeaturesKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FMA_FEATURES_KHR", 71}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderImageAtomicInt64FeaturesEXT-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_ATOMIC_INT64_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderImageFootprintFeaturesNV-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_FOOTPRINT_FEATURES_NV", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderInstrumentationFeaturesARM-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INSTRUMENTATION_FEATURES_ARM", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderInstrumentationPropertiesARM-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INSTRUMENTATION_PROPERTIES_ARM", 85}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderIntegerDotProductFeatures-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_FEATURES", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderIntegerDotProductProperties-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_PROPERTIES", 85}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceShaderIntegerFunctions2FeaturesINTEL-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_FUNCTIONS_2_FEATURES_INTEL", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderLongVectorFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_LONG_VECTOR_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderLongVectorPropertiesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_LONG_VECTOR_PROPERTIES_EXT", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderMaximalReconvergenceFeaturesKHR-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MAXIMAL_RECONVERGENCE_FEATURES_KHR", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderMixedFloatDotProductFeaturesVALVE-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MIXED_FLOAT_DOT_PRODUCT_FEATURES_VALVE", 93}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderModuleIdentifierFeaturesEXT-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_FEATURES_EXT", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderModuleIdentifierPropertiesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_PROPERTIES_EXT", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderMultipleWaitQueuesFeaturesQCOM-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MULTIPLE_WAIT_QUEUES_FEATURES_QCOM", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderMultipleWaitQueuesPropertiesQCOM-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MULTIPLE_WAIT_QUEUES_PROPERTIES_QCOM", 91}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderOCPMicroscalingTypesFeaturesEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OCP_MICROSCALING_TYPES_FEATURES_EXT", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderObjectFeaturesEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_FEATURES_EXT", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderObjectPropertiesEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_PROPERTIES_EXT", 76}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderQuadControlFeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_QUAD_CONTROL_FEATURES_KHR", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderRelaxedExtendedInstructionFeaturesKHR-sType-sType", 76}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_RELAXED_EXTENDED_INSTRUCTION_FEATURES_KHR", 96}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderReplicatedCompositesFeaturesEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_REPLICATED_COMPOSITES_FEATURES_EXT", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSMBuiltinsFeaturesNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSMBuiltinsPropertiesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderSplitBarrierFeaturesEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SPLIT_BARRIER_FEATURES_EXT", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSplitBarrierPropertiesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SPLIT_BARRIER_PROPERTIES_EXT", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceShaderSubgroupExtendedTypesFeatures-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_EXTENDED_TYPES_FEATURES", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSubgroupPartitionedFeaturesEXT-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_PARTITIONED_FEATURES_EXT", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSubgroupRotateFeatures-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_ROTATE_FEATURES", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderSubgroupUniformControlFlowFeaturesKHR-sType-sType", 76}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_UNIFORM_CONTROL_FLOW_FEATURES_KHR", 97}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderTerminateInvocationFeatures-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TERMINATE_INVOCATION_FEATURES", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderTileImageFeaturesEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_FEATURES_EXT", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderTileImagePropertiesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_PROPERTIES_EXT", 80}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceShaderUniformBufferUnsizedArrayFeaturesEXT-sType-sType", 75}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_UNIFORM_BUFFER_UNSIZED_ARRAY_FEATURES_EXT", 96}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShaderUntypedPointersFeaturesKHR-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_UNTYPED_POINTERS_FEATURES_KHR", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShadingRateImageFeaturesNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_FEATURES_NV", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceShadingRateImagePropertiesNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_PROPERTIES_NV", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-format-parameter", 60}, { {"format must be a valid VkFormat value", 37}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-pNext-pNext", 55}, { {"pNext must be NULL or a pointer to a valid instance of VkImageUsageFlags2CreateInfoKHR", 86}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SPARSE_IMAGE_FORMAT_INFO_2", 74}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-sType-unique", 56}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-samples-01095", 57}, { {"samples must be a valid VkSampleCountFlagBits value that is set in VkImageFormatProperties::sampleCounts returned by vkGetPhysicalDeviceImageFormatProperties with format, type, tiling, and usage equal to those in this command", 225}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-samples-parameter", 61}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-tiling-parameter", 60}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-type-parameter", 58}, { {"type must be a valid VkImageType value", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-usage-parameter", 59}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSparseImageFormatInfo2-usage-requiredbitmask", 65}, { {"usage must not be 0", 19}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkPhysicalDeviceSubgroupProperties-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES", 67}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSubgroupSizeControlFeatures-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_FEATURES", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSubgroupSizeControlProperties-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_PROPERTIES", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSubpassMergeFeedbackFeaturesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_MERGE_FEEDBACK_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSubpassShadingFeaturesHUAWEI-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_FEATURES_HUAWEI", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSubpassShadingPropertiesHUAWEI-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_PROPERTIES_HUAWEI", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceSurfaceInfo2KHR-pNext-02672", 48}, { {"If the pNext chain includes a VkSurfaceFullScreenExclusiveInfoEXT structure with its fullScreenExclusive member set to VK_FULL_SCREEN_EXCLUSIVE_APPLICATION_CONTROLLED_EXT, and surface was created using vkCreateWin32SurfaceKHR, a VkSurfaceFullScreenExclusiveWin32InfoEXT structure must be included in the pNext chain", 315}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPhysicalDeviceSurfaceInfo2KHR-pNext-pNext", 48}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkSurfaceFullScreenExclusiveInfoEXT, VkSurfaceFullScreenExclusiveWin32InfoEXT, or VkSurfacePresentModeKHR", 236}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPhysicalDeviceSurfaceInfo2KHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SURFACE_INFO_2_KHR", 66}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPhysicalDeviceSurfaceInfo2KHR-sType-unique", 49}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPhysicalDeviceSurfaceInfo2KHR-surface-07919", 50}, { {"If surface is not VK_NULL_HANDLE, and the VK_GOOGLE_surfaceless_query extension is not enabled, surface must be a valid VkSurfaceKHR handle", 139}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPhysicalDeviceSwapchainMaintenance1FeaturesKHR-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SWAPCHAIN_MAINTENANCE_1_FEATURES_KHR", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceSynchronization2Features-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SYNCHRONIZATION_2_FEATURES", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTensorFeaturesARM-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TENSOR_FEATURES_ARM", 67}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTensorPropertiesARM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TENSOR_PROPERTIES_ARM", 69}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceTexelBufferAlignmentFeaturesEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_FEATURES_EXT", 83}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTexelBufferAlignmentProperties-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_PROPERTIES", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceTextureCompressionASTC3DFeaturesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_3D_FEATURES_EXT", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTextureCompressionASTCHDRFeatures-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_HDR_FEATURES", 85}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceThrottleHintFeaturesSEC-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_THROTTLE_HINT_FEATURES_SEC", 74}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkPhysicalDeviceTileMemoryHeapFeaturesQCOM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_MEMORY_HEAP_FEATURES_QCOM", 78}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTileMemoryHeapPropertiesQCOM-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_MEMORY_HEAP_PROPERTIES_QCOM", 80}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceTilePropertiesFeaturesQCOM-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_PROPERTIES_FEATURES_QCOM", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTileShadingFeaturesQCOM-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_SHADING_FEATURES_QCOM", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTileShadingPropertiesQCOM-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_SHADING_PROPERTIES_QCOM", 76}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceTimelineSemaphoreFeatures-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTimelineSemaphoreProperties-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_PROPERTIES", 77}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceToolProperties-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkPhysicalDeviceToolProperties-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TOOL_PROPERTIES", 63}, { "chapters/debugging.html", 23} } },
        { {"VUID-VkPhysicalDeviceTransformFeedbackFeaturesEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceTransformFeedbackPropertiesEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_PROPERTIES_EXT", 81}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceUnifiedImageLayoutsFeaturesKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFIED_IMAGE_LAYOUTS_FEATURES_KHR", 82}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceUniformBufferStandardLayoutFeatures-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFORM_BUFFER_STANDARD_LAYOUT_FEATURES", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVariablePointersFeatures-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VARIABLE_POINTERS_FEATURES", 74}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVariablePointersFeatures-variablePointers-01431", 68}, { {"If variablePointers is enabled then variablePointersStorageBuffer must also be enabled", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVertexAttributeDivisorFeatures-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_FEATURES", 81}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVertexAttributeDivisorProperties-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES", 83}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceVertexAttributeDivisorPropertiesEXT-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_EXT", 87}, { "chapters/limits.html", 20} } },
        { {"VUID-VkPhysicalDeviceVertexAttributeRobustnessFeaturesEXT-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_ROBUSTNESS_FEATURES_EXT", 88}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVertexInputDynamicStateFeaturesEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_INPUT_DYNAMIC_STATE_FEATURES_EXT", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoDecodeVP9FeaturesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_DECODE_VP9_FEATURES_KHR", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeAV1FeaturesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_AV1_FEATURES_KHR", 77}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeFeedback2FeaturesKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_FEEDBACK_2_FEATURES_KHR", 84}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeIntraRefreshFeaturesKHR-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_INTRA_REFRESH_FEATURES_KHR", 87}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-pNext-pNext", 63}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-pVideoProfile-08259", 71}, { {"pVideoProfile must be a supported video profile", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-pVideoProfile-08260", 71}, { {"pVideoProfile->videoCodecOperation must specify an encode operation", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-pVideoProfile-parameter", 75}, { {"pVideoProfile must be a valid pointer to a valid VkVideoProfileInfoKHR structure", 80}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-qualityLevel-08261", 70}, { {"qualityLevel must be less than VkVideoEncodeCapabilitiesKHR::maxQualityLevels, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in pVideoProfile", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeQuantizationMapFeaturesKHR-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUANTIZATION_MAP_FEATURES_KHR", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoEncodeRgbConversionFeaturesVALVE-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_RGB_CONVERSION_FEATURES_VALVE", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoFormatInfoKHR-imageUsage-parameter", 60}, { {"imageUsage must be a valid combination of VkImageUsageFlagBits values", 69}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoFormatInfoKHR-imageUsage-requiredbitmask", 66}, { {"imageUsage must not be 0", 24}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoFormatInfoKHR-pNext-pNext", 51}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkImageUsageFlags2CreateInfoKHR or VkVideoProfileListInfoKHR", 191}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoFormatInfoKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_FORMAT_INFO_KHR", 69}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoFormatInfoKHR-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkPhysicalDeviceVideoMaintenance1FeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_1_FEATURES_KHR", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVideoMaintenance2FeaturesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_2_FEATURES_KHR", 80}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVulkan11Features-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES", 67}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVulkan11Properties-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_PROPERTIES", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceVulkan12Features-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES", 67}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVulkan12Properties-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_PROPERTIES", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceVulkan13Features-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES", 67}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVulkan13Properties-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_PROPERTIES", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceVulkan14Features-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_4_FEATURES", 67}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceVulkan14Properties-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_4_PROPERTIES", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkPhysicalDeviceVulkanMemoryModelFeatures-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_MEMORY_MODEL_FEATURES", 76}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceWorkgroupMemoryExplicitLayoutFeaturesKHR-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_WORKGROUP_MEMORY_EXPLICIT_LAYOUT_FEATURES_KHR", 93}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceYcbcr2Plane444FormatsFeaturesEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_2_PLANE_444_FORMATS_FEATURES_EXT", 86}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceYcbcrDegammaFeaturesQCOM-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_DEGAMMA_FEATURES_QCOM", 75}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceYcbcrImageArraysFeaturesEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_IMAGE_ARRAYS_FEATURES_EXT", 79}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceZeroInitializeDeviceMemoryFeaturesEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_DEVICE_MEMORY_FEATURES_EXT", 90}, { "chapters/features.html", 22} } },
        { {"VUID-VkPhysicalDeviceZeroInitializeWorkgroupMemoryFeatures-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_WORKGROUP_MEMORY_FEATURES", 89}, { "chapters/features.html", 22} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-device-09610", 47}, { {"If device was created with VkDevicePipelineBinaryInternalCacheControlKHR::disableInternalCache set to VK_TRUE, pPipelineCreateInfo must be NULL", 143}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pKeysAndDataInfo-09619", 57}, { {"One and only one of pKeysAndDataInfo, pipeline, or pPipelineCreateInfo must be non-NULL", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pKeysAndDataInfo-parameter", 61}, { {"If pKeysAndDataInfo is not NULL, pKeysAndDataInfo must be a valid pointer to a valid VkPipelineBinaryKeysAndDataKHR structure", 125}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pPipelineCreateInfo-09606", 60}, { {"If pPipelineCreateInfo is not NULL, the pNext chain of pPipelineCreateInfo must not set VkPipelineBinaryInfoKHR::binaryCount to a value greater than 0", 150}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pPipelineCreateInfo-parameter", 64}, { {"If pPipelineCreateInfo is not NULL, pPipelineCreateInfo must be a valid pointer to a valid VkPipelineCreateInfoKHR structure", 124}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pipeline-09607", 49}, { {"If pipeline is not VK_NULL_HANDLE, pipeline must have been created with VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pipeline-09608", 49}, { {"If pipeline is not VK_NULL_HANDLE, vkReleaseCapturedPipelineDataKHR must not have been called on pipeline prior to this command", 127}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pipeline-parameter", 53}, { {"If pipeline is not VK_NULL_HANDLE, pipeline must be a valid VkPipeline handle", 77}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-pipelineBinaryInternalCache-09609", 68}, { {"If pipelineBinaryInternalCache is VK_FALSE pPipelineCreateInfo must be NULL", 75}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryCreateInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_BINARY_CREATE_INFO_KHR", 63}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryDataInfoKHR-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryDataInfoKHR-pipelineBinary-parameter", 57}, { {"pipelineBinary must be a valid VkPipelineBinaryKHR handle", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryDataInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_BINARY_DATA_INFO_KHR", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryDataKHR-dataSize-arraylength", 49}, { {"dataSize must be greater than 0", 31}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryDataKHR-pData-parameter", 44}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryHandlesInfoKHR-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryHandlesInfoKHR-pPipelineBinaries-parameter", 63}, { {"If pipelineBinaryCount is not 0, and pPipelineBinaries is not NULL, pPipelineBinaries must be a valid pointer to an array of pipelineBinaryCount VkPipelineBinaryKHR handles", 172}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryHandlesInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_BINARY_HANDLES_INFO_KHR", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryInfoKHR-binaryCount-09603", 46}, { {"binaryCount and the order of the elements in pPipelineBinaries must exactly match that returned by vkCreatePipelineBinariesKHR for the matching Vk*PipelineCreateInfo structure and its pNext chain, ignoring the presence of the VkPipelineBinaryInfoKHR structure, the presence of the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag, and absence of any shader module identifiers, shader modules, or VkShaderModuleCreateInfo structures, for the same global pipeline key, from either: VkPipelineBinaryCreateInfoKHR::pPipelineCreateInfo, or VkPipelineBinaryCreateInfoKHR::pipeline", 572}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryInfoKHR-pPipelineBinaries-parameter", 56}, { {"If binaryCount is not 0, pPipelineBinaries must be a valid pointer to an array of binaryCount valid VkPipelineBinaryKHR handles", 127}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_BINARY_INFO_KHR", 56}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryKeyKHR-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryKeyKHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_BINARY_KEY_KHR", 55}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryKeysAndDataKHR-binaryCount-arraylength", 59}, { {"binaryCount must be greater than 0", 34}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryKeysAndDataKHR-pPipelineBinaryData-parameter", 65}, { {"pPipelineBinaryData must be a valid pointer to an array of binaryCount valid VkPipelineBinaryDataKHR structures", 111}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineBinaryKeysAndDataKHR-pPipelineBinaryKeys-parameter", 65}, { {"pPipelineBinaryKeys must be a valid pointer to an array of binaryCount valid VkPipelineBinaryKeyKHR structures", 110}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-flags-10201", 42}, { {"If flags includes VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, it must not include VK_PIPELINE_CACHE_CREATE_INTERNALLY_SYNCHRONIZED_MERGE_BIT_KHR", 154}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-flags-parameter", 46}, { {"flags must be a valid combination of VkPipelineCacheCreateFlagBits values", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-initialDataSize-00768", 52}, { {"If initialDataSize is not 0, it must be equal to the size of pInitialData, as returned by vkGetPipelineCacheData when pInitialData was originally retrieved", 155}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-initialDataSize-00769", 52}, { {"If initialDataSize is not 0, pInitialData must have been retrieved from a previous call to vkGetPipelineCacheData", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-maintenance8-10200", 49}, { {"If the maintenance8 feature is not enabled, flags must not include VK_PIPELINE_CACHE_CREATE_INTERNALLY_SYNCHRONIZED_MERGE_BIT_KHR", 129}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-pInitialData-parameter", 53}, { {"If initialDataSize is not 0, pInitialData must be a valid pointer to an array of initialDataSize bytes", 102}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-pipelineCreationCacheControl-02892", 65}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheCreateInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_CACHE_CREATE_INFO", 58}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-None-11835", 57}, { {"The dataGraphModel feature must be enabled", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-cacheType-parameter", 66}, { {"cacheType must be a valid VkDataGraphModelCacheTypeQCOM value", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-headerSize-11836", 63}, { {"headerSize must be 28", 21}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-headerSize-11838", 63}, { {"headerSize must not exceed the size of the pipeline cache", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-headerVersion-11837", 66}, { {"headerVersion must be VK_PIPELINE_CACHE_HEADER_VERSION_DATA_GRAPH_QCOM", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionDataGraphQCOM-headerVersion-parameter", 70}, { {"headerVersion must be a valid VkPipelineCacheHeaderVersion value", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionOne-headerSize-04967", 53}, { {"headerSize must be 32", 21}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionOne-headerSize-08990", 53}, { {"headerSize must not exceed the size of the pipeline cache", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionOne-headerVersion-04968", 56}, { {"headerVersion must be VK_PIPELINE_CACHE_HEADER_VERSION_ONE", 58}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCacheHeaderVersionOne-headerVersion-parameter", 60}, { {"headerVersion must be a valid VkPipelineCacheHeaderVersion value", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineColorBlendAdvancedStateCreateInfoEXT-blendOverlap-01426", 70}, { {"If the correlated overlap property is not supported, blendOverlap must be VK_BLEND_OVERLAP_UNCORRELATED_EXT", 107}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAdvancedStateCreateInfoEXT-blendOverlap-parameter", 74}, { {"blendOverlap must be a valid VkBlendOverlapEXT value", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAdvancedStateCreateInfoEXT-dstPremultiplied-01425", 74}, { {"If the non-premultiplied destination color property is not supported, dstPremultiplied must be VK_TRUE", 102}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAdvancedStateCreateInfoEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_ADVANCED_STATE_CREATE_INFO_EXT", 83}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAdvancedStateCreateInfoEXT-srcPremultiplied-01424", 74}, { {"If the non-premultiplied source color property is not supported, srcPremultiplied must be VK_TRUE", 97}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-advancedBlendAllOperations-01409", 73}, { {"If VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT::advancedBlendAllOperations is VK_FALSE, then colorBlendOp must not be VK_BLEND_OP_ZERO_EXT, VK_BLEND_OP_SRC_EXT, VK_BLEND_OP_DST_EXT, VK_BLEND_OP_SRC_OVER_EXT, VK_BLEND_OP_DST_OVER_EXT, VK_BLEND_OP_SRC_IN_EXT, VK_BLEND_OP_DST_IN_EXT, VK_BLEND_OP_SRC_OUT_EXT, VK_BLEND_OP_DST_OUT_EXT, VK_BLEND_OP_SRC_ATOP_EXT, VK_BLEND_OP_DST_ATOP_EXT, VK_BLEND_OP_XOR_EXT, VK_BLEND_OP_INVERT_EXT, VK_BLEND_OP_INVERT_RGB_EXT, VK_BLEND_OP_LINEARDODGE_EXT, VK_BLEND_OP_LINEARBURN_EXT, VK_BLEND_OP_VIVIDLIGHT_EXT, VK_BLEND_OP_LINEARLIGHT_EXT, VK_BLEND_OP_PINLIGHT_EXT, VK_BLEND_OP_HARDMIX_EXT, VK_BLEND_OP_PLUS_EXT, VK_BLEND_OP_PLUS_CLAMPED_EXT, VK_BLEND_OP_PLUS_CLAMPED_ALPHA_EXT, VK_BLEND_OP_PLUS_DARKER_EXT, VK_BLEND_OP_MINUS_EXT, VK_BLEND_OP_MINUS_CLAMPED_EXT, VK_BLEND_OP_CONTRAST_EXT, VK_BLEND_OP_INVERT_OVG_EXT, VK_BLEND_OP_RED_EXT, VK_BLEND_OP_GREEN_EXT, or VK_BLEND_OP_BLUE_EXT", 922}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-advancedBlendIndependentBlend-01407", 76}, { {"If VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT::advancedBlendIndependentBlend is VK_FALSE and colorBlendOp is an advanced blend operation, then colorBlendOp must be the same for all attachments", 201}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-advancedBlendIndependentBlend-01408", 76}, { {"If VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT::advancedBlendIndependentBlend is VK_FALSE and alphaBlendOp is an advanced blend operation, then alphaBlendOp must be the same for all attachments", 201}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-alphaBlendOp-parameter", 63}, { {"alphaBlendOp must be a valid VkBlendOp value", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-colorBlendOp-01406", 59}, { {"If either of colorBlendOp or alphaBlendOp is an advanced blend operation, then colorBlendOp must equal alphaBlendOp", 115}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-colorBlendOp-01410", 59}, { {"If colorBlendOp or alphaBlendOp is an advanced blend operation, then colorAttachmentCount of the subpass this pipeline is compiled against must be less than or equal to VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT::advancedBlendMaxColorAttachments", 254}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-colorBlendOp-parameter", 63}, { {"colorBlendOp must be a valid VkBlendOp value", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-colorWriteMask-parameter", 65}, { {"colorWriteMask must be a valid combination of VkColorComponentFlagBits values", 77}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-constantAlphaColorBlendFactors-04454", 77}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::constantAlphaColorBlendFactors is VK_FALSE, srcColorBlendFactor must not be VK_BLEND_FACTOR_CONSTANT_ALPHA or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA", 255}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-constantAlphaColorBlendFactors-04455", 77}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::constantAlphaColorBlendFactors is VK_FALSE, dstColorBlendFactor must not be VK_BLEND_FACTOR_CONSTANT_ALPHA or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA", 255}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-dstAlphaBlendFactor-00611", 66}, { {"If the dualSrcBlend feature is not enabled, dstAlphaBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-dstAlphaBlendFactor-parameter", 70}, { {"dstAlphaBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-dstColorBlendFactor-00609", 66}, { {"If the dualSrcBlend feature is not enabled, dstColorBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-dstColorBlendFactor-parameter", 70}, { {"dstColorBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-srcAlphaBlendFactor-00610", 66}, { {"If the dualSrcBlend feature is not enabled, srcAlphaBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-srcAlphaBlendFactor-parameter", 70}, { {"srcAlphaBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-srcColorBlendFactor-00608", 66}, { {"If the dualSrcBlend feature is not enabled, srcColorBlendFactor must not be VK_BLEND_FACTOR_SRC1_COLOR, VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR, VK_BLEND_FACTOR_SRC1_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA", 209}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendAttachmentState-srcColorBlendFactor-parameter", 70}, { {"srcColorBlendFactor must be a valid VkBlendFactor value", 55}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-flags-parameter", 56}, { {"flags must be a valid combination of VkPipelineColorBlendStateCreateFlagBits values", 83}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-logicOpEnable-00606", 60}, { {"If the logicOp feature is not enabled, logicOpEnable must be VK_FALSE", 69}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-logicOpEnable-00607", 60}, { {"If logicOpEnable is VK_TRUE, logicOp must be a valid VkLogicOp value", 68}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-pAttachments-00605", 59}, { {"If the independentBlend feature is not enabled, all elements of pAttachments must be identical", 94}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-pAttachments-07353", 59}, { {"If attachmentCount is not 0 , and any of VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT, or VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT are not set, pAttachments must be a valid pointer to an array of attachmentCount valid VkPipelineColorBlendAttachmentState structures", 342}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-pAttachments-parameter", 63}, { {"If attachmentCount is not 0, and pAttachments is not NULL, pAttachments must be a valid pointer to an array of attachmentCount valid VkPipelineColorBlendAttachmentState structures", 179}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-pNext-pNext", 52}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineColorBlendAdvancedStateCreateInfoEXT or VkPipelineColorWriteCreateInfoEXT", 214}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-rasterizationOrderColorAttachmentAccess-06465", 86}, { {"If the rasterizationOrderColorAttachmentAccess feature is not enabled, flags must not include VK_PIPELINE_COLOR_BLEND_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_BIT_EXT", 176}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO", 70}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorBlendStateCreateInfo-sType-unique", 53}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorWriteCreateInfoEXT-attachmentCount-06655", 60}, { {"attachmentCount must be less than or equal to the maxColorAttachments member of VkPhysicalDeviceLimits", 102}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorWriteCreateInfoEXT-attachmentCount-07608", 60}, { {"If the pipeline is being created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT, VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT, or VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states not set, attachmentCount must be equal to the attachmentCount member of the VkPipelineColorBlendStateCreateInfo structure specified during pipeline creation", 377}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorWriteCreateInfoEXT-pAttachments-04801", 57}, { {"If the colorWriteEnable feature is not enabled, all elements of pColorWriteEnables must be VK_TRUE", 98}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorWriteCreateInfoEXT-pColorWriteEnables-parameter", 67}, { {"If attachmentCount is not 0, pColorWriteEnables must be a valid pointer to an array of attachmentCount VkBool32 values", 118}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineColorWriteCreateInfoEXT-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COLOR_WRITE_CREATE_INFO_EXT", 68}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-VkPipelineCompilerControlCreateInfoAMD-compilerControlFlags-zerobitmask", 76}, { {"compilerControlFlags must be 0", 30}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCompilerControlCreateInfoAMD-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COMPILER_CONTROL_CREATE_INFO_AMD", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCoverageModulationStateCreateInfoNV-coverageModulationMode-parameter", 83}, { {"coverageModulationMode must be a valid VkCoverageModulationModeNV value", 71}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageModulationStateCreateInfoNV-coverageModulationTableEnable-01405", 86}, { {"If coverageModulationTableEnable is VK_TRUE, coverageModulationTableCount must be equal to the number of rasterization samples divided by the number of color samples in the subpass", 180}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageModulationStateCreateInfoNV-flags-zerobitmask", 68}, { {"flags must be 0", 15}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageModulationStateCreateInfoNV-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_MODULATION_STATE_CREATE_INFO_NV", 81}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageReductionStateCreateInfoNV-coverageReductionMode-parameter", 81}, { {"coverageReductionMode must be a valid VkCoverageReductionModeNV value", 69}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageReductionStateCreateInfoNV-flags-zerobitmask", 67}, { {"flags must be 0", 15}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageReductionStateCreateInfoNV-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_REDUCTION_STATE_CREATE_INFO_NV", 80}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageToColorStateCreateInfoNV-coverageToColorEnable-01404", 75}, { {"If coverageToColorEnable is VK_TRUE, then the render pass subpass indicated by VkGraphicsPipelineCreateInfo::renderPass and VkGraphicsPipelineCreateInfo::subpass must have a color attachment at the location selected by coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 382}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageToColorStateCreateInfoNV-flags-zerobitmask", 65}, { {"flags must be 0", 15}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCoverageToColorStateCreateInfoNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_TO_COLOR_STATE_CREATE_INFO_NV", 79}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineCreateFlags2CreateInfo-flags-parameter", 53}, { {"flags must be a valid combination of VkPipelineCreateFlagBits2 values", 69}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreateFlags2CreateInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_CREATE_FLAGS_2_CREATE_INFO", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreateInfoKHR-pNext-09604", 40}, { {"pNext must be pointer to a valid instance of VkGraphicsPipelineCreateInfo, VkExecutionGraphPipelineCreateInfoAMDX, VkRayTracingPipelineCreateInfoKHR, or VkComputePipelineCreateInfo", 180}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreateInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_CREATE_INFO_KHR", 56}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreationFeedbackCreateInfo-pPipelineCreationFeedback-parameter", 77}, { {"pPipelineCreationFeedback must be a valid pointer to a VkPipelineCreationFeedback structure", 91}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreationFeedbackCreateInfo-pPipelineStageCreationFeedbacks-parameter", 83}, { {"If pipelineStageCreationFeedbackCount is not 0, pPipelineStageCreationFeedbacks must be a valid pointer to an array of pipelineStageCreationFeedbackCount VkPipelineCreationFeedback structures", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineCreationFeedbackCreateInfo-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_CREATION_FEEDBACK_CREATE_INFO", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-back-parameter", 57}, { {"back must be a valid VkStencilOpState structure", 47}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-depthBoundsTestEnable-00598", 70}, { {"If the depthBounds feature is not enabled, depthBoundsTestEnable must be VK_FALSE", 81}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-depthCompareOp-parameter", 67}, { {"depthCompareOp must be a valid VkCompareOp value", 48}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-flags-parameter", 58}, { {"flags must be a valid combination of VkPipelineDepthStencilStateCreateFlagBits values", 85}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-front-parameter", 58}, { {"front must be a valid VkStencilOpState structure", 48}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-rasterizationOrderDepthAttachmentAccess-06463", 88}, { {"If the rasterizationOrderDepthAttachmentAccess feature is not enabled, flags must not include VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT", 184}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-rasterizationOrderStencilAttachmentAccess-06464", 90}, { {"If the rasterizationOrderStencilAttachmentAccess feature is not enabled, flags must not include VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT", 188}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO", 72}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDepthStencilStateCreateInfo-separateStencilMaskRef-04453", 71}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::separateStencilMaskRef is VK_FALSE, and the value of VkPipelineDepthStencilStateCreateInfo::stencilTestEnable is VK_TRUE, and the value of VkPipelineRasterizationStateCreateInfo::cullMode is VK_CULL_MODE_NONE, the value of reference in each of the VkStencilOpState structs in front and back must be the same", 412}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDiscardRectangleStateCreateInfoEXT-discardRectangleCount-00582", 77}, { {"discardRectangleCount must be less than or equal to VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 119}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDiscardRectangleStateCreateInfoEXT-discardRectangleMode-parameter", 80}, { {"discardRectangleMode must be a valid VkDiscardRectangleModeEXT value", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDiscardRectangleStateCreateInfoEXT-flags-zerobitmask", 67}, { {"flags must be 0", 15}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDiscardRectangleStateCreateInfoEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_DISCARD_RECTANGLE_STATE_CREATE_INFO_EXT", 80}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineDynamicStateCreateInfo-flags-zerobitmask", 55}, { {"flags must be 0", 15}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineDynamicStateCreateInfo-pDynamicStates-01442", 58}, { {"Each element of pDynamicStates must be unique", 45}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineDynamicStateCreateInfo-pDynamicStates-parameter", 62}, { {"If dynamicStateCount is not 0, pDynamicStates must be a valid pointer to an array of dynamicStateCount valid VkDynamicState values", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineDynamicStateCreateInfo-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineDynamicStateCreateInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInfoKHR-executableIndex-03275", 54}, { {"executableIndex must be less than the number of pipeline executables associated with pipeline as returned in the pExecutableCount parameter of vkGetPipelineExecutablePropertiesKHR", 179}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInfoKHR-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInfoKHR-pipeline-parameter", 51}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INFO_KHR", 60}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInternalRepresentationKHR-pNext-pNext", 62}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableInternalRepresentationKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INTERNAL_REPRESENTATION_KHR", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutablePropertiesKHR-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutablePropertiesKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_PROPERTIES_KHR", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableStatisticKHR-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineExecutableStatisticKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_STATISTIC_KHR", 65}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineFragmentDensityMapLayeredCreateInfoVALVE-maxFragmentDensityMapLayers-10825", 89}, { {"maxFragmentDensityMapLayers must be less than or equal to maxFragmentDensityMapLayers", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineFragmentDensityMapLayeredCreateInfoVALVE-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_DENSITY_MAP_LAYERED_CREATE_INFO_VALVE", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineFragmentShadingRateEnumStateCreateInfoNV-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_ENUM_STATE_CREATE_INFO_NV", 88}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineFragmentShadingRateStateCreateInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_STATE_CREATE_INFO_KHR", 84}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pipeline-09080", 57}, { {"pipeline must have been created with flag VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV set", 89}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pipeline-09081", 57}, { {"pipeline must have been created with a VkComputePipelineIndirectBufferInfoNV structure specifying a valid address where its metadata will be saved", 146}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pipeline-parameter", 61}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pipelineBindPoint-09079", 66}, { {"The provided pipelineBindPoint must be of type VK_PIPELINE_BIND_POINT_COMPUTE", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-pipelineBindPoint-parameter", 70}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineIndirectDeviceAddressInfoNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_INDIRECT_DEVICE_ADDRESS_INFO_NV", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkPipelineInfoKHR-pNext-pNext", 34}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineInfoKHR-pipeline-parameter", 41}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineInfoKHR-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_INFO_KHR", 49}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-flags-zerobitmask", 61}, { {"flags must be 0", 15}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-topology-00429", 58}, { {"If the geometryShader feature is not enabled, topology must not be any of VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, VK_PRIMITIVE_TOPOLOGY_LINE_STRIP_WITH_ADJACENCY, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP_WITH_ADJACENCY", 276}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-topology-00430", 58}, { {"If the tessellationShader feature is not enabled, topology must not be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST", 103}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-topology-06252", 58}, { {"If the primitiveTopologyListRestart feature is not enabled, and topology is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, primitiveRestartEnable must be VK_FALSE", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-topology-06253", 58}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, and topology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, primitiveRestartEnable must be VK_FALSE", 154}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-topology-parameter", 62}, { {"topology must be a valid VkPrimitiveTopology value", 50}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineInputAssemblyStateCreateInfo-triangleFans-04452", 62}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::triangleFans is VK_FALSE, topology must not be VK_PRIMITIVE_TOPOLOGY_TRIANGLE_FAN", 186}, { "chapters/drawing.html", 21} } },
        { {"VUID-VkPipelineLayoutCreateInfo-None-10005", 42}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetTotalBuffersDynamic", 440}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-02214", 52}, { {"The total number of bindings in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set and with a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK accessible to any given shader stage across all elements of pSetLayouts, must be less than or equal to VkPhysicalDeviceInlineUniformBlockProperties::maxPerStageDescriptorInlineUniformBlocks", 400}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-02215", 52}, { {"If the descriptorBindingInlineUniformBlockUpdateAfterBind feature is supported on the device, the total number of bindings with a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceInlineUniformBlockProperties::maxPerStageDescriptorUpdateAfterBindInlineUniformBlocks", 391}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-02216", 52}, { {"The total number of bindings in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceInlineUniformBlockProperties::maxDescriptorSetInlineUniformBlocks", 393}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-02217", 52}, { {"If the descriptorBindingInlineUniformBlockUpdateAfterBind feature is supported on the device, the total number of bindings with a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceInlineUniformBlockProperties::maxDescriptorSetUpdateAfterBindInlineUniformBlocks", 389}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-02381", 52}, { {"The total number of bindings with a descriptorType of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxDescriptorSetAccelerationStructures", 282}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03016", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_SAMPLER and VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorSamplers", 398}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03017", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER and VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorUniformBuffers", 411}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03018", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_BUFFER and VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorStorageBuffers", 411}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03020", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorStorageImages", 408}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03021", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorInputAttachments", 369}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03022", 52}, { {"If the descriptorBindingSampledImageUpdateAfterBind feature is supported on the device, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_SAMPLER and VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindSamplers", 410}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03023", 52}, { {"If the descriptorBindingUniformBufferUpdateAfterBind feature is supported on the device, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER and VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindUniformBuffers", 424}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03024", 52}, { {"If the descriptorBindingStorageBufferUpdateAfterBind feature is supported on the device, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_BUFFER and VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindStorageBuffers", 424}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03025", 52}, { {"If the descriptorBindingSampledImageUpdateAfterBind feature is supported on the device, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindSampledImages", 463}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03026", 52}, { {"If the descriptorBindingStorageImageUpdateAfterBind feature is supported on the device, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindStorageImages", 420}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03027", 52}, { {"If any element of pSetLayouts is created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set, the total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxPerStageDescriptorUpdateAfterBindInputAttachments", 411}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03028", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_SAMPLER and VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetSamplers", 396}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03029", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetUniformBuffers", 363}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03030", 52}, { {"If the maintenance7 feature is not enabled, the total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetUniformBuffersDynamic", 422}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03031", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetStorageBuffers", 363}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03032", 52}, { {"If the maintenance7 feature is not enabled, the total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetStorageBuffersDynamic", 422}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03033", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetSampledImages", 449}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03034", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetStorageImages", 406}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03035", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetInputAttachments", 367}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03571", 52}, { {"The total number of bindings in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxPerStageDescriptorAccelerationStructures", 410}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03572", 52}, { {"The total number of bindings with a descriptorType of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxPerStageDescriptorUpdateAfterBindAccelerationStructures", 312}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03573", 52}, { {"The total number of bindings in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxDescriptorSetAccelerationStructures", 408}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-03574", 52}, { {"The total number of bindings with a descriptorType of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxDescriptorSetUpdateAfterBindAccelerationStructures", 310}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-06531", 52}, { {"The total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceVulkan13Properties::maxInlineUniformTotalSize", 263}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-06939", 52}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, and VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxPerStageDescriptorSampledImages", 540}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-descriptorType-09879", 52}, { {"The total number of descriptors with a descriptorType of VK_DESCRIPTOR_TYPE_TENSOR_ARM accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxPerStageDescriptorUpdateAfterBindStorageTensors", 276}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-flags-13352", 43}, { {"If flags has VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set, then VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT must also be set", 139}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-flags-parameter", 47}, { {"flags must be a valid combination of VkPipelineLayoutCreateFlagBits values", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-graphicsPipelineLibrary-06753", 61}, { {"If the graphicsPipelineLibrary feature is not enabled, elements of pSetLayouts must be valid VkDescriptorSetLayout objects", 122}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-maintenance7-10003", 50}, { {"If the maintenance7 feature is enabled, the total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetTotalUniformBuffersDynamic", 442}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-maintenance7-10004", 50}, { {"If the maintenance7 feature is enabled, the total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetTotalStorageBuffersDynamic", 442}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-maintenance7-10007", 50}, { {"If the descriptorBindingUniformBufferUpdateAfterBind feature is supported on the device, and the maintenance7 feature is enabled, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetUpdateAfterBindTotalUniformBuffersDynamic", 421}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-maintenance7-10008", 50}, { {"If the descriptorBindingStorageBufferUpdateAfterBind feature is supported on the device, and if the maintenance7 feature is enabled, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetUpdateAfterBindTotalStorageBuffersDynamic", 424}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pImmutableSamplers-03566", 56}, { {"The total number of pImmutableSamplers created with flags containing VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT or VK_SAMPLER_CREATE_SUBSAMPLED_COARSE_RECONSTRUCTION_BIT_EXT across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceFragmentDensityMap2PropertiesEXT::maxDescriptorSetSubsampledSamplers", 346}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pPushConstantRanges-00292", 57}, { {"Any two elements of pPushConstantRanges must not include the same stage in stageFlags", 85}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pPushConstantRanges-parameter", 61}, { {"If pushConstantRangeCount is not 0, pPushConstantRanges must be a valid pointer to an array of pushConstantRangeCount valid VkPushConstantRange structures", 154}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-00293", 49}, { {"pSetLayouts must not contain more than one descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT set", 142}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03036", 49}, { {"If the descriptorBindingSampledImageUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_SAMPLER and VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindSamplers", 395}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03037", 49}, { {"If the descriptorBindingUniformBufferUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindUniformBuffers", 363}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03038", 49}, { {"If the descriptorBindingUniformBufferUpdateAfterBind feature is supported on the device, and if the maintenance7 feature is not enabled, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetUpdateAfterBindUniformBuffersDynamic", 404}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03039", 49}, { {"If the descriptorBindingStorageBufferUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindStorageBuffers", 363}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03040", 49}, { {"If the descriptorBindingStorageBufferUpdateAfterBind feature is supported on the device, and if the maintenance7 feature is not enabled, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceLimits::maxDescriptorSetUpdateAfterBindStorageBuffersDynamic", 404}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03041", 49}, { {"If the descriptorBindingSampledImageUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, and VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindSampledImages", 448}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03042", 49}, { {"If the descriptorBindingStorageImageUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindStorageImages", 405}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-03043", 49}, { {"If any element of pSetLayouts is created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceDescriptorIndexingProperties::maxDescriptorSetUpdateAfterBindInputAttachments", 396}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-04606", 49}, { {"Any element of pSetLayouts must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT bit set", 125}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-08008", 49}, { {"If any element of pSetLayouts was created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT bit set, all elements of pSetLayouts must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT bit set", 243}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-09698", 49}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_TENSOR_ARM accessible to any given shader stage across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxPerStageDescriptorSetStorageTensors", 377}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-09699", 49}, { {"The total number of descriptors in descriptor set layouts created without the VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT bit set with a descriptorType of VK_DESCRIPTOR_TYPE_TENSOR_ARM accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxDescriptorSetStorageTensors", 372}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-09878", 49}, { {"The total number of descriptors of the type VK_DESCRIPTOR_TYPE_TENSOR_ARM accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxDescriptorSetUpdateAfterBindStorageTensors", 261}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-10006", 49}, { {"If either the descriptorBindingStorageBufferUpdateAfterBind or descriptorBindingUniformBufferUpdateAfterBind feature is supported on the device, the total number of descriptors of the type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC accessible across all shader stages and across all elements of pSetLayouts must be less than or equal to VkPhysicalDeviceMaintenance7PropertiesKHR::maxDescriptorSetUpdateAfterBindTotalBuffersDynamic", 474}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-pSetLayouts-parameter", 53}, { {"If setLayoutCount is not 0, pSetLayouts must be a valid pointer to an array of setLayoutCount valid or VK_NULL_HANDLE VkDescriptorSetLayout handles", 147}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO", 59}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLayoutCreateInfo-setLayoutCount-00286", 52}, { {"setLayoutCount must be less than or equal to VkPhysicalDeviceLimits::maxBoundDescriptorSets", 91}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPipelineLibraryCreateInfoKHR-pLibraries-03381", 52}, { {"Each element of pLibraries must have been created with VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineLibraryCreateInfoKHR-pLibraries-06855", 52}, { {"If any library in pLibraries was created with a shader stage with VkPipelineShaderStageModuleIdentifierCreateInfoEXT and identifierSize not equal to 0, the pipeline must be created with the VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT flag set", 255}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineLibraryCreateInfoKHR-pLibraries-parameter", 56}, { {"If libraryCount is not 0, pLibraries must be a valid pointer to an array of libraryCount valid VkPipeline handles", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineLibraryCreateInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_LIBRARY_CREATE_INFO_KHR", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-alphaToOneEnable-00785", 64}, { {"If the alphaToOne feature is not enabled, alphaToOneEnable must be VK_FALSE", 75}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-flags-zerobitmask", 59}, { {"flags must be 0", 15}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-minSampleShading-00786", 64}, { {"minSampleShading must be in the range [0,1]", 43}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-pNext-pNext", 53}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineCoverageModulationStateCreateInfoNV, VkPipelineCoverageReductionStateCreateInfoNV, VkPipelineCoverageToColorStateCreateInfoNV, or VkPipelineSampleLocationsStateCreateInfoEXT", 314}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-rasterizationSamples-01415", 68}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the pNext chain does not contain VkPipelineCoverageReductionStateCreateInfoNV, or VkPipelineCoverageReductionStateCreateInfoNV::coverageReductionMode is not set to VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the subpass has any color attachments, and rasterizationSamples is greater than the number of color samples, then sample shading must not be enabled", 474}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-rasterizationSamples-parameter", 72}, { {"rasterizationSamples must be a valid VkSampleCountFlagBits value", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO", 70}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-sType-unique", 54}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineMultisampleStateCreateInfo-sampleShadingEnable-00784", 67}, { {"If the sampleRateShading feature is not enabled, sampleShadingEnable must be VK_FALSE", 85}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelinePropertiesIdentifierEXT-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelinePropertiesIdentifierEXT-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_PROPERTIES_IDENTIFIER_EXT", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRasterizationConservativeStateCreateInfoEXT-conservativeRasterizationMode-parameter", 98}, { {"conservativeRasterizationMode must be a valid VkConservativeRasterizationModeEXT value", 86}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationConservativeStateCreateInfoEXT-extraPrimitiveOverestimationSize-01769", 97}, { {"extraPrimitiveOverestimationSize must be in the range of 0.0 to VkPhysicalDeviceConservativeRasterizationPropertiesEXT::maxExtraPrimitiveOverestimationSize inclusive", 165}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationConservativeStateCreateInfoEXT-flags-zerobitmask", 76}, { {"flags must be 0", 15}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationConservativeStateCreateInfoEXT-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_CONSERVATIVE_STATE_CREATE_INFO_EXT", 89}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationDepthClipStateCreateInfoEXT-flags-zerobitmask", 73}, { {"flags must be 0", 15}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationDepthClipStateCreateInfoEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_DEPTH_CLIP_STATE_CREATE_INFO_EXT", 87}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-lineRasterizationMode-02768", 75}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the rectangularLines feature must be enabled", 117}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-lineRasterizationMode-02769", 75}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the bresenhamLines feature must be enabled", 113}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-lineRasterizationMode-02770", 75}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the smoothLines feature must be enabled", 119}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-lineRasterizationMode-parameter", 79}, { {"lineRasterizationMode must be a valid VkLineRasterizationMode value", 67}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_LINE_STATE_CREATE_INFO", 77}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-stippledLineEnable-02771", 72}, { {"If stippledLineEnable is VK_TRUE and lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 159}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-stippledLineEnable-02772", 72}, { {"If stippledLineEnable is VK_TRUE and lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 155}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-stippledLineEnable-02773", 72}, { {"If stippledLineEnable is VK_TRUE and lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 161}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationLineStateCreateInfo-stippledLineEnable-02774", 72}, { {"If stippledLineEnable is VK_TRUE and lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 211}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationProvokingVertexStateCreateInfoEXT-provokingVertexMode-04883", 87}, { {"If provokingVertexMode is VK_PROVOKING_VERTEX_MODE_LAST_VERTEX_EXT, then the provokingVertexLast feature must be enabled", 120}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineRasterizationProvokingVertexStateCreateInfoEXT-provokingVertexMode-parameter", 91}, { {"provokingVertexMode must be a valid VkProvokingVertexModeEXT value", 66}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineRasterizationProvokingVertexStateCreateInfoEXT-sType-sType", 73}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_PROVOKING_VERTEX_STATE_CREATE_INFO_EXT", 93}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-cullMode-parameter", 62}, { {"cullMode must be a valid combination of VkCullModeFlagBits values", 65}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-depthClampEnable-00782", 66}, { {"If the depthClamp feature is not enabled, depthClampEnable must be VK_FALSE", 75}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-flags-zerobitmask", 61}, { {"flags must be 0", 15}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-frontFace-parameter", 63}, { {"frontFace must be a valid VkFrontFace value", 43}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-pNext-pNext", 55}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDepthBiasRepresentationInfoEXT, VkPipelineRasterizationConservativeStateCreateInfoEXT, VkPipelineRasterizationDepthClipStateCreateInfoEXT, VkPipelineRasterizationLineStateCreateInfo, VkPipelineRasterizationProvokingVertexStateCreateInfoEXT, VkPipelineRasterizationStateRasterizationOrderAMD, or VkPipelineRasterizationStateStreamCreateInfoEXT", 475}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-pointPolygons-04458", 63}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::pointPolygons is VK_FALSE, and rasterizerDiscardEnable is VK_FALSE, polygonMode must not be VK_POLYGON_MODE_POINT", 218}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-polygonMode-01414", 61}, { {"If the VK_NV_fill_rectangle extension is not enabled, polygonMode must not be VK_POLYGON_MODE_FILL_RECTANGLE_NV", 111}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-polygonMode-01507", 61}, { {"If the fillModeNonSolid feature is not enabled, polygonMode must be VK_POLYGON_MODE_FILL or VK_POLYGON_MODE_FILL_RECTANGLE_NV", 125}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-polygonMode-parameter", 65}, { {"polygonMode must be a valid VkPolygonMode value", 47}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO", 72}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateCreateInfo-sType-unique", 56}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateRasterizationOrderAMD-rasterizationOrder-parameter", 83}, { {"rasterizationOrder must be a valid VkRasterizationOrderAMD value", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateRasterizationOrderAMD-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_RASTERIZATION_ORDER_AMD", 84}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateStreamCreateInfoEXT-flags-zerobitmask", 70}, { {"flags must be 0", 15}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateStreamCreateInfoEXT-geometryStreams-02324", 74}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::geometryStreams must be enabled", 77}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateStreamCreateInfoEXT-rasterizationStream-02325", 78}, { {"rasterizationStream must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 113}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateStreamCreateInfoEXT-rasterizationStream-02326", 78}, { {"rasterizationStream must be zero if VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackRasterizationStreamSelect is VK_FALSE", 138}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRasterizationStateStreamCreateInfoEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_STREAM_CREATE_INFO_EXT", 83}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineRenderingCreateInfo-colorAttachmentCount-09533", 61}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRenderingCreateInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_RENDERING_CREATE_INFO", 62}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRepresentativeFragmentTestStateCreateInfoNV-sType-sType", 70}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_REPRESENTATIVE_FRAGMENT_TEST_STATE_CREATE_INFO_NV", 90}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-images-10638", 48}, { {"If images is VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS or VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS_2, and either the descriptorBindingStorageImageUpdateAfterBind feature or the descriptorBindingSampledImageUpdateAfterBind feature is enabled on the device, robustBufferAccessUpdateAfterBind must be VK_TRUE", 338}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-images-parameter", 52}, { {"images must be a valid VkPipelineRobustnessImageBehavior value", 62}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-pipelineRobustness-06926", 60}, { {"If the pipelineRobustness feature is not enabled, storageBuffers must be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DEVICE_DEFAULT", 126}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-pipelineRobustness-06927", 60}, { {"If the pipelineRobustness feature is not enabled, uniformBuffers must be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DEVICE_DEFAULT", 126}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-pipelineRobustness-06928", 60}, { {"If the pipelineRobustness feature is not enabled, vertexInputs must be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DEVICE_DEFAULT", 124}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-pipelineRobustness-06929", 60}, { {"If the pipelineRobustness feature is not enabled, images must be VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_DEVICE_DEFAULT", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-robustBufferAccess2-06931", 61}, { {"If the robustBufferAccess2 feature is not supported, storageBuffers must not be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2", 141}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-robustBufferAccess2-06932", 61}, { {"If the robustBufferAccess2 feature is not supported, uniformBuffers must not be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2", 141}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-robustBufferAccess2-06933", 61}, { {"If the robustBufferAccess2 feature is not supported, vertexInputs must not be VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2", 139}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-robustImageAccess-06930", 59}, { {"If the robustImageAccess feature is not supported, images must not be VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS", 127}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-robustImageAccess2-06934", 60}, { {"If the robustImageAccess2 feature is not supported, images must not be VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS_2", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_ROBUSTNESS_CREATE_INFO", 63}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-storageBuffers-10636", 56}, { {"If storageBuffers is VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2, and either the descriptorBindingStorageBufferUpdateAfterBind feature or the descriptorBindingStorageTexelBufferUpdateAfterBind feature is enabled on the device, robustBufferAccessUpdateAfterBind must be VK_TRUE", 357}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-storageBuffers-parameter", 60}, { {"storageBuffers must be a valid VkPipelineRobustnessBufferBehavior value", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-uniformBuffers-10637", 56}, { {"If uniformBuffers is VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2, and either the descriptorBindingInlineUniformBlockUpdateAfterBind feature, the descriptorBindingUniformBufferUpdateAfterBind feature, or the descriptorBindingUniformTexelBufferUpdateAfterBind feature is enabled on the device, robustBufferAccessUpdateAfterBind must be VK_TRUE", 422}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-uniformBuffers-parameter", 60}, { {"uniformBuffers must be a valid VkPipelineRobustnessBufferBehavior value", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineRobustnessCreateInfo-vertexInputs-parameter", 58}, { {"vertexInputs must be a valid VkPipelineRobustnessBufferBehavior value", 69}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineSampleLocationsStateCreateInfoEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_SAMPLE_LOCATIONS_STATE_CREATE_INFO_EXT", 79}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineSampleLocationsStateCreateInfoEXT-sampleLocationsInfo-parameter", 78}, { {"sampleLocationsInfo must be a valid VkSampleLocationsInfoEXT structure", 70}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-clustercullingShader-07813", 63}, { {"If the clustercullingShader feature is not enabled, stage must not be VK_SHADER_STAGE_CLUSTER_CULLING_BIT_HUAWEI", 112}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-descriptorHeap-11314", 57}, { {"If the descriptorHeap feature is not enabled, VkShaderDescriptorSetAndBindingMappingInfoEXT::mappingCount must be 0", 115}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-02758", 48}, { {"If flags has both the VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT and VK_PIPELINE_SHADER_STAGE_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT flags set, the local workgroup size in the X dimension of the pipeline must be a multiple of maxSubgroupSize", 257}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-02759", 48}, { {"If flags has the VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT flag set and flags does not have the VK_PIPELINE_SHADER_STAGE_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT flag set and no VkPipelineShaderStageRequiredSubgroupSizeCreateInfo structure is included in the pNext chain, the local workgroup size in the X dimension of the pipeline must be a multiple of subgroupSize", 381}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-02784", 48}, { {"If flags has the VK_PIPELINE_SHADER_STAGE_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT flag set, the subgroupSizeControl feature must be enabled", 138}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-02785", 48}, { {"If flags has the VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT flag set, the computeFullSubgroups feature must be enabled", 134}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-08988", 48}, { {"If flags includes VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT, stage must be one of VK_SHADER_STAGE_MESH_BIT_EXT, VK_SHADER_STAGE_TASK_BIT_EXT, or VK_SHADER_STAGE_COMPUTE_BIT", 189}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-flags-parameter", 52}, { {"flags must be a valid combination of VkPipelineShaderStageCreateFlagBits values", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-maxClipDistances-00708", 59}, { {"If the identified entry point includes any variable in its interface that is declared with the ClipDistance BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxClipDistances", 223}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-maxCombinedClipAndCullDistances-00710", 74}, { {"If the identified entry point includes variables in its interface that are declared with the ClipDistance BuiltIn decoration and variables in its interface that are declared with the CullDistance BuiltIn decoration, those variables must not have array sizes which sum to more than VkPhysicalDeviceLimits::maxCombinedClipAndCullDistances", 336}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-maxCullDistances-00709", 59}, { {"If the identified entry point includes any variable in its interface that is declared with the CullDistance BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxCullDistances", 223}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-maxSampleMaskWords-00711", 61}, { {"If the identified entry point includes any variable in its interface that is declared with the SampleMask BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxSampleMaskWords", 223}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-module-08987", 49}, { {"If module uses the OpTypeCooperativeMatrixKHR instruction with a Scope equal to Subgroup, then the local workgroup size in the X dimension of the pipeline must be a multiple of the effective subgroup size", 204}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-module-10169", 49}, { {"If module uses the OpTypeCooperativeMatrixKHR instruction with a Scope equal to Workgroup, then the local workgroup size in the X dimension of the pipeline must be a multiple of the effective subgroup size and the total local workgroup size must be a power of two multiple of the effective subgroup size and must be less than or equal to cooperativeMatrixWorkgroupScopeMaxWorkgroupSize", 385}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-module-parameter", 53}, { {"If module is not VK_NULL_HANDLE, module must be a valid VkShaderModule handle", 77}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pName-00707", 48}, { {"pName must be the name of an OpEntryPoint in module with an execution model that matches stage", 94}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pName-parameter", 52}, { {"pName must be a null-terminated UTF-8 string", 44}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-02754", 48}, { {"If a VkPipelineShaderStageRequiredSubgroupSizeCreateInfo structure is included in the pNext chain, flags must not have the VK_PIPELINE_SHADER_STAGE_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT flag set", 195}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-02755", 48}, { {"If a VkPipelineShaderStageRequiredSubgroupSizeCreateInfo structure is included in the pNext chain, the subgroupSizeControl feature must be enabled, and stage must be a valid bit specified in requiredSubgroupSizeStages", 217}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-02756", 48}, { {"If a VkPipelineShaderStageRequiredSubgroupSizeCreateInfo structure is included in the pNext chain and stage is VK_SHADER_STAGE_COMPUTE_BIT, VK_SHADER_STAGE_MESH_BIT_EXT, or VK_SHADER_STAGE_TASK_BIT_EXT, the local workgroup size of the shader must be less than or equal to the product of VkPipelineShaderStageRequiredSubgroupSizeCreateInfo::requiredSubgroupSize and maxComputeWorkgroupSubgroups", 393}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-02757", 48}, { {"If a VkPipelineShaderStageRequiredSubgroupSizeCreateInfo structure is included in the pNext chain, and flags has the VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT flag set, the local workgroup size in the X dimension of the pipeline must be a multiple of VkPipelineShaderStageRequiredSubgroupSizeCreateInfo::requiredSubgroupSize", 341}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11315", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, the mapped resource in the shader must be a variable with a structure type decorated with Block in the Uniform Storage Class", 333}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11316", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, the mapped structure must not be larger than maxPushDataSize minus the pushDataOffset used in the mapping", 205}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11317", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, the sum of mapped structure size and shaderRecordDataOffset used in the mapping must not be larger than maxShaderGroupStride", 233}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11318", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT the mapped resource in the shader must be one of: A variable with a structure type decorated with Block in the Uniform Storage Class A variable with a structure type decorated with BufferBlock in the Uniform Storage Class A variable with a structure type decorated with Block in the StorageBuffer Storage Class A OpTypeAccelerationStructureKHR variable A OpTypeAccelerationStructureNV variable", 605}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11378", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the OpArrayLength or OpUntypedArrayLengthKHR instruction must not be used on that resource", 303}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-11399", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and the mapped resource declaration is an array, the pEmbeddedSampler member of the corresponding mapping structure must be NULL", 490}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-12454", 48}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, and there is a variable in the shader that matches the mapping, the stage must be a ray tracing stage", 333}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pNext-pNext", 48}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDebugUtilsObjectNameInfoEXT, VkPipelineRobustnessCreateInfo, VkPipelineShaderStageModuleIdentifierCreateInfoEXT, VkPipelineShaderStageNodeCreateInfoAMDX, VkPipelineShaderStageRequiredSubgroupSizeCreateInfo, VkShaderDescriptorSetAndBindingMappingInfoEXT, VkShaderModuleCreateInfo, VkShaderModuleValidationCacheCreateInfoEXT, or VkValidationFeaturesEXT", 483}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pSpecializationInfo-06849", 62}, { {"If a shader module identifier is not specified, the shader code used by the pipeline must be valid as described by the Khronos SPIR-V Specification after applying the specializations provided in pSpecializationInfo, if any, and then converting all specialization constants into fixed constants", 293}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-pSpecializationInfo-parameter", 66}, { {"If pSpecializationInfo is not NULL, pSpecializationInfo must be a valid pointer to a valid VkSpecializationInfo structure", 121}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO", 65}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-sType-unique", 49}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00704", 48}, { {"If the geometryShader feature is not enabled, stage must not be VK_SHADER_STAGE_GEOMETRY_BIT", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00705", 48}, { {"If the tessellationShader feature is not enabled, stage must not be VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 155}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00706", 48}, { {"stage must not be VK_SHADER_STAGE_ALL_GRAPHICS, or VK_SHADER_STAGE_ALL", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00713", 48}, { {"If stage is VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, and the identified entry point has an OpExecutionMode instruction specifying a patch size with OutputVertices, the patch size must be greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxTessellationPatchSize", 324}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00714", 48}, { {"If stage is VK_SHADER_STAGE_GEOMETRY_BIT, the identified entry point must have an OpExecutionMode instruction specifying a maximum output vertex count that is greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxGeometryOutputVertices", 249}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-00715", 48}, { {"If stage is VK_SHADER_STAGE_GEOMETRY_BIT, the identified entry point must have an OpExecutionMode instruction specifying an invocation count that is greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxGeometryShaderInvocations", 242}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-02091", 48}, { {"If the meshShaders feature is not enabled, stage must not be VK_SHADER_STAGE_MESH_BIT_EXT", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-02092", 48}, { {"If the taskShaders feature is not enabled, stage must not be VK_SHADER_STAGE_TASK_BIT_EXT", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-02596", 48}, { {"If stage is either VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, or VK_SHADER_STAGE_GEOMETRY_BIT, and the identified entry point writes to Layer for any primitive, it must write the same value to Layer for all vertices of a given primitive", 308}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-02597", 48}, { {"If stage is either VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, or VK_SHADER_STAGE_GEOMETRY_BIT, and the identified entry point writes to ViewportIndex for any primitive, it must write the same value to ViewportIndex for all vertices of a given primitive", 324}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-06685", 48}, { {"If stage is VK_SHADER_STAGE_FRAGMENT_BIT, and the identified entry point writes to FragDepth in any execution path, all execution paths that are not exclusive to helper invocations must either discard the fragment, or write or initialize the value of FragDepth", 260}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-06686", 48}, { {"If stage is VK_SHADER_STAGE_FRAGMENT_BIT, and the identified entry point writes to FragStencilRefEXT in any execution path, all execution paths that are not exclusive to helper invocations must either discard the fragment, or write or initialize the value of FragStencilRefEXT", 276}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-06844", 48}, { {"If a shader module identifier is specified for this stage, the pNext chain must not include a VkShaderModuleCreateInfo structure", 128}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-06845", 48}, { {"If a shader module identifier is not specified for this stage, module must be a valid VkShaderModule, or there must be a valid VkShaderModuleCreateInfo structure in the pNext chain , or the pNext chain of the parent Vk*CreateInfo structure must set VkPipelineBinaryInfoKHR::binaryCount to a value greater than 0,", 312}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-06848", 48}, { {"If a shader module identifier is specified for this stage, module must be VK_NULL_HANDLE", 88}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-08771", 48}, { {"If a shader module identifier is not specified for this stage, module must be a valid VkShaderModule , or the pNext chain of the parent Vk*CreateInfo structure must set VkPipelineBinaryInfoKHR::binaryCount to a value greater than 0, if none of the following features are enabled: graphicsPipelineLibrary maintenance5", 316}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageCreateInfo-stage-parameter", 52}, { {"stage must be a valid VkShaderStageFlagBits value", 49}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageModuleIdentifierCreateInfoEXT-identifierSize-06852", 76}, { {"identifierSize must be less-or-equal to VK_MAX_SHADER_MODULE_IDENTIFIER_SIZE_EXT", 80}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageModuleIdentifierCreateInfoEXT-pIdentifier-parameter", 77}, { {"If identifierSize is not 0, pIdentifier must be a valid pointer to an array of identifierSize uint8_t values", 108}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageModuleIdentifierCreateInfoEXT-pNext-06850", 67}, { {"If this structure is included in a pNext chain and identifierSize is not equal to 0, the shaderModuleIdentifier feature must be enabled", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageModuleIdentifierCreateInfoEXT-pNext-06851", 67}, { {"If this structure is included in a pNext chain of VkPipelineShaderStageCreateInfo and identifierSize is not equal to 0, the pipeline must be created with the VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT flag set", 223}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageModuleIdentifierCreateInfoEXT-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_MODULE_IDENTIFIER_CREATE_INFO_EXT", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageNodeCreateInfoAMDX-pName-parameter", 60}, { {"If pName is not NULL, pName must be a null-terminated UTF-8 string", 66}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkPipelineShaderStageNodeCreateInfoAMDX-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_NODE_CREATE_INFO_AMDX", 75}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-VkPipelineShaderStageRequiredSubgroupSizeCreateInfo-requiredSubgroupSize-02760", 83}, { {"requiredSubgroupSize must be a power-of-two integer", 51}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageRequiredSubgroupSizeCreateInfo-requiredSubgroupSize-02761", 83}, { {"requiredSubgroupSize must be greater or equal to minSubgroupSize", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageRequiredSubgroupSizeCreateInfo-requiredSubgroupSize-02762", 83}, { {"requiredSubgroupSize must be less than or equal to maxSubgroupSize", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineShaderStageRequiredSubgroupSizeCreateInfo-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_REQUIRED_SUBGROUP_SIZE_CREATE_INFO", 88}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkPipelineTessellationDomainOriginStateCreateInfo-domainOrigin-parameter", 77}, { {"domainOrigin must be a valid VkTessellationDomainOrigin value", 61}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationDomainOriginStateCreateInfo-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_DOMAIN_ORIGIN_STATE_CREATE_INFO", 85}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationStateCreateInfo-flags-zerobitmask", 60}, { {"flags must be 0", 15}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationStateCreateInfo-pNext-pNext", 54}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineTessellationDomainOriginStateCreateInfo", 104}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationStateCreateInfo-patchControlPoints-01214", 67}, { {"patchControlPoints must be greater than zero and less than or equal to VkPhysicalDeviceLimits::maxTessellationPatchSize", 119}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationStateCreateInfo-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO", 71}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineTessellationStateCreateInfo-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/tessellation.html", 26} } },
        { {"VUID-VkPipelineVertexInputDivisorStateCreateInfo-pVertexBindingDivisors-parameter", 81}, { {"pVertexBindingDivisors must be a valid pointer to an array of vertexBindingDivisorCount VkVertexInputBindingDivisorDescription structures", 137}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputDivisorStateCreateInfo-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_DIVISOR_STATE_CREATE_INFO", 79}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputDivisorStateCreateInfo-vertexBindingDivisorCount-arraylength", 86}, { {"vertexBindingDivisorCount must be greater than 0", 48}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-binding-00615", 55}, { {"For every binding specified by each element of pVertexAttributeDescriptions, a VkVertexInputBindingDescription must exist in pVertexBindingDescriptions with the same value of binding", 182}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-flags-zerobitmask", 59}, { {"flags must be 0", 15}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-pNext-pNext", 53}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineVertexInputDivisorStateCreateInfo", 98}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-pVertexAttributeDescriptions-00617", 76}, { {"All elements of pVertexAttributeDescriptions must describe distinct attribute locations", 87}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-pVertexAttributeDescriptions-parameter", 80}, { {"If vertexAttributeDescriptionCount is not 0, pVertexAttributeDescriptions must be a valid pointer to an array of vertexAttributeDescriptionCount valid VkVertexInputAttributeDescription structures", 195}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-pVertexBindingDescriptions-00616", 74}, { {"All elements of pVertexBindingDescriptions must describe distinct binding numbers", 81}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-pVertexBindingDescriptions-parameter", 78}, { {"If vertexBindingDescriptionCount is not 0, pVertexBindingDescriptions must be a valid pointer to an array of vertexBindingDescriptionCount valid VkVertexInputBindingDescription structures", 187}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO", 71}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-sType-unique", 54}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-vertexAttributeDescriptionCount-00614", 79}, { {"vertexAttributeDescriptionCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputAttributes", 110}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineVertexInputStateCreateInfo-vertexBindingDescriptionCount-00613", 77}, { {"vertexBindingDescriptionCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindings", 106}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkPipelineViewportCoarseSampleOrderStateCreateInfoNV-pCustomSampleOrders-02234", 83}, { {"The array pCustomSampleOrders must not contain two structures with matching values for both the shadingRate and sampleCount members", 131}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportCoarseSampleOrderStateCreateInfoNV-pCustomSampleOrders-parameter", 87}, { {"If customSampleOrderCount is not 0, pCustomSampleOrders must be a valid pointer to an array of customSampleOrderCount valid VkCoarseSampleOrderCustomNV structures", 162}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportCoarseSampleOrderStateCreateInfoNV-sType-sType", 69}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_COARSE_SAMPLE_ORDER_STATE_CREATE_INFO_NV", 90}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportCoarseSampleOrderStateCreateInfoNV-sampleOrderType-02072", 79}, { {"If sampleOrderType is not VK_COARSE_SAMPLE_ORDER_TYPE_CUSTOM_NV, customSamplerOrderCount must be 0", 98}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportCoarseSampleOrderStateCreateInfoNV-sampleOrderType-parameter", 83}, { {"sampleOrderType must be a valid VkCoarseSampleOrderTypeNV value", 63}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportDepthClampControlCreateInfoEXT-depthClampMode-parameter", 78}, { {"depthClampMode must be a valid VkDepthClampModeEXT value", 56}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportDepthClampControlCreateInfoEXT-pDepthClampRange-09646", 76}, { {"If depthClampMode is VK_DEPTH_CLAMP_MODE_USER_DEFINED_RANGE_EXT, and the pipeline is not created with VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT, then pDepthClampRange must be a valid pointer to a valid VkDepthClampRangeEXT structure", 229}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportDepthClampControlCreateInfoEXT-pDepthClampRange-parameter", 80}, { {"If pDepthClampRange is not NULL, pDepthClampRange must be a valid pointer to a valid VkDepthClampRangeEXT structure", 115}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportDepthClampControlCreateInfoEXT-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLAMP_CONTROL_CREATE_INFO_EXT", 85}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportDepthClipControlCreateInfoEXT-negativeOneToOne-06470", 75}, { {"If the depthClipControl feature is not enabled, negativeOneToOne must be VK_FALSE", 81}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportDepthClipControlCreateInfoEXT-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLIP_CONTROL_CREATE_INFO_EXT", 84}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportExclusiveScissorStateCreateInfoNV-exclusiveScissorCount-02027", 84}, { {"If the multiViewport feature is not enabled, exclusiveScissorCount must be 0 or 1", 81}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportExclusiveScissorStateCreateInfoNV-exclusiveScissorCount-02028", 84}, { {"exclusiveScissorCount must be less than or equal to VkPhysicalDeviceLimits::maxViewports", 88}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportExclusiveScissorStateCreateInfoNV-exclusiveScissorCount-02029", 84}, { {"exclusiveScissorCount must be 0 or greater than or equal to the viewportCount member of VkPipelineViewportStateCreateInfo", 121}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportExclusiveScissorStateCreateInfoNV-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_EXCLUSIVE_SCISSOR_STATE_CREATE_INFO_NV", 88}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkPipelineViewportShadingRateImageStateCreateInfoNV-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SHADING_RATE_IMAGE_STATE_CREATE_INFO_NV", 89}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportShadingRateImageStateCreateInfoNV-shadingRateImageEnable-02056", 85}, { {"If shadingRateImageEnable is VK_TRUE, viewportCount must be greater or equal to the viewportCount member of VkPipelineViewportStateCreateInfo", 141}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportShadingRateImageStateCreateInfoNV-viewportCount-02054", 76}, { {"If the multiViewport feature is not enabled, viewportCount must be 0 or 1", 73}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportShadingRateImageStateCreateInfoNV-viewportCount-02055", 76}, { {"viewportCount must be less than or equal to VkPhysicalDeviceLimits::maxViewports", 80}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-flags-zerobitmask", 56}, { {"flags must be 0", 15}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-offset-02822", 51}, { {"Evaluation of (offset.x + extent.width) must not cause a signed integer addition overflow for any element of pScissors", 118}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-offset-02823", 51}, { {"Evaluation of (offset.y + extent.height) must not cause a signed integer addition overflow for any element of pScissors", 119}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-pNext-pNext", 50}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineViewportCoarseSampleOrderStateCreateInfoNV, VkPipelineViewportDepthClampControlCreateInfoEXT, VkPipelineViewportDepthClipControlCreateInfoEXT, VkPipelineViewportExclusiveScissorStateCreateInfoNV, VkPipelineViewportShadingRateImageStateCreateInfoNV, VkPipelineViewportSwizzleStateCreateInfoNV, or VkPipelineViewportWScalingStateCreateInfoNV", 480}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO", 67}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-sType-unique", 51}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-scissorCount-01217", 57}, { {"If the multiViewport feature is not enabled, scissorCount must not be greater than 1", 84}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-scissorCount-01219", 57}, { {"scissorCount must be less than or equal to VkPhysicalDeviceLimits::maxViewports", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-scissorCount-04134", 57}, { {"If scissorCount and viewportCount are both not dynamic, then scissorCount and viewportCount must be identical", 109}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-scissorCount-04136", 57}, { {"If the graphics pipeline is being created with VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT set then scissorCount must be 0, otherwise scissorCount must be greater than 0", 161}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-viewportCount-01216", 58}, { {"If the multiViewport feature is not enabled, viewportCount must not be greater than 1", 85}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-viewportCount-01218", 58}, { {"viewportCount must be less than or equal to VkPhysicalDeviceLimits::maxViewports", 80}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-viewportCount-04135", 58}, { {"If the graphics pipeline is being created with VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT set then viewportCount must be 0, otherwise viewportCount must be greater than 0", 164}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-viewportWScalingEnable-01726", 67}, { {"If the viewportWScalingEnable member of a VkPipelineViewportWScalingStateCreateInfoNV structure included in the pNext chain is VK_TRUE, the viewportCount member of the VkPipelineViewportWScalingStateCreateInfoNV structure must be greater than or equal to VkPipelineViewportStateCreateInfo::viewportCount", 303}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportStateCreateInfo-x-02821", 46}, { {"The x and y members of offset member of any element of pScissors must be greater than or equal to 0", 99}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportSwizzleStateCreateInfoNV-flags-zerobitmask", 65}, { {"flags must be 0", 15}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportSwizzleStateCreateInfoNV-pViewportSwizzles-parameter", 75}, { {"pViewportSwizzles must be a valid pointer to an array of viewportCount valid VkViewportSwizzleNV structures", 107}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportSwizzleStateCreateInfoNV-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SWIZZLE_STATE_CREATE_INFO_NV", 78}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportSwizzleStateCreateInfoNV-viewportCount-01215", 67}, { {"viewportCount must be greater than or equal to the viewportCount set in VkPipelineViewportStateCreateInfo", 105}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportSwizzleStateCreateInfoNV-viewportCount-arraylength", 73}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportWScalingStateCreateInfoNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_W_SCALING_STATE_CREATE_INFO_NV", 80}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPipelineViewportWScalingStateCreateInfoNV-viewportCount-arraylength", 74}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkPresentFrameTokenGGP-frameToken-02680", 44}, { {"frameToken must be a valid GgpFrameToken", 40}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentFrameTokenGGP-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_FRAME_TOKEN_GGP", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-None-10820", 31}, { {"The swapchain must have been created with VK_SWAPCHAIN_CREATE_PRESENT_ID_2_BIT_KHR bit set in the VkSwapchainCreateFlagBitsKHR field", 132}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-pPresentIds-parameter", 42}, { {"If pPresentIds is not NULL, pPresentIds must be a valid pointer to an array of swapchainCount uint64_t values", 109}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-presentIds-10819", 37}, { {"Each non-zero entry in presentIds must be greater than all previously submitted present ids for the associated swapchain in VkPresentInfoKHR::pSwapchains", 153}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_ID_2_KHR", 48}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-swapchainCount-10818", 41}, { {"swapchainCount must be the same value as VkPresentInfoKHR::swapchainCount, where this VkPresentId2KHR is in the pNext chain of the VkPresentInfoKHR structure", 157}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentId2KHR-swapchainCount-arraylength", 47}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentIdKHR-pPresentIds-parameter", 41}, { {"If pPresentIds is not NULL, pPresentIds must be a valid pointer to an array of swapchainCount uint64_t values", 109}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentIdKHR-presentIds-04999", 36}, { {"Each non-zero entry in presentIds must be greater than all previously submitted present ids for the associated swapchain in VkPresentInfoKHR::pSwapchains", 153}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentIdKHR-sType-sType", 31}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_ID_KHR", 46}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentIdKHR-swapchainCount-04998", 40}, { {"swapchainCount must be the same value as VkPresentInfoKHR::swapchainCount, where this VkPresentIdKHR is in the pNext chain of the VkPresentInfoKHR structure", 156}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentIdKHR-swapchainCount-arraylength", 46}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-commonparent", 34}, { {"Both of the elements of pSwapchains, and the elements of pWaitSemaphores that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 192}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pImageIndices-01430", 41}, { {"Each element of pImageIndices must be the index of a presentable image acquired from the swapchain specified by the corresponding element of the pSwapchains array, and the presented image subresource must be in the VK_IMAGE_LAYOUT_PRESENT_SRC_KHR or VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR layout at the time the operation is executed on a VkDevice", 343}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pImageIndices-parameter", 45}, { {"pImageIndices must be a valid pointer to an array of swapchainCount uint32_t values", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pNext-06235", 33}, { {"If a VkPresentIdKHR structure is included in the pNext chain, and the presentId feature is not enabled, each presentIds entry in that structure must be NULL", 156}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pNext-09759", 33}, { {"If the pNext chain of this structure includes a VkFrameBoundaryTensorsARM structure then it must also include a VkFrameBoundaryEXT structure", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pNext-10821", 33}, { {"If a VkPresentId2KHR structure is included in the pNext chain, and the presentId2 feature is not enabled, each presentIds entry in that structure must be zero", 158}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pNext-pNext", 33}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceGroupPresentInfoKHR, VkDisplayPresentInfoKHR, VkFrameBoundaryEXT, VkFrameBoundaryTensorsARM, VkPresentFrameTokenGGP, VkPresentId2KHR, VkPresentIdKHR, VkPresentRegionsKHR, VkPresentTimesInfoGOOGLE, VkPresentTimingsInfoEXT, VkSetPresentConfigNV, VkSwapchainPresentFenceInfoKHR, or VkSwapchainPresentModeInfoKHR", 447}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pResults-parameter", 40}, { {"If pResults is not NULL, pResults must be a valid pointer to an array of swapchainCount VkResult values", 103}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pSwapchain-09231", 38}, { {"Elements of pSwapchain must be unique", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pSwapchains-09199", 39}, { {"If any element of the pSwapchains array has been created with VkSwapchainPresentModesCreateInfoKHR, all of the elements of this array must be created with VkSwapchainPresentModesCreateInfoKHR", 191}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pSwapchains-parameter", 43}, { {"pSwapchains must be a valid pointer to an array of swapchainCount valid VkSwapchainKHR handles", 94}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-pWaitSemaphores-parameter", 47}, { {"If waitSemaphoreCount is not 0, pWaitSemaphores must be a valid pointer to an array of waitSemaphoreCount valid VkSemaphore handles", 131}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-presentId2Supported-10822", 47}, { {"If a VkPresentId2KHR structure is included and contains non-zero presentIds, presentId2Supported must be VK_TRUE in the VkSurfaceCapabilitiesPresentId2KHR structure returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for the surface", 235}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_INFO_KHR", 48}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-swapchainCount-arraylength", 48}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentInfoKHR-swapchainMaintenance1-10158", 49}, { {"If the swapchainMaintenance1 feature is not enabled, then the pNext chain must not include a VkSwapchainPresentFenceInfoKHR structure", 133}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentRegionKHR-pRectangles-parameter", 45}, { {"If rectangleCount is not 0, and pRectangles is not NULL, pRectangles must be a valid pointer to an array of rectangleCount valid VkRectLayerKHR structures", 154}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentRegionsKHR-pRegions-parameter", 43}, { {"If pRegions is not NULL, pRegions must be a valid pointer to an array of swapchainCount valid VkPresentRegionKHR structures", 123}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentRegionsKHR-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_REGIONS_KHR", 51}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentRegionsKHR-swapchainCount-01260", 45}, { {"swapchainCount must be the same value as VkPresentInfoKHR::swapchainCount, where VkPresentInfoKHR is included in the pNext chain of this VkPresentRegionsKHR structure", 166}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentRegionsKHR-swapchainCount-arraylength", 51}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimesInfoGOOGLE-pTimes-parameter", 46}, { {"If pTimes is not NULL, pTimes must be a valid pointer to an array of swapchainCount VkPresentTimeGOOGLE structures", 114}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimesInfoGOOGLE-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_TIMES_INFO_GOOGLE", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimesInfoGOOGLE-swapchainCount-01247", 50}, { {"swapchainCount must be the same value as VkPresentInfoKHR::swapchainCount, where VkPresentInfoKHR is included in the pNext chain of this VkPresentTimesInfoGOOGLE structure", 171}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimesInfoGOOGLE-swapchainCount-arraylength", 56}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-flags-parameter", 43}, { {"flags must be a valid combination of VkPresentTimingInfoFlagBitsEXT values", 74}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-pNext-pNext", 39}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-presentStageQueries-parameter", 57}, { {"presentStageQueries must be a valid combination of VkPresentStageFlagBitsEXT values", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_TIMING_INFO_EXT", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-targetTime-12236", 44}, { {"If targetTime is not zero and flags does not contain VK_PRESENT_TIMING_INFO_PRESENT_AT_RELATIVE_TIME_BIT_EXT, the presentAtAbsoluteTime feature must be enabled and the presentAtAbsoluteTimeSupported member of the VkPresentTimingSurfaceCapabilitiesEXT returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for the surface associated with the swapchain must be VK_TRUE", 367}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-targetTime-12237", 44}, { {"If targetTime is not zero and flags contains VK_PRESENT_TIMING_INFO_PRESENT_AT_RELATIVE_TIME_BIT_EXT, the presentAtRelativeTime feature must be enabled and the presentAtRelativeTimeSupported member of the VkPresentTimingSurfaceCapabilitiesEXT returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for the surface associated with the swapchain must be VK_TRUE", 359}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-targetTimeDomainPresentStage-parameter", 66}, { {"targetTimeDomainPresentStage must be a valid combination of VkPresentStageFlagBitsEXT values", 92}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-timeDomainId-12238", 46}, { {"If timeDomainId is associated with a VK_TIME_DOMAIN_PRESENT_STAGE_LOCAL_EXT time domain, and targetTime is not zero, targetTimeDomainPresentStage must be a single VkPresentStageFlagsEXT value", 191}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingInfoEXT-timeDomainId-12400", 46}, { {"timeDomainId must be an id previously returned by vkGetSwapchainTimeDomainPropertiesEXT", 87}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingSurfaceCapabilitiesEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_TIMING_SURFACE_CAPABILITIES_EXT", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-pSwapchains-12234", 46}, { {"All swapchains in VkPresentInfoKHR::pSwapchains must have been created with the VkSwapchainCreateInfoKHR::flags field containing VK_SWAPCHAIN_CREATE_PRESENT_TIMING_BIT_EXT", 171}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-pSwapchains-12235", 46}, { {"For each member of VkPresentInfoKHR::pSwapchains, if the associated VkPresentTimingInfoEXT::targetTime is not zero, the swapchain's current present mode must be VK_PRESENT_MODE_FIFO_LATEST_READY_KHR, VK_PRESENT_MODE_FIFO_KHR or VK_PRESENT_MODE_FIFO_RELAXED_KHR", 260}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-pTimingInfos-parameter", 51}, { {"If pTimingInfos is not NULL, pTimingInfos must be a valid pointer to an array of swapchainCount valid VkPresentTimingInfoEXT structures", 135}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_TIMINGS_INFO_EXT", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-swapchainCount-12233", 49}, { {"swapchainCount must be equal to VkPresentInfoKHR::swapchainCount", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentTimingsInfoEXT-swapchainCount-arraylength", 55}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentWait2InfoKHR-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPresentWait2InfoKHR-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_PRESENT_WAIT_2_INFO_KHR", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkPrivateDataSlotCreateInfo-flags-zerobitmask", 50}, { {"flags must be 0", 15}, { "chapters/private_data.html", 26} } },
        { {"VUID-VkPrivateDataSlotCreateInfo-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/private_data.html", 26} } },
        { {"VUID-VkPrivateDataSlotCreateInfo-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_PRIVATE_DATA_SLOT_CREATE_INFO", 61}, { "chapters/private_data.html", 26} } },
        { {"VUID-VkProtectedSubmitInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_PROTECTED_SUBMIT_INFO", 53}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12342", 40}, { {"When chained to VkPushDataInfoEXT, if the command buffer is executing graphics operations, bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxGraphicsPushDataBanks", 184}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12343", 40}, { {"When chained to VkPushDataInfoEXT, if the command buffer is executing compute operations, bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxComputePushDataBanks", 182}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12344", 40}, { {"When chained to VkPushConstantsInfo, if VkPushConstantsInfo::stageFlags includes a graphics stage then bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxGraphicsPushConstantBanks", 200}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12345", 40}, { {"When chained to VkPushConstantsInfo, if VkPushConstantsInfo::stageFlags includes a compute stage then bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxComputePushConstantBanks", 198}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12346", 40}, { {"When chained to VkDescriptorSetAndBindingMappingEXT for a graphics shader stage, bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxGraphicsPushDataBanks", 174}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-bank-12347", 40}, { {"When chained to VkDescriptorSetAndBindingMappingEXT for a compute shader stage, bank must be less than VkPhysicalDevicePushConstantBankPropertiesNV::maxComputePushDataBanks", 172}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantBankInfoNV-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_PUSH_CONSTANT_BANK_INFO_NV", 58}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushConstantRange-offset-00294", 37}, { {"offset must be less than VkPhysicalDeviceLimits::maxPushConstantsSize", 69}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-offset-00295", 37}, { {"offset must be a multiple of 4", 30}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-size-00296", 35}, { {"size must be greater than 0", 27}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-size-00297", 35}, { {"size must be a multiple of 4", 28}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-size-00298", 35}, { {"size must be less than or equal to VkPhysicalDeviceLimits::maxPushConstantsSize minus offset", 92}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-stageFlags-parameter", 45}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantRange-stageFlags-requiredbitmask", 51}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-None-09495", 35}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-layout-09496", 37}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-layout-parameter", 41}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-offset-00368", 37}, { {"offset must be a multiple of 4", 30}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-offset-00370", 37}, { {"offset must be less than VkPhysicalDeviceLimits::maxPushConstantsSize", 69}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-offset-01795", 37}, { {"For each byte in the range specified by offset and size and for each shader stage in stageFlags, there must be a push constant range in layout that includes that byte and that stage", 181}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-offset-01796", 37}, { {"For each byte in the range specified by offset and size and for each push constant range that overlaps that byte, stageFlags must include all stages in that push constant range's VkPushConstantRange::stageFlags", 210}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-pNext-pNext", 36}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo or VkPushConstantBankInfoNV", 185}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-pValues-parameter", 42}, { {"pValues must be a valid pointer to an array of size bytes", 57}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_PUSH_CONSTANTS_INFO", 51}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-sType-unique", 37}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-size-00369", 35}, { {"size must be a multiple of 4", 28}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-size-00371", 35}, { {"size must be less than or equal to VkPhysicalDeviceLimits::maxPushConstantsSize minus offset", 92}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-size-arraylength", 41}, { {"size must be greater than 0", 27}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-stageFlags-parameter", 45}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushConstantsInfo-stageFlags-requiredbitmask", 51}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDataInfoEXT-data-11419", 33}, { {"data.size must be a multiple of 4", 33}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-data-parameter", 37}, { {"data must be a valid VkHostAddressRangeConstEXT structure", 57}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-offset-11243", 35}, { {"The sum of offset and data.size must be less than or equal to maxPushDataSize", 77}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-offset-11418", 35}, { {"offset must be a multiple of 4", 30}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-pNext-pNext", 34}, { {"pNext must be NULL or a pointer to a valid instance of VkPushConstantBankInfoNV", 79}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-sType-sType", 34}, { {"sType must be VK_STRUCTURE_TYPE_PUSH_DATA_INFO_EXT", 50}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDataInfoEXT-sType-unique", 35}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkPushDescriptorSetInfo-None-09495", 39}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-descriptorWriteCount-arraylength", 61}, { {"descriptorWriteCount must be greater than 0", 43}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-layout-09496", 41}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-layout-parameter", 45}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-pDescriptorWrites-06494", 52}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, pDescriptorWrites[i].pImageInfo must be a valid pointer to an array of pDescriptorWrites[i].descriptorCount valid VkDescriptorImageInfo structures", 389}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-pDescriptorWrites-parameter", 56}, { {"pDescriptorWrites must be a valid pointer to an array of descriptorWriteCount valid VkWriteDescriptorSet structures", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-pNext-pNext", 40}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_INFO", 56}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-set-00364", 38}, { {"set must be less than VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-set-00365", 38}, { {"set must be the unique set number in the pipeline layout that uses a descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-stageFlags-parameter", 49}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetInfo-stageFlags-requiredbitmask", 55}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-None-09495", 51}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-None-10359", 51}, { {"If the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-commandBuffer-00366", 60}, { {"The pipelineBindPoint specified during the creation of the descriptor update template must be supported by the commandBuffer's parent VkCommandPool's queue family", 162}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-commonparent", 53}, { {"Both of descriptorUpdateTemplate, and layout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-descriptorUpdateTemplate-07994", 71}, { {"descriptorUpdateTemplate must have been created with a templateType of VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS", 122}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-descriptorUpdateTemplate-parameter", 75}, { {"descriptorUpdateTemplate must be a valid VkDescriptorUpdateTemplate handle", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-layout-07993", 53}, { {"layout must be compatible with the layout used to create descriptorUpdateTemplate", 81}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-layout-09496", 53}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-layout-parameter", 57}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-pData-01686", 52}, { {"pData must be a valid pointer to a memory containing one or more valid instances of VkDescriptorImageInfo, VkDescriptorBufferInfo, or VkBufferView in a layout defined by descriptorUpdateTemplate when it was created with vkCreateDescriptorUpdateTemplate", 252}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-pData-parameter", 56}, { {"pData must be a pointer value", 29}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-pNext-pNext", 52}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_WITH_TEMPLATE_INFO", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-sType-unique", 53}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-set-07304", 50}, { {"set must be less than VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-set-07305", 50}, { {"set must be the unique set number in the pipeline layout that uses a descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-set-07995", 50}, { {"set must be the same value used to create descriptorUpdateTemplate", 66}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkPushDescriptorSetWithTemplateInfo-set-11854", 50}, { {"set must reference a valid VkDescriptorSetLayout handle in layout", 65}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkQueryLowLatencySupportNV-pQueriedLowLatencyData-parameter", 64}, { {"pQueriedLowLatencyData must be a pointer value", 46}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkQueryLowLatencySupportNV-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_LOW_LATENCY_SUPPORT_NV", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkQueryPoolCreateInfo-flags-parameter", 42}, { {"flags must be a valid combination of VkQueryPoolCreateFlagBits values", 69}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-meshShaderQueries-07068", 50}, { {"If the meshShaderQueries feature is not enabled, queryType must not be VK_QUERY_TYPE_MESH_PRIMITIVES_GENERATED_EXT", 114}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-meshShaderQueries-07069", 50}, { {"If the meshShaderQueries feature is not enabled, and queryType is VK_QUERY_TYPE_PIPELINE_STATISTICS, pipelineStatistics must not contain VK_QUERY_PIPELINE_STATISTIC_TASK_SHADER_INVOCATIONS_BIT_EXT or VK_QUERY_PIPELINE_STATISTIC_MESH_SHADER_INVOCATIONS_BIT_EXT", 259}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-pNext-10248", 38}, { {"If the pNext chain includes a VkVideoProfileInfoKHR structure and its videoCodecOperation member is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the videoEncodeAV1 feature must be enabled", 192}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-pNext-10779", 38}, { {"If the pNext chain includes a VkVideoProfileInfoKHR structure and its videoCodecOperation member is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the videoDecodeVP9 feature must be enabled", 192}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-pNext-10918", 38}, { {"If the pNext chain includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be enabled", 141}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkQueryPoolPerformanceCreateInfoKHR, VkQueryPoolPerformanceQueryCreateInfoINTEL, VkQueryPoolVideoEncodeFeedbackCreateInfoKHR, VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR, VkVideoDecodeAV1ProfileInfoKHR, VkVideoDecodeH264ProfileInfoKHR, VkVideoDecodeH265ProfileInfoKHR, VkVideoDecodeUsageInfoKHR, VkVideoDecodeVP9ProfileInfoKHR, VkVideoEncodeAV1ProfileInfoKHR, VkVideoEncodeH264ProfileInfoKHR, VkVideoEncodeH265ProfileInfoKHR, VkVideoEncodeUsageInfoKHR, or VkVideoProfileInfoKHR", 620}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryCount-02763", 43}, { {"queryCount must be greater than 0", 33}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-00791", 42}, { {"If the pipelineStatisticsQuery feature is not enabled, queryType must not be VK_QUERY_TYPE_PIPELINE_STATISTICS", 110}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-00792", 42}, { {"If queryType is VK_QUERY_TYPE_PIPELINE_STATISTICS, pipelineStatistics must be a valid combination of VkQueryPipelineStatisticFlagBits values", 140}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-03222", 42}, { {"If queryType is VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the pNext chain must include a VkQueryPoolPerformanceCreateInfoKHR structure", 129}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-07133", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then the pNext chain must include a VkVideoProfileInfoKHR structure with videoCodecOperation specifying an encode operation", 180}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-07906", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then the pNext chain must include a VkQueryPoolVideoEncodeFeedbackCreateInfoKHR structure", 146}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-07907", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, and the pNext chain includes a VkVideoProfileInfoKHR structure and a VkQueryPoolVideoEncodeFeedbackCreateInfoKHR structure, then VkQueryPoolVideoEncodeFeedbackCreateInfoKHR::encodeFeedbackFlags must not contain any bits that are not set in VkVideoEncodeCapabilitiesKHR::supportedEncodeFeedbackFlags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile described by VkVideoProfileInfoKHR and its pNext chain", 488}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-09534", 42}, { {"If queryType is VK_QUERY_TYPE_PIPELINE_STATISTICS, pipelineStatistics must not be zero", 86}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-11839", 42}, { {"If queryType is VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then at least one of the queue families of the device must support result status queries, as indicated by VkQueueFamilyQueryResultStatusPropertiesKHR::queryResultStatusSupport", 229}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12416", 42}, { {"If the elapsedTimerQuery feature is not enabled, queryType must not be VK_QUERY_TYPE_TIME_ELAPSED_QCOM", 102}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12437", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and the pNext chain includes a VkQueryPoolVideoEncodeFeedbackCreateInfoKHR structure with its encodeFeedbackFlags member containing VK_VIDEO_ENCODE_FEEDBACK_AVERAGE_QUANTIZATION_BIT_KHR, VK_VIDEO_ENCODE_FEEDBACK_MIN_QUANTIZATION_BIT_KHR, VK_VIDEO_ENCODE_FEEDBACK_MAX_QUANTIZATION_BIT_KHR, VK_VIDEO_ENCODE_FEEDBACK_INTRA_PIXELS_BIT_KHR, VK_VIDEO_ENCODE_FEEDBACK_INTER_PIXELS_BIT_KHR, VK_VIDEO_ENCODE_FEEDBACK_SKIPPED_PIXELS_BIT_KHR, or VK_VIDEO_ENCODE_FEEDBACK_PICTURE_PARTITION_COUNT_BIT_KHR, then videoEncodeFeedback2 must be enabled", 590}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12438", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and the pNext chain includes a VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR structure with a non-zero maxPerPartitionFeedbackEntries member, then videoEncodeFeedback2 must be enabled", 249}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12439", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and the pNext chain includes a VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR structure with a non-zero maxPerPartitionFeedbackEntries member, then VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR::perPartitionEncodeFeedbackFlags must not be zero", 318}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12440", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and the pNext chain includes a VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR structure, then its maxPerPartitionFeedbackEntries must be less than or equal to VkVideoEncodeFeedback2CapabilitiesKHR::maxPerPartitionFeedbackEntries, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile described by VkVideoProfileInfoKHR and its pNext chain", 426}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-12441", 42}, { {"If queryType is VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and the pNext chain includes a VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR structure, then its perPartitionEncodeFeedbackFlags must not contain any bits that are not set in VkVideoEncodeFeedback2CapabilitiesKHR::supportedPerPartitionEncodeFeedbackFlags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile described by VkVideoProfileInfoKHR and its pNext chain", 453}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-queryType-parameter", 46}, { {"queryType must be a valid VkQueryType value", 43}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO", 54}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolCreateInfo-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-counterIndexCount-arraylength", 70}, { {"counterIndexCount must be greater than 0", 40}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-pCounterIndices-03321", 62}, { {"Each element of pCounterIndices must be in the range of counters reported by vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR for the queue family specified in queueFamilyIndex", 191}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-pCounterIndices-parameter", 66}, { {"pCounterIndices must be a valid pointer to an array of counterIndexCount uint32_t values", 88}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-performanceCounterQueryPools-03237", 75}, { {"The performanceCounterQueryPools feature must be enabled", 56}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-queueFamilyIndex-03236", 63}, { {"queueFamilyIndex must be a valid queue family index of the device", 65}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceCreateInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_CREATE_INFO_KHR", 70}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceQueryCreateInfoINTEL-performanceCountersSampling-parameter", 85}, { {"performanceCountersSampling must be a valid VkQueryPoolSamplingModeINTEL value", 78}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolPerformanceQueryCreateInfoINTEL-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_QUERY_CREATE_INFO_INTEL", 78}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolVideoEncodeFeedbackCreateInfoKHR-encodeFeedbackFlags-parameter", 78}, { {"encodeFeedbackFlags must be a valid combination of VkVideoEncodeFeedbackFlagBitsKHR values", 90}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolVideoEncodeFeedbackCreateInfoKHR-encodeFeedbackFlags-requiredbitmask", 84}, { {"encodeFeedbackFlags must not be 0", 33}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolVideoEncodeFeedbackCreateInfoKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_FEEDBACK_CREATE_INFO_KHR", 80}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR-perPartitionEncodeFeedbackFlags-parameter", 102}, { {"perPartitionEncodeFeedbackFlags must be a valid combination of VkVideoEncodePerPartitionFeedbackFlagBitsKHR values", 114}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR-sType-sType", 72}, { {"sType must be VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_PER_PARTITION_FEEDBACK_CREATE_INFO_KHR", 94}, { "chapters/queries.html", 21} } },
        { {"VUID-VkQueueFamilyCheckpointProperties2NV-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_2_NV", 71}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyCheckpointPropertiesNV-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_NV", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyDataGraphOpticalFlowPropertiesARM-pNext-pNext", 63}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphOpticalFlowPropertiesARM-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_OPTICAL_FLOW_PROPERTIES_ARM", 83}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphProcessingEnginePropertiesARM-pNext-pNext", 68}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphProcessingEnginePropertiesARM-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_PROCESSING_ENGINE_PROPERTIES_ARM", 88}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphPropertiesARM-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphPropertiesARM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_PROPERTIES_ARM", 70}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphTOSAPropertiesARM-pNext-pNext", 56}, { {"pNext must be NULL", 18}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyDataGraphTOSAPropertiesARM-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_TOSA_PROPERTIES_ARM", 75}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-VkQueueFamilyGlobalPriorityProperties-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_GLOBAL_PRIORITY_PROPERTIES", 71}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyOptimalImageTransferGranularityPropertiesKHR-sType-sType", 74}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_OPTIMAL_IMAGE_TRANSFER_GRANULARITY_PROPERTIES_KHR", 94}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyOwnershipTransferPropertiesKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_OWNERSHIP_TRANSFER_PROPERTIES_KHR", 78}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyProperties2-pNext-pNext", 41}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkQueueFamilyCheckpointProperties2NV, VkQueueFamilyCheckpointPropertiesNV, VkQueueFamilyGlobalPriorityProperties, VkQueueFamilyOptimalImageTransferGranularityPropertiesKHR, VkQueueFamilyOwnershipTransferPropertiesKHR, VkQueueFamilyQueryResultStatusPropertiesKHR, or VkQueueFamilyVideoPropertiesKHR", 428}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyProperties2-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_PROPERTIES_2", 57}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyProperties2-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyQueryResultStatusPropertiesKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_QUERY_RESULT_STATUS_PROPERTIES_KHR", 79}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkQueueFamilyVideoPropertiesKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_QUEUE_FAMILY_VIDEO_PROPERTIES_KHR", 65}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV-clusterAccelerationStructure-10576", 100}, { {"The VkPhysicalDeviceClusterAccelerationStructureFeaturesNV::clusterAccelerationStructure feature must be enabled", 112}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV-sType-sType", 77}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CLUSTER_ACCELERATION_STRUCTURE_CREATE_INFO_NV", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-None-09497", 49}, { {"If the pNext chain does not include a VkPipelineCreateFlags2CreateInfo structure, flags must be a valid combination of VkPipelineCreateFlagBits values", 150}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-None-10391", 49}, { {"If a resource variables is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-None-11369", 49}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must not be VK_NULL_HANDLE", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-commonparent", 51}, { {"Both of basePipelineHandle, and layout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 158}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-02904", 50}, { {"flags must not include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-03465", 50}, { {"If flags includes VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, pLibraryInterface must not be NULL", 88}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-03470", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR, for each element of pGroups with a type of VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR or VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR, the anyHitShader of that element must not be VK_SHADER_UNUSED_KHR", 309}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-03471", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR, for each element of pGroups with a type of VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR or VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR, the closestHitShader of that element must not be VK_SHADER_UNUSED_KHR", 317}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-03594", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR bit set", 247}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-03598", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR, rayTracingPipelineShaderGroupHandleCaptureReplay must be enabled", 157}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04718", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR bit set", 199}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04719", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR bit set", 207}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04720", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR bit set", 225}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04721", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR bit set", 233}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04722", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR bit set", 235}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-04723", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR bit set", 219}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-06546", 50}, { {"flags must not include both VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR and VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-07369", 50}, { {"flags must not include both VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT and VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-07984", 50}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineIndex is -1, basePipelineHandle must be a valid ray tracing VkPipeline handle", 155}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-07985", 50}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineHandle is VK_NULL_HANDLE, basePipelineIndex must be a valid index into the calling command's pCreateInfos parameter", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-07986", 50}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, basePipelineIndex must be -1 or basePipelineHandle must be VK_NULL_HANDLE", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-08700", 50}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR and either pLibraryInfo is NULL or its libraryCount is 0, groupCount must not be 0", 143}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-08701", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV bit set", 219}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-10392", 50}, { {"If the pNext chain includes a VkPipelineCreateFlags2CreateInfo structure, flags must not include both VK_PIPELINE_CREATE_2_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR and VK_PIPELINE_CREATE_2_DISALLOW_OPACITY_MICROMAP_BIT_ARM", 218}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11275", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also have that flag set", 167}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11276", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all libraries linked to this pipeline must also not have that flag set", 179}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11311", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be VK_NULL_HANDLE", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11312", 50}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 287}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11597", 50}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR, at least one of the following features must be enabled VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap", 235}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11599", 50}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature is not enabled or flags does not include VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR, flags must not include VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 263}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-11798", 50}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-12341", 50}, { {"flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-12361", 50}, { {"If flags includes VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, all libraries linked to this pipeline must also have that flag set", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-flags-12362", 50}, { {"If flags does not include VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, all libraries linked to this pipeline must also not have that flag set", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-03428", 51}, { {"The number of resources in layout accessible to each shader stage that is used by the pipeline must be less than or equal to VkPhysicalDeviceLimits::maxPerStageResources", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-07987", 51}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, a push constant range in layout must match the shader stage", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-07988", 51}, { {"If a resource variable is declared in a shader and layout is not VK_NULL_HANDLE, the corresponding descriptor set in layout must match the shader stage", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-07990", 51}, { {"If a resource variable is declared in a shader, layout is not VK_NULL_HANDLE, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in layout must match the descriptor type", 214}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-07991", 51}, { {"If a resource variable is declared in a shader as an array and layout is not VK_NULL_HANDLE, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 236}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-10069", 51}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, the block must be contained inside the push constant range in layout that matches the stage", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-12402", 51}, { {"If layout is not VK_NULL_HANDLE, it must not have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-layout-parameter", 55}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-maxPipelineRayRecursionDepth-03589", 73}, { {"maxPipelineRayRecursionDepth must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxRayRecursionDepth", 128}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pDynamicState-parameter", 62}, { {"If pDynamicState is not NULL, pDynamicState must be a valid pointer to a valid VkPipelineDynamicStateCreateInfo structure", 121}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pDynamicStates-03602", 59}, { {"Any element of the pDynamicStates member of pDynamicState must be VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pGroups-parameter", 56}, { {"If groupCount is not 0, pGroups must be a valid pointer to an array of groupCount valid VkRayTracingShaderGroupCreateInfoKHR structures", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraries-03591", 55}, { {"Each element of pLibraryInfo->pLibraries must have been created with the value of maxPipelineRayRecursionDepth equal to that in this pipeline", 141}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraries-11598", 55}, { {"If flags includes VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR, each element of pLibraryInfo->pLibraries must have been created with the VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR bit set", 211}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-03590", 57}, { {"If pLibraryInfo is not NULL and its libraryCount member is greater than 0, pLibraryInterface must not be NULL", 109}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-03592", 57}, { {"If pLibraryInfo is not NULL and layout is not VK_NULL_HANDLE, each element of pLibraryInfo->pLibraries must have been created with a layout that is compatible with the layout in this pipeline", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-03593", 57}, { {"If pLibraryInfo is not NULL, each element of its pLibraries member must have been created with values of the maxPipelineRayPayloadSize and maxPipelineRayHitAttributeSize members of pLibraryInterface equal to those in this pipeline", 230}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-03595", 57}, { {"If the VK_KHR_pipeline_library extension is not enabled, pLibraryInfo and pLibraryInterface must be NULL", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-07999", 57}, { {"If pLibraryInfo is NULL or its libraryCount is 0, stageCount must not be 0", 74}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInfo-parameter", 61}, { {"If pLibraryInfo is not NULL, pLibraryInfo must be a valid pointer to a valid VkPipelineLibraryCreateInfoKHR structure", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pLibraryInterface-parameter", 66}, { {"If pLibraryInterface is not NULL, pLibraryInterface must be a valid pointer to a valid VkRayTracingPipelineInterfaceCreateInfoKHR structure", 139}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pNext-pNext", 50}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineBinaryInfoKHR, VkPipelineCreateFlags2CreateInfo, VkPipelineCreationFeedbackCreateInfo, VkPipelineRobustnessCreateInfo, VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV, or VkValidationFeaturesEXT", 348}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pStages-03426", 52}, { {"The shader code for the entry points identified by pStages, and the rest of the state identified by this structure must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 199}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pStages-parameter", 56}, { {"If stageCount is not 0, pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pipelineCreationCacheControl-02878", 73}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT nor VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pipelineProtectedAccess-07368", 68}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-pipelineStageCreationFeedbackCount-06652", 79}, { {"If VkPipelineCreationFeedbackCreateInfo::pipelineStageCreationFeedbackCount is not 0, it must be equal to stageCount", 116}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-rayTracingPipelineShaderGroupHandleCaptureReplay-03599", 93}, { {"If VkPhysicalDeviceRayTracingPipelineFeaturesKHR::rayTracingPipelineShaderGroupHandleCaptureReplay is VK_TRUE and the pShaderGroupCaptureReplayHandle member of any element of pGroups is not NULL, flags must include VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 288}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-rayTraversalPrimitiveCulling-03596", 73}, { {"If the rayTraversalPrimitiveCulling feature is not enabled, flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 132}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-rayTraversalPrimitiveCulling-03597", 73}, { {"If the rayTraversalPrimitiveCulling feature is not enabled, flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR", 136}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_KHR", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-sType-unique", 51}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-stage-03425", 50}, { {"If flags does not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR, the stage member of at least one element of pStages, including those implicitly added by pLibraryInfo, must be VK_SHADER_STAGE_RAYGEN_BIT_KHR", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoKHR-stage-06899", 50}, { {"The stage value in all pStages elements must be one of VK_SHADER_STAGE_RAYGEN_BIT_KHR, VK_SHADER_STAGE_ANY_HIT_BIT_KHR, VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR, VK_SHADER_STAGE_MISS_BIT_KHR, VK_SHADER_STAGE_INTERSECTION_BIT_KHR, or VK_SHADER_STAGE_CALLABLE_BIT_KHR", 260}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-None-09497", 48}, { {"If the pNext chain does not include a VkPipelineCreateFlags2CreateInfo structure, flags must be a valid combination of VkPipelineCreateFlagBits values", 150}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-None-10391", 48}, { {"If a resource variables is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-None-11368", 48}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags does not include VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must not be VK_NULL_HANDLE", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-commonparent", 50}, { {"Both of basePipelineHandle, and layout that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 158}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-02904", 49}, { {"flags must not include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-02957", 49}, { {"flags must not include both VK_PIPELINE_CREATE_DEFER_COMPILE_BIT_NV and VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT at the same time", 145}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03456", 49}, { {"flags must not include VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03458", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03459", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03460", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03461", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03462", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03463", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-03588", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 96}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-04948", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07369", 49}, { {"flags must not include both VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT and VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 119}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07402", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_EXT", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07984", 49}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineIndex is -1, basePipelineHandle must be a valid ray tracing VkPipeline handle", 155}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07985", 49}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and basePipelineHandle is VK_NULL_HANDLE, basePipelineIndex must be a valid index into the calling command's pCreateInfos parameter", 193}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07986", 49}, { {"If flags contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, basePipelineIndex must be -1 or basePipelineHandle must be VK_NULL_HANDLE", 135}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-07998", 49}, { {"flags must not include VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-11008", 49}, { {"flags must not include VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT", 69}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-11311", 49}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, layout must be VK_NULL_HANDLE", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-11312", 49}, { {"If VkPipelineCreateFlags2CreateInfoKHR::flags includes VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 287}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-11596", 49}, { {"flags must not include any of the following bits VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 179}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-11798", 49}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_PIPELINE_CREATE_2_64_BIT_INDEXING_BIT_EXT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-flags-12341", 49}, { {"flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-groupCount-arraylength", 60}, { {"groupCount must be greater than 0", 33}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-03428", 50}, { {"The number of resources in layout accessible to each shader stage that is used by the pipeline must be less than or equal to VkPhysicalDeviceLimits::maxPerStageResources", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-07987", 50}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, a push constant range in layout must match the shader stage", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-07988", 50}, { {"If a resource variable is declared in a shader and layout is not VK_NULL_HANDLE, the corresponding descriptor set in layout must match the shader stage", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-07990", 50}, { {"If a resource variable is declared in a shader, layout is not VK_NULL_HANDLE, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in layout must match the descriptor type", 214}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-07991", 50}, { {"If a resource variable is declared in a shader as an array and layout is not VK_NULL_HANDLE, the corresponding descriptor binding used to create layout must have a descriptorCount that is greater than or equal to the length of the array", 236}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-10069", 50}, { {"If a push constant block is declared in a shader and layout is not VK_NULL_HANDLE, the block must be contained inside the push constant range in layout that matches the stage", 174}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-12402", 50}, { {"If layout is not VK_NULL_HANDLE, it must not have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-layout-parameter", 54}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-maxRecursionDepth-03457", 61}, { {"maxRecursionDepth must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxRecursionDepth", 105}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pGroups-parameter", 55}, { {"pGroups must be a valid pointer to an array of groupCount valid VkRayTracingShaderGroupCreateInfoNV structures", 110}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pNext-pNext", 49}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkPipelineCreateFlags2CreateInfo or VkPipelineCreationFeedbackCreateInfo", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pStages-03426", 51}, { {"The shader code for the entry points identified by pStages, and the rest of the state identified by this structure must adhere to the pipeline linking rules described in the Shader Interfaces chapter", 199}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pStages-parameter", 55}, { {"pStages must be a valid pointer to an array of stageCount valid VkPipelineShaderStageCreateInfo structures", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pipelineCreationCacheControl-02878", 72}, { {"If the pipelineCreationCacheControl feature is not enabled, flags must not include VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT nor VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 190}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pipelineProtectedAccess-07368", 67}, { {"If the pipelineProtectedAccess feature is not enabled, flags must not include VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT nor VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 169}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-pipelineStageCreationFeedbackCount-06651", 78}, { {"If VkPipelineCreationFeedbackCreateInfo::pipelineStageCreationFeedbackCount is not 0, it must be equal to stageCount", 116}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_NV", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-sType-unique", 50}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-stage-06232", 49}, { {"The stage member of at least one element of pStages must be VK_SHADER_STAGE_RAYGEN_BIT_KHR", 90}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-stage-06898", 49}, { {"The stage value in all pStages elements must be one of VK_SHADER_STAGE_RAYGEN_BIT_KHR, VK_SHADER_STAGE_ANY_HIT_BIT_KHR, VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR, VK_SHADER_STAGE_MISS_BIT_KHR, VK_SHADER_STAGE_INTERSECTION_BIT_KHR, or VK_SHADER_STAGE_CALLABLE_BIT_KHR", 260}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineCreateInfoNV-stageCount-arraylength", 60}, { {"stageCount must be greater than 0", 33}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineInterfaceCreateInfoKHR-maxPipelineRayHitAttributeSize-03605", 84}, { {"maxPipelineRayHitAttributeSize must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxRayHitAttributeSize", 132}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineInterfaceCreateInfoKHR-pNext-pNext", 59}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingPipelineInterfaceCreateInfoKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_INTERFACE_CREATE_INFO_KHR", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-anyHitShader-03479", 60}, { {"anyHitShader must be either VK_SHADER_UNUSED_KHR or a valid index into the list of shaders, formed by shaders in VkRayTracingPipelineCreateInfoKHR::pStages and the shaders imported from pipeline libraries, referring to a shader of VK_SHADER_STAGE_ANY_HIT_BIT_KHR", 262}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-closestHitShader-03478", 64}, { {"closestHitShader must be either VK_SHADER_UNUSED_KHR or a valid index into the list of shaders, formed by shaders in VkRayTracingPipelineCreateInfoKHR::pStages and the shaders imported from pipeline libraries, referring to a shader of VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR", 270}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-rayTracingPipelineShaderGroupHandleCaptureReplayMixed-03603", 101}, { {"If VkPhysicalDeviceRayTracingPipelineFeaturesKHR::rayTracingPipelineShaderGroupHandleCaptureReplayMixed is VK_FALSE then pShaderGroupCaptureReplayHandle must not be provided if it has not been provided on a previous call to ray tracing pipeline creation", 253}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-rayTracingPipelineShaderGroupHandleCaptureReplayMixed-03604", 101}, { {"If VkPhysicalDeviceRayTracingPipelineFeaturesKHR::rayTracingPipelineShaderGroupHandleCaptureReplayMixed is VK_FALSE then the caller must guarantee that no ray tracing pipeline creation commands with pShaderGroupCaptureReplayHandle provided execute simultaneously with ray tracing pipeline creation commands without pShaderGroupCaptureReplayHandle provided", 355}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_KHR", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-type-03474", 52}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_GENERAL_KHR then generalShader must be a valid index into the list of shaders, formed by shaders in VkRayTracingPipelineCreateInfoKHR::pStages and the shaders imported from pipeline libraries, referring to a shader of VK_SHADER_STAGE_RAYGEN_BIT_KHR, VK_SHADER_STAGE_MISS_BIT_KHR, or VK_SHADER_STAGE_CALLABLE_BIT_KHR", 359}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-type-03475", 52}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_GENERAL_KHR then closestHitShader, anyHitShader, and intersectionShader must be VK_SHADER_UNUSED_KHR", 144}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-type-03476", 52}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR then intersectionShader must be a valid index into the list of shaders, formed by shaders in VkRayTracingPipelineCreateInfoKHR::pStages and the shaders imported from pipeline libraries, referring to a shader of VK_SHADER_STAGE_INTERSECTION_BIT_KHR", 316}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-type-03477", 52}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR then intersectionShader must be VK_SHADER_UNUSED_KHR", 120}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoKHR-type-parameter", 56}, { {"type must be a valid VkRayTracingShaderGroupTypeKHR value", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-anyHitShader-02418", 59}, { {"anyHitShader must be either VK_SHADER_UNUSED_NV or a valid index into VkRayTracingPipelineCreateInfoNV::pStages referring to a shader of VK_SHADER_STAGE_ANY_HIT_BIT_NV", 167}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-closestHitShader-02417", 63}, { {"closestHitShader must be either VK_SHADER_UNUSED_NV or a valid index into VkRayTracingPipelineCreateInfoNV::pStages referring to a shader of VK_SHADER_STAGE_CLOSEST_HIT_BIT_NV", 175}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_NV", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-type-02413", 51}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_GENERAL_NV then generalShader must be a valid index into VkRayTracingPipelineCreateInfoNV::pStages referring to a shader of VK_SHADER_STAGE_RAYGEN_BIT_NV, VK_SHADER_STAGE_MISS_BIT_NV, or VK_SHADER_STAGE_CALLABLE_BIT_NV", 262}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-type-02414", 51}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_GENERAL_NV then closestHitShader, anyHitShader, and intersectionShader must be VK_SHADER_UNUSED_NV", 142}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-type-02415", 51}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_NV then intersectionShader must be a valid index into VkRayTracingPipelineCreateInfoNV::pStages referring to a shader of VK_SHADER_STAGE_INTERSECTION_BIT_NV", 221}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-type-02416", 51}, { {"If type is VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_NV then intersectionShader must be VK_SHADER_UNUSED_NV", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRayTracingShaderGroupCreateInfoNV-type-parameter", 55}, { {"type must be a valid VkRayTracingShaderGroupTypeKHR value", 57}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkRectLayerKHR-layer-01262", 31}, { {"layer must be less than the imageArrayLayers member of the VkSwapchainCreateInfoKHR structure passed to vkCreateSwapchainKHR", 124}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkRectLayerKHR-offset-04864", 32}, { {"The sum of offset and extent, after being transformed according to the preTransform member of the VkSwapchainCreateInfoKHR structure, must be no greater than the imageExtent member of the VkSwapchainCreateInfoKHR structure passed to vkCreateSwapchainKHR", 253}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseCapturedPipelineDataInfoKHR-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkReleaseCapturedPipelineDataInfoKHR-pipeline-09613", 56}, { {"pipeline must have been created with VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkReleaseCapturedPipelineDataInfoKHR-pipeline-09618", 56}, { {"pipeline must not have been used in a previous call to vkReleaseCapturedPipelineDataKHR", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkReleaseCapturedPipelineDataInfoKHR-pipeline-parameter", 60}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkReleaseCapturedPipelineDataInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_RELEASE_CAPTURED_PIPELINE_DATA_INFO_KHR", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-imageIndexCount-arraylength", 64}, { {"imageIndexCount must be greater than 0", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-pImageIndices-07785", 56}, { {"Each element of pImageIndices must be the index of a presentable image acquired from the swapchain specified by swapchain", 121}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-pImageIndices-07786", 56}, { {"All uses of presentable images identified by elements of pImageIndices must have completed execution", 100}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-pImageIndices-parameter", 60}, { {"pImageIndices must be a valid pointer to an array of imageIndexCount uint32_t values", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-pNext-pNext", 48}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_RELEASE_SWAPCHAIN_IMAGES_INFO_KHR", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkReleaseSwapchainImagesInfoKHR-swapchain-parameter", 56}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-pAttachments-03218", 55}, { {"Each element of pAttachments must only specify a single mip level", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-pAttachments-03219", 55}, { {"Each element of pAttachments must have been created with the identity swizzle", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-pAttachments-04114", 55}, { {"Each element of pAttachments must have been created with VkImageViewCreateInfo::viewType not equal to VK_IMAGE_VIEW_TYPE_3D", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-pAttachments-07010", 55}, { {"If multisampled-render-to-single-sampled is enabled for any subpass, all element of pAttachments which have a sample count equal to VK_SAMPLE_COUNT_1_BIT must have a format that supports the sample count specified in VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 281}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-pAttachments-parameter", 59}, { {"If attachmentCount is not 0, pAttachments must be a valid pointer to an array of attachmentCount valid VkImageView handles", 122}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassAttachmentBeginInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_ATTACHMENT_BEGIN_INFO", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-None-08996", 37}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.extent.width must be greater than 0", 166}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-None-08997", 37}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.extent.height must be greater than 0", 167}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-clearValueCount-00902", 48}, { {"clearValueCount must be greater than the largest attachment index in renderPass specifying a loadOp (or stencilLoadOp, if the attachment has a depth/stencil format) of VK_ATTACHMENT_LOAD_OP_CLEAR", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-clearValueCount-04962", 48}, { {"If clearValueCount is not 0, pClearValues must be a valid pointer to an array of clearValueCount VkClearValue unions", 116}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-commonparent", 39}, { {"Both of framebuffer, and renderPass must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-02780", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must have been created on the same VkDevice as framebuffer and renderPass", 310}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03207", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that did not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, and the pNext chain includes a VkRenderPassAttachmentBeginInfo structure, its attachmentCount must be zero", 235}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03208", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, the attachmentCount of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be equal to the value of VkFramebufferAttachmentsCreateInfo::attachmentImageInfoCount used to create framebuffer", 334}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03209", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkImageCreateInfo::flags equal to the flags member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 462}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03211", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView with a width equal to the width member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 414}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03212", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView with a height equal to the height member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 416}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03213", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkImageViewCreateInfo::subresourceRange.layerCount equal to the layerCount member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 493}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03214", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkImageFormatListCreateInfo::viewFormatCount equal to the viewFormatCount member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 492}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03215", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a set of elements in VkImageFormatListCreateInfo::pViewFormats equal to the set of elements in the pViewFormats member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 519}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-03216", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkImageViewCreateInfo::format equal to the corresponding value of VkAttachmentDescription::format in renderPass", 406}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-04627", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView with an inherited usage equal to the usage member of the corresponding element of VkFramebufferAttachmentsCreateInfo::pAttachmentImageInfos used to create framebuffer", 425}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-09047", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkImageCreateInfo::samples equal to the corresponding value of VkAttachmentDescription::samples in renderPass , or VK_SAMPLE_COUNT_1_BIT if renderPass was created with VkMultisampledRenderToSingleSampledInfoEXT structure in the pNext chain with multisampledRenderToSingleSampledEnable equal to VK_TRUE", 596}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-09353", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, and the nullColorAttachmentWithExternalFormatResolve is VK_FALSE, the format of the color attachment for each subpass that includes an external format image as a resolve attachment must have a format equal to the value of VkAndroidHardwareBufferFormatResolvePropertiesANDROID::colorAttachmentFormat as returned by a call to vkGetAndroidHardwareBufferPropertiesANDROID for the Android hardware buffer that was used to create the image view use as its resolve attachment", 590}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-09354", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each element of the pAttachments member of a VkRenderPassAttachmentBeginInfo structure included in the pNext chain must be a VkImageView of an image created with a value of VkExternalFormatANDROID::externalFormat equal to VkExternalFormatANDROID::externalFormat in the pNext chain of the corresponding VkAttachmentDescription2 structure used to create renderPass", 484}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-12328", 44}, { {"If framebuffer was created with a VkFramebufferCreateInfo::flags value that included VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, each pAttachments member of VkRenderPassAttachmentBeginInfo that is used as a resolve attachment by renderPass must not be bound to a VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 361}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-framebuffer-parameter", 48}, { {"framebuffer must be a valid VkFramebuffer handle", 48}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02850", 38}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.x must be greater than or equal to 0", 174}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02851", 38}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.y must be greater than or equal to 0", 174}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02852", 38}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.x + renderArea.extent.width must be less than or equal to VkFramebufferCreateInfo::width the framebuffer was created with", 259}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02853", 38}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.y + renderArea.extent.height must be less than or equal to VkFramebufferCreateInfo::height the framebuffer was created with", 261}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02856", 38}, { {"If the pNext chain contains VkDeviceGroupRenderPassBeginInfo, offset.x + extent.width of each element of pDeviceRenderAreas must be less than or equal to VkFramebufferCreateInfo::width the framebuffer was created with", 217}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02857", 38}, { {"If the pNext chain contains VkDeviceGroupRenderPassBeginInfo, offset.y + extent.height of each element of pDeviceRenderAreas must be less than or equal to VkFramebufferCreateInfo::height the framebuffer was created with", 219}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02869", 38}, { {"If the pNext chain includes VkRenderPassTransformBeginInfoQCOM, renderArea.offset must equal (0,0)", 98}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-02870", 38}, { {"If the pNext chain includes VkRenderPassTransformBeginInfoQCOM, renderArea.extent transformed by VkRenderPassTransformBeginInfoQCOM::transform must equal the framebuffer dimensions", 180}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-09539", 38}, { {"If the pNext chain contains a VkRenderPassStripeBeginInfoARM structure, the union of stripe areas defined by the elements of VkRenderPassStripeBeginInfoARM::pStripeInfos must cover the renderArea", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceGroupRenderPassBeginInfo, VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM, VkRenderPassAttachmentBeginInfo, VkRenderPassPerformanceCountersByRegionBeginInfoARM, VkRenderPassSampleLocationsBeginInfoEXT, VkRenderPassStripeBeginInfoARM, or VkRenderPassTransformBeginInfoQCOM", 415}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-perViewRenderAreaCount-07859", 55}, { {"If the perViewRenderAreaCount member of a VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM structure included in the pNext chain is not 0, then the multiviewPerViewRenderAreas feature must be enabled", 203}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-perViewRenderAreaCount-07860", 55}, { {"If the perViewRenderAreaCount member of a VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM structure included in the pNext chain is not 0, then renderArea must specify a render area that includes the union of all per view render areas", 238}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-renderPass-00904", 43}, { {"renderPass must be compatible with the renderPass member of the VkFramebufferCreateInfo structure specified when creating framebuffer", 133}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-renderPass-parameter", 47}, { {"renderPass must be a valid VkRenderPass handle", 46}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassBeginInfo-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-None-10915", 38}, { {"If any subpass preserves an attachment, there must be a subpass dependency from a prior subpass which uses or preserves that attachment", 135}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-attachment-00834", 44}, { {"If the attachment member of any element of pInputAttachments, pColorAttachments, pResolveAttachments or pDepthStencilAttachment, or any element of pPreserveAttachments in any element of pSubpasses is not VK_ATTACHMENT_UNUSED, then it must be less than attachmentCount", 267}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-flags-parameter", 43}, { {"flags must be a valid combination of VkRenderPassCreateFlagBits values", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-fragmentDensityMapAttachment-06471", 62}, { {"If the pNext chain includes a VkRenderPassFragmentDensityMapCreateInfoEXT structure and the fragmentDensityMapAttachment member is not VK_ATTACHMENT_UNUSED, then attachment must be less than attachmentCount", 206}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-fragmentDensityMapAttachment-10648", 62}, { {"If VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-fragmentDensityMapLayered-10828", 59}, { {"If the fragmentDensityMapLayered feature is not enabled, flags must not contain VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pAttachments-00836", 46}, { {"For any member of pAttachments with a loadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 242}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pAttachments-01566", 46}, { {"For any member of pAttachments with a loadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 209}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pAttachments-01567", 46}, { {"For any member of pAttachments with a stencilLoadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL", 216}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pAttachments-02511", 46}, { {"For any member of pAttachments with a stencilLoadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 249}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pAttachments-parameter", 50}, { {"If attachmentCount is not 0, pAttachments must be a valid pointer to an array of attachmentCount valid VkAttachmentDescription structures", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pDependencies-00837", 47}, { {"For each element of pDependencies, if the srcSubpass is not VK_SUBPASS_EXTERNAL, all stage flags included in the srcStageMask member of that dependency must be VK_PIPELINE_STAGE_ALL_COMMANDS_BIT or a pipeline stage supported by the pipeline identified by the pipelineBindPoint member of the source subpass", 305}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pDependencies-00838", 47}, { {"For each element of pDependencies, if the dstSubpass is not VK_SUBPASS_EXTERNAL, all stage flags included in the dstStageMask member of that dependency must be VK_PIPELINE_STAGE_ALL_COMMANDS_BIT or a pipeline stage supported by the pipeline identified by the pipelineBindPoint member of the destination subpass", 310}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pDependencies-06866", 47}, { {"For each element of pDependencies, if its srcSubpass is not VK_SUBPASS_EXTERNAL, it must be less than subpassCount", 114}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pDependencies-06867", 47}, { {"For each element of pDependencies, if its dstSubpass is not VK_SUBPASS_EXTERNAL, it must be less than subpassCount", 114}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pDependencies-parameter", 51}, { {"If dependencyCount is not 0, pDependencies must be a valid pointer to an array of dependencyCount valid VkSubpassDependency structures", 134}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01926", 39}, { {"If the pNext chain includes a VkRenderPassInputAttachmentAspectCreateInfo structure, the subpass member of each element of its pAspectReferences member must be less than subpassCount", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01927", 39}, { {"If the pNext chain includes a VkRenderPassInputAttachmentAspectCreateInfo structure, the inputAttachmentIndex member of each element of its pAspectReferences member must be less than the value of inputAttachmentCount in the element of pSubpasses identified by its subpass member", 278}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01928", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, and its subpassCount member is not zero, that member must be equal to the value of subpassCount", 168}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01929", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, if its dependencyCount member is not zero, it must be equal to dependencyCount", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01930", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, for each non-zero element of pViewOffsets, the srcSubpass and dstSubpass members of pDependencies at the same index must not be equal", 206}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-01963", 39}, { {"If the pNext chain includes a VkRenderPassInputAttachmentAspectCreateInfo structure, for any element of the pInputAttachments member of any element of pSubpasses where the attachment member is not VK_ATTACHMENT_UNUSED, the aspectMask member of the corresponding element of VkRenderPassInputAttachmentAspectCreateInfo::pAspectReferences must only include aspects that are present in images of the format specified by the element of pAttachments at attachment", 457}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-02512", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, for each element of pDependencies with a dependencyFlags member that does not include VK_DEPENDENCY_VIEW_LOCAL_BIT, the corresponding element of the pViewOffsets member of that VkRenderPassMultiviewCreateInfo instance must be 0", 300}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-02513", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, elements of its pViewMasks member must either all be 0, or all not be 0", 144}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-02514", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, and each element of its pViewMasks member is 0, the dependencyFlags member of each element of pDependencies must not include VK_DEPENDENCY_VIEW_LOCAL_BIT", 226}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-02515", 39}, { {"If the pNext chain includes a VkRenderPassMultiviewCreateInfo structure, and each element of its pViewMasks member is 0, its correlationMaskCount member must be 0", 162}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pNext-pNext", 39}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkRenderPassFragmentDensityMapCreateInfoEXT, VkRenderPassInputAttachmentAspectCreateInfo, VkRenderPassMultiviewCreateInfo, VkRenderPassTileShadingCreateInfoQCOM, or VkTileMemorySizeInfoQCOM", 320}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pResolveAttachments-10647", 53}, { {"If any element of pResolveAttachments of any element of pSubpasses references an attachment description with a format of VK_FORMAT_UNDEFINED, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 254}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-pSubpasses-parameter", 48}, { {"pSubpasses must be a valid pointer to an array of subpassCount valid VkSubpassDescription structures", 100}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO", 55}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo-subpassCount-arraylength", 52}, { {"subpassCount must be greater than 0", 35}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-None-03049", 39}, { {"If any two subpasses operate on attachments with overlapping ranges of the same VkDeviceMemory object, and at least one subpass writes to that area of VkDeviceMemory, a subpass dependency must be included (either directly or via some intermediate subpasses) between them", 270}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-None-10916", 39}, { {"If any subpass preserves an attachment, there must be a subpass dependency from a prior subpass which uses or preserves that attachment", 135}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-02525", 45}, { {"If the attachment member of any element of the pInputAttachments member of any element of pSubpasses is not VK_ATTACHMENT_UNUSED, the aspectMask member of that element of pInputAttachments must only include aspects that are present in images of the format specified by the element of pAttachments specified by attachment", 320}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-03050", 45}, { {"If the attachment member of any element of pInputAttachments, pColorAttachments, pResolveAttachments or pDepthStencilAttachment, or the attachment indexed by any element of pPreserveAttachments in any element of pSubpasses is bound to a range of a VkDeviceMemory object that overlaps with any other attachment in any subpass (including the same subpass), the VkAttachmentDescription2 structures describing them must include VK_ATTACHMENT_DESCRIPTION_MAY_ALIAS_BIT in flags", 472}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-03051", 45}, { {"If the attachment member of any element of pInputAttachments, pColorAttachments, pResolveAttachments or pDepthStencilAttachment, or any element of pPreserveAttachments in any element of pSubpasses is not VK_ATTACHMENT_UNUSED, then it must be less than attachmentCount", 267}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-06244", 45}, { {"If the attachment member of the pDepthStencilAttachment member of an element of pSubpasses is not VK_ATTACHMENT_UNUSED, the layout member of that same structure is either VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, and the pNext chain of that structure does not include a VkAttachmentReferenceStencilLayout structure, then the element of pAttachments with an index equal to attachment must not have a format that includes both depth and stencil components", 496}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-06245", 45}, { {"If the attachment member of the pDepthStencilAttachment member of an element of pSubpasses is not VK_ATTACHMENT_UNUSED and the layout member of that same structure is either VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, then the element of pAttachments with an index equal to attachment must have a format that includes only a stencil component", 387}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-attachment-06246", 45}, { {"If the attachment member of the pDepthStencilAttachment member of an element of pSubpasses is not VK_ATTACHMENT_UNUSED and the layout member of that same structure is either VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, then the element of pAttachments with an index equal to attachment must not have a format that includes only a stencil component", 387}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-dstSubpass-02527", 45}, { {"The dstSubpass member of each element of pDependencies must be less than subpassCount", 85}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-flags-04521", 40}, { {"If flags includes VK_RENDER_PASS_CREATE_TRANSFORM_BIT_QCOM, an element of pSubpasses includes an instance of VkFragmentShadingRateAttachmentInfoKHR in its pNext chain, and the pFragmentShadingRateAttachment member of that structure is not equal to NULL, the attachment member of pFragmentShadingRateAttachment must be VK_ATTACHMENT_UNUSED", 338}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-flags-parameter", 44}, { {"flags must be a valid combination of VkRenderPassCreateFlagBits values", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-fragmentDensityMapAttachment-06472", 63}, { {"If the pNext chain includes a VkRenderPassFragmentDensityMapCreateInfoEXT structure and the fragmentDensityMapAttachment member is not VK_ATTACHMENT_UNUSED, then attachment must be less than attachmentCount", 206}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-fragmentDensityMapAttachment-10651", 63}, { {"If VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-fragmentDensityMapLayered-10829", 60}, { {"If the fragmentDensityMapLayered feature is not enabled, flags must not contain VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-02522", 47}, { {"For any member of pAttachments with a loadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 303}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-02523", 47}, { {"For any member of pAttachments with a stencilLoadOp equal to VK_ATTACHMENT_LOAD_OP_CLEAR, the first use of that attachment must not specify a layout equal to VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL", 310}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-04585", 47}, { {"If any element of pAttachments is used as a fragment shading rate attachment in any subpass, it must not be used as any other attachment in the render pass", 155}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-04586", 47}, { {"If any element of pAttachments is used as a fragment shading rate attachment in any subpass, it must have an image format whose potential format features contain VK_FORMAT_FEATURE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 220}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-09387", 47}, { {"If any element of pAttachments is used as a fragment shading rate attachment, the loadOp for that attachment must not be VK_ATTACHMENT_LOAD_OP_CLEAR", 148}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pAttachments-parameter", 51}, { {"If attachmentCount is not 0, pAttachments must be a valid pointer to an array of attachmentCount valid VkAttachmentDescription2 structures", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pCorrelatedViewMasks-03056", 55}, { {"The set of bits included in any element of pCorrelatedViewMasks must not overlap with the set of bits included in any other element of pCorrelatedViewMasks", 155}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pCorrelatedViewMasks-parameter", 59}, { {"If correlatedViewMaskCount is not 0, pCorrelatedViewMasks must be a valid pointer to an array of correlatedViewMaskCount uint32_t values", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pDependencies-03054", 48}, { {"For each element of pDependencies, if the srcSubpass is not VK_SUBPASS_EXTERNAL, all stage flags included in the srcStageMask member of that dependency must be VK_PIPELINE_STAGE_ALL_COMMANDS_BIT or a pipeline stage supported by the pipeline identified by the pipelineBindPoint member of the source subpass", 305}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pDependencies-03055", 48}, { {"For each element of pDependencies, if the dstSubpass is not VK_SUBPASS_EXTERNAL, all stage flags included in the dstStageMask member of that dependency must be VK_PIPELINE_STAGE_ALL_COMMANDS_BIT or a pipeline stage supported by the pipeline identified by the pipelineBindPoint member of the destination subpass", 310}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pDependencies-03060", 48}, { {"For each element of pDependencies where its srcSubpass member equals its dstSubpass member, if the viewMask member of the corresponding element of pSubpasses includes more than one bit, its dependencyFlags member must include VK_DEPENDENCY_VIEW_LOCAL_BIT", 254}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pDependencies-parameter", 52}, { {"If dependencyCount is not 0, pDependencies must be a valid pointer to an array of dependencyCount valid VkSubpassDependency2 structures", 135}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pNext-pNext", 40}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkRenderPassCreationControlEXT, VkRenderPassCreationFeedbackCreateInfoEXT, VkRenderPassFragmentDensityMapCreateInfoEXT, VkRenderPassTileShadingCreateInfoQCOM, or VkTileMemorySizeInfoQCOM", 317}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pResolveAttachments-09331", 54}, { {"If any element of pResolveAttachments of any element of pSubpasses references an attachment description with a format of VK_FORMAT_UNDEFINED, VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment->attachment must be VK_ATTACHMENT_UNUSED", 256}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pResolveAttachments-10650", 54}, { {"If any element of pResolveAttachments of any element of pSubpasses references an attachment description with a format of VK_FORMAT_UNDEFINED, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 254}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pSubpasses-06473", 45}, { {"If the pSubpasses pNext chain includes a VkSubpassDescriptionDepthStencilResolve structure and the pDepthStencilResolveAttachment member is not NULL and does not have the value VK_ATTACHMENT_UNUSED, then attachment must be less than attachmentCount", 248}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-pSubpasses-parameter", 49}, { {"pSubpasses must be a valid pointer to an array of subpassCount valid VkSubpassDescription2 structures", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO_2", 57}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-sType-unique", 41}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-srcSubpass-02526", 45}, { {"The srcSubpass member of each element of pDependencies must be less than subpassCount", 85}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-subpassCount-arraylength", 53}, { {"subpassCount must be greater than 0", 35}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-viewMask-03057", 43}, { {"If the VkSubpassDescription2::viewMask member of all elements of pSubpasses is 0, correlatedViewMaskCount must be 0", 115}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-viewMask-03058", 43}, { {"The VkSubpassDescription2::viewMask member of all elements of pSubpasses must either all be 0, or all not be 0", 110}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreateInfo2-viewMask-03059", 43}, { {"If the VkSubpassDescription2::viewMask member of all elements of pSubpasses is 0, the dependencyFlags member of any element of pDependencies must not include VK_DEPENDENCY_VIEW_LOCAL_BIT", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreationControlEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_CONTROL_EXT", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreationFeedbackCreateInfoEXT-pRenderPassFeedback-parameter", 76}, { {"pRenderPassFeedback must be a valid pointer to a VkRenderPassCreationFeedbackInfoEXT structure", 94}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassCreationFeedbackCreateInfoEXT-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_FEEDBACK_CREATE_INFO_EXT", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-fragmentDensityMapAttachment-02548", 83}, { {"If fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, fragmentDensityMapAttachment must not be an element of VkSubpassDescription::pInputAttachments, VkSubpassDescription::pColorAttachments, VkSubpassDescription::pResolveAttachments, VkSubpassDescription::pDepthStencilAttachment, or VkSubpassDescription::pPreserveAttachments for any subpass", 349}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-fragmentDensityMapAttachment-02549", 83}, { {"If fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, layout must be equal to VK_IMAGE_LAYOUT_FRAGMENT_DENSITY_MAP_OPTIMAL_EXT, or VK_IMAGE_LAYOUT_GENERAL", 161}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-fragmentDensityMapAttachment-02550", 83}, { {"If fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, fragmentDensityMapAttachment must reference an attachment with a loadOp equal to VK_ATTACHMENT_LOAD_OP_LOAD or VK_ATTACHMENT_LOAD_OP_DONT_CARE", 203}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-fragmentDensityMapAttachment-02551", 83}, { {"If fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED, fragmentDensityMapAttachment must reference an attachment with a storeOp equal to VK_ATTACHMENT_STORE_OP_DONT_CARE", 175}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-fragmentDensityMapAttachment-parameter", 87}, { {"fragmentDensityMapAttachment must be a valid VkAttachmentReference structure", 76}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapCreateInfoEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_CREATE_INFO_EXT", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-fragmentDensityMapAttachment-06504", 86}, { {"If VkRenderPassFragmentDensityMapCreateInfoEXT::fragmentDensityMapAttachment is not VK_ATTACHMENT_UNUSED and was not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 222}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-fragmentDensityMapOffsets-06503", 83}, { {"If the fragmentDensityMapOffset feature is not enabled or fragment density map is not enabled in the render pass, fragmentDensityOffsetCount must equal 0", 153}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-fragmentDensityOffsetCount-06510", 84}, { {"If fragmentDensityOffsetCount is not 0 and multiview is enabled for the render pass, fragmentDensityOffsetCount must equal the layerCount that was specified in creating the fragment density map attachment view", 209}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-fragmentDensityOffsetCount-06511", 84}, { {"If fragmentDensityOffsetCount is not 0 and multiview is not enabled for the render pass, fragmentDensityOffsetCount must equal 1", 128}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pColorAttachments-06507", 75}, { {"If any used color attachments for the render pass were not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pDepthStencilAttachment-06505", 81}, { {"If the depth or stencil attachments for the render pass are used and were not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pFragmentDensityOffsets-10730", 81}, { {"Each element of pFragmentDensityOffsets must be identical for every vkCmdEndRendering2KHR call made in a render pass", 116}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pFragmentDensityOffsets-parameter", 85}, { {"If fragmentDensityOffsetCount is not 0, pFragmentDensityOffsets must be a valid pointer to an array of fragmentDensityOffsetCount VkOffset2D structures", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pInputAttachments-06506", 75}, { {"If any used input attachments for the render pass were not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pPreserveAttachments-06509", 78}, { {"If any used preserve attachments for the render pass were not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 167}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-pResolveAttachments-06508", 77}, { {"If any used resolve attachments for the render pass were not created with VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_EXT, fragmentDensityOffsetCount must equal 0", 166}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_OFFSET_END_INFO_EXT", 84}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-x-06512", 59}, { {"The x component of each element of pFragmentDensityOffsets must be an integer multiple of fragmentDensityOffsetGranularity.width", 128}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassFragmentDensityMapOffsetEndInfoEXT-y-06513", 59}, { {"The y component of each element of pFragmentDensityOffsets must be an integer multiple of fragmentDensityOffsetGranularity.height", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassInputAttachmentAspectCreateInfo-aspectReferenceCount-arraylength", 81}, { {"aspectReferenceCount must be greater than 0", 43}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassInputAttachmentAspectCreateInfo-pAspectReferences-parameter", 76}, { {"pAspectReferences must be a valid pointer to an array of aspectReferenceCount valid VkInputAttachmentAspectReference structures", 127}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassInputAttachmentAspectCreateInfo-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_INPUT_ATTACHMENT_ASPECT_CREATE_INFO", 79}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-multiview-06555", 52}, { {"If the multiview feature is not enabled, each element of pViewMasks must be 0", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-pCorrelationMasks-00841", 60}, { {"Each view index must not be set in more than one element of pCorrelationMasks", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-pCorrelationMasks-parameter", 64}, { {"If correlationMaskCount is not 0, pCorrelationMasks must be a valid pointer to an array of correlationMaskCount uint32_t values", 127}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-pViewMasks-06697", 53}, { {"The index of the most significant bit in each element of pViewMasks must be less than maxMultiviewViewCount", 107}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-pViewMasks-parameter", 57}, { {"If subpassCount is not 0, pViewMasks must be a valid pointer to an array of subpassCount uint32_t values", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-pViewOffsets-parameter", 59}, { {"If dependencyCount is not 0, pViewOffsets must be a valid pointer to an array of dependencyCount int32_t values", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassMultiviewCreateInfo-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_MULTIVIEW_CREATE_INFO", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-counterAddressCount-11815", 82}, { {"counterAddressCount must be equal to S x L, where S is the number of subpasses in the current render pass instance, and L is the number of layers in the current render pass instance or the index of the most significant bit of any view mask in the current render pass instance", 275}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-counterAddressCount-arraylength", 88}, { {"counterAddressCount must be greater than 0", 42}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-counterIndexCount-11818", 80}, { {"counterIndexCount must be less than or equal to VkPhysicalDevicePerformanceCountersByRegionPropertiesARM::maxPerRegionPerformanceCounters", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-counterIndexCount-arraylength", 86}, { {"counterIndexCount must be greater than 0", 40}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-pCounterAddresses-11816", 80}, { {"For each element of pCounterAddresses that is not 0, if the buffer from which it was queried is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 190}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-pCounterAddresses-11817", 80}, { {"For each element of pCounterAddresses[i], all device addresses between pCounterAddresses[i] and pCounterAddresses[i] plus N - 1, must be in the buffer device address range of the same buffer, where N is given by \\(N = \\mathbin{align}\\left(\\left\\lceil w/rsw \\right\\rceil \\times \\mathbin{align}\\left(c \\times sizeof(uint32\\_t), ra\\right), rsa\\right) \\times \\left\\lceil h/rsh \\right\\rceil\\), where w is the value of VkRenderingInfo::renderArea.extent.width, h is the value of VkRenderingInfo::renderArea.extent.height, rsw is the value of VkPhysicalDevicePerformanceCountersByRegionPropertiesARM::performanceCounterRegionSize.width, rsh is the value of VkPhysicalDevicePerformanceCountersByRegionPropertiesARM::performanceCounterRegionSize.height, c is the value of counterIndexCount, ra is the value of VkPhysicalDevicePerformanceCountersByRegionPropertiesARM::regionAlignment, and rsa is the value of VkPhysicalDevicePerformanceCountersByRegionPropertiesARM::rowStrideAlignment", 996}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-pCounterAddresses-parameter", 84}, { {"pCounterAddresses must be a valid pointer to an array of counterAddressCount VkDeviceAddress values", 99}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-pCounterIndices-parameter", 82}, { {"pCounterIndices must be a valid pointer to an array of counterIndexCount uint32_t values", 88}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassPerformanceCountersByRegionBeginInfoARM-sType-sType", 68}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_PERFORMANCE_COUNTERS_BY_REGION_BEGIN_INFO_ARM", 89}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassSampleLocationsBeginInfoEXT-pAttachmentInitialSampleLocations-parameter", 88}, { {"If attachmentInitialSampleLocationsCount is not 0, pAttachmentInitialSampleLocations must be a valid pointer to an array of attachmentInitialSampleLocationsCount valid VkAttachmentSampleLocationsEXT structures", 209}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassSampleLocationsBeginInfoEXT-pPostSubpassSampleLocations-parameter", 82}, { {"If postSubpassSampleLocationsCount is not 0, pPostSubpassSampleLocations must be a valid pointer to an array of postSubpassSampleLocationsCount valid VkSubpassSampleLocationsEXT structures", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassSampleLocationsBeginInfoEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_SAMPLE_LOCATIONS_BEGIN_INFO_EXT", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeBeginInfoARM-pStripeInfos-parameter", 58}, { {"pStripeInfos must be a valid pointer to an array of stripeInfoCount valid VkRenderPassStripeInfoARM structures", 110}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeBeginInfoARM-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_BEGIN_INFO_ARM", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeBeginInfoARM-stripeArea-09451", 52}, { {"The stripeArea defined by each element of pStripeInfos must not overlap the stripeArea of any other element", 107}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeBeginInfoARM-stripeInfoCount-09450", 57}, { {"stripeInfoCount must be less than or equal to VkPhysicalDeviceRenderPassStripedPropertiesARM::maxRenderPassStripes", 114}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeBeginInfoARM-stripeInfoCount-arraylength", 63}, { {"stripeInfoCount must be greater than 0", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_INFO_ARM", 59}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-stripeArea-09452", 47}, { {"stripeArea.offset.x must be a multiple of VkPhysicalDeviceRenderPassStripedPropertiesARM::renderPassStripeGranularity.width", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-stripeArea-09453", 47}, { {"stripeArea.extent.width must be a multiple of VkPhysicalDeviceRenderPassStripedPropertiesARM::renderPassStripeGranularity.width, or the sum of stripeArea.offset.x and stripeArea.extent.width must be equal to the renderArea.extent.width of the render pass instance", 263}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-stripeArea-09454", 47}, { {"stripeArea.offset.y must be a multiple of VkPhysicalDeviceRenderPassStripedPropertiesARM::renderPassStripeGranularity.height", 124}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeInfoARM-stripeArea-09455", 47}, { {"stripeArea.extent.height must be a multiple of VkPhysicalDeviceRenderPassStripedPropertiesARM::renderPassStripeGranularity.height, or the sum of stripeArea.offset.y and stripeArea.extent.height must be equal to the renderArea.extent.height of the render pass instance", 267}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassStripeSubmitInfoARM-pStripeSemaphoreInfos-parameter", 68}, { {"pStripeSemaphoreInfos must be a valid pointer to an array of stripeSemaphoreInfoCount valid VkSemaphoreSubmitInfo structures", 124}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkRenderPassStripeSubmitInfoARM-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_SUBMIT_INFO_ARM", 66}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkRenderPassStripeSubmitInfoARM-semaphore-09447", 52}, { {"The semaphore member of each element of pStripeSemaphoreInfos must have been created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 135}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkRenderPassStripeSubmitInfoARM-stripeSemaphoreInfoCount-arraylength", 73}, { {"stripeSemaphoreInfoCount must be greater than 0", 47}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkRenderPassSubpassFeedbackCreateInfoEXT-pSubpassFeedback-parameter", 72}, { {"pSubpassFeedback must be a valid pointer to a VkRenderPassSubpassFeedbackInfoEXT structure", 90}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassSubpassFeedbackCreateInfoEXT-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_SUBPASS_FEEDBACK_CREATE_INFO_EXT", 76}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-flags-10659", 54}, { {"If VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM is not included in flags or the tileShadingApron feature is not enabled, tileApronSize must be (0,0)", 147}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-flags-10660", 54}, { {"If VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM is not included in flags, or neither the tileShadingPerTileDispatch and tileShadingPerTileDraw features are enabled, flags must not include VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM", 242}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-flags-parameter", 58}, { {"flags must be a valid combination of VkTileShadingRenderPassFlagBitsQCOM values", 79}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_TILE_SHADING_CREATE_INFO_QCOM", 73}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-tileApronSize-10662", 62}, { {"tileApronSize.x must be less than or equal to maxApronSize", 58}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-tileApronSize-10663", 62}, { {"tileApronSize.y must be less than or equal to maxApronSize", 58}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-tileShading-10658", 60}, { {"If the tileShading feature is not enabled, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in flags", 116}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTileShadingCreateInfoQCOM-tileShadingAnisotropicApron-10661", 76}, { {"If the tileShadingAnisotropicApron feature is not enabled, tileApronSize.x and must be equal to tileApronSize.y", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTransformBeginInfoQCOM-flags-02872", 51}, { {"The renderpass must have been created with VkRenderPassCreateInfo::flags containing VK_RENDER_PASS_CREATE_TRANSFORM_BIT_QCOM", 124}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTransformBeginInfoQCOM-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_RENDER_PASS_TRANSFORM_BEGIN_INFO_QCOM", 69}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderPassTransformBeginInfoQCOM-transform-02871", 55}, { {"transform must be VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR", 180}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAreaInfo-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAreaInfo-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_AREA_INFO", 51}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentFlagsInfoKHR-flags-11755", 50}, { {"flags must not include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR if the dynamicRenderingLocalRead feature is not enabled", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentFlagsInfoKHR-flags-11756", 50}, { {"If flags includes VK_RENDERING_ATTACHMENT_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR, flags must not include VK_RENDERING_ATTACHMENT_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR", 169}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentFlagsInfoKHR-flags-11757", 50}, { {"If flags includes VK_RENDERING_ATTACHMENT_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RENDERING_ATTACHMENT_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveSrgbFormatSupportsTransferFunctionControl must be VK_TRUE", 214}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentFlagsInfoKHR-flags-parameter", 54}, { {"flags must be a valid combination of VkRenderingAttachmentFlagBitsKHR values", 76}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentFlagsInfoKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_FLAGS_INFO_KHR", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-None-12256", 41}, { {"If all of the following are true: imageView is not VK_NULL_HANDLE imageView has a sample count of VK_SAMPLE_COUNT_1_BIT the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with the multisampledRenderToSingleSampledEnable field equal to VK_TRUE then resolveMode must not be VK_RESOLVE_MODE_NONE", 340}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-clearValue-parameter", 51}, { {"clearValue must be a valid VkClearValue union", 45}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-commonparent", 43}, { {"Both of imageView, and resolveImageView that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-externalFormatResolve-09323", 58}, { {"If the externalFormatResolve feature is not enabled, resolveMode must not be VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID", 131}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageLayout-parameter", 52}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06129", 46}, { {"If imageView is not VK_NULL_HANDLE and has a non-integer color format, resolveMode must be VK_RESOLVE_MODE_NONE or VK_RESOLVE_MODE_CUSTOM_BIT_EXT or VK_RESOLVE_MODE_AVERAGE_BIT", 176}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06130", 46}, { {"If imageView is not VK_NULL_HANDLE and has an integer color format, resolveMode must be VK_RESOLVE_MODE_NONE or VK_RESOLVE_MODE_CUSTOM_BIT_EXT or VK_RESOLVE_MODE_SAMPLE_ZERO_BIT", 177}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06135", 46}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must not be VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_PREINITIALIZED", 276}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06136", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, VK_IMAGE_LAYOUT_ZERO_INITIALIZED_EXT, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_PREINITIALIZED", 376}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06137", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06138", 46}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must not be VK_IMAGE_LAYOUT_SHADING_RATE_OPTIMAL_NV", 99}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06139", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_SHADING_RATE_OPTIMAL_NV", 150}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06140", 46}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must not be VK_IMAGE_LAYOUT_FRAGMENT_DENSITY_MAP_OPTIMAL_EXT", 108}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06141", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_FRAGMENT_DENSITY_MAP_OPTIMAL_EXT", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06142", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 148}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06143", 46}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must not be VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR", 120}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06144", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR", 171}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06145", 46}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must not be VK_IMAGE_LAYOUT_PRESENT_SRC_KHR", 91}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06146", 46}, { {"If imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageLayout must not be VK_IMAGE_LAYOUT_PRESENT_SRC_KHR", 142}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06861", 46}, { {"If all of the following are true: imageView is not VK_NULL_HANDLE resolveMode is not VK_RESOLVE_MODE_NONE the pNext chain of VkRenderingInfo does not include a VkMultisampledRenderToSingleSampledInfoEXT structure with the multisampledRenderToSingleSampledEnable field equal to VK_TRUE imageView must not have a sample count of VK_SAMPLE_COUNT_1_BIT", 348}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06862", 46}, { {"If all of the following are true: imageView is not VK_NULL_HANDLE resolveMode is not VK_RESOLVE_MODE_NONE the pNext chain of VkRenderingInfo does not include a VkMultisampledRenderToSingleSampledInfoEXT structure with the multisampledRenderToSingleSampledEnable field equal to VK_TRUE resolveImageView must not be VK_NULL_HANDLE", 328}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06863", 46}, { {"If imageView is not VK_NULL_HANDLE, resolveMode is not VK_RESOLVE_MODE_NONE, the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with the multisampledRenderToSingleSampledEnable field equal to VK_TRUE, and imageView has a sample count of VK_SAMPLE_COUNT_1_BIT, resolveImageView must be VK_NULL_HANDLE", 347}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06864", 46}, { {"If imageView is not VK_NULL_HANDLE, resolveImageView is not VK_NULL_HANDLE, and resolveMode is not VK_RESOLVE_MODE_NONE, resolveImageView must have a sample count of VK_SAMPLE_COUNT_1_BIT", 187}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-06865", 46}, { {"If imageView is not VK_NULL_HANDLE, resolveImageView is not VK_NULL_HANDLE, and resolveMode is neither VK_RESOLVE_MODE_CUSTOM_BIT_EXT nor VK_RESOLVE_MODE_NONE, imageView and resolveImageView must have the same VkFormat", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-10780", 46}, { {"If feedback loop is enabled for the attachment identified by imageView, then imageView must have been created with a usage value including VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT, either VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, and either VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT", 355}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-12470", 46}, { {"If imageView is not VK_NULL_HANDLE, it must have been created with a VkImageViewCreateInfo::viewType not equal to VK_IMAGE_VIEW_TYPE_3D", 135}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-imageView-parameter", 50}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-loadOp-parameter", 47}, { {"loadOp must be a valid VkAttachmentLoadOp value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-nullColorAttachmentWithExternalFormatResolve-09325", 81}, { {"If the nullColorAttachmentWithExternalFormatResolve property is VK_TRUE and resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, resolveImageView must have been created with an image with a samples value of VK_SAMPLE_COUNT_1_BIT", 246}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-pNext-11752", 42}, { {"If the pNext chain includes a VkRenderingAttachmentFlagsInfoKHR structure, and flags includes VK_RENDERING_ATTACHMENT_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RENDERING_ATTACHMENT_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, imageView must have a format using sRGB encoding", 274}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-pNext-11753", 42}, { {"If the pNext chain includes a VkRenderingAttachmentFlagsInfoKHR structure, and flags includes VK_RENDERING_ATTACHMENT_RESOLVE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RENDERING_ATTACHMENT_RESOLVE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveMode must be equal to VK_RESOLVE_MODE_AVERAGE_BIT", 282}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-pNext-11754", 42}, { {"If the pNext chain includes a VkRenderingAttachmentFlagsInfoKHR structure, and flags includes VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, imageView must have an image that was created with the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 258}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-pNext-pNext", 42}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAttachmentFeedbackLoopInfoEXT or VkRenderingAttachmentFlagsInfoKHR", 199}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveImageLayout-parameter", 59}, { {"resolveImageLayout must be a valid VkImageLayout value", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveImageView-10728", 53}, { {"If resolveImageView is not VK_NULL_HANDLE, the underlying resource must not be bound to a VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveImageView-12471", 53}, { {"If resolveImageView is not VK_NULL_HANDLE, it must have been created with a VkImageViewCreateInfo::viewType not equal to VK_IMAGE_VIEW_TYPE_3D", 142}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveImageView-parameter", 57}, { {"If resolveImageView is not VK_NULL_HANDLE, resolveImageView must be a valid VkImageView handle", 94}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09324", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, resolveImageView must be a valid image view", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09326", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, resolveImageView must have been created with an external format specified by VkExternalFormatANDROID", 174}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09327", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, resolveImageView must have been created with a subresourceRange.layerCount of 1", 153}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09328", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID and nullColorAttachmentWithExternalFormatResolve is VK_TRUE, imageView must be VK_NULL_HANDLE", 166}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09329", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID and nullColorAttachmentWithExternalFormatResolve is VK_FALSE, imageView must be a valid VkImageView", 172}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-09330", 48}, { {"If resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID and nullColorAttachmentWithExternalFormatResolve is VK_FALSE, imageView must have a format equal to the value of VkAndroidHardwareBufferFormatResolvePropertiesANDROID::colorAttachmentFormat as returned by a call to vkGetAndroidHardwareBufferPropertiesANDROID for the Android hardware buffer that was used to create resolveImageView", 404}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-resolveMode-parameter", 52}, { {"If resolveMode is not 0, resolveMode must be a valid VkResolveModeFlagBits value", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO", 57}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentInfo-storeOp-parameter", 48}, { {"storeOp must be a valid VkAttachmentStoreOp value", 49}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingAttachmentLocationInfo-colorAttachmentCount-09514", 65}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingAttachmentLocationInfo-dynamicRenderingLocalRead-09512", 70}, { {"If the dynamicRenderingLocalRead feature is not enabled, and pColorAttachmentLocations is not NULL, each element must be the value of its index within the array", 160}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingAttachmentLocationInfo-pColorAttachmentLocations-09513", 70}, { {"Elements of pColorAttachmentLocations that are not VK_ATTACHMENT_UNUSED must each be unique", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingAttachmentLocationInfo-pColorAttachmentLocations-09515", 70}, { {"Each element of pColorAttachmentLocations must be less than maxColorAttachments", 79}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingAttachmentLocationInfo-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_LOCATION_INFO", 66}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingEndInfoKHR-pNext-pNext", 38}, { {"pNext must be NULL or a pointer to a valid instance of VkRenderPassFragmentDensityMapOffsetEndInfoEXT", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingEndInfoKHR-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_END_INFO_KHR", 54}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingEndInfoKHR-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-apiVersion-07908", 68}, { {"If the multiview feature is not enabled, VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and imageView is not VK_NULL_HANDLE, it must have a layerCount equal to 1", 178}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-imageLayout-parameter", 73}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-imageView-06157", 67}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_FRAGMENT_DENSITY_MAP_OPTIMAL_EXT", 131}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-imageView-06158", 67}, { {"If imageView is not VK_NULL_HANDLE, it must have been created with the VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT usage flag set", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-imageView-06159", 67}, { {"If imageView is not VK_NULL_HANDLE, it must not have been created with VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 105}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-imageView-parameter", 71}, { {"imageView must be a valid VkImageView handle", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentDensityMapAttachmentInfoEXT-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_INFO_EXT", 82}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageLayout-parameter", 74}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06147", 68}, { {"If imageView is not VK_NULL_HANDLE, layout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_FRAGMENT_SHADING_RATE_ATTACHMENT_OPTIMAL_KHR", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06148", 68}, { {"If imageView is not VK_NULL_HANDLE, it must have been created with the VK_IMAGE_USAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR usage flag set", 141}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06149", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.width must be a power of two value", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06150", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.width must be less than or equal to maxFragmentShadingRateAttachmentTexelSize.width", 150}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06151", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.width must be greater than or equal to minFragmentShadingRateAttachmentTexelSize.width", 153}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06152", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.height must be a power of two value", 102}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06153", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.height must be less than or equal to maxFragmentShadingRateAttachmentTexelSize.height", 152}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06154", 68}, { {"If imageView is not VK_NULL_HANDLE, shadingRateAttachmentTexelSize.height must be greater than or equal to minFragmentShadingRateAttachmentTexelSize.height", 155}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06155", 68}, { {"If imageView is not VK_NULL_HANDLE, the quotient of shadingRateAttachmentTexelSize.width and shadingRateAttachmentTexelSize.height must be less than or equal to maxFragmentShadingRateAttachmentTexelSizeAspectRatio", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-06156", 68}, { {"If imageView is not VK_NULL_HANDLE, the quotient of shadingRateAttachmentTexelSize.height and shadingRateAttachmentTexelSize.width must be less than or equal to maxFragmentShadingRateAttachmentTexelSizeAspectRatio", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-imageView-parameter", 72}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingFragmentShadingRateAttachmentInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-None-08994", 31}, { {"If VkDeviceGroupRenderPassBeginInfo::deviceRenderAreaCount is 0, renderArea.extent.width must be greater than 0", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-None-08995", 31}, { {"If VkDeviceGroupRenderPassBeginInfo::deviceRenderAreaCount is 0, renderArea.extent.height must be greater than 0", 112}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-None-09044", 31}, { {"Valid attachments specified by this structure must not be bound to memory locations that are bound to any other valid attachments specified by this structure", 157}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06087", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, that imageView must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06090", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, the layout member of that element of pColorAttachments must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 282}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06091", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, if the resolveMode member of that element of pColorAttachments is not VK_RESOLVE_MODE_NONE, its resolveImageLayout member must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL", 349}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06096", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, the layout member of that element of pColorAttachments must not be VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL", 303}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06097", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, if the resolveMode member of that element of pColorAttachments is not VK_RESOLVE_MODE_NONE, its resolveImageLayout member must not be VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL", 370}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06100", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, the layout member of that element of pColorAttachments must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 354}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06101", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, if the resolveMode member of that element of pColorAttachments is not VK_RESOLVE_MODE_NONE, its resolveImageLayout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 421}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-06106", 47}, { {"colorAttachmentCount must be less than or equal to VkPhysicalDeviceLimits::maxColorAttachments", 94}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-09320", 47}, { {"If colorAttachmentCount is not 1, the resolveMode member of any element of pColorAttachments must not be VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-09476", 47}, { {"If colorAttachmentCount is not 0 and there is an element of pColorAttachments with either its resolveMode member set to VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, or its imageView member not VK_NULL_HANDLE, its resolveMode member not set to VK_RESOLVE_MODE_NONE, and resolveImageView is not VK_NULL_HANDLE, then the resolveImageView member of that element of pColorAttachments must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 472}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-09479", 47}, { {"If colorAttachmentCount is not 0 and the imageView member of an element of pColorAttachments is not VK_NULL_HANDLE, that imageView must have been created with the identity swizzle", 179}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-colorAttachmentCount-09480", 47}, { {"If colorAttachmentCount is not 0, and there is an element of pColorAttachments with either its resolveMode member set to VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, or its imageView member not set to VK_NULL_HANDLE, its resolveMode member not set to VK_RESOLVE_MODE_NONE, and resolveImageView is not VK_NULL_HANDLE, then the resolveImageView member of that element of pColorAttachments must have been created with the identity swizzle", 446}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-flags-10012", 32}, { {"If flags includes VK_RENDERING_CONTENTS_INLINE_BIT_KHR then at least one of the following features must be enabled maintenance7 nestedCommandBuffer", 147}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-flags-10826", 32}, { {"If flags contains VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then layerCount must be less than or equal to maxFragmentDensityMapLayers", 142}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-flags-11514", 32}, { {"If flags contains VK_RENDERING_CUSTOM_RESOLVE_BIT_EXT or VK_RENDERING_FRAGMENT_REGION_BIT_EXT, then the customResolve feature must enabled", 138}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-flags-11516", 32}, { {"If flags contains VK_RENDERING_CUSTOM_RESOLVE_BIT_EXT, then for any element of pColorAttachments, pDepthAttachment, or pStencilAttachment, resolveMode must be VK_RESOLVE_MODE_CUSTOM_BIT_EXT or VK_RESOLVE_MODE_NONE", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-flags-parameter", 36}, { {"flags must be a valid combination of VkRenderingFlagBits values", 63}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-fragmentDensityMapLayered-10827", 52}, { {"If the fragmentDensityMapLayered feature is not enabled, flags must not contain VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06107", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, and the fragmentDensityMapNonSubsampledImages feature is not enabled, valid imageView and resolveImageView members of pDepthAttachment, pStencilAttachment, and each element of pColorAttachments must be a VkImageView created with VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 400}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06108", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, and viewMask is not 0, imageView must have a layerCount greater than the index of the most significant bit in viewMask", 255}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06109", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, and viewMask is 0, imageView must have a layerCount equal to 1", 199}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06116", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, it must not be equal to the imageView or resolveImageView member of pDepthAttachment, pStencilAttachment, or any element of pColorAttachments", 278}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06123", 36}, { {"If the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, and viewMask is 0, imageView must have a layerCount that is either equal to 1 or greater than or equal to VkRenderingInfo::layerCount", 271}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06124", 36}, { {"If the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, and viewMask is not 0, imageView must have a layerCount that is either equal to 1 or greater than the index of the most significant bit in viewMask", 285}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06125", 36}, { {"If the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, it must not be equal to the imageView or resolveImageView member of pDepthAttachment, pStencilAttachment, or any element of pColorAttachments", 279}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06126", 36}, { {"If the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, it must not be equal to the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain", 272}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06858", 36}, { {"If multisampled-render-to-single-sampled is enabled, then all attachments referenced by imageView members of pDepthAttachment, pStencilAttachment, and elements of pColorAttachments that are not VK_NULL_HANDLE must have a sample count that is either VK_SAMPLE_COUNT_1_BIT or equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 347}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-06859", 36}, { {"If multisampled-render-to-single-sampled is enabled, then all attachments referenced by imageView members of pDepthAttachment, pStencilAttachment, and elements of pColorAttachments that are not VK_NULL_HANDLE and have a sample count of VK_SAMPLE_COUNT_1_BIT must have been created with VK_IMAGE_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT in their VkImageCreateInfo::flags", 381}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-09429", 36}, { {"imageView members of elements of pColorAttachments that are not VK_NULL_HANDLE must have been created with the same sampleCount , if the multisampledRenderToSingleSampled feature is not enabled", 193}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-09485", 36}, { {"If the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, it must have been created with the identity swizzle", 189}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-09486", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, it must have been created with the identity swizzle", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-imageView-10643", 36}, { {"If the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 249}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-layerCount-07817", 37}, { {"layerCount must be less than or equal to maxFramebufferLayers", 61}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-multisampledRenderToSingleSampled-06857", 60}, { {"If none of the following are enabled: The VK_AMD_mixed_attachment_samples extension The VK_NV_framebuffer_mixed_samples extension The multisampledRenderToSingleSampled feature imageView members of pDepthAttachment, pStencilAttachment, and elements of pColorAttachments that are not VK_NULL_HANDLE must have been created with the same sampleCount", 345}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-multiview-06127", 36}, { {"If the multiview feature is not enabled, viewMask must be 0", 59}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pColorAttachments-11515", 44}, { {"For any element of pColorAttachments, pDepthAttachment, or pStencilAttachment, if resolveMode contains VK_RESOLVE_MODE_CUSTOM_BIT_EXT, then flags must contain VK_RENDERING_CUSTOM_RESOLVE_BIT_EXT", 194}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pColorAttachments-parameter", 48}, { {"If colorAttachmentCount is not 0, pColorAttachments must be a valid pointer to an array of colorAttachmentCount valid VkRenderingAttachmentInfo structures", 154}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06085", 43}, { {"If neither pDepthAttachment or pStencilAttachment are NULL and the imageView member of either structure is not VK_NULL_HANDLE, the imageView member of each structure must be the same", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06086", 43}, { {"If neither pDepthAttachment or pStencilAttachment are NULL, and the resolveMode member of each is not VK_RESOLVE_MODE_NONE, the resolveImageView member of each structure must be the same", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06088", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->imageView must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 205}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06092", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->layout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06093", 43}, { {"If pDepthAttachment is not NULL, pDepthAttachment->imageView is not VK_NULL_HANDLE, and pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pDepthAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 236}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06098", 43}, { {"If pDepthAttachment is not NULL, pDepthAttachment->imageView is not VK_NULL_HANDLE, and pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pDepthAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 254}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06102", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->resolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedDepthResolveModes", 221}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06104", 43}, { {"If pDepthAttachment or pStencilAttachment are both not NULL, pDepthAttachment->imageView and pStencilAttachment->imageView are both not VK_NULL_HANDLE, and VkPhysicalDeviceDepthStencilResolveProperties::independentResolveNone is VK_FALSE, the resolveMode of both structures must be the same value", 296}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06105", 43}, { {"If pDepthAttachment or pStencilAttachment are both not NULL, pDepthAttachment->imageView and pStencilAttachment->imageView are both not VK_NULL_HANDLE, VkPhysicalDeviceDepthStencilResolveProperties::independentResolve is VK_FALSE, and the resolveMode of neither structure is VK_RESOLVE_MODE_NONE, the resolveMode of both structures must be the same value", 354}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-06547", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->imageView must have been created with a format that includes a depth component", 183}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-07732", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->layout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-07733", 43}, { {"If pDepthAttachment is not NULL, pDepthAttachment->imageView is not VK_NULL_HANDLE, and pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pDepthAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 283}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-09318", 43}, { {"pDepthAttachment->resolveMode must not be VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID", 96}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-09477", 43}, { {"If pDepthAttachment is not NULL, pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pDepthAttachment->resolveImageView is not VK_NULL_HANDLE, then pDepthAttachment->resolveImageView must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 284}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-09481", 43}, { {"If pDepthAttachment is not NULL and pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->imageView must have been created with the identity swizzle", 163}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-09482", 43}, { {"If pDepthAttachment is not NULL, pDepthAttachment->imageView is not VK_NULL_HANDLE, pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pDepthAttachment->resolveImageView is not VK_NULL_HANDLE, pDepthAttachment->resolveImageView must have been created with the identity swizzle", 288}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pDepthAttachment-parameter", 47}, { {"If pDepthAttachment is not NULL, pDepthAttachment must be a valid pointer to a valid VkRenderingAttachmentInfo structure", 120}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06077", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.x must be greater than or equal to 0", 174}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06078", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, renderArea.offset.y must be greater than or equal to 0", 174}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06079", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, the width of the imageView member of each element of pColorAttachments, pDepthAttachment, or pStencilAttachment that is not VK_NULL_HANDLE must be greater than or equal to renderArea.offset.x + renderArea.extent.width", 337}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06080", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, the height of the imageView member of each element of pColorAttachments, pDepthAttachment, or pStencilAttachment that is not VK_NULL_HANDLE must be greater than or equal to renderArea.offset.y + renderArea.extent.height", 339}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06083", 32}, { {"If the pNext chain contains VkDeviceGroupRenderPassBeginInfo, the width of the imageView member of any element of pColorAttachments, pDepthAttachment, or pStencilAttachment that is not VK_NULL_HANDLE must be greater than or equal to the sum of the offset.x and extent.width members of each element of pDeviceRenderAreas", 319}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06084", 32}, { {"If the pNext chain contains VkDeviceGroupRenderPassBeginInfo, the height of the imageView member of any element of pColorAttachments, pDepthAttachment, or pStencilAttachment that is not VK_NULL_HANDLE must be greater than or equal to the sum of the offset.y and extent.height members of each element of pDeviceRenderAreas", 321}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06112", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0 and the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a width greater than or equal to \\(\\left\\lceil{\\frac{renderArea_{x}+renderArea_{width}}{maxFragmentDensityTexelSize_{width}}}\\right\\rceil\\)", 423}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06113", 32}, { {"If the pNext chain contains a VkDeviceGroupRenderPassBeginInfo structure, its deviceRenderAreaCount member is not 0, and the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a width greater than or equal to \\(\\left\\lceil{\\frac{pDeviceRenderAreas_{x}+pDeviceRenderAreas_{width}}{maxFragmentDensityTexelSize_{width}}}\\right\\rceil\\) for each element of pDeviceRenderAreas", 476}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06114", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0 and the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a height greater than or equal to \\(\\left\\lceil{\\frac{renderArea_{y}+renderArea_{height}}{maxFragmentDensityTexelSize_{height}}}\\right\\rceil\\)", 426}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06115", 32}, { {"If the pNext chain contains a VkDeviceGroupRenderPassBeginInfo structure, its deviceRenderAreaCount member is not 0, and the imageView member of a VkRenderingFragmentDensityMapAttachmentInfoEXT structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a height greater than or equal to \\(\\left\\lceil{\\frac{pDeviceRenderAreas_{y}+pDeviceRenderAreas_{height}}{maxFragmentDensityTexelSize_{height}}}\\right\\rceil\\) for each element of pDeviceRenderAreas", 479}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06119", 32}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, and the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a width greater than or equal to \\(\\left\\lceil{\\frac{renderArea_{x}+renderArea_{width}}{shadingRateAttachmentTexelSize_{width}}}\\right\\rceil\\)", 693}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06120", 32}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, the pNext chain contains a VkDeviceGroupRenderPassBeginInfo structure, its deviceRenderAreaCount member is not 0, and the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a width greater than or equal to \\(\\left\\lceil{\\frac{pDeviceRenderAreas_{x}+pDeviceRenderAreas_{width}}{shadingRateAttachmentTexelSize_{width}}}\\right\\rceil\\) for each element of pDeviceRenderAreas", 745}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06121", 32}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0 and the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a height greater than or equal to \\(\\left\\lceil{\\frac{renderArea_{y}+renderArea_{height}}{shadingRateAttachmentTexelSize_{height}}}\\right\\rceil\\)", 695}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-06122", 32}, { {"If the maintenance7 feature is not enabled or the robustFragmentShadingRateAttachmentAccess limit is VK_FALSE or the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure was created with VkImageSubresourceRange::baseMipLevel greater than 0, the pNext chain contains a VkDeviceGroupRenderPassBeginInfo structure, its deviceRenderAreaCount member is not 0, and the imageView member of a VkRenderingFragmentShadingRateAttachmentInfoKHR structure included in the pNext chain is not VK_NULL_HANDLE, imageView must have a height greater than or equal to \\(\\left\\lceil{\\frac{pDeviceRenderAreas_{y}+pDeviceRenderAreas_{height}}{shadingRateAttachmentTexelSize_{height}}}\\right\\rceil\\) for each element of pDeviceRenderAreas", 748}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-07815", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, the sum of renderArea.extent.width and renderArea.offset.x must be less than or equal to maxFramebufferWidth", 228}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-07816", 32}, { {"If the pNext chain does not contain VkDeviceGroupRenderPassBeginInfo or its deviceRenderAreaCount member is equal to 0, the sum of renderArea.extent.height and renderArea.offset.y must be less than or equal to maxFramebufferHeight", 230}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-09535", 32}, { {"If the pNext chain contains a VkRenderPassStripeBeginInfoARM structure, the union of stripe areas defined by the elements of VkRenderPassStripeBeginInfoARM::pStripeInfos must cover the renderArea", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pNext-pNext", 32}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceGroupRenderPassBeginInfo, VkMultisampledRenderToSingleSampledInfoEXT, VkMultiviewPerViewAttributesInfoNVX, VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM, VkRenderPassPerformanceCountersByRegionBeginInfoARM, VkRenderPassStripeBeginInfoARM, VkRenderPassTileShadingCreateInfoQCOM, VkRenderingFragmentDensityMapAttachmentInfoEXT, VkRenderingFragmentShadingRateAttachmentInfoKHR, or VkTileMemorySizeInfoQCOM", 548}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06089", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->imageView must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06094", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->layout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 170}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06095", 45}, { {"If pStencilAttachment is not NULL, pStencilAttachment->imageView is not VK_NULL_HANDLE, and pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pStencilAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL", 244}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06099", 45}, { {"If pStencilAttachment is not NULL, pStencilAttachment->imageView is not VK_NULL_HANDLE, and pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pStencilAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL", 262}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06103", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->resolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedStencilResolveModes", 229}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-06548", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->imageView must have been created with a format that includes a stencil aspect", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-07734", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->layout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-07735", 45}, { {"If pStencilAttachment is not NULL, pStencilAttachment->imageView is not VK_NULL_HANDLE, and pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, pStencilAttachment->resolveImageLayout must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL", 287}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-09319", 45}, { {"pStencilAttachment->resolveMode must not be VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID", 98}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-09478", 45}, { {"If pStencilAttachment is not NULL, pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pStencilAttachment->resolveImageView is not VK_NULL_HANDLE, then pStencilAttachment->resolveImageView must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-09483", 45}, { {"If pStencilAttachment is not NULL and pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->imageView must have been created with the identity swizzle", 169}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-09484", 45}, { {"If pStencilAttachment is not NULL, pStencilAttachment->imageView is not VK_NULL_HANDLE, pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pStencilAttachment->resolveImageView is not VK_NULL_HANDLE, pStencilAttachment->resolveImageView must have been created with the identity swizzle", 298}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-pStencilAttachment-parameter", 49}, { {"If pStencilAttachment is not NULL, pStencilAttachment must be a valid pointer to a valid VkRenderingAttachmentInfo structure", 124}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-perViewRenderAreaCount-07857", 49}, { {"If the perViewRenderAreaCount member of a VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM structure included in the pNext chain is not 0, then the multiviewPerViewRenderAreas feature must be enabled", 203}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-perViewRenderAreaCount-07858", 49}, { {"If the perViewRenderAreaCount member of a VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM structure included in the pNext chain is not 0, then renderArea must specify a render area that includes the union of all per view render areas", 238}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-resolveMode-09321", 38}, { {"If the resolveMode of any element of pColorAttachments is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView must be VK_NULL_HANDLE", 194}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-resolveMode-09322", 38}, { {"If the resolveMode of any element of pColorAttachments is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView must be VK_NULL_HANDLE", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-resolveMode-10644", 38}, { {"If the resolveMode of any element of pColorAttachments is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM must not be included in VkRenderPassTileShadingCreateInfoQCOM::flags", 226}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_INFO", 46}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-sType-unique", 33}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-06069", 35}, { {"If viewMask is 0, layerCount must not be 0", 42}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-06128", 35}, { {"The index of the most significant bit in viewMask must be less than maxMultiviewViewCount", 89}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-10859", 35}, { {"If viewMask is 0, each pColorAttachment->imageView and pColorAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than or equal to VkRenderingInfo::layerCount", 200}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-10860", 35}, { {"If viewMask is 0, each pDepthAttachment->imageView and pDepthAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than or equal to VkRenderingInfo::layerCount", 200}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-10861", 35}, { {"If viewMask is 0, each pStencilAttachment->imageView and pStencilAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than or equal to VkRenderingInfo::layerCount", 204}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-12403", 35}, { {"If viewMask is not 0, each pColorAttachment->imageView and pColorAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than the index of the most significant bit in viewMask", 214}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-12404", 35}, { {"If viewMask is not 0, each pDepthAttachment->imageView and pDepthAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than the index of the most significant bit in viewMask", 214}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInfo-viewMask-12405", 35}, { {"If viewMask is not 0, each pStencilAttachment->imageView and pStencilAttachment->resolveImageView that is not VK_NULL_HANDLE must have a layerCount that is greater than the index of the most significant bit in viewMask", 218}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-colorAttachmentCount-09525", 67}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-dynamicRenderingLocalRead-09519", 72}, { {"If the dynamicRenderingLocalRead feature is not enabled, and pColorAttachmentInputIndices is not NULL, each element must be VK_ATTACHMENT_UNUSED", 144}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-dynamicRenderingLocalRead-09520", 72}, { {"If the dynamicRenderingLocalRead feature is not enabled, pDepthInputAttachmentIndex must be a valid pointer to a value of VK_ATTACHMENT_UNUSED", 142}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-dynamicRenderingLocalRead-09521", 72}, { {"If the dynamicRenderingLocalRead feature is not enabled, pStencilInputAttachmentIndex must be a valid pointer to a value of VK_ATTACHMENT_UNUSED", 144}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pColorAttachmentInputIndices-09522", 75}, { {"Elements of pColorAttachmentInputIndices that are not VK_ATTACHMENT_UNUSED must each be unique", 94}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pColorAttachmentInputIndices-09523", 75}, { {"Elements of pColorAttachmentInputIndices that are not VK_ATTACHMENT_UNUSED must not take the same value as the content of pDepthInputAttachmentIndex", 148}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pColorAttachmentInputIndices-09524", 75}, { {"Elements of pColorAttachmentInputIndices that are not VK_ATTACHMENT_UNUSED must not take the same value as the content of pStencilInputAttachmentIndex", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pColorAttachmentInputIndices-parameter", 79}, { {"If colorAttachmentCount is not 0, and pColorAttachmentInputIndices is not NULL, pColorAttachmentInputIndices must be a valid pointer to an array of colorAttachmentCount uint32_t values", 184}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pDepthInputAttachmentIndex-12274", 73}, { {"Elements of pDepthInputAttachmentIndex, pStencilInputAttachmentIndex, and pColorAttachmentInputIndices that are not VK_ATTACHMENT_UNUSED must be less than maxPerStageDescriptorInputAttachments", 192}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pDepthInputAttachmentIndex-parameter", 77}, { {"If pDepthInputAttachmentIndex is not NULL, pDepthInputAttachmentIndex must be a valid pointer to a valid uint32_t value", 119}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-pStencilInputAttachmentIndex-parameter", 79}, { {"If pStencilInputAttachmentIndex is not NULL, pStencilInputAttachmentIndex must be a valid pointer to a valid uint32_t value", 123}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkRenderingInputAttachmentIndexInfo-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_RENDERING_INPUT_ATTACHMENT_INDEX_INFO", 69}, { "chapters/interfaces.html", 24} } },
        { {"VUID-VkResolveImageInfo2-commonparent", 37}, { {"Both of dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-00258", 39}, { {"If dstImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-00259", 39}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-00276", 39}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-00278", 39}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-02003", 39}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT if the maintenance10 feature is not enabled", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-02546", 39}, { {"dstImage and srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-06764", 39}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-06765", 39}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImage-parameter", 43}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImageLayout-00262", 45}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImageLayout-01401", 45}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstImageLayout-parameter", 49}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstOffset-00274", 40}, { {"For each element of pRegions, dstOffset.x and (extent.width + dstOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstOffset-00275", 40}, { {"For each element of pRegions, dstOffset.y and (extent.height + dstOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstOffset-00277", 40}, { {"For each element of pRegions, dstOffset.z and (extent.depth + dstOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstSubresource-01710", 45}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstSubresource-01712", 45}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-dstSubresource-11801", 45}, { {"dstSubresource.aspectMask for each element in pRegions must not specify an aspect which is not part of the image format of dstImage", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-linearColorAttachment-06519", 52}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, the format features of dstImage must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 197}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-maintenance10-11799", 44}, { {"If the maintenance10 feature is enabled, the format features of dstImage must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 174}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-pNext-10982", 36}, { {"If VkResolveImageModeInfoKHR is included in the pNext chain, flags includes VK_RESOLVE_IMAGE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RESOLVE_IMAGE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, then the format of srcImage and dstImage must use sRGB encoding", 241}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-pNext-pNext", 36}, { {"pNext must be NULL or a pointer to a valid instance of VkResolveImageModeInfoKHR", 80}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-pRegions-00255", 39}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-pRegions-parameter", 43}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageResolve2 structures", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-regionCount-arraylength", 48}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_RESOLVE_IMAGE_INFO_2", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-sType-unique", 37}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-00256", 39}, { {"If srcImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-00257", 39}, { {"srcImage must have a sample count equal to any valid sample count value other than VK_SAMPLE_COUNT_1_BIT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-00271", 39}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-00273", 39}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-01386", 39}, { {"srcImage and dstImage must have been created with the same image format", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-04446", 39}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.layerCount must be 1", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-04447", 39}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0 and dstSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-06762", 39}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-06763", 39}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10983", 39}, { {"If srcImage has a color format and VkResolveImageModeInfoKHR is included in the pNext chain, its resolveMode must not be VK_RESOLVE_MODE_NONE", 141}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10984", 39}, { {"If srcImage has a non-integer color format, and VkResolveImageModeInfoKHR is included in the pNext chain, its resolveMode must be VK_RESOLVE_MODE_AVERAGE_BIT", 157}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10985", 39}, { {"If srcImage has an integer color format, and VkResolveImageModeInfoKHR is included in the pNext chain, its resolveMode must be VK_RESOLVE_MODE_SAMPLE_ZERO_BIT", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10986", 39}, { {"If srcImage has a depth-stencil format, VkResolveImageModeInfoKHR must be included in the pNext chain", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10987", 39}, { {"If srcImage has a depth-stencil format, and a depth aspect is referenced by pRegions, VkResolveImageModeInfoKHR::resolveMode must not be VK_RESOLVE_MODE_NONE", 157}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10988", 39}, { {"If srcImage has a depth-stencil format, and a stencil aspect is referenced by pRegions, VkResolveImageModeInfoKHR::stencilResolveMode must not be VK_RESOLVE_MODE_NONE", 166}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10989", 39}, { {"If srcImage has a depth-stencil format, and a depth aspect is referenced by pRegions, VkResolveImageModeInfoKHR::resolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedDepthResolveModes", 229}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10990", 39}, { {"If srcImage has a depth-stencil format, and a stencil aspect is referenced by pRegions, VkResolveImageModeInfoKHR::stencilResolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedStencilResolveModes", 240}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10991", 39}, { {"If srcImage has a depth-stencil format, and both a depth aspect and stencil aspect is referenced by pRegions, and VkPhysicalDeviceDepthStencilResolveProperties::indepdendentResolve is VK_FALSE, VkResolveImageModeInfoKHR::resolveMode must be equal to VkResolveImageModeInfoKHR::stencilResolveMode", 295}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-10992", 39}, { {"If srcImage has a depth-stencil format containing both a depth aspect and stencil aspect, and VkPhysicalDeviceDepthStencilResolveProperties::indepdendentResolveNone is VK_FALSE, every element of pRegions must contain both depth and stencil aspects", 247}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImage-parameter", 43}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImageLayout-00260", 45}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImageLayout-01400", 45}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcImageLayout-parameter", 49}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcOffset-00269", 40}, { {"For each element of pRegions, srcOffset.x and (extent.width + srcOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcOffset-00270", 40}, { {"For each element of pRegions, srcOffset.y and (extent.height + srcOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcOffset-00272", 40}, { {"For each element of pRegions, srcOffset.z and (extent.depth + srcOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcSubresource-01709", 45}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcSubresource-01711", 45}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcSubresource-11800", 45}, { {"srcSubresource.aspectMask for each element in pRegions must not specify an aspect which is not part of the image format of srcImage", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageInfo2-srcSubresource-11802", 45}, { {"srcSubresource.aspectMask must equal dstSubresource.aspectMask for each element in pRegions", 91}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-flags-10995", 42}, { {"If flags includes VK_RESOLVE_IMAGE_SKIP_TRANSFER_FUNCTION_BIT_KHR, flags must not include VK_RESOLVE_IMAGE_ENABLE_TRANSFER_FUNCTION_BIT_KHR", 139}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-flags-10996", 42}, { {"If flags includes VK_RESOLVE_IMAGE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RESOLVE_IMAGE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveSrgbFormatSupportsTransferFunctionControl must be VK_TRUE", 184}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-flags-10997", 42}, { {"If flags includes VK_RESOLVE_IMAGE_SKIP_TRANSFER_FUNCTION_BIT_KHR or VK_RESOLVE_IMAGE_ENABLE_TRANSFER_FUNCTION_BIT_KHR, resolveMode must be equal to VK_RESOLVE_MODE_AVERAGE_BIT", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-flags-parameter", 46}, { {"flags must be a valid combination of VkResolveImageFlagBitsKHR values", 69}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-resolveMode-parameter", 52}, { {"If resolveMode is not 0, resolveMode must be a valid VkResolveModeFlagBits value", 80}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_RESOLVE_IMAGE_MODE_INFO_KHR", 59}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResolveImageModeInfoKHR-stencilResolveMode-parameter", 59}, { {"If stencilResolveMode is not 0, stencilResolveMode must be a valid VkResolveModeFlagBits value", 94}, { "chapters/copies.html", 20} } },
        { {"VUID-VkResourceDescriptorInfoEXT-None-11211", 43}, { {"If the nullDescriptor feature is not enabled, and type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE or VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, data->pImage must not be NULL", 157}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-None-11212", 43}, { {"If the nullDescriptor feature is not enabled, and type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, data->pTexelBuffer must not be NULL", 177}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-None-11213", 43}, { {"If the nullDescriptor feature is not enabled, and type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, data->pAddressRange must not be NULL", 318}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-None-11457", 43}, { {"If the nullDescriptor feature is not enabled, and type is VK_DESCRIPTOR_TYPE_TENSOR_ARM, data->pTensorARM must not be NULL", 122}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pAddressRange-parameter", 56}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, and if pAddressRange is not NULL, the pAddressRange member of data must be a valid pointer to a valid VkDeviceAddressRangeEXT structure", 266}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pImage-parameter", 49}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, and if pImage is not NULL, the pImage member of data must be a valid pointer to a valid VkImageDescriptorInfoEXT structure", 329}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pNext-11401", 44}, { {"If there is a VkDebugUtilsObjectNameInfoEXT structure in the pNext chain, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 119}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pNext-pNext", 44}, { {"pNext must be NULL or a pointer to a valid instance of VkDebugUtilsObjectNameInfoEXT", 84}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pTensorARM-parameter", 53}, { {"If type is VK_DESCRIPTOR_TYPE_TENSOR_ARM, and if pTensorARM is not NULL, the pTensorARM member of data must be a valid pointer to a valid VkTensorViewCreateInfoARM structure", 173}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-pTexelBuffer-parameter", 55}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, and if pTexelBuffer is not NULL, the pTexelBuffer member of data must be a valid pointer to a valid VkTexelBufferDescriptorInfoEXT structure", 235}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_RESOURCE_DESCRIPTOR_INFO_EXT", 60}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11210", 43}, { {"type must be one of VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_TENSOR_ARM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT", 548}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11422", 43}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, data.pImage->pView->image must not have been created with an imageType of VK_IMAGE_TYPE_3D", 229}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11424", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, data.pImage->pView->viewType is VK_IMAGE_VIEW_TYPE_2D, and the image2DViewOf3D feature is not enabled, data.pImage->pView->image must not have been created with an imageType of VK_IMAGE_TYPE_3D", 238}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11425", 43}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, data.pImage->pView->viewType is VK_IMAGE_VIEW_TYPE_2D, and the sampler2DViewOf3D feature is not enabled, data.pImage->pView->image must not have been created with an imageType of VK_IMAGE_TYPE_3D", 240}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11433", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER or VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, data.pAddressRange->size must not be 0", 121}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11454", 43}, { {"If type is one of VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV data->pAddressRange->address must be a multiple of 256", 220}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11458", 43}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE and data.pImage is not NULL, data.pImage->pView->image must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT usage flag set", 172}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11459", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE and data.pImage is not NULL, data.pImage->pView->image must have been created with the VK_IMAGE_USAGE_STORAGE_BIT usage flag set", 172}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11460", 43}, { {"If type is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT and data.pImage is not NULL, data.pImage->pView->image must have been created with the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 184}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11461", 43}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER and data.pAddressRange is not NULL, data.pAddressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 240}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11462", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER and data.pAddressRange is not NULL, data.pAddressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 240}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11463", 43}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER and data.pTexelBuffer is not NULL, data.pTexelBuffer->addressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT usage flag set", 264}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11464", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER and data.pTexelBuffer is not NULL, data.pTexelBuffer->addressRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT usage flag set", 264}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11467", 43}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV and data.pAddressRange is not NULL, data.pAddressRange->address must be an acceleration structure handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 249}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11468", 43}, { {"If type is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV and data.pAddressRange is not NULL, data.pAddressRange->size must be 0", 126}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11469", 43}, { {"If type is VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, data->pImage must not be NULL", 168}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11483", 43}, { {"If type is VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV or VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, and data.pAddressRange is not NULL, data.pAddressRange->address must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR", 321}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-11484", 43}, { {"If type is VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV or VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, data.pAddressRange is not NULL, and data.pAddressRange->size is not 0, data.pAddressRange must be a device address range allocated to the application from the buffer used to create the acceleration structure that data.pAddressRange->address was retrieved from, and within the buffer range bound to that acceleration structure", 443}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-12349", 43}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, data->pTexelBuffer->addressRange.address must be a multiple of the effective alignment requirement of data->pTexelBuffer->format as determined by minTexelBufferOffsetAlignment", 270}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-12350", 43}, { {"If type is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, data->pAddressRange->address must be a multiple of minUniformBufferOffsetAlignment", 128}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-12351", 43}, { {"If type is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, data->pAddressRange->address must be a multiple of minStorageBufferOffsetAlignment", 128}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkResourceDescriptorInfoEXT-type-parameter", 47}, { {"type must be a valid VkDescriptorType value", 43}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkSampleLocationsInfoEXT-pSampleLocations-parameter", 56}, { {"If sampleLocationsCount is not 0, pSampleLocations must be a valid pointer to an array of sampleLocationsCount VkSampleLocationEXT structures", 141}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkSampleLocationsInfoEXT-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLE_LOCATIONS_INFO_EXT", 57}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkSampleLocationsInfoEXT-sampleLocationsCount-01527", 56}, { {"sampleLocationsCount must equal sampleLocationsPerPixel x sampleLocationGridSize.width x sampleLocationGridSize.height", 118}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkSampleLocationsInfoEXT-sampleLocationsPerPixel-01526", 59}, { {"sampleLocationsPerPixel must be a valid VkSampleCountFlagBits value that is set in VkPhysicalDeviceSampleLocationsPropertiesEXT::sampleLocationSampleCounts", 155}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkSamplerBlockMatchWindowCreateInfoQCOM-WindowExtent-09210", 63}, { {"WindowExtent must not be larger than VkPhysicalDeviceImageProcessing2PropertiesQCOM::maxBlockMatchWindow", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerBlockMatchWindowCreateInfoQCOM-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_BLOCK_MATCH_WINDOW_CREATE_INFO_QCOM", 75}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerBlockMatchWindowCreateInfoQCOM-windowCompareMode-parameter", 72}, { {"windowCompareMode must be a valid VkBlockMatchWindowCompareModeQCOM value", 73}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerBorderColorComponentMappingCreateInfoEXT-borderColorSwizzle-06437", 79}, { {"The borderColorSwizzle feature must be enabled", 46}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerBorderColorComponentMappingCreateInfoEXT-components-parameter", 75}, { {"components must be a valid VkComponentMapping structure", 55}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerBorderColorComponentMappingCreateInfoEXT-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_BORDER_COLOR_COMPONENT_MAPPING_CREATE_INFO_EXT", 86}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCaptureDescriptorDataInfoEXT-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSamplerCaptureDescriptorDataInfoEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT", 72}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSamplerCaptureDescriptorDataInfoEXT-sampler-08087", 56}, { {"sampler must have been created with VK_SAMPLER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT set in VkSamplerCreateInfo::flags", 128}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSamplerCaptureDescriptorDataInfoEXT-sampler-parameter", 60}, { {"sampler must be a valid VkSampler handle", 40}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSamplerCreateInfo-None-01647", 35}, { {"If sampler Y'CBCR conversion is enabled and the pNext chain includes a VkSamplerReductionModeCreateInfo structure, then the sampler reduction mode must be VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE", 197}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-None-04012", 35}, { {"The maximum number of samplers with custom border colors which can be simultaneously created on a device is implementation-dependent and specified by the maxCustomBorderColorSamplers member of the VkPhysicalDeviceCustomBorderColorPropertiesEXT structure", 253}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeU-01078", 43}, { {"If any of addressModeU, addressModeV or addressModeW are VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER, borderColor must be a valid VkBorderColor value", 145}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeU-01079", 43}, { {"If the samplerMirrorClampToEdge feature is not enabled, and if the VK_KHR_sampler_mirror_clamp_to_edge extension is not enabled, addressModeU, addressModeV and addressModeW must not be VK_SAMPLER_ADDRESS_MODE_MIRROR_CLAMP_TO_EDGE", 229}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeU-01646", 43}, { {"If sampler Y'CBCR conversion is enabled, addressModeU, addressModeV, and addressModeW must be VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, anisotropyEnable must be VK_FALSE, and unnormalizedCoordinates must be VK_FALSE", 212}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeU-parameter", 47}, { {"addressModeU must be a valid VkSamplerAddressMode value", 55}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeV-parameter", 47}, { {"addressModeV must be a valid VkSamplerAddressMode value", 55}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-addressModeW-parameter", 47}, { {"addressModeW must be a valid VkSamplerAddressMode value", 55}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-anisotropyEnable-01070", 47}, { {"If the samplerAnisotropy feature is not enabled, anisotropyEnable must be VK_FALSE", 82}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-anisotropyEnable-01071", 47}, { {"If anisotropyEnable is VK_TRUE, maxAnisotropy must be between 1.0 and VkPhysicalDeviceLimits::maxSamplerAnisotropy, inclusive", 125}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-borderColor-04011", 42}, { {"If borderColor is one of VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT, then a VkSamplerCustomBorderColorCreateInfoEXT must be included in the pNext chain", 175}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-borderColor-04442", 42}, { {"If borderColor is one of VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT, and VkSamplerCustomBorderColorCreateInfoEXT::format is not VK_FORMAT_UNDEFINED, VkSamplerCustomBorderColorCreateInfoEXT::customBorderColor must be within the range of values representable in format", 290}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-compareEnable-01080", 44}, { {"If compareEnable is VK_TRUE, compareOp must be a valid VkCompareOp value", 72}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-compareEnable-01423", 44}, { {"If compareEnable is VK_TRUE, the reductionMode member of VkSamplerReductionModeCreateInfo must be VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE", 140}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-customBorderColors-04085", 49}, { {"If the customBorderColors feature is not enabled, borderColor must not be VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT", 140}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02574", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then minFilter and magFilter must be equal", 98}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02575", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then mipmapMode must be VK_SAMPLER_MIPMAP_MODE_NEAREST", 110}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02576", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then minLod and maxLod must be zero", 91}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02577", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then addressModeU and addressModeV must each be either VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE or VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER", 191}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02578", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then anisotropyEnable must be VK_FALSE", 94}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02579", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then compareEnable must be VK_FALSE", 91}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-02580", 36}, { {"If flags includes VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT, then unnormalizedCoordinates must be VK_FALSE", 101}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06964", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then minFilter and magFilter must be VK_FILTER_NEAREST", 117}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06965", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then mipmapMode must be VK_SAMPLER_MIPMAP_MODE_NEAREST", 117}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06966", 36}, { {"[If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then minLod and maxLod must be zero", 99}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06967", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then addressModeU and addressModeV must each be either VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE or VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER", 198}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06968", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, and if addressModeU or addressModeV is VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER, then borderColor must be VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK", 207}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06969", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then anisotropyEnable must be VK_FALSE", 101}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-06970", 36}, { {"If flags includes VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM, then compareEnable must be VK_FALSE", 98}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-08110", 36}, { {"If flags includes VK_SAMPLER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT, the descriptorBufferCaptureReplay feature must be enabled", 135}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-flags-parameter", 40}, { {"flags must be a valid combination of VkSamplerCreateFlagBits values", 67}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-magFilter-01081", 40}, { {"If either magFilter or minFilter is VK_FILTER_CUBIC_EXT, anisotropyEnable must be VK_FALSE", 90}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-magFilter-07911", 40}, { {"If the VK_EXT_filter_cubic extension is not enabled and either magFilter or minFilter is VK_FILTER_CUBIC_IMG, the reductionMode member of VkSamplerReductionModeCreateInfo must be VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE", 221}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-magFilter-parameter", 44}, { {"magFilter must be a valid VkFilter value", 40}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-maxLod-01973", 37}, { {"maxLod must be greater than or equal to minLod", 46}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-minFilter-01645", 40}, { {"If sampler Y'CBCR conversion is enabled and the potential format features of the sampler Y'CBCR conversion do not support VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_SEPARATE_RECONSTRUCTION_FILTER_BIT, minFilter and magFilter must be equal to the sampler Y'CBCR conversion's chromaFilter", 292}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-minFilter-parameter", 44}, { {"minFilter must be a valid VkFilter value", 40}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-mipLodBias-01069", 41}, { {"The absolute value of mipLodBias must be less than or equal to VkPhysicalDeviceLimits::maxSamplerLodBias", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-mipmapMode-parameter", 45}, { {"mipmapMode must be a valid VkSamplerMipmapMode value", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-nonSeamlessCubeMap-06788", 49}, { {"If the nonSeamlessCubeMap feature is not enabled, flags must not include VK_SAMPLER_CREATE_NON_SEAMLESS_CUBE_MAP_BIT_EXT", 120}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-pNext-06726", 36}, { {"If the samplerFilterMinmax feature is not enabled and the pNext chain includes a VkSamplerReductionModeCreateInfo structure, then the sampler reduction mode must be VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE", 207}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-pNext-08111", 36}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_SAMPLER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT", 161}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-pNext-pNext", 36}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDebugUtilsObjectNameInfoEXT, VkOpaqueCaptureDescriptorDataCreateInfoEXT, VkSamplerBlockMatchWindowCreateInfoQCOM, VkSamplerBorderColorComponentMappingCreateInfoEXT, VkSamplerCubicWeightsCreateInfoQCOM, VkSamplerCustomBorderColorCreateInfoEXT, VkSamplerCustomBorderColorIndexCreateInfoEXT, VkSamplerReductionModeCreateInfo, or VkSamplerYcbcrConversionInfo", 487}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO", 51}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-sType-unique", 37}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-samplerMipLodBias-04467", 48}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::samplerMipLodBias is VK_FALSE, mipLodBias must be zero", 159}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01072", 54}, { {"If unnormalizedCoordinates is VK_TRUE, minFilter and magFilter must be equal", 76}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01073", 54}, { {"If unnormalizedCoordinates is VK_TRUE, mipmapMode must be VK_SAMPLER_MIPMAP_MODE_NEAREST", 88}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01074", 54}, { {"If unnormalizedCoordinates is VK_TRUE, minLod and maxLod must be zero", 69}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01075", 54}, { {"If unnormalizedCoordinates is VK_TRUE, addressModeU and addressModeV must each be either VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE or VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER", 169}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01076", 54}, { {"If unnormalizedCoordinates is VK_TRUE, anisotropyEnable must be VK_FALSE", 72}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCreateInfo-unnormalizedCoordinates-01077", 54}, { {"If unnormalizedCoordinates is VK_TRUE, compareEnable must be VK_FALSE", 69}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCubicWeightsCreateInfoQCOM-cubicWeights-parameter", 63}, { {"cubicWeights must be a valid VkCubicFilterWeightsQCOM value", 59}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCubicWeightsCreateInfoQCOM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_CUBIC_WEIGHTS_CREATE_INFO_QCOM", 70}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorCreateInfoEXT-format-04014", 57}, { {"If the customBorderColorWithoutFormat feature is not enabled then format must not be VK_FORMAT_UNDEFINED", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorCreateInfoEXT-format-04015", 57}, { {"If the sampler is used to sample an image view of VK_FORMAT_B4G4R4A4_UNORM_PACK16, VK_FORMAT_B5G6R5_UNORM_PACK16, VK_FORMAT_A1B5G5R5_UNORM_PACK16, or VK_FORMAT_B5G5R5A1_UNORM_PACK16 format then format must not be VK_FORMAT_UNDEFINED", 232}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorCreateInfoEXT-format-07605", 57}, { {"If format is not VK_FORMAT_UNDEFINED and format is not a depth/stencil format then the VkSamplerCreateInfo::borderColor type must match the sampled type of the provided format, as shown in the SPIR-V Type column of the Interpretation of Numeric Format table", 257}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorCreateInfoEXT-format-parameter", 61}, { {"format must be a valid VkFormat value", 37}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorCreateInfoEXT-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_CREATE_INFO_EXT", 75}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorIndexCreateInfoEXT-index-11289", 61}, { {"index must be less than maxCustomBorderColorSamplers", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerCustomBorderColorIndexCreateInfoEXT-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_INDEX_CREATE_INFO_EXT", 81}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerReductionModeCreateInfo-reductionMode-parameter", 61}, { {"reductionMode must be a valid VkSamplerReductionMode value", 58}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerReductionModeCreateInfo-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_REDUCTION_MODE_CREATE_INFO", 66}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-chromaFilter-01657", 58}, { {"If the potential format features of the sampler Y'CBCR conversion do not support VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_LINEAR_FILTER_BIT, chromaFilter must not be VK_FILTER_LINEAR", 190}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-chromaFilter-parameter", 62}, { {"chromaFilter must be a valid VkFilter value", 43}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-02581", 56}, { {"If the format has a _422 or _420 suffix, then components.g must be the identity swizzle", 87}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-02582", 56}, { {"If the format has a _422 or _420 suffix, then components.a must be the identity swizzle, VK_COMPONENT_SWIZZLE_ONE, or VK_COMPONENT_SWIZZLE_ZERO", 143}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-02583", 56}, { {"If the format has a _422 or _420 suffix, then components.r must be the identity swizzle or VK_COMPONENT_SWIZZLE_B", 113}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-02584", 56}, { {"If the format has a _422 or _420 suffix, then components.b must be the identity swizzle or VK_COMPONENT_SWIZZLE_R", 113}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-02585", 56}, { {"If the format has a _422 or _420 suffix, and if either components.r or components.b is the identity swizzle, both values must be the identity swizzle", 149}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-components-parameter", 60}, { {"components must be a valid VkComponentMapping structure", 55}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-forceExplicitReconstruction-01656", 73}, { {"If the potential format features of the sampler Y'CBCR conversion do not support VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_CHROMA_RECONSTRUCTION_EXPLICIT_FORCEABLE_BIT forceExplicitReconstruction must be VK_FALSE", 219}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-format-01650", 52}, { {"The potential format features of the sampler Y'CBCR conversion must support VK_FORMAT_FEATURE_MIDPOINT_CHROMA_SAMPLES_BIT or VK_FORMAT_FEATURE_COSITED_CHROMA_SAMPLES_BIT", 169}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-format-01904", 52}, { {"If an external format conversion is being created, format must be VK_FORMAT_UNDEFINED", 85}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-format-04061", 52}, { {"If an external format conversion is not being created, format must represent unsigned normalized values (i.e. the format must be a UNORM format)", 144}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-format-parameter", 56}, { {"format must be a valid VkFormat value", 37}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-pNext-09207", 51}, { {"If the pNext chain includes a VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM structure, and if the ycbcrDegamma feature is not enabled, then VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM::enableYDegamma must be VK_FALSE", 228}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-pNext-09208", 51}, { {"If the pNext chain includes a VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM structure, and if the ycbcrDegamma feature is not enabled, then VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM::enableCbCrDegamma must be VK_FALSE", 231}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-pNext-09209", 51}, { {"If the pNext chain includes a VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM structure, format must be a format with 8-bit R, G, and B components", 149}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-pNext-pNext", 51}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExternalFormatANDROID, VkExternalFormatOHOS, VkExternalFormatQNX, or VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM", 252}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_CREATE_INFO", 68}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-sType-unique", 52}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-xChromaOffset-01651", 59}, { {"If the potential format features of the sampler Y'CBCR conversion do not support VK_FORMAT_FEATURE_COSITED_CHROMA_SAMPLES_BIT, xChromaOffset and yChromaOffset must not be VK_CHROMA_LOCATION_COSITED_EVEN if the corresponding components are downsampled", 250}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-xChromaOffset-01652", 59}, { {"If the potential format features of the sampler Y'CBCR conversion do not support VK_FORMAT_FEATURE_MIDPOINT_CHROMA_SAMPLES_BIT, xChromaOffset and yChromaOffset must not be VK_CHROMA_LOCATION_MIDPOINT if the corresponding components are downsampled", 247}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-xChromaOffset-parameter", 63}, { {"xChromaOffset must be a valid VkChromaLocation value", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-yChromaOffset-parameter", 63}, { {"yChromaOffset must be a valid VkChromaLocation value", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-ycbcrModel-01655", 56}, { {"If ycbcrModel is not VK_SAMPLER_YCBCR_MODEL_CONVERSION_RGB_IDENTITY, then components.r, components.g, and components.b must correspond to components of the format; that is, components.r, components.g, and components.b must not be VK_COMPONENT_SWIZZLE_ZERO or VK_COMPONENT_SWIZZLE_ONE, and must not correspond to a component containing zero or one as a consequence of component substitution", 389}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-ycbcrModel-parameter", 60}, { {"ycbcrModel must be a valid VkSamplerYcbcrModelConversion value", 62}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-ycbcrRange-02748", 56}, { {"If ycbcrRange is VK_SAMPLER_YCBCR_RANGE_ITU_NARROW, then the R, G, and B components obtained by applying the component swizzle to format must each have a bit-depth greater than or equal to 8", 190}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionCreateInfo-ycbcrRange-parameter", 60}, { {"ycbcrRange must be a valid VkSamplerYcbcrRange value", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionImageFormatProperties-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_IMAGE_FORMAT_PROPERTIES", 80}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkSamplerYcbcrConversionInfo-conversion-parameter", 54}, { {"conversion must be a valid VkSamplerYcbcrConversion handle", 58}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionInfo-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_INFO", 61}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM-sType-sType", 67}, { {"sType must be VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_YCBCR_DEGAMMA_CREATE_INFO_QCOM", 87}, { "chapters/samplers.html", 22} } },
        { {"VUID-VkScreenBufferFormatPropertiesQNX-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_SCREEN_BUFFER_FORMAT_PROPERTIES_QNX", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkScreenBufferPropertiesQNX-pNext-pNext", 44}, { {"pNext must be NULL or a pointer to a valid instance of VkScreenBufferFormatPropertiesQNX", 88}, { "chapters/memory.html", 20} } },
        { {"VUID-VkScreenBufferPropertiesQNX-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_SCREEN_BUFFER_PROPERTIES_QNX", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-VkScreenBufferPropertiesQNX-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-VkScreenSurfaceCreateInfoQNX-context-04741", 47}, { {"context must point to a valid QNX Screen struct _screen_context", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkScreenSurfaceCreateInfoQNX-flags-zerobitmask", 51}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkScreenSurfaceCreateInfoQNX-pNext-pNext", 45}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkScreenSurfaceCreateInfoQNX-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_SCREEN_SURFACE_CREATE_INFO_QNX", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkScreenSurfaceCreateInfoQNX-window-04742", 46}, { {"window must point to a valid QNX Screen struct _screen_window", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSemaphoreCreateInfo-flags-zerobitmask", 44}, { {"flags must be 0", 15}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreCreateInfo-pNext-06789", 38}, { {"If the pNext chain includes a VkExportMetalObjectCreateInfoEXT structure, its exportObjectType member must be VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreCreateInfo-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExportMetalObjectCreateInfoEXT, VkExportSemaphoreCreateInfo, VkExportSemaphoreWin32HandleInfoKHR, VkImportMetalSharedEventInfoEXT, VkQueryLowLatencySupportNV, or VkSemaphoreTypeCreateInfo", 320}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreCreateInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreCreateInfo-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique, with the exception of structures of type VkExportMetalObjectCreateInfoEXT", 142}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-01132", 45}, { {"handleType must have been included in VkExportSemaphoreCreateInfo::handleTypes when semaphore's current payload was created", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-01134", 45}, { {"If handleType refers to a handle type with copy payload transference semantics, as defined below in Importing Semaphore Payloads, there must be no queue waiting on semaphore", 173}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-01135", 45}, { {"If handleType refers to a handle type with copy payload transference semantics, semaphore must be signaled, or have an associated semaphore signal operation pending execution", 174}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-01136", 45}, { {"handleType must be defined as a POSIX file descriptor handle", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-03253", 45}, { {"If handleType refers to a handle type with copy payload transference semantics, semaphore must have been created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-03254", 45}, { {"If handleType refers to a handle type with copy payload transference semantics, semaphore must have an associated semaphore signal operation that has been submitted for execution and any semaphore signal operations on which it depends must have also been submitted for execution", 278}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-handleType-parameter", 49}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR", 57}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-semaphore-01133", 44}, { {"semaphore must not currently have its payload replaced by an imported payload as described below in Importing Semaphore Payloads unless that imported payload's handle type was included in VkExternalSemaphoreProperties::exportFromImportedHandleTypes for handleType", 263}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetFdInfoKHR-semaphore-parameter", 48}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-01126", 54}, { {"handleType must have been included in VkExportSemaphoreCreateInfo::handleTypes when the semaphore's current payload was created", 127}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-01127", 54}, { {"If handleType is defined as an NT handle, vkGetSemaphoreWin32HandleKHR must be called no more than once for each valid unique combination of semaphore and handleType", 165}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-01129", 54}, { {"If handleType refers to a handle type with copy payload transference semantics, as defined below in Importing Semaphore Payloads, there must be no queue waiting on semaphore", 173}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-01130", 54}, { {"If handleType refers to a handle type with copy payload transference semantics, semaphore must be signaled, or have an associated semaphore signal operation pending execution", 174}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-01131", 54}, { {"handleType must be defined as an NT handle or a global share handle", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-handleType-parameter", 58}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-semaphore-01128", 53}, { {"semaphore must not currently have its payload replaced by an imported payload as described below in Importing Semaphore Payloads unless that imported payload's handle type was included in VkExternalSemaphoreProperties::exportFromImportedHandleTypes for handleType", 263}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetWin32HandleInfoKHR-semaphore-parameter", 57}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-handleType-04758", 59}, { {"handleType must have been included in VkExportSemaphoreCreateInfo::handleTypes when semaphore's current payload was created", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-handleType-04760", 59}, { {"If handleType refers to a handle type with copy payload transference semantics, as defined below in Importing Semaphore Payloads, there must be no queue waiting on semaphore", 173}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-handleType-04761", 59}, { {"If handleType refers to a handle type with copy payload transference semantics, semaphore must be signaled, or have an associated semaphore signal operation pending execution", 174}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-handleType-04762", 59}, { {"handleType must be defined as a Zircon event handle", 51}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-handleType-parameter", 63}, { {"handleType must be a valid VkExternalSemaphoreHandleTypeFlagBits value", 70}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-pNext-pNext", 54}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_GET_ZIRCON_HANDLE_INFO_FUCHSIA", 72}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-semaphore-04759", 58}, { {"semaphore must not currently have its payload replaced by an imported payload as described below in Importing Semaphore Payloads unless that imported payload's handle type was included in VkExternalSemaphoreProperties::exportFromImportedHandleTypes for handleType", 263}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-semaphore-04763", 58}, { {"semaphore must have been created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreGetZirconHandleInfoFUCHSIA-semaphore-parameter", 62}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_SIGNAL_INFO", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-semaphore-03257", 42}, { {"semaphore must have been created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE", 85}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-semaphore-parameter", 46}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-value-03258", 38}, { {"value must have a value greater than the current value of the semaphore", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-value-03259", 38}, { {"value must be less than the value of any pending semaphore signal operations", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSignalInfo-value-03260", 38}, { {"value must have a value which does not differ from the current value of the semaphore or the value of any outstanding semaphore wait or signal operation on semaphore by more than maxTimelineSemaphoreValueDifference", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreSubmitInfo-device-03888", 39}, { {"If the device that semaphore was created on is not a device group, deviceIndex must be 0", 88}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-device-03889", 39}, { {"If the device that semaphore was created on is a device group, deviceIndex must be a valid device index", 103}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-pNext-pNext", 38}, { {"pNext must be NULL", 18}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO", 53}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-semaphore-parameter", 46}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03929", 42}, { {"If the geometryShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 112}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03930", 42}, { {"If the tessellationShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 186}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03931", 42}, { {"If the conditionalRendering feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 128}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03932", 42}, { {"If the fragmentDensityMap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 129}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03933", 42}, { {"If the transformFeedback feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 122}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03934", 42}, { {"If the meshShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 108}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-03935", 42}, { {"If the taskShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 108}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-04957", 42}, { {"If the subpassShading feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 118}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-04995", 42}, { {"If the invocationMask feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 119}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-07316", 42}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 181}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-07946", 42}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 163}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-10751", 42}, { {"If the accelerationStructure feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 136}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-10752", 42}, { {"If the rayTracingMaintenance1 feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 136}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-11541", 42}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 153}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreSubmitInfo-stageMask-parameter", 46}, { {"stageMask must be a valid combination of VkPipelineStageFlagBits2 values", 72}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSemaphoreTypeCreateInfo-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO", 58}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreTypeCreateInfo-semaphoreType-03279", 50}, { {"If semaphoreType is VK_SEMAPHORE_TYPE_BINARY, initialValue must be zero", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreTypeCreateInfo-semaphoreType-parameter", 54}, { {"semaphoreType must be a valid VkSemaphoreType value", 51}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreTypeCreateInfo-timelineSemaphore-03252", 54}, { {"If the timelineSemaphore feature is not enabled, semaphoreType must not equal VK_SEMAPHORE_TYPE_TIMELINE", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-flags-parameter", 40}, { {"flags must be a valid combination of VkSemaphoreWaitFlagBits values", 67}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-pNext-pNext", 36}, { {"pNext must be NULL", 18}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-pSemaphores-03256", 42}, { {"All of the elements of pSemaphores must reference a semaphore that was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE", 131}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-pSemaphores-parameter", 46}, { {"pSemaphores must be a valid pointer to an array of semaphoreCount valid VkSemaphore handles", 91}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-pValues-parameter", 42}, { {"pValues must be a valid pointer to an array of semaphoreCount uint64_t values", 77}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_SEMAPHORE_WAIT_INFO", 51}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSemaphoreWaitInfo-semaphoreCount-arraylength", 51}, { {"semaphoreCount must be greater than 0", 37}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-None-09495", 51}, { {"If the dynamicPipelineLayout feature is not enabled, layout must be a valid VkPipelineLayout handle", 99}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-firstSet-08066", 55}, { {"The sum of firstSet and setCount must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 138}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-firstSet-09006", 55}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT bit set", 197}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-firstSet-11803", 55}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT_KHR bit set", 199}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-firstSet-11804", 55}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT bit set", 211}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-layout-09496", 53}, { {"If layout is VK_NULL_HANDLE, the pNext chain must include a valid VkPipelineLayoutCreateInfo structure", 102}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-layout-parameter", 57}, { {"If layout is not VK_NULL_HANDLE, layout must be a valid VkPipelineLayout handle", 79}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pBufferIndices-08064", 61}, { {"Each element of pBufferIndices must be less than VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxDescriptorBufferBindings", 123}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pBufferIndices-08065", 61}, { {"Each element of pBufferIndices must reference a valid descriptor buffer binding set by a previous call to vkCmdBindDescriptorBuffersEXT in commandBuffer", 152}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pBufferIndices-parameter", 65}, { {"pBufferIndices must be a valid pointer to an array of setCount uint32_t values", 78}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pNext-pNext", 52}, { {"pNext must be NULL or a pointer to a valid instance of VkPipelineLayoutCreateInfo", 81}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pOffsets-08061", 55}, { {"The offsets in pOffsets must be aligned to VkPhysicalDeviceDescriptorBufferPropertiesEXT::descriptorBufferOffsetAlignment", 121}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pOffsets-08063", 55}, { {"The offsets in pOffsets must be small enough such that any descriptor binding referenced by layout without the VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT flag computes a valid address inside the underlying VkBuffer", 223}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pOffsets-08126", 55}, { {"The offsets in pOffsets must be small enough such that any location accessed by a shader as a sampler descriptor must be within VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxSamplerDescriptorBufferRange of the sampler descriptor buffer binding", 247}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pOffsets-08127", 55}, { {"The offsets in pOffsets must be small enough such that any location accessed by a shader as a resource descriptor must be within VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxResourceDescriptorBufferRange of the resource descriptor buffer binding", 250}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-pOffsets-parameter", 59}, { {"pOffsets must be a valid pointer to an array of setCount VkDeviceSize values", 76}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_SET_DESCRIPTOR_BUFFER_OFFSETS_INFO_EXT", 70}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-sType-unique", 53}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-setCount-arraylength", 61}, { {"setCount must be greater than 0", 31}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-stageFlags-parameter", 61}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetDescriptorBufferOffsetsInfoEXT-stageFlags-requiredbitmask", 67}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkSetLatencyMarkerInfoNV-marker-parameter", 46}, { {"marker must be a valid VkLatencyMarkerNV value", 46}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSetLatencyMarkerInfoNV-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_SET_LATENCY_MARKER_INFO_NV", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSetPresentConfigNV-numFramesPerBatch-10581", 49}, { {"numFramesPerBatch must not be larger than 8", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSetPresentConfigNV-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_SET_PRESENT_CONFIG_NV", 53}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkShaderCreateInfoEXT-codeSize-08735", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, codeSize must be a multiple of 4", 78}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeSize-arraylength", 47}, { {"codeSize must be greater than 0", 31}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-08872", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, pCode must contain an OpExecutionMode instruction specifying the type of subdivision", 188}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10064", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and if a push constant block is declared in a shader, then an element of pPushConstantRanges->stageFlags must match stage", 231}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10065", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and if a push constant block is declared in a shader, the block must be contained inside the element of pPushConstantRanges that matches the stage", 256}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10383", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and a resource variable is declared in a shader, the corresponding descriptor set in pSetLayouts must match the shader stage", 234}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10384", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and a resource variable is declared in a shader, and the descriptor type is not VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the corresponding descriptor set in pSetLayouts must match the descriptor type", 300}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10385", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and a resource variable is declared in a shader as an array, the corresponding descriptor set in pSetLayouts must match the descriptor count", 250}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-10386", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, flags does not include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and a resource variable is declared in a shader as an array of descriptors, then the descriptor type of that variable must not be VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK", 279}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-12226", 41}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, pCode must contain an OpExecutionMode instruction specifying the output patch size", 183}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-codeType-parameter", 45}, { {"codeType must be a valid VkShaderCodeTypeEXT value", 50}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-descriptorHeap-11314", 47}, { {"If the descriptorHeap feature is not enabled, VkShaderDescriptorSetAndBindingMappingInfoEXT::mappingCount must be 0", 115}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08412", 38}, { {"If stage is not VK_SHADER_STAGE_TASK_BIT_EXT, VK_SHADER_STAGE_MESH_BIT_EXT, VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, or VK_SHADER_STAGE_FRAGMENT_BIT, flags must not include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT", 312}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08414", 38}, { {"If stage is not VK_SHADER_STAGE_MESH_BIT_EXT, flags must not include VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT", 108}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08416", 38}, { {"If flags includes both VK_SHADER_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT_EXT and VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT, the local workgroup size in the X dimension of the shader must be a multiple of maxSubgroupSize", 224}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08417", 38}, { {"If flags includes VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT but not VK_SHADER_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT_EXT and no VkShaderRequiredSubgroupSizeCreateInfoEXT structure is included in the pNext chain, the local workgroup size in the X dimension of the shader must be a multiple of subgroupSize", 310}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08485", 38}, { {"If stage is not VK_SHADER_STAGE_COMPUTE_BIT, flags must not include VK_SHADER_CREATE_DISPATCH_BASE_BIT_EXT", 106}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08486", 38}, { {"If stage is not VK_SHADER_STAGE_FRAGMENT_BIT, flags must not include VK_SHADER_CREATE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_EXT", 126}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08487", 38}, { {"If the attachmentFragmentShadingRate feature is not enabled, flags must not include VK_SHADER_CREATE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_EXT", 141}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08488", 38}, { {"If stage is not VK_SHADER_STAGE_FRAGMENT_BIT, flags must not include VK_SHADER_CREATE_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 125}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08489", 38}, { {"If the fragmentDensityMap feature is not enabled, flags must not include VK_SHADER_CREATE_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 129}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-08992", 38}, { {"If flags includes VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT, stage must be one of VK_SHADER_STAGE_MESH_BIT_EXT, VK_SHADER_STAGE_TASK_BIT_EXT, or VK_SHADER_STAGE_COMPUTE_BIT", 178}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-09404", 38}, { {"If flags includes VK_SHADER_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT_EXT, the subgroupSizeControl feature must be enabled", 119}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-09405", 38}, { {"If flags includes VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT, the computeFullSubgroups feature must be enabled", 115}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11005", 38}, { {"If flags includes VK_SHADER_CREATE_INDIRECT_BINDABLE_BIT_EXT, then the VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 170}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11006", 38}, { {"If flags includes VK_SHADER_CREATE_INDIRECT_BINDABLE_BIT_EXT, then the identified entry point must not specify Xfb execution mode", 129}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11290", 38}, { {"If flags includes VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, setLayoutCount must be 0", 84}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11291", 38}, { {"If flags includes VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, pSetLayouts must be NULL", 84}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11292", 38}, { {"If flags includes VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, and codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, all shader variables in the shader resource interface with a DescriptorSet and Binding decoration must have a mapping declared in VkShaderDescriptorSetAndBindingMappingInfoEXT::pMappings", 293}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11370", 38}, { {"If flags includes VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, pushConstantRangeCount must be 0", 92}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11371", 38}, { {"If flags includes VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, pPushConstantRanges must be NULL", 92}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-11758", 38}, { {"If shader64BitIndexing feature is not enabled, flags must not contain VK_SHADER_CREATE_64_BIT_INDEXING_BIT_EXT", 110}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-flags-parameter", 42}, { {"flags must be a valid combination of VkShaderCreateFlagBitsEXT values", 69}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-micromap-11623", 41}, { {"If flags includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT, the VkPhysicalDeviceOpacityMicromapFeaturesKHR::micromap feature must be enabled", 170}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08427", 42}, { {"If stage is VK_SHADER_STAGE_VERTEX_BIT, nextStage must not include any bits other than VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, and VK_SHADER_STAGE_FRAGMENT_BIT", 191}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08428", 42}, { {"If the tessellationShader feature is not enabled, nextStage must not include VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 164}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08429", 42}, { {"If the geometryShader feature is not enabled, nextStage must not include VK_SHADER_STAGE_GEOMETRY_BIT", 101}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08430", 42}, { {"If stage is VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, nextStage must not include any bits other than VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 144}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08431", 42}, { {"If stage is VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, nextStage must not include any bits other than VK_SHADER_STAGE_GEOMETRY_BIT and VK_SHADER_STAGE_FRAGMENT_BIT", 165}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08433", 42}, { {"If stage is VK_SHADER_STAGE_GEOMETRY_BIT, nextStage must not include any bits other than VK_SHADER_STAGE_FRAGMENT_BIT", 117}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08434", 42}, { {"If stage is VK_SHADER_STAGE_FRAGMENT_BIT or VK_SHADER_STAGE_COMPUTE_BIT, nextStage must be 0", 92}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08435", 42}, { {"If stage is VK_SHADER_STAGE_TASK_BIT_EXT, nextStage must not include any bits other than VK_SHADER_STAGE_MESH_BIT_EXT", 117}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-08436", 42}, { {"If stage is VK_SHADER_STAGE_MESH_BIT_EXT, nextStage must not include any bits other than VK_SHADER_STAGE_FRAGMENT_BIT", 117}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-nextStage-parameter", 46}, { {"nextStage must be a valid combination of VkShaderStageFlagBits values", 69}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08448", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and the identified entry point includes any variable in its interface that is declared with the ClipDistance BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxClipDistances", 270}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08449", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and the identified entry point includes any variable in its interface that is declared with the CullDistance BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxCullDistances", 270}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08450", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and the identified entry point includes variables in its interface that are declared with the ClipDistance BuiltIn decoration and variables in its interface that are declared with the CullDistance BuiltIn decoration, those variables must not have array sizes which sum to more than VkPhysicalDeviceLimits::maxCombinedClipAndCullDistances", 383}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08451", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and the identified entry point includes any variable in its interface that is declared with the SampleMask BuiltIn decoration, that variable must not have an array size greater than VkPhysicalDeviceLimits::maxSampleMaskWords", 270}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08453", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, and the identified entry point has an OpExecutionMode instruction specifying a patch size with OutputVertices, the patch size must be greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxTessellationPatchSize", 371}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08454", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_GEOMETRY_BIT, the identified entry point must have an OpExecutionMode instruction specifying a maximum output vertex count that is greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxGeometryOutputVertices", 296}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08455", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_GEOMETRY_BIT, the identified entry point must have an OpExecutionMode instruction specifying an invocation count that is greater than 0 and less than or equal to VkPhysicalDeviceLimits::maxGeometryShaderInvocations", 289}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08456", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is a pre-rasterization shader stage, and the identified entry point writes to Layer for any primitive, it must write the same value to Layer for all vertices of a given primitive", 234}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08457", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is a pre-rasterization shader stage, and the identified entry point writes to ViewportIndex for any primitive, it must write the same value to ViewportIndex for all vertices of a given primitive", 250}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08459", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and stage is VK_SHADER_STAGE_FRAGMENT_BIT, and the identified entry point writes to FragDepth in any execution path, all execution paths that are not exclusive to helper invocations must either discard the fragment, or write or initialize the value of FragDepth", 307}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08460", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, the shader code in pCode must be valid as described by the Khronos SPIR-V Specification after applying the specializations provided in pSpecializationInfo, if any, and then converting all specialization constants into fixed constants", 279}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08492", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_BINARY_EXT, pCode must be aligned to 16 bytes", 80}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08493", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must be aligned to 4 bytes", 78}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08736", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must point to valid SPIR-V code, formatted and packed as described by the Khronos SPIR-V Specification", 154}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08737", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must adhere to the validation rules described by the Validation Rules within a Module section of the SPIR-V Environment appendix", 180}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08738", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must declare the Shader capability for SPIR-V code", 102}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08739", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must not declare any capability that is not supported by the API, as described by the Capabilities section of the SPIR-V Environment appendix", 193}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08740", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and pCode declares any of the capabilities listed in the SPIR-V Environment appendix, one of the corresponding requirements must be satisfied", 187}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08741", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pCode must not declare any SPIR-V extension that is not supported by the API, as described by the Extension section of the SPIR-V Environment appendix", 196}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-08742", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and pCode declares any of the SPIR-V extensions listed in the SPIR-V Environment appendix, one of the corresponding requirements must be satisfied", 192}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pCode-parameter", 42}, { {"pCode must be a valid pointer to an array of codeSize bytes", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pName-08440", 38}, { {"If codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, pName must be the name of an OpEntryPoint in pCode with an execution model that matches stage", 139}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pName-parameter", 42}, { {"If pName is not NULL, pName must be a null-terminated UTF-8 string", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11315", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, the mapped resource in the shader must be a variable with a structure type decorated with Block in the Uniform Storage Class", 333}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11316", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, the mapped structure must not be larger than maxPushDataSize minus the pushDataOffset used in the mapping", 205}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11317", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, the sum of mapped structure size and shaderRecordDataOffset used in the mapping must not be larger than maxShaderGroupStride", 233}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11318", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT the mapped resource in the shader must be one of: A variable with a structure type decorated with Block in the Uniform Storage Class A variable with a structure type decorated with BufferBlock in the Uniform Storage Class A variable with a structure type decorated with Block in the StorageBuffer Storage Class A OpTypeAccelerationStructureKHR variable A OpTypeAccelerationStructureNV variable", 605}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11378", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the OpArrayLength or OpUntypedArrayLengthKHR instruction must not be used on that resource", 303}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-11399", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, and the mapped resource declaration is an array, the pEmbeddedSampler member of the corresponding mapping structure must be NULL", 490}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-12454", 38}, { {"If the pNext chain specifies a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, and there is a variable in the shader that matches the mapping, the stage must be a ray tracing stage", 333}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkCustomResolveCreateInfoEXT, VkPipelineShaderStageRequiredSubgroupSizeCreateInfo, VkShaderDescriptorSetAndBindingMappingInfoEXT, or VkValidationFeaturesEXT", 287}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pPushConstantRanges-10063", 52}, { {"Any two elements of pPushConstantRanges must not include the same stage in stageFlags", 85}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pPushConstantRanges-parameter", 56}, { {"If pushConstantRangeCount is not 0, and pPushConstantRanges is not NULL, pPushConstantRanges must be a valid pointer to an array of pushConstantRangeCount valid VkPushConstantRange structures", 191}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-pSpecializationInfo-parameter", 56}, { {"If pSpecializationInfo is not NULL, pSpecializationInfo must be a valid pointer to a valid VkSpecializationInfo structure", 121}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_CREATE_INFO_EXT", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-setLayoutCount-12257", 47}, { {"setLayoutCount must be less than or equal to VkPhysicalDeviceLimits::maxBoundDescriptorSets", 91}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-setLayoutCount-13359", 47}, { {"If setLayoutCount is not 0, pSetLayouts is not NULL, and flags does not include VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR, then pSetLayouts must be a valid pointer to an array of setLayoutCount valid VkDescriptorSetLayout handles", 229}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-setLayoutCount-13360", 47}, { {"If setLayoutCount is not 0, pSetLayouts is not NULL, and flags includes VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR, then pSetLayouts must be a valid pointer to an array of setLayoutCount VkDescriptorSetLayout or VK_NULL_HANDLE handles", 233}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08418", 38}, { {"stage must not be VK_SHADER_STAGE_ALL_GRAPHICS or VK_SHADER_STAGE_ALL", 69}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08419", 38}, { {"If the tessellationShader feature is not enabled, stage must not be VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 155}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08420", 38}, { {"If the geometryShader feature is not enabled, stage must not be VK_SHADER_STAGE_GEOMETRY_BIT", 92}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08421", 38}, { {"If the taskShader feature is not enabled, stage must not be VK_SHADER_STAGE_TASK_BIT_EXT", 88}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08422", 38}, { {"If the meshShader feature is not enabled, stage must not be VK_SHADER_STAGE_MESH_BIT_EXT", 88}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08425", 38}, { {"stage must not be VK_SHADER_STAGE_SUBPASS_SHADING_BIT_HUAWEI", 60}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-08426", 38}, { {"stage must not be VK_SHADER_STAGE_CLUSTER_CULLING_BIT_HUAWEI", 60}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-12428", 38}, { {"If stage is VK_SHADER_STAGE_COMPUTE_BIT, then flags must not include VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR", 110}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-12445", 38}, { {"stage must not contain VK_SHADER_STAGE_RAYGEN_BIT_KHR, VK_SHADER_STAGE_ANY_HIT_BIT_KHR, VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR, VK_SHADER_STAGE_MISS_BIT_KHR, VK_SHADER_STAGE_INTERSECTION_BIT_KHR, or VK_SHADER_STAGE_CALLABLE_BIT_KHR", 228}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderCreateInfoEXT-stage-parameter", 42}, { {"stage must be a valid VkShaderStageFlagBits value", 49}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderDescriptorSetAndBindingMappingInfoEXT-pMappings-11244", 66}, { {"Any two elements of pMappings must not have the same value of descriptorSet, an overlapping range specified by firstBinding and bindingCount, and any overlapping bits in resourceMask", 182}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkShaderDescriptorSetAndBindingMappingInfoEXT-pMappings-parameter", 70}, { {"If mappingCount is not 0, pMappings must be a valid pointer to an array of mappingCount valid VkDescriptorSetAndBindingMappingEXT structures", 140}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkShaderDescriptorSetAndBindingMappingInfoEXT-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_DESCRIPTOR_SET_AND_BINDING_MAPPING_INFO_EXT", 82}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkShaderInstrumentationCreateInfoARM-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationCreateInfoARM-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_INSTRUMENTATION_CREATE_INFO_ARM", 70}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDataHeaderARM-stages-parameter", 64}, { {"stages must be a valid combination of VkShaderStageFlagBits values", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDataHeaderARM-stages-requiredbitmask", 70}, { {"stages must not be 0", 20}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDescriptionARM-description-parameter", 70}, { {"description must be a null-terminated UTF-8 string whose length is less than or equal to VK_MAX_DESCRIPTION_SIZE", 112}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDescriptionARM-name-parameter", 63}, { {"name must be a null-terminated UTF-8 string whose length is less than or equal to VK_MAX_DESCRIPTION_SIZE", 105}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDescriptionARM-pNext-pNext", 60}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderInstrumentationMetricDescriptionARM-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_INSTRUMENTATION_METRIC_DESCRIPTION_ARM", 77}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-codeSize-01085", 44}, { {"codeSize must be greater than 0", 31}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-codeSize-08735", 44}, { {"If pCode is a pointer to SPIR-V code, codeSize must be a multiple of 4", 70}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-flags-zerobitmask", 47}, { {"flags must be 0", 15}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-01379", 41}, { {"If pCode is a pointer to GLSL code, it must be valid GLSL code written to the GL_KHR_vulkan_glsl GLSL extension specification", 125}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-07912", 41}, { {"If the VK_NV_glsl_shader extension is not enabled, pCode must be a pointer to SPIR-V code", 89}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08736", 41}, { {"If pCode is a pointer to SPIR-V code, pCode must point to valid SPIR-V code, formatted and packed as described by the Khronos SPIR-V Specification", 146}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08737", 41}, { {"If pCode is a pointer to SPIR-V code, pCode must adhere to the validation rules described by the Validation Rules within a Module section of the SPIR-V Environment appendix", 172}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08738", 41}, { {"If pCode is a pointer to SPIR-V code, pCode must declare the Shader capability for SPIR-V code", 94}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08739", 41}, { {"If pCode is a pointer to SPIR-V code, pCode must not declare any capability that is not supported by the API, as described by the Capabilities section of the SPIR-V Environment appendix", 185}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08740", 41}, { {"If pCode is a pointer to SPIR-V code, and pCode declares any of the capabilities listed in the SPIR-V Environment appendix, one of the corresponding requirements must be satisfied", 179}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08741", 41}, { {"If pCode is a pointer to SPIR-V code, pCode must not declare any SPIR-V extension that is not supported by the API, as described by the Extension section of the SPIR-V Environment appendix", 188}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-08742", 41}, { {"If pCode is a pointer to SPIR-V code, and pCode declares any of the SPIR-V extensions listed in the SPIR-V Environment appendix, one of the corresponding requirements must be satisfied", 184}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-pCode-parameter", 45}, { {"pCode must be a valid pointer to an array of (codeSize/4) uint32_t values", 73}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleCreateInfo-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO", 57}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleIdentifierEXT-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleIdentifierEXT-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_MODULE_IDENTIFIER_EXT", 60}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleValidationCacheCreateInfoEXT-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_SHADER_MODULE_VALIDATION_CACHE_CREATE_INFO_EXT", 78}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShaderModuleValidationCacheCreateInfoEXT-validationCache-parameter", 73}, { {"validationCache must be a valid VkValidationCacheEXT handle", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkShadingRatePaletteNV-pShadingRatePaletteEntries-parameter", 64}, { {"pShadingRatePaletteEntries must be a valid pointer to an array of shadingRatePaletteEntryCount valid VkShadingRatePaletteEntryNV values", 135}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkShadingRatePaletteNV-shadingRatePaletteEntryCount-02071", 62}, { {"shadingRatePaletteEntryCount must be between 1 and VkPhysicalDeviceShadingRateImagePropertiesNV::shadingRatePaletteSize, inclusive", 130}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkShadingRatePaletteNV-shadingRatePaletteEntryCount-arraylength", 68}, { {"shadingRatePaletteEntryCount must be greater than 0", 51}, { "chapters/primsrast.html", 23} } },
        { {"VUID-VkSharedPresentSurfaceCapabilities2KHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_2_KHR", 73}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSharedPresentSurfaceCapabilitiesKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_KHR", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSparseBufferMemoryBindInfo-bindCount-arraylength", 55}, { {"bindCount must be greater than 0", 32}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseBufferMemoryBindInfo-buffer-parameter", 50}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseBufferMemoryBindInfo-pBinds-parameter", 50}, { {"pBinds must be a valid pointer to an array of bindCount valid VkSparseMemoryBind structures", 91}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageFormatProperties2-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageFormatProperties2-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_SPARSE_IMAGE_FORMAT_PROPERTIES_2", 64}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-01108", 41}, { {"extent.width must either be a multiple of the sparse image block width of the image, or else (extent.width + offset.x) must equal the width of the image subresource", 164}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-01110", 41}, { {"extent.height must either be a multiple of the sparse image block height of the image, or else (extent.height + offset.y) must equal the height of the image subresource", 168}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-01112", 41}, { {"extent.depth must either be a multiple of the sparse image block depth of the image, or else (extent.depth + offset.z) must equal the depth of the image subresource", 164}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-09388", 41}, { {"extent.width must be greater than 0", 35}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-09389", 41}, { {"extent.height must be greater than 0", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-extent-09390", 41}, { {"extent.depth must be greater than 0", 35}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-flags-parameter", 44}, { {"flags must be a valid combination of VkSparseMemoryBindFlagBits values", 70}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-memory-01104", 41}, { {"If the sparseResidencyAliased feature is not enabled, and if any other resources are bound to ranges of memory, the range of memory being bound must not overlap with those bound ranges", 184}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-memory-01105", 41}, { {"memory and memoryOffset must match the memory requirements of the calling command's image, as described in section Resource Memory Association", 142}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-memory-02732", 41}, { {"If memory was created with VkExportMemoryAllocateInfo::handleTypes not equal to 0, at least one handle type it contained must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when the image was created", 219}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-memory-02733", 41}, { {"If memory was created by a memory import operation, the external handle type of the imported memory must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 194}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-memory-parameter", 45}, { {"If memory is not VK_NULL_HANDLE, memory must be a valid VkDeviceMemory handle", 77}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-offset-01107", 41}, { {"offset.x must be a multiple of the sparse image block width (VkSparseImageFormatProperties::imageGranularity.width) of the image", 128}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-offset-01109", 41}, { {"offset.y must be a multiple of the sparse image block height (VkSparseImageFormatProperties::imageGranularity.height) of the image", 130}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-offset-01111", 41}, { {"offset.z must be a multiple of the sparse image block depth (VkSparseImageFormatProperties::imageGranularity.depth) of the image", 128}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBind-subresource-parameter", 50}, { {"subresource must be a valid VkImageSubresource structure", 56}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-bindCount-arraylength", 54}, { {"bindCount must be greater than 0", 32}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-image-02901", 44}, { {"image must have been created with VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT set", 74}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-image-parameter", 48}, { {"image must be a valid VkImage handle", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-pBinds-parameter", 49}, { {"pBinds must be a valid pointer to an array of bindCount valid VkSparseImageMemoryBind structures", 96}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-subresource-01106", 50}, { {"The subresource.aspectMask member of each element of pBinds must be valid for the format specified in VkImageCreateInfo when image was created", 142}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-subresource-01722", 50}, { {"The subresource.mipLevel member of each element of pBinds must be less than the mipLevels specified in VkImageCreateInfo when image was created", 143}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryBindInfo-subresource-01723", 50}, { {"The subresource.arrayLayer member of each element of pBinds must be less than the arrayLayers specified in VkImageCreateInfo when image was created", 147}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryRequirements2-pNext-pNext", 49}, { {"pNext must be NULL", 18}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageMemoryRequirements2-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_SPARSE_IMAGE_MEMORY_REQUIREMENTS_2", 66}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageOpaqueMemoryBindInfo-bindCount-arraylength", 60}, { {"bindCount must be greater than 0", 32}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageOpaqueMemoryBindInfo-image-parameter", 54}, { {"image must be a valid VkImage handle", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageOpaqueMemoryBindInfo-pBinds-01103", 51}, { {"If the flags member of any element of pBinds contains VK_SPARSE_MEMORY_BIND_METADATA_BIT, the binding range defined must be within the mip tail region of the metadata aspect of image", 182}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseImageOpaqueMemoryBindInfo-pBinds-parameter", 55}, { {"pBinds must be a valid pointer to an array of bindCount valid VkSparseMemoryBind structures", 91}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-flags-parameter", 39}, { {"flags must be a valid combination of VkSparseMemoryBindFlagBits values", 70}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memory-01096", 36}, { {"If memory is not VK_NULL_HANDLE, memory and memoryOffset must match the memory requirements of the resource, as described in section Resource Memory Association", 160}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memory-01097", 36}, { {"If memory is not VK_NULL_HANDLE, memory must not have been created with a memory type that reports VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT bit set", 146}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memory-02730", 36}, { {"If memory was created with VkExportMemoryAllocateInfo::handleTypes not equal to 0, at least one handle type it contained must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes or VkExternalMemoryImageCreateInfo::handleTypes when the resource was created", 271}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memory-02731", 36}, { {"If memory was created by a memory import operation, the external handle type of the imported memory must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes or VkExternalMemoryImageCreateInfo::handleTypes when the resource was created", 250}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memory-parameter", 40}, { {"If memory is not VK_NULL_HANDLE, memory must be a valid VkDeviceMemory handle", 77}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-memoryOffset-01101", 42}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-resourceOffset-01099", 44}, { {"resourceOffset must be less than the size of the resource", 57}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-resourceOffset-09491", 44}, { {"If the resource being bound is a VkBuffer, resourceOffset, memoryOffset and size must be an integer multiple of the alignment of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with the buffer resource", 242}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-resourceOffset-09492", 44}, { {"If the resource being bound is a VkImage, resourceOffset and memoryOffset must be an integer multiple of the alignment of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements with the image resource", 233}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-size-01098", 34}, { {"size must be greater than 0", 27}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-size-01100", 34}, { {"size must be less than or equal to the size of the resource minus resourceOffset", 80}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSparseMemoryBind-size-01102", 34}, { {"size must be less than or equal to the size of memory minus memoryOffset", 72}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-VkSpecializationInfo-constantID-04911", 42}, { {"The constantID value of each element of pMapEntries must be unique within pMapEntries", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSpecializationInfo-offset-00773", 38}, { {"The offset member of each element of pMapEntries must be less than dataSize", 75}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSpecializationInfo-pData-parameter", 41}, { {"If dataSize is not 0, pData must be a valid pointer to an array of dataSize bytes", 81}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSpecializationInfo-pMapEntries-00774", 43}, { {"The size member of each element of pMapEntries must be less than or equal to dataSize minus offset", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSpecializationInfo-pMapEntries-parameter", 47}, { {"If mapEntryCount is not 0, pMapEntries must be a valid pointer to an array of mapEntryCount valid VkSpecializationMapEntry structures", 133}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSpecializationMapEntry-constantID-00776", 46}, { {"For a constantID specialization constant declared in a shader, size must match the byte size of the constantID. If the specialization constant is of type boolean, size must be the byte size of VkBool32", 201}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkStencilOpState-compareOp-parameter", 41}, { {"compareOp must be a valid VkCompareOp value", 43}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkStencilOpState-depthFailOp-parameter", 43}, { {"depthFailOp must be a valid VkStencilOp value", 45}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkStencilOpState-failOp-parameter", 38}, { {"failOp must be a valid VkStencilOp value", 40}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkStencilOpState-passOp-parameter", 38}, { {"passOp must be a valid VkStencilOp value", 40}, { "chapters/fragops.html", 21} } },
        { {"VUID-VkStreamDescriptorSurfaceCreateInfoGGP-flags-zerobitmask", 61}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkStreamDescriptorSurfaceCreateInfoGGP-pNext-pNext", 55}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkStreamDescriptorSurfaceCreateInfoGGP-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_STREAM_DESCRIPTOR_SURFACE_CREATE_INFO_GGP", 73}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkStreamDescriptorSurfaceCreateInfoGGP-streamDescriptor-02681", 66}, { {"streamDescriptor must be a valid GgpStreamDescriptor", 52}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkStridedDeviceAddressNV-startAddress-parameter", 52}, { {"startAddress must be a valid VkDeviceAddress value", 50}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkStridedDeviceAddressRangeKHR-address-11365", 49}, { {"The sum of address and size must be less than or equal to the sum of an address retrieved from a VkBuffer and the value of VkBufferCreateInfo::size used to create that VkBuffer", 176}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkStridedDeviceAddressRangeKHR-address-parameter", 53}, { {"If address is not 0, address must be a valid VkDeviceAddress value", 66}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkStridedDeviceAddressRangeKHR-size-11411", 46}, { {"If size is not 0, address must not be 0", 39}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkStridedDeviceAddressRangeKHR-stride-10957", 48}, { {"stride must be less than or equal to size", 41}, { "chapters/fundamentals.html", 26} } },
        { {"VUID-VkStridedDeviceAddressRegionKHR-deviceAddress-parameter", 60}, { {"If deviceAddress is not 0, deviceAddress must be a valid VkDeviceAddress value", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-VkStridedDeviceAddressRegionKHR-size-04631", 47}, { {"If size is not zero, all addresses between deviceAddress and deviceAddress + size - 1 must be in the buffer device address range of the same buffer", 147}, { "chapters/resources.html", 23} } },
        { {"VUID-VkStridedDeviceAddressRegionKHR-size-04632", 47}, { {"If size is not zero, stride must be less than or equal to the size of the buffer from which deviceAddress was queried", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSubmitInfo-commonparent", 30}, { {"Each of the elements of pCommandBuffers, the elements of pSignalSemaphores, and the elements of pWaitSemaphores that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 231}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBufferInfos-09942", 43}, { {"If at least one VkCommandBufferSubmitInfo structure in pCommandBufferInfos references a commandBuffer allocated from a pool that was created with a VkDataGraphProcessingEngineCreateInfoARM structure in the pNext chain of VkCommandPoolCreateInfo that included a foreign data graph processing engine in its pProcessingEngines member, then pWaitSemaphoreInfos and pSignalSemaphoreInfos must only reference semaphore objects that were created from external handle types reported as supported in a VkQueueFamilyDataGraphProcessingEnginePropertiesARM::foreignSemaphoreHandleTypes structure via vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM with a queueFamilyIndex matching the one the command pool was created for, for all the foreign data graph processing engines that were part of the VkDataGraphProcessingEngineCreateInfoARM used to create the command pool", 876}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-00075", 39}, { {"Each element of pCommandBuffers must not have been allocated with VK_COMMAND_BUFFER_LEVEL_SECONDARY", 99}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-06014", 39}, { {"If pCommandBuffers contains any suspended render pass instances, they must be resumed by a render pass instance later in submission order within pCommandBuffers", 160}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-06015", 39}, { {"If pCommandBuffers contains any suspended render pass instances, there must be no action or synchronization commands executed in a primary or secondary command buffer between that render pass instance and the render pass instance that resumes it", 245}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-06016", 39}, { {"If pCommandBuffers contains any suspended render pass instances, there must be no render pass instances between that render pass instance and the render pass instance that resumes it", 182}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-06193", 39}, { {"If pCommandBuffers contains any resumed render pass instances, they must be suspended by a render pass instance earlier in submission order within pCommandBuffers", 162}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pCommandBuffers-parameter", 43}, { {"If commandBufferCount is not 0, pCommandBuffers must be a valid pointer to an array of commandBufferCount valid VkCommandBuffer handles", 135}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-03240", 29}, { {"If the pNext chain of this structure includes a VkTimelineSemaphoreSubmitInfo structure and any element of pWaitSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE, then its waitSemaphoreValueCount member must equal waitSemaphoreCount", 258}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-03241", 29}, { {"If the pNext chain of this structure includes a VkTimelineSemaphoreSubmitInfo structure and any element of pSignalSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE, then its signalSemaphoreValueCount member must equal signalSemaphoreCount", 264}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-04120", 29}, { {"If the pNext chain of this structure does not include a VkProtectedSubmitInfo structure with protectedSubmit set to VK_TRUE, then each element of the pCommandBuffers array must be an unprotected command buffer", 209}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-04148", 29}, { {"If the pNext chain of this structure includes a VkProtectedSubmitInfo structure with protectedSubmit set to VK_TRUE, then each element of the pCommandBuffers array must be a protected command buffer", 198}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-09683", 29}, { {"If the pNext chain of this structure includes a VkFrameBoundaryTensorsARM structure then it must also include a VkFrameBoundaryEXT structure", 140}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pNext-pNext", 29}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkAmigoProfilingSubmitInfoSEC, VkD3D12FenceSubmitInfoKHR, VkDeviceGroupSubmitInfo, VkFrameBoundaryEXT, VkFrameBoundaryTensorsARM, VkLatencySubmissionPresentIdNV, VkPerformanceQuerySubmitInfoKHR, VkProtectedSubmitInfo, VkThrottleHintSubmitInfoSEC, VkTimelineSemaphoreSubmitInfo, VkWin32KeyedMutexAcquireReleaseInfoKHR, or VkWin32KeyedMutexAcquireReleaseInfoNV", 489}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pSignalSemaphores-03242", 41}, { {"For each element of pSignalSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pSignalSemaphoreValues must have a value greater than the current value of the semaphore when the semaphore signal operation is executed", 295}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pSignalSemaphores-03244", 41}, { {"For each element of pSignalSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pSignalSemaphoreValues must have a value which does not differ from the current value of the semaphore or the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 395}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pSignalSemaphores-parameter", 45}, { {"If signalSemaphoreCount is not 0, pSignalSemaphores must be a valid pointer to an array of signalSemaphoreCount valid VkSemaphore handles", 137}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-00078", 41}, { {"Each element of pWaitDstStageMask must not include VK_PIPELINE_STAGE_HOST_BIT", 77}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-03937", 41}, { {"If the synchronization2 feature is not enabled, pWaitDstStageMask must not be 0", 79}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04090", 41}, { {"If the geometryShader feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 118}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04091", 41}, { {"If the tessellationShader feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 190}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04092", 41}, { {"If the conditionalRendering feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 134}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04093", 41}, { {"If the fragmentDensityMap feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 135}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04094", 41}, { {"If the transformFeedback feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 128}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04095", 41}, { {"If the meshShader feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 114}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-04096", 41}, { {"If the taskShader feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 114}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-07318", 41}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 187}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-07949", 41}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 169}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-10754", 41}, { {"If the accelerationStructure feature is not enabled, pWaitDstStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 142}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitDstStageMask-parameter", 45}, { {"If waitSemaphoreCount is not 0, pWaitDstStageMask must be a valid pointer to an array of waitSemaphoreCount valid combinations of VkPipelineStageFlagBits values", 160}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitSemaphores-03239", 39}, { {"If any element of pWaitSemaphores or pSignalSemaphores was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE, then the pNext chain must include a VkTimelineSemaphoreSubmitInfo structure", 196}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitSemaphores-03243", 39}, { {"For each element of pWaitSemaphores created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE the corresponding element of VkTimelineSemaphoreSubmitInfo::pWaitSemaphoreValues must have a value which does not differ from the current value of the semaphore or the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 391}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-pWaitSemaphores-parameter", 43}, { {"If waitSemaphoreCount is not 0, pWaitSemaphores must be a valid pointer to an array of waitSemaphoreCount valid VkSemaphore handles", 131}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-sType-sType", 29}, { {"sType must be VK_STRUCTURE_TYPE_SUBMIT_INFO", 43}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-sType-unique", 30}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo-variableSampleLocations-06017", 47}, { {"If the variableSampleLocations limit is not supported, and any element of pCommandBuffers contains any suspended render pass instances, where a graphics pipeline has been bound, any pipelines bound in the render pass instance that resumes it, or any subsequent render pass instances that resume from that one and so on, must use the same sample locations", 354}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-commandBuffer-06010", 38}, { {"If any commandBuffer member of an element of pCommandBufferInfos contains any suspended render pass instances, they must be resumed by a render pass instance later in submission order within pCommandBufferInfos", 210}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-commandBuffer-06011", 38}, { {"If any commandBuffer member of an element of pCommandBufferInfos contains any suspended render pass instances, there must be no action or synchronization commands between that render pass instance and the render pass instance that resumes it", 241}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-commandBuffer-06012", 38}, { {"If any commandBuffer member of an element of pCommandBufferInfos contains any suspended render pass instances, there must be no render pass instances between that render pass instance and the render pass instance that resumes it", 228}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-commandBuffer-06192", 38}, { {"If any commandBuffer member of an element of pCommandBufferInfos contains any resumed render pass instances, they must be suspended by a render pass instance earlier in submission order within pCommandBufferInfos", 212}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-flags-03886", 30}, { {"If flags includes VK_SUBMIT_PROTECTED_BIT, all elements of pCommandBuffers must be protected command buffers", 108}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-flags-03887", 30}, { {"If flags does not include VK_SUBMIT_PROTECTED_BIT, each element of pCommandBuffers must not be a protected command buffer", 121}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-flags-parameter", 34}, { {"flags must be a valid combination of VkSubmitFlagBits values", 60}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pCommandBufferInfos-09933", 44}, { {"If at least one VkCommandBufferSubmitInfo structure in pCommandBufferInfos references a commandBuffer allocated from a pool that was created with a VkDataGraphProcessingEngineCreateInfoARM structure in the pNext chain of VkCommandPoolCreateInfo that included a foreign data graph processing engine in its pProcessingEngines member, then pWaitSemaphoreInfos and pSignalSemaphoreInfos must only reference semaphore objects that were created from external handle types reported as supported in a VkQueueFamilyDataGraphProcessingEnginePropertiesARM::foreignSemaphoreHandleTypes structure via vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM with a queueFamilyIndex matching the one the command pool was created for, for all the foreign data graph processing engines that were part of the VkDataGraphProcessingEngineCreateInfoARM used to create the command pool", 876}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pCommandBufferInfos-parameter", 48}, { {"If commandBufferInfoCount is not 0, pCommandBufferInfos must be a valid pointer to an array of commandBufferInfoCount valid VkCommandBufferSubmitInfo structures", 160}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pNext-09682", 30}, { {"If the pNext chain of this structure includes a VkFrameBoundaryTensorsARM structure then it must also include a VkFrameBoundaryEXT structure", 140}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pNext-pNext", 30}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkFrameBoundaryEXT, VkFrameBoundaryTensorsARM, VkLatencySubmissionPresentIdNV, VkPerformanceQuerySubmitInfoKHR, VkWin32KeyedMutexAcquireReleaseInfoKHR, or VkWin32KeyedMutexAcquireReleaseInfoNV", 323}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pSignalSemaphoreInfos-parameter", 50}, { {"If signalSemaphoreInfoCount is not 0, pSignalSemaphoreInfos must be a valid pointer to an array of signalSemaphoreInfoCount valid VkSemaphoreSubmitInfo structures", 162}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-pWaitSemaphoreInfos-parameter", 48}, { {"If waitSemaphoreInfoCount is not 0, pWaitSemaphoreInfos must be a valid pointer to an array of waitSemaphoreInfoCount valid VkSemaphoreSubmitInfo structures", 156}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-sType-sType", 30}, { {"sType must be VK_STRUCTURE_TYPE_SUBMIT_INFO_2", 45}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-sType-unique", 31}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-semaphore-03881", 34}, { {"If the same semaphore is used as the semaphore member of both an element of pSignalSemaphoreInfos and pWaitSemaphoreInfos, and that semaphore is a timeline semaphore, the value member of the pSignalSemaphoreInfos element must be greater than the value member of the pWaitSemaphoreInfos element", 293}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-semaphore-03882", 34}, { {"If the semaphore member of any element of pSignalSemaphoreInfos is a timeline semaphore, the value member of that element must have a value greater than the current value of the semaphore when the semaphore signal operation is executed", 235}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-semaphore-03883", 34}, { {"If the semaphore member of any element of pSignalSemaphoreInfos is a timeline semaphore, the value member of that element must have a value which does not differ from the current value of the semaphore or the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 335}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-semaphore-03884", 34}, { {"If the semaphore member of any element of pWaitSemaphoreInfos is a timeline semaphore, the value member of that element must have a value which does not differ from the current value of the semaphore or the value of any outstanding semaphore wait or signal operation on that semaphore by more than maxTimelineSemaphoreValueDifference", 333}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubmitInfo2-variableSampleLocations-06013", 48}, { {"If the variableSampleLocations limit is not supported, and any commandBuffer member of an element of pCommandBufferInfos contains any suspended render pass instances, where a graphics pipeline has been bound, any pipelines bound in the render pass instance that resumes it, or any subsequent render pass instances that resume from that one and so on, must use the same sample locations", 385}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkSubpassBeginInfo-contents-09382", 38}, { {"If contents is VK_SUBPASS_CONTENTS_INLINE_AND_SECONDARY_COMMAND_BUFFERS_KHR, then at least one of the following features must be enabled: maintenance7 nestedCommandBuffer", 170}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassBeginInfo-contents-parameter", 42}, { {"contents must be a valid VkSubpassContents value", 48}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassBeginInfo-pNext-pNext", 35}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassBeginInfo-sType-sType", 35}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_BEGIN_INFO", 50}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dependencyFlags-02520", 46}, { {"If dependencyFlags includes VK_DEPENDENCY_VIEW_LOCAL_BIT, srcSubpass must not be equal to VK_SUBPASS_EXTERNAL", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dependencyFlags-02521", 46}, { {"If dependencyFlags includes VK_DEPENDENCY_VIEW_LOCAL_BIT, dstSubpass must not be equal to VK_SUBPASS_EXTERNAL", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dependencyFlags-10203", 46}, { {"dependencyFlags must not include VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dependencyFlags-parameter", 50}, { {"dependencyFlags must be a valid combination of VkDependencyFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstAccessMask-00869", 44}, { {"Any access flag included in dstAccessMask must be supported by one of the pipeline stages in dstStageMask, as specified in the table of supported access types", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstAccessMask-parameter", 48}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-03937", 43}, { {"If the synchronization2 feature is not enabled, dstStageMask must not be 0", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04090", 43}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04091", 43}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04092", 43}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04093", 43}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04094", 43}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04095", 43}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-04096", 43}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-07318", 43}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-07949", 43}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-10754", 43}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-dstStageMask-parameter", 47}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcAccessMask-00868", 44}, { {"Any access flag included in srcAccessMask must be supported by one of the pipeline stages in srcStageMask, as specified in the table of supported access types", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcAccessMask-parameter", 48}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-03937", 43}, { {"If the synchronization2 feature is not enabled, srcStageMask must not be 0", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04090", 43}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04091", 43}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04092", 43}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04093", 43}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04094", 43}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04095", 43}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-04096", 43}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-07318", 43}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-07949", 43}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-10754", 43}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcStageMask-parameter", 47}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcSubpass-00864", 41}, { {"srcSubpass must be less than or equal to dstSubpass, unless one of them is VK_SUBPASS_EXTERNAL, to avoid cyclic dependencies and ensure a valid execution order", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcSubpass-00865", 41}, { {"srcSubpass and dstSubpass must not both be equal to VK_SUBPASS_EXTERNAL", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcSubpass-00872", 41}, { {"If srcSubpass equals dstSubpass and that subpass has more than one bit set in the view mask, then dependencyFlags must include VK_DEPENDENCY_VIEW_LOCAL_BIT", 155}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcSubpass-02243", 41}, { {"If srcSubpass equals dstSubpass, and srcStageMask and dstStageMask both include a framebuffer-space stage, then dependencyFlags must include VK_DEPENDENCY_BY_REGION_BIT", 168}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency-srcSubpass-06809", 41}, { {"If srcSubpass is equal to dstSubpass and srcStageMask includes a framebuffer-space stage, dstStageMask must only contain framebuffer-space stages", 145}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dependencyFlags-03090", 47}, { {"If dependencyFlags includes VK_DEPENDENCY_VIEW_LOCAL_BIT, srcSubpass must not be equal to VK_SUBPASS_EXTERNAL", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dependencyFlags-03091", 47}, { {"If dependencyFlags includes VK_DEPENDENCY_VIEW_LOCAL_BIT, dstSubpass must not be equal to VK_SUBPASS_EXTERNAL", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dependencyFlags-03092", 47}, { {"If dependencyFlags does not include VK_DEPENDENCY_VIEW_LOCAL_BIT, viewOffset must be 0", 86}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dependencyFlags-10204", 47}, { {"dependencyFlags must not include VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dependencyFlags-parameter", 51}, { {"dependencyFlags must be a valid combination of VkDependencyFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstAccessMask-03089", 45}, { {"Any access flag included in dstAccessMask must be supported by one of the pipeline stages in dstStageMask, as specified in the table of supported access types", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstAccessMask-parameter", 49}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-03937", 44}, { {"If the synchronization2 feature is not enabled, dstStageMask must not be 0", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04090", 44}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04091", 44}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04092", 44}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04093", 44}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04094", 44}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04095", 44}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-04096", 44}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-07318", 44}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-07949", 44}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-10754", 44}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-dstStageMask-parameter", 48}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkMemoryBarrier2 or VkMemoryBarrierAccessFlags3KHR", 181}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_DEPENDENCY_2", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcAccessMask-03088", 45}, { {"Any access flag included in srcAccessMask must be supported by one of the pipeline stages in srcStageMask, as specified in the table of supported access types", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcAccessMask-parameter", 49}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits values", 68}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-03937", 44}, { {"If the synchronization2 feature is not enabled, srcStageMask must not be 0", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04090", 44}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04091", 44}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04092", 44}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04093", 44}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04094", 44}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04095", 44}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-04096", 44}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-07318", 44}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-07949", 44}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-10754", 44}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcStageMask-parameter", 48}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcSubpass-02245", 42}, { {"If srcSubpass equals dstSubpass, and srcStageMask and dstStageMask both include a framebuffer-space stage, then dependencyFlags must include VK_DEPENDENCY_BY_REGION_BIT", 168}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcSubpass-03084", 42}, { {"srcSubpass must be less than or equal to dstSubpass, unless one of them is VK_SUBPASS_EXTERNAL, to avoid cyclic dependencies and ensure a valid execution order", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcSubpass-03085", 42}, { {"srcSubpass and dstSubpass must not both be equal to VK_SUBPASS_EXTERNAL", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-srcSubpass-06810", 42}, { {"If srcSubpass is equal to dstSubpass and srcStageMask includes a framebuffer-space stage, dstStageMask must only contain framebuffer-space stages", 145}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDependency2-viewOffset-02530", 42}, { {"If viewOffset is not equal to 0, srcSubpass must not be equal to dstSubpass", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-None-09431", 36}, { {"If either of the following is enabled: The VK_AMD_mixed_attachment_samples extension The VK_NV_framebuffer_mixed_samples extension all attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have a sample count that is smaller than or equal to the sample count of pDepthStencilAttachment if it is not VK_ATTACHMENT_UNUSED", 337}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-00853", 42}, { {"Each element of pPreserveAttachments must not be VK_ATTACHMENT_UNUSED", 69}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06912", 42}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06913", 42}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06914", 42}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06915", 42}, { {"If the attachment member of pDepthStencilAttachment is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06916", 42}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 239}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06917", 42}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 241}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06918", 42}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL", 205}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06919", 42}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 290}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06920", 42}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06921", 42}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR", 157}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06922", 42}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 156}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-attachment-06923", 42}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-colorAttachmentCount-00845", 52}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-flags-00856", 37}, { {"If flags includes VK_SUBPASS_DESCRIPTION_PER_VIEW_POSITION_X_ONLY_BIT_NVX, it must also include VK_SUBPASS_DESCRIPTION_PER_VIEW_ATTRIBUTES_BIT_NVX", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-flags-03341", 37}, { {"If flags includes VK_SUBPASS_DESCRIPTION_CUSTOM_RESOLVE_BIT_EXT, and if pResolveAttachments is not NULL, then each resolve attachment must be VK_ATTACHMENT_UNUSED", 162}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-flags-03343", 37}, { {"If flags includes VK_SUBPASS_DESCRIPTION_CUSTOM_RESOLVE_BIT_EXT, then the subpass must be the last subpass in a subpass dependency chain", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-flags-10683", 37}, { {"If flags includes VK_SUBPASS_DESCRIPTION_TILE_SHADING_APRON_BIT_QCOM, the render pass must have been created with a VkRenderPassTileShadingCreateInfoQCOM::tileApronSize greater than (0,0)", 187}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-flags-parameter", 41}, { {"flags must be a valid combination of VkSubpassDescriptionFlagBits values", 72}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-inputAttachmentCount-12293", 52}, { {"inputAttachmentCount must be less than or equal to maxPerStageDescriptorInputAttachments", 88}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-layout-02519", 38}, { {"If any attachment is used by more than one VkAttachmentReference member, then each use must use the same layout", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-linearColorAttachment-06496", 53}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pInputAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-linearColorAttachment-06497", 53}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-linearColorAttachment-06498", 53}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pResolveAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 294}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-loadOp-00846", 38}, { {"If the first use of an attachment in this render pass is as an input attachment, and the attachment is not also used as a color or depth/stencil attachment in the same subpass, then loadOp must not be VK_ATTACHMENT_LOAD_OP_CLEAR", 228}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pColorAttachments-02648", 49}, { {"All attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pColorAttachments-09430", 49}, { {"All attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have the same sample count", 102}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pColorAttachments-parameter", 53}, { {"If colorAttachmentCount is not 0, pColorAttachments must be a valid pointer to an array of colorAttachmentCount valid VkAttachmentReference structures", 150}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pDepthStencilAttachment-01418", 55}, { {"If none of the following are enabled: The VK_AMD_mixed_attachment_samples extension The VK_NV_framebuffer_mixed_samples extension all attachments in pDepthStencilAttachment and pColorAttachments that are not VK_ATTACHMENT_UNUSED must have the same sample count", 260}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pDepthStencilAttachment-02650", 55}, { {"If pDepthStencilAttachment is not NULL and the attachment is not VK_ATTACHMENT_UNUSED then it must have an image format whose potential format features contain VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 206}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pDepthStencilAttachment-04438", 55}, { {"pDepthStencilAttachment and pColorAttachments must not contain references to the same attachment", 96}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pDepthStencilAttachment-parameter", 59}, { {"If pDepthStencilAttachment is not NULL, pDepthStencilAttachment must be a valid pointer to a valid VkAttachmentReference structure", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pInputAttachments-02647", 49}, { {"All attachments in pInputAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features contain at least VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 232}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pInputAttachments-02868", 49}, { {"If the render pass is created with VK_RENDER_PASS_CREATE_TRANSFORM_BIT_QCOM each of the elements of pInputAttachments must be VK_ATTACHMENT_UNUSED", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pInputAttachments-parameter", 53}, { {"If inputAttachmentCount is not 0, pInputAttachments must be a valid pointer to an array of inputAttachmentCount valid VkAttachmentReference structures", 150}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pPreserveAttachments-00854", 52}, { {"Each element of pPreserveAttachments must not also be an element of any other member of the subpass description", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pPreserveAttachments-parameter", 56}, { {"If preserveAttachmentCount is not 0, pPreserveAttachments must be a valid pointer to an array of preserveAttachmentCount uint32_t values", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-00847", 51}, { {"If pResolveAttachments is not NULL, for each resolve attachment that is not VK_ATTACHMENT_UNUSED, the corresponding color attachment must not be VK_ATTACHMENT_UNUSED", 165}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-00848", 51}, { {"If pResolveAttachments is not NULL, for each resolve attachment that is not VK_ATTACHMENT_UNUSED, the corresponding color attachment must not have a sample count of VK_SAMPLE_COUNT_1_BIT", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-00849", 51}, { {"If pResolveAttachments is not NULL, each resolve attachment that is not VK_ATTACHMENT_UNUSED must have a sample count of VK_SAMPLE_COUNT_1_BIT", 142}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-00850", 51}, { {"If pResolveAttachments is not NULL, each resolve attachment that is not VK_ATTACHMENT_UNUSED must have the same VkFormat as its corresponding color attachment", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-02649", 51}, { {"All attachments in pResolveAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT", 175}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pResolveAttachments-parameter", 55}, { {"If colorAttachmentCount is not 0, and pResolveAttachments is not NULL, pResolveAttachments must be a valid pointer to an array of colorAttachmentCount valid VkAttachmentReference structures", 189}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pipelineBindPoint-04952", 49}, { {"pipelineBindPoint must be VK_PIPELINE_BIND_POINT_GRAPHICS or VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI", 106}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription-pipelineBindPoint-parameter", 53}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-None-09456", 37}, { {"If either of the following is enabled: The VK_AMD_mixed_attachment_samples extension The VK_NV_framebuffer_mixed_samples extension all attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have a sample count that is smaller than or equal to the sample count of pDepthStencilAttachment if it is not VK_ATTACHMENT_UNUSED", 337}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-02799", 43}, { {"If the attachment member of any element of pInputAttachments is not VK_ATTACHMENT_UNUSED, then the aspectMask member must be a valid combination of VkImageAspectFlagBits", 169}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-02800", 43}, { {"If the attachment member of any element of pInputAttachments is not VK_ATTACHMENT_UNUSED, then the aspectMask member must not be 0", 130}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-02801", 43}, { {"If the attachment member of any element of pInputAttachments is not VK_ATTACHMENT_UNUSED, then the aspectMask member must not include VK_IMAGE_ASPECT_METADATA_BIT", 162}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-03073", 43}, { {"Each element of pPreserveAttachments must not be VK_ATTACHMENT_UNUSED", 69}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-04563", 43}, { {"If the attachment member of any element of pInputAttachments is not VK_ATTACHMENT_UNUSED, then the aspectMask member must not include VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT for any index i", 188}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06251", 43}, { {"If the attachment member of pDepthStencilAttachment is not VK_ATTACHMENT_UNUSED and its pNext chain includes a VkAttachmentReferenceStencilLayout structure, the layout member of pDepthStencilAttachment must not be VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 301}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06912", 43}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06913", 43}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 211}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06914", 43}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 213}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06915", 43}, { {"If the attachment member of pDepthStencilAttachment is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL", 195}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06916", 43}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 239}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06917", 43}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL", 241}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06918", 43}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL or VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL", 205}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06919", 43}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 290}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06920", 43}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06921", 43}, { {"If the attachment member of an element of pInputAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL_KHR", 157}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06922", 43}, { {"If the attachment member of an element of pColorAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 156}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-attachment-06923", 43}, { {"If the attachment member of an element of pResolveAttachments is not VK_ATTACHMENT_UNUSED, its layout member must not be VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL_KHR", 158}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-colorAttachmentCount-00845", 53}, { {"colorAttachmentCount must be less than or equal to maxColorAttachments", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09335", 54}, { {"If the externalFormatResolve feature is not enabled and pResolveAttachments is not NULL, for each resolve attachment that does not have the value VK_ATTACHMENT_UNUSED, the corresponding color attachment must not have the value VK_ATTACHMENT_UNUSED", 247}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09338", 54}, { {"If the externalFormatResolve feature is not enabled and pResolveAttachments is not NULL, for each resolve attachment that is not VK_ATTACHMENT_UNUSED, the corresponding color attachment must not have a sample count of VK_SAMPLE_COUNT_1_BIT", 239}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09339", 54}, { {"If the externalFormatResolve feature is not enabled, each element of pResolveAttachments must have the same VkFormat as its corresponding color attachment", 154}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09344", 54}, { {"If the externalFormatResolve feature is enabled, pResolveAttachments is not NULL, and colorAttachmentCount is not 1, any element of pResolveAttachments that is not VK_ATTACHMENT_UNUSED, must not have a format of VK_FORMAT_UNDEFINED", 231}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09345", 54}, { {"If the externalFormatResolve feature is enabled, pResolveAttachments is not NULL, any element of pResolveAttachments is not VK_ATTACHMENT_UNUSED and has a format of VK_FORMAT_UNDEFINED, and the corresponding element of pColorAttachments is not VK_ATTACHMENT_UNUSED, the color attachment must have a samples value of 1", 317}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09346", 54}, { {"If the externalFormatResolve feature is enabled, pResolveAttachments is not NULL, and any element of pResolveAttachments is not VK_ATTACHMENT_UNUSED and has a format of VK_FORMAT_UNDEFINED, viewMask must be 0", 208}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09347", 54}, { {"If the externalFormatResolve feature is enabled, pResolveAttachments is not NULL, and any element of pResolveAttachments is not VK_ATTACHMENT_UNUSED and has a format of VK_FORMAT_UNDEFINED, VkFragmentShadingRateAttachmentInfoKHR::pFragmentShadingRateAttachment must either be NULL or a VkAttachmentReference2 structure with an attachment value of VK_ATTACHMENT_UNUSED", 367}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-externalFormatResolve-09348", 54}, { {"If the externalFormatResolve feature is enabled, pResolveAttachments is not NULL, and any element of pResolveAttachments is not VK_ATTACHMENT_UNUSED and has a format of VK_FORMAT_UNDEFINED, elements of pInputAttachments referencing either a color attachment or resolve attachment used in this subpass must not include VK_IMAGE_ASPECT_PLANE_i_BIT for any index i in its aspectMask", 379}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-03076", 38}, { {"If flags includes VK_SUBPASS_DESCRIPTION_PER_VIEW_POSITION_X_ONLY_BIT_NVX, it must also include VK_SUBPASS_DESCRIPTION_PER_VIEW_ATTRIBUTES_BIT_NVX", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-04907", 38}, { {"If flags includes VK_SUBPASS_DESCRIPTION_CUSTOM_RESOLVE_BIT_EXT, and if pResolveAttachments is not NULL, then each resolve attachment must be VK_ATTACHMENT_UNUSED", 162}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-04908", 38}, { {"If flags includes VK_SUBPASS_DESCRIPTION_CUSTOM_RESOLVE_BIT_EXT, and if pDepthStencilResolveAttachment is not NULL, then the depth/stencil resolve attachment must be VK_ATTACHMENT_UNUSED", 186}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-04909", 38}, { {"If flags includes VK_SUBPASS_DESCRIPTION_CUSTOM_RESOLVE_BIT_EXT, then the subpass must be the last subpass in a subpass dependency chain", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-10683", 38}, { {"If flags includes VK_SUBPASS_DESCRIPTION_TILE_SHADING_APRON_BIT_QCOM, the render pass must have been created with a VkRenderPassTileShadingCreateInfoQCOM::tileApronSize greater than (0,0)", 187}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-flags-parameter", 42}, { {"flags must be a valid combination of VkSubpassDescriptionFlagBits values", 72}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-inputAttachmentCount-12293", 53}, { {"inputAttachmentCount must be less than or equal to maxPerStageDescriptorInputAttachments", 88}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-layout-02528", 39}, { {"If any attachment is used by more than one VkAttachmentReference2 member, then each use must use the same layout", 112}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-linearColorAttachment-06499", 54}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pInputAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-linearColorAttachment-06500", 54}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 292}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-linearColorAttachment-06501", 54}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, all attachments in pResolveAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 294}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-loadOp-00846", 39}, { {"If the first use of an attachment in this render pass is as an input attachment, and the attachment is not also used as a color or depth/stencil attachment in the same subpass, then loadOp must not be VK_ATTACHMENT_LOAD_OP_CLEAR", 228}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-multisampledRenderToSingleSampled-06869", 66}, { {"If the multisampledRenderToSingleSampled feature is not enabled, all attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have the same sample count", 167}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-multisampledRenderToSingleSampled-06872", 66}, { {"If none of the following are enabled: The VK_AMD_mixed_attachment_samples extension The VK_NV_framebuffer_mixed_samples extension The multisampledRenderToSingleSampled feature all attachments in pDepthStencilAttachment and pColorAttachments that are not VK_ATTACHMENT_UNUSED must have the same sample count", 306}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-multiview-06558", 42}, { {"If the multiview feature is not enabled, viewMask must be 0", 59}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-nullColorAttachmentWithExternalFormatResolve-09336", 77}, { {"If the nullColorAttachmentWithExternalFormatResolve property is VK_FALSE and pResolveAttachments is not NULL, for each resolve attachment that has a format of VK_FORMAT_UNDEFINED, the corresponding color attachment must not have the value VK_ATTACHMENT_UNUSED", 259}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-nullColorAttachmentWithExternalFormatResolve-09337", 77}, { {"If the nullColorAttachmentWithExternalFormatResolve property is VK_TRUE and pResolveAttachments is not NULL, for each resolve attachment that has a format of VK_FORMAT_UNDEFINED, the corresponding color attachment must have the value VK_ATTACHMENT_UNUSED", 254}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pColorAttachments-02898", 50}, { {"All attachments in pColorAttachments that are not VK_ATTACHMENT_UNUSED must have image formats whose potential format features contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pColorAttachments-parameter", 54}, { {"If colorAttachmentCount is not 0, pColorAttachments must be a valid pointer to an array of colorAttachmentCount valid VkAttachmentReference2 structures", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pDepthStencilAttachment-02900", 56}, { {"If pDepthStencilAttachment is not NULL and the attachment is not VK_ATTACHMENT_UNUSED then it must have an image format whose potential format features contain VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 206}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pDepthStencilAttachment-04440", 56}, { {"An attachment must not be used in both pDepthStencilAttachment and pColorAttachments", 84}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pDepthStencilAttachment-parameter", 60}, { {"If pDepthStencilAttachment is not NULL, pDepthStencilAttachment must be a valid pointer to a valid VkAttachmentReference2 structure", 131}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pInputAttachments-02897", 50}, { {"All attachments in pInputAttachments that are not VK_ATTACHMENT_UNUSED and any of the following is true: the externalFormatResolve feature is not enabled the nullColorAttachmentWithExternalFormatResolve property is VK_FALSE does not have a non-zero value of VkExternalFormatANDROID::externalFormat must have image formats whose potential format features contain at least VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 459}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pInputAttachments-parameter", 54}, { {"If inputAttachmentCount is not 0, pInputAttachments must be a valid pointer to an array of inputAttachmentCount valid VkAttachmentReference2 structures", 151}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pNext-06870", 38}, { {"If the pNext chain includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then all attachments in pColorAttachments and pDepthStencilAttachment that are not VK_ATTACHMENT_UNUSED must have a sample count that is either VK_SAMPLE_COUNT_1_BIT or equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 388}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pNext-06871", 38}, { {"If the pNext chain includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, and pDepthStencilAttachment is not NULL, does not have the value VK_ATTACHMENT_UNUSED, and has a sample count of VK_SAMPLE_COUNT_1_BIT, the pNext chain must also include a VkSubpassDescriptionDepthStencilResolve structure with pDepthStencilResolveAttachment that is either NULL or has the value VK_ATTACHMENT_UNUSED", 461}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkFragmentShadingRateAttachmentInfoKHR, VkMultisampledRenderToSingleSampledInfoEXT, VkRenderPassCreationControlEXT, VkRenderPassSubpassFeedbackCreateInfoEXT, or VkSubpassDescriptionDepthStencilResolve", 331}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pPreserveAttachments-03074", 53}, { {"Each element of pPreserveAttachments must not also be an element of any other member of the subpass description", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pPreserveAttachments-parameter", 57}, { {"If preserveAttachmentCount is not 0, pPreserveAttachments must be a valid pointer to an array of preserveAttachmentCount uint32_t values", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pResolveAttachments-03067", 52}, { {"If pResolveAttachments is not NULL, each resolve attachment that is not VK_ATTACHMENT_UNUSED must have a sample count of VK_SAMPLE_COUNT_1_BIT", 142}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pResolveAttachments-09343", 52}, { {"All attachments in pResolveAttachments that are not VK_ATTACHMENT_UNUSED and do not have an image format of VK_FORMAT_UNDEFINED must have image formats whose potential format features contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT", 230}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pResolveAttachments-parameter", 56}, { {"If colorAttachmentCount is not 0, and pResolveAttachments is not NULL, pResolveAttachments must be a valid pointer to an array of colorAttachmentCount valid VkAttachmentReference2 structures", 190}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pipelineBindPoint-04953", 50}, { {"pipelineBindPoint must be VK_PIPELINE_BIND_POINT_GRAPHICS or VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI", 106}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-pipelineBindPoint-parameter", 54}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_2", 53}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescription2-viewMask-06706", 41}, { {"The index of the most significant bit in viewMask must be less than maxMultiviewViewCount", 89}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-depthResolveMode-03183", 67}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED and the VkFormat of pDepthStencilResolveAttachment has a depth component, then the value of depthResolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedDepthResolveModes or VK_RESOLVE_MODE_NONE", 332}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-02651", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED then it must have an image format whose potential format features contain VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 215}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03177", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, pDepthStencilAttachment must not be NULL or have the value VK_ATTACHMENT_UNUSED", 175}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03178", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, depthResolveMode and stencilResolveMode must not both be VK_RESOLVE_MODE_NONE", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03179", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, pDepthStencilAttachment must not have a sample count of VK_SAMPLE_COUNT_1_BIT", 173}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03180", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, pDepthStencilResolveAttachment must have a sample count of VK_SAMPLE_COUNT_1_BIT", 176}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03181", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED and VkFormat of pDepthStencilResolveAttachment has a depth component, then the VkFormat of pDepthStencilAttachment must have a depth component with the same number of bits and numeric format", 285}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03182", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, and VkFormat of pDepthStencilResolveAttachment has a stencil component, then the VkFormat of pDepthStencilAttachment must have a stencil component with the same number of bits and numeric format", 290}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03185", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, the VkFormat of pDepthStencilResolveAttachment has both depth and stencil components, VkPhysicalDeviceDepthStencilResolveProperties::independentResolve is VK_FALSE, and VkPhysicalDeviceDepthStencilResolveProperties::independentResolveNone is VK_FALSE, then the values of depthResolveMode and stencilResolveMode must be identical", 424}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-03186", 81}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, the VkFormat of pDepthStencilResolveAttachment has both depth and stencil components, VkPhysicalDeviceDepthStencilResolveProperties::independentResolve is VK_FALSE and VkPhysicalDeviceDepthStencilResolveProperties::independentResolveNone is VK_TRUE, then the values of depthResolveMode and stencilResolveMode must be identical or one of them must be VK_RESOLVE_MODE_NONE", 466}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pDepthStencilResolveAttachment-parameter", 85}, { {"If pDepthStencilResolveAttachment is not NULL, pDepthStencilResolveAttachment must be a valid pointer to a valid VkAttachmentReference2 structure", 145}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pNext-06873", 56}, { {"If the pNext chain of VkSubpassDescription2 includes a VkMultisampledRenderToSingleSampledInfoEXT structure, the multisampledRenderToSingleSampledEnable field is VK_TRUE, and pDepthStencilAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED, depthResolveMode and stencilResolveMode must not both be VK_RESOLVE_MODE_NONE", 338}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pNext-06874", 56}, { {"If the pNext chain of VkSubpassDescription2 includes a VkMultisampledRenderToSingleSampledInfoEXT structure whose multisampledRenderToSingleSampledEnable field is VK_TRUE, and pDepthStencilAttachment is not NULL, does not have the value VK_ATTACHMENT_UNUSED, and has a VkFormat that has a depth component, then the value of depthResolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedDepthResolveModes or VK_RESOLVE_MODE_NONE", 469}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pNext-06875", 56}, { {"If the pNext chain of VkSubpassDescription2 includes a VkMultisampledRenderToSingleSampledInfoEXT structure whose multisampledRenderToSingleSampledEnable field is VK_TRUE, and pDepthStencilAttachment is not NULL, does not have the value VK_ATTACHMENT_UNUSED, and has a VkFormat with a stencil component, then the value of stencilResolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedStencilResolveModes or VK_RESOLVE_MODE_NONE", 471}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pNext-06876", 56}, { {"If the pNext chain of VkSubpassDescription2 includes a VkMultisampledRenderToSingleSampledInfoEXT structure whose multisampledRenderToSingleSampledEnable field is VK_TRUE, pDepthStencilAttachment is not NULL, does not have the value VK_ATTACHMENT_UNUSED, and has a VkFormat with both depth and stencil components, and both VkPhysicalDeviceDepthStencilResolveProperties::independentResolve and VkPhysicalDeviceDepthStencilResolveProperties::independentResolveNone are VK_FALSE, then the values of depthResolveMode and stencilResolveMode must be identical", 553}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-pNext-06877", 56}, { {"If the pNext chain of VkSubpassDescription2 includes a VkMultisampledRenderToSingleSampledInfoEXT structure whose multisampledRenderToSingleSampledEnable field is VK_TRUE, pDepthStencilAttachment is not NULL, does not have the value VK_ATTACHMENT_UNUSED, and has a VkFormat with both depth and stencil components, VkPhysicalDeviceDepthStencilResolveProperties::independentResolve is VK_FALSE, and VkPhysicalDeviceDepthStencilResolveProperties::independentResolveNone is VK_TRUE, then the values of depthResolveMode and stencilResolveMode must be identical or one of them must be VK_RESOLVE_MODE_NONE", 599}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_DEPTH_STENCIL_RESOLVE", 73}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassDescriptionDepthStencilResolve-stencilResolveMode-03184", 69}, { {"If pDepthStencilResolveAttachment is not NULL and does not have the value VK_ATTACHMENT_UNUSED and the VkFormat of pDepthStencilResolveAttachment has a stencil component, then the value of stencilResolveMode must be one of the bits set in VkPhysicalDeviceDepthStencilResolveProperties::supportedStencilResolveModes or VK_RESOLVE_MODE_NONE", 338}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassEndInfo-pNext-pNext", 33}, { {"pNext must be NULL or a pointer to a valid instance of VkRenderPassFragmentDensityMapOffsetEndInfoEXT", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassEndInfo-sType-sType", 33}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_END_INFO", 48}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassEndInfo-sType-unique", 34}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassResolvePerformanceQueryEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_RESOLVE_PERFORMANCE_QUERY_EXT", 69}, { "chapters/formats.html", 21} } },
        { {"VUID-VkSubpassSampleLocationsEXT-sampleLocationsInfo-parameter", 62}, { {"sampleLocationsInfo must be a valid VkSampleLocationsInfoEXT structure", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassSampleLocationsEXT-subpassIndex-01532", 51}, { {"subpassIndex must be less than the subpassCount specified in VkRenderPassCreateInfo the render pass specified by VkRenderPassBeginInfo::renderPass was created with", 163}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkSubpassShadingPipelineCreateInfoHUAWEI-renderPass-parameter", 66}, { {"renderPass must be a valid VkRenderPass handle", 46}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSubpassShadingPipelineCreateInfoHUAWEI-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_SUBPASS_SHADING_PIPELINE_CREATE_INFO_HUAWEI", 75}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSubpassShadingPipelineCreateInfoHUAWEI-subpass-04946", 59}, { {"subpass must be created with VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI bind point", 85}, { "chapters/pipelines.html", 23} } },
        { {"VUID-VkSubresourceHostMemcpySize-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_SUBRESOURCE_HOST_MEMCPY_SIZE", 60}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSubresourceLayout2-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkImageCompressionPropertiesEXT or VkSubresourceHostMemcpySize", 193}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSubresourceLayout2-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_SUBRESOURCE_LAYOUT_2", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSubresourceLayout2-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSubsampledImageFormatPropertiesEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_SUBSAMPLED_IMAGE_FORMAT_PROPERTIES_EXT", 70}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkSurfaceCapabilities2EXT-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilities2EXT-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_EXT", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilities2KHR-pNext-pNext", 42}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDisplayNativeHdrSurfaceCapabilitiesAMD, VkImageUsageFlags2CreateInfoKHR, VkLatencySurfaceCapabilitiesNV, VkPresentTimingSurfaceCapabilitiesEXT, VkSharedPresentSurfaceCapabilities2KHR, VkSharedPresentSurfaceCapabilitiesKHR, VkSurfaceCapabilitiesFullScreenExclusiveEXT, VkSurfaceCapabilitiesPresentBarrierNV, VkSurfaceCapabilitiesPresentId2KHR, VkSurfaceCapabilitiesPresentWait2KHR, VkSurfacePresentModeCompatibilityKHR, VkSurfacePresentScalingCapabilitiesKHR, VkSurfaceProtectedCapabilitiesKHR, or VkSwapchainFlagsSurfaceCapabilitiesEXT", 668}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilities2KHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_KHR", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilities2KHR-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilitiesFullScreenExclusiveEXT-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_FULL_SCREEN_EXCLUSIVE_EXT", 78}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilitiesPresentBarrierNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_BARRIER_NV", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilitiesPresentId2KHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_ID_2_KHR", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCapabilitiesPresentWait2KHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_WAIT_2_KHR", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCreateInfoOHOS-flags-zerobitmask", 46}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCreateInfoOHOS-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceCreateInfoOHOS-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_CREATE_INFO_OHOS", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFormat2KHR-pNext-06750", 36}, { {"If the VK_EXT_image_compression_control_swapchain extension is not supported, the pNext chain must not include an VkImageCompressionPropertiesEXT structure", 155}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFormat2KHR-pNext-pNext", 36}, { {"pNext must be NULL or a pointer to a valid instance of VkImageCompressionPropertiesEXT", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFormat2KHR-sType-sType", 36}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_FORMAT_2_KHR", 52}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFormat2KHR-sType-unique", 37}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFullScreenExclusiveInfoEXT-fullScreenExclusive-parameter", 70}, { {"fullScreenExclusive must be a valid VkFullScreenExclusiveEXT value", 66}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFullScreenExclusiveInfoEXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_INFO_EXT", 70}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFullScreenExclusiveWin32InfoEXT-hmonitor-02673", 60}, { {"hmonitor must be a valid HMONITOR", 33}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceFullScreenExclusiveWin32InfoEXT-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_WIN32_INFO_EXT", 76}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentModeCompatibilityKHR-pPresentModes-parameter", 65}, { {"If presentModeCount is not 0, and pPresentModes is not NULL, pPresentModes must be a valid pointer to an array of presentModeCount VkPresentModeKHR values", 154}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentModeCompatibilityKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_COMPATIBILITY_KHR", 70}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentModeKHR-presentMode-07780", 46}, { {"presentMode must be a value reported by vkGetPhysicalDeviceSurfacePresentModesKHR for the specified surface", 107}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentModeKHR-presentMode-parameter", 50}, { {"presentMode must be a valid VkPresentModeKHR value", 50}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentModeKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_KHR", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentScalingCapabilitiesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_PRESENT_SCALING_CAPABILITIES_KHR", 72}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentScalingCapabilitiesKHR-supportedPresentGravityX-parameter", 78}, { {"supportedPresentGravityX must be a valid combination of VkPresentGravityFlagBitsKHR values", 90}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentScalingCapabilitiesKHR-supportedPresentGravityY-parameter", 78}, { {"supportedPresentGravityY must be a valid combination of VkPresentGravityFlagBitsKHR values", 90}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfacePresentScalingCapabilitiesKHR-supportedPresentScaling-parameter", 77}, { {"supportedPresentScaling must be a valid combination of VkPresentScalingFlagBitsKHR values", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSurfaceProtectedCapabilitiesKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_SURFACE_PROTECTED_CAPABILITIES_KHR", 66}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCalibratedTimestampInfoEXT-presentStage-parameter", 65}, { {"presentStage must be a valid combination of VkPresentStageFlagBitsEXT values", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSwapchainCalibratedTimestampInfoEXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_CALIBRATED_TIMESTAMP_INFO_EXT", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSwapchainCalibratedTimestampInfoEXT-swapchain-parameter", 62}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSwapchainCalibratedTimestampInfoEXT-timeDomain-12228", 59}, { {"If the timeDomain member of the VkCalibratedTimestampInfoKHR structure in this structure's pNext chain is VK_TIME_DOMAIN_PRESENT_STAGE_LOCAL_EXT, presentStage must specify one and only one present stage", 202}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkSwapchainCounterCreateInfoEXT-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_COUNTER_CREATE_INFO_EXT", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCounterCreateInfoEXT-surfaceCounters-01244", 58}, { {"The bits in surfaceCounters must be supported by VkSwapchainCreateInfoKHR::surface, as reported by vkGetPhysicalDeviceSurfaceCapabilities2EXT", 141}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCounterCreateInfoEXT-surfaceCounters-parameter", 62}, { {"surfaceCounters must be a valid combination of VkSurfaceCounterFlagBitsEXT values", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-commonparent", 42}, { {"Both of oldSwapchain, and surface that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkInstance", 155}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-compositeAlpha-01280", 50}, { {"compositeAlpha must be one of the bits present in the supportedCompositeAlpha member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface", 196}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-compositeAlpha-parameter", 54}, { {"compositeAlpha must be a valid VkCompositeAlphaFlagBitsKHR value", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-flags-03168", 41}, { {"If flags contains VK_SWAPCHAIN_CREATE_MUTABLE_FORMAT_BIT_KHR then the pNext chain must include a VkImageFormatListCreateInfo structure with a viewFormatCount greater than zero and pViewFormats must have an element equal to imageFormat", 234}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-flags-03187", 41}, { {"If flags contains VK_SWAPCHAIN_CREATE_PROTECTED_BIT_KHR, then VkSurfaceProtectedCapabilitiesKHR::supportsProtected must be VK_TRUE in the VkSurfaceProtectedCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for surface", 248}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-flags-04100", 41}, { {"If flags does not contain VK_SWAPCHAIN_CREATE_MUTABLE_FORMAT_BIT_KHR and the pNext chain include a VkImageFormatListCreateInfo structure then VkImageFormatListCreateInfo::viewFormatCount must be 0 or 1", 201}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-flags-12448", 41}, { {"flags must be a subset of flags returned in VkSwapchainFlagsSurfaceCapabilitiesEXT::swapchainSupportedFlags , given VkSwapchainCreateInfoKHR::presentMode in VkSurfacePresentModeKHR", 180}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-flags-parameter", 45}, { {"flags must be a valid combination of VkSwapchainCreateFlagBitsKHR values", 72}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageArrayLayers-01275", 52}, { {"imageArrayLayers must be greater than 0 and less than or equal to the maxImageArrayLayers member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface", 208}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageColorSpace-parameter", 55}, { {"imageColorSpace must be a valid VkColorSpaceKHR value", 53}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageExtent-01689", 47}, { {"imageExtent members width and height must both be non-zero", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageFormat-01273", 47}, { {"imageFormat and imageColorSpace must match the format and colorSpace members, respectively, of one of the VkSurfaceFormatKHR structures returned by vkGetPhysicalDeviceSurfaceFormatsKHR for the surface", 200}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageFormat-01778", 47}, { {"The implied image creation parameters of the swapchain must be supported as reported by vkGetPhysicalDeviceImageFormatProperties", 128}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageFormat-parameter", 51}, { {"imageFormat must be a valid VkFormat value", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageSharingMode-01277", 52}, { {"If imageSharingMode is VK_SHARING_MODE_CONCURRENT, pQueueFamilyIndices must be a valid pointer to an array of queueFamilyIndexCount uint32_t values", 147}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageSharingMode-01278", 52}, { {"If imageSharingMode is VK_SHARING_MODE_CONCURRENT, queueFamilyIndexCount must be greater than 1", 95}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageSharingMode-01428", 52}, { {"If imageSharingMode is VK_SHARING_MODE_CONCURRENT, each element of pQueueFamilyIndices must be unique and must be less than pQueueFamilyPropertyCount returned by either vkGetPhysicalDeviceQueueFamilyProperties or vkGetPhysicalDeviceQueueFamilyProperties2 for the physicalDevice that was used to create device", 308}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageSharingMode-parameter", 56}, { {"imageSharingMode must be a valid VkSharingMode value", 52}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageUsage-01384", 46}, { {"If presentMode is VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR or VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR, imageUsage must be a subset of the supported usage flags present in the sharedPresentSupportedUsageFlags member of the VkSharedPresentSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for surface", 343}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageUsage-parameter", 50}, { {"imageUsage must be a valid combination of VkImageUsageFlagBits values", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-imageUsage-requiredbitmask", 56}, { {"imageUsage must not be 0", 24}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-minImageCount-01272", 49}, { {"minImageCount must be less than or equal to the value returned in the maxImageCount member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface if the returned maxImageCount is not zero", 244}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-minImageCount-01383", 49}, { {"minImageCount must be 1 if presentMode is either VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR or VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR", 139}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-multisampledRenderToSwapchain-12447", 65}, { {"If the multisampledRenderToSwapchain feature is not enabled, flags must not contain VK_SWAPCHAIN_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT", 149}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-oldSwapchain-01933", 48}, { {"If oldSwapchain is not VK_NULL_HANDLE, oldSwapchain must be a non-retired swapchain associated with native window referred to by surface", 136}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-oldSwapchain-parameter", 52}, { {"If oldSwapchain is not VK_NULL_HANDLE, oldSwapchain must be a valid VkSwapchainKHR handle", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-02679", 41}, { {"If the pNext chain includes a VkSurfaceFullScreenExclusiveInfoEXT structure with its fullScreenExclusive member set to VK_FULL_SCREEN_EXCLUSIVE_APPLICATION_CONTROLLED_EXT, and surface was created using vkCreateWin32SurfaceKHR, a VkSurfaceFullScreenExclusiveWin32InfoEXT structure must be included in the pNext chain", 315}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-04099", 41}, { {"If a VkImageFormatListCreateInfo structure was included in the pNext chain and VkImageFormatListCreateInfo::viewFormatCount is not zero then all of the formats in VkImageFormatListCreateInfo::pViewFormats must be compatible with the format as described in the compatibility table", 279}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-06752", 41}, { {"If the imageCompressionControlSwapchain feature is not enabled, the pNext chain must not include an VkImageCompressionControlEXT structure", 138}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-07781", 41}, { {"If a VkSwapchainPresentScalingCreateInfoKHR structure was not included in the pNext chain, or it is included and VkSwapchainPresentScalingCreateInfoKHR::scalingBehavior is zero then imageExtent must be between minImageExtent and maxImageExtent, inclusive, where minImageExtent and maxImageExtent are members of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface", 419}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-07782", 41}, { {"If a VkSwapchainPresentScalingCreateInfoKHR structure was included in the pNext chain and VkSwapchainPresentScalingCreateInfoKHR::scalingBehavior is not zero then imageExtent must be between minScaledImageExtent and maxScaledImageExtent, inclusive, where minScaledImageExtent and maxScaledImageExtent are members of the VkSurfacePresentScalingCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilities2KHR for the surface and presentMode", 455}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-pNext-pNext", 41}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkDeviceGroupSwapchainCreateInfoKHR, VkImageCompressionControlEXT, VkImageFormatListCreateInfo, VkImageUsageFlags2CreateInfoKHR, VkSurfaceFullScreenExclusiveInfoEXT, VkSurfaceFullScreenExclusiveWin32InfoEXT, VkSwapchainCounterCreateInfoEXT, VkSwapchainDisplayNativeHdrCreateInfoAMD, VkSwapchainLatencyCreateInfoNV, VkSwapchainPresentBarrierCreateInfoNV, VkSwapchainPresentModesCreateInfoKHR, or VkSwapchainPresentScalingCreateInfoKHR", 564}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-physicalDeviceCount-01429", 55}, { {"If the logical device was created with VkDeviceGroupDeviceCreateInfo::physicalDeviceCount equal to 1, flags must not contain VK_SWAPCHAIN_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT_KHR", 180}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-preTransform-01279", 48}, { {"preTransform must be one of the bits present in the supportedTransforms member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface", 190}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-preTransform-parameter", 52}, { {"preTransform must be a valid VkSurfaceTransformFlagBitsKHR value", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentMode-01281", 47}, { {"presentMode must be one of the VkPresentModeKHR values returned by vkGetPhysicalDeviceSurfacePresentModesKHR for the surface", 124}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentMode-01427", 47}, { {"If presentMode is VK_PRESENT_MODE_FIFO_LATEST_READY_KHR, VK_PRESENT_MODE_IMMEDIATE_KHR, VK_PRESENT_MODE_MAILBOX_KHR, VK_PRESENT_MODE_FIFO_KHR or VK_PRESENT_MODE_FIFO_RELAXED_KHR, imageUsage must be a subset of the supported usage flags present in the supportedUsageFlags member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for surface", 385}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentMode-02839", 47}, { {"If presentMode is not VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR nor VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR, then minImageCount must be greater than or equal to the value returned in the minImageCount member of the VkSurfaceCapabilitiesKHR structure returned by vkGetPhysicalDeviceSurfaceCapabilitiesKHR for the surface", 325}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentMode-parameter", 51}, { {"presentMode must be a valid VkPresentModeKHR value", 50}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentModeFifoLatestReady-10161", 62}, { {"If the presentModeFifoLatestReady feature is not enabled, presentMode must not be VK_PRESENT_MODE_FIFO_LATEST_READY_KHR", 119}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-presentTiming-12232", 49}, { {"If none of the presentTiming, presentAtAbsoluteTime, or presentAtRelativeTime features are enabled, flags must not contain VK_SWAPCHAIN_CREATE_PRESENT_TIMING_BIT_EXT", 165}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-sType-unique", 42}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-surface-01270", 43}, { {"surface must be a surface that is supported by the device as determined using vkGetPhysicalDeviceSurfaceSupportKHR", 114}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-surface-parameter", 47}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-swapchainMaintenance1-10155", 57}, { {"If the swapchainMaintenance1 feature is not enabled, then the pNext chain must not include a VkSwapchainPresentModesCreateInfoKHR structure", 139}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainCreateInfoKHR-swapchainMaintenance1-10157", 57}, { {"If the swapchainMaintenance1 feature is not enabled, then flags must not include VK_SWAPCHAIN_CREATE_DEFERRED_MEMORY_ALLOCATION_BIT_KHR", 135}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainDisplayNativeHdrCreateInfoAMD-localDimmingEnable-04449", 70}, { {"It is only valid to set localDimmingEnable to VK_TRUE if VkDisplayNativeHdrSurfaceCapabilitiesAMD::localDimmingSupport is supported", 131}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainDisplayNativeHdrCreateInfoAMD-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_DISPLAY_NATIVE_HDR_CREATE_INFO_AMD", 76}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainFlagsSurfaceCapabilitiesEXT-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_FLAGS_SURFACE_CAPABILITIES_EXT", 72}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainLatencyCreateInfoNV-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_LATENCY_CREATE_INFO_NV", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentBarrierCreateInfoNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_BARRIER_CREATE_INFO_NV", 72}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-pFences-07758", 49}, { {"Each element of pFences that is not VK_NULL_HANDLE must be unsignaled", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-pFences-07759", 49}, { {"Each element of pFences that is not VK_NULL_HANDLE must not be associated with any other queue command that has not yet completed execution on that queue", 153}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-pFences-parameter", 53}, { {"pFences must be a valid pointer to an array of swapchainCount valid or VK_NULL_HANDLE VkFence handles", 101}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_FENCE_INFO_KHR", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-swapchainCount-07757", 56}, { {"swapchainCount must be equal to VkPresentInfoKHR::swapchainCount", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentFenceInfoKHR-swapchainCount-arraylength", 62}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModeInfoKHR-pPresentModes-07761", 54}, { {"Each entry in pPresentModes must be a presentation mode specified in VkSwapchainPresentModesCreateInfoKHR::pPresentModes when creating the entry's corresponding swapchain", 170}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModeInfoKHR-pPresentModes-parameter", 58}, { {"pPresentModes must be a valid pointer to an array of swapchainCount valid VkPresentModeKHR values", 97}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModeInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODE_INFO_KHR", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModeInfoKHR-swapchainCount-07760", 55}, { {"swapchainCount must be equal to VkPresentInfoKHR::swapchainCount", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModeInfoKHR-swapchainCount-arraylength", 61}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-None-07762", 52}, { {"Each entry in pPresentModes must be one of the VkPresentModeKHR values returned by vkGetPhysicalDeviceSurfacePresentModesKHR for the surface", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-pPresentModes-07763", 61}, { {"The entries in pPresentModes must be a subset of the present modes returned in VkSurfacePresentModeCompatibilityKHR::pPresentModes, given VkSwapchainCreateInfoKHR::presentMode in VkSurfacePresentModeKHR", 202}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-pPresentModes-parameter", 65}, { {"pPresentModes must be a valid pointer to an array of presentModeCount valid VkPresentModeKHR values", 99}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-presentMode-07764", 59}, { {"VkSwapchainCreateInfoKHR::presentMode must be included in pPresentModes", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-presentModeCount-arraylength", 70}, { {"presentModeCount must be greater than 0", 39}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-presentModeFifoLatestReady-10160", 74}, { {"If the presentModeFifoLatestReady feature is not enabled, pPresentModes must not contain VK_PRESENT_MODE_FIFO_LATEST_READY_KHR", 126}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentModesCreateInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODES_CREATE_INFO_KHR", 71}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-07765", 65}, { {"If presentGravityX is 0, presentGravityY must be 0", 50}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-07766", 65}, { {"If presentGravityX is not 0, presentGravityY must not be 0", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-07768", 65}, { {"presentGravityX must not have more than one bit set", 51}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-07772", 65}, { {"presentGravityX must be 0 or a valid x-axis present gravity for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentGravityX, given VkSwapchainCreateInfoKHR::presentMode in VkSurfacePresentModeKHR", 227}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-07773", 65}, { {"If the swapchain is created with VkSwapchainPresentModesCreateInfoKHR, presentGravityX must be 0 or a valid x-axis present gravity for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentGravityX, given each present mode in VkSwapchainPresentModesCreateInfoKHR::pPresentModes in VkSurfacePresentModeKHR", 333}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityX-parameter", 69}, { {"presentGravityX must be a valid combination of VkPresentGravityFlagBitsKHR values", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityY-07769", 65}, { {"presentGravityY must not have more than one bit set", 51}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityY-07774", 65}, { {"presentGravityY must be 0 or a valid y-axis present gravity for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentGravityY, given VkSwapchainCreateInfoKHR::presentMode in VkSurfacePresentModeKHR", 227}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityY-07775", 65}, { {"If the swapchain is created with VkSwapchainPresentModesCreateInfoKHR, presentGravityY must be 0 or a valid y-axis present gravity for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentGravityY, given each present mode in VkSwapchainPresentModesCreateInfoKHR::pPresentModes in VkSurfacePresentModeKHR", 333}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-presentGravityY-parameter", 69}, { {"presentGravityY must be a valid combination of VkPresentGravityFlagBitsKHR values", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_SCALING_CREATE_INFO_KHR", 73}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-scalingBehavior-07767", 65}, { {"scalingBehavior must not have more than one bit set", 51}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-scalingBehavior-07770", 65}, { {"scalingBehavior must be 0 or a valid scaling method for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentScaling, given VkSwapchainCreateInfoKHR::presentMode in VkSurfacePresentModeKHR", 218}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-scalingBehavior-07771", 65}, { {"If the swapchain is created with VkSwapchainPresentModesCreateInfoKHR, scalingBehavior must be 0 or a valid scaling method for the surface as returned in VkSurfacePresentScalingCapabilitiesKHR::supportedPresentScaling, given each present mode in VkSwapchainPresentModesCreateInfoKHR::pPresentModes in VkSurfacePresentModeKHR", 324}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-scalingBehavior-parameter", 69}, { {"scalingBehavior must be a valid combination of VkPresentScalingFlagBitsKHR values", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainPresentScalingCreateInfoKHR-swapchainMaintenance1-10154", 71}, { {"If the swapchainMaintenance1 feature is not enabled, then scalingBehavior, presentGravityX, and presentGravityY must be 0", 121}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimeDomainPropertiesEXT-pNext-pNext", 51}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimeDomainPropertiesEXT-pTimeDomains-12370", 58}, { {"pTimeDomains and pTimeDomainIds must both be NULL or both not be NULL", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimeDomainPropertiesEXT-pTimeDomains-12371", 58}, { {"If pTimeDomains and pTimeDomainIds are not NULL, then timeDomainCount must not be zero", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimeDomainPropertiesEXT-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_TIME_DOMAIN_PROPERTIES_EXT", 68}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimingPropertiesEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSwapchainTimingPropertiesEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_SWAPCHAIN_TIMING_PROPERTIES_EXT", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkSysmemColorSpaceFUCHSIA-colorSpace-06402", 47}, { {"colorSpace must be a ColorSpaceType as defined in fuchsia.sysmem/image_formats.fidl", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSysmemColorSpaceFUCHSIA-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkSysmemColorSpaceFUCHSIA-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_SYSMEM_COLOR_SPACE_FUCHSIA", 58}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCaptureDescriptorDataInfoARM-pNext-pNext", 53}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorCaptureDescriptorDataInfoARM-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_CAPTURE_DESCRIPTOR_DATA_INFO_ARM", 71}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorCaptureDescriptorDataInfoARM-tensor-09705", 54}, { {"If tensor is not VK_NULL_HANDLE then tensor must have been created with VK_TENSOR_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM set in VkTensorCreateInfoARM::flags", 165}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorCaptureDescriptorDataInfoARM-tensor-parameter", 58}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorCopyARM-dimensionCount-09955", 41}, { {"dimensionCount must be greater than 0 if pSrcOffset, pDstOffset, or pExtent is not NULL", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCopyARM-pDstOffset-parameter", 41}, { {"If dimensionCount is not 0, and pDstOffset is not NULL, pDstOffset must be a valid pointer to an array of dimensionCount uint64_t values", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCopyARM-pExtent-parameter", 38}, { {"If dimensionCount is not 0, and pExtent is not NULL, pExtent must be a valid pointer to an array of dimensionCount uint64_t values", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCopyARM-pNext-pNext", 32}, { {"pNext must be NULL", 18}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCopyARM-pSrcOffset-parameter", 41}, { {"If dimensionCount is not 0, and pSrcOffset is not NULL, pSrcOffset must be a valid pointer to an array of dimensionCount uint64_t values", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCopyARM-sType-sType", 32}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_COPY_ARM", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-VkTensorCreateInfoARM-flags-09726", 38}, { {"If flags includes VK_TENSOR_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM, the descriptorBufferCaptureReplay feature must be enabled", 134}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-flags-11395", 38}, { {"If VkOpaqueCaptureDataCreateInfoEXT::pData is not NULL, flags must contain VK_TENSOR_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_ARM", 130}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-flags-11396", 38}, { {"If flags contains VK_TENSOR_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_ARM, VkOpaqueCaptureDataCreateInfoEXT::pData->size must be equal to imageCaptureReplayOpaqueDataSize", 170}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-flags-parameter", 42}, { {"flags must be a valid combination of VkTensorCreateFlagBitsARM values", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-maintenance11-13358", 46}, { {"If the maintenance11 feature is enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 0", 136}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pDescription-09720", 45}, { {"If pDescription->tiling is VK_TENSOR_TILING_OPTIMAL_ARM, pDescription->pStrides must be NULL", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pDescription-09728", 45}, { {"If pDescription->usage does not have any of the following bits set (i.e. if it is not possible to create a tensor view for this tensor), then the format features must contain the format feature flags required by the usage flags for pDescription->format as indicated in the Format Feature Dependent Usage Flags section VK_TENSOR_USAGE_SHADER_BIT_ARM VK_TENSOR_USAGE_DATA_GRAPH_BIT_ARM", 383}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pDescription-parameter", 49}, { {"pDescription must be a valid pointer to a valid VkTensorDescriptionARM structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pNext-09727", 38}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_TENSOR_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pNext-09833", 38}, { {"If a VkTensorRollingBackingCreateInfoARM structure is present in the pNext chain of VkTensorCreateInfoARM, then pDescription->tiling must be VK_TENSOR_TILING_LINEAR_ARM, VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, or VK_TENSOR_TILING_BRICK_4_WIDE_ARM", 277}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pNext-09834", 38}, { {"If a VkTensorRollingBackingCreateInfoARM structure is present in the pNext chain of VkTensorCreateInfoARM, then pDescription->dimensionCount must be less than or equal to 4", 172}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pNext-09864", 38}, { {"If the pNext chain includes a VkExternalMemoryTensorCreateInfoARM structure, its handleTypes member must only contain bits that are also in VkExternalTensorPropertiesARM::externalMemoryProperties.compatibleHandleTypes, as returned by vkGetPhysicalDeviceExternalTensorPropertiesARM with pExternalTensorInfo->handleType equal to any one of the handle types specified in VkExternalMemoryTensorCreateInfoARM::handleTypes", 416}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-pNext-pNext", 38}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkExternalMemoryTensorCreateInfoARM, VkOpaqueCaptureDataCreateInfoEXT, VkOpaqueCaptureDescriptorDataCreateInfoEXT, or VkTensorRollingBackingCreateInfoARM", 284}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-protectedMemory-09729", 48}, { {"If the protectedMemory feature is not enabled, flags must not contain VK_TENSOR_CREATE_PROTECTED_BIT_ARM", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_CREATE_INFO_ARM", 54}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sType-unique", 39}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sharingMode-09722", 44}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, pQueueFamilyIndices must be a valid pointer to an array of queueFamilyIndexCount uint32_t values", 142}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sharingMode-09723", 44}, { {"If the maintenance11 feature is not enabled and sharingMode is VK_SHARING_MODE_CONCURRENT, then queueFamilyIndexCount must be greater than 1", 140}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sharingMode-09725", 44}, { {"If sharingMode is VK_SHARING_MODE_CONCURRENT, each element of pQueueFamilyIndices must be unique and must be less than pQueueFamilyPropertyCount returned by either vkGetPhysicalDeviceQueueFamilyProperties or vkGetPhysicalDeviceQueueFamilyProperties2 for the physicalDevice that was used to create device", 303}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-sharingMode-parameter", 48}, { {"sharingMode must be a valid VkSharingMode value", 47}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorCreateInfoARM-tensorElements-09721", 47}, { {"tensorElements (as defined in Tensor Creation Limits) must not be greater than VkPhysicalDeviceTensorPropertiesARM::maxTensorElements", 133}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDependencyInfoARM-pTensorMemoryBarriers-parameter", 62}, { {"pTensorMemoryBarriers must be a valid pointer to an array of tensorMemoryBarrierCount valid VkTensorMemoryBarrierARM structures", 127}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorDependencyInfoARM-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_DEPENDENCY_INFO_ARM", 58}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorDependencyInfoARM-tensorMemoryBarrierCount-arraylength", 67}, { {"tensorMemoryBarrierCount must be greater than 0", 47}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorDescriptionARM-None-09740", 38}, { {"If the tensorNonPacked feature is not enabled, then the members of VkTensorDescriptionARM must describe a packed tensor", 119}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-dimensionCount-09733", 48}, { {"dimensionCount must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxTensorDimensionCount", 105}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-dimensionCount-arraylength", 54}, { {"dimensionCount must be greater than 0", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-format-09735", 40}, { {"format must not be VK_FORMAT_UNDEFINED and must be a one-component VkFormat", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-format-parameter", 44}, { {"format must be a valid VkFormat value", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pDimensions-09734", 45}, { {"For each i where i <= dimensionCount-1, pDimensions[i] must be greater than 0", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pDimensions-09883", 45}, { {"For each i where i <= dimensionCount-1, pDimensions[i] must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxPerDimensionTensorElements", 151}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pDimensions-parameter", 49}, { {"pDimensions must be a valid pointer to an array of dimensionCount int64_t values", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-09736", 42}, { {"pStrides[dimensionCount-1] must equal the size in bytes of a tensor element", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-09737", 42}, { {"For each i, pStrides[i] must be a multiple of the element size", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-09738", 42}, { {"For each i, pStrides[i] must be greater than 0 and less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxTensorStride", 125}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-09739", 42}, { {"For each i greater than 0, pStrides[i-1] must be greater than or equal to pStrides[i] x pDimensions[i] so that no two elements of the tensor reference the same memory address", 174}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-09884", 42}, { {"pStrides[0] x pDimensions[0] must be less than or equal to VkPhysicalDeviceTensorPropertiesARM::maxTensorSize", 109}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-pStrides-parameter", 46}, { {"If pStrides is not NULL, pStrides must be a valid pointer to an array of dimensionCount int64_t values", 102}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_DESCRIPTION_ARM", 54}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-tiling-09741", 40}, { {"If tiling is VK_TENSOR_TILING_OPTIMAL_ARM and usage is VK_TENSOR_USAGE_IMAGE_ALIASING_BIT_ARM then the size of the tensor along its innermost dimension, i.e. pDimensions[dimensionCount - 1], must be less than or equal to 4", 222}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-tiling-09742", 40}, { {"If tiling is VK_TENSOR_TILING_LINEAR_ARM then VK_TENSOR_USAGE_IMAGE_ALIASING_BIT_ARM must not be set in usage", 109}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-tiling-09842", 40}, { {"If tiling is VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM then pDimensions[dimensionCount-1] must be less than or equal to 4", 168}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-tiling-09843", 40}, { {"If tiling is VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM, then pStrides must be NULL", 236}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-tiling-parameter", 44}, { {"tiling must be a valid VkTensorTilingARM value", 46}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-usage-parameter", 43}, { {"usage must be a valid combination of VkTensorUsageFlagBitsARM values", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorDescriptionARM-usage-requiredbitmask", 49}, { {"usage must not be 0", 19}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorExplicitTilingFormatPropertiesARM-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_EXPLICIT_TILING_FORMAT_PROPERTIES_ARM", 76}, { "chapters/formats.html", 21} } },
        { {"VUID-VkTensorFormatPropertiesARM-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_FORMAT_PROPERTIES_ARM", 60}, { "chapters/formats.html", 21} } },
        { {"VUID-VkTensorMemoryBarrierARM-dstAccessMask-parameter", 53}, { {"dstAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-dstStageMask-parameter", 52}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_MEMORY_BARRIER_ARM", 57}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-srcAccessMask-parameter", 53}, { {"srcAccessMask must be a valid combination of VkAccessFlagBits2 values", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-srcStageMask-parameter", 52}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits2 values", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-tensor-09755", 42}, { {"If tensor was created with a sharing mode of VK_SHARING_MODE_CONCURRENT, srcQueueFamilyIndex and dstQueueFamilyIndex must both be VK_QUEUE_FAMILY_IGNORED", 153}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-tensor-09756", 42}, { {"If tensor was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, srcQueueFamilyIndex and dstQueueFamilyIndex must both be either VK_QUEUE_FAMILY_IGNORED, or a valid queue family (see Queue Family Properties)", 214}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-tensor-09757", 42}, { {"If tensor was created with a sharing mode of VK_SHARING_MODE_EXCLUSIVE, and srcQueueFamilyIndex and dstQueueFamilyIndex are not VK_QUEUE_FAMILY_IGNORED, at least one of them must be the same as the family of the queue that will execute this barrier", 248}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-tensor-09758", 42}, { {"If tensor is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryBarrierARM-tensor-parameter", 46}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/synchronization.html", 29} } },
        { {"VUID-VkTensorMemoryRequirementsInfoARM-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorMemoryRequirementsInfoARM-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_MEMORY_REQUIREMENTS_INFO_ARM", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorMemoryRequirementsInfoARM-tensor-parameter", 55}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_ROLLING_BACKING_CREATE_INFO_ARM", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-wraps-09835", 52}, { {"For each i where i < VkTensorDescriptionARM::dimensionCount, wraps[i] must be less than or equal to VkTensorDescriptionARM::pDimensions[i]", 138}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-wraps-09836", 52}, { {"For each i where i < VkTensorDescriptionARM::dimensionCount, wraps[i] must either be equal to VkTensorDescriptionARM::pDimensions[i] or must be less than 2^16", 158}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-wraps-09837", 52}, { {"For each i where i < VkTensorDescriptionARM::dimensionCount, if wraps[i] is not equal to VkTensorDescriptionARM::pDimensions[i] and one element wraps[j] with j < i is not a power of two, then wraps[i] must be a power of two", 223}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-wraps-09838", 52}, { {"If VkTensorDescriptionARM::dimensionCount is greater than 1, then wraps[dimensionCount - 2] must be a multiple of brickOuterSize or must be equal to VkTensorDescriptionARM::pDimensions[dimensionCount - 2]", 204}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorRollingBackingCreateInfoARM-wraps-09839", 52}, { {"If brickOuterSize is not equal to 1, then wraps[dimensionCount - 1] must be a multiple of 64 / brickOuterSize / elementSize or be equal to VkTensorDescriptionARM::pDimensions[dimensionCount - 1]", 194}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCaptureDescriptorDataInfoARM-pNext-pNext", 57}, { {"pNext must be NULL", 18}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorViewCaptureDescriptorDataInfoARM-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_ARM", 76}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorViewCaptureDescriptorDataInfoARM-tensorView-09709", 62}, { {"If tensorView is not VK_NULL_HANDLE then tensorView must have been created with VK_TENSOR_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM set in VkTensorViewCreateInfoARM::flags", 182}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorViewCaptureDescriptorDataInfoARM-tensorView-parameter", 66}, { {"tensorView must be a valid VkTensorViewARM handle", 49}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-VkTensorViewCreateInfoARM-flags-09745", 42}, { {"If flags includes VK_TENSOR_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM, the descriptorBufferCaptureReplay feature must be enabled", 139}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-flags-parameter", 46}, { {"flags must be a valid combination of VkTensorViewCreateFlagBitsARM values", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-format-parameter", 47}, { {"format must be a valid VkFormat value", 37}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-pNext-09746", 42}, { {"If the pNext chain includes a VkOpaqueCaptureDescriptorDataCreateInfoEXT structure, flags must contain VK_TENSOR_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_ARM", 165}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-pNext-pNext", 42}, { {"pNext must be NULL or a pointer to a valid instance of VkOpaqueCaptureDescriptorDataCreateInfoEXT", 97}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_TENSOR_VIEW_CREATE_INFO_ARM", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-tensor-09743", 43}, { {"If tensor was not created with VK_TENSOR_CREATE_MUTABLE_FORMAT_BIT_ARM flag, format must be identical to the format used to create tensor", 137}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-tensor-09744", 43}, { {"If tensor was created with VK_TENSOR_CREATE_MUTABLE_FORMAT_BIT_ARM flag, format must be compatible with the format used to create tensor, as defined in Format Compatibility Classes", 180}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-tensor-09749", 43}, { {"If tensor is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-tensor-parameter", 47}, { {"tensor must be a valid VkTensorARM handle", 41}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-usage-09747", 42}, { {"The usage flags of tensor must have at least one of the following bits set: VK_TENSOR_USAGE_SHADER_BIT_ARM VK_TENSOR_USAGE_DATA_GRAPH_BIT_ARM", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTensorViewCreateInfoARM-usage-09748", 42}, { {"The tensor view's format features must contain the format feature flags required by the usage flags of tensor for format as indicated in the Format Feature Dependent Usage Flags section", 185}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTexelBufferDescriptorInfoEXT-format-parameter", 52}, { {"format must be a valid VkFormat value", 37}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkTexelBufferDescriptorInfoEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkTexelBufferDescriptorInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_TEXEL_BUFFER_DESCRIPTOR_INFO_EXT", 64}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-VkTextureLODGatherFormatPropertiesAMD-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_TEXTURE_LOD_GATHER_FORMAT_PROPERTIES_AMD", 72}, { "chapters/capabilities.html", 26} } },
        { {"VUID-VkThrottleHintSubmitInfoSEC-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_THROTTLE_HINT_SUBMIT_INFO_SEC", 61}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkThrottleHintSubmitInfoSEC-throttleHint-parameter", 55}, { {"throttleHint must be a valid VkThrottleHintTypeSEC value", 56}, { "appendices/extensions.html", 26} } },
        { {"VUID-VkTileMemoryBindInfoQCOM-memory-10726", 42}, { {"memory must have been allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 105}, { "chapters/memory.html", 20} } },
        { {"VUID-VkTileMemoryBindInfoQCOM-memory-parameter", 46}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-VkTileMemoryBindInfoQCOM-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_TILE_MEMORY_BIND_INFO_QCOM", 58}, { "chapters/memory.html", 20} } },
        { {"VUID-VkTileMemoryRequirementsQCOM-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_TILE_MEMORY_REQUIREMENTS_QCOM", 61}, { "chapters/resources.html", 23} } },
        { {"VUID-VkTileMemorySizeInfoQCOM-sType-sType", 41}, { {"sType must be VK_STRUCTURE_TYPE_TILE_MEMORY_SIZE_INFO_QCOM", 58}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkTileMemorySizeInfoQCOM-size-10729", 40}, { {"size must be less than or equal to the largest size memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkTilePropertiesQCOM-pNext-pNext", 37}, { {"pNext must be NULL", 18}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkTilePropertiesQCOM-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_TILE_PROPERTIES_QCOM", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-VkTimelineSemaphoreSubmitInfo-pSignalSemaphoreValues-parameter", 67}, { {"If signalSemaphoreValueCount is not 0, and pSignalSemaphoreValues is not NULL, pSignalSemaphoreValues must be a valid pointer to an array of signalSemaphoreValueCount uint64_t values", 182}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkTimelineSemaphoreSubmitInfo-pWaitSemaphoreValues-parameter", 65}, { {"If waitSemaphoreValueCount is not 0, and pWaitSemaphoreValues is not NULL, pWaitSemaphoreValues must be a valid pointer to an array of waitSemaphoreValueCount uint64_t values", 174}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkTimelineSemaphoreSubmitInfo-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO", 62}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-callableShaderBindingTableAddress-parameter", 79}, { {"callableShaderBindingTableAddress must be a valid VkDeviceAddress value", 71}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-depth-03640", 47}, { {"depth must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[2]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-flags-03511", 47}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR, the entries in the table identified by missShaderBindingTableAddress accessed as a result of this command in order to execute a miss shader must not be zero", 288}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-flags-03512", 47}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR, entries in the table identified by hitShaderBindingTableAddress accessed as a result of this command in order to execute an any-hit shader must not be zero", 290}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-flags-03513", 47}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR, entries in the table identified by hitShaderBindingTableAddress accessed as a result of this command in order to execute a closest hit shader must not be zero", 297}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-flags-03514", 47}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR, entries in the table identified by hitShaderBindingTableAddress accessed as a result of this command in order to execute an intersection shader must not be zero", 300}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-height-03639", 48}, { {"height must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[1]", 141}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-hitShaderBindingTableAddress-parameter", 74}, { {"hitShaderBindingTableAddress must be a valid VkDeviceAddress value", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-missShaderBindingTableAddress-parameter", 75}, { {"missShaderBindingTableAddress must be a valid VkDeviceAddress value", 67}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pCallableShaderBindingTable-03692", 69}, { {"callableShaderBindingTableAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 178}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pCallableShaderBindingTable-03693", 69}, { {"callableShaderBindingTableAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 129}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pHitShaderBindingTable-03688", 64}, { {"hitShaderBindingTableAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 173}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pHitShaderBindingTable-03689", 64}, { {"hitShaderBindingTableAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 124}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pHitShaderBindingTable-04735", 64}, { {"Any non-zero hit shader group entries in the table identified by hitShaderBindingTableAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR", 270}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pHitShaderBindingTable-04736", 64}, { {"Any non-zero hit shader group entries in the table identified by hitShaderBindingTableAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR", 267}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pMissShaderBindingTable-03684", 65}, { {"missShaderBindingTableAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 174}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pMissShaderBindingTable-03685", 65}, { {"missShaderBindingTableAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 125}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pRayGenShaderBindingTable-03681", 67}, { {"raygenShaderRecordAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 170}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-pRayGenShaderBindingTable-03682", 67}, { {"raygenShaderRecordAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 121}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-raygenShaderRecordAddress-parameter", 71}, { {"raygenShaderRecordAddress must be a valid VkDeviceAddress value", 63}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-03686", 48}, { {"missShaderBindingTableStride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 126}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-03690", 48}, { {"hitShaderBindingTableStride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 125}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-03694", 48}, { {"callableShaderBindingTableStride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 130}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-04029", 48}, { {"missShaderBindingTableStride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 128}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-04035", 48}, { {"hitShaderBindingTableStride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 127}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-stride-04041", 48}, { {"callableShaderBindingTableStride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 132}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-width-03638", 47}, { {"width must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[0]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommand2KHR-width-03641", 47}, { {"width x height x depth must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxRayDispatchInvocationCount", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommandKHR-depth-03640", 46}, { {"depth must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[2]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommandKHR-height-03639", 47}, { {"height must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[1]", 141}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommandKHR-width-03638", 46}, { {"width must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[0]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTraceRaysIndirectCommandKHR-width-03641", 46}, { {"width x height x depth must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxRayDispatchInvocationCount", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-VkTransformMatrixKHR-matrix-03799", 38}, { {"The first three columns of matrix must define an invertible 3x3 matrix", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-VkUbmSurfaceCreateInfoSEC-device-12366", 43}, { {"device must point to a valid UBM ubm_device", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkUbmSurfaceCreateInfoSEC-flags-zerobitmask", 48}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkUbmSurfaceCreateInfoSEC-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkUbmSurfaceCreateInfoSEC-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_UBM_SURFACE_CREATE_INFO_SEC", 59}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkUbmSurfaceCreateInfoSEC-surface-12367", 44}, { {"surface must point to a valid UBM ubm_surface", 45}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-flags-zerobitmask", 53}, { {"flags must be 0", 15}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-initialDataSize-01534", 57}, { {"If initialDataSize is not 0, it must be equal to the size of pInitialData, as returned by vkGetValidationCacheDataEXT when pInitialData was originally retrieved", 160}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-initialDataSize-01535", 57}, { {"If initialDataSize is not 0, pInitialData must have been retrieved from a previous call to vkGetValidationCacheDataEXT", 118}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-pInitialData-parameter", 58}, { {"If initialDataSize is not 0, pInitialData must be a valid pointer to an array of initialDataSize bytes", 102}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-pNext-pNext", 47}, { {"pNext must be NULL", 18}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationCacheCreateInfoEXT-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VALIDATION_CACHE_CREATE_INFO_EXT", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-VkValidationFeaturesEXT-pDisabledValidationFeatures-parameter", 66}, { {"If disabledValidationFeatureCount is not 0, pDisabledValidationFeatures must be a valid pointer to an array of disabledValidationFeatureCount valid VkValidationFeatureDisableEXT values", 184}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFeaturesEXT-pEnabledValidationFeatures-02967", 61}, { {"If the pEnabledValidationFeatures array contains VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_RESERVE_BINDING_SLOT_EXT, then it must also contain VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_EXT or VK_VALIDATION_FEATURE_ENABLE_DEBUG_PRINTF_EXT", 237}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFeaturesEXT-pEnabledValidationFeatures-parameter", 65}, { {"If enabledValidationFeatureCount is not 0, pEnabledValidationFeatures must be a valid pointer to an array of enabledValidationFeatureCount valid VkValidationFeatureEnableEXT values", 180}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFeaturesEXT-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT", 55}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFlagsEXT-disabledValidationCheckCount-arraylength", 66}, { {"disabledValidationCheckCount must be greater than 0", 51}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFlagsEXT-pDisabledValidationChecks-parameter", 61}, { {"pDisabledValidationChecks must be a valid pointer to an array of disabledValidationCheckCount valid VkValidationCheckEXT values", 127}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkValidationFlagsEXT-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_VALIDATION_FLAGS_EXT", 52}, { "chapters/initialization.html", 28} } },
        { {"VUID-VkVertexInputAttributeDescription-binding-00621", 52}, { {"binding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription-format-00623", 51}, { {"The format features of format must contain VK_FORMAT_FEATURE_VERTEX_BUFFER_BIT", 78}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription-format-parameter", 55}, { {"format must be a valid VkFormat value", 37}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription-location-00620", 53}, { {"location must be less than VkPhysicalDeviceLimits::maxVertexInputAttributes", 75}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription-offset-00622", 51}, { {"offset must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputAttributeOffset", 90}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription-vertexAttributeAccessBeyondStride-04457", 78}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::vertexAttributeAccessBeyondStride is VK_FALSE, the sum of offset plus the size of the vertex attribute data described by format must not be greater than stride in the VkVertexInputBindingDescription referenced in binding", 325}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-binding-06229", 56}, { {"binding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-format-04805", 55}, { {"The format features of format must contain VK_FORMAT_FEATURE_VERTEX_BUFFER_BIT", 78}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-format-parameter", 59}, { {"format must be a valid VkFormat value", 37}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-location-06228", 57}, { {"location must be less than VkPhysicalDeviceLimits::maxVertexInputAttributes", 75}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-offset-06230", 55}, { {"offset must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputAttributeOffset", 90}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VERTEX_INPUT_ATTRIBUTE_DESCRIPTION_2_EXT", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputAttributeDescription2EXT-vertexAttributeAccessBeyondStride-04806", 82}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::vertexAttributeAccessBeyondStride is VK_FALSE, the sum of offset plus the size of the vertex attribute data described by format must not be greater than stride in the VkVertexInputBindingDescription2EXT referenced in binding", 329}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription-binding-00618", 50}, { {"binding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription-inputRate-parameter", 56}, { {"inputRate must be a valid VkVertexInputRate value", 49}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription-stride-00619", 49}, { {"stride must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindingStride", 88}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription-stride-04456", 49}, { {"If the VK_KHR_portability_subset extension is enabled, stride must be a multiple of, and at least as large as, VkPhysicalDevicePortabilitySubsetPropertiesKHR::minVertexInputBindingStrideAlignment", 195}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-binding-04796", 54}, { {"binding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-divisor-04798", 54}, { {"If the vertexAttributeInstanceRateZeroDivisor feature is not enabled, divisor must not be 0", 91}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-divisor-04799", 54}, { {"If the vertexAttributeInstanceRateDivisor feature is not enabled, divisor must be 1", 83}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-divisor-06226", 54}, { {"divisor must be a value between 0 and VkPhysicalDeviceVertexAttributeDivisorPropertiesEXT::maxVertexAttribDivisor, inclusive", 124}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-divisor-06227", 54}, { {"If divisor is not 1 then inputRate must be of type VK_VERTEX_INPUT_RATE_INSTANCE", 80}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-inputRate-parameter", 60}, { {"inputRate must be a valid VkVertexInputRate value", 49}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VERTEX_INPUT_BINDING_DESCRIPTION_2_EXT", 70}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDescription2EXT-stride-04797", 53}, { {"stride must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindingStride", 88}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDivisorDescription-binding-01869", 57}, { {"binding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 72}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDivisorDescription-divisor-01870", 57}, { {"divisor must be a value between 0 and VkPhysicalDeviceVertexAttributeDivisorProperties::maxVertexAttribDivisor, inclusive", 121}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDivisorDescription-inputRate-01871", 59}, { {"VkVertexInputBindingDescription::inputRate must be of type VK_VERTEX_INPUT_RATE_INSTANCE for this binding", 105}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDivisorDescription-vertexAttributeInstanceRateDivisor-02229", 84}, { {"If the vertexAttributeInstanceRateDivisor feature is not enabled, divisor must be 1", 83}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkVertexInputBindingDivisorDescription-vertexAttributeInstanceRateZeroDivisor-02228", 88}, { {"If the vertexAttributeInstanceRateZeroDivisor feature is not enabled, divisor must not be 0", 91}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-VkViSurfaceCreateInfoNN-flags-zerobitmask", 46}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkViSurfaceCreateInfoNN-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkViSurfaceCreateInfoNN-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_VI_SURFACE_CREATE_INFO_NN", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkViSurfaceCreateInfoNN-window-01318", 41}, { {"window must be a valid nn::vi::NativeWindowHandle", 49}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-commonparent", 43}, { {"Both of videoSession, and videoSessionParameters that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 168}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-flags-07244", 42}, { {"If VkVideoCapabilitiesKHR::flags does not include VK_VIDEO_CAPABILITY_SEPARATE_REFERENCE_IMAGES_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSession was created with, then pPictureResource->imageViewBinding of all elements of pReferenceSlots with a non-NULL pPictureResource member must specify image views created from the same image", 380}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-flags-zerobitmask", 48}, { {"flags must be 0", 15}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pNext-pNext", 42}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1GopRemainingFrameInfoKHR, VkVideoEncodeAV1RateControlInfoKHR, VkVideoEncodeH264GopRemainingFrameInfoKHR, VkVideoEncodeH264RateControlInfoKHR, VkVideoEncodeH265GopRemainingFrameInfoKHR, VkVideoEncodeH265RateControlInfoKHR, or VkVideoEncodeRateControlInfoKHR", 403}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pPictureResource-07238", 53}, { {"Each video picture resource corresponding to any non-NULL pPictureResource member specified in the elements of pReferenceSlots must be unique within pReferenceSlots", 164}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pPictureResource-07240", 53}, { {"If the pPictureResource member of any element of pReferenceSlots is not NULL, then the image view specified in pPictureResource->imageViewBinding for that element must be compatible with the video profile videoSession was created with", 234}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pPictureResource-07241", 53}, { {"If the pPictureResource member of any element of pReferenceSlots is not NULL, then the format of the image view specified in pPictureResource->imageViewBinding for that element must match the VkVideoSessionCreateInfoKHR::referencePictureFormat videoSession was created with", 273}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pPictureResource-07242", 53}, { {"If the pPictureResource member of any element of pReferenceSlots is not NULL, then its codedOffset member must be an integer multiple of codedOffsetGranularity", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pPictureResource-07243", 53}, { {"If the pPictureResource member of any element of pReferenceSlots is not NULL, then its codedExtent member must be between minCodedExtent and maxCodedExtent, inclusive, videoSession was created with", 197}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-pReferenceSlots-parameter", 56}, { {"If referenceSlotCount is not 0, pReferenceSlots must be a valid pointer to an array of referenceSlotCount valid VkVideoReferenceSlotInfoKHR structures", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_BEGIN_CODING_INFO_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-sType-unique", 43}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-slotIndex-04856", 46}, { {"Each non-negative VkVideoReferenceSlotInfoKHR::slotIndex specified in the elements of pReferenceSlots must be less than the VkVideoSessionCreateInfoKHR::maxDpbSlots specified when videoSession was created", 204}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-slotIndex-07245", 46}, { {"If videoSession was created with a decode operation and the slotIndex member of any element of pReferenceSlots is not negative, then the image view specified in pPictureResource->imageViewBinding for that element must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR usage flag set", 299}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-slotIndex-07246", 46}, { {"If videoSession was created with an encode operation and the slotIndex member of any element of pReferenceSlots is not negative, then the image view specified in pPictureResource->imageViewBinding for that element must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR usage flag set", 300}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-07237", 49}, { {"videoSession must have memory bound to all of its memory bindings returned by vkGetVideoSessionMemoryRequirementsKHR for videoSession", 133}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-07247", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE , unless the videoMaintenance2 feature is enabled", 209}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-07248", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE , unless the videoMaintenance2 feature is enabled", 209}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-07249", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-07250", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-09261", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE , unless the videoMaintenance2 feature is enabled", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-10283", 49}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then videoSessionParameters must not be VK_NULL_HANDLE", 158}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSession-parameter", 53}, { {"videoSession must be a valid VkVideoSessionKHR handle", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSessionParameters-04857", 59}, { {"If videoSessionParameters is not VK_NULL_HANDLE, it must have been created with videoSession specified in VkVideoSessionParametersCreateInfoKHR::videoSession", 157}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoBeginCodingInfoKHR-videoSessionParameters-parameter", 63}, { {"If videoSessionParameters is not VK_NULL_HANDLE, videoSessionParameters must be a valid VkVideoSessionParametersKHR handle", 122}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCapabilitiesKHR-pNext-pNext", 39}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoDecodeAV1CapabilitiesKHR, VkVideoDecodeCapabilitiesKHR, VkVideoDecodeH264CapabilitiesKHR, VkVideoDecodeH265CapabilitiesKHR, VkVideoDecodeVP9CapabilitiesKHR, VkVideoEncodeAV1CapabilitiesKHR, VkVideoEncodeAV1QuantizationMapCapabilitiesKHR, VkVideoEncodeCapabilitiesKHR, VkVideoEncodeFeedback2CapabilitiesKHR, VkVideoEncodeH264CapabilitiesKHR, VkVideoEncodeH264QuantizationMapCapabilitiesKHR, VkVideoEncodeH265CapabilitiesKHR, VkVideoEncodeH265QuantizationMapCapabilitiesKHR, VkVideoEncodeIntraRefreshCapabilitiesKHR, VkVideoEncodeQuantizationMapCapabilitiesKHR, or VkVideoEncodeRgbConversionCapabilitiesVALVE", 744}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCapabilitiesKHR-sType-sType", 39}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_CAPABILITIES_KHR", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCapabilitiesKHR-sType-unique", 40}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-flags-07018", 44}, { {"If flags includes VK_VIDEO_CODING_CONTROL_ENCODE_RATE_CONTROL_BIT_KHR, then the pNext chain must include a VkVideoEncodeRateControlInfoKHR structure", 148}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-flags-08349", 44}, { {"If flags includes VK_VIDEO_CODING_CONTROL_ENCODE_QUALITY_LEVEL_BIT_KHR, then the pNext chain must include a VkVideoEncodeQualityLevelInfoKHR structure", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-flags-parameter", 48}, { {"flags must be a valid combination of VkVideoCodingControlFlagBitsKHR values", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-flags-requiredbitmask", 54}, { {"flags must not be 0", 19}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1RateControlInfoKHR, VkVideoEncodeH264RateControlInfoKHR, VkVideoEncodeH265RateControlInfoKHR, VkVideoEncodeQualityLevelInfoKHR, or VkVideoEncodeRateControlInfoKHR", 309}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_CODING_CONTROL_INFO_KHR", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoCodingControlInfoKHR-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1CapabilitiesKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_CAPABILITIES_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1DpbSlotInfoKHR-pStdReferenceInfo-parameter", 63}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoDecodeAV1ReferenceInfo value", 89}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1DpbSlotInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_DPB_SLOT_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1InlineSessionParametersInfoKHR-pStdSequenceHeader-parameter", 80}, { {"If pStdSequenceHeader is not NULL, pStdSequenceHeader must be a valid pointer to a valid StdVideoAV1SequenceHeader value", 120}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1InlineSessionParametersInfoKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_INLINE_SESSION_PARAMETERS_INFO_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1PictureInfoKHR-pStdPictureInfo-parameter", 61}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoDecodeAV1PictureInfo value", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1PictureInfoKHR-pTileOffsets-parameter", 58}, { {"pTileOffsets must be a valid pointer to an array of tileCount uint32_t values", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1PictureInfoKHR-pTileSizes-parameter", 56}, { {"pTileSizes must be a valid pointer to an array of tileCount uint32_t values", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1PictureInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_PICTURE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1PictureInfoKHR-tileCount-arraylength", 57}, { {"tileCount must be greater than 0", 32}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1ProfileInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_PROFILE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1SessionParametersCreateInfoKHR-pStdSequenceHeader-parameter", 80}, { {"pStdSequenceHeader must be a valid pointer to a valid StdVideoAV1SequenceHeader value", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeAV1SessionParametersCreateInfoKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_SESSION_PARAMETERS_CREATE_INFO_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeCapabilitiesKHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_CAPABILITIES_KHR", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264CapabilitiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_CAPABILITIES_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264DpbSlotInfoKHR-pStdReferenceInfo-parameter", 64}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoDecodeH264ReferenceInfo value", 90}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264DpbSlotInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_DPB_SLOT_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264InlineSessionParametersInfoKHR-pStdPPS-parameter", 70}, { {"If pStdPPS is not NULL, pStdPPS must be a valid pointer to a valid StdVideoH264PictureParameterSet value", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264InlineSessionParametersInfoKHR-pStdSPS-parameter", 70}, { {"If pStdSPS is not NULL, pStdSPS must be a valid pointer to a valid StdVideoH264SequenceParameterSet value", 105}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264InlineSessionParametersInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_INLINE_SESSION_PARAMETERS_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264PictureInfoKHR-pSliceOffsets-parameter", 60}, { {"pSliceOffsets must be a valid pointer to an array of sliceCount uint32_t values", 79}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264PictureInfoKHR-pStdPictureInfo-parameter", 62}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoDecodeH264PictureInfo value", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264PictureInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PICTURE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264PictureInfoKHR-sliceCount-arraylength", 59}, { {"sliceCount must be greater than 0", 33}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264ProfileInfoKHR-pictureLayout-parameter", 60}, { {"If pictureLayout is not 0, pictureLayout must be a valid VkVideoDecodeH264PictureLayoutFlagBitsKHR value", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264ProfileInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PROFILE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersAddInfoKHR-None-04825", 60}, { {"The seq_parameter_set_id member of each StdVideoH264SequenceParameterSet structure specified in the elements of pStdSPSs must be unique within pStdSPSs", 151}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersAddInfoKHR-None-04826", 60}, { {"The pair constructed from the seq_parameter_set_id and pic_parameter_set_id members of each StdVideoH264PictureParameterSet structure specified in the elements of pStdPPSs must be unique within pStdPPSs", 202}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersAddInfoKHR-pStdPPSs-parameter", 68}, { {"If stdPPSCount is not 0, pStdPPSs must be a valid pointer to an array of stdPPSCount StdVideoH264PictureParameterSet values", 123}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersAddInfoKHR-pStdSPSs-parameter", 68}, { {"If stdSPSCount is not 0, pStdSPSs must be a valid pointer to an array of stdSPSCount StdVideoH264SequenceParameterSet values", 124}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersAddInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersCreateInfoKHR-pParametersAddInfo-parameter", 81}, { {"If pParametersAddInfo is not NULL, pParametersAddInfo must be a valid pointer to a valid VkVideoDecodeH264SessionParametersAddInfoKHR structure", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH264SessionParametersCreateInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265CapabilitiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_CAPABILITIES_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265DpbSlotInfoKHR-pStdReferenceInfo-parameter", 64}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoDecodeH265ReferenceInfo value", 90}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265DpbSlotInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_DPB_SLOT_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265InlineSessionParametersInfoKHR-pStdPPS-parameter", 70}, { {"If pStdPPS is not NULL, pStdPPS must be a valid pointer to a valid StdVideoH265PictureParameterSet value", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265InlineSessionParametersInfoKHR-pStdSPS-parameter", 70}, { {"If pStdSPS is not NULL, pStdSPS must be a valid pointer to a valid StdVideoH265SequenceParameterSet value", 105}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265InlineSessionParametersInfoKHR-pStdVPS-parameter", 70}, { {"If pStdVPS is not NULL, pStdVPS must be a valid pointer to a valid StdVideoH265VideoParameterSet value", 102}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265InlineSessionParametersInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_INLINE_SESSION_PARAMETERS_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265PictureInfoKHR-pSliceSegmentOffsets-parameter", 67}, { {"pSliceSegmentOffsets must be a valid pointer to an array of sliceSegmentCount uint32_t values", 93}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265PictureInfoKHR-pStdPictureInfo-parameter", 62}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoDecodeH265PictureInfo value", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265PictureInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PICTURE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265PictureInfoKHR-sliceSegmentCount-arraylength", 66}, { {"sliceSegmentCount must be greater than 0", 40}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265ProfileInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PROFILE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-None-04833", 60}, { {"The vps_video_parameter_set_id member of each StdVideoH265VideoParameterSet structure specified in the elements of pStdVPSs must be unique within pStdVPSs", 154}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-None-04834", 60}, { {"The pair constructed from the sps_video_parameter_set_id and sps_seq_parameter_set_id members of each StdVideoH265SequenceParameterSet structure specified in the elements of pStdSPSs must be unique within pStdSPSs", 213}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-None-04835", 60}, { {"The triplet constructed from the sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id members of each StdVideoH265PictureParameterSet structure specified in the elements of pStdPPSs must be unique within pStdPPSs", 242}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-pStdPPSs-parameter", 68}, { {"If stdPPSCount is not 0, pStdPPSs must be a valid pointer to an array of stdPPSCount StdVideoH265PictureParameterSet values", 123}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-pStdSPSs-parameter", 68}, { {"If stdSPSCount is not 0, pStdSPSs must be a valid pointer to an array of stdSPSCount StdVideoH265SequenceParameterSet values", 124}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-pStdVPSs-parameter", 68}, { {"If stdVPSCount is not 0, pStdVPSs must be a valid pointer to an array of stdVPSCount StdVideoH265VideoParameterSet values", 121}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersAddInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersCreateInfoKHR-pParametersAddInfo-parameter", 81}, { {"If pParametersAddInfo is not NULL, pParametersAddInfo must be a valid pointer to a valid VkVideoDecodeH265SessionParametersAddInfoKHR structure", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeH265SessionParametersCreateInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-dstPictureResource-parameter", 54}, { {"dstPictureResource must be a valid VkVideoPictureResourceInfoKHR structure", 74}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-flags-zerobitmask", 43}, { {"flags must be 0", 15}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoDecodeAV1InlineSessionParametersInfoKHR, VkVideoDecodeAV1PictureInfoKHR, VkVideoDecodeH264InlineSessionParametersInfoKHR, VkVideoDecodeH264PictureInfoKHR, VkVideoDecodeH265InlineSessionParametersInfoKHR, VkVideoDecodeH265PictureInfoKHR, VkVideoDecodeVP9PictureInfoKHR, or VkVideoInlineQueryInfoKHR", 435}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pPictureResource-07172", 48}, { {"The pPictureResource member of each element of pReferenceSlots must not be NULL", 79}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pReferenceSlots-parameter", 51}, { {"If referenceSlotCount is not 0, pReferenceSlots must be a valid pointer to an array of referenceSlotCount valid VkVideoReferenceSlotInfoKHR structures", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pSetupReferenceSlot-07168", 51}, { {"If pSetupReferenceSlot is not NULL, then its slotIndex member must not be negative", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pSetupReferenceSlot-07169", 51}, { {"If pSetupReferenceSlot is not NULL, then its pPictureResource must not be NULL", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-pSetupReferenceSlot-parameter", 55}, { {"If pSetupReferenceSlot is not NULL, pSetupReferenceSlot must be a valid pointer to a valid VkVideoReferenceSlotInfoKHR structure", 128}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_INFO_KHR", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-slotIndex-07171", 41}, { {"The slotIndex member of each element of pReferenceSlots must not be negative", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-srcBuffer-07165", 41}, { {"srcBuffer must have been created with the VK_BUFFER_USAGE_VIDEO_DECODE_SRC_BIT_KHR usage flag set", 97}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-srcBuffer-parameter", 45}, { {"srcBuffer must be a valid VkBuffer handle", 41}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-srcBufferOffset-07166", 47}, { {"srcBufferOffset must be less than the size of srcBuffer", 55}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeInfoKHR-srcBufferRange-07167", 46}, { {"srcBufferRange must be less than or equal to the size of srcBuffer minus srcBufferOffset", 88}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeUsageInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_USAGE_INFO_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeUsageInfoKHR-videoUsageHints-parameter", 56}, { {"videoUsageHints must be a valid combination of VkVideoDecodeUsageFlagBitsKHR values", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeVP9CapabilitiesKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_CAPABILITIES_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeVP9PictureInfoKHR-pStdPictureInfo-parameter", 61}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoDecodeVP9PictureInfo value", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeVP9PictureInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_PICTURE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoDecodeVP9ProfileInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_PROFILE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1CapabilitiesKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_CAPABILITIES_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1DpbSlotInfoKHR-pStdReferenceInfo-parameter", 63}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoEncodeAV1ReferenceInfo value", 89}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1DpbSlotInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_DPB_SLOT_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1GopRemainingFrameInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_GOP_REMAINING_FRAME_INFO_KHR", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-flags-10289", 47}, { {"If VkVideoEncodeAV1CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_PRIMARY_REFERENCE_CDF_ONLY_BIT_KHR, then primaryReferenceCdfOnly must be VK_FALSE", 255}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-flags-10292", 47}, { {"If VkVideoEncodeAV1CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_GENERATE_OBU_EXTENSION_HEADER_BIT_KHR, then generateObuExtensionHeader must be VK_FALSE", 261}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-generateObuExtensionHeader-10293", 68}, { {"If generateObuExtensionHeader is set to VK_TRUE, then pStdPictureInfo->pExtensionHeader must not be NULL", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-pStdPictureInfo-10291", 57}, { {"If pStdPictureInfo->primary_ref_frame is less than VK_MAX_VIDEO_AV1_REFERENCES_PER_FRAME_KHR, then referenceNameSlotIndices[pStdPictureInfo->primary_ref_frame] must not be negative", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-pStdPictureInfo-parameter", 61}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoEncodeAV1PictureInfo value", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-predictionMode-parameter", 60}, { {"predictionMode must be a valid VkVideoEncodeAV1PredictionModeKHR value", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-primaryReferenceCdfOnly-10290", 65}, { {"If primaryReferenceCdfOnly is set to VK_TRUE, then pStdPictureInfo->primary_ref_frame must be less than VK_MAX_VIDEO_AV1_REFERENCES_PER_FRAME_KHR", 145}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-rateControlGroup-parameter", 62}, { {"rateControlGroup must be a valid VkVideoEncodeAV1RateControlGroupKHR value", 74}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1PictureInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_PICTURE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1ProfileInfoKHR-sType-sType", 47}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_PROFILE_INFO_KHR", 65}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1QualityLevelPropertiesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_QUALITY_LEVEL_PROPERTIES_KHR", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1QuantizationMapCapabilitiesKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_QUANTIZATION_MAP_CAPABILITIES_KHR", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-consecutiveBipredictiveFrameCount-10298", 79}, { {"If consecutiveBipredictiveFrameCount is not 0, then it must be less than gopFrameCount", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-flags-10294", 51}, { {"If flags contains VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR or VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR, then it must also contain VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REGULAR_GOP_BIT_KHR", 230}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-flags-10295", 51}, { {"If flags contains VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR, then it must not also contain VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR", 178}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-flags-10296", 51}, { {"If flags contains VK_VIDEO_ENCODE_AV1_RATE_CONTROL_REGULAR_GOP_BIT_KHR, then gopFrameCount must be greater than 0", 113}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-flags-parameter", 55}, { {"flags must be a valid combination of VkVideoEncodeAV1RateControlFlagBitsKHR values", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-keyFramePeriod-10297", 60}, { {"If keyFramePeriod is not 0, then it must be greater than or equal to gopFrameCount", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-sType-sType", 51}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_RATE_CONTROL_INFO_KHR", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlInfoKHR-temporalLayerCount-10299", 64}, { {"temporalLayerCount must be less than or equal to VkVideoEncodeAV1CapabilitiesKHR::maxTemporalLayerCount, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 186}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-maxFrameSize-parameter", 67}, { {"maxFrameSize must be a valid VkVideoEncodeAV1FrameSizeKHR structure", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-maxQIndex-parameter", 64}, { {"maxQIndex must be a valid VkVideoEncodeAV1QIndexKHR structure", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-minQIndex-parameter", 64}, { {"minQIndex must be a valid VkVideoEncodeAV1QIndexKHR structure", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-sType-sType", 56}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_RATE_CONTROL_LAYER_INFO_KHR", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-useMaxQIndex-10302", 63}, { {"If useMaxQIndex is VK_TRUE, then the intraQIndex, predictiveQIndex, and bipredictiveQIndex members of maxQIndex must all be between VkVideoEncodeAV1CapabilitiesKHR::minQIndex and VkVideoEncodeAV1CapabilitiesKHR::maxQIndex, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 304}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-useMaxQIndex-10303", 63}, { {"If useMaxQIndex is VK_TRUE and VkVideoEncodeAV1CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_PER_RATE_CONTROL_GROUP_MIN_MAX_Q_INDEX_BIT_KHR, then the intraQIndex, predictiveQIndex, and bipredictiveQIndex members of maxQIndex must all specify the same value", 365}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-useMinQIndex-10300", 63}, { {"If useMinQIndex is VK_TRUE, then the intraQIndex, predictiveQIndex, and bipredictiveQIndex members of minQIndex must all be between VkVideoEncodeAV1CapabilitiesKHR::minQIndex and VkVideoEncodeAV1CapabilitiesKHR::maxQIndex, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 304}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-useMinQIndex-10301", 63}, { {"If useMinQIndex is VK_TRUE and VkVideoEncodeAV1CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_PER_RATE_CONTROL_GROUP_MIN_MAX_Q_INDEX_BIT_KHR, then the intraQIndex, predictiveQIndex, and bipredictiveQIndex members of minQIndex must all specify the same value", 365}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1RateControlLayerInfoKHR-useMinQIndex-10304", 63}, { {"If useMinQIndex and useMaxQIndex are both VK_TRUE, then the intraQIndex, predictiveQIndex, and bipredictiveQIndex members of minQIndex must all be less than or equal to the respective members of maxQIndex", 204}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionCreateInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_SESSION_CREATE_INFO_KHR", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionParametersCreateInfoKHR-pStdDecoderModelInfo-parameter", 82}, { {"If pStdDecoderModelInfo is not NULL, pStdDecoderModelInfo must be a valid pointer to a valid StdVideoEncodeAV1DecoderModelInfo value", 132}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionParametersCreateInfoKHR-pStdOperatingPoints-parameter", 81}, { {"If stdOperatingPointCount is not 0, and pStdOperatingPoints is not NULL, pStdOperatingPoints must be a valid pointer to an array of stdOperatingPointCount StdVideoEncodeAV1OperatingPointInfo values", 197}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionParametersCreateInfoKHR-pStdSequenceHeader-10288", 76}, { {"pStdSequenceHeader->flags.film_grain_params_present must be zero", 64}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionParametersCreateInfoKHR-pStdSequenceHeader-parameter", 80}, { {"pStdSequenceHeader must be a valid pointer to a valid StdVideoAV1SequenceHeader value", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeAV1SessionParametersCreateInfoKHR-sType-sType", 63}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_SESSION_PARAMETERS_CREATE_INFO_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeCapabilitiesKHR-sType-sType", 45}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_CAPABILITIES_KHR", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeFeedback2CapabilitiesKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_FEEDBACK_2_CAPABILITIES_KHR", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeFeedback2CapabilitiesKHR-supportedPerPartitionEncodeFeedbackFlags-parameter", 93}, { {"supportedPerPartitionEncodeFeedbackFlags must be a valid combination of VkVideoEncodePerPartitionFeedbackFlagBitsKHR values", 123}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeFeedback2CapabilitiesKHR-supportedPerPartitionEncodeFeedbackFlags-requiredbitmask", 99}, { {"supportedPerPartitionEncodeFeedbackFlags must not be 0", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264CapabilitiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_CAPABILITIES_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264DpbSlotInfoKHR-pStdReferenceInfo-parameter", 64}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoEncodeH264ReferenceInfo value", 90}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264DpbSlotInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_DPB_SLOT_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264GopRemainingFrameInfoKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_GOP_REMAINING_FRAME_INFO_KHR", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264NaluSliceInfoKHR-pNext-pNext", 50}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264NaluSliceInfoKHR-pStdSliceHeader-parameter", 64}, { {"pStdSliceHeader must be a valid pointer to a valid StdVideoEncodeH264SliceHeader value", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264NaluSliceInfoKHR-sType-sType", 50}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_NALU_SLICE_INFO_KHR", 69}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-flags-08304", 48}, { {"If VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_GENERATE_PREFIX_NALU_BIT_KHR, then generatePrefixNalu must be VK_FALSE", 246}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-flags-08314", 48}, { {"If VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_PREDICTION_WEIGHT_TABLE_GENERATED_BIT_KHR and the slice corresponding to any element of pNaluSliceEntries uses explicit weighted sample prediction, then VkVideoEncodeH264NaluSliceInfoKHR::pStdSliceHeader->pWeightTable must not be NULL for that element of pNaluSliceEntries", 448}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-naluSliceEntryCount-08301", 62}, { {"naluSliceEntryCount must be between 1 and VkVideoEncodeH264CapabilitiesKHR::maxSliceCount, inclusive, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 183}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-naluSliceEntryCount-arraylength", 68}, { {"naluSliceEntryCount must be greater than 0", 42}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-pNaluSliceEntries-parameter", 64}, { {"pNaluSliceEntries must be a valid pointer to an array of naluSliceEntryCount valid VkVideoEncodeH264NaluSliceInfoKHR structures", 127}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-pStdPictureInfo-parameter", 62}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoEncodeH264PictureInfo value", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264PictureInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PICTURE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264ProfileInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PROFILE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264QualityLevelPropertiesKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUALITY_LEVEL_PROPERTIES_KHR", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264QuantizationMapCapabilitiesKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUANTIZATION_MAP_CAPABILITIES_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-consecutiveBFrameCount-08285", 69}, { {"If consecutiveBFrameCount is not 0, then it must be less than gopFrameCount", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-flags-08280", 52}, { {"If VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_HRD_COMPLIANCE_BIT_KHR, then flags must not contain VK_VIDEO_ENCODE_H264_RATE_CONTROL_ATTEMPT_HRD_COMPLIANCE_BIT_KHR", 292}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-flags-08281", 52}, { {"If flags contains VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR or VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR, then it must also contain VK_VIDEO_ENCODE_H264_RATE_CONTROL_REGULAR_GOP_BIT_KHR", 233}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-flags-08282", 52}, { {"If flags contains VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR, then it must not also contain VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-flags-08283", 52}, { {"If flags contains VK_VIDEO_ENCODE_H264_RATE_CONTROL_REGULAR_GOP_BIT_KHR, then gopFrameCount must be greater than 0", 114}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-flags-parameter", 56}, { {"flags must be a valid combination of VkVideoEncodeH264RateControlFlagBitsKHR values", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-idrPeriod-08284", 56}, { {"If idrPeriod is not 0, then it must be greater than or equal to gopFrameCount", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_INFO_KHR", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-maxFrameSize-parameter", 68}, { {"maxFrameSize must be a valid VkVideoEncodeH264FrameSizeKHR structure", 68}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-maxQp-parameter", 61}, { {"maxQp must be a valid VkVideoEncodeH264QpKHR structure", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-minQp-parameter", 61}, { {"minQp must be a valid VkVideoEncodeH264QpKHR structure", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_LAYER_INFO_KHR", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-useMaxQp-08287", 60}, { {"If useMaxQp is VK_TRUE, then the qpI, qpP, and qpB members of maxQp must all be between VkVideoEncodeH264CapabilitiesKHR::minQp and VkVideoEncodeH264CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 254}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-useMaxQp-08289", 60}, { {"If useMaxQp is VK_TRUE and VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR, then the qpI, qpP, and qpB members of maxQp must all specify the same value", 312}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-useMinQp-08286", 60}, { {"If useMinQp is VK_TRUE, then the qpI, qpP, and qpB members of minQp must all be between VkVideoEncodeH264CapabilitiesKHR::minQp and VkVideoEncodeH264CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 254}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-useMinQp-08288", 60}, { {"If useMinQp is VK_TRUE and VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR, then the qpI, qpP, and qpB members of minQp must all specify the same value", 312}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264RateControlLayerInfoKHR-useMinQp-08374", 60}, { {"If useMinQp and useMaxQp are both VK_TRUE, then the qpI, qpP, and qpB members of minQp must all be less than or equal to the respective members of maxQp", 152}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionCreateInfoKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_CREATE_INFO_KHR", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersAddInfoKHR-None-04837", 60}, { {"The seq_parameter_set_id member of each StdVideoH264SequenceParameterSet structure specified in the elements of pStdSPSs must be unique within pStdSPSs", 151}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersAddInfoKHR-None-04838", 60}, { {"The pair constructed from the seq_parameter_set_id and pic_parameter_set_id members of each StdVideoH264PictureParameterSet structure specified in the elements of pStdPPSs must be unique within pStdPPSs", 202}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersAddInfoKHR-pStdPPSs-parameter", 68}, { {"If stdPPSCount is not 0, and pStdPPSs is not NULL, pStdPPSs must be a valid pointer to an array of stdPPSCount StdVideoH264PictureParameterSet values", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersAddInfoKHR-pStdSPSs-parameter", 68}, { {"If stdSPSCount is not 0, and pStdSPSs is not NULL, pStdSPSs must be a valid pointer to an array of stdSPSCount StdVideoH264SequenceParameterSet values", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersAddInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersCreateInfoKHR-pParametersAddInfo-parameter", 81}, { {"If pParametersAddInfo is not NULL, pParametersAddInfo must be a valid pointer to a valid VkVideoEncodeH264SessionParametersAddInfoKHR structure", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersCreateInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersFeedbackInfoKHR-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_FEEDBACK_INFO_KHR", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersGetInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_GET_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH264SessionParametersGetInfoKHR-writeStdSPS-08279", 67}, { {"At least one of writeStdSPS and writeStdPPS must be VK_TRUE", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265CapabilitiesKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_CAPABILITIES_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265DpbSlotInfoKHR-pStdReferenceInfo-parameter", 64}, { {"pStdReferenceInfo must be a valid pointer to a valid StdVideoEncodeH265ReferenceInfo value", 90}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265DpbSlotInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_DPB_SLOT_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265GopRemainingFrameInfoKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_GOP_REMAINING_FRAME_INFO_KHR", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265NaluSliceSegmentInfoKHR-pNext-pNext", 57}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265NaluSliceSegmentInfoKHR-pStdSliceSegmentHeader-parameter", 78}, { {"pStdSliceSegmentHeader must be a valid pointer to a valid StdVideoEncodeH265SliceSegmentHeader value", 100}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265NaluSliceSegmentInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_NALU_SLICE_SEGMENT_INFO_KHR", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-flags-08316", 48}, { {"If VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_PREDICTION_WEIGHT_TABLE_GENERATED_BIT_KHR and the slice segment corresponding to any element of pNaluSliceSegmentEntries uses explicit weighted sample prediction, then VkVideoEncodeH265NaluSliceSegmentInfoKHR::pStdSliceSegmentHeader->pWeightTable must not be NULL for that element of pNaluSliceSegmentEntries", 484}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-flags-08323", 48}, { {"If VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_MULTIPLE_TILES_PER_SLICE_SEGMENT_BIT_KHR, then naluSliceSegmentEntryCount must be greater than or equal to the number of H.265 tiles in the picture", 323}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-flags-08324", 48}, { {"If VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_MULTIPLE_SLICE_SEGMENTS_PER_TILE_BIT_KHR, then naluSliceSegmentEntryCount must be less than or equal to the number of H.265 tiles in the picture", 320}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-naluSliceSegmentEntryCount-08306", 69}, { {"naluSliceSegmentEntryCount must be between 1 and VkVideoEncodeH265CapabilitiesKHR::maxSliceSegmentCount, inclusive, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 197}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-naluSliceSegmentEntryCount-arraylength", 75}, { {"naluSliceSegmentEntryCount must be greater than 0", 49}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-pNaluSliceSegmentEntries-parameter", 71}, { {"pNaluSliceSegmentEntries must be a valid pointer to an array of naluSliceSegmentEntryCount valid VkVideoEncodeH265NaluSliceSegmentInfoKHR structures", 148}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-pStdPictureInfo-parameter", 62}, { {"pStdPictureInfo must be a valid pointer to a valid StdVideoEncodeH265PictureInfo value", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265PictureInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PICTURE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265ProfileInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PROFILE_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265QualityLevelPropertiesKHR-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUALITY_LEVEL_PROPERTIES_KHR", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265QuantizationMapCapabilitiesKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUANTIZATION_MAP_CAPABILITIES_KHR", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-consecutiveBFrameCount-08296", 69}, { {"If consecutiveBFrameCount is not 0, then it must be less than gopFrameCount", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-flags-08291", 52}, { {"If VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_HRD_COMPLIANCE_BIT_KHR, then flags must not contain VK_VIDEO_ENCODE_H265_RATE_CONTROL_ATTEMPT_HRD_COMPLIANCE_BIT_KHR", 292}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-flags-08292", 52}, { {"If flags contains VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR or VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR, then it must also contain VK_VIDEO_ENCODE_H265_RATE_CONTROL_REGULAR_GOP_BIT_KHR", 233}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-flags-08293", 52}, { {"If flags contains VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR, then it must not also contain VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-flags-08294", 52}, { {"If flags contains VK_VIDEO_ENCODE_H265_RATE_CONTROL_REGULAR_GOP_BIT_KHR, then gopFrameCount must be greater than 0", 114}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-flags-parameter", 56}, { {"flags must be a valid combination of VkVideoEncodeH265RateControlFlagBitsKHR values", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-idrPeriod-08295", 56}, { {"If idrPeriod is not 0, then it must be greater than or equal to gopFrameCount", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_INFO_KHR", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-maxFrameSize-parameter", 68}, { {"maxFrameSize must be a valid VkVideoEncodeH265FrameSizeKHR structure", 68}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-maxQp-parameter", 61}, { {"maxQp must be a valid VkVideoEncodeH265QpKHR structure", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-minQp-parameter", 61}, { {"minQp must be a valid VkVideoEncodeH265QpKHR structure", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_LAYER_INFO_KHR", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-useMaxQp-08298", 60}, { {"If useMaxQp is VK_TRUE, then the qpI, qpP, and qpB members of maxQp must all be between VkVideoEncodeH265CapabilitiesKHR::minQp and VkVideoEncodeH265CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 254}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-useMaxQp-08300", 60}, { {"If useMaxQp is VK_TRUE and VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR, then the qpI, qpP, and qpB members of maxQp must all specify the same value", 312}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-useMinQp-08297", 60}, { {"If useMinQp is VK_TRUE, then the qpI, qpP, and qpB members of minQp must all be between VkVideoEncodeH265CapabilitiesKHR::minQp and VkVideoEncodeH265CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 254}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-useMinQp-08299", 60}, { {"If useMinQp is VK_TRUE and VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR, then the qpI, qpP, and qpB members of minQp must all specify the same value", 312}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265RateControlLayerInfoKHR-useMinQp-08375", 60}, { {"If useMinQp and useMaxQp are both VK_TRUE, then the qpI, qpP, and qpB members of minQp must all be less than or equal to the respective members of maxQp", 152}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionCreateInfoKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_CREATE_INFO_KHR", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-None-06438", 60}, { {"The vps_video_parameter_set_id member of each StdVideoH265VideoParameterSet structure specified in the elements of pStdVPSs must be unique within pStdVPSs", 154}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-None-06439", 60}, { {"The pair constructed from the sps_video_parameter_set_id and sps_seq_parameter_set_id members of each StdVideoH265SequenceParameterSet structure specified in the elements of pStdSPSs must be unique within pStdSPSs", 213}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-None-06440", 60}, { {"The triplet constructed from the sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id members of each StdVideoH265PictureParameterSet structure specified in the elements of pStdPPSs must be unique within pStdPPSs", 242}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-pStdPPSs-parameter", 68}, { {"If stdPPSCount is not 0, and pStdPPSs is not NULL, pStdPPSs must be a valid pointer to an array of stdPPSCount StdVideoH265PictureParameterSet values", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-pStdSPSs-parameter", 68}, { {"If stdSPSCount is not 0, and pStdSPSs is not NULL, pStdSPSs must be a valid pointer to an array of stdSPSCount StdVideoH265SequenceParameterSet values", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-pStdVPSs-parameter", 68}, { {"If stdVPSCount is not 0, and pStdVPSs is not NULL, pStdVPSs must be a valid pointer to an array of stdVPSCount StdVideoH265VideoParameterSet values", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersAddInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersCreateInfoKHR-pParametersAddInfo-parameter", 81}, { {"If pParametersAddInfo is not NULL, pParametersAddInfo must be a valid pointer to a valid VkVideoEncodeH265SessionParametersAddInfoKHR structure", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersCreateInfoKHR-sType-sType", 64}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersFeedbackInfoKHR-sType-sType", 66}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_FEEDBACK_INFO_KHR", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersGetInfoKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_GET_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeH265SessionParametersGetInfoKHR-writeStdVPS-08290", 67}, { {"At least one of writeStdVPS, writeStdSPS, and writeStdPPS must be VK_TRUE", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-dstBuffer-08236", 41}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_VIDEO_ENCODE_DST_BIT_KHR usage flag set", 97}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-dstBuffer-parameter", 45}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-dstBufferOffset-08237", 47}, { {"dstBufferOffset must be less than the size of dstBuffer", 55}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-dstBufferRange-08238", 46}, { {"dstBufferRange must be less than or equal to the size of dstBuffer minus dstBufferOffset", 88}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-flags-parameter", 41}, { {"flags must be a valid combination of VkVideoEncodeFlagBitsKHR values", 68}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1PictureInfoKHR, VkVideoEncodeH264PictureInfoKHR, VkVideoEncodeH265PictureInfoKHR, VkVideoEncodeIntraRefreshInfoKHR, VkVideoEncodeQuantizationMapInfoKHR, or VkVideoInlineQueryInfoKHR", 328}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pPictureResource-08242", 48}, { {"The pPictureResource member of each element of pReferenceSlots must not be NULL", 79}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pReferenceSlots-parameter", 51}, { {"If referenceSlotCount is not 0, pReferenceSlots must be a valid pointer to an array of referenceSlotCount valid VkVideoReferenceSlotInfoKHR structures", 150}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pSetupReferenceSlot-08239", 51}, { {"If pSetupReferenceSlot is not NULL, then its slotIndex member must not be negative", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pSetupReferenceSlot-08240", 51}, { {"If pSetupReferenceSlot is not NULL, then its pPictureResource must not be NULL", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-pSetupReferenceSlot-parameter", 55}, { {"If pSetupReferenceSlot is not NULL, pSetupReferenceSlot must be a valid pointer to a valid VkVideoReferenceSlotInfoKHR structure", 128}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_INFO_KHR", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-slotIndex-08241", 41}, { {"The slotIndex member of each element of pReferenceSlots must not be negative", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeInfoKHR-srcPictureResource-parameter", 54}, { {"srcPictureResource must be a valid VkVideoPictureResourceInfoKHR structure", 74}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeIntraRefreshCapabilitiesKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_INTRA_REFRESH_CAPABILITIES_KHR", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeIntraRefreshInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_INTRA_REFRESH_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeProfileRgbConversionInfoVALVE-sType-sType", 59}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_PROFILE_RGB_CONVERSION_INFO_VALVE", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQualityLevelInfoKHR-qualityLevel-08311", 56}, { {"qualityLevel must be less than VkVideoEncodeCapabilitiesKHR::maxQualityLevels, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 160}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQualityLevelInfoKHR-sType-sType", 49}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQualityLevelPropertiesKHR-pNext-pNext", 55}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1QualityLevelPropertiesKHR, VkVideoEncodeH264QualityLevelPropertiesKHR, or VkVideoEncodeH265QualityLevelPropertiesKHR", 263}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQualityLevelPropertiesKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_PROPERTIES_KHR", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQualityLevelPropertiesKHR-sType-unique", 56}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapCapabilitiesKHR-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_CAPABILITIES_KHR", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapInfoKHR-quantizationMap-parameter", 66}, { {"If quantizationMap is not VK_NULL_HANDLE, quantizationMap must be a valid VkImageView handle", 92}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapInfoKHR-quantizationMapExtent-10352", 68}, { {"quantizationMapExtent.width must be less than or equal to the width of quantizationMap", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapInfoKHR-quantizationMapExtent-10353", 68}, { {"quantizationMapExtent.height must be less than or equal to the height of quantizationMap", 88}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_INFO_KHR", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR-sType-sType", 75}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_SESSION_PARAMETERS_CREATE_INFO_KHR", 96}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-flags-zerobitmask", 54}, { {"flags must be 0", 15}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-layerCount-08245", 53}, { {"layerCount member must be less than or equal to VkVideoEncodeCapabilitiesKHR::maxRateControlLayers, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 181}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-layerCount-08357", 53}, { {"If layerCount is not zero, then virtualBufferSizeInMs must be greater than zero", 79}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-layerCount-08358", 53}, { {"If layerCount is not zero, then initialVirtualBufferSizeInMs must be less than or equal to virtualBufferSizeInMs", 112}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-pLayers-08276", 50}, { {"For each element of pLayers, its averageBitrate member must be between 1 and VkVideoEncodeCapabilitiesKHR::maxBitrate, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 200}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-pLayers-08277", 50}, { {"For each element of pLayers, its maxBitrate member must be between 1 and VkVideoEncodeCapabilitiesKHR::maxBitrate, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 196}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-pLayers-parameter", 54}, { {"If layerCount is not 0, pLayers must be a valid pointer to an array of layerCount valid VkVideoEncodeRateControlLayerInfoKHR structures", 135}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-08244", 58}, { {"If rateControlMode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR, then it must specify one of the bits included in VkVideoEncodeCapabilitiesKHR::rateControlModes, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile", 251}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-08248", 58}, { {"If rateControlMode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then layerCount must be 0", 148}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-08275", 58}, { {"If rateControlMode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_CBR_BIT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_VBR_BIT_KHR, then layerCount must be greater than 0", 156}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-08278", 58}, { {"If rateControlMode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_VBR_BIT_KHR, then for each element of pLayers, its averageBitrate member must be less than or equal to its maxBitrate member", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-08356", 58}, { {"If rateControlMode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_CBR_BIT_KHR, then for each element of pLayers, its averageBitrate member must equal its maxBitrate member", 161}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-rateControlMode-parameter", 62}, { {"If rateControlMode is not 0, rateControlMode must be a valid VkVideoEncodeRateControlModeFlagBitsKHR value", 106}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-sType-sType", 48}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_INFO_KHR", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-videoCodecOperation-07022", 62}, { {"If the videoCodecOperation of the used video profile is VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, the pNext chain this structure is included in also includes an instance of the VkVideoEncodeH264RateControlInfoKHR structure, and layerCount is greater than 1, then layerCount must equal VkVideoEncodeH264RateControlInfoKHR::temporalLayerCount", 344}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-videoCodecOperation-07025", 62}, { {"If the videoCodecOperation of the used video profile is VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, the pNext chain this structure is included in also includes an instance of the VkVideoEncodeH265RateControlInfoKHR structure, and layerCount is greater than 1, then layerCount must equal VkVideoEncodeH265RateControlInfoKHR::subLayerCount", 339}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlInfoKHR-videoCodecOperation-10351", 62}, { {"If the videoCodecOperation of the used video profile is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, the pNext chain this structure is included in also includes an instance of the VkVideoEncodeAV1RateControlInfoKHR structure, and layerCount is greater than 1, then layerCount must equal VkVideoEncodeAV1RateControlInfoKHR::temporalLayerCount", 341}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlLayerInfoKHR-frameRateDenominator-08351", 68}, { {"frameRateDenominator must be greater than zero", 46}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlLayerInfoKHR-frameRateNumerator-08350", 66}, { {"frameRateNumerator must be greater than zero", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlLayerInfoKHR-pNext-pNext", 53}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1RateControlLayerInfoKHR, VkVideoEncodeH264RateControlLayerInfoKHR, or VkVideoEncodeH265RateControlLayerInfoKHR", 257}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlLayerInfoKHR-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_LAYER_INFO_KHR", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRateControlLayerInfoKHR-sType-unique", 54}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeRgbConversionCapabilitiesVALVE-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_RGB_CONVERSION_CAPABILITIES_VALVE", 78}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionIntraRefreshCreateInfoKHR-intraRefreshMode-parameter", 77}, { {"If intraRefreshMode is not 0, intraRefreshMode must be a valid VkVideoEncodeIntraRefreshModeFlagBitsKHR value", 109}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionIntraRefreshCreateInfoKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_INTRA_REFRESH_CREATE_INFO_KHR", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersFeedbackInfoKHR-pNext-pNext", 62}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeH264SessionParametersFeedbackInfoKHR or VkVideoEncodeH265SessionParametersFeedbackInfoKHR", 233}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersFeedbackInfoKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_FEEDBACK_INFO_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersFeedbackInfoKHR-sType-unique", 63}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersGetInfoKHR-pNext-pNext", 57}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeH264SessionParametersGetInfoKHR or VkVideoEncodeH265SessionParametersGetInfoKHR", 223}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersGetInfoKHR-sType-sType", 57}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_GET_INFO_KHR", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersGetInfoKHR-sType-unique", 58}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionParametersGetInfoKHR-videoSessionParameters-parameter", 78}, { {"videoSessionParameters must be a valid VkVideoSessionParametersKHR handle", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-rgbModel-10930", 68}, { {"rgbModel must only be a bit set in VkVideoEncodeRgbConversionCapabilitiesVALVE::rgbModels as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR with VkVideoEncodeRgbConversionCapabilitiesVALVE in the pNext chain of VkVideoCapabilitiesKHR with the given pVideoProfile", 267}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-rgbModel-parameter", 72}, { {"rgbModel must be a valid VkVideoEncodeRgbModelConversionFlagBitsVALVE value", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-rgbRange-10931", 68}, { {"rgbRange must only be a bit set in VkVideoEncodeRgbConversionCapabilitiesVALVE::rgbRanges as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR with VkVideoEncodeRgbConversionCapabilitiesVALVE in the pNext chain of VkVideoCapabilitiesKHR with the given pVideoProfile", 267}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-rgbRange-parameter", 72}, { {"rgbRange must be a valid VkVideoEncodeRgbRangeCompressionFlagBitsVALVE value", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-sType-sType", 65}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_RGB_CONVERSION_CREATE_INFO_VALVE", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-xChromaOffset-10932", 73}, { {"xChromaOffset must only be a bit set in VkVideoEncodeRgbConversionCapabilitiesVALVE::xChromaOffsets as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR with VkVideoEncodeRgbConversionCapabilitiesVALVE in the pNext chain of VkVideoCapabilitiesKHR with the given pVideoProfile", 277}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-xChromaOffset-parameter", 77}, { {"xChromaOffset must be a valid VkVideoEncodeRgbChromaOffsetFlagBitsVALVE value", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-yChromaOffset-10933", 73}, { {"yChromaOffset must only be a bit set in VkVideoEncodeRgbConversionCapabilitiesVALVE::yChromaOffsets as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR with VkVideoEncodeRgbConversionCapabilitiesVALVE in the pNext chain of VkVideoCapabilitiesKHR with the given pVideoProfile", 277}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeSessionRgbConversionCreateInfoVALVE-yChromaOffset-parameter", 77}, { {"yChromaOffset must be a valid VkVideoEncodeRgbChromaOffsetFlagBitsVALVE value", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeUsageInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_ENCODE_USAGE_INFO_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeUsageInfoKHR-tuningMode-parameter", 51}, { {"If tuningMode is not 0, tuningMode must be a valid VkVideoEncodeTuningModeKHR value", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeUsageInfoKHR-videoContentHints-parameter", 58}, { {"videoContentHints must be a valid combination of VkVideoEncodeContentFlagBitsKHR values", 87}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEncodeUsageInfoKHR-videoUsageHints-parameter", 56}, { {"videoUsageHints must be a valid combination of VkVideoEncodeUsageFlagBitsKHR values", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEndCodingInfoKHR-flags-zerobitmask", 46}, { {"flags must be 0", 15}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEndCodingInfoKHR-pNext-pNext", 40}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoEndCodingInfoKHR-sType-sType", 40}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_END_CODING_INFO_KHR", 57}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatAV1QuantizationMapPropertiesKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_FORMAT_AV1_QUANTIZATION_MAP_PROPERTIES_KHR", 80}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatH265QuantizationMapPropertiesKHR-sType-sType", 62}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_FORMAT_H265_QUANTIZATION_MAP_PROPERTIES_KHR", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatPropertiesKHR-pNext-pNext", 43}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkImageCreateFlags2CreateInfoKHR, VkImageUsageFlags2CreateInfoKHR, VkVideoFormatAV1QuantizationMapPropertiesKHR, VkVideoFormatH265QuantizationMapPropertiesKHR, or VkVideoFormatQuantizationMapPropertiesKHR", 335}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatPropertiesKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_FORMAT_PROPERTIES_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatPropertiesKHR-sType-unique", 44}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoFormatQuantizationMapPropertiesKHR-sType-sType", 58}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_FORMAT_QUANTIZATION_MAP_PROPERTIES_KHR", 76}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoInlineQueryInfoKHR-queryPool-08372", 46}, { {"If queryPool is not VK_NULL_HANDLE, then firstQuery must be less than the number of queries in queryPool", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoInlineQueryInfoKHR-queryPool-08373", 46}, { {"If queryPool is not VK_NULL_HANDLE, then the sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 142}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoInlineQueryInfoKHR-queryPool-parameter", 50}, { {"If queryPool is not VK_NULL_HANDLE, queryPool must be a valid VkQueryPool handle", 80}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoInlineQueryInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_INLINE_QUERY_INFO_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoPictureResourceInfoKHR-baseArrayLayer-07175", 55}, { {"baseArrayLayer must be less than the VkImageViewCreateInfo::subresourceRange.layerCount specified when the image view imageViewBinding was created", 146}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoPictureResourceInfoKHR-imageViewBinding-parameter", 61}, { {"imageViewBinding must be a valid VkImageView handle", 51}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoPictureResourceInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoPictureResourceInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_INFO_KHR", 63}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-chromaBitDepth-parameter", 51}, { {"chromaBitDepth must be a valid combination of VkVideoComponentBitDepthFlagBitsKHR values", 88}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-chromaSubsampling-07013", 50}, { {"chromaSubsampling must have a single bit set", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-chromaSubsampling-07015", 50}, { {"If chromaSubsampling is not VK_VIDEO_CHROMA_SUBSAMPLING_MONOCHROME_BIT_KHR, then chromaBitDepth must have a single bit set", 122}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-chromaSubsampling-parameter", 54}, { {"chromaSubsampling must be a valid combination of VkVideoChromaSubsamplingFlagBitsKHR values", 91}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-chromaSubsampling-requiredbitmask", 60}, { {"chromaSubsampling must not be 0", 31}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-lumaBitDepth-07014", 45}, { {"lumaBitDepth must have a single bit set", 39}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-lumaBitDepth-parameter", 49}, { {"lumaBitDepth must be a valid combination of VkVideoComponentBitDepthFlagBitsKHR values", 86}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-lumaBitDepth-requiredbitmask", 55}, { {"lumaBitDepth must not be 0", 26}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-sType-sType", 38}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_PROFILE_INFO_KHR", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-07179", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the pNext chain must include a VkVideoDecodeH264ProfileInfoKHR structure", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-07180", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the pNext chain must include a VkVideoDecodeH265ProfileInfoKHR structure", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-07181", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain must include a VkVideoEncodeH264ProfileInfoKHR structure", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-07182", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain must include a VkVideoEncodeH265ProfileInfoKHR structure", 149}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-09256", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the pNext chain must include a VkVideoDecodeAV1ProfileInfoKHR structure", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-10262", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain must include a VkVideoEncodeAV1ProfileInfoKHR structure", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-10791", 52}, { {"If videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the pNext chain must include a VkVideoDecodeVP9ProfileInfoKHR structure", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileInfoKHR-videoCodecOperation-parameter", 56}, { {"videoCodecOperation must be a valid VkVideoCodecOperationFlagBitsKHR value", 74}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileListInfoKHR-pProfiles-06813", 46}, { {"pProfiles must not contain more than one element whose videoCodecOperation member specifies a decode operation", 110}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileListInfoKHR-pProfiles-parameter", 50}, { {"If profileCount is not 0, pProfiles must be a valid pointer to an array of profileCount valid VkVideoProfileInfoKHR structures", 126}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoProfileListInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_PROFILE_LIST_INFO_KHR", 59}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoReferenceIntraRefreshInfoKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_REFERENCE_INTRA_REFRESH_INFO_KHR", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoReferenceSlotInfoKHR-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoDecodeAV1DpbSlotInfoKHR, VkVideoDecodeH264DpbSlotInfoKHR, VkVideoDecodeH265DpbSlotInfoKHR, VkVideoEncodeAV1DpbSlotInfoKHR, VkVideoEncodeH264DpbSlotInfoKHR, VkVideoEncodeH265DpbSlotInfoKHR, or VkVideoReferenceIntraRefreshInfoKHR", 365}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoReferenceSlotInfoKHR-pPictureResource-parameter", 59}, { {"If pPictureResource is not NULL, pPictureResource must be a valid pointer to a valid VkVideoPictureResourceInfoKHR structure", 124}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoReferenceSlotInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_INFO_KHR", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoReferenceSlotInfoKHR-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-08371", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, then videoMaintenance1 must be enabled", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10264", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR, then the videoEncodeQuantizationMap feature must be enabled", 207}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10265", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR, then pVideoProfile->videoCodecOperation must specify an encode operation", 220}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10266", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR, then it must not also include VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR", 174}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10267", 44}, { {"If VkVideoEncodeCapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_CAPABILITY_QUANTIZATION_DELTA_MAP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile, then flags must not include VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR", 315}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10268", 44}, { {"If VkVideoEncodeCapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_CAPABILITY_EMPHASIS_MAP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile, then flags must not include VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR", 295}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10398", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, then videoMaintenance2 must be enabled", 115}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-10399", 44}, { {"If flags includes VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, then pVideoProfile->videoCodecOperation must specify a decode operation", 148}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-flags-parameter", 48}, { {"flags must be a valid combination of VkVideoSessionCreateFlagBitsKHR values", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-maxActiveReferencePictures-04849", 65}, { {"maxActiveReferencePictures must be less than or equal to VkVideoCapabilitiesKHR::maxActiveReferencePictures, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 212}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-maxCodedExtent-04851", 53}, { {"maxCodedExtent must be between VkVideoCapabilitiesKHR::minCodedExtent and VkVideoCapabilitiesKHR::maxCodedExtent, inclusive, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 228}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-maxDpbSlots-04847", 50}, { {"maxDpbSlots must be less than or equal to VkVideoCapabilitiesKHR::maxDpbSlots, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-maxDpbSlots-04850", 50}, { {"If either maxDpbSlots or maxActiveReferencePictures is 0, then both must be 0", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pNext-10924", 44}, { {"If a VkVideoEncodeProfileRgbConversionInfoVALVE structure is included in the pNext chain of pVideoProfile and VkVideoEncodeProfileRgbConversionInfoVALVE::performEncodeRgbConversion is enabled, a VkVideoEncodeSessionRgbConversionCreateInfoVALVE structure must be included in the pNext chain of VkVideoSessionCreateInfoKHR", 320}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pNext-pNext", 44}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoEncodeAV1SessionCreateInfoKHR, VkVideoEncodeH264SessionCreateInfoKHR, VkVideoEncodeH265SessionCreateInfoKHR, VkVideoEncodeSessionIntraRefreshCreateInfoKHR, or VkVideoEncodeSessionRgbConversionCreateInfoVALVE", 345}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pStdHeaderVersion-07190", 56}, { {"pStdHeaderVersion->extensionName must match VkVideoCapabilitiesKHR::stdHeaderVersion.extensionName, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 203}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pStdHeaderVersion-07191", 56}, { {"pStdHeaderVersion->specVersion must be less than or equal to VkVideoCapabilitiesKHR::stdHeaderVersion.specVersion, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 218}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pStdHeaderVersion-parameter", 60}, { {"pStdHeaderVersion must be a valid pointer to a valid VkExtensionProperties structure", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-04845", 52}, { {"pVideoProfile must be a supported video profile", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-08251", 52}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and the pNext chain of this structure includes a VkVideoEncodeH264SessionCreateInfoKHR structure, then its maxLevelIdc member must be less than or equal to VkVideoEncodeH264CapabilitiesKHR::maxLevelIdc, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in pVideoProfile", 392}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-08252", 52}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of this structure includes a VkVideoEncodeH265SessionCreateInfoKHR structure, then its maxLevelIdc member must be less than or equal to VkVideoEncodeH265CapabilitiesKHR::maxLevelIdc, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in pVideoProfile", 392}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10269", 52}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the videoEncodeAV1 feature must be enabled", 133}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10270", 52}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pNext chain of this structure includes a VkVideoEncodeAV1SessionCreateInfoKHR structure, then its maxLevel member must be less than or equal to VkVideoEncodeAV1CapabilitiesKHR::maxLevel, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified in pVideoProfile", 383}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10793", 52}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the videoDecodeVP9 feature must be enabled", 133}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10835", 52}, { {"If pVideoProfile->videoCodecOperation specifies an encode operation, the pNext chain of this structure includes a VkVideoEncodeSessionIntraRefreshCreateInfoKHR structure, and its intraRefreshMode member is not VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_NONE_KHR, then the videoEncodeIntraRefresh feature must be enabled", 311}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10836", 52}, { {"If pVideoProfile->videoCodecOperation specifies an encode operation, the pNext chain of this structure includes a VkVideoEncodeSessionIntraRefreshCreateInfoKHR structure, and its intraRefreshMode member is not VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_NONE_KHR, then intraRefreshMode must specify one of the bits included in VkVideoEncodeIntraRefreshCapabilitiesKHR::intraRefreshModes, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile", 482}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-10923", 52}, { {"If the pVideoProfile->pNext chain includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be enabled", 156}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-11759", 52}, { {"pVideoProfile->videoCodecOperation must be supported by the queue family index specified in queueFamilyIndex, as reported in VkQueueFamilyVideoPropertiesKHR::videoCodecOperations", 178}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pVideoProfile-parameter", 56}, { {"pVideoProfile must be a valid pointer to a valid VkVideoProfileInfoKHR structure", 80}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pictureFormat-04853", 52}, { {"If pVideoProfile->videoCodecOperation specifies a decode operation, then pictureFormat must be one of the supported decode output formats, as returned by vkGetPhysicalDeviceVideoFormatPropertiesKHR in VkVideoFormatPropertiesKHR::format when called with the imageUsage member of its pVideoFormatInfo parameter containing VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR, and with a VkVideoProfileListInfoKHR structure specified in the pNext chain of its pVideoFormatInfo parameter whose pProfiles member contains an element matching pVideoProfile", 536}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pictureFormat-04854", 52}, { {"If pVideoProfile->videoCodecOperation specifies an encode operation, then pictureFormat must be one of the supported encode input formats, as returned by vkGetPhysicalDeviceVideoFormatPropertiesKHR in VkVideoFormatPropertiesKHR::format when called with the imageUsage member of its pVideoFormatInfo parameter containing VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR, and with a VkVideoProfileListInfoKHR structure specified in the pNext chain of its pVideoFormatInfo parameter whose pProfiles member contains an element matching pVideoProfile", 536}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-pictureFormat-parameter", 56}, { {"pictureFormat must be a valid VkFormat value", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-protectedMemory-07189", 54}, { {"If the protectedMemory feature is not enabled or if VkVideoCapabilitiesKHR::flags does not include VK_VIDEO_CAPABILITY_PROTECTED_CONTENT_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile specified by pVideoProfile, then flags must not include VK_VIDEO_SESSION_CREATE_PROTECTED_CONTENT_BIT_KHR", 328}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-referencePictureFormat-04852", 61}, { {"If pVideoProfile->videoCodecOperation specifies a decode operation and maxActiveReferencePictures is greater than 0, then referencePictureFormat must be one of the supported decode DPB formats, as returned by vkGetPhysicalDeviceVideoFormatPropertiesKHR in VkVideoFormatPropertiesKHR::format when called with the imageUsage member of its pVideoFormatInfo parameter containing VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR, and with a VkVideoProfileListInfoKHR structure specified in the pNext chain of its pVideoFormatInfo parameter whose pProfiles member contains an element matching pVideoProfile", 591}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-referencePictureFormat-06814", 61}, { {"If pVideoProfile->videoCodecOperation specifies an encode operation and maxActiveReferencePictures is greater than 0, then referencePictureFormat must be one of the supported decode DPB formats, as returned by then referencePictureFormat must be one of the supported encode DPB formats, as returned by vkGetPhysicalDeviceVideoFormatPropertiesKHR in VkVideoFormatPropertiesKHR::format when called with the imageUsage member of its pVideoFormatInfo parameter containing VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR, and with a VkVideoProfileListInfoKHR structure specified in the pNext chain of its pVideoFormatInfo parameter whose pProfiles member contains an element matching pVideoProfile", 684}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-referencePictureFormat-parameter", 65}, { {"referencePictureFormat must be a valid VkFormat value", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_SESSION_CREATE_INFO_KHR", 61}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionCreateInfoKHR-sType-unique", 45}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionMemoryRequirementsKHR-pNext-pNext", 52}, { {"pNext must be NULL", 18}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionMemoryRequirementsKHR-sType-sType", 52}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_SESSION_MEMORY_REQUIREMENTS_KHR", 69}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-commonparent", 55}, { {"Both of videoSession, and videoSessionParametersTemplate that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 176}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-flags-10271", 54}, { {"If flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then videoSession must have been created with VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR", 264}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-flags-10272", 54}, { {"If flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then the pNext chain must include a VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR structure", 194}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-flags-10273", 54}, { {"If flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR and videoSession was created with VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR, then the list of video format properties supported for the image usage flag VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR must have an element for which VkVideoFormatQuantizationMapPropertiesKHR::quantizationMapTexelSize equals the quantizationMapTexelSize member of the VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR structure included in the pNext chain", 572}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-flags-10274", 54}, { {"If flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR and videoSession was created with VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR, then the list of video format properties supported for the image usage flag VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR must have an element for which VkVideoFormatQuantizationMapPropertiesKHR::quantizationMapTexelSize equals the quantizationMapTexelSize member of the VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR structure included in the pNext chain", 552}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-flags-parameter", 58}, { {"flags must be a valid combination of VkVideoSessionParametersCreateFlagBitsKHR values", 85}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-pNext-pNext", 54}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoDecodeAV1SessionParametersCreateInfoKHR, VkVideoDecodeH264SessionParametersCreateInfoKHR, VkVideoDecodeH265SessionParametersCreateInfoKHR, VkVideoEncodeAV1SessionParametersCreateInfoKHR, VkVideoEncodeH264SessionParametersCreateInfoKHR, VkVideoEncodeH265SessionParametersCreateInfoKHR, VkVideoEncodeQualityLevelInfoKHR, or VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR", 518}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_CREATE_INFO_KHR", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-04839", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the number of elements of spsAddList must be less than or equal to the maxStdSPSCount specified in the VkVideoEncodeH264SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-04840", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the number of elements of ppsAddList must be less than or equal to the maxStdPPSCount specified in the VkVideoEncodeH264SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-04841", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of elements of vpsAddList must be less than or equal to the maxStdVPSCount specified in the VkVideoEncodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-04842", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of elements of spsAddList must be less than or equal to the maxStdSPSCount specified in the VkVideoEncodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-04843", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of elements of ppsAddList must be less than or equal to the maxStdPPSCount specified in the VkVideoEncodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07203", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the pNext chain must include a VkVideoDecodeH264SessionParametersCreateInfoKHR structure", 198}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07204", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the number of elements of spsAddList must be less than or equal to the maxStdSPSCount specified in the VkVideoDecodeH264SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07205", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the number of elements of ppsAddList must be less than or equal to the maxStdPPSCount specified in the VkVideoDecodeH264SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07206", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the pNext chain must include a VkVideoDecodeH265SessionParametersCreateInfoKHR structure", 198}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07207", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of elements of vpsAddList must be less than or equal to the maxStdVPSCount specified in the VkVideoDecodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07208", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of elements of spsAddList must be less than or equal to the maxStdSPSCount specified in the VkVideoDecodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07209", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of elements of ppsAddList must be less than or equal to the maxStdPPSCount specified in the VkVideoDecodeH265SessionParametersCreateInfoKHR structure included in the pNext chain", 298}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07210", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain must include a VkVideoEncodeH264SessionParametersCreateInfoKHR structure", 198}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-07211", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain must include a VkVideoEncodeH265SessionParametersCreateInfoKHR structure", 198}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-08319", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then num_tile_columns_minus1 must be less than VkVideoEncodeH265CapabilitiesKHR::maxTiles.width, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSession was created with, for each element of ppsAddList", 340}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-08320", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then num_tile_rows_minus1 must be less than VkVideoEncodeH265CapabilitiesKHR::maxTiles.height, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSession was created with, for each element of ppsAddList", 338}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-09258", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then videoSessionParametersTemplate must be VK_NULL_HANDLE", 162}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-09259", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the pNext chain must include a VkVideoDecodeAV1SessionParametersCreateInfoKHR structure", 196}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-10278", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then videoSessionParametersTemplate must be VK_NULL_HANDLE", 162}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-10279", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain must include a VkVideoEncodeAV1SessionParametersCreateInfoKHR structure", 196}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-10280", 61}, { {"If videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the stdOperatingPointCount member of the VkVideoEncodeAV1SessionParametersCreateInfoKHR structure included in the pNext chain must be less than or equal to VkVideoEncodeAV1CapabilitiesKHR::maxOperatingPoints, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSession was created with", 424}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-10794", 61}, { {"videoSession must not have been created with the codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR", 108}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSession-parameter", 65}, { {"videoSession must be a valid VkVideoSessionKHR handle", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-04855", 79}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE, it must have been created against videoSession", 103}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-08310", 79}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE and videoSession was created with an encode operation, then qualityLevel must equal the video encode quality level videoSessionParametersTemplate was created with", 218}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-10275", 79}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE and flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then videoSessionParametersTemplate must have been created with VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR", 281}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-10276", 79}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE and flags includes VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then videoSessionParametersTemplate must have been created with the same quantization map texel size as the one specified in the quantizationMapTexelSize member of the VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR structure included in the pNext chain", 411}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-10277", 79}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE and flags does not include VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then videoSessionParametersTemplate must have been created without VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR", 292}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersCreateInfoKHR-videoSessionParametersTemplate-parameter", 83}, { {"If videoSessionParametersTemplate is not VK_NULL_HANDLE, videoSessionParametersTemplate must be a valid VkVideoSessionParametersKHR handle", 138}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersUpdateInfoKHR-pNext-pNext", 54}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkVideoDecodeH264SessionParametersAddInfoKHR, VkVideoDecodeH265SessionParametersAddInfoKHR, VkVideoEncodeH264SessionParametersAddInfoKHR, or VkVideoEncodeH265SessionParametersAddInfoKHR", 316}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersUpdateInfoKHR-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_UPDATE_INFO_KHR", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkVideoSessionParametersUpdateInfoKHR-sType-unique", 55}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-VkViewport-apiVersion-07917", 32}, { {"If the VK_KHR_maintenance1 extension is not enabled, the VK_AMD_negative_viewport_height extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, height must be greater than 0.0", 214}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-height-01773", 28}, { {"The absolute value of height must be less than or equal to VkPhysicalDeviceLimits::maxViewportDimensions[1]", 107}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-maxDepth-01235", 30}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled, maxDepth must be between 0.0 and 1.0, inclusive", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-minDepth-01234", 30}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled, minDepth must be between 0.0 and 1.0, inclusive", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-width-01770", 27}, { {"width must be greater than 0.0", 30}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-width-01771", 27}, { {"width must be less than or equal to VkPhysicalDeviceLimits::maxViewportDimensions[0]", 84}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-x-01232", 23}, { {"(x + width) must be less than or equal to viewportBoundsRange[1]", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-x-01774", 23}, { {"x must be greater than or equal to viewportBoundsRange[0]", 57}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-y-01233", 23}, { {"(y + height) must be less than or equal to viewportBoundsRange[1]", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-y-01775", 23}, { {"y must be greater than or equal to viewportBoundsRange[0]", 57}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-y-01776", 23}, { {"y must be less than or equal to viewportBoundsRange[1]", 54}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewport-y-01777", 23}, { {"(y + height) must be greater than or equal to viewportBoundsRange[0]", 68}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewportSwizzleNV-w-parameter", 36}, { {"w must be a valid VkViewportCoordinateSwizzleNV value", 53}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewportSwizzleNV-x-parameter", 36}, { {"x must be a valid VkViewportCoordinateSwizzleNV value", 53}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewportSwizzleNV-y-parameter", 36}, { {"y must be a valid VkViewportCoordinateSwizzleNV value", 53}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkViewportSwizzleNV-z-parameter", 36}, { {"z must be a valid VkViewportCoordinateSwizzleNV value", 53}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-VkWaylandSurfaceCreateInfoKHR-display-01304", 48}, { {"display must point to a valid Wayland wl_display", 48}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWaylandSurfaceCreateInfoKHR-flags-zerobitmask", 52}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWaylandSurfaceCreateInfoKHR-pNext-pNext", 46}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWaylandSurfaceCreateInfoKHR-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_WAYLAND_SURFACE_CREATE_INFO_KHR", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWaylandSurfaceCreateInfoKHR-surface-01305", 48}, { {"surface must point to a valid Wayland wl_surface", 48}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-commonparent", 56}, { {"Both of the elements of pAcquireSyncs, and the elements of pReleaseSyncs that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 192}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pAcquireKeys-parameter", 66}, { {"If acquireCount is not 0, pAcquireKeys must be a valid pointer to an array of acquireCount uint64_t values", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pAcquireSyncs-00081", 63}, { {"Each member of pAcquireSyncs and pReleaseSyncs must be a device memory object imported by setting VkImportMemoryWin32HandleInfoKHR::handleType to VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT", 250}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pAcquireSyncs-parameter", 67}, { {"If acquireCount is not 0, pAcquireSyncs must be a valid pointer to an array of acquireCount valid VkDeviceMemory handles", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pAcquireTimeouts-parameter", 70}, { {"If acquireCount is not 0, pAcquireTimeouts must be a valid pointer to an array of acquireCount uint32_t values", 110}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pReleaseKeys-parameter", 66}, { {"If releaseCount is not 0, pReleaseKeys must be a valid pointer to an array of releaseCount uint64_t values", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-pReleaseSyncs-parameter", 67}, { {"If releaseCount is not 0, pReleaseSyncs must be a valid pointer to an array of releaseCount valid VkDeviceMemory handles", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoKHR-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_KHR", 74}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-commonparent", 55}, { {"Both of the elements of pAcquireSyncs, and the elements of pReleaseSyncs that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 192}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-pAcquireKeys-parameter", 65}, { {"If acquireCount is not 0, pAcquireKeys must be a valid pointer to an array of acquireCount uint64_t values", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-pAcquireSyncs-parameter", 66}, { {"If acquireCount is not 0, pAcquireSyncs must be a valid pointer to an array of acquireCount valid VkDeviceMemory handles", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-pAcquireTimeoutMilliseconds-parameter", 80}, { {"If acquireCount is not 0, pAcquireTimeoutMilliseconds must be a valid pointer to an array of acquireCount uint32_t values", 121}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-pReleaseKeys-parameter", 65}, { {"If releaseCount is not 0, pReleaseKeys must be a valid pointer to an array of releaseCount uint64_t values", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-pReleaseSyncs-parameter", 66}, { {"If releaseCount is not 0, pReleaseSyncs must be a valid pointer to an array of releaseCount valid VkDeviceMemory handles", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32KeyedMutexAcquireReleaseInfoNV-sType-sType", 54}, { {"sType must be VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_NV", 73}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-VkWin32SurfaceCreateInfoKHR-flags-zerobitmask", 50}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWin32SurfaceCreateInfoKHR-hinstance-01307", 48}, { {"hinstance must be a valid Win32 HINSTANCE", 41}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWin32SurfaceCreateInfoKHR-hwnd-01308", 43}, { {"hwnd must be a valid Win32 HWND", 31}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWin32SurfaceCreateInfoKHR-pNext-pNext", 44}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWin32SurfaceCreateInfoKHR-sType-sType", 44}, { {"sType must be VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkWriteDescriptorSet-commonparent", 38}, { {"Both of dstSet, and the elements of pTexelBufferView that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 172}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorCount-00317", 47}, { {"All consecutive bindings updated via a single VkWriteDescriptorSet structure, except those with a descriptorCount of zero, must have identical descriptorType", 157}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorCount-00318", 47}, { {"All consecutive bindings updated via a single VkWriteDescriptorSet structure, except those with a descriptorCount of zero, must all either use immutable samplers or must all not use immutable samplers", 200}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorCount-10776", 47}, { {"All consecutive bindings updated via a single VkWriteDescriptorSet structure, except those with a descriptorCount of zero, must have identical stageFlags", 153}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorCount-10777", 47}, { {"All consecutive bindings updated via a single VkWriteDescriptorSet structure, except those with a descriptorCount of zero, must have identical VkDescriptorBindingFlagBits", 170}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorCount-arraylength", 53}, { {"descriptorCount must be greater than 0", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00319", 46}, { {"descriptorType must match the type of dstBinding within dstSet", 62}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00324", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, pBufferInfo must be a valid pointer to an array of descriptorCount valid VkDescriptorBufferInfo structures", 286}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00325", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and dstSet was not allocated with a layout that included immutable samplers for dstBinding with descriptorType, the sampler member of each element of pImageInfo must be a valid VkSampler object", 287}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00327", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER or VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, the offset member of each element of pBufferInfo must be a multiple of VkPhysicalDeviceLimits::minUniformBufferOffsetAlignment", 227}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00328", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, the offset member of each element of pBufferInfo must be a multiple of VkPhysicalDeviceLimits::minStorageBufferOffsetAlignment", 227}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00329", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, and the buffer member of any element of pBufferInfo is the handle of a non-sparse buffer, then that buffer must be bound completely and contiguously to a single VkDeviceMemory object", 362}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00330", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER or VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, the buffer member of each element of pBufferInfo must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 231}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00331", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, the buffer member of each element of pBufferInfo must have been created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 231}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00332", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER or VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, the range member of each element of pBufferInfo, or the effective range if range is VK_WHOLE_SIZE, must be less than or equal to VkPhysicalDeviceLimits::maxUniformBufferRange", 275}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00333", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, and the shader64BitIndexing feature is not enabled, the range member of each element of pBufferInfo, or the effective range if range is VK_WHOLE_SIZE, must be less than or equal to VkPhysicalDeviceLimits::maxStorageBufferRange", 327}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00336", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the imageView member of each element of pImageInfo must have been created with the identity swizzle", 193}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00337", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, the imageView member of each element of pImageInfo must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT usage flag set", 224}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00338", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the imageView member of each element of pImageInfo must have been created with the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flag set", 191}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-00339", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, the imageView member of each element of pImageInfo must have been created with the VK_IMAGE_USAGE_STORAGE_BIT usage flag set", 179}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-01946", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, then the imageView member of each pImageInfo element must have been created without a VkSamplerYcbcrConversionInfo structure in its pNext chain", 198}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-01948", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and dstSet was allocated with a layout that included immutable samplers for dstBinding, then the imageView member of each element of pImageInfo which corresponds to an immutable sampler that enables sampler Y'CBCR conversion must have been created with a VkSamplerYcbcrConversionInfo structure in its pNext chain with an identically defined VkSamplerYcbcrConversionInfo to the corresponding immutable sampler", 472}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02219", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, dstArrayElement must be an integer multiple of 4", 110}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02220", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, descriptorCount must be an integer multiple of 4", 110}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02221", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, the pNext chain must include a VkWriteDescriptorSetInlineUniformBlock structure whose dataSize member equals descriptorCount", 186}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02382", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, the pNext chain must include a VkWriteDescriptorSetAccelerationStructureKHR structure whose accelerationStructureCount member equals descriptorCount", 216}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02738", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and if any element of pImageInfo has an imageView member that was created with a VkSamplerYcbcrConversionInfo structure in its pNext chain, then dstSet must have been allocated with a layout that included immutable samplers for dstBinding, and the corresponding immutable sampler must have been created with an identically defined VkSamplerYcbcrConversionInfo object", 430}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02752", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER, then dstSet must not have been allocated with a layout that included immutable samplers for dstBinding", 151}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02994", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, each element of pTexelBufferView must be either a valid VkBufferView handle or VK_NULL_HANDLE", 198}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02995", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER and the nullDescriptor feature is not enabled, each element of pTexelBufferView must not be VK_NULL_HANDLE", 210}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02996", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, or VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, the imageView member of each element of pImageInfo must be either a valid VkImageView handle or VK_NULL_HANDLE", 245}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-02997", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, or VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, and the nullDescriptor feature is not enabled, the imageView member of each element of pImageInfo must not be VK_NULL_HANDLE", 259}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-03817", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, the pNext chain must include a VkWriteDescriptorSetAccelerationStructureNV structure whose accelerationStructureCount member equals descriptorCount", 214}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-04149", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE the imageLayout member of each element of pImageInfo must be a member of the list given in Sampled Image", 158}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-04150", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER the imageLayout member of each element of pImageInfo must be a member of the list given in Combined Image Sampler", 176}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-04151", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT the imageLayout member of each element of pImageInfo must be a member of the list given in Input Attachment", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-04152", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_IMAGE the imageLayout member of each element of pImageInfo must be a member of the list given in Storage Image", 158}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-06450", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, the imageView member of each element of pImageInfo must have either been created without a VkImageViewMinLodCreateInfoEXT included in the pNext chain or with a VkImageViewMinLodCreateInfoEXT::minLod of 0.0", 263}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-06942", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, the imageView member of each element of pImageInfo must have been created with a view created with an image created with the VK_IMAGE_USAGE_SAMPLE_WEIGHT_BIT_QCOM usage flag set", 243}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-06943", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, the imageView member of each element of pImageInfo must have been created with a view created with an image created with the VK_IMAGE_USAGE_SAMPLE_BLOCK_MATCH_BIT_QCOM usage flag set", 246}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-07683", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, then the imageView member of each element of pImageInfo must not be VK_NULL_HANDLE", 231}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-08765", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, the pTexelBufferView buffer view usage must include VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT", 154}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-08766", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, the pTexelBufferView buffer view usage must include VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT", 154}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-09506", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, dstSet was allocated with a layout that included immutable samplers for dstBinding, and those samplers enable sampler Y'CBCR conversion, then imageView must not be VK_NULL_HANDLE", 242}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-09945", 46}, { {"If descriptorType is VK_DESCRIPTOR_TYPE_TENSOR_ARM, the pNext chain must include a VkWriteDescriptorSetTensorARM structure whose tensorViewCount member equals descriptorCount", 174}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-descriptorType-parameter", 50}, { {"descriptorType must be a valid VkDescriptorType value", 53}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstArrayElement-00321", 47}, { {"The sum of dstArrayElement and descriptorCount must be less than or equal to the number of array elements in the descriptor set binding specified by dstBinding, and all applicable consecutive bindings", 200}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstBinding-00315", 42}, { {"dstBinding must be less than or equal to the maximum value of binding of all VkDescriptorSetLayoutBinding structures specified when dstSet's descriptor set layout was created", 174}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstBinding-00316", 42}, { {"dstBinding must be a binding with a non-zero descriptorCount", 60}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstBinding-10009", 42}, { {"dstBinding must be a binding with a non-zero VkDescriptorSetLayoutCreateInfo::bindingCount", 90}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstSet-00320", 38}, { {"dstSet must be a valid VkDescriptorSet handle", 45}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-dstSet-04611", 38}, { {"If the VkDescriptorSetLayoutBinding for dstSet at dstBinding is VK_DESCRIPTOR_TYPE_MUTABLE_EXT, the new active descriptor type descriptorType must exist in the corresponding pMutableDescriptorTypeLists list for dstBinding", 221}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-pNext-pNext", 37}, { {"Each pNext member of any structure (including this one) in the pNext chain must be either NULL or a pointer to a valid instance of VkWriteDescriptorSetAccelerationStructureKHR, VkWriteDescriptorSetAccelerationStructureNV, VkWriteDescriptorSetInlineUniformBlock, VkWriteDescriptorSetPartitionedAccelerationStructureNV, or VkWriteDescriptorSetTensorARM", 350}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-sType-sType", 37}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSet-sType-unique", 38}, { {"The sType value of each structure in the pNext chain must be unique", 67}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureKHR-accelerationStructureCount-arraylength", 88}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureKHR-pAccelerationStructures-03579", 79}, { {"Each acceleration structure in pAccelerationStructures must have been created with a type of VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 183}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureKHR-pAccelerationStructures-03580", 79}, { {"If the nullDescriptor feature is not enabled, each element of pAccelerationStructures must not be VK_NULL_HANDLE", 112}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureKHR-pAccelerationStructures-parameter", 83}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount valid or VK_NULL_HANDLE VkAccelerationStructureKHR handles", 148}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureKHR-sType-sType", 61}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_KHR", 79}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureNV-accelerationStructureCount-arraylength", 87}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureNV-pAccelerationStructures-03748", 78}, { {"Each acceleration structure in pAccelerationStructures must have been created with VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR", 127}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureNV-pAccelerationStructures-03749", 78}, { {"If the nullDescriptor feature is not enabled, each member of pAccelerationStructures must not be VK_NULL_HANDLE", 111}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureNV-pAccelerationStructures-parameter", 82}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount valid or VK_NULL_HANDLE VkAccelerationStructureNV handles", 147}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetAccelerationStructureNV-sType-sType", 60}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_NV", 78}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetInlineUniformBlock-dataSize-02222", 58}, { {"dataSize must be an integer multiple of 4", 41}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetInlineUniformBlock-dataSize-arraylength", 64}, { {"dataSize must be greater than 0", 31}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetInlineUniformBlock-pData-parameter", 59}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetInlineUniformBlock-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_INLINE_UNIFORM_BLOCK", 73}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetPartitionedAccelerationStructureNV-accelerationStructureCount-10511", 92}, { {"accelerationStructureCount must be equal to descriptorCount in the extended structure", 85}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetPartitionedAccelerationStructureNV-accelerationStructureCount-arraylength", 98}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetPartitionedAccelerationStructureNV-pAccelerationStructures-10512", 89}, { {"Each entry in pAccelerationStructures must be a valid address of a PTLAS", 72}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetPartitionedAccelerationStructureNV-pAccelerationStructures-parameter", 93}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount or VK_NULL_HANDLE VkDeviceAddress values", 130}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetPartitionedAccelerationStructureNV-sType-sType", 71}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_PARTITIONED_ACCELERATION_STRUCTURE_NV", 90}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetTensorARM-nullDescriptor-09898", 55}, { {"If the nullDescriptor feature is not enabled, each element of pTensorViews must not be VK_NULL_HANDLE", 101}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetTensorARM-pTensorViews-parameter", 57}, { {"pTensorViews must be a valid pointer to an array of tensorViewCount valid or VK_NULL_HANDLE VkTensorViewARM handles", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetTensorARM-sType-sType", 46}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_TENSOR_ARM", 63}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteDescriptorSetTensorARM-tensorViewCount-arraylength", 62}, { {"tensorViewCount must be greater than 0", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-index-11026", 55}, { {"index must be less than the value of VkIndirectExecutionSetPipelineInfoEXT::maxPipelineCount used to create the set", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-index-11029", 55}, { {"index must not be referenced by submitted command buffers", 57}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-pipeline-11027", 58}, { {"pipeline must have been created with VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_EXT", 83}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-pipeline-11030", 58}, { {"The shader stages contained in pipeline must be supported by VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStagesPipelineBinding", 167}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-pipeline-parameter", 62}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetPipelineEXT-sType-sType", 55}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_INDIRECT_EXECUTION_SET_PIPELINE_EXT", 73}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-index-11031", 53}, { {"index must be less than VkIndirectExecutionSetShaderInfoEXT::maxShaderCount", 75}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-index-11034", 53}, { {"index must not be in use by submitted command buffers", 53}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-pInitialShaders-11033", 63}, { {"A shader created with the same VkShaderStageFlagBits must have been passed in the VkIndirectExecutionSetShaderInfoEXT::pInitialShaders array", 140}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-sType-sType", 53}, { {"sType must be VK_STRUCTURE_TYPE_WRITE_INDIRECT_EXECUTION_SET_SHADER_EXT", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-shader-11032", 54}, { {"shader must have been created with VK_SHADER_CREATE_INDIRECT_BINDABLE_BIT_EXT", 77}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkWriteIndirectExecutionSetShaderEXT-shader-parameter", 58}, { {"shader must be a valid VkShaderEXT handle", 41}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-VkXcbSurfaceCreateInfoKHR-connection-01310", 47}, { {"connection must point to a valid X11 xcb_connection_t", 53}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXcbSurfaceCreateInfoKHR-flags-zerobitmask", 48}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXcbSurfaceCreateInfoKHR-pNext-pNext", 42}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXcbSurfaceCreateInfoKHR-sType-sType", 42}, { {"sType must be VK_STRUCTURE_TYPE_XCB_SURFACE_CREATE_INFO_KHR", 59}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXcbSurfaceCreateInfoKHR-window-01311", 43}, { {"window must be a valid X11 xcb_window_t", 39}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXlibSurfaceCreateInfoKHR-dpy-01313", 41}, { {"dpy must point to a valid Xlib Display", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXlibSurfaceCreateInfoKHR-flags-zerobitmask", 49}, { {"flags must be 0", 15}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXlibSurfaceCreateInfoKHR-pNext-pNext", 43}, { {"pNext must be NULL", 18}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXlibSurfaceCreateInfoKHR-sType-sType", 43}, { {"sType must be VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR", 60}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-VkXlibSurfaceCreateInfoKHR-window-01314", 44}, { {"window must be a valid Xlib Window", 34}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-WarpIDARM-WarpIDARM-07603", 30}, { {"The variable decorated with WarpIDARM must be declared using the Input Storage Class", 84}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpIDARM-WarpIDARM-07604", 30}, { {"The variable decorated with WarpIDARM must be declared as a scalar 32-bit integer value", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpIDNV-WarpIDNV-04420", 28}, { {"The variable decorated with WarpIDNV must be declared using the Input Storage Class", 83}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpIDNV-WarpIDNV-04421", 28}, { {"The variable decorated with WarpIDNV must be declared as a scalar 32-bit integer value", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpMaxIDARM-WarpMaxIDARM-07601", 36}, { {"The variable decorated with WarpMaxIDARM must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpMaxIDARM-WarpMaxIDARM-07602", 36}, { {"The variable decorated with WarpMaxIDARM must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpsPerSMNV-WarpsPerSMNV-04418", 36}, { {"The variable decorated with WarpsPerSMNV must be declared using the Input Storage Class", 87}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WarpsPerSMNV-WarpsPerSMNV-04419", 36}, { {"The variable decorated with WarpsPerSMNV must be declared as a scalar 32-bit integer value", 90}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupId-WorkgroupId-04422", 34}, { {"The WorkgroupId decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 118}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupId-WorkgroupId-04423", 34}, { {"The variable decorated with WorkgroupId must be declared using the Input Storage Class", 86}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupId-WorkgroupId-04424", 34}, { {"The variable decorated with WorkgroupId must be declared as a three-component vector of 32-bit integer values", 109}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupSize-TileShadingRateQCOM-10635", 44}, { {"If the TileShadingRateQCOM Execution Mode is used, variables decorated with WorkgroupSize must be declared using the Input Storage Class", 136}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupSize-WorkgroupSize-04425", 38}, { {"The WorkgroupSize decoration must be used only within the GLCompute, MeshEXT, TaskEXT, MeshNV, or TaskNV Execution Model", 120}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupSize-WorkgroupSize-04426", 38}, { {"The variable decorated with WorkgroupSize must be a specialization constant or a constant", 89}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorkgroupSize-WorkgroupSize-04427", 38}, { {"The variable decorated with WorkgroupSize must be declared as a three-component vector of 32-bit integer values", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayDirectionKHR-WorldRayDirectionKHR-04428", 52}, { {"The WorldRayDirectionKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 134}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayDirectionKHR-WorldRayDirectionKHR-04429", 52}, { {"The variable decorated with WorldRayDirectionKHR must be declared using the Input Storage Class", 95}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayDirectionKHR-WorldRayDirectionKHR-04430", 52}, { {"The variable decorated with WorldRayDirectionKHR must be declared as a three-component vector of 32-bit floating-point values", 125}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayOriginKHR-WorldRayOriginKHR-04431", 46}, { {"The WorldRayOriginKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, ClosestHitKHR, or MissKHR Execution Model", 131}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayOriginKHR-WorldRayOriginKHR-04432", 46}, { {"The variable decorated with WorldRayOriginKHR must be declared using the Input Storage Class", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldRayOriginKHR-WorldRayOriginKHR-04433", 46}, { {"The variable decorated with WorldRayOriginKHR must be declared as a three-component vector of 32-bit floating-point values", 122}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldToObjectKHR-WorldToObjectKHR-04434", 44}, { {"The WorldToObjectKHR decoration must be used only within the IntersectionKHR, AnyHitKHR, or ClosestHitKHR Execution Model", 121}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldToObjectKHR-WorldToObjectKHR-04435", 44}, { {"The variable decorated with WorldToObjectKHR must be declared using the Input Storage Class", 91}, { "chapters/interfaces.html", 24} } },
        { {"VUID-WorldToObjectKHR-WorldToObjectKHR-04436", 44}, { {"The variable decorated with WorldToObjectKHR must be declared as a matrix with four columns of three-component vectors of 32-bit floating-point values", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkAcquireDrmDisplayEXT-display-parameter", 45}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireDrmDisplayEXT-display-parent", 42}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireDrmDisplayEXT-physicalDevice-parameter", 52}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-swapchain-02674", 56}, { {"swapchain must not be in the retired state", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-swapchain-02675", 56}, { {"swapchain must be a swapchain created with a VkSurfaceFullScreenExclusiveInfoEXT structure, with fullScreenExclusive set to VK_FULL_SCREEN_EXCLUSIVE_APPLICATION_CONTROLLED_EXT", 175}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-swapchain-02676", 56}, { {"swapchain must not currently have exclusive full-screen access", 62}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-swapchain-parameter", 60}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireFullScreenExclusiveModeEXT-swapchain-parent", 57}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImage2KHR-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImage2KHR-pAcquireInfo-parameter", 50}, { {"pAcquireInfo must be a valid pointer to a valid VkAcquireNextImageInfoKHR structure", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImage2KHR-pImageIndex-parameter", 49}, { {"pImageIndex must be a valid pointer to a uint32_t value", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImage2KHR-surface-07784", 41}, { {"If forward progress cannot be guaranteed for the surface used to create swapchain, the timeout member of pAcquireInfo must not be UINT64_MAX", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-fence-01287", 38}, { {"If fence is not VK_NULL_HANDLE, fence must be unsignaled", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-fence-10066", 38}, { {"If fence is not VK_NULL_HANDLE, fence must not be associated with any other queue command that has not yet completed execution on that queue", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-fence-parameter", 42}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-fence-parent", 39}, { {"If fence is a valid handle, it must have been created, allocated, or retrieved from device", 90}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-pImageIndex-parameter", 48}, { {"pImageIndex must be a valid pointer to a uint32_t value", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-01286", 42}, { {"If semaphore is not VK_NULL_HANDLE, it must be unsignaled", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-01779", 42}, { {"If semaphore is not VK_NULL_HANDLE, it must not have any uncompleted signal or wait operations pending", 102}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-01780", 42}, { {"semaphore and fence must not both be equal to VK_NULL_HANDLE", 60}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-03265", 42}, { {"semaphore must have a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 65}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-parameter", 46}, { {"If semaphore is not VK_NULL_HANDLE, semaphore must be a valid VkSemaphore handle", 80}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-semaphore-parent", 43}, { {"If semaphore is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-surface-07783", 40}, { {"If forward progress cannot be guaranteed for the surface used to create the swapchain member of pAcquireInfo, timeout must not be UINT64_MAX", 140}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-swapchain-01285", 42}, { {"swapchain must not be in the retired state", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-swapchain-parameter", 46}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireNextImageKHR-swapchain-parent", 43}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquirePerformanceConfigurationINTEL-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkAcquirePerformanceConfigurationINTEL-pAcquireInfo-parameter", 66}, { {"pAcquireInfo must be a valid pointer to a valid VkPerformanceConfigurationAcquireInfoINTEL structure", 100}, { "chapters/queries.html", 21} } },
        { {"VUID-vkAcquirePerformanceConfigurationINTEL-pConfiguration-parameter", 68}, { {"pConfiguration must be a valid pointer to a VkPerformanceConfigurationINTEL handle", 82}, { "chapters/queries.html", 21} } },
        { {"VUID-vkAcquireProfilingLockKHR-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkAcquireProfilingLockKHR-pInfo-parameter", 46}, { {"pInfo must be a valid pointer to a valid VkAcquireProfilingLockInfoKHR structure", 80}, { "chapters/queries.html", 21} } },
        { {"VUID-vkAcquireWinrtDisplayNV-display-parameter", 46}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireWinrtDisplayNV-display-parent", 43}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireWinrtDisplayNV-physicalDevice-parameter", 53}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireXlibDisplayEXT-display-parameter", 46}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireXlibDisplayEXT-display-parent", 43}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireXlibDisplayEXT-dpy-parameter", 42}, { {"dpy must be a valid pointer to a Display value", 46}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAcquireXlibDisplayEXT-physicalDevice-parameter", 53}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAllocateCommandBuffers-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkAllocateCommandBuffers-device-queuecount", 47}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkAllocateCommandBuffers-pAllocateInfo-parameter", 53}, { {"pAllocateInfo must be a valid pointer to a valid VkCommandBufferAllocateInfo structure", 86}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkAllocateCommandBuffers-pAllocateInfo::commandBufferCount-arraylength", 75}, { {"pAllocateInfo->commandBufferCount must be greater than 0", 56}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkAllocateCommandBuffers-pCommandBuffers-parameter", 55}, { {"pCommandBuffers must be a valid pointer to an array of pAllocateInfo->commandBufferCount VkCommandBuffer handles", 112}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkAllocateDescriptorSets-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkAllocateDescriptorSets-device-queuecount", 47}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkAllocateDescriptorSets-pAllocateInfo-parameter", 53}, { {"pAllocateInfo must be a valid pointer to a valid VkDescriptorSetAllocateInfo structure", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkAllocateDescriptorSets-pAllocateInfo::descriptorSetCount-arraylength", 75}, { {"pAllocateInfo->descriptorSetCount must be greater than 0", 56}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkAllocateDescriptorSets-pDescriptorSets-parameter", 55}, { {"pDescriptorSets must be a valid pointer to an array of pAllocateInfo->descriptorSetCount VkDescriptorSet handles", 112}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkAllocateMemory-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-device-queuecount", 39}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-deviceCoherentMemory-02790", 48}, { {"If the deviceCoherentMemory feature is not enabled, pAllocateInfo->memoryTypeIndex must not identify a memory type supporting VK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD", 168}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-maxMemoryAllocationCount-04101", 52}, { {"There must be less than VkPhysicalDeviceLimits::maxMemoryAllocationCount device memory allocations currently allocated on the device", 132}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-pAllocateInfo-01713", 41}, { {"pAllocateInfo->allocationSize must be less than or equal to VkPhysicalDeviceMemoryProperties::memoryHeaps[memindex].size where memindex = VkPhysicalDeviceMemoryProperties::memoryTypes[pAllocateInfo->memoryTypeIndex].heapIndex as returned by vkGetPhysicalDeviceMemoryProperties for the VkPhysicalDevice that device was created from", 330}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-pAllocateInfo-01714", 41}, { {"pAllocateInfo->memoryTypeIndex must be less than VkPhysicalDeviceMemoryProperties::memoryTypeCount as returned by vkGetPhysicalDeviceMemoryProperties for the VkPhysicalDevice that device was created from", 203}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-pAllocateInfo-parameter", 45}, { {"pAllocateInfo must be a valid pointer to a valid VkMemoryAllocateInfo structure", 79}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-pAllocator-parameter", 42}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-pMemory-parameter", 39}, { {"pMemory must be a valid pointer to a VkDeviceMemory handle", 58}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAllocateMemory-tileMemoryHeap-10976", 42}, { {"If the tileMemoryHeap feature is not enabled, pAllocateInfo->memoryTypeIndex must not identify a memory type that corresponds to a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property", 197}, { "chapters/memory.html", 20} } },
        { {"VUID-vkAntiLagUpdateAMD-antiLag-10061", 37}, { {"The antiLag feature must be enabled", 35}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAntiLagUpdateAMD-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkAntiLagUpdateAMD-pData-parameter", 39}, { {"pData must be a valid pointer to a valid VkAntiLagDataAMD structure", 67}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-00049", 45}, { {"commandBuffer must not be in the recording or pending state", 59}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-00050", 45}, { {"If commandBuffer was allocated from a VkCommandPool which did not have the VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT flag set, commandBuffer must be in the initial state", 175}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-00051", 45}, { {"If commandBuffer is a secondary command buffer, the pInheritanceInfo member of pBeginInfo must be a valid VkCommandBufferInheritanceInfo structure", 146}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-00052", 45}, { {"If commandBuffer is a secondary command buffer and either the occlusionQueryEnable member of the pInheritanceInfo member of pBeginInfo is VK_FALSE, or the occlusionQueryPrecise feature is not enabled, then pBeginInfo->pInheritanceInfo->queryFlags must not contain VK_QUERY_CONTROL_PRECISE_BIT", 292}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-02840", 45}, { {"If commandBuffer is a primary command buffer, then pBeginInfo->flags must not set both the VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT and the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT flags", 193}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBeginCommandBuffer-pBeginInfo-parameter", 46}, { {"pBeginInfo must be a valid pointer to a valid VkCommandBufferBeginInfo structure", 80}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkBindAccelerationStructureMemoryNV-bindInfoCount-arraylength", 66}, { {"bindInfoCount must be greater than 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindAccelerationStructureMemoryNV-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindAccelerationStructureMemoryNV-pBindInfos-parameter", 61}, { {"pBindInfos must be a valid pointer to an array of bindInfoCount valid VkBindAccelerationStructureMemoryInfoNV structures", 120}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-None-01898", 34}, { {"If buffer was created with the VK_BUFFER_CREATE_PROTECTED_BIT bit set, the buffer must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 189}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-None-01899", 34}, { {"If buffer was created with the VK_BUFFER_CREATE_PROTECTED_BIT bit not set, the buffer must not be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 197}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-None-10739", 34}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer", 274}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-None-10741", 34}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer must be less than or equal to the size of memory minus memoryOffset", 289}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-apiVersion-07920", 40}, { {"If the VK_KHR_dedicated_allocation extension is not enabled, VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and buffer was not created with VkDedicatedAllocationBufferCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must not have been allocated dedicated for a specific buffer or image", 312}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-01030", 36}, { {"buffer must not have been created with any sparse memory binding flags", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-01038", 36}, { {"If buffer was created with VkDedicatedAllocationBufferCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must have been allocated with VkDedicatedAllocationMemoryAllocateInfoNV::buffer equal to a buffer handle created with identical creation parameters to buffer and memoryOffset must be zero", 300}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-01444", 36}, { {"If buffer requires a dedicated allocation (as reported by vkGetBufferMemoryRequirements2 in VkMemoryDedicatedRequirements::requiresDedicatedAllocation for buffer), memory must have been allocated with VkMemoryDedicatedAllocateInfo::buffer equal to buffer", 254}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-06408", 36}, { {"If buffer was created with VkBufferCollectionBufferCreateInfoFUCHSIA chained to VkBufferCreateInfo::pNext, memory must be allocated with a VkImportMemoryBufferCollectionFUCHSIA chained to VkMemoryAllocateInfo::pNext", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-07459", 36}, { {"buffer must not have been bound to a memory object", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-09201", 36}, { {"If the buffer was created with the VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 203}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-11408", 36}, { {"If the buffer was created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT or VK_BUFFER_USAGE_2_DESCRIPTOR_HEAP_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 215}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-parameter", 40}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-buffer-parent", 37}, { {"buffer must have been created, allocated, or retrieved from device", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-bufferDeviceAddress-03339", 49}, { {"If the VkPhysicalDeviceBufferDeviceAddressFeatures::bufferDeviceAddress feature is enabled and buffer was created with the VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT usage flag set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 267}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-bufferDeviceAddressCaptureReplay-09200", 62}, { {"If the VkPhysicalDeviceBufferDeviceAddressFeatures::bufferDeviceAddressCaptureReplay feature is enabled and buffer was created with the VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 297}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-descriptorBufferCaptureReplay-08112", 59}, { {"If the buffer was created with the VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 188}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-01035", 36}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetBufferMemoryRequirements with buffer", 202}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-01508", 36}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::buffer was not VK_NULL_HANDLE, then buffer must equal VkMemoryDedicatedAllocateInfo::buffer, and memoryOffset must be zero", 291}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-02726", 36}, { {"If the value of VkExportMemoryAllocateInfo::handleTypes used to allocate memory is not 0, it must include at least one of the handles set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 210}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-02985", 36}, { {"If memory was allocated by a memory import operation, that is not VkImportAndroidHardwareBufferInfoANDROID with a non-NULL buffer value, the external handle type of the imported memory must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 281}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-02986", 36}, { {"If memory was allocated with the VkImportAndroidHardwareBufferInfoANDROID memory import operation with a non-NULL buffer value, VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must also have been set in VkExternalMemoryBufferCreateInfo::handleTypes when buffer was created", 291}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-10740", 36}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetBufferMemoryRequirements with buffer", 278}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-10742", 36}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetBufferMemoryRequirements with buffer must be less than or equal to the size of memory minus memoryOffset", 293}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-10925", 36}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, VkMemoryDedicatedAllocateInfo::image must have been VK_NULL_HANDLE", 200}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-parameter", 40}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memory-parent", 37}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory-memoryOffset-01031", 42}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory2-bindInfoCount-arraylength", 50}, { {"bindInfoCount must be greater than 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory2-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindBufferMemory2-pBindInfos-parameter", 45}, { {"pBindInfos must be a valid pointer to an array of bindInfoCount valid VkBindBufferMemoryInfo structures", 103}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindDataGraphPipelineSessionMemoryARM-bindInfoCount-arraylength", 70}, { {"bindInfoCount must be greater than 0", 36}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkBindDataGraphPipelineSessionMemoryARM-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkBindDataGraphPipelineSessionMemoryARM-pBindInfos-parameter", 65}, { {"pBindInfos must be a valid pointer to an array of bindInfoCount valid VkBindDataGraphPipelineSessionMemoryInfoARM structures", 124}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkBindImageMemory-None-01901", 33}, { {"If image was created with the VK_IMAGE_CREATE_PROTECTED_BIT bit set, the image must be bound to a memory object allocated with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 186}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-None-01902", 33}, { {"If image was created with the VK_IMAGE_CREATE_PROTECTED_BIT bit not set, the image must not be bound to a memory object created with a memory type that reports VK_MEMORY_PROPERTY_PROTECTED_BIT", 192}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-None-10735", 33}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements with image", 272}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-None-10737", 33}, { {"If memory was not allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements with image must be less than or equal to the size of memory minus memoryOffset", 287}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-apiVersion-07921", 39}, { {"If the VK_KHR_dedicated_allocation extension is not enabled, VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and image was not created with VkDedicatedAllocationImageCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must not have been allocated dedicated for a specific buffer or image", 310}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-01045", 34}, { {"image must not have been created with any sparse memory binding flags", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-01050", 34}, { {"If image was created with VkDedicatedAllocationImageCreateInfoNV::dedicatedAllocation equal to VK_TRUE, memory must have been created with VkDedicatedAllocationMemoryAllocateInfoNV::image equal to an image handle created with identical creation parameters to image and memoryOffset must be zero", 294}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-01445", 34}, { {"If image requires a dedicated allocation (as reported by vkGetImageMemoryRequirements2 in VkMemoryDedicatedRequirements::requiresDedicatedAllocation for image), memory must have been created with VkMemoryDedicatedAllocateInfo::image equal to image", 247}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-01608", 34}, { {"image must not have been created with the VK_IMAGE_CREATE_DISJOINT_BIT set", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-06392", 34}, { {"If image was created with VkBufferCollectionImageCreateInfoFUCHSIA chained to VkImageCreateInfo::pNext, memory must be allocated with a VkImportMemoryBufferCollectionFUCHSIA chained to VkMemoryAllocateInfo::pNext", 212}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-07460", 34}, { {"image must not have been bound to a memory object", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-08113", 34}, { {"If the image was created with the VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT bit set", 184}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-09202", 34}, { {"If the image was created with the VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT bit set, memory must have been allocated with the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT bit set", 199}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-parameter", 38}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-image-parent", 35}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-01047", 35}, { {"memory must have been allocated using one of the memory types allowed in the memoryTypeBits member of the VkMemoryRequirements structure returned from a call to vkGetImageMemoryRequirements with image", 200}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-02628", 35}, { {"If the dedicatedAllocationImageAliasing feature is not enabled, and the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::image was not VK_NULL_HANDLE, then image must equal VkMemoryDedicatedAllocateInfo::image and memoryOffset must be zero", 352}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-02629", 35}, { {"If the dedicatedAllocationImageAliasing feature is enabled, and the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, and VkMemoryDedicatedAllocateInfo::image was not VK_NULL_HANDLE, then memoryOffset must be zero, and image must be either equal to VkMemoryDedicatedAllocateInfo::image or an image that was created using the same parameters in VkImageCreateInfo, with the exception that extent and arrayLayers may differ subject to the following restrictions: every dimension in the extent parameter of the image being bound must be equal to or smaller than the original image for which the allocation was created; and the arrayLayers parameter of the image being bound must be equal to or smaller than the original image for which the allocation was created", 835}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-02728", 35}, { {"If the value of VkExportMemoryAllocateInfo::handleTypes used to allocate memory is not 0, it must include at least one of the handles set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 208}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-02989", 35}, { {"If memory was created by a memory import operation, that is not VkImportAndroidHardwareBufferInfoANDROID with a non-NULL buffer value, the external handle type of the imported memory must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 277}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-02990", 35}, { {"If memory was created with the VkImportAndroidHardwareBufferInfoANDROID memory import operation with a non-NULL buffer value, VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID must also have been set in VkExternalMemoryImageCreateInfo::handleTypes when image was created", 287}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-10736", 35}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, memoryOffset must be an integer multiple of the alignment member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetImageMemoryRequirements2 with image", 277}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-10738", 35}, { {"If memory was allocated from a memory heap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property set, size member of the VkTileMemoryRequirementsQCOM structure returned from a call to vkGetImageMemoryRequirements2 with image must be less than or equal to the size of memory minus memoryOffset", 292}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-10926", 35}, { {"If the VkMemoryAllocateInfo provided when memory was allocated included a VkMemoryDedicatedAllocateInfo structure in its pNext chain, VkMemoryDedicatedAllocateInfo::buffer must have been VK_NULL_HANDLE", 201}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-parameter", 39}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memory-parent", 36}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory-memoryOffset-01046", 41}, { {"memoryOffset must be less than the size of memory", 49}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory2-bindInfoCount-arraylength", 49}, { {"bindInfoCount must be greater than 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory2-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory2-pBindInfos-02858", 40}, { {"If any VkBindImageMemoryInfo::image was created with VK_IMAGE_CREATE_DISJOINT_BIT then all planes of VkBindImageMemoryInfo::image must be bound individually in separate pBindInfos", 179}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory2-pBindInfos-04006", 40}, { {"pBindInfos must not refer to the same image subresource more than once", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindImageMemory2-pBindInfos-parameter", 44}, { {"pBindInfos must be a valid pointer to an array of bindInfoCount valid VkBindImageMemoryInfo structures", 102}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-bindingPoint-parameter", 59}, { {"bindingPoint must be a valid VkOpticalFlowSessionBindingPointNV value", 69}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-layout-parameter", 53}, { {"layout must be a valid VkImageLayout value", 42}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-session-parameter", 54}, { {"session must be a valid VkOpticalFlowSessionNV handle", 53}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-session-parent", 51}, { {"session must have been created, allocated, or retrieved from device", 67}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-view-parameter", 51}, { {"If view is not VK_NULL_HANDLE, view must be a valid VkImageView handle", 70}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindOpticalFlowSessionImageNV-view-parent", 48}, { {"If view is a valid handle, it must have been created, allocated, or retrieved from device", 89}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkBindTensorMemoryARM-bindInfoCount-arraylength", 52}, { {"bindInfoCount must be greater than 0", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindTensorMemoryARM-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindTensorMemoryARM-pBindInfos-parameter", 47}, { {"pBindInfos must be a valid pointer to an array of bindInfoCount valid VkBindTensorMemoryInfoARM structures", 106}, { "chapters/resources.html", 23} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-bindSessionMemoryInfoCount-arraylength", 71}, { {"bindSessionMemoryInfoCount must be greater than 0", 49}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-memoryBindIndex-07196", 54}, { {"The memoryBindIndex member of each element of pBindSessionMemoryInfos must be unique within pBindSessionMemoryInfos", 115}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-pBindSessionMemoryInfos-07197", 62}, { {"Each element of pBindSessionMemoryInfos must have a corresponding VkMemoryRequirements structure", 96}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-pBindSessionMemoryInfos-07198", 62}, { {"If an element of pBindSessionMemoryInfos has a corresponding VkMemoryRequirements structure, then the memory member of that element of pBindSessionMemoryInfos must have been allocated using one of the memory types allowed in the memoryTypeBits member of the corresponding VkMemoryRequirements structure", 302}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-pBindSessionMemoryInfos-07199", 62}, { {"If an element of pBindSessionMemoryInfos has a corresponding VkMemoryRequirements structure, then the memoryOffset member of that element of pBindSessionMemoryInfos must be an integer multiple of the alignment member of the corresponding VkMemoryRequirements structure", 268}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-pBindSessionMemoryInfos-07200", 62}, { {"If an element of pBindSessionMemoryInfos has a corresponding VkMemoryRequirements structure, then the memorySize member of that element of pBindSessionMemoryInfos must equal the size member of the corresponding VkMemoryRequirements structure", 241}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-pBindSessionMemoryInfos-parameter", 66}, { {"pBindSessionMemoryInfos must be a valid pointer to an array of bindSessionMemoryInfoCount valid VkBindVideoSessionMemoryInfoKHR structures", 138}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-videoSession-07195", 51}, { {"The memory binding of videoSession identified by the memoryBindIndex member of any element of pBindSessionMemoryInfos must not already be backed by a memory object", 163}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-videoSession-parameter", 55}, { {"videoSession must be a valid VkVideoSessionKHR handle", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBindVideoSessionMemoryKHR-videoSession-parent", 52}, { {"videoSession must have been created, allocated, or retrieved from device", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-None-03407", 48}, { {"The dstAccelerationStructure member of any element of pInfos must not be referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-accelerationStructureHostCommands-03581", 77}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureHostCommands feature must be enabled", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-deferredOperation-03678", 61}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-deferredOperation-parameter", 65}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-deferredOperation-parent", 62}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03698", 68}, { {"The dstAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03701", 68}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any other element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, which is accessed by this command", 328}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03702", 68}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the dstAccelerationStructure member of any other element of pInfos, which is accessed by this command", 259}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03703", 68}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any element of pInfos (including the same element), which is accessed by this command", 269}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03706", 68}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing any acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos, which is accessed by this command", 399}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-dstAccelerationStructure-03800", 68}, { {"The dstAccelerationStructure member of any element of pInfos must be a valid VkAccelerationStructureKHR handle", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-infoCount-arraylength", 59}, { {"infoCount must be greater than 0", 32}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-mode-04628", 48}, { {"The mode member of each element of pInfos must be a valid VkBuildAccelerationStructureModeKHR value", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03403", 50}, { {"The srcAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03663", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, inactive primitives in its srcAccelerationStructure member must not be made active", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03664", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, active primitives in its srcAccelerationStructure member must not be made inactive", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03667", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must have previously been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_KHR set in VkAccelerationStructureBuildGeometryInfoKHR::flags in the build", 301}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03668", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure and dstAccelerationStructure members must either be the same VkAccelerationStructureKHR, or not have any memory aliasing", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03699", 50}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03700", 50}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 326}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03707", 50}, { {"For each element of pInfos, its dstAccelerationStructure member must be fully backed by physical memory", 103}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03708", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, srcAccelerationStructure must be fully backed by physical memory", 163}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03709", 50}, { {"For each element of pInfos, if an acceleration structure is referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, it must be fully backed by physical memory", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03722", 50}, { {"For each element of pInfos, its dstAccelerationStructure member must be bound to host-visible device memory", 107}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03723", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR the buffer used to create its srcAccelerationStructure member must be bound to host-visible device memory", 203}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03724", 50}, { {"For each element of pInfos, each acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR must be bound to host-visible device memory", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03758", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its geometryCount member must have the same value which was specified when srcAccelerationStructure was last built", 213}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03759", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 205}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03760", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its type member must have the same value which was specified when srcAccelerationStructure was last built", 204}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03761", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its geometryType member must have the same value which was specified when srcAccelerationStructure was last built", 327}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03762", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03763", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.vertexFormat member must have the same value which was specified when srcAccelerationStructure was last built", 397}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03764", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.maxVertex member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03765", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.indexType member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03766", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was NULL when srcAccelerationStructure was last built, then it must be NULL", 388}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03767", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was not NULL when srcAccelerationStructure was last built, then it must not be NULL", 396}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03768", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, then the value of each index referenced must be the same as the corresponding index value when srcAccelerationStructure was last built", 463}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03771", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, geometry.triangles.vertexData.hostAddress must be a valid host address", 193}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03772", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, geometry.triangles.indexData.hostAddress must be a valid host address", 255}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03773", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.transformData.hostAddress is not 0, it must be a valid host address", 212}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03774", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR, geometry.aabbs.data.hostAddress must be a valid host address", 179}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03775", 50}, { {"For each element of pInfos, its dstAccelerationStructure member must be bound to memory that was not allocated with multiple instances", 134}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03776", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR the buffer used to create its srcAccelerationStructure member must be bound to memory that was not allocated with multiple instances", 230}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03777", 50}, { {"For each element of pInfos, each acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR must be bound to memory that was not allocated with multiple instances", 275}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03778", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, geometry.instances.data.hostAddress must be a valid host address", 187}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03779", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, each VkAccelerationStructureInstanceKHR::accelerationStructureReference value in geometry.instances.data.hostAddress must be a valid VkAccelerationStructureKHR object", 289}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-03801", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, the corresponding ppBuildRangeInfos[i][j].primitiveCount must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxInstanceCount", 278}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-04630", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must not be VK_NULL_HANDLE", 161}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-04930", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR with VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV set, each accelerationStructureReference in any structure in VkAccelerationStructureMotionInstanceNV value in geometry.instances.data.hostAddress must be a valid VkAccelerationStructureKHR object", 368}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10126", 50}, { {"For each pInfos[i], dstAccelerationStructure must have been created with a value of VkAccelerationStructureCreateInfoKHR::size greater than or equal to either: the memory size required by the build operation, as returned by vkGetAccelerationStructureBuildSizesKHR with pBuildInfo = pInfos[i] and with each element of the pMaxPrimitiveCounts array greater than or equal to the equivalent ppBuildRangeInfos[i][j].primitiveCount values for j in [0,pInfos[i].geometryCount) or, the result of querying the corresponding VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, if updating a compacted acceleration structure", 618}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10892", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.hostAddress must be 0", 327}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10893", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, geometryType must not be VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX", 147}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10898", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numTriangles member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 497}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10899", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numVertices member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 496}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10900", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxPrimitiveIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 502}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10901", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxGeometryIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 501}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10902", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the format member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 491}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-10903", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the dataSize member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 493}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11702", 50}, { {"Each element of ppBuildRangeInfos[i] must be a valid pointer to an array of pInfos[i].geometryCount VkAccelerationStructureBuildRangeInfoKHR structures", 151}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11703", 50}, { {"For each element of pInfos, its dstAccelerationStructure member must have been created with vkCreateAccelerationStructureKHR", 124}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11704", 50}, { {"For each element of pInfos, the acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR must have been created with vkCreateAccelerationStructureKHR", 264}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11705", 50}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in the geometry.triangles.pNext chain its micromap member must be VK_NULL_HANDLE", 280}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11706", 50}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must have been created with vkCreateAccelerationStructureKHR", 195}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11820", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureGeometryMotionTrianglesDataNV in the geometry.triangles.pNext chain, then its vertexData.hostAddress must not be 0", 287}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11821", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.hostAddress must not be 0", 335}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11822", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.hostAddress must be 0", 243}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11823", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.hostAddress must not be 0", 251}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11824", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::vertexData.hostAddress must not be 0", 202}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11825", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::radiusData.hostAddress must not be 0", 202}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11826", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.hostAddress must be 0", 267}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11827", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.hostAddress must not be 0", 275}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11828", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::vertexData.hostAddress must not be 0", 226}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-11829", 50}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::radiusData.hostAddress must not be 0", 226}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-12244", 50}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR, and N is not 0, all addresses between pInfos[i].scratchData.hostAddress and pInfos[i].scratchData.hostAddress + N - 1 must be valid host memory, where N is given by the buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 475}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-12245", 50}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, and N is not 0, all addresses between pInfos[i].scratchData.hostAddress and pInfos[i].scratchData.hostAddress + N - 1 must be valid host memory, where N is given by the updateScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 477}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-pInfos-parameter", 54}, { {"pInfos must be a valid pointer to an array of infoCount valid VkAccelerationStructureBuildGeometryInfoKHR structures", 116}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-ppBuildRangeInfos-parameter", 65}, { {"ppBuildRangeInfos must be a valid pointer to an array of infoCount VkAccelerationStructureBuildRangeInfoKHR structures", 118}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-primitiveCount-03769", 58}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, the primitiveCount member of its corresponding VkAccelerationStructureBuildRangeInfoKHR structure must have the same value which was specified when srcAccelerationStructure was last built", 401}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-scratchData-03704", 55}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any other element of pInfos, which is accessed by this command", 233}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-scratchData-03705", 55}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR (including the same element), which is accessed by this command", 338}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildAccelerationStructuresKHR-srcAccelerationStructure-04629", 68}, { {"If the srcAccelerationStructure member of any element of pInfos is not VK_NULL_HANDLE, the srcAccelerationStructure member must be a valid VkAccelerationStructureKHR handle", 172}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkBuildMicromapsEXT-deferredOperation-parameter", 52}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-deferredOperation-parent", 49}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-dstMicromap-07463", 42}, { {"The dstMicromap member of any element of pInfos must be a valid VkMicromapEXT handle", 84}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-dstMicromap-11667", 42}, { {"The range of memory backing the dstMicromap member of any element of pInfos that is accessed by this command must not overlap the memory backing the dstMicromap member of any other element of pInfos, which is accessed by this command", 233}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-dstMicromap-11668", 42}, { {"The range of memory backing the dstMicromap member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any element of pInfos (including the same element), which is accessed by this command", 256}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-infoCount-arraylength", 46}, { {"infoCount must be greater than 0", 32}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-micromapHostCommands-07555", 51}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapHostCommands feature must be enabled", 92}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-mode-07462", 35}, { {"The mode member of each element of pInfos must be a valid VkBuildMicromapModeEXT value", 86}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07461", 37}, { {"For each pInfos[i], dstMicromap must have been created with a value of VkMicromapCreateInfoEXT::size greater than or equal to the memory size required by the build operation, as returned by vkGetMicromapBuildSizesEXT with pBuildInfo = pInfos[i]", 244}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07464", 37}, { {"For each element of pInfos its type member must match the value of VkMicromapCreateInfoEXT::type when its dstMicromap was created", 129}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07552", 37}, { {"For each element of pInfos, the buffer used to create its dstMicromap member must be bound to host-visible device memory", 120}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07553", 37}, { {"For each element of pInfos, all referenced addresses of pInfos[i].data.hostAddress must be valid host memory", 108}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07554", 37}, { {"For each element of pInfos, all referenced addresses of pInfos[i].triangleArray.hostAddress must be valid host memory", 117}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07556", 37}, { {"If pInfos[i].mode is VK_BUILD_MICROMAP_MODE_BUILD_EXT, and N is not 0, then all addresses between pInfos[i].scratchData.hostAddress and pInfos[i].scratchData.hostAddress + N - 1 must be valid host memory, where N is given by the value of VkMicromapBuildSizesInfoEXT::buildScratchSize returned from a call to vkGetMicromapBuildSizesEXT with an identical VkMicromapBuildInfoEXT structure", 385}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-07557", 37}, { {"For each element of pInfos, the buffer used to create its dstMicromap member must be bound to memory that was not allocated with multiple instances", 147}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-pInfos-parameter", 41}, { {"pInfos must be a valid pointer to an array of infoCount valid VkMicromapBuildInfoEXT structures", 95}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkBuildMicromapsEXT-scratchData-11669", 42}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any other element of pInfos, which is accessed by this command", 233}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkClearShaderInstrumentationMetricsARM-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkClearShaderInstrumentationMetricsARM-instrumentation-parameter", 69}, { {"instrumentation must be a valid VkShaderInstrumentationARM handle", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkClearShaderInstrumentationMetricsARM-instrumentation-parent", 66}, { {"instrumentation must have been created, allocated, or retrieved from device", 75}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-None-13063", 50}, { {"Conditional rendering must not already be active", 48}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-pConditionalRenderingBegin-parameter", 76}, { {"pConditionalRenderingBegin must be a valid pointer to a valid VkConditionalRenderingBeginInfo2EXT structure", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-suspended", 49}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRendering2EXT-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-None-01980", 49}, { {"Conditional rendering must not already be active", 48}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-pConditionalRenderingBegin-parameter", 75}, { {"pConditionalRenderingBegin must be a valid pointer to a valid VkConditionalRenderingBeginInfoEXT structure", 106}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginConditionalRenderingEXT-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-None-11518", 42}, { {"vkCmdBeginCustomResolveEXT must not have already been recorded in the current render pass instance", 98}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-None-11519", 42}, { {"The current render pass instance must have specified VK_RENDERING_CUSTOM_RESOLVE_BIT_EXT", 88}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-None-11520", 42}, { {"The current render pass instance must not have specified VK_RENDERING_SUSPENDING_BIT", 84}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-commandBuffer-11517", 51}, { {"The current render pass instance must have been started or resumed by vkCmdBeginRendering in this commandBuffer", 111}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-pBeginCustomResolveInfo-parameter", 65}, { {"If pBeginCustomResolveInfo is not NULL, pBeginCustomResolveInfo must be a valid pointer to a valid VkBeginCustomResolveInfoEXT structure", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-renderpass", 42}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginCustomResolveEXT-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginDebugUtilsLabelEXT-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdBeginDebugUtilsLabelEXT-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdBeginDebugUtilsLabelEXT-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdBeginDebugUtilsLabelEXT-pLabelInfo-parameter", 54}, { {"pLabelInfo must be a valid pointer to a valid VkDebugUtilsLabelEXT structure", 76}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-commonparent", 40}, { {"Both of commandBuffer, and gpaSession must have been created, allocated, or retrieved from the same VkDevice", 108}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-gpaSession-parameter", 48}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-pGpaSampleBeginInfo-parameter", 57}, { {"pGpaSampleBeginInfo must be a valid pointer to a valid VkGpaSampleBeginInfoAMD structure", 88}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-pSampleID-parameter", 47}, { {"pSampleID must be a valid pointer to a uint32_t value", 53}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSampleAMD-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-commandBuffer-12410", 48}, { {"If another GPA session has been started with vkCmdBeginGpaSessionAMD in commandBuffer, it must have been ended using vkCmdEndGpaSessionAMD before this call", 155}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-commonparent", 41}, { {"Both of commandBuffer, and gpaSession must have been created, allocated, or retrieved from the same VkDevice", 108}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-gpaSession-12409", 45}, { {"If gpaSession has been used previously to begin and end a session, vkResetGpaSessionAMD must have first been called", 115}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-gpaSession-parameter", 49}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginGpaSessionAMD-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-None-10664", 46}, { {"The current render pass must be a tile shading render pass", 58}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-None-10665", 46}, { {"The tileShadingPerTileDispatch or tileShadingPerTileDraw feature must be enabled", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-pPerTileBeginInfo-parameter", 63}, { {"pPerTileBeginInfo must be a valid pointer to a valid VkPerTileBeginInfoQCOM structure", 85}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-renderpass", 46}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginPerTileExecutionQCOM-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginQuery-None-00807", 31}, { {"All queries used by the command must be unavailable", 51}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-None-02863", 31}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, this command must not be recorded in a command buffer that, either directly or through secondary command buffers, also contains a vkCmdResetQueryPool command affecting the same query", 264}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-None-07127", 31}, { {"If there is a bound video session, then there must be no active queries", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-None-08370", 31}, { {"If there is a bound video session, then it must not have been created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR", 121}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-None-10681", 31}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-commandBuffer-01885", 40}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-commonparent", 33}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-flags-parameter", 36}, { {"flags must be a valid combination of VkQueryControlFlagBits values", 66}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-query-00802", 32}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-query-00808", 32}, { {"If called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 184}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-01922", 36}, { {"queryPool must have been created with a queryType that differs from that of any queries that are active within commandBuffer", 124}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-03223", 36}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the profiling lock must have been held before vkBeginCommandBuffer was called on commandBuffer", 176}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-03224", 36}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_COMMAND_BUFFER_KHR, the query begin must be the first recorded command in commandBuffer", 250}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-03225", 36}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_RENDER_PASS_KHR, the begin command must not be recorded within a render pass instance", 248}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-03226", 36}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and another query pool with a queryType VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR has been used within commandBuffer, its parent primary command buffer or secondary command buffer recorded within the same parent primary command buffer as commandBuffer, the performanceCounterMultipleQueryPools feature must be enabled", 392}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-07289", 36}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then the VkQueryPoolPerformanceCreateInfoKHR::queueFamilyIndex queryPool was created with must equal the queue family index of the VkCommandPool that commandBuffer was allocated from", 264}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryPool-parameter", 40}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-00800", 36}, { {"If the occlusionQueryPrecise feature is not enabled, or the queryType used to create queryPool was not VK_QUERY_TYPE_OCCLUSION, flags must not contain VK_QUERY_CONTROL_PRECISE_BIT", 179}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-00803", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_OCCLUSION, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 159}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-00804", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PIPELINE_STATISTICS and any of the pipelineStatistics indicate graphics operations, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 232}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-00805", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PIPELINE_STATISTICS and any of the pipelineStatistics indicate compute operations, the VkCommandPool that commandBuffer was allocated from must support compute operations", 230}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-02327", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT the VkCommandPool that commandBuffer was allocated from must support graphics operations", 178}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-02328", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT then VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackQueries must be supported", 185}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-02804", 36}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_TIMESTAMP", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-04728", 36}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR", 169}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-04729", 36}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_NV", 105}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-06687", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT the VkCommandPool that commandBuffer was allocated from must support graphics operations", 173}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-06688", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT then primitivesGeneratedQuery must be enabled", 130}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-06741", 36}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR", 176}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07070", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_MESH_PRIMITIVES_GENERATED_EXT the VkCommandPool that commandBuffer was allocated from must support graphics operations", 178}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07126", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then the VkCommandPool that commandBuffer was allocated from must have been created with a queue family index that supports result status queries, as indicated by VkQueueFamilyQueryResultStatusPropertiesKHR::queryResultStatusSupport", 316}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07128", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR and there is a bound video session, then queryPool must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 366}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07129", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then there must be a bound video session", 127}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07130", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and there is a bound video session, then queryPool must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 369}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-07131", 36}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then there must be no bound video session", 172}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-08972", 36}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT, or VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT", 142}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-queryType-12423", 36}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIME_ELAPSED_QCOM, the command pool's queue family must support a non-zero timestampValidBits", 153}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQuery-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-None-00807", 41}, { {"All queries used by the command must be unavailable", 51}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-None-02863", 41}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, this command must not be recorded in a command buffer that, either directly or through secondary command buffers, also contains a vkCmdResetQueryPool command affecting the same query", 264}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-None-07127", 41}, { {"If there is a bound video session, then there must be no active queries", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-None-08370", 41}, { {"If there is a bound video session, then it must not have been created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR", 121}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-None-10681", 41}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-commandBuffer-01885", 50}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-commonparent", 43}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-flags-parameter", 46}, { {"flags must be a valid combination of VkQueryControlFlagBits values", 66}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-query-00802", 42}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-query-00808", 42}, { {"If called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 184}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-03223", 46}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the profiling lock must have been held before vkBeginCommandBuffer was called on commandBuffer", 176}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-03224", 46}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_COMMAND_BUFFER_KHR, the query begin must be the first recorded command in commandBuffer", 250}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-03225", 46}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_RENDER_PASS_KHR, the begin command must not be recorded within a render pass instance", 248}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-03226", 46}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and another query pool with a queryType VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR has been used within commandBuffer, its parent primary command buffer or secondary command buffer recorded within the same parent primary command buffer as commandBuffer, the performanceCounterMultipleQueryPools feature must be enabled", 392}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-04753", 46}, { {"If the queryPool was created with the same queryType as that of another active query within commandBuffer, then index must not match the index used for the active query", 168}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-07289", 46}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then the VkQueryPoolPerformanceCreateInfoKHR::queueFamilyIndex queryPool was created with must equal the queue family index of the VkCommandPool that commandBuffer was allocated from", 264}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryPool-parameter", 50}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-00800", 46}, { {"If the occlusionQueryPrecise feature is not enabled, or the queryType used to create queryPool was not VK_QUERY_TYPE_OCCLUSION, flags must not contain VK_QUERY_CONTROL_PRECISE_BIT", 179}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-00803", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_OCCLUSION, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 159}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-00804", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PIPELINE_STATISTICS and any of the pipelineStatistics indicate graphics operations, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 232}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-00805", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PIPELINE_STATISTICS and any of the pipelineStatistics indicate compute operations, the VkCommandPool that commandBuffer was allocated from must support compute operations", 230}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-02338", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT the VkCommandPool that commandBuffer was allocated from must support graphics operations", 178}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-02339", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT the index parameter must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 203}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-02341", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT then VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackQueries must be supported", 185}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-02804", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_TIMESTAMP", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-04728", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR", 169}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-04729", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_NV", 105}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06689", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT the VkCommandPool that commandBuffer was allocated from must support graphics operations", 173}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06690", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT the index parameter must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 198}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06691", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT and the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled, the index parameter must be zero", 192}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06692", 46}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT and not VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT, the index must be zero", 164}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06693", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT then primitivesGeneratedQuery must be enabled", 130}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-06741", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR", 176}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07071", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_MESH_PRIMITIVES_GENERATED_EXT", 94}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07126", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then the VkCommandPool that commandBuffer was allocated from must have been created with a queue family index that supports result status queries, as indicated by VkQueueFamilyQueryResultStatusPropertiesKHR::queryResultStatusSupport", 316}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07128", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR and there is a bound video session, then queryPool must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 366}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07129", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then there must be a bound video session", 127}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07130", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR and there is a bound video session, then queryPool must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 369}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-07131", 46}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then there must be no bound video session", 172}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-08972", 46}, { {"The queryType used to create queryPool must not be VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT, or VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT", 142}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-queryType-12423", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIME_ELAPSED_QCOM, the command pool's queue family must support a non-zero timestampValidBits", 153}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginQueryIndexedEXT-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdBeginRenderPass-bufferlevel", 37}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-contents-09640", 40}, { {"If contents is VK_SUBPASS_CONTENTS_INLINE_AND_SECONDARY_COMMAND_BUFFERS_KHR, then at least one of the following features must be enabled: maintenance7 nestedCommandBuffer", 170}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-contents-parameter", 44}, { {"contents must be a valid VkSubpassContents value", 48}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-dstStageMask-06452", 44}, { {"The dstStageMask members of any element of the pDependencies member of VkRenderPassCreateInfo used to create renderPass must be supported by the capabilities of the queue family identified by the queueFamilyIndex member of the VkCommandPoolCreateInfo used to create the command pool which commandBuffer was allocated from", 321}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-framebuffer-02532", 43}, { {"For any attachment in framebuffer that is used by renderPass and is bound to memory locations that are also bound to another attachment used by renderPass, and if at least one of those uses causes either attachment to be written to, both attachments must have had the VK_ATTACHMENT_DESCRIPTION_MAY_ALIAS_BIT set", 311}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-framebuffer-09045", 43}, { {"If any attachments specified in framebuffer are used by renderPass and are bound to overlapping memory locations, there must be only one that is used as a color attachment, depth/stencil, or resolve attachment in any subpass", 224}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-00895", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 494}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-00897", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 525}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-00898", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 486}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-00899", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 486}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-00900", 45}, { {"If the initialLayout member of any of the VkAttachmentDescription structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is not VK_IMAGE_LAYOUT_UNDEFINED, then each such initialLayout must be equal to the current layout of the corresponding attachment image subresource of the framebuffer specified in the framebuffer member of pRenderPassBegin", 395}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-01758", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 682}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-02842", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 636}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-07000", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with either the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set, and either the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT usage flags set", 659}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-07001", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT usage flag set", 518}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-initialLayout-09537", 45}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with either the VK_IMAGE_USAGE_STORAGE_BIT usage flag set, or both the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT and either of VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set", 650}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-pRenderPassBegin-parameter", 52}, { {"pRenderPassBegin must be a valid pointer to a valid VkRenderPassBeginInfo structure", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-renderpass", 36}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-srcStageMask-06451", 44}, { {"The srcStageMask members of any element of the pDependencies member of VkRenderPassCreateInfo used to create renderPass must be supported by the capabilities of the queue family identified by the queueFamilyIndex member of the VkCommandPoolCreateInfo used to create the command pool which commandBuffer was allocated from", 321}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-stencilInitialLayout-02843", 52}, { {"If any of the stencilInitialLayout or stencilFinalLayout member of the VkAttachmentDescriptionStencilLayout structures or the stencilLayout member of the VkAttachmentReferenceStencilLayout structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 597}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-bufferlevel", 38}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-dstStageMask-06454", 45}, { {"The dstStageMask members of any element of the pDependencies member of VkRenderPassCreateInfo used to create renderPass must be supported by the capabilities of the queue family identified by the queueFamilyIndex member of the VkCommandPoolCreateInfo used to create the command pool which commandBuffer was allocated from", 321}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-flags-10652", 38}, { {"If VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM was included in the VkRenderPassTileShadingCreateInfoQCOM::flags used to create the renderPass, commandBuffer must not have been recorded with VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT", 234}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-framebuffer-02533", 44}, { {"For any attachment in framebuffer that is used by renderPass and is bound to memory locations that are also bound to another attachment used by renderPass, and if at least one of those uses causes either attachment to be written to, both attachments must have had the VK_ATTACHMENT_DESCRIPTION_MAY_ALIAS_BIT set", 311}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-framebuffer-02779", 44}, { {"Both the framebuffer and renderPass members of pRenderPassBegin must have been created on the same VkDevice that commandBuffer was allocated on", 143}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-framebuffer-09046", 44}, { {"If any attachments specified in framebuffer are used by renderPass and are bound to overlapping memory locations, there must be only one that is used as a color attachment, depth/stencil, or resolve attachment in any subpass", 224}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-02844", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 636}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03094", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT usage flag set", 494}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03096", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 682}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03097", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_SAMPLED_BIT or VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT usage flags set", 525}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03098", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 486}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03099", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 486}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-03100", 46}, { {"If the initialLayout member of any of the VkAttachmentDescription structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is not VK_IMAGE_LAYOUT_UNDEFINED, then each such initialLayout must be equal to the current layout of the corresponding attachment image subresource of the framebuffer specified in the framebuffer member of pRenderPassBegin", 395}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-07002", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with either the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set, and either the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT or VK_IMAGE_USAGE_SAMPLED_BIT usage flags set", 659}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-07003", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT usage flag set", 518}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-initialLayout-09538", 46}, { {"If any of the initialLayout or finalLayout member of the VkAttachmentDescription structures or the layout member of the VkAttachmentReference structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with either the VK_IMAGE_USAGE_STORAGE_BIT usage flag set, or both the VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT and either of the VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT or VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flags set", 654}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-pRenderPassBegin-parameter", 53}, { {"pRenderPassBegin must be a valid pointer to a valid VkRenderPassBeginInfo structure", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-pSubpassBeginInfo-parameter", 54}, { {"pSubpassBeginInfo must be a valid pointer to a valid VkSubpassBeginInfo structure", 81}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-renderpass", 37}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-srcStageMask-06453", 45}, { {"The srcStageMask members of any element of the pDependencies member of VkRenderPassCreateInfo used to create renderPass must be supported by the capabilities of the queue family identified by the queueFamilyIndex member of the VkCommandPoolCreateInfo used to create the command pool which commandBuffer was allocated from", 321}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-stencilInitialLayout-02845", 53}, { {"If any of the stencilInitialLayout or stencilFinalLayout member of the VkAttachmentDescriptionStencilLayout structures or the stencilLayout member of the VkAttachmentReferenceStencilLayout structures specified when creating the render pass specified in the renderPass member of pRenderPassBegin is VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL, or VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL then the corresponding attachment image view of the framebuffer specified in the framebuffer member of pRenderPassBegin must have been created with the VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT usage flag set", 597}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRenderPass2-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-commandBuffer-06068", 44}, { {"If commandBuffer is a secondary command buffer, and the nestedCommandBuffer feature is not enabled, pRenderingInfo->flags must not include VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT", 190}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-commandBuffer-10914", 44}, { {"If commandBuffer is a secondary command buffer, VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT must not have been set in VkCommandBufferBeginInfo::flags when commandBuffer began", 179}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-dynamicRendering-06446", 47}, { {"The dynamicRendering feature must be enabled", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-flags-10641", 36}, { {"If VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM is included in VkRenderPassTileShadingCreateInfoQCOM::flags, commandBuffer must not have been recorded with VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT", 199}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-flags-10642", 36}, { {"VkRenderPassTileShadingCreateInfoQCOM::flags must not include VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-imageView-12276", 40}, { {"If VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView is not equal to VK_NULL_HANDLE, when imageView is accessed it must be in the layout specified by VkRenderingFragmentDensityMapAttachmentInfoEXT::imageLayout", 217}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-imageView-12277", 40}, { {"If VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView is not equal to VK_NULL_HANDLE, when imageView is accessed it must be in the layout specified by VkRenderingFragmentShadingRateAttachmentInfoKHR::imageLayout", 219}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09588", 45}, { {"If pRenderingInfo->pDepthAttachment is not NULL and pRenderingInfo->pDepthAttachment->imageView is not VK_NULL_HANDLE, when pRenderingInfo->pDepthAttachment->imageView is accessed it must be in the layout specified by pRenderingInfo->pDepthAttachment->imageLayout", 263}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09589", 45}, { {"If pRenderingInfo->pDepthAttachment is not NULL, pRenderingInfo->pDepthAttachment->imageView is not VK_NULL_HANDLE, pRenderingInfo->pDepthAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pRenderingInfo->pDepthAttachment->resolveImageView is not VK_NULL_HANDLE, pRenderingInfo->pDepthAttachment->resolveImageView must be in the layout specified by pRenderingInfo->pDepthAttachment->resolveImageLayout", 407}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09590", 45}, { {"If pRenderingInfo->pStencilAttachment is not NULL and pRenderingInfo->pStencilAttachment->imageView is not VK_NULL_HANDLE, when pRenderingInfo->pStencilAttachment->imageView is accessed it must be in the layout specified by pRenderingInfo->pStencilAttachment->imageLayout", 271}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09591", 45}, { {"If pRenderingInfo->pStencilAttachment is not NULL, pRenderingInfo->pStencilAttachment->imageView is not VK_NULL_HANDLE, pRenderingInfo->pStencilAttachment->resolveMode is not VK_RESOLVE_MODE_NONE, and pRenderingInfo->pStencilAttachment->resolveImageView is not VK_NULL_HANDLE, pRenderingInfo->pStencilAttachment->resolveImageView must be in the layout specified by pRenderingInfo->pStencilAttachment->resolveImageLayout", 419}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09592", 45}, { {"For each element of pRenderingInfo->pColorAttachments, if imageView is not VK_NULL_HANDLE, when that image view is accessed it must be in the layout specified by the imageLayout member of that same element of pRenderingInfo->pColorAttachments", 242}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-09593", 45}, { {"For each element of pRenderingInfo->pColorAttachments, if either resolveMode is VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, or imageView is not VK_NULL_HANDLE and resolveMode is not VK_RESOLVE_MODE_NONE, and resolveImageView is not VK_NULL_HANDLE, resolveImageView must be in the layout specified by resolveImageLayout", 330}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-11750", 45}, { {"If pRenderingInfo->flags contains VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR, maintenance10 must be enabled", 122}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-11751", 45}, { {"If pRenderingInfo->flags does not contain VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR, attachments must not specify VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 187}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-pRenderingInfo-parameter", 49}, { {"pRenderingInfo must be a valid pointer to a valid VkRenderingInfo structure", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-renderpass", 35}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginRendering-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commandBuffer-12375", 59}, { {"This command must not be recorded while shader instrumentation is active within commandBuffer", 93}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commandBuffer-12376", 59}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-commonparent", 52}, { {"Both of commandBuffer, and instrumentation must have been created, allocated, or retrieved from the same VkDevice", 113}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-instrumentation-parameter", 65}, { {"instrumentation must be a valid VkShaderInstrumentationARM handle", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-suspended", 49}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginShaderInstrumentationARM-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-None-02367", 47}, { {"Transform feedback must not be active", 37}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-None-02373", 47}, { {"Transform feedback must not be made active in a render pass instance with multiview enabled", 91}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-None-04128", 47}, { {"The last pre-rasterization shader stage of the bound graphics pipeline must have been declared with the Xfb execution mode", 122}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-None-06233", 47}, { {"If the shaderObject feature is not enabled, a valid graphics pipeline must be bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 118}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-None-10656", 47}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-firstCounter-02368", 55}, { {"firstCounterRange must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 111}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-firstCounter-02369", 55}, { {"The sum of firstCounterRange and counterRangeCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 156}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-firstCounter-09630", 55}, { {"The sum of firstCounterRange and counterRangeCount must be less than or equal to the number of transform feedback buffers bound by vkCmdBindTransformFeedbackBuffers2EXT", 168}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-pCounterInfos-13093", 56}, { {"Each element of pCounterInfos must have a addressRange.size that is either 0 or greater than or equal to 4", 106}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-pCounterInfos-13094", 56}, { {"For each element of pCounterInfos with a non-zero address, the buffer the address was queried from must have been created with the VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT usage flag set", 203}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-pCounterInfos-parameter", 60}, { {"If counterRangeCount is not 0, and pCounterInfos is not NULL, pCounterInfos must be a valid pointer to an array of counterRangeCount valid VkBindTransformFeedbackBuffer2InfoEXT structures", 187}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-renderpass", 47}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-transformFeedback-02366", 60}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedback2EXT-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-None-02367", 46}, { {"Transform feedback must not be active", 37}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-None-02373", 46}, { {"Transform feedback must not be made active in a render pass instance with multiview enabled", 91}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-None-04128", 46}, { {"The last pre-rasterization shader stage of the bound graphics pipeline must have been declared with the Xfb execution mode", 122}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-None-06233", 46}, { {"If the shaderObject feature is not enabled, a valid graphics pipeline must be bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 118}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-None-10656", 46}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-commonparent", 48}, { {"Both of commandBuffer, and the elements of pCounterBuffers that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 178}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-counterBufferCount-02607", 60}, { {"If counterBufferCount is not 0, and pCounterBuffers is not NULL, pCounterBuffers must be a valid pointer to an array of counterBufferCount VkBuffer handles that are either valid or VK_NULL_HANDLE", 195}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-firstCounter-02368", 54}, { {"firstCounterBuffer must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-firstCounter-02369", 54}, { {"The sum of firstCounterBuffer and counterBufferCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 158}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-firstCounter-09630", 54}, { {"The sum of firstCounterBuffer and counterBufferCount must be less than or equal to the number of transform feedback buffers bound by vkCmdBindTransformFeedbackBuffersEXT", 169}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-pCounterBuffer-02371", 56}, { {"If pCounterBuffer is NULL, then pCounterBufferOffsets must also be NULL", 71}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-pCounterBufferOffsets-02370", 63}, { {"For each buffer handle in the array, if it is not VK_NULL_HANDLE it must reference a buffer large enough to hold 4 bytes at the corresponding offset from the pCounterBufferOffsets array", 185}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-pCounterBufferOffsets-parameter", 67}, { {"If counterBufferCount is not 0, and pCounterBufferOffsets is not NULL, pCounterBufferOffsets must be a valid pointer to an array of counterBufferCount VkDeviceSize values", 170}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-pCounterBuffers-02372", 57}, { {"For each buffer handle in the pCounterBuffers array that is not VK_NULL_HANDLE it must have been with the VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT usage flag set", 178}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-renderpass", 46}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-transformFeedback-02366", 59}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginTransformFeedbackEXT-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-None-07232", 40}, { {"There must be no active queries", 31}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-bufferlevel", 41}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-07233", 49}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, then pBeginInfo->videoSession must not have been created with VK_VIDEO_SESSION_CREATE_PROTECTED_CONTENT_BIT_KHR", 200}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-07234", 49}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, then pBeginInfo->videoSession must have been created with VK_VIDEO_SESSION_CREATE_PROTECTED_CONTENT_BIT_KHR", 193}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-07235", 49}, { {"If commandBuffer is an unprotected command buffer, protectedNoFault is not supported, and the pPictureResource member of any element of pBeginInfo->pReferenceSlots is not NULL, then pPictureResource->imageViewBinding for that element must not specify an image view created from a protected image", 295}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-07236", 49}, { {"If commandBuffer is a protected command buffer protectedNoFault is not supported, and the pPictureResource member of any element of pBeginInfo->pReferenceSlots is not NULL, then pPictureResource->imageViewBinding for that element must specify an image view created from a protected image", 287}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-11760", 49}, { {"The VkCommandPool that commandBuffer was allocated from must have been created with the same queue family index that pBeginInfo->videoSession was created with", 158}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-08253", 46}, { {"If pBeginInfo->videoSession was created with a video encode operation and the pNext chain of pBeginInfo does not include an instance of the VkVideoEncodeRateControlInfoKHR structure, then the rate control mode configured for pBeginInfo->videoSession at the time the command is executed on the device must be VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR", 353}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-08254", 46}, { {"If pBeginInfo->videoSession was created with a video encode operation and the pNext chain of pBeginInfo includes an instance of the VkVideoEncodeRateControlInfoKHR structure, then it must match the rate control state configured for pBeginInfo->videoSession at the time the command is executed on the device", 306}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-08255", 46}, { {"If pBeginInfo->videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, and VkVideoEncodeH264CapabilitiesKHR::requiresGopRemainingFrames is VK_TRUE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the pBeginInfo->videoSession was created with, then the pNext chain of pBeginInfo must include an instance of the VkVideoEncodeH264GopRemainingFrameInfoKHR with its useGopRemainingFrames member set to VK_TRUE", 617}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-08256", 46}, { {"If pBeginInfo->videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, and VkVideoEncodeH265CapabilitiesKHR::requiresGopRemainingFrames is VK_TRUE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the pBeginInfo->videoSession was created with, then the pNext chain of pBeginInfo must include an instance of the VkVideoEncodeH265GopRemainingFrameInfoKHR with its useGopRemainingFrames member set to VK_TRUE", 617}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-10282", 46}, { {"If pBeginInfo->videoSession was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, and VkVideoEncodeAV1CapabilitiesKHR::requiresGopRemainingFrames is VK_TRUE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the pBeginInfo->videoSession was created with, then the pNext chain of pBeginInfo must include an instance of the VkVideoEncodeAV1GopRemainingFrameInfoKHR with its useGopRemainingFrames member set to VK_TRUE", 614}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pBeginInfo-parameter", 50}, { {"pBeginInfo must be a valid pointer to a valid VkVideoBeginCodingInfoKHR structure", 81}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-pPictureResource-07265", 52}, { {"Each video picture resource specified by any non-NULL pPictureResource member specified in the elements of pBeginInfo->pReferenceSlots for which slotIndex is not negative must match one of the video picture resources currently associated with the DPB slot index of pBeginInfo->videoSession specified by slotIndex at the time the command is executed on the device", 362}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-renderpass", 40}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-slotIndex-07239", 45}, { {"If the slotIndex member of any element of pBeginInfo->pReferenceSlots is not negative, then it must specify the index of a DPB slot that is in the active state in pBeginInfo->videoSession at the time the command is executed on the device", 237}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-suspended", 39}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBeginVideoCodingKHR-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-commandBuffer-11295", 70}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-commandBuffer-11296", 70}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-commandBuffer-cmdpool", 72}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-commandBuffer-parameter", 74}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-commandBuffer-recording", 74}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-descriptorBuffer-09472", 73}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-pBindDescriptorBufferEmbeddedSamplersInfo-09473", 98}, { {"Each bit in pBindDescriptorBufferEmbeddedSamplersInfo->stageFlags must be a stage supported by the commandBuffer's parent VkCommandPool's queue family", 150}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-pBindDescriptorBufferEmbeddedSamplersInfo-parameter", 102}, { {"pBindDescriptorBufferEmbeddedSamplersInfo must be a valid pointer to a valid VkBindDescriptorBufferEmbeddedSamplersInfoEXT structure", 132}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplers2EXT-videocoding", 62}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-None-08068", 60}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commandBuffer-11295", 69}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commandBuffer-11296", 69}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commandBuffer-cmdpool", 71}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commandBuffer-parameter", 73}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commandBuffer-recording", 73}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-commonparent", 62}, { {"Both of commandBuffer, and layout must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-layout-parameter", 66}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-pipelineBindPoint-08069", 73}, { {"pipelineBindPoint must be supported by the commandBuffer's parent VkCommandPool's queue family", 94}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-pipelineBindPoint-parameter", 77}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-set-08070", 59}, { {"The VkDescriptorSetLayout at index set when layout was created must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT bit set", 170}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-set-08071", 59}, { {"set must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 109}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBufferEmbeddedSamplersEXT-videocoding", 61}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-None-08047", 45}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-None-08050", 45}, { {"There must be no more than 1 element in pBindingInfos with VkDescriptorBufferBindingInfoEXT::usage containing VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT", 168}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-bufferCount-08051", 52}, { {"bufferCount must be less than or equal to VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxDescriptorBufferBindings", 116}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-bufferCount-arraylength", 58}, { {"bufferCount must be greater than 0", 34}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-commandBuffer-11295", 54}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-commandBuffer-11296", 54}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-maxResourceDescriptorBufferBindings-08049", 76}, { {"There must be no more than VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxResourceDescriptorBufferBindings elements in pBindingInfos with VkDescriptorBufferBindingInfoEXT::usage containing VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT", 242}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-maxSamplerDescriptorBufferBindings-08048", 75}, { {"There must be no more than VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxSamplerDescriptorBufferBindings elements in pBindingInfos with VkDescriptorBufferBindingInfoEXT::usage containing VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT", 240}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-pBindingInfos-08053", 54}, { {"For each element of pBindingInfos, the buffer from which address was queried must have been created with the VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT usage flag set if it contains sampler descriptor data", 212}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-pBindingInfos-08054", 54}, { {"For each element of pBindingInfos, the buffer from which address was queried must have been created with the VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT usage flag set if it contains resource descriptor data", 214}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-pBindingInfos-08055", 54}, { {"For each element of pBindingInfos, at least one buffer from which address was queried must contain usage", 104}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-pBindingInfos-09947", 54}, { {"For all elements of pBindingInfos, the buffer from which address was queried must have been created with the VK_BUFFER_USAGE_2_DATA_GRAPH_FOREIGN_DESCRIPTOR_BIT_ARM usage flag set if the command pool from which commandBuffer was allocated from was created with any element of VkDataGraphProcessingEngineCreateInfoARM::pProcessingEngines with isForeign set to VK_TRUE", 366}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-pBindingInfos-parameter", 58}, { {"pBindingInfos must be a valid pointer to an array of bufferCount valid VkDescriptorBufferBindingInfoEXT structures", 114}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorBuffersEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdBindDescriptorSets-commandBuffer-11295", 48}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-commandBuffer-11296", 48}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-commonparent", 41}, { {"Each of commandBuffer, layout, and the elements of pDescriptorSets that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 186}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-descriptorSetCount-arraylength", 59}, { {"descriptorSetCount must be greater than 0", 41}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-dynamicOffsetCount-00359", 53}, { {"dynamicOffsetCount must be equal to the total number of dynamic descriptors in pDescriptorSets", 94}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-firstSet-00360", 43}, { {"The sum of firstSet and descriptorSetCount must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 148}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-layout-parameter", 45}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-00358", 50}, { {"Each element of pDescriptorSets that is not VK_NULL_HANDLE must have been allocated with a VkDescriptorSetLayout that matches (is the same as, or identically defined as) the VkDescriptorSetLayout at set n in layout, where n is the sum of firstSet and the index into pDescriptorSets", 281}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-01979", 50}, { {"For each dynamic uniform or storage buffer binding in pDescriptorSets, the sum of the effective offset and the range of the binding must be less than or equal to the size of the buffer", 184}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-04616", 50}, { {"Each element of pDescriptorSets must not have been allocated from a VkDescriptorPool with the VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT flag set", 146}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-06563", 50}, { {"If the graphicsPipelineLibrary feature is not enabled, each element of pDescriptorSets must be a valid VkDescriptorSet", 118}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-06715", 50}, { {"For each dynamic uniform or storage buffer binding in pDescriptorSets, if the range was set with VK_WHOLE_SIZE then pDynamicOffsets which corresponds to the descriptor binding must be 0", 185}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-08010", 50}, { {"Each element of pDescriptorSets must have been allocated with a VkDescriptorSetLayout which was not created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 170}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-09914", 50}, { {"If any element of pDescriptorSets was allocated from a descriptor pool created with a VkDescriptorPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure specifying foreign data processing engines in its pNext chain, then the command pool from which commandBuffer was allocated must have been created with a VkCommandPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain specifying a superset of all the foreign data processing engines specified when creating the descriptor pools from which the elements of pDescriptorSets were allocated", 621}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-09915", 50}, { {"If none of the elements of pDescriptorSets were allocated from a descriptor pool created with a VkDescriptorPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure specifying foreign data processing engines in its pNext chain, then the command pool from which commandBuffer was allocated must not have been created with a VkCommandPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain", 467}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDescriptorSets-parameter", 54}, { {"pDescriptorSets must be a valid pointer to an array of descriptorSetCount valid or VK_NULL_HANDLE VkDescriptorSet handles", 121}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDynamicOffsets-01971", 50}, { {"Each element of pDynamicOffsets which corresponds to a descriptor binding with type VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC must be a multiple of VkPhysicalDeviceLimits::minUniformBufferOffsetAlignment", 203}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDynamicOffsets-01972", 50}, { {"Each element of pDynamicOffsets which corresponds to a descriptor binding with type VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC must be a multiple of VkPhysicalDeviceLimits::minStorageBufferOffsetAlignment", 203}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pDynamicOffsets-parameter", 54}, { {"If dynamicOffsetCount is not 0, pDynamicOffsets must be a valid pointer to an array of dynamicOffsetCount uint32_t values", 121}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pipelineBindPoint-00361", 52}, { {"pipelineBindPoint must be supported by the commandBuffer's parent VkCommandPool's queue family", 94}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-pipelineBindPoint-parameter", 56}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-commandBuffer-11295", 49}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-commandBuffer-11296", 49}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-pBindDescriptorSetsInfo-09467", 59}, { {"Each bit in pBindDescriptorSetsInfo->stageFlags must be a stage supported by the commandBuffer's parent VkCommandPool's queue family", 132}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-pBindDescriptorSetsInfo-parameter", 63}, { {"pBindDescriptorSetsInfo must be a valid pointer to a valid VkBindDescriptorSetsInfo structure", 93}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindDescriptorSets2-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdBindIndexBuffer-None-09493", 36}, { {"If the maintenance6 feature is not enabled, buffer must not be VK_NULL_HANDLE", 77}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-buffer-08784", 38}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDEX_BUFFER_BIT usage flag set", 86}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-buffer-08785", 38}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-buffer-09494", 38}, { {"If buffer is VK_NULL_HANDLE, offset must be zero", 48}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-buffer-parameter", 42}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-commonparent", 38}, { {"Both of buffer, and commandBuffer that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 153}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-indexType-08786", 41}, { {"indexType must not be VK_INDEX_TYPE_NONE_KHR", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-indexType-08787", 41}, { {"If indexType is VK_INDEX_TYPE_UINT8, the indexTypeUint8 feature must be enabled", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-indexType-parameter", 45}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-offset-08782", 38}, { {"offset must be less than the size of buffer", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-offset-08783", 38}, { {"The sum of offset and the base address of the range of VkDeviceMemory object that is backing buffer, must be a multiple of the size of the type indicated by indexType", 166}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-None-09493", 37}, { {"If the maintenance6 feature is not enabled, buffer must not be VK_NULL_HANDLE", 77}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-buffer-08784", 39}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDEX_BUFFER_BIT usage flag set", 86}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-buffer-08785", 39}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-buffer-09494", 39}, { {"If buffer is VK_NULL_HANDLE, offset must be zero", 48}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-buffer-parameter", 43}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-commonparent", 39}, { {"Both of buffer, and commandBuffer that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 153}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-indexType-08786", 42}, { {"indexType must not be VK_INDEX_TYPE_NONE_KHR", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-indexType-08787", 42}, { {"If indexType is VK_INDEX_TYPE_UINT8, the indexTypeUint8 feature must be enabled", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-indexType-parameter", 46}, { {"indexType must be a valid VkIndexType value", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-offset-08782", 39}, { {"offset must be less than the size of buffer", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-offset-08783", 39}, { {"The sum of offset and the base address of the range of VkDeviceMemory object that is backing buffer, must be a multiple of the size of the type indicated by indexType", 166}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-size-08767", 37}, { {"If size is not VK_WHOLE_SIZE, size must be a multiple of the size of the type indicated by indexType", 100}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-size-08768", 37}, { {"If size is not VK_WHOLE_SIZE, the sum of offset and size must be less than or equal to the size of buffer", 105}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer2-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer3KHR-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer3KHR-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer3KHR-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer3KHR-pInfo-parameter", 45}, { {"pInfo must be a valid pointer to a valid VkBindIndexBuffer3InfoKHR structure", 76}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindIndexBuffer3KHR-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-None-04976", 45}, { {"The invocationMask feature must be enabled", 42}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-None-04982", 45}, { {"Each element in the invocation mask image must have the value 0 or 1. The value 1 means the invocation is active", 112}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-commonparent", 47}, { {"Both of commandBuffer, and imageView that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 156}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-depth-04983", 46}, { {"depth in vkCmdTraceRaysKHR must be 1", 36}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageLayout-parameter", 56}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageView-04977", 50}, { {"If imageView is not VK_NULL_HANDLE, it must be a valid VkImageView handle of type VK_IMAGE_VIEW_TYPE_2D", 103}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageView-04978", 50}, { {"If imageView is not VK_NULL_HANDLE, it must have a format of VK_FORMAT_R8_UINT", 78}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageView-04979", 50}, { {"If imageView is not VK_NULL_HANDLE, it must have been created with the VK_IMAGE_USAGE_INVOCATION_MASK_BIT_HUAWEI usage flag set", 127}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageView-04980", 50}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must be VK_IMAGE_LAYOUT_GENERAL", 79}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-imageView-parameter", 54}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-renderpass", 45}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindInvocationMaskHUAWEI-width-04981", 46}, { {"Thread mask image resolution must match the width and height in vkCmdTraceRaysKHR", 81}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdBindPipeline-None-02323", 33}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commandBuffer-04808", 42}, { {"If commandBuffer is a secondary command buffer with VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled and pipelineBindPoint is VK_PIPELINE_BIND_POINT_GRAPHICS, then the pipeline must have been created with VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT or VK_DYNAMIC_STATE_VIEWPORT, and VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT or VK_DYNAMIC_STATE_SCISSOR enabled", 373}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commandBuffer-04809", 42}, { {"If commandBuffer is a secondary command buffer with VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled and pipelineBindPoint is VK_PIPELINE_BIND_POINT_GRAPHICS and pipeline was created with VkPipelineDiscardRectangleStateCreateInfoEXT structure and its discardRectangleCount member is not 0, or the pipeline was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT enabled, then the pipeline must have been created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT enabled", 496}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-commonparent", 35}, { {"Both of commandBuffer, and pipeline must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipeline-00781", 37}, { {"If the variableMultisampleRate feature is not supported, pipeline is a graphics pipeline, the current subpass uses no attachments, and this is not the first call to this function with a graphics pipeline after transitioning to the current subpass, then the sample count specified by this pipeline must match that set in the previous pipeline", 341}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipeline-03382", 37}, { {"pipeline must not have been created with VK_PIPELINE_CREATE_LIBRARY_BIT_KHR set", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipeline-09912", 37}, { {"If pipeline is a data graph pipeline and the VkDataGraphPipelineCreateInfoARM structure used to create it had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain that specified any foreign data processing engines, then the command pool from which commandBuffer was allocated must have been created with a VkCommandPoolCreateInfo structure that had a VkDataGraphProcessingEngineCreateInfoARM structure specifying a superset of the foreign data graph processing engines specified at pipeline creation time in its pNext chain", 542}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipeline-09913", 37}, { {"If pipeline is a data graph pipeline and the VkDataGraphPipelineCreateInfoARM structure used to create it did not have a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain, then the command pool from which commandBuffer was allocated must not have been created with a VkCommandPoolCreateInfo that had a VkDataGraphProcessingEngineCreateInfoARM structure in its pNext chain", 391}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipeline-parameter", 41}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-00777", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_COMPUTE, the VkCommandPool that commandBuffer was allocated from must support compute operations", 143}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-00778", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_GRAPHICS, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 145}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-00779", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_COMPUTE, pipeline must be a compute pipeline", 91}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-00780", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline must be a graphics pipeline", 93}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-02391", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_RAY_TRACING_KHR, the VkCommandPool that commandBuffer was allocated from must support compute operations", 151}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-02392", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_RAY_TRACING_KHR, pipeline must be a ray tracing pipeline", 103}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-04881", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_GRAPHICS and the provokingVertexModePerPipeline limit is VK_FALSE, then pipeline's VkPipelineRasterizationProvokingVertexStateCreateInfoEXT::provokingVertexMode must be the same as that of any other pipelines previously bound to this bind point within the current render pass instance, including any pipeline already bound when beginning the render pass instance", 409}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-04949", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI, the VkCommandPool that commandBuffer was allocated from must support compute operations", 158}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-04950", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI, pipeline must be a subpass shading pipeline", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-06721", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_RAY_TRACING_KHR, commandBuffer must not be a protected command buffer", 116}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-09910", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_DATA_GRAPH_ARM, the VkCommandPool that commandBuffer was allocated from must have been created for a queue family that supports VK_QUEUE_DATA_GRAPH_BIT_ARM", 202}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-09911", 46}, { {"If pipelineBindPoint is VK_PIPELINE_BIND_POINT_DATA_GRAPH_ARM, pipeline must be a data graph pipeline", 101}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineBindPoint-parameter", 50}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineProtectedAccess-07408", 52}, { {"If the pipelineProtectedAccess feature is enabled, and commandBuffer is a protected command buffer, pipeline must have been created without VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT", 182}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-pipelineProtectedAccess-07409", 52}, { {"If the pipelineProtectedAccess feature is enabled, and commandBuffer is not a protected command buffer, pipeline must have been created without VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT", 188}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-variableSampleLocations-01525", 52}, { {"If VkPhysicalDeviceSampleLocationsPropertiesEXT::variableSampleLocations is VK_FALSE, and pipeline is a graphics pipeline created with a renderPass that is not VK_NULL_HANDLE and with a VkPipelineSampleLocationsStateCreateInfoEXT structure having its sampleLocationsEnable member set to VK_TRUE but without VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT enabled then the current render pass instance must have been begun by specifying a VkRenderPassSampleLocationsBeginInfoEXT structure whose pPostSubpassSampleLocations member contains an element with a subpassIndex matching the current subpass index and the sampleLocationsInfo member of that element must match the sampleLocationsInfo specified in VkPipelineSampleLocationsStateCreateInfoEXT when the pipeline was created", 766}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipeline-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-commonparent", 48}, { {"Both of commandBuffer, and pipeline must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-deviceGeneratedCommands-02896", 65}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-groupIndex-02893", 52}, { {"groupIndex must be 0 or less than the effective VkGraphicsPipelineShaderGroupsCreateInfoNV::groupCount including the referenced pipelines", 137}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-groupIndex-02895", 52}, { {"The same restrictions as vkCmdBindPipeline apply as if the bound pipeline was created only with the Shader Group from the groupIndex information", 144}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-pipeline-parameter", 54}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-pipelineBindPoint-02894", 59}, { {"The pipelineBindPoint must be VK_PIPELINE_BIND_POINT_GRAPHICS", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-pipelineBindPoint-parameter", 63}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindPipelineShaderGroupNV-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdBindResourceHeapEXT-commandBuffer-11238", 49}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-heapRange-11237", 45}, { {"heapRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT usage flag set", 155}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11232", 45}, { {"The sum of pBindInfo->reservedRangeOffset and pBindInfo->reservedRangeSize must be less than or equal to pBindInfo->heapRange.size", 130}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11233", 45}, { {"pBindInfo->reservedRangeSize must be greater than or equal to minResourceHeapReservedRange", 90}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11234", 45}, { {"pBindInfo->heapRange.size must less than or equal to maxResourceHeapSize", 72}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11235", 45}, { {"pBindInfo->heapRange.address must be a multiple of resourceHeapAlignment", 72}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11236", 45}, { {"Memory bound to addresses in the range [pBindInfo->heapRange.address + pBindInfo->reservedRangeOffset, pBindInfo->heapRange.address + pBindInfo->reservedRangeOffset + pBindInfo->reservedRangeSize) must not be bound to any other command buffer as a reserved range for any heap unless the reserved range matches exactly and it is the same heap type", 346}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11435", 45}, { {"pBindInfo->reservedRangeOffset must be a multiple of bufferDescriptorAlignment", 78}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-11436", 45}, { {"pBindInfo->reservedRangeOffset must be a multiple of imageDescriptorAlignment", 77}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-pBindInfo-parameter", 49}, { {"pBindInfo must be a valid pointer to a valid VkBindHeapInfoEXT structure", 72}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindResourceHeapEXT-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-commandBuffer-11231", 48}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-heapRange-11230", 44}, { {"heapRange must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_DESCRIPTOR_HEAP_BIT_EXT usage flag set", 155}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11223", 44}, { {"The sum of pBindInfo->reservedRangeOffset and pBindInfo->reservedRangeSize must be less than or equal to pBindInfo->heapRange.size", 130}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11224", 44}, { {"pBindInfo->reservedRangeSize must be greater than or equal to minSamplerHeapReservedRange", 89}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11225", 44}, { {"pBindInfo->heapRange.size must less than or equal to maxSamplerHeapSize", 71}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11226", 44}, { {"pBindInfo->heapRange.address must be a multiple of samplerHeapAlignment", 71}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11228", 44}, { {"Memory bound to addresses in the range [pBindInfo->heapRange.address + pBindInfo->reservedRangeOffset, pBindInfo->heapRange.address + pBindInfo->reservedRangeOffset + pBindInfo->reservedRangeSize) must not be bound to any other command buffer as a reserved range for any heap unless the reserved range matches exactly and it is the same heap type", 346}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-11434", 44}, { {"pBindInfo->reservedRangeOffset must be a multiple of samplerDescriptorAlignment", 79}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-pBindInfo-parameter", 48}, { {"pBindInfo must be a valid pointer to a valid VkBindHeapInfoEXT structure", 72}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindSamplerHeapEXT-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdBindShadersEXT-None-08462", 35}, { {"The shaderObject feature must be enabled", 40}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-commonparent", 37}, { {"Both of commandBuffer, and the elements of pShaders that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 171}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08469", 39}, { {"For each element of pStages, if pShaders is not NULL, and the element of the pShaders array with the same index is not VK_NULL_HANDLE, it must have been created with a stage equal to the corresponding element of pStages", 219}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08470", 39}, { {"If pStages contains both VK_SHADER_STAGE_TASK_BIT_EXT and VK_SHADER_STAGE_VERTEX_BIT, and pShaders is not NULL, and the same index in pShaders as VK_SHADER_STAGE_TASK_BIT_EXT in pStages is not VK_NULL_HANDLE, the same index in pShaders as VK_SHADER_STAGE_VERTEX_BIT in pStages must be VK_NULL_HANDLE", 299}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08471", 39}, { {"If pStages contains both VK_SHADER_STAGE_MESH_BIT_EXT and VK_SHADER_STAGE_VERTEX_BIT, and pShaders is not NULL, and the same index in pShaders as VK_SHADER_STAGE_MESH_BIT_EXT in pStages is not VK_NULL_HANDLE, the same index in pShaders as VK_SHADER_STAGE_VERTEX_BIT in pStages must be VK_NULL_HANDLE", 299}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08476", 39}, { {"If pStages contains VK_SHADER_STAGE_COMPUTE_BIT, the VkCommandPool that commandBuffer was allocated from must support compute operations", 136}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08477", 39}, { {"If pStages contains VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, or VK_SHADER_STAGE_FRAGMENT_BIT, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 286}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-08478", 39}, { {"If pStages contains VK_SHADER_STAGE_MESH_BIT_EXT or VK_SHADER_STAGE_TASK_BIT_EXT, the VkCommandPool that commandBuffer was allocated from must support graphics operations", 170}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pShaders-parameter", 43}, { {"If pShaders is not NULL, pShaders must be a valid pointer to an array of stageCount valid or VK_NULL_HANDLE VkShaderEXT handles", 127}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-08463", 38}, { {"Every element of pStages must be unique", 39}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-08464", 38}, { {"pStages must not contain VK_SHADER_STAGE_ALL_GRAPHICS or VK_SHADER_STAGE_ALL", 76}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-08465", 38}, { {"pStages must not contain VK_SHADER_STAGE_RAYGEN_BIT_KHR, VK_SHADER_STAGE_ANY_HIT_BIT_KHR, VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR, VK_SHADER_STAGE_MISS_BIT_KHR, VK_SHADER_STAGE_INTERSECTION_BIT_KHR, or VK_SHADER_STAGE_CALLABLE_BIT_KHR", 230}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-08467", 38}, { {"pStages must not contain VK_SHADER_STAGE_SUBPASS_SHADING_BIT_HUAWEI", 67}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-08468", 38}, { {"pStages must not contain VK_SHADER_STAGE_CLUSTER_CULLING_BIT_HUAWEI", 67}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-pStages-parameter", 42}, { {"pStages must be a valid pointer to an array of stageCount valid VkShaderStageFlagBits values", 92}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-stageCount-arraylength", 47}, { {"stageCount must be greater than 0", 33}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadersEXT-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdBindShadingRateImageNV-None-02058", 43}, { {"The shadingRateImage feature must be enabled", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-commonparent", 45}, { {"Both of commandBuffer, and imageView that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 156}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageLayout-02063", 50}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must be VK_IMAGE_LAYOUT_SHADING_RATE_OPTIMAL_NV or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageLayout-parameter", 54}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageView-02059", 48}, { {"If imageView is not VK_NULL_HANDLE, it must be a valid VkImageView handle of type VK_IMAGE_VIEW_TYPE_2D or VK_IMAGE_VIEW_TYPE_2D_ARRAY", 134}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageView-02060", 48}, { {"If imageView is not VK_NULL_HANDLE, it must have a format of VK_FORMAT_R8_UINT", 78}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageView-02061", 48}, { {"If imageView is not VK_NULL_HANDLE, it must have been created with the VK_IMAGE_USAGE_SHADING_RATE_IMAGE_BIT_NV usage flag set", 126}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageView-02062", 48}, { {"If imageView is not VK_NULL_HANDLE, imageLayout must match the actual VkImageLayout of each subresource accessible from imageView at the time the subresource is accessed", 169}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-imageView-parameter", 52}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindShadingRateImageNV-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-pTileMemoryBindInfo-parameter", 58}, { {"If pTileMemoryBindInfo is not NULL, pTileMemoryBindInfo must be a valid pointer to a valid VkTileMemoryBindInfoQCOM structure", 125}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-renderpass", 39}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTileMemoryQCOM-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-None-02365", 53}, { {"Transform feedback must not be active when the vkCmdBindTransformFeedbackBuffers2EXT command is recorded", 104}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-addressRange-13090", 61}, { {"The addressRange.address member of all elements of pBindingInfos must be a multiple of 4", 88}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-addressRange-13091", 61}, { {"The buffer that the addressRange of each element of pBindingInfos was queried from must have been created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT", 160}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-addressRange-13092", 61}, { {"The addressRange.size member of all elements of pBindingInfos must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBufferSize", 170}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-bindingCount-arraylength", 67}, { {"If pBindingInfos is not NULL, bindingCount must be greater than 0", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-commandBuffer-cmdpool", 64}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-commandBuffer-parameter", 66}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-commandBuffer-recording", 66}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-firstBinding-02356", 61}, { {"firstBinding must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 106}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-firstBinding-02357", 61}, { {"The sum of firstBinding and bindingCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 146}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-pBindingInfos-parameter", 66}, { {"If pBindingInfos is not NULL, pBindingInfos must be a valid pointer to an array of bindingCount valid VkBindTransformFeedbackBuffer2InfoEXT structures", 150}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-transformFeedback-02355", 66}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffers2EXT-videocoding", 54}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-None-02365", 52}, { {"Transform feedback must not be active when the vkCmdBindTransformFeedbackBuffersEXT command is recorded", 103}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-bindingCount-arraylength", 66}, { {"bindingCount must be greater than 0", 35}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-commonparent", 54}, { {"Both of commandBuffer, and the elements of pBuffers must have been created, allocated, or retrieved from the same VkDevice", 122}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-firstBinding-02356", 60}, { {"firstBinding must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 106}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-firstBinding-02357", 60}, { {"The sum of firstBinding and bindingCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 146}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pBuffers-02360", 56}, { {"All elements of pBuffers must have been created with the VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT usage flag set", 121}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pBuffers-02364", 56}, { {"Each element of pBuffers that is non-sparse must be bound completely and contiguously to a single VkDeviceMemory object", 119}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pBuffers-parameter", 60}, { {"pBuffers must be a valid pointer to an array of bindingCount valid VkBuffer handles", 83}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pOffsets-02358", 56}, { {"All elements of pOffsets must be less than the size of the corresponding element in pBuffers", 92}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pOffsets-02359", 56}, { {"All elements of pOffsets must be a multiple of 4", 48}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pOffsets-02363", 56}, { {"All elements of pOffsets plus the effective size of the element, must be less than or equal to the size of the corresponding buffer in pBuffers", 143}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-pOffsets-parameter", 60}, { {"pOffsets must be a valid pointer to an array of bindingCount VkDeviceSize values", 80}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-transformFeedback-02355", 65}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindTransformFeedbackBuffersEXT-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdBindVertexBuffers-bindingCount-arraylength", 52}, { {"bindingCount must be greater than 0", 35}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-commonparent", 40}, { {"Both of commandBuffer, and the elements of pBuffers that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 171}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-firstBinding-00624", 46}, { {"firstBinding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 77}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-firstBinding-00625", 46}, { {"The sum of firstBinding and bindingCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindings", 117}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pBuffers-00627", 42}, { {"All elements of pBuffers must have been created with the VK_BUFFER_USAGE_VERTEX_BUFFER_BIT flag", 95}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pBuffers-00628", 42}, { {"Each element of pBuffers that is non-sparse must be bound completely and contiguously to a single VkDeviceMemory object", 119}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pBuffers-04001", 42}, { {"If the nullDescriptor feature is not enabled, all elements of pBuffers must not be VK_NULL_HANDLE", 97}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pBuffers-04002", 42}, { {"If an element of pBuffers is VK_NULL_HANDLE, then the corresponding element of pOffsets must be zero", 100}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pBuffers-parameter", 46}, { {"pBuffers must be a valid pointer to an array of bindingCount valid or VK_NULL_HANDLE VkBuffer handles", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pOffsets-00626", 42}, { {"All elements of pOffsets must be less than the size of the corresponding element in pBuffers", 92}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-pOffsets-parameter", 46}, { {"pOffsets must be a valid pointer to an array of bindingCount VkDeviceSize values", 80}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-None-08971", 39}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-bindingCount-arraylength", 53}, { {"If any of pSizes, or pStrides are not NULL, bindingCount must be greater than 0", 79}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-commonparent", 41}, { {"Both of commandBuffer, and the elements of pBuffers that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 171}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-firstBinding-03355", 47}, { {"firstBinding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 77}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-firstBinding-03356", 47}, { {"The sum of firstBinding and bindingCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindings", 117}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pBuffers-03359", 43}, { {"All elements of pBuffers must have been created with the VK_BUFFER_USAGE_VERTEX_BUFFER_BIT usage flag set", 105}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pBuffers-03360", 43}, { {"Each element of pBuffers that is non-sparse must be bound completely and contiguously to a single VkDeviceMemory object", 119}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pBuffers-04111", 43}, { {"If the nullDescriptor feature is not enabled, all elements of pBuffers must not be VK_NULL_HANDLE", 97}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pBuffers-04112", 43}, { {"If an element of pBuffers is VK_NULL_HANDLE, then the corresponding element of pOffsets must be zero", 100}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pBuffers-parameter", 47}, { {"pBuffers must be a valid pointer to an array of bindingCount valid or VK_NULL_HANDLE VkBuffer handles", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pOffsets-03357", 43}, { {"If pSizes is not NULL, all elements of pOffsets must be less than the size of the corresponding element in pBuffers", 115}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pOffsets-parameter", 47}, { {"pOffsets must be a valid pointer to an array of bindingCount VkDeviceSize values", 80}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pSizes-03358", 41}, { {"If pSizes is not NULL, all elements of pOffsets plus pSizes , where pSizes is not VK_WHOLE_SIZE, must be less than or equal to the size of the corresponding element in pBuffers", 176}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pSizes-parameter", 45}, { {"If pSizes is not NULL, pSizes must be a valid pointer to an array of bindingCount VkDeviceSize values", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pStrides-03362", 43}, { {"If pStrides is not NULL each element of pStrides must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindingStride", 130}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pStrides-06209", 43}, { {"If pStrides is not NULL each element of pStrides must be either 0 or greater than or equal to the maximum extent of all vertex input attributes fetched from the corresponding binding, where the extent is calculated as the VkVertexInputAttributeDescription::offset plus VkVertexInputAttributeDescription::format size", 315}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-pStrides-parameter", 47}, { {"If pStrides is not NULL, pStrides must be a valid pointer to an array of bindingCount VkDeviceSize values", 105}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers2-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-addressRange-13073", 50}, { {"The addressRange.stride of all elements of pInfos must be either 0 or greater than or equal to the maximum extent of all vertex input attributes fetched from the corresponding binding, where the extent is calculated as the VkVertexInputAttributeDescription::offset plus VkVertexInputAttributeDescription::format size", 316}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-bindingCount-arraylength", 56}, { {"bindingCount must be greater than 0", 35}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-firstBinding-13070", 50}, { {"firstBinding must be less than VkPhysicalDeviceLimits::maxVertexInputBindings", 77}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-firstBinding-13071", 50}, { {"The sum of firstBinding and bindingCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindings", 117}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-pBindingInfos-parameter", 55}, { {"pBindingInfos must be a valid pointer to an array of bindingCount valid VkBindVertexBuffer3InfoKHR structures", 109}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBindVertexBuffers3KHR-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdBlitImage-aspectMask-00241", 36}, { {"For each element of pRegions, srcSubresource.aspectMask must specify aspects present in srcImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-aspectMask-00242", 36}, { {"For each element of pRegions, dstSubresource.aspectMask must specify aspects present in dstImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-01834", 39}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-01835", 39}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-01836", 39}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-cmdpool", 41}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-parameter", 43}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commandBuffer-recording", 43}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-commonparent", 32}, { {"Each of commandBuffer, dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-00224", 34}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-00225", 34}, { {"If dstImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-00234", 34}, { {"dstImage must have been created with a samples value of VK_SAMPLE_COUNT_1_BIT", 77}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-00250", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffsets[0].y must be 0 and dstOffsets[1].y must be 1", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-00252", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffsets[0].z must be 0 and dstOffsets[1].z must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-02000", 34}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_BLIT_DST_BIT", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-02545", 34}, { {"dstImage and srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-06422", 34}, { {"dstImage must not use a format that requires a sampler Y'CBCR conversion", 72}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImage-parameter", 38}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImageLayout-00226", 40}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImageLayout-01399", 40}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstImageLayout-parameter", 44}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstOffset-00248", 35}, { {"For each element of pRegions, dstOffsets[0].x and dstOffsets[1].x must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstOffset-00249", 35}, { {"For each element of pRegions, dstOffsets[0].y and dstOffsets[1].y must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 186}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstOffset-00251", 35}, { {"For each element of pRegions, dstOffsets[0].z and dstOffsets[1].z must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstSubresource-01706", 40}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-dstSubresource-01708", 40}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-filter-00237", 32}, { {"If filter is VK_FILTER_CUBIC_EXT, srcImage must be of type VK_IMAGE_TYPE_2D", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-filter-02001", 32}, { {"If filter is VK_FILTER_LINEAR, then the format features of srcImage must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 130}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-filter-02002", 32}, { {"If filter is VK_FILTER_CUBIC_EXT, then the format features of srcImage must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 136}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-filter-parameter", 36}, { {"filter must be a valid VkFilter value", 37}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-maintenance8-10207", 38}, { {"If the maintenance8 feature is enabled and srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer must be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 231}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-maintenance8-10208", 38}, { {"If the maintenance8 feature is enabled and dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 231}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-maintenance8-10579", 38}, { {"If the maintenance8 feature is enabled, dstImage is VK_IMAGE_TYPE_3D, and srcImage is not of type VK_IMAGE_TYPE_3D, then for each element of pRegions, the absolute difference of the z member of each member of dstOffsets must equal srcSubresource.layerCount", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-maintenance8-10580", 38}, { {"If the maintenance8 feature is enabled, srcImage is VK_IMAGE_TYPE_3D, and dstImage is not of type VK_IMAGE_TYPE_3D, then for each element of pRegions, the absolute difference of the z member of each member of srcOffsets must equal dstSubresource.layerCount", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-pRegions-00217", 34}, { {"The union of all destination regions, specified by the elements of pRegions, must not overlap in memory with any texel that may be sampled during the blit operation", 164}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-pRegions-parameter", 38}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageBlit structures", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-regionCount-arraylength", 43}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-renderpass", 30}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00219", 34}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00220", 34}, { {"If srcImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00229", 34}, { {"If either of srcImage or dstImage was created with a signed integer VkFormat, the other must also have been created with a signed integer VkFormat", 146}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00230", 34}, { {"If either of srcImage or dstImage was created with an unsigned integer color VkFormat, the other must also have been created with an unsigned integer color VkFormat", 164}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00231", 34}, { {"If either of srcImage or dstImage was created with a depth/stencil format, the other must have exactly the same format", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00232", 34}, { {"If srcImage was created with a depth/stencil format, filter must be VK_FILTER_NEAREST", 85}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00233", 34}, { {"srcImage must have been created with a samples value of VK_SAMPLE_COUNT_1_BIT", 77}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00240", 34}, { {"If the maintenance8 feature is not enabled and either srcImage or dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer and dstSubresource.baseArrayLayer must each be 0, and srcSubresource.layerCount and dstSubresource.layerCount must each be 1", 293}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00245", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffsets[0].y must be 0 and srcOffsets[1].y must be 1", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-00247", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffsets[0].z must be 0 and srcOffsets[1].z must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-01999", 34}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_BLIT_SRC_BIT", 75}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-06421", 34}, { {"srcImage must not use a format that requires a sampler Y'CBCR conversion", 72}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-09459", 34}, { {"If srcImage and dstImage are the same, and an element of pRegions contains the srcSubresource and dstSubresource with matching mipLevel and overlapping array layers, then the srcImageLayout and dstImageLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR", 278}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImage-parameter", 38}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImageLayout-00221", 40}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImageLayout-01398", 40}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcImageLayout-parameter", 44}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcOffset-00243", 35}, { {"For each element of pRegions, srcOffsets[0].x and srcOffsets[1].x must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcOffset-00244", 35}, { {"For each element of pRegions, srcOffsets[0].y and srcOffsets[1].y must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 186}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcOffset-00246", 35}, { {"For each element of pRegions, srcOffsets[0].z and srcOffsets[1].z must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcSubresource-01705", 40}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-srcSubresource-01707", 40}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-suspended", 29}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage-videocoding", 31}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-01834", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-01835", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-01836", 40}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-pBlitImageInfo-parameter", 45}, { {"pBlitImageInfo must be a valid pointer to a valid VkBlitImageInfo2 structure", 76}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-renderpass", 31}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBlitImage2-videocoding", 32}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-accelerationStructureReference-12264", 75}, { {"Each VkAccelerationStructureInstanceKHR::accelerationStructureReference value in instanceData must be a value obtained from vkGetAccelerationStructureHandleNV for a bottom level acceleration structure", 200}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-commonparent", 51}, { {"Each of commandBuffer, dst, instanceData, scratch, and src that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 178}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-dst-02488", 48}, { {"dst must have been created with compatible VkAccelerationStructureInfoNV where VkAccelerationStructureInfoNV::type and VkAccelerationStructureInfoNV::flags are identical, VkAccelerationStructureInfoNV::instanceCount and VkAccelerationStructureInfoNV::geometryCount for dst are greater than or equal to the build size and each geometry in VkAccelerationStructureInfoNV::pGeometries for dst has greater than or equal to the number of vertices, indices, and AABBs", 460}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-dst-07787", 48}, { {"dst must be bound completely and contiguously to a single VkDeviceMemory object via vkBindAccelerationStructureMemoryNV", 119}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-dst-parameter", 52}, { {"dst must be a valid VkAccelerationStructureNV handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-geometryCount-02241", 58}, { {"geometryCount must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxGeometryCount", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-instanceData-03523", 57}, { {"If instanceData is not VK_NULL_HANDLE, instanceData must have been created with the VK_BUFFER_USAGE_RAY_TRACING_BIT_NV usage flag set", 133}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-instanceData-parameter", 61}, { {"If instanceData is not VK_NULL_HANDLE, instanceData must be a valid VkBuffer handle", 83}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-pInfo-parameter", 54}, { {"pInfo must be a valid pointer to a valid VkAccelerationStructureInfoNV structure", 80}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-renderpass", 49}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-scratch-03522", 52}, { {"scratch must have been created with the VK_BUFFER_USAGE_RAY_TRACING_BIT_NV usage flag set", 89}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-scratch-parameter", 56}, { {"scratch must be a valid VkBuffer handle", 39}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-src-parameter", 52}, { {"If src is not VK_NULL_HANDLE, src must be a valid VkAccelerationStructureNV handle", 82}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-02489", 51}, { {"If update is VK_TRUE, src must not be VK_NULL_HANDLE", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-02490", 51}, { {"If update is VK_TRUE, src must have previously been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_NV set in VkAccelerationStructureInfoNV::flags in the original build", 186}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-02491", 51}, { {"If update is VK_FALSE, the size member of the VkMemoryRequirements structure returned from a call to vkGetAccelerationStructureMemoryRequirementsNV with VkAccelerationStructureMemoryRequirementsInfoNV::accelerationStructure set to dst and VkAccelerationStructureMemoryRequirementsInfoNV::type set to VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_BUILD_SCRATCH_NV must be less than or equal to the size of scratch minus scratchOffset", 437}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-02492", 51}, { {"If update is VK_TRUE, the size member of the VkMemoryRequirements structure returned from a call to vkGetAccelerationStructureMemoryRequirementsNV with VkAccelerationStructureMemoryRequirementsInfoNV::accelerationStructure set to dst and VkAccelerationStructureMemoryRequirementsInfoNV::type set to VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_UPDATE_SCRATCH_NV must be less than or equal to the size of scratch minus scratchOffset", 437}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-03524", 51}, { {"If update is VK_TRUE, then objects that were previously active must not be made inactive as per Inactive Primitives and Instances", 129}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-03525", 51}, { {"If update is VK_TRUE, then objects that were previously inactive must not be made active as per Inactive Primitives and Instances", 129}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-update-03526", 51}, { {"If update is VK_TRUE, the src and dst objects must either be the same object or not have any memory aliasing", 108}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructureNV-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-None-03407", 59}, { {"The dstAccelerationStructure member of any element of pInfos must not be referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-accelerationStructureIndirectBuild-03650", 89}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureIndirectBuild feature must be enabled", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-commandBuffer-09547", 68}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-commandBuffer-cmdpool", 70}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-commandBuffer-parameter", 72}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-commandBuffer-recording", 72}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03698", 79}, { {"The dstAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03701", 79}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any other element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, which is accessed by this command", 328}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03702", 79}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the dstAccelerationStructure member of any other element of pInfos, which is accessed by this command", 259}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03703", 79}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any element of pInfos (including the same element), which is accessed by this command", 269}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03706", 79}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing any acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos, which is accessed by this command", 399}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-dstAccelerationStructure-03800", 79}, { {"The dstAccelerationStructure member of any element of pInfos must be a valid VkAccelerationStructureKHR handle", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-geometry-03673", 63}, { {"The buffers from which the buffer device addresses for all of the geometry.triangles.vertexData, geometry.triangles.indexData, geometry.triangles.transformData, geometry.aabbs.data, and geometry.instances.data members of all pInfos[i].pGeometries and pInfos[i].ppGeometries are queried must have been created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag set", 401}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-infoCount-arraylength", 70}, { {"infoCount must be greater than 0", 32}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11624", 63}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in the geometry.triangles.pNext chain with a valid micromap member, then the micromap member must have been built with a type of VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 380}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11625", 63}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with a valid micromap member", 647}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11626", 63}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with its micromap member equal to VK_NULL_HANDLE, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with its micromap member equal to VK_NULL_HANDLE", 687}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11627", 63}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was not built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must not be included in the in the geometry.triangles.pNext chain", 597}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11628", 63}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure was originally built without the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR flag, and srcAccelerationStructure was constructed by deserialization with vkCmdCopyMemoryToAccelerationStructureKHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with a micromap that was deserialized from the serialized data of the micromap in the original build of srcAccelerationStructure", 997}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11629", 63}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure was originally built without either the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_DATA_UPDATE_BIT_EXT or the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR flags, and srcAccelerationStructure was not constructed by deserialization with vkCmdCopyMemoryToAccelerationStructureKHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with the same micromap handle as the original build", 1014}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11632", 63}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, then its micromap member must have been constructed prior to this command's execution on the device", 338}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-micromap-11701", 63}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in the geometry.triangles.pNext chain with a valid micromap member, then the micromap member must have been constructed prior to this command's execution on the device", 367}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-mode-04628", 59}, { {"The mode member of each element of pInfos must be a valid VkBuildAccelerationStructureModeKHR value", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectDeviceAddresses-03646", 79}, { {"For each element of pIndirectDeviceAddresses[i], all device addresses between pIndirectDeviceAddresses[i] and pIndirectDeviceAddresses[i] + (pInfos[i].geometryCount x pIndirectStrides[i]) - 1 must be in the buffer device address range of the same buffer", 253}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectDeviceAddresses-03647", 79}, { {"For each element of pIndirectDeviceAddresses, the buffer from which it was queried must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 165}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectDeviceAddresses-03648", 79}, { {"Each element of pIndirectDeviceAddresses must be a multiple of 4", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectDeviceAddresses-03651", 79}, { {"Each VkAccelerationStructureBuildRangeInfoKHR structure referenced by any element of pIndirectDeviceAddresses must be a valid VkAccelerationStructureBuildRangeInfoKHR structure", 176}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectDeviceAddresses-parameter", 83}, { {"pIndirectDeviceAddresses must be a valid pointer to an array of infoCount VkDeviceAddress values", 96}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectStrides-03787", 71}, { {"Each element of pIndirectStrides must be a multiple of 4", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pIndirectStrides-parameter", 75}, { {"pIndirectStrides must be a valid pointer to an array of infoCount uint32_t values", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03403", 61}, { {"The srcAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03652", 61}, { {"pInfos[i].dstAccelerationStructure must have been created with a value of VkAccelerationStructureCreateInfoKHR::size greater than or equal to the memory size required by the build operation, as returned by vkGetAccelerationStructureBuildSizesKHR with pBuildInfo = pInfos[i] and pMaxPrimitiveCounts = ppMaxPrimitiveCounts[i]", 323}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03663", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, inactive primitives in its srcAccelerationStructure member must not be made active", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03664", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, active primitives in its srcAccelerationStructure member must not be made inactive", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03667", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must have previously been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_KHR set in VkAccelerationStructureBuildGeometryInfoKHR::flags in the build", 301}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03668", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure and dstAccelerationStructure members must either be the same VkAccelerationStructureKHR, or not have any memory aliasing", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03699", 61}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03700", 61}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 326}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03707", 61}, { {"For each element of pInfos, its dstAccelerationStructure member must be fully backed by physical memory", 103}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03708", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, srcAccelerationStructure must be fully backed by physical memory", 163}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03709", 61}, { {"For each element of pInfos, if an acceleration structure is referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, it must be fully backed by physical memory", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03710", 61}, { {"For each element of pInfos, its scratchData.deviceAddress member must be a multiple of VkPhysicalDeviceAccelerationStructurePropertiesKHR::minAccelerationStructureScratchOffsetAlignment", 185}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03711", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, geometry.triangles.vertexData.deviceAddress must be aligned to: the size of the format specified in vertexFormat, in bytes, if that format is a packed format the component size of the format specified in vertexFormat, in bytes, if that format is not a packed format", 388}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03712", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, and with geometry.triangles.indexType not equal to VK_INDEX_TYPE_NONE_KHR, geometry.triangles.indexData.deviceAddress must be aligned to the size in bytes of the type in indexType", 302}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03714", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR, geometry.aabbs.data.deviceAddress must be aligned to 8 bytes", 179}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03715", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_FALSE, geometry.instances.data.deviceAddress must be aligned to 16 bytes", 229}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03716", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_TRUE, geometry.instances.data.deviceAddress must be aligned to 8 bytes", 227}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03717", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_TRUE, each element of geometry.instances.data.deviceAddress in device memory must be aligned to 16 bytes", 261}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03758", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its geometryCount member must have the same value which was specified when srcAccelerationStructure was last built", 213}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03759", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 205}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03760", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its type member must have the same value which was specified when srcAccelerationStructure was last built", 204}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03761", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its geometryType member must have the same value which was specified when srcAccelerationStructure was last built", 327}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03762", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03763", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.vertexFormat member must have the same value which was specified when srcAccelerationStructure was last built", 397}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03764", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.maxVertex member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03765", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.indexType member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03766", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was NULL when srcAccelerationStructure was last built, then it must be NULL", 388}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03767", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was not NULL when srcAccelerationStructure was last built, then it must not be NULL", 396}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03768", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, then the value of each index referenced must be the same as the corresponding index value when srcAccelerationStructure was last built", 463}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03801", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, the corresponding ppMaxPrimitiveCounts[i][j] must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxInstanceCount", 266}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03804", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, geometry.triangles.vertexData.deviceAddress must be a valid VkDeviceAddress", 198}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03806", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, geometry.triangles.indexData.deviceAddress must be a valid VkDeviceAddress", 260}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03808", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.transformData.deviceAddress is not 0, it must be a valid VkDeviceAddress", 217}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03810", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.transformData.deviceAddress is not 0, it must be aligned to 16 bytes", 213}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03811", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR, geometry.aabbs.data.deviceAddress must be a valid VkDeviceAddress", 184}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-03813", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, geometry.instances.data.deviceAddress must be a valid VkDeviceAddress", 192}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-04630", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must not be VK_NULL_HANDLE", 161}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10607", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if VK_GEOMETRY_INSTANCE_DISABLE_OPACITY_MICROMAPS_BIT_KHR is set in VkAccelerationStructureInstanceKHR::flags then geometry.instances.data.deviceAddress must refer to an acceleration structure that was built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISABLE_OPACITY_MICROMAPS_BIT_KHR set in VkAccelerationStructureBuildGeometryInfoKHR::flags", 465}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10898", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numTriangles member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 497}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10899", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numVertices member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 496}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10900", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxPrimitiveIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 502}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10901", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxGeometryIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 501}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10902", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the format member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 491}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10903", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the dataSize member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 493}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10904", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.deviceAddress must be 0", 329}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-10905", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.deviceAddress must be a valid VkDeviceAddress", 355}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11846", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, if VkAccelerationStructureGeometrySpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be 0", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11847", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, if VkAccelerationStructureGeometrySpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be a valid VkDeviceAddress", 274}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11848", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::vertexData.deviceAddress must be a valid VkDeviceAddress", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11849", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::radiusData.deviceAddress must be a valid VkDeviceAddress", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11850", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, if VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be 0", 272}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11851", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, if VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be a valid VkDeviceAddress", 298}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11852", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::vertexData.deviceAddress must be a valid VkDeviceAddress", 246}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-11853", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::radiusData.deviceAddress must be a valid VkDeviceAddress", 246}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-12258", 61}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR, and N is not 0, all addresses between pInfos[i].scratchData.deviceAddress and pInfos[i].scratchData.deviceAddress + N - 1 must be in the buffer device address range of the same buffer, where N is given by the buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 515}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-12259", 61}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, and N is not 0, all addresses between pInfos[i].scratchData.deviceAddress and pInfos[i].scratchData.deviceAddress + N - 1 must be in the buffer device address range of the same buffer, where N is given by the updateScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 517}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-12260", 61}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, and the size reported by updateScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count is non-zero, pInfos[i].scratchData.deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 517}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-12261", 61}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR, and the size reported by buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count is non-zero, pInfos[i].scratchData.deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 515}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-12281", 61}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, each VkAccelerationStructureInstanceKHR::accelerationStructureReference value in geometry.instances.data.deviceAddress must be 0 or a value obtained from vkGetAccelerationStructureDeviceAddressKHR for a valid bottom level acceleration structure", 367}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-pInfos-parameter", 65}, { {"pInfos must be a valid pointer to an array of infoCount valid VkAccelerationStructureBuildGeometryInfoKHR structures", 116}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-ppMaxPrimitiveCounts-03653", 75}, { {"Each ppMaxPrimitiveCounts[i][j] must be greater than or equal to the primitiveCount value specified by the VkAccelerationStructureBuildRangeInfoKHR structure located at pIndirectDeviceAddresses[i] + (j x pIndirectStrides[i])", 224}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-ppMaxPrimitiveCounts-parameter", 79}, { {"ppMaxPrimitiveCounts must be a valid pointer to an array of infoCount uint32_t values", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-primitiveCount-03769", 69}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, the primitiveCount member of its corresponding VkAccelerationStructureBuildRangeInfoKHR structure must have the same value which was specified when srcAccelerationStructure was last built", 401}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-renderpass", 59}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-scratchData-03704", 66}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any other element of pInfos, which is accessed by this command", 233}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-scratchData-03705", 66}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR (including the same element), which is accessed by this command", 338}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-srcAccelerationStructure-04629", 79}, { {"If the srcAccelerationStructure member of any element of pInfos is not VK_NULL_HANDLE, the srcAccelerationStructure member must be a valid VkAccelerationStructureKHR handle", 172}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-suspended", 58}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-type-11557", 59}, { {"For any element of pInfos, its type member must not be VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 106}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresIndirectKHR-videocoding", 60}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-None-03407", 51}, { {"The dstAccelerationStructure member of any element of pInfos must not be referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-accelerationStructure-08923", 68}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-commandBuffer-09547", 60}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03698", 71}, { {"The dstAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03701", 71}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any other element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, which is accessed by this command", 328}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03702", 71}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the dstAccelerationStructure member of any other element of pInfos, which is accessed by this command", 259}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03703", 71}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any element of pInfos (including the same element), which is accessed by this command", 269}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03706", 71}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing any acceleration structure referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR in any other element of pInfos, which is accessed by this command", 399}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-03800", 71}, { {"The dstAccelerationStructure member of any element of pInfos must be a valid VkAccelerationStructureKHR handle", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-11547", 71}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 334}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-11548", 71}, { {"The dstAccelerationStructure member of any element of pInfos must not be referenced by any VkAccelerationStructureTrianglesOpacityMicromapKHR::micromap member included in the pNext chain of any geometry.triangles member of any element of pGeometries or ppGeometries in any other element of pInfos", 296}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-dstAccelerationStructure-11549", 71}, { {"The range of memory backing the dstAccelerationStructure member of any element of pInfos that is accessed by this command must not overlap the memory backing any VkAccelerationStructureTrianglesOpacityMicromapKHR::micromap member included in the pNext chain of any geometry.triangles member of any element of pGeometries or ppGeometries in any other element of pInfos, which is accessed by this command", 402}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-geometry-03673", 55}, { {"The buffers from which the buffer device addresses for all of the geometry.triangles.vertexData, geometry.triangles.indexData, geometry.triangles.transformData, geometry.aabbs.data, and geometry.instances.data members of all pInfos[i].pGeometries and pInfos[i].ppGeometries are queried must have been created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag set", 401}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-indexBuffer-11575", 58}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType equal to VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles member includes a VkAccelerationStructureTrianglesOpacityMicromapKHR structure in its pNext chain, the micromap's indexType member is not VK_INDEX_TYPE_NONE_KHR, and micromap is not VK_NULL_HANDLE, for each index in indexBuffer[j] greater than zero where j is within the range [0, numTrianglesInGeometry - 1] where numTrianglesInGeometry is equal to ppBuildRangeInfos[i]->primitiveCount, index plus baseTriangle must be less than or equal to numTriangles, where numTriangles is given by the sum of all count parameters in the pUsageCounts or ppUsageCounts provided to the micromap's build command", 751}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-indexBuffer-11577", 58}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType equal to VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles member includes a VkAccelerationStructureTrianglesOpacityMicromapKHR structure in its pNext chain, and the micromap's indexType member is not VK_INDEX_TYPE_NONE_KHR, indexBuffer (indexStride * numTrianglesInGeometry) must be within the range of the VkBuffer from which the indexBuffer device address was obtained from where numTrianglesInGeometry is equal to ppBuildRangeInfos[i]->primitiveCount", 550}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-indexBuffer-11578", 58}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType equal to VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles member includes a VkAccelerationStructureTrianglesOpacityMicromapKHR structure in its pNext chain, the micromap's indexType member is not VK_INDEX_TYPE_NONE_KHR, and micromap is VK_NULL_HANDLE, for each index in indexBuffer[j] where j is within the range from [0, numTrianglesInGeometry - 1], index must be a valid VkOpacityMicromapSpecialIndexKHR value where numTrianglesInGeometry is equal to ppBuildRangeInfos[i]->primitiveCount", 586}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-infoCount-arraylength", 62}, { {"infoCount must be greater than 0", 32}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11550", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, its data and triangleArray members must be a valid VkDeviceAddress", 229}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11551", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, the buffers from which the buffer device addresses for its data and triangleArray members are queried must have been created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag", 376}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11552", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, its data and triangleArray members must be a multiple of 128", 223}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11553", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, each element of its triangleArray member must specify an access within the range of the VkBuffer that data was retrieved from, given by [dataOffset, dataOffset triangleSize] where triangleSize is calculated using its format and subdivisionLevel members rounded up to the nearest byte", 446}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11554", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, the range of [triangleArray, triangleArray + (triangleArrayStride * numTriangles)] must be within the range of the VkBuffer from which the triangleArray member was retrieved from, where numTriangles is the sum of count of all elements of pUsageCounts or ppUsageCounts", 430}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11555", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, each element of its geometry.triangleArray member must have a format with a valid value from VkOpacityMicromapFormatKHR", 282}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11556", 55}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries, if it includes a VkAccelerationStructureGeometryMicromapDataKHR structure in its pNext chain, for each element of its triangleArray member with a format equal to VK_OPACITY_MICROMAP_FORMAT_2_STATE_KHR, subdivisionLevel must be less than or equal to maxOpacity2StateSubdivisionLevel", 350}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11624", 55}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in the geometry.triangles.pNext chain with a valid micromap member, then the micromap member must have been built with a type of VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 380}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11625", 55}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with a valid micromap member", 647}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11626", 55}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with its micromap member equal to VK_NULL_HANDLE, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with its micromap member equal to VK_NULL_HANDLE", 687}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11627", 55}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was not built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must not be included in the in the geometry.triangles.pNext chain", 597}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11628", 55}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure was originally built without the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR flag, and srcAccelerationStructure was constructed by deserialization with vkCmdCopyMemoryToAccelerationStructureKHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with a micromap that was deserialized from the serialized data of the micromap in the original build of srcAccelerationStructure", 997}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11629", 55}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure was originally built without either the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_DATA_UPDATE_BIT_EXT or the VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_BIT_KHR flags, and srcAccelerationStructure was not constructed by deserialization with vkCmdCopyMemoryToAccelerationStructureKHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and the same geometry in the original build of srcAccelerationStructure was built with an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in its geometry.triangles.pNext chain with a valid micromap member, then an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR must be included in the in the geometry.triangles.pNext chain with the same micromap handle as the original build", 1014}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11632", 55}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, then its micromap member must have been constructed prior to this command's execution on the device", 338}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-micromap-11701", 55}, { {"For any element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapKHR in the geometry.triangles.pNext chain with a valid micromap member, then the micromap member must have been constructed prior to this command's execution on the device", 367}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-mode-04628", 51}, { {"The mode member of each element of pInfos must be a valid VkBuildAccelerationStructureModeKHR value", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-mode-11545", 51}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, its mode must equal VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR", 169}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03403", 53}, { {"The srcAccelerationStructure member of any element of pInfos must not be the same acceleration structure as the dstAccelerationStructure member of any other element of pInfos", 174}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03663", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, inactive primitives in its srcAccelerationStructure member must not be made active", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03664", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, active primitives in its srcAccelerationStructure member must not be made inactive", 181}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03667", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must have previously been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_KHR set in VkAccelerationStructureBuildGeometryInfoKHR::flags in the build", 301}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03668", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure and dstAccelerationStructure members must either be the same VkAccelerationStructureKHR, or not have any memory aliasing", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03699", 53}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03700", 53}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR, its dstAccelerationStructure member must have been created with a value of VkAccelerationStructureCreateInfoKHR::type equal to either VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR or VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR", 326}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03707", 53}, { {"For each element of pInfos, its dstAccelerationStructure member must be fully backed by physical memory", 103}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03708", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, srcAccelerationStructure must be fully backed by physical memory", 163}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03709", 53}, { {"For each element of pInfos, if an acceleration structure is referenced by the geometry.instances.data member of any element of pGeometries or ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, it must be fully backed by physical memory", 252}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03710", 53}, { {"For each element of pInfos, its scratchData.deviceAddress member must be a multiple of VkPhysicalDeviceAccelerationStructurePropertiesKHR::minAccelerationStructureScratchOffsetAlignment", 185}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03711", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, geometry.triangles.vertexData.deviceAddress must be aligned to: the size of the format specified in vertexFormat, in bytes, if that format is a packed format the component size of the format specified in vertexFormat, in bytes, if that format is not a packed format", 388}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03712", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, and with geometry.triangles.indexType not equal to VK_INDEX_TYPE_NONE_KHR, geometry.triangles.indexData.deviceAddress must be aligned to the size in bytes of the type in indexType", 302}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03714", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR, geometry.aabbs.data.deviceAddress must be aligned to 8 bytes", 179}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03715", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_FALSE, geometry.instances.data.deviceAddress must be aligned to 16 bytes", 229}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03716", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_TRUE, geometry.instances.data.deviceAddress must be aligned to 8 bytes", 227}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03717", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if geometry.arrayOfPointers is VK_TRUE, each element of geometry.instances.data.deviceAddress in device memory must be aligned to 16 bytes", 261}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03758", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its geometryCount member must have the same value which was specified when srcAccelerationStructure was last built", 213}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03759", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 205}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03760", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its type member must have the same value which was specified when srcAccelerationStructure was last built", 204}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03761", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its geometryType member must have the same value which was specified when srcAccelerationStructure was last built", 327}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03762", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, its flags member must have the same value which was specified when srcAccelerationStructure was last built", 320}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03763", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.vertexFormat member must have the same value which was specified when srcAccelerationStructure was last built", 397}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03764", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.maxVertex member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03765", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, its geometry.triangles.indexType member must have the same value which was specified when srcAccelerationStructure was last built", 394}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03766", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was NULL when srcAccelerationStructure was last built, then it must be NULL", 388}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03767", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, if its geometry.triangles.transformData address was not NULL when srcAccelerationStructure was last built, then it must not be NULL", 396}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03768", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_TRIANGLES_KHR, and geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, then the value of each index referenced must be the same as the corresponding index value when srcAccelerationStructure was last built", 463}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03801", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, the corresponding ppBuildRangeInfos[i][j].primitiveCount must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxInstanceCount", 278}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03804", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, geometry.triangles.vertexData.deviceAddress must be a valid VkDeviceAddress", 198}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03806", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.indexType is not VK_INDEX_TYPE_NONE_KHR, geometry.triangles.indexData.deviceAddress must be a valid VkDeviceAddress", 260}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03808", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.transformData.deviceAddress is not 0, it must be a valid VkDeviceAddress", 217}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03810", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if geometry.triangles.transformData.deviceAddress is not 0, it must be aligned to 16 bytes", 213}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03811", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR, geometry.aabbs.data.deviceAddress must be a valid VkDeviceAddress", 184}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-03813", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, geometry.instances.data.deviceAddress must be a valid VkDeviceAddress", 192}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-04630", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, its srcAccelerationStructure member must not be VK_NULL_HANDLE", 161}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10126", 53}, { {"For each pInfos[i], dstAccelerationStructure must have been created with a value of VkAccelerationStructureCreateInfoKHR::size greater than or equal to either: the memory size required by the build operation, as returned by vkGetAccelerationStructureBuildSizesKHR with pBuildInfo = pInfos[i] and with each element of the pMaxPrimitiveCounts array greater than or equal to the equivalent ppBuildRangeInfos[i][j].primitiveCount values for j in [0,pInfos[i].geometryCount) or, the result of querying the corresponding VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, if updating a compacted acceleration structure", 618}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10607", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, if VK_GEOMETRY_INSTANCE_DISABLE_OPACITY_MICROMAPS_BIT_KHR is set in VkAccelerationStructureInstanceKHR::flags then geometry.instances.data.deviceAddress must refer to an acceleration structure that was built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISABLE_OPACITY_MICROMAPS_BIT_KHR set in VkAccelerationStructureBuildGeometryInfoKHR::flags", 465}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10898", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numTriangles member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 497}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10899", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the numVertices member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 496}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10900", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxPrimitiveIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 502}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10901", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the maxGeometryIndex member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 501}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10902", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the format member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 491}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10903", 53}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, if geometryType is VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, the dataSize member of the VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX structure in the pNext chain must have the same value which was specified when srcAccelerationStructure was last built", 493}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10904", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.deviceAddress must be 0", 329}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10905", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureTrianglesOpacityMicromapEXT in the geometry.triangles.pNext chain, and its indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexBuffer.deviceAddress must be a valid VkDeviceAddress", 355}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-10906", 53}, { {"For each element of pInfos[i] whose pGeometries or ppGeometries members have a geometryType of VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX, ppBuildRangeInfos[i] must be NULL", 183}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11845", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR, if there is an instance of VkAccelerationStructureGeometryMotionTrianglesDataNV in the geometry.triangles.pNext chain, then its vertexData.deviceAddress must be a valid VkDeviceAddress", 307}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11846", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, if VkAccelerationStructureGeometrySpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be 0", 248}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11847", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, if VkAccelerationStructureGeometrySpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be a valid VkDeviceAddress", 274}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11848", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::vertexData.deviceAddress must be a valid VkDeviceAddress", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11849", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_SPHERES_NV, VkAccelerationStructureGeometrySpheresDataNV::radiusData.deviceAddress must be a valid VkDeviceAddress", 222}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11850", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, if VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be 0", 272}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11851", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, if VkAccelerationStructureGeometryLinearSweptSpheresDataNV::indexType is not VK_INDEX_TYPE_NONE_KHR, then its indexData.deviceAddress must be a valid VkDeviceAddress", 298}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11852", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::vertexData.deviceAddress must be a valid VkDeviceAddress", 246}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-11853", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_LINEAR_SWEPT_SPHERES_NV, VkAccelerationStructureGeometryLinearSweptSpheresDataNV::radiusData.deviceAddress must be a valid VkDeviceAddress", 246}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-12258", 53}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR, and N is not 0, all addresses between pInfos[i].scratchData.deviceAddress and pInfos[i].scratchData.deviceAddress + N - 1 must be in the buffer device address range of the same buffer, where N is given by the buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 515}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-12259", 53}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, and N is not 0, all addresses between pInfos[i].scratchData.deviceAddress and pInfos[i].scratchData.deviceAddress + N - 1 must be in the buffer device address range of the same buffer, where N is given by the updateScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count", 517}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-12260", 53}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, and the size reported by updateScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count is non-zero, pInfos[i].scratchData.deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 517}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-12261", 53}, { {"If pInfos[i].mode is VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR, and the size reported by buildScratchSize member of the VkAccelerationStructureBuildSizesInfoKHR structure returned from a call to vkGetAccelerationStructureBuildSizesKHR with an identical VkAccelerationStructureBuildGeometryInfoKHR structure and primitive count is non-zero, pInfos[i].scratchData.deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 515}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-12281", 53}, { {"For each element of pInfos[i].pGeometries or pInfos[i].ppGeometries with a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, each VkAccelerationStructureInstanceKHR::accelerationStructureReference value in geometry.instances.data.deviceAddress must be 0 or a value obtained from vkGetAccelerationStructureDeviceAddressKHR for a valid bottom level acceleration structure", 367}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-pInfos-parameter", 57}, { {"pInfos must be a valid pointer to an array of infoCount valid VkAccelerationStructureBuildGeometryInfoKHR structures", 116}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-ppBuildRangeInfos-11543", 64}, { {"Each element of ppBuildRangeInfos[i] must be a valid pointer to an array of pInfos[i].geometryCount VkAccelerationStructureBuildRangeInfoKHR structures if the pInfos[i].type is not one of: VK_GEOMETRY_TYPE_DENSE_GEOMETRY_FORMAT_TRIANGLES_AMDX VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR", 294}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-ppBuildRangeInfos-11544", 64}, { {"For each element of pInfos[i] whose pGeometries or ppGeometries members have a geometryType of VK_GEOMETRY_TYPE_MICROMAP_KHR, ppBuildRangeInfos[i] must be NULL", 159}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-ppBuildRangeInfos-parameter", 68}, { {"ppBuildRangeInfos must be a valid pointer to an array of infoCount VkAccelerationStructureBuildRangeInfoKHR structures", 118}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-primitiveCount-03769", 61}, { {"For each element of pInfos, if its mode member is VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR, then for each VkAccelerationStructureGeometryKHR structure referred to by its pGeometries or ppGeometries members, the primitiveCount member of its corresponding VkAccelerationStructureBuildRangeInfoKHR structure must have the same value which was specified when srcAccelerationStructure was last built", 401}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-renderpass", 51}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-scratchData-03704", 58}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the scratchData member of any other element of pInfos, which is accessed by this command", 233}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-scratchData-03705", 58}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by this command must not overlap the memory backing the srcAccelerationStructure member of any element of pInfos with a mode equal to VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR (including the same element), which is accessed by this command", 338}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-srcAccelerationStructure-04629", 71}, { {"If the srcAccelerationStructure member of any element of pInfos is not VK_NULL_HANDLE, the srcAccelerationStructure member must be a valid VkAccelerationStructureKHR handle", 172}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-srcAccelerationStructure-11546", 71}, { {"For each element of pInfos, if its type member is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, its srcAccelerationStructure must be VK_NULL_HANDLE", 154}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-suspended", 50}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildAccelerationStructuresKHR-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-clusterAccelerationStructure-10443", 88}, { {"The VkPhysicalDeviceClusterAccelerationStructureFeaturesNV::clusterAccelerationStructure feature must be enabled", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-commandBuffer-cmdpool", 75}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-commandBuffer-parameter", 77}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-commandBuffer-recording", 77}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-dstAddressesArray-12302", 77}, { {"The memory regions specified in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must not overlap with each other or with pCommandInfos->scratchData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV", 705}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-dstImplicitData-12303", 75}, { {"The memory region specified in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, for multiple acceleration structure builds must not overlap with pCommandInfos->scratchData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV", 724}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10445", 73}, { {"pCommandInfos->input.maxAccelerationStructureCount must be less than or equal to the value used in pInfo->maxAccelerationStructureCount in vkGetClusterAccelerationStructureBuildSizesNV to determine the memory requirements for the build operation", 245}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10448", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_MOVE_OBJECTS_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureMoveObjectsInfoNV structures", 200}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10449", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_CLUSTERS_BOTTOM_LEVEL_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV structures", 228}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10450", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureBuildTriangleClusterInfoNV structures", 219}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10451", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_BUILD_TRIANGLE_CLUSTER_TEMPLATE_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV structures", 236}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10452", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_INSTANTIATE_TRIANGLE_CLUSTER_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureInstantiateClusterInfoNV structures", 223}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10453", 73}, { {"The value in pCommandInfos->srcInfosCount must be less than or equal to pCommandInfos->input.maxAccelerationStructureCount", 122}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10454", 73}, { {"The number of inputs specified in pCommandInfos->srcInfosArray must be greater than or equal to pCommandInfos->srcInfosCount", 124}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-10832", 73}, { {"If pCommandInfos->input.opType is VK_CLUSTER_ACCELERATION_STRUCTURE_OP_TYPE_GET_CLUSTER_TEMPLATE_INDICES_NV, pCommandInfos->srcInfosArray must be an array of VkClusterAccelerationStructureGetTemplateIndicesInfoNV structures", 223}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12304", 73}, { {"pCommandInfos->scratchData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV, must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 428}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12305", 73}, { {"pCommandInfos->srcInfosArray, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::srcInfosArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SRC_INFOS_ARRAY_BIT_NV, must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag set", 475}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12306", 73}, { {"pCommandInfos->srcInfosCount, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::srcInfosCount when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SRC_INFOS_COUNT_BIT_NV, must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag set", 469}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12308", 73}, { {"pCommandInfos->dstImplicitData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstImplicitData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_IMPLICIT_DATA_BIT_NV, must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR usage flag set", 461}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12381", 73}, { {"pCommandInfos->dstAddressesArray, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be a device address range allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 451}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-12382", 73}, { {"All destination addresses contained in pCommandInfos->dstAddressesArray, or in the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::dstAddressesArray when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_DST_ADDRESS_ARRAY_BIT_NV, must be device address ranges allocated to the application from buffers created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR usage flag set", 510}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pCommandInfos-parameter", 77}, { {"pCommandInfos must be a valid pointer to a valid VkClusterAccelerationStructureCommandsInfoNV structure", 103}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-pNext-10444", 65}, { {"The pNext chain of the bound ray tracing pipeline must include a VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV structure", 135}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-renderpass", 64}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-scratchData-12300", 71}, { {"The scratch memory of the cluster acceleration structure specified in VkClusterAccelerationStructureCommandsInfoNV::scratchData, or the memory indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV, must be larger than or equal to the scratch size queried with vkGetClusterAccelerationStructureBuildSizesNV", 501}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-scratchData-12301", 71}, { {"The scratch address of the cluster acceleration structure specified in VkClusterAccelerationStructureCommandsInfoNV::scratchData, or the address indirectly referenced in VkClusterAccelerationStructureCommandsInfoNV::scratchData when VkClusterAccelerationStructureCommandsInfoNV::addressResolutionFlags contains VK_CLUSTER_ACCELERATION_STRUCTURE_ADDRESS_RESOLUTION_INDIRECTED_SCRATCH_DATA_BIT_NV, must be aligned based on the cluster acceleration structure type and its alignment properties as queried with VkPhysicalDeviceClusterAccelerationStructurePropertiesNV", 562}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-suspended", 63}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildClusterAccelerationStructureIndirectNV-videocoding", 65}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildMicromapsEXT-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-data-07510", 38}, { {"The buffers from which the buffer device addresses for all of the data and triangleArray members of all pInfos[i] are queried must have been created with the VK_BUFFER_USAGE_MICROMAP_BUILD_INPUT_READ_ONLY_BIT_EXT usage flag set", 227}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-dstMicromap-07463", 45}, { {"The dstMicromap member of any element of pInfos must be a valid VkMicromapEXT handle", 84}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-dstMicromap-11649", 45}, { {"The range of memory backing the dstMicromap member of any element of pInfos that is accessed by the execution of this command on the device must not overlap the memory backing the dstMicromap member of any other element of pInfos, which is accessed by the execution of this command on the device", 295}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-dstMicromap-11650", 45}, { {"The range of memory backing the dstMicromap member of any element of pInfos that is accessed by the execution of this command on the device must not overlap the memory backing the scratchData member of any element of pInfos (including the same element), which is accessed by the execution of this command on the device", 318}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-infoCount-arraylength", 49}, { {"infoCount must be greater than 0", 32}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-micromap-11648", 42}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-mode-07462", 38}, { {"The mode member of each element of pInfos must be a valid VkBuildMicromapModeEXT value", 86}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07461", 40}, { {"For each pInfos[i], dstMicromap must have been created with a value of VkMicromapCreateInfoEXT::size greater than or equal to the memory size required by the build operation, as returned by vkGetMicromapBuildSizesEXT with pBuildInfo = pInfos[i]", 244}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07464", 40}, { {"For each element of pInfos its type member must match the value of VkMicromapCreateInfoEXT::type when its dstMicromap was created", 129}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07508", 40}, { {"For each element of pInfos, the buffer used to create its dstMicromap member must be bound to device memory", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07509", 40}, { {"If pInfos[i].mode is VK_BUILD_MICROMAP_MODE_BUILD_EXT, all addresses between pInfos[i].scratchData.deviceAddress and pInfos[i].scratchData.deviceAddress + N - 1 must be in the buffer device address range of the same buffer, where N is given by the buildScratchSize member of the VkMicromapBuildSizesInfoEXT structure returned from a call to vkGetMicromapBuildSizesEXT with an identical VkMicromapBuildInfoEXT structure and primitive count", 438}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07511", 40}, { {"For each element of pInfos[i] the buffer from which the buffer device address pInfos[i].scratchData.deviceAddress is queried must have been created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 206}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07512", 40}, { {"For each element of pInfos, if the value of VkMicromapBuildSizesInfoEXT::buildScratchSize, returned from a call to vkGetMicromapBuildSizesEXT with an identical VkMicromapBuildInfoEXT structure, is not 0, scratchData.deviceAddress must be a valid VkDeviceAddress", 261}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07514", 40}, { {"For each element of pInfos, its scratchData.deviceAddress member must be a multiple of VkPhysicalDeviceAccelerationStructurePropertiesKHR::minAccelerationStructureScratchOffsetAlignment", 185}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-07515", 40}, { {"For each element of pInfos, its triangleArray.deviceAddress and data.deviceAddress members must be a multiple of 256", 116}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-10896", 40}, { {"For each element of pInfos, data.deviceAddress must be a valid VkDeviceAddress", 78}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-10897", 40}, { {"For each element of pInfos, triangleArray.deviceAddress must be a valid VkDeviceAddress", 87}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-pInfos-parameter", 44}, { {"pInfos must be a valid pointer to an array of infoCount valid VkMicromapBuildInfoEXT structures", 95}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-renderpass", 38}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-scratchData-11651", 45}, { {"The range of memory backing the scratchData member of any element of pInfos that is accessed by the execution of this command on the device must not overlap the memory backing the scratchData member of any other element of pInfos, which is accessed by the execution of this command on the device", 295}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildMicromapsEXT-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-commandBuffer-cmdpool", 72}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-commandBuffer-parameter", 74}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-commandBuffer-recording", 74}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10537", 67}, { {"The count specified in pBuildInfo->input.instanceCount for the build operation must not exceed the value provided in pInfo->instanceCount when calling vkGetPartitionedAccelerationStructuresBuildSizesNV to determine the memory size", 230}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10538", 67}, { {"The count specified in pBuildInfo->input.maxInstancePerPartitionCount for the build operation must not exceed the value provided in pInfo->maxInstancePerPartitionCount when calling vkGetPartitionedAccelerationStructuresBuildSizesNV to determine the memory size", 260}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10539", 67}, { {"The count specified in pBuildInfo->input.partitionCount for the build operation must not exceed the value provided in pInfo->partitionCount when calling vkGetPartitionedAccelerationStructuresBuildSizesNV to determine the memory size", 232}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10540", 67}, { {"The count specified in pBuildInfo->input.maxInstanceInGlobalPartitionCount for the build operation must not exceed the value provided in pInfo->maxInstanceInGlobalPartitionCount when calling vkGetPartitionedAccelerationStructuresBuildSizesNV to determine the memory size", 270}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10541", 67}, { {"The scratch memory for the partitioned acceleration structure build specified in pBuildInfo->scratchData must be larger than or equal to the scratch size queried with vkGetPartitionedAccelerationStructuresBuildSizesNV", 217}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10542", 67}, { {"pBuildInfo->scratchData must be aligned to 256 bytes", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10543", 67}, { {"The destination memory of the partitioned acceleration structure build specified in pBuildInfo->dstAccelerationStructureData must be larger than or equal to the size queried with vkGetPartitionedAccelerationStructuresBuildSizesNV", 229}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10544", 67}, { {"pBuildInfo->srcAccelerationStructureData must be aligned to 256 bytes", 69}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10545", 67}, { {"pBuildInfo->dstAccelerationStructureData must be aligned to 256 bytes", 69}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10546", 67}, { {"The number of inputs specified in pBuildInfo->srcInfos must be greater than or equal to pBuildInfo->srcInfosCount", 113}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10547", 67}, { {"The memory region containing the acceleration structure at address pBuildInfo->srcAccelerationStructureData must not overlap with scratch memory region at address pBuildInfo->scratchData", 186}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10548", 67}, { {"The memory region containing the acceleration structure at address pBuildInfo->dstAccelerationStructureData must not overlap with scratch memory region at address pBuildInfo->scratchData", 186}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10549", 67}, { {"If the source and destination acceleration structures are not the same, the memory regions containing the acceleration structures at addresses pBuildInfo->srcAccelerationStructureData and pBuildInfo->dstAccelerationStructureData must not overlap with each other", 261}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10550", 67}, { {"pBuildInfo->scratchData must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 158}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10551", 67}, { {"pBuildInfo->srcInfos and pBuildInfo->srcInfosCount must be device addresses allocated to the application from buffers created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR usage flag set", 218}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-10552", 67}, { {"pBuildInfo->srcAccelerationStructureData and pBuildInfo->dstAccelerationStructureData must be a device addresses allocated to the application from buffers created with the VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR usage flag set", 241}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-pBuildInfo-parameter", 71}, { {"pBuildInfo must be a valid pointer to a valid VkBuildPartitionedAccelerationStructureInfoNV structure", 101}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-partitionedAccelerationStructure-10536", 89}, { {"The VkPhysicalDevicePartitionedAccelerationStructureFeaturesNV::partitionedAccelerationStructure feature must be enabled", 120}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-renderpass", 61}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-suspended", 60}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdBuildPartitionedAccelerationStructuresNV-videocoding", 62}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdClearAttachments-None-09679", 37}, { {"If the attachment format has components other than R and G, it must not have a 64-bit component width", 101}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-None-10616", 37}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-aspectMask-07271", 43}, { {"If the aspectMask member of any element of pAttachments contains VK_IMAGE_ASPECT_COLOR_BIT, the colorAttachment must be a valid color attachment index in the current render pass instance", 186}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-aspectMask-07884", 43}, { {"If the current render pass instance does not use dynamic rendering, and the aspectMask member of any element of pAttachments contains VK_IMAGE_ASPECT_DEPTH_BIT, the current subpass instance's depth-stencil attachment must be either VK_ATTACHMENT_UNUSED or the attachment format must contain a depth component", 308}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-aspectMask-07885", 43}, { {"If the current render pass instance does not use dynamic rendering, and the aspectMask member of any element of pAttachments contains VK_IMAGE_ASPECT_STENCIL_BIT, the current subpass instance's depth-stencil attachment must be either VK_ATTACHMENT_UNUSED or the attachment format must contain a stencil component", 312}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-aspectMask-09298", 43}, { {"If the subpass this is recorded in performs an external format resolve, the aspectMask member of any element of pAttachments must not include VK_IMAGE_ASPECT_PLANE_i_BIT for any index i", 185}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-attachmentCount-arraylength", 54}, { {"attachmentCount must be greater than 0", 38}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-baseArrayLayer-00018", 47}, { {"If the render pass instance this is recorded in uses multiview, then baseArrayLayer must be zero and layerCount must be one", 123}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-colorAttachment-09503", 48}, { {"The colorAttachment member of each element of pAttachments must not identify a color attachment that is currently mapped to VK_ATTACHMENT_UNUSED in commandBuffer via VkRenderingAttachmentLocationInfo", 199}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-commandBuffer-02504", 46}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, each attachment to be cleared must not be a protected image", 148}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-commandBuffer-02505", 46}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, each attachment to be cleared must not be an unprotected image", 148}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-layerCount-01934", 43}, { {"The layerCount member of each element of pRects must not be 0", 61}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-pAttachments-parameter", 49}, { {"pAttachments must be a valid pointer to an array of attachmentCount valid VkClearAttachment structures", 102}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-pRects-00016", 39}, { {"The rectangular region specified by each element of pRects must be contained within the render area of the current render pass instance", 135}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-pRects-06937", 39}, { {"The layers specified by each element of pRects must be contained within every attachment that pAttachments refers to, i.e. for each element of pRects, VkClearRect::baseArrayLayer + VkClearRect::layerCount must be less than or equal to the number of layers rendered to in the current render pass instance", 303}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-pRects-parameter", 43}, { {"pRects must be a valid pointer to an array of rectCount VkClearRect structures", 78}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-rect-02682", 37}, { {"The rect member of each element of pRects must have an extent.width greater than 0", 82}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-rect-02683", 37}, { {"The rect member of each element of pRects must have an extent.height greater than 0", 83}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-rectCount-arraylength", 48}, { {"rectCount must be greater than 0", 32}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-renderpass", 37}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearAttachments-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-aspectMask-02498", 42}, { {"The VkImageSubresourceRange::aspectMask members of the elements of the pRanges array must each only include VK_IMAGE_ASPECT_COLOR_BIT", 133}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-baseArrayLayer-01472", 46}, { {"The VkImageSubresourceRange::baseArrayLayer members of the elements of the pRanges array must each be less than the arrayLayers specified in VkImageCreateInfo when image was created", 181}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-baseMipLevel-01470", 44}, { {"The VkImageSubresourceRange::baseMipLevel members of the elements of the pRanges array must each be less than the mipLevels specified in VkImageCreateInfo when image was created", 177}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commandBuffer-01805", 45}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, image must not be a protected image", 124}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commandBuffer-01806", 45}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, must not be an unprotected image", 118}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-commonparent", 38}, { {"Both of commandBuffer, and image must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-00002", 37}, { {"image must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 84}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-00003", 37}, { {"If image is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 106}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-00007", 37}, { {"image must not have a compressed or depth/stencil format", 56}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-01545", 37}, { {"image must not use any of the formats that require a sampler Y'CBCR conversion", 78}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-01993", 37}, { {"The format features of image must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 76}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-09678", 37}, { {"If image's format has components other than R and G, it must not have a 64-bit component width", 94}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-image-parameter", 41}, { {"image must be a valid VkImage handle", 36}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-imageLayout-00004", 43}, { {"imageLayout must specify the layout of the image subresource ranges of image specified in pRanges at the time this command is executed on a VkDevice", 148}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-imageLayout-01394", 43}, { {"imageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 119}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-imageLayout-parameter", 47}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-pColor-04961", 38}, { {"pColor must be a valid pointer to a VkClearColorValue union", 59}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-pRanges-01692", 39}, { {"For each VkImageSubresourceRange element of pRanges, if the levelCount member is not VK_REMAINING_MIP_LEVELS, then baseMipLevel + levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 238}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-pRanges-01693", 39}, { {"For each VkImageSubresourceRange element of pRanges, if the layerCount member is not VK_REMAINING_ARRAY_LAYERS, then baseArrayLayer + layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 244}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-pRanges-parameter", 43}, { {"pRanges must be a valid pointer to an array of rangeCount valid VkImageSubresourceRange structures", 98}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-rangeCount-arraylength", 48}, { {"rangeCount must be greater than 0", 33}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-renderpass", 36}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearColorImage-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-aspectMask-02824", 49}, { {"The VkImageSubresourceRange::aspectMask member of each element of the pRanges array must not include bits other than VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 173}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-baseArrayLayer-01476", 53}, { {"The VkImageSubresourceRange::baseArrayLayer members of the elements of the pRanges array must each be less than the arrayLayers specified in VkImageCreateInfo when image was created", 181}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-baseMipLevel-01474", 51}, { {"The VkImageSubresourceRange::baseMipLevel members of the elements of the pRanges array must each be less than the mipLevels specified in VkImageCreateInfo when image was created", 177}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commandBuffer-01807", 52}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, image must not be a protected image", 124}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commandBuffer-01808", 52}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, image must not be an unprotected image", 124}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-commonparent", 45}, { {"Both of commandBuffer, and image must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-00010", 44}, { {"If image is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 106}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-00014", 44}, { {"image must have a depth/stencil format", 38}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-01994", 44}, { {"The format features of image must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 76}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-02825", 44}, { {"If the image's format does not have a stencil component, then the VkImageSubresourceRange::aspectMask member of each element of the pRanges array must not include the VK_IMAGE_ASPECT_STENCIL_BIT bit", 198}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-02826", 44}, { {"If the image's format does not have a depth component, then the VkImageSubresourceRange::aspectMask member of each element of the pRanges array must not include the VK_IMAGE_ASPECT_DEPTH_BIT bit", 194}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-image-parameter", 48}, { {"image must be a valid VkImage handle", 36}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-imageLayout-00011", 50}, { {"imageLayout must specify the layout of the image subresource ranges of image specified in pRanges at the time this command is executed on a VkDevice", 148}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-imageLayout-00012", 50}, { {"imageLayout must be either of VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 93}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-imageLayout-parameter", 54}, { {"imageLayout must be a valid VkImageLayout value", 47}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pDepthStencil-parameter", 56}, { {"pDepthStencil must be a valid pointer to a valid VkClearDepthStencilValue structure", 83}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-01694", 46}, { {"For each VkImageSubresourceRange element of pRanges, if the levelCount member is not VK_REMAINING_MIP_LEVELS, then baseMipLevel + levelCount must be less than or equal to the mipLevels specified in VkImageCreateInfo when image was created", 238}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-01695", 46}, { {"For each VkImageSubresourceRange element of pRanges, if the layerCount member is not VK_REMAINING_ARRAY_LAYERS, then baseArrayLayer + layerCount must be less than or equal to the arrayLayers specified in VkImageCreateInfo when image was created", 244}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-02658", 46}, { {"If the aspect member of any element of pRanges includes VK_IMAGE_ASPECT_STENCIL_BIT, and image was created with separate stencil usage, it must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 217}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-02659", 46}, { {"If the aspect member of any element of pRanges includes VK_IMAGE_ASPECT_STENCIL_BIT, and image was not created with separate stencil usage, the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag must have been set in the VkImageCreateInfo::usage used to create image", 258}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-02660", 46}, { {"If the aspect member of any element of pRanges includes VK_IMAGE_ASPECT_DEPTH_BIT, image must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 167}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-pRanges-parameter", 50}, { {"pRanges must be a valid pointer to an array of rangeCount valid VkImageSubresourceRange structures", 98}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-rangeCount-arraylength", 55}, { {"rangeCount must be greater than 0", 33}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-renderpass", 43}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-suspended", 42}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdClearDepthStencilImage-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdControlVideoCodingKHR-bufferlevel", 43}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-flags-07017", 43}, { {"If pCodingControlInfo->flags does not include VK_VIDEO_CODING_CONTROL_RESET_BIT_KHR, then the bound video session must not be in uninitialized state at the time the command is executed on the device", 198}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-pCodingControlInfo-08243", 56}, { {"If the bound video session was not created with an encode operation, then pCodingControlInfo->flags must not include VK_VIDEO_CODING_CONTROL_ENCODE_RATE_CONTROL_BIT_KHR or VK_VIDEO_CODING_CONTROL_ENCODE_QUALITY_LEVEL_BIT_KHR", 224}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-pCodingControlInfo-parameter", 60}, { {"pCodingControlInfo must be a valid pointer to a valid VkVideoCodingControlInfoKHR structure", 91}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-renderpass", 42}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdControlVideoCodingKHR-videocoding", 43}, { {"This command must only be called inside of a video coding scope", 63}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-None-10088", 53}, { {"Memory accessed by the sources and destinations of all of the conversions must not overlap", 90}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-commandBuffer-cmdpool", 64}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-commandBuffer-parameter", 66}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-commandBuffer-recording", 66}, { {"commandBuffer must be in the recording state", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-infoCount-arraylength", 64}, { {"infoCount must be greater than 0", 32}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10083", 54}, { {"For each element of pInfo, srcData.deviceAddress must be a valid VkDeviceAddress", 80}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10084", 54}, { {"For each element of pInfo, srcData.deviceAddress must be 64 byte aligned", 72}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10085", 54}, { {"For each element of pInfo, dstData.deviceAddress must be 64 byte aligned", 72}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10086", 54}, { {"For each element of pInfo, srcSize must be large enough to contain the source matrix, based either on the standard matrix layout or based on the size filled out by vkConvertCooperativeVectorMatrixNV", 198}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10087", 54}, { {"For each element of pInfo, the value pointed to by pDstSize must be large enough to contain the destination matrix, based either on the standard matrix layout or based on the size filled out by vkConvertCooperativeVectorMatrixNV", 228}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfo-10895", 54}, { {"For each element of pInfo, dstData.deviceAddress must be a valid VkDeviceAddress", 80}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-pInfos-parameter", 59}, { {"pInfos must be a valid pointer to an array of infoCount valid VkConvertCooperativeVectorMatrixInfoNV structures", 111}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-renderpass", 53}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-suspended", 52}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdConvertCooperativeVectorMatrixNV-videocoding", 54}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-accelerationStructure-08925", 66}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-pInfo-parameter", 54}, { {"pInfo must be a valid pointer to a valid VkCopyAccelerationStructureInfoKHR structure", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-renderpass", 49}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-src-11633", 48}, { {"The source acceleration structure pInfo->src must have been constructed prior to the execution of this command on the device", 124}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-src-11634", 48}, { {"If the source acceleration structure pInfo->src was constructed through deserialization, all micromap arrays the acceleration structure references that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 403}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureKHR-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-buffer-03718", 50}, { {"The range of src accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-buffer-03719", 50}, { {"The range of dst accessed by this command must be fully backed by physical memory", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-commonparent", 50}, { {"Each of commandBuffer, dst, and src must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-dst-07791", 47}, { {"The range of memory backing dst that is accessed by this command must not overlap the memory backing src that is accessed by this command", 137}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-dst-parameter", 51}, { {"dst must be a valid VkAccelerationStructureNV handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-mode-03410", 48}, { {"mode must be VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR or VK_COPY_ACCELERATION_STRUCTURE_MODE_CLONE_KHR", 109}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-mode-12418", 48}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_CLONE_KHR, the memory pointed to by dst must be at least as large as the size of src, as reported by vkWriteAccelerationStructuresPropertiesKHR or vkCmdWriteAccelerationStructuresPropertiesKHR with a query type of VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR", 306}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-mode-12419", 48}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR, the memory pointed to by dst must be at least as large as the compacted size of src, as reported by vkWriteAccelerationStructuresPropertiesKHR or vkCmdWriteAccelerationStructuresPropertiesKHR with a query type of VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR", 328}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-mode-parameter", 52}, { {"mode must be a valid VkCopyAccelerationStructureModeKHR value", 61}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-renderpass", 48}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-src-03411", 47}, { {"If mode is VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_KHR, src must have been constructed with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR in the build", 165}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-src-parameter", 51}, { {"src must be a valid VkAccelerationStructureNV handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-suspended", 47}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-accelerationStructure-08926", 74}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-commandBuffer-cmdpool", 68}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-commandBuffer-parameter", 70}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-commandBuffer-recording", 70}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-pInfo-03739", 58}, { {"pInfo->dst.deviceAddress must be a valid VkDeviceAddress", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-pInfo-03740", 58}, { {"pInfo->dst.deviceAddress must be aligned to 256 bytes", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-pInfo-11699", 58}, { {"If pInfo->src was built with a type of VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, pInfo->dst.deviceAddress must be aligned to 128 bytes", 145}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-pInfo-11700", 58}, { {"If the serialized acceleration structure in pInfo->src was originally built with a type of VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, pInfo->src.deviceAddress must be aligned to 128 bytes", 197}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-pInfo-parameter", 62}, { {"pInfo must be a valid pointer to a valid VkCopyAccelerationStructureToMemoryInfoKHR structure", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-renderpass", 57}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-suspended", 56}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyAccelerationStructureToMemoryKHR-videocoding", 58}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-01822", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-01823", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-01824", 40}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-commonparent", 33}, { {"Each of commandBuffer, dstBuffer, and srcBuffer must have been created, allocated, or retrieved from the same VkDevice", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-dstBuffer-00120", 36}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-dstBuffer-00121", 36}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-dstBuffer-parameter", 40}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-dstOffset-00114", 36}, { {"The dstOffset member of each element of pRegions must be less than the size of dstBuffer", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-pRegions-00117", 35}, { {"The union of the source regions, and the union of the destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-pRegions-parameter", 39}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferCopy structures", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-regionCount-arraylength", 44}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-renderpass", 31}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-size-00115", 31}, { {"The size member of each element of pRegions must be less than or equal to the size of srcBuffer minus srcOffset", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-size-00116", 31}, { {"The size member of each element of pRegions must be less than or equal to the size of dstBuffer minus dstOffset", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-srcBuffer-00118", 36}, { {"srcBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-srcBuffer-00119", 36}, { {"If srcBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-srcBuffer-parameter", 40}, { {"srcBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-srcOffset-00113", 36}, { {"The srcOffset member of each element of pRegions must be less than the size of srcBuffer", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer-videocoding", 32}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-01822", 41}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-01823", 41}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-01824", 41}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-pCopyBufferInfo-parameter", 47}, { {"pCopyBufferInfo must be a valid pointer to a valid VkCopyBufferInfo2 structure", 78}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-renderpass", 32}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBuffer2-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-bufferImageHeight-09107", 51}, { {"For each element of pRegions, bufferImageHeight must be a multiple of the texel block extent height of the VkFormat of dstImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-bufferRowLength-09106", 49}, { {"For each element of pRegions, bufferRowLength must be a multiple of the texel block extent width of the VkFormat of dstImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-bufferRowLength-09108", 49}, { {"For each element of pRegions, bufferRowLength divided by the texel block extent width and then multiplied by the texel block size of dstImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-01828", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-01829", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-01830", 47}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-07737", 47}, { {"If the maintenance11 feature is not enabled, and the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the bufferOffset member of any element of pRegions must be a multiple of 4", 276}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-07739", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 318}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-11778", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 378}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-11779", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 406}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-11780", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 382}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-11781", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 410}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-commonparent", 40}, { {"Each of commandBuffer, dstImage, and srcBuffer must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-00177", 42}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-00207", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-00208", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-00209", 42}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of dstImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-01997", 42}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07274", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07275", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07276", 42}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07966", 42}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07969", 42}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07973", 42}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07975", 42}, { {"If dstImage does not have either a depth/stencil format or a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the texel block size", 172}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07976", 42}, { {"If dstImage has a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 275}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07978", 42}, { {"If dstImage has a depth/stencil format, the bufferOffset member of any element of pRegions must be a multiple of 4", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07979", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07980", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07981", 42}, { {"If dstImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-07983", 42}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImage-parameter", 46}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImageLayout-00180", 48}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImageLayout-01396", 48}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-dstImageLayout-parameter", 52}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-07738", 45}, { {"The imageOffset and imageExtent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 212}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-09104", 45}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of dstImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10051", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10052", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10053", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10054", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of dstImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10055", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of dstImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10056", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10057", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of dstImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageOffset-10058", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of dstImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageSubresource-07967", 50}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageSubresource-07968", 50}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageSubresource-07971", 50}, { {"For each element of pRegions, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of dstImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageSubresource-07972", 50}, { {"For each element of pRegions, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of dstImage", 207}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-imageSubresource-09105", 50}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in dstImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-pRegions-00171", 42}, { {"srcBuffer must be large enough to contain all buffer locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-pRegions-00173", 42}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-pRegions-07931", 42}, { {"If VK_EXT_depth_range_unrestricted is not enabled, for each element of pRegions whose imageSubresource contains a depth aspect, the data in srcBuffer must be in the range [0,1]", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-pRegions-parameter", 46}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferImageCopy structures", 94}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-regionCount-arraylength", 51}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-renderpass", 38}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-srcBuffer-00174", 43}, { {"srcBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_SRC_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-srcBuffer-00176", 43}, { {"If srcBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-srcBuffer-parameter", 47}, { {"srcBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-01828", 48}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-01829", 48}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-01830", 48}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-07737", 48}, { {"If the maintenance11 feature is not enabled, and the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the bufferOffset member of any element of pCopyBufferToImageInfo->pRegions must be a multiple of 4", 300}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-07739", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pCopyBufferToImageInfo->pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-11778", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyBufferToImageInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 402}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-11779", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyBufferToImageInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 430}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-11780", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyBufferToImageInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 406}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-11781", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyBufferToImageInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 434}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-imageOffset-07738", 46}, { {"The imageOffset and imageExtent members of each element of pCopyBufferToImageInfo->pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 236}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-pCopyBufferToImageInfo-parameter", 61}, { {"pCopyBufferToImageInfo must be a valid pointer to a valid VkCopyBufferToImageInfo2 structure", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-renderpass", 39}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyBufferToImage2-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-commonparent", 47}, { {"Both of commandBuffer, and gpaSession must have been created, allocated, or retrieved from the same VkDevice", 108}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-gpaSession-parameter", 55}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-suspended", 44}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyGpaSessionResultsAMD-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdCopyImage-None-01549", 30}, { {"In a copy to or from a plane of a multi-planar image, the VkFormat of the image and plane must be compatible according to the description of compatible planes for the plane being copied", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-apiVersion-07932", 36}, { {"If the VK_KHR_maintenance1 extension is not enabled, or VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, and either srcImage or dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer and dstSubresource.baseArrayLayer must both be 0, and srcSubresource.layerCount and dstSubresource.layerCount must both be 1", 370}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-apiVersion-07933", 36}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, srcImage and dstImage must have the same VkImageType", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-apiVersion-08969", 36}, { {"If the VK_KHR_maintenance1 extension is not enabled, and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, srcImage or dstImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, extent.depth must be 1", 228}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspect-06662", 32}, { {"If the aspect member of any element of pRegions includes any flag other than VK_IMAGE_ASPECT_STENCIL_BIT or srcImage was not created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 249}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspect-06663", 32}, { {"If the aspect member of any element of pRegions includes any flag other than VK_IMAGE_ASPECT_STENCIL_BIT or dstImage was not created with separate stencil usage, dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 249}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspect-06664", 32}, { {"If the aspect member of any element of pRegions includes VK_IMAGE_ASPECT_STENCIL_BIT, and srcImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 227}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspect-06665", 32}, { {"If the aspect member of any element of pRegions includes VK_IMAGE_ASPECT_STENCIL_BIT, and dstImage was created with separate stencil usage, srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 227}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspectMask-00142", 36}, { {"For each element of pRegions, srcSubresource.aspectMask must specify aspects present in srcImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-aspectMask-00143", 36}, { {"For each element of pRegions, dstSubresource.aspectMask must specify aspects present in dstImage", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-01825", 39}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-01826", 39}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-01827", 39}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-10217", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pRegions, where the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT, the aspectMask of dstSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 385}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-10218", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pRegions, where the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT then the aspectMask of srcSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 389}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11782", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 442}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11783", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 470}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11784", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 446}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11785", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 474}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11786", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 442}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11787", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 470}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11788", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 446}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-11789", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 474}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-12449", 39}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and srcSubresource.aspectMask or dstSubresource.aspectMask of any element of pRegions contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, then both srcImage and dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 370}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-cmdpool", 41}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-parameter", 43}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commandBuffer-recording", 43}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-commonparent", 32}, { {"Each of commandBuffer, dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-00152", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.y must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-01557", 34}, { {"If dstImage has a multi-planar format and the srcImage does not have a multi-planar image format, then for each element of pRegions, srcSubresource.aspectMask must be VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-01786", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-01788", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-01792", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_2D, and srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, extent.depth must equal dstSubresource.layerCount", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-01996", 34}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-04444", 34}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0 and dstSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-07966", 34}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-07969", 34}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-08714", 34}, { {"If dstImage has a multi-planar format, then for each element of pRegions, dstSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImage-parameter", 38}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImageLayout-00133", 40}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImageLayout-01395", 40}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstImageLayout-parameter", 44}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstOffset-00150", 35}, { {"For each element of pRegions, dstOffset.x and (extent.width + dstOffset.x), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 244}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstOffset-00151", 35}, { {"For each element of pRegions, dstOffset.y and (extent.height + dstOffset.y), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 246}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstOffset-00153", 35}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstOffset.z and (extent.depth + dstOffset.z), where extent is adjusted for size-compatibility, must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 290}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstOffset-01784", 35}, { {"The dstOffset and extent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstSubresource-07967", 40}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstSubresource-07968", 40}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-dstSubresource-10215", 40}, { {"If dstSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then srcSubresource.aspectMask must not contain both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 169}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-00124", 34}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07278", 34}, { {"For each element of pRegions, srcOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07279", 34}, { {"For each element of pRegions, srcOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 121}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07280", 34}, { {"For each element of pRegions, srcOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07281", 34}, { {"For each element of pRegions, dstOffset.x must be a multiple of the texel block extent width of the VkFormat of dstImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07282", 34}, { {"For each element of pRegions, dstOffset.y must be a multiple of the texel block extent height of the VkFormat of dstImage", 121}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-07283", 34}, { {"For each element of pRegions, dstOffset.z must be a multiple of the texel block extent depth of the VkFormat of dstImage", 120}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-12201", 34}, { {"For each element of pRegions where srcSubresource.aspectMask and dstSubresource.aspectMask each contain at least one of VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, srcSubresource.aspectMask and dstSubresource.aspectMask must match", 244}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-pRegions-parameter", 38}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageCopy structures", 88}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-regionCount-arraylength", 43}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-renderpass", 30}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-00136", 34}, { {"The sample count of srcImage and dstImage must match", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-00146", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01548", 34}, { {"If the VkFormat of each of srcImage and dstImage is not a multi-planar format, the VkFormat of each of srcImage and dstImage must be size-compatible", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01551", 34}, { {"If neither srcImage nor dstImage has a multi-planar format and the maintenance8 feature is not enabled then for each element of pRegions, srcSubresource.aspectMask and dstSubresource.aspectMask must match", 204}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01556", 34}, { {"If srcImage has a multi-planar format and the dstImage does not have a multi-planar image format, then for each element of pRegions, dstSubresource.aspectMask must be VK_IMAGE_ASPECT_COLOR_BIT", 192}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01728", 34}, { {"For each element of pRegions, if the sum of srcOffset.x and extent.width does not equal the width of the subresource specified by srcSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01729", 34}, { {"For each element of pRegions, if the sum of srcOffset.y and extent.height does not equal the height of the subresource specified by srcSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 241}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01730", 34}, { {"For each element of pRegions, if the sum of srcOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 237}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01785", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.z must be 0 and extent.depth must be 1", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01787", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffset.z must be 0", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01790", 34}, { {"If srcImage and dstImage are both of type VK_IMAGE_TYPE_2D, then for each element of pRegions, extent.depth must be 1", 117}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01791", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_2D, and dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, extent.depth must equal srcSubresource.layerCount", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-01995", 34}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-04443", 34}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.baseArrayLayer must be 0 and srcSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-07743", 34}, { {"If srcImage and dstImage have a different VkImageType, and the maintenance5 feature is not enabled, one must be VK_IMAGE_TYPE_3D and the other must be VK_IMAGE_TYPE_2D", 167}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-07966", 34}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-07969", 34}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-08713", 34}, { {"If srcImage has a multi-planar format, then for each element of pRegions, srcSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-08793", 34}, { {"If srcImage and dstImage have the same VkImageType, for each element of pRegions, if neither of the layerCount members of srcSubresource or dstSubresource are VK_REMAINING_ARRAY_LAYERS, the layerCount members of srcSubresource or dstSubresource must match", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-08794", 34}, { {"If srcImage and dstImage have the same VkImageType, and one of the layerCount members of srcSubresource or dstSubresource is VK_REMAINING_ARRAY_LAYERS, the other member must be either VK_REMAINING_ARRAY_LAYERS or equal to the arrayLayers member of the VkImageCreateInfo used to create the image minus baseArrayLayer", 315}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-09247", 34}, { {"If the VkFormat of each of srcImage and dstImage is a compressed image format, the formats must have the same texel block extent", 128}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-09460", 34}, { {"If srcImage and dstImage are the same, and any elements of pRegions contains the srcSubresource and dstSubresource with matching mipLevel and overlapping array layers, then the srcImageLayout and dstImageLayout must be VK_IMAGE_LAYOUT_GENERAL or VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR", 280}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-10907", 34}, { {"If either the VkFormat of each of srcImage and dstImage is not a compressed image format, and dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, extent.depth must be 1", 189}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-10908", 34}, { {"If either the VkFormat of each of srcImage and dstImage is not a compressed image format, and dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, extent.height must be 1, where extent is adjusted for size-compatibility", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImage-parameter", 38}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImageLayout-00128", 40}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImageLayout-01917", 40}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcImageLayout-parameter", 44}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcOffset-00144", 35}, { {"For each element of pRegions, srcOffset.x and (extent.width + srcOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcOffset-00145", 35}, { {"For each element of pRegions, srcOffset.y and (extent.height + srcOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcOffset-00147", 35}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcOffset.z and (extent.depth + srcOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 240}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcOffset-01783", 35}, { {"The srcOffset and extent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcSubresource-07967", 40}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcSubresource-07968", 40}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcSubresource-10211", 40}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT and dstSubresource.aspectMask is VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, then the VkFormat values of srcImage and dstImage must be compatible according to the list of compatible depth-stencil and color formats", 285}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcSubresource-10212", 40}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT and dstSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then the VkFormat values of srcImage and dstImage must be compatible according to the list of compatible depth-stencil and color formats", 285}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-srcSubresource-10214", 40}, { {"If srcSubresource.aspectMask is VK_IMAGE_ASPECT_COLOR_BIT, then dstSubresource.aspectMask must not contain both VK_IMAGE_ASPECT_DEPTH_BIT and VK_IMAGE_ASPECT_STENCIL_BIT", 169}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-suspended", 29}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage-videocoding", 31}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-01825", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-01826", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-01827", 40}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-10217", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pCopyImageInfo->pRegions, where the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT, the aspectMask of dstSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 401}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-10218", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pCopyImageInfo->pRegions, where the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT then the aspectMask of srcSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 405}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11782", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 458}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11783", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 486}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11784", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 462}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11785", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of srcSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of dstSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of dstImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 490}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11786", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 458}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11787", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 486}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11788", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 462}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-11789", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageInfo->pRegions the aspectMask member of dstSubresource is VK_IMAGE_ASPECT_COLOR_BIT and the aspectMask of srcSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 490}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-12449", 40}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and srcSubresource.aspectMask or dstSubresource.aspectMask of any element of pCopyImageInfo->pRegions contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT, then both srcImage and dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 386}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-pCopyImageInfo-parameter", 45}, { {"pCopyImageInfo must be a valid pointer to a valid VkCopyImageInfo2 structure", 76}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-renderpass", 31}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImage2-videocoding", 32}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-bufferImageHeight-09107", 51}, { {"For each element of pRegions, bufferImageHeight must be a multiple of the texel block extent height of the VkFormat of srcImage", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-bufferRowLength-09106", 49}, { {"For each element of pRegions, bufferRowLength must be a multiple of the texel block extent width of the VkFormat of srcImage", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-bufferRowLength-09108", 49}, { {"For each element of pRegions, bufferRowLength divided by the texel block extent width and then multiplied by the texel block size of srcImage must be less than or equal to 2^31-1", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-01831", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-01832", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-01833", 47}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-07746", 47}, { {"If the maintenance11 feature is not enabled, and the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the bufferOffset member of any element of pRegions must be a multiple of 4", 276}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-10216", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 318}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-11790", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 378}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-11791", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 406}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-11792", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 382}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-11793", 47}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 410}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-commonparent", 40}, { {"Each of commandBuffer, dstBuffer, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-dstBuffer-00191", 43}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-dstBuffer-00192", 43}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-dstBuffer-parameter", 47}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-07747", 45}, { {"The imageOffset and imageExtent members of each element of pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 212}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-09104", 45}, { {"For each element of pRegions, imageOffset.z and (imageExtent.depth + imageOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified imageSubresource of srcImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10051", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, and imageOffset.x does not equal the width of the subresource specified by imageSubresource, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 355}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10052", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, and imageOffset.y does not equal the height of the subresource specified by imageSubresource, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 358}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10053", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the difference of imageOffset.x and extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10054", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.x and extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 255}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10055", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the sum of imageOffset.x and extent.height does not equal the width of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent width of the VkFormat of srcImage", 338}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10056", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, the sum of imageOffset.y and extent.width does not equal the height of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 337}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10057", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR, the difference of imageOffset.y and extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageOffset-10058", 45}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, the difference of imageOffset.y and extent.width must be a multiple of the texel block extent height of the VkFormat of srcImage", 256}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageSubresource-07967", 50}, { {"The imageSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageSubresource-07968", 50}, { {"If imageSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, imageSubresource.baseArrayLayer + imageSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 257}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageSubresource-07971", 50}, { {"For each element of pRegions, imageOffset.x and (imageExtent.width + imageOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified imageSubresource of srcImage", 205}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageSubresource-07972", 50}, { {"For each element of pRegions, imageOffset.y and (imageExtent.height + imageOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified imageSubresource of srcImage", 207}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-imageSubresource-09105", 50}, { {"For each element of pRegions, imageSubresource.aspectMask must specify aspects present in srcImage", 98}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-pRegions-00183", 42}, { {"dstBuffer must be large enough to contain all buffer locations that are accessed according to Buffer and Image Addressing, for each element of pRegions", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-pRegions-00184", 42}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-pRegions-parameter", 46}, { {"pRegions must be a valid pointer to an array of regionCount valid VkBufferImageCopy structures", 94}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-regionCount-arraylength", 51}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-renderpass", 38}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-00186", 42}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-00207", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, the sum of imageOffset.x and extent.width does not equal the width of the subresource specified by imageSubresource, extent.width must be a multiple of the texel block extent width of the VkFormat of srcImage", 334}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-00208", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR, and the sum of imageOffset.y and extent.height does not equal the height of the subresource specified by imageSubresource, extent.height must be a multiple of the texel block extent height of the VkFormat of srcImage", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-00209", 42}, { {"For each element of pRegions, if the sum of imageOffset.z and extent.depth does not equal the depth of the subresource specified by srcSubresource, extent.depth must be a multiple of the texel block extent depth of the VkFormat of srcImage", 239}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-01998", 42}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07274", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR, imageOffset.x must be a multiple of the texel block extent width of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07275", 42}, { {"For each element of pRegions, if VkCopyCommandTransformInfoQCOM::transform is equal to VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR or VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR, imageOffset.y must be a multiple of the texel block extent height of the VkFormat of srcImage", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07276", 42}, { {"For each element of pRegions, imageOffset.z must be a multiple of the texel block extent depth of the VkFormat of srcImage", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07966", 42}, { {"If srcImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07969", 42}, { {"srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07973", 42}, { {"srcImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07975", 42}, { {"If srcImage does not have either a depth/stencil format or a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the texel block size", 172}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07976", 42}, { {"If srcImage has a multi-planar format, then for each element of pRegions, bufferOffset must be a multiple of the element size of the compatible format for the format and the aspectMask of the imageSubresource as defined in Compatible Formats of Planes of Multi-Planar Formats", 275}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07978", 42}, { {"If srcImage has a depth/stencil format, the bufferOffset member of any element of pRegions must be a multiple of 4", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07979", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, imageOffset.y must be 0 and imageExtent.height must be 1", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07980", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, imageOffset.z must be 0 and imageExtent.depth must be 1", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07981", 42}, { {"If srcImage has a multi-planar format, then for each element of pRegions, imageSubresource.aspectMask must be a single valid multi-planar aspect mask bit", 153}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-07983", 42}, { {"If srcImage is of type VK_IMAGE_TYPE_3D, for each element of pRegions, imageSubresource.baseArrayLayer must be 0 and imageSubresource.layerCount must be 1", 154}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImage-parameter", 46}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImageLayout-00189", 48}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImageLayout-01397", 48}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-srcImageLayout-parameter", 52}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-01831", 48}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-01832", 48}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-01833", 48}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-07746", 48}, { {"If the maintenance11 feature is not enabled, and the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the bufferOffset member of any element of pCopyImageToBufferInfo->pRegions must be a multiple of 4", 300}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-10216", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, and the maintenance10 feature is not enabled, for each element of pCopyImageToBufferInfo->pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 342}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-11790", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageToBufferInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 402}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-11791", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageToBufferInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_DEPTH_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_DEPTH_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 430}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-11792", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT but does support VK_QUEUE_COMPUTE_BIT, and in any element of pCopyImageToBufferInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_COMPUTE_QUEUE_BIT_KHR", 406}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-11793", 48}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT and VK_QUEUE_COMPUTE_BIT, but does support VK_QUEUE_TRANSFER_BIT, and in any element of pCopyImageToBufferInfo->pRegions the aspectMask member of imageSubresource is VK_IMAGE_ASPECT_STENCIL_BIT, then the format features of srcImage must contain VK_FORMAT_FEATURE_2_STENCIL_COPY_ON_TRANSFER_QUEUE_BIT_KHR", 434}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-imageOffset-07747", 46}, { {"The imageOffset and imageExtent members of each element of pCopyImageToBufferInfo->pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 236}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-pCopyImageToBufferInfo-parameter", 61}, { {"pCopyImageToBufferInfo must be a valid pointer to a valid VkCopyImageToBufferInfo2 structure", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-renderpass", 39}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToBuffer2-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-addressRange-13129", 49}, { {"At least one buffer containing the total address range specified by the addressRange member of each element of pCopyMemoryInfo->pRegions must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT", 197}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-13025", 50}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, the addressCopyFlags member of all elements of pCopyMemoryInfo->pRegions must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 208}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-13102", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, the addressCopyFlags member of all elements of pCopyMemoryInfo->pRegions must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 215}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-13103", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, pCopyMemoryInfo->image must not have been created with VK_IMAGE_CREATE_PROTECTED_BIT", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-13104", 50}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the addressRange.pname:address member of any element of pCopyMemoryInfo->pRegions must be a multiple of 4", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-13106", 50}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, for each element of pCopyMemoryInfo->pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 289}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-imageLayout-13023", 48}, { {"The imageLayout member of each element of pCopyMemoryInfo->pRegions must be VK_IMAGE_LAYOUT_GENERAL, VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, or VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-imageOffset-13105", 48}, { {"The imageOffset and imageExtent members of each element of pCopyMemoryInfo->pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 229}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-pCopyMemoryInfo-13024", 52}, { {"pCopyMemoryInfo->image must have been created with VK_IMAGE_USAGE_TRANSFER_SRC_BIT", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-pCopyMemoryInfo-parameter", 56}, { {"If pCopyMemoryInfo is not NULL, pCopyMemoryInfo must be a valid pointer to a valid VkCopyDeviceMemoryImageInfoKHR structure", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyImageToMemoryKHR-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-commandBuffer-10936", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support at least one of the queue types specified in VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR::supportedQueues", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-commandBuffer-10937", 51}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-indirectMemoryCopy-10935", 56}, { {"The indirectMemoryCopy feature must be enabled", 46}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-pCopyMemoryIndirectInfo-parameter", 65}, { {"pCopyMemoryIndirectInfo must be a valid pointer to a valid VkCopyMemoryIndirectInfoKHR structure", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-renderpass", 42}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectKHR-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-None-07653", 41}, { {"The indirectCopy feature must be enabled", 40}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-commandBuffer-07656", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support at least one of the queue types specified in VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR::supportedQueues", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-copyBufferAddress-07654", 54}, { {"copyBufferAddress must be 4 byte aligned", 40}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-copyBufferAddress-10946", 54}, { {"Any of the source or destination memory regions specified in copyBufferAddress must not overlap with any of the specified destination memory regions", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-copyBufferAddress-parameter", 58}, { {"copyBufferAddress must be a valid VkDeviceAddress value", 55}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-stride-07655", 43}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkCopyMemoryIndirectCommandNV)", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryIndirectNV-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-13012", 43}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, the srcCopyFlags member of all elements of pCopyMemoryInfo->pRegions must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-13013", 43}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, the dstCopyFlags member of all elements of pCopyMemoryInfo->pRegions must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-13014", 43}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, the dstCopyFlags member of all elements of pCopyMemoryInfo->pRegions must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 204}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-pCopyMemoryInfo-parameter", 49}, { {"If pCopyMemoryInfo is not NULL, pCopyMemoryInfo must be a valid pointer to a valid VkCopyDeviceMemoryInfoKHR structure", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-renderpass", 34}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-suspended", 33}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryKHR-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-accelerationStructure-08927", 74}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-commandBuffer-cmdpool", 68}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-commandBuffer-parameter", 70}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-commandBuffer-recording", 70}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-pInfo-03742", 58}, { {"pInfo->src.deviceAddress must be a valid VkDeviceAddress", 56}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-pInfo-03743", 58}, { {"pInfo->src.deviceAddress must be aligned to 256 bytes", 53}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-pInfo-parameter", 62}, { {"pInfo must be a valid pointer to a valid VkCopyMemoryToAccelerationStructureInfoKHR structure", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-renderpass", 57}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-suspended", 56}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToAccelerationStructureKHR-videocoding", 58}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-commandBuffer-10948", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support at least one of the queue types specified in VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR::supportedQueues", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-commandBuffer-10949", 58}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-indirectMemoryToImageCopy-10947", 70}, { {"The indirectMemoryToImageCopy feature must be enabled", 53}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-pCopyMemoryToImageIndirectInfo-parameter", 79}, { {"pCopyMemoryToImageIndirectInfo must be a valid pointer to a valid VkCopyMemoryToImageIndirectInfoKHR structure", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-renderpass", 49}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectKHR-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-None-07660", 48}, { {"The indirectCopy feature must be enabled", 40}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-aspectMask-07662", 54}, { {"The aspectMask member for every subresource in pImageSubresources must only have a single bit set", 97}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-aspectMask-12287", 54}, { {"The aspectMask member for every subresource in pImageSubresources must specify an aspect present in dstImage", 108}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commandBuffer-07674", 57}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, for each region, the aspectMask member of pImageSubresources must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commandBuffer-10956", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support at least one of the queue types specified in VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR::supportedQueues", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-commonparent", 50}, { {"Both of commandBuffer, and dstImage must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-copyBufferAddress-10975", 61}, { {"Any of the source or destination memory regions specified in copyBufferAddress must not overlap with any of the specified destination memory regions at the time this command is executed on device", 195}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-copyBufferAddress-parameter", 65}, { {"copyBufferAddress must be a valid VkDeviceAddress value", 55}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-copyCount-arraylength", 59}, { {"copyCount must be greater than 0", 32}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-07661", 52}, { {"dstImage must not be a protected image", 38}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-07664", 52}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-07665", 52}, { {"If dstImage is non-sparse then the image or each specified disjoint plane must be bound completely and contiguously to a single VkDeviceMemory object", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-07673", 52}, { {"dstImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 92}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-07973", 52}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-10974", 52}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-12288", 52}, { {"If dstImage is not of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, the specified baseArrayLayer + layerCount of each region in pImageSubresources must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 326}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-12289", 52}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, for each destination region, (imageSubresource.baseArrayLayer + imageSubresource.layerCount) must be less than or equal to the depth of the specified subresource", 302}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-12290", 52}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, and the specified layerCount of each region in pImageSubresources is not VK_REMAINING_ARRAY_LAYERS, for each destination region, if (imageSubresource.baseArrayLayer + imageSubresource.layerCount) does not equal the depth of the specified subresource, imageSubresource.layerCount must be a multiple of the texel block extent depth of the VkFormat of dstImage", 398}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-12291", 52}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each destination region, imageSubresource.baseArrayLayer must be a multiple of the texel block extent depth of the VkFormat of dstImage", 180}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-12292", 52}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, for each destination region, imageSubresource.baseArrayLayer must be less than or equal to the depth of the specified subresource", 170}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImage-parameter", 56}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImageLayout-07667", 58}, { {"dstImageLayout must specify the layout of the image subresources of dstImage at the time this command is executed on a VkDevice", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImageLayout-07669", 58}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, or VK_IMAGE_LAYOUT_GENERAL", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-dstImageLayout-parameter", 62}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-mipLevel-07670", 52}, { {"The specified mipLevel of each region in pImageSubresources must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 148}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-offset-07676", 50}, { {"copyBufferAddress must be 4 byte aligned", 40}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-pImageSubresources-parameter", 66}, { {"pImageSubresources must be a valid pointer to an array of copyCount valid VkImageSubresourceLayers structures", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-renderpass", 48}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-stride-07677", 50}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkCopyMemoryToImageIndirectCommandNV)", 112}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-suspended", 47}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageIndirectNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-None-13022", 41}, { {"If VK_EXT_depth_range_unrestricted is not enabled, for each element of pCopyMemoryInfo->pRegions whose imageSubresource contains a depth aspect, each data element in its addressRange must be in the range [0,1]", 209}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-addressRange-13128", 49}, { {"At least one buffer containing the total address range specified by the addressRange member of each element of pCopyMemoryInfo->pRegions must have been created with VK_BUFFER_USAGE_TRANSFER_SRC_BIT", 197}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-13021", 50}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pCopyMemoryInfo->image must have been created with VK_IMAGE_CREATE_PROTECTED_BIT", 166}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-13102", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, the addressCopyFlags member of all elements of pCopyMemoryInfo->pRegions must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 215}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-13103", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, pCopyMemoryInfo->image must not have been created with VK_IMAGE_CREATE_PROTECTED_BIT", 173}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-13104", 50}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT or VK_QUEUE_COMPUTE_BIT, the addressRange.pname:address member of any element of pCopyMemoryInfo->pRegions must be a multiple of 4", 261}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-13106", 50}, { {"If the queue family used to create the VkCommandPool which commandBuffer was allocated from does not support VK_QUEUE_GRAPHICS_BIT, for each element of pCopyMemoryInfo->pRegions, the aspectMask member of imageSubresource must not be VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 289}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-imageLayout-13019", 48}, { {"The imageLayout member of each element of pCopyMemoryInfo->pRegions must be VK_IMAGE_LAYOUT_GENERAL, VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, or VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL", 176}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-imageOffset-13105", 48}, { {"The imageOffset and imageExtent members of each element of pCopyMemoryInfo->pRegions must respect the image transfer granularity requirements of commandBuffer's command pool's queue family, as described in VkQueueFamilyProperties", 229}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-pCopyMemoryInfo-13020", 52}, { {"pCopyMemoryInfo->image must have been created with VK_IMAGE_USAGE_TRANSFER_DST_BIT", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-pCopyMemoryInfo-parameter", 56}, { {"If pCopyMemoryInfo is not NULL, pCopyMemoryInfo must be a valid pointer to a valid VkCopyDeviceMemoryImageInfoKHR structure", 123}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToImageKHR-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-buffer-07546", 46}, { {"The buffer used to create pInfo->dst must be bound to device memory", 67}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-pInfo-07543", 45}, { {"pInfo->src.deviceAddress must be a valid VkDeviceAddress", 56}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-pInfo-07544", 45}, { {"pInfo->src.deviceAddress must be aligned to 256 bytes", 53}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-pInfo-parameter", 49}, { {"pInfo must be a valid pointer to a valid VkCopyMemoryToMicromapInfoEXT structure", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-renderpass", 44}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-suspended", 43}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMemoryToMicromapEXT-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-buffer-07529", 38}, { {"The buffer used to create pInfo->src must be bound to device memory", 67}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-buffer-07530", 38}, { {"The buffer used to create pInfo->dst must must be bound to device memory", 72}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-pInfo-parameter", 41}, { {"pInfo must be a valid pointer to a valid VkCopyMicromapInfoEXT structure", 72}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-renderpass", 36}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-src-07532", 35}, { {"The source micromap src must have been constructed prior to the execution of this command on the device", 103}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapEXT-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-buffer-07539", 46}, { {"The buffer used to create pInfo->src must be bound to device memory", 67}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-pInfo-07536", 45}, { {"pInfo->dst.deviceAddress must be a valid VkDeviceAddress", 56}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-pInfo-07537", 45}, { {"pInfo->dst.deviceAddress must be aligned to 256 bytes", 53}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-pInfo-parameter", 49}, { {"pInfo must be a valid pointer to a valid VkCopyMicromapToMemoryInfoEXT structure", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-renderpass", 44}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-suspended", 43}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyMicromapToMemoryEXT-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdCopyQueryPoolResults-None-07429", 41}, { {"All queries used by the command must not be active", 50}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-None-08752", 41}, { {"All queries used by the command must have been made available by prior executed commands", 88}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-None-09402", 41}, { {"All queries used by the command must not be uninitialized when the command is executed", 86}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-commonparent", 43}, { {"Each of commandBuffer, dstBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 118}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-dstBuffer-00824", 46}, { {"dstBuffer must have enough storage, from dstOffset, to contain the result of each query, as described here", 106}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-dstBuffer-00825", 46}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-dstBuffer-00826", 46}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-dstBuffer-parameter", 50}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-dstOffset-00819", 46}, { {"dstOffset must be less than the size of dstBuffer", 49}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-firstQuery-09436", 47}, { {"firstQuery must be less than the number of queries in queryPool", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-firstQuery-09437", 47}, { {"The sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-flags-00822", 42}, { {"If VK_QUERY_RESULT_64_BIT is not set in flags then dstOffset must be a multiple of 4", 84}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-flags-00823", 42}, { {"If VK_QUERY_RESULT_64_BIT is set in flags then dstOffset must be a multiple of 8", 80}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-flags-09443", 42}, { {"If flags includes VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, then it must not include VK_QUERY_RESULT_WITH_AVAILABILITY_BIT", 117}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-flags-parameter", 46}, { {"flags must be a valid combination of VkQueryResultFlagBits values", 65}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryCount-09438", 47}, { {"If queryCount is greater than 1, stride must not be zero", 56}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryCount-12254", 47}, { {"If queryCount is greater than 1 and VK_QUERY_RESULT_64_BIT is not set in flags then stride must be a multiple of 4", 114}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryCount-12255", 47}, { {"If queryCount is greater than 1 and VK_QUERY_RESULT_64_BIT is set in flags then stride must be a multiple of 8", 110}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryPool-parameter", 50}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-02734", 46}, { {"vkCmdCopyQueryPoolResults must not be called if the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_INTEL", 128}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-03232", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, VkPhysicalDevicePerformanceQueryPropertiesKHR::allowCommandBufferQueryCopies must be VK_TRUE", 175}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-09439", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIMESTAMP, flags must not contain VK_QUERY_RESULT_PARTIAL_BIT", 121}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-09440", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, flags must not contain VK_QUERY_RESULT_WITH_AVAILABILITY_BIT, VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, VK_QUERY_RESULT_PARTIAL_BIT, or VK_QUERY_RESULT_64_BIT", 236}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-09441", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the queryPool must have been recorded once for each pass as retrieved via a call to vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR", 222}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-09442", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then flags must include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 143}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-11874", 46}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then flags must not include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 194}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-queryType-12424", 46}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIME_ELAPSED_QCOM, flags must not contain VK_QUERY_RESULT_PARTIAL_BIT or VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 168}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResults-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-None-13076", 52}, { {"All queries used by the command must not be uninitialized when the command is executed", 86}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-None-13083", 52}, { {"All queries used by the command must not be active", 50}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-None-13084", 52}, { {"All queries used by the command must have been made available by prior executed commands", 88}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-commonparent", 54}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-dstFlags-13085", 56}, { {"dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 62}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-dstFlags-13100", 56}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-dstFlags-13101", 56}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-dstFlags-parameter", 60}, { {"dstFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-firstQuery-09436", 58}, { {"firstQuery must be less than the number of queries in queryPool", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-firstQuery-09437", 58}, { {"The sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-flags-09443", 53}, { {"If queryResultFlags includes VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, then it must not include VK_QUERY_RESULT_WITH_AVAILABILITY_BIT", 128}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-flags-13077", 53}, { {"If VK_QUERY_RESULT_64_BIT is not set in flags then pDstRange->address and pDstRange->stride must be multiples of 4", 114}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-flags-13078", 53}, { {"If VK_QUERY_RESULT_64_BIT is set in flags then pDstRange->address and pDstRange->stride must be multiples of 8", 110}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13079", 57}, { {"pDstRange->size must be large enough to contain the result of each query, as described here", 91}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13080", 57}, { {"The buffer from which the range defined by pDstRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag", 136}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13097", 57}, { {"If the range specified by pDstRange is not bound completely to memory when accessed, dstFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 149}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13098", 57}, { {"If the buffer from which the range specified by pDstRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 204}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13099", 57}, { {"If the buffer from which the range specified by pDstRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13122", 57}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 283}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13123", 57}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 231}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13124", 57}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 320}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-13125", 57}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 257}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-pDstRange-parameter", 61}, { {"pDstRange must be a valid pointer to a valid VkStridedDeviceAddressRangeKHR structure", 85}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryCount-09438", 58}, { {"If queryCount is greater than 1, stride must not be zero", 56}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryPool-parameter", 61}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryResultFlags-parameter", 68}, { {"queryResultFlags must be a valid combination of VkQueryResultFlagBits values", 76}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-09439", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIMESTAMP, queryResultFlags must not contain VK_QUERY_RESULT_PARTIAL_BIT", 132}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-09440", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, queryResultFlags must not contain VK_QUERY_RESULT_WITH_AVAILABILITY_BIT, VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, VK_QUERY_RESULT_PARTIAL_BIT, or VK_QUERY_RESULT_64_BIT", 247}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-09441", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the queryPool must have been recorded once for each pass as retrieved via a call to vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR", 222}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-09442", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then queryResultFlags must include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 154}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-11874", 57}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then queryResultFlags must not include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 205}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-12424", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIME_ELAPSED_QCOM, flags must not contain VK_QUERY_RESULT_PARTIAL_BIT or VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 168}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-13081", 57}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, VkPhysicalDevicePerformanceQueryPropertiesKHR::allowCommandBufferQueryCopies must be VK_TRUE", 175}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-queryType-13082", 57}, { {"vkCmdCopyQueryPoolResultsToMemoryKHR must not be called if the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_INTEL", 139}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-renderpass", 52}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-suspended", 51}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyQueryPoolResultsToMemoryKHR-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdCopyTensorARM-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-pCopyTensorInfo-parameter", 49}, { {"pCopyTensorInfo must be a valid pointer to a valid VkCopyTensorInfoARM structure", 80}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-renderpass", 34}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-suspended", 33}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCopyTensorARM-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-pLaunchInfo-parameter", 49}, { {"pLaunchInfo must be a valid pointer to a valid VkCuLaunchInfoNVX structure", 74}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCuLaunchKernelNVX-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-pLaunchInfo-parameter", 50}, { {"pLaunchInfo must be a valid pointer to a valid VkCudaLaunchInfoNV structure", 75}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdCudaLaunchKernelNV-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDebugMarkerBeginEXT-None-10614", 40}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerBeginEXT-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerBeginEXT-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerBeginEXT-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerBeginEXT-pMarkerInfo-parameter", 51}, { {"pMarkerInfo must be a valid pointer to a valid VkDebugMarkerMarkerInfoEXT structure", 83}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-None-10615", 38}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-commandBuffer-01239", 47}, { {"There must be an outstanding vkCmdDebugMarkerBeginEXT command prior to the vkCmdDebugMarkerEndEXT on the queue that commandBuffer is submitted to", 145}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-commandBuffer-01240", 47}, { {"If commandBuffer is a secondary command buffer, there must be an outstanding vkCmdDebugMarkerBeginEXT command recorded to commandBuffer that has not previously been ended by a call to vkCmdDebugMarkerEndEXT", 206}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerEndEXT-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerInsertEXT-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerInsertEXT-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerInsertEXT-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDebugMarkerInsertEXT-pMarkerInfo-parameter", 52}, { {"pMarkerInfo must be a valid pointer to a valid VkDebugMarkerMarkerInfoEXT structure", 83}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-07011", 35}, { {"The bound video session must not be in uninitialized state at the time the command is executed on the device", 108}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-07258", 35}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR but was not created with interlaced frame support, then the decode output picture must represent a frame", 219}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-08249", 35}, { {"The bound video session must have been created with a decode operation", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-10400", 35}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then there must be a bound video session parameters object if any of the following conditions are not met: the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR the pNext chain of pDecodeInfo includes a VkVideoDecodeH264InlineSessionParametersInfoKHR structure the pStdSPS member of the VkVideoDecodeH264InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL the pStdPPS member of the VkVideoDecodeH264InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL", 699}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-10403", 35}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then there must be a bound video session parameters object if any of the following conditions are not met: the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure the pStdVPS member of the VkVideoDecodeH265InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL the pStdSPS member of the VkVideoDecodeH265InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL the pStdPPS member of the VkVideoDecodeH265InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL", 838}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-None-10407", 35}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then there must be a bound video session parameters object if any of the following conditions are not met: the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR the pNext chain of pDecodeInfo includes a VkVideoDecodeAV1InlineSessionParametersInfoKHR structure the pStdSequenceHeader member of the VkVideoDecodeAV1InlineSessionParametersInfoKHR structure included in the pNext chain of pDecodeInfo is not NULL", 568}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-StdVideoH264PictureParameterSet-07155", 62}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the bound video session parameters object must contain a StdVideoH264PictureParameterSet entry with seq_parameter_set_id and pic_parameter_set_id matching StdVideoDecodeH264PictureInfo::seq_parameter_set_id and StdVideoDecodeH264PictureInfo::pic_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoDecodeH264PictureInfoKHR structure included in the pNext chain of pDecodeInfo , unless the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR and the pNext chain of pDecodeInfo includes a VkVideoDecodeH264InlineSessionParametersInfoKHR structure, and its pStdPPS member is not equal to NULL", 795}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-StdVideoH264SequenceParameterSet-07154", 63}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the bound video session parameters object must contain a StdVideoH264SequenceParameterSet entry with seq_parameter_set_id matching StdVideoDecodeH264PictureInfo::seq_parameter_set_id that is provided in the pStdPictureInfo member of the VkVideoDecodeH264PictureInfoKHR structure included in the pNext chain of pDecodeInfo , unless the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR and the pNext chain of pDecodeInfo includes a VkVideoDecodeH264InlineSessionParametersInfoKHR structure, and its pStdSPS member is not equal to NULL", 699}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-StdVideoH265PictureParameterSet-07162", 62}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265PictureParameterSet entry with sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id matching StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id, StdVideoDecodeH265PictureInfo::pps_seq_parameter_set_id, and StdVideoDecodeH265PictureInfo::pps_pic_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo , unless the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure, and its pStdPPS member is not equal to NULL", 900}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-StdVideoH265SequenceParameterSet-07161", 63}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265SequenceParameterSet entry with sps_video_parameter_set_id and sps_seq_parameter_set_id matching StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id and StdVideoDecodeH265PictureInfo::pps_seq_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo , unless the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure, and its pStdSPS member is not equal to NULL", 816}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-StdVideoH265VideoParameterSet-07160", 60}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265VideoParameterSet entry with vps_video_parameter_set_id matching StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id that is provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo , unless the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure, and its pStdVPS member is not equal to NULL", 708}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-activeReferencePictureCount-07150", 58}, { {"activeReferencePictureCount must be less than or equal to the VkVideoSessionCreateInfoKHR::maxActiveReferencePictures specified when the bound video session was created", 168}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-bufferlevel", 36}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-codedOffset-07257", 42}, { {"The codedOffset member of the VkVideoPictureResourceInfoKHR structure pointed to by the pPictureResource member of each element of pDecodeInfo->pReferenceSlots must be an integer multiple of codedOffsetGranularity", 213}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-07136", 44}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, then pDecodeInfo->srcBuffer must not be a protected buffer", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-07147", 44}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, then pDecodeInfo->dstPictureResource.imageViewBinding must not have been created from a protected image", 192}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-07148", 44}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, then pDecodeInfo->dstPictureResource.imageViewBinding must have been created from a protected image", 185}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_VIDEO_DECODE_BIT_KHR operations", 109}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-compressedHeaderOffset-10807", 53}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the compressedHeaderOffset member of the VkVideoDecodeVP9PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 290}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-dpbBottomFieldUseCount-07178", 53}, { {"All elements of dpbBottomFieldUseCount must be less than or equal to 1", 70}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-dpbFrameUseCount-07176", 47}, { {"All elements of dpbFrameUseCount must be less than or equal to 1", 64}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-dpbTopFieldUseCount-07177", 50}, { {"All elements of dpbTopFieldUseCount must be less than or equal to 1", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-filmGrainSupport-09248", 47}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR and VkVideoDecodeAV1ProfileInfoKHR::filmGrainSupport set to VK_FALSE, then film grain must not be enabled for the decoded picture", 243}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-frameHeaderOffset-09251", 48}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the frameHeaderOffset member of the VkVideoDecodeAV1PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 285}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-opCount-07134", 38}, { {"For each active query, the active query index corresponding to the query type of that query plus opCount must be less than or equal to the last activatable query index corresponding to the query type of that query plus one", 222}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07135", 42}, { {"pDecodeInfo->srcBuffer must be compatible with the video profile the bound video session was created with", 105}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07138", 42}, { {"pDecodeInfo->srcBufferOffset must be an integer multiple of VkVideoCapabilitiesKHR::minBitstreamBufferOffsetAlignment, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 236}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07139", 42}, { {"pDecodeInfo->srcBufferRange must be an integer multiple of VkVideoCapabilitiesKHR::minBitstreamBufferSizeAlignment, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 233}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07140", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL and VkVideoDecodeCapabilitiesKHR::flags does not include VK_VIDEO_DECODE_CAPABILITY_DPB_AND_OUTPUT_COINCIDE_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the video picture resources specified by pDecodeInfo->dstPictureResource and pDecodeInfo->pSetupReferenceSlot->pPictureResource must not match", 431}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07141", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL and none of the following is true: VkVideoDecodeCapabilitiesKHR::flags includes VK_VIDEO_DECODE_CAPABILITY_DPB_AND_OUTPUT_DISTINCT_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR and VkVideoDecodeAV1ProfileInfoKHR::filmGrainSupport set to VK_TRUE, and film grain is enabled for the decoded picture then the video picture resources specified by pDecodeInfo->dstPictureResource and pDecodeInfo->pSetupReferenceSlot->pPictureResource must match", 679}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07142", 42}, { {"pDecodeInfo->dstPictureResource.imageViewBinding must be compatible with the video profile the bound video session was created with", 131}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07143", 42}, { {"The format of pDecodeInfo->dstPictureResource.imageViewBinding must match the VkVideoSessionCreateInfoKHR::pictureFormat the bound video session was created with", 161}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07144", 42}, { {"pDecodeInfo->dstPictureResource.codedOffset must be an integer multiple of codedOffsetGranularity", 97}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07145", 42}, { {"pDecodeInfo->dstPictureResource.codedExtent must be between minCodedExtent and maxCodedExtent, inclusive, the bound video session was created with", 146}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07146", 42}, { {"pDecodeInfo->dstPictureResource.imageViewBinding must have been created with the VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR usage flag set", 135}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07149", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL, then pDecodeInfo->pSetupReferenceSlot->pPictureResource must match one of the bound reference picture resource", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07151", 42}, { {"The pPictureResource member of each element of pDecodeInfo->pReferenceSlots must match one of the bound reference picture resource associated with the DPB slot index specified in the slotIndex member of that element", 215}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07156", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and pDecodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pDecodeInfo->pSetupReferenceSlot must include a VkVideoDecodeH264DpbSlotInfoKHR structure", 278}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07163", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and pDecodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pDecodeInfo->pSetupReferenceSlot must include a VkVideoDecodeH265DpbSlotInfoKHR structure", 278}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07170", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL, then pDecodeInfo->pSetupReferenceSlot->slotIndex must be less than the VkVideoSessionCreateInfoKHR::maxDpbSlots specified when the bound video session was created", 211}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07173", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL, then pDecodeInfo->pSetupReferenceSlot->pPictureResource->codedOffset must be an integer multiple of codedOffsetGranularity", 171}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07259", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR but was not created with interlaced frame support, and pDecodeInfo->pSetupReferenceSlot is not NULL, then the reconstructed picture must represent a frame", 269}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07260", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR but was not created with interlaced frame support, then each active reference picture corresponding to the elements of pDecodeInfo->pReferenceSlots must represent a frame", 285}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07261", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, pDecodeInfo->pSetupReferenceSlot is not NULL, and the decode output picture represents a frame, then the reconstructed picture must also represent a frame", 270}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07262", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, pDecodeInfo->pSetupReferenceSlot is not NULL, and the decode output picture represents a top field, then the reconstructed picture must also represent a top field", 278}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07263", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, pDecodeInfo->pSetupReferenceSlot is not NULL, and the decode output picture represents a bottom field, then the reconstructed picture must also represent a bottom field", 284}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07264", 42}, { {"Each video picture resource corresponding to the pPictureResource member specified in the elements of pDecodeInfo->pReferenceSlots must be unique within pDecodeInfo->pReferenceSlots", 181}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07266", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and an active reference picture corresponding to any element of pDecodeInfo->pReferenceSlots represents a frame, then the DPB slot index of the bound video session specified by the slotIndex member of that element must be currently associated with a frame picture matching the video picture resource specified by the pPictureResource member of the same element at the time the command is executed on the device", 525}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07267", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and an active reference picture corresponding to any element of pDecodeInfo->pReferenceSlots represents a top field, then the DPB slot index of the bound video session specified by the slotIndex member of that element must be currently associated with a top field picture matching the video picture resource specified by the pPictureResource member of the same element at the time the command is executed on the device", 533}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-07268", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and an active reference picture corresponding to any element of pDecodeInfo->pReferenceSlots represents a bottom field, then the DPB slot index of the bound video session specified by the slotIndex member of that element must be currently associated with a bottom field picture matching the video picture resource specified by the pPictureResource member of the same element at the time the command is executed on the device", 539}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-08376", 42}, { {"pDecodeInfo->pSetupReferenceSlot must not be NULL unless the bound video session was created with VkVideoSessionCreateInfoKHR::maxDpbSlots equal to zero", 152}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-09249", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, pDecodeInfo->pSetupReferenceSlot is not NULL, and film grain is enabled for the decoded picture, then the video picture resources specified by pDecodeInfo->dstPictureResource and pDecodeInfo->pSetupReferenceSlot->pPictureResource must not match", 359}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-09254", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR and pDecodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pDecodeInfo->pSetupReferenceSlot must include a VkVideoDecodeAV1DpbSlotInfoKHR structure", 276}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-10801", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is NULL or pDecodeInfo->pSetupReferenceSlot->pPictureResource does not refer to the same image subresource as pDecodeInfo->dstPictureResource, then the image subresource referred to by pDecodeInfo->dstPictureResource must be in the VK_IMAGE_LAYOUT_VIDEO_DECODE_DST_KHR layout at the time the video decode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 496}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-10802", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL and pDecodeInfo->pSetupReferenceSlot->pPictureResource refers to the same image subresource as pDecodeInfo->dstPictureResource, then the image subresource referred to by pDecodeInfo->dstPictureResource must be in the VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR layout at the time the video decode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 493}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-10803", 42}, { {"If pDecodeInfo->pSetupReferenceSlot is not NULL, then the image subresource referred to by pDecodeInfo->pSetupReferenceSlot->pPictureResource must be in the VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR layout at the time the video decode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 385}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pDecodeInfo-parameter", 46}, { {"pDecodeInfo must be a valid pointer to a valid VkVideoDecodeInfoKHR structure", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-07152", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the pNext chain of pDecodeInfo must include a VkVideoDecodeH264PictureInfoKHR structure", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-07157", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the pNext chain of each element of pDecodeInfo->pReferenceSlots must include a VkVideoDecodeH264DpbSlotInfoKHR structure", 241}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-07158", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the pNext chain of pDecodeInfo must include a VkVideoDecodeH265PictureInfoKHR structure", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-07164", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the pNext chain of each element of pDecodeInfo->pReferenceSlots must include a VkVideoDecodeH265DpbSlotInfoKHR structure", 241}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-08365", 36}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoInlineQueryInfoKHR structure with its queryPool member specifying a valid VkQueryPool handle, then VkVideoInlineQueryInfoKHR::queryCount must equal opCount", 300}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-08366", 36}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoInlineQueryInfoKHR structure with its queryPool member specifying a valid VkQueryPool handle, then all the queries used by the command, as specified by the VkVideoInlineQueryInfoKHR structure, must be unavailable", 357}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-09250", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the pNext chain of pDecodeInfo must include a VkVideoDecodeAV1PictureInfoKHR structure", 206}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-09255", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the pNext chain of each element of pDecodeInfo->pReferenceSlots must include a VkVideoDecodeAV1DpbSlotInfoKHR structure", 239}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10401", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoDecodeH264InlineSessionParametersInfoKHR structure with a non-NULL pStdSPS member, then pStdSPS->seq_parameter_set_id must equal StdVideoDecodeH264PictureInfo::seq_parameter_set_id provided in the pStdPictureInfo member of the VkVideoDecodeH264PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 547}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10402", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoDecodeH264InlineSessionParametersInfoKHR structure with a non-NULL pStdPPS member, then pStdPPS->seq_parameter_set_id and pStdPPS->pic_parameter_set_id must equal StdVideoDecodeH264PictureInfo::seq_parameter_set_id and StdVideoDecodeH264PictureInfo::pic_parameter_set_id, respectively, provided in the pStdPictureInfo member of the VkVideoDecodeH264PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 652}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10404", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure with a non-NULL pStdVPS member, then pStdVPS->vps_video_parameter_set_id must equal StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 559}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10405", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure with a non-NULL pStdSPS member, then pStdSPS->sps_video_parameter_set_id and pStdSPS->sps_seq_parameter_set_id must equal StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id and StdVideoDecodeH265PictureInfo::pps_seq_parameter_set_id, respectively, provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 672}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10406", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and with VK_VIDEO_SESSION_CREATE_INLINE_SESSION_PARAMETERS_BIT_KHR, and the pNext chain of pDecodeInfo includes a VkVideoDecodeH265InlineSessionParametersInfoKHR structure with a non-NULL pStdPPS member, then pStdPPS->sps_video_parameter_set_id, pStdPPS->pps_seq_parameter_set_id, and pStdPPS->pps_pic_parameter_set_id must equal StdVideoDecodeH265PictureInfo::sps_video_parameter_set_id, StdVideoDecodeH265PictureInfo::pps_seq_parameter_set_id, and StdVideoDecodeH265PictureInfo::pps_pic_parameter_set_id, respectively, provided in the pStdPictureInfo member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 766}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pNext-10805", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the pNext chain of pDecodeInfo must include a VkVideoDecodeVP9PictureInfoKHR structure", 206}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pPictureResource-10804", 47}, { {"The image subresource referred to by the pPictureResource member of each element of pDecodeInfo->pReferenceSlots must be in the VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR layout at the time the video decode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 356}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pSliceOffsets-07153", 44}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then all elements of the pSliceOffsets member of the VkVideoDecodeH264PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 299}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pSliceSegmentOffsets-07159", 51}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then all elements of the pSliceSegmentOffsets member of the VkVideoDecodeH265PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 306}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pTileOffsets-09252", 43}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then for each element i of the pTileOffsets and pTileSizes members of the VkVideoDecodeAV1PictureInfoKHR structure included in the pNext chain of pDecodeInfo the sum of pTileOffsets[i] and pTileSizes[i] must be less than or equal to pDecodeInfo->srcBufferRange", 375}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-pTileOffsets-09253", 43}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then all elements of the pTileOffsets member of the VkVideoDecodeAV1PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 296}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-queryPool-08368", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, then the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pDecodeInfo must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 439}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-queryType-08367", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, then the queryType used to create the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pDecodeInfo must be VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR", 280}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-queryType-08369", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the queryType used to create the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pDecodeInfo is VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then the VkCommandPool that commandBuffer was allocated from must have been created with a queue family index that supports result status queries, as indicated by VkQueueFamilyQueryResultStatusPropertiesKHR::queryResultStatusSupport", 508}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-referenceNameSlotIndices-09262", 55}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then each element of the referenceNameSlotIndices array member of the VkVideoDecodeAV1PictureInfoKHR structure included in the pNext chain of pDecodeInfo must either be negative or must equal the slotIndex member of one of the elements of pDecodeInfo->pReferenceSlots", 382}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-referenceNameSlotIndices-10809", 55}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then each element of the referenceNameSlotIndices array member of the VkVideoDecodeVP9PictureInfoKHR structure included in the pNext chain of pDecodeInfo must either be negative or must equal the slotIndex member of one of the elements of pDecodeInfo->pReferenceSlots", 382}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-renderpass", 35}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-slotIndex-07256", 40}, { {"The slotIndex member of each element of pDecodeInfo->pReferenceSlots must be less than the VkVideoSessionCreateInfoKHR::maxDpbSlots specified when the bound video session was created", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-slotIndex-09263", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the slotIndex member of each element of pDecodeInfo->pReferenceSlots must equal one of the elements of the referenceNameSlotIndices array member of the VkVideoDecodeAV1PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 355}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-slotIndex-10810", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the slotIndex member of each element of pDecodeInfo->pReferenceSlots must equal one of the elements of the referenceNameSlotIndices array member of the VkVideoDecodeVP9PictureInfoKHR structure included in the pNext chain of pDecodeInfo", 355}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-tilesOffset-10808", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the tilesOffset member of the VkVideoDecodeVP9PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 279}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-uncompressedHeaderOffset-10806", 55}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the uncompressedHeaderOffset member of the VkVideoDecodeVP9PictureInfoKHR structure included in the pNext chain of pDecodeInfo must be less than pDecodeInfo->srcBufferRange", 292}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecodeVideoKHR-videocoding", 36}, { {"This command must only be called inside of a video coding scope", 63}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdDecompressMemoryEXT-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-memoryDecompression-11761", 55}, { {"The memoryDecompression feature must be enabled", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-pDecompressMemoryInfoEXT-parameter", 64}, { {"pDecompressMemoryInfoEXT must be a valid pointer to a valid VkDecompressMemoryInfoEXT structure", 95}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-renderpass", 40}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-suspended", 39}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryEXT-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-None-07692", 53}, { {"The memoryDecompression feature must be enabled", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-commandBuffer-cmdpool", 64}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-commandBuffer-parameter", 66}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-commandBuffer-recording", 66}, { {"commandBuffer must be in the recording state", 44}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-decompressionMethod-07690", 68}, { {"The decompressionMethod must have a single bit set", 50}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-decompressionMethod-11769", 68}, { {"If decompressionMethod is VK_MEMORY_DECOMPRESSION_METHOD_GDEFLATE_1_0_BIT_EXT, then all values in VkDecompressMemoryRegionEXT::decompressedSize must be less than or equal to 65536 bytes", 185}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-decompressionMethod-11810", 68}, { {"decompressionMethod must be a valid bit specified in VkPhysicalDeviceMemoryDecompressionPropertiesEXT::decompressionMethods", 123}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-decompressionMethod-parameter", 72}, { {"decompressionMethod must be a valid combination of VkMemoryDecompressionMethodFlagBitsEXT values", 96}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-decompressionMethod-requiredbitmask", 78}, { {"decompressionMethod must not be 0", 33}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsAddress-07694", 72}, { {"indirectCommandsAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 159}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsAddress-07695", 72}, { {"indirectCommandsAddress must be a multiple of 4", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsAddress-11794", 72}, { {"All device addresses between indirectCommandsAddress and indirectCommandsAddress + (stride x (count stored in indirectCommandsCountAddress)) - 1 must be in the buffer device address range of the same buffer", 206}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsAddress-parameter", 76}, { {"indirectCommandsAddress must be a valid VkDeviceAddress value", 61}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsCountAddress-07697", 77}, { {"indirectCommandsCountAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 164}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsCountAddress-07698", 77}, { {"indirectCommandsCountAddress must be a multiple of 4", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsCountAddress-07699", 77}, { {"The count stored in indirectCommandsCountAddress must be less than or equal to VkPhysicalDeviceMemoryDecompressionPropertiesEXT::maxDecompressionIndirectCount", 158}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-indirectCommandsCountAddress-parameter", 81}, { {"indirectCommandsCountAddress must be a valid VkDeviceAddress value", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-maxDecompressionCount-11768", 70}, { {"maxDecompressionCount must be less than or equal to VkPhysicalDeviceMemoryDecompressionPropertiesEXT::maxDecompressionIndirectCount", 131}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-renderpass", 53}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-stride-11767", 55}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDecompressMemoryRegionEXT)", 103}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-suspended", 52}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountEXT-videocoding", 54}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-None-07692", 52}, { {"The memoryDecompression feature must be enabled", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsAddress-07694", 71}, { {"indirectCommandsAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 159}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsAddress-07695", 71}, { {"indirectCommandsAddress must be a multiple of 4", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsAddress-11794", 71}, { {"All device addresses between indirectCommandsAddress and indirectCommandsAddress + (stride x (count stored in indirectCommandsCountAddress)) - 1 must be in the buffer device address range of the same buffer", 206}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsAddress-parameter", 75}, { {"indirectCommandsAddress must be a valid VkDeviceAddress value", 61}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsCountAddress-07697", 76}, { {"indirectCommandsCountAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 164}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsCountAddress-07698", 76}, { {"indirectCommandsCountAddress must be a multiple of 4", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsCountAddress-07699", 76}, { {"The count stored in indirectCommandsCountAddress must be less than or equal to VkPhysicalDeviceMemoryDecompressionPropertiesEXT::maxDecompressionIndirectCount", 158}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-indirectCommandsCountAddress-parameter", 80}, { {"indirectCommandsCountAddress must be a valid VkDeviceAddress value", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-renderpass", 52}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-stride-11770", 54}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDecompressMemoryRegionNV)", 102}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-suspended", 51}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryIndirectCountNV-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-None-07684", 39}, { {"The memoryDecompression feature must be enabled", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-decompressRegionCount-arraylength", 62}, { {"decompressRegionCount must be greater than 0", 44}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-memoryDecompression-11766", 54}, { {"The memoryDecompression feature must be enabled", 47}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-pDecompressMemoryRegions-parameter", 63}, { {"pDecompressMemoryRegions must be a valid pointer to an array of decompressRegionCount valid VkDecompressMemoryRegionNV structures", 129}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-renderpass", 39}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDecompressMemoryNV-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/memory_decompression.html", 34} } },
        { {"VUID-vkCmdDispatch-None-02691", 29}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-02692", 29}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-02693", 29}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-06479", 29}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-07288", 29}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-07888", 29}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08114", 29}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08115", 29}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08116", 29}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08117", 29}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08119", 29}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08600", 29}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08601", 29}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08605", 29}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08606", 29}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08608", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08609", 29}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08610", 29}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08611", 29}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08612", 29}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-08613", 29}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-09600", 29}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10068", 29}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10672", 29}, { {"If the per-tile execution model is not enabled, this command must be called outside of a render pass instance", 109}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10674", 29}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDispatch feature must be enabled", 98}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10675", 29}, { {"Memory backing image subresources used as tile attachments in the current render pass must not be written in any way other than as a tile attachment by this command", 164}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10676", 29}, { {"If any recorded command in the current subpass will write to an image subresource as a tile attachment, this command must not read from the memory backing that image subresource in any other way than as a tile attachment", 220}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10678", 29}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10679", 29}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-10743", 29}, { {"If there is no bound compute pipeline, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_COMPUTE_BIT stage", 113}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11297", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11298", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11299", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11300", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11301", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11302", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11304", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11305", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11306", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11308", 29}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11309", 29}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11372", 29}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11373", 29}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11374", 29}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11376", 29}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11397", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11398", 29}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11437", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11438", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11439", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11441", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11442", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-11485", 29}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-None-12472", 29}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchSADQCOM-06976", 49}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchSADQCOM-12420", 49}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchSSDQCOM-06974", 49}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchWindow-09215", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchWindow-09216", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchWindow-09217", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBlockMatchWindow-12421", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageBoxFilterQCOM-06973", 45}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageSampleWeightedQCOM-06971", 50}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageSampleWeightedQCOM-06972", 50}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageSampleWeightedQCOM-06977", 50}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageSampleWeightedQCOM-06978", 50}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageWrite-04469", 37}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageWrite-08795", 37}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpImageWrite-08796", 37}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTensorReadARM-12462", 40}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTypeImage-07027", 36}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTypeImage-07028", 36}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTypeImage-07029", 36}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTypeImage-07030", 36}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-OpTypeTensorARM-09906", 40}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-SampledType-04470", 36}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-SampledType-04471", 36}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-SampledType-04472", 36}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-SampledType-04473", 36}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-addressModeU-12422", 37}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-aspectMask-10673", 35}, { {"If this command is recorded where per-tile execution model is enabled, and if the VkPipeline object bound to the pipeline bind point used by this command writes to a variable of storage class Storage Class TileAttachmentQCOM, the corresponding VkImageView using must not have been created with an aspectMask that contains VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 378}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-02707", 38}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-02712", 38}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-02713", 38}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-04617", 38}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-10746", 38}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-cmdpool", 40}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-parameter", 42}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-commandBuffer-recording", 42}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-cubicRangeClamp-09212", 40}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-dimensionCount-09905", 39}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-filterCubic-02694", 36}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-filterCubicMinmax-02695", 42}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-flags-02696", 30}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-format-07753", 31}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-groupCountX-00386", 36}, { {"groupCountX must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0]", 93}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-groupCountY-00387", 36}, { {"groupCountY must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1]", 93}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-groupCountZ-00388", 36}, { {"groupCountZ must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2]", 93}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-imageLayout-00344", 36}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-index-11450", 30}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-magFilter-04553", 34}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-magFilter-09598", 34}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-micromap-11636", 33}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-micromap-11637", 33}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-micromap-11638", 33}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-micromap-11639", 33}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-mipmapMode-04770", 35}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-mipmapMode-09599", 35}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-pBindInfo-11375", 34}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-pDescription-09900", 37}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-pNext-12463", 30}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-protectedNoFault-11455", 41}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-protectedNoFault-11456", 41}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-reductionMode-09213", 38}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-selectableCubicWeights-09214", 47}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-sparseImageInt64Atomics-04474", 48}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-sparseImageInt64Atomics-04475", 48}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-storageBuffers-06936", 39}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-suspended", 28}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-uniformBuffers-06935", 39}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-unnormalizedCoordinates-09635", 48}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-videocoding", 30}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatch-viewType-07752", 33}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-02691", 33}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-02692", 33}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-02693", 33}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-06479", 33}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-07288", 33}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-07888", 33}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08114", 33}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08115", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08116", 33}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08117", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08119", 33}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08600", 33}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08601", 33}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08605", 33}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08606", 33}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08608", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08609", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08610", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08611", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08612", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-08613", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-09600", 33}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10068", 33}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10672", 33}, { {"If the per-tile execution model is not enabled, this command must be called outside of a render pass instance", 109}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10674", 33}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDispatch feature must be enabled", 98}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10675", 33}, { {"Memory backing image subresources used as tile attachments in the current render pass must not be written in any way other than as a tile attachment by this command", 164}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10676", 33}, { {"If any recorded command in the current subpass will write to an image subresource as a tile attachment, this command must not read from the memory backing that image subresource in any other way than as a tile attachment", 220}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10678", 33}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10679", 33}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-10743", 33}, { {"If there is no bound compute pipeline, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_COMPUTE_BIT stage", 113}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11297", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11298", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11299", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11300", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11301", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11302", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11304", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11305", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11306", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11308", 33}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11309", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11372", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11373", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11374", 33}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11376", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11397", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11398", 33}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11437", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11438", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11439", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11441", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11442", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-11485", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-None-12472", 33}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchSADQCOM-06976", 53}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchSADQCOM-12420", 53}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchSSDQCOM-06974", 53}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchWindow-09215", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchWindow-09216", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchWindow-09217", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBlockMatchWindow-12421", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageBoxFilterQCOM-06973", 49}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageSampleWeightedQCOM-06971", 54}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageSampleWeightedQCOM-06972", 54}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageSampleWeightedQCOM-06977", 54}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageSampleWeightedQCOM-06978", 54}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageWrite-04469", 41}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageWrite-08795", 41}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpImageWrite-08796", 41}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTensorReadARM-12462", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTypeImage-07027", 40}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTypeImage-07028", 40}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTypeImage-07029", 40}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTypeImage-07030", 40}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-OpTypeTensorARM-09906", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-SampledType-04470", 40}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-SampledType-04471", 40}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-SampledType-04472", 40}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-SampledType-04473", 40}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-addressModeU-12422", 41}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-aspectMask-10673", 39}, { {"If this command is recorded where per-tile execution model is enabled, and if the VkPipeline object bound to the pipeline bind point used by this command writes to a variable of storage class Storage Class TileAttachmentQCOM, the corresponding VkImageView using must not have been created with an aspectMask that contains VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 378}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-baseGroupX-00421", 39}, { {"baseGroupX must be less than VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0]", 80}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-baseGroupX-00422", 39}, { {"baseGroupY must be less than VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1]", 80}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-baseGroupX-00427", 39}, { {"If any of baseGroupX, baseGroupY, or baseGroupZ are not zero, then the bound compute pipeline must have been created with the VK_PIPELINE_CREATE_DISPATCH_BASE_BIT flag or the bound compute shader object must have been created with the VK_SHADER_CREATE_DISPATCH_BASE_BIT_EXT flag", 278}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-baseGroupZ-00423", 39}, { {"baseGroupZ must be less than VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2]", 80}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-02707", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-02712", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-02713", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-04617", 42}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-10746", 42}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-cubicRangeClamp-09212", 44}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-dimensionCount-09905", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-filterCubic-02694", 40}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-filterCubicMinmax-02695", 46}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-flags-02696", 34}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-format-07753", 35}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-groupCountX-00424", 40}, { {"groupCountX must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0] minus baseGroupX", 110}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-groupCountY-00425", 40}, { {"groupCountY must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1] minus baseGroupY", 110}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-groupCountZ-00426", 40}, { {"groupCountZ must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2] minus baseGroupZ", 110}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-imageLayout-00344", 40}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-index-11450", 34}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-magFilter-04553", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-magFilter-09598", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-micromap-11636", 37}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-micromap-11637", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-micromap-11638", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-micromap-11639", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-mipmapMode-04770", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-mipmapMode-09599", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-pBindInfo-11375", 38}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-pDescription-09900", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-pNext-12463", 34}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-protectedNoFault-11455", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-protectedNoFault-11456", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-reductionMode-09213", 42}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-selectableCubicWeights-09214", 51}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-sparseImageInt64Atomics-04474", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-sparseImageInt64Atomics-04475", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-storageBuffers-06936", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-uniformBuffers-06935", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-unnormalizedCoordinates-09635", 52}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchBase-viewType-07752", 37}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09797", 41}, { {"For each set n that is statically used by a bound data graph pipeline, a descriptor set must have been bound to n at the same pipeline bind point, with a VkPipelineLayout that is compatible for set n, with the VkPipelineLayout used to create the current VkPipeline, as described in Pipeline Layout Compatibility", 311}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09799", 41}, { {"A valid data graph pipeline must be bound to the VK_PIPELINE_BIND_POINT_DATA_GRAPH_ARM pipeline bind point used by this command", 127}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09935", 41}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid as described by descriptor validity if they are statically used by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 321}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09936", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09937", 41}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically used by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 292}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09938", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-None-09939", 41}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-09800", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound data graph pipelines must not be a protected resource", 173}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-09801", 50}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the bind point used by this command must not be an unprotected resource", 215}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-09941", 50}, { {"All the operations used by the bound data graph pipeline must be supported on the queue family for which the command pool out of which commandBuffer was allocated, as reported by vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM", 231}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_DATA_GRAPH_BIT_ARM operations", 107}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-commonparent", 43}, { {"Both of commandBuffer, and session must have been created, allocated, or retrieved from the same VkDevice", 105}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-dataGraphPipeline-09951", 54}, { {"The VkPipeline bound to the pipeline bind point used by this command must be identical to the dataGraphPipeline used to create session", 134}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-nodeType-09980", 45}, { {"If the VkPipeline bound to the VK_PIPELINE_BIND_POINT_DATA_GRAPH_ARM pipeline bind point was not created with a VkDataGraphPipelineSingleNodeCreateInfoARM structure whose nodeType member is VK_DATA_GRAPH_PIPELINE_NODE_TYPE_OPTICAL_FLOW_ARM included in the pNext chain of VkDataGraphPipelineCreateInfoARM, then a VkDataGraphPipelineOpticalFlowDispatchInfoARM structure must not be included in the pNext chain of pInfo", 416}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-nodeType-09981", 45}, { {"If the VkPipeline bound to the VK_PIPELINE_BIND_POINT_DATA_GRAPH_ARM pipeline bind point was created with a VkDataGraphPipelineSingleNodeCreateInfoARM structure whose nodeType member is VK_DATA_GRAPH_PIPELINE_NODE_TYPE_OPTICAL_FLOW_ARM included in the pNext chain of VkDataGraphPipelineCreateInfoARM, then all image subresources that are accessed via a connection point in VkDataGraphPipelineSingleNodeCreateInfoARM::pConnections must be in the layout specified by the layout member of the VkDataGraphPipelineResourceInfoImageLayoutARM that extends the VkDataGraphPipelineResourceInfoARM structure whose descriptorSet and binding members match the set and binding members of the VkDataGraphPipelineSingleNodeConnectionARM structure for the connection point , unless the unifiedImageLayouts feature is enabled, in which case they may be in the VK_IMAGE_LAYOUT_GENERAL layout", 873}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-pDescription-09930", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with a VkTensorCreateInfoARM::pDescription whose usage member contained VK_TENSOR_USAGE_DATA_GRAPH_BIT_ARM", 255}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-pInfo-09964", 42}, { {"If pInfo is not NULL and its pNext chain includes a VkDataGraphPipelineOpticalFlowDispatchInfoARM structure, then the flags member of that latter structure must not contain VK_DATA_GRAPH_OPTICAL_FLOW_EXECUTE_INPUT_UNCHANGED_BIT_ARM, VK_DATA_GRAPH_OPTICAL_FLOW_EXECUTE_REFERENCE_UNCHANGED_BIT_ARM, VK_DATA_GRAPH_OPTICAL_FLOW_EXECUTE_INPUT_IS_PREVIOUS_REFERENCE_BIT_ARM, or VK_DATA_GRAPH_OPTICAL_FLOW_EXECUTE_REFERENCE_IS_PREVIOUS_INPUT_BIT_ARM if session was not created with VK_DATA_GRAPH_PIPELINE_SESSION_CREATE_OPTICAL_FLOW_CACHE_BIT_ARM", 539}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-pInfo-parameter", 46}, { {"If pInfo is not NULL, pInfo must be a valid pointer to a valid VkDataGraphPipelineDispatchInfoARM structure", 107}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-pNext-09952", 42}, { {"If the VkPipeline bound to the pipeline bind point used by this command was created with a VkDataGraphProcessingEngineCreateInfoARM structure in the pNext chain of VkDataGraphPipelineCreateInfoARM that included a foreign data graph processing engine in its pProcessingEngines member, then all session bound memory must have been allocated with VkExportMemoryAllocateInfo::handleTypes with set bits that are a subset of the bits in VkQueueFamilyDataGraphProcessingEnginePropertiesARM::foreignMemoryHandleTypes structure queried via vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM with a queueFamilyIndex matching the one the command pool used to create commandBuffer was created for and an identical engineType, for all the foreign data graph processing engines that were part of the VkDataGraphProcessingEngineCreateInfoARM used to create the VkPipeline", 874}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-pipeline-09940", 45}, { {"If the VkPipeline bound to the pipeline bind point used by this command was created with a VkDataGraphProcessingEngineCreateInfoARM structure in the pNext chain of VkDataGraphPipelineCreateInfoARM that included a foreign data graph processing engine in its pProcessingEngines member, then all VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptors accessed as a result of this command must be VkTensorARM objects that have been bound to memory allocated with VkExportMemoryAllocateInfo::handleTypes with set bits that are a subset of the bits in VkQueueFamilyDataGraphProcessingEnginePropertiesARM::foreignMemoryHandleTypes structure queried via vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM with a queueFamilyIndex matching the one the command pool used to create commandBuffer was created for and an identical engineType, for all the foreign data graph processing engines that were part of the VkDataGraphProcessingEngineCreateInfoARM used to create the VkPipeline", 976}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-session-09796", 44}, { {"For each of the session bind point requirements returned by vkGetDataGraphPipelineSessionBindPointRequirementsARM for session, VkDataGraphPipelineSessionBindPointRequirementARM::numObjects objects must have been bound to session", 228}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-session-parameter", 48}, { {"session must be a valid VkDataGraphPipelineSessionARM handle", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchDataGraphARM-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCmdDispatchGraphAMDX-NodeMaxPayloadsAMDX-09188", 53}, { {"Execution of this command must not cause any shader that declares NodeMaxPayloadsAMDX to initialize more output payloads than specified by the max number of payloads for that decoration. This requirement applies to each NodeMaxPayloadsAMDX decoration separately", 261}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-02691", 38}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-02692", 38}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-02693", 38}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-06479", 38}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-07288", 38}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-07888", 38}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08114", 38}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08115", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08116", 38}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08117", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08119", 38}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08600", 38}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08601", 38}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08605", 38}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08606", 38}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08608", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08609", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08610", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08611", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08612", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-08613", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-09600", 38}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-10068", 38}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-10195", 38}, { {"If the bound execution graph pipeline includes draw nodes, this command must be called within a render pass instance that is compatible with the graphics pipeline used to create each of those nodes", 197}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-10678", 38}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-10679", 38}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11297", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11298", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11299", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11300", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11301", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11302", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11304", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11305", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11306", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11308", 38}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11309", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11372", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11373", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11374", 38}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11376", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11397", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11398", 38}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11437", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11438", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11439", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11441", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11442", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-11485", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-None-12472", 38}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchSADQCOM-06976", 58}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchSADQCOM-12420", 58}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchSSDQCOM-06974", 58}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchWindow-09215", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchWindow-09216", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchWindow-09217", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBlockMatchWindow-12421", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageBoxFilterQCOM-06973", 54}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageSampleWeightedQCOM-06971", 59}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageSampleWeightedQCOM-06972", 59}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageSampleWeightedQCOM-06977", 59}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageSampleWeightedQCOM-06978", 59}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageWrite-04469", 46}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageWrite-08795", 46}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpImageWrite-08796", 46}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTensorReadARM-12462", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTypeImage-07027", 45}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTypeImage-07028", 45}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTypeImage-07029", 45}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTypeImage-07030", 45}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-OpTypeTensorARM-09906", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-SampledType-04470", 45}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-SampledType-04471", 45}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-SampledType-04472", 45}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-SampledType-04473", 45}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-addressModeU-12422", 46}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-bufferlevel", 39}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-02707", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-09181", 47}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-09182", 47}, { {"commandBuffer must be a primary command buffer", 46}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-10746", 47}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-cubicRangeClamp-09212", 49}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-dimensionCount-09905", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-filterCubic-02694", 45}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-filterCubicMinmax-02695", 51}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-flags-02696", 39}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-format-07753", 40}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-imageLayout-00344", 45}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-index-11450", 39}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-infos-09146", 39}, { {"Host memory locations at indexes in the range [infos, infos + (count*stride)), at a granularity of stride must contain valid VkDispatchGraphInfoAMDX structures in the first 24 bytes", 181}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-magFilter-04553", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-magFilter-09598", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-maxComputeWorkGroupCount-09186", 58}, { {"Execution of this command must not cause a node to be dispatched with a larger number of workgroups than that specified by either a MaxNumWorkgroupsAMDX decoration in the dispatched node or maxComputeWorkGroupCount", 214}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-maxExecutionGraphShaderPayloadCount-09187", 69}, { {"Execution of this command must not cause any shader to initialize more than maxExecutionGraphShaderPayloadCount output payloads", 127}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-micromap-11636", 42}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-micromap-11637", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-micromap-11638", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-micromap-11639", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-mipmapMode-04770", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-mipmapMode-09599", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pBindInfo-11375", 43}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pCountInfo-09145", 44}, { {"pCountInfo->infos must be a host pointer to a memory allocation at least as large as the product of count and stride", 116}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pCountInfo-09147", 44}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, payloads must be a host pointer to a memory allocation at least as large as the product of payloadCount and payloadStride", 186}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pCountInfo-09148", 44}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, nodeIndex must be a valid node index in the bound execution graph pipeline, as returned by vkGetExecutionGraphPipelineNodeIndexAMDX", 196}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pCountInfo-09149", 44}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, host memory locations at indexes in the range [payloads, payloads + (payloadCount * payloadStride)), at a granularity of payloadStride must contain a payload matching the size of the input payload expected by the node in nodeIndex in the first bytes", 314}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pCountInfo-parameter", 48}, { {"pCountInfo must be a valid pointer to a valid VkDispatchGraphCountInfoAMDX structure", 84}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pDescription-09900", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-pNext-12463", 39}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-protectedNoFault-11455", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-protectedNoFault-11456", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-reductionMode-09213", 47}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-scratch-09184", 41}, { {"scratch must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_EXECUTION_GRAPH_SCRATCH_BIT_AMDX or VK_BUFFER_USAGE_2_EXECUTION_GRAPH_SCRATCH_BIT_AMDX usage flags set", 186}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-scratch-10192", 41}, { {"scratch must be the device address of an allocated memory range at least as large as scratchSize", 96}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-scratch-10194", 41}, { {"The device memory range [scratch,scratch scratchSize] must have been initialized with vkCmdInitializeGraphScratchMemoryAMDX using the bound execution graph pipeline, and not modified after that by anything other than another execution graph dispatch command", 257}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-scratch-parameter", 45}, { {"scratch must be a valid VkDeviceAddress value", 45}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-scratchSize-10193", 45}, { {"scratchSize must be greater than or equal to VkExecutionGraphPipelineScratchSizeAMDX::minSize returned by vkGetExecutionGraphPipelineScratchSizeAMDX for the bound execution graph pipeline", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-selectableCubicWeights-09214", 56}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-sparseImageInt64Atomics-04474", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-sparseImageInt64Atomics-04475", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-storageBuffers-06936", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-uniformBuffers-06935", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-unnormalizedCoordinates-09635", 57}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphAMDX-viewType-07752", 42}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-NodeMaxPayloadsAMDX-09188", 61}, { {"Execution of this command must not cause any shader that declares NodeMaxPayloadsAMDX to initialize more output payloads than specified by the max number of payloads for that decoration. This requirement applies to each NodeMaxPayloadsAMDX decoration separately", 261}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-02691", 46}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-02692", 46}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-02693", 46}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-06479", 46}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-07288", 46}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-07888", 46}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08114", 46}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08115", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08116", 46}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08117", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08119", 46}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08600", 46}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08601", 46}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08605", 46}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08606", 46}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08608", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08609", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08610", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08611", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08612", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-08613", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-09600", 46}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-10068", 46}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-10195", 46}, { {"If the bound execution graph pipeline includes draw nodes, this command must be called within a render pass instance that is compatible with the graphics pipeline used to create each of those nodes", 197}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-10678", 46}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-10679", 46}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11297", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11298", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11299", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11300", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11301", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11302", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11304", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11305", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11306", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11308", 46}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11309", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11372", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11373", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11374", 46}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11376", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11397", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11398", 46}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11437", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11438", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11439", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11441", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11442", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-11485", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-None-12472", 46}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchSADQCOM-06976", 66}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchSADQCOM-12420", 66}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchSSDQCOM-06974", 66}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchWindow-09215", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchWindow-09216", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchWindow-09217", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBlockMatchWindow-12421", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageBoxFilterQCOM-06973", 62}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageSampleWeightedQCOM-06971", 67}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageSampleWeightedQCOM-06972", 67}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageSampleWeightedQCOM-06977", 67}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageSampleWeightedQCOM-06978", 67}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageWrite-04469", 54}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageWrite-08795", 54}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpImageWrite-08796", 54}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTensorReadARM-12462", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTypeImage-07027", 53}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTypeImage-07028", 53}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTypeImage-07029", 53}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTypeImage-07030", 53}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-OpTypeTensorARM-09906", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-SampledType-04470", 53}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-SampledType-04471", 53}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-SampledType-04472", 53}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-SampledType-04473", 53}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-addressModeU-12422", 54}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-bufferlevel", 47}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-02707", 55}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-09181", 55}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-09182", 55}, { {"commandBuffer must be a primary command buffer", 46}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-10746", 55}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-cubicRangeClamp-09212", 57}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-dimensionCount-09905", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-filterCubic-02694", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-filterCubicMinmax-02695", 59}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-flags-02696", 47}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-format-07753", 48}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-imageLayout-00344", 53}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-index-11450", 47}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-infos-09153", 47}, { {"Device memory locations at indexes in the range [infos, infos + (count*stride)), at a granularity of stride must contain valid VkDispatchGraphInfoAMDX structures in the first 24 bytes when this command is executed on the device", 227}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-magFilter-04553", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-magFilter-09598", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-maxComputeWorkGroupCount-09186", 66}, { {"Execution of this command must not cause a node to be dispatched with a larger number of workgroups than that specified by either a MaxNumWorkgroupsAMDX decoration in the dispatched node or maxComputeWorkGroupCount", 214}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-maxExecutionGraphShaderPayloadCount-09187", 77}, { {"Execution of this command must not cause any shader to initialize more than maxExecutionGraphShaderPayloadCount output payloads", 127}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-micromap-11636", 50}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-micromap-11637", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-micromap-11638", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-micromap-11639", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-mipmapMode-04770", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-mipmapMode-09599", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pBindInfo-11375", 51}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09150", 52}, { {"pCountInfo->infos must be a device pointer to a memory allocation at least as large as the product of count and stride when this command is executed on the device", 162}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09151", 52}, { {"pCountInfo->infos must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 128}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09152", 52}, { {"pCountInfo->infos must be a multiple of executionGraphDispatchAddressAlignment", 78}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09154", 52}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, payloads must be a device pointer to a memory allocation at least as large as the product of payloadCount and payloadStride when this command is executed on the device", 232}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09155", 52}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, payloads must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 184}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09156", 52}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, payloads must be a multiple of executionGraphDispatchAddressAlignment", 134}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09157", 52}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, nodeIndex must be a valid node index in the bound execution graph pipeline, as returned by vkGetExecutionGraphPipelineNodeIndexAMDX when this command is executed on the device", 240}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-09158", 52}, { {"For each VkDispatchGraphInfoAMDX structure in pCountInfo->infos, device memory locations at indexes in the range [payloads, payloads + (payloadCount * payloadStride)), at a granularity of payloadStride must contain a payload matching the size of the input payload expected by the node in nodeIndex in the first bytes when this command is executed on the device", 360}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pCountInfo-parameter", 56}, { {"pCountInfo must be a valid pointer to a valid VkDispatchGraphCountInfoAMDX structure", 84}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pDescription-09900", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-pNext-12463", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-protectedNoFault-11455", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-protectedNoFault-11456", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-reductionMode-09213", 55}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-scratch-09184", 49}, { {"scratch must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_EXECUTION_GRAPH_SCRATCH_BIT_AMDX or VK_BUFFER_USAGE_2_EXECUTION_GRAPH_SCRATCH_BIT_AMDX usage flags set", 186}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-scratch-10192", 49}, { {"scratch must be the device address of an allocated memory range at least as large as scratchSize", 96}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-scratch-10194", 49}, { {"The device memory range [scratch,scratch scratchSize] must have been initialized with vkCmdInitializeGraphScratchMemoryAMDX using the bound execution graph pipeline, and not modified after that by anything other than another execution graph dispatch command", 257}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-scratch-parameter", 53}, { {"scratch must be a valid VkDeviceAddress value", 45}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-scratchSize-10193", 53}, { {"scratchSize must be greater than or equal to VkExecutionGraphPipelineScratchSizeAMDX::minSize returned by vkGetExecutionGraphPipelineScratchSizeAMDX for the bound execution graph pipeline", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-selectableCubicWeights-09214", 64}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-sparseImageInt64Atomics-04474", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-sparseImageInt64Atomics-04475", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-storageBuffers-06936", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-suspended", 45}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-uniformBuffers-06935", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-unnormalizedCoordinates-09635", 65}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectAMDX-viewType-07752", 50}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-NodeMaxPayloadsAMDX-09188", 66}, { {"Execution of this command must not cause any shader that declares NodeMaxPayloadsAMDX to initialize more output payloads than specified by the max number of payloads for that decoration. This requirement applies to each NodeMaxPayloadsAMDX decoration separately", 261}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-02691", 51}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-02692", 51}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-02693", 51}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-06479", 51}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-07288", 51}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-07888", 51}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08114", 51}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08115", 51}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08116", 51}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08117", 51}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08119", 51}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08600", 51}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08601", 51}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08605", 51}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08606", 51}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08608", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08609", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08610", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08611", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08612", 51}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-08613", 51}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-09600", 51}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-10068", 51}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-10195", 51}, { {"If the bound execution graph pipeline includes draw nodes, this command must be called within a render pass instance that is compatible with the graphics pipeline used to create each of those nodes", 197}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-10678", 51}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-10679", 51}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11297", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11298", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11299", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11300", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11301", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11302", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11304", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11305", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11306", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11308", 51}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11309", 51}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11372", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11373", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11374", 51}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11376", 51}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11397", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11398", 51}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11437", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11438", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11439", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11441", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11442", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-11485", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-None-12472", 51}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchSADQCOM-06976", 71}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchSADQCOM-12420", 71}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchSSDQCOM-06974", 71}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchWindow-09215", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchWindow-09216", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchWindow-09217", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBlockMatchWindow-12421", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageBoxFilterQCOM-06973", 67}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageSampleWeightedQCOM-06971", 72}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageSampleWeightedQCOM-06972", 72}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageSampleWeightedQCOM-06977", 72}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageSampleWeightedQCOM-06978", 72}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageWrite-04469", 59}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageWrite-08795", 59}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpImageWrite-08796", 59}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTensorReadARM-12462", 62}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTypeImage-07027", 58}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTypeImage-07028", 58}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTypeImage-07029", 58}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTypeImage-07030", 58}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-OpTypeTensorARM-09906", 62}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-SampledType-04470", 58}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-SampledType-04471", 58}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-SampledType-04472", 58}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-SampledType-04473", 58}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-addressModeU-12422", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-bufferlevel", 52}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-02707", 60}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-09181", 60}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-09182", 60}, { {"commandBuffer must be a primary command buffer", 46}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-10746", 60}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09159", 56}, { {"countInfo must be a device pointer to a memory allocation containing a valid VkDispatchGraphCountInfoAMDX structure when this command is executed on the device", 159}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09160", 56}, { {"countInfo must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 120}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09161", 56}, { {"countInfo must be a multiple of executionGraphDispatchAddressAlignment", 70}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09162", 56}, { {"countInfo->infos must be a device pointer to a memory allocation at least as large as the product of count and stride when this command is executed on the device", 161}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09163", 56}, { {"countInfo->infos must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 127}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09164", 56}, { {"countInfo->infos must be a multiple of executionGraphDispatchAddressAlignment", 77}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09166", 56}, { {"For each VkDispatchGraphInfoAMDX structure in countInfo->infos, payloads must be a device pointer to a memory allocation at least as large as the product of payloadCount and payloadStride when this command is executed on the device", 231}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09167", 56}, { {"For each VkDispatchGraphInfoAMDX structure in countInfo->infos, payloads must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 183}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09168", 56}, { {"For each VkDispatchGraphInfoAMDX structure in countInfo->infos, payloads must be a multiple of executionGraphDispatchAddressAlignment", 133}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09169", 56}, { {"For each VkDispatchGraphInfoAMDX structure in countInfo->infos, nodeIndex must be a valid node index in the bound execution graph pipeline, as returned by vkGetExecutionGraphPipelineNodeIndexAMDX when this command is executed on the device", 239}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-09170", 56}, { {"For each VkDispatchGraphInfoAMDX structure in countInfo->infos, device memory locations at indexes in the range [payloads, payloads + (payloadCount * payloadStride)), at a granularity of payloadStride must contain a payload matching the size of the input payload expected by the node in nodeIndex in the first bytes when this command is executed on the device", 359}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-countInfo-parameter", 60}, { {"countInfo must be a valid VkDeviceAddress value", 47}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-cubicRangeClamp-09212", 62}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-dimensionCount-09905", 61}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-filterCubic-02694", 58}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-filterCubicMinmax-02695", 64}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-flags-02696", 52}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-format-07753", 53}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-imageLayout-00344", 58}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-index-11450", 52}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-infos-09165", 52}, { {"Device memory locations at indexes in the range [infos, infos + (count*stride)), at a granularity of stride must contain valid VkDispatchGraphInfoAMDX structures in the first 24 bytes when this command is executed on the device", 227}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-magFilter-04553", 56}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-magFilter-09598", 56}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-maxComputeWorkGroupCount-09186", 71}, { {"Execution of this command must not cause a node to be dispatched with a larger number of workgroups than that specified by either a MaxNumWorkgroupsAMDX decoration in the dispatched node or maxComputeWorkGroupCount", 214}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-maxExecutionGraphShaderPayloadCount-09187", 82}, { {"Execution of this command must not cause any shader to initialize more than maxExecutionGraphShaderPayloadCount output payloads", 127}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-micromap-11636", 55}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-micromap-11637", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-micromap-11638", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-micromap-11639", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-mipmapMode-04770", 57}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-mipmapMode-09599", 57}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-pBindInfo-11375", 56}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-pDescription-09900", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-pNext-12463", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-protectedNoFault-11455", 63}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-protectedNoFault-11456", 63}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-reductionMode-09213", 60}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-scratch-09184", 54}, { {"scratch must be a device address within a VkBuffer created with the VK_BUFFER_USAGE_EXECUTION_GRAPH_SCRATCH_BIT_AMDX or VK_BUFFER_USAGE_2_EXECUTION_GRAPH_SCRATCH_BIT_AMDX usage flags set", 186}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-scratch-10192", 54}, { {"scratch must be the device address of an allocated memory range at least as large as scratchSize", 96}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-scratch-10194", 54}, { {"The device memory range [scratch,scratch scratchSize] must have been initialized with vkCmdInitializeGraphScratchMemoryAMDX using the bound execution graph pipeline, and not modified after that by anything other than another execution graph dispatch command", 257}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-scratch-parameter", 58}, { {"scratch must be a valid VkDeviceAddress value", 45}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-scratchSize-10193", 58}, { {"scratchSize must be greater than or equal to VkExecutionGraphPipelineScratchSizeAMDX::minSize returned by vkGetExecutionGraphPipelineScratchSizeAMDX for the bound execution graph pipeline", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-selectableCubicWeights-09214", 69}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-sparseImageInt64Atomics-04474", 70}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-sparseImageInt64Atomics-04475", 70}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-storageBuffers-06936", 61}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-suspended", 50}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-uniformBuffers-06935", 61}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-unnormalizedCoordinates-09635", 70}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchGraphIndirectCountAMDX-viewType-07752", 55}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdDispatchIndirect-None-02691", 37}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-02692", 37}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-02693", 37}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-06479", 37}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-07288", 37}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-07888", 37}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08114", 37}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08115", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08116", 37}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08117", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08119", 37}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08600", 37}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08601", 37}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08605", 37}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08606", 37}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08608", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08609", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08610", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08611", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08612", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-08613", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-09600", 37}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10068", 37}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10672", 37}, { {"If the per-tile execution model is not enabled, this command must be called outside of a render pass instance", 109}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10674", 37}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDispatch feature must be enabled", 98}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10675", 37}, { {"Memory backing image subresources used as tile attachments in the current render pass must not be written in any way other than as a tile attachment by this command", 164}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10676", 37}, { {"If any recorded command in the current subpass will write to an image subresource as a tile attachment, this command must not read from the memory backing that image subresource in any other way than as a tile attachment", 220}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10678", 37}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10679", 37}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-10743", 37}, { {"If there is no bound compute pipeline, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_COMPUTE_BIT stage", 113}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11297", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11298", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11299", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11300", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11301", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11302", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11304", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11305", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11306", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11308", 37}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11309", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11372", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11373", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11374", 37}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11376", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11397", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11398", 37}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11437", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11438", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11439", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11441", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11442", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-11485", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-None-12472", 37}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchSADQCOM-06976", 57}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchSADQCOM-12420", 57}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchSSDQCOM-06974", 57}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchWindow-09215", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchWindow-09216", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchWindow-09217", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBlockMatchWindow-12421", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageBoxFilterQCOM-06973", 53}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageSampleWeightedQCOM-06971", 58}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageSampleWeightedQCOM-06972", 58}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageSampleWeightedQCOM-06977", 58}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageSampleWeightedQCOM-06978", 58}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageWrite-04469", 45}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageWrite-08795", 45}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpImageWrite-08796", 45}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTensorReadARM-12462", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTypeImage-07027", 44}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTypeImage-07028", 44}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTypeImage-07029", 44}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTypeImage-07030", 44}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-OpTypeTensorARM-09906", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-SampledType-04470", 44}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-SampledType-04471", 44}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-SampledType-04472", 44}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-SampledType-04473", 44}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-addressModeU-12422", 45}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-aspectMask-10673", 43}, { {"If this command is recorded where per-tile execution model is enabled, and if the VkPipeline object bound to the pipeline bind point used by this command writes to a variable of storage class Storage Class TileAttachmentQCOM, the corresponding VkImageView using must not have been created with an aspectMask that contains VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 378}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-buffer-02708", 39}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-buffer-02709", 39}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-buffer-parameter", 43}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-02707", 46}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-02711", 46}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-10746", 46}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-commonparent", 39}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-cubicRangeClamp-09212", 48}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-dimensionCount-09905", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-filterCubic-02694", 44}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-filterCubicMinmax-02695", 50}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-flags-02696", 38}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-format-07753", 39}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-imageLayout-00344", 44}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-index-11450", 38}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-magFilter-04553", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-magFilter-09598", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-micromap-11636", 41}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-micromap-11637", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-micromap-11638", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-micromap-11639", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-mipmapMode-04770", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-mipmapMode-09599", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-offset-00407", 39}, { {"The sum of offset and the size of VkDispatchIndirectCommand must be less than or equal to the size of buffer", 108}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-offset-02710", 39}, { {"offset must be a multiple of 4", 30}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-pBindInfo-11375", 42}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-pDescription-09900", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-pNext-12463", 38}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-protectedNoFault-11455", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-protectedNoFault-11456", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-reductionMode-09213", 46}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-selectableCubicWeights-09214", 55}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-sparseImageInt64Atomics-04474", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-sparseImageInt64Atomics-04475", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-storageBuffers-06936", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-uniformBuffers-06935", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-unnormalizedCoordinates-09635", 56}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect-viewType-07752", 41}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-02691", 41}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-02692", 41}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-02693", 41}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-06479", 41}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-07288", 41}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-07888", 41}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08114", 41}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08115", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08116", 41}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08117", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08119", 41}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08600", 41}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08601", 41}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08605", 41}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08606", 41}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08608", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08609", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08610", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08611", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08612", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-08613", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-09600", 41}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-10068", 41}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-10678", 41}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-10679", 41}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11297", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11298", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11299", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11300", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11301", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11302", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11304", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11305", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11306", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11308", 41}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11309", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11372", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11373", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11374", 41}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11376", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11397", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11398", 41}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11437", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11438", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11439", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11441", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11442", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-11485", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-None-12472", 41}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchSADQCOM-06976", 61}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchSADQCOM-12420", 61}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchSSDQCOM-06974", 61}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchWindow-09215", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchWindow-09216", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchWindow-09217", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBlockMatchWindow-12421", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageBoxFilterQCOM-06973", 57}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageSampleWeightedQCOM-06971", 62}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageSampleWeightedQCOM-06972", 62}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageSampleWeightedQCOM-06977", 62}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageSampleWeightedQCOM-06978", 62}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageWrite-04469", 49}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageWrite-08795", 49}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpImageWrite-08796", 49}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTensorReadARM-12462", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTypeImage-07027", 48}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTypeImage-07028", 48}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTypeImage-07029", 48}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTypeImage-07030", 48}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-OpTypeTensorARM-09906", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-SampledType-04470", 48}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-SampledType-04471", 48}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-SampledType-04472", 48}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-SampledType-04473", 48}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-addressModeU-12422", 49}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-02707", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-10746", 50}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-13049", 50}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-cubicRangeClamp-09212", 52}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-dimensionCount-09905", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-filterCubic-02694", 48}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-filterCubicMinmax-02695", 54}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-flags-02696", 42}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-format-07753", 43}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-imageLayout-00344", 48}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-index-11450", 42}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-magFilter-04553", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-magFilter-09598", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-micromap-11636", 45}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-micromap-11637", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-micromap-11638", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-micromap-11639", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-mipmapMode-04770", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-mipmapMode-09599", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-pBindInfo-11375", 46}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-pDescription-09900", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-pInfo-13050", 42}, { {"pInfo->addressRange.size must be greater than or equal to sizeof(VkDispatchIndirectCommand)", 91}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-pInfo-parameter", 46}, { {"pInfo must be a valid pointer to a valid VkDispatchIndirect2InfoKHR structure", 77}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-pNext-12463", 42}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-protectedNoFault-11455", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-protectedNoFault-11456", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-reductionMode-09213", 50}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-selectableCubicWeights-09214", 59}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-sparseImageInt64Atomics-04474", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-sparseImageInt64Atomics-04475", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-storageBuffers-06936", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-uniformBuffers-06935", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-unnormalizedCoordinates-09635", 60}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchIndirect2KHR-viewType-07752", 45}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-02691", 37}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-02692", 37}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-02693", 37}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-06479", 37}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-07288", 37}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-07888", 37}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08114", 37}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08115", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08116", 37}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08117", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08119", 37}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08600", 37}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08601", 37}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08605", 37}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08606", 37}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08608", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08609", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08610", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08611", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08612", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-08613", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-09600", 37}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10068", 37}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10668", 37}, { {"When this command is recorded per-tile execution model must be enabled", 70}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10669", 37}, { {"The tileShadingDispatchTile must enabled", 40}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10672", 37}, { {"If the per-tile execution model is not enabled, this command must be called outside of a render pass instance", 109}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10674", 37}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDispatch feature must be enabled", 98}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10675", 37}, { {"Memory backing image subresources used as tile attachments in the current render pass must not be written in any way other than as a tile attachment by this command", 164}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10676", 37}, { {"If any recorded command in the current subpass will write to an image subresource as a tile attachment, this command must not read from the memory backing that image subresource in any other way than as a tile attachment", 220}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10678", 37}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10679", 37}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-10743", 37}, { {"If there is no bound compute pipeline, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_COMPUTE_BIT stage", 113}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11297", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11298", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11299", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11300", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11301", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11302", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11304", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11305", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11306", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11308", 37}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11309", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11372", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11373", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11374", 37}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11376", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11397", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11398", 37}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11437", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11438", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11439", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11441", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11442", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-11485", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-None-12472", 37}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchSADQCOM-06976", 57}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchSADQCOM-12420", 57}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchSSDQCOM-06974", 57}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchWindow-09215", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchWindow-09216", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchWindow-09217", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBlockMatchWindow-12421", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageBoxFilterQCOM-06973", 53}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageSampleWeightedQCOM-06971", 58}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageSampleWeightedQCOM-06972", 58}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageSampleWeightedQCOM-06977", 58}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageSampleWeightedQCOM-06978", 58}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageWrite-04469", 45}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageWrite-08795", 45}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpImageWrite-08796", 45}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTensorReadARM-12462", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTypeImage-07027", 44}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTypeImage-07028", 44}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTypeImage-07029", 44}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTypeImage-07030", 44}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-OpTypeTensorARM-09906", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-SampledType-04470", 44}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-SampledType-04471", 44}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-SampledType-04472", 44}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-SampledType-04473", 44}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-addressModeU-12422", 45}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-aspectMask-10673", 43}, { {"If this command is recorded where per-tile execution model is enabled, and if the VkPipeline object bound to the pipeline bind point used by this command writes to a variable of storage class Storage Class TileAttachmentQCOM, the corresponding VkImageView using must not have been created with an aspectMask that contains VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 378}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-02707", 46}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-02712", 46}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-02713", 46}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-04617", 46}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-10746", 46}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-cubicRangeClamp-09212", 48}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-dimensionCount-09905", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-filterCubic-02694", 44}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-filterCubicMinmax-02695", 50}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-flags-02696", 38}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-format-07753", 39}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-imageLayout-00344", 44}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-index-11450", 38}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-magFilter-04553", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-magFilter-09598", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-micromap-11636", 41}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-micromap-11637", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-micromap-11638", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-micromap-11639", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-mipmapMode-04770", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-mipmapMode-09599", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-pBindInfo-11375", 42}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-pDescription-09900", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-pDispatchTileInfo-parameter", 54}, { {"pDispatchTileInfo must be a valid pointer to a valid VkDispatchTileInfoQCOM structure", 85}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-pNext-12463", 38}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-protectedNoFault-11455", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-protectedNoFault-11456", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-reductionMode-09213", 46}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-renderpass", 37}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-selectableCubicWeights-09214", 55}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-sparseImageInt64Atomics-04474", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-sparseImageInt64Atomics-04475", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-storageBuffers-06936", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-uniformBuffers-06935", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-unnormalizedCoordinates-09635", 56}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDispatchTileQCOM-viewType-07752", 41}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdDraw-Input-07939", 26}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-Input-08734", 26}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-InputAttachmentIndex-09597", 41}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-02691", 25}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-02692", 25}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-02693", 25}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-02721", 25}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04007", 25}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04008", 25}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04875", 25}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04876", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04877", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04879", 25}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-04914", 25}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-06479", 25}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-06537", 25}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-06666", 25}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-06886", 25}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-06887", 25}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07288", 25}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07469", 25}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07619", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07620", 25}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07621", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07622", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07623", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07624", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07625", 25}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07626", 25}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07627", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07629", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07630", 25}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07631", 25}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07632", 25}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07633", 25}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07634", 25}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07636", 25}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07639", 25}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07640", 25}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07641", 25}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07642", 25}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07643", 25}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07644", 25}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07645", 25}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07646", 25}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07647", 25}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07648", 25}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07649", 25}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07749", 25}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07751", 25}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07831", 25}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07832", 25}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07834", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07835", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07836", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07837", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07838", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07839", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07840", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07841", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07842", 25}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07843", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07844", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07845", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07846", 25}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07847", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07848", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07849", 25}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07850", 25}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07878", 25}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07879", 25}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07880", 25}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07881", 25}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-07888", 25}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08114", 25}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08115", 25}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08116", 25}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08117", 25}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08119", 25}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08600", 25}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08601", 25}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08605", 25}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08606", 25}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08608", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08609", 25}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08610", 25}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08611", 25}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08612", 25}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08613", 25}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08617", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08636", 25}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08637", 25}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08644", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08666", 25}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08669", 25}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08684", 25}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08685", 25}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08686", 25}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08687", 25}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08688", 25}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08689", 25}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08690", 25}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08693", 25}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08696", 25}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08698", 25}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08699", 25}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08876", 25}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08877", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08878", 25}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08879", 25}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-08885", 25}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09003", 25}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09116", 25}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09203", 25}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09211", 25}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09363", 25}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09364", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09365", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09368", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09369", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09462", 25}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09548", 25}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09549", 25}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09600", 25}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09637", 25}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09642", 25}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09643", 25}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-09650", 25}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10068", 25}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10608", 25}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10677", 25}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10678", 25}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10679", 25}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10772", 25}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10795", 25}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10796", 25}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10797", 25}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10862", 25}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10864", 25}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10909", 25}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10927", 25}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-10928", 25}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11297", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11298", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11299", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11300", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11301", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11302", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11304", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11305", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11306", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11308", 25}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11309", 25}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11372", 25}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11373", 25}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11374", 25}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11376", 25}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11397", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11398", 25}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11437", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11438", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11439", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11441", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11442", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11485", 25}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11522", 25}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11523", 25}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11861", 25}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11862", 25}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11863", 25}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11865", 25}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11866", 25}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11868", 25}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-11869", 25}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-12338", 25}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-12339", 25}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-12340", 25}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-12472", 25}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-None-13118", 25}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpExecutionMode-12239", 36}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpExecutionMode-12240", 36}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpExecutionMode-12241", 36}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpExecutionMode-12242", 36}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchSADQCOM-06976", 45}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchSADQCOM-12420", 45}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchSSDQCOM-06974", 45}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchWindow-09215", 44}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchWindow-09216", 44}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchWindow-09217", 44}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBlockMatchWindow-12421", 44}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageBoxFilterQCOM-06973", 41}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageSampleWeightedQCOM-06971", 46}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageSampleWeightedQCOM-06972", 46}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageSampleWeightedQCOM-06977", 46}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageSampleWeightedQCOM-06978", 46}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageWrite-04469", 33}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageWrite-08795", 33}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpImageWrite-08796", 33}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTensorReadARM-12462", 36}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeImage-07027", 32}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeImage-07028", 32}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeImage-07029", 32}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeImage-07030", 32}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeImage-07468", 32}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-OpTypeTensorARM-09906", 36}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-SampledType-04470", 32}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-SampledType-04471", 32}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-SampledType-04472", 32}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-SampledType-04473", 32}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-VkPipelineVieportCreateInfo-04141", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-VkPipelineVieportCreateInfo-04142", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-addressModeU-12422", 33}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-advancedBlendMaxColorAttachments-07480", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-alphaToCoverageEnable-08919", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-alphaToCoverageEnable-08920", 42}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-attachmentCount-07750", 36}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-blendEnable-04727", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-colorAttachmentCount-06179", 41}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-colorAttachmentCount-06185", 41}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-colorAttachmentCount-09362", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-colorAttachmentCount-09372", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-colorAttachmentCount-11871", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-02707", 34}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-02712", 34}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-02713", 34}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-04617", 34}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-10746", 34}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-cmdpool", 36}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-parameter", 38}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-commandBuffer-recording", 38}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-conservativePointAndLineRasterization-07499", 58}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-coverageModulationTableEnable-07488", 50}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-coverageReductionMode-07491", 42}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-coverageToColorEnable-07490", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-cubicRangeClamp-09212", 36}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-customResolve-11524", 34}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-customResolve-11525", 34}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-customResolve-11529", 34}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-customResolve-11530", 34}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dimensionCount-09905", 35}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicPrimitiveTopologyUnrestricted-07500", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingLocalRead-11797", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08910", 54}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08911", 54}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08912", 54}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08913", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08914", 54}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08915", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08916", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08917", 54}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-08918", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-11864", 54}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-11867", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-dynamicRenderingUnusedAttachments-11870", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-filterCubic-02694", 32}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-filterCubicMinmax-02695", 38}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-firstAttachment-07476", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-firstAttachment-07478", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-02696", 26}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-10582", 26}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-11521", 26}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-13361", 26}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-13362", 26}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-13363", 26}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-13364", 26}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-flags-13365", 26}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-format-07753", 27}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-format-08936", 27}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-format-08937", 27}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-format-10389", 27}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-format-10390", 27}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageLayout-00344", 32}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06172", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06173", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06174", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06175", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06176", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06177", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06183", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-imageView-06184", 30}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-index-11450", 26}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-layers-10831", 27}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-logicOp-04878", 28}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-magFilter-04553", 30}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-magFilter-09598", 30}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-maxFragmentDualSrcAttachments-09239", 50}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-maxMultiviewInstanceIndex-02688", 46}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-micromap-11636", 29}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-micromap-11637", 29}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-micromap-11638", 29}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-micromap-11639", 29}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-mipmapMode-04770", 31}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-mipmapMode-09599", 31}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-multisampledRenderToSingleSampled-07285", 54}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-multisampledRenderToSingleSampled-07286", 54}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-multisampledRenderToSingleSampled-07287", 54}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-multiviewPerViewViewports-12262", 46}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-multiviewPerViewViewports-12263", 46}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-nextStage-10745", 30}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pBindInfo-11375", 30}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pColorAttachments-08963", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pColorAttachments-11539", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthAttachment-06186", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthAttachment-08964", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthAttachment-11540", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthAttachment-11872", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthInputAttachmentIndex-09595", 47}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDepthInputAttachmentIndex-09596", 47}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDescription-09900", 33}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDynamicStates-08715", 35}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pDynamicStates-08716", 35}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pNext-07935", 26}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pNext-09461", 26}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pNext-12463", 26}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pStencilAttachment-06187", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pStencilAttachment-08965", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pStencilAttachment-11860", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pStencilAttachment-11873", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pStrides-04913", 29}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-pipelineFragmentShadingRate-09238", 48}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitiveFragmentShadingRateWithMultipleViewports-04552", 70}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitiveFragmentShadingRateWithMultipleViewports-08642", 70}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitiveTopology-10286", 38}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitiveTopology-10747", 38}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitiveTopology-10748", 38}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitivesGeneratedQueryWithNonZeroStreams-06709", 63}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitivesGeneratedQueryWithNonZeroStreams-07481", 63}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-primitivesGeneratedQueryWithRasterizerDiscard-06708", 66}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-protectedNoFault-11455", 37}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-protectedNoFault-11456", 37}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizationSamples-07471", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizationSamples-07489", 41}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizationSamples-07494", 41}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizerDiscardEnable-09236", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizerDiscardEnable-09420", 44}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-rasterizerDiscardEnable-10863", 44}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-reductionMode-09213", 34}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-renderPass-02684", 31}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-renderPass-06198", 31}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-renderpass", 25}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-02689", 42}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07484", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07485", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07486", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07487", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07936", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07937", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsEnable-07938", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsPerPixel-07482", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sampleLocationsPerPixel-07483", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-samples-07472", 28}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-samples-07473", 28}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-scissorCount-03418", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-selectableCubicWeights-09214", 43}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-shadingRateImage-09233", 37}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-shadingRateImage-09234", 37}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sparseImageInt64Atomics-04474", 44}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-sparseImageInt64Atomics-04475", 44}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stage-06481", 26}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stage-07073", 26}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stippledLineEnable-07495", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stippledLineEnable-07496", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stippledLineEnable-07497", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-stippledLineEnable-07498", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-storageBuffers-06936", 35}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-subpass-02685", 28}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-suspended", 24}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-uniformBuffers-06935", 35}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-unnormalizedCoordinates-09635", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-videocoding", 26}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewMask-06178", 29}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewType-07752", 29}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-03417", 34}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-03419", 34}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-04137", 34}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-04138", 34}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-04139", 34}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-07492", 34}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-07493", 34}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDraw-viewportCount-09421", 34}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-ClusterCullingHUAWEI-07823", 54}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the ClusterCullingHUAWEI Execution Model", 136}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-InputAttachmentIndex-09597", 54}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-02691", 38}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-02692", 38}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-02693", 38}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-04876", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-04877", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-06479", 38}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-06537", 38}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-06666", 38}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-06886", 38}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-06887", 38}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07074", 38}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07075", 38}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07288", 38}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07469", 38}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07620", 38}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07621", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07622", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07623", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07624", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07625", 38}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07626", 38}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07627", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07629", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07630", 38}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07631", 38}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07632", 38}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07633", 38}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07634", 38}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07636", 38}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07639", 38}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07640", 38}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07641", 38}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07642", 38}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07643", 38}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07644", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07645", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07646", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07647", 38}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07648", 38}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07649", 38}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07749", 38}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07751", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07819", 38}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_TASK_SHADER_INVOCATIONS_BIT_EXT, or VK_QUERY_PIPELINE_STATISTIC_MESH_SHADER_INVOCATIONS_BIT_EXT", 222}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07831", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07832", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07834", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07835", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07836", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07837", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07838", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07839", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07840", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07841", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07843", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07844", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07845", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07846", 38}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07847", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07848", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07849", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07850", 38}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07878", 38}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07879", 38}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07880", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07881", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-07888", 38}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08114", 38}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08115", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08116", 38}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08117", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08119", 38}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08600", 38}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08601", 38}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08605", 38}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08606", 38}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08608", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08609", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08610", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08611", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08612", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08613", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08617", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08636", 38}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08637", 38}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08644", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08666", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08669", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08684", 38}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08685", 38}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08686", 38}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08687", 38}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08688", 38}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08689", 38}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08690", 38}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08693", 38}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08694", 38}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08695", 38}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08696", 38}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08698", 38}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08699", 38}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08876", 38}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08877", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08878", 38}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-08879", 38}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09003", 38}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09116", 38}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09211", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09363", 38}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09364", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09365", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09368", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09369", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09548", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09549", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09600", 38}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09642", 38}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09643", 38}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-09650", 38}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10068", 38}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10608", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10677", 38}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10678", 38}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10679", 38}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10772", 38}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10795", 38}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10796", 38}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10797", 38}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10862", 38}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10864", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10927", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-10928", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11297", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11298", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11299", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11300", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11301", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11302", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11304", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11305", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11306", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11308", 38}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11309", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11372", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11373", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11374", 38}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11376", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11397", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11398", 38}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11437", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11438", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11439", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11441", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11442", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11485", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11522", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11523", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11861", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11862", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11863", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11865", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11866", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11868", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-11869", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-12338", 38}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-12339", 38}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-12340", 38}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-None-12472", 38}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchSADQCOM-06976", 58}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchSADQCOM-12420", 58}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchSSDQCOM-06974", 58}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchWindow-09215", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchWindow-09216", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchWindow-09217", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBlockMatchWindow-12421", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageBoxFilterQCOM-06973", 54}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageSampleWeightedQCOM-06971", 59}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageSampleWeightedQCOM-06972", 59}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageSampleWeightedQCOM-06977", 59}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageSampleWeightedQCOM-06978", 59}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageWrite-04469", 46}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageWrite-08795", 46}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpImageWrite-08796", 46}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTensorReadARM-12462", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeImage-07027", 45}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeImage-07028", 45}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeImage-07029", 45}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeImage-07030", 45}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeImage-07468", 45}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-OpTypeTensorARM-09906", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-SampledType-04470", 45}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-SampledType-04471", 45}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-SampledType-04472", 45}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-SampledType-04473", 45}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-VkPipelineVieportCreateInfo-04141", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-VkPipelineVieportCreateInfo-04142", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-addressModeU-12422", 46}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-advancedBlendMaxColorAttachments-07480", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-alphaToCoverageEnable-08919", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-alphaToCoverageEnable-08920", 55}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-attachmentCount-07750", 49}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-blendEnable-04727", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-colorAttachmentCount-06179", 54}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-colorAttachmentCount-06185", 54}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-colorAttachmentCount-09362", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-colorAttachmentCount-09372", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-colorAttachmentCount-11871", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-commandBuffer-02707", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-commandBuffer-10746", 47}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-conservativePointAndLineRasterization-07499", 71}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-coverageModulationTableEnable-07488", 63}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-coverageReductionMode-07491", 55}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-coverageToColorEnable-07490", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-cubicRangeClamp-09212", 49}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-customResolve-11524", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-customResolve-11525", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-customResolve-11529", 47}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-customResolve-11530", 47}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dimensionCount-09905", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingLocalRead-11797", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08910", 67}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08911", 67}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08912", 67}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08913", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08914", 67}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08915", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08916", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08917", 67}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-08918", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-11864", 67}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-11867", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-dynamicRenderingUnusedAttachments-11870", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-filterCubic-02694", 45}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-filterCubicMinmax-02695", 51}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-firstAttachment-07476", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-firstAttachment-07478", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-02696", 39}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-10582", 39}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-11521", 39}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-13361", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-13362", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-13363", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-13364", 39}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-flags-13365", 39}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-format-07753", 40}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-groupCountX-07820", 45}, { {"groupCountX must be less than or equal to VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI::maxWorkGroupCount[0]", 116}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-groupCountY-07821", 45}, { {"groupCountY must be less than or equal to VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI::maxWorkGroupCount[1]", 116}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-groupCountZ-07822", 45}, { {"groupCountZ must be less than or equal to VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI::maxWorkGroupCount[2]", 116}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageLayout-00344", 45}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06172", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06173", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06174", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06175", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06176", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06177", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06183", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-imageView-06184", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-index-11450", 39}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-layers-10831", 40}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-logicOp-04878", 41}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-magFilter-04553", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-magFilter-09598", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-maxFragmentDualSrcAttachments-09239", 63}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-maxMultiviewInstanceIndex-02688", 59}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-micromap-11636", 42}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-micromap-11637", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-micromap-11638", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-micromap-11639", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-mipmapMode-04770", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-mipmapMode-09599", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-multisampledRenderToSingleSampled-07285", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-multisampledRenderToSingleSampled-07286", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-multisampledRenderToSingleSampled-07287", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-multiviewPerViewViewports-12262", 59}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-multiviewPerViewViewports-12263", 59}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-nextStage-10745", 43}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pBindInfo-11375", 43}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pColorAttachments-08963", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pColorAttachments-11539", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthAttachment-06186", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthAttachment-08964", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthAttachment-11540", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthAttachment-11872", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthInputAttachmentIndex-09595", 60}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDepthInputAttachmentIndex-09596", 60}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDescription-09900", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDynamicStates-08715", 48}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pDynamicStates-08716", 48}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pNext-07935", 39}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pNext-12463", 39}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pStages-10680", 41}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pStencilAttachment-06187", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pStencilAttachment-08965", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pStencilAttachment-11860", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pStencilAttachment-11873", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pipelineFragmentShadingRate-09238", 61}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-pipelineStatistics-07076", 52}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-primitiveFragmentShadingRateWithMultipleViewports-04552", 83}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-primitiveFragmentShadingRateWithMultipleViewports-08642", 83}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-primitivesGeneratedQueryWithNonZeroStreams-06709", 76}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-primitivesGeneratedQueryWithNonZeroStreams-07481", 76}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-primitivesGeneratedQueryWithRasterizerDiscard-06708", 79}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-protectedNoFault-11455", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-protectedNoFault-11456", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizationSamples-07471", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizationSamples-07489", 54}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizationSamples-07494", 54}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizerDiscardEnable-09236", 57}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizerDiscardEnable-09420", 57}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-rasterizerDiscardEnable-10863", 57}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-reductionMode-09213", 47}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-renderPass-02684", 44}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-renderPass-06198", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-renderpass", 38}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-02689", 55}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07484", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07485", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07486", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07487", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07936", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07937", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsEnable-07938", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsPerPixel-07482", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sampleLocationsPerPixel-07483", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-samples-07472", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-samples-07473", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-scissorCount-03418", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-selectableCubicWeights-09214", 56}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-shadingRateImage-09233", 50}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-shadingRateImage-09234", 50}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sparseImageInt64Atomics-04474", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-sparseImageInt64Atomics-04475", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stage-06480", 39}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stage-07073", 39}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stippledLineEnable-07495", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stippledLineEnable-07496", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stippledLineEnable-07497", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-stippledLineEnable-07498", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-storageBuffers-06936", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-subpass-02685", 41}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-uniformBuffers-06935", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-unnormalizedCoordinates-09635", 57}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewMask-06178", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewType-07752", 42}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-03417", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-03419", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-04137", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-04138", 47}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-04139", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-07492", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-07493", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterHUAWEI-viewportCount-09421", 47}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-ClusterCullingHUAWEI-07824", 62}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the ClusterCullingHUAWEI Execution Model", 136}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-InputAttachmentIndex-09597", 62}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-02691", 46}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-02692", 46}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-02693", 46}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-04876", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-04877", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-06479", 46}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-06537", 46}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-06666", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-06886", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-06887", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07074", 46}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07075", 46}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07288", 46}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07469", 46}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07620", 46}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07621", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07622", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07623", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07624", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07625", 46}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07626", 46}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07627", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07629", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07630", 46}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07631", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07632", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07633", 46}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07634", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07636", 46}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07639", 46}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07640", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07641", 46}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07642", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07643", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07644", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07645", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07646", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07647", 46}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07648", 46}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07649", 46}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07749", 46}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07751", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07831", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07832", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07834", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07835", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07836", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07837", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07838", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07839", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07840", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07841", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07843", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07844", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07845", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07846", 46}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07847", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07848", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07849", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07850", 46}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07878", 46}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07879", 46}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07880", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07881", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-07888", 46}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08114", 46}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08115", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08116", 46}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08117", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08119", 46}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08600", 46}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08601", 46}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08605", 46}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08606", 46}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08608", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08609", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08610", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08611", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08612", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08613", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08617", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08636", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08637", 46}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08644", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08666", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08669", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08684", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08685", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08686", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08687", 46}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08688", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08689", 46}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08690", 46}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08693", 46}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08694", 46}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08695", 46}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08696", 46}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08698", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08699", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08876", 46}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08877", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08878", 46}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-08879", 46}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09003", 46}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09116", 46}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09211", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09363", 46}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09364", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09365", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09368", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09369", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09548", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09549", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09600", 46}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09642", 46}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09643", 46}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-09650", 46}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10068", 46}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10608", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10677", 46}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10678", 46}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10679", 46}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10772", 46}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10795", 46}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10796", 46}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10797", 46}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10862", 46}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10864", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10927", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-10928", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11297", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11298", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11299", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11300", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11301", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11302", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11304", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11305", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11306", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11308", 46}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11309", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11372", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11373", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11374", 46}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11376", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11397", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11398", 46}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11437", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11438", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11439", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11441", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11442", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11485", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11522", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11523", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11861", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11862", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11863", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11865", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11866", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11868", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-11869", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-12338", 46}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-12339", 46}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-12340", 46}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-None-12472", 46}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchSADQCOM-06976", 66}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchSADQCOM-12420", 66}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchSSDQCOM-06974", 66}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchWindow-09215", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchWindow-09216", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchWindow-09217", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBlockMatchWindow-12421", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageBoxFilterQCOM-06973", 62}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageSampleWeightedQCOM-06971", 67}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageSampleWeightedQCOM-06972", 67}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageSampleWeightedQCOM-06977", 67}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageSampleWeightedQCOM-06978", 67}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageWrite-04469", 54}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageWrite-08795", 54}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpImageWrite-08796", 54}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTensorReadARM-12462", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeImage-07027", 53}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeImage-07028", 53}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeImage-07029", 53}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeImage-07030", 53}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeImage-07468", 53}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-OpTypeTensorARM-09906", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-SampledType-04470", 53}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-SampledType-04471", 53}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-SampledType-04472", 53}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-SampledType-04473", 53}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-VkPipelineVieportCreateInfo-04141", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-VkPipelineVieportCreateInfo-04142", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-addressModeU-12422", 54}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-advancedBlendMaxColorAttachments-07480", 74}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-alphaToCoverageEnable-08919", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-alphaToCoverageEnable-08920", 63}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-attachmentCount-07750", 57}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-blendEnable-04727", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-buffer-parameter", 52}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-colorAttachmentCount-06179", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-colorAttachmentCount-06185", 62}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-colorAttachmentCount-09362", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-colorAttachmentCount-09372", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-colorAttachmentCount-11871", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commandBuffer-02707", 55}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commandBuffer-10746", 55}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-commonparent", 48}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-conservativePointAndLineRasterization-07499", 79}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-coverageModulationTableEnable-07488", 71}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-coverageReductionMode-07491", 63}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-coverageToColorEnable-07490", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-cubicRangeClamp-09212", 57}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-customResolve-11524", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-customResolve-11525", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-customResolve-11529", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-customResolve-11530", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dimensionCount-09905", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-drawCount-02718", 51}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-drawCount-02719", 51}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingLocalRead-11797", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08910", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08911", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08912", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08913", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08914", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08915", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08916", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08917", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-08918", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-11864", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-11867", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-dynamicRenderingUnusedAttachments-11870", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-filterCubic-02694", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-filterCubicMinmax-02695", 59}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-firstAttachment-07476", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-firstAttachment-07478", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-02696", 47}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-10582", 47}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-11521", 47}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-13361", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-13362", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-13363", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-13364", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-flags-13365", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-format-07753", 48}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageLayout-00344", 53}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06172", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06173", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06174", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06175", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06176", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06177", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06183", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-imageView-06184", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-index-11450", 47}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-layers-10831", 48}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-logicOp-04878", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-magFilter-04553", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-magFilter-09598", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-maxFragmentDualSrcAttachments-09239", 71}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-maxMultiviewInstanceIndex-02688", 67}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-micromap-11636", 50}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-micromap-11637", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-micromap-11638", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-micromap-11639", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-mipmapMode-04770", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-mipmapMode-09599", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-multisampledRenderToSingleSampled-07285", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-multisampledRenderToSingleSampled-07286", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-multisampledRenderToSingleSampled-07287", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-multiviewPerViewViewports-12262", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-multiviewPerViewViewports-12263", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-nextStage-10745", 51}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-offset-07918", 48}, { {"offset must be a multiple of VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI::indirectBufferOffsetAlignment", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pBindInfo-11375", 51}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pColorAttachments-08963", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pColorAttachments-11539", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthAttachment-06186", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthAttachment-08964", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthAttachment-11540", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthAttachment-11872", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthInputAttachmentIndex-09595", 68}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDepthInputAttachmentIndex-09596", 68}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDescription-09900", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDynamicStates-08715", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pDynamicStates-08716", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pNext-07935", 47}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pNext-12463", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pStages-10680", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pStencilAttachment-06187", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pStencilAttachment-08965", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pStencilAttachment-11860", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pStencilAttachment-11873", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pipelineFragmentShadingRate-09238", 69}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-pipelineStatistics-07076", 60}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-primitiveFragmentShadingRateWithMultipleViewports-04552", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-primitiveFragmentShadingRateWithMultipleViewports-08642", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-primitivesGeneratedQueryWithNonZeroStreams-06709", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-primitivesGeneratedQueryWithNonZeroStreams-07481", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-primitivesGeneratedQueryWithRasterizerDiscard-06708", 87}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-protectedNoFault-11455", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-protectedNoFault-11456", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizationSamples-07471", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizationSamples-07489", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizationSamples-07494", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizerDiscardEnable-09236", 65}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizerDiscardEnable-09420", 65}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-rasterizerDiscardEnable-10863", 65}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-reductionMode-09213", 55}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-renderPass-02684", 52}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-renderPass-06198", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-renderpass", 46}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-02689", 63}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07484", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07485", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07486", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07487", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07936", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07937", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsEnable-07938", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsPerPixel-07482", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sampleLocationsPerPixel-07483", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-samples-07472", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-samples-07473", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-scissorCount-03418", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-selectableCubicWeights-09214", 64}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-shadingRateImage-09233", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-shadingRateImage-09234", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sparseImageInt64Atomics-04474", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-sparseImageInt64Atomics-04475", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stage-06480", 47}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stage-07073", 47}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stippledLineEnable-07495", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stippledLineEnable-07496", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stippledLineEnable-07497", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-stippledLineEnable-07498", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-storageBuffers-06936", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-subpass-02685", 49}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-suspended", 45}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-uniformBuffers-06935", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-unnormalizedCoordinates-09635", 65}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewMask-06178", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewType-07752", 50}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-03417", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-03419", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-04137", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-04138", 55}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-04139", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-07492", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-07493", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawClusterIndirectHUAWEI-viewportCount-09421", 55}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-Input-07939", 33}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-Input-08734", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-InputAttachmentIndex-09597", 48}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-02691", 32}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-02692", 32}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-02693", 32}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-02721", 32}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04007", 32}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04008", 32}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04875", 32}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04876", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04877", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04879", 32}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-04914", 32}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-06479", 32}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-06537", 32}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-06666", 32}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-06886", 32}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-06887", 32}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07288", 32}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07312", 32}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07469", 32}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07619", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07620", 32}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07621", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07622", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07623", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07624", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07625", 32}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07626", 32}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07627", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07629", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07630", 32}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07631", 32}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07632", 32}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07633", 32}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07634", 32}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07636", 32}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07639", 32}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07640", 32}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07641", 32}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07642", 32}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07643", 32}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07644", 32}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07645", 32}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07646", 32}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07647", 32}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07648", 32}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07649", 32}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07749", 32}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07751", 32}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07831", 32}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07832", 32}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07834", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07835", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07836", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07837", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07838", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07839", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07840", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07841", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07842", 32}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07843", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07844", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07845", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07846", 32}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07847", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07848", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07849", 32}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07850", 32}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07878", 32}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07879", 32}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07880", 32}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07881", 32}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-07888", 32}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08114", 32}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08115", 32}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08116", 32}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08117", 32}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08119", 32}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08600", 32}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08601", 32}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08605", 32}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08606", 32}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08608", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08609", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08610", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08611", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08612", 32}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08613", 32}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08617", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08636", 32}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08637", 32}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08644", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08666", 32}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08669", 32}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08684", 32}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08685", 32}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08686", 32}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08687", 32}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08688", 32}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08689", 32}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08690", 32}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08693", 32}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08696", 32}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08698", 32}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08699", 32}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08876", 32}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08877", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08878", 32}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08879", 32}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-08885", 32}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09003", 32}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09116", 32}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09203", 32}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09211", 32}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09363", 32}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09364", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09365", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09368", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09369", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09462", 32}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09548", 32}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09549", 32}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09600", 32}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09637", 32}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09642", 32}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09643", 32}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-09650", 32}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10068", 32}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10608", 32}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10677", 32}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10678", 32}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10679", 32}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10772", 32}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10795", 32}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10796", 32}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10797", 32}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10862", 32}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10864", 32}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10909", 32}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10927", 32}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-10928", 32}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11297", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11298", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11299", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11300", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11301", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11302", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11304", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11305", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11306", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11308", 32}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11309", 32}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11372", 32}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11373", 32}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11374", 32}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11376", 32}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11397", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11398", 32}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11437", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11438", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11439", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11441", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11442", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11485", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11522", 32}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11523", 32}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11861", 32}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11862", 32}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11863", 32}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11865", 32}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11866", 32}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11868", 32}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-11869", 32}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-12338", 32}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-12339", 32}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-12340", 32}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-12472", 32}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-None-13118", 32}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpExecutionMode-12239", 43}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpExecutionMode-12240", 43}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpExecutionMode-12241", 43}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpExecutionMode-12242", 43}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchSADQCOM-06976", 52}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchSADQCOM-12420", 52}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchSSDQCOM-06974", 52}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchWindow-09215", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchWindow-09216", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchWindow-09217", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBlockMatchWindow-12421", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageBoxFilterQCOM-06973", 48}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageSampleWeightedQCOM-06971", 53}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageSampleWeightedQCOM-06972", 53}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageSampleWeightedQCOM-06977", 53}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageSampleWeightedQCOM-06978", 53}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageWrite-04469", 40}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageWrite-08795", 40}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpImageWrite-08796", 40}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTensorReadARM-12462", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeImage-07027", 39}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeImage-07028", 39}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeImage-07029", 39}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeImage-07030", 39}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeImage-07468", 39}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-OpTypeTensorARM-09906", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-SampledType-04470", 39}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-SampledType-04471", 39}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-SampledType-04472", 39}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-SampledType-04473", 39}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-VkPipelineVieportCreateInfo-04141", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-VkPipelineVieportCreateInfo-04142", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-addressModeU-12422", 40}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-advancedBlendMaxColorAttachments-07480", 60}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-alphaToCoverageEnable-08919", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-alphaToCoverageEnable-08920", 49}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-attachmentCount-07750", 43}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-blendEnable-04727", 39}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-colorAttachmentCount-06179", 48}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-colorAttachmentCount-06185", 48}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-colorAttachmentCount-09362", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-colorAttachmentCount-09372", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-colorAttachmentCount-11871", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-02707", 41}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-02712", 41}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-02713", 41}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-04617", 41}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-10746", 41}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-conservativePointAndLineRasterization-07499", 65}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-coverageModulationTableEnable-07488", 57}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-coverageReductionMode-07491", 49}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-coverageToColorEnable-07490", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-cubicRangeClamp-09212", 43}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-customResolve-11524", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-customResolve-11525", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-customResolve-11529", 41}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-customResolve-11530", 41}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dimensionCount-09905", 42}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicPrimitiveTopologyUnrestricted-07500", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingLocalRead-11797", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08910", 61}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08911", 61}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08912", 61}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08913", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08914", 61}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08915", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08916", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08917", 61}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-08918", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-11864", 61}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-11867", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-dynamicRenderingUnusedAttachments-11870", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-filterCubic-02694", 39}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-filterCubicMinmax-02695", 45}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-firstAttachment-07476", 43}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-firstAttachment-07478", 43}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-02696", 33}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-10582", 33}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-11521", 33}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-13361", 33}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-13362", 33}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-13363", 33}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-13364", 33}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-flags-13365", 33}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-format-07753", 34}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-format-08936", 34}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-format-08937", 34}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-format-10389", 34}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-format-10390", 34}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageLayout-00344", 39}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06172", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06173", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06174", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06175", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06176", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06177", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06183", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-imageView-06184", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-index-11450", 33}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-layers-10831", 34}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-logicOp-04878", 35}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-magFilter-04553", 37}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-magFilter-09598", 37}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-maxFragmentDualSrcAttachments-09239", 57}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-maxMultiviewInstanceIndex-02688", 53}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-micromap-11636", 36}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-micromap-11637", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-micromap-11638", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-micromap-11639", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-mipmapMode-04770", 38}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-mipmapMode-09599", 38}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-multisampledRenderToSingleSampled-07285", 61}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-multisampledRenderToSingleSampled-07286", 61}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-multisampledRenderToSingleSampled-07287", 61}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-multiviewPerViewViewports-12262", 53}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-multiviewPerViewViewports-12263", 53}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-nextStage-10745", 37}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pBindInfo-11375", 37}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pColorAttachments-08963", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pColorAttachments-11539", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthAttachment-06186", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthAttachment-08964", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthAttachment-11540", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthAttachment-11872", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthInputAttachmentIndex-09595", 54}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDepthInputAttachmentIndex-09596", 54}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDescription-09900", 40}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDynamicStates-08715", 42}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pDynamicStates-08716", 42}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pNext-07935", 33}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pNext-09461", 33}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pNext-12463", 33}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pStencilAttachment-06187", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pStencilAttachment-08965", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pStencilAttachment-11860", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pStencilAttachment-11873", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pStrides-04913", 36}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-pipelineFragmentShadingRate-09238", 55}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveFragmentShadingRateWithMultipleViewports-04552", 77}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveFragmentShadingRateWithMultipleViewports-08642", 77}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveRestartIndex-12401", 49}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveTopology-10286", 45}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveTopology-10747", 45}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitiveTopology-10748", 45}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitivesGeneratedQueryWithNonZeroStreams-06709", 70}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitivesGeneratedQueryWithNonZeroStreams-07481", 70}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-primitivesGeneratedQueryWithRasterizerDiscard-06708", 73}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-protectedNoFault-11455", 44}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-protectedNoFault-11456", 44}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizationSamples-07471", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizationSamples-07489", 48}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizationSamples-07494", 48}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizerDiscardEnable-09236", 51}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizerDiscardEnable-09420", 51}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-rasterizerDiscardEnable-10863", 51}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-reductionMode-09213", 41}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-renderPass-02684", 38}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-renderPass-06198", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-renderpass", 32}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-robustBufferAccess2-08798", 47}, { {"If the robustBufferAccess2 feature is not enabled, (indexSize x (firstIndex + indexCount)) must be less than or equal to the size of the bound index buffer range, with indexSize being based on the type specified by indexType, and the other parameters sourced from this command", 276}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-02689", 49}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07484", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07485", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07486", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07487", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07936", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07937", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsEnable-07938", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsPerPixel-07482", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sampleLocationsPerPixel-07483", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-samples-07472", 35}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-samples-07473", 35}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-scissorCount-03418", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-selectableCubicWeights-09214", 50}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-shadingRateImage-09233", 44}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-shadingRateImage-09234", 44}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sparseImageInt64Atomics-04474", 51}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-sparseImageInt64Atomics-04475", 51}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stage-06481", 33}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stage-07073", 33}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stippledLineEnable-07495", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stippledLineEnable-07496", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stippledLineEnable-07497", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-stippledLineEnable-07498", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-storageBuffers-06936", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-subpass-02685", 35}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-uniformBuffers-06935", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-unnormalizedCoordinates-09635", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewMask-06178", 36}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewType-07752", 36}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-03417", 41}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-03419", 41}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-04137", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-04138", 41}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-04139", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-07492", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-07493", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexed-viewportCount-09421", 41}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-Input-07939", 41}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-Input-08734", 41}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-InputAttachmentIndex-09597", 56}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-02691", 40}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-02692", 40}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-02693", 40}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-02721", 40}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04007", 40}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04008", 40}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04875", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04876", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04877", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04879", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-04914", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-06479", 40}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-06537", 40}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-06666", 40}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-06886", 40}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-06887", 40}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07288", 40}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07312", 40}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07469", 40}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07619", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07620", 40}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07621", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07622", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07623", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07624", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07625", 40}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07626", 40}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07627", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07629", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07630", 40}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07631", 40}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07632", 40}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07633", 40}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07634", 40}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07636", 40}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07639", 40}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07640", 40}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07641", 40}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07642", 40}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07643", 40}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07644", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07645", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07646", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07647", 40}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07648", 40}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07649", 40}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07749", 40}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07751", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07831", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07832", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07834", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07835", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07836", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07837", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07838", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07839", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07840", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07841", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07842", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07843", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07844", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07845", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07846", 40}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07847", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07848", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07849", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07850", 40}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07878", 40}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07879", 40}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07880", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07881", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-07888", 40}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08114", 40}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08115", 40}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08116", 40}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08117", 40}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08119", 40}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08600", 40}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08601", 40}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08605", 40}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08606", 40}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08608", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08609", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08610", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08611", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08612", 40}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08613", 40}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08617", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08636", 40}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08637", 40}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08644", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08666", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08669", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08684", 40}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08685", 40}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08686", 40}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08687", 40}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08688", 40}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08689", 40}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08690", 40}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08693", 40}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08696", 40}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08698", 40}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08699", 40}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08876", 40}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08877", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08878", 40}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08879", 40}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-08885", 40}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09003", 40}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09116", 40}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09203", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09211", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09363", 40}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09364", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09365", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09368", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09369", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09548", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09549", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09600", 40}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09637", 40}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09642", 40}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09643", 40}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-09650", 40}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10068", 40}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10608", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10677", 40}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10678", 40}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10679", 40}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10772", 40}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10795", 40}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10796", 40}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10797", 40}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10862", 40}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10864", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10909", 40}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10927", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-10928", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11297", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11298", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11299", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11300", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11301", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11302", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11304", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11305", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11306", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11308", 40}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11309", 40}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11372", 40}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11373", 40}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11374", 40}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11376", 40}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11397", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11398", 40}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11437", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11438", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11439", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11441", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11442", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11485", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11522", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11523", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11861", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11862", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11863", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11865", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11866", 40}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11868", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-11869", 40}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-12338", 40}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-12339", 40}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-12340", 40}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-12472", 40}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-None-13118", 40}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpExecutionMode-12239", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpExecutionMode-12240", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpExecutionMode-12241", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpExecutionMode-12242", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchSADQCOM-06976", 60}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchSADQCOM-12420", 60}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchSSDQCOM-06974", 60}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchWindow-09215", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchWindow-09216", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchWindow-09217", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBlockMatchWindow-12421", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageBoxFilterQCOM-06973", 56}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageSampleWeightedQCOM-06971", 61}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageSampleWeightedQCOM-06972", 61}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageSampleWeightedQCOM-06977", 61}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageSampleWeightedQCOM-06978", 61}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageWrite-04469", 48}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageWrite-08795", 48}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpImageWrite-08796", 48}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTensorReadARM-12462", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeImage-07027", 47}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeImage-07028", 47}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeImage-07029", 47}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeImage-07030", 47}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeImage-07468", 47}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-OpTypeTensorARM-09906", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-SampledType-04470", 47}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-SampledType-04471", 47}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-SampledType-04472", 47}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-SampledType-04473", 47}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-VkPipelineVieportCreateInfo-04141", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-VkPipelineVieportCreateInfo-04142", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-addressModeU-12422", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-advancedBlendMaxColorAttachments-07480", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-alphaToCoverageEnable-08919", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-alphaToCoverageEnable-08920", 57}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-attachmentCount-07750", 51}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-blendEnable-04727", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-buffer-02708", 42}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-buffer-02709", 42}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-buffer-parameter", 46}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-colorAttachmentCount-06179", 56}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-colorAttachmentCount-06185", 56}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-colorAttachmentCount-09362", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-colorAttachmentCount-09372", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-colorAttachmentCount-11871", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-02707", 49}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-02711", 49}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-10746", 49}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-commonparent", 42}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-conservativePointAndLineRasterization-07499", 73}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-coverageModulationTableEnable-07488", 65}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-coverageReductionMode-07491", 57}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-coverageToColorEnable-07490", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-cubicRangeClamp-09212", 51}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-customResolve-11524", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-customResolve-11525", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-customResolve-11529", 49}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-customResolve-11530", 49}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dimensionCount-09905", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-drawCount-00528", 45}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawIndexedIndirectCommand)", 136}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-drawCount-00539", 45}, { {"If drawCount is equal to 1, (offset + sizeof(VkDrawIndexedIndirectCommand)) must be less than or equal to the size of buffer", 124}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-drawCount-00540", 45}, { {"If drawCount is greater than 1, (stride x (drawCount - 1) + offset + sizeof(VkDrawIndexedIndirectCommand)) must be less than or equal to the size of buffer", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-drawCount-02718", 45}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-drawCount-02719", 45}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicPrimitiveTopologyUnrestricted-07500", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingLocalRead-11797", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08910", 69}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08911", 69}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08912", 69}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08913", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08914", 69}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08915", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08916", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08917", 69}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-08918", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-11864", 69}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-11867", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-dynamicRenderingUnusedAttachments-11870", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-filterCubic-02694", 47}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-filterCubicMinmax-02695", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-firstAttachment-07476", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-firstAttachment-07478", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-02696", 41}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-10582", 41}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-11521", 41}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-13361", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-13362", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-13363", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-13364", 41}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-flags-13365", 41}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-format-07753", 42}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-format-08936", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-format-08937", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-format-10389", 42}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-format-10390", 42}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageLayout-00344", 47}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06172", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06173", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06174", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06175", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06176", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06177", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06183", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-imageView-06184", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-index-11450", 41}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-layers-10831", 42}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-logicOp-04878", 43}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-magFilter-04553", 45}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-magFilter-09598", 45}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-maxFragmentDualSrcAttachments-09239", 65}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-maxMultiviewInstanceIndex-02688", 61}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-micromap-11636", 44}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-micromap-11637", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-micromap-11638", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-micromap-11639", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-mipmapMode-04770", 46}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-mipmapMode-09599", 46}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-multisampledRenderToSingleSampled-07285", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-multisampledRenderToSingleSampled-07286", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-multisampledRenderToSingleSampled-07287", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-multiviewPerViewViewports-12262", 61}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-multiviewPerViewViewports-12263", 61}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-nextStage-10745", 45}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-offset-02710", 42}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pBindInfo-11375", 45}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pColorAttachments-08963", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pColorAttachments-11539", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthAttachment-06186", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthAttachment-08964", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthAttachment-11540", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthAttachment-11872", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthInputAttachmentIndex-09595", 62}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDepthInputAttachmentIndex-09596", 62}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDescription-09900", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDynamicStates-08715", 50}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pDynamicStates-08716", 50}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pNext-07935", 41}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pNext-12463", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pStencilAttachment-06187", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pStencilAttachment-08965", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pStencilAttachment-11860", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pStencilAttachment-11873", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pStrides-04913", 44}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-pipelineFragmentShadingRate-09238", 63}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveFragmentShadingRateWithMultipleViewports-04552", 85}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveFragmentShadingRateWithMultipleViewports-08642", 85}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveRestartIndex-12401", 57}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveTopology-10286", 53}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveTopology-10747", 53}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitiveTopology-10748", 53}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitivesGeneratedQueryWithNonZeroStreams-06709", 78}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitivesGeneratedQueryWithNonZeroStreams-07481", 78}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-primitivesGeneratedQueryWithRasterizerDiscard-06708", 81}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-protectedNoFault-11455", 52}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-protectedNoFault-11456", 52}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizationSamples-07471", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizationSamples-07489", 56}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizationSamples-07494", 56}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizerDiscardEnable-09236", 59}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizerDiscardEnable-09420", 59}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-rasterizerDiscardEnable-10863", 59}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-reductionMode-09213", 49}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-renderPass-02684", 46}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-renderPass-06198", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-renderpass", 40}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-02689", 57}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07484", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07485", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07486", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07487", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07936", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07937", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsEnable-07938", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsPerPixel-07482", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sampleLocationsPerPixel-07483", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-samples-07472", 43}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-samples-07473", 43}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-scissorCount-03418", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-selectableCubicWeights-09214", 58}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-shadingRateImage-09233", 52}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-shadingRateImage-09234", 52}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sparseImageInt64Atomics-04474", 59}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-sparseImageInt64Atomics-04475", 59}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stage-06481", 41}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stage-07073", 41}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stippledLineEnable-07495", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stippledLineEnable-07496", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stippledLineEnable-07497", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-stippledLineEnable-07498", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-storageBuffers-06936", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-subpass-02685", 43}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-suspended", 39}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-uniformBuffers-06935", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-unnormalizedCoordinates-09635", 59}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewMask-06178", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewType-07752", 44}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-03417", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-03419", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-04137", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-04138", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-04139", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-07492", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-07493", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect-viewportCount-09421", 49}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-Input-07939", 45}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-Input-08734", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-InputAttachmentIndex-09597", 60}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-02691", 44}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-02692", 44}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-02693", 44}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-02721", 44}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04007", 44}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04008", 44}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04875", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04876", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04877", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04879", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-04914", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-06479", 44}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-06537", 44}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-06666", 44}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-06886", 44}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-06887", 44}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07288", 44}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07312", 44}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07469", 44}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07619", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07620", 44}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07621", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07622", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07623", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07624", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07625", 44}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07626", 44}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07627", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07629", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07630", 44}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07631", 44}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07632", 44}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07633", 44}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07634", 44}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07636", 44}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07639", 44}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07640", 44}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07641", 44}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07642", 44}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07643", 44}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07644", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07645", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07646", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07647", 44}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07648", 44}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07649", 44}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07749", 44}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07751", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07831", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07832", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07834", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07835", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07836", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07837", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07838", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07839", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07840", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07841", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07842", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07843", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07844", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07845", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07846", 44}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07847", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07848", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07849", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07850", 44}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07878", 44}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07879", 44}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07880", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07881", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-07888", 44}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08114", 44}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08115", 44}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08116", 44}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08117", 44}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08119", 44}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08600", 44}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08601", 44}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08605", 44}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08606", 44}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08608", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08609", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08610", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08611", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08612", 44}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08613", 44}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08617", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08636", 44}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08637", 44}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08644", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08666", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08669", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08684", 44}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08685", 44}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08686", 44}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08687", 44}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08688", 44}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08689", 44}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08690", 44}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08693", 44}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08696", 44}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08698", 44}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08699", 44}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08876", 44}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08877", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08878", 44}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08879", 44}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-08885", 44}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09003", 44}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09116", 44}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09203", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09211", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09363", 44}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09364", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09365", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09368", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09369", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09548", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09549", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09600", 44}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09637", 44}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09642", 44}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09643", 44}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-09650", 44}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10068", 44}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10608", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10677", 44}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10678", 44}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10679", 44}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10772", 44}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10795", 44}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10796", 44}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10797", 44}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10862", 44}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10864", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10909", 44}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10927", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-10928", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11297", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11298", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11299", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11300", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11301", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11302", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11304", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11305", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11306", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11308", 44}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11309", 44}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11372", 44}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11373", 44}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11374", 44}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11376", 44}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11397", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11398", 44}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11437", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11438", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11439", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11441", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11442", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11485", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11522", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11523", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11861", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11862", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11863", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11865", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11866", 44}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11868", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-11869", 44}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-12338", 44}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-12339", 44}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-12340", 44}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-12472", 44}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-None-13118", 44}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpExecutionMode-12239", 55}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpExecutionMode-12240", 55}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpExecutionMode-12241", 55}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpExecutionMode-12242", 55}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchSADQCOM-06976", 64}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchSADQCOM-12420", 64}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchSSDQCOM-06974", 64}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchWindow-09215", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchWindow-09216", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchWindow-09217", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBlockMatchWindow-12421", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageBoxFilterQCOM-06973", 60}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageSampleWeightedQCOM-06971", 65}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageSampleWeightedQCOM-06972", 65}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageSampleWeightedQCOM-06977", 65}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageSampleWeightedQCOM-06978", 65}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageWrite-04469", 52}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageWrite-08795", 52}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpImageWrite-08796", 52}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTensorReadARM-12462", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeImage-07027", 51}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeImage-07028", 51}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeImage-07029", 51}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeImage-07030", 51}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeImage-07468", 51}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-OpTypeTensorARM-09906", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-SampledType-04470", 51}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-SampledType-04471", 51}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-SampledType-04472", 51}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-SampledType-04473", 51}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-VkPipelineVieportCreateInfo-04141", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-VkPipelineVieportCreateInfo-04142", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-addressModeU-12422", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-advancedBlendMaxColorAttachments-07480", 72}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-alphaToCoverageEnable-08919", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-alphaToCoverageEnable-08920", 61}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-attachmentCount-07750", 55}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-blendEnable-04727", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-colorAttachmentCount-06179", 60}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-colorAttachmentCount-06185", 60}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-colorAttachmentCount-09362", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-colorAttachmentCount-09372", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-colorAttachmentCount-11871", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-02707", 53}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-10746", 53}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-13059", 53}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-conservativePointAndLineRasterization-07499", 77}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-coverageModulationTableEnable-07488", 69}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-coverageReductionMode-07491", 61}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-coverageToColorEnable-07490", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-cubicRangeClamp-09212", 55}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-customResolve-11524", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-customResolve-11525", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-customResolve-11529", 53}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-customResolve-11530", 53}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dimensionCount-09905", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicPrimitiveTopologyUnrestricted-07500", 76}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingLocalRead-11797", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08910", 73}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08911", 73}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08912", 73}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08913", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08914", 73}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08915", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08916", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08917", 73}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-08918", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-11864", 73}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-11867", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-dynamicRenderingUnusedAttachments-11870", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-filterCubic-02694", 51}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-filterCubicMinmax-02695", 57}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-firstAttachment-07476", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-firstAttachment-07478", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-02696", 45}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-10582", 45}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-11521", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-13361", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-13362", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-13363", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-13364", 45}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-flags-13365", 45}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-format-07753", 46}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-format-08936", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-format-08937", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-format-10389", 46}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-format-10390", 46}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageLayout-00344", 51}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06172", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06173", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06174", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06175", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06176", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06177", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06183", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-imageView-06184", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-index-11450", 45}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-layers-10831", 46}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-logicOp-04878", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-magFilter-04553", 49}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-magFilter-09598", 49}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-maxFragmentDualSrcAttachments-09239", 69}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-maxMultiviewInstanceIndex-02688", 65}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-micromap-11636", 48}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-micromap-11637", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-micromap-11638", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-micromap-11639", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-mipmapMode-04770", 50}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-mipmapMode-09599", 50}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-multisampledRenderToSingleSampled-07285", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-multisampledRenderToSingleSampled-07286", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-multisampledRenderToSingleSampled-07287", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-multiviewPerViewViewports-12262", 65}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-multiviewPerViewViewports-12263", 65}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-nextStage-10745", 49}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pBindInfo-11375", 49}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pColorAttachments-08963", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pColorAttachments-11539", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthAttachment-06186", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthAttachment-08964", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthAttachment-11540", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthAttachment-11872", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthInputAttachmentIndex-09595", 66}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDepthInputAttachmentIndex-09596", 66}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDescription-09900", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDynamicStates-08715", 54}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pDynamicStates-08716", 54}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pInfo-13110", 45}, { {"If pInfo->drawCount is greater than 0, pInfo->addressRange.size must be greater than or equal to (pInfo->drawCount - 1) x pInfo->addressRange.stride + sizeof(VkDrawIndexedIndirectCommand)", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pInfo-13111", 45}, { {"If pInfo->drawCount is greater than 1, pInfo->addressRange.stride must not be 0", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pInfo-13112", 45}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be greater than or equal to sizeof(VkDrawIndexedIndirectCommand)", 136}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pInfo-13113", 45}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be a multiple of 4", 90}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pInfo-parameter", 49}, { {"pInfo must be a valid pointer to a valid VkDrawIndirect2InfoKHR structure", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pNext-07935", 45}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pNext-12463", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pStencilAttachment-06187", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pStencilAttachment-08965", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pStencilAttachment-11860", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pStencilAttachment-11873", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pStrides-04913", 48}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-pipelineFragmentShadingRate-09238", 67}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveFragmentShadingRateWithMultipleViewports-04552", 89}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveFragmentShadingRateWithMultipleViewports-08642", 89}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveRestartIndex-12401", 61}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveTopology-10286", 57}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveTopology-10747", 57}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitiveTopology-10748", 57}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitivesGeneratedQueryWithNonZeroStreams-06709", 82}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitivesGeneratedQueryWithNonZeroStreams-07481", 82}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-primitivesGeneratedQueryWithRasterizerDiscard-06708", 85}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-protectedNoFault-11455", 56}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-protectedNoFault-11456", 56}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizationSamples-07471", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizationSamples-07489", 60}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizationSamples-07494", 60}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizerDiscardEnable-09236", 63}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizerDiscardEnable-09420", 63}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-rasterizerDiscardEnable-10863", 63}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-reductionMode-09213", 53}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-renderPass-02684", 50}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-renderPass-06198", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-renderpass", 44}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-02689", 61}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07484", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07485", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07486", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07487", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07936", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07937", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsEnable-07938", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsPerPixel-07482", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sampleLocationsPerPixel-07483", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-samples-07472", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-samples-07473", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-scissorCount-03418", 52}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-selectableCubicWeights-09214", 62}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-shadingRateImage-09233", 56}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-shadingRateImage-09234", 56}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sparseImageInt64Atomics-04474", 63}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-sparseImageInt64Atomics-04475", 63}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stage-06481", 45}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stage-07073", 45}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stippledLineEnable-07495", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stippledLineEnable-07496", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stippledLineEnable-07497", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-stippledLineEnable-07498", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-storageBuffers-06936", 54}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-subpass-02685", 47}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-suspended", 43}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-uniformBuffers-06935", 54}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-unnormalizedCoordinates-09635", 63}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewMask-06178", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewType-07752", 48}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-03417", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-03419", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-04137", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-04138", 53}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-04139", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-07492", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-07493", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirect2KHR-viewportCount-09421", 53}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-Input-07939", 46}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-Input-08734", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-InputAttachmentIndex-09597", 61}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-02691", 45}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-02692", 45}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-02693", 45}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-02721", 45}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04007", 45}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04008", 45}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04445", 45}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04875", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04876", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04877", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04879", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-04914", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-06479", 45}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-06537", 45}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-06666", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-06886", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-06887", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07288", 45}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07312", 45}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07469", 45}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07619", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07620", 45}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07621", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07622", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07623", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07624", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07625", 45}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07626", 45}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07627", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07629", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07630", 45}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07631", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07632", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07633", 45}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07634", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07636", 45}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07639", 45}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07640", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07641", 45}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07642", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07643", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07644", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07645", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07646", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07647", 45}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07648", 45}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07649", 45}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07749", 45}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07751", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07831", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07832", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07834", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07835", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07836", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07837", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07838", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07839", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07840", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07841", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07842", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07843", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07844", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07845", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07846", 45}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07847", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07848", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07849", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07850", 45}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07878", 45}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07879", 45}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07880", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07881", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-07888", 45}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08114", 45}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08115", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08116", 45}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08117", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08119", 45}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08600", 45}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08601", 45}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08605", 45}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08606", 45}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08608", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08609", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08610", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08611", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08612", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08613", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08617", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08636", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08637", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08644", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08666", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08669", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08684", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08685", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08686", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08687", 45}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08688", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08689", 45}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08690", 45}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08693", 45}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08696", 45}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08698", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08699", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08876", 45}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08877", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08878", 45}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08879", 45}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-08885", 45}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09003", 45}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09116", 45}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09203", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09211", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09363", 45}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09364", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09365", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09368", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09369", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09548", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09549", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09600", 45}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09637", 45}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09642", 45}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09643", 45}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-09650", 45}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10068", 45}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10608", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10677", 45}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10678", 45}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10679", 45}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10772", 45}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10795", 45}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10796", 45}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10797", 45}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10862", 45}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10864", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10909", 45}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10927", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-10928", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11297", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11298", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11299", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11300", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11301", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11302", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11304", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11305", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11306", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11308", 45}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11309", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11372", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11373", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11374", 45}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11376", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11397", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11398", 45}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11437", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11438", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11439", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11441", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11442", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11485", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11522", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11523", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11861", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11862", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11863", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11865", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11866", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11868", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-11869", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-12338", 45}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-12339", 45}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-12340", 45}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-12472", 45}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-None-13118", 45}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpExecutionMode-12239", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpExecutionMode-12240", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpExecutionMode-12241", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpExecutionMode-12242", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchSADQCOM-06976", 65}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchSADQCOM-12420", 65}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchSSDQCOM-06974", 65}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchWindow-09215", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchWindow-09216", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchWindow-09217", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBlockMatchWindow-12421", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageBoxFilterQCOM-06973", 61}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageSampleWeightedQCOM-06971", 66}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageSampleWeightedQCOM-06972", 66}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageSampleWeightedQCOM-06977", 66}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageSampleWeightedQCOM-06978", 66}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageWrite-04469", 53}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageWrite-08795", 53}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpImageWrite-08796", 53}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTensorReadARM-12462", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeImage-07027", 52}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeImage-07028", 52}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeImage-07029", 52}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeImage-07030", 52}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeImage-07468", 52}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-OpTypeTensorARM-09906", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-SampledType-04470", 52}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-SampledType-04471", 52}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-SampledType-04472", 52}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-SampledType-04473", 52}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-VkPipelineVieportCreateInfo-04141", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-VkPipelineVieportCreateInfo-04142", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-addressModeU-12422", 53}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-advancedBlendMaxColorAttachments-07480", 73}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-alphaToCoverageEnable-08919", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-alphaToCoverageEnable-08920", 62}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-attachmentCount-07750", 56}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-blendEnable-04727", 52}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-buffer-02708", 47}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-buffer-02709", 47}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-buffer-parameter", 51}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-colorAttachmentCount-06179", 61}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-colorAttachmentCount-06185", 61}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-colorAttachmentCount-09362", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-colorAttachmentCount-09372", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-colorAttachmentCount-11871", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-02707", 54}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-02711", 54}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-10746", 54}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-commonparent", 47}, { {"Each of buffer, commandBuffer, and countBuffer must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-conservativePointAndLineRasterization-07499", 78}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBuffer-02714", 52}, { {"If countBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBuffer-02715", 52}, { {"countBuffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBuffer-02717", 52}, { {"The count stored in countBuffer must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 106}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBuffer-parameter", 56}, { {"countBuffer must be a valid VkBuffer handle", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBufferOffset-02716", 58}, { {"countBufferOffset must be a multiple of 4", 41}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-countBufferOffset-04129", 58}, { {"(countBufferOffset + sizeof(uint32_t)) must be less than or equal to the size of countBuffer", 92}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-coverageModulationTableEnable-07488", 70}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-coverageReductionMode-07491", 62}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-coverageToColorEnable-07490", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-cubicRangeClamp-09212", 56}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-customResolve-11524", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-customResolve-11525", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-customResolve-11529", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-customResolve-11530", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dimensionCount-09905", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicPrimitiveTopologyUnrestricted-07500", 77}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingLocalRead-11797", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08910", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08911", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08912", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08913", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08914", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08915", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08916", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08917", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-08918", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-11864", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-11867", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-dynamicRenderingUnusedAttachments-11870", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-filterCubic-02694", 52}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-filterCubicMinmax-02695", 58}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-firstAttachment-07476", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-firstAttachment-07478", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-02696", 46}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-10582", 46}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-11521", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-13361", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-13362", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-13363", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-13364", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-flags-13365", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-format-07753", 47}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-format-08936", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-format-08937", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-format-10389", 47}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-format-10390", 47}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageLayout-00344", 52}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06172", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06173", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06174", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06175", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06176", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06177", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06183", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-imageView-06184", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-index-11450", 46}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-layers-10831", 47}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-logicOp-04878", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-magFilter-04553", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-magFilter-09598", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-maxDrawCount-03143", 53}, { {"If maxDrawCount is greater than or equal to 1, (stride x (maxDrawCount - 1) + offset + sizeof(VkDrawIndexedIndirectCommand)) must be less than or equal to the size of buffer", 173}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-maxFragmentDualSrcAttachments-09239", 70}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-maxMultiviewInstanceIndex-02688", 66}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-micromap-11636", 49}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-micromap-11637", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-micromap-11638", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-micromap-11639", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-mipmapMode-04770", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-mipmapMode-09599", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-multisampledRenderToSingleSampled-07285", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-multisampledRenderToSingleSampled-07286", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-multisampledRenderToSingleSampled-07287", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-multiviewPerViewViewports-12262", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-multiviewPerViewViewports-12263", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-nextStage-10745", 50}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-offset-02710", 47}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pBindInfo-11375", 50}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pColorAttachments-08963", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pColorAttachments-11539", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthAttachment-06186", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthAttachment-08964", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthAttachment-11540", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthAttachment-11872", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthInputAttachmentIndex-09595", 67}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDepthInputAttachmentIndex-09596", 67}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDescription-09900", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDynamicStates-08715", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pDynamicStates-08716", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pNext-07935", 46}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pNext-12463", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pStencilAttachment-06187", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pStencilAttachment-08965", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pStencilAttachment-11860", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pStencilAttachment-11873", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pStrides-04913", 49}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-pipelineFragmentShadingRate-09238", 68}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveFragmentShadingRateWithMultipleViewports-04552", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveFragmentShadingRateWithMultipleViewports-08642", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveRestartIndex-12401", 62}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveTopology-10286", 58}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveTopology-10747", 58}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitiveTopology-10748", 58}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitivesGeneratedQueryWithNonZeroStreams-06709", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitivesGeneratedQueryWithNonZeroStreams-07481", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-primitivesGeneratedQueryWithRasterizerDiscard-06708", 86}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-protectedNoFault-11455", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-protectedNoFault-11456", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizationSamples-07471", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizationSamples-07489", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizationSamples-07494", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizerDiscardEnable-09236", 64}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizerDiscardEnable-09420", 64}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-rasterizerDiscardEnable-10863", 64}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-reductionMode-09213", 54}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-renderPass-02684", 51}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-renderPass-06198", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-renderpass", 45}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-02689", 62}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07484", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07485", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07486", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07487", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07936", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07937", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsEnable-07938", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsPerPixel-07482", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sampleLocationsPerPixel-07483", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-samples-07472", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-samples-07473", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-scissorCount-03418", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-selectableCubicWeights-09214", 63}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-shadingRateImage-09233", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-shadingRateImage-09234", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sparseImageInt64Atomics-04474", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-sparseImageInt64Atomics-04475", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stage-06481", 46}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stage-07073", 46}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stippledLineEnable-07495", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stippledLineEnable-07496", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stippledLineEnable-07497", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stippledLineEnable-07498", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-storageBuffers-06936", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-stride-03142", 47}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawIndexedIndirectCommand)", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-subpass-02685", 48}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-suspended", 44}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-uniformBuffers-06935", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-unnormalizedCoordinates-09635", 64}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewMask-06178", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewType-07752", 49}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-03417", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-03419", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-04137", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-04138", 54}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-04139", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-07492", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-07493", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount-viewportCount-09421", 54}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-Input-07939", 50}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-Input-08734", 50}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-InputAttachmentIndex-09597", 65}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-02691", 49}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-02692", 49}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-02693", 49}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-02721", 49}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04007", 49}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04008", 49}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04875", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04876", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04877", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04879", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-04914", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-06479", 49}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-06537", 49}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-06666", 49}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-06886", 49}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-06887", 49}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07288", 49}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07312", 49}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07469", 49}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07619", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07620", 49}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07621", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07622", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07623", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07624", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07625", 49}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07626", 49}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07627", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07629", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07630", 49}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07631", 49}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07632", 49}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07633", 49}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07634", 49}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07636", 49}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07639", 49}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07640", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07641", 49}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07642", 49}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07643", 49}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07644", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07645", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07646", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07647", 49}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07648", 49}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07649", 49}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07749", 49}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07751", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07831", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07832", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07834", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07835", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07836", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07837", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07838", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07839", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07840", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07841", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07842", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07843", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07844", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07845", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07846", 49}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07847", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07848", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07849", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07850", 49}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07878", 49}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07879", 49}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07880", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07881", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-07888", 49}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08114", 49}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08115", 49}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08116", 49}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08117", 49}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08119", 49}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08600", 49}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08601", 49}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08605", 49}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08606", 49}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08608", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08609", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08610", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08611", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08612", 49}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08613", 49}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08617", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08636", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08637", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08644", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08666", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08669", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08684", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08685", 49}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08686", 49}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08687", 49}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08688", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08689", 49}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08690", 49}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08693", 49}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08696", 49}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08698", 49}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08699", 49}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08876", 49}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08877", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08878", 49}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08879", 49}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-08885", 49}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09003", 49}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09116", 49}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09203", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09211", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09363", 49}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09364", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09365", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09368", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09369", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09548", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09549", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09600", 49}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09637", 49}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09642", 49}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09643", 49}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-09650", 49}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10068", 49}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10608", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10677", 49}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10678", 49}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10679", 49}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10772", 49}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10795", 49}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10796", 49}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10797", 49}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10862", 49}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10864", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10909", 49}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10927", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-10928", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11297", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11298", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11299", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11300", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11301", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11302", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11304", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11305", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11306", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11308", 49}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11309", 49}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11372", 49}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11373", 49}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11374", 49}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11376", 49}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11397", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11398", 49}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11437", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11438", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11439", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11441", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11442", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11485", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11522", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11523", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11861", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11862", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11863", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11865", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11866", 49}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11868", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-11869", 49}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-12338", 49}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-12339", 49}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-12340", 49}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-12472", 49}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-None-13118", 49}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpExecutionMode-12239", 60}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpExecutionMode-12240", 60}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpExecutionMode-12241", 60}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpExecutionMode-12242", 60}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchSADQCOM-06976", 69}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchSADQCOM-12420", 69}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchSSDQCOM-06974", 69}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchWindow-09215", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchWindow-09216", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchWindow-09217", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBlockMatchWindow-12421", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageBoxFilterQCOM-06973", 65}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageSampleWeightedQCOM-06971", 70}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageSampleWeightedQCOM-06972", 70}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageSampleWeightedQCOM-06977", 70}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageSampleWeightedQCOM-06978", 70}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageWrite-04469", 57}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageWrite-08795", 57}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpImageWrite-08796", 57}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTensorReadARM-12462", 60}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeImage-07027", 56}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeImage-07028", 56}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeImage-07029", 56}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeImage-07030", 56}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeImage-07468", 56}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-OpTypeTensorARM-09906", 60}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-SampledType-04470", 56}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-SampledType-04471", 56}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-SampledType-04472", 56}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-SampledType-04473", 56}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-VkPipelineVieportCreateInfo-04141", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-VkPipelineVieportCreateInfo-04142", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-addressModeU-12422", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-advancedBlendMaxColorAttachments-07480", 77}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-alphaToCoverageEnable-08919", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-alphaToCoverageEnable-08920", 66}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-attachmentCount-07750", 60}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-blendEnable-04727", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-colorAttachmentCount-06179", 65}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-colorAttachmentCount-06185", 65}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-colorAttachmentCount-09362", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-colorAttachmentCount-09372", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-colorAttachmentCount-11871", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-02707", 58}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-10746", 58}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-13060", 58}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-conservativePointAndLineRasterization-07499", 82}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-coverageModulationTableEnable-07488", 74}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-coverageReductionMode-07491", 66}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-coverageToColorEnable-07490", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-cubicRangeClamp-09212", 60}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-customResolve-11524", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-customResolve-11525", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-customResolve-11529", 58}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-customResolve-11530", 58}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dimensionCount-09905", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-drawIndirectCount-13120", 62}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicPrimitiveTopologyUnrestricted-07500", 81}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingLocalRead-11797", 70}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08910", 78}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08911", 78}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08912", 78}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08913", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08914", 78}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08915", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08916", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08917", 78}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-08918", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-11864", 78}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-11867", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-dynamicRenderingUnusedAttachments-11870", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-filterCubic-02694", 56}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-filterCubicMinmax-02695", 62}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-firstAttachment-07476", 60}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-firstAttachment-07478", 60}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-02696", 50}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-10582", 50}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-11521", 50}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-13361", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-13362", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-13363", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-13364", 50}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-flags-13365", 50}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-format-07753", 51}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-format-08936", 51}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-format-08937", 51}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-format-10389", 51}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-format-10390", 51}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageLayout-00344", 56}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06172", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06173", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06174", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06175", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06176", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06177", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06183", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-imageView-06184", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-index-11450", 50}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-layers-10831", 51}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-logicOp-04878", 52}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-magFilter-04553", 54}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-magFilter-09598", 54}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-maxFragmentDualSrcAttachments-09239", 74}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-maxMultiviewInstanceIndex-02688", 70}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-micromap-11636", 53}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-micromap-11637", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-micromap-11638", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-micromap-11639", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-mipmapMode-04770", 55}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-mipmapMode-09599", 55}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-multisampledRenderToSingleSampled-07285", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-multisampledRenderToSingleSampled-07286", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-multisampledRenderToSingleSampled-07287", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-multiviewPerViewViewports-12262", 70}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-multiviewPerViewViewports-12263", 70}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-nextStage-10745", 54}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pBindInfo-11375", 54}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pColorAttachments-08963", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pColorAttachments-11539", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthAttachment-06186", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthAttachment-08964", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthAttachment-11540", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthAttachment-11872", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthInputAttachmentIndex-09595", 71}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDepthInputAttachmentIndex-09596", 71}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDescription-09900", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDynamicStates-08715", 59}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pDynamicStates-08716", 59}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pInfo-13110", 50}, { {"If pInfo->maxDrawCount is greater than 0, pInfo->addressRange.size must be greater than or equal to (pInfo->maxDrawCount - 1) x pInfo->addressRange.stride + sizeof(VkDrawIndexedIndirectCommand)", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pInfo-13111", 50}, { {"If pInfo->maxDrawCount is greater than 1, pInfo->addressRange.stride must not be 0", 82}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pInfo-13112", 50}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be greater than or equal to sizeof(VkDrawIndexedIndirectCommand)", 136}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pInfo-13113", 50}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be a multiple of 4", 90}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pInfo-parameter", 54}, { {"pInfo must be a valid pointer to a valid VkDrawIndirectCount2InfoKHR structure", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pNext-07935", 50}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pNext-12463", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pStencilAttachment-06187", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pStencilAttachment-08965", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pStencilAttachment-11860", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pStencilAttachment-11873", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pStrides-04913", 53}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-pipelineFragmentShadingRate-09238", 72}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveFragmentShadingRateWithMultipleViewports-04552", 94}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveFragmentShadingRateWithMultipleViewports-08642", 94}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveRestartIndex-12401", 66}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveTopology-10286", 62}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveTopology-10747", 62}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitiveTopology-10748", 62}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitivesGeneratedQueryWithNonZeroStreams-06709", 87}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitivesGeneratedQueryWithNonZeroStreams-07481", 87}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-primitivesGeneratedQueryWithRasterizerDiscard-06708", 90}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-protectedNoFault-11455", 61}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-protectedNoFault-11456", 61}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizationSamples-07471", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizationSamples-07489", 65}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizationSamples-07494", 65}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizerDiscardEnable-09236", 68}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizerDiscardEnable-09420", 68}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-rasterizerDiscardEnable-10863", 68}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-reductionMode-09213", 58}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-renderPass-02684", 55}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-renderPass-06198", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-renderpass", 49}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-02689", 66}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07484", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07485", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07486", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07487", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07936", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07937", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsEnable-07938", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsPerPixel-07482", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sampleLocationsPerPixel-07483", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-samples-07472", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-samples-07473", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-scissorCount-03418", 57}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-selectableCubicWeights-09214", 67}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-shadingRateImage-09233", 61}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-shadingRateImage-09234", 61}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sparseImageInt64Atomics-04474", 68}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-sparseImageInt64Atomics-04475", 68}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stage-06481", 50}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stage-07073", 50}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stippledLineEnable-07495", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stippledLineEnable-07496", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stippledLineEnable-07497", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-stippledLineEnable-07498", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-storageBuffers-06936", 59}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-subpass-02685", 52}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-uniformBuffers-06935", 59}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-unnormalizedCoordinates-09635", 68}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewMask-06178", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewType-07752", 53}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-03417", 58}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-03419", 58}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-04137", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-04138", 58}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-04139", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-07492", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-07493", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndexedIndirectCount2KHR-viewportCount-09421", 58}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-Input-07939", 34}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-Input-08734", 34}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-InputAttachmentIndex-09597", 49}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-02691", 33}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-02692", 33}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-02693", 33}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-02721", 33}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04007", 33}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04008", 33}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04875", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04876", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04877", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04879", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-04914", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-06479", 33}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-06537", 33}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-06666", 33}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-06886", 33}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-06887", 33}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07288", 33}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07469", 33}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07619", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07620", 33}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07621", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07622", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07623", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07624", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07625", 33}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07626", 33}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07627", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07629", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07630", 33}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07631", 33}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07632", 33}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07633", 33}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07634", 33}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07636", 33}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07639", 33}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07640", 33}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07641", 33}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07642", 33}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07643", 33}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07644", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07645", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07646", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07647", 33}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07648", 33}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07649", 33}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07749", 33}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07751", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07831", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07832", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07834", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07835", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07836", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07837", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07838", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07839", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07840", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07841", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07842", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07843", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07844", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07845", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07846", 33}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07847", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07848", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07849", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07850", 33}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07878", 33}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07879", 33}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07880", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07881", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-07888", 33}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08114", 33}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08115", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08116", 33}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08117", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08119", 33}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08600", 33}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08601", 33}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08605", 33}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08606", 33}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08608", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08609", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08610", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08611", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08612", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08613", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08617", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08636", 33}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08637", 33}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08644", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08666", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08669", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08684", 33}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08685", 33}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08686", 33}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08687", 33}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08688", 33}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08689", 33}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08690", 33}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08693", 33}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08696", 33}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08698", 33}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08699", 33}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08876", 33}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08877", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08878", 33}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08879", 33}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-08885", 33}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09003", 33}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09116", 33}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09203", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09211", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09363", 33}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09364", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09365", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09368", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09369", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09548", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09549", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09600", 33}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09637", 33}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09642", 33}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09643", 33}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-09650", 33}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10068", 33}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10608", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10677", 33}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10678", 33}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10679", 33}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10772", 33}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10795", 33}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10796", 33}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10797", 33}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10862", 33}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10864", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10909", 33}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10927", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-10928", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11297", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11298", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11299", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11300", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11301", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11302", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11304", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11305", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11306", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11308", 33}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11309", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11372", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11373", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11374", 33}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11376", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11397", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11398", 33}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11437", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11438", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11439", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11441", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11442", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11485", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11522", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11523", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11861", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11862", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11863", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11865", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11866", 33}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11868", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-11869", 33}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-12338", 33}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-12339", 33}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-12340", 33}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-12472", 33}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-None-13118", 33}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpExecutionMode-12239", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpExecutionMode-12240", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpExecutionMode-12241", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpExecutionMode-12242", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchSADQCOM-06976", 53}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchSADQCOM-12420", 53}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchSSDQCOM-06974", 53}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchWindow-09215", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchWindow-09216", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchWindow-09217", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBlockMatchWindow-12421", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageBoxFilterQCOM-06973", 49}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageSampleWeightedQCOM-06971", 54}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageSampleWeightedQCOM-06972", 54}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageSampleWeightedQCOM-06977", 54}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageSampleWeightedQCOM-06978", 54}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageWrite-04469", 41}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageWrite-08795", 41}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpImageWrite-08796", 41}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTensorReadARM-12462", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeImage-07027", 40}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeImage-07028", 40}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeImage-07029", 40}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeImage-07030", 40}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeImage-07468", 40}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-OpTypeTensorARM-09906", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-SampledType-04470", 40}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-SampledType-04471", 40}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-SampledType-04472", 40}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-SampledType-04473", 40}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-VkPipelineVieportCreateInfo-04141", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-VkPipelineVieportCreateInfo-04142", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-addressModeU-12422", 41}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-advancedBlendMaxColorAttachments-07480", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-alphaToCoverageEnable-08919", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-alphaToCoverageEnable-08920", 50}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-attachmentCount-07750", 44}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-blendEnable-04727", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-buffer-02708", 35}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-buffer-02709", 35}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-buffer-parameter", 39}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-colorAttachmentCount-06179", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-colorAttachmentCount-06185", 49}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-colorAttachmentCount-09362", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-colorAttachmentCount-09372", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-colorAttachmentCount-11871", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-02707", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-02711", 42}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-10746", 42}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-commonparent", 35}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-conservativePointAndLineRasterization-07499", 66}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-coverageModulationTableEnable-07488", 58}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-coverageReductionMode-07491", 50}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-coverageToColorEnable-07490", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-cubicRangeClamp-09212", 44}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-customResolve-11524", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-customResolve-11525", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-customResolve-11529", 42}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-customResolve-11530", 42}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dimensionCount-09905", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-drawCount-00476", 38}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawIndirectCommand)", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-drawCount-00487", 38}, { {"If drawCount is equal to 1, (offset + sizeof(VkDrawIndirectCommand)) must be less than or equal to the size of buffer", 117}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-drawCount-00488", 38}, { {"If drawCount is greater than 1, (stride x (drawCount - 1) + offset + sizeof(VkDrawIndirectCommand)) must be less than or equal to the size of buffer", 148}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-drawCount-02718", 38}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-drawCount-02719", 38}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicPrimitiveTopologyUnrestricted-07500", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingLocalRead-11797", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08910", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08911", 62}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08912", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08913", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08914", 62}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08915", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08916", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08917", 62}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-08918", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-11864", 62}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-11867", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-dynamicRenderingUnusedAttachments-11870", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-filterCubic-02694", 40}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-filterCubicMinmax-02695", 46}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-firstAttachment-07476", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-firstAttachment-07478", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-02696", 34}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-10582", 34}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-11521", 34}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-13361", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-13362", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-13363", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-13364", 34}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-flags-13365", 34}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-format-07753", 35}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-format-08936", 35}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-format-08937", 35}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-format-10389", 35}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-format-10390", 35}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageLayout-00344", 40}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06172", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06173", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06174", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06175", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06176", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06177", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06183", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-imageView-06184", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-index-11450", 34}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-layers-10831", 35}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-logicOp-04878", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-magFilter-04553", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-magFilter-09598", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-maxFragmentDualSrcAttachments-09239", 58}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-maxMultiviewInstanceIndex-02688", 54}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-micromap-11636", 37}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-micromap-11637", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-micromap-11638", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-micromap-11639", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-mipmapMode-04770", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-mipmapMode-09599", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-multisampledRenderToSingleSampled-07285", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-multisampledRenderToSingleSampled-07286", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-multisampledRenderToSingleSampled-07287", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-multiviewPerViewViewports-12262", 54}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-multiviewPerViewViewports-12263", 54}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-nextStage-10745", 38}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-offset-02710", 35}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pBindInfo-11375", 38}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pColorAttachments-08963", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pColorAttachments-11539", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthAttachment-06186", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthAttachment-08964", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthAttachment-11540", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthAttachment-11872", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthInputAttachmentIndex-09595", 55}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDepthInputAttachmentIndex-09596", 55}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDescription-09900", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDynamicStates-08715", 43}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pDynamicStates-08716", 43}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pNext-07935", 34}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pNext-12463", 34}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pStencilAttachment-06187", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pStencilAttachment-08965", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pStencilAttachment-11860", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pStencilAttachment-11873", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pStrides-04913", 37}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-pipelineFragmentShadingRate-09238", 56}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitiveFragmentShadingRateWithMultipleViewports-04552", 78}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitiveFragmentShadingRateWithMultipleViewports-08642", 78}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitiveTopology-10286", 46}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitiveTopology-10747", 46}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitiveTopology-10748", 46}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitivesGeneratedQueryWithNonZeroStreams-06709", 71}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitivesGeneratedQueryWithNonZeroStreams-07481", 71}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-primitivesGeneratedQueryWithRasterizerDiscard-06708", 74}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-protectedNoFault-11455", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-protectedNoFault-11456", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizationSamples-07471", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizationSamples-07489", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizationSamples-07494", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizerDiscardEnable-09236", 52}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizerDiscardEnable-09420", 52}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-rasterizerDiscardEnable-10863", 52}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-reductionMode-09213", 42}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-renderPass-02684", 39}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-renderPass-06198", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-renderpass", 33}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-02689", 50}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07484", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07485", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07486", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07487", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07936", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07937", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsEnable-07938", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsPerPixel-07482", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sampleLocationsPerPixel-07483", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-samples-07472", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-samples-07473", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-scissorCount-03418", 41}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-selectableCubicWeights-09214", 51}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-shadingRateImage-09233", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-shadingRateImage-09234", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sparseImageInt64Atomics-04474", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-sparseImageInt64Atomics-04475", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stage-06481", 34}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stage-07073", 34}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stippledLineEnable-07495", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stippledLineEnable-07496", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stippledLineEnable-07497", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-stippledLineEnable-07498", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-storageBuffers-06936", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-subpass-02685", 36}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-uniformBuffers-06935", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-unnormalizedCoordinates-09635", 52}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewMask-06178", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewType-07752", 37}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-03417", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-03419", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-04137", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-04138", 42}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-04139", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-07492", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-07493", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect-viewportCount-09421", 42}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-Input-07939", 38}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-Input-08734", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-InputAttachmentIndex-09597", 53}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-02691", 37}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-02692", 37}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-02693", 37}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-02721", 37}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04007", 37}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04008", 37}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04875", 37}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04876", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04877", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04879", 37}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-04914", 37}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-06479", 37}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-06537", 37}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-06666", 37}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-06886", 37}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-06887", 37}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07288", 37}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07469", 37}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07619", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07620", 37}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07621", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07622", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07623", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07624", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07625", 37}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07626", 37}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07627", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07629", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07630", 37}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07631", 37}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07632", 37}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07633", 37}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07634", 37}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07636", 37}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07639", 37}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07640", 37}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07641", 37}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07642", 37}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07643", 37}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07644", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07645", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07646", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07647", 37}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07648", 37}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07649", 37}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07749", 37}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07751", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07831", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07832", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07834", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07835", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07836", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07837", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07838", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07839", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07840", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07841", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07842", 37}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07843", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07844", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07845", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07846", 37}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07847", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07848", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07849", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07850", 37}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07878", 37}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07879", 37}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07880", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07881", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-07888", 37}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08114", 37}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08115", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08116", 37}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08117", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08119", 37}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08600", 37}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08601", 37}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08605", 37}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08606", 37}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08608", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08609", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08610", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08611", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08612", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08613", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08617", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08636", 37}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08637", 37}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08644", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08666", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08669", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08684", 37}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08685", 37}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08686", 37}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08687", 37}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08688", 37}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08689", 37}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08690", 37}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08693", 37}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08696", 37}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08698", 37}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08699", 37}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08876", 37}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08877", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08878", 37}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08879", 37}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-08885", 37}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09003", 37}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09116", 37}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09203", 37}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09211", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09363", 37}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09364", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09365", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09368", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09369", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09548", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09549", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09600", 37}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09637", 37}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09642", 37}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09643", 37}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-09650", 37}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10068", 37}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10608", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10677", 37}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10678", 37}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10679", 37}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10772", 37}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10795", 37}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10796", 37}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10797", 37}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10862", 37}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10864", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10909", 37}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10927", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-10928", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11297", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11298", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11299", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11300", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11301", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11302", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11304", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11305", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11306", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11308", 37}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11309", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11372", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11373", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11374", 37}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11376", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11397", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11398", 37}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11437", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11438", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11439", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11441", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11442", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11485", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11522", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11523", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11861", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11862", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11863", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11865", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11866", 37}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11868", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-11869", 37}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-12338", 37}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-12339", 37}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-12340", 37}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-12472", 37}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-None-13118", 37}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpExecutionMode-12239", 48}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpExecutionMode-12240", 48}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpExecutionMode-12241", 48}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpExecutionMode-12242", 48}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchSADQCOM-06976", 57}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchSADQCOM-12420", 57}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchSSDQCOM-06974", 57}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchWindow-09215", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchWindow-09216", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchWindow-09217", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBlockMatchWindow-12421", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageBoxFilterQCOM-06973", 53}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageSampleWeightedQCOM-06971", 58}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageSampleWeightedQCOM-06972", 58}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageSampleWeightedQCOM-06977", 58}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageSampleWeightedQCOM-06978", 58}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageWrite-04469", 45}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageWrite-08795", 45}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpImageWrite-08796", 45}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTensorReadARM-12462", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeImage-07027", 44}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeImage-07028", 44}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeImage-07029", 44}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeImage-07030", 44}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeImage-07468", 44}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-OpTypeTensorARM-09906", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-SampledType-04470", 44}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-SampledType-04471", 44}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-SampledType-04472", 44}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-SampledType-04473", 44}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-VkPipelineVieportCreateInfo-04141", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-VkPipelineVieportCreateInfo-04142", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-addressModeU-12422", 45}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-advancedBlendMaxColorAttachments-07480", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-alphaToCoverageEnable-08919", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-alphaToCoverageEnable-08920", 54}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-attachmentCount-07750", 48}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-blendEnable-04727", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-colorAttachmentCount-06179", 53}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-colorAttachmentCount-06185", 53}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-colorAttachmentCount-09362", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-colorAttachmentCount-09372", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-colorAttachmentCount-11871", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-02707", 46}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-10746", 46}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-13057", 46}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-conservativePointAndLineRasterization-07499", 70}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-coverageModulationTableEnable-07488", 62}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-coverageReductionMode-07491", 54}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-coverageToColorEnable-07490", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-cubicRangeClamp-09212", 48}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-customResolve-11524", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-customResolve-11525", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-customResolve-11529", 46}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-customResolve-11530", 46}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dimensionCount-09905", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicPrimitiveTopologyUnrestricted-07500", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingLocalRead-11797", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08910", 66}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08911", 66}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08912", 66}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08913", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08914", 66}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08915", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08916", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08917", 66}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-08918", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-11864", 66}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-11867", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-dynamicRenderingUnusedAttachments-11870", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-filterCubic-02694", 44}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-filterCubicMinmax-02695", 50}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-firstAttachment-07476", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-firstAttachment-07478", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-02696", 38}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-10582", 38}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-11521", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-13361", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-13362", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-13363", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-13364", 38}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-flags-13365", 38}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-format-07753", 39}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-format-08936", 39}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-format-08937", 39}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-format-10389", 39}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-format-10390", 39}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageLayout-00344", 44}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06172", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06173", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06174", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06175", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06176", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06177", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06183", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-imageView-06184", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-index-11450", 38}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-layers-10831", 39}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-logicOp-04878", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-magFilter-04553", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-magFilter-09598", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-maxFragmentDualSrcAttachments-09239", 62}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-maxMultiviewInstanceIndex-02688", 58}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-micromap-11636", 41}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-micromap-11637", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-micromap-11638", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-micromap-11639", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-mipmapMode-04770", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-mipmapMode-09599", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-multisampledRenderToSingleSampled-07285", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-multisampledRenderToSingleSampled-07286", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-multisampledRenderToSingleSampled-07287", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-multiviewPerViewViewports-12262", 58}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-multiviewPerViewViewports-12263", 58}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-nextStage-10745", 42}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pBindInfo-11375", 42}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pColorAttachments-08963", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pColorAttachments-11539", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthAttachment-06186", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthAttachment-08964", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthAttachment-11540", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthAttachment-11872", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthInputAttachmentIndex-09595", 59}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDepthInputAttachmentIndex-09596", 59}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDescription-09900", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDynamicStates-08715", 47}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pDynamicStates-08716", 47}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pInfo-13110", 38}, { {"If pInfo->drawCount is greater than 0, pInfo->addressRange.size must be greater than or equal to (pInfo->drawCount - 1) x pInfo->addressRange.stride + sizeof(VkDrawIndirectCommand)", 180}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pInfo-13111", 38}, { {"If pInfo->drawCount is greater than 1, pInfo->addressRange.stride must not be 0", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pInfo-13112", 38}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be greater than or equal to sizeof(VkDrawIndirectCommand)", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pInfo-13113", 38}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be a multiple of 4", 90}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pInfo-parameter", 42}, { {"pInfo must be a valid pointer to a valid VkDrawIndirect2InfoKHR structure", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pNext-07935", 38}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pNext-12463", 38}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pStencilAttachment-06187", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pStencilAttachment-08965", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pStencilAttachment-11860", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pStencilAttachment-11873", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pStrides-04913", 41}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-pipelineFragmentShadingRate-09238", 60}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitiveFragmentShadingRateWithMultipleViewports-04552", 82}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitiveFragmentShadingRateWithMultipleViewports-08642", 82}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitiveTopology-10286", 50}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitiveTopology-10747", 50}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitiveTopology-10748", 50}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitivesGeneratedQueryWithNonZeroStreams-06709", 75}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitivesGeneratedQueryWithNonZeroStreams-07481", 75}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-primitivesGeneratedQueryWithRasterizerDiscard-06708", 78}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-protectedNoFault-11455", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-protectedNoFault-11456", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizationSamples-07471", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizationSamples-07489", 53}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizationSamples-07494", 53}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizerDiscardEnable-09236", 56}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizerDiscardEnable-09420", 56}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-rasterizerDiscardEnable-10863", 56}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-reductionMode-09213", 46}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-renderPass-02684", 43}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-renderPass-06198", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-renderpass", 37}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-02689", 54}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07484", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07485", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07486", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07487", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07936", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07937", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsEnable-07938", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsPerPixel-07482", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sampleLocationsPerPixel-07483", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-samples-07472", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-samples-07473", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-scissorCount-03418", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-selectableCubicWeights-09214", 55}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-shadingRateImage-09233", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-shadingRateImage-09234", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sparseImageInt64Atomics-04474", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-sparseImageInt64Atomics-04475", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stage-06481", 38}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stage-07073", 38}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stippledLineEnable-07495", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stippledLineEnable-07496", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stippledLineEnable-07497", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-stippledLineEnable-07498", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-storageBuffers-06936", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-subpass-02685", 40}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-uniformBuffers-06935", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-unnormalizedCoordinates-09635", 56}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewMask-06178", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewType-07752", 41}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-03417", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-03419", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-04137", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-04138", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-04139", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-07492", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-07493", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirect2KHR-viewportCount-09421", 46}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-Input-07939", 47}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-Input-08734", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-InputAttachmentIndex-09597", 62}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-02691", 46}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-02692", 46}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-02693", 46}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-02721", 46}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04007", 46}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04008", 46}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04875", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04876", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04877", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04879", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-04914", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-06479", 46}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-06537", 46}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-06666", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-06886", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-06887", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07288", 46}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07469", 46}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07619", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07620", 46}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07621", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07622", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07623", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07624", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07625", 46}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07626", 46}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07627", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07629", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07630", 46}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07631", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07632", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07633", 46}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07634", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07636", 46}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07639", 46}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07640", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07641", 46}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07642", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07643", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07644", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07645", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07646", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07647", 46}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07648", 46}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07649", 46}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07749", 46}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07751", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07831", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07832", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07834", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07835", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07836", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07837", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07838", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07839", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07840", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07841", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07842", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07843", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07844", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07845", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07846", 46}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07847", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07848", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07849", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07850", 46}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07878", 46}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07879", 46}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07880", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07881", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-07888", 46}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08114", 46}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08115", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08116", 46}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08117", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08119", 46}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08600", 46}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08601", 46}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08605", 46}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08606", 46}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08608", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08609", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08610", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08611", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08612", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08613", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08617", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08636", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08637", 46}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08644", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08666", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08669", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08684", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08685", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08686", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08687", 46}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08688", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08689", 46}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08690", 46}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08693", 46}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08696", 46}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08698", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08699", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08876", 46}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08877", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08878", 46}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08879", 46}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-08885", 46}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09003", 46}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09116", 46}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09203", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09211", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09363", 46}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09364", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09365", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09368", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09369", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09462", 46}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09548", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09549", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09600", 46}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09637", 46}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09642", 46}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09643", 46}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-09650", 46}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10068", 46}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10608", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10677", 46}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10678", 46}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10679", 46}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10772", 46}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10795", 46}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10796", 46}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10797", 46}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10862", 46}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10864", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10909", 46}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10927", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-10928", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11297", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11298", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11299", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11300", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11301", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11302", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11304", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11305", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11306", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11308", 46}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11309", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11372", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11373", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11374", 46}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11376", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11397", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11398", 46}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11437", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11438", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11439", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11441", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11442", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11485", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11522", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11523", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11861", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11862", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11863", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11865", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11866", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11868", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-11869", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-12338", 46}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-12339", 46}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-12340", 46}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-12472", 46}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-None-13118", 46}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpExecutionMode-12239", 57}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpExecutionMode-12240", 57}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpExecutionMode-12241", 57}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpExecutionMode-12242", 57}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchSADQCOM-06976", 66}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchSADQCOM-12420", 66}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchSSDQCOM-06974", 66}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchWindow-09215", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchWindow-09216", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchWindow-09217", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBlockMatchWindow-12421", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageBoxFilterQCOM-06973", 62}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageSampleWeightedQCOM-06971", 67}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageSampleWeightedQCOM-06972", 67}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageSampleWeightedQCOM-06977", 67}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageSampleWeightedQCOM-06978", 67}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageWrite-04469", 54}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageWrite-08795", 54}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpImageWrite-08796", 54}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTensorReadARM-12462", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeImage-07027", 53}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeImage-07028", 53}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeImage-07029", 53}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeImage-07030", 53}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeImage-07468", 53}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-OpTypeTensorARM-09906", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-SampledType-04470", 53}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-SampledType-04471", 53}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-SampledType-04472", 53}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-SampledType-04473", 53}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-VkPipelineVieportCreateInfo-04141", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-VkPipelineVieportCreateInfo-04142", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-addressModeU-12422", 54}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-advancedBlendMaxColorAttachments-07480", 74}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-alphaToCoverageEnable-08919", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-alphaToCoverageEnable-08920", 63}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-attachmentCount-07750", 57}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-blendEnable-04727", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-colorAttachmentCount-06179", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-colorAttachmentCount-06185", 62}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-colorAttachmentCount-09362", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-colorAttachmentCount-09372", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-colorAttachmentCount-11871", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-02646", 55}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-02707", 55}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-10746", 55}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-conservativePointAndLineRasterization-07499", 79}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-counterOffset-09474", 55}, { {"counterOffset must be a multiple of 4", 37}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-coverageModulationTableEnable-07488", 71}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-coverageReductionMode-07491", 63}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-coverageToColorEnable-07490", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-cubicRangeClamp-09212", 57}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-customResolve-11524", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-customResolve-11525", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-customResolve-11529", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-customResolve-11530", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dimensionCount-09905", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicPrimitiveTopologyUnrestricted-07500", 78}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingLocalRead-11797", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08910", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08911", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08912", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08913", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08914", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08915", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08916", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08917", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-08918", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-11864", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-11867", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-dynamicRenderingUnusedAttachments-11870", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-filterCubic-02694", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-filterCubicMinmax-02695", 59}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-firstAttachment-07476", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-firstAttachment-07478", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-02696", 47}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-10582", 47}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-11521", 47}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-13361", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-13362", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-13363", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-13364", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-flags-13365", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-format-07753", 48}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-format-08936", 48}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-format-08937", 48}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-format-10389", 48}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-format-10390", 48}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageLayout-00344", 53}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06172", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06173", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06174", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06175", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06176", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06177", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06183", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-imageView-06184", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-index-11450", 47}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-layers-10831", 48}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-logicOp-04878", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-magFilter-04553", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-magFilter-09598", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-maxFragmentDualSrcAttachments-09239", 71}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-maxMultiviewInstanceIndex-02688", 67}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-micromap-11636", 50}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-micromap-11637", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-micromap-11638", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-micromap-11639", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-mipmapMode-04770", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-mipmapMode-09599", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-multisampledRenderToSingleSampled-07285", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-multisampledRenderToSingleSampled-07286", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-multisampledRenderToSingleSampled-07287", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-multiviewPerViewViewports-12262", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-multiviewPerViewViewports-12263", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-nextStage-10745", 51}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pBindInfo-11375", 51}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pColorAttachments-08963", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pColorAttachments-11539", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pCounterInfo-parameter", 58}, { {"pCounterInfo must be a valid pointer to a valid VkBindTransformFeedbackBuffer2InfoEXT structure", 95}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthAttachment-06186", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthAttachment-08964", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthAttachment-11540", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthAttachment-11872", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthInputAttachmentIndex-09595", 68}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDepthInputAttachmentIndex-09596", 68}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDescription-09900", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDynamicStates-08715", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pDynamicStates-08716", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pInfo-13061", 47}, { {"The VkBuffer that pInfo->counterRange was queried from must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT bit set", 130}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pInfo-13062", 47}, { {"pInfo->counterRange.address must be a multiple of 4", 51}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pNext-07935", 47}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pNext-09461", 47}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pNext-12463", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pStencilAttachment-06187", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pStencilAttachment-08965", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pStencilAttachment-11860", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pStencilAttachment-11873", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pStrides-04913", 50}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-pipelineFragmentShadingRate-09238", 69}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitiveTopology-10286", 59}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitiveTopology-10747", 59}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitiveTopology-10748", 59}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 87}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-protectedNoFault-11455", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-protectedNoFault-11456", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizationSamples-07471", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizationSamples-07489", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizationSamples-07494", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizerDiscardEnable-09236", 65}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizerDiscardEnable-09420", 65}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-rasterizerDiscardEnable-10863", 65}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-reductionMode-09213", 55}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-renderPass-02684", 52}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-renderPass-06198", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-renderpass", 46}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-02689", 63}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07484", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07485", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07486", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07487", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07936", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07937", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsEnable-07938", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsPerPixel-07482", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sampleLocationsPerPixel-07483", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-samples-07472", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-samples-07473", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-scissorCount-03418", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-selectableCubicWeights-09214", 64}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-shadingRateImage-09233", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-shadingRateImage-09234", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sparseImageInt64Atomics-04474", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-sparseImageInt64Atomics-04475", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stage-06481", 47}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stage-07073", 47}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stippledLineEnable-07495", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stippledLineEnable-07496", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stippledLineEnable-07497", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-stippledLineEnable-07498", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-storageBuffers-06936", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-subpass-02685", 49}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-suspended", 45}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-transformFeedback-02287", 59}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-transformFeedbackDraw-02288", 63}, { {"The implementation must support VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackDraw", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-uniformBuffers-06935", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-unnormalizedCoordinates-09635", 65}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-vertexStride-02289", 54}, { {"vertexStride must be greater than 0 and less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBufferDataStride", 146}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-vertexStride-09475", 54}, { {"vertexStride must be a multiple of 4", 36}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewMask-06178", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewType-07752", 50}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-03417", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-03419", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-04137", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-04138", 55}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-04139", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-07492", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-07493", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCount2EXT-viewportCount-09421", 55}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-Input-07939", 46}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-Input-08734", 46}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-InputAttachmentIndex-09597", 61}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-02691", 45}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-02692", 45}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-02693", 45}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-02721", 45}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04007", 45}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04008", 45}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04875", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04876", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04877", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04879", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-04914", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-06479", 45}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-06537", 45}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-06666", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-06886", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-06887", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07288", 45}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07469", 45}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07619", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07620", 45}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07621", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07622", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07623", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07624", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07625", 45}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07626", 45}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07627", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07629", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07630", 45}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07631", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07632", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07633", 45}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07634", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07636", 45}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07639", 45}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07640", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07641", 45}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07642", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07643", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07644", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07645", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07646", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07647", 45}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07648", 45}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07649", 45}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07749", 45}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07751", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07831", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07832", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07834", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07835", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07836", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07837", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07838", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07839", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07840", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07841", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07842", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07843", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07844", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07845", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07846", 45}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07847", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07848", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07849", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07850", 45}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07878", 45}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07879", 45}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07880", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07881", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-07888", 45}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08114", 45}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08115", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08116", 45}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08117", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08119", 45}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08600", 45}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08601", 45}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08605", 45}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08606", 45}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08608", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08609", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08610", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08611", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08612", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08613", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08617", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08636", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08637", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08644", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08666", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08669", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08684", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08685", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08686", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08687", 45}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08688", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08689", 45}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08690", 45}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08693", 45}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08696", 45}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08698", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08699", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08876", 45}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08877", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08878", 45}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08879", 45}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-08885", 45}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09003", 45}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09116", 45}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09203", 45}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09211", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09363", 45}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09364", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09365", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09368", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09369", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09462", 45}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09548", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09549", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09600", 45}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09637", 45}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09642", 45}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09643", 45}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-09650", 45}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10068", 45}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10608", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10677", 45}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10678", 45}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10679", 45}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10772", 45}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10795", 45}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10796", 45}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10797", 45}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10862", 45}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10864", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10909", 45}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10927", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-10928", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11297", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11298", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11299", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11300", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11301", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11302", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11304", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11305", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11306", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11308", 45}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11309", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11372", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11373", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11374", 45}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11376", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11397", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11398", 45}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11437", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11438", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11439", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11441", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11442", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11485", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11522", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11523", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11861", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11862", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11863", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11865", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11866", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11868", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-11869", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-12338", 45}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-12339", 45}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-12340", 45}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-12472", 45}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-None-13118", 45}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpExecutionMode-12239", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpExecutionMode-12240", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpExecutionMode-12241", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpExecutionMode-12242", 56}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchSADQCOM-06976", 65}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchSADQCOM-12420", 65}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchSSDQCOM-06974", 65}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchWindow-09215", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchWindow-09216", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchWindow-09217", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBlockMatchWindow-12421", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageBoxFilterQCOM-06973", 61}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageSampleWeightedQCOM-06971", 66}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageSampleWeightedQCOM-06972", 66}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageSampleWeightedQCOM-06977", 66}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageSampleWeightedQCOM-06978", 66}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageWrite-04469", 53}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageWrite-08795", 53}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpImageWrite-08796", 53}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTensorReadARM-12462", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeImage-07027", 52}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeImage-07028", 52}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeImage-07029", 52}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeImage-07030", 52}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeImage-07468", 52}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-OpTypeTensorARM-09906", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-SampledType-04470", 52}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-SampledType-04471", 52}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-SampledType-04472", 52}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-SampledType-04473", 52}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-VkPipelineVieportCreateInfo-04141", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-VkPipelineVieportCreateInfo-04142", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-addressModeU-12422", 53}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-advancedBlendMaxColorAttachments-07480", 73}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-alphaToCoverageEnable-08919", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-alphaToCoverageEnable-08920", 62}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-attachmentCount-07750", 56}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-blendEnable-04727", 52}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-colorAttachmentCount-06179", 61}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-colorAttachmentCount-06185", 61}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-colorAttachmentCount-09362", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-colorAttachmentCount-09372", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-colorAttachmentCount-11871", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-02646", 54}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-02707", 54}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-10746", 54}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-commonparent", 47}, { {"Both of commandBuffer, and counterBuffer must have been created, allocated, or retrieved from the same VkDevice", 111}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-conservativePointAndLineRasterization-07499", 78}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-counterBuffer-02290", 54}, { {"counterBuffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 96}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-counterBuffer-04567", 54}, { {"If counterBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 114}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-counterBuffer-parameter", 58}, { {"counterBuffer must be a valid VkBuffer handle", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-counterBufferOffset-04568", 60}, { {"counterBufferOffset must be a multiple of 4", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-counterOffset-09474", 54}, { {"counterOffset must be a multiple of 4", 37}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-coverageModulationTableEnable-07488", 70}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-coverageReductionMode-07491", 62}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-coverageToColorEnable-07490", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-cubicRangeClamp-09212", 56}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-customResolve-11524", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-customResolve-11525", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-customResolve-11529", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-customResolve-11530", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dimensionCount-09905", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicPrimitiveTopologyUnrestricted-07500", 77}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingLocalRead-11797", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08910", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08911", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08912", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08913", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08914", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08915", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08916", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08917", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-08918", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-11864", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-11867", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-dynamicRenderingUnusedAttachments-11870", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-filterCubic-02694", 52}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-filterCubicMinmax-02695", 58}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-firstAttachment-07476", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-firstAttachment-07478", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-02696", 46}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-10582", 46}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-11521", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-13361", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-13362", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-13363", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-13364", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-flags-13365", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-format-07753", 47}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-format-08936", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-format-08937", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-format-10389", 47}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-format-10390", 47}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageLayout-00344", 52}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06172", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06173", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06174", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06175", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06176", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06177", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06183", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-imageView-06184", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-index-11450", 46}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-layers-10831", 47}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-logicOp-04878", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-magFilter-04553", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-magFilter-09598", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-maxFragmentDualSrcAttachments-09239", 70}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-maxMultiviewInstanceIndex-02688", 66}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-micromap-11636", 49}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-micromap-11637", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-micromap-11638", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-micromap-11639", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-mipmapMode-04770", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-mipmapMode-09599", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-multisampledRenderToSingleSampled-07285", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-multisampledRenderToSingleSampled-07286", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-multisampledRenderToSingleSampled-07287", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-multiviewPerViewViewports-12262", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-multiviewPerViewViewports-12263", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-nextStage-10745", 50}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pBindInfo-11375", 50}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pColorAttachments-08963", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pColorAttachments-11539", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthAttachment-06186", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthAttachment-08964", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthAttachment-11540", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthAttachment-11872", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthInputAttachmentIndex-09595", 67}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDepthInputAttachmentIndex-09596", 67}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDescription-09900", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDynamicStates-08715", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pDynamicStates-08716", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pNext-07935", 46}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pNext-09461", 46}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pNext-12463", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pStencilAttachment-06187", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pStencilAttachment-08965", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pStencilAttachment-11860", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pStencilAttachment-11873", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pStrides-04913", 49}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-pipelineFragmentShadingRate-09238", 68}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitiveTopology-10286", 58}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitiveTopology-10747", 58}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitiveTopology-10748", 58}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 86}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-protectedNoFault-11455", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-protectedNoFault-11456", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizationSamples-07471", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizationSamples-07489", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizationSamples-07494", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizerDiscardEnable-09236", 64}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizerDiscardEnable-09420", 64}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-rasterizerDiscardEnable-10863", 64}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-reductionMode-09213", 54}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-renderPass-02684", 51}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-renderPass-06198", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-renderpass", 45}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-02689", 62}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07484", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07485", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07486", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07487", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07936", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07937", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsEnable-07938", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsPerPixel-07482", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sampleLocationsPerPixel-07483", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-samples-07472", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-samples-07473", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-scissorCount-03418", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-selectableCubicWeights-09214", 63}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-shadingRateImage-09233", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-shadingRateImage-09234", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sparseImageInt64Atomics-04474", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-sparseImageInt64Atomics-04475", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stage-06481", 46}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stage-07073", 46}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stippledLineEnable-07495", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stippledLineEnable-07496", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stippledLineEnable-07497", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-stippledLineEnable-07498", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-storageBuffers-06936", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-subpass-02685", 48}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-suspended", 44}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-transformFeedback-02287", 58}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-transformFeedbackDraw-02288", 62}, { {"The implementation must support VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackDraw", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-uniformBuffers-06935", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-unnormalizedCoordinates-09635", 64}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-vertexStride-02289", 53}, { {"vertexStride must be greater than 0 and less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBufferDataStride", 146}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-vertexStride-09475", 53}, { {"vertexStride must be a multiple of 4", 36}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewMask-06178", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewType-07752", 49}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-03417", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-03419", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-04137", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-04138", 54}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-04139", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-07492", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-07493", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectByteCountEXT-viewportCount-09421", 54}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-Input-07939", 39}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-Input-08734", 39}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-InputAttachmentIndex-09597", 54}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-02691", 38}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-02692", 38}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-02693", 38}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-02721", 38}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04007", 38}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04008", 38}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04445", 38}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04875", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04876", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04877", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04879", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-04914", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-06479", 38}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-06537", 38}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-06666", 38}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-06886", 38}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-06887", 38}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07288", 38}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07469", 38}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07619", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07620", 38}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07621", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07622", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07623", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07624", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07625", 38}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07626", 38}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07627", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07629", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07630", 38}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07631", 38}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07632", 38}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07633", 38}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07634", 38}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07636", 38}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07639", 38}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07640", 38}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07641", 38}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07642", 38}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07643", 38}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07644", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07645", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07646", 38}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07647", 38}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07648", 38}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07649", 38}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07749", 38}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07751", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07831", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07832", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07834", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07835", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07836", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07837", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07838", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07839", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07840", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07841", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07842", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07843", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07844", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07845", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07846", 38}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07847", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07848", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07849", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07850", 38}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07878", 38}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07879", 38}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07880", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07881", 38}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-07888", 38}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08114", 38}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08115", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08116", 38}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08117", 38}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08119", 38}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08600", 38}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08601", 38}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08605", 38}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08606", 38}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08608", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08609", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08610", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08611", 38}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08612", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08613", 38}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08617", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08636", 38}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08637", 38}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08644", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08666", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08669", 38}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08684", 38}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08685", 38}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08686", 38}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08687", 38}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08688", 38}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08689", 38}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08690", 38}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08693", 38}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08696", 38}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08698", 38}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08699", 38}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08876", 38}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08877", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08878", 38}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08879", 38}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-08885", 38}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09003", 38}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09116", 38}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09203", 38}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09211", 38}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09363", 38}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09364", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09365", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09368", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09369", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09548", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09549", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09600", 38}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09637", 38}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09642", 38}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09643", 38}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-09650", 38}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10068", 38}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10608", 38}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10677", 38}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10678", 38}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10679", 38}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10772", 38}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10795", 38}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10796", 38}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10797", 38}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10862", 38}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10864", 38}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10909", 38}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10927", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-10928", 38}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11297", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11298", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11299", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11300", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11301", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11302", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11304", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11305", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11306", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11308", 38}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11309", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11372", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11373", 38}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11374", 38}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11376", 38}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11397", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11398", 38}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11437", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11438", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11439", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11441", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11442", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11485", 38}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11522", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11523", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11861", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11862", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11863", 38}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11865", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11866", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11868", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-11869", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-12338", 38}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-12339", 38}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-12340", 38}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-12472", 38}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-None-13118", 38}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpExecutionMode-12239", 49}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpExecutionMode-12240", 49}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpExecutionMode-12241", 49}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpExecutionMode-12242", 49}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchSADQCOM-06976", 58}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchSADQCOM-12420", 58}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchSSDQCOM-06974", 58}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchWindow-09215", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchWindow-09216", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchWindow-09217", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBlockMatchWindow-12421", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageBoxFilterQCOM-06973", 54}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageSampleWeightedQCOM-06971", 59}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageSampleWeightedQCOM-06972", 59}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageSampleWeightedQCOM-06977", 59}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageSampleWeightedQCOM-06978", 59}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageWrite-04469", 46}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageWrite-08795", 46}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpImageWrite-08796", 46}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTensorReadARM-12462", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeImage-07027", 45}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeImage-07028", 45}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeImage-07029", 45}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeImage-07030", 45}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeImage-07468", 45}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-OpTypeTensorARM-09906", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-SampledType-04470", 45}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-SampledType-04471", 45}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-SampledType-04472", 45}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-SampledType-04473", 45}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-VkPipelineVieportCreateInfo-04141", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-VkPipelineVieportCreateInfo-04142", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-addressModeU-12422", 46}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-advancedBlendMaxColorAttachments-07480", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-alphaToCoverageEnable-08919", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-alphaToCoverageEnable-08920", 55}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-attachmentCount-07750", 49}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-blendEnable-04727", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-buffer-02708", 40}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-buffer-02709", 40}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-buffer-parameter", 44}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-colorAttachmentCount-06179", 54}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-colorAttachmentCount-06185", 54}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-colorAttachmentCount-09362", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-colorAttachmentCount-09372", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-colorAttachmentCount-11871", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-02707", 47}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-02711", 47}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-10746", 47}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-commonparent", 40}, { {"Each of buffer, commandBuffer, and countBuffer must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-conservativePointAndLineRasterization-07499", 71}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBuffer-02714", 45}, { {"If countBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBuffer-02715", 45}, { {"countBuffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBuffer-02717", 45}, { {"The count stored in countBuffer must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 106}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBuffer-parameter", 49}, { {"countBuffer must be a valid VkBuffer handle", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBufferOffset-02716", 51}, { {"countBufferOffset must be a multiple of 4", 41}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-countBufferOffset-04129", 51}, { {"(countBufferOffset + sizeof(uint32_t)) must be less than or equal to the size of countBuffer", 92}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-coverageModulationTableEnable-07488", 63}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-coverageReductionMode-07491", 55}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-coverageToColorEnable-07490", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-cubicRangeClamp-09212", 49}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-customResolve-11524", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-customResolve-11525", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-customResolve-11529", 47}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-customResolve-11530", 47}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dimensionCount-09905", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicPrimitiveTopologyUnrestricted-07500", 70}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingLocalRead-11797", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08910", 67}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08911", 67}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08912", 67}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08913", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08914", 67}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08915", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08916", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08917", 67}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-08918", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-11864", 67}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-11867", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-dynamicRenderingUnusedAttachments-11870", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-filterCubic-02694", 45}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-filterCubicMinmax-02695", 51}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-firstAttachment-07476", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-firstAttachment-07478", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-02696", 39}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-10582", 39}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-11521", 39}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-13361", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-13362", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-13363", 39}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-13364", 39}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-flags-13365", 39}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-format-07753", 40}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-format-08936", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-format-08937", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-format-10389", 40}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-format-10390", 40}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageLayout-00344", 45}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06172", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06173", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06174", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06175", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06176", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06177", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06183", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-imageView-06184", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-index-11450", 39}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-layers-10831", 40}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-logicOp-04878", 41}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-magFilter-04553", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-magFilter-09598", 43}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-maxDrawCount-03111", 46}, { {"If maxDrawCount is greater than or equal to 1, (stride x (maxDrawCount - 1) + offset + sizeof(VkDrawIndirectCommand)) must be less than or equal to the size of buffer", 166}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-maxFragmentDualSrcAttachments-09239", 63}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-maxMultiviewInstanceIndex-02688", 59}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-micromap-11636", 42}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-micromap-11637", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-micromap-11638", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-micromap-11639", 42}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-mipmapMode-04770", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-mipmapMode-09599", 44}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-multisampledRenderToSingleSampled-07285", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-multisampledRenderToSingleSampled-07286", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-multisampledRenderToSingleSampled-07287", 67}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-multiviewPerViewViewports-12262", 59}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-multiviewPerViewViewports-12263", 59}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-nextStage-10745", 43}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-offset-02710", 40}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pBindInfo-11375", 43}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pColorAttachments-08963", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pColorAttachments-11539", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthAttachment-06186", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthAttachment-08964", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthAttachment-11540", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthAttachment-11872", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthInputAttachmentIndex-09595", 60}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDepthInputAttachmentIndex-09596", 60}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDescription-09900", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDynamicStates-08715", 48}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pDynamicStates-08716", 48}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pNext-07935", 39}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pNext-12463", 39}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pStencilAttachment-06187", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pStencilAttachment-08965", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pStencilAttachment-11860", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pStencilAttachment-11873", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pStrides-04913", 42}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-pipelineFragmentShadingRate-09238", 61}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitiveFragmentShadingRateWithMultipleViewports-04552", 83}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitiveFragmentShadingRateWithMultipleViewports-08642", 83}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitiveTopology-10286", 51}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitiveTopology-10747", 51}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitiveTopology-10748", 51}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitivesGeneratedQueryWithNonZeroStreams-06709", 76}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitivesGeneratedQueryWithNonZeroStreams-07481", 76}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-primitivesGeneratedQueryWithRasterizerDiscard-06708", 79}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-protectedNoFault-11455", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-protectedNoFault-11456", 50}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizationSamples-07471", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizationSamples-07489", 54}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizationSamples-07494", 54}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizerDiscardEnable-09236", 57}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizerDiscardEnable-09420", 57}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-rasterizerDiscardEnable-10863", 57}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-reductionMode-09213", 47}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-renderPass-02684", 44}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-renderPass-06198", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-renderpass", 38}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-02689", 55}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07484", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07485", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07486", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07487", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07936", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07937", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsEnable-07938", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsPerPixel-07482", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sampleLocationsPerPixel-07483", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-samples-07472", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-samples-07473", 41}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-scissorCount-03418", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-selectableCubicWeights-09214", 56}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-shadingRateImage-09233", 50}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-shadingRateImage-09234", 50}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sparseImageInt64Atomics-04474", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-sparseImageInt64Atomics-04475", 57}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stage-06481", 39}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stage-07073", 39}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stippledLineEnable-07495", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stippledLineEnable-07496", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stippledLineEnable-07497", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stippledLineEnable-07498", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-storageBuffers-06936", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-stride-03110", 40}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawIndirectCommand)", 97}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-subpass-02685", 41}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-uniformBuffers-06935", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-unnormalizedCoordinates-09635", 57}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewMask-06178", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewType-07752", 42}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-03417", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-03419", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-04137", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-04138", 47}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-04139", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-07492", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-07493", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount-viewportCount-09421", 47}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-Input-07939", 43}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-Input-08734", 43}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-InputAttachmentIndex-09597", 58}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-02691", 42}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-02692", 42}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-02693", 42}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-02721", 42}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04007", 42}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04008", 42}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04875", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04876", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04877", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04879", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-04914", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-06479", 42}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-06537", 42}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-06666", 42}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-06886", 42}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-06887", 42}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07288", 42}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07469", 42}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07619", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07620", 42}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07621", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07622", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07623", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07624", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07625", 42}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07626", 42}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07627", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07629", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07630", 42}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07631", 42}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07632", 42}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07633", 42}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07634", 42}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07636", 42}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07639", 42}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07640", 42}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07641", 42}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07642", 42}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07643", 42}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07644", 42}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07645", 42}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07646", 42}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07647", 42}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07648", 42}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07649", 42}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07749", 42}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07751", 42}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07831", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07832", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07834", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07835", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07836", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07837", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07838", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07839", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07840", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07841", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07842", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07843", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07844", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07845", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07846", 42}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07847", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07848", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07849", 42}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07850", 42}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07878", 42}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07879", 42}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07880", 42}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07881", 42}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-07888", 42}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08114", 42}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08115", 42}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08116", 42}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08117", 42}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08119", 42}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08600", 42}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08601", 42}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08605", 42}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08606", 42}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08608", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08609", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08610", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08611", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08612", 42}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08613", 42}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08617", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08636", 42}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08637", 42}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08644", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08666", 42}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08669", 42}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08684", 42}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08685", 42}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08686", 42}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08687", 42}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08688", 42}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08689", 42}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08690", 42}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08693", 42}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08696", 42}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08698", 42}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08699", 42}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08876", 42}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08877", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08878", 42}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08879", 42}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-08885", 42}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09003", 42}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09116", 42}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09203", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09211", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09363", 42}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09364", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09365", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09368", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09369", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09548", 42}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09549", 42}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09600", 42}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09637", 42}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09642", 42}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09643", 42}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-09650", 42}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10068", 42}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10608", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10677", 42}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10678", 42}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10679", 42}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10772", 42}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10795", 42}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10796", 42}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10797", 42}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10862", 42}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10864", 42}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10909", 42}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10927", 42}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-10928", 42}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11297", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11298", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11299", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11300", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11301", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11302", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11304", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11305", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11306", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11308", 42}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11309", 42}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11372", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11373", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11374", 42}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11376", 42}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11397", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11398", 42}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11437", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11438", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11439", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11441", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11442", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11485", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11522", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11523", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11861", 42}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11862", 42}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11863", 42}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11865", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11866", 42}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11868", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-11869", 42}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-12338", 42}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-12339", 42}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-12340", 42}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-12472", 42}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-None-13118", 42}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpExecutionMode-12239", 53}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpExecutionMode-12240", 53}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpExecutionMode-12241", 53}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpExecutionMode-12242", 53}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchSADQCOM-06976", 62}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchSADQCOM-12420", 62}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchSSDQCOM-06974", 62}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchWindow-09215", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchWindow-09216", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchWindow-09217", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBlockMatchWindow-12421", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageBoxFilterQCOM-06973", 58}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageSampleWeightedQCOM-06971", 63}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageSampleWeightedQCOM-06972", 63}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageSampleWeightedQCOM-06977", 63}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageSampleWeightedQCOM-06978", 63}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageWrite-04469", 50}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageWrite-08795", 50}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpImageWrite-08796", 50}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTensorReadARM-12462", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeImage-07027", 49}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeImage-07028", 49}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeImage-07029", 49}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeImage-07030", 49}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeImage-07468", 49}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-OpTypeTensorARM-09906", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-SampledType-04470", 49}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-SampledType-04471", 49}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-SampledType-04472", 49}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-SampledType-04473", 49}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-VkPipelineVieportCreateInfo-04141", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-VkPipelineVieportCreateInfo-04142", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-addressModeU-12422", 50}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-advancedBlendMaxColorAttachments-07480", 70}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-alphaToCoverageEnable-08919", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-alphaToCoverageEnable-08920", 59}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-attachmentCount-07750", 53}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-blendEnable-04727", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-colorAttachmentCount-06179", 58}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-colorAttachmentCount-06185", 58}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-colorAttachmentCount-09362", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-colorAttachmentCount-09372", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-colorAttachmentCount-11871", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-02707", 51}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-10746", 51}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-13058", 51}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-conservativePointAndLineRasterization-07499", 75}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-coverageModulationTableEnable-07488", 67}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-coverageReductionMode-07491", 59}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-coverageToColorEnable-07490", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-cubicRangeClamp-09212", 53}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-customResolve-11524", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-customResolve-11525", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-customResolve-11529", 51}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-customResolve-11530", 51}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dimensionCount-09905", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-drawIndirectCount-13119", 55}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicPrimitiveTopologyUnrestricted-07500", 74}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingLocalRead-11797", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08910", 71}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08911", 71}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08912", 71}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08913", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08914", 71}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08915", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08916", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08917", 71}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-08918", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-11864", 71}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-11867", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-dynamicRenderingUnusedAttachments-11870", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-filterCubic-02694", 49}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-filterCubicMinmax-02695", 55}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-firstAttachment-07476", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-firstAttachment-07478", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-02696", 43}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-10582", 43}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-11521", 43}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-13361", 43}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-13362", 43}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-13363", 43}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-13364", 43}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-flags-13365", 43}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-format-07753", 44}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-format-08936", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-format-08937", 44}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-format-10389", 44}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-format-10390", 44}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageLayout-00344", 49}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06172", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06173", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06174", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06175", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06176", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06177", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06183", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-imageView-06184", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-index-11450", 43}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-layers-10831", 44}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-logicOp-04878", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-magFilter-04553", 47}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-magFilter-09598", 47}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-maxFragmentDualSrcAttachments-09239", 67}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-maxMultiviewInstanceIndex-02688", 63}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-micromap-11636", 46}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-micromap-11637", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-micromap-11638", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-micromap-11639", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-mipmapMode-04770", 48}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-mipmapMode-09599", 48}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-multisampledRenderToSingleSampled-07285", 71}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-multisampledRenderToSingleSampled-07286", 71}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-multisampledRenderToSingleSampled-07287", 71}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-multiviewPerViewViewports-12262", 63}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-multiviewPerViewViewports-12263", 63}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-nextStage-10745", 47}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pBindInfo-11375", 47}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pColorAttachments-08963", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pColorAttachments-11539", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthAttachment-06186", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthAttachment-08964", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthAttachment-11540", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthAttachment-11872", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthInputAttachmentIndex-09595", 64}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDepthInputAttachmentIndex-09596", 64}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDescription-09900", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDynamicStates-08715", 52}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pDynamicStates-08716", 52}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pInfo-13110", 43}, { {"If pInfo->maxDrawCount is greater than 0, pInfo->addressRange.size must be greater than or equal to (pInfo->maxDrawCount - 1) x pInfo->addressRange.stride + sizeof(VkDrawIndirectCommand)", 186}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pInfo-13111", 43}, { {"If pInfo->maxDrawCount is greater than 1, pInfo->addressRange.stride must not be 0", 82}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pInfo-13112", 43}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be greater than or equal to sizeof(VkDrawIndirectCommand)", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pInfo-13113", 43}, { {"If pInfo->addressRange.stride is not 0, pInfo->addressRange.stride must be a multiple of 4", 90}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pInfo-parameter", 47}, { {"pInfo must be a valid pointer to a valid VkDrawIndirectCount2InfoKHR structure", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pNext-07935", 43}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pNext-12463", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pStencilAttachment-06187", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pStencilAttachment-08965", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pStencilAttachment-11860", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pStencilAttachment-11873", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pStrides-04913", 46}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-pipelineFragmentShadingRate-09238", 65}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitiveFragmentShadingRateWithMultipleViewports-04552", 87}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitiveFragmentShadingRateWithMultipleViewports-08642", 87}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitiveTopology-10286", 55}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitiveTopology-10747", 55}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitiveTopology-10748", 55}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitivesGeneratedQueryWithNonZeroStreams-06709", 80}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitivesGeneratedQueryWithNonZeroStreams-07481", 80}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-primitivesGeneratedQueryWithRasterizerDiscard-06708", 83}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-protectedNoFault-11455", 54}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-protectedNoFault-11456", 54}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizationSamples-07471", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizationSamples-07489", 58}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizationSamples-07494", 58}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizerDiscardEnable-09236", 61}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizerDiscardEnable-09420", 61}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-rasterizerDiscardEnable-10863", 61}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-reductionMode-09213", 51}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-renderPass-02684", 48}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-renderPass-06198", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-renderpass", 42}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-02689", 59}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07484", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07485", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07486", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07487", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07936", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07937", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsEnable-07938", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsPerPixel-07482", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sampleLocationsPerPixel-07483", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-samples-07472", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-samples-07473", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-scissorCount-03418", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-selectableCubicWeights-09214", 60}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-shadingRateImage-09233", 54}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-shadingRateImage-09234", 54}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sparseImageInt64Atomics-04474", 61}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-sparseImageInt64Atomics-04475", 61}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stage-06481", 43}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stage-07073", 43}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stippledLineEnable-07495", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stippledLineEnable-07496", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stippledLineEnable-07497", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-stippledLineEnable-07498", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-storageBuffers-06936", 52}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-subpass-02685", 45}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-uniformBuffers-06935", 52}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-unnormalizedCoordinates-09635", 61}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewMask-06178", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewType-07752", 46}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-03417", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-03419", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-04137", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-04138", 51}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-04139", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-07492", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-07493", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawIndirectCount2KHR-viewportCount-09421", 51}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-InputAttachmentIndex-09597", 53}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-MeshEXT-07087", 40}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshEXT Execution Model", 123}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-02691", 37}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-02692", 37}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-02693", 37}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-04876", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-04877", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-06479", 37}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-06537", 37}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-06666", 37}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-06886", 37}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-06887", 37}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07074", 37}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07075", 37}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07288", 37}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07469", 37}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07620", 37}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07621", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07622", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07623", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07624", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07625", 37}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07626", 37}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07627", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07629", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07630", 37}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07631", 37}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07632", 37}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07633", 37}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07634", 37}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07636", 37}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07639", 37}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07640", 37}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07641", 37}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07642", 37}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07643", 37}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07644", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07645", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07646", 37}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07647", 37}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07648", 37}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07649", 37}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07749", 37}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07751", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07831", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07832", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07834", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07835", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07836", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07837", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07838", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07839", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07840", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07841", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07843", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07844", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07845", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07846", 37}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07847", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07848", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07849", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07850", 37}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07878", 37}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07879", 37}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07880", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07881", 37}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-07888", 37}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08114", 37}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08115", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08116", 37}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08117", 37}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08119", 37}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08600", 37}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08601", 37}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08605", 37}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08606", 37}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08608", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08609", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08610", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08611", 37}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08612", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08613", 37}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08617", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08636", 37}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08637", 37}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08644", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08666", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08669", 37}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08684", 37}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08685", 37}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08686", 37}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08687", 37}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08688", 37}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08689", 37}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08690", 37}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08693", 37}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08694", 37}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08695", 37}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08696", 37}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08698", 37}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08699", 37}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08876", 37}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08877", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08878", 37}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-08879", 37}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09003", 37}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09116", 37}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09211", 37}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09363", 37}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09364", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09365", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09368", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09369", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09548", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09549", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09600", 37}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09642", 37}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09643", 37}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-09650", 37}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10068", 37}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10608", 37}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10677", 37}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10678", 37}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10679", 37}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10772", 37}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10795", 37}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10796", 37}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10797", 37}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10862", 37}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10864", 37}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10927", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-10928", 37}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11297", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11298", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11299", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11300", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11301", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11302", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11304", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11305", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11306", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11308", 37}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11309", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11372", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11373", 37}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11374", 37}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11376", 37}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11397", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11398", 37}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11437", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11438", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11439", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11441", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11442", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11485", 37}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11522", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11523", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11861", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11862", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11863", 37}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11865", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11866", 37}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11868", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-11869", 37}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-12338", 37}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-12339", 37}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-12340", 37}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-None-12472", 37}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchSADQCOM-06976", 57}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchSADQCOM-12420", 57}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchSSDQCOM-06974", 57}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchWindow-09215", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchWindow-09216", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchWindow-09217", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBlockMatchWindow-12421", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageBoxFilterQCOM-06973", 53}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageSampleWeightedQCOM-06971", 58}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageSampleWeightedQCOM-06972", 58}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageSampleWeightedQCOM-06977", 58}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageSampleWeightedQCOM-06978", 58}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageWrite-04469", 45}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageWrite-08795", 45}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpImageWrite-08796", 45}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTensorReadARM-12462", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeImage-07027", 44}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeImage-07028", 44}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeImage-07029", 44}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeImage-07030", 44}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeImage-07468", 44}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-OpTypeTensorARM-09906", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-SampledType-04470", 44}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-SampledType-04471", 44}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-SampledType-04472", 44}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-SampledType-04473", 44}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07322", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountX must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[0]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07323", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountY must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[1]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07324", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupCount[2]", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07325", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS contains a shader using the TaskEXT Execution Model, The product of groupCountX, groupCountY and groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxTaskWorkGroupTotalCount", 271}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07326", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountX must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[0]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07327", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountY must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[1]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07328", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupCount[2]", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-TaskEXT-07329", 40}, { {"If the current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS does not contain a shader using the TaskEXT Execution Model, The product of groupCountX, groupCountY and groupCountZ must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesEXT::maxMeshWorkGroupTotalCount", 279}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-VkPipelineVieportCreateInfo-04141", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-VkPipelineVieportCreateInfo-04142", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-addressModeU-12422", 45}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-advancedBlendMaxColorAttachments-07480", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-alphaToCoverageEnable-08919", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-alphaToCoverageEnable-08920", 54}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-attachmentCount-07750", 48}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-blendEnable-04727", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-colorAttachmentCount-06179", 53}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-colorAttachmentCount-06185", 53}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-colorAttachmentCount-09362", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-colorAttachmentCount-09372", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-colorAttachmentCount-11871", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-commandBuffer-02707", 46}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-commandBuffer-10746", 46}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-conservativePointAndLineRasterization-07499", 70}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-coverageModulationTableEnable-07488", 62}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-coverageReductionMode-07491", 54}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-coverageToColorEnable-07490", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-cubicRangeClamp-09212", 48}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-customResolve-11524", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-customResolve-11525", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-customResolve-11529", 46}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-customResolve-11530", 46}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dimensionCount-09905", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingLocalRead-11797", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08910", 66}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08911", 66}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08912", 66}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08913", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08914", 66}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08915", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08916", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08917", 66}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-08918", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-11864", 66}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-11867", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-dynamicRenderingUnusedAttachments-11870", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-filterCubic-02694", 44}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-filterCubicMinmax-02695", 50}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-firstAttachment-07476", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-firstAttachment-07478", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-02696", 38}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-10582", 38}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-11521", 38}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-13361", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-13362", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-13363", 38}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-13364", 38}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-flags-13365", 38}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-format-07753", 39}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageLayout-00344", 44}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06172", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06173", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06174", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06175", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06176", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06177", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06183", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-imageView-06184", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-index-11450", 38}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-layers-10831", 39}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-logicOp-04878", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-magFilter-04553", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-magFilter-09598", 42}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-maxFragmentDualSrcAttachments-09239", 62}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-maxMultiviewInstanceIndex-02688", 58}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-micromap-11636", 41}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-micromap-11637", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-micromap-11638", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-micromap-11639", 41}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-mipmapMode-04770", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-mipmapMode-09599", 43}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-multisampledRenderToSingleSampled-07285", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-multisampledRenderToSingleSampled-07286", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-multisampledRenderToSingleSampled-07287", 66}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-multiviewPerViewViewports-12262", 58}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-multiviewPerViewViewports-12263", 58}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-nextStage-10745", 42}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pBindInfo-11375", 42}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pColorAttachments-08963", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pColorAttachments-11539", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthAttachment-06186", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthAttachment-08964", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthAttachment-11540", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthAttachment-11872", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthInputAttachmentIndex-09595", 59}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDepthInputAttachmentIndex-09596", 59}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDescription-09900", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDynamicStates-08715", 47}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pDynamicStates-08716", 47}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pNext-07935", 38}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pNext-12463", 38}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pStages-10680", 40}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pStencilAttachment-06187", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pStencilAttachment-08965", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pStencilAttachment-11860", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pStencilAttachment-11873", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pipelineFragmentShadingRate-09238", 60}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-pipelineStatistics-07076", 51}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 82}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 82}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 75}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 75}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 78}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-protectedNoFault-11455", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-protectedNoFault-11456", 49}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizationSamples-07471", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizationSamples-07489", 53}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizationSamples-07494", 53}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizerDiscardEnable-09236", 56}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizerDiscardEnable-09420", 56}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-rasterizerDiscardEnable-10863", 56}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-reductionMode-09213", 46}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-renderPass-02684", 43}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-renderPass-06198", 43}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-renderpass", 37}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-02689", 54}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07484", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07485", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07486", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07487", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07936", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07937", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsEnable-07938", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsPerPixel-07482", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sampleLocationsPerPixel-07483", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-samples-07472", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-samples-07473", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-scissorCount-03418", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-selectableCubicWeights-09214", 55}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-shadingRateImage-09233", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-shadingRateImage-09234", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sparseImageInt64Atomics-04474", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-sparseImageInt64Atomics-04475", 56}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stage-06480", 38}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stage-07073", 38}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stippledLineEnable-07495", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stippledLineEnable-07496", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stippledLineEnable-07497", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-stippledLineEnable-07498", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-storageBuffers-06936", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-subpass-02685", 40}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-uniformBuffers-06935", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-unnormalizedCoordinates-09635", 56}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewMask-06178", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewType-07752", 41}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-03417", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-03419", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-04137", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-04138", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-04139", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-07492", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-07493", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksEXT-viewportCount-09421", 46}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-InputAttachmentIndex-09597", 62}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-02691", 46}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-02692", 46}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-02693", 46}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-04876", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-04877", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-06479", 46}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-06537", 46}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-06666", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-06886", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-06887", 46}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07074", 46}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07075", 46}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07288", 46}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07469", 46}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07620", 46}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07621", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07622", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07623", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07624", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07625", 46}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07626", 46}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07627", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07629", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07630", 46}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07631", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07632", 46}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07633", 46}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07634", 46}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07636", 46}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07639", 46}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07640", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07641", 46}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07642", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07643", 46}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07644", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07645", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07646", 46}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07647", 46}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07648", 46}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07649", 46}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07749", 46}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07751", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07831", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07832", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07834", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07835", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07836", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07837", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07838", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07839", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07840", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07841", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07843", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07844", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07845", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07846", 46}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07847", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07848", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07849", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07850", 46}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07878", 46}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07879", 46}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07880", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07881", 46}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-07888", 46}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08114", 46}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08115", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08116", 46}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08117", 46}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08119", 46}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08600", 46}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08601", 46}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08605", 46}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08606", 46}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08608", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08609", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08610", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08611", 46}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08612", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08613", 46}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08617", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08636", 46}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08637", 46}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08644", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08666", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08669", 46}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08684", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08685", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08686", 46}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08687", 46}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08688", 46}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08689", 46}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08690", 46}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08693", 46}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08694", 46}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08695", 46}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08696", 46}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08698", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08699", 46}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08876", 46}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08877", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08878", 46}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-08879", 46}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09003", 46}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09116", 46}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09211", 46}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09363", 46}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09364", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09365", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09368", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09369", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09548", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09549", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09600", 46}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09642", 46}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09643", 46}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-09650", 46}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10068", 46}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10608", 46}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10677", 46}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10678", 46}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10679", 46}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10772", 46}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10795", 46}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10796", 46}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10797", 46}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10862", 46}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10864", 46}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10927", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-10928", 46}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11297", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11298", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11299", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11300", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11301", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11302", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11304", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11305", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11306", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11308", 46}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11309", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11372", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11373", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11374", 46}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11376", 46}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11397", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11398", 46}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11437", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11438", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11439", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11441", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11442", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11485", 46}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11522", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11523", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11861", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11862", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11863", 46}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11865", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11866", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11868", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-11869", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-12338", 46}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-12339", 46}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-12340", 46}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-None-12472", 46}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchSADQCOM-06976", 66}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchSADQCOM-12420", 66}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchSSDQCOM-06974", 66}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchWindow-09215", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchWindow-09216", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchWindow-09217", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBlockMatchWindow-12421", 65}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageBoxFilterQCOM-06973", 62}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageSampleWeightedQCOM-06971", 67}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageSampleWeightedQCOM-06972", 67}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageSampleWeightedQCOM-06977", 67}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageSampleWeightedQCOM-06978", 67}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageWrite-04469", 54}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageWrite-08795", 54}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpImageWrite-08796", 54}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTensorReadARM-12462", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeImage-07027", 53}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeImage-07028", 53}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeImage-07029", 53}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeImage-07030", 53}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeImage-07468", 53}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-OpTypeTensorARM-09906", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-SampledType-04470", 53}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-SampledType-04471", 53}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-SampledType-04472", 53}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-SampledType-04473", 53}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-VkPipelineVieportCreateInfo-04141", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-VkPipelineVieportCreateInfo-04142", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-addressModeU-12422", 54}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-advancedBlendMaxColorAttachments-07480", 74}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-alphaToCoverageEnable-08919", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-alphaToCoverageEnable-08920", 63}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-attachmentCount-07750", 57}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-blendEnable-04727", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-colorAttachmentCount-06179", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-colorAttachmentCount-06185", 62}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-colorAttachmentCount-09362", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-colorAttachmentCount-09372", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-colorAttachmentCount-11871", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-02707", 55}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-10746", 55}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-13067", 55}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-conservativePointAndLineRasterization-07499", 79}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-coverageModulationTableEnable-07488", 71}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-coverageReductionMode-07491", 63}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-coverageToColorEnable-07490", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-cubicRangeClamp-09212", 57}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-customResolve-11524", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-customResolve-11525", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-customResolve-11529", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-customResolve-11530", 55}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dimensionCount-09905", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingLocalRead-11797", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08910", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08911", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08912", 75}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08913", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08914", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08915", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08916", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08917", 75}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-08918", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-11864", 75}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-11867", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-dynamicRenderingUnusedAttachments-11870", 75}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-filterCubic-02694", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-filterCubicMinmax-02695", 59}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-firstAttachment-07476", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-firstAttachment-07478", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-02696", 47}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-10582", 47}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-11521", 47}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-13361", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-13362", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-13363", 47}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-13364", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-flags-13365", 47}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-format-07753", 48}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageLayout-00344", 53}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06172", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06173", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06174", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06175", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06176", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06177", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06183", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-imageView-06184", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-index-11450", 47}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-layers-10831", 48}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-logicOp-04878", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-magFilter-04553", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-magFilter-09598", 51}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-maxFragmentDualSrcAttachments-09239", 71}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-maxMultiviewInstanceIndex-02688", 67}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-micromap-11636", 50}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-micromap-11637", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-micromap-11638", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-micromap-11639", 50}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-mipmapMode-04770", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-mipmapMode-09599", 52}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-multisampledRenderToSingleSampled-07285", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-multisampledRenderToSingleSampled-07286", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-multisampledRenderToSingleSampled-07287", 75}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-multiviewPerViewViewports-12262", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-multiviewPerViewViewports-12263", 67}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-nextStage-10745", 51}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pBindInfo-11375", 51}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pColorAttachments-08963", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pColorAttachments-11539", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthAttachment-06186", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthAttachment-08964", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthAttachment-11540", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthAttachment-11872", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthInputAttachmentIndex-09595", 68}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDepthInputAttachmentIndex-09596", 68}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDescription-09900", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDynamicStates-08715", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pDynamicStates-08716", 56}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pInfo-parameter", 51}, { {"pInfo must be a valid pointer to a valid VkDrawIndirect2InfoKHR structure", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pNext-07935", 47}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pNext-12463", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pStages-10680", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pStencilAttachment-06187", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pStencilAttachment-08965", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pStencilAttachment-11860", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pStencilAttachment-11873", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pipelineFragmentShadingRate-09238", 69}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-pipelineStatistics-07076", 60}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 91}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 84}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 87}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-protectedNoFault-11455", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-protectedNoFault-11456", 58}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizationSamples-07471", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizationSamples-07489", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizationSamples-07494", 62}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizerDiscardEnable-09236", 65}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizerDiscardEnable-09420", 65}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-rasterizerDiscardEnable-10863", 65}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-reductionMode-09213", 55}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-renderPass-02684", 52}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-renderPass-06198", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-renderpass", 46}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-02689", 63}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07484", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07485", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07486", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07487", 63}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07936", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07937", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsEnable-07938", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsPerPixel-07482", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sampleLocationsPerPixel-07483", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-samples-07472", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-samples-07473", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-scissorCount-03418", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-selectableCubicWeights-09214", 64}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-shadingRateImage-09233", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-shadingRateImage-09234", 58}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sparseImageInt64Atomics-04474", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-sparseImageInt64Atomics-04475", 65}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stage-06480", 47}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stage-07073", 47}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stippledLineEnable-07495", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stippledLineEnable-07496", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stippledLineEnable-07497", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-stippledLineEnable-07498", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-storageBuffers-06936", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-subpass-02685", 49}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-suspended", 45}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-uniformBuffers-06935", 56}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-unnormalizedCoordinates-09635", 65}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewMask-06178", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewType-07752", 50}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-03417", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-03419", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-04137", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-04138", 55}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-04139", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-07492", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-07493", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirect2EXT-viewportCount-09421", 55}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-InputAttachmentIndex-09597", 67}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-02691", 51}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-02692", 51}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-02693", 51}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-04876", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-04877", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-06479", 51}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-06537", 51}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-06666", 51}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-06886", 51}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-06887", 51}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07074", 51}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07075", 51}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07288", 51}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07469", 51}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07620", 51}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07621", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07622", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07623", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07624", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07625", 51}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07626", 51}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07627", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07629", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07630", 51}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07631", 51}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07632", 51}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07633", 51}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07634", 51}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07636", 51}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07639", 51}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07640", 51}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07641", 51}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07642", 51}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07643", 51}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07644", 51}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07645", 51}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07646", 51}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07647", 51}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07648", 51}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07649", 51}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07749", 51}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07751", 51}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07831", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07832", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07834", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07835", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07836", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07837", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07838", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07839", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07840", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07841", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07843", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07844", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07845", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07846", 51}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07847", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07848", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07849", 51}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07850", 51}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07878", 51}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07879", 51}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07880", 51}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07881", 51}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-07888", 51}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08114", 51}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08115", 51}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08116", 51}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08117", 51}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08119", 51}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08600", 51}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08601", 51}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08605", 51}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08606", 51}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08608", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08609", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08610", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08611", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08612", 51}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08613", 51}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08617", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08636", 51}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08637", 51}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08644", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08666", 51}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08669", 51}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08684", 51}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08685", 51}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08686", 51}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08687", 51}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08688", 51}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08689", 51}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08690", 51}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08693", 51}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08694", 51}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08695", 51}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08696", 51}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08698", 51}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08699", 51}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08876", 51}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08877", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08878", 51}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-08879", 51}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09003", 51}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09116", 51}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09211", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09363", 51}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09364", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09365", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09368", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09369", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09548", 51}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09549", 51}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09600", 51}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09642", 51}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09643", 51}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-09650", 51}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10068", 51}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10608", 51}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10677", 51}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10678", 51}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10679", 51}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10772", 51}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10795", 51}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10796", 51}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10797", 51}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10862", 51}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10864", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10927", 51}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-10928", 51}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11297", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11298", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11299", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11300", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11301", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11302", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11304", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11305", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11306", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11308", 51}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11309", 51}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11372", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11373", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11374", 51}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11376", 51}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11397", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11398", 51}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11437", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11438", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11439", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11441", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11442", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11485", 51}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11522", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11523", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11861", 51}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11862", 51}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11863", 51}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11865", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11866", 51}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11868", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-11869", 51}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-12338", 51}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-12339", 51}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-12340", 51}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-None-12472", 51}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchSADQCOM-06976", 71}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchSADQCOM-12420", 71}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchSSDQCOM-06974", 71}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchWindow-09215", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchWindow-09216", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchWindow-09217", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBlockMatchWindow-12421", 70}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageBoxFilterQCOM-06973", 67}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageSampleWeightedQCOM-06971", 72}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageSampleWeightedQCOM-06972", 72}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageSampleWeightedQCOM-06977", 72}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageSampleWeightedQCOM-06978", 72}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageWrite-04469", 59}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageWrite-08795", 59}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpImageWrite-08796", 59}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTensorReadARM-12462", 62}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeImage-07027", 58}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeImage-07028", 58}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeImage-07029", 58}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeImage-07030", 58}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeImage-07468", 58}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-OpTypeTensorARM-09906", 62}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-SampledType-04470", 58}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-SampledType-04471", 58}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-SampledType-04472", 58}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-SampledType-04473", 58}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-VkPipelineVieportCreateInfo-04141", 74}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-VkPipelineVieportCreateInfo-04142", 74}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-addressModeU-12422", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-advancedBlendMaxColorAttachments-07480", 79}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-alphaToCoverageEnable-08919", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-alphaToCoverageEnable-08920", 68}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-attachmentCount-07750", 62}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-blendEnable-04727", 58}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-colorAttachmentCount-06179", 67}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-colorAttachmentCount-06185", 67}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-colorAttachmentCount-09362", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-colorAttachmentCount-09372", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-colorAttachmentCount-11871", 67}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-02707", 60}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-10746", 60}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-13068", 60}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-conservativePointAndLineRasterization-07499", 84}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-coverageModulationTableEnable-07488", 76}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-coverageReductionMode-07491", 68}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-coverageToColorEnable-07490", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-cubicRangeClamp-09212", 62}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-customResolve-11524", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-customResolve-11525", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-customResolve-11529", 60}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-customResolve-11530", 60}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dimensionCount-09905", 61}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-drawIndirectCount-13069", 64}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingLocalRead-11797", 72}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08910", 80}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08911", 80}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08912", 80}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08913", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08914", 80}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08915", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08916", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08917", 80}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-08918", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-11864", 80}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-11867", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-dynamicRenderingUnusedAttachments-11870", 80}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-filterCubic-02694", 58}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-filterCubicMinmax-02695", 64}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-firstAttachment-07476", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-firstAttachment-07478", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-02696", 52}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-10582", 52}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-11521", 52}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-13361", 52}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-13362", 52}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-13363", 52}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-13364", 52}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-flags-13365", 52}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-format-07753", 53}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageLayout-00344", 58}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06172", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06173", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06174", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06175", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06176", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06177", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06183", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-imageView-06184", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-index-11450", 52}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-layers-10831", 53}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-logicOp-04878", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-magFilter-04553", 56}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-magFilter-09598", 56}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-maxFragmentDualSrcAttachments-09239", 76}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-maxMultiviewInstanceIndex-02688", 72}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-micromap-11636", 55}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-micromap-11637", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-micromap-11638", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-micromap-11639", 55}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-mipmapMode-04770", 57}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-mipmapMode-09599", 57}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-multisampledRenderToSingleSampled-07285", 80}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-multisampledRenderToSingleSampled-07286", 80}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-multisampledRenderToSingleSampled-07287", 80}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-multiviewPerViewViewports-12262", 72}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-multiviewPerViewViewports-12263", 72}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-nextStage-10745", 56}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pBindInfo-11375", 56}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pColorAttachments-08963", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pColorAttachments-11539", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthAttachment-06186", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthAttachment-08964", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthAttachment-11540", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthAttachment-11872", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthInputAttachmentIndex-09595", 73}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDepthInputAttachmentIndex-09596", 73}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDescription-09900", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDynamicStates-08715", 61}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pDynamicStates-08716", 61}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pInfo-parameter", 56}, { {"pInfo must be a valid pointer to a valid VkDrawIndirectCount2InfoKHR structure", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pNext-07935", 52}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pNext-12463", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pStages-10680", 54}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pStencilAttachment-06187", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pStencilAttachment-08965", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pStencilAttachment-11860", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pStencilAttachment-11873", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pipelineFragmentShadingRate-09238", 74}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-pipelineStatistics-07076", 65}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 96}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 96}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 89}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 89}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 92}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-protectedNoFault-11455", 63}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-protectedNoFault-11456", 63}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizationSamples-07471", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizationSamples-07489", 67}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizationSamples-07494", 67}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizerDiscardEnable-09236", 70}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizerDiscardEnable-09420", 70}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-rasterizerDiscardEnable-10863", 70}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-reductionMode-09213", 60}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-renderPass-02684", 57}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-renderPass-06198", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-renderpass", 51}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-02689", 68}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07484", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07485", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07486", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07487", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07936", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07937", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsEnable-07938", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsPerPixel-07482", 70}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sampleLocationsPerPixel-07483", 70}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-samples-07472", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-samples-07473", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-scissorCount-03418", 59}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-selectableCubicWeights-09214", 69}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-shadingRateImage-09233", 63}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-shadingRateImage-09234", 63}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sparseImageInt64Atomics-04474", 70}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-sparseImageInt64Atomics-04475", 70}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stage-06480", 52}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stage-07073", 52}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stippledLineEnable-07495", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stippledLineEnable-07496", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stippledLineEnable-07497", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-stippledLineEnable-07498", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-storageBuffers-06936", 61}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-subpass-02685", 54}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-suspended", 50}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-uniformBuffers-06935", 61}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-unnormalizedCoordinates-09635", 70}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewMask-06178", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewType-07752", 55}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-03417", 60}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-03419", 60}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-04137", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-04138", 60}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-04139", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-07492", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-07493", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCount2EXT-viewportCount-09421", 60}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-InputAttachmentIndex-09597", 66}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-MeshEXT-07100", 53}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshEXT Execution Model", 123}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-02691", 50}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-02692", 50}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-02693", 50}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-04445", 50}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-04876", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-04877", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-06479", 50}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-06537", 50}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-06666", 50}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-06886", 50}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-06887", 50}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07074", 50}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07075", 50}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07288", 50}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07469", 50}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07620", 50}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07621", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07622", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07623", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07624", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07625", 50}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07626", 50}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07627", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07629", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07630", 50}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07631", 50}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07632", 50}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07633", 50}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07634", 50}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07636", 50}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07639", 50}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07640", 50}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07641", 50}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07642", 50}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07643", 50}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07644", 50}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07645", 50}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07646", 50}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07647", 50}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07648", 50}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07649", 50}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07749", 50}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07751", 50}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07831", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07832", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07834", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07835", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07836", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07837", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07838", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07839", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07840", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07841", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07843", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07844", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07845", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07846", 50}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07847", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07848", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07849", 50}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07850", 50}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07878", 50}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07879", 50}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07880", 50}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07881", 50}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-07888", 50}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08114", 50}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08115", 50}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08116", 50}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08117", 50}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08119", 50}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08600", 50}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08601", 50}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08605", 50}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08606", 50}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08608", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08609", 50}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08610", 50}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08611", 50}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08612", 50}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08613", 50}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08617", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08636", 50}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08637", 50}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08644", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08666", 50}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08669", 50}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08684", 50}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08685", 50}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08686", 50}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08687", 50}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08688", 50}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08689", 50}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08690", 50}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08693", 50}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08694", 50}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08695", 50}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08696", 50}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08698", 50}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08699", 50}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08876", 50}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08877", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08878", 50}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-08879", 50}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09003", 50}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09116", 50}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09211", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09363", 50}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09364", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09365", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09368", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09369", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09548", 50}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09549", 50}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09600", 50}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09642", 50}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09643", 50}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-09650", 50}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10068", 50}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10608", 50}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10677", 50}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10678", 50}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10679", 50}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10772", 50}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10795", 50}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10796", 50}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10797", 50}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10862", 50}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10864", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10927", 50}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-10928", 50}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11297", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11298", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11299", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11300", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11301", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11302", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11304", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11305", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11306", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11308", 50}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11309", 50}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11372", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11373", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11374", 50}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11376", 50}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11397", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11398", 50}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11437", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11438", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11439", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11441", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11442", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11485", 50}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11522", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11523", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11861", 50}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11862", 50}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11863", 50}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11865", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11866", 50}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11868", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-11869", 50}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-12338", 50}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-12339", 50}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-12340", 50}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-None-12472", 50}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchSADQCOM-06976", 70}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchSADQCOM-12420", 70}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchSSDQCOM-06974", 70}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchWindow-09215", 69}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchWindow-09216", 69}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchWindow-09217", 69}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBlockMatchWindow-12421", 69}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageBoxFilterQCOM-06973", 66}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageSampleWeightedQCOM-06971", 71}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageSampleWeightedQCOM-06972", 71}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageSampleWeightedQCOM-06977", 71}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageSampleWeightedQCOM-06978", 71}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageWrite-04469", 58}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageWrite-08795", 58}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpImageWrite-08796", 58}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTensorReadARM-12462", 61}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeImage-07027", 57}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeImage-07028", 57}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeImage-07029", 57}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeImage-07030", 57}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeImage-07468", 57}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-OpTypeTensorARM-09906", 61}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-SampledType-04470", 57}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-SampledType-04471", 57}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-SampledType-04472", 57}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-SampledType-04473", 57}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-VkPipelineVieportCreateInfo-04141", 73}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-VkPipelineVieportCreateInfo-04142", 73}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-addressModeU-12422", 58}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-advancedBlendMaxColorAttachments-07480", 78}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-alphaToCoverageEnable-08919", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-alphaToCoverageEnable-08920", 67}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-attachmentCount-07750", 61}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-blendEnable-04727", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-buffer-02708", 52}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-buffer-02709", 52}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-buffer-parameter", 56}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-colorAttachmentCount-06179", 66}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-colorAttachmentCount-06185", 66}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-colorAttachmentCount-09362", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-colorAttachmentCount-09372", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-colorAttachmentCount-11871", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-02707", 59}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-02711", 59}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-10746", 59}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-commonparent", 52}, { {"Each of buffer, commandBuffer, and countBuffer must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-conservativePointAndLineRasterization-07499", 83}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBuffer-02714", 57}, { {"If countBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBuffer-02715", 57}, { {"countBuffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBuffer-02717", 57}, { {"The count stored in countBuffer must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 106}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBuffer-parameter", 61}, { {"countBuffer must be a valid VkBuffer handle", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBufferOffset-02716", 63}, { {"countBufferOffset must be a multiple of 4", 41}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-countBufferOffset-04129", 63}, { {"(countBufferOffset + sizeof(uint32_t)) must be less than or equal to the size of countBuffer", 92}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-coverageModulationTableEnable-07488", 75}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-coverageReductionMode-07491", 67}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-coverageToColorEnable-07490", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-cubicRangeClamp-09212", 61}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-customResolve-11524", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-customResolve-11525", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-customResolve-11529", 59}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-customResolve-11530", 59}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dimensionCount-09905", 60}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingLocalRead-11797", 71}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08910", 79}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08911", 79}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08912", 79}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08913", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08914", 79}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08915", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08916", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08917", 79}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-08918", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-11864", 79}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-11867", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-dynamicRenderingUnusedAttachments-11870", 79}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-filterCubic-02694", 57}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-filterCubicMinmax-02695", 63}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-firstAttachment-07476", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-firstAttachment-07478", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-02696", 51}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-10582", 51}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-11521", 51}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-13361", 51}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-13362", 51}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-13363", 51}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-13364", 51}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-flags-13365", 51}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-format-07753", 52}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageLayout-00344", 57}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06172", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06173", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06174", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06175", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06176", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06177", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06183", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-imageView-06184", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-index-11450", 51}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-layers-10831", 52}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-logicOp-04878", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-magFilter-04553", 55}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-magFilter-09598", 55}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-maxDrawCount-07097", 58}, { {"If maxDrawCount is greater than or equal to 1, (stride x (maxDrawCount - 1) + offset + sizeof(VkDrawMeshTasksIndirectCommandEXT)) must be less than or equal to the size of buffer", 178}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-maxFragmentDualSrcAttachments-09239", 75}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-maxMultiviewInstanceIndex-02688", 71}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-micromap-11636", 54}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-micromap-11637", 54}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-micromap-11638", 54}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-micromap-11639", 54}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-mipmapMode-04770", 56}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-mipmapMode-09599", 56}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-multisampledRenderToSingleSampled-07285", 79}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-multisampledRenderToSingleSampled-07286", 79}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-multisampledRenderToSingleSampled-07287", 79}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-multiviewPerViewViewports-12262", 71}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-multiviewPerViewViewports-12263", 71}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-nextStage-10745", 55}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-offset-02710", 52}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pBindInfo-11375", 55}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pColorAttachments-08963", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pColorAttachments-11539", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthAttachment-06186", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthAttachment-08964", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthAttachment-11540", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthAttachment-11872", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthInputAttachmentIndex-09595", 72}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDepthInputAttachmentIndex-09596", 72}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDescription-09900", 58}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDynamicStates-08715", 60}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pDynamicStates-08716", 60}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pNext-07935", 51}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pNext-12463", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pStages-10680", 53}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pStencilAttachment-06187", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pStencilAttachment-08965", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pStencilAttachment-11860", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pStencilAttachment-11873", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pipelineFragmentShadingRate-09238", 73}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-pipelineStatistics-07076", 64}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 95}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 95}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 88}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 88}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 91}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-protectedNoFault-11455", 62}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-protectedNoFault-11456", 62}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizationSamples-07471", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizationSamples-07489", 66}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizationSamples-07494", 66}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizerDiscardEnable-09236", 69}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizerDiscardEnable-09420", 69}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-rasterizerDiscardEnable-10863", 69}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-reductionMode-09213", 59}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-renderPass-02684", 56}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-renderPass-06198", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-renderpass", 50}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-02689", 67}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07484", 67}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07485", 67}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07486", 67}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07487", 67}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07936", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07937", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsEnable-07938", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsPerPixel-07482", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sampleLocationsPerPixel-07483", 69}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-samples-07472", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-samples-07473", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-scissorCount-03418", 58}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-selectableCubicWeights-09214", 68}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-shadingRateImage-09233", 62}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-shadingRateImage-09234", 62}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sparseImageInt64Atomics-04474", 69}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-sparseImageInt64Atomics-04475", 69}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stage-06480", 51}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stage-07073", 51}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stippledLineEnable-07495", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stippledLineEnable-07496", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stippledLineEnable-07497", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stippledLineEnable-07498", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-storageBuffers-06936", 60}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-stride-07096", 52}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawMeshTasksIndirectCommandEXT)", 109}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-subpass-02685", 53}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-suspended", 49}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-uniformBuffers-06935", 60}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-unnormalizedCoordinates-09635", 69}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewMask-06178", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewType-07752", 54}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-03417", 59}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-03419", 59}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-04137", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-04138", 59}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-04139", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-07492", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-07493", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountEXT-viewportCount-09421", 59}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-InputAttachmentIndex-09597", 65}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-MeshNV-07082", 51}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshNV Execution Model", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-02691", 49}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-02692", 49}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-02693", 49}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-04445", 49}, { {"If the drawIndirectCount feature is not enabled this function must not be used", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-04876", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-04877", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-06479", 49}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-06537", 49}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-06666", 49}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-06886", 49}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-06887", 49}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07074", 49}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07075", 49}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07288", 49}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07469", 49}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07620", 49}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07621", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07622", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07623", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07624", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07625", 49}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07626", 49}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07627", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07629", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07630", 49}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07631", 49}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07632", 49}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07633", 49}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07634", 49}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07636", 49}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07639", 49}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07640", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07641", 49}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07642", 49}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07643", 49}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07644", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07645", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07646", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07647", 49}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07648", 49}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07649", 49}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07749", 49}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07751", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07831", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07832", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07834", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07835", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07836", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07837", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07838", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07839", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07840", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07841", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07843", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07844", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07845", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07846", 49}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07847", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07848", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07849", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07850", 49}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07878", 49}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07879", 49}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07880", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07881", 49}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-07888", 49}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08114", 49}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08115", 49}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08116", 49}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08117", 49}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08119", 49}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08600", 49}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08601", 49}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08605", 49}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08606", 49}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08608", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08609", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08610", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08611", 49}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08612", 49}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08613", 49}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08617", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08636", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08637", 49}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08644", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08666", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08669", 49}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08684", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08685", 49}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08686", 49}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08687", 49}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08688", 49}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08689", 49}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08690", 49}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08693", 49}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08694", 49}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08695", 49}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08696", 49}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08698", 49}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08699", 49}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08876", 49}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08877", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08878", 49}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-08879", 49}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09003", 49}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09116", 49}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09211", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09363", 49}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09364", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09365", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09368", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09369", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09548", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09549", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09600", 49}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09642", 49}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09643", 49}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-09650", 49}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10068", 49}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10608", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10677", 49}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10678", 49}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10679", 49}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10772", 49}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10795", 49}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10796", 49}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10797", 49}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10862", 49}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10864", 49}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10927", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-10928", 49}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11297", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11298", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11299", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11300", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11301", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11302", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11304", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11305", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11306", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11308", 49}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11309", 49}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11372", 49}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11373", 49}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11374", 49}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11376", 49}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11397", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11398", 49}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11437", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11438", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11439", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11441", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11442", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11485", 49}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11522", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11523", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11861", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11862", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11863", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11865", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11866", 49}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11868", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-11869", 49}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-12338", 49}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-12339", 49}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-12340", 49}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-None-12472", 49}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchSADQCOM-06976", 69}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchSADQCOM-12420", 69}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchSSDQCOM-06974", 69}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchWindow-09215", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchWindow-09216", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchWindow-09217", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBlockMatchWindow-12421", 68}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageBoxFilterQCOM-06973", 65}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageSampleWeightedQCOM-06971", 70}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageSampleWeightedQCOM-06972", 70}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageSampleWeightedQCOM-06977", 70}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageSampleWeightedQCOM-06978", 70}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageWrite-04469", 57}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageWrite-08795", 57}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpImageWrite-08796", 57}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTensorReadARM-12462", 60}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeImage-07027", 56}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeImage-07028", 56}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeImage-07029", 56}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeImage-07030", 56}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeImage-07468", 56}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-OpTypeTensorARM-09906", 60}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-SampledType-04470", 56}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-SampledType-04471", 56}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-SampledType-04472", 56}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-SampledType-04473", 56}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-VkPipelineVieportCreateInfo-04141", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-VkPipelineVieportCreateInfo-04142", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-addressModeU-12422", 57}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-advancedBlendMaxColorAttachments-07480", 77}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-alphaToCoverageEnable-08919", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-alphaToCoverageEnable-08920", 66}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-attachmentCount-07750", 60}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-blendEnable-04727", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-buffer-02708", 51}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-buffer-02709", 51}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-buffer-parameter", 55}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-colorAttachmentCount-06179", 65}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-colorAttachmentCount-06185", 65}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-colorAttachmentCount-09362", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-colorAttachmentCount-09372", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-colorAttachmentCount-11871", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-02707", 58}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-02711", 58}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-10746", 58}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-commonparent", 51}, { {"Each of buffer, commandBuffer, and countBuffer must have been created, allocated, or retrieved from the same VkDevice", 117}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-conservativePointAndLineRasterization-07499", 82}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBuffer-02714", 56}, { {"If countBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 112}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBuffer-02715", 56}, { {"countBuffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBuffer-02717", 56}, { {"The count stored in countBuffer must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 106}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBuffer-parameter", 60}, { {"countBuffer must be a valid VkBuffer handle", 43}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBufferOffset-02716", 62}, { {"countBufferOffset must be a multiple of 4", 41}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-countBufferOffset-04129", 62}, { {"(countBufferOffset + sizeof(uint32_t)) must be less than or equal to the size of countBuffer", 92}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-coverageModulationTableEnable-07488", 74}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-coverageReductionMode-07491", 66}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-coverageToColorEnable-07490", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-cubicRangeClamp-09212", 60}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-customResolve-11524", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-customResolve-11525", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-customResolve-11529", 58}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-customResolve-11530", 58}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dimensionCount-09905", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingLocalRead-11797", 70}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08910", 78}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08911", 78}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08912", 78}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08913", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08914", 78}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08915", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08916", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08917", 78}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-08918", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-11864", 78}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-11867", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-dynamicRenderingUnusedAttachments-11870", 78}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-filterCubic-02694", 56}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-filterCubicMinmax-02695", 62}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-firstAttachment-07476", 60}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-firstAttachment-07478", 60}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-02696", 50}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-10582", 50}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-11521", 50}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-13361", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-13362", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-13363", 50}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-13364", 50}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-flags-13365", 50}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-format-07753", 51}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageLayout-00344", 56}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06172", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06173", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06174", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06175", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06176", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06177", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06183", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-imageView-06184", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-index-11450", 50}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-layers-10831", 51}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-logicOp-04878", 52}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-magFilter-04553", 54}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-magFilter-09598", 54}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-maxDrawCount-02183", 57}, { {"If maxDrawCount is greater than or equal to 1, (stride x (maxDrawCount - 1) + offset + sizeof(VkDrawMeshTasksIndirectCommandNV)) must be less than or equal to the size of buffer", 177}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-maxFragmentDualSrcAttachments-09239", 74}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-maxMultiviewInstanceIndex-02688", 70}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-micromap-11636", 53}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-micromap-11637", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-micromap-11638", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-micromap-11639", 53}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-mipmapMode-04770", 55}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-mipmapMode-09599", 55}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-multisampledRenderToSingleSampled-07285", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-multisampledRenderToSingleSampled-07286", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-multisampledRenderToSingleSampled-07287", 78}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-multiviewPerViewViewports-12262", 70}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-multiviewPerViewViewports-12263", 70}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-nextStage-10745", 54}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-offset-02710", 51}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pBindInfo-11375", 54}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pColorAttachments-08963", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pColorAttachments-11539", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthAttachment-06186", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthAttachment-08964", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthAttachment-11540", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthAttachment-11872", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthInputAttachmentIndex-09595", 71}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDepthInputAttachmentIndex-09596", 71}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDescription-09900", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDynamicStates-08715", 59}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pDynamicStates-08716", 59}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pNext-07935", 50}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pNext-12463", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pStages-10680", 52}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pStencilAttachment-06187", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pStencilAttachment-08965", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pStencilAttachment-11860", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pStencilAttachment-11873", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pipelineFragmentShadingRate-09238", 72}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-pipelineStatistics-07076", 63}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-primitiveFragmentShadingRateWithMultipleViewports-04552", 94}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-primitiveFragmentShadingRateWithMultipleViewports-08642", 94}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-primitivesGeneratedQueryWithNonZeroStreams-06709", 87}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-primitivesGeneratedQueryWithNonZeroStreams-07481", 87}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-primitivesGeneratedQueryWithRasterizerDiscard-06708", 90}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-protectedNoFault-11455", 61}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-protectedNoFault-11456", 61}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizationSamples-07471", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizationSamples-07489", 65}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizationSamples-07494", 65}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizerDiscardEnable-09236", 68}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizerDiscardEnable-09420", 68}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-rasterizerDiscardEnable-10863", 68}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-reductionMode-09213", 58}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-renderPass-02684", 55}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-renderPass-06198", 55}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-renderpass", 49}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-02689", 66}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07484", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07485", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07486", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07487", 66}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07936", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07937", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsEnable-07938", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsPerPixel-07482", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sampleLocationsPerPixel-07483", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-samples-07472", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-samples-07473", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-scissorCount-03418", 57}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-selectableCubicWeights-09214", 67}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-shadingRateImage-09233", 61}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-shadingRateImage-09234", 61}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sparseImageInt64Atomics-04474", 68}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-sparseImageInt64Atomics-04475", 68}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stage-06480", 50}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stage-07073", 50}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stippledLineEnable-07495", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stippledLineEnable-07496", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stippledLineEnable-07497", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stippledLineEnable-07498", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-storageBuffers-06936", 59}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-stride-02182", 51}, { {"stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawMeshTasksIndirectCommandNV)", 108}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-subpass-02685", 52}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-suspended", 48}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-uniformBuffers-06935", 59}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-unnormalizedCoordinates-09635", 68}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewMask-06178", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewType-07752", 53}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-03417", 58}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-03419", 58}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-04137", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-04138", 58}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-04139", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-07492", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-07493", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectCountNV-viewportCount-09421", 58}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-InputAttachmentIndex-09597", 61}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-MeshEXT-07091", 48}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshEXT Execution Model", 123}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-02691", 45}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-02692", 45}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-02693", 45}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-04876", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-04877", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-06479", 45}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-06537", 45}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-06666", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-06886", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-06887", 45}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07074", 45}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07075", 45}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07288", 45}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07469", 45}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07620", 45}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07621", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07622", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07623", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07624", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07625", 45}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07626", 45}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07627", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07629", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07630", 45}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07631", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07632", 45}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07633", 45}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07634", 45}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07636", 45}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07639", 45}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07640", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07641", 45}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07642", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07643", 45}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07644", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07645", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07646", 45}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07647", 45}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07648", 45}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07649", 45}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07749", 45}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07751", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07831", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07832", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07834", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07835", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07836", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07837", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07838", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07839", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07840", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07841", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07843", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07844", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07845", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07846", 45}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07847", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07848", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07849", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07850", 45}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07878", 45}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07879", 45}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07880", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07881", 45}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-07888", 45}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08114", 45}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08115", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08116", 45}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08117", 45}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08119", 45}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08600", 45}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08601", 45}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08605", 45}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08606", 45}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08608", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08609", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08610", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08611", 45}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08612", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08613", 45}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08617", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08636", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08637", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08644", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08666", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08669", 45}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08684", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08685", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08686", 45}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08687", 45}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08688", 45}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08689", 45}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08690", 45}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08693", 45}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08694", 45}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08695", 45}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08696", 45}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08698", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08699", 45}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08876", 45}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08877", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08878", 45}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-08879", 45}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09003", 45}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09116", 45}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09211", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09363", 45}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09364", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09365", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09368", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09369", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09548", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09549", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09600", 45}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09642", 45}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09643", 45}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-09650", 45}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10068", 45}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10608", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10677", 45}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10678", 45}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10679", 45}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10772", 45}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10795", 45}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10796", 45}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10797", 45}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10862", 45}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10864", 45}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10927", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-10928", 45}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11297", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11298", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11299", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11300", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11301", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11302", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11304", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11305", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11306", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11308", 45}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11309", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11372", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11373", 45}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11374", 45}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11376", 45}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11397", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11398", 45}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11437", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11438", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11439", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11441", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11442", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11485", 45}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11522", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11523", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11861", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11862", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11863", 45}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11865", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11866", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11868", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-11869", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-12338", 45}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-12339", 45}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-12340", 45}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-None-12472", 45}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchSADQCOM-06976", 65}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchSADQCOM-12420", 65}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchSSDQCOM-06974", 65}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchWindow-09215", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchWindow-09216", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchWindow-09217", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBlockMatchWindow-12421", 64}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageBoxFilterQCOM-06973", 61}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageSampleWeightedQCOM-06971", 66}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageSampleWeightedQCOM-06972", 66}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageSampleWeightedQCOM-06977", 66}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageSampleWeightedQCOM-06978", 66}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageWrite-04469", 53}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageWrite-08795", 53}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpImageWrite-08796", 53}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTensorReadARM-12462", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeImage-07027", 52}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeImage-07028", 52}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeImage-07029", 52}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeImage-07030", 52}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeImage-07468", 52}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-OpTypeTensorARM-09906", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-SampledType-04470", 52}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-SampledType-04471", 52}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-SampledType-04472", 52}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-SampledType-04473", 52}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-VkPipelineVieportCreateInfo-04141", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-VkPipelineVieportCreateInfo-04142", 68}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-addressModeU-12422", 53}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-advancedBlendMaxColorAttachments-07480", 73}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-alphaToCoverageEnable-08919", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-alphaToCoverageEnable-08920", 62}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-attachmentCount-07750", 56}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-blendEnable-04727", 52}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-buffer-02708", 47}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-buffer-02709", 47}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-buffer-parameter", 51}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-colorAttachmentCount-06179", 61}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-colorAttachmentCount-06185", 61}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-colorAttachmentCount-09362", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-colorAttachmentCount-09372", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-colorAttachmentCount-11871", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-02707", 54}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-02711", 54}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-10746", 54}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-commonparent", 47}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-conservativePointAndLineRasterization-07499", 78}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-coverageModulationTableEnable-07488", 70}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-coverageReductionMode-07491", 62}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-coverageToColorEnable-07490", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-cubicRangeClamp-09212", 56}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-customResolve-11524", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-customResolve-11525", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-customResolve-11529", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-customResolve-11530", 54}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dimensionCount-09905", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-drawCount-02718", 50}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-drawCount-02719", 50}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-drawCount-07088", 50}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawMeshTasksIndirectCommandEXT)", 141}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-drawCount-07089", 50}, { {"If drawCount is equal to 1, (offset + sizeof(VkDrawMeshTasksIndirectCommandEXT)) must be less than or equal to the size of buffer", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-drawCount-07090", 50}, { {"If drawCount is greater than 1, (stride x (drawCount - 1) + offset + sizeof(VkDrawMeshTasksIndirectCommandEXT)) must be less than or equal to the size of buffer", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingLocalRead-11797", 66}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08910", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08911", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08912", 74}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08913", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08914", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08915", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08916", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08917", 74}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-08918", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-11864", 74}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-11867", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-dynamicRenderingUnusedAttachments-11870", 74}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-filterCubic-02694", 52}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-filterCubicMinmax-02695", 58}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-firstAttachment-07476", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-firstAttachment-07478", 56}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-02696", 46}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-10582", 46}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-11521", 46}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-13361", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-13362", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-13363", 46}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-13364", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-flags-13365", 46}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-format-07753", 47}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageLayout-00344", 52}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06172", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06173", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06174", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06175", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06176", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06177", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06183", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-imageView-06184", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-index-11450", 46}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-layers-10831", 47}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-logicOp-04878", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-magFilter-04553", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-magFilter-09598", 50}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-maxFragmentDualSrcAttachments-09239", 70}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-maxMultiviewInstanceIndex-02688", 66}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-micromap-11636", 49}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-micromap-11637", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-micromap-11638", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-micromap-11639", 49}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-mipmapMode-04770", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-mipmapMode-09599", 51}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-multisampledRenderToSingleSampled-07285", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-multisampledRenderToSingleSampled-07286", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-multisampledRenderToSingleSampled-07287", 74}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-multiviewPerViewViewports-12262", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-multiviewPerViewViewports-12263", 66}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-nextStage-10745", 50}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-offset-02710", 47}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pBindInfo-11375", 50}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pColorAttachments-08963", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pColorAttachments-11539", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthAttachment-06186", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthAttachment-08964", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthAttachment-11540", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthAttachment-11872", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthInputAttachmentIndex-09595", 67}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDepthInputAttachmentIndex-09596", 67}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDescription-09900", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDynamicStates-08715", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pDynamicStates-08716", 55}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pNext-07935", 46}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pNext-12463", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pStages-10680", 48}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pStencilAttachment-06187", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pStencilAttachment-08965", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pStencilAttachment-11860", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pStencilAttachment-11873", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pipelineFragmentShadingRate-09238", 68}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-pipelineStatistics-07076", 59}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 90}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 83}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 86}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-protectedNoFault-11455", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-protectedNoFault-11456", 57}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizationSamples-07471", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizationSamples-07489", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizationSamples-07494", 61}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizerDiscardEnable-09236", 64}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizerDiscardEnable-09420", 64}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-rasterizerDiscardEnable-10863", 64}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-reductionMode-09213", 54}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-renderPass-02684", 51}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-renderPass-06198", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-renderpass", 45}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-02689", 62}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07484", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07485", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07486", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07487", 62}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07936", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07937", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsEnable-07938", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsPerPixel-07482", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sampleLocationsPerPixel-07483", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-samples-07472", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-samples-07473", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-scissorCount-03418", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-selectableCubicWeights-09214", 63}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-shadingRateImage-09233", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-shadingRateImage-09234", 57}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sparseImageInt64Atomics-04474", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-sparseImageInt64Atomics-04475", 64}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stage-06480", 46}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stage-07073", 46}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stippledLineEnable-07495", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stippledLineEnable-07496", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stippledLineEnable-07497", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-stippledLineEnable-07498", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-storageBuffers-06936", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-subpass-02685", 48}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-suspended", 44}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-uniformBuffers-06935", 55}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-unnormalizedCoordinates-09635", 64}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewMask-06178", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewType-07752", 49}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-03417", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-03419", 54}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-04137", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-04138", 54}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-04139", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-07492", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-07493", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectEXT-viewportCount-09421", 54}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-InputAttachmentIndex-09597", 60}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-MeshNV-07081", 46}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshNV Execution Model", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-02691", 44}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-02692", 44}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-02693", 44}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-04876", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-04877", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-06479", 44}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-06537", 44}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-06666", 44}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-06886", 44}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-06887", 44}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07074", 44}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07075", 44}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07288", 44}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07469", 44}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07620", 44}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07621", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07622", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07623", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07624", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07625", 44}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07626", 44}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07627", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07629", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07630", 44}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07631", 44}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07632", 44}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07633", 44}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07634", 44}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07636", 44}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07639", 44}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07640", 44}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07641", 44}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07642", 44}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07643", 44}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07644", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07645", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07646", 44}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07647", 44}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07648", 44}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07649", 44}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07749", 44}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07751", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07831", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07832", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07834", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07835", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07836", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07837", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07838", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07839", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07840", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07841", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07843", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07844", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07845", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07846", 44}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07847", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07848", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07849", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07850", 44}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07878", 44}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07879", 44}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07880", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07881", 44}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-07888", 44}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08114", 44}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08115", 44}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08116", 44}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08117", 44}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08119", 44}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08600", 44}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08601", 44}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08605", 44}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08606", 44}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08608", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08609", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08610", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08611", 44}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08612", 44}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08613", 44}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08617", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08636", 44}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08637", 44}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08644", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08666", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08669", 44}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08684", 44}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08685", 44}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08686", 44}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08687", 44}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08688", 44}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08689", 44}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08690", 44}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08693", 44}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08694", 44}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08695", 44}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08696", 44}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08698", 44}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08699", 44}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08876", 44}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08877", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08878", 44}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-08879", 44}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09003", 44}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09116", 44}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09211", 44}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09363", 44}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09364", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09365", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09368", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09369", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09548", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09549", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09600", 44}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09642", 44}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09643", 44}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-09650", 44}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10068", 44}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10608", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10677", 44}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10678", 44}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10679", 44}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10772", 44}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10795", 44}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10796", 44}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10797", 44}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10862", 44}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10864", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10927", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-10928", 44}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11297", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11298", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11299", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11300", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11301", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11302", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11304", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11305", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11306", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11308", 44}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11309", 44}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11372", 44}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11373", 44}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11374", 44}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11376", 44}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11397", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11398", 44}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11437", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11438", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11439", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11441", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11442", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11485", 44}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11522", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11523", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11861", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11862", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11863", 44}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11865", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11866", 44}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11868", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-11869", 44}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-12338", 44}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-12339", 44}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-12340", 44}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-None-12472", 44}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchSADQCOM-06976", 64}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchSADQCOM-12420", 64}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchSSDQCOM-06974", 64}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchWindow-09215", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchWindow-09216", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchWindow-09217", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBlockMatchWindow-12421", 63}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageBoxFilterQCOM-06973", 60}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageSampleWeightedQCOM-06971", 65}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageSampleWeightedQCOM-06972", 65}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageSampleWeightedQCOM-06977", 65}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageSampleWeightedQCOM-06978", 65}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageWrite-04469", 52}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageWrite-08795", 52}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpImageWrite-08796", 52}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTensorReadARM-12462", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeImage-07027", 51}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeImage-07028", 51}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeImage-07029", 51}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeImage-07030", 51}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeImage-07468", 51}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-OpTypeTensorARM-09906", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-SampledType-04470", 51}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-SampledType-04471", 51}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-SampledType-04472", 51}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-SampledType-04473", 51}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-VkPipelineVieportCreateInfo-04141", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-VkPipelineVieportCreateInfo-04142", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-addressModeU-12422", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-advancedBlendMaxColorAttachments-07480", 72}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-alphaToCoverageEnable-08919", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-alphaToCoverageEnable-08920", 61}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-attachmentCount-07750", 55}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-blendEnable-04727", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-buffer-02708", 46}, { {"If buffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 107}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-buffer-02709", 46}, { {"buffer must have been created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 89}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-buffer-parameter", 50}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-colorAttachmentCount-06179", 60}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-colorAttachmentCount-06185", 60}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-colorAttachmentCount-09362", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-colorAttachmentCount-09372", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-colorAttachmentCount-11871", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-02707", 53}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-02711", 53}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-10746", 53}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-commonparent", 46}, { {"Both of buffer, and commandBuffer must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-conservativePointAndLineRasterization-07499", 77}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-coverageModulationTableEnable-07488", 69}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-coverageReductionMode-07491", 61}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-coverageToColorEnable-07490", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-cubicRangeClamp-09212", 55}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-customResolve-11524", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-customResolve-11525", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-customResolve-11529", 53}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-customResolve-11530", 53}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dimensionCount-09905", 54}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-drawCount-02146", 49}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkDrawMeshTasksIndirectCommandNV)", 140}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-drawCount-02156", 49}, { {"If drawCount is equal to 1, (offset + sizeof(VkDrawMeshTasksIndirectCommandNV)) must be less than or equal to the size of buffer", 128}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-drawCount-02157", 49}, { {"If drawCount is greater than 1, (stride x (drawCount - 1) + offset + sizeof(VkDrawMeshTasksIndirectCommandNV)) must be less than or equal to the size of buffer", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-drawCount-02718", 49}, { {"If the multiDrawIndirect feature is not enabled, drawCount must be 0 or 1", 73}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-drawCount-02719", 49}, { {"drawCount must be less than or equal to VkPhysicalDeviceLimits::maxDrawIndirectCount", 84}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingLocalRead-11797", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08910", 73}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08911", 73}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08912", 73}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08913", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08914", 73}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08915", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08916", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08917", 73}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-08918", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-11864", 73}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-11867", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-dynamicRenderingUnusedAttachments-11870", 73}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-filterCubic-02694", 51}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-filterCubicMinmax-02695", 57}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-firstAttachment-07476", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-firstAttachment-07478", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-02696", 45}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-10582", 45}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-11521", 45}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-13361", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-13362", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-13363", 45}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-13364", 45}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-flags-13365", 45}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-format-07753", 46}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageLayout-00344", 51}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06172", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06173", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06174", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06175", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06176", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06177", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06183", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-imageView-06184", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-index-11450", 45}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-layers-10831", 46}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-logicOp-04878", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-magFilter-04553", 49}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-magFilter-09598", 49}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-maxFragmentDualSrcAttachments-09239", 69}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-maxMultiviewInstanceIndex-02688", 65}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-micromap-11636", 48}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-micromap-11637", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-micromap-11638", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-micromap-11639", 48}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-mipmapMode-04770", 50}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-mipmapMode-09599", 50}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-multisampledRenderToSingleSampled-07285", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-multisampledRenderToSingleSampled-07286", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-multisampledRenderToSingleSampled-07287", 73}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-multiviewPerViewViewports-12262", 65}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-multiviewPerViewViewports-12263", 65}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-nextStage-10745", 49}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-offset-02710", 46}, { {"offset must be a multiple of 4", 30}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pBindInfo-11375", 49}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pColorAttachments-08963", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pColorAttachments-11539", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthAttachment-06186", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthAttachment-08964", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthAttachment-11540", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthAttachment-11872", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthInputAttachmentIndex-09595", 66}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDepthInputAttachmentIndex-09596", 66}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDescription-09900", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDynamicStates-08715", 54}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pDynamicStates-08716", 54}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pNext-07935", 45}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pNext-12463", 45}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pStages-10680", 47}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pStencilAttachment-06187", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pStencilAttachment-08965", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pStencilAttachment-11860", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pStencilAttachment-11873", 58}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pipelineFragmentShadingRate-09238", 67}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-pipelineStatistics-07076", 58}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-primitiveFragmentShadingRateWithMultipleViewports-04552", 89}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-primitiveFragmentShadingRateWithMultipleViewports-08642", 89}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-primitivesGeneratedQueryWithNonZeroStreams-06709", 82}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-primitivesGeneratedQueryWithNonZeroStreams-07481", 82}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-primitivesGeneratedQueryWithRasterizerDiscard-06708", 85}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-protectedNoFault-11455", 56}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-protectedNoFault-11456", 56}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizationSamples-07471", 60}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizationSamples-07489", 60}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizationSamples-07494", 60}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizerDiscardEnable-09236", 63}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizerDiscardEnable-09420", 63}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-rasterizerDiscardEnable-10863", 63}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-reductionMode-09213", 53}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-renderPass-02684", 50}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-renderPass-06198", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-renderpass", 44}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-02689", 61}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07484", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07485", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07486", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07487", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07936", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07937", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsEnable-07938", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsPerPixel-07482", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sampleLocationsPerPixel-07483", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-samples-07472", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-samples-07473", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-scissorCount-03418", 52}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-selectableCubicWeights-09214", 62}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-shadingRateImage-09233", 56}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-shadingRateImage-09234", 56}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sparseImageInt64Atomics-04474", 63}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-sparseImageInt64Atomics-04475", 63}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stage-06480", 45}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stage-07073", 45}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stippledLineEnable-07495", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stippledLineEnable-07496", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stippledLineEnable-07497", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-stippledLineEnable-07498", 58}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-storageBuffers-06936", 54}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-subpass-02685", 47}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-suspended", 43}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-uniformBuffers-06935", 54}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-unnormalizedCoordinates-09635", 63}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewMask-06178", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewType-07752", 48}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-03417", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-03419", 53}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-04137", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-04138", 53}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-04139", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-07492", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-07493", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksIndirectNV-viewportCount-09421", 53}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-InputAttachmentIndex-09597", 52}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-MeshNV-07080", 38}, { {"The current pipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS must contain a shader stage using the MeshNV Execution Model", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-02691", 36}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-02692", 36}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-02693", 36}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-04876", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-04877", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-06479", 36}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-06537", 36}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-06666", 36}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-06886", 36}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-06887", 36}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07074", 36}, { {"Transform Feedback Queries must not be active", 45}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07075", 36}, { {"Primitives Generated Queries must not be active", 47}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07288", 36}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07469", 36}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07620", 36}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07621", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07622", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07623", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07624", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07625", 36}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07626", 36}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07627", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07629", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07630", 36}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07631", 36}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07632", 36}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07633", 36}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07634", 36}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07636", 36}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07639", 36}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07640", 36}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07641", 36}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07642", 36}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07643", 36}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07644", 36}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07645", 36}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07646", 36}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07647", 36}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07648", 36}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07649", 36}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07749", 36}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07751", 36}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07831", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07832", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07834", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07835", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07836", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07837", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07838", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07839", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07840", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07841", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07843", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07844", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07845", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07846", 36}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07847", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07848", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07849", 36}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07850", 36}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07878", 36}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07879", 36}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07880", 36}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07881", 36}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-07888", 36}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08114", 36}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08115", 36}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08116", 36}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08117", 36}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08119", 36}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08600", 36}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08601", 36}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08605", 36}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08606", 36}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08608", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08609", 36}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08610", 36}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08611", 36}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08612", 36}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08613", 36}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08617", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08636", 36}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08637", 36}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08644", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08666", 36}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08669", 36}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08684", 36}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08685", 36}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08686", 36}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08687", 36}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08688", 36}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08689", 36}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08690", 36}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08693", 36}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08694", 36}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created without the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, a valid VkShaderEXT must be bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08695", 36}, { {"If there is no bound graphics pipeline, and both the taskShader and meshShader features are enabled, and a valid VkShaderEXT is bound the to the VK_SHADER_STAGE_MESH_BIT_EXT stage, and that VkShaderEXT was created with the VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT flag, there must be no VkShaderEXT bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08696", 36}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08698", 36}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08699", 36}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08876", 36}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08877", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08878", 36}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-08879", 36}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09003", 36}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09116", 36}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09211", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09363", 36}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09364", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09365", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09368", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09369", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09548", 36}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09549", 36}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09600", 36}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09642", 36}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09643", 36}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-09650", 36}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10068", 36}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10608", 36}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10677", 36}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10678", 36}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10679", 36}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10772", 36}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10795", 36}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10796", 36}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10797", 36}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10862", 36}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10864", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10927", 36}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-10928", 36}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11297", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11298", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11299", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11300", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11301", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11302", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11304", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11305", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11306", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11308", 36}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11309", 36}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11372", 36}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11373", 36}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11374", 36}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11376", 36}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11397", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11398", 36}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11437", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11438", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11439", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11441", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11442", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11485", 36}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11522", 36}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11523", 36}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11861", 36}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11862", 36}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11863", 36}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11865", 36}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11866", 36}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11868", 36}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-11869", 36}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-12338", 36}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-12339", 36}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-12340", 36}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-None-12472", 36}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchSADQCOM-06976", 56}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchSADQCOM-12420", 56}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchSSDQCOM-06974", 56}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchWindow-09215", 55}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchWindow-09216", 55}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchWindow-09217", 55}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBlockMatchWindow-12421", 55}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageBoxFilterQCOM-06973", 52}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageSampleWeightedQCOM-06971", 57}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageSampleWeightedQCOM-06972", 57}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageSampleWeightedQCOM-06977", 57}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageSampleWeightedQCOM-06978", 57}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageWrite-04469", 44}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageWrite-08795", 44}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpImageWrite-08796", 44}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTensorReadARM-12462", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeImage-07027", 43}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeImage-07028", 43}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeImage-07029", 43}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeImage-07030", 43}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeImage-07468", 43}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-OpTypeTensorARM-09906", 47}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-SampledType-04470", 43}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-SampledType-04471", 43}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-SampledType-04472", 43}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-SampledType-04473", 43}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-VkPipelineVieportCreateInfo-04141", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-VkPipelineVieportCreateInfo-04142", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-addressModeU-12422", 44}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-advancedBlendMaxColorAttachments-07480", 64}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-alphaToCoverageEnable-08919", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-alphaToCoverageEnable-08920", 53}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-attachmentCount-07750", 47}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-blendEnable-04727", 43}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-colorAttachmentCount-06179", 52}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-colorAttachmentCount-06185", 52}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-colorAttachmentCount-09362", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-colorAttachmentCount-09372", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-colorAttachmentCount-11871", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-commandBuffer-02707", 45}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-commandBuffer-10746", 45}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-conservativePointAndLineRasterization-07499", 69}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-coverageModulationTableEnable-07488", 61}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-coverageReductionMode-07491", 53}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-coverageToColorEnable-07490", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-cubicRangeClamp-09212", 47}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-customResolve-11524", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-customResolve-11525", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-customResolve-11529", 45}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-customResolve-11530", 45}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dimensionCount-09905", 46}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingLocalRead-11797", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08910", 65}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08911", 65}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08912", 65}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08913", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08914", 65}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08915", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08916", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08917", 65}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-08918", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-11864", 65}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-11867", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-dynamicRenderingUnusedAttachments-11870", 65}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-filterCubic-02694", 43}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-filterCubicMinmax-02695", 49}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-firstAttachment-07476", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-firstAttachment-07478", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-02696", 37}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-10582", 37}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-11521", 37}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-13361", 37}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-13362", 37}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-13363", 37}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-13364", 37}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-flags-13365", 37}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-format-07753", 38}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageLayout-00344", 43}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06172", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06173", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06174", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06175", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06176", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06177", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06183", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-imageView-06184", 41}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-index-11450", 37}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-layers-10831", 38}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-logicOp-04878", 39}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-magFilter-04553", 41}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-magFilter-09598", 41}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-maxFragmentDualSrcAttachments-09239", 61}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-maxMultiviewInstanceIndex-02688", 57}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-micromap-11636", 40}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-micromap-11637", 40}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-micromap-11638", 40}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-micromap-11639", 40}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-mipmapMode-04770", 42}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-mipmapMode-09599", 42}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-multisampledRenderToSingleSampled-07285", 65}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-multisampledRenderToSingleSampled-07286", 65}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-multisampledRenderToSingleSampled-07287", 65}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-multiviewPerViewViewports-12262", 57}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-multiviewPerViewViewports-12263", 57}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-nextStage-10745", 41}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pBindInfo-11375", 41}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pColorAttachments-08963", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pColorAttachments-11539", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthAttachment-06186", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthAttachment-08964", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthAttachment-11540", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthAttachment-11872", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthInputAttachmentIndex-09595", 58}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDepthInputAttachmentIndex-09596", 58}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDescription-09900", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDynamicStates-08715", 46}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pDynamicStates-08716", 46}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pNext-07935", 37}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pNext-12463", 37}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pStages-10680", 39}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must not have last bound the pStages element of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT with a valid shader object other than VK_NULL_HANDLE", 306}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pStencilAttachment-06187", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pStencilAttachment-08965", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pStencilAttachment-11860", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pStencilAttachment-11873", 50}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pipelineFragmentShadingRate-09238", 59}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-pipelineStatistics-07076", 50}, { {"The pipelineStatistics member used to create any active Pipeline Statistics Query must not contain VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT, VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT, VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT, VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT, or VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-primitiveFragmentShadingRateWithMultipleViewports-04552", 81}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-primitiveFragmentShadingRateWithMultipleViewports-08642", 81}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-primitivesGeneratedQueryWithNonZeroStreams-06709", 74}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-primitivesGeneratedQueryWithNonZeroStreams-07481", 74}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-primitivesGeneratedQueryWithRasterizerDiscard-06708", 77}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-protectedNoFault-11455", 48}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-protectedNoFault-11456", 48}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizationSamples-07471", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizationSamples-07489", 52}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizationSamples-07494", 52}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizerDiscardEnable-09236", 55}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizerDiscardEnable-09420", 55}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-rasterizerDiscardEnable-10863", 55}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-reductionMode-09213", 45}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-renderPass-02684", 42}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-renderPass-06198", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-renderpass", 36}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-02689", 53}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07484", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07485", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07486", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07487", 53}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07936", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07937", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsEnable-07938", 53}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsPerPixel-07482", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sampleLocationsPerPixel-07483", 55}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-samples-07472", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-samples-07473", 39}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-scissorCount-03418", 44}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-selectableCubicWeights-09214", 54}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-shadingRateImage-09233", 48}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-shadingRateImage-09234", 48}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sparseImageInt64Atomics-04474", 55}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-sparseImageInt64Atomics-04475", 55}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stage-06480", 37}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stage-07073", 37}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stippledLineEnable-07495", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stippledLineEnable-07496", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stippledLineEnable-07497", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-stippledLineEnable-07498", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-storageBuffers-06936", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-subpass-02685", 39}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-taskCount-02119", 41}, { {"taskCount must be less than or equal to VkPhysicalDeviceMeshShaderPropertiesNV::maxDrawMeshTasksCount", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-uniformBuffers-06935", 46}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-unnormalizedCoordinates-09635", 55}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewMask-06178", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewType-07752", 40}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-03417", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-03419", 45}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-04137", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-04138", 45}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-04139", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-07492", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-07493", 45}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMeshTasksNV-viewportCount-09421", 45}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-Input-07939", 34}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-Input-08734", 34}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-InputAttachmentIndex-09597", 49}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-02691", 33}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-02692", 33}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-02693", 33}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-02721", 33}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04007", 33}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04008", 33}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04875", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04876", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04877", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04879", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04914", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-04933", 33}, { {"The multiDraw feature must be enabled", 37}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-06479", 33}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-06537", 33}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-06666", 33}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-06886", 33}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-06887", 33}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07288", 33}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07469", 33}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07619", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07620", 33}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07621", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07622", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07623", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07624", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07625", 33}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07626", 33}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07627", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07629", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07630", 33}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07631", 33}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07632", 33}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07633", 33}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07634", 33}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07636", 33}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07639", 33}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07640", 33}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07641", 33}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07642", 33}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07643", 33}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07644", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07645", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07646", 33}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07647", 33}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07648", 33}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07649", 33}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07749", 33}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07751", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07831", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07832", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07834", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07835", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07836", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07837", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07838", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07839", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07840", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07841", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07842", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07843", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07844", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07845", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07846", 33}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07847", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07848", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07849", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07850", 33}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07878", 33}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07879", 33}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07880", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07881", 33}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-07888", 33}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08114", 33}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08115", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08116", 33}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08117", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08119", 33}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08600", 33}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08601", 33}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08605", 33}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08606", 33}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08608", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08609", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08610", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08611", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08612", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08613", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08617", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08636", 33}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08637", 33}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08644", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08666", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08669", 33}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08684", 33}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08685", 33}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08686", 33}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08687", 33}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08688", 33}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08689", 33}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08690", 33}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08693", 33}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08696", 33}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08698", 33}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08699", 33}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08876", 33}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08877", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08878", 33}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08879", 33}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-08885", 33}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09003", 33}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09116", 33}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09203", 33}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09211", 33}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09363", 33}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09364", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09365", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09368", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09369", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09462", 33}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09548", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09549", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09600", 33}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09637", 33}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09642", 33}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09643", 33}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-09650", 33}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10068", 33}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10608", 33}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10677", 33}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10678", 33}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10679", 33}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10772", 33}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10795", 33}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10796", 33}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10797", 33}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10862", 33}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10864", 33}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10909", 33}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10927", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-10928", 33}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11297", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11298", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11299", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11300", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11301", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11302", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11304", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11305", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11306", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11308", 33}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11309", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11372", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11373", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11374", 33}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11376", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11397", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11398", 33}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11437", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11438", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11439", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11441", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11442", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11485", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11522", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11523", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11861", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11862", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11863", 33}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11865", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11866", 33}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11868", 33}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-11869", 33}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-12338", 33}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-12339", 33}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-12340", 33}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-12472", 33}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-None-13118", 33}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpExecutionMode-12239", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpExecutionMode-12240", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpExecutionMode-12241", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpExecutionMode-12242", 44}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchSADQCOM-06976", 53}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchSADQCOM-12420", 53}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchSSDQCOM-06974", 53}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchWindow-09215", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchWindow-09216", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchWindow-09217", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBlockMatchWindow-12421", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageBoxFilterQCOM-06973", 49}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageSampleWeightedQCOM-06971", 54}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageSampleWeightedQCOM-06972", 54}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageSampleWeightedQCOM-06977", 54}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageSampleWeightedQCOM-06978", 54}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageWrite-04469", 41}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageWrite-08795", 41}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpImageWrite-08796", 41}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTensorReadARM-12462", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeImage-07027", 40}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeImage-07028", 40}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeImage-07029", 40}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeImage-07030", 40}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeImage-07468", 40}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-OpTypeTensorARM-09906", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-SampledType-04470", 40}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-SampledType-04471", 40}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-SampledType-04472", 40}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-SampledType-04473", 40}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-VkPipelineVieportCreateInfo-04141", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-VkPipelineVieportCreateInfo-04142", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-addressModeU-12422", 41}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-advancedBlendMaxColorAttachments-07480", 61}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-alphaToCoverageEnable-08919", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-alphaToCoverageEnable-08920", 50}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-attachmentCount-07750", 44}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-blendEnable-04727", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-colorAttachmentCount-06179", 49}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-colorAttachmentCount-06185", 49}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-colorAttachmentCount-09362", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-colorAttachmentCount-09372", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-colorAttachmentCount-11871", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-02707", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-02712", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-02713", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-04617", 42}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-10746", 42}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-conservativePointAndLineRasterization-07499", 66}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-coverageModulationTableEnable-07488", 58}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-coverageReductionMode-07491", 50}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-coverageToColorEnable-07490", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-cubicRangeClamp-09212", 44}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-customResolve-11524", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-customResolve-11525", 42}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-customResolve-11529", 42}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-customResolve-11530", 42}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dimensionCount-09905", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-drawCount-04934", 38}, { {"drawCount must be less than VkPhysicalDeviceMultiDrawPropertiesEXT::maxMultiDrawCount", 85}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-drawCount-04935", 38}, { {"If drawCount is greater than zero, pVertexInfo must be a valid pointer to memory containing one or more valid instances of VkMultiDrawInfoEXT structures", 152}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-drawCount-09628", 38}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkMultiDrawInfoEXT)", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicPrimitiveTopologyUnrestricted-07500", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingLocalRead-11797", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08910", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08911", 62}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08912", 62}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08913", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08914", 62}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08915", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08916", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08917", 62}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-08918", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-11864", 62}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-11867", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-dynamicRenderingUnusedAttachments-11870", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-filterCubic-02694", 40}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-filterCubicMinmax-02695", 46}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-firstAttachment-07476", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-firstAttachment-07478", 44}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-02696", 34}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-10582", 34}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-11521", 34}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-13361", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-13362", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-13363", 34}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-13364", 34}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-flags-13365", 34}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-format-07753", 35}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-format-08936", 35}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-format-08937", 35}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-format-10389", 35}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-format-10390", 35}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageLayout-00344", 40}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06172", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06173", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06174", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06175", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06176", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06177", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06183", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-imageView-06184", 38}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-index-11450", 34}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-layers-10831", 35}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-logicOp-04878", 36}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-magFilter-04553", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-magFilter-09598", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-maxFragmentDualSrcAttachments-09239", 58}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-maxMultiviewInstanceIndex-02688", 54}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-micromap-11636", 37}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-micromap-11637", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-micromap-11638", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-micromap-11639", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-mipmapMode-04770", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-mipmapMode-09599", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-multisampledRenderToSingleSampled-07285", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-multisampledRenderToSingleSampled-07286", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-multisampledRenderToSingleSampled-07287", 62}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-multiviewPerViewViewports-12262", 54}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-multiviewPerViewViewports-12263", 54}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-nextStage-10745", 38}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pBindInfo-11375", 38}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pColorAttachments-08963", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pColorAttachments-11539", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthAttachment-06186", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthAttachment-08964", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthAttachment-11540", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthAttachment-11872", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthInputAttachmentIndex-09595", 55}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDepthInputAttachmentIndex-09596", 55}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDescription-09900", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDynamicStates-08715", 43}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pDynamicStates-08716", 43}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pNext-07935", 34}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pNext-09461", 34}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pNext-12463", 34}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pStencilAttachment-06187", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pStencilAttachment-08965", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pStencilAttachment-11860", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pStencilAttachment-11873", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pStrides-04913", 37}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-pipelineFragmentShadingRate-09238", 56}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 78}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 78}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitiveTopology-10286", 46}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitiveTopology-10747", 46}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitiveTopology-10748", 46}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 71}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 71}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 74}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-protectedNoFault-11455", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-protectedNoFault-11456", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizationSamples-07471", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizationSamples-07489", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizationSamples-07494", 49}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizerDiscardEnable-09236", 52}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizerDiscardEnable-09420", 52}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-rasterizerDiscardEnable-10863", 52}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-reductionMode-09213", 42}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-renderPass-02684", 39}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-renderPass-06198", 39}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-renderpass", 33}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-02689", 50}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07484", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07485", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07486", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07487", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07936", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07937", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsEnable-07938", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsPerPixel-07482", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sampleLocationsPerPixel-07483", 52}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-samples-07472", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-samples-07473", 36}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-scissorCount-03418", 41}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-selectableCubicWeights-09214", 51}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-shadingRateImage-09233", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-shadingRateImage-09234", 45}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sparseImageInt64Atomics-04474", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-sparseImageInt64Atomics-04475", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stage-06481", 34}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stage-07073", 34}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stippledLineEnable-07495", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stippledLineEnable-07496", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stippledLineEnable-07497", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-stippledLineEnable-07498", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-storageBuffers-06936", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-subpass-02685", 36}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-uniformBuffers-06935", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-unnormalizedCoordinates-09635", 52}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewMask-06178", 37}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewType-07752", 37}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-03417", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-03419", 42}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-04137", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-04138", 42}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-04139", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-07492", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-07493", 42}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiEXT-viewportCount-09421", 42}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-Input-07939", 41}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-Input-08734", 41}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-InputAttachmentIndex-09597", 56}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-02691", 40}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-02692", 40}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-02693", 40}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-02721", 40}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04007", 40}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04008", 40}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04875", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04876", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04877", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04879", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04914", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-04937", 40}, { {"The multiDraw feature must be enabled", 37}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-06479", 40}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-06537", 40}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-06666", 40}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-06886", 40}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-06887", 40}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07288", 40}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07312", 40}, { {"If the maintenance6 feature is not enabled, a valid index buffer must be bound", 78}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07469", 40}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07619", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07620", 40}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07621", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07622", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07623", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07624", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07625", 40}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07626", 40}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07627", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07629", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07630", 40}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07631", 40}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07632", 40}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07633", 40}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07634", 40}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07636", 40}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07639", 40}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07640", 40}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07641", 40}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07642", 40}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07643", 40}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07644", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07645", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07646", 40}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07647", 40}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07648", 40}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07649", 40}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07749", 40}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07751", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07831", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07832", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07834", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07835", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07836", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07837", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07838", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07839", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07840", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07841", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07842", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07843", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07844", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07845", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07846", 40}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07847", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07848", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07849", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07850", 40}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07878", 40}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07879", 40}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07880", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07881", 40}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-07888", 40}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08114", 40}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08115", 40}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08116", 40}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08117", 40}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08119", 40}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08600", 40}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08601", 40}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08605", 40}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08606", 40}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08608", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08609", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08610", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08611", 40}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08612", 40}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08613", 40}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08617", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08636", 40}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08637", 40}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08644", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08666", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08669", 40}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08684", 40}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08685", 40}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08686", 40}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08687", 40}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08688", 40}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08689", 40}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08690", 40}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08693", 40}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08696", 40}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08698", 40}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08699", 40}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08876", 40}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08877", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08878", 40}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08879", 40}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-08885", 40}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09003", 40}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09116", 40}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09203", 40}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09211", 40}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09363", 40}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09364", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09365", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09368", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09369", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09462", 40}, { {"If shader objects are used for drawing or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, any member of the pVertexBindingDescriptions parameter to the vkCmdSetVertexInputEXT call that sets this dynamic state has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 434}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09548", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09549", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09600", 40}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09637", 40}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09642", 40}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09643", 40}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-09650", 40}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10068", 40}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10608", 40}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10677", 40}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10678", 40}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10679", 40}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10772", 40}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10795", 40}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10796", 40}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10797", 40}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10862", 40}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10864", 40}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10909", 40}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10927", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-10928", 40}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11297", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11298", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11299", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11300", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11301", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11302", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11304", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11305", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11306", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11308", 40}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11309", 40}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11372", 40}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11373", 40}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11374", 40}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11376", 40}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11397", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11398", 40}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11437", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11438", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11439", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11441", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11442", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11485", 40}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11522", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11523", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11861", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11862", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11863", 40}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11865", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11866", 40}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11868", 40}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-11869", 40}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-12338", 40}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-12339", 40}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-12340", 40}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-12472", 40}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-None-13118", 40}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpExecutionMode-12239", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpExecutionMode-12240", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpExecutionMode-12241", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpExecutionMode-12242", 51}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchSADQCOM-06976", 60}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchSADQCOM-12420", 60}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchSSDQCOM-06974", 60}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchWindow-09215", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchWindow-09216", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchWindow-09217", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBlockMatchWindow-12421", 59}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageBoxFilterQCOM-06973", 56}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageSampleWeightedQCOM-06971", 61}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageSampleWeightedQCOM-06972", 61}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageSampleWeightedQCOM-06977", 61}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageSampleWeightedQCOM-06978", 61}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageWrite-04469", 48}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageWrite-08795", 48}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpImageWrite-08796", 48}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTensorReadARM-12462", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeImage-07027", 47}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeImage-07028", 47}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeImage-07029", 47}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeImage-07030", 47}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeImage-07468", 47}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-OpTypeTensorARM-09906", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-SampledType-04470", 47}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-SampledType-04471", 47}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-SampledType-04472", 47}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-SampledType-04473", 47}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-VkPipelineVieportCreateInfo-04141", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-VkPipelineVieportCreateInfo-04142", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-addressModeU-12422", 48}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-advancedBlendMaxColorAttachments-07480", 68}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-alphaToCoverageEnable-08919", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-alphaToCoverageEnable-08920", 57}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-attachmentCount-07750", 51}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-blendEnable-04727", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-colorAttachmentCount-06179", 56}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-colorAttachmentCount-06185", 56}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-colorAttachmentCount-09362", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-colorAttachmentCount-09372", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-colorAttachmentCount-11871", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-02707", 49}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-02712", 49}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, any resource written to by the VkPipeline object bound to the pipeline bind point used by this command must not be an unprotected resource", 224}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-02713", 49}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pipeline stages other than the framebuffer-space and compute stages in the VkPipeline object bound to the pipeline bind point used by this command must not write to any resource", 263}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-04617", 49}, { {"If any of the shader stages of the VkPipeline bound to the pipeline bind point used by this command uses the RayQueryKHR capability, then commandBuffer must not be a protected command buffer", 190}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-10746", 49}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-conservativePointAndLineRasterization-07499", 73}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-coverageModulationTableEnable-07488", 65}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-coverageReductionMode-07491", 57}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-coverageToColorEnable-07490", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-cubicRangeClamp-09212", 51}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-customResolve-11524", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-customResolve-11525", 49}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-customResolve-11529", 49}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-customResolve-11530", 49}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dimensionCount-09905", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-drawCount-04939", 45}, { {"drawCount must be less than VkPhysicalDeviceMultiDrawPropertiesEXT::maxMultiDrawCount", 85}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-drawCount-04940", 45}, { {"If drawCount is greater than zero, pIndexInfo must be a valid pointer to memory containing one or more valid instances of VkMultiDrawIndexedInfoEXT structures", 158}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-drawCount-09629", 45}, { {"If drawCount is greater than 1, stride must be a multiple of 4 and must be greater than or equal to sizeof(VkMultiDrawIndexedInfoEXT)", 133}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicPrimitiveTopologyUnrestricted-07500", 72}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingLocalRead-11797", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08910", 69}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08911", 69}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08912", 69}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08913", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08914", 69}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08915", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08916", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08917", 69}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-08918", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-11864", 69}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-11867", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-dynamicRenderingUnusedAttachments-11870", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-filterCubic-02694", 47}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-filterCubicMinmax-02695", 53}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-firstAttachment-07476", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-firstAttachment-07478", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-02696", 41}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-10582", 41}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-11521", 41}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-13361", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-13362", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-13363", 41}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-13364", 41}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-flags-13365", 41}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-format-07753", 42}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-format-08936", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-format-08937", 42}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-format-10389", 42}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-format-10390", 42}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageLayout-00344", 47}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06172", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06173", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06174", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06175", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06176", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06177", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06183", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-imageView-06184", 45}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-index-11450", 41}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-layers-10831", 42}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-logicOp-04878", 43}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-magFilter-04553", 45}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-magFilter-09598", 45}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-maxFragmentDualSrcAttachments-09239", 65}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-maxMultiviewInstanceIndex-02688", 61}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-micromap-11636", 44}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-micromap-11637", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-micromap-11638", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-micromap-11639", 44}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-mipmapMode-04770", 46}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-mipmapMode-09599", 46}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-multisampledRenderToSingleSampled-07285", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-multisampledRenderToSingleSampled-07286", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-multisampledRenderToSingleSampled-07287", 69}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-multiviewPerViewViewports-12262", 61}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-multiviewPerViewViewports-12263", 61}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-nextStage-10745", 45}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pBindInfo-11375", 45}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pColorAttachments-08963", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pColorAttachments-11539", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthAttachment-06186", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthAttachment-08964", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthAttachment-11540", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthAttachment-11872", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthInputAttachmentIndex-09595", 62}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDepthInputAttachmentIndex-09596", 62}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDescription-09900", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDynamicStates-08715", 50}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pDynamicStates-08716", 50}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pNext-07935", 41}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pNext-09461", 41}, { {"If the bound graphics pipeline state was created with VkPipelineVertexInputDivisorStateCreateInfo in the pNext chain of VkGraphicsPipelineCreateInfo::pVertexInputState, any member of VkPipelineVertexInputDivisorStateCreateInfo::pVertexBindingDivisors has a value other than 1 in divisor, and VkPhysicalDeviceVertexAttributeDivisorProperties::supportsNonZeroFirstInstance is VK_FALSE, then firstInstance must be 0", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pNext-12463", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pStencilAttachment-06187", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pStencilAttachment-08965", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pStencilAttachment-11860", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pStencilAttachment-11873", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pStrides-04913", 44}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pVertexOffset-parameter", 53}, { {"If pVertexOffset is not NULL, pVertexOffset must be a valid pointer to a valid int32_t value", 92}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-pipelineFragmentShadingRate-09238", 63}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 85}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 85}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveRestartIndex-12401", 57}, { {"If the primitive restart index for this draw operation was set with vkCmdSetPrimitiveRestartIndexEXT, then primitiveRestartIndex must not be greater than the maximum representable value for the bound VkIndexType", 211}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveTopology-10286", 53}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveTopology-10747", 53}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitiveTopology-10748", 53}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 78}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 78}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 81}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-protectedNoFault-11455", 52}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-protectedNoFault-11456", 52}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizationSamples-07471", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizationSamples-07489", 56}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizationSamples-07494", 56}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizerDiscardEnable-09236", 59}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizerDiscardEnable-09420", 59}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-rasterizerDiscardEnable-10863", 59}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-reductionMode-09213", 49}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-renderPass-02684", 46}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-renderPass-06198", 46}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-renderpass", 40}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-robustBufferAccess2-08798", 55}, { {"If the robustBufferAccess2 feature is not enabled, (indexSize x (firstIndex + indexCount)) must be less than or equal to the size of the bound index buffer range, with indexSize being based on the type specified by indexType, and the other parameters sourced from this command", 276}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-02689", 57}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07484", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07485", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07486", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07487", 57}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07936", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07937", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsEnable-07938", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsPerPixel-07482", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sampleLocationsPerPixel-07483", 59}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-samples-07472", 43}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-samples-07473", 43}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-scissorCount-03418", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-selectableCubicWeights-09214", 58}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-shadingRateImage-09233", 52}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-shadingRateImage-09234", 52}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sparseImageInt64Atomics-04474", 59}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-sparseImageInt64Atomics-04475", 59}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stage-06481", 41}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stage-07073", 41}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stippledLineEnable-07495", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stippledLineEnable-07496", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stippledLineEnable-07497", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-stippledLineEnable-07498", 54}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-storageBuffers-06936", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-subpass-02685", 43}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-suspended", 39}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-uniformBuffers-06935", 50}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-unnormalizedCoordinates-09635", 59}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewMask-06178", 44}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewType-07752", 44}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-03417", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-03419", 49}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-04137", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-04138", 49}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-04139", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-07492", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-07493", 49}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdDrawMultiIndexedEXT-viewportCount-09421", 49}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEncodeVideoKHR-None-07012", 35}, { {"The bound video session must not be in uninitialized state at the time the command is executed on the device", 108}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-None-08250", 35}, { {"The bound video session must have been created with an encode operation", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-None-08318", 35}, { {"If there is a bound video session parameters object, then it must have been created with the currently set video encode quality level for the bound video session at the time the command is executed on the device", 211}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-StdVideoH264PictureParameterSet-08227", 62}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the bound video session parameters object must contain a StdVideoH264PictureParameterSet entry with seq_parameter_set_id and pic_parameter_set_id matching StdVideoEncodeH264PictureInfo::seq_parameter_set_id and StdVideoEncodeH264PictureInfo::pic_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 538}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-StdVideoH264SequenceParameterSet-08226", 63}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the bound video session parameters object must contain a StdVideoH264SequenceParameterSet entry with seq_parameter_set_id matching StdVideoEncodeH264PictureInfo::seq_parameter_set_id that is provided in the pStdPictureInfo member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 442}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-StdVideoH265PictureParameterSet-08233", 62}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265PictureParameterSet entry with sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id matching StdVideoEncodeH265PictureInfo::sps_video_parameter_set_id, StdVideoEncodeH265PictureInfo::pps_seq_parameter_set_id, and StdVideoEncodeH265PictureInfo::pps_pic_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 643}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-StdVideoH265SequenceParameterSet-08232", 63}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265SequenceParameterSet entry with sps_video_parameter_set_id and sps_seq_parameter_set_id matching StdVideoEncodeH265PictureInfo::sps_video_parameter_set_id and StdVideoEncodeH265PictureInfo::pps_seq_parameter_set_id, respectively, that are provided in the pStdPictureInfo member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 559}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-StdVideoH265VideoParameterSet-08231", 60}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the bound video session parameters object must contain a StdVideoH265VideoParameterSet entry with vps_video_parameter_set_id matching StdVideoEncodeH265PictureInfo::sps_video_parameter_set_id that is provided in the pStdPictureInfo member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 451}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-activeReferencePictureCount-08216", 58}, { {"activeReferencePictureCount must be less than or equal to the VkVideoSessionCreateInfoKHR::maxActiveReferencePictures specified when the bound video session was created", 168}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-bufferlevel", 36}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-codedOffset-08218", 42}, { {"The codedOffset member of the VkVideoPictureResourceInfoKHR structure pointed to by the pPictureResource member of each element of pEncodeInfo->pReferenceSlots must be an integer multiple of codedOffsetGranularity", 213}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-08202", 44}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, then pEncodeInfo->dstBuffer must not be a protected buffer", 147}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-08203", 44}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, then pEncodeInfo->dstBuffer must be a protected buffer", 140}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-08211", 44}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, then pEncodeInfo->srcPictureResource.imageViewBinding must not have been created from a protected image", 192}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 109}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQIndex-10320", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then the constantQIndex member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be zero", 341}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQIndex-10321", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the current rate control mode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then the constantQIndex member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be between VkVideoEncodeAV1CapabilitiesKHR::minQIndex and VkVideoEncodeAV1CapabilitiesKHR::maxQIndex, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 549}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08269", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then VkVideoEncodeH264NaluSliceInfoKHR::constantQp must be zero for each element of the pNaluSliceEntries member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 412}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08270", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and the current rate control mode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then VkVideoEncodeH264NaluSliceInfoKHR::constantQp must be between VkVideoEncodeH264CapabilitiesKHR::minQp and VkVideoEncodeH264CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, for each element of the pNaluSliceEntries member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 615}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08271", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H264_CAPABILITY_PER_SLICE_CONSTANT_QP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then VkVideoEncodeH264NaluSliceInfoKHR::constantQp must have the same value for each element of the pNaluSliceEntries member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 574}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08272", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the current rate control mode is not VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then VkVideoEncodeH265NaluSliceSegmentInfoKHR::constantQp must be zero for each element of the pNaluSliceSegmentEntries member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 426}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08273", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the current rate control mode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then VkVideoEncodeH265NaluSliceSegmentInfoKHR::constantQp must be between VkVideoEncodeH265CapabilitiesKHR::minQp and VkVideoEncodeH265CapabilitiesKHR::maxQp, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, for each element of the pNaluSliceSegmentEntries member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 629}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-constantQp-08274", 41}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H265_CAPABILITY_PER_SLICE_SEGMENT_CONSTANT_QP_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then VkVideoEncodeH265NaluSliceSegmentInfoKHR::constantQp must have the same value for each element of the pNaluSliceSegmentEntries member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 596}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-dpbFrameUseCount-08221", 47}, { {"All elements of dpbFrameUseCount must be less than or equal to 1", 64}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-08342", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H264_CAPABILITY_B_FRAME_IN_L0_LIST_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then each element of h264L0PictureTypes must not be STD_VIDEO_H264_PICTURE_TYPE_B", 436}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-08343", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H264_CAPABILITY_B_FRAME_IN_L1_LIST_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then each element of h264L1PictureTypes must not be STD_VIDEO_H264_PICTURE_TYPE_B", 436}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-08347", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H265_CAPABILITY_B_FRAME_IN_L0_LIST_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then each element of h265L0PictureTypes must not be STD_VIDEO_H264_PICTURE_TYPE_B", 436}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-08348", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H265_CAPABILITY_B_FRAME_IN_L1_LIST_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then each element of h265L1PictureTypes must not be STD_VIDEO_H265_PICTURE_TYPE_B", 436}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10322", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_FRAME_SIZE_OVERRIDE_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo->flags.frame_size_override_flag must be zero for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 542}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10323", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_FRAME_SIZE_OVERRIDE_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then pEncodeInfo->srcPictureResource.codedExtent.width must equal StdVideoAV1SequenceHeader::max_frame_width_minus_1 + 1 of the active AV1 sequence header", 507}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10324", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_FRAME_SIZE_OVERRIDE_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then pEncodeInfo->srcPictureResource.codedExtent.height must equal StdVideoAV1SequenceHeader::max_frame_height_minus_1 + 1 of the active AV1 sequence header", 509}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10325", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_MOTION_VECTOR_SCALING_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then for each element i of pEncodeInfo->pReferenceSlots pEncodeInfo->pReferenceSlots[i].pPictureResource->codedExtent must match pEncodeInfo->srcPictureResource.codedExtent", 527}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10850", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_DIFFERENT_SLICE_TYPE_BIT_KHR, then pNaluSliceEntries[i].pStdSliceHeader->slice_type must be identical for all elements i of the pNaluSliceEntries member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo , except for element index i equal to intraRefreshH264SliceIndex, if intraRefreshH264SliceIndex is defined", 679}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-flags-10856", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the used video profile, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_DIFFERENT_SLICE_SEGMENT_TYPE_BIT_KHR, then pNaluSliceSegmentEntries[i].pStdSliceSegmentHeader->slice_type must be identical for all elements i of the pNaluSliceSegmentEntries member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo , except for element index i equal to intraRefreshH265SliceSegmentIndex, if intraRefreshH265SliceSegmentIndex is defined", 686}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-h264PictureType-10849", 46}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, h264PictureType is STD_VIDEO_H264_PICTURE_TYPE_B, and VkVideoEncodeH264CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, does not include VK_VIDEO_ENCODE_H264_CAPABILITY_B_PICTURE_INTRA_REFRESH_BIT_KHR, then pEncodeInfo->flags must not include VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR", 490}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-h265PictureType-10855", 46}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, h265PictureType is STD_VIDEO_H265_PICTURE_TYPE_B, and VkVideoEncodeH265CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, does not include VK_VIDEO_ENCODE_H265_CAPABILITY_B_PICTURE_INTRA_REFRESH_BIT_KHR, then pEncodeInfo->flags must not include VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR", 490}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-intraRefreshH264SliceIndex-10847", 57}, { {"If intraRefreshH264SliceIndex is defined, then pNaluSliceEntries[intraRefreshH264SliceIndex].pStdSliceHeader->slice_type must be STD_VIDEO_H264_SLICE_TYPE_I in the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 248}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-intraRefreshH265SliceSegmentIndex-10853", 64}, { {"If intraRefreshH265SliceSegmentIndex is defined, then pNaluSliceSegmentEntries[intraRefreshH265SliceSegmentIndex].pStdSliceSegmentHeader->slice_type must be STD_VIDEO_H265_SLICE_TYPE_I in the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 276}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxBPictureL0ReferenceCount-08341", 58}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::maxBPictureL0ReferenceCount and VkVideoEncodeH264CapabilitiesKHR::maxL1ReferenceCount are both zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then h264PictureType and each element of h264L0PictureTypes and h264L1PictureTypes must not be STD_VIDEO_H264_PICTURE_TYPE_B", 497}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxBPictureL0ReferenceCount-08346", 58}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::maxBPictureL0ReferenceCount and VkVideoEncodeH265CapabilitiesKHR::maxL1ReferenceCount are both zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then h265PictureType and each element of h265L0PictureTypes and h265L1PictureTypes must not be STD_VIDEO_H265_PICTURE_TYPE_B", 497}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxBidirectionalCompoundReferenceCount-10332", 69}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::maxBidirectionalCompoundReferenceCount is zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must not be VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_BIDIRECTIONAL_COMPOUND_KHR", 514}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxPPictureL0ReferenceCount-08340", 58}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::maxPPictureL0ReferenceCount is zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then h264PictureType and each element of h264L0PictureTypes and h264L1PictureTypes must not be STD_VIDEO_H264_PICTURE_TYPE_P", 433}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxPPictureL0ReferenceCount-08345", 58}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::maxPPictureL0ReferenceCount is zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then h265PictureType and each element of h265L0PictureTypes and h265L1PictureTypes must not be STD_VIDEO_H265_PICTURE_TYPE_P", 433}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxSingleReferenceCount-10328", 54}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::maxSingleReferenceCount is zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must not be VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_SINGLE_REFERENCE_KHR", 493}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-maxUnidirectionalCompoundReferenceCount-10330", 70}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and VkVideoEncodeAV1CapabilitiesKHR::maxUnidirectionalCompoundReferenceCount is zero, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must not be VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_UNIDIRECTIONAL_COMPOUND_KHR", 516}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-naluSliceEntryCount-08302", 50}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the naluSliceEntryCount member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.width multiplied by minCodingBlockExtent.height", 342}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-naluSliceEntryCount-08312", 50}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and VkVideoEncodeH264CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H264_CAPABILITY_ROW_UNALIGNED_SLICE_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceEntryCount member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.height", 541}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-naluSliceSegmentEntryCount-08307", 57}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the naluSliceSegmentEntryCount member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.width multiplied by minCodingBlockExtent.height", 349}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-naluSliceSegmentEntryCount-08313", 57}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and VkVideoEncodeH265CapabilitiesKHR::flags does not include VK_VIDEO_ENCODE_H265_CAPABILITY_ROW_UNALIGNED_SLICE_SEGMENT_BIT_KHR, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceSegmentEntryCount member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.height", 556}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-opCount-07174", 38}, { {"For each active query, the active query index corresponding to the query type of that query plus opCount must be less than or equal to the last activatable query index corresponding to the query type of that query plus one", 222}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08201", 42}, { {"pEncodeInfo->dstBuffer must be compatible with the video profile the bound video session was created with", 105}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08204", 42}, { {"pEncodeInfo->dstBufferOffset must be an integer multiple of VkVideoCapabilitiesKHR::minBitstreamBufferOffsetAlignment, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 236}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08205", 42}, { {"pEncodeInfo->dstBufferRange must be an integer multiple of VkVideoCapabilitiesKHR::minBitstreamBufferSizeAlignment, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 233}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08206", 42}, { {"pEncodeInfo->srcPictureResource.imageViewBinding must be compatible with the video profile the bound video session was created with", 131}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08207", 42}, { {"The format of pEncodeInfo->srcPictureResource.imageViewBinding must match the VkVideoSessionCreateInfoKHR::pictureFormat the bound video session was created with", 161}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08208", 42}, { {"pEncodeInfo->srcPictureResource.codedOffset must be an integer multiple of codedOffsetGranularity", 97}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08209", 42}, { {"pEncodeInfo->srcPictureResource.codedExtent must be between minCodedExtent and maxCodedExtent, inclusive, the bound video session was created with", 146}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08210", 42}, { {"pEncodeInfo->srcPictureResource.imageViewBinding must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR usage flag set", 135}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08213", 42}, { {"If pEncodeInfo->pSetupReferenceSlot is not NULL, then pEncodeInfo->pSetupReferenceSlot->slotIndex must be less than the VkVideoSessionCreateInfoKHR::maxDpbSlots specified when the bound video session was created", 211}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08214", 42}, { {"If pEncodeInfo->pSetupReferenceSlot is not NULL, then pEncodeInfo->pSetupReferenceSlot->pPictureResource->codedOffset must be an integer multiple of codedOffsetGranularity", 171}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08215", 42}, { {"If pEncodeInfo->pSetupReferenceSlot is not NULL, then pEncodeInfo->pSetupReferenceSlot->pPictureResource must match one of the bound reference picture resource", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08228", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and pEncodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pEncodeInfo->pSetupReferenceSlot must include a VkVideoEncodeH264DpbSlotInfoKHR structure", 278}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08234", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and pEncodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pEncodeInfo->pSetupReferenceSlot must include a VkVideoEncodeH265DpbSlotInfoKHR structure", 278}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-08377", 42}, { {"pEncodeInfo->pSetupReferenceSlot must not be NULL unless the bound video session was created with VkVideoSessionCreateInfoKHR::maxDpbSlots equal to zero", 152}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10306", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_QUANTIZATION_DELTA_MAP_BIT_KHR, then the bound video session must have been created with VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10307", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_EMPHASIS_MAP_BIT_KHR, then the bound video session must have been created with VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_EMPHASIS_MAP_BIT_KHR", 188}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10308", 42}, { {"If the current rate control mode is VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR or VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR, then pEncodeInfo->flags must not include VK_VIDEO_ENCODE_WITH_EMPHASIS_MAP_BIT_KHR", 219}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10309", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_VIDEO_ENCODE_WITH_EMPHASIS_MAP_BIT_KHR, then the pNext chain of pEncodeInfo must include a VkVideoEncodeQuantizationMapInfoKHR structure with its quantizationMap member specifying a valid VkImageView handle", 295}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10310", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_QUANTIZATION_DELTA_MAP_BIT_KHR or VK_VIDEO_ENCODE_WITH_EMPHASIS_MAP_BIT_KHR, then the VkImageView specified by the quantizationMap member of the VkVideoEncodeQuantizationMapInfoKHR structure included in the pNext chain must be compatible with the video profile the bound video session was created with", 353}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10311", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_QUANTIZATION_DELTA_MAP_BIT_KHR, then the VkImageView specified by the quantizationMap member of the VkVideoEncodeQuantizationMapInfoKHR structure included in the pNext chain of pEncodeInfo must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_QUANTIZATION_DELTA_MAP_BIT_KHR usage flag set", 346}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10312", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_WITH_EMPHASIS_MAP_BIT_KHR, then the VkImageView specified by the quantizationMap member of the VkVideoEncodeQuantizationMapInfoKHR structure included in the pNext chain of pEncodeInfo must have been created with the VK_IMAGE_USAGE_VIDEO_ENCODE_EMPHASIS_MAP_BIT_KHR usage flag set", 326}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10318", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and pEncodeInfo->pSetupReferenceSlot is not NULL, then the pNext chain of pEncodeInfo->pSetupReferenceSlot must include a VkVideoEncodeAV1DpbSlotInfoKHR structure", 276}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10338", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and pEncodeInfo->pSetupReferenceSlot is not NULL, then the pExtensionHeader member of VkVideoEncodeAV1DpbSlotInfoKHR::pStdReferenceInfo for the VkVideoEncodeAV1DpbSlotInfoKHR structure included in the pNext chain of pEncodeInfo->pSetupReferenceSlot and the pExtensionHeader member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must both be NULL or not NULL", 567}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10339", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, pEncodeInfo->pSetupReferenceSlot is not NULL, and the pExtensionHeader member of VkVideoEncodeAV1DpbSlotInfoKHR::pStdReferenceInfo for the VkVideoEncodeAV1DpbSlotInfoKHR structure included in the pNext chain of pEncodeInfo->pSetupReferenceSlot is not NULL, then pExtensionHeader->temporal_id must equal pStdPictureInfo->pExtensionHeader->temporal_id in the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 555}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10340", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, pEncodeInfo->pSetupReferenceSlot is not NULL, and the pExtensionHeader member of VkVideoEncodeAV1DpbSlotInfoKHR::pStdReferenceInfo for the VkVideoEncodeAV1DpbSlotInfoKHR structure included in the pNext chain of pEncodeInfo->pSetupReferenceSlot is not NULL, then pExtensionHeader->spatial_id must equal pStdPictureInfo->pExtensionHeader->spatial_id in the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 553}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10811", 42}, { {"The image subresource referred to by pEncodeInfo->srcPictureResource must be in the VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR layout at the time the video encode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 312}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10812", 42}, { {"If pEncodeInfo->pSetupReferenceSlot is not NULL, then the image subresource referred to by pEncodeInfo->pSetupReferenceSlot->pPictureResource must be in the VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR layout at the time the video encode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 385}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10837", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the intra refresh mode the bound video session was created with must not be VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_NONE_KHR", 194}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10838", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then pEncodeInfo->activeReferencePictureCount must be less than or equal to VkVideoEncodeIntraRefreshCapabilitiesKHR::maxIntraRefreshActiveReferencePictures, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 345}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10839", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the pNext chain of pEncodeInfo must include a VkVideoEncodeIntraRefreshInfoKHR structure", 163}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10840", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the intraRefreshCycleDuration member of the VkVideoEncodeIntraRefreshInfoKHR structure included in the pNext chain of pEncodeInfo must not be zero", 221}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10841", 42}, { {"If pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the intraRefreshIndex member of the VkVideoEncodeIntraRefreshInfoKHR structure included in the pNext chain of pEncodeInfo must be less than VkVideoEncodeIntraRefreshInfoKHR::intraRefreshCycleDuration", 274}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10845", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and with an intra refresh mode other than VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, and VkVideoEncodeIntraRefreshCapabilitiesKHR::partitionIndependentIntraRefreshRegions is VK_FALSE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceEntryCount member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo must equal 1", 648}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10846", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and with the intra refresh mode VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, and pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the naluSliceEntryCount member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo must equal the intraRefreshCycleDuration member of the VkVideoEncodeIntraRefreshInfoKHR structure included in the pNext chain of pEncodeInfo", 552}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10848", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and with the intra refresh mode VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, and VkVideoEncodeIntraRefreshCapabilitiesKHR::nonRectangularIntraRefreshRegions is VK_FALSE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceEntryCount member of the VkVideoEncodeH264PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.height", 677}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10851", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and with an intra refresh mode other than VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, and VkVideoEncodeIntraRefreshCapabilitiesKHR::partitionIndependentIntraRefreshRegions is VK_FALSE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceSegmentEntryCount member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo must equal 1", 655}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10852", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and with the intra refresh mode VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, and pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, then the naluSliceSegmentEntryCount member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo must equal VkVideoEncodeIntraRefreshInfoKHR::intraRefreshCycleDuration", 489}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-10854", 42}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and with the intra refresh mode VK_VIDEO_ENCODE_INTRA_REFRESH_MODE_PER_PICTURE_PARTITION_BIT_KHR, pEncodeInfo->flags includes VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR, and VkVideoEncodeIntraRefreshCapabilitiesKHR::nonRectangularIntraRefreshRegions is VK_FALSE, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, then the naluSliceSegmentEntryCount member of the VkVideoEncodeH265PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be less than or equal to minCodingBlockExtent.height", 684}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pEncodeInfo-parameter", 46}, { {"pEncodeInfo must be a valid pointer to a valid VkVideoEncodeInfoKHR structure", 77}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pExtensionHeader-10336", 47}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pExtensionHeader member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pExtensionHeader->temporal_id must be less than VkVideoEncodeAV1CapabilitiesKHR::maxTemporalLayerCount, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 528}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pExtensionHeader-10337", 47}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pExtensionHeader member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pExtensionHeader->spatial_id must be less than VkVideoEncodeAV1CapabilitiesKHR::maxSpatialLayerCount, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 526}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pExtensionHeader-10341", 47}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pExtensionHeader member of VkVideoEncodeAV1DpbSlotInfoKHR::pStdReferenceInfo for any of the VkVideoEncodeAV1DpbSlotInfoKHR structures included in the pNext chain of any element of pEncodeInfo->pReferenceSlots is not NULL, then pExtensionHeader->temporal_id must be less than VkVideoEncodeAV1CapabilitiesKHR::maxTemporalLayerCount, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 570}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pExtensionHeader-10342", 47}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pExtensionHeader member of VkVideoEncodeAV1DpbSlotInfoKHR::pStdReferenceInfo for any of the VkVideoEncodeAV1DpbSlotInfoKHR structures included in the pNext chain of any element of pEncodeInfo->pReferenceSlots is not NULL, then pExtensionHeader->spatial_id must be less than VkVideoEncodeAV1CapabilitiesKHR::maxSpatialLayerCount, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 568}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08225", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain of pEncodeInfo must include a VkVideoEncodeH264PictureInfoKHR structure", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08229", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain of each element of pEncodeInfo->pReferenceSlots must include a VkVideoEncodeH264DpbSlotInfoKHR structure", 241}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08230", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain of pEncodeInfo must include a VkVideoEncodeH265PictureInfoKHR structure", 208}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08235", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain of each element of pEncodeInfo->pReferenceSlots must include a VkVideoEncodeH265DpbSlotInfoKHR structure", 241}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08339", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH264PictureInfoKHR structure, and VkVideoEncodeH264PictureInfoKHR::pStdPictureInfo->pRefLists is not NULL, then each element of the RefPicList0 and RefPicList1 array members of the StdVideoEncodeH264ReferenceListsInfo structure pointed to by VkVideoEncodeH264PictureInfoKHR::pStdPictureInfo->pRefLists must either be STD_VIDEO_H264_NO_REFERENCE_PICTURE or must equal the slotIndex member of one of the elements of pEncodeInfo->pReferenceSlots", 613}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08344", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH265PictureInfoKHR structure, and VkVideoEncodeH265PictureInfoKHR::pStdPictureInfo->pRefLists is not NULL, then each element of the RefPicList0 and RefPicList1 array members of the StdVideoEncodeH265ReferenceListsInfo structure pointed to by VkVideoEncodeH265PictureInfoKHR::pStdPictureInfo->pRefLists must either be STD_VIDEO_H265_NO_REFERENCE_PICTURE or must equal the slotIndex member of one of the elements of pEncodeInfo->pReferenceSlots", 613}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08352", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH264PictureInfoKHR structure, and pEncodeInfo->referenceSlotCount is greater than zero, then VkVideoEncodeH264PictureInfoKHR::pStdPictureInfo->pRefLists must not be NULL", 340}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08353", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH264PictureInfoKHR structure, and pEncodeInfo->referenceSlotCount is greater than zero, then the slotIndex member of each element of pEncodeInfo->pReferenceSlots must equal one of the elements of the RefPicList0 or RefPicList1 array members of the StdVideoEncodeH264ReferenceListsInfo structure pointed to by VkVideoEncodeH264PictureInfoKHR::pStdPictureInfo->pRefLists", 539}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08354", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH265PictureInfoKHR structure, and pEncodeInfo->referenceSlotCount is greater than zero, then VkVideoEncodeH265PictureInfoKHR::pStdPictureInfo->pRefLists must not be NULL", 340}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08355", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, the pNext chain of pEncodeInfo includes a VkVideoEncodeH265PictureInfoKHR structure, and pEncodeInfo->referenceSlotCount is greater than zero, then the slotIndex member of each element of pEncodeInfo->pReferenceSlots must equal one of the elements of the RefPicList0 or RefPicList1 array members of the StdVideoEncodeH265ReferenceListsInfo structure pointed to by VkVideoEncodeH265PictureInfoKHR::pStdPictureInfo->pRefLists", 539}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08360", 36}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the pNext chain of pEncodeInfo includes a VkVideoInlineQueryInfoKHR structure with its queryPool member specifying a valid VkQueryPool handle, then VkVideoInlineQueryInfoKHR::queryCount must equal opCount", 300}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-08361", 36}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the pNext chain of pEncodeInfo includes a VkVideoInlineQueryInfoKHR structure with its queryPool member specifying a valid VkQueryPool handle, then all the queries used by the command, as specified by the VkVideoInlineQueryInfoKHR structure, must be unavailable", 357}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10313", 36}, { {"If an instance of the VkVideoEncodeQuantizationMapInfoKHR structure is included in the pNext chain of pEncodeInfo, its quantizationMap member is not VK_NULL_HANDLE, commandBuffer is an unprotected command buffer, and protectedNoFault is not supported, then VkVideoEncodeQuantizationMapInfoKHR::quantizationMap must not have been created from a protected image", 359}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10314", 36}, { {"If an instance of the VkVideoEncodeQuantizationMapInfoKHR structure is included in the pNext chain of pEncodeInfo and its quantizationMap member is not VK_NULL_HANDLE, then the image subresource range referenced by quantizationMap must be in the VK_IMAGE_LAYOUT_VIDEO_ENCODE_QUANTIZATION_MAP_KHR layout at the time the video encode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 487}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10315", 36}, { {"If an instance of the VkVideoEncodeQuantizationMapInfoKHR structure is included in the pNext chain of pEncodeInfo and its quantizationMap member is not VK_NULL_HANDLE, and there is a bound video session parameters object, then the bound video session parameters object must have been created with VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR", 367}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10316", 36}, { {"If an instance of the VkVideoEncodeQuantizationMapInfoKHR structure is included in the pNext chain of pEncodeInfo, its quantizationMap member is not VK_NULL_HANDLE, and there is a bound video session parameters object created with VK_VIDEO_SESSION_PARAMETERS_CREATE_QUANTIZATION_MAP_COMPATIBLE_BIT_KHR, then quantizationMapExtent must equal ceil(pEncodeInfo->srcPictureResource.codedExtent / quantizationMapTexelSize)", 417}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10317", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain of pEncodeInfo must include a VkVideoEncodeAV1PictureInfoKHR structure", 206}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10319", 36}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain of each element of pEncodeInfo->pReferenceSlots must include a VkVideoEncodeAV1DpbSlotInfoKHR structure", 239}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10842", 36}, { {"If the pNext chain of any element of pEncodeInfo->pReferenceSlots includes a VkVideoReferenceIntraRefreshInfoKHR structure with a non-zero dirtyIntraRefreshRegions member, then pEncodeInfo->flags must include VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR", 246}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10843", 36}, { {"If the pNext chain of any element of pEncodeInfo->pReferenceSlots includes a VkVideoReferenceIntraRefreshInfoKHR structure with a non-zero dirtyIntraRefreshRegions member, then for each such element dirtyIntraRefreshRegions must equal intraRefreshCycleDuration minus intraRefreshIndex where intraRefreshCycleDuration and intraRefreshIndex are the members of the VkVideoEncodeIntraRefreshInfoKHR structure included in the pNext chain of pEncodeInfo", 447}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pNext-10844", 36}, { {"If the pNext chain of pEncodeInfo includes a VkVideoEncodeIntraRefreshInfoKHR structure, then its intraRefreshCycleDuration member must be zero or must be between 2 and VkVideoEncodeIntraRefreshCapabilitiesKHR::maxIntraRefreshCycleDuration, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 358}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pPictureResource-08219", 47}, { {"The pPictureResource member of each element of pEncodeInfo->pReferenceSlots must match one of the bound reference picture resource associated with the DPB slot index specified in the slotIndex member of that element", 215}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pPictureResource-08220", 47}, { {"Each video picture resource corresponding to the pPictureResource member specified in the elements of pEncodeInfo->pReferenceSlots must be unique within pEncodeInfo->pReferenceSlots", 181}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pPictureResource-10813", 47}, { {"The image subresource referred to by the pPictureResource member of each element of pEncodeInfo->pReferenceSlots must be in the VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR layout at the time the video encode operation is executed on the device , unless the unifiedImageLayoutsVideo feature is enabled, in which case it may be in the VK_IMAGE_LAYOUT_GENERAL layout", 356}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pStdPictureInfo-10327", 46}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and pStdPictureInfo->frame_type for the pStdPictureInfo member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is STD_VIDEO_AV1_FRAME_TYPE_KEY or STD_VIDEO_AV1_FRAME_TYPE_INTRA_ONLY, then VkVideoEncodeAV1PictureInfoKHR::predictionMode must be VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_INTRA_ONLY_KHR", 450}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pStdPictureInfo-10349", 46}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo->flags.segmentation_enabled for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be zero", 300}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pStdPictureInfo-10350", 46}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo->pSegmentation for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must be NULL", 287}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10343", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pTileInfo->TileCols must be greater than 0", 342}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10344", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pTileInfo->TileRows must be greater than 0", 342}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10345", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pTileInfo->TileCols must be less than or equal to VkVideoEncodeAV1CapabilitiesKHR::maxTiles.width, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 516}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10346", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then pTileInfo->TileRows must be less than or equal to VkVideoEncodeAV1CapabilitiesKHR::maxTiles.height, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 517}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10347", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then ceil(pEncodeInfo->srcPictureResource.codedExtent.width / pTileInfo->TileCols) must be between VkVideoEncodeAV1CapabilitiesKHR::minTileSize.width and VkVideoEncodeAV1CapabilitiesKHR::maxTileSize.width, inclusive, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 629}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-pTileInfo-10348", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the pTileInfo member of VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo for the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is not NULL, then ceil(pEncodeInfo->srcPictureResource.codedExtent.height / pTileInfo->TileRows) must be between VkVideoEncodeAV1CapabilitiesKHR::minTileSize.height and VkVideoEncodeAV1CapabilitiesKHR::maxTileSize.height, inclusive, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 632}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-predictionMode-10326", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_INTRA_ONLY_KHR, then VkVideoEncodeAV1PictureInfoKHR::pStdPictureInfo->frame_type must be STD_VIDEO_AV1_FRAME_TYPE_KEY or STD_VIDEO_AV1_FRAME_TYPE_INTRA_ONLY", 430}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-predictionMode-10329", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_SINGLE_REFERENCE_KHR, then there must be at least one non-negative element of VkVideoEncodeAV1PictureInfoKHR::referenceNameSlotIndices with element index i that does not equal cdfOnlyReferenceIndex and for which bit index i is set in VkVideoEncodeAV1CapabilitiesKHR::singleReferenceNameMask, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 683}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-predictionMode-10331", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_UNIDIRECTIONAL_COMPOUND_KHR, then there must be at least two non-negative elements of VkVideoEncodeAV1PictureInfoKHR::referenceNameSlotIndices with element indices i and j where (i,j) E {(0,1),(0,2),(0,3),(4,6)}, such that neither element equals cdfOnlyReferenceIndex and for which bit indices i and j are set in VkVideoEncodeAV1CapabilitiesKHR::unidirectionalCompoundReferenceNameMask, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 778}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-predictionMode-10333", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR and the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_BIDIRECTIONAL_COMPOUND_KHR, then there must be at least two non-negative elements of VkVideoEncodeAV1PictureInfoKHR::referenceNameSlotIndices with element indices i E {0,1,2,3} and j E {4,5,6}, respectively, such that neither element equals cdfOnlyReferenceIndex, and for which bit indices i and j are set in VkVideoEncodeAV1CapabilitiesKHR::bidirectionalCompoundReferenceNameMask, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with", 773}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-predictionMode-10857", 45}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, the predictionMode member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo is VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_UNIDIRECTIONAL_COMPOUND_KHR or VK_VIDEO_ENCODE_AV1_PREDICTION_MODE_BIDIRECTIONAL_COMPOUND_KHR, and VkVideoEncodeAV1CapabilitiesKHR::flags, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile the bound video session was created with, does not include VK_VIDEO_ENCODE_AV1_CAPABILITY_COMPOUND_PREDICTION_INTRA_REFRESH_BIT_KHR, then pEncodeInfo->flags must not include VK_VIDEO_ENCODE_INTRA_REFRESH_BIT_KHR", 698}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-queryPool-08363", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, then the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pEncodeInfo must have been created with a VkVideoProfileInfoKHR structure included in the pNext chain of VkQueryPoolCreateInfo identical to the one specified in VkVideoSessionCreateInfoKHR::pVideoProfile the bound video session was created with", 439}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-queryType-08362", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, then the queryType used to create the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pEncodeInfo must be VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR", 323}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-queryType-08364", 40}, { {"If the bound video session was created with VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR, and the queryType used to create the queryPool specified in the VkVideoInlineQueryInfoKHR structure included in the pNext chain of pEncodeInfo is VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then the VkCommandPool that commandBuffer was allocated from must have been created with a queue family index that supports result status queries, as indicated by VkQueueFamilyQueryResultStatusPropertiesKHR::queryResultStatusSupport", 508}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-referenceNameSlotIndices-10334", 55}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then each element of the referenceNameSlotIndices array member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo must either be negative or must equal the slotIndex member of one of the elements of pEncodeInfo->pReferenceSlots", 382}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-renderpass", 35}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-slotIndex-08217", 40}, { {"The slotIndex member of each element of pEncodeInfo->pReferenceSlots must be less than the VkVideoSessionCreateInfoKHR::maxDpbSlots specified when the bound video session was created", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-slotIndex-10335", 40}, { {"If the bound video session was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the slotIndex member of each element of pEncodeInfo->pReferenceSlots must equal one of the elements of the referenceNameSlotIndices array member of the VkVideoEncodeAV1PictureInfoKHR structure included in the pNext chain of pEncodeInfo", 355}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEncodeVideoKHR-videocoding", 36}, { {"This command must only be called inside of a video coding scope", 63}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-None-01985", 47}, { {"Conditional rendering must be active", 36}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-None-01986", 47}, { {"If conditional rendering was made active outside of a render pass instance, it must not be ended inside a render pass instance", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-None-01987", 47}, { {"If conditional rendering was made active within a subpass it must be ended in the same subpass", 94}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-suspended", 46}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndConditionalRenderingEXT-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdEndDebugUtilsLabelEXT-commandBuffer-01912", 51}, { {"There must be an outstanding vkCmdBeginDebugUtilsLabelEXT command prior to the vkCmdEndDebugUtilsLabelEXT on the queue that commandBuffer is submitted to", 153}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdEndDebugUtilsLabelEXT-commandBuffer-01913", 51}, { {"If commandBuffer is a secondary command buffer, there must be an outstanding vkCmdBeginDebugUtilsLabelEXT command recorded to commandBuffer that has not previously been ended by a call to vkCmdEndDebugUtilsLabelEXT", 214}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdEndDebugUtilsLabelEXT-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdEndDebugUtilsLabelEXT-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdEndDebugUtilsLabelEXT-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdEndGpaSampleAMD-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-commonparent", 38}, { {"Both of commandBuffer, and gpaSession must have been created, allocated, or retrieved from the same VkDevice", 108}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-gpaSession-parameter", 46}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSampleAMD-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-commonparent", 39}, { {"Both of commandBuffer, and gpaSession must have been created, allocated, or retrieved from the same VkDevice", 108}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-gpaSession-12411", 43}, { {"gpaSession must have previously begun using vkCmdBeginGpaSessionAMD", 67}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-gpaSession-parameter", 47}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndGpaSessionAMD-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-None-10666", 44}, { {"The per-tile execution model must have been enabled in the current render pass", 78}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-None-10667", 44}, { {"The current render pass must be a tile shading render pass", 58}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-pPerTileEndInfo-parameter", 59}, { {"pPerTileEndInfo must be a valid pointer to a valid VkPerTileEndInfoQCOM structure", 81}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-renderpass", 44}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndPerTileExecutionQCOM-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndQuery-None-01923", 29}, { {"All queries used by the command must be active", 46}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-None-07007", 29}, { {"If called within a subpass of a render pass instance, the corresponding vkCmdBeginQuery* command must have been called previously within the same subpass", 153}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-None-10682", 29}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-commandBuffer-01886", 38}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-commandBuffer-cmdpool", 40}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-commandBuffer-parameter", 42}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-commandBuffer-recording", 42}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-commonparent", 31}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-query-00810", 30}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-query-00812", 30}, { {"If vkCmdEndQuery is called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 201}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-queryPool-03227", 34}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one or more of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_COMMAND_BUFFER_KHR, the vkCmdEndQuery must be the last recorded command in commandBuffer", 259}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-queryPool-03228", 34}, { {"If queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR and one or more of the counters used to create queryPool was VK_PERFORMANCE_COUNTER_SCOPE_RENDER_PASS_KHR, the vkCmdEndQuery must not be recorded within a render pass instance", 256}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-queryPool-parameter", 38}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQuery-suspended", 28}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-None-02342", 39}, { {"All queries used by the command must be active", 46}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-None-07007", 39}, { {"If called within a subpass of a render pass instance, the corresponding vkCmdBeginQuery* command must have been called previously within the same subpass", 153}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-None-10682", 39}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-commandBuffer-02344", 48}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-commonparent", 41}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-query-02343", 40}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-query-02345", 40}, { {"If vkCmdEndQueryIndexedEXT is called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 211}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-queryPool-parameter", 48}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-queryType-06694", 44}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT or VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT, the index parameter must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 246}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-queryType-06695", 44}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT and not VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT, the index must be zero", 164}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-queryType-06696", 44}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT or VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT, index must equal the index used to begin the query", 183}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-suspended", 38}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndQueryIndexedEXT-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdEndRenderPass-None-00910", 34}, { {"The current subpass index must be equal to the number of subpasses in the render pass minus one", 95}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-None-02351", 34}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-None-06170", 34}, { {"The current render pass instance must not have been begun with vkCmdBeginRendering", 82}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-None-07004", 34}, { {"If vkCmdBeginQuery* was called within a subpass of the render pass, the corresponding vkCmdEndQuery* must have been called subsequently within the same subpass", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-None-10653", 34}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-bufferlevel", 35}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-renderpass", 34}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-suspended", 33}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-None-02352", 35}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-None-03103", 35}, { {"The current subpass index must be equal to the number of subpasses in the render pass minus one", 95}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-None-06171", 35}, { {"The current render pass instance must not have been begun with vkCmdBeginRendering", 82}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-None-07005", 35}, { {"If vkCmdBeginQuery* was called within a subpass of the render pass, the corresponding vkCmdEndQuery* must have been called subsequently within the same subpass", 159}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-bufferlevel", 36}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-pSubpassEndInfo-parameter", 50}, { {"pSubpassEndInfo must be a valid pointer to a valid VkSubpassEndInfo structure", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-renderpass", 35}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRenderPass2-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-None-06161", 33}, { {"The current render pass instance must have been begun with vkCmdBeginRendering", 78}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-None-06781", 33}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-None-06999", 33}, { {"If vkCmdBeginQuery* was called within the render pass, the corresponding vkCmdEndQuery* must have been called subsequently within the same subpass", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-None-10645", 33}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-commandBuffer-06162", 42}, { {"The current render pass instance must have been begun in commandBuffer", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-renderpass", 33}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-None-10610", 37}, { {"The current render pass instance must have been begun with vkCmdBeginRendering", 78}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-None-10612", 37}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-None-10613", 37}, { {"If vkCmdBeginQuery* was called within the render pass, the corresponding vkCmdEndQuery* must have been called subsequently within the same subpass", 146}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-commandBuffer-10611", 46}, { {"The current render pass instance must have been begun in commandBuffer", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-pRenderingEndInfo-parameter", 54}, { {"If pRenderingEndInfo is not NULL, pRenderingEndInfo must be a valid pointer to a valid VkRenderingEndInfoKHR structure", 118}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-renderpass", 37}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndRendering2KHR-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-commandBuffer-12377", 57}, { {"Shader instrumentation must be active within commandBuffer", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-commandBuffer-12378", 57}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-suspended", 47}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndShaderInstrumentationARM-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-None-02375", 45}, { {"Transform feedback must be active", 33}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-None-10657", 45}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-firstCounterBuffer-02376", 59}, { {"firstCounterBuffer must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-firstCounterBuffer-02377", 59}, { {"The sum of firstCounterBuffer and counterRangeCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 157}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-pCounterInfos-13095", 54}, { {"Each element of pCounterInfos must have a addressRange.size that is either 0 or greater than or equal to 4", 106}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-pCounterInfos-13096", 54}, { {"For each element of pCounterInfos with a non-zero address, the buffer the address was queried from must have been created with the VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT usage flag set", 203}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-pCounterInfos-parameter", 58}, { {"If counterRangeCount is not 0, and pCounterInfos is not NULL, pCounterInfos must be a valid pointer to an array of counterRangeCount valid VkBindTransformFeedbackBuffer2InfoEXT structures", 187}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-renderpass", 45}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-transformFeedback-02374", 58}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedback2EXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-None-02375", 44}, { {"Transform feedback must be active", 33}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-None-10657", 44}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-commonparent", 46}, { {"Both of commandBuffer, and the elements of pCounterBuffers that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 178}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-counterBufferCount-02608", 58}, { {"If counterBufferCount is not 0, and pCounterBuffers is not NULL, pCounterBuffers must be a valid pointer to an array of counterBufferCount VkBuffer handles that are either valid or VK_NULL_HANDLE", 195}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-firstCounterBuffer-02376", 58}, { {"firstCounterBuffer must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-firstCounterBuffer-02377", 58}, { {"The sum of firstCounterBuffer and counterBufferCount must be less than or equal to VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackBuffers", 158}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-pCounterBuffer-02379", 54}, { {"If pCounterBuffer is NULL, then pCounterBufferOffsets must also be NULL", 71}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-pCounterBufferOffsets-02378", 61}, { {"For each buffer handle in the array, if it is not VK_NULL_HANDLE it must reference a buffer large enough to hold 4 bytes at the corresponding offset from the pCounterBufferOffsets array", 185}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-pCounterBufferOffsets-parameter", 65}, { {"If counterBufferCount is not 0, and pCounterBufferOffsets is not NULL, pCounterBufferOffsets must be a valid pointer to an array of counterBufferCount VkDeviceSize values", 170}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-pCounterBuffers-02380", 55}, { {"For each buffer handle in the pCounterBuffers array that is not VK_NULL_HANDLE it must have been created with the VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT usage flag set", 186}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-renderpass", 44}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-transformFeedback-02374", 57}, { {"VkPhysicalDeviceTransformFeedbackFeaturesEXT::transformFeedback must be enabled", 79}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndTransformFeedbackEXT-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdEndVideoCodingKHR-None-07251", 38}, { {"There must be no active queries", 31}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-bufferlevel", 39}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 143}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-pEndCodingInfo-parameter", 52}, { {"pEndCodingInfo must be a valid pointer to a valid VkVideoEndCodingInfoKHR structure", 83}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-renderpass", 38}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-suspended", 37}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdEndVideoCodingKHR-videocoding", 39}, { {"This command must only be called inside of a video coding scope", 63}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCmdExecuteCommands-None-02286", 36}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-None-09681", 36}, { {"If this command is called within a render pass instance begun with vkCmdBeginRenderPass, and vkCmdNextSubpass has been called in the current render pass instance, the contents parameter of the last call to vkCmdNextSubpass must have been VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS , or VK_SUBPASS_CONTENTS_INLINE_AND_SECONDARY_COMMAND_BUFFERS_KHR", 349}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-colorAttachmentCount-06027", 52}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the colorAttachmentCount member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the VkRenderingInfo::colorAttachmentCount parameter to vkCmdBeginRendering", 398}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-colorAttachmentCount-11532", 52}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve, the colorAttachmentCount member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the VkRenderingInfo::colorAttachmentCount parameter to vkCmdBeginRendering", 418}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-00101", 45}, { {"If the inheritedQueries feature is not enabled, commandBuffer must not have any queries active", 94}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-00102", 45}, { {"If commandBuffer has a VK_QUERY_TYPE_OCCLUSION query active, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceInfo::occlusionQueryEnable set to VK_TRUE", 194}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-00103", 45}, { {"If commandBuffer has a VK_QUERY_TYPE_OCCLUSION query active, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceInfo::queryFlags having all bits set that are set for the query", 216}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-00104", 45}, { {"If commandBuffer has a VK_QUERY_TYPE_PIPELINE_STATISTICS query active, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceInfo::pipelineStatistics having all bits set that are set in the VkQueryPool the query uses", 254}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-01820", 45}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, each element of pCommandBuffers must be a protected command buffer", 152}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-01821", 45}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, each element of pCommandBuffers must be an unprotected command buffer", 158}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-06533", 45}, { {"If this command is called within a render pass instance and any recorded command in commandBuffer in the current subpass will write to an image subresource as an attachment, commands recorded in elements of pCommandBuffers must not read from the memory backing that image subresource in any other way", 300}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-06534", 45}, { {"If this command is called within a render pass instance and any recorded command in commandBuffer in the current subpass will read from an image subresource used as an attachment in any way other than as an attachment, commands recorded in elements of pCommandBuffers must not write to that image subresource as an attachment", 325}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-07594", 45}, { {"commandBuffer must not have any queries other than VK_QUERY_TYPE_OCCLUSION and VK_QUERY_TYPE_PIPELINE_STATISTICS active", 119}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-09375", 45}, { {"commandBuffer must not be a secondary command buffer unless the nestedCommandBuffer feature is enabled", 102}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-11351", 45}, { {"If there is a sampler descriptor heap bound to commandBuffer, each element of pCommandBuffers must have been recorded with a value of VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo that is either NULL or a pointer to a bind info that is identical to that set via the last call to vkCmdBindSamplerHeapEXT", 326}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-11352", 45}, { {"If there is a resource descriptor heap bound to commandBuffer, each element of pCommandBuffers must have been recorded with a value of VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo that is either NULL or a pointer to a bind info that is identical to that set via the last call to vkCmdBindResourceHeapEXT", 329}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-11473", 45}, { {"If there is no sampler descriptor heap bound to commandBuffer, each element of pCommandBuffers must have been recorded with a value of VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo set to NULL", 216}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-11474", 45}, { {"If there is no resource descriptor heap bound to commandBuffer, each element of pCommandBuffers must have been recorded with a value of VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo set to NULL", 218}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-12373", 45}, { {"commandBuffer must not have any shader instrumentation active", 61}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commandBufferCount-arraylength", 56}, { {"commandBufferCount must be greater than 0", 41}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-commonparent", 38}, { {"Both of commandBuffer, and the elements of pCommandBuffers must have been created, allocated, or retrieved from the same VkDevice", 129}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-contents-09680", 40}, { {"If this command is called within a render pass instance begun with vkCmdBeginRenderPass, and vkCmdNextSubpass has not been called in the current render pass instance, the contents parameter of vkCmdBeginRenderPass must have been VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS , or VK_SUBPASS_CONTENTS_INLINE_AND_SECONDARY_COMMAND_BUFFERS_EXT", 340}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-flags-06024", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, its VkRenderingInfo::flags parameter must have included VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT", 195}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-flags-06026", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the flags member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the VkRenderingInfo::flags parameter to vkCmdBeginRendering, excluding VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT", 431}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-imageView-06028", 41}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, if the imageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the corresponding element of the pColorAttachmentFormats member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 533}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-imageView-07606", 41}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, if the imageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is VK_NULL_HANDLE, the corresponding element of the pColorAttachmentFormats member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 498}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-memory-10724", 38}, { {"If this command is called within a render pass instance, the size of memory member of the VkTileMemoryBindInfoQCOM structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the active bound bound tile memory object in commandBuffer", 330}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-nestedCommandBuffer-09376", 51}, { {"If the nestedCommandBuffer feature is enabled, and commandBuffer is a secondary command buffer, the command buffer nesting level of each element of pCommandBuffers must be less than maxCommandBufferNestingLevel", 210}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-nestedCommandBufferRendering-09377", 60}, { {"If the nestedCommandBufferRendering feature is not enabled, and commandBuffer is a secondary command buffer, commandBuffer must not have been recorded with VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT", 204}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-nestedCommandBufferSimultaneousUse-09378", 66}, { {"If the nestedCommandBufferSimultaneousUse feature is not enabled, and commandBuffer is a secondary command buffer, each element of pCommandBuffers must not have been recorded with VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT", 224}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pBeginInfo-06020", 42}, { {"If this command is called within a render pass instance begun with vkCmdBeginRenderPass, the render passes specified in the pBeginInfo->pInheritanceInfo->renderPass members of the vkBeginCommandBuffer commands used to begin recording each element of pCommandBuffers must be compatible with the current render pass", 313}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pBeginInfo-06025", 42}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the render passes specified in the pBeginInfo->pInheritanceInfo->renderPass members of the vkBeginCommandBuffer commands used to begin recording each element of pCommandBuffers must be VK_NULL_HANDLE", 287}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00088", 47}, { {"Each element of pCommandBuffers must have been allocated with a level of VK_COMMAND_BUFFER_LEVEL_SECONDARY", 106}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00089", 47}, { {"Each element of pCommandBuffers must be in the pending or executable state", 74}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00091", 47}, { {"If any element of pCommandBuffers was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT flag, it must not be in the pending state", 146}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00092", 47}, { {"If any element of pCommandBuffers was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT flag, it must not have already been recorded to commandBuffer", 166}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00093", 47}, { {"If any element of pCommandBuffers was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT flag, it must not appear more than once in pCommandBuffers", 163}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00094", 47}, { {"Each element of pCommandBuffers must have been allocated from a VkCommandPool that was created for the same queue family as the VkCommandPool from which commandBuffer was allocated", 180}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00096", 47}, { {"If this command is called within a render pass instance, each element of pCommandBuffers must have been recorded with the VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT", 170}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00099", 47}, { {"If this command is called within a render pass instance, and any element of pCommandBuffers was recorded with VkCommandBufferInheritanceInfo::framebuffer not equal to VK_NULL_HANDLE, that VkFramebuffer must match the VkFramebuffer used in the current render pass instance", 271}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00100", 47}, { {"If vkCmdExecuteCommands is not being called within a render pass instance, each element of pCommandBuffers must not have been recorded with the VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT", 192}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-00105", 47}, { {"Each element of pCommandBuffers must not begin any query types that are active in commandBuffer", 95}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-06019", 47}, { {"If this command is called within a render pass instance begun with vkCmdBeginRenderPass, each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceInfo::subpass set to the index of the subpass which the given command buffer will be executed in", 272}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-06021", 47}, { {"If pCommandBuffers contains any suspended render pass instances, there must be no action or synchronization commands between that render pass instance and any render pass instance that resumes it", 195}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-06022", 47}, { {"If pCommandBuffers contains any suspended render pass instances, there must be no render pass instances between that render pass instance and any render pass instance that resumes it", 182}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-06535", 47}, { {"If this command is called within a render pass instance and any recorded command in a given element of pCommandBuffers will write to an image subresource as an attachment, commands recorded in elements of pCommandBuffers at a higher index must not read from the memory backing that image subresource in any other way", 316}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-06536", 47}, { {"If this command is called within a render pass instance and any recorded command in a given element of pCommandBuffers will read from an image subresource used as an attachment in any way other than as an attachment, commands recorded in elements of pCommandBuffers at a higher index must not write to that image subresource as an attachment", 341}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-09504", 47}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the color attachment mapping state specified by VkRenderingAttachmentLocationInfo in the inheritance info of each element of pCommandBuffers and in the current state of commandBuffer must match", 281}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-09505", 47}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the input attachment mapping state specified by VkRenderingInputAttachmentIndexInfo in the inheritance info of each element of pCommandBuffers and in the current state of commandBuffer must match", 283}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-10620", 47}, { {"If this command is being recorded within a render pass instance with tile shading enabled, all elements of pCommandBuffers must have been recorded with VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM included in VkRenderPassTileShadingCreateInfoQCOM::flags", 252}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-10621", 47}, { {"If the per-tile execution model is enabled, all elements of pCommandBuffers must have been recorded with VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM included in VkRenderPassTileShadingCreateInfoQCOM::flags", 217}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-10623", 47}, { {"If any element of pCommandBuffers was recorded with VK_TILE_SHADING_RENDER_PASS_ENABLE_BIT_QCOM included in VkRenderPassTileShadingCreateInfoQCOM::flags, this command must be recorded in a render pass that has tile shading enabled", 230}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-10624", 47}, { {"If any element of pCommandBuffers was recorded with VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM included in VkRenderPassTileShadingCreateInfoQCOM::flags, per-tile execution model must be enabled", 206}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-11500", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, then each element of pCommandBuffers must have been recorded with a VkCustomResolveCreateInfoEXT struct chained to its VkCommandBufferInheritanceInfo", 251}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-11501", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, then each element of pCommandBuffers must not have been recorded with a VkCustomResolveCreateInfoEXT struct chained to its VkCommandBufferInheritanceInfo", 263}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-11502", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then each element of pCommandBuffers must have been recorded with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 269}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-11503", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, then each element of pCommandBuffers must have been recorded with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 305}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pCommandBuffers-parameter", 51}, { {"pCommandBuffers must be a valid pointer to an array of commandBufferCount valid VkCommandBuffer handles", 103}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pDepthAttachment-06029", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, if the VkRenderingInfo::pDepthAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the depthAttachmentFormat member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 487}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pDepthAttachment-06774", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the VkRenderingInfo::pDepthAttachment->imageView parameter to vkCmdBeginRendering was VK_NULL_HANDLE, the value of the depthAttachmentFormat member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 453}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pDepthAttachment-11527", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and the VkRenderingInfo::pDepthAttachment->resolveImageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that image view", 417}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pDepthAttachment-11535", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, if the VkRenderingInfo::pDepthAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the depthAttachmentFormat member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 487}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pDepthAttachment-11537", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve and the VkRenderingInfo::pDepthAttachment->resolveImageView parameter to vkCmdBeginRendering was VK_NULL_HANDLE, the value of the depthAttachmentFormat member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 480}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-02865", 37}, { {"If this command is called within a render pass instance that included VkRenderPassTransformBeginInfoQCOM in the pNext chain of VkRenderPassBeginInfo, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceRenderPassTransformInfoQCOM in the pNext chain of VkCommandBufferBeginInfo", 316}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-02866", 37}, { {"If this command is called within a render pass instance that included VkRenderPassTransformBeginInfoQCOM in the pNext chain of VkRenderPassBeginInfo, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceRenderPassTransformInfoQCOM::transform identical to VkRenderPassTransformBeginInfoQCOM::transform", 339}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-02867", 37}, { {"If this command is called within a render pass instance that included VkRenderPassTransformBeginInfoQCOM in the pNext chain of VkRenderPassBeginInfo, then each element of pCommandBuffers must have been recorded with VkCommandBufferInheritanceRenderPassTransformInfoQCOM::renderArea identical to VkRenderPassBeginInfo::renderArea", 328}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06032", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the pNext chain of VkCommandBufferInheritanceInfo includes a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, if the imageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the corresponding element of the pColorAttachmentSamples member of the VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the sample count used to create that image view", 702}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06033", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the pNext chain of VkCommandBufferInheritanceInfo includes a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, if the VkRenderingInfo::pDepthAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of the VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the sample count used to create that image view", 664}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06034", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the pNext chain of VkCommandBufferInheritanceInfo includes a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, if the VkRenderingInfo::pStencilAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of the VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the sample count used to create that image view", 666}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06035", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, and the pNext chain of VkCommandBufferInheritanceInfo does not include a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and if the imageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that image view", 597}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06036", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the pNext chain of VkCommandBufferInheritanceInfo does not include a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and if the VkRenderingInfo::pDepthAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that image view", 568}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-06037", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the pNext chain of VkCommandBufferInheritanceInfo does not include a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and if the VkRenderingInfo::pStencilAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that image view", 570}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-09299", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, with any color attachment using a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, the pNext chain of VkCommandBufferInheritanceInfo used to create each element of pCommandBuffers must include a VkExternalFormatANDROID structure with an externalFormat matching that used to create the resolve attachment in the render pass", 433}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pNext-09300", 37}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering with any color attachment using a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and the pNext chain of VkCommandBufferInheritanceInfo does not include a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be VK_SAMPLE_COUNT_1_BIT", 445}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pStencilAttachment-06030", 50}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, if the VkRenderingInfo::pStencilAttachment->imageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the stencilAttachmentFormat member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 491}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pStencilAttachment-06775", 50}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering and the VkRenderingInfo::pStencilAttachment->imageView parameter to vkCmdBeginRendering was VK_NULL_HANDLE, the value of the stencilAttachmentFormat member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 457}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pStencilAttachment-11528", 50}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and the VkRenderingInfo::pStencilAttachment->resolveImageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that image view", 419}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pStencilAttachment-11536", 50}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve, if the VkRenderingInfo::pStencilAttachment->resolveImageView parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of the stencilAttachmentFormat member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 518}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-pStencilAttachment-11538", 50}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve and the VkRenderingInfo::pStencilAttachment->resolveImageView parameter to vkCmdBeginRendering was VK_NULL_HANDLE, the value of the stencilAttachmentFormat member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 484}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-resolveImageView-11526", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and the resolveImageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the value of VkCommandBufferInheritanceRenderingInfo::rasterizationSamples must be equal to the sample count used to create that resolveImageView", 451}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-resolveImageView-11533", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve, if the resolveImageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is not VK_NULL_HANDLE, the corresponding element of the pColorAttachmentFormats member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the format used to create that image view", 560}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-resolveImageView-11534", 48}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering that contains a custom resolve, if the resolveImageView member of an element of the VkRenderingInfo::pColorAttachments parameter to vkCmdBeginRendering is VK_NULL_HANDLE, the corresponding element of the pColorAttachmentFormats member of the VkCustomResolveCreateInfoEXT structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be VK_FORMAT_UNDEFINED", 525}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-tileApronSize-10622", 45}, { {"If this command is being recorded within a render pass instance, the tileApronSize used to create the render pass instance must equal the VkRenderPassTileShadingCreateInfoQCOM::tileApronSize used to record all elements of pCommandBuffers", 237}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-tileApronSize-10625", 45}, { {"If this command is not being recorded into a render pass instance, the VkRenderPassTileShadingCreateInfoQCOM::tileApronSize that was recorded into all elements of pCommandBuffers must equal (0,0)", 195}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-variableSampleLocations-06023", 55}, { {"If the variableSampleLocations limit is not supported, and any element of pCommandBuffers contains any suspended render pass instances, where a graphics pipeline has been bound, any pipelines bound in the render pass instance that resumes it, or any subsequent render pass instances that resume from that one and so on, must use the same sample locations", 354}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteCommands-viewMask-06031", 40}, { {"If this command is called within a render pass instance begun with vkCmdBeginRendering, the viewMask member of the VkCommandBufferInheritanceRenderingInfo structure included in the pNext chain of VkCommandBufferBeginInfo::pInheritanceInfo used to begin recording each element of pCommandBuffers must be equal to the VkRenderingInfo::viewMask parameter to vkCmdBeginRendering", 374}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-InputAttachmentIndex-09597", 64}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-02691", 48}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-02692", 48}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-02693", 48}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-04876", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-04877", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-06479", 48}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-06537", 48}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-06666", 48}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-06886", 48}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-06887", 48}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07288", 48}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07469", 48}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07620", 48}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07621", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07622", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07623", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07624", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07625", 48}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07626", 48}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07627", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07629", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07630", 48}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07631", 48}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07632", 48}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07633", 48}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07634", 48}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07636", 48}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07639", 48}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07640", 48}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07641", 48}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07642", 48}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07643", 48}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07644", 48}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07645", 48}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07646", 48}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07647", 48}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07648", 48}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07649", 48}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07749", 48}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07751", 48}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07831", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07832", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07834", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07835", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07836", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07837", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07838", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07839", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07840", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07841", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07843", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07844", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07845", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07846", 48}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07847", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07848", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07849", 48}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07850", 48}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07878", 48}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07879", 48}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07880", 48}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07881", 48}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-07888", 48}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08114", 48}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08115", 48}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08116", 48}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08117", 48}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08119", 48}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08600", 48}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08601", 48}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08605", 48}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08606", 48}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08608", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08609", 48}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08610", 48}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08611", 48}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08612", 48}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08613", 48}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08617", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08636", 48}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08637", 48}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08644", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08666", 48}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08669", 48}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08684", 48}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08685", 48}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08686", 48}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08687", 48}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08688", 48}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08689", 48}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08690", 48}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08693", 48}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08696", 48}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08698", 48}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08699", 48}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08876", 48}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08877", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08878", 48}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-08879", 48}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09003", 48}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09116", 48}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09211", 48}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09363", 48}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09364", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09365", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09368", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09369", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09548", 48}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09549", 48}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09600", 48}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09642", 48}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09643", 48}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-09650", 48}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10068", 48}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10608", 48}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10677", 48}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10678", 48}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10679", 48}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10772", 48}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10795", 48}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10796", 48}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10797", 48}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10862", 48}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10864", 48}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10927", 48}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-10928", 48}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11062", 48}, { {"If a rendering pass is currently active, either the view mask must be 0 or indirectExecutionSet must be VK_NULL_HANDLE", 118}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11297", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11298", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11299", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11300", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11301", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11302", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11304", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11305", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11306", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11308", 48}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11309", 48}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11372", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11373", 48}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11374", 48}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11376", 48}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11397", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11398", 48}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11437", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11438", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11439", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11441", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11442", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11485", 48}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11522", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11523", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11861", 48}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11862", 48}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11863", 48}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11865", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11866", 48}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11868", 48}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-11869", 48}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-12338", 48}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-12339", 48}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-12340", 48}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-None-12472", 48}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchSADQCOM-06976", 68}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchSADQCOM-12420", 68}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchSSDQCOM-06974", 68}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchWindow-09215", 67}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchWindow-09216", 67}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchWindow-09217", 67}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBlockMatchWindow-12421", 67}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageBoxFilterQCOM-06973", 64}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageSampleWeightedQCOM-06971", 69}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageSampleWeightedQCOM-06972", 69}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageSampleWeightedQCOM-06977", 69}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageSampleWeightedQCOM-06978", 69}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageWrite-04469", 56}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageWrite-08795", 56}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpImageWrite-08796", 56}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTensorReadARM-12462", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeImage-07027", 55}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeImage-07028", 55}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeImage-07029", 55}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeImage-07030", 55}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeImage-07468", 55}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-OpTypeTensorARM-09906", 59}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-SampledType-04470", 55}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-SampledType-04471", 55}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-SampledType-04472", 55}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-SampledType-04473", 55}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-VkPipelineVieportCreateInfo-04141", 71}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-VkPipelineVieportCreateInfo-04142", 71}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-addressModeU-12422", 56}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-advancedBlendMaxColorAttachments-07480", 76}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-alphaToCoverageEnable-08919", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-alphaToCoverageEnable-08920", 65}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-attachmentCount-07750", 59}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-blendEnable-04727", 55}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-bufferlevel", 49}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-colorAttachmentCount-06179", 64}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-colorAttachmentCount-06185", 64}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-colorAttachmentCount-09362", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-colorAttachmentCount-09372", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-colorAttachmentCount-11871", 64}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-02707", 57}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-10746", 57}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-11045", 57}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-11143", 57}, { {"commandBuffer must not have been recorded with VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT", 91}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-conservativePointAndLineRasterization-07499", 81}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-coverageModulationTableEnable-07488", 73}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-coverageReductionMode-07491", 65}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-coverageToColorEnable-07490", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-cubicRangeClamp-09212", 59}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-customResolve-11524", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-customResolve-11525", 57}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-customResolve-11529", 57}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-customResolve-11530", 57}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-deviceGeneratedCommands-11059", 67}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-deviceGeneratedCommandsTransformFeedback-11057", 84}, { {"If deviceGeneratedCommandsTransformFeedback is not supported on device, transform feedback must not be active", 109}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dimensionCount-09905", 58}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingLocalRead-11797", 69}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08910", 77}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08911", 77}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08912", 77}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08913", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08914", 77}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08915", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08916", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08917", 77}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-08918", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-11864", 77}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-11867", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-dynamicRenderingUnusedAttachments-11870", 77}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-filterCubic-02694", 55}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-filterCubicMinmax-02695", 61}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-firstAttachment-07476", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-firstAttachment-07478", 59}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-02696", 49}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-10582", 49}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-11521", 49}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-13361", 49}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-13362", 49}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-13363", 49}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-13364", 49}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-flags-13365", 49}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-format-07753", 50}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageLayout-00344", 55}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06172", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06173", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06174", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06175", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06176", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06177", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06183", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-imageView-06184", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-index-11450", 49}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-10769", 66}, { {"If the indirectCommandsLayout member of pGeneratedCommandsInfo contains a draw token, this command must not be called outside a render pass instance", 148}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-11004", 66}, { {"If indirectCommandsLayout was created with a token sequence that contained the VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token and indirectExecutionSet was created using VK_INDIRECT_EXECUTION_SET_INFO_TYPE_SHADER_OBJECTS_EXT, every executed VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token must bind all the shader stages set in the VkIndirectCommandsExecutionSetTokenEXT::shaderStages used to create indirectCommandsLayout", 438}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-11053", 66}, { {"If the token sequence of the passed VkGeneratedCommandsInfoEXT::indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the initial shader state of VkGeneratedCommandsInfoEXT::indirectExecutionSet must be bound on commandBuffer", 262}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-11056", 66}, { {"If the indirectCommandsLayout member of pGeneratedCommandsInfo contains a draw token, then the active render pass must not have a specified fragment density map", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-11141", 66}, { {"If the indirectCommandsLayout member of pGeneratedCommandsInfo was created with the VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_EXT bit set, then isPreprocessed must be VK_TRUE", 190}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectCommandsLayout-12202", 66}, { {"If the indirectCommandsLayout member of pGeneratedCommandsInfo does not contain a draw token, this command must not be called inside a render pass instance", 155}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-indirectExecutionSet-11058", 64}, { {"If transform feedback is active, VkGeneratedCommandsInfoEXT::indirectExecutionSet must be VK_NULL_HANDLE", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-10198", 58}, { {"If isPreprocessed is VK_TRUE then the conditional render state and its predicate value must match identically with the state and value set during execution of the corresponding call to vkCmdPreprocessGeneratedCommandsEXT", 220}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11046", 58}, { {"If isPreprocessed is VK_TRUE and vkGetGeneratedCommandsMemoryRequirementsEXT did not return a required size of zero then vkCmdPreprocessGeneratedCommandsEXT must have already been executed on the device before this command executes, and the preprocessing command must have used the same pGeneratedCommandsInfo content as well as the content of the input buffers it references (all except VkGeneratedCommandsInfoEXT::preprocessAddress)", 434}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11047", 58}, { {"If isPreprocessed is VK_TRUE then the indirectCommandsLayout member of pGeneratedCommandsInfo must have been created with the VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_EXT bit set", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11048", 58}, { {"If isPreprocessed is VK_TRUE then the bound descriptor sets and push constants must match identically with those bound during recording of the corresponding call to vkCmdPreprocessGeneratedCommandsEXT", 200}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11049", 58}, { {"If isPreprocessed is VK_TRUE and the indirectCommandsLayout member of pGeneratedCommandsInfo contains a draw token, then the graphics state bound on commandBuffer must match identically with the graphics state bound on the stateCommandBuffer passed to vkCmdPreprocessGeneratedCommandsEXT", 287}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11051", 58}, { {"If isPreprocessed is VK_TRUE and the indirectCommandsLayout member of pGeneratedCommandsInfo contains a dispatch token, then the compute state bound on commandBuffer must match identically with the compute state bound on the stateCommandBuffer passed to vkCmdPreprocessGeneratedCommandsEXT", 289}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11052", 58}, { {"If isPreprocessed is VK_TRUE and the indirectCommandsLayout member of pGeneratedCommandsInfo contains a ray tracing token, then the ray tracing state bound on commandBuffer must match identically with the ray tracing state bound on the stateCommandBuffer passed to vkCmdPreprocessGeneratedCommandsEXT", 300}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11055", 58}, { {"If isPreprocessed is VK_TRUE and the token sequence of the passed VkGeneratedCommandsInfoEXT::indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the members of VkGeneratedCommandsInfoEXT::indirectExecutionSet accessed by this command must not have been modified since the preprocess buffer was generated", 343}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-isPreprocessed-11149", 58}, { {"If isPreprocessed is VK_TRUE, then the queue family index of commandBuffer must be the same as the queue family index used to allocate the stateCommandBuffer passed to vkCmdPreprocessGeneratedCommandsEXT", 203}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-layers-10831", 50}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-logicOp-04878", 51}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-magFilter-04553", 53}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-magFilter-09598", 53}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-maxFragmentDualSrcAttachments-09239", 73}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-maxMultiviewInstanceIndex-02688", 69}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-micromap-11636", 52}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-micromap-11637", 52}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-micromap-11638", 52}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-micromap-11639", 52}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-mipmapMode-04770", 54}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-mipmapMode-09599", 54}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-multisampledRenderToSingleSampled-07285", 77}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-multisampledRenderToSingleSampled-07286", 77}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-multisampledRenderToSingleSampled-07287", 77}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-multiviewPerViewViewports-12262", 69}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-multiviewPerViewViewports-12263", 69}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-nextStage-10745", 53}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pBindInfo-11375", 53}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pColorAttachments-08963", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pColorAttachments-11539", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthAttachment-06186", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthAttachment-08964", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthAttachment-11540", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthAttachment-11872", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthInputAttachmentIndex-09595", 70}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDepthInputAttachmentIndex-09596", 70}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDescription-09900", 56}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDynamicStates-08715", 58}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pDynamicStates-08716", 58}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pGeneratedCommandsInfo-parameter", 70}, { {"pGeneratedCommandsInfo must be a valid pointer to a valid VkGeneratedCommandsInfoEXT structure", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pNext-07935", 49}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pNext-12463", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pStencilAttachment-06187", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pStencilAttachment-08965", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pStencilAttachment-11860", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pStencilAttachment-11873", 62}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-pipelineFragmentShadingRate-09238", 71}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-preprocessAddress-11142", 61}, { {"The contents of the preprocessAddress member of pGeneratedCommandsInfo must not have been previously used to record another vkCmdExecuteGeneratedCommandsEXT", 156}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-primitiveFragmentShadingRateWithMultipleViewports-04552", 93}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-primitiveFragmentShadingRateWithMultipleViewports-08642", 93}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-primitivesGeneratedQueryWithNonZeroStreams-06709", 86}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-primitivesGeneratedQueryWithNonZeroStreams-07481", 86}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-primitivesGeneratedQueryWithRasterizerDiscard-06708", 89}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-protectedNoFault-11455", 60}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-protectedNoFault-11456", 60}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizationSamples-07471", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizationSamples-07489", 64}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizationSamples-07494", 64}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizerDiscardEnable-09236", 67}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizerDiscardEnable-09420", 67}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-rasterizerDiscardEnable-10863", 67}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-reductionMode-09213", 57}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-renderPass-02684", 54}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-renderPass-06198", 54}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-02689", 65}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07484", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07485", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07486", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07487", 65}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07936", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07937", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsEnable-07938", 65}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsPerPixel-07482", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sampleLocationsPerPixel-07483", 67}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-samples-07472", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-samples-07473", 51}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-scissorCount-03418", 56}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-selectableCubicWeights-09214", 66}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-shadingRateImage-09233", 60}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-shadingRateImage-09234", 60}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sparseImageInt64Atomics-04474", 67}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-sparseImageInt64Atomics-04475", 67}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-stage-07073", 49}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-stippledLineEnable-07495", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-stippledLineEnable-07496", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-stippledLineEnable-07497", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-stippledLineEnable-07498", 62}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-storageBuffers-06936", 58}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-subpass-02685", 51}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-supportedIndirectCommandsShaderStages-11060", 81}, { {"The bound shader stages must be supported by VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStages", 136}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-supportedIndirectCommandsShaderStages-11061", 81}, { {"Only stages specified in VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStages can be set in pGeneratedCommandsInfo->shaderStages", 167}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-suspended", 47}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-uniformBuffers-06935", 58}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-unnormalizedCoordinates-09635", 67}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewMask-06178", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewType-07752", 52}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-03417", 57}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-03419", 57}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-04137", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-04138", 57}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-04139", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-07492", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-07493", 57}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsEXT-viewportCount-09421", 57}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-Input-07939", 48}, { {"If the vertexAttributeRobustness feature is not enabled, and the maintenance9 feature is not enabled, and there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then all variables with the Input storage class decorated with Location in the Vertex Execution Model OpEntryPoint must contain a location in VkVertexInputAttributeDescription2EXT::location", 480}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-Input-08734", 48}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and either the legacyVertexAttributes feature is not enabled or the SPIR-V Type associated with a given Input variable of the corresponding Location in the Vertex Execution Model OpEntryPoint is 64-bit, then the numeric type associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be the same as VkVertexInputAttributeDescription2EXT::format", 587}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-InputAttachmentIndex-09597", 63}, { {"If an input attachment view accessed in a dynamic render pass via a shader object has an InputAttachmentIndex, the InputAttachmentIndex must match an index in VkRenderingInputAttachmentIndexInfo", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-02691", 47}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-02692", 47}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-02693", 47}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-02721", 47}, { {"If the robustBufferAccess feature is not enabled, and that pipeline was created without enabling VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS for vertexInputs, then for a given vertex buffer binding, any attribute data fetched must be entirely contained within the corresponding vertex buffer binding, as described in Vertex Input Description", 357}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-02910", 47}, { {"Transform feedback must not be active", 37}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04007", 47}, { {"All vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must have either valid or VK_NULL_HANDLE buffers bound", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04008", 47}, { {"If the nullDescriptor feature is not enabled, all vertex input bindings accessed via vertex input variables declared in the vertex shader entry point's interface must not be VK_NULL_HANDLE", 188}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04875", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with both a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage and the VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT dynamic state enabled, and the current value of primitiveTopology is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, then vkCmdSetPatchControlPointsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 498}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04876", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE dynamic state enabled, then vkCmdSetRasterizerDiscardEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 326}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04877", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBiasEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04879", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-04914", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled then vkCmdSetVertexInputEXT must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 327}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-06479", 47}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-06537", 47}, { {"Memory backing image subresources used as attachments in the current render pass must not be written in any way other than as an attachment by this command", 155}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-06666", 47}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of sampleLocationsEnable is VK_TRUE, then vkCmdSetSampleLocationsEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 483}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-06886", 47}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the depth aspect, depth writes must be disabled", 143}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-06887", 47}, { {"If the current render pass instance uses a depth/stencil attachment with a read-only layout for the stencil aspect, both front and back writeMask are not zero, and stencil test is enabled, all stencil ops must be VK_STENCIL_OP_KEEP", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07288", 47}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07469", 47}, { {"Input attachment views accessed in a subpass must be created with the same VkFormat as the corresponding subpass definition, and be created with a VkImageView that is compatible with the attachment referenced by the subpass' pInputAttachments[InputAttachmentIndex] in the bound VkFramebuffer as specified by Fragment Input Attachment Compatibility", 347}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07619", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stage and the VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT dynamic state enabled, then vkCmdSetTessellationDomainOriginEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 431}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07620", 47}, { {"If the depthClamp feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthClampEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 416}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07621", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_POLYGON_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetPolygonModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 370}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07622", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRasterizationSamplesEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 388}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07623", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07624", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToCoverageEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07625", 47}, { {"If the alphaToOne feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAlphaToOneEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07626", 47}, { {"If the logicOp feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetLogicOpEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07627", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorBlendEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 432}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07629", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and there are color attachments bound, then vkCmdSetColorWriteMaskEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 428}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07630", 47}, { {"If the geometryStreams feature is enabled, and a shader object is bound to the VK_SHADER_STAGE_GEOMETRY_BIT stage or a graphics pipeline is bound which was created with both a VK_SHADER_STAGE_GEOMETRY_BIT stage and the VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT dynamic state enabled, then vkCmdSetRasterizationStreamEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 434}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07631", 47}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetConservativeRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 467}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07632", 47}, { {"If the VK_EXT_conservative_rasterization extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of conservativeRasterizationMode is VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT, then vkCmdSetExtraPrimitiveOverestimationSizeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 581}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07633", 47}, { {"If the depthClipEnable feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT dynamic state, then vkCmdSetDepthClipEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 353}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07634", 47}, { {"If the VK_EXT_sample_locations extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetSampleLocationsEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07636", 47}, { {"If the VK_EXT_provoking_vertex extension is enabled, a shader object is bound to the VK_SHADER_STAGE_VERTEX_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetProvokingVertexModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07639", 47}, { {"If the depthClipControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT dynamic state enabled, then vkCmdSetDepthClipNegativeOneToOneEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 385}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07640", 47}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV dynamic state enabled, then vkCmdSetViewportWScalingEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 387}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07641", 47}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then vkCmdSetViewportSwizzleNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07642", 47}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageToColorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07643", 47}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageToColorEnable is VK_TRUE, then vkCmdSetCoverageToColorLocationNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 508}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07644", 47}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageModulationModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 449}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07645", 47}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationMode is any value other than VK_COVERAGE_MODULATION_MODE_NONE_NV, then vkCmdSetCoverageModulationTableEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 569}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07646", 47}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of coverageModulationTableEnable is VK_TRUE, then vkCmdSetCoverageModulationTableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 514}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07647", 47}, { {"If the shadingRateImage feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetShadingRateImageEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07648", 47}, { {"If the representativeFragmentTest feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetRepresentativeFragmentTestEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 463}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07649", 47}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoverageReductionModeNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 435}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07749", 47}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorWriteEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 442}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07751", 47}, { {"If the VK_EXT_discard_rectangles extension is enabled, a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo included a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPipelineDiscardRectangleStateCreateInfoEXT::discardRectangleCount", 654}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07831", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT dynamic state enabled then vkCmdSetViewport must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 242}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07832", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SCISSOR dynamic state enabled then vkCmdSetScissor must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 240}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07834", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BIAS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBiasEnable is VK_TRUE, then vkCmdSetDepthBias or vkCmdSetDepthBias2EXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07835", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_BLEND_CONSTANTS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE with a blend equations where any VkBlendFactor member is VK_BLEND_FACTOR_CONSTANT_COLOR, VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR, VK_BLEND_FACTOR_CONSTANT_ALPHA, or VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA, then vkCmdSetBlendConstants must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 664}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07836", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthBoundsTestEnable is VK_TRUE, then vkCmdSetDepthBounds must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 418}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07837", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilCompareMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 429}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07838", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_WRITE_MASK dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilWriteMask must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 425}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07839", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_REFERENCE dynamic state enabled, the current value of and rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilReference must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 424}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07840", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_CULL_MODE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCullMode must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 357}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07841", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRONT_FACE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFrontFace must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 359}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07842", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled then vkCmdSetPrimitiveTopology must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 335}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07843", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, vkCmdSetDepthTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07844", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthWriteEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 427}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07845", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_COMPARE_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of depthTestEnable is VK_TRUE, then vkCmdSetDepthCompareOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 419}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07846", 47}, { {"If the depthBounds feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDepthBoundsTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 421}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07847", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetStencilTestEnable must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 376}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07848", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_STENCIL_OP dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, the current value of stencilTestEnable is VK_TRUE, then vkCmdSetStencilOp must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 406}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07849", 47}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled and a shader object is bound to any graphics stage, or a bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of stippledLineEnable is VK_TRUE, then vkCmdSetLineStipple must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 521}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07850", 47}, { {"If dynamic state was inherited from VkCommandBufferInheritanceViewportScissorInfoNV, it must be set in the current command buffer prior to this drawing command", 159}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07878", 47}, { {"If the exclusiveScissor feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV dynamic state enabled, then vkCmdSetExclusiveScissorEnableNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 374}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07879", 47}, { {"If the exclusiveScissor feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV dynamic state enabled, and the most recent call to vkCmdSetExclusiveScissorEnableNV in the current command buffer set any element of pExclusiveScissorEnables to VK_TRUE, then vkCmdSetExclusiveScissorNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 504}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07880", 47}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetDiscardRectangleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 441}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07881", 47}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-07888", 47}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08114", 47}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08115", 47}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08116", 47}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08117", 47}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08119", 47}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08600", 47}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08601", 47}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08605", 47}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08606", 47}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08608", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08609", 47}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08610", 47}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08611", 47}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08612", 47}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08613", 47}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08617", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_WIDTH dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineWidth must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 433}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08636", 47}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportWScalingNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 511}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08637", 47}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then the viewportCount parameter in the last call to vkCmdSetViewportShadingRatePaletteNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 582}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08644", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and none of the following is enabled: the VK_AMD_mixed_attachment_samples extension the VK_NV_framebuffer_mixed_samples extension the multisampledRenderToSingleSampled feature then the current value of rasterizationSamples must be the same as the current color and/or depth/stencil attachments", 434}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08666", 47}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineRasterizationModeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 578}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08669", 47}, { {"If any of the stippledRectangularLines, stippledBresenhamLines or stippledSmoothLines features are enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line topology class, then vkCmdSetLineStippleEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 570}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08684", 47}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_VERTEX_BIT", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08685", 47}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT", 231}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08686", 47}, { {"If there is no bound graphics pipeline, and the tessellationShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 234}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08687", 47}, { {"If there is no bound graphics pipeline, and the geometryShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_GEOMETRY_BIT", 215}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08688", 47}, { {"If there is no bound graphics pipeline, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_FRAGMENT_BIT", 172}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08689", 47}, { {"If there is no bound graphics pipeline, and the taskShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_TASK_BIT_EXT", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08690", 47}, { {"If there is no bound graphics pipeline, and the meshShader feature is enabled, vkCmdBindShadersEXT must have been called in the current command buffer with pStages with an element of VK_SHADER_STAGE_MESH_BIT_EXT", 211}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08693", 47}, { {"If there is no bound graphics pipeline, and at least one of the taskShader and meshShader features is enabled, one of the VK_SHADER_STAGE_VERTEX_BIT or VK_SHADER_STAGE_MESH_BIT_EXT stages must have a valid VkShaderEXT bound, and the other must have no VkShaderEXT bound", 269}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08696", 47}, { {"If there is no bound graphics pipeline, and a valid VkShaderEXT is bound to the VK_SHADER_STAGE_VERTEX_BIT stage, there must be no VkShaderEXT bound to either the VK_SHADER_STAGE_TASK_BIT_EXT stage or the VK_SHADER_STAGE_MESH_BIT_EXT stage", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08698", 47}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, then all shaders created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag in the same vkCreateShadersEXT call must also be bound", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08699", 47}, { {"If any graphics shader is bound which was created with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag, any stages in between stages whose shaders which did not create a shader with the VK_SHADER_CREATE_LINK_STAGE_BIT_EXT flag as part of the same vkCreateShadersEXT call must not have any VkShaderEXT bound", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08876", 47}, { {"If a shader object is bound to any graphics stage, the current render pass instance must have been begun with vkCmdBeginRendering", 129}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08877", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetAttachmentFeedbackLoopEnableEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 426}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08878", 47}, { {"All bound graphics shader objects must have been created with identical or identically defined push constant ranges", 115}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08879", 47}, { {"All bound graphics shader objects must have either been created with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag set, or with identical or identically defined arrays of descriptor set layouts", 197}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-08885", 47}, { {"There must be no shader object bound to either of the VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT stages", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09003", 47}, { {"If an attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, it must not be accessed in any way other than as an attachment, storage image, or sampled image by this command", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09116", 47}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline was created with VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT, and the format of any color attachment is VK_FORMAT_E5B9G9R9_UFLOAT_PACK32, the corresponding element of the pColorWriteMasks parameter of vkCmdSetColorWriteMaskEXT must either include all of VK_COLOR_COMPONENT_R_BIT, VK_COLOR_COMPONENT_G_BIT, VK_COLOR_COMPONENT_B_BIT, and VK_COLOR_COMPONENT_A_BIT, or none of them", 452}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09203", 47}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then all Input variables at the corresponding Location in the Vertex Execution Model OpEntryPoint must not use components that are not present in the format", 418}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09211", 47}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, or a shader object is bound to any graphics stage, and the current render pass instance includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must be the same as the rasterizationSamples member of that structure", 482}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09363", 47}, { {"If there is no shader object bound to any graphics stage, the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 589}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09364", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetColorBlendEnableEXT must have set the blend enable to VK_FALSE prior to this drawing command", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09365", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetRasterizationSamplesEXT must have set rasterizationSamples to VK_SAMPLE_COUNT_1_BIT prior to this drawing command", 505}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09368", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->width to 1 prior to this drawing command", 484}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09369", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, the current render pass instance was begun with vkCmdBeginRendering, and it includes a color attachment that uses the VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID resolve mode, then the last call to vkCmdSetFragmentShadingRateKHR must have set pFragmentSize->height to 1 prior to this drawing command", 485}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09548", 47}, { {"If the current render pass was begun with vkCmdBeginRendering, there is no shader object bound to any graphics stage, the value of each element of VkRenderingAttachmentLocationInfo::pColorAttachmentLocations in the bound pipeline must match the value for the corresponding locations set currently in the current render pass instance", 332}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09549", 47}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of each element of VkRenderingInputAttachmentIndexInfo::pColorAttachmentInputIndices in the bound pipeline must match the value for the corresponding index set currently in the current render pass instance", 337}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09600", 47}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09637", 47}, { {"If the primitiveTopologyListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_POINT_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY, or VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY, there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled, then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 535}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09642", 47}, { {"If the current render pass was begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT", 230}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09643", 47}, { {"If the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_ENABLE_LEGACY_DITHERING_BIT_EXT, the current render pass must have begun with vkCmdBeginRendering with the VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT flag", 225}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-09650", 47}, { {"If the depthClampControl feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_DEPTH_CLAMP_RANGE_EXT dynamic state enabled, and the current value of depthClampEnable is VK_TRUE, then vkCmdSetDepthClampRangeEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 417}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10068", 47}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10608", 47}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic state enabled, the effective rasterization input topology is in the line topology class, and the current lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM or VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the current alphaToCoverageEnable, alphaToOneEnable and sampleShadingEnable states must all be VK_FALSE", 487}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10677", 47}, { {"If the per-tile execution model is enabled, the tileShadingPerTileDraw feature must be enabled", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10678", 47}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10679", 47}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10772", 47}, { {"If a shader object is bound to any graphics stage, multiview functionality must not be enabled in the current render pass", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10795", 47}, { {"If a color attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10796", 47}, { {"If a depth attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 294}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10797", 47}, { {"If a stencil attachment is written by any prior command in this subpass or by the load, store, or resolve operations for this subpass, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, it must not be accessed in any way other than as an attachment by this command", 298}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10862", 47}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT , but not the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10864", 47}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then either vkCmdSetColorBlendAdvancedEXT or vkCmdSetColorBlendEquationEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 560}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10909", 47}, { {"If the primitiveTopologyPatchListRestart feature is not enabled, the input assembly is VK_PRIMITIVE_TOPOLOGY_PATCH_LIST, there is a shader object bound to the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE dynamic state enabled then vkCmdSetPrimitiveRestartEnable must be VK_FALSE", 380}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10927", 47}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 291}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-10928", 47}, { {"If the current render pass was begun with vkCmdBeginRendering, and there is no shader object bound to any graphics stage, the value of VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex in the bound pipeline must match the value set currently in the current render pass instance", 293}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11297", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11298", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11299", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11300", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11301", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11302", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11304", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11305", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11306", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11308", 47}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11309", 47}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11372", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11373", 47}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11374", 47}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11376", 47}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11397", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11398", 47}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11437", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11438", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11439", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11441", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11442", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11485", 47}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11522", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must have been created with a VkCustomResolveCreateInfoEXT", 254}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11523", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering and does not contain a custom resolve, and the dynamicRenderingUnusedAttachments feature is not enabled, the graphics pipeline bound must not have been created with a VkCustomResolveCreateInfoEXT", 266}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11861", 47}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, the bound graphics pipeline must have been created with a VkCustomResolveCreateInfoEXT::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 322}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11862", 47}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline", 518}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11863", 47}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView equal to VK_NULL_HANDLE must have the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 496}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11865", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 384}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11866", 47}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 444}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11868", 47}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 388}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-11869", 47}, { {"If current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 450}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-12338", 47}, { {"If a color attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_COLOR_BIT on that attachment, the color attachment must not be written to by this command", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-12339", 47}, { {"If a depth attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_DEPTH_BIT on that attachment, the depth attachment must not be written to by this command", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-12340", 47}, { {"If a stencil attachment is read in this command in any way other than as an attachment, or has been read by any prior command in this subpass as a non-attachment, and feedback loop is not enabled for VK_IMAGE_ASPECT_STENCIL_BIT on that attachment, the stencil attachment must not be written to by this command", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-12472", 47}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-None-13118", 47}, { {"If the bound graphics pipeline was created without the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, and vkCmdBindVertexBuffers3KHR was called and not subsequently invalidated in the current command buffer prior to this draw command, the value of setStride in each of its pBindInfos elements must have been VK_FALSE", 406}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpExecutionMode-12239", 58}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 245}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpExecutionMode-12240", 58}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 250}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpExecutionMode-12241", 58}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the segment spacing, they must be the same", 241}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpExecutionMode-12242", 58}, { {"If a shader is bound to both the VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT stages, and if both stages contain an OpExecutionMode instruction specifying the output patch size, they must be the same", 243}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchSADQCOM-06976", 67}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchSADQCOM-12420", 67}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchSSDQCOM-06974", 67}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchWindow-09215", 66}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchWindow-09216", 66}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchWindow-09217", 66}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBlockMatchWindow-12421", 66}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageBoxFilterQCOM-06973", 63}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageSampleWeightedQCOM-06971", 68}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageSampleWeightedQCOM-06972", 68}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageSampleWeightedQCOM-06977", 68}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageSampleWeightedQCOM-06978", 68}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageWrite-04469", 55}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageWrite-08795", 55}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpImageWrite-08796", 55}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTensorReadARM-12462", 58}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeImage-07027", 54}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeImage-07028", 54}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeImage-07029", 54}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeImage-07030", 54}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeImage-07468", 54}, { {"If any shader executed by this pipeline accesses an OpTypeImage variable with a Dim operand of SubpassData, it must be decorated with an InputAttachmentIndex that corresponds to a valid input attachment in the current subpass", 225}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-OpTypeTensorARM-09906", 58}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-SampledType-04470", 54}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-SampledType-04471", 54}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-SampledType-04472", 54}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-SampledType-04473", 54}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-VkPipelineVieportCreateInfo-04141", 70}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportSwizzleStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 436}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-VkPipelineVieportCreateInfo-04142", 70}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled and a VkPipelineViewportExclusiveScissorStateCreateInfoNV structure chained from VkPipelineViewportStateCreateInfo, then the bound graphics pipeline must have been created with VkPipelineViewportExclusiveScissorStateCreateInfoNV::exclusiveScissorCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 462}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-addressModeU-12422", 55}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-advancedBlendMaxColorAttachments-07480", 75}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT and VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and an active color attachment current value of blendEnable is VK_TRUE, then the number of active color attachments must not exceed advancedBlendMaxColorAttachments", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-alphaToCoverageEnable-08919", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT dynamic state enabled, and alphaToCoverageEnable was VK_TRUE in the last call to vkCmdSetAlphaToCoverageEnableEXT, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 331}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-alphaToCoverageEnable-08920", 64}, { {"If a shader object is bound to any graphics stage, and the most recent call to vkCmdSetAlphaToCoverageEnableEXT in the current command buffer set alphaToCoverageEnable to VK_TRUE, then the Fragment Output Interface must contain a variable for the alpha Component word in Location 0 at Index 0", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-attachmentCount-07750", 58}, { {"If the colorWriteEnable feature is enabled, a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the attachmentCount parameter of most recent call to vkCmdSetColorWriteEnableEXT in the current command buffer must be greater than or equal to the number of active color attachments", 482}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-blendEnable-04727", 54}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, then for each color attachment, if the corresponding image view's format features do not contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT, then the corresponding current value of blendEnable must be VK_FALSE", 457}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-colorAttachmentCount-06179", 63}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled and the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::colorAttachmentCount equal to VkRenderingInfo::colorAttachmentCount", 293}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-colorAttachmentCount-06185", 63}, { {"If the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the corresponding element of the pColorAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline", 595}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-colorAttachmentCount-09362", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering, with a VkRenderingInfo::colorAttachmentCount equal to 1, there is no shader object bound to any graphics stage, and a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE must have been created with an image created with a VkExternalFormatANDROID::externalFormat value equal to the VkExternalFormatANDROID::externalFormat value used to create the bound graphics pipeline", 598}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-colorAttachmentCount-09372", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering and a VkRenderingInfo::colorAttachmentCount equal to 1, a color attachment with a resolve mode of VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_BIT_ANDROID, and a fragment shader is bound, it must not declare the DepthReplacing or StencilRefReplacingEXT execution modes", 337}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-colorAttachmentCount-11871", 63}, { {"If the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0 and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, then for each element of the VkRenderingInfo::pColorAttachments array with a resolveImageView not equal to VK_NULL_HANDLE, the resolveImageView must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 480}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-02707", 56}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-02970", 56}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-10746", 56}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-conservativePointAndLineRasterization-07499", 80}, { {"If the VK_EXT_conservative_rasterization extension is enabled, conservativePointAndLineRasterization is not supported, a shader object is bound to any graphics stage or a graphics pipeline is bound, the current value of rasterizerDiscardEnable is VK_FALSE, and the effective rasterization input topology is in the line or point topology class, then the current value of conservativeRasterizationMode must be VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT", 455}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-coverageModulationTableEnable-07488", 72}, { {"If a shader object is bound to any graphics stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV state enabled, and the last call to vkCmdSetCoverageModulationTableEnableNV set coverageModulationTableEnable to VK_TRUE, then the coverageModulationTableCount parameter in the last call to vkCmdSetCoverageModulationTableNV must equal the current rasterizationSamples divided by the number of color samples in the current active color attachment", 506}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-coverageReductionMode-07491", 64}, { {"If the coverageReductionMode feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV or VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT dynamic states enabled, then the current values of coverageReductionMode, rasterizationSamples, the sample counts for the color and depth/stencil attachments (if the subpass has them) must be a valid combination returned by vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 534}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-coverageToColorEnable-07490", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV state enabled and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 500}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-cubicRangeClamp-09212", 58}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-customResolve-11524", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 259}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-customResolve-11525", 56}, { {"If the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, and vkCmdBeginCustomResolveEXT has not been recorded in the render pass instance, the graphics pipeline bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 295}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-customResolve-11529", 56}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_TRUE", 333}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-customResolve-11530", 56}, { {"If a shader object is bound to the fragment stage, the current render pass instance was begun with vkCmdBeginRendering and contains a custom resolve, a fragment density map attachment is active, and vkCmdBeginCustomResolveEXT has not yet been called, then the fragment shader object bound must have been created with VkCustomResolveCreateInfoEXT::customResolve as VK_FALSE", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-deviceGeneratedCommands-02911", 66}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dimensionCount-09905", 57}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicPrimitiveTopologyUnrestricted-07500", 79}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY dynamic state enabled and the dynamicPrimitiveTopologyUnrestricted is VK_FALSE, then the primitiveTopology parameter of vkCmdSetPrimitiveTopology must be of the same topology class as the pipeline VkPipelineInputAssemblyStateCreateInfo::topology state", 345}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingLocalRead-11797", 68}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingLocalRead feature is enabled, the VK_RENDERING_LOCAL_READ_CONCURRENT_ACCESS_CONTROL_BIT_KHR flag is specified, and an attachment is being used as a feedback loop as specified by VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR, VkRenderingAttachmentFlagsInfoKHR::flags for that attachment must include VK_RENDERING_ATTACHMENT_INPUT_ATTACHMENT_FEEDBACK_BIT_KHR", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08910", 76}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08911", 76}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 608}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08912", 76}, { {"If the dynamicRenderingUnusedAttachments feature is not enabled, and the current render pass instance was begun with vkCmdBeginRendering and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an imageView equal to VK_NULL_HANDLE must have the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the bound pipeline equal to VK_FORMAT_UNDEFINED", 459}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08913", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 348}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08914", 76}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView", 401}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08915", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 454}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08916", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to VK_FORMAT_UNDEFINED", 352}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08917", 76}, { {"If current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is not enabled, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline must be equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView", 407}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-08918", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, and the value of VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->imageView, the value of the format must be VK_FORMAT_UNDEFINED", 460}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-11864", 76}, { {"If the dynamicRenderingUnusedAttachments feature is enabled, the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, and VkRenderingInfo::colorAttachmentCount greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with an resolveImageView not equal to VK_NULL_HANDLE must have been created with a VkFormat equal to the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the bound graphics pipeline, or the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats, if it exists, must be VK_FORMAT_UNDEFINED", 640}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-11867", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pDepthAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 497}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-dynamicRenderingUnusedAttachments-11870", 76}, { {"If the current render pass instance was begun with vkCmdBeginRendering, it contains a custom resolve, the dynamicRenderingUnusedAttachments feature is enabled, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, and the value of VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the bound graphics pipeline was not equal to the VkFormat used to create VkRenderingInfo::pStencilAttachment->resolveImageView, the value of the format must be VK_FORMAT_UNDEFINED", 503}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-filterCubic-02694", 54}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-filterCubicMinmax-02695", 60}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-firstAttachment-07476", 58}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorBlendEnableEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 413}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-firstAttachment-07478", 58}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage or a graphics pipeline is bound was created with the VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT dynamic states enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then the last call to vkCmdSetColorWriteMaskEXT in the current command buffer prior to this drawing command must have set a value for all active color attachments", 409}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-02696", 48}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-10582", 48}, { {"If the current render pass instance was begun with a vkCmdBeginRendering call in commandBuffer, its VkRenderingInfo::flags parameter must not have VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT set unless VK_RENDERING_CONTENTS_INLINE_BIT_KHR is also set", 258}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-11521", 48}, { {"If current render pass instance was begun with vkCmdBeginRendering with VkRenderingInfo::flags which includes VK_RENDERING_FRAGMENT_REGION_BIT_EXT, and if sample shading is enabled (explicitly or implicitly), then the minimum fraction for sample shading must equal 0.0", 268}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-13361", 48}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then all bound shader objects must have been created with the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit", 239}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-13362", 48}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, then the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT set", 268}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-13363", 48}, { {"If there are any shader objects bound that were created with flags that includes the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR bit, and any of the descriptor set layouts used to create the pipeline layout of the bound descriptor sets were created with any binding's VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then there must be no binding with VkDescriptorSetLayoutBinding::stageFlags set with VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 608}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-13364", 48}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, the pipeline layout of the bound descriptor sets must have been created with VK_PIPELINE_LAYOUT_CREATE_NO_TASK_SHADER_BIT_KHR set", 311}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-flags-13365", 48}, { {"If there are any shader objects bound that were created with flags that includes both the VK_SHADER_CREATE_INDEPENDENT_SETS_BIT_KHR and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT bits, there must be no shader object bound to the VK_SHADER_STAGE_TASK_BIT_EXT stage", 260}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-format-07753", 49}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-format-08936", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and VkVertexInputAttributeDescription2EXT::format has a 64-bit component, then the scalar width associated with all Input variables of the corresponding Location in the Vertex Execution Model OpEntryPoint must be 64-bit", 407}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-format-08937", 49}, { {"If there is a shader object bound to the VK_SHADER_STAGE_VERTEX_BIT stage or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled and the scalar width associated with a Location decorated Input variable in the Vertex Execution Model OpEntryPoint is 64-bit, then the corresponding VkVertexInputAttributeDescription2EXT::format must have a 64-bit component", 412}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-format-10389", 49}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is a packed format, and the legacyVertexAttributes feature is not enabled, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the size of the format", 350}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-format-10390", 49}, { {"For each vertex attribute accessed by this command, if its VkVertexInputAttributeDescription::format or VkVertexInputAttributeDescription2EXT::format is not a packed format, and either the legacyVertexAttributes feature is not enabled or format has 64-bit components, the value of attribAddress, calculated as described in Vertex Input Calculation, must be a multiple of the component size of the format", 403}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageLayout-00344", 54}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06172", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06173", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 298}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06174", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL, this command must not write any values to the depth attachment", 303}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06175", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 309}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06176", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pDepthAttachment is not VK_NULL_HANDLE, and the layout member of pDepthAttachment is VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL, this command must not write any values to the depth attachment", 284}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06177", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the imageView member of pStencilAttachment is not VK_NULL_HANDLE, and the layout member of pStencilAttachment is VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL, this command must not write any values to the stencil attachment", 292}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06183", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentShadingRateAttachmentInfoKHR::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 283}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-imageView-06184", 52}, { {"If the current render pass instance was begun with vkCmdBeginRendering and VkRenderingFragmentDensityMapAttachmentInfoEXT::imageView was not VK_NULL_HANDLE, the bound graphics pipeline must have been created with VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT", 281}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-index-11450", 48}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-isPreprocessed-02908", 57}, { {"If isPreprocessed is VK_TRUE then vkCmdPreprocessGeneratedCommandsNV must have already been executed on the device, using the same pGeneratedCommandsInfo content as well as the content of the input buffers it references (all except VkGeneratedCommandsInfoNV::preprocessBuffer). Furthermore, pGeneratedCommandsInfo's indirectCommandsLayout must have been created with the VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_NV bit set", 439}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-layers-10831", 49}, { {"If the current render pass instance was created with VK_RENDERING_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE or VK_RENDER_PASS_CREATE_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, and the bound graphics pipeline was created with VK_PIPELINE_CREATE_2_PER_LAYER_FRAGMENT_DENSITY_BIT_VALVE, then the current render pass instance must have a layers value less than or equal to VkPipelineFragmentDensityMapLayeredCreateInfoVALVE::maxFragmentDensityMapLayers", 438}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-logicOp-04878", 50}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_LOGIC_OP_EXT dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of logicOpEnable is VK_TRUE, then vkCmdSetLogicOpEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 423}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-magFilter-04553", 52}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-magFilter-09598", 52}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-maxFragmentDualSrcAttachments-09239", 72}, { {"If blending is enabled for any attachment where either the source or destination blend factors for that attachment use the secondary color input, the maximum value of Location for any output attachment statically used in the Fragment Execution Model executed by this command must be less than maxFragmentDualSrcAttachments", 322}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-maxMultiviewInstanceIndex-02688", 68}, { {"If the draw is recorded in a render pass instance with multiview enabled, the maximum instance index must be less than or equal to VkPhysicalDeviceMultiviewProperties::maxMultiviewInstanceIndex", 193}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-micromap-11636", 51}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-micromap-11637", 51}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-micromap-11638", 51}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-micromap-11639", 51}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-mipmapMode-04770", 53}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-mipmapMode-09599", 53}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-multisampledRenderToSingleSampled-07285", 76}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and the current render pass instance was begun with vkCmdBeginRendering with a VkRenderingInfo::colorAttachmentCount parameter greater than 0, then each element of the VkRenderingInfo::pColorAttachments array with a imageView not equal to VK_NULL_HANDLE must have been created with a sample count equal to the value of rasterizationSamples for the bound graphics pipeline", 642}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-multisampledRenderToSingleSampled-07286", 76}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 503}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-multisampledRenderToSingleSampled-07287", 76}, { {"If the bound pipeline was created without a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and the multisampledRenderToSingleSampled feature is not enabled, and vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-multiviewPerViewViewports-12262", 68}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of viewportCount", 189}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-multiviewPerViewViewports-12263", 68}, { {"If the multiviewPerViewViewports feature is enabled, then the index of the most significant bit in current render pass instance viewMask must be less than the current value of scissorCount", 188}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-nextStage-10745", 52}, { {"For each shader object bound to a graphics stage, except for shader object bound to the last graphics stage in the logical pipeline, it must have been created with a nextStage including the corresponding bit to the shader object bound to the following graphics stage in the logical pipeline", 290}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pBindInfo-11375", 52}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pColorAttachments-08963", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->imageView was not VK_NULL_HANDLE, then the corresponding element of VkPipelineRenderingCreateInfo::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 573}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pColorAttachments-11539", 60}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound with a fragment shader that statically writes to a color attachment, the color write mask is not zero, color writes are enabled, and the corresponding element of the VkRenderingInfo::pColorAttachments->resolveImageView was not VK_NULL_HANDLE, then the corresponding element of VkCustomResolveCreateInfoEXT::pColorAttachmentFormats used to create the pipeline must not be VK_FORMAT_UNDEFINED", 571}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthAttachment-06186", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->imageView", 515}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthAttachment-08964", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 411}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthAttachment-11540", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, depth test is enabled, and the VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::depthAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 409}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthAttachment-11872", 59}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, and VkRenderingInfo::pDepthAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pDepthAttachment->resolveImageView", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthInputAttachmentIndex-09595", 69}, { {"Input attachment views accessed in a dynamic render pass with a InputAttachmentIndex referenced by VkRenderingInputAttachmentIndexInfo, or no InputAttachmentIndex if VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex or VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are NULL, must be created with a VkImageView that is compatible with the corresponding color, depth, or stencil attachment in VkRenderingInfo", 440}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDepthInputAttachmentIndex-09596", 69}, { {"Input attachment views accessed in a dynamic render pass via a shader object must have an InputAttachmentIndex if both VkRenderingInputAttachmentIndexInfo::pDepthInputAttachmentIndex and VkRenderingInputAttachmentIndexInfo::pStencilInputAttachmentIndex are non-NULL", 265}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDescription-09900", 55}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDynamicStates-08715", 57}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpDepthAttachmentReadEXT, the depthWriteEnable parameter in the last call to vkCmdSetDepthWriteEnable must be VK_FALSE", 379}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pDynamicStates-08716", 57}, { {"If the bound graphics pipeline state includes a fragment shader stage, was created with VK_DYNAMIC_STATE_STENCIL_WRITE_MASK set in VkPipelineDynamicStateCreateInfo::pDynamicStates, and the fragment shader declares the EarlyFragmentTests execution mode and uses OpStencilAttachmentReadEXT, the writeMask parameter in the last call to vkCmdSetStencilWriteMask must be 0", 367}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pGeneratedCommandsInfo-parameter", 69}, { {"pGeneratedCommandsInfo must be a valid pointer to a valid VkGeneratedCommandsInfoNV structure", 93}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pNext-07935", 48}, { {"If this command is called inside a render pass instance started with vkCmdBeginRendering, and the pNext chain of VkRenderingInfo includes a VkMultisampledRenderToSingleSampledInfoEXT structure with multisampledRenderToSingleSampledEnable equal to VK_TRUE, then the value of rasterizationSamples for the bound graphics pipeline must be equal to VkMultisampledRenderToSingleSampledInfoEXT::rasterizationSamples", 408}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pNext-12463", 48}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pStencilAttachment-06187", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline was created with a VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV structure, and VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, the value of the depthStencilAttachmentSamples member of VkAttachmentSampleCountInfoAMD or VkAttachmentSampleCountInfoNV used to create the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->imageView", 519}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pStencilAttachment-08965", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has not yet been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->imageView was not VK_NULL_HANDLE, then the VkPipelineRenderingCreateInfo::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 416}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pStencilAttachment-11860", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, there is a graphics pipeline bound, stencil test is enabled and the VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, then the VkCustomResolveCreateInfoEXT::stencilAttachmentFormat used to create the pipeline must not be VK_FORMAT_UNDEFINED", 414}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pStencilAttachment-11873", 61}, { {"If the current render pass instance was begun with vkCmdBeginRendering, vkCmdBeginCustomResolveEXT has been recorded in the render pass instance, VkRenderingInfo::pStencilAttachment->resolveImageView was not VK_NULL_HANDLE, the value of rasterizationSamples for the bound graphics pipeline must be equal to the sample count used to create VkRenderingInfo::pStencilAttachment->resolveImageView", 392}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pStrides-04913", 51}, { {"If the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE dynamic state enabled, but without the VK_DYNAMIC_STATE_VERTEX_INPUT_EXT dynamic state enabled, then vkCmdBindVertexBuffers2 with a non-NULL pStrides parameter or vkCmdBindVertexBuffers3KHR with setStride set to VK_TRUE must have been called and not subsequently invalidated in the current command buffer prior to this draw command", 428}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pipeline-02909", 51}, { {"VkGeneratedCommandsInfoNV::pipeline must match the current bound pipeline at VkGeneratedCommandsInfoNV::pipelineBindPoint", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-pipelineFragmentShadingRate-09238", 70}, { {"If the pipelineFragmentShadingRate feature is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetFragmentShadingRateKHR must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 439}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitiveFragmentShadingRateWithMultipleViewports-04552", 92}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, the bound graphics pipeline was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and any of the shader stages of the bound graphics pipeline write to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 481}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitiveFragmentShadingRateWithMultipleViewports-08642", 92}, { {"If the primitiveFragmentShadingRateWithMultipleViewports limit is not supported, and any shader object bound to a graphics stage writes to the PrimitiveShadingRateKHR built-in, then vkCmdSetViewportWithCount must have been called in the current command buffer prior to this drawing command, and the viewportCount parameter of vkCmdSetViewportWithCount must be 1", 361}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitiveTopology-10286", 60}, { {"If a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage is bound, then the current value of primitiveTopology must be VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command", 176}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitiveTopology-10747", 60}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_PATCH_LIST prior to this drawing command, then a VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT stage must be bound", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitiveTopology-10748", 60}, { {"If vkCmdSetPrimitiveTopology set primitiveTopology to VK_PRIMITIVE_TOPOLOGY_POINT_LIST prior to this drawing command, the maintenance5 feature is not enabled, both a VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT and VK_SHADER_STAGE_GEOMETRY_BIT stage are not bound, then the Vertex Execution Model must have a PointSize decorated variable that is statically written to", 366}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitivesGeneratedQueryWithNonZeroStreams-06709", 85}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, the bound graphics pipeline must not have been created with a non-zero value in VkPipelineRasterizationStateStreamCreateInfoEXT::rasterizationStream", 285}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitivesGeneratedQueryWithNonZeroStreams-07481", 85}, { {"If the primitivesGeneratedQueryWithNonZeroStreams feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, and the bound graphics pipeline was created with VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT state enabled, the last call to vkCmdSetRasterizationStreamEXT must have set the rasterizationStream to zero", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-primitivesGeneratedQueryWithRasterizerDiscard-06708", 88}, { {"If the primitivesGeneratedQueryWithRasterizerDiscard feature is not enabled and the VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT query is active, rasterization discard must not be enabled", 181}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-protectedNoFault-11455", 59}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-protectedNoFault-11456", 59}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizationSamples-07471", 63}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current subpass does not use any color and/or depth/stencil attachments, then the rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT must follow the rules for a zero-attachment subpass", 330}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizationSamples-07489", 63}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and if current subpass has a depth/stencil attachment and depth test, stencil test, or depth bounds test are enabled in the bound pipeline, then the current rasterizationSamples must be the same as the sample count of the depth/stencil attachment", 307}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizationSamples-07494", 63}, { {"If the VK_NV_framebuffer_mixed_samples extension is enabled, and the coverageReductionMode feature is not enabled, or the current value of coverageReductionMode is not VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV, and the current value of rasterizationSamples is greater than sample count of the color attachment, then sample shading must be disabled", 344}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizerDiscardEnable-09236", 66}, { {"If the VK_EXT_discard_rectangles extension is enabled, a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT dynamic state enabled and the pNext chain of VkGraphicsPipelineCreateInfo did not include a VkPipelineDiscardRectangleStateCreateInfoEXT structure, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of discardRectangleEnable is VK_TRUE, then vkCmdSetDiscardRectangleEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for each discard rectangle in VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 707}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizerDiscardEnable-09420", 66}, { {"If the VK_NV_fragment_coverage_to_color extension is enabled, and a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, and the most recent call to vkCmdSetRasterizerDiscardEnable in the current command buffer set rasterizerDiscardEnable to VK_FALSE, and the last call to vkCmdSetCoverageToColorEnableNV set the coverageToColorEnable to VK_TRUE, then there must be an active color attachment at the location selected by the last call to vkCmdSetCoverageToColorLocationNV coverageToColorLocation, with a VkFormat of VK_FORMAT_R8_UINT, VK_FORMAT_R8_SINT, VK_FORMAT_R16_UINT, VK_FORMAT_R16_SINT, VK_FORMAT_R32_UINT, or VK_FORMAT_R32_SINT", 648}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-rasterizerDiscardEnable-10863", 66}, { {"If a graphics pipeline is bound was created with VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT, but not the VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT dynamic state enabled, and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetColorBlendAdvancedEXT must have been called and not subsequently invalidated in the current command buffer prior to this drawing command for all active color attachments with the blendEnable current value of VK_TRUE", 459}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-reductionMode-09213", 56}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-renderPass-02684", 53}, { {"The current render pass must be compatible with the renderPass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 195}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-renderPass-06198", 53}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound pipeline must have been created with a VkGraphicsPipelineCreateInfo::renderPass equal to VK_NULL_HANDLE", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-renderpass", 47}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-02689", 64}, { {"If the bound graphics pipeline was created with VkPipelineSampleLocationsStateCreateInfoEXT::sampleLocationsEnable set to VK_TRUE, then the active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 261}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07484", 64}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT then the current active depth attachment must have been created with the VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT bit set", 412}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07485", 64}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.width in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 644}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07486", 64}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, then the sampleLocationsInfo.maxSampleLocationGridSize.height in the last call to vkCmdSetSampleLocationsEXT must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling rasterizationSamples", 646}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07487", 64}, { {"If a shader object is bound to the VK_SHADER_STAGE_FRAGMENT_BIT stage, or the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT state enabled, and if sampleLocationsEnable was VK_TRUE in the last call to vkCmdSetSampleLocationsEnableEXT, the fragment shader code must not statically use the extended instruction InterpolateAtSample", 377}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07936", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.width must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.width as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 547}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07937", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationGridSize.height must evenly divide VkMultisamplePropertiesEXT::maxSampleLocationGridSize.height as returned by vkGetPhysicalDeviceMultisamplePropertiesEXT with a samples parameter equaling the value of rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 549}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsEnable-07938", 64}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state disabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then sampleLocationsInfo.sampleLocationsPerPixel must equal rasterizationSamples in the last call to vkCmdSetRasterizationSamplesEXT", 368}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsPerPixel-07482", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples member of the VkPipelineMultisampleStateCreateInfo structure the bound graphics pipeline has been created with", 493}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sampleLocationsPerPixel-07483", 66}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state enabled, and the current value of sampleLocationsEnable is VK_TRUE, then the sampleLocationsPerPixel member of pSampleLocationsInfo in the last call to vkCmdSetSampleLocationsEXT must equal the rasterizationSamples parameter of the last call to vkCmdSetRasterizationSamplesEXT", 443}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-samples-07472", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state enabled and the VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT state disabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the VkPipelineMultisampleStateCreateInfo::rasterizationSamples parameter used to create the bound graphics pipeline", 384}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-samples-07473", 50}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_SAMPLE_MASK_EXT state and VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT states enabled, then the samples parameter in the last call to vkCmdSetSampleMaskEXT must be greater or equal to the rasterizationSamples parameter in the last call to vkCmdSetRasterizationSamplesEXT", 343}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-scissorCount-03418", 55}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetScissorWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing", 336}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-selectableCubicWeights-09214", 65}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-shadingRateImage-09233", 59}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV and the current value of rasterizerDiscardEnable is VK_FALSE, then vkCmdSetCoarseSampleOrderNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 412}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-shadingRateImage-09234", 59}, { {"If the shadingRateImage feature is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, the current value of rasterizerDiscardEnable is VK_FALSE, and the current value of shadingRateImageEnable is VK_TRUE, then vkCmdSetViewportShadingRatePaletteNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 542}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sparseImageInt64Atomics-04474", 66}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-sparseImageInt64Atomics-04475", 66}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stage-06481", 48}, { {"The bound graphics pipeline must not have been created with the VkPipelineShaderStageCreateInfo::stage member of any element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT", 233}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stage-07073", 48}, { {"If the bound pipeline was created with the VkPipelineShaderStageCreateInfo::stage member of an element of VkGraphicsPipelineCreateInfo::pStages set to VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT or VK_SHADER_STAGE_GEOMETRY_BIT, then Mesh Shader Queries must not be active", 341}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stippledLineEnable-07495", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the stippledRectangularLines feature must be enabled", 370}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stippledLineEnable-07496", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the stippledBresenhamLines feature must be enabled", 366}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stippledLineEnable-07497", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the stippledSmoothLines feature must be enabled", 372}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-stippledLineEnable-07498", 61}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT or VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT dynamic states enabled, and if the current stippledLineEnable state is VK_TRUE and the current lineRasterizationMode state is VK_LINE_RASTERIZATION_MODE_DEFAULT, then the stippledRectangularLines feature must be enabled and VkPhysicalDeviceLimits::strictLines must be VK_TRUE", 422}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-storageBuffers-06936", 57}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-subpass-02685", 50}, { {"The subpass index of the current render pass must be equal to the subpass member of the VkGraphicsPipelineCreateInfo structure specified when creating the VkPipeline bound to VK_PIPELINE_BIND_POINT_GRAPHICS", 206}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-suspended", 46}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-uniformBuffers-06935", 57}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-unnormalizedCoordinates-09635", 66}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewMask-06178", 51}, { {"If the current render pass instance was begun with vkCmdBeginRendering, the bound graphics pipeline must have been created with a VkPipelineRenderingCreateInfo::viewMask equal to VkRenderingInfo::viewMask", 204}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewType-07752", 51}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-03417", 56}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, and the state is not inherited, then vkCmdSetViewportWithCount must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 346}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-03419", 56}, { {"If a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with both the VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic states enabled, and the state is not inherited, then the viewportCount parameter of vkCmdSetViewportWithCount must match the scissorCount parameter of vkCmdSetScissorWithCount", 374}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-04137", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportWScalingStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 405}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-04138", 56}, { {"If the VK_NV_clip_space_w_scaling extension is enabled, and a shader object is bound to any graphics stage or a graphics pipeline is bound which was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV dynamic state enabled, the current value of viewportWScalingEnable is VK_TRUE, then vkCmdSetViewportWScalingNV must have been called and not subsequently invalidated in the current command buffer prior to this drawing command", 471}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-04139", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportShadingRateImageStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 424}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-07492", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT dynamic state enabled, but not the VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic state enabled, then the bound graphics pipeline must have been created with VkPipelineViewportSwizzleStateCreateInfoNV::viewportCount greater or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 402}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-07493", 56}, { {"If the bound graphics pipeline state was created with the VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT and VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV dynamic states enabled then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 344}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdExecuteGeneratedCommandsNV-viewportCount-09421", 56}, { {"If the VK_NV_viewport_swizzle extension is enabled, and a shader object is bound to any graphics stage, then the viewportCount parameter in the last call to vkCmdSetViewportSwizzleNV must be greater than or equal to the viewportCount parameter in the last call to vkCmdSetViewportWithCount", 289}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdFillBuffer-apiVersion-07894", 37}, { {"If the VK_KHR_maintenance1 extension is not enabled and VkPhysicalDeviceProperties::apiVersion is less than Vulkan 1.1, the VkCommandPool that commandBuffer was allocated from must support graphics or compute operations", 219}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commandBuffer-01811", 40}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commandBuffer-01812", 40}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-commonparent", 33}, { {"Both of commandBuffer, and dstBuffer must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-dstBuffer-00029", 36}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-dstBuffer-00031", 36}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-dstBuffer-parameter", 40}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-dstOffset-00024", 36}, { {"dstOffset must be less than the size of dstBuffer", 49}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-dstOffset-00025", 36}, { {"dstOffset must be a multiple of 4", 33}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-renderpass", 31}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-size-00026", 31}, { {"If size is not equal to VK_WHOLE_SIZE, size must be greater than 0", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-size-00027", 31}, { {"If size is not equal to VK_WHOLE_SIZE, size must be less than or equal to the size of dstBuffer minus dstOffset", 111}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-size-00028", 31}, { {"If size is not equal to VK_WHOLE_SIZE, size must be a multiple of 4", 67}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillBuffer-videocoding", 32}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-commandBuffer-13003", 43}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 151}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-commandBuffer-13004", 43}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 144}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-dstFlags-13100", 38}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-dstFlags-13101", 38}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-dstFlags-parameter", 42}, { {"dstFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-dstRange-13000", 38}, { {"The buffer from which dstRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag", 114}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13001", 39}, { {"pDstRange->address must be a multiple of 4", 42}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13002", 39}, { {"pDstRange->size must be a multiple of 4", 39}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13097", 39}, { {"If the range specified by pDstRange is not bound completely to memory when accessed, dstFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 149}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13098", 39}, { {"If the buffer from which the range specified by pDstRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 204}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13099", 39}, { {"If the buffer from which the range specified by pDstRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13122", 39}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 283}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13123", 39}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 231}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13124", 39}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 320}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-13125", 39}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 257}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-pDstRange-parameter", 43}, { {"pDstRange must be a valid pointer to a valid VkDeviceAddressRangeKHR structure", 78}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-renderpass", 34}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-suspended", 33}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdFillMemoryKHR-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-bufferlevel", 54}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-commandBuffer-cmdpool", 64}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-commandBuffer-parameter", 66}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-commandBuffer-recording", 66}, { {"commandBuffer must be in the recording state", 44}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-commonparent", 55}, { {"Both of commandBuffer, and executionGraph must have been created, allocated, or retrieved from the same VkDevice", 112}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-executionGraph-parameter", 67}, { {"executionGraph must be a valid VkPipeline handle", 48}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-scratch-09144", 56}, { {"scratch must be a multiple of 64", 32}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-scratch-10185", 56}, { {"scratch must be the device address of an allocated memory range at least as large as scratchSize", 96}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-scratch-parameter", 60}, { {"scratch must be a valid VkDeviceAddress value", 45}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-scratchSize-10186", 60}, { {"scratchSize must be greater than or equal to VkExecutionGraphPipelineScratchSizeAMDX::minSize returned by vkGetExecutionGraphPipelineScratchSizeAMDX for the bound execution graph pipeline", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-suspended", 52}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInitializeGraphScratchMemoryAMDX-videocoding", 54}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCmdInsertDebugUtilsLabelEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdInsertDebugUtilsLabelEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdInsertDebugUtilsLabelEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdInsertDebugUtilsLabelEXT-pLabelInfo-parameter", 55}, { {"pLabelInfo must be a valid pointer to a valid VkDebugUtilsLabelEXT structure", 76}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdNextSubpass-None-00909", 32}, { {"The current subpass index must be less than the number of subpasses in the render pass minus one", 96}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-None-02349", 32}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-bufferlevel", 33}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-contents-parameter", 40}, { {"contents must be a valid VkSubpassContents value", 48}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-renderpass", 32}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-None-02350", 33}, { {"This command must not be recorded when transform feedback is active", 67}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-None-03102", 33}, { {"The current subpass index must be less than the number of subpasses in the render pass minus one", 96}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-bufferlevel", 34}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-pSubpassBeginInfo-parameter", 50}, { {"pSubpassBeginInfo must be a valid pointer to a valid VkSubpassBeginInfo structure", 81}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-pSubpassEndInfo-parameter", 48}, { {"pSubpassEndInfo must be a valid pointer to a valid VkSubpassEndInfo structure", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-renderpass", 33}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdNextSubpass2-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_OPTICAL_FLOW_BIT_NV operations", 108}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-commonparent", 43}, { {"Both of commandBuffer, and session must have been created, allocated, or retrieved from the same VkDevice", 105}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-pExecuteInfo-parameter", 53}, { {"pExecuteInfo must be a valid pointer to a valid VkOpticalFlowExecuteInfoNV structure", 84}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-session-parameter", 48}, { {"session must be a valid VkOpticalFlowSessionNV handle", 53}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdOpticalFlowExecuteNV-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCmdPipelineBarrier-None-07889", 36}, { {"If vkCmdPipelineBarrier is called within a render pass instance using a VkRenderPass object, the render pass must have been created with at least one subpass dependency that expresses a dependency from the current subpass to itself, does not include VK_DEPENDENCY_BY_REGION_BIT if this command does not, does not include VK_DEPENDENCY_VIEW_LOCAL_BIT if this command does not, and has synchronization scopes and access scopes that are all supersets of the scopes defined in this command", 485}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-07890", 36}, { {"If vkCmdPipelineBarrier is called within a render pass instance, and the source stage masks of any memory barriers include framebuffer-space stages, destination stage masks of all memory barriers must only include framebuffer-space stages", 238}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-07892", 36}, { {"If vkCmdPipelineBarrier is called within a render pass instance, the source and destination stage masks of any memory barriers must only include graphics pipeline stages", 169}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-07893", 36}, { {"If vkCmdPipelineBarrier is called inside a render pass instance, and there is more than one view in the current subpass, dependency flags must include VK_DEPENDENCY_VIEW_LOCAL_BIT", 179}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-09553", 36}, { {"If none of the shaderTileImageColorReadAccess, shaderTileImageStencilReadAccess, or shaderTileImageDepthReadAccess features are enabled, and the dynamicRenderingLocalRead feature is not enabled, vkCmdPipelineBarrier must not be called within a render pass instance started with vkCmdBeginRendering", 297}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-09554", 36}, { {"If the dynamicRenderingLocalRead feature is not enabled, and vkCmdPipelineBarrier is called within a render pass instance started with vkCmdBeginRendering, there must be no buffer or image memory barriers specified by this command", 230}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-None-09586", 36}, { {"If the dynamicRenderingLocalRead feature is not enabled, and vkCmdPipelineBarrier is called within a render pass instance started with vkCmdBeginRendering, memory barriers specified by this command must only include VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, or VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT in their access masks", 413}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-bufferMemoryBarrierCount-01178", 56}, { {"If vkCmdPipelineBarrier is called within a render pass instance using a VkRenderPass object, it must not include any buffer memory barriers", 139}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dependencyFlags-01186", 47}, { {"If vkCmdPipelineBarrier is called outside of a render pass instance, the dependency flags must not include VK_DEPENDENCY_VIEW_LOCAL_BIT", 135}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dependencyFlags-07891", 47}, { {"If vkCmdPipelineBarrier is called within a render pass instance, and the source stage masks of any memory barriers include framebuffer-space stages, then dependencyFlags must include VK_DEPENDENCY_BY_REGION_BIT", 210}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dependencyFlags-parameter", 51}, { {"dependencyFlags must be a valid combination of VkDependencyFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstAccessMask-02816", 45}, { {"The dstAccessMask member of each element of pMemoryBarriers must only include access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstAccessMask-06257", 45}, { {"If the rayQuery feature is not enabled and a memory barrier dstAccessMask includes VK_ACCESS_ACCELERATION_STRUCTURE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, or VK_PIPELINE_STAGE_ALL_COMMANDS_BIT", 295}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-03937", 44}, { {"If the synchronization2 feature is not enabled, dstStageMask must not be 0", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04090", 44}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04091", 44}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04092", 44}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04093", 44}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04094", 44}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04095", 44}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-04096", 44}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-06462", 44}, { {"Any pipeline stage included in dstStageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 319}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-07318", 44}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-07949", 44}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-10754", 44}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-dstStageMask-parameter", 48}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-image-04073", 37}, { {"If vkCmdPipelineBarrier is called within a render pass instance using a VkRenderPass object, the image member of any image memory barrier included in this command must be an attachment used in the current subpass both as an input attachment, and as either a color, color resolve, or depth/stencil attachment", 307}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-image-09373", 37}, { {"If vkCmdPipelineBarrier is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, the corresponding color attachment must be VK_ATTACHMENT_UNUSED", 236}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-image-09374", 37}, { {"If vkCmdPipelineBarrier is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, it must have been created with a non-zero VkExternalFormatANDROID::externalFormat value", 260}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-image-09555", 37}, { {"If vkCmdPipelineBarrier is called within a render pass instance started with vkCmdBeginRendering, and the image member of any image memory barrier is used as an attachment in the current render pass instance, it must be in the VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ or VK_IMAGE_LAYOUT_GENERAL layout", 297}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-maintenance8-10206", 44}, { {"If the maintenance8 feature is not enabled, dependencyFlags must not include VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR", 145}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-oldLayout-01181", 41}, { {"If vkCmdPipelineBarrier is called within a render pass instance, the oldLayout and newLayout members of any image memory barrier included in this command must be equal", 167}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-oldLayout-10758", 41}, { {"If called within a render pass instance using a VkRenderPass object, the oldLayout member of any image memory barrier included in this command must be equal to the layout that the corresponding attachment uses during the subpass", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-oldLayout-10759", 41}, { {"If called within a render pass instance started with vkCmdBeginRendering, the oldLayout member of any image memory barrier included in this command must be equal to the layout that the corresponding attachment uses during the render pass instance", 246}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pBufferMemoryBarriers-02817", 53}, { {"For each element of pBufferMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its srcQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its srcAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 435}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pBufferMemoryBarriers-02818", 53}, { {"For each element of pBufferMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its dstQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its dstAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 435}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pBufferMemoryBarriers-parameter", 57}, { {"If bufferMemoryBarrierCount is not 0, pBufferMemoryBarriers must be a valid pointer to an array of bufferMemoryBarrierCount valid VkBufferMemoryBarrier structures", 162}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pImageMemoryBarriers-02819", 52}, { {"For each element of pImageMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its srcQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its srcAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 434}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pImageMemoryBarriers-02820", 52}, { {"For each element of pImageMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its dstQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its dstAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 434}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pImageMemoryBarriers-parameter", 56}, { {"If imageMemoryBarrierCount is not 0, pImageMemoryBarriers must be a valid pointer to an array of imageMemoryBarrierCount valid VkImageMemoryBarrier structures", 158}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-pMemoryBarriers-parameter", 51}, { {"If memoryBarrierCount is not 0, pMemoryBarriers must be a valid pointer to an array of memoryBarrierCount valid VkMemoryBarrier structures", 138}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcAccessMask-02815", 45}, { {"The srcAccessMask member of each element of pMemoryBarriers must only include access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcAccessMask-06257", 45}, { {"If the rayQuery feature is not enabled and a memory barrier srcAccessMask includes VK_ACCESS_ACCELERATION_STRUCTURE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, or VK_PIPELINE_STAGE_ALL_COMMANDS_BIT", 295}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcQueueFamilyIndex-01182", 51}, { {"If vkCmdPipelineBarrier is called within a render pass instance, the srcQueueFamilyIndex and dstQueueFamilyIndex members of any memory barrier included in this command must be equal", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcQueueFamilyIndex-10388", 51}, { {"If a buffer or image memory barrier specifies a queue family ownership transfer operation, either the srcQueueFamilyIndex or dstQueueFamilyIndex member and the queue family index that was used to create the command pool that commandBuffer was allocated from must be equal", 271}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-03937", 44}, { {"If the synchronization2 feature is not enabled, srcStageMask must not be 0", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04090", 44}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04091", 44}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04092", 44}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04093", 44}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04094", 44}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04095", 44}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-04096", 44}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-06461", 44}, { {"Any pipeline stage included in srcStageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 319}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-07318", 44}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-07949", 44}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-09556", 44}, { {"If vkCmdPipelineBarrier is called within a render pass instance started with vkCmdBeginRendering, this command must only specify framebuffer-space stages in srcStageMask and dstStageMask", 186}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-09633", 44}, { {"If either srcStageMask or dstStageMask includes VK_PIPELINE_STAGE_HOST_BIT, for each element of pImageMemoryBarriers, srcQueueFamilyIndex and dstQueueFamilyIndex must be equal", 175}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-09634", 44}, { {"If either srcStageMask or dstStageMask includes VK_PIPELINE_STAGE_HOST_BIT, for each element of pBufferMemoryBarriers, srcQueueFamilyIndex and dstQueueFamilyIndex must be equal", 176}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-10754", 44}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-srcStageMask-parameter", 48}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-07889", 37}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance using a VkRenderPass object, the render pass must have been created with at least one subpass dependency that expresses a dependency from the current subpass to itself, does not include VK_DEPENDENCY_BY_REGION_BIT if this command does not, does not include VK_DEPENDENCY_VIEW_LOCAL_BIT if this command does not, and has synchronization scopes and access scopes that are all supersets of the scopes defined in this command", 486}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-07890", 37}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance, and the source stage masks of any memory barriers include framebuffer-space stages, destination stage masks of all memory barriers must only include framebuffer-space stages", 239}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-07892", 37}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance, the source and destination stage masks of any memory barriers must only include graphics pipeline stages", 170}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-07893", 37}, { {"If vkCmdPipelineBarrier2 is called inside a render pass instance, and there is more than one view in the current subpass, dependency flags must include VK_DEPENDENCY_VIEW_LOCAL_BIT", 180}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-09553", 37}, { {"If none of the shaderTileImageColorReadAccess, shaderTileImageStencilReadAccess, or shaderTileImageDepthReadAccess features are enabled, and the dynamicRenderingLocalRead feature is not enabled, vkCmdPipelineBarrier2 must not be called within a render pass instance started with vkCmdBeginRendering", 298}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-09554", 37}, { {"If the dynamicRenderingLocalRead feature is not enabled, and vkCmdPipelineBarrier2 is called within a render pass instance started with vkCmdBeginRendering, there must be no buffer or image memory barriers specified by this command", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-None-09586", 37}, { {"If the dynamicRenderingLocalRead feature is not enabled, and vkCmdPipelineBarrier2 is called within a render pass instance started with vkCmdBeginRendering, memory barriers specified by this command must only include VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT, VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT, VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT, or VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT in their access masks", 414}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-bufferMemoryBarrierCount-01178", 57}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance using a VkRenderPass object, it must not include any buffer memory barriers", 140}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-dependencyFlags-01186", 48}, { {"If vkCmdPipelineBarrier2 is called outside of a render pass instance, the dependency flags must not include VK_DEPENDENCY_VIEW_LOCAL_BIT", 136}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-dependencyFlags-07891", 48}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance, and the source stage masks of any memory barriers include framebuffer-space stages, then dependencyFlags must include VK_DEPENDENCY_BY_REGION_BIT", 211}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-dstStageMask-09674", 45}, { {"The dstStageMask member of any element of the pMemoryBarriers member of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 227}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-dstStageMask-09676", 45}, { {"If a buffer or image memory barrier does not specify an release operation, or if it does but pDependencyInfo->dependencyFlags includes VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR, the respective dstStageMask member of the element of the pBufferMemoryBarriers or pImageMemoryBarriers members of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 474}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-image-04073", 38}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance using a VkRenderPass object, the image member of any image memory barrier included in this command must be an attachment used in the current subpass both as an input attachment, and as either a color, color resolve, or depth/stencil attachment", 308}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-image-09373", 38}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, the corresponding color attachment must be VK_ATTACHMENT_UNUSED", 237}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-image-09374", 38}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, it must have been created with a non-zero VkExternalFormatANDROID::externalFormat value", 261}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-image-09555", 38}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance started with vkCmdBeginRendering, and the image member of any image memory barrier is used as an attachment in the current render pass instance, it must be in the VK_IMAGE_LAYOUT_RENDERING_LOCAL_READ or VK_IMAGE_LAYOUT_GENERAL layout", 298}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-oldLayout-01181", 42}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance, the oldLayout and newLayout members of any image memory barrier included in this command must be equal", 168}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-pDependencyInfo-parameter", 52}, { {"pDependencyInfo must be a valid pointer to a valid VkDependencyInfo structure", 77}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-srcQueueFamilyIndex-01182", 52}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance, the srcQueueFamilyIndex and dstQueueFamilyIndex members of any memory barrier included in this command must be equal", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-srcQueueFamilyIndex-10387", 52}, { {"If a buffer or image memory barrier specifies a queue family ownership transfer operation, either the srcQueueFamilyIndex or dstQueueFamilyIndex member of the element of the pBufferMemoryBarriers or pImageMemoryBarriers members of pDependencyInfo and the queue family index that was used to create the command pool that commandBuffer was allocated from must be equal", 366}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-srcStageMask-09556", 45}, { {"If vkCmdPipelineBarrier2 is called within a render pass instance started with vkCmdBeginRendering, this command must only specify framebuffer-space stages in srcStageMask and dstStageMask", 187}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-srcStageMask-09673", 45}, { {"The srcStageMask member of any element of the pMemoryBarriers member of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 227}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-srcStageMask-09675", 45}, { {"If a buffer or image memory barrier does not specify an acquire operation, or if it does but pDependencyInfo->dependencyFlags includes VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR, the respective srcStageMask member of the element of the pBufferMemoryBarriers or pImageMemoryBarriers members of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 474}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-suspended", 36}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPipelineBarrier2-synchronization2-03848", 49}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-bufferlevel", 52}, { {"commandBuffer must be a primary VkCommandBuffer", 47}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-commandBuffer-11081", 60}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-commonparent", 53}, { {"Both of commandBuffer, and stateCommandBuffer must have been created, allocated, or retrieved from the same VkDevice", 116}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-deviceGeneratedCommands-11087", 70}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-indirectCommandsLayout-11084", 69}, { {"If the token sequence of the passed VkGeneratedCommandsInfoEXT::indirectCommandsLayout contains a VK_INDIRECT_COMMANDS_TOKEN_TYPE_EXECUTION_SET_EXT token, the initial shader state of VkGeneratedCommandsInfoEXT::indirectExecutionSet must be bound on stateCommandBuffer", 267}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-pGeneratedCommandsInfo-11082", 69}, { {"pGeneratedCommandsInfo's indirectCommandsLayout must have been created with the VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_EXT bit set", 149}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-pGeneratedCommandsInfo-parameter", 73}, { {"pGeneratedCommandsInfo must be a valid pointer to a valid VkGeneratedCommandsInfoEXT structure", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-renderpass", 51}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-stateCommandBuffer-11138", 65}, { {"stateCommandBuffer must be in the recording state", 49}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-stateCommandBuffer-parameter", 69}, { {"stateCommandBuffer must be a valid VkCommandBuffer handle", 57}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-supportedIndirectCommandsShaderStages-11088", 84}, { {"Only stages specified in VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT::supportedIndirectCommandsShaderStages can be set in pGeneratedCommandsInfo->shaderStages", 167}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-suspended", 50}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsEXT-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-commandBuffer-02974", 59}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-deviceGeneratedCommands-02928", 69}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-pGeneratedCommandsInfo-02927", 68}, { {"pGeneratedCommandsInfo->indirectCommandsLayout must have been created with the VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_NV bit set", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-pGeneratedCommandsInfo-parameter", 72}, { {"pGeneratedCommandsInfo must be a valid pointer to a valid VkGeneratedCommandsInfoNV structure", 93}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-renderpass", 50}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-suspended", 49}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPreprocessGeneratedCommandsNV-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCmdPushConstants-commandBuffer-11295", 43}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-commandBuffer-11296", 43}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-commonparent", 36}, { {"Both of commandBuffer, and layout must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-layout-parameter", 40}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-offset-00368", 36}, { {"offset must be a multiple of 4", 30}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-offset-00370", 36}, { {"offset must be less than VkPhysicalDeviceLimits::maxPushConstantsSize", 69}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-offset-01795", 36}, { {"For each byte in the range specified by offset and size and for each shader stage in stageFlags, there must be a push constant range in layout that includes that byte and that stage", 181}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-offset-01796", 36}, { {"For each byte in the range specified by offset and size and for each push constant range that overlaps that byte, stageFlags must include all stages in that push constant range's VkPushConstantRange::stageFlags", 210}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-pValues-parameter", 41}, { {"pValues must be a valid pointer to an array of size bytes", 57}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-size-00369", 34}, { {"size must be a multiple of 4", 28}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-size-00371", 34}, { {"size must be less than or equal to VkPhysicalDeviceLimits::maxPushConstantsSize minus offset", 92}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-size-arraylength", 40}, { {"size must be greater than 0", 27}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-stageFlags-parameter", 44}, { {"stageFlags must be a valid combination of VkShaderStageFlagBits values", 70}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-stageFlags-requiredbitmask", 50}, { {"stageFlags must not be 0", 24}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-commandBuffer-11295", 44}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-commandBuffer-11296", 44}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-pPushConstantsInfo-parameter", 53}, { {"pPushConstantsInfo must be a valid pointer to a valid VkPushConstantsInfo structure", 83}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushConstants2-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDataEXT-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdPushDataEXT-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdPushDataEXT-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdPushDataEXT-pPushDataInfo-parameter", 45}, { {"pPushDataInfo must be a valid pointer to a valid VkPushDataInfoEXT structure", 76}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdPushDataEXT-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkCmdPushDescriptorSet-None-10356", 38}, { {"If the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commandBuffer-11295", 47}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commandBuffer-11296", 47}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-commonparent", 40}, { {"Both of commandBuffer, and layout must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-descriptorWriteCount-arraylength", 60}, { {"descriptorWriteCount must be greater than 0", 43}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-layout-parameter", 44}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-pDescriptorWrites-06494", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, pDescriptorWrites[i].pImageInfo must be a valid pointer to an array of pDescriptorWrites[i].descriptorCount valid VkDescriptorImageInfo structures", 389}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-pDescriptorWrites-parameter", 55}, { {"pDescriptorWrites must be a valid pointer to an array of descriptorWriteCount valid VkWriteDescriptorSet structures", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-pipelineBindPoint-00363", 51}, { {"pipelineBindPoint must be supported by the commandBuffer's parent VkCommandPool's queue family", 94}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-pipelineBindPoint-parameter", 55}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-set-00364", 37}, { {"set must be less than VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-set-00365", 37}, { {"set must be the unique set number in the pipeline layout that uses a descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-None-10357", 39}, { {"If the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-commandBuffer-11295", 48}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-commandBuffer-11296", 48}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-pPushDescriptorSetInfo-09468", 57}, { {"Each bit in pPushDescriptorSetInfo->stageFlags must be a stage supported by the commandBuffer's parent VkCommandPool's queue family", 131}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-pPushDescriptorSetInfo-parameter", 61}, { {"pPushDescriptorSetInfo must be a valid pointer to a valid VkPushDescriptorSetInfo structure", 91}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSet2-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-None-10358", 50}, { {"If the VK_KHR_push_descriptor extension is not enabled, pushDescriptor must be enabled", 86}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-00366", 59}, { {"The pipelineBindPoint specified during the creation of the descriptor update template must be supported by the commandBuffer's parent VkCommandPool's queue family", 162}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-11295", 59}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-11296", 59}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-commonparent", 52}, { {"Each of commandBuffer, descriptorUpdateTemplate, and layout must have been created, allocated, or retrieved from the same VkDevice", 130}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-descriptorUpdateTemplate-07994", 70}, { {"descriptorUpdateTemplate must have been created with a templateType of VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS", 122}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-descriptorUpdateTemplate-parameter", 74}, { {"descriptorUpdateTemplate must be a valid VkDescriptorUpdateTemplate handle", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-layout-07993", 52}, { {"layout must be compatible with the layout used to create descriptorUpdateTemplate", 81}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-layout-parameter", 56}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-pData-01686", 51}, { {"pData must be a valid pointer to a memory containing one or more valid instances of VkDescriptorImageInfo, VkDescriptorBufferInfo, or VkBufferView in a layout defined by descriptorUpdateTemplate when it was created with vkCreateDescriptorUpdateTemplate", 252}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-set-07304", 49}, { {"set must be less than VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-set-07305", 49}, { {"set must be the unique set number in the pipeline layout that uses a descriptor set layout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT", 164}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-set-07995", 49}, { {"set must be the same value used to create descriptorUpdateTemplate", 66}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-set-11854", 49}, { {"set must reference a valid VkDescriptorSetLayout handle in layout", 65}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-commandBuffer-11295", 60}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-commandBuffer-11296", 60}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT operations", 126}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-pPushDescriptorSetWithTemplateInfo-parameter", 85}, { {"pPushDescriptorSetWithTemplateInfo must be a valid pointer to a valid VkPushDescriptorSetWithTemplateInfo structure", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdPushDescriptorSetWithTemplate2-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCmdResetEvent-commandBuffer-01157", 40}, { {"commandBuffer's current device mask must include exactly one physical device", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-commonparent", 33}, { {"Both of commandBuffer, and event must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-event-03834", 32}, { {"There must be an execution dependency between vkCmdResetEvent and the execution of any vkCmdWaitEvents that includes event in its pEvents parameter", 147}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-event-03835", 32}, { {"There must be an execution dependency between vkCmdResetEvent and the execution of any vkCmdWaitEvents2 that includes event in its pEvents parameter", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-event-parameter", 36}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-renderpass", 31}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-01153", 36}, { {"stageMask must not include VK_PIPELINE_STAGE_HOST_BIT", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-03937", 36}, { {"If the synchronization2 feature is not enabled, stageMask must not be 0", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04090", 36}, { {"If the geometryShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04091", 36}, { {"If the tessellationShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04092", 36}, { {"If the conditionalRendering feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 126}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04093", 36}, { {"If the fragmentDensityMap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 127}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04094", 36}, { {"If the transformFeedback feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 120}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04095", 36}, { {"If the meshShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-04096", 36}, { {"If the taskShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-06458", 36}, { {"Any pipeline stage included in stageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 316}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-07318", 36}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 179}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-07949", 36}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 161}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-10754", 36}, { {"If the accelerationStructure feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 134}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-stageMask-parameter", 40}, { {"stageMask must be a valid combination of VkPipelineStageFlagBits values", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-commandBuffer-03833", 41}, { {"commandBuffer's current device mask must include exactly one physical device", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-commonparent", 34}, { {"Both of commandBuffer, and event must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-event-03831", 33}, { {"There must be an execution dependency between vkCmdResetEvent2 and the execution of any vkCmdWaitEvents that includes event in its pEvents parameter", 148}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-event-03832", 33}, { {"There must be an execution dependency between vkCmdResetEvent2 and the execution of any vkCmdWaitEvents2 that includes event in its pEvents parameter", 149}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-event-parameter", 37}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-renderpass", 32}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03830", 37}, { {"stageMask must not include VK_PIPELINE_STAGE_2_HOST_BIT", 55}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03929", 37}, { {"If the geometryShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 112}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03930", 37}, { {"If the tessellationShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 186}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03931", 37}, { {"If the conditionalRendering feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 128}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03932", 37}, { {"If the fragmentDensityMap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03933", 37}, { {"If the transformFeedback feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 122}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03934", 37}, { {"If the meshShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-03935", 37}, { {"If the taskShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 108}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-04957", 37}, { {"If the subpassShading feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 118}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-04995", 37}, { {"If the invocationMask feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 119}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-07316", 37}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stageMask must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 181}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-07946", 37}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stageMask must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-10751", 37}, { {"If the accelerationStructure feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 136}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-10752", 37}, { {"If the rayTracingMaintenance1 feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 136}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-11541", 37}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 153}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-stageMask-parameter", 41}, { {"stageMask must be a valid combination of VkPipelineStageFlagBits2 values", 72}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetEvent2-synchronization2-03829", 44}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdResetQueryPool-None-02841", 35}, { {"All queries used by the command must not be active", 50}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 218}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-commonparent", 37}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-firstQuery-02862", 41}, { {"If queryPool was created with VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, this command must not be recorded in a command buffer that, either directly or through secondary command buffers, also contains begin commands for a query from the set of queries [firstQuery, firstQuery + queryCount - 1]", 287}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-firstQuery-09436", 41}, { {"firstQuery must be less than the number of queries in queryPool", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-firstQuery-09437", 41}, { {"The sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-queryPool-parameter", 44}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-renderpass", 35}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResetQueryPool-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-01837", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-01838", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-01839", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-commonparent", 35}, { {"Each of commandBuffer, dstImage, and srcImage must have been created, allocated, or retrieved from the same VkDevice", 116}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-00258", 37}, { {"If dstImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-00259", 37}, { {"dstImage must have a sample count equal to VK_SAMPLE_COUNT_1_BIT", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-00276", 37}, { {"If dstImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, dstOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-00278", 37}, { {"If dstImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, dstOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-02003", 37}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT if the maintenance10 feature is not enabled", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-02546", 37}, { {"dstImage and srcImage must not have been created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-06764", 37}, { {"dstImage must have been created with the VK_IMAGE_USAGE_TRANSFER_DST_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-06765", 37}, { {"The format features of dstImage must contain VK_FORMAT_FEATURE_TRANSFER_DST_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImage-parameter", 41}, { {"dstImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImageLayout-00262", 43}, { {"dstImageLayout must specify the layout of the image subresources of dstImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImageLayout-01401", 43}, { {"dstImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstImageLayout-parameter", 47}, { {"dstImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstOffset-00274", 38}, { {"For each element of pRegions, dstOffset.x and (extent.width + dstOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstOffset-00275", 38}, { {"For each element of pRegions, dstOffset.y and (extent.height + dstOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified dstSubresource of dstImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstOffset-00277", 38}, { {"For each element of pRegions, dstOffset.z and (extent.depth + dstOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified dstSubresource of dstImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstSubresource-01710", 43}, { {"The dstSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when dstImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstSubresource-01712", 43}, { {"If dstSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, dstSubresource.baseArrayLayer + dstSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when dstImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-dstSubresource-11801", 43}, { {"dstSubresource.aspectMask for each element in pRegions must not specify an aspect which is not part of the image format of dstImage", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-linearColorAttachment-06519", 50}, { {"If the linearColorAttachment feature is enabled and the image is created with VK_IMAGE_TILING_LINEAR, the format features of dstImage must contain VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV", 197}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-maintenance10-11799", 42}, { {"If the maintenance10 feature is enabled, the format features of dstImage must contain VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT or VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT", 174}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-pRegions-00255", 37}, { {"The union of all source regions, and the union of all destination regions, specified by the elements of pRegions, must not overlap in memory", 140}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-pRegions-parameter", 41}, { {"pRegions must be a valid pointer to an array of regionCount valid VkImageResolve structures", 91}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-regionCount-arraylength", 46}, { {"regionCount must be greater than 0", 34}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-renderpass", 33}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-00256", 37}, { {"If srcImage is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-00257", 37}, { {"srcImage must have a sample count equal to any valid sample count value other than VK_SAMPLE_COUNT_1_BIT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-00271", 37}, { {"If srcImage is of type VK_IMAGE_TYPE_1D, then for each element of pRegions, srcOffset.y must be 0 and extent.height must be 1", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-00273", 37}, { {"If srcImage is of type VK_IMAGE_TYPE_1D or VK_IMAGE_TYPE_2D, then for each element of pRegions, srcOffset.z must be 0 and extent.depth must be 1", 144}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-01386", 37}, { {"srcImage and dstImage must have been created with the same image format", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-04446", 37}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, srcSubresource.layerCount must be 1", 111}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-04447", 37}, { {"If dstImage is of type VK_IMAGE_TYPE_3D, then for each element of pRegions, dstSubresource.baseArrayLayer must be 0 and dstSubresource.layerCount must be 1", 155}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-06762", 37}, { {"srcImage must have been created with the VK_IMAGE_USAGE_TRANSFER_SRC_BIT usage flag set", 87}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-06763", 37}, { {"The format features of srcImage must contain VK_FORMAT_FEATURE_TRANSFER_SRC_BIT", 79}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImage-parameter", 41}, { {"srcImage must be a valid VkImage handle", 39}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImageLayout-00260", 43}, { {"srcImageLayout must specify the layout of the image subresources of srcImage specified in pRegions at the time this command is executed on a VkDevice", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImageLayout-01400", 43}, { {"srcImageLayout must be VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL or VK_IMAGE_LAYOUT_GENERAL", 122}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcImageLayout-parameter", 47}, { {"srcImageLayout must be a valid VkImageLayout value", 50}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcOffset-00269", 38}, { {"For each element of pRegions, srcOffset.x and (extent.width + srcOffset.x) must both be greater than or equal to 0 and less than or equal to the width of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcOffset-00270", 38}, { {"For each element of pRegions, srcOffset.y and (extent.height + srcOffset.y) must both be greater than or equal to 0 and less than or equal to the height of the specified srcSubresource of srcImage", 196}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcOffset-00272", 38}, { {"For each element of pRegions, srcOffset.z and (extent.depth + srcOffset.z) must both be greater than or equal to 0 and less than or equal to the depth of the specified srcSubresource of srcImage", 194}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcSubresource-01709", 43}, { {"The srcSubresource.mipLevel member of each element of pRegions must be less than the mipLevels specified in VkImageCreateInfo when srcImage was created", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcSubresource-01711", 43}, { {"If srcSubresource.layerCount is not VK_REMAINING_ARRAY_LAYERS, srcSubresource.baseArrayLayer + srcSubresource.layerCount of each element of pRegions must be less than or equal to the arrayLayers specified in VkImageCreateInfo when srcImage was created", 251}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcSubresource-11800", 43}, { {"srcSubresource.aspectMask for each element in pRegions must not specify an aspect which is not part of the image format of srcImage", 131}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-srcSubresource-11802", 43}, { {"srcSubresource.aspectMask must equal dstSubresource.aspectMask for each element in pRegions", 91}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-01837", 43}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, srcImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-01838", 43}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstImage must not be a protected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-01839", 43}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstImage must not be an unprotected image", 127}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-pResolveImageInfo-parameter", 51}, { {"pResolveImageInfo must be a valid pointer to a valid VkResolveImageInfo2 structure", 82}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-renderpass", 34}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-suspended", 33}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdResolveImage2-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdSetAlphaToCoverageEnableEXT-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3AlphaToCoverageEnable feature is enabled The shaderObject feature is enabled", 145}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToCoverageEnableEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToCoverageEnableEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToCoverageEnableEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToCoverageEnableEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-None-09423", 43}, { {"At least one of the following must be true: The extendedDynamicState3AlphaToOneEnable feature is enabled The shaderObject feature is enabled", 140}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-alphaToOne-07607", 49}, { {"If the alphaToOne feature is not enabled, alphaToOneEnable must be VK_FALSE", 75}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAlphaToOneEnableEXT-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-aspectMask-08863", 61}, { {"aspectMask must only include VK_IMAGE_ASPECT_NONE, VK_IMAGE_ASPECT_COLOR_BIT, VK_IMAGE_ASPECT_DEPTH_BIT, and VK_IMAGE_ASPECT_STENCIL_BIT", 136}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-aspectMask-parameter", 65}, { {"aspectMask must be a valid combination of VkImageAspectFlagBits values", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-attachmentFeedbackLoopDynamicState-08862", 85}, { {"The attachmentFeedbackLoopDynamicState feature must be enabled", 62}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-attachmentFeedbackLoopLayout-08864", 79}, { {"If the attachmentFeedbackLoopLayout feature is not enabled, aspectMask must be VK_IMAGE_ASPECT_NONE", 99}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-commandBuffer-cmdpool", 66}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-commandBuffer-parameter", 68}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-commandBuffer-recording", 68}, { {"commandBuffer must be in the recording state", 44}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetAttachmentFeedbackLoopEnableEXT-videocoding", 56}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCmdSetBlendConstants-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetBlendConstants-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetBlendConstants-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetBlendConstants-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetCheckpointNV-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdSetCheckpointNV-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdSetCheckpointNV-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdSetCheckpointNV-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdSetCheckpointNV-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-pCustomSampleOrders-02235", 58}, { {"The array pCustomSampleOrders must not contain two structures with matching values for both the shadingRate and sampleCount members", 131}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-pCustomSampleOrders-parameter", 62}, { {"If customSampleOrderCount is not 0, pCustomSampleOrders must be a valid pointer to an array of customSampleOrderCount valid VkCoarseSampleOrderCustomNV structures", 162}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-sampleOrderType-02081", 54}, { {"If sampleOrderType is not VK_COARSE_SAMPLE_ORDER_TYPE_CUSTOM_NV, customSamplerOrderCount must be 0", 98}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-sampleOrderType-parameter", 58}, { {"sampleOrderType must be a valid VkCoarseSampleOrderTypeNV value", 63}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoarseSampleOrderNV-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-None-09423", 45}, { {"At least one of the following must be true: The extendedDynamicState3ColorBlendAdvanced feature is enabled The shaderObject feature is enabled", 142}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-attachmentCount-arraylength", 62}, { {"attachmentCount must be greater than 0", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-pColorBlendAdvanced-parameter", 64}, { {"pColorBlendAdvanced must be a valid pointer to an array of attachmentCount valid VkColorBlendAdvancedEXT structures", 115}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendAdvancedEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-None-09423", 43}, { {"At least one of the following must be true: The extendedDynamicState3ColorBlendEnable feature is enabled The shaderObject feature is enabled", 140}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-attachmentCount-arraylength", 60}, { {"attachmentCount must be greater than 0", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-pColorBlendEnables-parameter", 61}, { {"pColorBlendEnables must be a valid pointer to an array of attachmentCount VkBool32 values", 89}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEnableEXT-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-None-09423", 45}, { {"At least one of the following must be true: The extendedDynamicState3ColorBlendEquation feature is enabled The shaderObject feature is enabled", 142}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-attachmentCount-arraylength", 62}, { {"attachmentCount must be greater than 0", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-pColorBlendEquations-parameter", 65}, { {"pColorBlendEquations must be a valid pointer to an array of attachmentCount valid VkColorBlendEquationEXT structures", 116}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorBlendEquationEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-None-04803", 43}, { {"The colorWriteEnable feature must be enabled", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-attachmentCount-06656", 54}, { {"attachmentCount must be less than or equal to the maxColorAttachments member of VkPhysicalDeviceLimits", 102}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-attachmentCount-arraylength", 60}, { {"attachmentCount must be greater than 0", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-pColorWriteEnables-parameter", 61}, { {"pColorWriteEnables must be a valid pointer to an array of attachmentCount VkBool32 values", 89}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteEnableEXT-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-None-09423", 41}, { {"At least one of the following must be true: The extendedDynamicState3ColorWriteMask feature is enabled The shaderObject feature is enabled", 138}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-attachmentCount-arraylength", 58}, { {"attachmentCount must be greater than 0", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-pColorWriteMasks-parameter", 57}, { {"pColorWriteMasks must be a valid pointer to an array of attachmentCount valid combinations of VkColorComponentFlagBits values", 125}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetColorWriteMaskEXT-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetComputeOccupancyPriorityNV-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetComputeOccupancyPriorityNV-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetComputeOccupancyPriorityNV-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetComputeOccupancyPriorityNV-pParameters-parameter", 61}, { {"pParameters must be a valid pointer to a valid VkComputeOccupancyPriorityParametersNV structure", 95}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetComputeOccupancyPriorityNV-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-None-09423", 56}, { {"At least one of the following must be true: The extendedDynamicState3ConservativeRasterizationMode feature is enabled The shaderObject feature is enabled", 153}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-commandBuffer-cmdpool", 67}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-commandBuffer-parameter", 69}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-commandBuffer-recording", 69}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-conservativeRasterizationMode-parameter", 85}, { {"conservativeRasterizationMode must be a valid VkConservativeRasterizationModeEXT value", 86}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetConservativeRasterizationModeEXT-videocoding", 57}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3CoverageModulationMode feature is enabled The shaderObject feature is enabled", 146}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-coverageModulationMode-parameter", 70}, { {"coverageModulationMode must be a valid VkCoverageModulationModeNV value", 71}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationModeNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableEnableNV-None-09423", 55}, { {"At least one of the following must be true: The extendedDynamicState3CoverageModulationTableEnable feature is enabled The shaderObject feature is enabled", 153}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableEnableNV-commandBuffer-cmdpool", 66}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableEnableNV-commandBuffer-parameter", 68}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableEnableNV-commandBuffer-recording", 68}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableEnableNV-videocoding", 56}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-None-09423", 49}, { {"At least one of the following must be true: The extendedDynamicState3CoverageModulationTable feature is enabled The shaderObject feature is enabled", 147}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-coverageModulationTableCount-arraylength", 79}, { {"coverageModulationTableCount must be greater than 0", 51}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-pCoverageModulationTable-parameter", 73}, { {"pCoverageModulationTable must be a valid pointer to an array of coverageModulationTableCount float values", 105}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageModulationTableNV-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-None-09423", 47}, { {"At least one of the following must be true: The extendedDynamicState3CoverageReductionMode feature is enabled The shaderObject feature is enabled", 145}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-coverageReductionMode-parameter", 68}, { {"coverageReductionMode must be a valid VkCoverageReductionModeNV value", 69}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageReductionModeNV-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorEnableNV-None-09423", 47}, { {"At least one of the following must be true: The extendedDynamicState3CoverageToColorEnable feature is enabled The shaderObject feature is enabled", 145}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorEnableNV-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorEnableNV-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorEnableNV-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorEnableNV-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorLocationNV-None-09423", 49}, { {"At least one of the following must be true: The extendedDynamicState3CoverageToColorLocation feature is enabled The shaderObject feature is enabled", 147}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorLocationNV-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorLocationNV-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorLocationNV-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCoverageToColorLocationNV-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetCullMode-None-08971", 32}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCullMode-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCullMode-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCullMode-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCullMode-cullMode-parameter", 40}, { {"cullMode must be a valid combination of VkCullModeFlagBits values", 65}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetCullMode-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias-depthBiasClamp-00790", 43}, { {"If the depthBiasClamp feature is not enabled, depthBiasClamp must be 0.0", 72}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias2EXT-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias2EXT-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias2EXT-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias2EXT-pDepthBiasInfo-parameter", 51}, { {"pDepthBiasInfo must be a valid pointer to a valid VkDepthBiasInfoEXT structure", 78}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBias2EXT-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBiasEnable-None-08970", 39}, { {"At least one of the following must be true: the extendedDynamicState2 feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 261}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBiasEnable-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBiasEnable-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBiasEnable-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBiasEnable-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetDepthBounds-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-maxDepthBounds-00601", 45}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled maxDepthBounds must be between 0.0 and 1.0, inclusive", 117}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-minDepthBounds-00600", 45}, { {"If the VK_EXT_depth_range_unrestricted extension is not enabled minDepthBounds must be between 0.0 and 1.0, inclusive", 117}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-minDepthBounds-10912", 45}, { {"minDepthBounds must be less than or equal to maxDepthBounds", 59}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBounds-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-None-08971", 45}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-depthBounds-10010", 52}, { {"If the depthBounds feature is not enabled, depthBoundsTestEnable must be VK_FALSE", 81}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthBoundsTestEnable-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-None-09423", 43}, { {"At least one of the following must be true: The extendedDynamicState3DepthClampEnable feature is enabled The shaderObject feature is enabled", 140}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-depthClamp-07449", 49}, { {"If the depthClamp feature is not enabled, depthClampEnable must be VK_FALSE", 75}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampEnableEXT-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-depthClampMode-parameter", 56}, { {"depthClampMode must be a valid VkDepthClampModeEXT value", 56}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-pDepthClampRange-09647", 54}, { {"If depthClampMode is VK_DEPTH_CLAMP_MODE_USER_DEFINED_RANGE_EXT, then pDepthClampRange must be a valid pointer to a valid VkDepthClampRangeEXT structure", 152}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-pDepthClampRange-parameter", 58}, { {"If pDepthClampRange is not NULL, pDepthClampRange must be a valid pointer to a valid VkDepthClampRangeEXT structure", 115}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClampRangeEXT-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-None-09423", 42}, { {"At least one of the following must be true: The extendedDynamicState3DepthClipEnable feature is enabled The shaderObject feature is enabled", 139}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-depthClipEnable-07451", 53}, { {"The depthClipEnable feature must be enabled", 43}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipEnableEXT-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-None-09423", 52}, { {"At least one of the following must be true: The extendedDynamicState3DepthClipNegativeOneToOne feature is enabled The shaderObject feature is enabled", 149}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-depthClipControl-07453", 64}, { {"The depthClipControl feature must be enabled", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthClipNegativeOneToOneEXT-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetDepthCompareOp-None-08971", 38}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthCompareOp-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthCompareOp-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthCompareOp-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthCompareOp-depthCompareOp-parameter", 52}, { {"depthCompareOp must be a valid VkCompareOp value", 48}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthCompareOp-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthTestEnable-None-08971", 39}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthTestEnable-commandBuffer-cmdpool", 50}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthTestEnable-commandBuffer-parameter", 52}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthTestEnable-commandBuffer-recording", 52}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthTestEnable-videocoding", 40}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthWriteEnable-None-08971", 40}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthWriteEnable-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthWriteEnable-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthWriteEnable-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDepthWriteEnable-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-commandBuffer-11295", 60}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-commandBuffer-11296", 60}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-descriptorBuffer-09470", 63}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-pSetDescriptorBufferOffsetsInfo-09471", 78}, { {"Each bit in pSetDescriptorBufferOffsetsInfo->stageFlags must be a stage supported by the commandBuffer's parent VkCommandPool's queue family", 140}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-pSetDescriptorBufferOffsetsInfo-parameter", 82}, { {"pSetDescriptorBufferOffsetsInfo must be a valid pointer to a valid VkSetDescriptorBufferOffsetsInfoEXT structure", 112}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsets2EXT-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-None-08060", 50}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commandBuffer-11295", 59}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pSamplerHeapBindInfo equal to NULL", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commandBuffer-11296", 59}, { {"If commandBuffer is a secondary command buffer, it must have begun with VkCommandBufferInheritanceDescriptorHeapInfoEXT::pResourceHeapBindInfo equal to NULL", 156}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commandBuffer-cmdpool", 61}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_DATA_GRAPH_BIT_ARM, or VK_QUEUE_GRAPHICS_BIT operations", 155}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commandBuffer-parameter", 63}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commandBuffer-recording", 63}, { {"commandBuffer must be in the recording state", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-commonparent", 52}, { {"Both of commandBuffer, and layout must have been created, allocated, or retrieved from the same VkDevice", 104}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-firstSet-08066", 54}, { {"The sum of firstSet and setCount must be less than or equal to VkPipelineLayoutCreateInfo::setLayoutCount provided when layout was created", 138}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-firstSet-09006", 54}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT bit set", 197}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-firstSet-11803", 54}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT_KHR bit set", 199}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-firstSet-11804", 54}, { {"The VkDescriptorSetLayout for each set from firstSet to firstSet + setCount when layout was created must not have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT bit set", 211}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-layout-parameter", 56}, { {"layout must be a valid VkPipelineLayout handle", 46}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pBufferIndices-08064", 60}, { {"Each element of pBufferIndices must be less than VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxDescriptorBufferBindings", 123}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pBufferIndices-08065", 60}, { {"Each element of pBufferIndices must reference a valid descriptor buffer binding set by a previous call to vkCmdBindDescriptorBuffersEXT in commandBuffer", 152}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pBufferIndices-parameter", 64}, { {"pBufferIndices must be a valid pointer to an array of setCount uint32_t values", 78}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pOffsets-08061", 54}, { {"The offsets in pOffsets must be aligned to VkPhysicalDeviceDescriptorBufferPropertiesEXT::descriptorBufferOffsetAlignment", 121}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pOffsets-08063", 54}, { {"The offsets in pOffsets must be small enough such that any descriptor binding referenced by layout without the VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT flag computes a valid address inside the underlying VkBuffer", 223}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pOffsets-08126", 54}, { {"The offsets in pOffsets must be small enough such that any location accessed by a shader as a sampler descriptor must be within VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxSamplerDescriptorBufferRange of the sampler descriptor buffer binding", 247}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pOffsets-08127", 54}, { {"The offsets in pOffsets must be small enough such that any location accessed by a shader as a resource descriptor must be within VkPhysicalDeviceDescriptorBufferPropertiesEXT::maxResourceDescriptorBufferRange of the resource descriptor buffer binding", 250}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pOffsets-parameter", 58}, { {"pOffsets must be a valid pointer to an array of setCount VkDeviceSize values", 76}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pipelineBindPoint-08067", 63}, { {"pipelineBindPoint must be supported by the commandBuffer's parent VkCommandPool's queue family", 94}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-pipelineBindPoint-parameter", 67}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-setCount-arraylength", 60}, { {"setCount must be greater than 0", 31}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDescriptorBufferOffsetsEXT-videocoding", 51}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkCmdSetDeviceMask-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-deviceMask-00108", 40}, { {"deviceMask must be a valid device mask value", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-deviceMask-00109", 40}, { {"deviceMask must not be zero", 27}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-deviceMask-00110", 40}, { {"deviceMask must not include any set bits that were not in the VkDeviceGroupCommandBufferBeginInfo::deviceMask value when the command buffer began recording", 155}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDeviceMask-deviceMask-00111", 40}, { {"If vkCmdSetDeviceMask is called inside a render pass instance, deviceMask must not include any set bits that were not in the VkDeviceGroupRenderPassBeginInfo::deviceMask value when the render pass instance began recording", 221}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-discardRectangleCount-arraylength", 66}, { {"discardRectangleCount must be greater than 0", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-firstDiscardRectangle-00585", 60}, { {"The sum of firstDiscardRectangle and discardRectangleCount must be less than or equal to VkPhysicalDeviceDiscardRectanglePropertiesEXT::maxDiscardRectangles", 156}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-offset-00588", 45}, { {"Evaluation of (offset.x + extent.width) in each VkRect2D element of pDiscardRectangles must not cause a signed integer addition overflow", 136}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-offset-00589", 45}, { {"Evaluation of (offset.y + extent.height) in each VkRect2D element of pDiscardRectangles must not cause a signed integer addition overflow", 137}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-pDiscardRectangles-parameter", 61}, { {"pDiscardRectangles must be a valid pointer to an array of discardRectangleCount VkRect2D structures", 99}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-viewportScissor2D-04788", 56}, { {"If this command is recorded in a secondary command buffer with VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled, then this function must not be called", 176}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEXT-x-00587", 40}, { {"The x and y member of offset in each VkRect2D element of pDiscardRectangles must be greater than or equal to 0", 110}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEnableEXT-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEnableEXT-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEnableEXT-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEnableEXT-specVersion-07851", 56}, { {"The VK_EXT_discard_rectangles extension must be enabled, and the implementation must support at least specVersion 2 of this extension", 133}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleEnableEXT-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-discardRectangleMode-parameter", 67}, { {"discardRectangleMode must be a valid VkDiscardRectangleModeEXT value", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-specVersion-07852", 54}, { {"The VK_EXT_discard_rectangles extension must be enabled, and the implementation must support at least specVersion 2 of this extension", 133}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDiscardRectangleModeEXT-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetDispatchParametersARM-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-pDispatchParameters-parameter", 64}, { {"pDispatchParameters must be a valid pointer to a valid VkDispatchParametersARM structure", 88}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-renderpass", 45}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-schedulingControlsFlags-12391", 64}, { {"VkPhysicalDeviceSchedulingControlsPropertiesARM::schedulingControlsFlags must contain VK_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_DISPATCH_PARAMETERS_ARM", 148}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetDispatchParametersARM-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSetEvent-commandBuffer-01152", 38}, { {"The current device mask of commandBuffer must include exactly one physical device", 81}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-commandBuffer-cmdpool", 40}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-commandBuffer-parameter", 42}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-commandBuffer-recording", 42}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-commonparent", 31}, { {"Both of commandBuffer, and event must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-event-parameter", 34}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-renderpass", 29}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-01149", 34}, { {"stageMask must not include VK_PIPELINE_STAGE_HOST_BIT", 53}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-03937", 34}, { {"If the synchronization2 feature is not enabled, stageMask must not be 0", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04090", 34}, { {"If the geometryShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 110}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04091", 34}, { {"If the tessellationShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04092", 34}, { {"If the conditionalRendering feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 126}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04093", 34}, { {"If the fragmentDensityMap feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 127}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04094", 34}, { {"If the transformFeedback feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 120}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04095", 34}, { {"If the meshShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-04096", 34}, { {"If the taskShader feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-06457", 34}, { {"Any pipeline stage included in stageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 316}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-07318", 34}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 179}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-07949", 34}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 161}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-10754", 34}, { {"If the accelerationStructure feature is not enabled, stageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 134}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-stageMask-parameter", 38}, { {"stageMask must be a valid combination of VkPipelineStageFlagBits values", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent-suspended", 28}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-commandBuffer-03826", 39}, { {"The current device mask of commandBuffer must include exactly one physical device", 81}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-commandBuffer-cmdpool", 41}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-commandBuffer-parameter", 43}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-commandBuffer-recording", 43}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-commonparent", 32}, { {"Both of commandBuffer, and event must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dependencyFlags-03825", 41}, { {"The dependencyFlags member of pDependencyInfo must be 0 or VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR", 97}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dependencyFlags-10785", 41}, { {"If the dependencyFlags member of pDependencyInfo includes VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, the bufferMemoryBarrierCount and imageMemoryBarrierCount members of pDependencyInfo must be 0", 191}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dependencyFlags-10786", 41}, { {"If the dependencyFlags member of pDependencyInfo includes VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, the memoryBarrierCount member of pDependencyInfo must be 1", 156}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dependencyFlags-10787", 41}, { {"If the dependencyFlags member of pDependencyInfo includes VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, the srcAccessMask, dstStageMask, and dstAccessMask members of pDependencyInfo->pMemoryBarriers[0] must be 0", 205}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dstStageMask-03828", 38}, { {"The dstStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 276}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-dstStageMask-09392", 38}, { {"The dstStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfo must not include VK_PIPELINE_STAGE_2_HOST_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-event-parameter", 35}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-pDependencyInfo-parameter", 45}, { {"pDependencyInfo must be a valid pointer to a valid VkDependencyInfo structure", 77}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-renderpass", 30}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-srcStageMask-03827", 38}, { {"The srcStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfo must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 276}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-srcStageMask-09391", 38}, { {"The srcStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfo must not include VK_PIPELINE_STAGE_2_HOST_BIT", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-suspended", 29}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetEvent2-synchronization2-03824", 42}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-exclusiveScissor-07853", 60}, { {"The exclusiveScissor feature must be enabled, and the implementation must support at least specVersion 2 of the VK_NV_scissor_exclusive extension", 145}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-exclusiveScissorCount-arraylength", 71}, { {"exclusiveScissorCount must be greater than 0", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-pExclusiveScissorEnables-parameter", 72}, { {"pExclusiveScissorEnables must be a valid pointer to an array of exclusiveScissorCount VkBool32 values", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorEnableNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-None-02031", 42}, { {"The exclusiveScissor feature must be enabled", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-exclusiveScissorCount-02036", 59}, { {"If the multiViewport feature is not enabled, exclusiveScissorCount must be 1", 76}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-exclusiveScissorCount-arraylength", 65}, { {"exclusiveScissorCount must be greater than 0", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-firstExclusiveScissor-02034", 59}, { {"The sum of firstExclusiveScissor and exclusiveScissorCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 128}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-firstExclusiveScissor-02035", 59}, { {"If the multiViewport feature is not enabled, firstExclusiveScissor must be 0", 76}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-offset-02038", 44}, { {"Evaluation of (offset.x + extent.width) for each member of pExclusiveScissors must not cause a signed integer addition overflow", 127}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-offset-02039", 44}, { {"Evaluation of (offset.y + extent.height) for each member of pExclusiveScissors must not cause a signed integer addition overflow", 128}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-pExclusiveScissors-parameter", 60}, { {"pExclusiveScissors must be a valid pointer to an array of exclusiveScissorCount VkRect2D structures", 99}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExclusiveScissorNV-x-02037", 39}, { {"The x and y members of offset in each member of pExclusiveScissors must be greater than or equal to 0", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-None-09423", 59}, { {"At least one of the following must be true: The extendedDynamicState3ExtraPrimitiveOverestimationSize feature is enabled The shaderObject feature is enabled", 156}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-commandBuffer-cmdpool", 70}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-commandBuffer-parameter", 72}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-commandBuffer-recording", 72}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-extraPrimitiveOverestimationSize-07428", 87}, { {"extraPrimitiveOverestimationSize must be in the range of 0.0 to VkPhysicalDeviceConservativeRasterizationPropertiesEXT::maxExtraPrimitiveOverestimationSize inclusive", 165}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetExtraPrimitiveOverestimationSizeEXT-videocoding", 60}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-attachmentFragmentShadingRate-04582", 74}, { {"If the attachmentFragmentShadingRate feature is not enabled, combinerOps[1] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 129}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-combinerOps-parameter", 60}, { {"Each element of combinerOps must be a valid VkFragmentShadingRateCombinerOpKHR value", 84}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-commandBuffer-cmdpool", 60}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-commandBuffer-parameter", 62}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-commandBuffer-recording", 62}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-fragmentShadingRateEnums-04579", 69}, { {"The fragmentShadingRateEnums feature must be enabled", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-fragmentSizeNonTrivialCombinerOps-04583", 78}, { {"If the fragmentSizeNonTrivialCombinerOps limit is not supported, elements of combinerOps must be either VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR or VK_FRAGMENT_SHADING_RATE_COMBINER_OP_REPLACE_KHR", 201}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-noInvocationFragmentShadingRates-04578", 77}, { {"If the noInvocationFragmentShadingRates feature is not enabled, shadingRate must not be VK_FRAGMENT_SHADING_RATE_NO_INVOCATIONS_NV", 130}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-pipelineFragmentShadingRate-04576", 72}, { {"If the pipelineFragmentShadingRate feature is not enabled, shadingRate must be VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_PIXEL_NV", 129}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-pipelineFragmentShadingRate-04580", 72}, { {"One of the pipelineFragmentShadingRate, primitiveFragmentShadingRate, or attachmentFragmentShadingRate features must be enabled", 127}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-primitiveFragmentShadingRate-04581", 73}, { {"If the primitiveFragmentShadingRate feature is not enabled, combinerOps[0] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 128}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-shadingRate-parameter", 60}, { {"shadingRate must be a valid VkFragmentShadingRateNV value", 57}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-supersampleFragmentShadingRates-04577", 76}, { {"If the supersampleFragmentShadingRates feature is not enabled, shadingRate must not be VK_FRAGMENT_SHADING_RATE_2_INVOCATIONS_PER_PIXEL_NV, VK_FRAGMENT_SHADING_RATE_4_INVOCATIONS_PER_PIXEL_NV, VK_FRAGMENT_SHADING_RATE_8_INVOCATIONS_PER_PIXEL_NV, or VK_FRAGMENT_SHADING_RATE_16_INVOCATIONS_PER_PIXEL_NV", 301}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateEnumNV-videocoding", 50}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-attachmentFragmentShadingRate-04511", 71}, { {"If the attachmentFragmentShadingRate feature is not enabled, combinerOps[1] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 129}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-combinerOps-parameter", 57}, { {"Each element of combinerOps must be a valid VkFragmentShadingRateCombinerOpKHR value", 84}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-fragmentSizeNonTrivialCombinerOps-04512", 75}, { {"If the fragmentSizeNonTrivialCombinerOps limit is not supported, elements of combinerOps must be either VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR or VK_FRAGMENT_SHADING_RATE_COMBINER_OP_REPLACE_KHR", 201}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04513", 55}, { {"pFragmentSize->width must be greater than or equal to 1", 55}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04514", 55}, { {"pFragmentSize->height must be greater than or equal to 1", 56}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04515", 55}, { {"pFragmentSize->width must be a power-of-two value", 49}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04516", 55}, { {"pFragmentSize->height must be a power-of-two value", 50}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04517", 55}, { {"pFragmentSize->width must be less than or equal to 4", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-04518", 55}, { {"pFragmentSize->height must be less than or equal to 4", 53}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pFragmentSize-parameter", 59}, { {"pFragmentSize must be a valid pointer to a valid VkExtent2D structure", 69}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pipelineFragmentShadingRate-04507", 69}, { {"If the pipelineFragmentShadingRate feature is not enabled, pFragmentSize->width must be 1", 89}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pipelineFragmentShadingRate-04508", 69}, { {"If the pipelineFragmentShadingRate feature is not enabled, pFragmentSize->height must be 1", 90}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-pipelineFragmentShadingRate-04509", 69}, { {"One of the pipelineFragmentShadingRate, primitiveFragmentShadingRate, or attachmentFragmentShadingRate features must be enabled", 127}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-primitiveFragmentShadingRate-04510", 70}, { {"If the primitiveFragmentShadingRate feature is not enabled, combinerOps[0] must be VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR", 128}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFragmentShadingRateKHR-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-None-08971", 33}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-frontFace-parameter", 42}, { {"frontFace must be a valid VkFrontFace value", 43}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetFrontFace-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3LineRasterizationMode feature is enabled The shaderObject feature is enabled", 145}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-lineRasterizationMode-07418", 65}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR, then the rectangularLines feature must be enabled", 117}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-lineRasterizationMode-07419", 65}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_BRESENHAM, then the bresenhamLines feature must be enabled", 113}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-lineRasterizationMode-07420", 65}, { {"If lineRasterizationMode is VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH, then the smoothLines feature must be enabled", 119}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-lineRasterizationMode-parameter", 69}, { {"lineRasterizationMode must be a valid VkLineRasterizationModeEXT value", 70}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineRasterizationModeEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStipple-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStipple-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStipple-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStipple-lineStippleFactor-02776", 48}, { {"lineStippleFactor must be in the range [1,256]", 46}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStipple-videocoding", 36}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStippleEnableEXT-None-09423", 44}, { {"At least one of the following must be true: The extendedDynamicState3LineStippleEnable feature is enabled The shaderObject feature is enabled", 141}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStippleEnableEXT-commandBuffer-cmdpool", 55}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStippleEnableEXT-commandBuffer-parameter", 57}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStippleEnableEXT-commandBuffer-recording", 57}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineStippleEnableEXT-videocoding", 45}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineWidth-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineWidth-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineWidth-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineWidth-lineWidth-00788", 38}, { {"If the wideLines feature is not enabled, lineWidth must be 1.0", 62}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLineWidth-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetLogicOpEXT-None-09422", 34}, { {"At least one of the following must be true: The extendedDynamicState2LogicOp feature is enabled The shaderObject feature is enabled", 131}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEXT-commandBuffer-cmdpool", 45}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEXT-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEXT-commandBuffer-recording", 47}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEXT-logicOp-parameter", 41}, { {"logicOp must be a valid VkLogicOp value", 39}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEXT-videocoding", 35}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-None-09423", 40}, { {"At least one of the following must be true: The extendedDynamicState3LogicOpEnable feature is enabled The shaderObject feature is enabled", 137}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-logicOp-07366", 43}, { {"If the logicOp feature is not enabled, logicOpEnable must be VK_FALSE", 69}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetLogicOpEnableEXT-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-None-09422", 45}, { {"At least one of the following must be true: The extendedDynamicState2PatchControlPoints feature is enabled The shaderObject feature is enabled", 142}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-commandBuffer-cmdpool", 56}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-commandBuffer-parameter", 58}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-commandBuffer-recording", 58}, { {"commandBuffer must be in the recording state", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-patchControlPoints-04874", 59}, { {"patchControlPoints must be greater than zero and less than or equal to VkPhysicalDeviceLimits::maxTessellationPatchSize", 119}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPatchControlPointsEXT-videocoding", 46}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-pMarkerInfo-parameter", 57}, { {"pMarkerInfo must be a valid pointer to a valid VkPerformanceMarkerInfoINTEL structure", 85}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-suspended", 45}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceMarkerINTEL-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-pOverrideInfo-02736", 57}, { {"pOverrideInfo must not be used with a VkPerformanceOverrideTypeINTEL that is not reported available by vkGetPerformanceParameterINTEL", 133}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-pOverrideInfo-parameter", 61}, { {"pOverrideInfo must be a valid pointer to a valid VkPerformanceOverrideInfoINTEL structure", 89}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceOverrideINTEL-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-pMarkerInfo-parameter", 63}, { {"pMarkerInfo must be a valid pointer to a valid VkPerformanceStreamMarkerInfoINTEL structure", 91}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-suspended", 51}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPerformanceStreamMarkerINTEL-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdSetPolygonModeEXT-None-09423", 38}, { {"At least one of the following must be true: The extendedDynamicState3PolygonMode feature is enabled The shaderObject feature is enabled", 135}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-fillModeNonSolid-07424", 50}, { {"If the fillModeNonSolid feature is not enabled, polygonMode must be VK_POLYGON_MODE_FILL or VK_POLYGON_MODE_FILL_RECTANGLE_NV", 125}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-polygonMode-07425", 45}, { {"If the VK_NV_fill_rectangle extension is not enabled, polygonMode must not be VK_POLYGON_MODE_FILL_RECTANGLE_NV", 111}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-polygonMode-parameter", 49}, { {"polygonMode must be a valid VkPolygonMode value", 47}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPolygonModeEXT-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetPrimitiveRestartEnable-None-08970", 46}, { {"At least one of the following must be true: the extendedDynamicState2 feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 261}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartEnable-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartEnable-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartEnable-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartEnable-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartIndexEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartIndexEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartIndexEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartIndexEXT-primitiveRestartIndex-12395", 65}, { {"The primitiveRestartIndex feature must be enabled", 49}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveRestartIndexEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-None-08971", 41}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-primitiveTopology-parameter", 58}, { {"primitiveTopology must be a valid VkPrimitiveTopology value", 59}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetPrimitiveTopology-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/drawing.html", 21} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-None-09423", 46}, { {"At least one of the following must be true: The extendedDynamicState3ProvokingVertexMode feature is enabled The shaderObject feature is enabled", 143}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-provokingVertexMode-07447", 61}, { {"If provokingVertexMode is VK_PROVOKING_VERTEX_MODE_LAST_VERTEX_EXT, then the provokingVertexLast feature must be enabled", 120}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-provokingVertexMode-parameter", 65}, { {"provokingVertexMode must be a valid VkProvokingVertexModeEXT value", 66}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetProvokingVertexModeEXT-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-None-09423", 47}, { {"At least one of the following must be true: The extendedDynamicState3RasterizationSamples feature is enabled The shaderObject feature is enabled", 144}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-rasterizationSamples-parameter", 67}, { {"rasterizationSamples must be a valid VkSampleCountFlagBits value", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationSamplesEXT-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-None-09423", 46}, { {"At least one of the following must be true: The extendedDynamicState3RasterizationStream feature is enabled The shaderObject feature is enabled", 143}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-commandBuffer-cmdpool", 57}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-commandBuffer-parameter", 59}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-commandBuffer-recording", 59}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-rasterizationStream-07412", 61}, { {"rasterizationStream must be less than VkPhysicalDeviceTransformFeedbackPropertiesEXT::maxTransformFeedbackStreams", 113}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-rasterizationStream-07413", 61}, { {"rasterizationStream must be zero if VkPhysicalDeviceTransformFeedbackPropertiesEXT::transformFeedbackRasterizationStreamSelect is VK_FALSE", 138}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-transformFeedback-07411", 59}, { {"The transformFeedback feature must be enabled", 45}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizationStreamEXT-videocoding", 47}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizerDiscardEnable-None-08970", 47}, { {"At least one of the following must be true: the extendedDynamicState2 feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 261}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizerDiscardEnable-commandBuffer-cmdpool", 58}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizerDiscardEnable-commandBuffer-parameter", 60}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizerDiscardEnable-commandBuffer-recording", 60}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRasterizerDiscardEnable-videocoding", 48}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-commandBuffer-cmdpool", 65}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-commandBuffer-parameter", 67}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-commandBuffer-recording", 67}, { {"commandBuffer must be in the recording state", 44}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-pipelineStackSize-03610", 67}, { {"pipelineStackSize must be large enough for any dynamic execution through the shaders in the ray tracing pipeline used by a subsequent trace call", 144}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-renderpass", 54}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRayTracingPipelineStackSizeKHR-videocoding", 55}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-commandBuffer-09511", 61}, { {"The current render pass instance must have been started or resumed by vkCmdBeginRendering in this commandBuffer", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-dynamicRenderingLocalRead-09509", 73}, { {"dynamicRenderingLocalRead must be enabled", 41}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-pLocationInfo-09510", 61}, { {"pLocationInfo->colorAttachmentCount must be equal to the value of VkRenderingInfo::colorAttachmentCount used to begin the current render pass instance", 150}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-pLocationInfo-parameter", 65}, { {"pLocationInfo must be a valid pointer to a valid VkRenderingAttachmentLocationInfo structure", 92}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-renderpass", 52}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingAttachmentLocations-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-commandBuffer-09518", 64}, { {"The current render pass instance must have been started or resumed by vkCmdBeginRendering in this commandBuffer", 111}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-commandBuffer-cmdpool", 66}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-commandBuffer-parameter", 68}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-commandBuffer-recording", 68}, { {"commandBuffer must be in the recording state", 44}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-dynamicRenderingLocalRead-09516", 76}, { {"dynamicRenderingLocalRead must be enabled", 41}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-pInputAttachmentIndexInfo-09517", 76}, { {"pInputAttachmentIndexInfo->colorAttachmentCount must be equal to the value of VkRenderingInfo::colorAttachmentCount used to begin the current render pass instance", 162}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-pInputAttachmentIndexInfo-parameter", 80}, { {"pInputAttachmentIndexInfo must be a valid pointer to a valid VkRenderingInputAttachmentIndexInfo structure", 106}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-renderpass", 55}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRenderingInputAttachmentIndices-videocoding", 56}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/interfaces.html", 24} } },
        { {"VUID-vkCmdSetRepresentativeFragmentTestEnableNV-None-09423", 58}, { {"At least one of the following must be true: The extendedDynamicState3RepresentativeFragmentTestEnable feature is enabled The shaderObject feature is enabled", 156}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetRepresentativeFragmentTestEnableNV-commandBuffer-cmdpool", 69}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetRepresentativeFragmentTestEnableNV-commandBuffer-parameter", 71}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetRepresentativeFragmentTestEnableNV-commandBuffer-recording", 71}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetRepresentativeFragmentTestEnableNV-videocoding", 59}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-pSampleLocationsInfo-parameter", 62}, { {"pSampleLocationsInfo must be a valid pointer to a valid VkSampleLocationsInfoEXT structure", 90}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-variableSampleLocations-01530", 61}, { {"If VkPhysicalDeviceSampleLocationsPropertiesEXT::variableSampleLocations is VK_FALSE then the current render pass must have been begun by specifying a VkRenderPassSampleLocationsBeginInfoEXT structure whose pPostSubpassSampleLocations member contains an element with a subpassIndex matching the current subpass index and the sampleLocationsInfo member of that element must match the sample locations state pointed to by pSampleLocationsInfo", 440}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEXT-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEnableEXT-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3SampleLocationsEnable feature is enabled The shaderObject feature is enabled", 145}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEnableEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEnableEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEnableEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleLocationsEnableEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetSampleMaskEXT-None-09423", 37}, { {"At least one of the following must be true: The extendedDynamicState3SampleMask feature is enabled The shaderObject feature is enabled", 134}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-commandBuffer-cmdpool", 48}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-commandBuffer-parameter", 50}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-commandBuffer-recording", 50}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-pSampleMask-10999", 44}, { {"pSampleMask must not be NULL if the maintenance10 feature is not enabled", 72}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-pSampleMask-parameter", 48}, { {"If pSampleMask is not NULL, pSampleMask must be a valid pointer to an array of \\(\\lceil{\\mathit{samples} \\over 32}\\rceil\\) VkSampleMask values", 148}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-samples-parameter", 44}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetSampleMaskEXT-videocoding", 38}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-firstScissor-00592", 39}, { {"The sum of firstScissor and scissorCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 110}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-firstScissor-00593", 39}, { {"If the multiViewport feature is not enabled, firstScissor must be 0", 67}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-offset-00596", 33}, { {"Evaluation of (offset.x + extent.width) must not cause a signed integer addition overflow for any element of pScissors", 118}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-offset-00597", 33}, { {"Evaluation of (offset.y + extent.height) must not cause a signed integer addition overflow for any element of pScissors", 119}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-pScissors-parameter", 40}, { {"pScissors must be a valid pointer to an array of scissorCount VkRect2D structures", 81}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-scissorCount-00594", 39}, { {"If the multiViewport feature is not enabled, scissorCount must be 1", 67}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-scissorCount-arraylength", 45}, { {"scissorCount must be greater than 0", 35}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-videocoding", 32}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-viewportScissor2D-04789", 44}, { {"If this command is recorded in a secondary command buffer with VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled, then this function must not be called", 176}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissor-x-00595", 28}, { {"The x and y members of offset member of any element of pScissors must be greater than or equal to 0", 99}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetScissorWithCount-None-08971", 40}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-commandBuffer-04820", 49}, { {"commandBuffer must not have VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled", 102}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-offset-03400", 42}, { {"Evaluation of (offset.x + extent.width) must not cause a signed integer addition overflow for any element of pScissors", 118}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-offset-03401", 42}, { {"Evaluation of (offset.y + extent.height) must not cause a signed integer addition overflow for any element of pScissors", 119}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-pScissors-parameter", 49}, { {"pScissors must be a valid pointer to an array of scissorCount VkRect2D structures", 81}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-scissorCount-03397", 48}, { {"scissorCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 82}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-scissorCount-03398", 48}, { {"If the multiViewport feature is not enabled, scissorCount must be 1", 67}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-scissorCount-arraylength", 54}, { {"scissorCount must be greater than 0", 35}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetScissorWithCount-x-03399", 37}, { {"The x and y members of offset member of any element of pScissors must be greater than or equal to 0", 99}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetShadingRateImageEnableNV-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3ShadingRateImageEnable feature is enabled The shaderObject feature is enabled", 146}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetShadingRateImageEnableNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetShadingRateImageEnableNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetShadingRateImageEnableNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetShadingRateImageEnableNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetStencilCompareMask-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilCompareMask-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilCompareMask-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilCompareMask-faceMask-parameter", 50}, { {"faceMask must be a valid combination of VkStencilFaceFlagBits values", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilCompareMask-faceMask-requiredbitmask", 56}, { {"faceMask must not be 0", 22}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilCompareMask-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-None-08971", 33}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-compareOp-parameter", 42}, { {"compareOp must be a valid VkCompareOp value", 43}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-depthFailOp-parameter", 44}, { {"depthFailOp must be a valid VkStencilOp value", 45}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-faceMask-parameter", 41}, { {"faceMask must be a valid combination of VkStencilFaceFlagBits values", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-faceMask-requiredbitmask", 47}, { {"faceMask must not be 0", 22}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-failOp-parameter", 39}, { {"failOp must be a valid VkStencilOp value", 40}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-passOp-parameter", 39}, { {"passOp must be a valid VkStencilOp value", 40}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilOp-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-faceMask-parameter", 48}, { {"faceMask must be a valid combination of VkStencilFaceFlagBits values", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-faceMask-requiredbitmask", 54}, { {"faceMask must not be 0", 22}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilReference-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilTestEnable-None-08971", 41}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilTestEnable-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilTestEnable-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilTestEnable-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilTestEnable-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-commandBuffer-cmdpool", 51}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-commandBuffer-parameter", 53}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-commandBuffer-recording", 53}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-faceMask-parameter", 48}, { {"faceMask must be a valid combination of VkStencilFaceFlagBits values", 68}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-faceMask-requiredbitmask", 54}, { {"faceMask must not be 0", 22}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetStencilWriteMask-videocoding", 41}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-None-09423", 51}, { {"At least one of the following must be true: The extendedDynamicState3TessellationDomainOrigin feature is enabled The shaderObject feature is enabled", 148}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-domainOrigin-parameter", 63}, { {"domainOrigin must be a valid VkTessellationDomainOrigin value", 61}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetTessellationDomainOriginEXT-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/tessellation.html", 26} } },
        { {"VUID-vkCmdSetVertexInputEXT-None-08546", 38}, { {"Either the vertexInputDynamicState feature or the shaderObject feature or both must be enabled", 94}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-binding-04793", 41}, { {"For every binding specified by each element of pVertexAttributeDescriptions, a VkVertexInputBindingDescription2EXT must exist in pVertexBindingDescriptions with the same value of binding", 186}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-commandBuffer-cmdpool", 49}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-commandBuffer-parameter", 51}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-commandBuffer-recording", 51}, { {"commandBuffer must be in the recording state", 44}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-pVertexAttributeDescriptions-04795", 62}, { {"All elements of pVertexAttributeDescriptions must describe distinct attribute locations", 87}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-pVertexAttributeDescriptions-parameter", 66}, { {"If vertexAttributeDescriptionCount is not 0, pVertexAttributeDescriptions must be a valid pointer to an array of vertexAttributeDescriptionCount valid VkVertexInputAttributeDescription2EXT structures", 199}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-pVertexBindingDescriptions-04794", 60}, { {"All elements of pVertexBindingDescriptions must describe distinct binding numbers", 81}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-pVertexBindingDescriptions-parameter", 64}, { {"If vertexBindingDescriptionCount is not 0, pVertexBindingDescriptions must be a valid pointer to an array of vertexBindingDescriptionCount valid VkVertexInputBindingDescription2EXT structures", 191}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-vertexAttributeDescriptionCount-04792", 65}, { {"vertexAttributeDescriptionCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputAttributes", 110}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-vertexBindingDescriptionCount-04791", 63}, { {"vertexBindingDescriptionCount must be less than or equal to VkPhysicalDeviceLimits::maxVertexInputBindings", 106}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetVertexInputEXT-videocoding", 39}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/fxvertex.html", 22} } },
        { {"VUID-vkCmdSetViewport-commandBuffer-04821", 41}, { {"commandBuffer must not have VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled", 102}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-firstViewport-01223", 41}, { {"The sum of firstViewport and viewportCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-firstViewport-01224", 41}, { {"If the multiViewport feature is not enabled, firstViewport must be 0", 68}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-pViewports-parameter", 42}, { {"pViewports must be a valid pointer to an array of viewportCount valid VkViewport structures", 91}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-viewportCount-01225", 41}, { {"If the multiViewport feature is not enabled, viewportCount must be 1", 68}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewport-viewportCount-arraylength", 47}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-None-02064", 52}, { {"The shadingRateImage feature must be enabled", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-commandBuffer-cmdpool", 63}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-commandBuffer-parameter", 65}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-commandBuffer-recording", 65}, { {"commandBuffer must be in the recording state", 44}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-firstViewport-02067", 61}, { {"The sum of firstViewport and viewportCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 112}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-firstViewport-02068", 61}, { {"If the multiViewport feature is not enabled, firstViewport must be 0", 68}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-pShadingRatePalettes-parameter", 72}, { {"pShadingRatePalettes must be a valid pointer to an array of viewportCount valid VkShadingRatePaletteNV structures", 113}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-videocoding", 53}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-viewportCount-02069", 61}, { {"If the multiViewport feature is not enabled, viewportCount must be 1", 68}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportShadingRatePaletteNV-viewportCount-arraylength", 67}, { {"viewportCount must be greater than 0", 36}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-None-09423", 41}, { {"At least one of the following must be true: The extendedDynamicState3ViewportSwizzle feature is enabled The shaderObject feature is enabled", 139}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-pViewportSwizzles-parameter", 58}, { {"pViewportSwizzles must be a valid pointer to an array of viewportCount valid VkViewportSwizzleNV structures", 107}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportSwizzleNV-viewportCount-arraylength", 56}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingEnableNV-None-09423", 48}, { {"At least one of the following must be true: The extendedDynamicState3ViewportWScalingEnable feature is enabled The shaderObject feature is enabled", 146}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingEnableNV-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingEnableNV-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingEnableNV-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingEnableNV-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-firstViewport-01324", 51}, { {"The sum of firstViewport and viewportCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 112}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-pViewportWScalings-parameter", 60}, { {"pViewportWScalings must be a valid pointer to an array of viewportCount VkViewportWScalingNV structures", 103}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWScalingNV-viewportCount-arraylength", 57}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-None-08971", 41}, { {"At least one of the following must be true: the extendedDynamicState feature is enabled the shaderObject feature is enabled the value of VkApplicationInfo::apiVersion used to create the VkInstance parent of commandBuffer is greater than or equal to Version 1.3", 260}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-commandBuffer-04819", 50}, { {"commandBuffer must not have VkCommandBufferInheritanceViewportScissorInfoNV::viewportScissor2D enabled", 102}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-pViewports-parameter", 51}, { {"pViewports must be a valid pointer to an array of viewportCount valid VkViewport structures", 91}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-viewportCount-03394", 50}, { {"viewportCount must be between 1 and VkPhysicalDeviceLimits::maxViewports, inclusive", 83}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-viewportCount-03395", 50}, { {"If the multiViewport feature is not enabled, viewportCount must be 1", 68}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSetViewportWithCount-viewportCount-arraylength", 56}, { {"viewportCount must be greater than 0", 36}, { "chapters/vertexpostproc.html", 28} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-02691", 41}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-02692", 41}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-02693", 41}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-04931", 41}, { {"This command must be called in a subpass with bind point VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI. No draw commands can be called in the same subpass. Only one vkCmdSubpassShadingHUAWEI command can be called in a subpass", 225}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-06479", 41}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-07288", 41}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-07888", 41}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08114", 41}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08115", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08116", 41}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08117", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08119", 41}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08600", 41}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08601", 41}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08605", 41}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08606", 41}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08608", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08609", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08610", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08611", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08612", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-08613", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-09600", 41}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-10068", 41}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-10678", 41}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-10679", 41}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11297", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11298", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11299", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11300", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11301", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11302", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11304", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11305", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11306", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11308", 41}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11309", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11372", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11373", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11374", 41}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11376", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11397", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11398", 41}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11437", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11438", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11439", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11441", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11442", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-11485", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-None-12472", 41}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchSADQCOM-06976", 61}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchSADQCOM-12420", 61}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchSSDQCOM-06974", 61}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchWindow-09215", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchWindow-09216", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchWindow-09217", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBlockMatchWindow-12421", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageBoxFilterQCOM-06973", 57}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageSampleWeightedQCOM-06971", 62}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageSampleWeightedQCOM-06972", 62}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageSampleWeightedQCOM-06977", 62}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageSampleWeightedQCOM-06978", 62}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageWrite-04469", 49}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageWrite-08795", 49}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpImageWrite-08796", 49}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTensorReadARM-12462", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTypeImage-07027", 48}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTypeImage-07028", 48}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTypeImage-07029", 48}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTypeImage-07030", 48}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-OpTypeTensorARM-09906", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-SampledType-04470", 48}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-SampledType-04471", 48}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-SampledType-04472", 48}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-SampledType-04473", 48}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-addressModeU-12422", 49}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-commandBuffer-02707", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-commandBuffer-10746", 50}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_GRAPHICS_BIT operations", 101}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-cubicRangeClamp-09212", 52}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-dimensionCount-09905", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-filterCubic-02694", 48}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-filterCubicMinmax-02695", 54}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-flags-02696", 42}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-format-07753", 43}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-imageLayout-00344", 48}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-index-11450", 42}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-magFilter-04553", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-magFilter-09598", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-micromap-11636", 45}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-micromap-11637", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-micromap-11638", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-micromap-11639", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-mipmapMode-04770", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-mipmapMode-09599", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-pBindInfo-11375", 46}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-pDescription-09900", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-pNext-12463", 42}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-protectedNoFault-11455", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-protectedNoFault-11456", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-reductionMode-09213", 50}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-renderpass", 41}, { {"This command must only be called inside of a render pass instance", 65}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-selectableCubicWeights-09214", 59}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-sparseImageInt64Atomics-04474", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-sparseImageInt64Atomics-04475", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-storageBuffers-06936", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-uniformBuffers-06935", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-unnormalizedCoordinates-09635", 60}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdSubpassShadingHUAWEI-viewType-07752", 45}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/dispatch.html", 22} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-02691", 42}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-02692", 42}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-02693", 42}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-03429", 42}, { {"Any shader group handle referenced by this call must have been queried from the bound ray tracing pipeline", 106}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-06479", 42}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-07288", 42}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-07888", 42}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08114", 42}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08115", 42}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08116", 42}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08117", 42}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08119", 42}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08600", 42}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08601", 42}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08605", 42}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08606", 42}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08608", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08609", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08610", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08611", 42}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08612", 42}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-08613", 42}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-09458", 42}, { {"If the bound ray tracing pipeline state was created with the VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR dynamic state enabled then vkCmdSetRayTracingPipelineStackSizeKHR must have been called in the current command buffer prior to this trace command", 259}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-09600", 42}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-10068", 42}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-10678", 42}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-10679", 42}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11297", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11298", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11299", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11300", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11301", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11302", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11304", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11305", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11306", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11308", 42}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11309", 42}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11319", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, the value of the address at the expected location in shader record data must be a valid address", 333}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11372", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11373", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11374", 42}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11376", 42}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11397", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11398", 42}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11437", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11438", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11439", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11441", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11442", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-11485", 42}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-None-12472", 42}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchSADQCOM-06976", 62}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchSADQCOM-12420", 62}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchSSDQCOM-06974", 62}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchWindow-09215", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchWindow-09216", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchWindow-09217", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBlockMatchWindow-12421", 61}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageBoxFilterQCOM-06973", 58}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageSampleWeightedQCOM-06971", 63}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageSampleWeightedQCOM-06972", 63}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageSampleWeightedQCOM-06977", 63}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageSampleWeightedQCOM-06978", 63}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageWrite-04469", 50}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageWrite-08795", 50}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpImageWrite-08796", 50}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTensorReadARM-12462", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTypeImage-07027", 49}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTypeImage-07028", 49}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTypeImage-07029", 49}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTypeImage-07030", 49}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-OpTypeTensorARM-09906", 53}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-SampledType-04470", 49}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-SampledType-04471", 49}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-SampledType-04472", 49}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-SampledType-04473", 49}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-addressModeU-12422", 50}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-02707", 51}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-03635", 51}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-10746", 51}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-cubicRangeClamp-09212", 53}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-dimensionCount-09905", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-filterCubic-02694", 49}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-filterCubicMinmax-02695", 55}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-flags-02696", 43}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-format-07753", 44}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-imageLayout-00344", 49}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-index-11450", 43}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-indirectDeviceAddress-03633", 59}, { {"indirectDeviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 157}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-indirectDeviceAddress-03634", 59}, { {"indirectDeviceAddress must be a multiple of 4", 45}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-indirectDeviceAddress-03636", 59}, { {"All device addresses between indirectDeviceAddress and indirectDeviceAddress + sizeof(VkTraceRaysIndirectCommand2KHR) - 1 must be in the buffer device address range of the same buffer", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-indirectDeviceAddress-parameter", 63}, { {"indirectDeviceAddress must be a valid VkDeviceAddress value", 59}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-magFilter-04553", 47}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-magFilter-09598", 47}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-maxPipelineRayRecursionDepth-03679", 66}, { {"This command must not cause a shader call instruction to be executed from a shader invocation with a recursion depth greater than the value of maxPipelineRayRecursionDepth used to create the bound ray tracing pipeline", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-micromap-11636", 46}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-micromap-11637", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-micromap-11638", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-micromap-11639", 46}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-micromap-11640", 46}, { {"If a VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR descriptor is accessed as a result of this command, the access is not through an OpRayQueryProceedKHR operation, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR or VkAccelerationStructureTrianglesOpacityMicromapEXT structure included in its pNext chain, the VkPipeline bound to the pipeline bind point used by this command must have been created with flags that includes VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR", 542}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-mipmapMode-04770", 48}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-mipmapMode-09599", 48}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-pBindInfo-11375", 47}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-pDescription-09900", 50}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-pNext-12463", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-protectedNoFault-11455", 54}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-protectedNoFault-11456", 54}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-rayTracingMotionBlurPipelineTraceRaysIndirect-04951", 83}, { {"If the bound ray tracing pipeline was created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV VkPhysicalDeviceRayTracingMotionBlurFeaturesNV::rayTracingMotionBlurPipelineTraceRaysIndirect feature must be enabled", 219}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-rayTracingPipelineTraceRaysIndirect2-03637", 74}, { {"The rayTracingPipelineTraceRaysIndirect2 feature must be enabled", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-reductionMode-09213", 51}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-renderpass", 42}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-selectableCubicWeights-09214", 60}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-sparseImageInt64Atomics-04474", 61}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-sparseImageInt64Atomics-04475", 61}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-storageBuffers-06936", 52}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-uniformBuffers-06935", 52}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-unnormalizedCoordinates-09635", 61}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirect2KHR-viewType-07752", 46}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-02691", 41}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-02692", 41}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-02693", 41}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-03429", 41}, { {"Any shader group handle referenced by this call must have been queried from the bound ray tracing pipeline", 106}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-06479", 41}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-07288", 41}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-07888", 41}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08114", 41}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08115", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08116", 41}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08117", 41}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08119", 41}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08600", 41}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08601", 41}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08605", 41}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08606", 41}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08608", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08609", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08610", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08611", 41}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08612", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-08613", 41}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-09458", 41}, { {"If the bound ray tracing pipeline state was created with the VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR dynamic state enabled then vkCmdSetRayTracingPipelineStackSizeKHR must have been called in the current command buffer prior to this trace command", 259}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-09600", 41}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-10068", 41}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-10678", 41}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-10679", 41}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11297", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11298", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11299", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11300", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11301", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11302", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11304", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11305", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11306", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11308", 41}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11309", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11319", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, the value of the address at the expected location in shader record data must be a valid address", 333}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11372", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11373", 41}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11374", 41}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11376", 41}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11397", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11398", 41}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11437", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11438", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11439", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11441", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11442", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-11485", 41}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-None-12472", 41}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchSADQCOM-06976", 61}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchSADQCOM-12420", 61}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchSSDQCOM-06974", 61}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchWindow-09215", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchWindow-09216", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchWindow-09217", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBlockMatchWindow-12421", 60}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageBoxFilterQCOM-06973", 57}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageSampleWeightedQCOM-06971", 62}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageSampleWeightedQCOM-06972", 62}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageSampleWeightedQCOM-06977", 62}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageSampleWeightedQCOM-06978", 62}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageWrite-04469", 49}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageWrite-08795", 49}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpImageWrite-08796", 49}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTensorReadARM-12462", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTypeImage-07027", 48}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTypeImage-07028", 48}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTypeImage-07029", 48}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTypeImage-07030", 48}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-OpTypeTensorARM-09906", 52}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-SampledType-04470", 48}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-SampledType-04471", 48}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-SampledType-04472", 48}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-SampledType-04473", 48}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-addressModeU-12422", 49}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-02707", 50}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-03635", 50}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-10746", 50}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-cubicRangeClamp-09212", 52}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-dimensionCount-09905", 51}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-filterCubic-02694", 48}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-filterCubicMinmax-02695", 54}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-flags-02696", 42}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-flags-03511", 42}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR, the entries in the table identified by pMissShaderBindingTable->deviceAddress accessed as a result of this command in order to execute a miss shader must not be zero", 297}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-flags-03512", 42}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute an any-hit shader must not be zero", 299}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-flags-03513", 42}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute a closest hit shader must not be zero", 306}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-flags-03514", 42}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute an intersection shader must not be zero", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-format-07753", 43}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-imageLayout-00344", 48}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-index-11450", 42}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-indirectDeviceAddress-03633", 58}, { {"indirectDeviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT usage flag set", 157}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-indirectDeviceAddress-03634", 58}, { {"indirectDeviceAddress must be a multiple of 4", 45}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-indirectDeviceAddress-03636", 58}, { {"All device addresses between indirectDeviceAddress and indirectDeviceAddress + sizeof(VkTraceRaysIndirectCommandKHR) - 1 must be in the buffer device address range of the same buffer", 182}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-indirectDeviceAddress-parameter", 62}, { {"indirectDeviceAddress must be a valid VkDeviceAddress value", 59}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-magFilter-04553", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-magFilter-09598", 46}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-maxPipelineRayRecursionDepth-03679", 65}, { {"This command must not cause a shader call instruction to be executed from a shader invocation with a recursion depth greater than the value of maxPipelineRayRecursionDepth used to create the bound ray tracing pipeline", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-micromap-11636", 45}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-micromap-11637", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-micromap-11638", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-micromap-11639", 45}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-micromap-11640", 45}, { {"If a VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR descriptor is accessed as a result of this command, the access is not through an OpRayQueryProceedKHR operation, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR or VkAccelerationStructureTrianglesOpacityMicromapEXT structure included in its pNext chain, the VkPipeline bound to the pipeline bind point used by this command must have been created with flags that includes VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR", 542}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-mipmapMode-04770", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-mipmapMode-09599", 47}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pBindInfo-11375", 46}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pCallableShaderBindingTable-03692", 64}, { {"pCallableShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pCallableShaderBindingTable-03693", 64}, { {"pCallableShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 138}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pCallableShaderBindingTable-parameter", 68}, { {"pCallableShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 104}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pDescription-09900", 49}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pHitShaderBindingTable-03688", 59}, { {"pHitShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 182}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pHitShaderBindingTable-03689", 59}, { {"pHitShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 133}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pHitShaderBindingTable-04735", 59}, { {"Any non-zero hit shader group entries in the table identified by pHitShaderBindingTable->deviceAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR", 279}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pHitShaderBindingTable-04736", 59}, { {"Any non-zero hit shader group entries in the table identified by pHitShaderBindingTable->deviceAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR", 276}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pHitShaderBindingTable-parameter", 63}, { {"pHitShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 99}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pMissShaderBindingTable-03684", 60}, { {"pMissShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pMissShaderBindingTable-03685", 60}, { {"pMissShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 134}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pMissShaderBindingTable-parameter", 64}, { {"pMissShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pNext-12463", 42}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pRayGenShaderBindingTable-03681", 62}, { {"pRayGenShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pRayGenShaderBindingTable-03682", 62}, { {"pRayGenShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 136}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-pRaygenShaderBindingTable-parameter", 66}, { {"pRaygenShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 102}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-protectedNoFault-11455", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-protectedNoFault-11456", 53}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-rayTracingMotionBlurPipelineTraceRaysIndirect-04951", 82}, { {"If the bound ray tracing pipeline was created with VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV VkPhysicalDeviceRayTracingMotionBlurFeaturesNV::rayTracingMotionBlurPipelineTraceRaysIndirect feature must be enabled", 219}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-rayTracingPipelineTraceRaysIndirect-03637", 72}, { {"The rayTracingPipelineTraceRaysIndirect feature must be enabled", 63}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-reductionMode-09213", 50}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-renderpass", 41}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-selectableCubicWeights-09214", 59}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-size-04023", 41}, { {"The size member of pRayGenShaderBindingTable must be equal to its stride member", 79}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-sparseImageInt64Atomics-04474", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-sparseImageInt64Atomics-04475", 60}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-storageBuffers-06936", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-03686", 43}, { {"pMissShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 129}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-03690", 43}, { {"pHitShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 128}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-03694", 43}, { {"pCallableShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 133}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-04029", 43}, { {"pMissShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-04035", 43}, { {"pHitShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 130}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-stride-04041", 43}, { {"pCallableShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 135}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-uniformBuffers-06935", 51}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-unnormalizedCoordinates-09635", 60}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysIndirectKHR-viewType-07752", 45}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-02691", 33}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-02692", 33}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-02693", 33}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-03429", 33}, { {"Any shader group handle referenced by this call must have been queried from the bound ray tracing pipeline", 106}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-06479", 33}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-07288", 33}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-07888", 33}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08114", 33}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08115", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08116", 33}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08117", 33}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08119", 33}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08600", 33}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08601", 33}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08605", 33}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08606", 33}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08608", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08609", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08610", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08611", 33}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08612", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-08613", 33}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-09458", 33}, { {"If the bound ray tracing pipeline state was created with the VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR dynamic state enabled then vkCmdSetRayTracingPipelineStackSizeKHR must have been called in the current command buffer prior to this trace command", 259}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-09600", 33}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-10068", 33}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-10678", 33}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-10679", 33}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11297", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11298", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11299", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11300", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11301", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11302", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11304", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11305", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11306", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11308", 33}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11309", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11319", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, the value of the address at the expected location in shader record data must be a valid address", 333}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11372", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11373", 33}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11374", 33}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11376", 33}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11397", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11398", 33}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11437", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11438", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11439", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11441", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11442", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-11485", 33}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-None-12472", 33}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchSADQCOM-06976", 53}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchSADQCOM-12420", 53}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchSSDQCOM-06974", 53}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchWindow-09215", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchWindow-09216", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchWindow-09217", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBlockMatchWindow-12421", 52}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageBoxFilterQCOM-06973", 49}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageSampleWeightedQCOM-06971", 54}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageSampleWeightedQCOM-06972", 54}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageSampleWeightedQCOM-06977", 54}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageSampleWeightedQCOM-06978", 54}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageWrite-04469", 41}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageWrite-08795", 41}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpImageWrite-08796", 41}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTensorReadARM-12462", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTypeImage-07027", 40}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTypeImage-07028", 40}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTypeImage-07029", 40}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTypeImage-07030", 40}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-OpTypeTensorARM-09906", 44}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-SampledType-04470", 40}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-SampledType-04471", 40}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-SampledType-04472", 40}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-SampledType-04473", 40}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-addressModeU-12422", 41}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-allowClusterAccelerationStructure-10578", 62}, { {"If the traced geometry contains a cluster acceleration structure, then VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV::allowClusterAccelerationStructure must have been set for that pipeline", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-02707", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-03635", 42}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-10746", 42}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-cubicRangeClamp-09212", 44}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-depth-03640", 34}, { {"depth must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[2]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-dimensionCount-09905", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-filterCubic-02694", 40}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-filterCubicMinmax-02695", 46}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-flags-02696", 34}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-flags-03511", 34}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR, the entries in the table identified by pMissShaderBindingTable->deviceAddress accessed as a result of this command in order to execute a miss shader must not be zero", 297}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-flags-03512", 34}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute an any-hit shader must not be zero", 299}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-flags-03513", 34}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute a closest hit shader must not be zero", 306}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-flags-03514", 34}, { {"If the bound ray tracing pipeline was created with flags that included VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR, entries in the table identified by pHitShaderBindingTable->deviceAddress accessed as a result of this command in order to execute an intersection shader must not be zero", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-format-07753", 35}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-height-03639", 35}, { {"height must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[1]", 141}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-imageLayout-00344", 40}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-index-11450", 34}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-magFilter-04553", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-magFilter-09598", 38}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-maxPipelineRayRecursionDepth-03679", 57}, { {"This command must not cause a shader call instruction to be executed from a shader invocation with a recursion depth greater than the value of maxPipelineRayRecursionDepth used to create the bound ray tracing pipeline", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-micromap-11636", 37}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-micromap-11637", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-micromap-11638", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-micromap-11639", 37}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-micromap-11640", 37}, { {"If a VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR descriptor is accessed as a result of this command, the access is not through an OpRayQueryProceedKHR operation, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR or VkAccelerationStructureTrianglesOpacityMicromapEXT structure included in its pNext chain, the VkPipeline bound to the pipeline bind point used by this command must have been created with flags that includes VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_KHR", 542}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-mipmapMode-04770", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-mipmapMode-09599", 39}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pBindInfo-11375", 38}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pCallableShaderBindingTable-03692", 56}, { {"pCallableShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pCallableShaderBindingTable-03693", 56}, { {"pCallableShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 138}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pCallableShaderBindingTable-parameter", 60}, { {"pCallableShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 104}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pDescription-09900", 41}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pHitShaderBindingTable-03688", 51}, { {"pHitShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 182}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pHitShaderBindingTable-03689", 51}, { {"pHitShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 133}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pHitShaderBindingTable-04735", 51}, { {"Any non-zero hit shader group entries in the table identified by pHitShaderBindingTable->deviceAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_TRIANGLES_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_KHR", 279}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pHitShaderBindingTable-04736", 51}, { {"Any non-zero hit shader group entries in the table identified by pHitShaderBindingTable->deviceAddress accessed by this call from a geometry with a geometryType of VK_GEOMETRY_TYPE_AABBS_KHR must have been created with VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_KHR", 276}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pHitShaderBindingTable-parameter", 55}, { {"pHitShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 99}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pMissShaderBindingTable-03684", 52}, { {"pMissShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pMissShaderBindingTable-03685", 52}, { {"pMissShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 134}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pMissShaderBindingTable-parameter", 56}, { {"pMissShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pNext-12463", 34}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pRayGenShaderBindingTable-03681", 54}, { {"pRayGenShaderBindingTable->deviceAddress must be a device address allocated to the application from a buffer created with the VK_BUFFER_USAGE_SHADER_BINDING_TABLE_BIT_KHR usage flag set", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pRayGenShaderBindingTable-03682", 54}, { {"pRayGenShaderBindingTable->deviceAddress must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupBaseAlignment", 136}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-pRaygenShaderBindingTable-parameter", 58}, { {"pRaygenShaderBindingTable must be a valid pointer to a valid VkStridedDeviceAddressRegionKHR structure", 102}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-protectedNoFault-11455", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-protectedNoFault-11456", 45}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-reductionMode-09213", 42}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-renderpass", 33}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-selectableCubicWeights-09214", 51}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-size-04023", 33}, { {"The size member of pRayGenShaderBindingTable must be equal to its stride member", 79}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-sparseImageInt64Atomics-04474", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-sparseImageInt64Atomics-04475", 52}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-storageBuffers-06936", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-03686", 35}, { {"pMissShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 129}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-03690", 35}, { {"pHitShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 128}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-03694", 35}, { {"pCallableShaderBindingTable->stride must be a multiple of VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleAlignment", 133}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-04029", 35}, { {"pMissShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-04035", 35}, { {"pHitShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 130}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-stride-04041", 35}, { {"pCallableShaderBindingTable->stride must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxShaderGroupStride", 135}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-uniformBuffers-06935", 43}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-unnormalizedCoordinates-09635", 52}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-viewType-07752", 37}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-width-03638", 34}, { {"width must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0] x VkPhysicalDeviceLimits::maxComputeWorkGroupSize[0]", 140}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysKHR-width-03641", 34}, { {"width x height x depth must be less than or equal to VkPhysicalDeviceRayTracingPipelinePropertiesKHR::maxRayDispatchInvocationCount", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-02691", 32}, { {"If a VkImageView is accessed using atomic operations as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-02692", 32}, { {"If a VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT", 185}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-02693", 32}, { {"If the VK_EXT_filter_cubic extension is not enabled and any VkImageView is sampled with VK_FILTER_CUBIC_EXT as a result of this command, it must not have a VkImageViewType of VK_IMAGE_VIEW_TYPE_3D, VK_IMAGE_VIEW_TYPE_CUBE, or VK_IMAGE_VIEW_TYPE_CUBE_ARRAY", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-03429", 32}, { {"Any shader group handle referenced by this call must have been queried from the bound ray tracing pipeline", 106}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-06479", 32}, { {"If a VkImageView is sampled with depth comparison, the image view's format features must contain VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-07288", 32}, { {"Any shader invocation executed by this command must terminate", 61}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-07888", 32}, { {"If a VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER descriptor is accessed using atomic operations as a result of this command, then the storage texel buffer's format features must contain VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT", 231}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08114", 32}, { {"Descriptors in each bound descriptor set, specified via vkCmdBindDescriptorSets, must be valid if they are accessed as described by descriptor validity by the VkPipeline bound to the pipeline bind point used by this command and the bound VkPipeline was not created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 314}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08115", 32}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdBindDescriptorSets, the bound VkPipeline must have been created without VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08116", 32}, { {"Descriptors in bound descriptor buffers, specified via vkCmdSetDescriptorBufferOffsetsEXT, must be valid if they are dynamically accessed", 137}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08117", 32}, { {"If the descriptors used by the VkPipeline bound to the pipeline bind point were specified via vkCmdSetDescriptorBufferOffsetsEXT, the bound VkPipeline must have been created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08119", 32}, { {"If a descriptor is dynamically used with a VkPipeline created with VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 151}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08600", 32}, { {"If a a bound shader was created with a pipeline layout and statically uses a set n, a descriptor set must have been bound to n at the same pipeline bind point, with layouts compatible for set n with the layout(s) used to create the shader, as described in Pipeline Layout Compatibility", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08601", 32}, { {"If a a bound shader was created with a pipeline layout or specified push constant ranges and statically uses a push constant value, that value must have been set for the same pipeline bind point, with push constant ranges that are compatible with the push constant range used to create the shader", 296}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08605", 32}, { {"If a descriptor is dynamically used with a VkShaderEXT created with a VkDescriptorSetLayout that was created with VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT, the descriptor memory must be resident", 211}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08606", 32}, { {"If the shaderObject feature is not enabled, a valid pipeline must be bound to the pipeline bind point used by this command", 122}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08608", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, there must not have been any calls to dynamic state setting commands for any state specified statically in the VkPipeline object bound to the pipeline bind point used by this command, since that pipeline was bound", 285}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08609", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's viewType must be VK_IMAGE_VIEW_TYPE_1D or VK_IMAGE_VIEW_TYPE_2D", 215}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08610", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions with ImplicitLod, Dref or Proj in their name", 283}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08611", 32}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the sampler must not be used with any of the SPIR-V OpImageSample* or OpImageSparseSample* instructions that includes a LOD bias or any offset values", 284}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08612", 32}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-08613", 32}, { {"If the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a storage buffer, it must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 309}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-09458", 32}, { {"If the bound ray tracing pipeline state was created with the VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR dynamic state enabled then vkCmdSetRayTracingPipelineStackSizeKHR must have been called in the current command buffer prior to this trace command", 259}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-09600", 32}, { {"If a descriptor with type equal to any of VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT is accessed as a result of this command, all image subresources identified by that descriptor must be in the image layout identified when the descriptor was written", 401}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-10068", 32}, { {"For each array of resources that is used by a bound shader, the indices used to access members of the array must be less than the descriptor count for the identified binding in the descriptor sets used by this command", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-10678", 32}, { {"If this command is recorded inside a tile shading render pass instance, the stages corresponding to the pipeline bind point used by this command must only include VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_FRAGMENT_BIT, and/or VK_SHADER_STAGE_COMPUTE_BIT", 255}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-10679", 32}, { {"If this command is recorded where per-tile execution model is enabled, there must be no access to any image while the image was be transitioned to the VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT layout", 210}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11297", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a OpTypeStruct decorated with Block or BufferBlock using that mapping, the calculated offset for the resource heap must be a multiple of bufferDescriptorAlignment", 612}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11298", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeImage or OpTypeSampledImage using that mapping, the calculated offset for the resource heap must be a multiple of imageDescriptorAlignment", 597}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11299", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeSampler or OpTypeSampledImage using that mapping, the calculated offset for the sampler heap must be a multiple of samplerDescriptorAlignment", 600}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11300", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 4", 447}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11301", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 523}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11302", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress", 376}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11304", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a multiple of 8", 372}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11305", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address at the expected location in push data must be a valid VkDeviceAddress backed by physical memory at every offset specified by each mapping", 448}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11306", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the value of the address pointed to by the address in push data must be a valid VkDeviceAddress", 381}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11308", 32}, { {"For each descriptor heap that is statically used by a bound shader, either directly or via a descriptor mapping, a valid descriptor heap must be bound", 150}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11309", 32}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, execution of this command must not result in any descriptor read accessing data outside of the user range of the respective heap bound by vkCmdBind*HeapEXT commands", 351}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11319", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, the value of the address at the expected location in shader record data must be a valid address", 333}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11372", 32}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer or uniform texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified via VkDeviceAddressRangeKHR when the descriptor was written", 624}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11373", 32}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer or storage texel buffer through a descriptor in the bound resource heap, that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, the robustBufferAccess2 feature is not enabled, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the descriptor specified by VkDeviceAddressRangeKHR when the descriptor was written", 623}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11374", 32}, { {"If the robustBufferAccess2 feature is not enabled, the robustBufferAccess feature is not enabled, and any VkShaderEXT bound to a stage corresponding to the pipeline bind point used by this command accesses a uniform buffer, uniform texel buffer, storage buffer, or storage texel buffer, that shader must not access values outside of the range of the buffer as specified by VkDeviceAddressRangeKHR when the descriptor was written", 428}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11376", 32}, { {"If a bound shader was created as a VkShaderEXT with the VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT flag or as part of a pipeline with the VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT flag, and that shader statically uses a push constant value, that value must have been set by vkCmdPushDataEXT", 291}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11397", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, and a shader accesses an OpTypeTensorARM using that mapping, the calculated offset for the resource heap must be a multiple of tensorDescriptorAlignment", 580}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11398", 32}, { {"If a bound shader was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the access must not be out of bounds", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11437", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a resource using that mapping, the buffer from which the address in push data was queried must have been created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 553}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11438", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11439", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must have been queried from a buffer created with the VK_BUFFER_USAGE_STORAGE_BUFFER_BIT usage flag set", 550}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11441", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT and a shader accesses a uniform buffer using that mapping, the address that the uniform buffer is mapped to must be aligned to minUniformBufferOffsetAlignment", 549}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11442", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses a storage buffer using that mapping, the address that the storage buffer is mapped to must be aligned to minStorageBufferOffsetAlignment", 497}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-11485", 32}, { {"If a pipeline is bound to the pipeline bind point used by this command, or shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, and a shader accesses an acceleration structure using that mapping, the address that the acceleration structure is mapped to must be an acceleration structure address retrieved from a VkAccelerationStructureKHR object via vkGetAccelerationStructureDeviceAddressKHR or handle retrieved from a VkAccelerationStructureNV object via vkGetAccelerationStructureHandleNV", 702}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-None-12472", 32}, { {"Descriptors dynamically accessed in bound descriptor heaps must be valid", 72}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchSADQCOM-06976", 52}, { {"If OpImageBlockMatchSADQCOM or OpImageBlockMatchSSDQCOM is used to read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 397}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchSADQCOM-12420", 52}, { {"If OpImageBlockMatchSADQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchSSDQCOM-06974", 52}, { {"If OpImageBlockMatchSSDQCOM is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM or VK_FORMAT_FEATURE_2_BLOCK_MATCHING_SXD_BIT_QCOM", 238}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchWindow-09215", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM", 237}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchWindow-09216", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction is used to read from an VkImageView as a result of this command, then the image view's format must be a single-component format", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchWindow-09217", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM read from a reference image as result of this command, and the blockMatchExtendedClampToEdge feature is not enabled or the input sampler was not created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE then the specified reference coordinates must not fail integer texel coordinate validation", 396}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBlockMatchWindow-12421", 51}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the blockMatchExtendedClampToEdge feature is not enabled, the input sampler must not have been created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 318}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageBoxFilterQCOM-06973", 48}, { {"If OpImageBoxFilterQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM", 183}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageSampleWeightedQCOM-06971", 53}, { {"If OpImageSampleWeightedQCOM is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM", 190}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageSampleWeightedQCOM-06972", 53}, { {"If OpImageSampleWeightedQCOM uses a VkImageView as a sample weight image as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageSampleWeightedQCOM-06977", 53}, { {"If OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 368}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageSampleWeightedQCOM-06978", 53}, { {"If any command other than OpImageSampleWeightedQCOM, OpImageBoxFilterQCOM, OpImageBlockMatchWindowSSDQCOM, OpImageBlockMatchWindowSADQCOM, OpImageBlockMatchGatherSSDQCOM, OpImageBlockMatchGatherSADQCOM, OpImageBlockMatchSSDQCOM, or OpImageBlockMatchSADQCOM uses a VkSampler as a result of this command, then the sampler must not have been created with VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM", 395}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageWrite-04469", 40}, { {"If a VkBufferView is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the buffer view's format", 198}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageWrite-08795", 40}, { {"If a VkImageView created with a format other than VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have at least as many components as the image view's format", 248}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpImageWrite-08796", 40}, { {"If a VkImageView created with the format VK_FORMAT_A8_UNORM is accessed using OpImageWrite as a result of this command, then the Type of the Texel operand of that instruction must have four components", 200}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTensorReadARM-12462", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, using OpTensorReadARM or OpTensorWriteARM, and the underlying VkTensorARM object was created with a VK_TENSOR_TILING_BRICK_16_WIDE_ARM, VK_TENSOR_TILING_BRICK_8_WIDE_ARM, VK_TENSOR_TILING_BRICK_4_WIDE_ARM, VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_ARM, or VK_TENSOR_TILING_BLOCK_U_INTERLEAVED_64K_ARM VkTensorTilingARM, and the object read or written is an array, then the innermost coordinate provided to OpTensorReadARM or OpTensorWriteARM via the Coordinates operand must be a multiple of the largest power of 2 that is less than or equal to the number of components of the array", 664}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTypeImage-07027", 39}, { {"For any VkImageView being written as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTypeImage-07028", 39}, { {"For any VkImageView being read as a storage image where the image format field of the OpTypeImage is Unknown, the view's format features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 201}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTypeImage-07029", 39}, { {"For any VkBufferView being written as a storage texel buffer where the image format field of the OpTypeImage is Unknown, the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT", 213}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTypeImage-07030", 39}, { {"Any VkBufferView being read as a storage texel buffer where the image format field of the OpTypeImage is Unknown then the view's buffer features must contain VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT", 209}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-OpTypeTensorARM-09906", 43}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the element type of the OpTypeTensorARM of the tensor resource variable must be compatible with the VkFormat of the VkTensorViewARM used for the access", 243}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-SampledType-04470", 39}, { {"If a VkImageView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 194}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-SampledType-04471", 39}, { {"If a VkImageView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 204}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-SampledType-04472", 39}, { {"If a VkBufferView with a VkFormat that has a 64-bit component width is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 64", 195}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-SampledType-04473", 39}, { {"If a VkBufferView with a VkFormat that has a component width less than 64-bit is accessed as a result of this command, the SampledType of the OpTypeImage operand of that instruction must have a Width of 32", 205}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-addressModeU-12422", 40}, { {"If a OpImageBlockMatchWindow*QCOM or OpImageBlockMatchGather*QCOM instruction uses a VkSampler as a result of this command, and the input sampler was created with either addressModeU or addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, then the sampler must have been created with both addressModeU and addressModeV equal to VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 373}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-allowClusterAccelerationStructure-10577", 61}, { {"If the traced geometry contains a cluster acceleration structure, then VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV::allowClusterAccelerationStructure must have been set for that pipeline", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingOffset-02461", 55}, { {"callableShaderBindingOffset must be less than the size of callableShaderBindingTableBuffer", 90}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingOffset-02462", 55}, { {"callableShaderBindingOffset must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupBaseAlignment", 114}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingStride-02465", 55}, { {"callableShaderBindingStride must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupHandleSize", 111}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingStride-02468", 55}, { {"callableShaderBindingStride must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxShaderGroupStride", 118}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingTableBuffer-04045", 60}, { {"If callableShaderBindingTableBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 133}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-callableShaderBindingTableBuffer-parameter", 64}, { {"If callableShaderBindingTableBuffer is not VK_NULL_HANDLE, callableShaderBindingTableBuffer must be a valid VkBuffer handle", 123}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-02707", 41}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, any resource accessed by bound shaders must not be a protected resource", 160}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-04624", 41}, { {"commandBuffer must not be a protected command buffer", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-10746", 41}, { {"The VkDeviceMemory object allocated from a VkMemoryHeap with the VK_MEMORY_HEAP_TILE_MEMORY_BIT_QCOM property that is bound to a resource accessed as a result of this command must be the active bound tile memory object in commandBuffer", 235}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-commonparent", 34}, { {"Each of callableShaderBindingTableBuffer, commandBuffer, hitShaderBindingTableBuffer, missShaderBindingTableBuffer, and raygenShaderBindingTableBuffer that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 270}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-cubicRangeClamp-09212", 43}, { {"If the cubicRangeClamp feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must not have a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM", 266}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-depth-02471", 33}, { {"depth must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[2]", 87}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-dimensionCount-09905", 42}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the Rank of the OpTypeTensorARM of the tensor resource variable must be equal to the dimensionCount provided via VkTensorCreateInfoARM::pDescription when creating the underlying VkTensorARM object", 288}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-filterCubic-02694", 39}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have a VkImageViewType and format that supports cubic filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubic returned by vkGetPhysicalDeviceImageFormatProperties2", 282}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-filterCubicMinmax-02695", 45}, { {"Any VkImageView being sampled with VK_FILTER_CUBIC_EXT with a reduction mode of either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX as a result of this command must have a VkImageViewType and format that supports cubic filtering together with minmax filtering, as specified by VkFilterCubicImageViewImageFormatPropertiesEXT::filterCubicMinmax returned by vkGetPhysicalDeviceImageFormatProperties2", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-flags-02696", 33}, { {"Any VkImage created with a VkImageCreateInfo::flags containing VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV sampled as a result of this command must only be sampled using a VkSamplerAddressMode of VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE", 227}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-format-07753", 34}, { {"If a VkImageView or VkBufferView is accessed as a result of this command, then the numeric type of the view's format and the Sampled Type operand of the OpTypeImage must match", 175}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-height-02470", 34}, { {"height must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[1]", 88}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingOffset-02459", 50}, { {"hitShaderBindingOffset must be less than the size of hitShaderBindingTableBuffer", 80}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingOffset-02460", 50}, { {"hitShaderBindingOffset must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupBaseAlignment", 109}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingStride-02464", 50}, { {"hitShaderBindingStride must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupHandleSize", 106}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingStride-02467", 50}, { {"hitShaderBindingStride must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxShaderGroupStride", 113}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingTableBuffer-04044", 55}, { {"If hitShaderBindingTableBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 128}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-hitShaderBindingTableBuffer-parameter", 59}, { {"If hitShaderBindingTableBuffer is not VK_NULL_HANDLE, hitShaderBindingTableBuffer must be a valid VkBuffer handle", 113}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-imageLayout-00344", 39}, { {"If an image descriptor is accessed by a shader, the VkImageLayout must match the subresource accessible from the VkImageView as defined by the image layout matching rules", 170}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-index-11450", 33}, { {"If a shader uses a sampler descriptor to sample an image as a result of this command, and that sampler descriptor uses a custom border color with an index defined by VkSamplerCustomBorderColorIndexCreateInfoEXT, the value of VkSamplerCustomBorderColorIndexCreateInfoEXT::index must have been registered before this command was recorded, and still be registered during the sampling operation, with an identically defined color", 425}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-magFilter-04553", 37}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 342}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-magFilter-09598", 37}, { {"If a VkSampler created with magFilter or minFilter equal to VK_FILTER_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 335}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-maxRecursionDepth-03625", 45}, { {"This command must not cause a pipeline trace ray instruction to be executed from a shader invocation with a recursion depth greater than the value of maxRecursionDepth used to create the bound ray tracing pipeline", 213}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-micromap-11636", 36}, { {"If an acceleration structure is accessed as a result of this command through an OpRayQueryProceedKHR operation , the VK_EXT_opacity_micromap extension is not enabled, and the acceleration structure is built using geometry with a VkAccelerationStructureTrianglesOpacityMicromapKHR structure included in its pNext chain, the shader must have specified a OpacityMicromapIdKHR execution mode with its specialization constant operand equal to VK_TRUE", 445}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-micromap-11637", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure was constructed through deserialization, all micromap arrays referenced by the acceleration structure that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command on the device", 458}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-micromap-11638", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkPipeline bound to the pipeline bind point used by this command must not have been created with flags that includes VK_PIPELINE_CREATE_2_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_KHR", 442}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-micromap-11639", 36}, { {"If an acceleration structure is accessed as a result of this command and the acceleration structure is built using geometry with an indexBuffer containing both VkOpacityMicromapSpecialIndexKHR values and indices into an associated micromap array, any VkShaderEXT that accessed the descriptor must not have been created with flags that includes VK_SHADER_CREATE_OPACITY_MICROMAP_DISALLOW_MIXED_SPECIAL_INDEX_BIT_EXT", 414}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-mipmapMode-04770", 38}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR, reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE, and compareEnable equal to VK_FALSE is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT", 343}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-mipmapMode-09599", 38}, { {"If a VkSampler created with mipmapMode equal to VK_SAMPLER_MIPMAP_MODE_LINEAR and reductionMode equal to either VK_SAMPLER_REDUCTION_MODE_MIN or VK_SAMPLER_REDUCTION_MODE_MAX is used to sample a VkImageView as a result of this command, then the image view's format features must contain VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT", 336}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingOffset-02457", 51}, { {"missShaderBindingOffset must be less than the size of missShaderBindingTableBuffer", 82}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingOffset-02458", 51}, { {"missShaderBindingOffset must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupBaseAlignment", 110}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingStride-02463", 51}, { {"missShaderBindingStride must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupHandleSize", 107}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingStride-02466", 51}, { {"missShaderBindingStride must be less than or equal to VkPhysicalDeviceRayTracingPropertiesNV::maxShaderGroupStride", 114}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingTableBuffer-04043", 56}, { {"If missShaderBindingTableBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 129}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-missShaderBindingTableBuffer-parameter", 60}, { {"If missShaderBindingTableBuffer is not VK_NULL_HANDLE, missShaderBindingTableBuffer must be a valid VkBuffer handle", 115}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-pBindInfo-11375", 37}, { {"If any bound shader uses an embedded sampler via a descriptor mapping, the value of pBindInfo->reservedRangeSize set for vkCmdBindSamplerHeapEXT must be greater than or equal to minSamplerHeapReservedRangeWithEmbedded", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-pDescription-09900", 40}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command, then the underlying VkTensorARM object must have been created with the VK_TENSOR_USAGE_SHADER_BIT_ARM usage flag set", 203}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-pNext-12463", 33}, { {"If a VK_DESCRIPTOR_TYPE_TENSOR_ARM descriptor is accessed as a result of this command and the corresponding VkTensorARM object was created with a VkTensorRollingBackingCreateInfoARM structure in the pNext chain of VkTensorCreateInfoARM, then the number of elements accessed with one OpTensorReadARM or OpTensorWriteARM, must be less than or equal to the number of elements available from the starting coordinate in the rolling backing along the innermost dimension, defined as wraps[dimensionCount - 1] - InnermostCoordinate % wraps[dimensionCount - 1] where InnermostCoordinate is the last component of the coordinates array provided via the Coordinates operand", 662}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-protectedNoFault-11455", 44}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT, VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT, or VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT, the address that the resource is mapped to must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 522}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-protectedNoFault-11456", 44}, { {"If protectedNoFault is not supported, a pipeline is bound to the pipeline bind point used by this command, or a shader is bound to a shader stage used by this command, and it was created with a descriptor mapping using VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT or VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT, the address of the indirect memory must have been queried from a buffer created without the VK_BUFFER_CREATE_PROTECTED_BIT create flag set", 483}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-raygenShaderBindingOffset-02455", 53}, { {"raygenShaderBindingOffset must be less than the size of raygenShaderBindingTableBuffer", 86}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-raygenShaderBindingOffset-02456", 53}, { {"raygenShaderBindingOffset must be a multiple of VkPhysicalDeviceRayTracingPropertiesNV::shaderGroupBaseAlignment", 112}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-raygenShaderBindingTableBuffer-04042", 58}, { {"If raygenShaderBindingTableBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 131}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-raygenShaderBindingTableBuffer-parameter", 62}, { {"raygenShaderBindingTableBuffer must be a valid VkBuffer handle", 62}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-reductionMode-09213", 41}, { {"Any VkImageView being sampled with a VkSamplerReductionModeCreateInfo::reductionMode equal to VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM as a result of this command must sample with VK_FILTER_CUBIC_EXT", 217}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-renderpass", 32}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-selectableCubicWeights-09214", 50}, { {"If the selectableCubicWeights feature is not enabled, then any VkImageView being sampled with VK_FILTER_CUBIC_EXT as a result of this command must have VkSamplerCubicWeightsCreateInfoQCOM::cubicWeights equal to VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM", 251}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-sparseImageInt64Atomics-04474", 51}, { {"If the sparseImageInt64Atomics feature is not enabled, VkImage objects created with the VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 250}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-sparseImageInt64Atomics-04475", 51}, { {"If the sparseImageInt64Atomics feature is not enabled, VkBuffer objects created with the VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT flag must not be accessed by atomic instructions through an OpTypeImage with a SampledType with a Width of 64 by this command", 252}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-storageBuffers-06936", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a storage buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for storageBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-uniformBuffers-06935", 42}, { {"If any stage of the VkPipeline object bound to the pipeline bind point used by this command accesses a uniform buffer, and that stage was created without enabling either VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS or VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2 for uniformBuffers, and the robustBufferAccess feature is not enabled, that stage must not access values outside of the range of the buffer as specified in the descriptor set bound to the same pipeline bind point", 507}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-unnormalizedCoordinates-09635", 51}, { {"If a VkSampler created with unnormalizedCoordinates equal to VK_TRUE is used to sample a VkImageView as a result of this command, then the image view's levelCount and layerCount must be 1", 187}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-videocoding", 33}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-viewType-07752", 36}, { {"If a VkImageView is accessed as a result of this command, then the image view's viewType must match the Dim operand of the OpTypeImage as described in Compatibility Between SPIR-V Image Dimensions and Vulkan ImageView Types", 223}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdTraceRaysNV-width-02469", 33}, { {"width must be less than or equal to VkPhysicalDeviceLimits::maxComputeWorkGroupCount[0]", 87}, { "chapters/raytracing.html", 24} } },
        { {"VUID-vkCmdUpdateBuffer-commandBuffer-01813", 42}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstBuffer must not be a protected buffer", 129}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-commandBuffer-01814", 42}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstBuffer must not be an unprotected buffer", 129}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-commandBuffer-cmdpool", 44}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-commandBuffer-parameter", 46}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-commandBuffer-recording", 46}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-commonparent", 35}, { {"Both of commandBuffer, and dstBuffer must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dataSize-00033", 37}, { {"dataSize must be less than or equal to the size of dstBuffer minus dstOffset", 76}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dataSize-00037", 37}, { {"dataSize must be less than or equal to 65536", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dataSize-00038", 37}, { {"dataSize must be a multiple of 4", 32}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dataSize-arraylength", 43}, { {"dataSize must be greater than 0", 31}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dstBuffer-00034", 38}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dstBuffer-00035", 38}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dstBuffer-parameter", 42}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dstOffset-00032", 38}, { {"dstOffset must be less than the size of dstBuffer", 49}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-dstOffset-00036", 38}, { {"dstOffset must be a multiple of 4", 33}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-pData-parameter", 38}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-renderpass", 33}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-suspended", 32}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateBuffer-videocoding", 34}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-commandBuffer-13010", 45}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 151}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-commandBuffer-13011", 45}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 144}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_TRANSFER_BIT operations", 101}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dataSize-13008", 40}, { {"dataSize must be less than or equal to pDstRange->size", 54}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dataSize-13009", 40}, { {"dataSize must be a multiple of 4", 32}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dataSize-arraylength", 46}, { {"dataSize must be greater than 0", 31}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dstFlags-13100", 40}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 138}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dstFlags-13101", 40}, { {"dstFlags must not include both VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR and VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 160}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dstFlags-parameter", 44}, { {"dstFlags must be a valid combination of VkAddressCommandFlagBitsKHR values", 74}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-dstRange-13005", 40}, { {"The buffer from which dstRange was queried must have been created with VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag", 114}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pData-parameter", 41}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13006", 41}, { {"pDstRange->address must be a multiple of 4", 42}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13007", 41}, { {"pDstRange->size must be less than or equal to 65536", 51}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13097", 41}, { {"If the range specified by pDstRange is not bound completely to memory when accessed, dstFlags must not include VK_ADDRESS_COMMAND_FULLY_BOUND_BIT_KHR", 149}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13098", 41}, { {"If the buffer from which the range specified by pDstRange was created with VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 204}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13099", 41}, { {"If the buffer from which the range specified by pDstRange was created without VK_BUFFER_CREATE_PROTECTED_BIT, and protectedNoFault is not supported, dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 211}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13122", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_STORAGE_BUFFER_USAGE_BIT_KHR", 283}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13123", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, dstFlags must not include VK_ADDRESS_COMMAND_STORAGE_BUFFER_USAGE_BIT_KHR", 231}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13124", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created with VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR or VK_ADDRESS_COMMAND_UNKNOWN_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 320}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-13125", 41}, { {"If any buffer, which is bound to a range of VkDeviceMemory that overlaps the range backing pDstRange, was created without VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT, dstFlags must not include VK_ADDRESS_COMMAND_TRANSFORM_FEEDBACK_BUFFER_USAGE_BIT_KHR", 257}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-pDstRange-parameter", 45}, { {"pDstRange must be a valid pointer to a valid VkDeviceAddressRangeKHR structure", 78}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-renderpass", 36}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdateMemoryKHR-videocoding", 37}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/clears.html", 20} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-commandBuffer-cmdpool", 62}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-commandBuffer-parameter", 64}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-commandBuffer-recording", 64}, { {"commandBuffer must be in the recording state", 44}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-commonparent", 53}, { {"Both of commandBuffer, and pipeline must have been created, allocated, or retrieved from the same VkDevice", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-deviceGeneratedComputePipelines-09021", 78}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-pipeline-09019", 55}, { {"pipeline must have been created with VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV flag set", 89}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-pipeline-09020", 55}, { {"pipeline must have been created with VkComputePipelineIndirectBufferInfoNV structure specifying a valid address where its metadata will be saved", 144}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-pipeline-parameter", 59}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-pipelineBindPoint-09018", 64}, { {"pipelineBindPoint must be VK_PIPELINE_BIND_POINT_COMPUTE", 56}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-pipelineBindPoint-parameter", 68}, { {"pipelineBindPoint must be a valid VkPipelineBindPoint value", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-renderpass", 51}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-suspended", 50}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdUpdatePipelineIndirectBufferNV-videocoding", 52}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCmdWaitEvents-None-10655", 31}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-commandBuffer-01167", 40}, { {"commandBuffer's current device mask must include exactly one physical device", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-commandBuffer-cmdpool", 42}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-commandBuffer-parameter", 44}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-commandBuffer-recording", 44}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-commonparent", 33}, { {"Both of commandBuffer, and the elements of pEvents must have been created, allocated, or retrieved from the same VkDevice", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstAccessMask-02816", 40}, { {"The dstAccessMask member of each element of pMemoryBarriers must only include access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstAccessMask-06257", 40}, { {"If the rayQuery feature is not enabled and a memory barrier dstAccessMask includes VK_ACCESS_ACCELERATION_STRUCTURE_READ_BIT_KHR, dstStageMask must include VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, or VK_PIPELINE_STAGE_ALL_COMMANDS_BIT", 295}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-03937", 39}, { {"If the synchronization2 feature is not enabled, dstStageMask must not be 0", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04090", 39}, { {"If the geometryShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04091", 39}, { {"If the tessellationShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04092", 39}, { {"If the conditionalRendering feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04093", 39}, { {"If the fragmentDensityMap feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04094", 39}, { {"If the transformFeedback feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04095", 39}, { {"If the meshShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-04096", 39}, { {"If the taskShader feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-06460", 39}, { {"Any pipeline stage included in dstStageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 319}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-07318", 39}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-07949", 39}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, dstStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-10754", 39}, { {"If the accelerationStructure feature is not enabled, dstStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-dstStageMask-parameter", 43}, { {"dstStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-eventCount-arraylength", 43}, { {"eventCount must be greater than 0", 33}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-image-09373", 32}, { {"If vkCmdWaitEvents is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, the corresponding color attachment must be VK_ATTACHMENT_UNUSED", 231}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-image-09374", 32}, { {"If vkCmdWaitEvents is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, it must have been created with a non-zero VkExternalFormatANDROID::externalFormat value", 255}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-oldLayout-01181", 36}, { {"If vkCmdWaitEvents is called within a render pass instance, the oldLayout and newLayout members of any image memory barrier included in this command must be equal", 162}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pBufferMemoryBarriers-02817", 48}, { {"For each element of pBufferMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its srcQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its srcAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 435}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pBufferMemoryBarriers-02818", 48}, { {"For each element of pBufferMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its dstQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its dstAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 435}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pBufferMemoryBarriers-parameter", 52}, { {"If bufferMemoryBarrierCount is not 0, pBufferMemoryBarriers must be a valid pointer to an array of bufferMemoryBarrierCount valid VkBufferMemoryBarrier structures", 162}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pEvents-03847", 34}, { {"Elements of pEvents must not have been signaled by vkCmdSetEvent2", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pEvents-parameter", 38}, { {"pEvents must be a valid pointer to an array of eventCount valid VkEvent handles", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pImageMemoryBarriers-02819", 47}, { {"For each element of pImageMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its srcQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its srcAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 434}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pImageMemoryBarriers-02820", 47}, { {"For each element of pImageMemoryBarriers, if its srcQueueFamilyIndex and dstQueueFamilyIndex members are equal, or if its dstQueueFamilyIndex is the queue family index that was used to create the command pool that commandBuffer was allocated from, then its dstAccessMask member must only contain access flags that are supported by one or more of the pipeline stages in dstStageMask, as specified in the table of supported access types", 434}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pImageMemoryBarriers-parameter", 51}, { {"If imageMemoryBarrierCount is not 0, pImageMemoryBarriers must be a valid pointer to an array of imageMemoryBarrierCount valid VkImageMemoryBarrier structures", 158}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-pMemoryBarriers-parameter", 46}, { {"If memoryBarrierCount is not 0, pMemoryBarriers must be a valid pointer to an array of memoryBarrierCount valid VkMemoryBarrier structures", 138}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcAccessMask-02815", 40}, { {"The srcAccessMask member of each element of pMemoryBarriers must only include access flags that are supported by one or more of the pipeline stages in srcStageMask, as specified in the table of supported access types", 216}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcAccessMask-06257", 40}, { {"If the rayQuery feature is not enabled and a memory barrier srcAccessMask includes VK_ACCESS_ACCELERATION_STRUCTURE_READ_BIT_KHR, srcStageMask must include VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR, VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR, or VK_PIPELINE_STAGE_ALL_COMMANDS_BIT", 295}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcQueueFamilyIndex-01182", 46}, { {"If vkCmdWaitEvents is called within a render pass instance, the srcQueueFamilyIndex and dstQueueFamilyIndex members of any memory barrier included in this command must be equal", 176}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcQueueFamilyIndex-02803", 46}, { {"The srcQueueFamilyIndex and dstQueueFamilyIndex members of any element of pBufferMemoryBarriers or pImageMemoryBarriers must be equal", 133}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-01158", 39}, { {"srcStageMask must be the bitwise OR of the stageMask parameter used in previous calls to vkCmdSetEvent with any of the elements of pEvents and VK_PIPELINE_STAGE_HOST_BIT if any of the elements of pEvents was set using vkSetEvent", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-03937", 39}, { {"If the synchronization2 feature is not enabled, srcStageMask must not be 0", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04090", 39}, { {"If the geometryShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04091", 39}, { {"If the tessellationShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 185}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04092", 39}, { {"If the conditionalRendering feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 129}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04093", 39}, { {"If the fragmentDensityMap feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04094", 39}, { {"If the transformFeedback feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 123}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04095", 39}, { {"If the meshShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-04096", 39}, { {"If the taskShader feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 109}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-06459", 39}, { {"Any pipeline stage included in srcStageMask must be supported by the capabilities of the queue family specified by the queueFamilyIndex member of the VkCommandPoolCreateInfo structure that was used to create the VkCommandPool that commandBuffer was allocated from, as specified in the table of supported pipeline stages", 319}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-07308", 39}, { {"If this command is called inside a render pass instance, srcStageMask must not include VK_PIPELINE_STAGE_HOST_BIT", 113}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-07318", 39}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 182}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-07949", 39}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, srcStageMask must not contain VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 164}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-10754", 39}, { {"If the accelerationStructure feature is not enabled, srcStageMask must not contain VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 137}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-srcStageMask-parameter", 43}, { {"srcStageMask must be a valid combination of VkPipelineStageFlagBits values", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents-suspended", 30}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-None-10654", 32}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-commandBuffer-03846", 41}, { {"commandBuffer's current device mask must include exactly one physical device", 76}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-commandBuffer-cmdpool", 43}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 188}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-commandBuffer-parameter", 45}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-commandBuffer-recording", 45}, { {"commandBuffer must be in the recording state", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-commonparent", 34}, { {"Both of commandBuffer, and the elements of pEvents must have been created, allocated, or retrieved from the same VkDevice", 121}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-dependencyFlags-03844", 43}, { {"If this command is called inside a render pass instance, the srcStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfos must not include VK_PIPELINE_STAGE_2_HOST_BIT", 240}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-dependencyFlags-10394", 43}, { {"The dependencyFlags member of any element of pDependencyInfo must not include any of the following bits: VK_DEPENDENCY_BY_REGION_BIT VK_DEPENDENCY_DEVICE_GROUP_BIT VK_DEPENDENCY_VIEW_LOCAL_BIT VK_DEPENDENCY_FEEDBACK_LOOP_BIT_EXT", 228}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-dstStageMask-03843", 40}, { {"The dstStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfos must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 277}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-eventCount-arraylength", 44}, { {"eventCount must be greater than 0", 33}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-image-09373", 33}, { {"If vkCmdWaitEvents2 is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, the corresponding color attachment must be VK_ATTACHMENT_UNUSED", 232}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-image-09374", 33}, { {"If vkCmdWaitEvents2 is called within a render pass instance using a VkRenderPass object, and the image member of any image memory barrier is a color resolve attachment, it must have been created with a non-zero VkExternalFormatANDROID::externalFormat value", 256}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-maintenance8-10205", 40}, { {"If the maintenance8 feature is not enabled, the dependencyFlags members of any element of pDependencyInfos must not include VK_DEPENDENCY_QUEUE_FAMILY_OWNERSHIP_TRANSFER_USE_ALL_STAGES_BIT_KHR", 192}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-oldLayout-01181", 37}, { {"If vkCmdWaitEvents2 is called within a render pass instance, the oldLayout and newLayout members of any image memory barrier included in this command must be equal", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pDependencyInfos-parameter", 48}, { {"pDependencyInfos must be a valid pointer to an array of eventCount valid VkDependencyInfo structures", 100}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-03837", 35}, { {"Members of pEvents must not have been signaled by vkCmdSetEvent", 63}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-03839", 35}, { {"For each element i of pEvents, if that event is signaled by vkSetEvent, barriers in the ith element of pDependencyInfos must include only host operations in their first synchronization scope", 190}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-03840", 35}, { {"For each element i of pEvents, if barriers in the ith element of pDependencyInfos include only host operations, the ith element of pEvents must be signaled before vkCmdWaitEvents2 is executed", 191}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-03841", 35}, { {"For each element i of pEvents, if barriers in the ith element of pDependencyInfos do not include host operations, the ith element of pEvents must be signaled by a corresponding vkCmdSetEvent2 that occurred earlier in submission order", 233}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-10788", 35}, { {"For each element i of pEvents, if the dependencyFlags member of the ith element of pDependencyInfos does not include VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, and if that event is signaled by vkCmdSetEvent2, that command's dependencyInfo parameter must be exactly equal to the ith element of pDependencyInfos", 306}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-10789", 35}, { {"For each element i of pEvents, if the dependencyFlags member of the ith element of pDependencyInfos includes VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, that event must be signaled by vkCmdSetEvent2 with VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR", 238}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-10790", 35}, { {"For each element i of pEvents, if the dependencyFlags member of the ith element of pDependencyInfos includes VK_DEPENDENCY_ASYMMETRIC_EVENT_BIT_KHR, the union of srcStageMask members of all elements of pMemoryBarriers, pBufferMemoryBarriers, and pImageMemoryBarriers of the ith element of pDependencyInfos must equal pDependencyInfos->pMemoryBarriers[0].srcStageMask in the vkCmdSetEvent2 command", 396}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-pEvents-parameter", 39}, { {"pEvents must be a valid pointer to an array of eventCount valid VkEvent handles", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-srcQueueFamilyIndex-01182", 47}, { {"If vkCmdWaitEvents2 is called within a render pass instance, the srcQueueFamilyIndex and dstQueueFamilyIndex members of any memory barrier included in this command must be equal", 177}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-srcStageMask-03842", 40}, { {"The srcStageMask member of any element of the pMemoryBarriers, pBufferMemoryBarriers, or pImageMemoryBarriers members of pDependencyInfos must only include pipeline stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 277}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-suspended", 31}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWaitEvents2-synchronization2-03836", 44}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-accelerationStructure-08924", 78}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-accelerationStructureCount-arraylength", 89}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-accelerationStructures-03431", 79}, { {"All acceleration structures in pAccelerationStructures must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR if queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR", 209}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-buffer-03736", 63}, { {"Each element of pAccelerationStructures must be fully backed by physical memory", 79}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-commandBuffer-cmdpool", 72}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-commandBuffer-parameter", 74}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-commandBuffer-recording", 74}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-commonparent", 63}, { {"Each of commandBuffer, queryPool, and the elements of pAccelerationStructures must have been created, allocated, or retrieved from the same VkDevice", 148}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-11684", 80}, { {"All acceleration structures in pAccelerationStructures must have been constructed prior to the execution of this command on the device", 134}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-12425", 80}, { {"All acceleration structures in pAccelerationStructures must have been built with VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR if queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR", 218}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-parameter", 84}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount valid VkAccelerationStructureKHR handles", 130}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-query-04880", 62}, { {"The sum of firstQuery plus accelerationStructureCount must be less than or equal to the number of queries in queryPool", 118}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-queryPool-02493", 66}, { {"queryPool must have been created with a queryType matching queryType", 68}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-queryPool-02494", 66}, { {"The queries identified by queryPool and firstQuery must be unavailable", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-queryPool-parameter", 70}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-queryType-06742", 66}, { {"queryType must be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR, VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR, VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR", 262}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-queryType-parameter", 70}, { {"queryType must be a valid VkQueryType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-renderpass", 61}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-suspended", 60}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesKHR-videocoding", 62}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-accelerationStructure-03757", 77}, { {"accelerationStructure must be bound completely and contiguously to a single VkDeviceMemory object via vkBindAccelerationStructureMemoryNV", 137}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-accelerationStructureCount-arraylength", 88}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-commandBuffer-cmdpool", 71}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-commandBuffer-parameter", 73}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-commandBuffer-recording", 73}, { {"commandBuffer must be in the recording state", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-commonparent", 62}, { {"Each of commandBuffer, queryPool, and the elements of pAccelerationStructures must have been created, allocated, or retrieved from the same VkDevice", 148}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-pAccelerationStructures-04958", 79}, { {"All acceleration structures in pAccelerationStructures must have been built prior to the execution of this command", 114}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-pAccelerationStructures-06215", 79}, { {"All acceleration structures in pAccelerationStructures must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR if queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_NV", 208}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-pAccelerationStructures-11685", 79}, { {"All acceleration structures in pAccelerationStructures must have been constructed prior to the execution of this command on the device", 134}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-pAccelerationStructures-parameter", 83}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount valid VkAccelerationStructureNV handles", 129}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-queryPool-03755", 65}, { {"queryPool must have been created with a queryType matching queryType", 68}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-queryPool-03756", 65}, { {"The queries identified by queryPool and firstQuery must be unavailable", 70}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-queryPool-parameter", 69}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-queryType-06216", 65}, { {"queryType must be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_NV", 72}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-queryType-parameter", 69}, { {"queryType must be a valid VkQueryType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-renderpass", 60}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-suspended", 59}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteAccelerationStructuresPropertiesNV-videocoding", 61}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-commandBuffer-cmdpool", 53}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-commandBuffer-parameter", 55}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-commandBuffer-recording", 55}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-commonparent", 44}, { {"Both of commandBuffer, and dstBuffer must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-dstBuffer-03897", 47}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-dstBuffer-03898", 47}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-dstBuffer-parameter", 51}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-dstOffset-03896", 47}, { {"dstOffset must be less than or equal to the size of dstBuffer minus 4", 69}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-dstOffset-03899", 47}, { {"dstOffset must be a multiple of 4", 33}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03894", 43}, { {"stage must include only a single pipeline stage", 47}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03895", 43}, { {"stage must include only stages that are valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 145}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03929", 43}, { {"If the geometryShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 108}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03930", 43}, { {"If the tessellationShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 182}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03931", 43}, { {"If the conditionalRendering feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 124}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03932", 43}, { {"If the fragmentDensityMap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03933", 43}, { {"If the transformFeedback feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 118}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03934", 43}, { {"If the meshShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-03935", 43}, { {"If the taskShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 104}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-04957", 43}, { {"If the subpassShading feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 114}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-04995", 43}, { {"If the invocationMask feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 115}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-07316", 43}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 177}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-07946", 43}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stage must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 159}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-10751", 43}, { {"If the accelerationStructure feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-10752", 43}, { {"If the rayTracingMaintenance1 feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 132}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-11541", 43}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-stage-parameter", 47}, { {"stage must be a valid combination of VkPipelineStageFlagBits2 values", 68}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-suspended", 41}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-synchronization2-03893", 54}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarker2AMD-videocoding", 43}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-commandBuffer-cmdpool", 52}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-commandBuffer-parameter", 54}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-commandBuffer-recording", 54}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-commonparent", 43}, { {"Both of commandBuffer, and dstBuffer must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-dstBuffer-01799", 46}, { {"dstBuffer must have been created with the VK_BUFFER_USAGE_TRANSFER_DST_BIT usage flag set", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-dstBuffer-01800", 46}, { {"If dstBuffer is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 110}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-dstBuffer-parameter", 50}, { {"dstBuffer must be a valid VkBuffer handle", 41}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-dstOffset-01798", 46}, { {"dstOffset must be less than or equal to the size of dstBuffer minus 4", 69}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-dstOffset-01801", 46}, { {"dstOffset must be a multiple of 4", 33}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04074", 50}, { {"pipelineStage must be a valid stage for the queue family that was used to create the command pool that commandBuffer was allocated from", 135}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04075", 50}, { {"If the geometryShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 109}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04076", 50}, { {"If the tessellationShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 181}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04077", 50}, { {"If the conditionalRendering feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 125}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04078", 50}, { {"If the fragmentDensityMap feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 126}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04079", 50}, { {"If the transformFeedback feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 119}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-04080", 50}, { {"If the meshShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-07077", 50}, { {"If the taskShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 105}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-pipelineStage-parameter", 54}, { {"If pipelineStage is not 0, pipelineStage must be a valid VkPipelineStageFlagBits value", 86}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-rayTracingPipeline-07943", 55}, { {"If neither of the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, pipelineStage must not be VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 163}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-shadingRateImage-07314", 53}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, pipelineStage must not be VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 178}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-suspended", 40}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-synchronization2-06489", 53}, { {"If the synchronization2 feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_NONE", 96}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteBufferMarkerAMD-videocoding", 42}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-commandBuffer-13044", 52}, { {"If commandBuffer is an unprotected command buffer and protectedNoFault is not supported, pInfo->dstFlags must not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 158}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-commandBuffer-13045", 52}, { {"If commandBuffer is a protected command buffer and protectedNoFault is not supported, pInfo->dstFlags must include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR", 151}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-commandBuffer-cmdpool", 54}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, or VK_QUEUE_TRANSFER_BIT operations", 149}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-commandBuffer-parameter", 56}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-commandBuffer-recording", 56}, { {"commandBuffer must be in the recording state", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-pInfo-13043", 44}, { {"pInfo->stage must include only stages that are valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 152}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-pInfo-13046", 44}, { {"If pInfo->dstFlags includes VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR, the buffer from which pInfo->dstRange was queried must have been created with VK_BUFFER_CREATE_PROTECTED_BIT", 174}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-pInfo-13047", 44}, { {"If pInfo->dstFlags does not include VK_ADDRESS_COMMAND_PROTECTED_BIT_KHR, the buffer from which pInfo->dstRange was queried must have been created without VK_BUFFER_CREATE_PROTECTED_BIT", 185}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-pInfo-parameter", 48}, { {"pInfo must be a valid pointer to a valid VkMemoryMarkerInfoAMD structure", 72}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-suspended", 42}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-synchronization2-13042", 55}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMarkerToMemoryAMD-videocoding", 44}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-None-12417", 48}, { {"All queries used by this command must be unavailable", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-buffer-07527", 50}, { {"The buffer used to create each micromap in pMicromaps must be bound to device memory", 84}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-commandBuffer-cmdpool", 59}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT operations", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-commandBuffer-parameter", 61}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-commandBuffer-recording", 61}, { {"commandBuffer must be in the recording state", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-commonparent", 50}, { {"Each of commandBuffer, queryPool, and the elements of pMicromaps must have been created, allocated, or retrieved from the same VkDevice", 135}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-micromapCount-arraylength", 63}, { {"micromapCount must be greater than 0", 36}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-pMicromaps-07502", 54}, { {"All micromaps in pMicromaps must have been constructed with VK_BUILD_MICROMAP_ALLOW_COMPACTION_BIT_EXT if queryType is VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT", 160}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-pMicromaps-11662", 54}, { {"All micromaps in pMicromaps must have been constructed prior to the execution of this command on the device", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-pMicromaps-parameter", 58}, { {"pMicromaps must be a valid pointer to an array of micromapCount valid VkMicromapEXT handles", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-query-07528", 49}, { {"The sum of query plus micromapCount must be less than or equal to the number of queries in queryPool", 100}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-queryPool-07525", 53}, { {"queryPool must have been created with a queryType matching queryType", 68}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-queryPool-parameter", 57}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-queryType-07503", 53}, { {"queryType must be VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT, or VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT", 109}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-renderpass", 48}, { {"This command must only be called outside of a render pass instance", 66}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-suspended", 47}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteMicromapsPropertiesEXT-videocoding", 49}, { {"This command must only be called outside of a video coding scope", 64}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCmdWriteTimestamp-None-00830", 35}, { {"All queries used by the command must be unavailable", 51}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-None-10640", 35}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-commandBuffer-cmdpool", 46}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 241}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-commandBuffer-parameter", 48}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-commandBuffer-recording", 48}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-commonparent", 37}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04074", 44}, { {"pipelineStage must be a valid stage for the queue family that was used to create the command pool that commandBuffer was allocated from", 135}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04075", 44}, { {"If the geometryShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT", 109}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04076", 44}, { {"If the tessellationShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT", 181}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04077", 44}, { {"If the conditionalRendering feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT", 125}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04078", 44}, { {"If the fragmentDensityMap feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 126}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04079", 44}, { {"If the transformFeedback feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT", 119}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-04080", 44}, { {"If the meshShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_MESH_SHADER_BIT_EXT", 105}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-07077", 44}, { {"If the taskShader feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_TASK_SHADER_BIT_EXT", 105}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-pipelineStage-parameter", 48}, { {"pipelineStage must be a valid VkPipelineStageFlagBits value", 59}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-query-00831", 36}, { {"If vkCmdWriteTimestamp is called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 207}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-query-04904", 36}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-queryPool-01416", 40}, { {"queryPool must have been created with a queryType of VK_QUERY_TYPE_TIMESTAMP", 76}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-queryPool-parameter", 44}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-rayTracingPipeline-07943", 49}, { {"If neither of the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, pipelineStage must not be VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR", 163}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-shadingRateImage-07314", 47}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, pipelineStage must not be VK_PIPELINE_STAGE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 178}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-suspended", 34}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-synchronization2-06489", 47}, { {"If the synchronization2 feature is not enabled, pipelineStage must not be VK_PIPELINE_STAGE_NONE", 96}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp-timestampValidBits-00829", 49}, { {"The command pool's queue family must support a non-zero timestampValidBits", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-None-03864", 36}, { {"All queries used by the command must be unavailable", 51}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-None-10639", 36}, { {"This command must not be recorded when per-tile execution model is enabled", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-commandBuffer-cmdpool", 47}, { {"The VkCommandPool that commandBuffer was allocated from must support VK_QUEUE_COMPUTE_BIT, VK_QUEUE_GRAPHICS_BIT, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_VIDEO_DECODE_BIT_KHR, or VK_QUEUE_VIDEO_ENCODE_BIT_KHR operations", 211}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-commandBuffer-recording", 49}, { {"commandBuffer must be in the recording state", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-commonparent", 38}, { {"Both of commandBuffer, and queryPool must have been created, allocated, or retrieved from the same VkDevice", 107}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-query-03865", 37}, { {"If vkCmdWriteTimestamp2 is called within a render pass instance, the sum of query and the number of bits set in the current subpass's view mask must be less than or equal to the number of queries in queryPool", 208}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-query-04903", 37}, { {"query must be less than the number of queries in queryPool", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-queryPool-03861", 41}, { {"queryPool must have been created with a queryType of VK_QUERY_TYPE_TIMESTAMP", 76}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-queryPool-parameter", 45}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03859", 37}, { {"stage must only include a single pipeline stage", 47}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03860", 37}, { {"stage must only include stages valid for the queue family that was used to create the command pool that commandBuffer was allocated from", 136}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03929", 37}, { {"If the geometryShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT", 108}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03930", 37}, { {"If the tessellationShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT or VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT", 182}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03931", 37}, { {"If the conditionalRendering feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT", 124}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03932", 37}, { {"If the fragmentDensityMap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT", 125}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03933", 37}, { {"If the transformFeedback feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT", 118}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03934", 37}, { {"If the meshShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_EXT", 104}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-03935", 37}, { {"If the taskShader feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_EXT", 104}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-04957", 37}, { {"If the subpassShading feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI", 114}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-04995", 37}, { {"If the invocationMask feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI", 115}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-07316", 37}, { {"If neither of the shadingRateImage or the attachmentFragmentShadingRate features are enabled, stage must not contain VK_PIPELINE_STAGE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR", 177}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-07946", 37}, { {"If neither the VK_NV_ray_tracing extension or the rayTracingPipeline feature are enabled, stage must not contain VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_KHR", 159}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-10751", 37}, { {"If the accelerationStructure feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_KHR", 132}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-10752", 37}, { {"If the rayTracingMaintenance1 feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR", 132}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-11541", 37}, { {"If the VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature is not enabled, stage must not contain VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT", 149}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-stage-parameter", 41}, { {"stage must be a valid combination of VkPipelineStageFlagBits2 values", 68}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-suspended", 35}, { {"This command must not be called between suspended render pass instances", 71}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-synchronization2-03858", 48}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCmdWriteTimestamp2-timestampValidBits-03863", 50}, { {"The command pool's queue family must support a non-zero timestampValidBits", 74}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCompileDeferredNV-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCompileDeferredNV-pipeline-02237", 39}, { {"pipeline must have been created with VK_PIPELINE_CREATE_DEFER_COMPILE_BIT_NV", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCompileDeferredNV-pipeline-04621", 39}, { {"pipeline must be a ray tracing pipeline", 39}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCompileDeferredNV-pipeline-parameter", 43}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCompileDeferredNV-pipeline-parent", 40}, { {"pipeline must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCompileDeferredNV-shader-02238", 37}, { {"shader must not have been called as a deferred compile before", 61}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-pInfo-10073", 51}, { {"If pInfo->srcData.hostAddress is NULL, then pInfo->dstData.hostAddress must be NULL", 83}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-pInfo-10074", 51}, { {"If pInfo->srcData.hostAddress is not NULL, then pInfo->srcSize must be large enough to contain the source matrix, based either on the standard matrix layout or based on the size filled out by this command", 204}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-pInfo-10075", 51}, { {"If pInfo->dstData.hostAddress is not NULL, then the value pointed to by pInfo->pDstSize must be large enough to contain the destination matrix, based either on the standard matrix layout or based on the size filled out by this command", 234}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-pInfo-10076", 51}, { {"If pInfo->dstData.hostAddress is not NULL, the source and destination memory ranges must not overlap", 100}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkConvertCooperativeVectorMatrixNV-pInfo-parameter", 55}, { {"pInfo must be a valid pointer to a valid VkConvertCooperativeVectorMatrixInfoNV structure", 89}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCopyAccelerationStructureKHR-accelerationStructureHostCommands-03582", 75}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureHostCommands feature must be enabled", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-buffer-03727", 48}, { {"pInfo->src must be bound to host-visible device memory", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-buffer-03728", 48}, { {"pInfo->dst must be bound to host-visible device memory", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-buffer-03780", 48}, { {"pInfo->src must be bound to memory that was not allocated with multiple instances", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-buffer-03781", 48}, { {"pInfo->dst must be bound to memory that was not allocated with multiple instances", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-deferredOperation-03678", 59}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-deferredOperation-parameter", 63}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-deferredOperation-parent", 60}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-dst-11589", 45}, { {"pInfo->dst must have been created with vkCreateAccelerationStructureKHR", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-pInfo-parameter", 51}, { {"pInfo must be a valid pointer to a valid VkCopyAccelerationStructureInfoKHR structure", 85}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-src-11586", 45}, { {"The source acceleration structure pInfo->src must have been constructed prior to the execution of this command", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-src-11587", 45}, { {"If the source acceleration structure pInfo->src was constructed through deserialization, all micromap arrays the acceleration structure references that were not replaced by an acceleration structure update command must have been deserialized using the serialized data of the corresponding micromaps used to originally build the acceleration structure prior to the execution of this command", 389}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureKHR-src-11588", 45}, { {"pInfo->src must have been created with vkCreateAccelerationStructureKHR", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-accelerationStructureHostCommands-03584", 83}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureHostCommands feature must be enabled", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-buffer-03783", 56}, { {"pInfo->src must not be bound to memory that was allocated with multiple instances", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-deferredOperation-03678", 67}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-deferredOperation-parameter", 71}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-deferredOperation-parent", 68}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-pInfo-03731", 55}, { {"pInfo->src must be bound to host-visible device memory", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-pInfo-03732", 55}, { {"pInfo->dst.hostAddress must be a valid host pointer", 51}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-pInfo-03751", 55}, { {"pInfo->dst.hostAddress must be aligned to 16 bytes", 50}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-pInfo-parameter", 59}, { {"pInfo must be a valid pointer to a valid VkCopyAccelerationStructureToMemoryInfoKHR structure", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyAccelerationStructureToMemoryKHR-src-11678", 53}, { {"pInfo->src must have been created with vkCreateAccelerationStructureKHR", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyImageToImage-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyImageToImage-pCopyImageToImageInfo-parameter", 55}, { {"pCopyImageToImageInfo must be a valid pointer to a valid VkCopyImageToImageInfo structure", 89}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyImageToMemory-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyImageToMemory-pCopyImageToMemoryInfo-parameter", 57}, { {"pCopyImageToMemoryInfo must be a valid pointer to a valid VkCopyImageToMemoryInfo structure", 91}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-accelerationStructureHostCommands-03583", 83}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureHostCommands feature must be enabled", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-buffer-03730", 56}, { {"pInfo->dst must be bound to host-visible device memory", 54}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-buffer-03782", 56}, { {"pInfo->dst must be bound to memory that was not allocated with multiple instances", 81}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-deferredOperation-03678", 67}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-deferredOperation-parameter", 71}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-deferredOperation-parent", 68}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-dst-11677", 53}, { {"pInfo->dst must have been created with vkCreateAccelerationStructureKHR", 71}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-pInfo-03729", 55}, { {"pInfo->src.hostAddress must be a valid host pointer", 51}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-pInfo-03750", 55}, { {"pInfo->src.hostAddress must be aligned to 16 bytes", 50}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToAccelerationStructureKHR-pInfo-parameter", 59}, { {"pInfo must be a valid pointer to a valid VkCopyMemoryToAccelerationStructureInfoKHR structure", 93}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkCopyMemoryToImage-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyMemoryToImage-pCopyMemoryToImageInfo-parameter", 57}, { {"pCopyMemoryToImageInfo must be a valid pointer to a valid VkCopyMemoryToImageInfo structure", 91}, { "chapters/copies.html", 20} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-buffer-07565", 43}, { {"The buffer used to create pInfo->dst must be bound to host-visible device memory", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-buffer-07567", 43}, { {"The buffer used to create pInfo->dst must be bound to memory that was not allocated with multiple instances", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-deferredOperation-03678", 54}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-deferredOperation-parameter", 58}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-deferredOperation-parent", 55}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-micromapHostCommands-07566", 57}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapHostCommands feature must be enabled", 92}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-pInfo-07563", 42}, { {"pInfo->src.hostAddress must be a valid host pointer", 51}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-pInfo-07564", 42}, { {"pInfo->src.hostAddress must be aligned to 16 bytes", 50}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMemoryToMicromapEXT-pInfo-parameter", 46}, { {"pInfo must be a valid pointer to a valid VkCopyMemoryToMicromapInfoEXT structure", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-buffer-07558", 35}, { {"The buffer used to create pInfo->src must be bound to host-visible device memory", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-buffer-07559", 35}, { {"The buffer used to create pInfo->dst must be bound to host-visible device memory", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-buffer-07561", 35}, { {"The buffer used to create pInfo->src must be bound to memory that was not allocated with multiple instances", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-buffer-07562", 35}, { {"The buffer used to create pInfo->dst must be bound to memory that was not allocated with multiple instances", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-deferredOperation-03678", 46}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-deferredOperation-parameter", 50}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-deferredOperation-parent", 47}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-micromapHostCommands-07560", 49}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapHostCommands feature must be enabled", 92}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-pInfo-parameter", 38}, { {"pInfo must be a valid pointer to a valid VkCopyMicromapInfoEXT structure", 72}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapEXT-src-11670", 32}, { {"The source micromap src must have been constructed prior to the execution of this command", 89}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-buffer-07568", 43}, { {"The buffer used to create pInfo->src must be bound to host-visible device memory", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-buffer-07572", 43}, { {"The buffer used to create pInfo->src must be bound to memory that was not allocated with multiple instances", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-deferredOperation-03678", 54}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-deferredOperation-parameter", 58}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-deferredOperation-parent", 55}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-micromapHostCommands-07571", 57}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapHostCommands feature must be enabled", 92}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-pInfo-07569", 42}, { {"pInfo->dst.hostAddress must be a valid host pointer", 51}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-pInfo-07570", 42}, { {"pInfo->dst.hostAddress must be aligned to 16 bytes", 50}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCopyMicromapToMemoryEXT-pInfo-parameter", 46}, { {"pInfo must be a valid pointer to a valid VkCopyMicromapToMemoryInfoEXT structure", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-accelerationStructure-03611", 66}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-device-queuecount", 56}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-deviceAddressCommands-13086", 66}, { {"The VkPhysicalDeviceDeviceAddressCommandsFeaturesKHR::deviceAddressCommands feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-pAccelerationStructure-parameter", 71}, { {"pAccelerationStructure must be a valid pointer to a VkAccelerationStructureKHR handle", 85}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructure2KHR-pCreateInfo-parameter", 60}, { {"pCreateInfo must be a valid pointer to a valid VkAccelerationStructureCreateInfo2KHR structure", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-accelerationStructure-03611", 65}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-device-queuecount", 55}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-deviceAddress-03488", 57}, { {"If VkAccelerationStructureCreateInfoKHR::deviceAddress is not zero, the accelerationStructureCaptureReplay feature must be enabled", 130}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-pAccelerationStructure-parameter", 70}, { {"pAccelerationStructure must be a valid pointer to a VkAccelerationStructureKHR handle", 85}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureKHR-pCreateInfo-parameter", 59}, { {"pCreateInfo must be a valid pointer to a valid VkAccelerationStructureCreateInfoKHR structure", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureNV-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureNV-device-queuecount", 54}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureNV-pAccelerationStructure-parameter", 69}, { {"pAccelerationStructure must be a valid pointer to a VkAccelerationStructureNV handle", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureNV-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAccelerationStructureNV-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkAccelerationStructureCreateInfoNV structure", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateAndroidSurfaceKHR-instance-parameter", 49}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateAndroidSurfaceKHR-pAllocator-parameter", 51}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateAndroidSurfaceKHR-pCreateInfo-parameter", 52}, { {"pCreateInfo must be a valid pointer to a valid VkAndroidSurfaceCreateInfoKHR structure", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateAndroidSurfaceKHR-pSurface-parameter", 49}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateBuffer-device-09664", 32}, { {"device must support at least one queue family with one of the VK_QUEUE_VIDEO_ENCODE_BIT_KHR, VK_QUEUE_VIDEO_DECODE_BIT_KHR, VK_QUEUE_SPARSE_BINDING_BIT, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT capabilities", 235}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-device-parameter", 36}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-device-queuecount", 37}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-flags-00911", 31}, { {"If the flags member of pCreateInfo includes VK_BUFFER_CREATE_SPARSE_BINDING_BIT, and the extendedSparseAddressSpace feature is not enabled, creating this VkBuffer must not cause the total required sparse memory for all currently valid sparse resources on the device to exceed VkPhysicalDeviceLimits::sparseAddressSpaceSize", 322}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-flags-09383", 31}, { {"If the flags member of pCreateInfo includes VK_BUFFER_CREATE_SPARSE_BINDING_BIT, the extendedSparseAddressSpace feature is enabled, and the usage member of pCreateInfo contains bits not in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseBufferUsageFlags, creating this VkBuffer must not cause the total required sparse memory for all currently valid sparse resources on the device, excluding VkBuffer created with usage member of pCreateInfo containing bits in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseBufferUsageFlags and VkImage created with usage member of pCreateInfo containing bits in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseImageUsageFlags, to exceed VkPhysicalDeviceLimits::sparseAddressSpaceSize", 785}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-flags-09384", 31}, { {"If the flags member of pCreateInfo includes VK_BUFFER_CREATE_SPARSE_BINDING_BIT and the extendedSparseAddressSpace feature is enabled, creating this VkBuffer must not cause the total required sparse memory for all currently valid sparse resources on the device to exceed VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseAddressSpaceSize", 357}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-pAllocator-parameter", 40}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-pBuffer-parameter", 37}, { {"pBuffer must be a valid pointer to a VkBuffer handle", 52}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-pCreateInfo-parameter", 41}, { {"pCreateInfo must be a valid pointer to a valid VkBufferCreateInfo structure", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBuffer-pNext-06387", 31}, { {"If using the VkBuffer for an import operation from a VkBufferCollectionFUCHSIA where a VkBufferCollectionBufferCreateInfoFUCHSIA has been chained to pNext, pCreateInfo must match the VkBufferConstraintsInfoFUCHSIA::createInfo used when setting the constraints on the buffer collection with vkSetBufferCollectionBufferConstraintsFUCHSIA", 335}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferCollectionFUCHSIA-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferCollectionFUCHSIA-device-queuecount", 54}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferCollectionFUCHSIA-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferCollectionFUCHSIA-pCollection-parameter", 58}, { {"pCollection must be a valid pointer to a VkBufferCollectionFUCHSIA handle", 73}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferCollectionFUCHSIA-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkBufferCollectionCreateInfoFUCHSIA structure", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-device-09665", 36}, { {"device must support at least one queue family with one of the VK_QUEUE_COMPUTE_BIT or VK_QUEUE_GRAPHICS_BIT capabilities", 120}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-device-queuecount", 41}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-pCreateInfo-parameter", 45}, { {"pCreateInfo must be a valid pointer to a valid VkBufferViewCreateInfo structure", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateBufferView-pView-parameter", 39}, { {"pView must be a valid pointer to a VkBufferView handle", 54}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateCommandPool-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateCommandPool-device-queuecount", 42}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateCommandPool-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateCommandPool-pCommandPool-parameter", 47}, { {"pCommandPool must be a valid pointer to a VkCommandPool handle", 62}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateCommandPool-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkCommandPoolCreateInfo structure", 80}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateCommandPool-queueFamilyIndex-01937", 47}, { {"pCreateInfo->queueFamilyIndex must be the index of a queue family available in the logical device device", 104}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkCreateComputePipelines-binaryCount-09620", 47}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT must not be set in the flags of that element", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-binaryCount-09621", 47}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-binaryCount-09622", 47}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_EXT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-createInfoCount-arraylength", 57}, { {"createInfoCount must be greater than 0", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-device-09661", 42}, { {"device must support at least one queue family with the VK_QUEUE_COMPUTE_BIT capability", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-flags-00695", 41}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and the basePipelineIndex member of that same element is not -1, basePipelineIndex must be less than the index into pCreateInfos that corresponds to that element", 265}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-flags-00696", 41}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, the base pipeline must have been created with the VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT flag set", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pAllocator-parameter", 50}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pCreateInfos-11414", 48}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 290}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pCreateInfos-11429", 48}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pCreateInfos-parameter", 52}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkComputePipelineCreateInfo structures", 112}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pNext-09616", 41}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pNext-09617", 41}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure with the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag set is included in the pNext chain of any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 210}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pPipelines-parameter", 50}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pipelineCache-02873", 49}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pipelineCache-parameter", 53}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateComputePipelines-pipelineCache-parent", 50}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateCuFunctionNVX-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuFunctionNVX-device-queuecount", 44}, { {"The device must have been created with at least 1 queue", 55}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuFunctionNVX-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuFunctionNVX-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkCuFunctionCreateInfoNVX structure", 82}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuFunctionNVX-pFunction-parameter", 46}, { {"pFunction must be a valid pointer to a VkCuFunctionNVX handle", 61}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuModuleNVX-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuModuleNVX-device-queuecount", 42}, { {"The device must have been created with at least 1 queue", 55}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuModuleNVX-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuModuleNVX-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkCuModuleCreateInfoNVX structure", 80}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCuModuleNVX-pModule-parameter", 42}, { {"pModule must be a valid pointer to a VkCuModuleNVX handle", 57}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkCreateCudaFunctionNV-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaFunctionNV-device-queuecount", 45}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaFunctionNV-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaFunctionNV-pCreateInfo-parameter", 49}, { {"pCreateInfo must be a valid pointer to a valid VkCudaFunctionCreateInfoNV structure", 83}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaFunctionNV-pFunction-parameter", 47}, { {"pFunction must be a valid pointer to a VkCudaFunctionNV handle", 62}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaModuleNV-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaModuleNV-device-queuecount", 43}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaModuleNV-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaModuleNV-pCreateInfo-parameter", 47}, { {"pCreateInfo must be a valid pointer to a valid VkCudaModuleCreateInfoNV structure", 81}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateCudaModuleNV-pModule-parameter", 43}, { {"pModule must be a valid pointer to a VkCudaModuleNV handle", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateDataGraphPipelineSessionARM-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelineSessionARM-device-queuecount", 58}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelineSessionARM-pAllocator-parameter", 61}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelineSessionARM-pCreateInfo-parameter", 62}, { {"pCreateInfo must be a valid pointer to a valid VkDataGraphPipelineSessionCreateInfoARM structure", 96}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelineSessionARM-pSession-parameter", 59}, { {"pSession must be a valid pointer to a VkDataGraphPipelineSessionARM handle", 74}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-createInfoCount-arraylength", 62}, { {"createInfoCount must be greater than 0", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-dataGraph-09760", 50}, { {"The dataGraph feature must be enabled", 37}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-deferredOperation-09761", 58}, { {"deferredOperation must be VK_NULL_HANDLE", 40}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-deferredOperation-09916", 58}, { {"If deferredOperation is not VK_NULL_HANDLE, the flags member of elements of pCreateInfos must not include VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 152}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-deferredOperation-parameter", 62}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-deferredOperation-parent", 59}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-device-09927", 47}, { {"device must support at least one queue family with the VK_QUEUE_DATA_GRAPH_BIT_ARM capability", 93}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-device-queuecount", 52}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pAllocator-parameter", 55}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pCreateInfos-parameter", 57}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkDataGraphPipelineCreateInfoARM structures", 117}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pNext-09928", 46}, { {"If at least one of the VkDataGraphPipelineCreateInfoARM includes a VkDataGraphPipelineIdentifierCreateInfoARM structure in its pNext chain then pipelineCache must not be VK_NULL_HANDLE", 184}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pPipelines-parameter", 55}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pipelineCache-09762", 54}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pipelineCache-parameter", 58}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDataGraphPipelinesARM-pipelineCache-parent", 55}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkCreateDebugReportCallbackEXT-instance-parameter", 54}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugReportCallbackEXT-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugReportCallbackEXT-pCallback-parameter", 55}, { {"pCallback must be a valid pointer to a VkDebugReportCallbackEXT handle", 70}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugReportCallbackEXT-pCreateInfo-parameter", 57}, { {"pCreateInfo must be a valid pointer to a valid VkDebugReportCallbackCreateInfoEXT structure", 91}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugUtilsMessengerEXT-instance-parameter", 54}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugUtilsMessengerEXT-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugUtilsMessengerEXT-pCreateInfo-parameter", 57}, { {"pCreateInfo must be a valid pointer to a valid VkDebugUtilsMessengerCreateInfoEXT structure", 91}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDebugUtilsMessengerEXT-pMessenger-parameter", 56}, { {"pMessenger must be a valid pointer to a VkDebugUtilsMessengerEXT handle", 71}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkCreateDeferredOperationKHR-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkCreateDeferredOperationKHR-device-queuecount", 51}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkCreateDeferredOperationKHR-pAllocator-parameter", 54}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkCreateDeferredOperationKHR-pDeferredOperation-parameter", 62}, { {"pDeferredOperation must be a valid pointer to a VkDeferredOperationKHR handle", 77}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkCreateDescriptorPool-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorPool-device-queuecount", 45}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorPool-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorPool-pCreateInfo-parameter", 49}, { {"pCreateInfo must be a valid pointer to a valid VkDescriptorPoolCreateInfo structure", 83}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorPool-pDescriptorPool-parameter", 53}, { {"pDescriptorPool must be a valid pointer to a VkDescriptorPool handle", 68}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorSetLayout-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorSetLayout-pAllocator-parameter", 53}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorSetLayout-pCreateInfo-parameter", 54}, { {"pCreateInfo must be a valid pointer to a valid VkDescriptorSetLayoutCreateInfo structure", 88}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorSetLayout-pSetLayout-parameter", 53}, { {"pSetLayout must be a valid pointer to a VkDescriptorSetLayout handle", 68}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorSetLayout-support-09582", 46}, { {"If the descriptor layout exceeds the limits reported through the physical device limits, then vkGetDescriptorSetLayoutSupport must have returned VkDescriptorSetLayoutSupport with support equal to VK_TRUE for pCreateInfo", 219}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorUpdateTemplate-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorUpdateTemplate-device-queuecount", 55}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorUpdateTemplate-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorUpdateTemplate-pCreateInfo-parameter", 59}, { {"pCreateInfo must be a valid pointer to a valid VkDescriptorUpdateTemplateCreateInfo structure", 93}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDescriptorUpdateTemplate-pDescriptorUpdateTemplate-parameter", 73}, { {"pDescriptorUpdateTemplate must be a valid pointer to a VkDescriptorUpdateTemplate handle", 88}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreateDevice-pAllocator-parameter", 40}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkCreateDevice-pCreateInfo-parameter", 41}, { {"pCreateInfo must be a valid pointer to a valid VkDeviceCreateInfo structure", 75}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkCreateDevice-pDevice-parameter", 37}, { {"pDevice must be a valid pointer to a VkDevice handle", 52}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkCreateDevice-physicalDevice-parameter", 44}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkCreateDevice-ppEnabledExtensionNames-01387", 49}, { {"All required device extensions for each extension in the VkDeviceCreateInfo::ppEnabledExtensionNames list must also be present in that list", 139}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkCreateDirectFBSurfaceEXT-instance-parameter", 50}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDirectFBSurfaceEXT-pAllocator-parameter", 52}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDirectFBSurfaceEXT-pCreateInfo-parameter", 53}, { {"pCreateInfo must be a valid pointer to a valid VkDirectFBSurfaceCreateInfoEXT structure", 87}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDirectFBSurfaceEXT-pSurface-parameter", 50}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-display-parameter", 45}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-display-parent", 42}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-pCreateInfo-parameter", 49}, { {"pCreateInfo must be a valid pointer to a valid VkDisplayModeCreateInfoKHR structure", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-pMode-parameter", 43}, { {"pMode must be a valid pointer to a VkDisplayModeKHR handle", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayModeKHR-physicalDevice-parameter", 52}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayPlaneSurfaceKHR-instance-parameter", 54}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayPlaneSurfaceKHR-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayPlaneSurfaceKHR-pCreateInfo-parameter", 57}, { {"pCreateInfo must be a valid pointer to a valid VkDisplaySurfaceCreateInfoKHR structure", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateDisplayPlaneSurfaceKHR-pSurface-parameter", 54}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateEvent-device-09672", 31}, { {"device must support at least one queue family with one of the VK_QUEUE_VIDEO_ENCODE_BIT_KHR, VK_QUEUE_VIDEO_DECODE_BIT_KHR, VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT capabilities", 183}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-device-parameter", 35}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-device-queuecount", 36}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-events-04468", 31}, { {"If the VK_KHR_portability_subset extension is enabled, and VkPhysicalDevicePortabilitySubsetFeaturesKHR::events is VK_FALSE, then the implementation does not support events, and vkCreateEvent must not be used", 208}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-pAllocator-parameter", 39}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-pCreateInfo-parameter", 40}, { {"pCreateInfo must be a valid pointer to a valid VkEventCreateInfo structure", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateEvent-pEvent-parameter", 35}, { {"pEvent must be a valid pointer to a VkEvent handle", 50}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-binaryCount-09620", 58}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT must not be set in the flags of that element", 191}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-binaryCount-09621", 58}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT must not be set in the flags of that element", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-binaryCount-09622", 58}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_EXT must not be set in the flags of that element", 187}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-createInfoCount-arraylength", 68}, { {"createInfoCount must be greater than 0", 38}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-flags-09125", 52}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and the basePipelineIndex member of that same element is not -1, basePipelineIndex must be less than the index into pCreateInfos that corresponds to that element", 265}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-flags-09126", 52}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, the base pipeline must have been created with the VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT flag set", 203}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pAllocator-parameter", 61}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pCreateInfos-11414", 59}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 290}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pCreateInfos-11429", 59}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 277}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pCreateInfos-parameter", 63}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkExecutionGraphPipelineCreateInfoAMDX structures", 123}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pNext-09616", 52}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 118}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pNext-09617", 52}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure with the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag set is included in the pNext chain of any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 210}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pPipelines-parameter", 61}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pipelineCache-09127", 60}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pipelineCache-parameter", 64}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-pipelineCache-parent", 61}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExecutionGraphPipelinesAMDX-shaderEnqueue-09124", 60}, { {"The shaderEnqueue feature must be enabled", 41}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkCreateExternalComputeQueueNV-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkCreateExternalComputeQueueNV-device-queuecount", 53}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkCreateExternalComputeQueueNV-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkCreateExternalComputeQueueNV-pCreateInfo-parameter", 57}, { {"pCreateInfo must be a valid pointer to a valid VkExternalComputeQueueCreateInfoNV structure", 91}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkCreateExternalComputeQueueNV-pExternalQueue-parameter", 60}, { {"pExternalQueue must be a valid pointer to a VkExternalComputeQueueNV handle", 75}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkCreateFence-device-parameter", 35}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateFence-device-queuecount", 36}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateFence-pAllocator-parameter", 39}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateFence-pCreateInfo-parameter", 40}, { {"pCreateInfo must be a valid pointer to a valid VkFenceCreateInfo structure", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateFence-pFence-parameter", 35}, { {"pFence must be a valid pointer to a VkFence handle", 50}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateFramebuffer-device-10002", 37}, { {"device must support at least one queue family with the VK_QUEUE_GRAPHICS_BIT capability", 87}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-device-queuecount", 42}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-pCreateInfo-02777", 42}, { {"If pCreateInfo->flags does not include VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT, and attachmentCount is not 0, each element of pCreateInfo->pAttachments must have been created on device", 180}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkFramebufferCreateInfo structure", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateFramebuffer-pFramebuffer-parameter", 47}, { {"pFramebuffer must be a valid pointer to a VkFramebuffer handle", 62}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateGpaSessionAMD-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCreateGpaSessionAMD-device-queuecount", 44}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCreateGpaSessionAMD-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCreateGpaSessionAMD-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkGpaSessionCreateInfoAMD structure", 82}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCreateGpaSessionAMD-pGpaSession-parameter", 48}, { {"pGpaSession must be a valid pointer to a VkGpaSessionAMD handle", 63}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkCreateGraphicsPipelines-binaryCount-09620", 48}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT must not be set in the flags of that element", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-binaryCount-09621", 48}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-binaryCount-09622", 48}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_EXT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-createInfoCount-arraylength", 58}, { {"createInfoCount must be greater than 0", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-device-09662", 43}, { {"device must support at least one queue family with the VK_QUEUE_GRAPHICS_BIT capability", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-flags-00720", 42}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and the basePipelineIndex member of that same element is not -1, basePipelineIndex must be less than the index into pCreateInfos that corresponds to that element", 265}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-flags-00721", 42}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, the base pipeline must have been created with the VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT flag set", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pAllocator-parameter", 51}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pCreateInfos-11414", 49}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 290}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pCreateInfos-11429", 49}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pCreateInfos-parameter", 53}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkGraphicsPipelineCreateInfo structures", 113}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pNext-09616", 42}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pNext-09617", 42}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure with the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag set is included in the pNext chain of any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 210}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pPipelines-parameter", 51}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pipelineCache-02876", 50}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pipelineCache-parameter", 54}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateGraphicsPipelines-pipelineCache-parent", 51}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateHeadlessSurfaceEXT-instance-parameter", 50}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateHeadlessSurfaceEXT-pAllocator-parameter", 52}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateHeadlessSurfaceEXT-pCreateInfo-parameter", 53}, { {"pCreateInfo must be a valid pointer to a valid VkHeadlessSurfaceCreateInfoEXT structure", 87}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateHeadlessSurfaceEXT-pSurface-parameter", 50}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateIOSSurfaceMVK-instance-parameter", 45}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateIOSSurfaceMVK-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateIOSSurfaceMVK-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkIOSSurfaceCreateInfoMVK structure", 82}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateIOSSurfaceMVK-pSurface-parameter", 45}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateImage-device-09666", 31}, { {"device must support at least one queue family with one of the VK_QUEUE_VIDEO_ENCODE_BIT_KHR, VK_QUEUE_VIDEO_DECODE_BIT_KHR, VK_QUEUE_OPTICAL_FLOW_BIT_NV, VK_QUEUE_SPARSE_BINDING_BIT, VK_QUEUE_TRANSFER_BIT, VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT capabilities", 265}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-device-parameter", 35}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-device-queuecount", 36}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-flags-00939", 30}, { {"If the flags member of pCreateInfo includes VK_IMAGE_CREATE_SPARSE_BINDING_BIT, and the extendedSparseAddressSpace feature is not enabled, creating this VkImage must not cause the total required sparse memory for all currently valid sparse resources on the device to exceed VkPhysicalDeviceLimits::sparseAddressSpaceSize", 320}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-flags-09385", 30}, { {"If the flags member of pCreateInfo includes VK_IMAGE_CREATE_SPARSE_BINDING_BIT, the extendedSparseAddressSpace feature is enabled, and the usage member of pCreateInfo contains bits not in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseImageUsageFlags, creating this VkImage must not cause the total required sparse memory for all currently valid sparse resources on the device, excluding VkBuffer created with usage member of pCreateInfo containing bits in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseBufferUsageFlags and VkImage created with usage member of pCreateInfo containing bits in VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseImageUsageFlags, to exceed VkPhysicalDeviceLimits::sparseAddressSpaceSize", 782}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-flags-09386", 30}, { {"If the flags member of pCreateInfo includes VK_IMAGE_CREATE_SPARSE_BINDING_BIT and the extendedSparseAddressSpace feature is enabled, creating this VkImage must not cause the total required sparse memory for all currently valid sparse resources on the device to exceed VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV::extendedSparseAddressSpaceSize", 355}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-pAllocator-parameter", 39}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-pCreateInfo-parameter", 40}, { {"pCreateInfo must be a valid pointer to a valid VkImageCreateInfo structure", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-pImage-parameter", 35}, { {"pImage must be a valid pointer to a VkImage handle", 50}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImage-pNext-06389", 30}, { {"If a VkBufferCollectionImageCreateInfoFUCHSIA has been chained to pNext, pCreateInfo must match the Sysmem chosen VkImageCreateInfo excepting members VkImageCreateInfo::extent and VkImageCreateInfo::usage in the match criteria", 226}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImagePipeSurfaceFUCHSIA-instance-parameter", 55}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateImagePipeSurfaceFUCHSIA-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateImagePipeSurfaceFUCHSIA-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkImagePipeSurfaceCreateInfoFUCHSIA structure", 92}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateImagePipeSurfaceFUCHSIA-pSurface-parameter", 55}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateImageView-device-09667", 35}, { {"device must support at least one queue family with one of the VK_QUEUE_VIDEO_ENCODE_BIT_KHR, VK_QUEUE_VIDEO_DECODE_BIT_KHR, VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT capabilities", 183}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-device-queuecount", 40}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-image-09179", 34}, { {"VkImageViewCreateInfo::image must have been created from device", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-pCreateInfo-parameter", 44}, { {"pCreateInfo must be a valid pointer to a valid VkImageViewCreateInfo structure", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-pNext-12444", 34}, { {"The pNext chain of pCreateInfo must contain at most one of the structures VkImageViewUsageCreateInfo and VkImageViewUsage2CreateInfoKHR", 135}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateImageView-pView-parameter", 38}, { {"pView must be a valid pointer to a VkImageView handle", 53}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-device-queuecount", 56}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-deviceGeneratedCommands-11089", 68}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-pCreateInfo-parameter", 60}, { {"pCreateInfo must be a valid pointer to a valid VkIndirectCommandsLayoutCreateInfoEXT structure", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutEXT-pIndirectCommandsLayout-parameter", 72}, { {"pIndirectCommandsLayout must be a valid pointer to a VkIndirectCommandsLayoutEXT handle", 87}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-device-queuecount", 55}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-deviceGeneratedCommands-02929", 67}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-pCreateInfo-parameter", 59}, { {"pCreateInfo must be a valid pointer to a valid VkIndirectCommandsLayoutCreateInfoNV structure", 93}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectCommandsLayoutNV-pIndirectCommandsLayout-parameter", 71}, { {"pIndirectCommandsLayout must be a valid pointer to a VkIndirectCommandsLayoutNV handle", 86}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-device-queuecount", 54}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-deviceGeneratedCommands-11013", 66}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT::deviceGeneratedCommands feature must be enabled", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkIndirectExecutionSetCreateInfoEXT structure", 92}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateIndirectExecutionSetEXT-pIndirectExecutionSet-parameter", 68}, { {"pIndirectExecutionSet must be a valid pointer to a VkIndirectExecutionSetEXT handle", 83}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkCreateInstance-pAllocator-parameter", 42}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkCreateInstance-pCreateInfo-parameter", 43}, { {"pCreateInfo must be a valid pointer to a valid VkInstanceCreateInfo structure", 77}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkCreateInstance-pInstance-parameter", 41}, { {"pInstance must be a valid pointer to a VkInstance handle", 56}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkCreateInstance-ppEnabledExtensionNames-01388", 51}, { {"All required extensions for each extension in the VkInstanceCreateInfo::ppEnabledExtensionNames list must also be present in that list", 134}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkCreateMacOSSurfaceMVK-instance-parameter", 47}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMacOSSurfaceMVK-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMacOSSurfaceMVK-pCreateInfo-parameter", 50}, { {"pCreateInfo must be a valid pointer to a valid VkMacOSSurfaceCreateInfoMVK structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMacOSSurfaceMVK-pSurface-parameter", 47}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMetalSurfaceEXT-instance-parameter", 47}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMetalSurfaceEXT-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMetalSurfaceEXT-pCreateInfo-parameter", 50}, { {"pCreateInfo must be a valid pointer to a valid VkMetalSurfaceCreateInfoEXT structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMetalSurfaceEXT-pSurface-parameter", 47}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateMicromapEXT-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateMicromapEXT-device-queuecount", 42}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateMicromapEXT-micromap-11615", 39}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateMicromapEXT-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateMicromapEXT-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkMicromapCreateInfoEXT structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateMicromapEXT-pMicromap-parameter", 44}, { {"pMicromap must be a valid pointer to a VkMicromapEXT handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateOpticalFlowSessionNV-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCreateOpticalFlowSessionNV-device-queuecount", 51}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCreateOpticalFlowSessionNV-pAllocator-parameter", 54}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCreateOpticalFlowSessionNV-pCreateInfo-parameter", 55}, { {"pCreateInfo must be a valid pointer to a valid VkOpticalFlowSessionCreateInfoNV structure", 89}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCreateOpticalFlowSessionNV-pSession-parameter", 52}, { {"pSession must be a valid pointer to a VkOpticalFlowSessionNV handle", 67}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkCreatePipelineBinariesKHR-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineBinariesKHR-pAllocator-parameter", 53}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineBinariesKHR-pBinaries-parameter", 52}, { {"pBinaries must be a valid pointer to a VkPipelineBinaryHandlesInfoKHR structure", 79}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineBinariesKHR-pCreateInfo-parameter", 54}, { {"pCreateInfo must be a valid pointer to a valid VkPipelineBinaryCreateInfoKHR structure", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineCache-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineCache-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineCache-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkPipelineCacheCreateInfo structure", 82}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineCache-pPipelineCache-parameter", 51}, { {"pPipelineCache must be a valid pointer to a VkPipelineCache handle", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreatePipelineLayout-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreatePipelineLayout-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreatePipelineLayout-pCreateInfo-parameter", 49}, { {"pCreateInfo must be a valid pointer to a valid VkPipelineLayoutCreateInfo structure", 83}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreatePipelineLayout-pPipelineLayout-parameter", 53}, { {"pPipelineLayout must be a valid pointer to a VkPipelineLayout handle", 68}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkCreatePrivateDataSlot-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreatePrivateDataSlot-device-queuecount", 46}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreatePrivateDataSlot-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreatePrivateDataSlot-pCreateInfo-parameter", 50}, { {"pCreateInfo must be a valid pointer to a valid VkPrivateDataSlotCreateInfo structure", 84}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreatePrivateDataSlot-pPrivateDataSlot-parameter", 55}, { {"pPrivateDataSlot must be a valid pointer to a VkPrivateDataSlot handle", 70}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreatePrivateDataSlot-privateData-04564", 46}, { {"The privateData feature must be enabled", 39}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkCreateQueryPool-device-09663", 35}, { {"device must support at least one queue family with one of the VK_QUEUE_VIDEO_ENCODE_BIT_KHR, VK_QUEUE_VIDEO_DECODE_BIT_KHR, VK_QUEUE_COMPUTE_BIT, or VK_QUEUE_GRAPHICS_BIT capabilities", 183}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateQueryPool-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateQueryPool-device-queuecount", 40}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateQueryPool-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateQueryPool-pCreateInfo-parameter", 44}, { {"pCreateInfo must be a valid pointer to a valid VkQueryPoolCreateInfo structure", 78}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateQueryPool-pQueryPool-parameter", 43}, { {"pQueryPool must be a valid pointer to a VkQueryPool handle", 58}, { "chapters/queries.html", 21} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-binaryCount-09620", 53}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT must not be set in the flags of that element", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-binaryCount-09621", 53}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-binaryCount-09622", 53}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_EXT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-createInfoCount-arraylength", 63}, { {"createInfoCount must be greater than 0", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-deferredOperation-03587", 59}, { {"If deferredOperation is not VK_NULL_HANDLE, the flags member of elements of pCreateInfos must not include VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT", 152}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-deferredOperation-03678", 59}, { {"Any previous deferred operation that was associated with deferredOperation must be complete", 91}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-deferredOperation-parameter", 63}, { {"If deferredOperation is not VK_NULL_HANDLE, deferredOperation must be a valid VkDeferredOperationKHR handle", 107}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-deferredOperation-parent", 60}, { {"If deferredOperation is a valid handle, it must have been created, allocated, or retrieved from device", 102}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-device-09677", 48}, { {"device must support at least one queue family with the VK_QUEUE_COMPUTE_BIT capability", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-flags-03415", 47}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and the basePipelineIndex member of that same element is not -1, basePipelineIndex must be less than the index into pCreateInfos that corresponds to that element", 265}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-flags-03416", 47}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, the base pipeline must have been created with the VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT flag set", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-flags-03816", 47}, { {"flags must not contain the VK_PIPELINE_CREATE_DISPATCH_BASE_BIT flag", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pCreateInfos-11414", 54}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 290}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pCreateInfos-11429", 54}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pCreateInfos-parameter", 58}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkRayTracingPipelineCreateInfoKHR structures", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pNext-09616", 47}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pNext-09617", 47}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure with the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag set is included in the pNext chain of any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 210}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pPipelines-parameter", 56}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pipelineCache-02903", 55}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pipelineCache-parameter", 59}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-pipelineCache-parent", 56}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesKHR-rayTracingPipeline-03586", 60}, { {"The rayTracingPipeline feature must be enabled", 46}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-binaryCount-09620", 52}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT must not be set in the flags of that element", 191}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-binaryCount-09621", 52}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-binaryCount-09622", 52}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_EXT must not be set in the flags of that element", 187}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-createInfoCount-arraylength", 62}, { {"createInfoCount must be greater than 0", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-device-09677", 47}, { {"device must support at least one queue family with the VK_QUEUE_COMPUTE_BIT capability", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-flags-03415", 46}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, and the basePipelineIndex member of that same element is not -1, basePipelineIndex must be less than the index into pCreateInfos that corresponds to that element", 265}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-flags-03416", 46}, { {"If the flags member of any element of pCreateInfos contains the VK_PIPELINE_CREATE_DERIVATIVE_BIT flag, the base pipeline must have been created with the VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT flag set", 203}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-flags-03816", 46}, { {"flags must not contain the VK_PIPELINE_CREATE_DISPATCH_BASE_BIT flag", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pAllocator-parameter", 55}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pCreateInfos-11414", 53}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 290}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pCreateInfos-11429", 53}, { {"If any element of pCreateInfos sets VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 277}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pCreateInfos-parameter", 57}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkRayTracingPipelineCreateInfoNV structures", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pNext-09616", 46}, { {"If VkPipelineBinaryInfoKHR::binaryCount is not 0 for any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 118}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pNext-09617", 46}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure with the VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag set is included in the pNext chain of any element of pCreateInfos, pipelineCache must be VK_NULL_HANDLE", 210}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pNext-10150", 46}, { {"If a VkPipelineCreateFlags2CreateInfoKHR structure is included in the pNext chain of any element of pCreateInfos, VK_PIPELINE_CREATE_2_CAPTURE_DATA_BIT_KHR flag must not be set", 176}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pPipelines-parameter", 55}, { {"pPipelines must be a valid pointer to an array of createInfoCount VkPipeline handles", 84}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pipelineCache-02903", 54}, { {"If pipelineCache was created with VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT, host access to pipelineCache must be externally synchronized", 148}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pipelineCache-parameter", 58}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRayTracingPipelinesNV-pipelineCache-parent", 55}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkCreateRenderPass-device-10000", 36}, { {"device must support at least one queue family with the VK_QUEUE_GRAPHICS_BIT capability", 87}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass-flags-10646", 35}, { {"VkRenderPassTileShadingCreateInfoQCOM::flags must not include VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass-pCreateInfo-parameter", 45}, { {"pCreateInfo must be a valid pointer to a valid VkRenderPassCreateInfo structure", 79}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass-pRenderPass-parameter", 45}, { {"pRenderPass must be a valid pointer to a VkRenderPass handle", 60}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-device-10001", 37}, { {"device must support at least one queue family with the VK_QUEUE_GRAPHICS_BIT capability", 87}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-flags-10649", 36}, { {"VkRenderPassTileShadingCreateInfoQCOM::flags must not include VK_TILE_SHADING_RENDER_PASS_PER_TILE_EXECUTION_BIT_QCOM", 117}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkRenderPassCreateInfo2 structure", 80}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateRenderPass2-pRenderPass-parameter", 46}, { {"pRenderPass must be a valid pointer to a VkRenderPass handle", 60}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkCreateSampler-device-09668", 33}, { {"device must support at least one queue family with one of the VK_QUEUE_COMPUTE_BIT or VK_QUEUE_GRAPHICS_BIT capabilities", 120}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-device-parameter", 37}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-maxSamplerAllocationCount-04110", 52}, { {"There must be less than VkPhysicalDeviceLimits::maxSamplerAllocationCount VkSampler objects currently created on the device", 123}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-maxSamplerAllocationCount-11412", 52}, { {"If there are any pipelines or shaders with embedded samplers currently created on the device, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 263}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-pAllocator-parameter", 41}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-pCreateInfo-parameter", 42}, { {"pCreateInfo must be a valid pointer to a valid VkSamplerCreateInfo structure", 76}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSampler-pSampler-parameter", 39}, { {"pSampler must be a valid pointer to a VkSampler handle", 54}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSamplerYcbcrConversion-None-01648", 46}, { {"The samplerYcbcrConversion feature must be enabled", 50}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSamplerYcbcrConversion-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSamplerYcbcrConversion-pAllocator-parameter", 56}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSamplerYcbcrConversion-pCreateInfo-parameter", 57}, { {"pCreateInfo must be a valid pointer to a valid VkSamplerYcbcrConversionCreateInfo structure", 91}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateSamplerYcbcrConversion-pYcbcrConversion-parameter", 62}, { {"pYcbcrConversion must be a valid pointer to a VkSamplerYcbcrConversion handle", 77}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkCreateScreenSurfaceQNX-instance-parameter", 48}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateScreenSurfaceQNX-pAllocator-parameter", 50}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateScreenSurfaceQNX-pCreateInfo-parameter", 51}, { {"pCreateInfo must be a valid pointer to a valid VkScreenSurfaceCreateInfoQNX structure", 85}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateScreenSurfaceQNX-pSurface-parameter", 48}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSemaphore-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateSemaphore-device-queuecount", 40}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateSemaphore-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateSemaphore-pCreateInfo-parameter", 44}, { {"pCreateInfo must be a valid pointer to a valid VkSemaphoreCreateInfo structure", 78}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateSemaphore-pSemaphore-parameter", 43}, { {"pSemaphore must be a valid pointer to a VkSemaphore handle", 58}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkCreateShaderInstrumentationARM-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderInstrumentationARM-device-queuecount", 55}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderInstrumentationARM-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderInstrumentationARM-pCreateInfo-parameter", 59}, { {"pCreateInfo must be a valid pointer to a valid VkShaderInstrumentationCreateInfoARM structure", 93}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderInstrumentationARM-pInstrumentation-parameter", 64}, { {"pInstrumentation must be a valid pointer to a VkShaderInstrumentationARM handle", 79}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderModule-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderModule-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderModule-pCreateInfo-06904", 43}, { {"If pCreateInfo is not NULL, pCreateInfo->pNext must be NULL or a pointer to a valid instance of VkShaderModuleValidationCacheCreateInfoEXT VkValidationFeaturesEXT", 162}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderModule-pCreateInfo-parameter", 47}, { {"pCreateInfo must be a valid pointer to a valid VkShaderModuleCreateInfo structure", 81}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShaderModule-pShaderModule-parameter", 49}, { {"pShaderModule must be a valid pointer to a VkShaderModule handle", 64}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-None-08400", 34}, { {"The shaderObject feature must be enabled", 40}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-createInfoCount-arraylength", 51}, { {"createInfoCount must be greater than 0", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-flags-11472", 35}, { {"If the flags member of any element of pCreateInfos includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT and VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, the flags member of all other elements of pCreateInfos whose stage is VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, VK_SHADER_STAGE_TASK_BIT_EXT, VK_SHADER_STAGE_MESH_BIT_EXT, or VK_SHADER_STAGE_FRAGMENT_BIT must also include VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 507}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08402", 42}, { {"If the flags member of any element of pCreateInfos includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, the flags member of all other elements of pCreateInfos whose stage is VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, or VK_SHADER_STAGE_FRAGMENT_BIT must also include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT", 397}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08403", 42}, { {"If the flags member of any element of pCreateInfos includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, the flags member of all other elements of pCreateInfos whose stage is VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT must also include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT", 281}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08404", 42}, { {"If the flags member of any element of pCreateInfos whose stage is VK_SHADER_STAGE_TASK_BIT_EXT or VK_SHADER_STAGE_MESH_BIT_EXT includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, there must be no member of pCreateInfos whose flags member includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT and whose stage is any of VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, or VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT", 449}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08405", 42}, { {"If there is any element of pCreateInfos whose stage is VK_SHADER_STAGE_MESH_BIT_EXT and whose flags member includes both VK_SHADER_CREATE_LINK_STAGE_BIT_EXT and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT, there must be no element of pCreateInfos whose stage is VK_SHADER_STAGE_TASK_BIT_EXT and whose flags member includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT", 354}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08409", 42}, { {"For each element of pCreateInfos whose flags member includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, if there is any other element of pCreateInfos whose stage is logically later than the stage of the former and whose flags member also includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, the nextStage of the former must be equal to the stage of the element with the logically earliest stage following the stage of the former whose flags member also includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT", 484}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08410", 42}, { {"The stage member of each element of pCreateInfos whose flags member includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT must be unique", 127}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08411", 42}, { {"The codeType member of all elements of pCreateInfos whose flags member includes VK_SHADER_CREATE_LINK_STAGE_BIT_EXT must be the same", 132}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08867", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and both stages contains an OpExecutionMode instruction specifying the type of subdivision, they must be the same", 371}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08868", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and both stages contains an OpExecutionMode instruction specifying the orientation of triangles, they must be the same", 376}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08870", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and both stages contains an OpExecutionMode instruction specifying the spacing of segments on the edges of tessellated primitives, they must be the same", 410}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-08871", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, and both stages contains an OpExecutionMode instruction specifying the output patch size, they must be the same", 369}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-09632", 42}, { {"If pCreateInfos contains a VK_SHADER_STAGE_MESH_BIT_EXT with codeType of VK_SHADER_CODE_TYPE_SPIRV_EXT and VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT is not set, then the mesh shader's entry point must not declare a variable with a DrawIndex BuiltIn decoration", 257}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-11413", 42}, { {"If any element of pCreateInfos sets VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, there must be less than (maxSamplerAllocationCount - (minSamplerHeapReservedRangeWithEmbedded / samplerDescriptorSize)) VkSampler objects currently created on the device", 286}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-11428", 42}, { {"If any element of pCreateInfos sets VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT and includes embedded sampler mappings, this command must not cause the total number of unique embedded samplers in pipelines and shaders on this device to exceed maxDescriptorHeapEmbeddedSamplers", 273}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-12224", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, at least one must contain an OpExecutionMode instruction specifying the orientation of triangles generated by the tessellator", 383}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-12225", 42}, { {"If pCreateInfos contains elements with both VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT and VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, both elements' flags include VK_SHADER_CREATE_LINK_STAGE_BIT_EXT, both elements' codeType is VK_SHADER_CODE_TYPE_SPIRV_EXT, at least one must contain an OpExecutionMode instruction specifying the spacing of segments on the edges of tessellated primitives", 388}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pCreateInfos-parameter", 46}, { {"pCreateInfos must be a valid pointer to an array of createInfoCount valid VkShaderCreateInfoEXT structures", 106}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-pShaders-parameter", 42}, { {"pShaders must be a valid pointer to an array of createInfoCount VkShaderEXT handles", 83}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-stage-09670", 35}, { {"If the stage member of any element of pCreateInfos is VK_SHADER_STAGE_COMPUTE_BIT, device must support at least one queue family with the VK_QUEUE_COMPUTE_BIT capability", 169}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateShadersEXT-stage-09671", 35}, { {"If the stage member of any element of pCreateInfos is VK_SHADER_STAGE_TASK_BIT_EXT, VK_SHADER_STAGE_MESH_BIT_EXT, VK_SHADER_STAGE_VERTEX_BIT, VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT, VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT, VK_SHADER_STAGE_GEOMETRY_BIT, or VK_SHADER_STAGE_FRAGMENT_BIT, device must support at least one queue family with the VK_QUEUE_GRAPHICS_BIT capability", 379}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-device-queuecount", 50}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-pAllocator-parameter", 53}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-pCreateInfos-parameter", 55}, { {"pCreateInfos must be a valid pointer to an array of swapchainCount valid VkSwapchainCreateInfoKHR structures", 108}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-pSwapchains-parameter", 54}, { {"pSwapchains must be a valid pointer to an array of swapchainCount VkSwapchainKHR handles", 88}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSharedSwapchainsKHR-swapchainCount-arraylength", 59}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateStreamDescriptorSurfaceGGP-instance-parameter", 58}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateStreamDescriptorSurfaceGGP-pAllocator-parameter", 60}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateStreamDescriptorSurfaceGGP-pCreateInfo-parameter", 61}, { {"pCreateInfo must be a valid pointer to a valid VkStreamDescriptorSurfaceCreateInfoGGP structure", 95}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateStreamDescriptorSurfaceGGP-pSurface-parameter", 58}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSurfaceOHOS-instance-parameter", 43}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSurfaceOHOS-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSurfaceOHOS-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkSurfaceCreateInfoOHOS structure", 80}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSurfaceOHOS-pSurface-parameter", 43}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSwapchainKHR-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSwapchainKHR-device-queuecount", 43}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSwapchainKHR-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSwapchainKHR-pCreateInfo-parameter", 47}, { {"pCreateInfo must be a valid pointer to a valid VkSwapchainCreateInfoKHR structure", 81}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateSwapchainKHR-pSwapchain-parameter", 46}, { {"pSwapchain must be a valid pointer to a VkSwapchainKHR handle", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateTensorARM-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorARM-device-queuecount", 40}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorARM-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorARM-pCreateInfo-parameter", 44}, { {"pCreateInfo must be a valid pointer to a valid VkTensorCreateInfoARM structure", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorARM-pTensor-parameter", 40}, { {"pTensor must be a valid pointer to a VkTensorARM handle", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorARM-tensors-09832", 36}, { {"The tensors feature must be enabled", 35}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorViewARM-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorViewARM-device-queuecount", 44}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorViewARM-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorViewARM-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkTensorViewCreateInfoARM structure", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateTensorViewARM-pView-parameter", 42}, { {"pView must be a valid pointer to a VkTensorViewARM handle", 57}, { "chapters/resources.html", 23} } },
        { {"VUID-vkCreateUbmSurfaceSEC-instance-parameter", 45}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateUbmSurfaceSEC-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateUbmSurfaceSEC-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkUbmSurfaceCreateInfoSEC structure", 82}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateUbmSurfaceSEC-pSurface-parameter", 45}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateValidationCacheEXT-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateValidationCacheEXT-device-queuecount", 49}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateValidationCacheEXT-pAllocator-parameter", 52}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateValidationCacheEXT-pCreateInfo-parameter", 53}, { {"pCreateInfo must be a valid pointer to a valid VkValidationCacheCreateInfoEXT structure", 87}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateValidationCacheEXT-pValidationCache-parameter", 58}, { {"pValidationCache must be a valid pointer to a VkValidationCacheEXT handle", 73}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkCreateViSurfaceNN-instance-parameter", 43}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateViSurfaceNN-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateViSurfaceNN-pCreateInfo-parameter", 46}, { {"pCreateInfo must be a valid pointer to a valid VkViSurfaceCreateInfoNN structure", 80}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateViSurfaceNN-pSurface-parameter", 43}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateVideoSessionKHR-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionKHR-device-queuecount", 46}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionKHR-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionKHR-pCreateInfo-parameter", 50}, { {"pCreateInfo must be a valid pointer to a valid VkVideoSessionCreateInfoKHR structure", 84}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionKHR-pVideoSession-parameter", 52}, { {"pVideoSession must be a valid pointer to a VkVideoSessionKHR handle", 67}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionParametersKHR-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionParametersKHR-device-queuecount", 56}, { {"The device must have been created with at least 1 queue", 55}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionParametersKHR-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionParametersKHR-pCreateInfo-parameter", 60}, { {"pCreateInfo must be a valid pointer to a valid VkVideoSessionParametersCreateInfoKHR structure", 94}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateVideoSessionParametersKHR-pVideoSessionParameters-parameter", 72}, { {"pVideoSessionParameters must be a valid pointer to a VkVideoSessionParametersKHR handle", 87}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkCreateWaylandSurfaceKHR-instance-parameter", 49}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWaylandSurfaceKHR-pAllocator-parameter", 51}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWaylandSurfaceKHR-pCreateInfo-parameter", 52}, { {"pCreateInfo must be a valid pointer to a valid VkWaylandSurfaceCreateInfoKHR structure", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWaylandSurfaceKHR-pSurface-parameter", 49}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWin32SurfaceKHR-instance-parameter", 47}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWin32SurfaceKHR-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWin32SurfaceKHR-pCreateInfo-parameter", 50}, { {"pCreateInfo must be a valid pointer to a valid VkWin32SurfaceCreateInfoKHR structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateWin32SurfaceKHR-pSurface-parameter", 47}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXcbSurfaceKHR-instance-parameter", 45}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXcbSurfaceKHR-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXcbSurfaceKHR-pCreateInfo-parameter", 48}, { {"pCreateInfo must be a valid pointer to a valid VkXcbSurfaceCreateInfoKHR structure", 82}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXcbSurfaceKHR-pSurface-parameter", 45}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXlibSurfaceKHR-instance-parameter", 46}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXlibSurfaceKHR-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXlibSurfaceKHR-pCreateInfo-parameter", 49}, { {"pCreateInfo must be a valid pointer to a valid VkXlibSurfaceCreateInfoKHR structure", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkCreateXlibSurfaceKHR-pSurface-parameter", 46}, { {"pSurface must be a valid pointer to a VkSurfaceKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDebugMarkerSetObjectNameEXT-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugMarkerSetObjectNameEXT-pNameInfo-parameter", 54}, { {"pNameInfo must be a valid pointer to a valid VkDebugMarkerObjectNameInfoEXT structure", 85}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugMarkerSetObjectTagEXT-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugMarkerSetObjectTagEXT-pTagInfo-parameter", 52}, { {"pTagInfo must be a valid pointer to a valid VkDebugMarkerObjectTagInfoEXT structure", 83}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-flags-parameter", 44}, { {"flags must be a valid combination of VkDebugReportFlagBitsEXT values", 68}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-flags-requiredbitmask", 50}, { {"flags must not be 0", 19}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-instance-parameter", 47}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-object-01241", 41}, { {"object must be a Vulkan object or VK_NULL_HANDLE", 48}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-objectType-01498", 45}, { {"If objectType is not VK_DEBUG_REPORT_OBJECT_TYPE_UNKNOWN_EXT and object is not VK_NULL_HANDLE, object must be a Vulkan object of the corresponding type associated with objectType as defined in VkDebugReportObjectTypeEXT and Vulkan Handle Relationship", 250}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-objectType-parameter", 49}, { {"objectType must be a valid VkDebugReportObjectTypeEXT value", 59}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-pLayerPrefix-parameter", 51}, { {"pLayerPrefix must be a null-terminated UTF-8 string", 51}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDebugReportMessageEXT-pMessage-parameter", 47}, { {"pMessage must be a null-terminated UTF-8 string", 47}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDeferredOperationJoinKHR-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDeferredOperationJoinKHR-operation-parameter", 51}, { {"operation must be a valid VkDeferredOperationKHR handle", 55}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDeferredOperationJoinKHR-operation-parent", 48}, { {"operation must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-02442", 66}, { {"All submitted commands that refer to accelerationStructure must have completed execution", 88}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-02443", 66}, { {"If VkAllocationCallbacks were provided when accelerationStructure was created, a compatible set of callbacks must be provided here", 130}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-02444", 66}, { {"If no VkAllocationCallbacks were provided when accelerationStructure was created, pAllocator must be NULL", 105}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-08934", 66}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-parameter", 70}, { {"If accelerationStructure is not VK_NULL_HANDLE, accelerationStructure must be a valid VkAccelerationStructureKHR handle", 119}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-accelerationStructure-parent", 67}, { {"If accelerationStructure is a valid handle, it must have been created, allocated, or retrieved from device", 106}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureKHR-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-accelerationStructure-03752", 65}, { {"All submitted commands that refer to accelerationStructure must have completed execution", 88}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-accelerationStructure-03753", 65}, { {"If VkAllocationCallbacks were provided when accelerationStructure was created, a compatible set of callbacks must be provided here", 130}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-accelerationStructure-03754", 65}, { {"If no VkAllocationCallbacks were provided when accelerationStructure was created, pAllocator must be NULL", 105}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-accelerationStructure-parameter", 69}, { {"If accelerationStructure is not VK_NULL_HANDLE, accelerationStructure must be a valid VkAccelerationStructureNV handle", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-accelerationStructure-parent", 66}, { {"If accelerationStructure is a valid handle, it must have been created, allocated, or retrieved from device", 106}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyAccelerationStructureNV-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-buffer-00922", 33}, { {"All submitted commands that refer to buffer, either directly or via a VkBufferView, must have completed execution", 113}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-buffer-00923", 33}, { {"If VkAllocationCallbacks were provided when buffer was created, a compatible set of callbacks must be provided here", 115}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-buffer-00924", 33}, { {"If no VkAllocationCallbacks were provided when buffer was created, pAllocator must be NULL", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-buffer-parameter", 37}, { {"If buffer is not VK_NULL_HANDLE, buffer must be a valid VkBuffer handle", 71}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-buffer-parent", 34}, { {"If buffer is a valid handle, it must have been created, allocated, or retrieved from device", 91}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-device-parameter", 37}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBuffer-pAllocator-parameter", 41}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferCollectionFUCHSIA-collection-06407", 54}, { {"VkImage and VkBuffer objects that referenced collection upon creation by inclusion of a VkBufferCollectionImageCreateInfoFUCHSIA or VkBufferCollectionBufferCreateInfoFUCHSIA chained to their VkImageCreateInfo or VkBufferCreateInfo structures respectively, may outlive collection", 278}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferCollectionFUCHSIA-collection-parameter", 58}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferCollectionFUCHSIA-collection-parent", 55}, { {"collection must have been created, allocated, or retrieved from device", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferCollectionFUCHSIA-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferCollectionFUCHSIA-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-bufferView-00936", 41}, { {"All submitted commands that refer to bufferView must have completed execution", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-bufferView-00937", 41}, { {"If VkAllocationCallbacks were provided when bufferView was created, a compatible set of callbacks must be provided here", 119}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-bufferView-00938", 41}, { {"If no VkAllocationCallbacks were provided when bufferView was created, pAllocator must be NULL", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-bufferView-parameter", 45}, { {"If bufferView is not VK_NULL_HANDLE, bufferView must be a valid VkBufferView handle", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-bufferView-parent", 42}, { {"If bufferView is a valid handle, it must have been created, allocated, or retrieved from device", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyBufferView-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyCommandPool-commandPool-00041", 43}, { {"All VkCommandBuffer objects allocated from commandPool must not be in the pending state", 87}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-commandPool-00042", 43}, { {"If VkAllocationCallbacks were provided when commandPool was created, a compatible set of callbacks must be provided here", 120}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-commandPool-00043", 43}, { {"If no VkAllocationCallbacks were provided when commandPool was created, pAllocator must be NULL", 95}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-commandPool-parameter", 47}, { {"If commandPool is not VK_NULL_HANDLE, commandPool must be a valid VkCommandPool handle", 86}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-commandPool-parent", 44}, { {"If commandPool is a valid handle, it must have been created, allocated, or retrieved from device", 96}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCommandPool-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkDestroyCuFunctionNVX-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuFunctionNVX-function-parameter", 46}, { {"function must be a valid VkCuFunctionNVX handle", 47}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuFunctionNVX-function-parent", 43}, { {"function must have been created, allocated, or retrieved from device", 68}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuFunctionNVX-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuModuleNVX-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuModuleNVX-module-parameter", 42}, { {"module must be a valid VkCuModuleNVX handle", 43}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuModuleNVX-module-parent", 39}, { {"module must have been created, allocated, or retrieved from device", 66}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCuModuleNVX-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkDestroyCudaFunctionNV-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaFunctionNV-function-parameter", 47}, { {"function must be a valid VkCudaFunctionNV handle", 48}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaFunctionNV-function-parent", 44}, { {"function must have been created, allocated, or retrieved from device", 68}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaFunctionNV-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaModuleNV-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaModuleNV-module-parameter", 43}, { {"module must be a valid VkCudaModuleNV handle", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaModuleNV-module-parent", 40}, { {"module must have been created, allocated, or retrieved from device", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyCudaModuleNV-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-pAllocator-parameter", 62}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-session-09793", 55}, { {"All submitted commands that refer to session must have completed execution", 74}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-session-09794", 55}, { {"If VkAllocationCallbacks were provided when session was created, a compatible set of callbacks must be provided here", 116}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-session-09795", 55}, { {"If no VkAllocationCallbacks were provided when session was created, pAllocator must be NULL", 91}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-session-parameter", 59}, { {"session must be a valid VkDataGraphPipelineSessionARM handle", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDataGraphPipelineSessionARM-session-parent", 56}, { {"session must have been created, allocated, or retrieved from device", 67}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-callback-parameter", 55}, { {"If callback is not VK_NULL_HANDLE, callback must be a valid VkDebugReportCallbackEXT handle", 91}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-callback-parent", 52}, { {"If callback is a valid handle, it must have been created, allocated, or retrieved from instance", 95}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-instance-01242", 51}, { {"If VkAllocationCallbacks were provided when callback was created, a compatible set of callbacks must be provided here", 117}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-instance-01243", 51}, { {"If no VkAllocationCallbacks were provided when callback was created, pAllocator must be NULL", 92}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-instance-parameter", 55}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugReportCallbackEXT-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-instance-parameter", 55}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-messenger-01915", 52}, { {"If VkAllocationCallbacks were provided when messenger was created, a compatible set of callbacks must be provided here", 118}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-messenger-01916", 52}, { {"If no VkAllocationCallbacks were provided when messenger was created, pAllocator must be NULL", 93}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-messenger-parameter", 56}, { {"If messenger is not VK_NULL_HANDLE, messenger must be a valid VkDebugUtilsMessengerEXT handle", 93}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-messenger-parent", 53}, { {"If messenger is a valid handle, it must have been created, allocated, or retrieved from instance", 96}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDebugUtilsMessengerEXT-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkDestroyDeferredOperationKHR-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-operation-03434", 50}, { {"If VkAllocationCallbacks were provided when operation was created, a compatible set of callbacks must be provided here", 118}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-operation-03435", 50}, { {"If no VkAllocationCallbacks were provided when operation was created, pAllocator must be NULL", 93}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-operation-03436", 50}, { {"operation must be completed", 27}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-operation-parameter", 54}, { {"If operation is not VK_NULL_HANDLE, operation must be a valid VkDeferredOperationKHR handle", 91}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-operation-parent", 51}, { {"If operation is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDeferredOperationKHR-pAllocator-parameter", 55}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkDestroyDescriptorPool-descriptorPool-00303", 49}, { {"All submitted commands that refer to descriptorPool (via any allocated descriptor sets) must have completed execution", 117}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-descriptorPool-00304", 49}, { {"If VkAllocationCallbacks were provided when descriptorPool was created, a compatible set of callbacks must be provided here", 123}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-descriptorPool-00305", 49}, { {"If no VkAllocationCallbacks were provided when descriptorPool was created, pAllocator must be NULL", 98}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-descriptorPool-parameter", 53}, { {"If descriptorPool is not VK_NULL_HANDLE, descriptorPool must be a valid VkDescriptorPool handle", 95}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-descriptorPool-parent", 50}, { {"If descriptorPool is a valid handle, it must have been created, allocated, or retrieved from device", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorPool-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-descriptorSetLayout-00284", 59}, { {"If VkAllocationCallbacks were provided when descriptorSetLayout was created, a compatible set of callbacks must be provided here", 128}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-descriptorSetLayout-00285", 59}, { {"If no VkAllocationCallbacks were provided when descriptorSetLayout was created, pAllocator must be NULL", 103}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-descriptorSetLayout-parameter", 63}, { {"If descriptorSetLayout is not VK_NULL_HANDLE, descriptorSetLayout must be a valid VkDescriptorSetLayout handle", 110}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-descriptorSetLayout-parent", 60}, { {"If descriptorSetLayout is a valid handle, it must have been created, allocated, or retrieved from device", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorSetLayout-pAllocator-parameter", 54}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-descriptorSetLayout-00356", 64}, { {"If VkAllocationCallbacks were provided when descriptorUpdateTemplate was created, a compatible set of callbacks must be provided here", 133}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-descriptorSetLayout-00357", 64}, { {"If no VkAllocationCallbacks were provided when descriptorUpdateTemplate was created, pAllocator must be NULL", 108}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-descriptorUpdateTemplate-parameter", 73}, { {"If descriptorUpdateTemplate is not VK_NULL_HANDLE, descriptorUpdateTemplate must be a valid VkDescriptorUpdateTemplate handle", 125}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-descriptorUpdateTemplate-parent", 70}, { {"If descriptorUpdateTemplate is a valid handle, it must have been created, allocated, or retrieved from device", 109}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDescriptorUpdateTemplate-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyDevice-device-00379", 33}, { {"If VkAllocationCallbacks were provided when device was created, a compatible set of callbacks must be provided here", 115}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkDestroyDevice-device-00380", 33}, { {"If no VkAllocationCallbacks were provided when device was created, pAllocator must be NULL", 90}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkDestroyDevice-device-05137", 33}, { {"All child objects created on device that can be destroyed or freed must have been destroyed or freed prior to destroying device", 127}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkDestroyDevice-device-parameter", 37}, { {"If device is not NULL, device must be a valid VkDevice handle", 61}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkDestroyDevice-pAllocator-parameter", 41}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkDestroyEvent-device-parameter", 36}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-event-01145", 31}, { {"All submitted commands that refer to event must have completed execution", 72}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-event-01146", 31}, { {"If VkAllocationCallbacks were provided when event was created, a compatible set of callbacks must be provided here", 114}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-event-01147", 31}, { {"If no VkAllocationCallbacks were provided when event was created, pAllocator must be NULL", 89}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-event-parameter", 35}, { {"If event is not VK_NULL_HANDLE, event must be a valid VkEvent handle", 68}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-event-parent", 32}, { {"If event is a valid handle, it must have been created, allocated, or retrieved from device", 90}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyEvent-pAllocator-parameter", 40}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyExternalComputeQueueNV-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkDestroyExternalComputeQueueNV-externalQueue-parameter", 60}, { {"externalQueue must be a valid VkExternalComputeQueueNV handle", 61}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkDestroyExternalComputeQueueNV-externalQueue-parent", 57}, { {"externalQueue must have been created, allocated, or retrieved from device", 73}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkDestroyExternalComputeQueueNV-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkDestroyFence-device-parameter", 36}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-fence-01120", 31}, { {"All queue submission commands that refer to fence must have completed execution", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-fence-01121", 31}, { {"If VkAllocationCallbacks were provided when fence was created, a compatible set of callbacks must be provided here", 114}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-fence-01122", 31}, { {"If no VkAllocationCallbacks were provided when fence was created, pAllocator must be NULL", 89}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-fence-parameter", 35}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-fence-parent", 32}, { {"If fence is a valid handle, it must have been created, allocated, or retrieved from device", 90}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFence-pAllocator-parameter", 40}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyFramebuffer-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-framebuffer-00892", 43}, { {"All submitted commands that refer to framebuffer must have completed execution", 78}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-framebuffer-00893", 43}, { {"If VkAllocationCallbacks were provided when framebuffer was created, a compatible set of callbacks must be provided here", 120}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-framebuffer-00894", 43}, { {"If no VkAllocationCallbacks were provided when framebuffer was created, pAllocator must be NULL", 95}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-framebuffer-parameter", 47}, { {"If framebuffer is not VK_NULL_HANDLE, framebuffer must be a valid VkFramebuffer handle", 86}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-framebuffer-parent", 44}, { {"If framebuffer is a valid handle, it must have been created, allocated, or retrieved from device", 96}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyFramebuffer-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyGpaSessionAMD-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkDestroyGpaSessionAMD-gpaSession-12408", 44}, { {"All submitted commands that refer to gpaSession must have completed execution", 77}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkDestroyGpaSessionAMD-gpaSession-parameter", 48}, { {"If gpaSession is not VK_NULL_HANDLE, gpaSession must be a valid VkGpaSessionAMD handle", 86}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkDestroyGpaSessionAMD-gpaSession-parent", 45}, { {"If gpaSession is a valid handle, it must have been created, allocated, or retrieved from device", 95}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkDestroyGpaSessionAMD-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkDestroyImage-device-parameter", 36}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-01000", 31}, { {"All submitted commands that refer to image, either directly or via a VkImageView, must have completed execution", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-01001", 31}, { {"If VkAllocationCallbacks were provided when image was created, a compatible set of callbacks must be provided here", 114}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-01002", 31}, { {"If no VkAllocationCallbacks were provided when image was created, pAllocator must be NULL", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-04882", 31}, { {"image must not have been acquired from vkGetSwapchainImagesKHR", 62}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-parameter", 35}, { {"If image is not VK_NULL_HANDLE, image must be a valid VkImage handle", 68}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-image-parent", 32}, { {"If image is a valid handle, it must have been created, allocated, or retrieved from device", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImage-pAllocator-parameter", 40}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-imageView-01026", 39}, { {"All submitted commands that refer to imageView must have completed execution", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-imageView-01027", 39}, { {"If VkAllocationCallbacks were provided when imageView was created, a compatible set of callbacks must be provided here", 118}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-imageView-01028", 39}, { {"If no VkAllocationCallbacks were provided when imageView was created, pAllocator must be NULL", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-imageView-parameter", 43}, { {"If imageView is not VK_NULL_HANDLE, imageView must be a valid VkImageView handle", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-imageView-parent", 40}, { {"If imageView is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyImageView-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-indirectCommandsLayout-11114", 68}, { {"All submitted commands that refer to indirectCommandsLayout must have completed execution", 89}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-indirectCommandsLayout-11115", 68}, { {"If VkAllocationCallbacks were provided when indirectCommandsLayout was created, a compatible set of callbacks must be provided here", 131}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-indirectCommandsLayout-11116", 68}, { {"If no VkAllocationCallbacks were provided when indirectCommandsLayout was created, pAllocator must be NULL", 106}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-indirectCommandsLayout-parameter", 72}, { {"If indirectCommandsLayout is not VK_NULL_HANDLE, indirectCommandsLayout must be a valid VkIndirectCommandsLayoutEXT handle", 122}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-indirectCommandsLayout-parent", 69}, { {"If indirectCommandsLayout is a valid handle, it must have been created, allocated, or retrieved from device", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutEXT-pAllocator-parameter", 60}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-deviceGeneratedCommands-02941", 68}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-indirectCommandsLayout-02938", 67}, { {"All submitted commands that refer to indirectCommandsLayout must have completed execution", 89}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-indirectCommandsLayout-02939", 67}, { {"If VkAllocationCallbacks were provided when indirectCommandsLayout was created, a compatible set of callbacks must be provided here", 131}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-indirectCommandsLayout-02940", 67}, { {"If no VkAllocationCallbacks were provided when indirectCommandsLayout was created, pAllocator must be NULL", 106}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-indirectCommandsLayout-parameter", 71}, { {"If indirectCommandsLayout is not VK_NULL_HANDLE, indirectCommandsLayout must be a valid VkIndirectCommandsLayoutNV handle", 121}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-indirectCommandsLayout-parent", 68}, { {"If indirectCommandsLayout is a valid handle, it must have been created, allocated, or retrieved from device", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectCommandsLayoutNV-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectExecutionSetEXT-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectExecutionSetEXT-indirectExecutionSet-11025", 64}, { {"All submitted commands that refer to indirectExecutionSet must have completed execution", 87}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectExecutionSetEXT-indirectExecutionSet-parameter", 68}, { {"If indirectExecutionSet is not VK_NULL_HANDLE, indirectExecutionSet must be a valid VkIndirectExecutionSetEXT handle", 116}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectExecutionSetEXT-indirectExecutionSet-parent", 65}, { {"If indirectExecutionSet is a valid handle, it must have been created, allocated, or retrieved from device", 105}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyIndirectExecutionSetEXT-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkDestroyInstance-instance-00629", 37}, { {"All child objects that were created with instance or with a VkPhysicalDevice retrieved from it, and that can be destroyed or freed, must have been destroyed or freed prior to destroying instance", 194}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkDestroyInstance-instance-00630", 37}, { {"If VkAllocationCallbacks were provided when instance was created, a compatible set of callbacks must be provided here", 117}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkDestroyInstance-instance-00631", 37}, { {"If no VkAllocationCallbacks were provided when instance was created, pAllocator must be NULL", 92}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkDestroyInstance-instance-parameter", 41}, { {"If instance is not NULL, instance must be a valid VkInstance handle", 67}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkDestroyInstance-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkDestroyMicromapEXT-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-07441", 40}, { {"All submitted commands that refer to micromap must have completed execution", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-07442", 40}, { {"If VkAllocationCallbacks were provided when micromap was created, a compatible set of callbacks must be provided here", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-07443", 40}, { {"If no VkAllocationCallbacks were provided when micromap was created, pAllocator must be NULL", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-11619", 40}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-parameter", 44}, { {"If micromap is not VK_NULL_HANDLE, micromap must be a valid VkMicromapEXT handle", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-micromap-parent", 41}, { {"If micromap is a valid handle, it must have been created, allocated, or retrieved from device", 93}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyMicromapEXT-pAllocator-parameter", 46}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyOpticalFlowSessionNV-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkDestroyOpticalFlowSessionNV-pAllocator-parameter", 55}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkDestroyOpticalFlowSessionNV-session-parameter", 52}, { {"session must be a valid VkOpticalFlowSessionNV handle", 53}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkDestroyOpticalFlowSessionNV-session-parent", 49}, { {"session must have been created, allocated, or retrieved from device", 67}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkDestroyPipeline-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pAllocator-parameter", 43}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pipeline-00765", 37}, { {"All submitted commands that refer to pipeline must have completed execution", 75}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pipeline-00766", 37}, { {"If VkAllocationCallbacks were provided when pipeline was created, a compatible set of callbacks must be provided here", 117}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pipeline-00767", 37}, { {"If no VkAllocationCallbacks were provided when pipeline was created, pAllocator must be NULL", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pipeline-parameter", 41}, { {"If pipeline is not VK_NULL_HANDLE, pipeline must be a valid VkPipeline handle", 77}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipeline-pipeline-parent", 38}, { {"If pipeline is a valid handle, it must have been created, allocated, or retrieved from device", 93}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-pAllocator-parameter", 52}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-pipelineBinary-09614", 52}, { {"If VkAllocationCallbacks were provided when pipelineBinary was created, a compatible set of callbacks must be provided here", 123}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-pipelineBinary-09615", 52}, { {"If no VkAllocationCallbacks were provided when pipelineBinary was created, pAllocator must be NULL", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-pipelineBinary-parameter", 56}, { {"If pipelineBinary is not VK_NULL_HANDLE, pipelineBinary must be a valid VkPipelineBinaryKHR handle", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineBinaryKHR-pipelineBinary-parent", 53}, { {"If pipelineBinary is a valid handle, it must have been created, allocated, or retrieved from device", 99}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-pipelineCache-00771", 47}, { {"If VkAllocationCallbacks were provided when pipelineCache was created, a compatible set of callbacks must be provided here", 122}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-pipelineCache-00772", 47}, { {"If no VkAllocationCallbacks were provided when pipelineCache was created, pAllocator must be NULL", 97}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-pipelineCache-parameter", 51}, { {"If pipelineCache is not VK_NULL_HANDLE, pipelineCache must be a valid VkPipelineCache handle", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineCache-pipelineCache-parent", 48}, { {"If pipelineCache is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkDestroyPipelineLayout-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPipelineLayout-pAllocator-parameter", 49}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPipelineLayout-pipelineLayout-00299", 49}, { {"If VkAllocationCallbacks were provided when pipelineLayout was created, a compatible set of callbacks must be provided here", 123}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPipelineLayout-pipelineLayout-00300", 49}, { {"If no VkAllocationCallbacks were provided when pipelineLayout was created, pAllocator must be NULL", 98}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPipelineLayout-pipelineLayout-parameter", 53}, { {"If pipelineLayout is not VK_NULL_HANDLE, pipelineLayout must be a valid VkPipelineLayout handle", 95}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPipelineLayout-pipelineLayout-parent", 50}, { {"If pipelineLayout is a valid handle, it must have been created, allocated, or retrieved from device", 99}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkDestroyPrivateDataSlot-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyPrivateDataSlot-pAllocator-parameter", 50}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyPrivateDataSlot-privateDataSlot-04062", 51}, { {"If VkAllocationCallbacks were provided when privateDataSlot was created, a compatible set of callbacks must be provided here", 124}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyPrivateDataSlot-privateDataSlot-04063", 51}, { {"If no VkAllocationCallbacks were provided when privateDataSlot was created, pAllocator must be NULL", 99}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyPrivateDataSlot-privateDataSlot-parameter", 55}, { {"If privateDataSlot is not VK_NULL_HANDLE, privateDataSlot must be a valid VkPrivateDataSlot handle", 98}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyPrivateDataSlot-privateDataSlot-parent", 52}, { {"If privateDataSlot is a valid handle, it must have been created, allocated, or retrieved from device", 100}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkDestroyQueryPool-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-queryPool-00793", 39}, { {"All submitted commands that refer to queryPool must have completed execution", 76}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-queryPool-00794", 39}, { {"If VkAllocationCallbacks were provided when queryPool was created, a compatible set of callbacks must be provided here", 118}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-queryPool-00795", 39}, { {"If no VkAllocationCallbacks were provided when queryPool was created, pAllocator must be NULL", 93}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-queryPool-parameter", 43}, { {"If queryPool is not VK_NULL_HANDLE, queryPool must be a valid VkQueryPool handle", 80}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyQueryPool-queryPool-parent", 40}, { {"If queryPool is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/queries.html", 21} } },
        { {"VUID-vkDestroyRenderPass-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-renderPass-00873", 41}, { {"All submitted commands that refer to renderPass must have completed execution", 77}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-renderPass-00874", 41}, { {"If VkAllocationCallbacks were provided when renderPass was created, a compatible set of callbacks must be provided here", 119}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-renderPass-00875", 41}, { {"If no VkAllocationCallbacks were provided when renderPass was created, pAllocator must be NULL", 94}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-renderPass-parameter", 45}, { {"If renderPass is not VK_NULL_HANDLE, renderPass must be a valid VkRenderPass handle", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroyRenderPass-renderPass-parent", 42}, { {"If renderPass is a valid handle, it must have been created, allocated, or retrieved from device", 95}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkDestroySampler-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-pAllocator-parameter", 42}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-sampler-01082", 35}, { {"All submitted commands that refer to sampler must have completed execution", 74}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-sampler-01083", 35}, { {"If VkAllocationCallbacks were provided when sampler was created, a compatible set of callbacks must be provided here", 116}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-sampler-01084", 35}, { {"If no VkAllocationCallbacks were provided when sampler was created, pAllocator must be NULL", 91}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-sampler-parameter", 39}, { {"If sampler is not VK_NULL_HANDLE, sampler must be a valid VkSampler handle", 74}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySampler-sampler-parent", 36}, { {"If sampler is a valid handle, it must have been created, allocated, or retrieved from device", 92}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySamplerYcbcrConversion-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySamplerYcbcrConversion-pAllocator-parameter", 57}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySamplerYcbcrConversion-ycbcrConversion-parameter", 62}, { {"If ycbcrConversion is not VK_NULL_HANDLE, ycbcrConversion must be a valid VkSamplerYcbcrConversion handle", 105}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySamplerYcbcrConversion-ycbcrConversion-parent", 59}, { {"If ycbcrConversion is a valid handle, it must have been created, allocated, or retrieved from device", 100}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkDestroySemaphore-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-semaphore-01138", 39}, { {"If VkAllocationCallbacks were provided when semaphore was created, a compatible set of callbacks must be provided here", 118}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-semaphore-01139", 39}, { {"If no VkAllocationCallbacks were provided when semaphore was created, pAllocator must be NULL", 93}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-semaphore-05149", 39}, { {"All submitted batches that refer to semaphore must have completed execution", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-semaphore-parameter", 43}, { {"If semaphore is not VK_NULL_HANDLE, semaphore must be a valid VkSemaphore handle", 80}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroySemaphore-semaphore-parent", 40}, { {"If semaphore is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDestroyShaderEXT-None-08481", 34}, { {"The shaderObject feature must be enabled", 40}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-pAllocator-08483", 40}, { {"If VkAllocationCallbacks were provided when shader was created, a compatible set of callbacks must be provided here", 115}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-pAllocator-08484", 40}, { {"If no VkAllocationCallbacks were provided when shader was created, pAllocator must be NULL", 90}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-shader-08482", 36}, { {"All submitted commands that refer to shader must have completed execution", 73}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-shader-parameter", 40}, { {"If shader is not VK_NULL_HANDLE, shader must be a valid VkShaderEXT handle", 74}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderEXT-shader-parent", 37}, { {"If shader is a valid handle, it must have been created, allocated, or retrieved from device", 91}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderInstrumentationARM-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderInstrumentationARM-instrumentation-12374", 60}, { {"All submitted commands that refer to instrumentation must have completed execution", 82}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderInstrumentationARM-instrumentation-parameter", 64}, { {"If instrumentation is not VK_NULL_HANDLE, instrumentation must be a valid VkShaderInstrumentationARM handle", 107}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderInstrumentationARM-instrumentation-parent", 61}, { {"If instrumentation is a valid handle, it must have been created, allocated, or retrieved from device", 100}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderInstrumentationARM-pAllocator-parameter", 59}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-shaderModule-01092", 45}, { {"If VkAllocationCallbacks were provided when shaderModule was created, a compatible set of callbacks must be provided here", 121}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-shaderModule-01093", 45}, { {"If no VkAllocationCallbacks were provided when shaderModule was created, pAllocator must be NULL", 96}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-shaderModule-parameter", 49}, { {"If shaderModule is not VK_NULL_HANDLE, shaderModule must be a valid VkShaderModule handle", 89}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyShaderModule-shaderModule-parent", 46}, { {"If shaderModule is a valid handle, it must have been created, allocated, or retrieved from device", 97}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroySurfaceKHR-instance-parameter", 43}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-pAllocator-parameter", 45}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-surface-01266", 38}, { {"All VkSwapchainKHR objects created for surface must have been destroyed prior to destroying surface", 99}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-surface-01267", 38}, { {"If VkAllocationCallbacks were provided when surface was created, a compatible set of callbacks must be provided here", 116}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-surface-01268", 38}, { {"If no VkAllocationCallbacks were provided when surface was created, pAllocator must be NULL", 91}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-surface-parameter", 42}, { {"If surface is not VK_NULL_HANDLE, surface must be a valid VkSurfaceKHR handle", 77}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySurfaceKHR-surface-parent", 39}, { {"If surface is a valid handle, it must have been created, allocated, or retrieved from instance", 94}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-pAllocator-parameter", 47}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-swapchain-01282", 42}, { {"All uses of presentable images acquired from swapchain must have completed execution", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-swapchain-01283", 42}, { {"If VkAllocationCallbacks were provided when swapchain was created, a compatible set of callbacks must be provided here", 118}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-swapchain-01284", 42}, { {"If no VkAllocationCallbacks were provided when swapchain was created, pAllocator must be NULL", 93}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-swapchain-parameter", 46}, { {"If swapchain is not VK_NULL_HANDLE, swapchain must be a valid VkSwapchainKHR handle", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroySwapchainKHR-swapchain-parent", 43}, { {"If swapchain is a valid handle, it must have been created, allocated, or retrieved from device", 94}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDestroyTensorARM-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-pAllocator-parameter", 44}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-tensor-09730", 36}, { {"All submitted commands that refer to tensor, either directly or via a VkTensorViewARM, must have completed execution", 116}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-tensor-09731", 36}, { {"If VkAllocationCallbacks were provided when tensor was created, a compatible set of callbacks must be provided here", 115}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-tensor-09732", 36}, { {"If no VkAllocationCallbacks were provided when tensor was created, pAllocator must be NULL", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-tensor-parameter", 40}, { {"If tensor is not VK_NULL_HANDLE, tensor must be a valid VkTensorARM handle", 74}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorARM-tensor-parent", 37}, { {"If tensor is a valid handle, it must have been created, allocated, or retrieved from device", 91}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-pAllocator-parameter", 48}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-tensorView-09750", 44}, { {"All submitted commands that refer to tensorView must have completed execution", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-tensorView-09751", 44}, { {"If VkAllocationCallbacks were provided when tensorView was created, a compatible set of callbacks must be provided here", 119}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-tensorView-09752", 44}, { {"If no VkAllocationCallbacks were provided when tensorView was created, pAllocator must be NULL", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-tensorView-parameter", 48}, { {"If tensorView is not VK_NULL_HANDLE, tensorView must be a valid VkTensorViewARM handle", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyTensorViewARM-tensorView-parent", 45}, { {"If tensorView is a valid handle, it must have been created, allocated, or retrieved from device", 95}, { "chapters/resources.html", 23} } },
        { {"VUID-vkDestroyValidationCacheEXT-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyValidationCacheEXT-pAllocator-parameter", 53}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyValidationCacheEXT-validationCache-01537", 54}, { {"If VkAllocationCallbacks were provided when validationCache was created, a compatible set of callbacks must be provided here", 124}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyValidationCacheEXT-validationCache-01538", 54}, { {"If no VkAllocationCallbacks were provided when validationCache was created, pAllocator must be NULL", 99}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyValidationCacheEXT-validationCache-parameter", 58}, { {"If validationCache is not VK_NULL_HANDLE, validationCache must be a valid VkValidationCacheEXT handle", 101}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyValidationCacheEXT-validationCache-parent", 55}, { {"If validationCache is a valid handle, it must have been created, allocated, or retrieved from device", 100}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkDestroyVideoSessionKHR-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-pAllocator-parameter", 50}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-videoSession-07192", 48}, { {"All submitted commands that refer to videoSession must have completed execution", 79}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-videoSession-07193", 48}, { {"If VkAllocationCallbacks were provided when videoSession was created, a compatible set of callbacks must be provided here", 121}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-videoSession-07194", 48}, { {"If no VkAllocationCallbacks were provided when videoSession was created, pAllocator must be NULL", 96}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-videoSession-parameter", 52}, { {"If videoSession is not VK_NULL_HANDLE, videoSession must be a valid VkVideoSessionKHR handle", 92}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionKHR-videoSession-parent", 49}, { {"If videoSession is a valid handle, it must have been created, allocated, or retrieved from device", 97}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-pAllocator-parameter", 60}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-07212", 68}, { {"All submitted commands that refer to videoSessionParameters must have completed execution", 89}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-07213", 68}, { {"If VkAllocationCallbacks were provided when videoSessionParameters was created, a compatible set of callbacks must be provided here", 131}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-07214", 68}, { {"If no VkAllocationCallbacks were provided when videoSessionParameters was created, pAllocator must be NULL", 106}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-parameter", 72}, { {"If videoSessionParameters is not VK_NULL_HANDLE, videoSessionParameters must be a valid VkVideoSessionParametersKHR handle", 122}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-parent", 69}, { {"If videoSessionParameters is a valid handle, it must have been created, allocated, or retrieved from device", 107}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkDeviceWaitIdle-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkDisplayPowerControlEXT-commonparent", 42}, { {"Both of device, and display must have been created, allocated, or retrieved from the same VkPhysicalDevice", 106}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDisplayPowerControlEXT-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDisplayPowerControlEXT-display-parameter", 47}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkDisplayPowerControlEXT-pDisplayPowerInfo-parameter", 57}, { {"pDisplayPowerInfo must be a valid pointer to a valid VkDisplayPowerInfoEXT structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkEndCommandBuffer-None-01978", 34}, { {"Conditional rendering must not be active", 40}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-None-06991", 34}, { {"There must be no video session object bound", 43}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-00059", 43}, { {"commandBuffer must be in the recording state", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-00060", 43}, { {"If commandBuffer is a primary command buffer, there must not be an active render pass instance", 94}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-00061", 43}, { {"All queries made active during the recording of commandBuffer must have been made inactive", 90}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-00062", 43}, { {"If commandBuffer is a secondary command buffer, there must not be an outstanding vkCmdDebugMarkerBeginEXT command recorded to commandBuffer that has not previously been ended by a call to vkCmdDebugMarkerEndEXT", 210}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-01815", 43}, { {"If commandBuffer is a secondary command buffer, there must not be an outstanding vkCmdBeginDebugUtilsLabelEXT command recorded to commandBuffer that has not previously been ended by a call to vkCmdEndDebugUtilsLabelEXT", 218}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-12372", 43}, { {"commandBuffer must not have any shader instrumentation active", 61}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEndCommandBuffer-commandBuffer-parameter", 47}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkEnumerateDeviceExtensionProperties-pLayerName-parameter", 62}, { {"If pLayerName is not NULL, pLayerName must be a null-terminated UTF-8 string", 76}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceExtensionProperties-pProperties-parameter", 63}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkExtensionProperties structures", 179}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceExtensionProperties-pPropertyCount-parameter", 66}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceExtensionProperties-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceLayerProperties-pProperties-parameter", 59}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkLayerProperties structures", 175}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceLayerProperties-pPropertyCount-parameter", 62}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateDeviceLayerProperties-physicalDevice-parameter", 62}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceExtensionProperties-pLayerName-parameter", 64}, { {"If pLayerName is not NULL, pLayerName must be a null-terminated UTF-8 string", 76}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceExtensionProperties-pProperties-parameter", 65}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkExtensionProperties structures", 179}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceExtensionProperties-pPropertyCount-parameter", 68}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceLayerProperties-pProperties-parameter", 61}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkLayerProperties structures", 175}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceLayerProperties-pPropertyCount-parameter", 64}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/extensions.html", 24} } },
        { {"VUID-vkEnumerateInstanceVersion-pApiVersion-parameter", 53}, { {"pApiVersion must be a valid pointer to a uint32_t value", 55}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkEnumeratePhysicalDeviceGroups-instance-parameter", 55}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceGroups-pPhysicalDeviceGroupCount-parameter", 72}, { {"pPhysicalDeviceGroupCount must be a valid pointer to a uint32_t value", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceGroups-pPhysicalDeviceGroupProperties-parameter", 77}, { {"If the value referenced by pPhysicalDeviceGroupCount is not 0, and pPhysicalDeviceGroupProperties is not NULL, pPhysicalDeviceGroupProperties must be a valid pointer to an array of pPhysicalDeviceGroupCount VkPhysicalDeviceGroupProperties structures", 249}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM-pCounterCount-parameter", 95}, { {"pCounterCount must be a valid pointer to a uint32_t value", 57}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM-pCounterDescriptions-parameter", 102}, { {"If the value referenced by pCounterCount is not 0, and pCounterDescriptions is not NULL, pCounterDescriptions must be a valid pointer to an array of pCounterCount VkPerformanceCounterDescriptionARM structures", 208}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM-pCounters-parameter", 91}, { {"If the value referenced by pCounterCount is not 0, and pCounters is not NULL, pCounters must be a valid pointer to an array of pCounterCount VkPerformanceCounterARM structures", 175}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM-physicalDevice-parameter", 96}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR-pCounterCount-parameter", 92}, { {"pCounterCount must be a valid pointer to a uint32_t value", 57}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR-pCounterDescriptions-parameter", 99}, { {"If the value referenced by pCounterCount is not 0, and pCounterDescriptions is not NULL, pCounterDescriptions must be a valid pointer to an array of pCounterCount VkPerformanceCounterDescriptionKHR structures", 208}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR-pCounters-parameter", 88}, { {"If the value referenced by pCounterCount is not 0, and pCounters is not NULL, pCounters must be a valid pointer to an array of pCounterCount VkPerformanceCounterKHR structures", 175}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR-physicalDevice-parameter", 93}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDeviceShaderInstrumentationMetricsARM-pDescriptionCount-parameter", 89}, { {"pDescriptionCount must be a valid pointer to a uint32_t value", 61}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkEnumeratePhysicalDeviceShaderInstrumentationMetricsARM-pDescriptions-parameter", 85}, { {"If the value referenced by pDescriptionCount is not 0, and pDescriptions is not NULL, pDescriptions must be a valid pointer to an array of pDescriptionCount VkShaderInstrumentationMetricDescriptionARM structures", 211}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkEnumeratePhysicalDeviceShaderInstrumentationMetricsARM-physicalDevice-parameter", 86}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkEnumeratePhysicalDevices-instance-parameter", 50}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDevices-pPhysicalDeviceCount-parameter", 62}, { {"pPhysicalDeviceCount must be a valid pointer to a uint32_t value", 64}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkEnumeratePhysicalDevices-pPhysicalDevices-parameter", 58}, { {"If the value referenced by pPhysicalDeviceCount is not 0, and pPhysicalDevices is not NULL, pPhysicalDevices must be a valid pointer to an array of pPhysicalDeviceCount VkPhysicalDevice handles", 193}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkExportMetalObjectsEXT-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkExportMetalObjectsEXT-pMetalObjectsInfo-parameter", 56}, { {"pMetalObjectsInfo must be a valid pointer to a VkExportMetalObjectsInfoEXT structure", 84}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFlushMappedMemoryRanges-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFlushMappedMemoryRanges-memoryRangeCount-arraylength", 59}, { {"memoryRangeCount must be greater than 0", 39}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFlushMappedMemoryRanges-pMemoryRanges-parameter", 54}, { {"pMemoryRanges must be a valid pointer to an array of memoryRangeCount valid VkMappedMemoryRange structures", 106}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFreeCommandBuffers-commandBufferCount-arraylength", 56}, { {"commandBufferCount must be greater than 0", 41}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-commandPool-parameter", 47}, { {"commandPool must be a valid VkCommandPool handle", 48}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-commandPool-parent", 44}, { {"commandPool must have been created, allocated, or retrieved from device", 71}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-pCommandBuffers-00047", 47}, { {"All elements of pCommandBuffers must not be in the pending state", 64}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-pCommandBuffers-00048", 47}, { {"pCommandBuffers must be a valid pointer to an array of commandBufferCount VkCommandBuffer handles, each element of which must either be a valid handle or NULL", 158}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeCommandBuffers-pCommandBuffers-parent", 48}, { {"Each element of pCommandBuffers that is a valid handle must have been created, allocated, or retrieved from commandPool", 119}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkFreeDescriptorSets-descriptorPool-00312", 46}, { {"descriptorPool must have been created with the VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT flag", 101}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-descriptorPool-parameter", 50}, { {"descriptorPool must be a valid VkDescriptorPool handle", 54}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-descriptorPool-parent", 47}, { {"descriptorPool must have been created, allocated, or retrieved from device", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-descriptorSetCount-arraylength", 56}, { {"descriptorSetCount must be greater than 0", 41}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-pDescriptorSets-00309", 47}, { {"All submitted commands that refer to any element of pDescriptorSets must have completed execution", 97}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-pDescriptorSets-00310", 47}, { {"pDescriptorSets must be a valid pointer to an array of descriptorSetCount VkDescriptorSet handles, each element of which must either be a valid handle or VK_NULL_HANDLE", 168}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeDescriptorSets-pDescriptorSets-parent", 48}, { {"Each element of pDescriptorSets that is a valid handle must have been created, allocated, or retrieved from descriptorPool", 122}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkFreeMemory-device-parameter", 34}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFreeMemory-memory-00677", 30}, { {"All submitted commands that refer to memory (via images or buffers) must have completed execution", 97}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFreeMemory-memory-parameter", 34}, { {"If memory is not VK_NULL_HANDLE, memory must be a valid VkDeviceMemory handle", 77}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFreeMemory-memory-parent", 31}, { {"If memory is a valid handle, it must have been created, allocated, or retrieved from device", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-vkFreeMemory-pAllocator-parameter", 38}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-accelerationStructure-08933", 72}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-buildType-11614", 60}, { {"If pBuildInfo->type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, buildType must equal VK_ACCELERATION_STRUCTURE_BUILD_TYPE_DEVICE_KHR", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-buildType-parameter", 64}, { {"buildType must be a valid VkAccelerationStructureBuildTypeKHR value", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-device-03618", 57}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pBuildInfo-03785", 61}, { {"If pBuildInfo->pGeometries or pBuildInfo->ppGeometries has a geometryType of VK_GEOMETRY_TYPE_INSTANCES_KHR, each pMaxPrimitiveCounts[i] must be less than or equal to VkPhysicalDeviceAccelerationStructurePropertiesKHR::maxInstanceCount", 235}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pBuildInfo-parameter", 65}, { {"pBuildInfo must be a valid pointer to a valid VkAccelerationStructureBuildGeometryInfoKHR structure", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pMaxPrimitiveCounts-11612", 70}, { {"If pBuildInfo->geometryCount is not 0 and pBuildInfo->type is not VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR , pMaxPrimitiveCounts must be a valid pointer to an array of pBuildInfo->geometryCount uint32_t values", 220}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pMaxPrimitiveCounts-11613", 70}, { {"If pBuildInfo->type is VK_ACCELERATION_STRUCTURE_TYPE_OPACITY_MICROMAP_KHR, pMaxPrimitiveCounts must equal NULL", 111}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pMaxPrimitiveCounts-parameter", 74}, { {"If pMaxPrimitiveCounts is not NULL, pMaxPrimitiveCounts must be a valid pointer to an array of pBuildInfo->geometryCount uint32_t values", 136}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureBuildSizesKHR-pSizeInfo-parameter", 64}, { {"pSizeInfo must be a valid pointer to a VkAccelerationStructureBuildSizesInfoKHR structure", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-accelerationStructure-08935", 75}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-device-03504", 60}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-device-parameter", 64}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-pInfo-09541", 59}, { {"If the buffer on which pInfo->accelerationStructure was placed is non-sparse then it must be bound completely and contiguously to a single VkDeviceMemory object", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-pInfo-09542", 59}, { {"The buffer on which pInfo->accelerationStructure was placed must have been created with the VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT usage flag set", 148}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureDeviceAddressKHR-pInfo-parameter", 63}, { {"pInfo must be a valid pointer to a valid VkAccelerationStructureDeviceAddressInfoKHR structure", 94}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-accelerationStructure-02787", 67}, { {"accelerationStructure must be bound completely and contiguously to a single VkDeviceMemory object via vkBindAccelerationStructureMemoryNV", 137}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-accelerationStructure-parameter", 71}, { {"accelerationStructure must be a valid VkAccelerationStructureNV handle", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-accelerationStructure-parent", 68}, { {"accelerationStructure must have been created, allocated, or retrieved from device", 81}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-dataSize-02240", 54}, { {"dataSize must be large enough to contain the result of the query, as described above", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-dataSize-arraylength", 60}, { {"dataSize must be greater than 0", 31}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureHandleNV-pData-parameter", 55}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureMemoryRequirementsNV-device-parameter", 68}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureMemoryRequirementsNV-pInfo-parameter", 67}, { {"pInfo must be a valid pointer to a valid VkAccelerationStructureMemoryRequirementsInfoNV structure", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureMemoryRequirementsNV-pMemoryRequirements-parameter", 81}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-None-08088", 72}, { {"The descriptorBufferCaptureReplay feature must be enabled", 57}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-device-08090", 74}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-device-parameter", 78}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-pData-08089", 73}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferPropertiesEXT::accelerationStructureCaptureReplayDescriptorDataSize bytes in size", 159}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-pData-parameter", 77}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT-pInfo-parameter", 77}, { {"pInfo must be a valid pointer to a valid VkAccelerationStructureCaptureDescriptorDataInfoEXT structure", 102}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetAndroidHardwareBufferPropertiesANDROID-buffer-01884", 61}, { {"buffer must be a valid Android hardware buffer object with at least one of the AHARDWAREBUFFER_USAGE_GPU_* flags in its AHardwareBuffer_Desc::usage", 147}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetAndroidHardwareBufferPropertiesANDROID-buffer-parameter", 65}, { {"buffer must be a valid pointer to a valid AHardwareBuffer value", 63}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetAndroidHardwareBufferPropertiesANDROID-device-parameter", 65}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetAndroidHardwareBufferPropertiesANDROID-pProperties-parameter", 70}, { {"pProperties must be a valid pointer to a VkAndroidHardwareBufferPropertiesANDROID structure", 91}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetBufferCollectionPropertiesFUCHSIA-None-06405", 54}, { {"Prior to calling vkGetBufferCollectionPropertiesFUCHSIA, the constraints on the buffer collection must have been set by either vkSetBufferCollectionImageConstraintsFUCHSIA or vkSetBufferCollectionBufferConstraintsFUCHSIA", 220}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferCollectionPropertiesFUCHSIA-collection-parameter", 64}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferCollectionPropertiesFUCHSIA-collection-parent", 61}, { {"collection must have been created, allocated, or retrieved from device", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferCollectionPropertiesFUCHSIA-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferCollectionPropertiesFUCHSIA-pProperties-parameter", 65}, { {"pProperties must be a valid pointer to a VkBufferCollectionPropertiesFUCHSIA structure", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferDeviceAddress-bufferDeviceAddress-03324", 55}, { {"The bufferDeviceAddress feature or the VkPhysicalDeviceBufferDeviceAddressFeaturesEXT::bufferDeviceAddress feature must be enabled, and at least one of the following conditions must be met buffer is sparse buffer is bound completely and contiguously to a single VkDeviceMemory object buffer was created with the VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT flag and the VkPhysicalDeviceBufferDeviceAddressFeaturesEXT::bufferDeviceAddress feature is enabled on the device", 476}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferDeviceAddress-device-03325", 42}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice or VkPhysicalDeviceBufferDeviceAddressFeaturesEXT::bufferDeviceAddressMultiDevice feature must be enabled", 199}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferDeviceAddress-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferDeviceAddress-pInfo-parameter", 45}, { {"pInfo must be a valid pointer to a valid VkBufferDeviceAddressInfo structure", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements-buffer-parameter", 51}, { {"buffer must be a valid VkBuffer handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements-buffer-parent", 48}, { {"buffer must have been created, allocated, or retrieved from device", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements-pMemoryRequirements-parameter", 64}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements structure", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements2-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements2-pInfo-parameter", 51}, { {"pInfo must be a valid pointer to a valid VkBufferMemoryRequirementsInfo2 structure", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferMemoryRequirements2-pMemoryRequirements-parameter", 65}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureAddress-None-03326", 47}, { {"The bufferDeviceAddress and bufferDeviceAddressCaptureReplay features must be enabled", 85}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureAddress-device-03327", 49}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureAddress-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureAddress-pInfo-10725", 48}, { {"pInfo->buffer must have been created with the VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT flag", 101}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureAddress-pInfo-parameter", 52}, { {"pInfo must be a valid pointer to a valid VkBufferDeviceAddressInfo structure", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-None-08072", 57}, { {"The descriptorBufferCaptureReplay feature must be enabled", 57}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-device-08074", 59}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-device-parameter", 63}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-pData-08073", 58}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferPropertiesEXT::bufferCaptureReplayDescriptorDataSize bytes in size", 144}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-pData-parameter", 62}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetBufferOpaqueCaptureDescriptorDataEXT-pInfo-parameter", 62}, { {"pInfo must be a valid pointer to a valid VkBufferCaptureDescriptorDataInfoEXT structure", 87}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-pMaxDeviation-parameter", 57}, { {"pMaxDeviation must be a valid pointer to a uint64_t value", 57}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-pTimestampInfos-parameter", 59}, { {"pTimestampInfos must be a valid pointer to an array of timestampCount valid VkCalibratedTimestampInfoKHR structures", 115}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-pTimestamps-parameter", 55}, { {"pTimestamps must be a valid pointer to an array of timestampCount uint64_t values", 81}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-timeDomain-09246", 50}, { {"The timeDomain value of each VkCalibratedTimestampInfoKHR in pTimestampInfos must be unique except for those with a value of VK_TIME_DOMAIN_PRESENT_STAGE_LOCAL_EXT", 163}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetCalibratedTimestampsKHR-timestampCount-arraylength", 60}, { {"timestampCount must be greater than 0", 37}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetClusterAccelerationStructureBuildSizesNV-clusterAccelerationStructure-10438", 85}, { {"The VkPhysicalDeviceClusterAccelerationStructureFeaturesNV::clusterAccelerationStructure feature must be enabled", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetClusterAccelerationStructureBuildSizesNV-device-parameter", 67}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetClusterAccelerationStructureBuildSizesNV-pInfo-parameter", 66}, { {"pInfo must be a valid pointer to a valid VkClusterAccelerationStructureInputInfoNV structure", 92}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetClusterAccelerationStructureBuildSizesNV-pSizeInfo-parameter", 70}, { {"pSizeInfo must be a valid pointer to a VkAccelerationStructureBuildSizesInfoKHR structure", 89}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetCudaModuleCacheNV-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetCudaModuleCacheNV-module-parameter", 44}, { {"module must be a valid VkCudaModuleNV handle", 44}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetCudaModuleCacheNV-module-parent", 41}, { {"module must have been created, allocated, or retrieved from device", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetCudaModuleCacheNV-pCacheData-parameter", 48}, { {"If the value referenced by pCacheSize is not 0, and pCacheData is not NULL, pCacheData must be a valid pointer to an array of pCacheSize bytes", 142}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetCudaModuleCacheNV-pCacheSize-parameter", 48}, { {"pCacheSize must be a valid pointer to a size_t value", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetDataGraphPipelineAvailablePropertiesARM-dataGraphPipeline-09888", 73}, { {"The dataGraphPipeline member of pPipelineInfo must have been created with device", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineAvailablePropertiesARM-device-parameter", 66}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineAvailablePropertiesARM-pPipelineInfo-parameter", 73}, { {"pPipelineInfo must be a valid pointer to a valid VkDataGraphPipelineInfoARM structure", 85}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineAvailablePropertiesARM-pProperties-parameter", 71}, { {"If the value referenced by pPropertiesCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertiesCount VkDataGraphPipelinePropertyARM values", 188}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineAvailablePropertiesARM-pPropertiesCount-parameter", 76}, { {"pPropertiesCount must be a valid pointer to a uint32_t value", 60}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-dataGraphPipeline-09802", 64}, { {"The dataGraphPipeline member of pPipelineInfo must have been created with device", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-pPipelineInfo-parameter", 64}, { {"pPipelineInfo must be a valid pointer to a valid VkDataGraphPipelineInfoARM structure", 85}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-pProperties-09856", 58}, { {"If pProperties includes a VkDataGraphPipelinePropertyQueryResultARM whose property member is VK_DATA_GRAPH_PIPELINE_PROPERTY_NEURAL_ACCELERATOR_STATISTICS_INFO_ARM then dataGraphPipeline must have been created with a VkDataGraphPipelineNeuralStatisticsCreateInfoARM whose allowNeuralStatistics member was VK_TRUE included in the pNext chain of VkDataGraphPipelineCreateInfoARM.", 377}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-pProperties-09889", 58}, { {"There must not be two or more structures in the pProperties array with the same VkDataGraphPipelinePropertyQueryResultARM::property", 131}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-pProperties-parameter", 62}, { {"pProperties must be a valid pointer to an array of propertiesCount VkDataGraphPipelinePropertyQueryResultARM structures", 119}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelinePropertiesARM-propertiesCount-arraylength", 68}, { {"propertiesCount must be greater than 0", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionBindPointRequirementsARM-device-parameter", 75}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionBindPointRequirementsARM-pBindPointRequirementCount-parameter", 95}, { {"pBindPointRequirementCount must be a valid pointer to a uint32_t value", 70}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionBindPointRequirementsARM-pBindPointRequirements-parameter", 91}, { {"If the value referenced by pBindPointRequirementCount is not 0, and pBindPointRequirements is not NULL, pBindPointRequirements must be a valid pointer to an array of pBindPointRequirementCount VkDataGraphPipelineSessionBindPointRequirementARM structures", 253}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionBindPointRequirementsARM-pInfo-parameter", 74}, { {"pInfo must be a valid pointer to a valid VkDataGraphPipelineSessionBindPointRequirementsInfoARM structure", 105}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionBindPointRequirementsARM-session-09783", 72}, { {"The session member of pInfo must have been created with device", 62}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionMemoryRequirementsARM-bindPoint-09784", 71}, { {"The bindPoint member of pInfo must have been returned as part of a VkDataGraphPipelineSessionBindPointRequirementARM whose bindPointType member is VK_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_TYPE_MEMORY_ARM by a prior call to vkGetDataGraphPipelineSessionBindPointRequirementsARM for the session member of pInfo", 309}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionMemoryRequirementsARM-device-parameter", 72}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionMemoryRequirementsARM-pInfo-parameter", 71}, { {"pInfo must be a valid pointer to a valid VkDataGraphPipelineSessionMemoryRequirementsInfoARM structure", 102}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionMemoryRequirementsARM-pMemoryRequirements-parameter", 85}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDataGraphPipelineSessionMemoryRequirementsARM-session-09950", 69}, { {"The session member of pInfo must have been created with device", 62}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetDeferredOperationMaxConcurrencyKHR-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDeferredOperationMaxConcurrencyKHR-operation-parameter", 64}, { {"operation must be a valid VkDeferredOperationKHR handle", 55}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDeferredOperationMaxConcurrencyKHR-operation-parent", 61}, { {"operation must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDeferredOperationResultKHR-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDeferredOperationResultKHR-operation-parameter", 56}, { {"operation must be a valid VkDeferredOperationKHR handle", 55}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDeferredOperationResultKHR-operation-parent", 53}, { {"operation must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_deferred_host_operations/deferred_host_operations.html", 70} } },
        { {"VUID-vkGetDescriptorEXT-None-08015", 34}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-dataSize-08125", 38}, { {"If pDescriptorInfo->type is not VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER or pDescriptorInfo->data.pCombinedImageSampler has an imageView member that was not created with a VkSamplerYcbcrConversionInfo structure in its pNext chain, dataSize must equal the size of a descriptor of type VkDescriptorGetInfoEXT::type determined by the value in VkPhysicalDeviceDescriptorBufferPropertiesEXT , or determined by VkPhysicalDeviceDescriptorBufferDensityMapPropertiesEXT::combinedImageSamplerDensityMapDescriptorSize if pDescriptorInfo specifies a VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER whose VkSampler was created with VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT set", 654}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-dataSize-arraylength", 44}, { {"dataSize must be greater than 0", 31}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-descriptorType-09469", 44}, { {"If pDescriptorInfo->type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER and pDescriptorInfo->data.pCombinedImageSampler has an imageView member that was created with a VkSamplerYcbcrConversionInfo structure in its pNext chain, dataSize must equal the size of VkPhysicalDeviceDescriptorBufferPropertiesEXT::combinedImageSamplerDescriptorSize times VkSamplerYcbcrConversionImageFormatProperties::combinedImageSamplerDescriptorCount", 427}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-pDescriptor-parameter", 45}, { {"pDescriptor must be a valid pointer to an array of dataSize bytes", 65}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-pDescriptorInfo-09507", 45}, { {"If pDescriptorInfo->type is VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER and it has a imageView that is VK_NULL_HANDLE then dataSize must be equal to the size of VkPhysicalDeviceDescriptorBufferPropertiesEXT::combinedImageSamplerDescriptorSize", 240}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorEXT-pDescriptorInfo-parameter", 49}, { {"pDescriptorInfo must be a valid pointer to a valid VkDescriptorGetInfoEXT structure", 83}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetHostMappingVALVE-descriptorSet-parameter", 63}, { {"descriptorSet must be a valid VkDescriptorSet handle", 52}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetHostMappingVALVE-descriptorSet-parent", 60}, { {"descriptorSet must have been created, allocated, or retrieved from device", 73}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetHostMappingVALVE-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetHostMappingVALVE-ppData-parameter", 56}, { {"ppData must be a valid pointer to a pointer value", 49}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-None-08013", 56}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-layout-08014", 58}, { {"layout must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT flag set", 105}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-layout-11813", 58}, { {"layout must have not been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT flag set", 103}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-layout-11814", 58}, { {"layout must have not been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT flag set", 119}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-layout-parameter", 62}, { {"layout must be a valid VkDescriptorSetLayout handle", 51}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-layout-parent", 59}, { {"layout must have been created, allocated, or retrieved from device", 66}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutBindingOffsetEXT-pOffset-parameter", 63}, { {"pOffset must be a valid pointer to a VkDeviceSize value", 55}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutHostMappingInfoVALVE-device-parameter", 66}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetLayoutHostMappingInfoVALVE-pBindingReference-parameter", 77}, { {"pBindingReference must be a valid pointer to a valid VkDescriptorSetBindingReferenceVALVE structure", 99}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetLayoutHostMappingInfoVALVE-pHostMapping-parameter", 72}, { {"pHostMapping must be a valid pointer to a VkDescriptorSetLayoutHostMappingInfoVALVE structure", 93}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-None-08011", 47}, { {"The descriptorBuffer feature must be enabled", 44}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-layout-08012", 49}, { {"layout must have been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT flag set", 105}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-layout-11811", 49}, { {"layout must have not been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT flag set", 103}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-layout-11812", 49}, { {"layout must have not been created with the VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT flag set", 119}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-layout-parameter", 53}, { {"layout must be a valid VkDescriptorSetLayout handle", 51}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-layout-parent", 50}, { {"layout must have been created, allocated, or retrieved from device", 66}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSizeEXT-pLayoutSizeInBytes-parameter", 65}, { {"pLayoutSizeInBytes must be a valid pointer to a VkDeviceSize value", 66}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetDescriptorSetLayoutSupport-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkGetDescriptorSetLayoutSupport-pCreateInfo-parameter", 58}, { {"pCreateInfo must be a valid pointer to a valid VkDescriptorSetLayoutCreateInfo structure", 88}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkGetDescriptorSetLayoutSupport-pSupport-parameter", 55}, { {"pSupport must be a valid pointer to a VkDescriptorSetLayoutSupport structure", 76}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkGetDeviceAccelerationStructureCompatibilityKHR-accelerationStructure-08928", 81}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructure feature must be enabled", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetDeviceAccelerationStructureCompatibilityKHR-device-parameter", 70}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetDeviceAccelerationStructureCompatibilityKHR-pCompatibility-parameter", 78}, { {"pCompatibility must be a valid pointer to a VkAccelerationStructureCompatibilityKHR value", 89}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetDeviceAccelerationStructureCompatibilityKHR-pVersionInfo-parameter", 76}, { {"pVersionInfo must be a valid pointer to a valid VkAccelerationStructureVersionInfoKHR structure", 95}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetDeviceBufferMemoryRequirements-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceBufferMemoryRequirements-pInfo-parameter", 56}, { {"pInfo must be a valid pointer to a valid VkDeviceBufferMemoryRequirements structure", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceBufferMemoryRequirements-pMemoryRequirements-parameter", 70}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceCombinedImageSamplerIndexNVX-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceFaultDebugInfoKHR-device-12383", 46}, { {"device must be lost", 19}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultDebugInfoKHR-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultDebugInfoKHR-pDebugInfo-parameter", 54}, { {"pDebugInfo must be a valid pointer to a VkDeviceFaultDebugInfoKHR structure", 75}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-device-07336", 41}, { {"device must be in the lost state", 32}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-pFaultCounts-07337", 47}, { {"If the value referenced by pFaultCounts->addressInfoCount is not 0, and pFaultInfo->pAddressInfos is not NULL, pFaultInfo->pAddressInfos must be a valid pointer to an array of pFaultCounts->addressInfoCount VkDeviceFaultAddressInfoEXT structures", 245}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-pFaultCounts-07338", 47}, { {"If the value referenced by pFaultCounts->vendorInfoCount is not 0, and pFaultInfo->pVendorInfos is not NULL, pFaultInfo->pVendorInfos must be a valid pointer to an array of pFaultCounts->vendorInfoCount VkDeviceFaultVendorInfoEXT structures", 240}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-pFaultCounts-07339", 47}, { {"If the value referenced by pFaultCounts->vendorBinarySize is not 0, and pFaultInfo->pVendorBinaryData is not NULL, pFaultInfo->pVendorBinaryData must be a valid pointer to an array of pFaultCounts->vendorBinarySize bytes", 220}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-pFaultCounts-parameter", 51}, { {"pFaultCounts must be a valid pointer to a VkDeviceFaultCountsEXT structure", 74}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultInfoEXT-pFaultInfo-parameter", 49}, { {"If pFaultInfo is not NULL, pFaultInfo must be a valid pointer to a VkDeviceFaultInfoEXT structure", 97}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultReportsKHR-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultReportsKHR-pFaultCounts-arraylength", 56}, { {"If pFaultInfo is not NULL, the value referenced by pFaultCounts must be greater than 0", 86}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultReportsKHR-pFaultCounts-parameter", 54}, { {"pFaultCounts must be a valid pointer to a uint32_t value", 56}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceFaultReportsKHR-pFaultInfo-parameter", 52}, { {"If pFaultInfo is not NULL, pFaultInfo must be a valid pointer to an array of pFaultCounts VkDeviceFaultInfoKHR structures", 121}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-heapIndex-00691", 55}, { {"heapIndex must be less than memoryHeapCount", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-localDeviceIndex-00692", 62}, { {"localDeviceIndex must be a valid device index", 45}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-localDeviceIndex-00694", 62}, { {"localDeviceIndex must not equal remoteDeviceIndex", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-pPeerMemoryFeatures-parameter", 69}, { {"pPeerMemoryFeatures must be a valid pointer to a VkPeerMemoryFeatureFlags value", 79}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPeerMemoryFeatures-remoteDeviceIndex-00693", 63}, { {"remoteDeviceIndex must be a valid device index", 46}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceGroupPresentCapabilitiesKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupPresentCapabilitiesKHR-pDeviceGroupPresentCapabilities-parameter", 85}, { {"pDeviceGroupPresentCapabilities must be a valid pointer to a VkDeviceGroupPresentCapabilitiesKHR structure", 106}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModes2EXT-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModes2EXT-pModes-parameter", 61}, { {"pModes must be a valid pointer to a VkDeviceGroupPresentModeFlagsKHR value", 74}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModes2EXT-pSurfaceInfo-06213", 63}, { {"pSurfaceInfo->surface must be supported by all physical devices associated with device, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 184}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModes2EXT-pSurfaceInfo-parameter", 67}, { {"pSurfaceInfo must be a valid pointer to a valid VkPhysicalDeviceSurfaceInfo2KHR structure", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModesKHR-commonparent", 56}, { {"Both of device, and surface must have been created, allocated, or retrieved from the same VkInstance", 100}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModesKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModesKHR-pModes-parameter", 60}, { {"pModes must be a valid pointer to a VkDeviceGroupPresentModeFlagsKHR value", 74}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModesKHR-surface-06212", 57}, { {"surface must be supported by all physical devices associated with device, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 170}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceGroupSurfacePresentModesKHR-surface-parameter", 61}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDeviceImageMemoryRequirements-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceImageMemoryRequirements-pInfo-parameter", 55}, { {"pInfo must be a valid pointer to a valid VkDeviceImageMemoryRequirements structure", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceImageMemoryRequirements-pMemoryRequirements-parameter", 69}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceImageSparseMemoryRequirements-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetDeviceImageSparseMemoryRequirements-pInfo-parameter", 61}, { {"pInfo must be a valid pointer to a valid VkDeviceImageMemoryRequirements structure", 82}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetDeviceImageSparseMemoryRequirements-pSparseMemoryRequirementCount-parameter", 85}, { {"pSparseMemoryRequirementCount must be a valid pointer to a uint32_t value", 73}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetDeviceImageSparseMemoryRequirements-pSparseMemoryRequirements-parameter", 81}, { {"If the value referenced by pSparseMemoryRequirementCount is not 0, and pSparseMemoryRequirements is not NULL, pSparseMemoryRequirements must be a valid pointer to an array of pSparseMemoryRequirementCount VkSparseImageMemoryRequirements2 structures", 248}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetDeviceImageSubresourceLayout-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceImageSubresourceLayout-pInfo-parameter", 54}, { {"pInfo must be a valid pointer to a valid VkDeviceImageSubresourceInfo structure", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceImageSubresourceLayout-pLayout-parameter", 56}, { {"pLayout must be a valid pointer to a VkSubresourceLayout2 structure", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceMemoryCommitment-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryCommitment-memory-00690", 45}, { {"memory must have been created with a memory type that reports VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT", 101}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryCommitment-memory-parameter", 49}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryCommitment-memory-parent", 46}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryCommitment-pCommittedMemoryInBytes-parameter", 66}, { {"pCommittedMemoryInBytes must be a valid pointer to a VkDeviceSize value", 71}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryOpaqueCaptureAddress-None-03334", 53}, { {"The bufferDeviceAddress and bufferDeviceAddressCaptureReplay features must be enabled", 85}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryOpaqueCaptureAddress-device-03335", 55}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryOpaqueCaptureAddress-device-parameter", 59}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryOpaqueCaptureAddress-pInfo-10727", 54}, { {"pInfo->memory must have been allocated using the VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT flag", 106}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMemoryOpaqueCaptureAddress-pInfo-parameter", 58}, { {"pInfo must be a valid pointer to a valid VkDeviceMemoryOpaqueCaptureAddressInfo structure", 89}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetDeviceMicromapCompatibilityEXT-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkGetDeviceMicromapCompatibilityEXT-micromap-11666", 55}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 80}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkGetDeviceMicromapCompatibilityEXT-pCompatibility-parameter", 65}, { {"pCompatibility must be a valid pointer to a VkAccelerationStructureCompatibilityKHR value", 89}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkGetDeviceMicromapCompatibilityEXT-pVersionInfo-parameter", 63}, { {"pVersionInfo must be a valid pointer to a valid VkMicromapVersionInfoEXT structure", 82}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkGetDeviceProcAddr-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkGetDeviceProcAddr-pName-parameter", 40}, { {"pName must be a null-terminated UTF-8 string", 44}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkGetDeviceQueue-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue-flags-01841", 33}, { {"VkDeviceQueueCreateInfo::flags must have been zero when device was created", 74}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue-pQueue-parameter", 38}, { {"pQueue must be a valid pointer to a VkQueue handle", 50}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue-queueFamilyIndex-00384", 44}, { {"queueFamilyIndex must be one of the queue family indices specified when device was created, via the VkDeviceQueueCreateInfo structure", 133}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue-queueIndex-00385", 38}, { {"queueIndex must be less than the value of VkDeviceQueueCreateInfo::queueCount for the queue family indicated by queueFamilyIndex when device was created", 152}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue2-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue2-pQueue-parameter", 39}, { {"pQueue must be a valid pointer to a VkQueue handle", 50}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceQueue2-pQueueInfo-parameter", 43}, { {"pQueueInfo must be a valid pointer to a valid VkDeviceQueueInfo2 structure", 74}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI-device-parameter", 69}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI-pMaxWorkgroupSize-parameter", 80}, { {"pMaxWorkgroupSize must be a valid pointer to VkExtent2D structures", 66}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI-renderpass-parameter", 73}, { {"renderpass must be a valid VkRenderPass handle", 46}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI-renderpass-parent", 70}, { {"renderpass must have been created, allocated, or retrieved from device", 70}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetDeviceTensorMemoryRequirementsARM-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceTensorMemoryRequirementsARM-pInfo-parameter", 59}, { {"pInfo must be a valid pointer to a valid VkDeviceTensorMemoryRequirementsARM structure", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceTensorMemoryRequirementsARM-pMemoryRequirements-parameter", 73}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDeviceTensorMemoryRequirementsARM-tensors-09831", 57}, { {"The tensors feature must be enabled", 35}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetDisplayModeProperties2KHR-display-parameter", 53}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModeProperties2KHR-display-parent", 50}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModeProperties2KHR-pProperties-parameter", 57}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayModeProperties2KHR structures", 185}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModeProperties2KHR-pPropertyCount-parameter", 60}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModeProperties2KHR-physicalDevice-parameter", 60}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModePropertiesKHR-display-parameter", 52}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModePropertiesKHR-display-parent", 49}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModePropertiesKHR-pProperties-parameter", 56}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayModePropertiesKHR structures", 184}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModePropertiesKHR-pPropertyCount-parameter", 59}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayModePropertiesKHR-physicalDevice-parameter", 59}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilities2KHR-pCapabilities-parameter", 62}, { {"pCapabilities must be a valid pointer to a VkDisplayPlaneCapabilities2KHR structure", 83}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilities2KHR-pDisplayPlaneInfo-parameter", 66}, { {"pDisplayPlaneInfo must be a valid pointer to a valid VkDisplayPlaneInfo2KHR structure", 85}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilities2KHR-physicalDevice-parameter", 63}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilitiesKHR-mode-parameter", 52}, { {"mode must be a valid VkDisplayModeKHR handle", 44}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilitiesKHR-mode-parent", 49}, { {"mode must have been created, allocated, or retrieved from physicalDevice", 72}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilitiesKHR-pCapabilities-parameter", 61}, { {"pCapabilities must be a valid pointer to a VkDisplayPlaneCapabilitiesKHR structure", 82}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneCapabilitiesKHR-physicalDevice-parameter", 62}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneSupportedDisplaysKHR-pDisplayCount-parameter", 66}, { {"pDisplayCount must be a valid pointer to a uint32_t value", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneSupportedDisplaysKHR-pDisplays-parameter", 62}, { {"If the value referenced by pDisplayCount is not 0, and pDisplays is not NULL, pDisplays must be a valid pointer to an array of pDisplayCount VkDisplayKHR handles", 161}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneSupportedDisplaysKHR-physicalDevice-parameter", 67}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDisplayPlaneSupportedDisplaysKHR-planeIndex-01249", 59}, { {"planeIndex must be less than the number of display planes supported by the device as determined by calling vkGetPhysicalDeviceDisplayPlanePropertiesKHR", 151}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDrmDisplayEXT-display-parameter", 41}, { {"display must be a valid pointer to a VkDisplayKHR handle", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDrmDisplayEXT-physicalDevice-parameter", 48}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetDynamicRenderingTilePropertiesQCOM-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetDynamicRenderingTilePropertiesQCOM-pProperties-parameter", 66}, { {"pProperties must be a valid pointer to a VkTilePropertiesQCOM structure", 71}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetDynamicRenderingTilePropertiesQCOM-pRenderingInfo-parameter", 69}, { {"pRenderingInfo must be a valid pointer to a valid VkRenderingInfo structure", 75}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-device-parameter", 59}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pData-parameter", 58}, { {"If the value referenced by pDataSize is not 0, and pData is not NULL, pData must be a valid pointer to an array of pDataSize bytes", 130}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pDataSize-parameter", 62}, { {"pDataSize must be a valid pointer to a size_t value", 51}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pFeedbackInfo-parameter", 66}, { {"If pFeedbackInfo is not NULL, pFeedbackInfo must be a valid pointer to a VkVideoEncodeSessionParametersFeedbackInfoKHR structure", 128}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08262", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain of pVideoSessionParametersInfo must include a VkVideoEncodeH264SessionParametersGetInfoKHR structure", 265}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08263", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then for the VkVideoEncodeH264SessionParametersGetInfoKHR structure included in the pNext chain of pVideoSessionParametersInfo, if its writeStdSPS member is VK_TRUE, then pVideoSessionParametersInfo->videoSessionParameters must contain a StdVideoH264SequenceParameterSet entry with seq_parameter_set_id matching VkVideoEncodeH264SessionParametersGetInfoKHR::stdSPSId", 510}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08264", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then for the VkVideoEncodeH264SessionParametersGetInfoKHR structure included in the pNext chain of pVideoSessionParametersInfo, if its writeStdPPS member is VK_TRUE, then pVideoSessionParametersInfo->videoSessionParameters must contain a StdVideoH264PictureParameterSet entry with seq_parameter_set_id and pic_parameter_set_id matching VkVideoEncodeH264SessionParametersGetInfoKHR::stdSPSId and VkVideoEncodeH264SessionParametersGetInfoKHR::stdPPSId, respectively", 607}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08265", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain of pVideoSessionParametersInfo must include a VkVideoEncodeH265SessionParametersGetInfoKHR structure", 265}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08266", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then for the VkVideoEncodeH265SessionParametersGetInfoKHR structure included in the pNext chain of pVideoSessionParametersInfo, if its writeStdVPS member is VK_TRUE, then pVideoSessionParametersInfo->videoSessionParameters must contain a StdVideoH265VideoParameterSet entry with vps_video_parameter_set_id matching VkVideoEncodeH265SessionParametersGetInfoKHR::stdVPSId", 513}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08267", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then for the VkVideoEncodeH265SessionParametersGetInfoKHR structure included in the pNext chain of pVideoSessionParametersInfo, if its writeStdSPS member is VK_TRUE, then pVideoSessionParametersInfo->videoSessionParameters must contain a StdVideoH265SequenceParameterSet entry with sps_video_parameter_set_id and sps_seq_parameter_set_id matching VkVideoEncodeH265SessionParametersGetInfoKHR::stdVPSId and VkVideoEncodeH265SessionParametersGetInfoKHR::stdSPSId, respectively", 618}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08268", 76}, { {"If pVideoSessionParametersInfo->videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then for the VkVideoEncodeH265SessionParametersGetInfoKHR structure included in the pNext chain of pVideoSessionParametersInfo, if its writeStdPPS member is VK_TRUE, then pVideoSessionParametersInfo->videoSessionParameters must contain a StdVideoH265PictureParameterSet entry with sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id matching VkVideoEncodeH265SessionParametersGetInfoKHR::stdVPSId, VkVideoEncodeH265SessionParametersGetInfoKHR::stdSPSId, and VkVideoEncodeH265SessionParametersGetInfoKHR::stdPPSId, respectively", 701}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-08359", 76}, { {"pVideoSessionParametersInfo->videoSessionParameters must have been created with an encode operation", 99}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEncodedVideoSessionParametersKHR-pVideoSessionParametersInfo-parameter", 80}, { {"pVideoSessionParametersInfo must be a valid pointer to a valid VkVideoEncodeSessionParametersGetInfoKHR structure", 113}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetEventStatus-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetEventStatus-event-03940", 33}, { {"event must not have been created with VK_EVENT_CREATE_DEVICE_ONLY_BIT", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetEventStatus-event-parameter", 37}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetEventStatus-event-parent", 34}, { {"event must have been created, allocated, or retrieved from device", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-executionGraph-09142", 66}, { {"There must be a node in executionGraph with a shader name and index equal to pNodeInfo->pName and pNodeInfo->index", 114}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-executionGraph-parameter", 70}, { {"executionGraph must be a valid VkPipeline handle", 48}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-executionGraph-parent", 67}, { {"executionGraph must have been created, allocated, or retrieved from device", 74}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-pNodeIndex-parameter", 66}, { {"pNodeIndex must be a valid pointer to a uint32_t value", 54}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-pNodeInfo-09140", 61}, { {"pNodeInfo->pName must not be NULL", 33}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-pNodeInfo-09141", 61}, { {"pNodeInfo->index must not be VK_SHADER_INDEX_UNUSED_AMDX", 56}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineNodeIndexAMDX-pNodeInfo-parameter", 65}, { {"pNodeInfo must be a valid pointer to a valid VkPipelineShaderStageNodeCreateInfoAMDX structure", 94}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineScratchSizeAMDX-device-parameter", 64}, { {"device must be a valid VkDevice handle", 38}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineScratchSizeAMDX-executionGraph-parameter", 72}, { {"executionGraph must be a valid VkPipeline handle", 48}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineScratchSizeAMDX-executionGraph-parent", 69}, { {"executionGraph must have been created, allocated, or retrieved from device", 74}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExecutionGraphPipelineScratchSizeAMDX-pSizeInfo-parameter", 67}, { {"pSizeInfo must be a valid pointer to a VkExecutionGraphPipelineScratchSizeAMDX structure", 88}, { "chapters/executiongraphs.html", 29} } },
        { {"VUID-vkGetExternalComputeQueueDataNV-externalQueue-parameter", 60}, { {"externalQueue must be a valid VkExternalComputeQueueNV handle", 61}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkGetExternalComputeQueueDataNV-pData-08134", 48}, { {"pData must be at least the size specified by the externalDataSize field in the VkPhysicalDeviceExternalComputeQueuePropertiesNV structure", 137}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkGetExternalComputeQueueDataNV-pData-parameter", 52}, { {"pData must be a pointer value", 29}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkGetExternalComputeQueueDataNV-params-parameter", 53}, { {"params must be a valid pointer to a VkExternalComputeQueueDataParamsNV structure", 80}, { "chapters/VK_NV_external_compute_queue/VK_NV_external_compute_queue.html", 71} } },
        { {"VUID-vkGetFenceFdKHR-device-parameter", 37}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceFdKHR-pFd-parameter", 34}, { {"pFd must be a valid pointer to an int value", 43}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceFdKHR-pGetFdInfo-parameter", 41}, { {"pGetFdInfo must be a valid pointer to a valid VkFenceGetFdInfoKHR structure", 75}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceStatus-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceStatus-fence-parameter", 37}, { {"fence must be a valid VkFence handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceStatus-fence-parent", 34}, { {"fence must have been created, allocated, or retrieved from device", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceWin32HandleKHR-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceWin32HandleKHR-pGetWin32HandleInfo-parameter", 59}, { {"pGetWin32HandleInfo must be a valid pointer to a valid VkFenceGetWin32HandleInfoKHR structure", 93}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFenceWin32HandleKHR-pHandle-parameter", 47}, { {"pHandle must be a valid pointer to a HANDLE value", 49}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetFramebufferTilePropertiesQCOM-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkGetFramebufferTilePropertiesQCOM-framebuffer-parameter", 61}, { {"framebuffer must be a valid VkFramebuffer handle", 48}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkGetFramebufferTilePropertiesQCOM-framebuffer-parent", 58}, { {"framebuffer must have been created, allocated, or retrieved from device", 71}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkGetFramebufferTilePropertiesQCOM-pProperties-parameter", 61}, { {"If the value referenced by pPropertiesCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertiesCount VkTilePropertiesQCOM structures", 182}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkGetFramebufferTilePropertiesQCOM-pPropertiesCount-parameter", 66}, { {"pPropertiesCount must be a valid pointer to a uint32_t value", 60}, { "chapters/framebuffer.html", 25} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsEXT-device-parameter", 65}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsEXT-pInfo-parameter", 64}, { {"pInfo must be a valid pointer to a valid VkGeneratedCommandsMemoryRequirementsInfoEXT structure", 95}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsEXT-pMemoryRequirements-parameter", 78}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsNV-device-parameter", 64}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsNV-deviceGeneratedCommands-02906", 77}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV::deviceGeneratedCommands feature must be enabled", 102}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsNV-pInfo-09074", 59}, { {"If pInfo->pipelineBindPoint is of type VK_PIPELINE_BIND_POINT_COMPUTE, then the VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedCompute feature must be enabled", 184}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsNV-pInfo-parameter", 63}, { {"pInfo must be a valid pointer to a valid VkGeneratedCommandsMemoryRequirementsInfoNV structure", 94}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGeneratedCommandsMemoryRequirementsNV-pMemoryRequirements-parameter", 77}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetGpaDeviceClockInfoAMD-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaDeviceClockInfoAMD-pInfo-parameter", 47}, { {"pInfo must be a valid pointer to a VkGpaDeviceGetClockInfoAMD structure", 71}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionResultsAMD-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionResultsAMD-gpaSession-parameter", 51}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionResultsAMD-gpaSession-parent", 48}, { {"gpaSession must have been created, allocated, or retrieved from device", 70}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionResultsAMD-pData-parameter", 46}, { {"If the value referenced by pSizeInBytes is not 0, and pData is not NULL, pData must be a valid pointer to an array of pSizeInBytes bytes", 136}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionResultsAMD-pSizeInBytes-parameter", 53}, { {"pSizeInBytes must be a valid pointer to a size_t value", 54}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionStatusAMD-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionStatusAMD-gpaSession-parameter", 50}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetGpaSessionStatusAMD-gpaSession-parent", 47}, { {"gpaSession must have been created, allocated, or retrieved from device", 70}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkGetImageDrmFormatModifierPropertiesEXT-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageDrmFormatModifierPropertiesEXT-image-02272", 57}, { {"image must have been created with tiling equal to VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageDrmFormatModifierPropertiesEXT-image-parameter", 61}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageDrmFormatModifierPropertiesEXT-image-parent", 58}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageDrmFormatModifierPropertiesEXT-pProperties-parameter", 67}, { {"pProperties must be a valid pointer to a VkImageDrmFormatModifierPropertiesEXT structure", 88}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-image-01588", 45}, { {"image must not have been created with the VK_IMAGE_CREATE_DISJOINT_BIT flag set", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-image-04004", 45}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID external memory handle type, then image must be bound to memory", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-image-08960", 45}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX external memory handle type, then image must be bound to memory", 146}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-image-parameter", 49}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-image-parent", 46}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements-pMemoryRequirements-parameter", 63}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements structure", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements2-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements2-pInfo-parameter", 50}, { {"pInfo must be a valid pointer to a valid VkImageMemoryRequirementsInfo2 structure", 81}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageMemoryRequirements2-pMemoryRequirements-parameter", 64}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-descriptorHeapCaptureReplay-11282", 69}, { {"The descriptorHeapCaptureReplay feature must be enabled", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-device-11284", 48}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-imageCount-arraylength", 58}, { {"imageCount must be greater than 0", 33}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-pDatas-parameter", 52}, { {"pDatas must be a valid pointer to an array of imageCount VkHostAddressRangeEXT structures", 89}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-pImages-11285", 49}, { {"Each element of pImages must have been created with VK_IMAGE_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_EXT set in VkImageCreateInfo::flags", 138}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-pImages-parameter", 53}, { {"pImages must be a valid pointer to an array of imageCount valid VkImage handles", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-pImages-parent", 50}, { {"Each element of pImages must have been created, allocated, or retrieved from device", 83}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDataEXT-size-11283", 46}, { {"The size member of each element of pDatas must be equal to imageCaptureReplayOpaqueDataSize", 91}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-None-08076", 56}, { {"The descriptorBufferCaptureReplay feature must be enabled", 57}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-device-08078", 58}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-pData-08077", 57}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferPropertiesEXT::imageCaptureReplayDescriptorDataSize bytes in size", 143}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-pData-parameter", 61}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageOpaqueCaptureDescriptorDataEXT-pInfo-parameter", 61}, { {"pInfo must be a valid pointer to a valid VkImageCaptureDescriptorDataInfoEXT structure", 86}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageSparseMemoryRequirements-device-parameter", 56}, { {"device must be a valid VkDevice handle", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements-image-parameter", 55}, { {"image must be a valid VkImage handle", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements-image-parent", 52}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements-pSparseMemoryRequirementCount-parameter", 79}, { {"pSparseMemoryRequirementCount must be a valid pointer to a uint32_t value", 73}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements-pSparseMemoryRequirements-parameter", 75}, { {"If the value referenced by pSparseMemoryRequirementCount is not 0, and pSparseMemoryRequirements is not NULL, pSparseMemoryRequirements must be a valid pointer to an array of pSparseMemoryRequirementCount VkSparseImageMemoryRequirements structures", 247}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements2-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements2-pInfo-parameter", 56}, { {"pInfo must be a valid pointer to a valid VkImageSparseMemoryRequirementsInfo2 structure", 87}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements2-pSparseMemoryRequirementCount-parameter", 80}, { {"pSparseMemoryRequirementCount must be a valid pointer to a uint32_t value", 73}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSparseMemoryRequirements2-pSparseMemoryRequirements-parameter", 76}, { {"If the value referenced by pSparseMemoryRequirementCount is not 0, and pSparseMemoryRequirements is not NULL, pSparseMemoryRequirements must be a valid pointer to an array of pSparseMemoryRequirementCount VkSparseImageMemoryRequirements2 structures", 248}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-arrayLayer-01717", 49}, { {"The arrayLayer member of pSubresource must be less than the arrayLayers specified in image", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-aspectMask-00997", 49}, { {"The aspectMask member of pSubresource must only have a single bit set", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-format-04462", 45}, { {"If format of the image has a depth component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_DEPTH_BIT", 122}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-format-04463", 45}, { {"If format of the image has a stencil component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_STENCIL_BIT", 126}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-format-04464", 45}, { {"If format of the image does not contain a stencil or depth component, the aspectMask member of pSubresource must not contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 181}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-format-08886", 45}, { {"If format of the image is a color format that is not a multi-planar format, and tiling of the image is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, the aspectMask member of pSubresource must be VK_IMAGE_ASPECT_COLOR_BIT", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-image-07790", 44}, { {"image must have been created with tiling equal to VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT", 115}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-image-09432", 44}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID external memory handle type, then image must be bound to memory", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-image-parameter", 48}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-image-parent", 45}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-mipLevel-01716", 47}, { {"The mipLevel member of pSubresource must be less than the mipLevels specified in image", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-pLayout-parameter", 50}, { {"pLayout must be a valid pointer to a VkSubresourceLayout structure", 66}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-pSubresource-parameter", 55}, { {"pSubresource must be a valid pointer to a valid VkImageSubresource structure", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-tiling-08717", 45}, { {"If the tiling of the image is VK_IMAGE_TILING_LINEAR and has a multi-planar format, then the aspectMask member of pSubresource must be a single valid multi-planar aspect mask bit", 178}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout-tiling-09433", 45}, { {"If the tiling of the image is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then the aspectMask member of pSubresource must be VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT and the index i must be less than the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with the image's format and VkImageDrmFormatModifierPropertiesEXT::drmFormatModifier", 356}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-arrayLayer-01717", 50}, { {"The arrayLayer member of pSubresource must be less than the arrayLayers specified in image", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-aspectMask-00997", 50}, { {"The aspectMask member of pSubresource must only have a single bit set", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-format-04462", 46}, { {"If format of the image has a depth component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_DEPTH_BIT", 122}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-format-04463", 46}, { {"If format of the image has a stencil component, the aspectMask member of pSubresource must contain VK_IMAGE_ASPECT_STENCIL_BIT", 126}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-format-04464", 46}, { {"If format of the image does not contain a stencil or depth component, the aspectMask member of pSubresource must not contain VK_IMAGE_ASPECT_DEPTH_BIT or VK_IMAGE_ASPECT_STENCIL_BIT", 181}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-format-08886", 46}, { {"If format of the image is a color format that is not a multi-planar format, and tiling of the image is VK_IMAGE_TILING_LINEAR or VK_IMAGE_TILING_OPTIMAL, the aspectMask member of pSubresource must be VK_IMAGE_ASPECT_COLOR_BIT", 225}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-image-09434", 45}, { {"If image was created with the VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID external memory handle type, then image must be bound to memory", 160}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-image-parameter", 49}, { {"image must be a valid VkImage handle", 36}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-image-parent", 46}, { {"image must have been created, allocated, or retrieved from device", 65}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-mipLevel-01716", 48}, { {"The mipLevel member of pSubresource must be less than the mipLevels specified in image", 86}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-pLayout-parameter", 51}, { {"pLayout must be a valid pointer to a VkSubresourceLayout2 structure", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-pSubresource-parameter", 56}, { {"pSubresource must be a valid pointer to a valid VkImageSubresource2 structure", 77}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-tiling-08717", 46}, { {"If the tiling of the image is VK_IMAGE_TILING_LINEAR and has a multi-planar format, then the aspectMask member of pSubresource must be a single valid multi-planar aspect mask bit", 178}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageSubresourceLayout2-tiling-09435", 46}, { {"If the tiling of the image is VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT, then the aspectMask member of pSubresource must be VK_IMAGE_ASPECT_MEMORY_PLANE_i_BIT_EXT and the index i must be less than the VkDrmFormatModifierPropertiesEXT::drmFormatModifierPlaneCount associated with the image's format and VkImageDrmFormatModifierPropertiesEXT::drmFormatModifier", 356}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewAddressNVX-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewAddressNVX-imageView-parameter", 49}, { {"imageView must be a valid VkImageView handle", 44}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewAddressNVX-imageView-parent", 46}, { {"imageView must have been created, allocated, or retrieved from device", 69}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewAddressNVX-pProperties-parameter", 51}, { {"pProperties must be a valid pointer to a VkImageViewAddressPropertiesNVX structure", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewHandle64NVX-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewHandle64NVX-pInfo-parameter", 46}, { {"pInfo must be a valid pointer to a valid VkImageViewHandleInfoNVX structure", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewHandleNVX-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewHandleNVX-pInfo-parameter", 44}, { {"pInfo must be a valid pointer to a valid VkImageViewHandleInfoNVX structure", 75}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-None-08080", 60}, { {"The descriptorBufferCaptureReplay feature must be enabled", 57}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-device-08082", 62}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-device-parameter", 66}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-pData-08081", 61}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferPropertiesEXT::imageViewCaptureReplayDescriptorDataSize bytes in size", 147}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-pData-parameter", 65}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetImageViewOpaqueCaptureDescriptorDataEXT-pInfo-parameter", 65}, { {"pInfo must be a valid pointer to a valid VkImageViewCaptureDescriptorDataInfoEXT structure", 90}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetInstanceProcAddr-instance-parameter", 45}, { {"If instance is not NULL, instance must be a valid VkInstance handle", 67}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkGetInstanceProcAddr-pName-parameter", 42}, { {"pName must be a null-terminated UTF-8 string", 44}, { "chapters/initialization.html", 28} } },
        { {"VUID-vkGetLatencyTimingsLegacyNV-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetLatencyTimingsLegacyNV-pTimings-parameter", 51}, { {"pTimings must be a pointer value", 32}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetLatencyTimingsNV-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetLatencyTimingsNV-pLatencyMarkerInfo-parameter", 55}, { {"pLatencyMarkerInfo must be a valid pointer to a VkGetLatencyMarkerInfoNV structure", 82}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetLatencyTimingsNV-swapchain-parameter", 46}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetLatencyTimingsNV-swapchain-parent", 43}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetMemoryAndroidHardwareBufferANDROID-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryAndroidHardwareBufferANDROID-pBuffer-parameter", 62}, { {"pBuffer must be a valid pointer to a valid pointer to an AHardwareBuffer value", 78}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryAndroidHardwareBufferANDROID-pInfo-parameter", 60}, { {"pInfo must be a valid pointer to a valid VkMemoryGetAndroidHardwareBufferInfoANDROID structure", 94}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdKHR-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdKHR-pFd-parameter", 35}, { {"pFd must be a valid pointer to an int value", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdKHR-pGetFdInfo-parameter", 42}, { {"pGetFdInfo must be a valid pointer to a valid VkMemoryGetFdInfoKHR structure", 76}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdPropertiesKHR-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdPropertiesKHR-fd-00673", 40}, { {"fd must point to a valid POSIX file descriptor memory handle", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdPropertiesKHR-handleType-00674", 48}, { {"handleType must not be VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdPropertiesKHR-handleType-parameter", 52}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryFdPropertiesKHR-pMemoryFdProperties-parameter", 61}, { {"pMemoryFdProperties must be a valid pointer to a VkMemoryFdPropertiesKHR structure", 82}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-handleType-01752", 57}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT", 142}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-handleType-01754", 57}, { {"If handleType is VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT, pHostPointer must be a pointer to host memory", 118}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-handleType-01755", 57}, { {"If handleType is VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT, pHostPointer must be a pointer to host mapped foreign memory", 144}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-handleType-parameter", 61}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-pHostPointer-01753", 59}, { {"pHostPointer must be a pointer aligned to an integer multiple of VkPhysicalDeviceExternalMemoryHostPropertiesEXT::minImportedHostPointerAlignment", 145}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-pHostPointer-parameter", 63}, { {"pHostPointer must be a pointer value", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryHostPointerPropertiesEXT-pMemoryHostPointerProperties-parameter", 79}, { {"pMemoryHostPointerProperties must be a valid pointer to a VkMemoryHostPointerPropertiesEXT structure", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandleEXT-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandleEXT-pGetMetalHandleInfo-parameter", 60}, { {"pGetMetalHandleInfo must be a valid pointer to a valid VkMemoryGetMetalHandleInfoEXT structure", 94}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandleEXT-pHandle-parameter", 48}, { {"pHandle must be a valid pointer to a pointer value", 50}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-handle-10416", 53}, { {"pHandle must point to a valid id, id or id", 42}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-handleType-10417", 57}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLBUFFER_BIT_EXT, VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLTEXTURE_BIT_EXT or VK_EXTERNAL_MEMORY_HANDLE_TYPE_MTLHEAP_BIT_EXT", 168}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-handleType-parameter", 61}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-pHandle-parameter", 58}, { {"pHandle must be a pointer value", 31}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryMetalHandlePropertiesEXT-pMemoryMetalHandleProperties-parameter", 79}, { {"pMemoryMetalHandleProperties must be a valid pointer to a VkMemoryMetalHandlePropertiesEXT structure", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryNativeBufferOHOS-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMemoryNativeBufferOHOS-pBuffer-parameter", 50}, { {"pBuffer must be a valid pointer to a valid pointer to an OH_NativeBuffer value", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMemoryNativeBufferOHOS-pInfo-parameter", 48}, { {"pInfo must be a valid pointer to a valid VkMemoryGetNativeBufferInfoOHOS structure", 82}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMemoryRemoteAddressNV-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryRemoteAddressNV-pAddress-parameter", 50}, { {"pAddress must be a valid pointer to a VkRemoteAddressNV value", 61}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryRemoteAddressNV-pMemoryGetRemoteAddressInfo-parameter", 69}, { {"pMemoryGetRemoteAddressInfo must be a valid pointer to a valid VkMemoryGetRemoteAddressInfoNV structure", 103}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleKHR-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleKHR-pGetWin32HandleInfo-parameter", 60}, { {"pGetWin32HandleInfo must be a valid pointer to a valid VkMemoryGetWin32HandleInfoKHR structure", 94}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleKHR-pHandle-parameter", 48}, { {"pHandle must be a valid pointer to a HANDLE value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-handleType-01326", 46}, { {"handleType must be a flag specified in VkExportMemoryAllocateInfoNV::handleTypes when allocating memory", 103}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-handleType-parameter", 50}, { {"handleType must be a valid combination of VkExternalMemoryHandleTypeFlagBitsNV values", 85}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-handleType-requiredbitmask", 56}, { {"handleType must not be 0", 24}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-memory-parameter", 46}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-memory-parent", 43}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandleNV-pHandle-parameter", 47}, { {"pHandle must be a valid pointer to a HANDLE value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandlePropertiesKHR-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandlePropertiesKHR-handle-00665", 53}, { {"handle must point to a valid Windows memory handle", 50}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandlePropertiesKHR-handleType-00666", 57}, { {"handleType must not be one of the handle types defined as opaque", 64}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandlePropertiesKHR-handleType-parameter", 61}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryWin32HandlePropertiesKHR-pMemoryWin32HandleProperties-parameter", 79}, { {"pMemoryWin32HandleProperties must be a valid pointer to a VkMemoryWin32HandlePropertiesKHR structure", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandleFUCHSIA-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandleFUCHSIA-pGetZirconHandleInfo-parameter", 66}, { {"pGetZirconHandleInfo must be a valid pointer to a valid VkMemoryGetZirconHandleInfoFUCHSIA structure", 100}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandleFUCHSIA-pZirconHandle-parameter", 59}, { {"pZirconHandle must be a valid pointer to a zx_handle_t value", 60}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandlePropertiesFUCHSIA-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandlePropertiesFUCHSIA-handleType-04773", 62}, { {"handleType must be VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA", 72}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandlePropertiesFUCHSIA-handleType-parameter", 66}, { {"handleType must be a valid VkExternalMemoryHandleTypeFlagBits value", 67}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandlePropertiesFUCHSIA-pMemoryZirconHandleProperties-parameter", 85}, { {"pMemoryZirconHandleProperties must be a valid pointer to a VkMemoryZirconHandlePropertiesFUCHSIA structure", 106}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMemoryZirconHandlePropertiesFUCHSIA-zirconHandle-04774", 64}, { {"zirconHandle must reference a valid VMO", 39}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-buildType-parameter", 51}, { {"buildType must be a valid VkAccelerationStructureBuildTypeKHR value", 67}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-device-07440", 44}, { {"If device was created with multiple physical devices, the bufferDeviceAddressMultiDevice feature must be enabled", 112}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-micromap-11618", 46}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromap feature must be enabled", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-pBuildInfo-parameter", 52}, { {"pBuildInfo must be a valid pointer to a valid VkMicromapBuildInfoEXT structure", 78}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetMicromapBuildSizesEXT-pSizeInfo-parameter", 51}, { {"pSizeInfo must be a valid pointer to a VkMicromapBuildSizesInfoEXT structure", 76}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetNativeBufferPropertiesOHOS-buffer-parameter", 53}, { {"buffer must be a valid pointer to a valid OH_NativeBuffer value", 63}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetNativeBufferPropertiesOHOS-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetNativeBufferPropertiesOHOS-pProperties-parameter", 58}, { {"pProperties must be a valid pointer to a VkNativeBufferPropertiesOHOS structure", 79}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetPartitionedAccelerationStructuresBuildSizesNV-device-parameter", 72}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetPartitionedAccelerationStructuresBuildSizesNV-pInfo-parameter", 71}, { {"pInfo must be a valid pointer to a valid VkPartitionedAccelerationStructureInstancesInputNV structure", 101}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetPartitionedAccelerationStructuresBuildSizesNV-pSizeInfo-parameter", 75}, { {"pSizeInfo must be a valid pointer to a VkAccelerationStructureBuildSizesInfoKHR structure", 89}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetPartitionedAccelerationStructuresBuildSizesNV-partitionedAccelerationStructure-10534", 94}, { {"The VkPhysicalDevicePartitionedAccelerationStructureFeaturesNV::partitionedAccelerationStructure feature must be enabled", 120}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkGetPastPresentationTimingEXT-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingEXT-flags-12230", 47}, { {"If VK_PAST_PRESENTATION_TIMING_ALLOW_OUT_OF_ORDER_RESULTS_BIT_EXT is set in VkPastPresentationTimingInfoEXT::flags, the presentStageCount value of each element of VkPastPresentationTimingPropertiesEXT::pPresentationTimings must be at least the maximum number of present stages set in VkPresentTimingInfoEXT::presentStageQueries among all vkQueuePresentKHR calls, with a non-zero presentStageQueries, for which complete results have not been returned yet by a previous call", 472}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingEXT-flags-12231", 47}, { {"If VK_PAST_PRESENTATION_TIMING_ALLOW_OUT_OF_ORDER_RESULTS_BIT_EXT is not set in VkPastPresentationTimingInfoEXT::flags, the presentStageCount value of each element of VkPastPresentationTimingPropertiesEXT::pPresentationTimings must be at least the number of present stages set in VkPresentTimingInfoEXT::presentStageQueries for the earliest call to vkQueuePresentKHR, with a non-zero presentStageQueries, that corresponds to that element's index and for which complete results have not been returned yet by a previous call", 522}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingEXT-pPastPresentationTimingInfo-parameter", 73}, { {"pPastPresentationTimingInfo must be a valid pointer to a valid VkPastPresentationTimingInfoEXT structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingEXT-pPastPresentationTimingProperties-parameter", 79}, { {"pPastPresentationTimingProperties must be a valid pointer to a VkPastPresentationTimingPropertiesEXT structure", 110}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingGOOGLE-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingGOOGLE-pPresentationTimingCount-parameter", 73}, { {"pPresentationTimingCount must be a valid pointer to a uint32_t value", 68}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingGOOGLE-pPresentationTimings-parameter", 69}, { {"If the value referenced by pPresentationTimingCount is not 0, and pPresentationTimings is not NULL, pPresentationTimings must be a valid pointer to an array of pPresentationTimingCount VkPastPresentationTimingGOOGLE structures", 226}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingGOOGLE-swapchain-parameter", 58}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPastPresentationTimingGOOGLE-swapchain-parent", 55}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPerformanceParameterINTEL-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPerformanceParameterINTEL-pValue-parameter", 52}, { {"pValue must be a valid pointer to a VkPerformanceValueINTEL structure", 69}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPerformanceParameterINTEL-parameter-parameter", 55}, { {"parameter must be a valid VkPerformanceParameterTypeINTEL value", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCalibrateableTimeDomainsKHR-pTimeDomainCount-parameter", 78}, { {"pTimeDomainCount must be a valid pointer to a uint32_t value", 60}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceCalibrateableTimeDomainsKHR-pTimeDomains-parameter", 74}, { {"If the value referenced by pTimeDomainCount is not 0, and pTimeDomains is not NULL, pTimeDomains must be a valid pointer to an array of pTimeDomainCount VkTimeDomainKHR values", 175}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceCalibrateableTimeDomainsKHR-physicalDevice-parameter", 76}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV-pProperties-parameter", 93}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkCooperativeMatrixFlexibleDimensionsPropertiesNV structures", 207}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV-pPropertyCount-parameter", 96}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV-physicalDevice-parameter", 96}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesKHR-pProperties-parameter", 76}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkCooperativeMatrixPropertiesKHR structures", 190}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesKHR-pPropertyCount-parameter", 79}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesKHR-physicalDevice-parameter", 79}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesNV-pProperties-parameter", 75}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkCooperativeMatrixPropertiesNV structures", 189}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesNV-pPropertyCount-parameter", 78}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeMatrixPropertiesNV-physicalDevice-parameter", 78}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeVectorPropertiesNV-pProperties-parameter", 75}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkCooperativeVectorPropertiesNV structures", 189}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeVectorPropertiesNV-pPropertyCount-parameter", 78}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceCooperativeVectorPropertiesNV-physicalDevice-parameter", 78}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceDescriptorSizeEXT-descriptorType-parameter", 66}, { {"descriptorType must be a valid VkDescriptorType value", 53}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkGetPhysicalDeviceDescriptorSizeEXT-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkGetPhysicalDeviceDescriptorSizeEXT-type-11362", 52}, { {"type must be one of VK_DESCRIPTOR_TYPE_SAMPLER, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_NV, VK_DESCRIPTOR_TYPE_TENSOR_ARM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT", 576}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkGetPhysicalDeviceDirectFBPresentationSupportEXT-dfb-parameter", 68}, { {"dfb must be a valid pointer to an IDirectFB value", 49}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDirectFBPresentationSupportEXT-physicalDevice-parameter", 79}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDirectFBPresentationSupportEXT-queueFamilyIndex-04119", 77}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlaneProperties2KHR-pProperties-parameter", 72}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayPlaneProperties2KHR structures", 186}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlaneProperties2KHR-pPropertyCount-parameter", 75}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlaneProperties2KHR-physicalDevice-parameter", 75}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlanePropertiesKHR-pProperties-parameter", 71}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayPlanePropertiesKHR structures", 185}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlanePropertiesKHR-pPropertyCount-parameter", 74}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPlanePropertiesKHR-physicalDevice-parameter", 74}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayProperties2KHR-pProperties-parameter", 67}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayProperties2KHR structures", 181}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayProperties2KHR-pPropertyCount-parameter", 70}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayProperties2KHR-physicalDevice-parameter", 70}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPropertiesKHR-pProperties-parameter", 66}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkDisplayPropertiesKHR structures", 180}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPropertiesKHR-pPropertyCount-parameter", 69}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceDisplayPropertiesKHR-physicalDevice-parameter", 69}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceExternalBufferProperties-pExternalBufferInfo-parameter", 78}, { {"pExternalBufferInfo must be a valid pointer to a valid VkPhysicalDeviceExternalBufferInfo structure", 99}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalBufferProperties-pExternalBufferProperties-parameter", 84}, { {"pExternalBufferProperties must be a valid pointer to a VkExternalBufferProperties structure", 91}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalBufferProperties-physicalDevice-parameter", 73}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalFenceProperties-pExternalFenceInfo-parameter", 76}, { {"pExternalFenceInfo must be a valid pointer to a valid VkPhysicalDeviceExternalFenceInfo structure", 97}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalFenceProperties-pExternalFenceProperties-parameter", 82}, { {"pExternalFenceProperties must be a valid pointer to a VkExternalFenceProperties structure", 89}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalFenceProperties-physicalDevice-parameter", 72}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-externalHandleType-07721", 80}, { {"externalHandleType must not have more than one bit set", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-externalHandleType-parameter", 84}, { {"externalHandleType must be a valid combination of VkExternalMemoryHandleTypeFlagBitsNV values", 93}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-flags-parameter", 71}, { {"flags must be a valid combination of VkImageCreateFlagBits values", 65}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-format-parameter", 72}, { {"format must be a valid VkFormat value", 37}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-pExternalImageFormatProperties-parameter", 96}, { {"pExternalImageFormatProperties must be a valid pointer to a VkExternalImageFormatPropertiesNV structure", 103}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-physicalDevice-parameter", 80}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-tiling-parameter", 72}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-type-parameter", 70}, { {"type must be a valid VkImageType value", 38}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-usage-parameter", 71}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalImageFormatPropertiesNV-usage-requiredbitmask", 77}, { {"usage must not be 0", 19}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalSemaphoreProperties-pExternalSemaphoreInfo-parameter", 84}, { {"pExternalSemaphoreInfo must be a valid pointer to a valid VkPhysicalDeviceExternalSemaphoreInfo structure", 105}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalSemaphoreProperties-pExternalSemaphoreProperties-parameter", 90}, { {"pExternalSemaphoreProperties must be a valid pointer to a VkExternalSemaphoreProperties structure", 97}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalSemaphoreProperties-physicalDevice-parameter", 76}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalTensorPropertiesARM-pExternalTensorInfo-parameter", 81}, { {"pExternalTensorInfo must be a valid pointer to a valid VkPhysicalDeviceExternalTensorInfoARM structure", 102}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalTensorPropertiesARM-pExternalTensorProperties-parameter", 87}, { {"pExternalTensorProperties must be a valid pointer to a VkExternalTensorPropertiesARM structure", 94}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceExternalTensorPropertiesARM-physicalDevice-parameter", 76}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceFeatures-pFeatures-parameter", 52}, { {"pFeatures must be a valid pointer to a VkPhysicalDeviceFeatures structure", 73}, { "chapters/features.html", 22} } },
        { {"VUID-vkGetPhysicalDeviceFeatures-physicalDevice-parameter", 57}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/features.html", 22} } },
        { {"VUID-vkGetPhysicalDeviceFeatures2-pFeatures-parameter", 53}, { {"pFeatures must be a valid pointer to a VkPhysicalDeviceFeatures2 structure", 74}, { "chapters/features.html", 22} } },
        { {"VUID-vkGetPhysicalDeviceFeatures2-physicalDevice-parameter", 58}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/features.html", 22} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties-None-12272", 51}, { {"If Vulkan 1.3 is not supported, the maintenance5 feature is not supported, and the ycbcr2plane444Formats feature is not supported, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 332}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties-format-parameter", 57}, { {"format must be a valid VkFormat value", 37}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties-pFormatProperties-parameter", 68}, { {"pFormatProperties must be a valid pointer to a VkFormatProperties structure", 75}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties-physicalDevice-parameter", 65}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties2-None-12273", 52}, { {"If Vulkan 1.3 is not supported, the maintenance5 feature is not supported, and the ycbcr2plane444Formats feature is not supported, format must not be VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, or VK_FORMAT_G16_B16R16_2PLANE_444_UNORM", 332}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties2-format-parameter", 58}, { {"format must be a valid VkFormat value", 37}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties2-pFormatProperties-parameter", 69}, { {"pFormatProperties must be a valid pointer to a VkFormatProperties2 structure", 76}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFormatProperties2-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/formats.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceFragmentShadingRatesKHR-pFragmentShadingRateCount-parameter", 83}, { {"pFragmentShadingRateCount must be a valid pointer to a uint32_t value", 69}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceFragmentShadingRatesKHR-pFragmentShadingRates-parameter", 79}, { {"If the value referenced by pFragmentShadingRateCount is not 0, and pFragmentShadingRates is not NULL, pFragmentShadingRates must be a valid pointer to an array of pFragmentShadingRateCount VkPhysicalDeviceFragmentShadingRateKHR structures", 238}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceFragmentShadingRatesKHR-physicalDevice-parameter", 72}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/primsrast.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-flags-parameter", 61}, { {"flags must be a valid combination of VkImageCreateFlagBits values", 65}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-format-parameter", 62}, { {"format must be a valid VkFormat value", 37}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-pImageFormatProperties-parameter", 78}, { {"pImageFormatProperties must be a valid pointer to a VkImageFormatProperties structure", 85}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-physicalDevice-parameter", 70}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-tiling-02248", 58}, { {"tiling must not be VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT. (Use vkGetPhysicalDeviceImageFormatProperties2 instead)", 115}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-tiling-parameter", 62}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-type-parameter", 60}, { {"type must be a valid VkImageType value", 38}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-usage-parameter", 61}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties-usage-requiredbitmask", 67}, { {"usage must not be 0", 19}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties2-pImageFormatInfo-parameter", 73}, { {"pImageFormatInfo must be a valid pointer to a valid VkPhysicalDeviceImageFormatInfo2 structure", 94}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties2-pImageFormatProperties-parameter", 79}, { {"pImageFormatProperties must be a valid pointer to a VkImageFormatProperties2 structure", 86}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties2-pNext-01868", 58}, { {"If the pNext chain of pImageFormatProperties includes a VkAndroidHardwareBufferUsageANDROID structure, the pNext chain of pImageFormatInfo must include a VkPhysicalDeviceExternalImageFormatInfo structure with handleType set to VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID", 293}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties2-pNext-09004", 58}, { {"If the pNext chain of pImageFormatProperties includes a VkHostImageCopyDevicePerformanceQuery structure, pImageFormatInfo->usage must contain VK_IMAGE_USAGE_HOST_TRANSFER_BIT", 174}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceImageFormatProperties2-physicalDevice-parameter", 71}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/capabilities.html", 26} } },
        { {"VUID-vkGetPhysicalDeviceMemoryProperties-pMemoryProperties-parameter", 68}, { {"pMemoryProperties must be a valid pointer to a VkPhysicalDeviceMemoryProperties structure", 89}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMemoryProperties-physicalDevice-parameter", 65}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMemoryProperties2-pMemoryProperties-parameter", 69}, { {"pMemoryProperties must be a valid pointer to a VkPhysicalDeviceMemoryProperties2 structure", 90}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMemoryProperties2-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMultisamplePropertiesEXT-pMultisampleProperties-parameter", 81}, { {"pMultisampleProperties must be a valid pointer to a VkMultisamplePropertiesEXT structure", 88}, { "chapters/limits.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMultisamplePropertiesEXT-physicalDevice-parameter", 73}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/limits.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceMultisamplePropertiesEXT-samples-parameter", 66}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/limits.html", 20} } },
        { {"VUID-vkGetPhysicalDeviceOpticalFlowImageFormatsNV-pFormatCount-parameter", 72}, { {"pFormatCount must be a valid pointer to a uint32_t value", 56}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkGetPhysicalDeviceOpticalFlowImageFormatsNV-pImageFormatProperties-parameter", 82}, { {"If the value referenced by pFormatCount is not 0, and pImageFormatProperties is not NULL, pImageFormatProperties must be a valid pointer to an array of pFormatCount VkOpticalFlowImageFormatPropertiesNV structures", 212}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkGetPhysicalDeviceOpticalFlowImageFormatsNV-pOpticalFlowImageFormatInfo-parameter", 87}, { {"pOpticalFlowImageFormatInfo must be a valid pointer to a valid VkOpticalFlowImageFormatInfoNV structure", 103}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkGetPhysicalDeviceOpticalFlowImageFormatsNV-physicalDevice-parameter", 74}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_NV_optical_flow/optical_flow.html", 45} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-commonparent", 57}, { {"Both of physicalDevice, and surface must have been created, allocated, or retrieved from the same VkInstance", 108}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-pRectCount-parameter", 65}, { {"pRectCount must be a valid pointer to a uint32_t value", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-pRects-parameter", 61}, { {"If the value referenced by pRectCount is not 0, and pRects is not NULL, pRects must be a valid pointer to an array of pRectCount VkRect2D structures", 148}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-physicalDevice-parameter", 69}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-surface-06211", 58}, { {"surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 141}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDevicePresentRectanglesKHR-surface-parameter", 62}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceProperties-pProperties-parameter", 56}, { {"pProperties must be a valid pointer to a VkPhysicalDeviceProperties structure", 77}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceProperties-physicalDevice-parameter", 59}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceProperties2-pProperties-parameter", 57}, { {"pProperties must be a valid pointer to a VkPhysicalDeviceProperties2 structure", 78}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceProperties2-physicalDevice-parameter", 60}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM-pProperties-09958", 90}, { {"pProperties must be a valid structure of the appropriate type as defined in Valid property queries for specific engines and operations", 134}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM-pQueueFamilyDataGraphProperties-09957", 110}, { {"The VkQueueFamilyDataGraphPropertiesARM structure pointed to by pQueueFamilyDataGraphProperties must have been obtained by a prior call to vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM with the same physicalDevice and queueFamilyIndex", 241}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM-pQueueFamilyDataGraphProperties-parameter", 114}, { {"pQueueFamilyDataGraphProperties must be a valid pointer to a valid VkQueueFamilyDataGraphPropertiesARM structure", 112}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM-physicalDevice-parameter", 97}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-pFormatCount-parameter", 93}, { {"pFormatCount must be a valid pointer to a uint32_t value", 56}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-pImageFormatProperties-parameter", 103}, { {"If the value referenced by pFormatCount is not 0, and pImageFormatProperties is not NULL, pImageFormatProperties must be a valid pointer to an array of pFormatCount VkDataGraphOpticalFlowImageFormatPropertiesARM structures", 222}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-pOpticalFlowImageFormatInfo-parameter", 108}, { {"pOpticalFlowImageFormatInfo must be a valid pointer to a valid VkDataGraphOpticalFlowImageFormatInfoARM structure", 113}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-pQueueFamilyDataGraphProperties-09965", 108}, { {"pQueueFamilyDataGraphProperties must point to a structure whose operation member has its name member equal to OpticalFlow", 121}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-pQueueFamilyDataGraphProperties-parameter", 112}, { {"pQueueFamilyDataGraphProperties must be a valid pointer to a valid VkQueueFamilyDataGraphPropertiesARM structure", 112}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM-physicalDevice-parameter", 95}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM-pQueueFamilyDataGraphProcessingEngineInfo-parameter", 125}, { {"pQueueFamilyDataGraphProcessingEngineInfo must be a valid pointer to a valid VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM structure", 146}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM-pQueueFamilyDataGraphProcessingEngineProperties-parameter", 131}, { {"pQueueFamilyDataGraphProcessingEngineProperties must be a valid pointer to a VkQueueFamilyDataGraphProcessingEnginePropertiesARM structure", 138}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM-physicalDevice-parameter", 98}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM-pQueueFamilyDataGraphProperties-parameter", 99}, { {"If the value referenced by pQueueFamilyDataGraphPropertyCount is not 0, and pQueueFamilyDataGraphProperties is not NULL, pQueueFamilyDataGraphProperties must be a valid pointer to an array of pQueueFamilyDataGraphPropertyCount VkQueueFamilyDataGraphPropertiesARM structures", 273}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM-pQueueFamilyDataGraphPropertyCount-parameter", 102}, { {"pQueueFamilyDataGraphPropertyCount must be a valid pointer to a uint32_t value", 78}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM-physicalDevice-parameter", 82}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_ARM_data_graph/graphs.html", 38} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR-pNumPasses-parameter", 81}, { {"pNumPasses must be a valid pointer to a uint32_t value", 54}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR-pPerformanceQueryCreateInfo-parameter", 98}, { {"pPerformanceQueryCreateInfo must be a valid pointer to a valid VkQueryPoolPerformanceCreateInfoKHR structure", 108}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR-physicalDevice-parameter", 85}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties-pQueueFamilyProperties-parameter", 78}, { {"If the value referenced by pQueueFamilyPropertyCount is not 0, and pQueueFamilyProperties is not NULL, pQueueFamilyProperties must be a valid pointer to an array of pQueueFamilyPropertyCount VkQueueFamilyProperties structures", 225}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties-pQueueFamilyPropertyCount-parameter", 81}, { {"pQueueFamilyPropertyCount must be a valid pointer to a uint32_t value", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties-physicalDevice-parameter", 70}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties2-pQueueFamilyProperties-parameter", 79}, { {"If the value referenced by pQueueFamilyPropertyCount is not 0, and pQueueFamilyProperties is not NULL, pQueueFamilyProperties must be a valid pointer to an array of pQueueFamilyPropertyCount VkQueueFamilyProperties2 structures", 226}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties2-pQueueFamilyPropertyCount-parameter", 82}, { {"pQueueFamilyPropertyCount must be a valid pointer to a uint32_t value", 69}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceQueueFamilyProperties2-physicalDevice-parameter", 71}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkGetPhysicalDeviceScreenPresentationSupportQNX-physicalDevice-parameter", 77}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceScreenPresentationSupportQNX-queueFamilyIndex-04743", 75}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceScreenPresentationSupportQNX-window-parameter", 69}, { {"window must be a valid pointer to a _screen_window value", 56}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-format-parameter", 68}, { {"format must be a valid VkFormat value", 37}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-pProperties-parameter", 73}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkSparseImageFormatProperties structures", 187}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-pPropertyCount-parameter", 76}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-physicalDevice-parameter", 76}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-samples-01094", 65}, { {"samples must be a valid VkSampleCountFlagBits value that is set in VkImageFormatProperties::sampleCounts returned by vkGetPhysicalDeviceImageFormatProperties with format, type, tiling, and usage equal to those in this command", 225}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-samples-parameter", 69}, { {"samples must be a valid VkSampleCountFlagBits value", 51}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-tiling-parameter", 68}, { {"tiling must be a valid VkImageTiling value", 42}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-type-parameter", 66}, { {"type must be a valid VkImageType value", 38}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-usage-parameter", 67}, { {"usage must be a valid combination of VkImageUsageFlagBits values", 64}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties-usage-requiredbitmask", 73}, { {"usage must not be 0", 19}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties2-pFormatInfo-parameter", 74}, { {"pFormatInfo must be a valid pointer to a valid VkPhysicalDeviceSparseImageFormatInfo2 structure", 95}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties2-pProperties-parameter", 74}, { {"If the value referenced by pPropertyCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pPropertyCount VkSparseImageFormatProperties2 structures", 188}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties2-pPropertyCount-parameter", 77}, { {"pPropertyCount must be a valid pointer to a uint32_t value", 58}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSparseImageFormatProperties2-physicalDevice-parameter", 77}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV-pCombinationCount-parameter", 98}, { {"pCombinationCount must be a valid pointer to a uint32_t value", 61}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV-pCombinations-parameter", 94}, { {"If the value referenced by pCombinationCount is not 0, and pCombinations is not NULL, pCombinations must be a valid pointer to an array of pCombinationCount VkFramebufferMixedSamplesCombinationNV structures", 206}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV-physicalDevice-parameter", 95}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/fragops.html", 21} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2EXT-commonparent", 60}, { {"Both of physicalDevice, and surface must have been created, allocated, or retrieved from the same VkInstance", 108}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2EXT-pSurfaceCapabilities-parameter", 78}, { {"pSurfaceCapabilities must be a valid pointer to a VkSurfaceCapabilities2EXT structure", 85}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2EXT-physicalDevice-parameter", 72}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2EXT-surface-06211", 61}, { {"surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 141}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2EXT-surface-parameter", 65}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-02671", 59}, { {"If a VkSurfaceCapabilitiesFullScreenExclusiveEXT structure is included in the pNext chain of pSurfaceCapabilities, a VkSurfaceFullScreenExclusiveWin32InfoEXT structure must be included in the pNext chain of pSurfaceInfo", 219}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-07776", 59}, { {"If a VkSurfacePresentModeCompatibilityKHR structure is included in the pNext chain of pSurfaceCapabilities, a VkSurfacePresentModeKHR structure must be included in the pNext chain of pSurfaceInfo", 195}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-07777", 59}, { {"If a VkSurfacePresentScalingCapabilitiesKHR structure is included in the pNext chain of pSurfaceCapabilities, a VkSurfacePresentModeKHR structure must be included in the pNext chain of pSurfaceInfo", 197}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-07778", 59}, { {"If a VkSurfacePresentModeCompatibilityKHR structure is included in the pNext chain of pSurfaceCapabilities, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 165}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-07779", 59}, { {"If a VkSurfacePresentScalingCapabilitiesKHR structure is included in the pNext chain of pSurfaceCapabilities, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 167}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pNext-12446", 59}, { {"If a VkSwapchainFlagsSurfaceCapabilitiesEXT structure is included in the pNext chain of pSurfaceCapabilities, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 167}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pSurfaceCapabilities-parameter", 78}, { {"pSurfaceCapabilities must be a valid pointer to a VkSurfaceCapabilities2KHR structure", 85}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pSurfaceInfo-06521", 66}, { {"If the VK_GOOGLE_surfaceless_query extension is not enabled, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 118}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pSurfaceInfo-06522", 66}, { {"If pSurfaceInfo->surface is not VK_NULL_HANDLE, pSurfaceInfo->surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 203}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-pSurfaceInfo-parameter", 70}, { {"pSurfaceInfo must be a valid pointer to a valid VkPhysicalDeviceSurfaceInfo2KHR structure", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilities2KHR-physicalDevice-parameter", 72}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilitiesKHR-commonparent", 59}, { {"Both of physicalDevice, and surface must have been created, allocated, or retrieved from the same VkInstance", 108}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilitiesKHR-pSurfaceCapabilities-parameter", 77}, { {"pSurfaceCapabilities must be a valid pointer to a VkSurfaceCapabilitiesKHR structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilitiesKHR-physicalDevice-parameter", 71}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilitiesKHR-surface-06211", 60}, { {"surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 141}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceCapabilitiesKHR-surface-parameter", 64}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-pSurfaceFormatCount-parameter", 72}, { {"pSurfaceFormatCount must be a valid pointer to a uint32_t value", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-pSurfaceFormats-parameter", 68}, { {"If the value referenced by pSurfaceFormatCount is not 0, and pSurfaceFormats is not NULL, pSurfaceFormats must be a valid pointer to an array of pSurfaceFormatCount VkSurfaceFormat2KHR structures", 195}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-pSurfaceInfo-06521", 61}, { {"If the VK_GOOGLE_surfaceless_query extension is not enabled, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 118}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-pSurfaceInfo-06522", 61}, { {"If pSurfaceInfo->surface is not VK_NULL_HANDLE, pSurfaceInfo->surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 203}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-pSurfaceInfo-parameter", 65}, { {"pSurfaceInfo must be a valid pointer to a valid VkPhysicalDeviceSurfaceInfo2KHR structure", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormats2KHR-physicalDevice-parameter", 67}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-commonparent", 54}, { {"Both of physicalDevice, and surface that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkInstance", 157}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-pSurfaceFormatCount-parameter", 71}, { {"pSurfaceFormatCount must be a valid pointer to a uint32_t value", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-pSurfaceFormats-parameter", 67}, { {"If the value referenced by pSurfaceFormatCount is not 0, and pSurfaceFormats is not NULL, pSurfaceFormats must be a valid pointer to an array of pSurfaceFormatCount VkSurfaceFormatKHR structures", 194}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-surface-06524", 55}, { {"If the VK_GOOGLE_surfaceless_query extension is not enabled, surface must be a valid VkSurfaceKHR handle", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-surface-06525", 55}, { {"If surface is not VK_NULL_HANDLE, surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 175}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceFormatsKHR-surface-parameter", 59}, { {"If surface is not VK_NULL_HANDLE, surface must be a valid VkSurfaceKHR handle", 77}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-pPresentModeCount-parameter", 75}, { {"pPresentModeCount must be a valid pointer to a uint32_t value", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-pPresentModes-parameter", 71}, { {"If the value referenced by pPresentModeCount is not 0, and pPresentModes is not NULL, pPresentModes must be a valid pointer to an array of pPresentModeCount VkPresentModeKHR values", 180}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-pSurfaceInfo-06521", 66}, { {"If the VK_GOOGLE_surfaceless_query extension is not enabled, pSurfaceInfo->surface must be a valid VkSurfaceKHR handle", 118}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-pSurfaceInfo-06522", 66}, { {"If pSurfaceInfo->surface is not VK_NULL_HANDLE, pSurfaceInfo->surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 203}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-pSurfaceInfo-parameter", 70}, { {"pSurfaceInfo must be a valid pointer to a valid VkPhysicalDeviceSurfaceInfo2KHR structure", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModes2EXT-physicalDevice-parameter", 72}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-commonparent", 59}, { {"Both of physicalDevice, and surface that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkInstance", 157}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-pPresentModeCount-parameter", 74}, { {"pPresentModeCount must be a valid pointer to a uint32_t value", 61}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-pPresentModes-parameter", 70}, { {"If the value referenced by pPresentModeCount is not 0, and pPresentModes is not NULL, pPresentModes must be a valid pointer to an array of pPresentModeCount VkPresentModeKHR values", 180}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-physicalDevice-parameter", 71}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-surface-06524", 60}, { {"If the VK_GOOGLE_surfaceless_query extension is not enabled, surface must be a valid VkSurfaceKHR handle", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-surface-06525", 60}, { {"If surface is not VK_NULL_HANDLE, surface must be supported by physicalDevice, as reported by vkGetPhysicalDeviceSurfaceSupportKHR or an equivalent platform-specific mechanism", 175}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfacePresentModesKHR-surface-parameter", 64}, { {"If surface is not VK_NULL_HANDLE, surface must be a valid VkSurfaceKHR handle", 77}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceSupportKHR-commonparent", 54}, { {"Both of physicalDevice, and surface must have been created, allocated, or retrieved from the same VkInstance", 108}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceSupportKHR-pSupported-parameter", 62}, { {"pSupported must be a valid pointer to a VkBool32 value", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceSupportKHR-physicalDevice-parameter", 66}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceSupportKHR-queueFamilyIndex-01269", 64}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceSurfaceSupportKHR-surface-parameter", 59}, { {"surface must be a valid VkSurfaceKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceToolProperties-pToolCount-parameter", 59}, { {"pToolCount must be a valid pointer to a uint32_t value", 54}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceToolProperties-pToolProperties-parameter", 64}, { {"If the value referenced by pToolCount is not 0, and pToolProperties is not NULL, pToolProperties must be a valid pointer to an array of pToolCount VkPhysicalDeviceToolProperties structures", 188}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceToolProperties-physicalDevice-parameter", 63}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetPhysicalDeviceUbmPresentationSupportSEC-device-12368", 62}, { {"device must point to a valid UBM ubm_device", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceUbmPresentationSupportSEC-device-parameter", 66}, { {"device must be a valid pointer to a ubm_device value", 52}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceUbmPresentationSupportSEC-physicalDevice-parameter", 74}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceUbmPresentationSupportSEC-queueFamilyIndex-12369", 72}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pCapabilities-parameter", 68}, { {"pCapabilities must be a valid pointer to a VkVideoCapabilitiesKHR structure", 75}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pNext-10921", 56}, { {"If the pNext chain of pVideoProfile includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be supported", 160}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07183", 64}, { {"If pVideoProfile->videoCodecOperation specifies a decode operation, then the pNext chain of pCapabilities must include a VkVideoDecodeCapabilitiesKHR structure", 159}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07184", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoDecodeH264CapabilitiesKHR structure", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07185", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoDecodeH265CapabilitiesKHR structure", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07186", 64}, { {"If pVideoProfile->videoCodecOperation specifies an encode operation, then the pNext chain of pCapabilities must include a VkVideoEncodeCapabilitiesKHR structure", 160}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07187", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoEncodeH264CapabilitiesKHR structure", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-07188", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoEncodeH265CapabilitiesKHR structure", 182}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-09257", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoDecodeAV1CapabilitiesKHR structure", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-10263", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoEncodeAV1CapabilitiesKHR structure", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-10792", 64}, { {"If pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_DECODE_VP9_BIT_KHR, then the pNext chain of pCapabilities must include a VkVideoDecodeVP9CapabilitiesKHR structure", 180}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-pVideoProfile-parameter", 68}, { {"pVideoProfile must be a valid pointer to a valid VkVideoProfileInfoKHR structure", 80}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoCapabilitiesKHR-physicalDevice-parameter", 69}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-pQualityLevelInfo-08257", 84}, { {"If pQualityLevelInfo->pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the pNext chain of pQualityLevelProperties must include a VkVideoEncodeH264QualityLevelPropertiesKHR structure", 221}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-pQualityLevelInfo-08258", 84}, { {"If pQualityLevelInfo->pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the pNext chain of pQualityLevelProperties must include a VkVideoEncodeH265QualityLevelPropertiesKHR structure", 221}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-pQualityLevelInfo-10305", 84}, { {"If pQualityLevelInfo->pVideoProfile->videoCodecOperation is VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR, then the pNext chain of pQualityLevelProperties must include a VkVideoEncodeAV1QualityLevelPropertiesKHR structure", 219}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-pQualityLevelInfo-parameter", 88}, { {"pQualityLevelInfo must be a valid pointer to a valid VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR structure", 109}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-pQualityLevelProperties-parameter", 94}, { {"pQualityLevelProperties must be a valid pointer to a VkVideoEncodeQualityLevelPropertiesKHR structure", 101}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR-physicalDevice-parameter", 85}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-pNext-06812", 60}, { {"The pNext chain of pVideoFormatInfo must include a VkVideoProfileListInfoKHR structure with profileCount greater than 0", 119}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-pNext-10922", 60}, { {"If the pNext chain of pVideoFormatInfo includes a VkVideoEncodeProfileRgbConversionInfoVALVE structure, then the videoEncodeRgbConversion feature must be supported", 163}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-pVideoFormatInfo-parameter", 75}, { {"pVideoFormatInfo must be a valid pointer to a valid VkPhysicalDeviceVideoFormatInfoKHR structure", 96}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-pVideoFormatProperties-parameter", 81}, { {"If the value referenced by pVideoFormatPropertyCount is not 0, and pVideoFormatProperties is not NULL, pVideoFormatProperties must be a valid pointer to an array of pVideoFormatPropertyCount VkVideoFormatPropertiesKHR structures", 228}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-pVideoFormatPropertyCount-parameter", 84}, { {"pVideoFormatPropertyCount must be a valid pointer to a uint32_t value", 69}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceVideoFormatPropertiesKHR-physicalDevice-parameter", 73}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetPhysicalDeviceWaylandPresentationSupportKHR-display-parameter", 71}, { {"display must be a valid pointer to a wl_display value", 53}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceWaylandPresentationSupportKHR-physicalDevice-parameter", 78}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceWaylandPresentationSupportKHR-queueFamilyIndex-01306", 76}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceWin32PresentationSupportKHR-physicalDevice-parameter", 76}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceWin32PresentationSupportKHR-queueFamilyIndex-01309", 74}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXcbPresentationSupportKHR-connection-parameter", 70}, { {"connection must be a valid pointer to an xcb_connection_t value", 63}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXcbPresentationSupportKHR-physicalDevice-parameter", 74}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXcbPresentationSupportKHR-queueFamilyIndex-01312", 72}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXlibPresentationSupportKHR-dpy-parameter", 64}, { {"dpy must be a valid pointer to a Display value", 46}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXlibPresentationSupportKHR-physicalDevice-parameter", 75}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPhysicalDeviceXlibPresentationSupportKHR-queueFamilyIndex-01315", 73}, { {"queueFamilyIndex must be less than pQueueFamilyPropertyCount returned by vkGetPhysicalDeviceQueueFamilyProperties for the given physicalDevice", 142}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetPipelineBinaryDataKHR-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineBinaryDataKHR-pInfo-parameter", 47}, { {"pInfo must be a valid pointer to a valid VkPipelineBinaryDataInfoKHR structure", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineBinaryDataKHR-pPipelineBinaryData-parameter", 61}, { {"If the value referenced by pPipelineBinaryDataSize is not 0, and pPipelineBinaryData is not NULL, pPipelineBinaryData must be a valid pointer to an array of pPipelineBinaryDataSize bytes", 186}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineBinaryDataKHR-pPipelineBinaryDataSize-parameter", 65}, { {"pPipelineBinaryDataSize must be a valid pointer to a size_t value", 65}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineBinaryDataKHR-pPipelineBinaryKey-parameter", 60}, { {"pPipelineBinaryKey must be a valid pointer to a VkPipelineBinaryKeyKHR structure", 80}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-pData-parameter", 43}, { {"If the value referenced by pDataSize is not 0, and pData is not NULL, pData must be a valid pointer to an array of pDataSize bytes", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-pDataSize-parameter", 47}, { {"pDataSize must be a valid pointer to a size_t value", 51}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-pipelineCache-11834", 47}, { {"pipelineCache must not have been created with the headerVersion member of VkPipelineCacheCreateInfo::pInitialData equal to VK_PIPELINE_CACHE_HEADER_VERSION_DATA_GRAPH_QCOM", 171}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-pipelineCache-parameter", 51}, { {"pipelineCache must be a valid VkPipelineCache handle", 52}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineCacheData-pipelineCache-parent", 48}, { {"pipelineCache must have been created, allocated, or retrieved from device", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-device-parameter", 71}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pExecutableInfo-parameter", 80}, { {"pExecutableInfo must be a valid pointer to a valid VkPipelineExecutableInfoKHR structure", 88}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pInternalRepresentationCount-parameter", 93}, { {"pInternalRepresentationCount must be a valid pointer to a uint32_t value", 72}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pInternalRepresentations-parameter", 89}, { {"If the value referenced by pInternalRepresentationCount is not 0, and pInternalRepresentations is not NULL, pInternalRepresentations must be a valid pointer to an array of pInternalRepresentationCount VkPipelineExecutableInternalRepresentationKHR structures", 257}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pipeline-03277", 69}, { {"The pipeline member of pExecutableInfo must have been created with device", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pipeline-03278", 69}, { {"The pipeline member of pExecutableInfo must have been created with VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR", 126}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableInternalRepresentationsKHR-pipelineExecutableInfo-03276", 83}, { {"The pipelineExecutableInfo feature must be enabled", 50}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-pExecutableCount-parameter", 68}, { {"pExecutableCount must be a valid pointer to a uint32_t value", 60}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-pPipelineInfo-parameter", 65}, { {"pPipelineInfo must be a valid pointer to a valid VkPipelineInfoKHR structure", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-pProperties-parameter", 63}, { {"If the value referenced by pExecutableCount is not 0, and pProperties is not NULL, pProperties must be a valid pointer to an array of pExecutableCount VkPipelineExecutablePropertiesKHR structures", 195}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-pipeline-03271", 56}, { {"The pipeline member of pPipelineInfo must have been created with device", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutablePropertiesKHR-pipelineExecutableInfo-03270", 70}, { {"The pipelineExecutableInfo feature must be enabled", 50}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pExecutableInfo-parameter", 67}, { {"pExecutableInfo must be a valid pointer to a valid VkPipelineExecutableInfoKHR structure", 88}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pStatisticCount-parameter", 67}, { {"pStatisticCount must be a valid pointer to a uint32_t value", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pStatistics-parameter", 63}, { {"If the value referenced by pStatisticCount is not 0, and pStatistics is not NULL, pStatistics must be a valid pointer to an array of pStatisticCount VkPipelineExecutableStatisticKHR structures", 192}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pipeline-03273", 56}, { {"The pipeline member of pExecutableInfo must have been created with device", 73}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pipeline-03274", 56}, { {"The pipeline member of pExecutableInfo must have been created with VK_PIPELINE_CREATE_CAPTURE_STATISTICS_BIT_KHR", 112}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineExecutableStatisticsKHR-pipelineExecutableInfo-03272", 70}, { {"The pipelineExecutableInfo feature must be enabled", 50}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineIndirectDeviceAddressNV-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectDeviceAddressNV-deviceGeneratedComputePipelines-09078", 79}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectDeviceAddressNV-pInfo-parameter", 57}, { {"pInfo must be a valid pointer to a valid VkPipelineIndirectDeviceAddressInfoNV structure", 88}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectMemoryRequirementsNV-device-parameter", 63}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectMemoryRequirementsNV-deviceGeneratedComputePipelines-09082", 84}, { {"The VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV::deviceGeneratedComputePipelines feature must be enabled", 117}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectMemoryRequirementsNV-pCreateInfo-09083", 64}, { {"pCreateInfo->flags must include VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV", 75}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectMemoryRequirementsNV-pCreateInfo-parameter", 68}, { {"pCreateInfo must be a valid pointer to a valid VkComputePipelineCreateInfo structure", 84}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineIndirectMemoryRequirementsNV-pMemoryRequirements-parameter", 76}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkGetPipelineKeyKHR-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineKeyKHR-pNext-09605", 36}, { {"The pNext chain of pPipelineCreateInfo must not set VkPipelineBinaryInfoKHR::binaryCount to a value greater than 0", 114}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineKeyKHR-pPipelineCreateInfo-parameter", 54}, { {"If pPipelineCreateInfo is not NULL, pPipelineCreateInfo must be a valid pointer to a valid VkPipelineCreateInfoKHR structure", 124}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelineKeyKHR-pPipelineKey-parameter", 47}, { {"pPipelineKey must be a valid pointer to a VkPipelineBinaryKeyKHR structure", 74}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelinePropertiesEXT-None-06766", 42}, { {"The pipelinePropertiesIdentifier feature must be enabled", 56}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelinePropertiesEXT-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelinePropertiesEXT-pPipelineInfo-parameter", 55}, { {"pPipelineInfo must be a valid pointer to a valid VkPipelineInfoKHR structure", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelinePropertiesEXT-pPipelineProperties-06739", 57}, { {"pPipelineProperties must be a valid pointer to a VkPipelinePropertiesIdentifierEXT structure", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPipelinePropertiesEXT-pipeline-06738", 46}, { {"The pipeline member of pPipelineInfo must have been created with device", 71}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetPrivateData-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-objectHandle-09498", 40}, { {"objectHandle must be a valid handle to an object of type objectType", 67}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-objectType-04018", 38}, { {"objectHandle must be device or a child of device", 48}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-objectType-parameter", 42}, { {"objectType must be a valid VkObjectType value", 45}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-pData-parameter", 37}, { {"pData must be a valid pointer to a uint64_t value", 49}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-privateDataSlot-parameter", 47}, { {"privateDataSlot must be a valid VkPrivateDataSlot handle", 56}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetPrivateData-privateDataSlot-parent", 44}, { {"privateDataSlot must have been created, allocated, or retrieved from device", 75}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkGetQueryPoolResults-None-09401", 37}, { {"All queries used by the command must not be uninitialized", 57}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-dataSize-00817", 41}, { {"dataSize must be large enough to contain the result of each query, as described here", 84}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-dataSize-arraylength", 47}, { {"dataSize must be greater than 0", 31}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-firstQuery-09436", 43}, { {"firstQuery must be less than the number of queries in queryPool", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-firstQuery-09437", 43}, { {"The sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-flags-00815", 38}, { {"If VK_QUERY_RESULT_64_BIT is set in flags then pData must be aligned to a multiple of 8", 87}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-flags-02828", 38}, { {"If VK_QUERY_RESULT_64_BIT is not set in flags and the queryType used to create queryPool was not VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then pData must be aligned to a multiple of 4", 179}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-flags-09443", 38}, { {"If flags includes VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, then it must not include VK_QUERY_RESULT_WITH_AVAILABILITY_BIT", 117}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-flags-parameter", 42}, { {"flags must be a valid combination of VkQueryResultFlagBits values", 65}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-pData-parameter", 42}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryCount-09438", 43}, { {"If queryCount is greater than 1, stride must not be zero", 56}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryCount-12251", 43}, { {"If queryCount is greater than 1, VK_QUERY_RESULT_64_BIT is not set in flags and the queryType used to create queryPool was not VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then stride must be a multiple of 4", 199}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryCount-12252", 43}, { {"If queryCount is greater than 1 and VK_QUERY_RESULT_64_BIT is set in flags, then stride must be a multiple of 8", 111}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryCount-12253", 43}, { {"If queryCount is greater than 1 and the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then stride must be a multiple of the size of VkPerformanceCounterResultKHR", 191}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryPool-parameter", 46}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryPool-parent", 43}, { {"queryPool must have been created, allocated, or retrieved from device", 69}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-03229", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then pData must be aligned to a multiple of the size of VkPerformanceCounterResultKHR", 168}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-04519", 42}, { {"If queryCount is greater than 1 and the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, then stride must be large enough to contain the VkQueryPoolPerformanceCreateInfoKHR::counterIndexCount used to create queryPool times the size of VkPerformanceCounterResultKHR", 291}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-09439", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIMESTAMP, flags must not contain VK_QUERY_RESULT_PARTIAL_BIT", 121}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-09440", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, flags must not contain VK_QUERY_RESULT_WITH_AVAILABILITY_BIT, VK_QUERY_RESULT_WITH_STATUS_BIT_KHR, VK_QUERY_RESULT_PARTIAL_BIT, or VK_QUERY_RESULT_64_BIT", 236}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-09441", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the queryPool must have been recorded once for each pass as retrieved via a call to vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR", 222}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-09442", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR, then flags must include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 143}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-11874", 42}, { {"If the queryType used to create queryPool was not VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR or VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR, then flags must not include VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 194}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-queryType-12424", 42}, { {"If the queryType used to create queryPool was VK_QUERY_TYPE_TIME_ELAPSED_QCOM, flags must not contain VK_QUERY_RESULT_PARTIAL_BIT or VK_QUERY_RESULT_WITH_STATUS_BIT_KHR", 168}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueryPoolResults-stride-08993", 39}, { {"If queryCount is greater than 1 and VK_QUERY_RESULT_WITH_AVAILABILITY_BIT is set, stride must be large enough to contain the unsigned integer representing availability or status in addition to the query result", 209}, { "chapters/queries.html", 21} } },
        { {"VUID-vkGetQueueCheckpointData2NV-pCheckpointData-parameter", 58}, { {"If the value referenced by pCheckpointDataCount is not 0, and pCheckpointData is not NULL, pCheckpointData must be a valid pointer to an array of pCheckpointDataCount VkCheckpointData2NV structures", 197}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointData2NV-pCheckpointDataCount-parameter", 63}, { {"pCheckpointDataCount must be a valid pointer to a uint32_t value", 64}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointData2NV-queue-03892", 44}, { {"The device that queue belongs to must be in the lost state", 58}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointData2NV-queue-parameter", 48}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointDataNV-pCheckpointData-parameter", 57}, { {"If the value referenced by pCheckpointDataCount is not 0, and pCheckpointData is not NULL, pCheckpointData must be a valid pointer to an array of pCheckpointDataCount VkCheckpointDataNV structures", 196}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointDataNV-pCheckpointDataCount-parameter", 62}, { {"pCheckpointDataCount must be a valid pointer to a uint32_t value", 64}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointDataNV-queue-02025", 43}, { {"The device that queue belongs to must be in the lost state", 58}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetQueueCheckpointDataNV-queue-parameter", 47}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkGetRandROutputDisplayEXT-dpy-parameter", 45}, { {"dpy must be a valid pointer to a Display value", 46}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRandROutputDisplayEXT-pDisplay-parameter", 50}, { {"pDisplay must be a valid pointer to a VkDisplayKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRandROutputDisplayEXT-physicalDevice-parameter", 56}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-dataSize-03484", 69}, { {"dataSize must be at least VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleCaptureReplaySize x groupCount", 122}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-dataSize-arraylength", 75}, { {"dataSize must be greater than 0", 31}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-device-parameter", 71}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-firstGroup-03483", 71}, { {"The sum of firstGroup and groupCount must be less than or equal to the number of shader groups in pipeline", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-firstGroup-04051", 71}, { {"firstGroup must be less than the number of shader groups in pipeline", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pData-parameter", 70}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pipeline-03607", 69}, { {"pipeline must have been created with a flags that included VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR", 132}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pipeline-04620", 69}, { {"pipeline must be a ray tracing pipeline", 39}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pipeline-07829", 69}, { {"If the pipelineLibraryGroupHandles feature is not enabled, pipeline must not have been created with VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 134}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pipeline-parameter", 73}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-pipeline-parent", 70}, { {"pipeline must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingCaptureReplayShaderGroupHandlesKHR-rayTracingPipelineShaderGroupHandleCaptureReplay-03606", 109}, { {"VkPhysicalDeviceRayTracingPipelineFeaturesKHR::rayTracingPipelineShaderGroupHandleCaptureReplay must be enabled to call this function", 133}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-dataSize-02420", 56}, { {"dataSize must be at least VkPhysicalDeviceRayTracingPipelinePropertiesKHR::shaderGroupHandleSize x groupCount", 109}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-dataSize-arraylength", 62}, { {"dataSize must be greater than 0", 31}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-firstGroup-02419", 58}, { {"The sum of firstGroup and groupCount must be less than or equal to the number of shader groups in pipeline", 106}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-firstGroup-04050", 58}, { {"firstGroup must be less than the number of shader groups in pipeline", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-pData-parameter", 57}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-pipeline-04619", 56}, { {"pipeline must be a ray tracing pipeline", 39}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-pipeline-07828", 56}, { {"If the pipelineLibraryGroupHandles feature is not enabled, pipeline must not have been created with VK_PIPELINE_CREATE_LIBRARY_BIT_KHR", 134}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-pipeline-parameter", 60}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupHandlesKHR-pipeline-parent", 57}, { {"pipeline must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-group-03608", 55}, { {"The value of group must be less than the number of shader groups in pipeline", 76}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-groupShader-03609", 61}, { {"The shader identified by groupShader in group must not be VK_SHADER_UNUSED_KHR", 78}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-groupShader-parameter", 65}, { {"groupShader must be a valid VkShaderGroupShaderKHR value", 56}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-pipeline-04622", 58}, { {"pipeline must be a ray tracing pipeline", 39}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-pipeline-parameter", 62}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRayTracingShaderGroupStackSizeKHR-pipeline-parent", 59}, { {"pipeline must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetRefreshCycleDurationGOOGLE-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRefreshCycleDurationGOOGLE-pDisplayTimingProperties-parameter", 71}, { {"pDisplayTimingProperties must be a valid pointer to a VkRefreshCycleDurationGOOGLE structure", 92}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRefreshCycleDurationGOOGLE-swapchain-parameter", 56}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRefreshCycleDurationGOOGLE-swapchain-parent", 53}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetRenderAreaGranularity-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderAreaGranularity-pGranularity-parameter", 54}, { {"pGranularity must be a valid pointer to a VkExtent2D structure", 62}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderAreaGranularity-renderPass-parameter", 52}, { {"renderPass must be a valid VkRenderPass handle", 46}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderAreaGranularity-renderPass-parent", 49}, { {"renderPass must have been created, allocated, or retrieved from device", 70}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderingAreaGranularity-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderingAreaGranularity-pGranularity-parameter", 57}, { {"pGranularity must be a valid pointer to a VkExtent2D structure", 62}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetRenderingAreaGranularity-pRenderingAreaInfo-parameter", 63}, { {"pRenderingAreaInfo must be a valid pointer to a valid VkRenderingAreaInfo structure", 83}, { "chapters/renderpass.html", 24} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-None-08084", 58}, { {"The descriptorBufferCaptureReplay feature must be enabled", 57}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-device-08086", 60}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-device-parameter", 64}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-pData-08085", 59}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferPropertiesEXT::samplerCaptureReplayDescriptorDataSize bytes in size", 145}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-pData-parameter", 63}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetSamplerOpaqueCaptureDescriptorDataEXT-pInfo-parameter", 63}, { {"pInfo must be a valid pointer to a valid VkSamplerCaptureDescriptorDataInfoEXT structure", 88}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetScreenBufferPropertiesQNX-buffer-08968", 48}, { {"buffer must be a valid QNX Screen buffer", 40}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetScreenBufferPropertiesQNX-buffer-parameter", 52}, { {"buffer must be a valid pointer to a valid _screen_buffer value", 62}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetScreenBufferPropertiesQNX-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetScreenBufferPropertiesQNX-pProperties-parameter", 57}, { {"pProperties must be a valid pointer to a VkScreenBufferPropertiesQNX structure", 78}, { "chapters/memory.html", 20} } },
        { {"VUID-vkGetSemaphoreCounterValue-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreCounterValue-pValue-parameter", 48}, { {"pValue must be a valid pointer to a uint64_t value", 50}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreCounterValue-semaphore-03255", 47}, { {"semaphore must have been created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_TIMELINE", 85}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreCounterValue-semaphore-parameter", 51}, { {"semaphore must be a valid VkSemaphore handle", 44}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreCounterValue-semaphore-parent", 48}, { {"semaphore must have been created, allocated, or retrieved from device", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreFdKHR-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreFdKHR-pFd-parameter", 38}, { {"pFd must be a valid pointer to an int value", 43}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreFdKHR-pGetFdInfo-parameter", 45}, { {"pGetFdInfo must be a valid pointer to a valid VkSemaphoreGetFdInfoKHR structure", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreWin32HandleKHR-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreWin32HandleKHR-pGetWin32HandleInfo-parameter", 63}, { {"pGetWin32HandleInfo must be a valid pointer to a valid VkSemaphoreGetWin32HandleInfoKHR structure", 97}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreWin32HandleKHR-pHandle-parameter", 51}, { {"pHandle must be a valid pointer to a HANDLE value", 49}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreZirconHandleFUCHSIA-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreZirconHandleFUCHSIA-pGetZirconHandleInfo-parameter", 69}, { {"pGetZirconHandleInfo must be a valid pointer to a valid VkSemaphoreGetZirconHandleInfoFUCHSIA structure", 103}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetSemaphoreZirconHandleFUCHSIA-pZirconHandle-parameter", 62}, { {"pZirconHandle must be a valid pointer to a zx_handle_t value", 60}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkGetShaderBinaryDataEXT-None-08461", 40}, { {"The shaderObject feature must be enabled", 40}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-None-08499", 40}, { {"If pData is not NULL, it must be aligned to 16 bytes", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-pData-parameter", 45}, { {"If the value referenced by pDataSize is not 0, and pData is not NULL, pData must be a valid pointer to an array of pDataSize bytes", 130}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-pDataSize-parameter", 49}, { {"pDataSize must be a valid pointer to a size_t value", 51}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-shader-parameter", 46}, { {"shader must be a valid VkShaderEXT handle", 41}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderBinaryDataEXT-shader-parent", 43}, { {"shader must have been created, allocated, or retrieved from device", 66}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInfoAMD-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-infoType-parameter", 42}, { {"infoType must be a valid VkShaderInfoTypeAMD value", 50}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-pInfo-parameter", 39}, { {"If the value referenced by pInfoSize is not 0, and pInfo is not NULL, pInfo must be a valid pointer to an array of pInfoSize bytes", 130}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-pInfoSize-parameter", 43}, { {"pInfoSize must be a valid pointer to a size_t value", 51}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-pipeline-parameter", 42}, { {"pipeline must be a valid VkPipeline handle", 42}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-pipeline-parent", 39}, { {"pipeline must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInfoAMD-shaderStage-parameter", 45}, { {"shaderStage must be a valid VkShaderStageFlagBits value", 55}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-flags-zerobitmask", 58}, { {"flags must be 0", 15}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-instrumentation-parameter", 66}, { {"instrumentation must be a valid VkShaderInstrumentationARM handle", 65}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-instrumentation-parent", 63}, { {"instrumentation must have been created, allocated, or retrieved from device", 75}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-pMetricBlockCount-parameter", 68}, { {"pMetricBlockCount must be a valid pointer to a uint32_t value", 61}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderInstrumentationValuesARM-pMetricValues-parameter", 64}, { {"pMetricValues must be a pointer value", 37}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleCreateInfoIdentifierEXT-device-parameter", 62}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleCreateInfoIdentifierEXT-pCreateInfo-parameter", 67}, { {"pCreateInfo must be a valid pointer to a valid VkShaderModuleCreateInfo structure", 81}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleCreateInfoIdentifierEXT-pIdentifier-parameter", 67}, { {"pIdentifier must be a valid pointer to a VkShaderModuleIdentifierEXT structure", 78}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleCreateInfoIdentifierEXT-shaderModuleIdentifier-06885", 74}, { {"shaderModuleIdentifier feature must be enabled", 46}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleIdentifierEXT-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleIdentifierEXT-pIdentifier-parameter", 57}, { {"pIdentifier must be a valid pointer to a VkShaderModuleIdentifierEXT structure", 78}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleIdentifierEXT-shaderModule-parameter", 58}, { {"shaderModule must be a valid VkShaderModule handle", 50}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleIdentifierEXT-shaderModule-parent", 55}, { {"shaderModule must have been created, allocated, or retrieved from device", 72}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetShaderModuleIdentifierEXT-shaderModuleIdentifier-06884", 64}, { {"shaderModuleIdentifier feature must be enabled", 46}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetSleepStatusLegacyNV-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetSleepStatusLegacyNV-pLowLatencyMode-parameter", 55}, { {"pLowLatencyMode must be a valid pointer to a VkBool32 value", 59}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkGetSwapchainCounterEXT-counter-parameter", 47}, { {"counter must be a valid VkSurfaceCounterFlagBitsEXT value", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainCounterEXT-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainCounterEXT-pCounterValue-parameter", 53}, { {"pCounterValue must be a valid pointer to a uint64_t value", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainCounterEXT-swapchain-01245", 45}, { {"One or more present commands on swapchain must have been processed by the presentation engine", 93}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainCounterEXT-swapchain-parameter", 49}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainCounterEXT-swapchain-parent", 46}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainImagesKHR-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainImagesKHR-pSwapchainImageCount-parameter", 59}, { {"pSwapchainImageCount must be a valid pointer to a uint32_t value", 64}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainImagesKHR-pSwapchainImages-parameter", 55}, { {"If the value referenced by pSwapchainImageCount is not 0, and pSwapchainImages is not NULL, pSwapchainImages must be a valid pointer to an array of pSwapchainImageCount VkImage handles", 184}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainImagesKHR-swapchain-parameter", 48}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainImagesKHR-swapchain-parent", 45}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainStatusKHR-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainStatusKHR-swapchain-parameter", 48}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainStatusKHR-swapchain-parent", 45}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimeDomainPropertiesEXT-device-parameter", 59}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimeDomainPropertiesEXT-pSwapchainTimeDomainProperties-parameter", 83}, { {"pSwapchainTimeDomainProperties must be a valid pointer to a VkSwapchainTimeDomainPropertiesEXT structure", 104}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimeDomainPropertiesEXT-pTimeDomainsCounter-parameter", 72}, { {"If pTimeDomainsCounter is not NULL, pTimeDomainsCounter must be a valid pointer to a uint64_t value", 99}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimeDomainPropertiesEXT-swapchain-parameter", 62}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimeDomainPropertiesEXT-swapchain-parent", 59}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimingPropertiesEXT-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimingPropertiesEXT-pSwapchainTimingProperties-parameter", 75}, { {"pSwapchainTimingProperties must be a valid pointer to a VkSwapchainTimingPropertiesEXT structure", 96}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimingPropertiesEXT-pSwapchainTimingPropertiesCounter-parameter", 82}, { {"If pSwapchainTimingPropertiesCounter is not NULL, pSwapchainTimingPropertiesCounter must be a valid pointer to a uint64_t value", 127}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimingPropertiesEXT-swapchain-parameter", 58}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetSwapchainTimingPropertiesEXT-swapchain-parent", 55}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetTensorMemoryRequirementsARM-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorMemoryRequirementsARM-pInfo-parameter", 53}, { {"pInfo must be a valid pointer to a valid VkTensorMemoryRequirementsInfoARM structure", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorMemoryRequirementsARM-pMemoryRequirements-parameter", 67}, { {"pMemoryRequirements must be a valid pointer to a VkMemoryRequirements2 structure", 80}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-descriptorHeapCaptureReplay-11391", 70}, { {"The descriptorHeapCaptureReplay feature must be enabled", 55}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-device-11393", 49}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-pDatas-parameter", 53}, { {"pDatas must be a valid pointer to an array of tensorCount VkHostAddressRangeEXT structures", 90}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-pTensors-11394", 51}, { {"Each element of pTensors must have been created with VK_TENSOR_CREATE_DESCRIPTOR_HEAP_CAPTURE_REPLAY_BIT_ARM set in VkTensorCreateInfoARM::flags", 144}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-pTensors-parameter", 55}, { {"pTensors must be a valid pointer to an array of tensorCount valid VkTensorARM handles", 85}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-pTensors-parent", 52}, { {"Each element of pTensors must have been created, allocated, or retrieved from device", 84}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-size-11392", 47}, { {"The size member of each element of pDatas must be equal to tensorCaptureReplayOpaqueDataSize", 92}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDataARM-tensorCount-arraylength", 60}, { {"tensorCount must be greater than 0", 34}, { "chapters/resources.html", 23} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-descriptorBufferCaptureReplay-09702", 82}, { {"The descriptorBufferCaptureReplay and descriptorBufferTensorDescriptors features must be enabled", 96}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-device-09704", 59}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-device-parameter", 63}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-pData-09703", 58}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferTensorPropertiesARM::tensorCaptureReplayDescriptorDataSize bytes in size", 150}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-pData-parameter", 62}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorOpaqueCaptureDescriptorDataARM-pInfo-parameter", 62}, { {"pInfo must be a valid pointer to a valid VkTensorCaptureDescriptorDataInfoARM structure", 87}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-descriptorBufferCaptureReplay-09706", 86}, { {"The descriptorBufferCaptureReplay and descriptorBufferTensorDescriptors features must be enabled", 96}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-device-09708", 63}, { {"If device was created with multiple physical devices, then the bufferDeviceAddressMultiDevice feature must be enabled", 117}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-device-parameter", 67}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-pData-09707", 62}, { {"pData must point to a buffer that is at least VkPhysicalDeviceDescriptorBufferTensorPropertiesARM::tensorViewCaptureReplayDescriptorDataSize bytes in size", 154}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-pData-parameter", 66}, { {"pData must be a pointer value", 29}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetTensorViewOpaqueCaptureDescriptorDataARM-pInfo-parameter", 66}, { {"pInfo must be a valid pointer to a valid VkTensorViewCaptureDescriptorDataInfoARM structure", 91}, { "chapters/descriptorbuffers.html", 31} } },
        { {"VUID-vkGetValidationCacheDataEXT-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetValidationCacheDataEXT-pData-parameter", 48}, { {"If the value referenced by pDataSize is not 0, and pData is not NULL, pData must be a valid pointer to an array of pDataSize bytes", 130}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetValidationCacheDataEXT-pDataSize-parameter", 52}, { {"pDataSize must be a valid pointer to a size_t value", 51}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetValidationCacheDataEXT-validationCache-parameter", 58}, { {"validationCache must be a valid VkValidationCacheEXT handle", 59}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetValidationCacheDataEXT-validationCache-parent", 55}, { {"validationCache must have been created, allocated, or retrieved from device", 75}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkGetVideoSessionMemoryRequirementsKHR-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetVideoSessionMemoryRequirementsKHR-pMemoryRequirements-parameter", 73}, { {"If the value referenced by pMemoryRequirementsCount is not 0, and pMemoryRequirements is not NULL, pMemoryRequirements must be a valid pointer to an array of pMemoryRequirementsCount VkVideoSessionMemoryRequirementsKHR structures", 229}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetVideoSessionMemoryRequirementsKHR-pMemoryRequirementsCount-parameter", 78}, { {"pMemoryRequirementsCount must be a valid pointer to a uint32_t value", 68}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetVideoSessionMemoryRequirementsKHR-videoSession-parameter", 66}, { {"videoSession must be a valid VkVideoSessionKHR handle", 53}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetVideoSessionMemoryRequirementsKHR-videoSession-parent", 63}, { {"videoSession must have been created, allocated, or retrieved from device", 72}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkGetWinrtDisplayNV-pDisplay-parameter", 43}, { {"pDisplay must be a valid pointer to a VkDisplayKHR handle", 57}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkGetWinrtDisplayNV-physicalDevice-parameter", 49}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkImportFenceFdKHR-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportFenceFdKHR-fence-01463", 35}, { {"fence must not be associated with any queue command that has not yet completed execution on that queue", 102}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportFenceFdKHR-pImportFenceFdInfo-parameter", 52}, { {"pImportFenceFdInfo must be a valid pointer to a valid VkImportFenceFdInfoKHR structure", 86}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportFenceWin32HandleKHR-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportFenceWin32HandleKHR-fence-04448", 44}, { {"fence must not be associated with any queue command that has not yet completed execution on that queue", 102}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportFenceWin32HandleKHR-pImportFenceWin32HandleInfo-parameter", 70}, { {"pImportFenceWin32HandleInfo must be a valid pointer to a valid VkImportFenceWin32HandleInfoKHR structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreFdKHR-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreFdKHR-pImportSemaphoreFdInfo-parameter", 60}, { {"pImportSemaphoreFdInfo must be a valid pointer to a valid VkImportSemaphoreFdInfoKHR structure", 94}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreFdKHR-semaphore-01142", 43}, { {"semaphore must not be associated with any queue command that has not yet completed execution on that queue", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreWin32HandleKHR-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreWin32HandleKHR-pImportSemaphoreWin32HandleInfo-parameter", 78}, { {"pImportSemaphoreWin32HandleInfo must be a valid pointer to a valid VkImportSemaphoreWin32HandleInfoKHR structure", 112}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreZirconHandleFUCHSIA-device-parameter", 58}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreZirconHandleFUCHSIA-pImportSemaphoreZirconHandleInfo-parameter", 84}, { {"pImportSemaphoreZirconHandleInfo must be a valid pointer to a valid VkImportSemaphoreZirconHandleInfoFUCHSIA structure", 118}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkImportSemaphoreZirconHandleFUCHSIA-semaphore-04764", 57}, { {"semaphore must not be associated with any queue command that has not yet completed execution on that queue", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkInitializePerformanceApiINTEL-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkInitializePerformanceApiINTEL-pInitializeInfo-parameter", 62}, { {"pInitializeInfo must be a valid pointer to a valid VkInitializePerformanceApiInfoINTEL structure", 96}, { "chapters/queries.html", 21} } },
        { {"VUID-vkInvalidateMappedMemoryRanges-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkInvalidateMappedMemoryRanges-memoryRangeCount-arraylength", 64}, { {"memoryRangeCount must be greater than 0", 39}, { "chapters/memory.html", 20} } },
        { {"VUID-vkInvalidateMappedMemoryRanges-pMemoryRanges-parameter", 59}, { {"pMemoryRanges must be a valid pointer to an array of memoryRangeCount valid VkMappedMemoryRange structures", 106}, { "chapters/memory.html", 20} } },
        { {"VUID-vkLatencySleepLegacyNV-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkLatencySleepLegacyNV-signalSemaphore-parameter", 53}, { {"signalSemaphore must be a valid VkSemaphore handle", 50}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkLatencySleepLegacyNV-signalSemaphore-parent", 50}, { {"signalSemaphore must have been created, allocated, or retrieved from device", 75}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkLatencySleepNV-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkLatencySleepNV-pSleepInfo-parameter", 42}, { {"pSleepInfo must be a valid pointer to a valid VkLatencySleepInfoNV structure", 76}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkLatencySleepNV-swapchain-parameter", 41}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkLatencySleepNV-swapchain-parent", 38}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkMapMemory-device-parameter", 33}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-flags-09568", 28}, { {"VK_MEMORY_MAP_PLACED_BIT_EXT must not be set in flags", 53}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-flags-parameter", 32}, { {"flags must be a valid combination of VkMemoryMapFlagBits values", 63}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-memory-00678", 29}, { {"memory must not be currently host mapped", 40}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-memory-00682", 29}, { {"memory must have been created with a memory type that reports VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT", 97}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-memory-00683", 29}, { {"memory must not have been allocated with multiple instances", 59}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-memory-parameter", 33}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-memory-parent", 30}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-offset-00679", 29}, { {"offset must be less than the size of memory", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-ppData-parameter", 33}, { {"ppData must be a valid pointer to a pointer value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-size-00680", 27}, { {"If size is not equal to VK_WHOLE_SIZE, size must be greater than 0", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory-size-00681", 27}, { {"If size is not equal to VK_WHOLE_SIZE, size must be less than or equal to the size of the memory minus offset", 109}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory2-device-parameter", 34}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory2-pMemoryMapInfo-parameter", 42}, { {"pMemoryMapInfo must be a valid pointer to a valid VkMemoryMapInfo structure", 75}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMapMemory2-ppData-parameter", 34}, { {"ppData must be a valid pointer to a pointer value", 49}, { "chapters/memory.html", 20} } },
        { {"VUID-vkMergePipelineCaches-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-dstCache-00770", 41}, { {"dstCache must not appear in the list of source caches", 53}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-dstCache-10202", 41}, { {"Host access to dstCache must be externally synchronized if it was not created with VK_PIPELINE_CACHE_CREATE_INTERNALLY_SYNCHRONIZED_MERGE_BIT_KHR", 145}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-dstCache-11832", 41}, { {"dstCache must not have been created with the headerVersion member of VkPipelineCacheCreateInfo::pInitialData equal to VK_PIPELINE_CACHE_HEADER_VERSION_DATA_GRAPH_QCOM", 166}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-dstCache-parameter", 45}, { {"dstCache must be a valid VkPipelineCache handle", 47}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-dstCache-parent", 42}, { {"dstCache must have been created, allocated, or retrieved from device", 68}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-headerVersion-11833", 46}, { {"Each member of pSrcCaches must not have been created with the headerVersion member of VkPipelineCacheCreateInfo::pInitialData equal to VK_PIPELINE_CACHE_HEADER_VERSION_DATA_GRAPH_QCOM", 183}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-pSrcCaches-parameter", 47}, { {"pSrcCaches must be a valid pointer to an array of srcCacheCount valid VkPipelineCache handles", 93}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-pSrcCaches-parent", 44}, { {"Each element of pSrcCaches must have been created, allocated, or retrieved from device", 86}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergePipelineCaches-srcCacheCount-arraylength", 52}, { {"srcCacheCount must be greater than 0", 36}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkMergeValidationCachesEXT-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-dstCache-01536", 46}, { {"dstCache must not appear in the list of source caches", 53}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-dstCache-parameter", 50}, { {"dstCache must be a valid VkValidationCacheEXT handle", 52}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-dstCache-parent", 47}, { {"dstCache must have been created, allocated, or retrieved from device", 68}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-pSrcCaches-parameter", 52}, { {"pSrcCaches must be a valid pointer to an array of srcCacheCount valid VkValidationCacheEXT handles", 98}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-pSrcCaches-parent", 49}, { {"Each element of pSrcCaches must have been created, allocated, or retrieved from device", 86}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkMergeValidationCachesEXT-srcCacheCount-arraylength", 57}, { {"srcCacheCount must be greater than 0", 36}, { "chapters/shaders.html", 21} } },
        { {"VUID-vkQueueBeginDebugUtilsLabelEXT-pLabelInfo-parameter", 56}, { {"pLabelInfo must be a valid pointer to a valid VkDebugUtilsLabelEXT structure", 76}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueBeginDebugUtilsLabelEXT-queue-parameter", 51}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueBindSparse-commonparent", 35}, { {"Both of fence, and queue that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 144}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-fence-01113", 34}, { {"If fence is not VK_NULL_HANDLE, fence must be unsignaled", 56}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-fence-01114", 34}, { {"If fence is not VK_NULL_HANDLE, fence must not be associated with any other queue command that has not yet completed execution on that queue", 140}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-fence-parameter", 38}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-pBindInfo-parameter", 42}, { {"If bindInfoCount is not 0, pBindInfo must be a valid pointer to an array of bindInfoCount valid VkBindSparseInfo structures", 123}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-pSignalSemaphores-01115", 46}, { {"Each element of the pSignalSemaphores member of each element of pBindInfo must be unsignaled when the semaphore signal operation it defines is executed on the device", 165}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-pWaitSemaphores-01116", 44}, { {"When a semaphore wait operation referring to a binary semaphore defined by any element of the pWaitSemaphores member of any element of pBindInfo executes on queue, there must be no other queues waiting on the same semaphore", 223}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-pWaitSemaphores-03245", 44}, { {"All elements of the pWaitSemaphores member of all elements of pBindInfo referring to a semaphore created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY must reference a semaphore signal operation that has been submitted for execution and any semaphore signal operations on which it depends must have also been submitted for execution", 337}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-queue-parameter", 38}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueBindSparse-queuetype", 32}, { {"The queue must support VK_QUEUE_SPARSE_BINDING_BIT operations", 61}, { "chapters/sparsemem.html", 23} } },
        { {"VUID-vkQueueEndDebugUtilsLabelEXT-None-01911", 44}, { {"There must be an outstanding vkQueueBeginDebugUtilsLabelEXT command prior to the vkQueueEndDebugUtilsLabelEXT on the queue", 122}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueEndDebugUtilsLabelEXT-queue-parameter", 49}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueInsertDebugUtilsLabelEXT-pLabelInfo-parameter", 57}, { {"pLabelInfo must be a valid pointer to a valid VkDebugUtilsLabelEXT structure", 76}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueInsertDebugUtilsLabelEXT-queue-parameter", 52}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkQueueNotifyOutOfBandLegacyNV-queue-parameter", 51}, { {"queue must be a valid VkQueue handle", 36}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkQueueNotifyOutOfBandNV-pQueueTypeInfo-parameter", 54}, { {"pQueueTypeInfo must be a valid pointer to a valid VkOutOfBandQueueTypeInfoNV structure", 86}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueueNotifyOutOfBandNV-queue-parameter", 45}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pPresentInfo-parameter", 45}, { {"pPresentInfo must be a valid pointer to a valid VkPresentInfoKHR structure", 74}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pSwapchains-01292", 40}, { {"Each element of pSwapchains member of pPresentInfo must be a swapchain that is created for a surface for which presentation is supported from queue as determined using a call to vkGetPhysicalDeviceSurfaceSupportKHR", 214}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pSwapchains-01293", 40}, { {"If more than one member of pSwapchains was created from a display surface, all display surfaces referenced that refer to the same display must use the same display mode", 168}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pSwapchains-10285", 40}, { {"If more than one member of pSwapchains was created from a display surface, all display surfaces referenced that refer to the same display must use the same stereoType", 166}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pWaitSemaphores-01294", 44}, { {"When a semaphore wait operation referring to a binary semaphore defined by the elements of the pWaitSemaphores member of pPresentInfo executes on queue, there must be no other queues waiting on the same semaphore", 212}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pWaitSemaphores-03267", 44}, { {"All elements of the pWaitSemaphores member of pPresentInfo must be created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY", 125}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-pWaitSemaphores-03268", 44}, { {"All elements of the pWaitSemaphores member of pPresentInfo must reference a semaphore signal operation that has been submitted for execution and any semaphore signal operations on which it depends must have also been submitted for execution", 240}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueuePresentKHR-queue-parameter", 38}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkQueueSetPerfHintQCOM-pPerfHintInfo-parameter", 51}, { {"pPerfHintInfo must be a valid pointer to a valid VkPerfHintInfoQCOM structure", 77}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkQueueSetPerfHintQCOM-queue-12388", 39}, { {"queue must support at least one of the queue types specified in VkPhysicalDeviceQueuePerfHintPropertiesQCOM::supportedQueues", 124}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkQueueSetPerfHintQCOM-queue-parameter", 43}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkQueueSetPerfHintQCOM-queuePerfHint-12387", 47}, { {"The queuePerfHint feature must be enabled", 41}, { "chapters/devsandqueues.html", 27} } },
        { {"VUID-vkQueueSetPerformanceConfigurationINTEL-commonparent", 57}, { {"Both of configuration, and queue must have been created, allocated, or retrieved from the same VkDevice", 103}, { "chapters/queries.html", 21} } },
        { {"VUID-vkQueueSetPerformanceConfigurationINTEL-configuration-parameter", 68}, { {"configuration must be a valid VkPerformanceConfigurationINTEL handle", 68}, { "chapters/queries.html", 21} } },
        { {"VUID-vkQueueSetPerformanceConfigurationINTEL-queue-parameter", 60}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/queries.html", 21} } },
        { {"VUID-vkQueueSubmit-commonparent", 31}, { {"Both of fence, and queue that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 144}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-fence-00063", 30}, { {"If fence is not VK_NULL_HANDLE, fence must be unsignaled", 56}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-fence-00064", 30}, { {"If fence is not VK_NULL_HANDLE, fence must not be associated with any other queue command that has not yet completed execution on that queue", 140}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-fence-parameter", 34}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00065", 40}, { {"Any calls to vkCmdSetEvent, vkCmdResetEvent or vkCmdWaitEvents that have been recorded into any of the command buffer elements of the pCommandBuffers member of any element of pSubmits, must not reference any VkEvent that is referenced by any of those commands in a command buffer that has been submitted to another queue and is still in the pending state", 354}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00070", 40}, { {"Each element of the pCommandBuffers member of each element of pSubmits must be in the pending or executable state", 113}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00071", 40}, { {"If any element of the pCommandBuffers member of any element of pSubmits was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT, it must not be in the pending state", 179}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00072", 40}, { {"Any secondary command buffers recorded into any element of the pCommandBuffers member of any element of pSubmits must be in the pending or executable state", 155}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00073", 40}, { {"If any secondary command buffers recorded into any element of the pCommandBuffers member of any element of pSubmits was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT, it must not be in the pending state", 223}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-00074", 40}, { {"Each element of the pCommandBuffers member of each element of pSubmits must have been allocated from a VkCommandPool that was created for the same queue family queue belongs to", 176}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pCommandBuffers-03220", 40}, { {"If a command recorded into any element of pCommandBuffers was a vkCmdBeginQuery whose queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the profiling lock must have been held continuously on the VkDevice that queue was retrieved from, throughout recording of those command buffers", 309}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSignalSemaphores-00067", 42}, { {"Each binary semaphore element of the pSignalSemaphores member of any element of pSubmits must be unsignaled when the semaphore signal operation it defines is executed on the device", 180}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSubmits-02207", 33}, { {"If any element of pSubmits->pCommandBuffers includes a Queue Family Ownership Transfer Acquire Operation, there must exist a previously submitted Queue Family Ownership Transfer Release Operation on a queue in the queue family identified by the acquire operation, with parameters matching the acquire operation as defined in the definition of such acquire operations, and which happens-before the acquire operation", 414}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSubmits-02808", 33}, { {"Any resource created with VK_SHARING_MODE_EXCLUSIVE that is read by an operation specified by pSubmits must not be owned by any queue family other than the one which queue belongs to, at the time it is executed", 210}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSubmits-04626", 33}, { {"Any resource created with VK_SHARING_MODE_CONCURRENT that is accessed by an operation specified by pSubmits must have included the queue family of queue at resource creation time", 178}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSubmits-10911", 33}, { {"If any element of pSubmits->pCommandBuffers includes a Queue Family Ownership Transfer Acquire Operation, the affected resource must not be modified in any way between the last matching release operation and the acquire operation", 229}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pSubmits-parameter", 37}, { {"If submitCount is not 0, pSubmits must be a valid pointer to an array of submitCount valid VkSubmitInfo structures", 114}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pWaitDstStageMask-00066", 42}, { {"Any stage flag included in any element of the pWaitDstStageMask member of any element of pSubmits must be a pipeline stage supported by one of the capabilities of queue, as specified in the table of supported pipeline stages", 224}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pWaitSemaphores-00068", 40}, { {"When a semaphore wait operation referring to a binary semaphore defined by any element of the pWaitSemaphores member of any element of pSubmits executes on queue, there must be no other queues waiting on the same semaphore", 222}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-pWaitSemaphores-03238", 40}, { {"All elements of the pWaitSemaphores member of all elements of pSubmits created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY must reference a semaphore signal operation that has been submitted for execution and any semaphore signal operations on which it depends must have also been submitted for execution", 311}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-queue-06448", 30}, { {"If queue was not created with VK_DEVICE_QUEUE_CREATE_PROTECTED_BIT, there must be no element of pSubmits that includes a VkProtectedSubmitInfo structure in its pNext chain with protectedSubmit equal to VK_TRUE", 209}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit-queue-parameter", 34}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03867", 39}, { {"If a command recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits referenced a VkEvent, that event must not be referenced by a command that has been submitted to another queue and is still in the pending state", 271}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03874", 39}, { {"The commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits must be in the pending or executable state", 143}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03875", 39}, { {"If a command recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT, it must not be in the pending state", 235}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03876", 39}, { {"Any secondary command buffers recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits must be in the pending or executable state", 187}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03877", 39}, { {"If any secondary command buffers recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits was not recorded with the VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT, it must not be in the pending state", 255}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03878", 39}, { {"The commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits must have been allocated from a VkCommandPool that was created for the same queue family queue belongs to", 206}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03879", 39}, { {"If a command recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits includes a Queue Family Ownership Transfer Acquire Operation, there must exist a previously submitted Queue Family Ownership Transfer Release Operation on a queue in the queue family identified by the acquire operation, with parameters matching the acquire operation as defined in the definition of such acquire operations, and which happens before the acquire operation", 498}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-03880", 39}, { {"If a command recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits was a vkCmdBeginQuery whose queryPool was created with a queryType of VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR, the profiling lock must have been held continuously on the VkDevice that queue was retrieved from, throughout recording of those command buffers", 379}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commandBuffer-10910", 39}, { {"If a command recorded into the commandBuffer member of any element of the pCommandBufferInfos member of any element of pSubmits includes a Queue Family Ownership Transfer Acquire Operation, the affected resource must not be modified in any way between the last matching release operation and the acquire operation", 313}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-commonparent", 32}, { {"Both of fence, and queue that are valid handles of non-ignored parameters must have been created, allocated, or retrieved from the same VkDevice", 144}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-fence-04894", 31}, { {"If fence is not VK_NULL_HANDLE, fence must be unsignaled", 56}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-fence-04895", 31}, { {"If fence is not VK_NULL_HANDLE, fence must not be associated with any other queue command that has not yet completed execution on that queue", 140}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-fence-parameter", 35}, { {"If fence is not VK_NULL_HANDLE, fence must be a valid VkFence handle", 68}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-pSubmits-parameter", 38}, { {"If submitCount is not 0, pSubmits must be a valid pointer to an array of submitCount valid VkSubmitInfo2 structures", 115}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-queue-06447", 31}, { {"If queue was not created with VK_DEVICE_QUEUE_CREATE_PROTECTED_BIT, the flags member of any element of pSubmits must not include VK_SUBMIT_PROTECTED_BIT_KHR", 156}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-queue-parameter", 35}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-semaphore-03868", 35}, { {"The semaphore member of any binary semaphore element of the pSignalSemaphoreInfos member of any element of pSubmits must be unsignaled when the semaphore signal operation it defines is executed on the device", 207}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-semaphore-03871", 35}, { {"When a semaphore wait operation for a binary semaphore is executed, as defined by the semaphore member of any element of the pWaitSemaphoreInfos member of any element of pSubmits, there must be no other queues waiting on the same semaphore", 239}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-semaphore-03873", 35}, { {"The semaphore member of any element of the pWaitSemaphoreInfos member of any element of pSubmits that was created with a VkSemaphoreType of VK_SEMAPHORE_TYPE_BINARY must reference a semaphore signal operation that has been submitted for execution and any semaphore signal operations on which it depends must have also been submitted for execution", 346}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-stageMask-03869", 35}, { {"The stageMask member of any element of the pSignalSemaphoreInfos member of any element of pSubmits must only include pipeline stages that are supported by the queue family which queue belongs to", 194}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-stageMask-03870", 35}, { {"The stageMask member of any element of the pWaitSemaphoreInfos member of any element of pSubmits must only include pipeline stages that are supported by the queue family which queue belongs to", 192}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueSubmit2-synchronization2-03866", 42}, { {"The synchronization2 feature must be enabled", 44}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkQueueWaitIdle-queue-parameter", 36}, { {"queue must be a valid VkQueue handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterCustomBorderColorEXT-device-parameter", 52}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkRegisterCustomBorderColorEXT-pBorderColor-parameter", 58}, { {"pBorderColor must be a valid pointer to a valid VkSamplerCustomBorderColorCreateInfoEXT structure", 97}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkRegisterCustomBorderColorEXT-pIndex-parameter", 52}, { {"pIndex must be a valid pointer to a uint32_t value", 50}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkRegisterCustomBorderColorEXT-requestIndex-11287", 54}, { {"If requestIndex is VK_TRUE, the value stored in pIndex must be less than maxCustomBorderColorSamplers", 101}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkRegisterDeviceEventEXT-device-parameter", 46}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDeviceEventEXT-pAllocator-parameter", 50}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDeviceEventEXT-pDeviceEventInfo-parameter", 56}, { {"pDeviceEventInfo must be a valid pointer to a valid VkDeviceEventInfoEXT structure", 82}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDeviceEventEXT-pFence-parameter", 46}, { {"pFence must be a valid pointer to a VkFence handle", 50}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-commonparent", 43}, { {"Both of device, and display must have been created, allocated, or retrieved from the same VkPhysicalDevice", 106}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-display-parameter", 48}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-pAllocator-parameter", 51}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-pDisplayEventInfo-parameter", 58}, { {"pDisplayEventInfo must be a valid pointer to a valid VkDisplayEventInfoEXT structure", 84}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkRegisterDisplayEventEXT-pFence-parameter", 47}, { {"pFence must be a valid pointer to a VkFence handle", 50}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkReleaseCapturedPipelineDataKHR-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkReleaseCapturedPipelineDataKHR-pAllocator-parameter", 58}, { {"If pAllocator is not NULL, pAllocator must be a valid pointer to a valid VkAllocationCallbacks structure", 104}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkReleaseCapturedPipelineDataKHR-pInfo-parameter", 53}, { {"pInfo must be a valid pointer to a valid VkReleaseCapturedPipelineDataInfoKHR structure", 87}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkReleaseCapturedPipelineDataKHR-pipeline-09611", 52}, { {"If VkAllocationCallbacks were provided when pipeline was created, a compatible set of callbacks must be provided in pAllocator", 126}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkReleaseCapturedPipelineDataKHR-pipeline-09612", 52}, { {"If no VkAllocationCallbacks were provided when pipeline was created, pAllocator must be NULL", 92}, { "chapters/pipelines.html", 23} } },
        { {"VUID-vkReleaseDisplayEXT-display-parameter", 42}, { {"display must be a valid VkDisplayKHR handle", 43}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseDisplayEXT-display-parent", 39}, { {"display must have been created, allocated, or retrieved from physicalDevice", 75}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseDisplayEXT-physicalDevice-parameter", 49}, { {"physicalDevice must be a valid VkPhysicalDevice handle", 54}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseFullScreenExclusiveModeEXT-device-parameter", 57}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseFullScreenExclusiveModeEXT-swapchain-02677", 56}, { {"swapchain must not be in the retired state", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseFullScreenExclusiveModeEXT-swapchain-02678", 56}, { {"swapchain must be a swapchain created with a VkSurfaceFullScreenExclusiveInfoEXT structure, with fullScreenExclusive set to VK_FULL_SCREEN_EXCLUSIVE_APPLICATION_CONTROLLED_EXT", 175}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseFullScreenExclusiveModeEXT-swapchain-parameter", 60}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseFullScreenExclusiveModeEXT-swapchain-parent", 57}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleasePerformanceConfigurationINTEL-configuration-02737", 63}, { {"configuration must not be released before all command buffers submitted while the configuration was set are in pending state", 124}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleasePerformanceConfigurationINTEL-configuration-parameter", 67}, { {"If configuration is not VK_NULL_HANDLE, configuration must be a valid VkPerformanceConfigurationINTEL handle", 108}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleasePerformanceConfigurationINTEL-configuration-parent", 64}, { {"If configuration is a valid handle, it must have been created, allocated, or retrieved from device", 98}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleasePerformanceConfigurationINTEL-device-parameter", 60}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleaseProfilingLockKHR-device-03235", 43}, { {"The profiling lock of device must have been held via a previous successful call to vkAcquireProfilingLockKHR", 108}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleaseProfilingLockKHR-device-parameter", 47}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkReleaseSwapchainImagesKHR-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseSwapchainImagesKHR-pReleaseInfo-parameter", 55}, { {"pReleaseInfo must be a valid pointer to a valid VkReleaseSwapchainImagesInfoKHR structure", 89}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkReleaseSwapchainImagesKHR-swapchainMaintenance1-10159", 60}, { {"Feature swapchainMaintenance1 must be enabled", 45}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkResetCommandBuffer-commandBuffer-00045", 45}, { {"commandBuffer must not be in the pending state", 46}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandBuffer-commandBuffer-00046", 45}, { {"commandBuffer must have been allocated from a pool that was created with the VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT", 124}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandBuffer-commandBuffer-parameter", 49}, { {"commandBuffer must be a valid VkCommandBuffer handle", 52}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandBuffer-flags-parameter", 41}, { {"flags must be a valid combination of VkCommandBufferResetFlagBits values", 72}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandPool-commandPool-00040", 41}, { {"All VkCommandBuffer objects allocated from commandPool must not be in the pending state", 87}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandPool-commandPool-parameter", 45}, { {"commandPool must be a valid VkCommandPool handle", 48}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandPool-commandPool-parent", 42}, { {"commandPool must have been created, allocated, or retrieved from device", 71}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandPool-device-parameter", 40}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetCommandPool-flags-parameter", 39}, { {"flags must be a valid combination of VkCommandPoolResetFlagBits values", 70}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkResetDescriptorPool-descriptorPool-00313", 47}, { {"All uses of descriptorPool (via any allocated descriptor sets) must have completed execution", 92}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkResetDescriptorPool-descriptorPool-parameter", 51}, { {"descriptorPool must be a valid VkDescriptorPool handle", 54}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkResetDescriptorPool-descriptorPool-parent", 48}, { {"descriptorPool must have been created, allocated, or retrieved from device", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkResetDescriptorPool-device-parameter", 43}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkResetDescriptorPool-flags-zerobitmask", 44}, { {"flags must be 0", 15}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkResetEvent-device-parameter", 34}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetEvent-event-03821", 29}, { {"There must be an execution dependency between vkResetEvent and the execution of any vkCmdWaitEvents that includes event in its pEvents parameter", 144}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetEvent-event-03822", 29}, { {"There must be an execution dependency between vkResetEvent and the execution of any vkCmdWaitEvents2 that includes event in its pEvents parameter", 145}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetEvent-event-03823", 29}, { {"event must not have been created with VK_EVENT_CREATE_DEVICE_ONLY_BIT", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetEvent-event-parameter", 33}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetEvent-event-parent", 30}, { {"event must have been created, allocated, or retrieved from device", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetFences-device-parameter", 35}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetFences-fenceCount-arraylength", 41}, { {"fenceCount must be greater than 0", 33}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetFences-pFences-01123", 32}, { {"Each element of pFences must not be currently associated with any queue command that has not yet completed execution on that queue", 130}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetFences-pFences-parameter", 36}, { {"pFences must be a valid pointer to an array of fenceCount valid VkFence handles", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetFences-pFences-parent", 33}, { {"Each element of pFences must have been created, allocated, or retrieved from device", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkResetGpaSessionAMD-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkResetGpaSessionAMD-gpaSession-parameter", 46}, { {"gpaSession must be a valid VkGpaSessionAMD handle", 49}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkResetGpaSessionAMD-gpaSession-parent", 43}, { {"gpaSession must have been created, allocated, or retrieved from device", 70}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkResetQueryPool-None-02665", 32}, { {"The hostQueryReset feature must be enabled", 42}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-firstQuery-02741", 38}, { {"Submitted commands that refer to the range specified by firstQuery and queryCount in queryPool must have completed execution", 124}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-firstQuery-02742", 38}, { {"The range of queries specified by firstQuery and queryCount in queryPool must not be in use by calls to vkGetQueryPoolResults or vkResetQueryPool in other threads", 162}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-firstQuery-09436", 38}, { {"firstQuery must be less than the number of queries in queryPool", 63}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-firstQuery-09437", 38}, { {"The sum of firstQuery and queryCount must be less than or equal to the number of queries in queryPool", 101}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-queryPool-parameter", 41}, { {"queryPool must be a valid VkQueryPool handle", 44}, { "chapters/queries.html", 21} } },
        { {"VUID-vkResetQueryPool-queryPool-parent", 38}, { {"queryPool must have been created, allocated, or retrieved from device", 69}, { "chapters/queries.html", 21} } },
        { {"VUID-vkSetBufferCollectionBufferConstraintsFUCHSIA-collection-06403", 67}, { {"vkSetBufferCollectionImageConstraintsFUCHSIA or vkSetBufferCollectionBufferConstraintsFUCHSIA must not have already been called on collection", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionBufferConstraintsFUCHSIA-collection-parameter", 71}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionBufferConstraintsFUCHSIA-collection-parent", 68}, { {"collection must have been created, allocated, or retrieved from device", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionBufferConstraintsFUCHSIA-device-parameter", 67}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionBufferConstraintsFUCHSIA-pBufferConstraintsInfo-parameter", 83}, { {"pBufferConstraintsInfo must be a valid pointer to a valid VkBufferConstraintsInfoFUCHSIA structure", 98}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionImageConstraintsFUCHSIA-collection-06394", 66}, { {"vkSetBufferCollectionImageConstraintsFUCHSIA or vkSetBufferCollectionBufferConstraintsFUCHSIA must not have already been called on collection", 141}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionImageConstraintsFUCHSIA-collection-parameter", 70}, { {"collection must be a valid VkBufferCollectionFUCHSIA handle", 59}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionImageConstraintsFUCHSIA-collection-parent", 67}, { {"collection must have been created, allocated, or retrieved from device", 70}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionImageConstraintsFUCHSIA-device-parameter", 66}, { {"device must be a valid VkDevice handle", 38}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetBufferCollectionImageConstraintsFUCHSIA-pImageConstraintsInfo-parameter", 81}, { {"pImageConstraintsInfo must be a valid pointer to a valid VkImageConstraintsInfoFUCHSIA structure", 96}, { "chapters/resources.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-02587", 49}, { {"pNameInfo->objectType must not be VK_OBJECT_TYPE_UNKNOWN", 56}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-02588", 49}, { {"pNameInfo->objectHandle must not be VK_NULL_HANDLE", 50}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-07872", 49}, { {"If pNameInfo->objectHandle is the valid handle of an instance-level object, the VkDevice identified by device must be a descendent of the same VkInstance as the object identified by pNameInfo->objectHandle", 205}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-07873", 49}, { {"If pNameInfo->objectHandle is the valid handle of a physical-device-level object, the VkDevice identified by device must be a descendant of the same VkPhysicalDevice as the object identified by pNameInfo->objectHandle", 217}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-07874", 49}, { {"If pNameInfo->objectHandle is the valid handle of a device-level object, that object must be a descendent of the VkDevice identified by device", 142}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectNameEXT-pNameInfo-parameter", 53}, { {"pNameInfo must be a valid pointer to a valid VkDebugUtilsObjectNameInfoEXT structure", 84}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectTagEXT-device-parameter", 49}, { {"device must be a valid VkDevice handle", 38}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectTagEXT-pNameInfo-07875", 48}, { {"If pNameInfo->objectHandle is the valid handle of an instance-level object, the VkDevice identified by device must be a descendent of the same VkInstance as the object identified by pNameInfo->objectHandle", 205}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectTagEXT-pNameInfo-07876", 48}, { {"If pNameInfo->objectHandle is the valid handle of a physical-device-level object, the VkDevice identified by device must be a descendant of the same VkPhysicalDevice as the object identified by pNameInfo->objectHandle", 217}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectTagEXT-pNameInfo-07877", 48}, { {"If pNameInfo->objectHandle is the valid handle of a device-level object, that object must be a descendent of the VkDevice identified by device", 142}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDebugUtilsObjectTagEXT-pTagInfo-parameter", 51}, { {"pTagInfo must be a valid pointer to a valid VkDebugUtilsObjectTagInfoEXT structure", 82}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSetDeviceMemoryPriorityEXT-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkSetDeviceMemoryPriorityEXT-memory-parameter", 50}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkSetDeviceMemoryPriorityEXT-memory-parent", 47}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkSetDeviceMemoryPriorityEXT-priority-06258", 48}, { {"priority must be between 0 and 1, inclusive", 43}, { "chapters/memory.html", 20} } },
        { {"VUID-vkSetEvent-device-parameter", 32}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSetEvent-event-03941", 27}, { {"event must not have been created with VK_EVENT_CREATE_DEVICE_ONLY_BIT", 69}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSetEvent-event-09543", 27}, { {"event must not be waited on by a command buffer in the pending state", 68}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSetEvent-event-parameter", 31}, { {"event must be a valid VkEvent handle", 36}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSetEvent-event-parent", 28}, { {"event must have been created, allocated, or retrieved from device", 65}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSetGpaDeviceClockModeAMD-clockModes-12415", 48}, { {"The VkPhysicalDeviceGpaFeaturesAMD::clockModes feature must be enabled", 70}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkSetGpaDeviceClockModeAMD-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkSetGpaDeviceClockModeAMD-pInfo-parameter", 47}, { {"pInfo must be a valid pointer to a VkGpaDeviceClockModeInfoAMD structure", 72}, { "chapters/gpa_interface.html", 27} } },
        { {"VUID-vkSetHdrMetadataEXT-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetHdrMetadataEXT-pMetadata-parameter", 44}, { {"pMetadata must be a valid pointer to an array of swapchainCount valid VkHdrMetadataEXT structures", 97}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetHdrMetadataEXT-pSwapchains-parameter", 46}, { {"pSwapchains must be a valid pointer to an array of swapchainCount valid VkSwapchainKHR handles", 94}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetHdrMetadataEXT-pSwapchains-parent", 43}, { {"Each element of pSwapchains must have been created, allocated, or retrieved from device", 87}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetHdrMetadataEXT-swapchainCount-arraylength", 51}, { {"swapchainCount must be greater than 0", 37}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencyMarkerLegacyNV-device-parameter", 48}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkSetLatencyMarkerNV-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencyMarkerNV-pLatencyMarkerInfo-parameter", 54}, { {"pLatencyMarkerInfo must be a valid pointer to a valid VkSetLatencyMarkerInfoNV structure", 88}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencyMarkerNV-swapchain-parameter", 45}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencyMarkerNV-swapchain-parent", 42}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencySleepModeLegacyNV-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkSetLatencySleepModeNV-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencySleepModeNV-pSleepModeInfo-parameter", 53}, { {"pSleepModeInfo must be a valid pointer to a valid VkLatencySleepModeInfoNV structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencySleepModeNV-swapchain-parameter", 48}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLatencySleepModeNV-swapchain-parent", 45}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLocalDimmingAMD-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLocalDimmingAMD-localDimmingSupport-04618", 51}, { {"VkDisplayNativeHdrSurfaceCapabilitiesAMD::localDimmingSupport must be supported", 79}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLocalDimmingAMD-swapChain-parameter", 45}, { {"swapChain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetLocalDimmingAMD-swapChain-parent", 42}, { {"swapChain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetPrivateData-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetPrivateData-objectHandle-04016", 40}, { {"objectHandle must be device or a child of device", 48}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetPrivateData-objectHandle-04017", 40}, { {"objectHandle must be a valid handle to an object of type objectType", 67}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetPrivateData-objectType-parameter", 42}, { {"objectType must be a valid VkObjectType value", 45}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetPrivateData-privateDataSlot-parameter", 47}, { {"privateDataSlot must be a valid VkPrivateDataSlot handle", 56}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetPrivateData-privateDataSlot-parent", 44}, { {"privateDataSlot must have been created, allocated, or retrieved from device", 75}, { "chapters/private_data.html", 26} } },
        { {"VUID-vkSetSwapchainPresentTimingQueueSizeEXT-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetSwapchainPresentTimingQueueSizeEXT-swapchain-12229", 60}, { {"swapchain must have been created with VkSwapchainCreateInfoKHR::flags containing VK_SWAPCHAIN_CREATE_PRESENT_TIMING_BIT_EXT", 123}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetSwapchainPresentTimingQueueSizeEXT-swapchain-parameter", 64}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkSetSwapchainPresentTimingQueueSizeEXT-swapchain-parent", 61}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkShutdownLatencyDeviceLegacyNV-device-parameter", 53}, { {"device must be a valid VkDevice handle", 38}, { "appendices/extensions.html", 26} } },
        { {"VUID-vkSignalSemaphore-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSignalSemaphore-pSignalInfo-parameter", 44}, { {"pSignalInfo must be a valid pointer to a valid VkSemaphoreSignalInfo structure", 78}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-instance-parameter", 52}, { {"instance must be a valid VkInstance handle", 42}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-messageSeverity-parameter", 59}, { {"messageSeverity must be a valid VkDebugUtilsMessageSeverityFlagBitsEXT value", 76}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-messageTypes-parameter", 56}, { {"messageTypes must be a valid combination of VkDebugUtilsMessageTypeFlagBitsEXT values", 85}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-messageTypes-requiredbitmask", 62}, { {"messageTypes must not be 0", 26}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-objectType-02591", 50}, { {"The objectType member of each element of pCallbackData->pObjects must not be VK_OBJECT_TYPE_UNKNOWN", 99}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkSubmitDebugUtilsMessageEXT-pCallbackData-parameter", 57}, { {"pCallbackData must be a valid pointer to a valid VkDebugUtilsMessengerCallbackDataEXT structure", 95}, { "chapters/debugging.html", 23} } },
        { {"VUID-vkTransitionImageLayout-device-parameter", 45}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkTransitionImageLayout-pTransitions-parameter", 51}, { {"pTransitions must be a valid pointer to an array of transitionCount valid VkHostImageLayoutTransitionInfo structures", 116}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkTransitionImageLayout-transitionCount-arraylength", 56}, { {"transitionCount must be greater than 0", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkTrimCommandPool-commandPool-parameter", 44}, { {"commandPool must be a valid VkCommandPool handle", 48}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkTrimCommandPool-commandPool-parent", 41}, { {"commandPool must have been created, allocated, or retrieved from device", 71}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkTrimCommandPool-device-parameter", 39}, { {"device must be a valid VkDevice handle", 38}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkTrimCommandPool-flags-zerobitmask", 40}, { {"flags must be 0", 15}, { "chapters/cmdbuffers.html", 24} } },
        { {"VUID-vkUninitializePerformanceApiINTEL-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/queries.html", 21} } },
        { {"VUID-vkUnmapMemory-device-parameter", 35}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnmapMemory-memory-00689", 31}, { {"memory must be currently host mapped", 36}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnmapMemory-memory-parameter", 35}, { {"memory must be a valid VkDeviceMemory handle", 44}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnmapMemory-memory-parent", 32}, { {"memory must have been created, allocated, or retrieved from device", 66}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnmapMemory2-device-parameter", 36}, { {"device must be a valid VkDevice handle", 38}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnmapMemory2-pMemoryUnmapInfo-parameter", 46}, { {"pMemoryUnmapInfo must be a valid pointer to a valid VkMemoryUnmapInfo structure", 79}, { "chapters/memory.html", 20} } },
        { {"VUID-vkUnregisterCustomBorderColorEXT-device-parameter", 54}, { {"device must be a valid VkDevice handle", 38}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkUnregisterCustomBorderColorEXT-index-11288", 49}, { {"index must be less than maxCustomBorderColorSamplers", 52}, { "chapters/samplers.html", 22} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-descriptorSet-06995", 58}, { {"Host access to descriptorSet must be externally synchronized unless explicitly denoted otherwise for specific flags", 115}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-descriptorSet-parameter", 62}, { {"descriptorSet must be a valid VkDescriptorSet handle", 52}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-descriptorSet-parent", 59}, { {"descriptorSet must have been created, allocated, or retrieved from device", 73}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-descriptorUpdateTemplate-parameter", 73}, { {"descriptorUpdateTemplate must be a valid VkDescriptorUpdateTemplate handle", 74}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-descriptorUpdateTemplate-parent", 70}, { {"descriptorUpdateTemplate must have been created, allocated, or retrieved from device", 84}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSetWithTemplate-pData-01685", 50}, { {"pData must be a valid pointer to a memory containing one or more valid instances of VkAccelerationStructureKHR handles, VkAccelerationStructureNV handles, VkDescriptorImageInfo, VkDescriptorBufferInfo, or VkBufferView in a layout defined by descriptorUpdateTemplate when it was created with vkCreateDescriptorUpdateTemplate", 323}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-None-03047", 38}, { {"The dstSet member of each element of pDescriptorWrites or pDescriptorCopies for bindings which were created without the VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT or VK_DESCRIPTOR_BINDING_UPDATE_UNUSED_WHILE_PENDING_BIT bits set must not be used by any command that was recorded to a command buffer which is in the pending state", 329}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-device-parameter", 44}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorCopies-parameter", 55}, { {"If descriptorCopyCount is not 0, pDescriptorCopies must be a valid pointer to an array of descriptorCopyCount valid VkCopyDescriptorSet structures", 146}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06236", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER or VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, elements of the pTexelBufferView member of pDescriptorWrites[i] must have been created on device", 244}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06237", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, or VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, the buffer member of any element of the pBufferInfo member of pDescriptorWrites[i] must have been created on device", 338}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06238", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, and dstSet was not allocated with a layout that included immutable samplers for dstBinding with descriptorType, the sampler member of any element of the pImageInfo member of pDescriptorWrites[i] must have been created on device", 364}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06239", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER the imageView member of any element of pDescriptorWrites[i] must have been created on device", 394}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06240", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR, elements of the pAccelerationStructures member of a VkWriteDescriptorSetAccelerationStructureKHR structure in the pNext chain of pDescriptorWrites[i] must have been created on device", 293}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06241", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV, elements of the pAccelerationStructures member of a VkWriteDescriptorSetAccelerationStructureNV structure in the pNext chain of pDescriptorWrites[i] must have been created on device", 291}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06493", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_SAMPLER, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM, VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM, or VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, pDescriptorWrites[i].pImageInfo must be a valid pointer to an array of pDescriptorWrites[i].descriptorCount valid VkDescriptorImageInfo structures", 477}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-06993", 51}, { {"Host access to pDescriptorWrites[i].dstSet and pDescriptorCopies[i].dstSet must be externally synchronized unless explicitly denoted otherwise for specific flags", 161}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-12324", 51}, { {"For each element i where pDescriptorWrites[i].descriptorType is VK_DESCRIPTOR_TYPE_TENSOR_ARM, elements of the pTensorViews member of a VkWriteDescriptorSetTensorARM structure in the pNext chain of pDescriptorWrites[i] must have been created on device", 251}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateDescriptorSets-pDescriptorWrites-parameter", 55}, { {"If descriptorWriteCount is not 0, pDescriptorWrites must be a valid pointer to an array of descriptorWriteCount valid VkWriteDescriptorSet structures", 149}, { "chapters/descriptorsets.html", 28} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-None-11038", 55}, { {"Each member of the Indirect Execution Set referenced by the update command must not be in use by the device", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-None-11039", 55}, { {"If initialPipeline was created without VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, the layout of each pipeline in pExecutionSetWrites must be compatible with the initialPipeline used to create the Indirect Execution Set", 221}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-None-11040", 55}, { {"Each pipeline in the Indirect Execution Set must have identically defined static and dynamic state values to the initialPipeline used to create the Indirect Execution Set", 170}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-device-parameter", 61}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-executionSetWriteCount-11037", 73}, { {"executionSetWriteCount must be less than or equal to VkIndirectExecutionSetPipelineInfoEXT::maxPipelineCount", 108}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-executionSetWriteCount-arraylength", 79}, { {"executionSetWriteCount must be greater than 0", 45}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-indirectExecutionSet-11035", 71}, { {"indirectExecutionSet must have been created with type VK_INDIRECT_EXECUTION_SET_INFO_TYPE_PIPELINES_EXT", 103}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-indirectExecutionSet-parameter", 75}, { {"indirectExecutionSet must be a valid VkIndirectExecutionSetEXT handle", 69}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-indirectExecutionSet-parent", 72}, { {"indirectExecutionSet must have been created, allocated, or retrieved from device", 80}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11085", 66}, { {"Each pipeline in the Indirect Execution Set must match the initialPipeline used to create the Indirect Execution Set in its use of StencilExportEXT", 147}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11086", 66}, { {"Each pipeline in the Indirect Execution Set must match the initialPipeline used to create the Indirect Execution Set in its use of SampleMask", 141}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11098", 66}, { {"Each pipeline in the Indirect Execution Set must match the initialPipeline used to create the Indirect Execution Set in its use of FragDepth", 140}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11147", 66}, { {"Each pipeline in the Indirect Execution Set must have identically defined fragment outputs interface to the initialPipeline used to create the Indirect Execution Set", 165}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11152", 66}, { {"Each pipeline in the Indirect Execution Set must match the initialPipeline used to create the Indirect Execution Set in its included shader stages", 146}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11324", 66}, { {"If initialPipeline was created without VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, each pipeline in pExecutionSetWrites must also have been created without VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT", 202}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-initialPipeline-11325", 66}, { {"If initialPipeline was created with VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT, each pipeline in pExecutionSetWrites must also have been created with VK_PIPELINE_CREATE_2_DESCRIPTOR_HEAP_BIT_EXT", 196}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-pExecutionSetWrites-11042", 70}, { {"Each element in the pExecutionSetWrites array must have a unique VkWriteIndirectExecutionSetPipelineEXT::index", 110}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetPipelineEXT-pExecutionSetWrites-parameter", 74}, { {"pExecutionSetWrites must be a valid pointer to an array of executionSetWriteCount valid VkWriteIndirectExecutionSetPipelineEXT structures", 137}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-FragDepth-11054", 58}, { {"Each fragment shader element in the Indirect Execution Set must match the initial shader state used to create the Indirect Execution Set in its use of FragDepth", 160}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-None-11044", 53}, { {"Each member of the Indirect Execution Set referenced by the update command must not be in use by the device", 107}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-None-11148", 53}, { {"Each fragment shader element in the Indirect Execution Set must have identically defined fragment outputs interface to the initial shader state used to create the Indirect Execution Set", 185}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-SampleMask-11050", 59}, { {"Each fragment shader element in the Indirect Execution Set must match the initial shader state used to create the Indirect Execution Set in its use of SampleMask", 161}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-StencilExportEXT-11003", 65}, { {"Each fragment shader element in the Indirect Execution Set must match the initial shader state used to create the Indirect Execution Set in its use of StencilExportEXT", 167}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-device-parameter", 59}, { {"device must be a valid VkDevice handle", 38}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-executionSetWriteCount-arraylength", 77}, { {"executionSetWriteCount must be greater than 0", 45}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-indirectExecutionSet-11041", 69}, { {"indirectExecutionSet must have been created with type VK_INDIRECT_EXECUTION_SET_INFO_TYPE_SHADER_OBJECTS_EXT", 108}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-indirectExecutionSet-parameter", 73}, { {"indirectExecutionSet must be a valid VkIndirectExecutionSetEXT handle", 69}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-indirectExecutionSet-parent", 70}, { {"indirectExecutionSet must have been created, allocated, or retrieved from device", 80}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-pExecutionSetWrites-11043", 68}, { {"Each element in the pExecutionSetWrites array must have a unique VkWriteIndirectExecutionSetShaderEXT::index", 108}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-pExecutionSetWrites-11140", 68}, { {"If the shaders in pInitialShaders were created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, the descriptor layout of each shader in pExecutionSetWrites must be compatible with the initial layout info used to create the Indirect Execution Set", 246}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-pExecutionSetWrites-parameter", 72}, { {"pExecutionSetWrites must be a valid pointer to an array of executionSetWriteCount valid VkWriteIndirectExecutionSetShaderEXT structures", 135}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-pInitialShaders-11326", 64}, { {"If the shaders in pInitialShaders were created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, each shader in pExecutionSetWrites must also have been created without VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 208}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateIndirectExecutionSetShaderEXT-pInitialShaders-11327", 64}, { {"If the shaders in pInitialShaders were created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT, each pipeline in pExecutionSetWrites must also have been created with VK_SHADER_CREATE_DESCRIPTOR_HEAP_BIT_EXT", 204}, { "chapters/device_generated_commands/generatedcommands.html", 57} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-device-parameter", 55}, { {"device must be a valid VkDevice handle", 38}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-pUpdateInfo-07215", 56}, { {"pUpdateInfo->updateSequenceCount must equal the current update sequence counter of videoSessionParameters plus one", 114}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-pUpdateInfo-parameter", 60}, { {"pUpdateInfo must be a valid pointer to a valid VkVideoSessionParametersUpdateInfoKHR structure", 94}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-06441", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the number of StdVideoH264SequenceParameterSet entries already stored in it plus the value of the stdSPSCount member of the VkVideoEncodeH264SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoEncodeH264SessionParametersCreateInfoKHR::maxStdSPSCount videoSessionParameters was created with", 477}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-06442", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR, then the number of StdVideoH264PictureParameterSet entries already stored in it plus the value of the stdPPSCount member of the VkVideoEncodeH264SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoEncodeH264SessionParametersCreateInfoKHR::maxStdPPSCount videoSessionParameters was created with", 476}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-06443", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of StdVideoH265VideoParameterSet entries already stored in it plus the value of the stdVPSCount member of the VkVideoEncodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoEncodeH265SessionParametersCreateInfoKHR::maxStdVPSCount videoSessionParameters was created with", 474}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-06444", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of StdVideoH265SequenceParameterSet entries already stored in it plus the value of the stdSPSCount member of the VkVideoEncodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoEncodeH265SessionParametersCreateInfoKHR::maxStdSPSCount videoSessionParameters was created with", 477}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-06445", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR, then the number of StdVideoH265PictureParameterSet entries already stored in it plus the value of the stdPPSCount member of the VkVideoEncodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoEncodeH265SessionParametersCreateInfoKHR::maxStdPPSCount videoSessionParameters was created with", 476}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07216", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoDecodeH264SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH264SequenceParameterSet entry with seq_parameter_set_id matching any of the elements of VkVideoDecodeH264SessionParametersAddInfoKHR::pStdSPSs", 422}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07217", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the number of StdVideoH264SequenceParameterSet entries already stored in it plus the value of the stdSPSCount member of the VkVideoDecodeH264SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoDecodeH264SessionParametersCreateInfoKHR::maxStdSPSCount videoSessionParameters was created with", 477}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07218", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoDecodeH264SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH264PictureParameterSet entry with both seq_parameter_set_id and pic_parameter_set_id matching any of the elements of VkVideoDecodeH264SessionParametersAddInfoKHR::pStdPPSs", 451}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07219", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR, then the number of StdVideoH264PictureParameterSet entries already stored in it plus the value of the stdPPSCount member of the VkVideoDecodeH264SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoDecodeH264SessionParametersCreateInfoKHR::maxStdPPSCount videoSessionParameters was created with", 476}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07220", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoDecodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265VideoParameterSet entry with vps_video_parameter_set_id matching any of the elements of VkVideoDecodeH265SessionParametersAddInfoKHR::pStdVPSs", 425}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07221", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of StdVideoH265VideoParameterSet entries already stored in it plus the value of the stdVPSCount member of the VkVideoDecodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoDecodeH265SessionParametersCreateInfoKHR::maxStdVPSCount videoSessionParameters was created with", 474}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07222", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoDecodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265SequenceParameterSet entry with both sps_video_parameter_set_id and sps_seq_parameter_set_id matching any of the elements of VkVideoDecodeH265SessionParametersAddInfoKHR::pStdSPSs", 462}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07223", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of StdVideoH265SequenceParameterSet entries already stored in it plus the value of the stdSPSCount member of the VkVideoDecodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoDecodeH265SessionParametersCreateInfoKHR::maxStdSPSCount videoSessionParameters was created with", 477}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07224", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoDecodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265PictureParameterSet entry with sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id all matching any of the elements of VkVideoDecodeH265SessionParametersAddInfoKHR::pStdPPSs", 487}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07225", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR, then the number of StdVideoH265PictureParameterSet entries already stored in it plus the value of the stdPPSCount member of the VkVideoDecodeH265SessionParametersAddInfoKHR structure included in the pUpdateInfo->pNext chain must be less than or equal to the VkVideoDecodeH265SessionParametersCreateInfoKHR::maxStdPPSCount videoSessionParameters was created with", 476}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07226", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH264SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH264SequenceParameterSet entry with seq_parameter_set_id matching any of the elements of VkVideoEncodeH264SessionParametersAddInfoKHR::pStdSPSs", 422}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07227", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH264SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH264PictureParameterSet entry with both seq_parameter_set_id and pic_parameter_set_id matching any of the elements of VkVideoEncodeH264SessionParametersAddInfoKHR::pStdPPSs", 451}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07228", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265VideoParameterSet entry with vps_video_parameter_set_id matching any of the elements of VkVideoEncodeH265SessionParametersAddInfoKHR::pStdVPSs", 425}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07229", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265SequenceParameterSet entry with both sps_video_parameter_set_id and sps_seq_parameter_set_id matching any of the elements of VkVideoEncodeH265SessionParametersAddInfoKHR::pStdSPSs", 462}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-07230", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH265SessionParametersAddInfoKHR structure, then videoSessionParameters must not already contain a StdVideoH265PictureParameterSet entry with sps_video_parameter_set_id, pps_seq_parameter_set_id, and pps_pic_parameter_set_id all matching any of the elements of VkVideoEncodeH265SessionParametersAddInfoKHR::pStdPPSs", 487}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-08321", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH265SessionParametersAddInfoKHR structure, then num_tile_columns_minus1 must be less than VkVideoEncodeH265CapabilitiesKHR::maxTiles.width, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSessionParameters was created with, for each element of VkVideoEncodeH265SessionParametersAddInfoKHR::pStdPPSs", 505}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-08322", 67}, { {"If videoSessionParameters was created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR and the pNext chain of pUpdateInfo includes a VkVideoEncodeH265SessionParametersAddInfoKHR structure, then num_tile_rows_minus1 must be less than VkVideoEncodeH265CapabilitiesKHR::maxTiles.height, as returned by vkGetPhysicalDeviceVideoCapabilitiesKHR for the video profile videoSessionParameters was created with, for each element of VkVideoEncodeH265SessionParametersAddInfoKHR::pStdPPSs", 503}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-09260", 67}, { {"videoSessionParameters must not have been created with the video codec operation VK_VIDEO_CODEC_OPERATION_DECODE_AV1_BIT_KHR", 124}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-10281", 67}, { {"videoSessionParameters must not have been created with the video codec operation VK_VIDEO_CODEC_OPERATION_ENCODE_AV1_BIT_KHR", 124}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-parameter", 71}, { {"videoSessionParameters must be a valid VkVideoSessionParametersKHR handle", 73}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-parent", 68}, { {"videoSessionParameters must have been created, allocated, or retrieved from device", 82}, { "chapters/videocoding.html", 25} } },
        { {"VUID-vkWaitForFences-device-parameter", 37}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWaitForFences-fenceCount-arraylength", 43}, { {"fenceCount must be greater than 0", 33}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWaitForFences-pFences-parameter", 38}, { {"pFences must be a valid pointer to an array of fenceCount valid VkFence handles", 79}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWaitForFences-pFences-parent", 35}, { {"Each element of pFences must have been created, allocated, or retrieved from device", 83}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWaitForPresent2KHR-None-10815", 36}, { {"The VkSurfaceCapabilitiesPresentWait2KHR surface capability must be present for the underlying surface", 102}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-None-10816", 36}, { {"The swapchain must have been created with VK_SWAPCHAIN_CREATE_PRESENT_WAIT_2_BIT_KHR bit set in the VkSwapchainCreateFlagBitsKHR field", 134}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-device-parameter", 42}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-pPresentWait2Info-parameter", 53}, { {"pPresentWait2Info must be a valid pointer to a valid VkPresentWait2InfoKHR structure", 84}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-presentId-10817", 41}, { {"The presentId value must have been associated with a vkQueuePresentKHR request on the swapchain which returned a non-error value", 128}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-presentWait2-10814", 44}, { {"The presentWait2 feature must be enabled", 40}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-swapchain-parameter", 45}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresent2KHR-swapchain-parent", 42}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresentKHR-device-parameter", 41}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresentKHR-presentWait-06234", 42}, { {"The presentWait feature must be enabled", 39}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresentKHR-swapchain-04997", 40}, { {"swapchain must not be in the retired state", 42}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresentKHR-swapchain-parameter", 44}, { {"swapchain must be a valid VkSwapchainKHR handle", 47}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitForPresentKHR-swapchain-parent", 41}, { {"swapchain must have been created, allocated, or retrieved from device", 69}, { "chapters/VK_KHR_surface/wsi.html", 32} } },
        { {"VUID-vkWaitSemaphores-device-parameter", 38}, { {"device must be a valid VkDevice handle", 38}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWaitSemaphores-pWaitInfo-parameter", 41}, { {"pWaitInfo must be a valid pointer to a valid VkSemaphoreWaitInfo structure", 74}, { "chapters/synchronization.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-accelerationStructureCount-arraylength", 86}, { {"accelerationStructureCount must be greater than 0", 49}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-accelerationStructureHostCommands-03585", 87}, { {"The VkPhysicalDeviceAccelerationStructureFeaturesKHR::accelerationStructureHostCommands feature must be enabled", 111}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-accelerationStructures-03431", 76}, { {"All acceleration structures in pAccelerationStructures must have been built with VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_KHR if queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR", 209}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-buffer-03733", 60}, { {"Each element of pAccelerationStructures must be bound to host-visible device memory", 83}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-buffer-03736", 60}, { {"Each element of pAccelerationStructures must be fully backed by physical memory", 79}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-buffer-03784", 60}, { {"Each element of pAccelerationStructures must be bound to memory that was not allocated with multiple instances", 110}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-dataSize-03452", 62}, { {"dataSize must be greater than or equal to accelerationStructureCount*stride", 75}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-dataSize-arraylength", 68}, { {"dataSize must be greater than 0", 31}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-device-parameter", 64}, { {"device must be a valid VkDevice handle", 38}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-11591", 77}, { {"All acceleration structures in pAccelerationStructures must have been constructed prior to the execution of this command", 120}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-11592", 77}, { {"Each acceleration structure in pAccelerationStructures must have been created with vkCreateAccelerationStructureKHR", 115}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-12425", 77}, { {"All acceleration structures in pAccelerationStructures must have been built with VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR if queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR", 218}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-parameter", 81}, { {"pAccelerationStructures must be a valid pointer to an array of accelerationStructureCount valid VkAccelerationStructureKHR handles", 130}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pAccelerationStructures-parent", 78}, { {"Each element of pAccelerationStructures must have been created, allocated, or retrieved from device", 99}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-pData-parameter", 63}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-03448", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, then stride must be a multiple of the size of VkDeviceSize", 131}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-03449", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, then pData must point to a VkDeviceSize", 112}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-03450", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR, then stride must be a multiple of the size of VkDeviceSize", 135}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-03451", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR, then pData must point to a VkDeviceSize", 116}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-06731", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR, then stride must be a multiple of the size of VkDeviceSize", 121}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-06732", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR, then pData must point to a VkDeviceSize", 102}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-06733", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR, then stride must be a multiple of the size of VkDeviceSize", 152}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-06734", 63}, { {"If queryType is VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR, then pData must point to a VkDeviceSize", 133}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-06742", 63}, { {"queryType must be VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR, VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR, VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR, or VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR", 262}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteAccelerationStructuresPropertiesKHR-queryType-parameter", 67}, { {"queryType must be a valid VkQueryType value", 43}, { "chapters/accelstructures.html", 29} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-buffer-07577", 47}, { {"The buffer used to create each micromap in pMicromaps must be bound to host-visible device memory", 97}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-buffer-07579", 47}, { {"The buffer used to create each micromap in pMicromaps must be bound to memory that was not allocated with multiple instances", 124}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-dataSize-07576", 49}, { {"dataSize must be greater than or equal to micromapCount*stride", 62}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-dataSize-arraylength", 55}, { {"dataSize must be greater than 0", 31}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-micromapCount-arraylength", 60}, { {"micromapCount must be greater than 0", 36}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-micromapHostCommands-07578", 61}, { {"The VkPhysicalDeviceOpacityMicromapFeaturesEXT::micromapHostCommands feature must be enabled", 92}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-pData-parameter", 50}, { {"pData must be a valid pointer to an array of dataSize bytes", 59}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-pMicromaps-07502", 51}, { {"All micromaps in pMicromaps must have been constructed with VK_BUILD_MICROMAP_ALLOW_COMPACTION_BIT_EXT if queryType is VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT", 160}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-pMicromaps-11671", 51}, { {"All micromaps in pMicromaps must have been constructed prior to the execution of this command", 93}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-pMicromaps-parameter", 55}, { {"pMicromaps must be a valid pointer to an array of micromapCount valid VkMicromapEXT handles", 91}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-pMicromaps-parent", 52}, { {"Each element of pMicromaps must have been created, allocated, or retrieved from device", 86}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-queryType-07503", 50}, { {"queryType must be VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT, or VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT", 109}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-queryType-10071", 50}, { {"If queryType is VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT or VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT then stride must be a multiple of the size of VkDeviceSize", 165}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-queryType-10072", 50}, { {"If queryType is VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT or VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT then pData must point to a VkDeviceSize", 146}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteMicromapsPropertiesEXT-queryType-parameter", 54}, { {"queryType must be a valid VkQueryType value", 43}, { "chapters/VK_KHR_opacity_micromap/micromaps.html", 47} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-descriptorHeap-11206", 55}, { {"The descriptorHeap feature must be enabled", 42}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-device-parameter", 51}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-pDescriptors-parameter", 57}, { {"pDescriptors must be a valid pointer to an array of resourceCount valid VkHostAddressRangeEXT structures", 104}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-pResources-11208", 51}, { {"If any element of pResources specifies a VkImageViewCreateInfo structure with a VkSamplerYcbcrConversionInfo structure in its pNext chain, the corresponding element of pDescriptors must have a size member that is greater than or equal to the product of the value returned by vkGetPhysicalDeviceDescriptorSizeEXT with a descriptorType equal to type and VkSamplerYcbcrConversionImageFormatProperties::combinedImageSamplerDescriptorCount, as queried from VkPhysicalDeviceImageFormatInfo2 with image format info equivalent to the image view the descriptor is being created for", 572}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-pResources-11209", 51}, { {"If any element of pResources specifies a VkImageViewCreateInfo structure with an image created with flags containing VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT, the corresponding element of pDescriptors must have a size member that is greater than or equal to the product of the value returned by vkGetPhysicalDeviceDescriptorSizeEXT with a descriptorType equal to type and VkSubsampledImageFormatPropertiesEXT::subsampledImageDescriptorCount, as queried from VkPhysicalDeviceImageFormatInfo2 with image format info equivalent to the image view the descriptor is being created for", 572}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-pResources-parameter", 55}, { {"pResources must be a valid pointer to an array of resourceCount valid VkResourceDescriptorInfoEXT structures", 108}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-resourceCount-arraylength", 60}, { {"resourceCount must be greater than 0", 36}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteResourceDescriptorsEXT-size-11207", 45}, { {"The size member of each element of pDescriptors must be greater than or equal to the value returned by vkGetPhysicalDeviceDescriptorSizeEXT with a descriptorType equal to type", 175}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-borderColor-11205", 51}, { {"If the borderColor of any element of pSamplers is VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT, VkSamplerCustomBorderColorIndexCreateInfoEXT::index must be a value less than maxCustomBorderColorSamplers", 224}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-borderColor-11444", 51}, { {"If the borderColor of any element of pSamplers is VK_BORDER_COLOR_FLOAT_CUSTOM_EXT or VK_BORDER_COLOR_INT_CUSTOM_EXT, VkSamplerCustomBorderColorIndexCreateInfoEXT must be included in the pNext chain of that element", 214}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-descriptorHeap-11202", 54}, { {"The descriptorHeap feature must be enabled", 42}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-device-parameter", 50}, { {"device must be a valid VkDevice handle", 38}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-pDescriptors-parameter", 56}, { {"pDescriptors must be a valid pointer to an array of samplerCount valid VkHostAddressRangeEXT structures", 103}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-pNext-11400", 45}, { {"If there is a VkDebugUtilsObjectNameInfoEXT structure in the pNext chain of any element of pSamplers, its objectType must be VK_OBJECT_TYPE_UNKNOWN", 147}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-pSamplers-11204", 49}, { {"Elements of pSamplers must not include VkSamplerYcbcrConversionInfo structures in their pNext chains", 100}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-pSamplers-parameter", 53}, { {"pSamplers must be a valid pointer to an array of samplerCount valid VkSamplerCreateInfo structures", 98}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-samplerCount-arraylength", 58}, { {"samplerCount must be greater than 0", 35}, { "chapters/descriptorheaps.html", 29} } },
        { {"VUID-vkWriteSamplerDescriptorsEXT-size-11203", 44}, { {"The size member of each element of pDescriptors must be greater than or equal to the value returned by vkGetPhysicalDeviceDescriptorSizeEXT with a descriptorType equal to VK_DESCRIPTOR_TYPE_SAMPLER", 197}, { "chapters/descriptorheaps.html", 29} } },
    }};
    static const vvl::unordered_map<std::string_view, vuid_info> vuid_map(std::begin(vuid_array), std::end(vuid_array));
    return vuid_map;
}
